module ConjugateGradients
    use iso_c_binding
    use Globals

    private

    public  ::  ConjugateGradient

    !>  \brief  What a minimisable function should look like
    abstract interface
        function target_function(values,params) bind(c)
            import :: c_double, c_ptr   ! because we're in an abstract interface, we normally don't have access to all the stuff from the module, so we need to import
            type(c_ptr), value  ::  values
            type(c_ptr), value  ::  params
            real(kind=c_double) ::  target_function
        end function target_function
    end interface

    type ConjugateGradient
        private
        integer                                     ::  n                   !<  number of parameters to refine
        integer                                     ::  num_function_calls  !<  number of scoring function calls
        real(kind=8),   allocatable                 ::  best_values(:)      !<  final best parameters determined by conjugate gradient mimimizer
        real(kind=8),   allocatable                 ::  e(:)                !<  recalculated accuracy for subroutine va04
        real                                        ::  escale              !<  anticipated change in the parameters during conjugate gradient minimization
        real                                        ::  best_score          !<  final best score
        !
        type(c_ptr)                                 ::  params_for_scoring  !<  a pointer to an object containing all the information needed by the scoring function (other than the function's arguments)
        procedure(target_function), pointer, nopass :: function => null()   !<  a pointer to the function the minimizer aims to minimize.
                                                                            !!  no pass because there an assumption by default that type-bound procedures have an object of the type to which they are bound as first argument.
                                                                            !!  this default assumption is overriden by no pass.
        !
        logical                                     ::  print_progress = .false.
        !

        contains
            procedure,  public  ::  Init
            procedure,  public  ::  PrintInfo
            procedure,  public  ::  Run
            procedure,  public  ::  GetBestValues
            procedure,  public  ::  GetBestScore
    end type

    contains


        !>  \brief  initialise a conjugate gradient minimizer
        subroutine Init(self,num_dim,function_to_minimize,parameters,   &
                            starting_values,accuracy,print_progress)
            implicit none
            ! arguments
            class(ConjugateGradient),                       intent(inout)   ::  self
            integer,                                        intent(in)      ::  num_dim                 !<  number of dimensions to minimize
            procedure(target_function)                                      ::  function_to_minimize    !<  the function to minimize
            type(c_ptr),                                    intent(in)      ::  parameters              !<  parameters needed by the scoring function (a c pointer to an object containing...)
            real(kind=8),              allocatable, target, intent(in)      ::  starting_values(:)      !<  the starting point for the minimizer
            real(kind=8),                                   intent(in)      ::  accuracy(:)             !<  for each search dimension, the accuracy of the parameter values to be determined by the minimizer
            logical,                   optional,            intent(in)      ::  print_progress          !<  When the minimizer is running, print something at every iteration
            ! private variables
            type(c_ptr)  ::  start_val_c
            logical,	parameter	::	debug	=	.false.
            ! start work

            ! Checks
            if (size(accuracy) .ne. num_dim .or. size(starting_values) .ne. num_dim) then
                call this_program%TerminateWithFatalError('ConjugateGradient::Init',&
                                                        'Wrong number of starting values and/or accuracy')
            endif


            self%escale = 100.0
            self%best_values = starting_values
            ! self%e = accuracy/self%escale
            self%e = accuracy
            self%n = num_dim
            ! self%best_score = huge(1.0)
            self%num_function_calls = 0

            ! the function we'll be minimizing
            self%function => function_to_minimize

            ! parameters needed by the function
            self%params_for_scoring = parameters

            ! Print progress?
            self%print_progress = .false.
            if (present(print_progress)) self%print_progress = print_progress

            ! the score for the starting values
            start_val_c = c_loc(starting_values)
            self%best_score = self%function(start_val_c,self%params_for_scoring)
            if (debug) write(*,'(a,f0.5)') '**debug(ConjugateGradient::Init): starting score = ', self%best_score

            !
            if (self%print_progress) then
                write(*,'(a)') '**info(ConjugateGradient::Init): conjugate gradient minimizer has been initialised. summary:'
                call self%PrintInfo()
            endif
        end subroutine Init

        !>  \brief  run the conjugate gradient minimizer
        subroutine Run(self)
            class(ConjugateGradient),        intent(inout)    ::    self
            ! private variables
            integer  ::  iprint, icon, maxit
            ! start work

            iprint=0
            icon=1
            maxit=50
            if (self%print_progress) write(*,'(a)') '**info(ConjugateGradient::Run): starting conjugate-gradient minimiser'
            call va04a(self,iprint,icon,maxit)
            if (self%print_progress) then
                write(*,'(/a)') '**info(ConjugateGradient::Run): conjugate-gradient minimiser returned'
                call self%PrintInfo()
            endif
        end subroutine Run

        !>  \brief  Return the best values
        function GetBestValues(self)
            class(ConjugateGradient),   intent(in)  ::  self
            real(kind=8),   allocatable             ::  GetBestValues(:)
            GetBestValues = self%best_values
        end function

        !>  \brief  Return the best score
        pure real function GetBestScore(self)
            class(ConjugateGradient),   intent(in)  ::  self
            GetBestScore = self%best_score
        end function

        !>  \brief  Print summary information about the CG minimizer
        subroutine PrintInfo(self)
            class(ConjugateGradient), intent(in)  ::  self
            ! private variables
            character(len=6)    ::  size_e, size_best_values
            ! start work
            write(size_e,'(i6)') size(self%e)
            write(size_best_values,'(i6)') size(self%best_values)
            !
            write(*,'(/a)')         'Conjugate gradient minimizer, summary info:'
            write(*,'(a,i0)')       'Number of dimensions                    = ', self%n
            write(*,'(a,i0)')       'Number of function calls                = ', self%num_function_calls
            if (size(self%best_values) .lt. 64) then
                ! nag has maximum formatted output record length of 1024 characters
                write(*,'(a,'//size_best_values//'(g15.5,1x))') 'Parameters: ', self%best_values
                write(*,'(a,'//size_e//'(g15.5,1x))')           'Accuracy:   ', self%e
            endif
            write(*,'(a,g15.5)')                             'Score:      ', self%best_score
            write(*,'(a)')      ' '
        end subroutine PrintInfo

!**************************************************************************
      !>  \brief  Fortran 66 derived code for conjugate gradient minimizer
      SUBROUTINE VA04A(SELF,IPRINT,ICON,MAXIT)
!**************************************************************************
!  STANDARD FORTRAN 66 (A VERIFIED PFORT SUBROUTINE)  - Powell minimization
!  some changes were made to reduce diagnostic output and prevent occasional
!  crashes (Niko, 12. June 1998)
!**************************************************************************
#ifdef _OPENMP
      USE OMP_LIB
#endif
      IMPLICIT NONE

      TYPE(ConjugateGradient), INTENT(INOUT)    ::  SELF
      TYPE(C_PTR)                               ::  START_VAL_C

      INTEGER N,IPRINT,ICON,MAXIT,MAXX,I,J,K,JJ,JJJ,NFCC
      INTEGER IND,INN,ITERC,ISGRAD,ICNT,ITONE,IXP,IDIRN
      INTEGER ILINE,IS,JIL
      REAL(KIND=8), ALLOCATABLE  :: E(:), W(:), XS(:)
      REAL(KIND=8), ALLOCATABLE, TARGET :: X(:)
      REAL F,ESCALE
      REAL(KIND=8)  ::  DDMAG,SCER
      REAL(KIND=8)  ::  FKEEP,FP,SUM,DMAX,DACC,DMAG,FPREV
      REAL(KIND=8)  ::  FA,DA,DD,DL,D,FB,DB,FC,DC,FHOLD
      REAL(KIND=8)  ::  DI,FI,A,B,AAA,TMP,DDMAX
      CHARACTER(LEN=6) :: SIZE_XP
      integer :: thread_num, num_of_threads
      LOGICAL, PARAMETER :: DEBUG = .FALSE.
!**************************************************************************
!	W[N*(N+3)]


#ifdef _OPENMP
      thread_num = omp_get_thread_num()
      num_of_threads = omp_get_num_threads()
#else
      thread_num = 0
      num_of_threads = 1
#endif

      ALLOCATE(W(SELF%N*(SELF%N+3)))
      ALLOCATE(XS(SELF%N))

      X = SELF%BEST_VALUES
      E = SELF%E
      N = SELF%N
      ESCALE = SELF%ESCALE

      WRITE(SIZE_XP,'(I6)') SIZE(X)+1

      START_VAL_C = C_LOC(X)

      DDMAG=0.1*ESCALE
      SCER=0.05/ESCALE
      ICNT=0
      MAXX=100*MAXIT
      DO 999 I=1,N
        XS(I)=X(I)
  999 CONTINUE
      JJ=N*N+N
      JJJ=JJ+N
      K=N+1
      NFCC=1
      IND=1
      INN=1
      DO 1 I=1,N
      DO 2 J=1,N
      W(K)=0.
      IF(I-J)4,3,4
    3 W(K)=ABS(E(I))
      W(I)=ESCALE
    4 K=K+1
    2 CONTINUE
    1 CONTINUE
      ITERC=1
      ISGRAD=2
      ICNT=ICNT+1
      IF (ICNT.GT.MAXX) GOTO 998

      F = SELF%FUNCTION(START_VAL_C,SELF%PARAMS_FOR_SCORING)
      SELF%NUM_FUNCTION_CALLS = SELF%NUM_FUNCTION_CALLS + 1
      IF (DEBUG .AND. SIZE(X) .LT. 63) WRITE(*,'(A,'//SIZE_XP//'(G15.5,1X))') '**debug(VA04A): trying these values: ', X,F
	  IF (DEBUG .AND. SIZE(X) .GE. 63) WRITE(*,'(A,F0.6)') '**debug(VA04A): got this score: ', F
	  IF (self%print_progress) THEN
#ifdef __INTEL_COMPILER
        WRITE(*,'(3(A,I0),2(A,F0.4),20X,A,$)') &
#else
        WRITE(*,'(3(A,I0),2(A,F0.4),20X,A)',ADVANCE='NO') &
#endif
                        '**info(VA04): scoring function call #', SELF%NUM_FUNCTION_CALLS, &
                        ' on thread ', thread_num+1, ' of ', num_of_threads, &
                        '. Latest score = ', F, &
                        '. Current best score = ', SELF%BEST_SCORE, char(13)
      ENDIF
	  ! IF (F.
      ! IF (F.LT.SELF%BEST_SCORE) THEN
      !   SELF%BEST_SCORE = F
      !   SELF%BEST_VALUES = X
      ! ENDIF

      FKEEP=ABS(F)+ABS(F)
    5 ITONE=1
      FP=F
      SUM=0.
      IXP=JJ
      DO 6 I=1,N
      IXP=IXP+1
      W(IXP)=X(I)
    6 CONTINUE
      IDIRN=N+1
      ILINE=1
    7 DMAX=W(ILINE)
      DACC=DMAX*SCER
      DMAG=DMIN1(DDMAG,0.1*DMAX)
      DMAG=DMAX1(DMAG,20.*DACC)
      DDMAX=10.*DMAG
      GO TO (70,70,71),ITONE
   70 DL=0.
      D=DMAG
      FPREV=F
      IS=5
      FA=F
      DA=DL
    8 DD=D-DL
      DL=D
   58 K=IDIRN
      DO 9 I=1,N
      X(I)=X(I)+DD*W(K)
      K=K+1
    9 CONTINUE
      ICNT=ICNT+1
      IF (ICNT.GT.MAXX) GOTO 998

      F = SELF%FUNCTION(START_VAL_C,SELF%PARAMS_FOR_SCORING)
      SELF%NUM_FUNCTION_CALLS = SELF%NUM_FUNCTION_CALLS + 1
      IF (DEBUG .AND. SIZE(X) .LT. 63) WRITE(*,'(A,'//SIZE_XP//'(G15.5,1X))') '**debug(VA04A): trying these values: ', X,F
	  IF (DEBUG .AND. SIZE(X) .GE. 63) WRITE(*,'(A,F0.6)') '**debug(VA04A): got this score: ', F
	  IF (self%print_progress) THEN
#ifdef __INTEL_COMPILER
        WRITE(*,'(3(A,I0),2(A,F0.4),20X,A,$)') &
#else
        WRITE(*,'(3(A,I0),2(A,F0.4),20X,A)',ADVANCE='NO') &
#endif
                        '**info(VA04): scoring function call #', SELF%NUM_FUNCTION_CALLS, &
                        ' on thread ', thread_num+1, ' of ', num_of_threads, &
                        '. Latest score = ', F, &
                        '. Current best score = ', SELF%BEST_SCORE, char(13)
      ENDIF
	  ! IF (F.LT.SELF%BEST_SCORE) THEN
      !   SELF%BEST_SCORE = F
      !   SELF%BEST_VALUES = X
      ! ENDIF

      NFCC=NFCC+1
      GO TO (10,11,12,13,14,96),IS
   14 IF(F-FA)15,16,24
   16 IF (ABS(D)-DMAX) 17,17,18
   17 D=D+D
      GO TO 8
!   18 WRITE(6,19)
!   19 FORMAT(5X,50HPOWELL MIN: MAXIMUM CHANGE DOES NOT ALTER FUNCTION)
   19 FORMAT(5X,44HVA04A MAXIMUM CHANGE DOES NOT ALTER FUNCTION)
!   18 CALL TERMINATE('Fatal error in VA04: maximum change does not alter function')
   18 WRITE(*,'(A)') '**warning(VA04): maximum change does not alter function'
      GO TO 20
   15 FB=F
      DB=D
      GO TO 21
   24 FB=FA
      DB=DA
      FA=F
      DA=D
   21 GO TO (83,23),ISGRAD
   23 D=DB+DB-DA
      IS=1
      GO TO 8
   83 D=0.5*(DA+DB-(FA-FB)/(DA-DB))
      IS=4
      IF((DA-D)*(D-DB))25,8,8
   25 IS=1
      IF(ABS(D-DB)-DDMAX)8,8,26
   26 D=DB+SIGN(DDMAX,DB-DA)
      IS=1
      DDMAX=DDMAX+DDMAX
      DDMAG=DDMAG+DDMAG
      IF(DDMAX-DMAX)8,8,27
   27 DDMAX=DMAX
      GO TO 8
   13 IF(F-FA)28,23,23
   28 FC=FB
      DC=DB
   29 FB=F
      DB=D
      GO TO 30
   12 IF(F-FB)28,28,31
   31 FA=F
      DA=D
      GO TO 30
   11 IF(F-FB)32,10,10
   32 FA=FB
      DA=DB
      GO TO 29
   71 DL=1.
      DDMAX=5.
      FA=FP
      DA=-1.
      FB=FHOLD
      DB=0.
      D=1.
   10 FC=F
      DC=D
   30 A=(DB-DC)*(FA-FC)
      B=(DC-DA)*(FB-FC)
      IF((A+B)*(DA-DC))33,33,34
   33 FA=FB
      DA=DB
      FB=FC
      DB=DC
      GO TO 26
   34 D=0.5*(A*(DB+DC)+B*(DA+DC))/(A+B)
      DI=DB
      FI=FB
      IF(FB-FC)44,44,43
   43 DI=DC
      FI=FC
   44 GO TO (86,86,85),ITONE
   85 ITONE=2
      GO TO 45
   86 IF (ABS(D-DI)-DACC) 41,41,93
   93 IF (ABS(D-DI)-0.03*ABS(D)) 41,41,45
   45 IF ((DA-DC)*(DC-D)) 47,46,46
   46 FA=FB
      DA=DB
      FB=FC
      DB=DC
      GO TO 25
   47 IS=2
      IF ((DB-D)*(D-DC)) 48,8,8
   48 IS=3
      GO TO 8
   41 F=FI
      D=DI-DL
      DD=(DC-DB)*(DC-DA)*(DA-DB)/(A+B)
      IF (DD.LT.0.0) DD=1E-10
      DD=SQRT(DD)
      DO 49 I=1,N
      X(I)=X(I)+D*W(IDIRN)
      W(IDIRN)=DD*W(IDIRN)
      IDIRN=IDIRN+1
   49 CONTINUE
      IF (DD.EQ.0.0) DD=1E-10
      W(ILINE)=W(ILINE)/DD
      ILINE=ILINE+1
      IF(IPRINT-1)51,50,51
   50 CONTINUE
!   50 WRITE(6,52) ITERC,NFCC,F,(X(I),I=1,N)
   52 FORMAT (/1X,9HITERATION,I5,I15,16H FUNCTION VALUES,10X,3HF =,E21.14/(5E24.14))
      GO TO(51,53),IPRINT
   51 GO TO (55,38),ITONE
   55 IF (FPREV-F-SUM) 94,95,95
   95 SUM=FPREV-F
      JIL=ILINE
   94 IF (IDIRN-JJ) 7,7,84
   84 GO TO (92,72),IND
   92 FHOLD=F
      IS=6
      IXP=JJ
      DO 59 I=1,N
      IXP=IXP+1
      W(IXP)=X(I)-W(IXP)
   59 CONTINUE
      DD=1.
      GO TO 58
   96 GO TO (112,87),IND
  112 IF (FP-F) 37,37,91
   91 D=2.*(FP+F-2.*FHOLD)/(FP-F)**2
      IF (D*(FP-FHOLD-SUM)**2-SUM) 87,37,37
   87 J=JIL*N+1
      IF (J-JJ) 60,60,61
   60 DO 62 I=J,JJ
      K=I-N
      W(K)=W(I)
   62 CONTINUE
      DO 97 I=JIL,N
      W(I-1)=W(I)
   97 CONTINUE
   61 IDIRN=IDIRN-N
      ITONE=3
      K=IDIRN
      IXP=JJ
      AAA=0.
      DO 65 I=1,N
      IXP=IXP+1
      W(K)=W(IXP)
      IF (AAA-ABS(W(K)/E(I))) 66,67,67
   66 AAA=ABS(W(K)/E(I))
   67 K=K+1
   65 CONTINUE
      DDMAG=1.
      IF (AAA.EQ.0.0) AAA=1E-10
      W(N)=ESCALE/AAA
      ILINE=N
      GO TO 7
   37 IXP=JJ
      AAA=0.
      F=FHOLD
      DO 99 I=1,N
      IXP=IXP+1
      X(I)=X(I)-W(IXP)
      IF (AAA*ABS(E(I))-ABS(W(IXP))) 98,99,99
   98 AAA=ABS(W(IXP)/E(I))
   99 CONTINUE
      GO TO 72
   38 AAA=AAA*(1.+DI)
      GO TO (72,106),IND
   72 IF (IPRINT-2) 53,50,50
   53 GO TO (109,88),IND
  109 IF (AAA-0.1) 89,89,76
   89 GO TO (20,116),ICON
  116 IND=2
      GO TO (100,101),INN
  100 INN=2
      K=JJJ
      DO 102 I=1,N
      K=K+1
      W(K)=X(I)
      X(I)=X(I)+10.*E(I)
  102 CONTINUE
      FKEEP=F
      ICNT=ICNT+1
      IF (ICNT.GT.MAXX) GOTO 998

      F = SELF%FUNCTION(START_VAL_C,SELF%PARAMS_FOR_SCORING)
      SELF%NUM_FUNCTION_CALLS = SELF%NUM_FUNCTION_CALLS + 1

      IF (DEBUG .AND. SIZE(X) .LT. 63) WRITE(*,'(A,'//SIZE_XP//'(G15.5,1X))') '**debug(VA04A): trying these values: ', X,F
	  IF (DEBUG .AND. SIZE(X) .GE. 63) WRITE(*,'(A,F0.6)') '**debug(VA04A): got this score: ', F
	  IF (self%print_progress) THEN
#ifdef __INTEL_COMPILER
        WRITE(*,'(3(A,I0),2(A,F0.4),20X,A,$)') &
#else
        WRITE(*,'(3(A,I0),2(A,F0.4),20X,A)',ADVANCE='NO') &
#endif
                        '**info(VA04): scoring function call #', SELF%NUM_FUNCTION_CALLS, &
                        ' on thread ', thread_num+1, ' of ', num_of_threads, &
                        '. Latest score = ', F, &
                        '. Current best score = ', SELF%BEST_SCORE, char(13)
      ENDIF
	  ! IF (F.LT.SELF%BEST_SCORE) THEN
      !   SELF%BEST_SCORE = F
      !   SELF%BEST_VALUES = X
      ! ENDIF

      NFCC=NFCC+1
      DDMAG=0.
      GO TO 108
   76 IF (F-FP) 35,78,78
   78 CONTINUE
!   78 WRITE(6,80)
   80 FORMAT (5X,43HPOWELL MIN: ACCURACY LIMITED BY ERRORS IN F)
!   80 FORMAT (5X,37HVA04A ACCURACY LIMITED BY ERRORS IN F)
      GO TO 20
   88 IND=1
   35 TMP=FP-F
      IF (TMP.GT.0.0) THEN
      DDMAG=0.4*SQRT(TMP)
      ELSE
      DDMAG=0.0
      ENDIF
      ISGRAD=1
  108 ITERC=ITERC+1
      IF (ITERC-MAXIT) 5,5,81
81    CONTINUE
!   81 WRITE(6,82) MAXIT
   82 FORMAT(I5,30H ITERATIONS COMPLETED BY VA04A)
      IF (F-FKEEP) 20,20,110
  110 F=FKEEP
      DO 111 I=1,N
      JJJ=JJJ+1
      X(I)=W(JJJ)
  111 CONTINUE
      GO TO 20
  101 JIL=1
      FP=FKEEP
      IF (F-FKEEP) 105,78,104
  104 JIL=2
      FP=F
      F=FKEEP
  105 IXP=JJ
      DO 113 I=1,N
      IXP=IXP+1
      K=IXP+N
      GO TO (114,115),JIL
  114 W(IXP)=W(K)
      GO TO 113
  115 W(IXP)=X(I)
      X(I)=W(K)
  113 CONTINUE
      JIL=2
      GO TO 92
  106 IF (AAA-0.1) 20,20,107
   20 CONTINUE
      IF (F .LT. SELF%BEST_SCORE) THEN
        SELF%BEST_SCORE = F
        SELF%BEST_VALUES = X
      ENDIF
      RETURN
  107 INN=1
      GO TO 35
  998 CONTINUE
      DO 997 I=1,N
        X(I)=XS(I)
  997 CONTINUE

      F = SELF%FUNCTION(START_VAL_C,SELF%PARAMS_FOR_SCORING)
      SELF%NUM_FUNCTION_CALLS = SELF%NUM_FUNCTION_CALLS + 1
      IF (DEBUG .AND. SIZE(X) .LT. 63) WRITE(*,'(A,'//SIZE_XP//'(G15.5,1X))') '**debug(VA04A): trying these values: ', X,F
	  IF (DEBUG .AND. SIZE(X) .GE. 63) WRITE(*,'(A,F0.6)') '**debug(VA04A): got this score: ', F
	  IF (self%print_progress) THEN
#ifdef __INTEL_COMPILER
        WRITE(*,'(3(A,I0),2(A,F0.4),20X,A,$)') &
#else
        WRITE(*,'(3(A,I0),2(A,F0.4),20X,A)',ADVANCE='NO') &
#endif
                        '**info(VA04): scoring function call #', SELF%NUM_FUNCTION_CALLS, &
                        ' on thread ', thread_num+1, ' of ', num_of_threads, &
                        '. Latest score = ', F, &
                        '. Current best score = ', SELF%BEST_SCORE, char(13)
      ENDIF
	  ! IF (F.LT.SELF%BEST_SCORE) THEN
      !   SELF%BEST_SCORE = F
      !   SELF%BEST_VALUES = X
      ! ENDIF

      PRINT *,'VA04A ENDLESS LOOP SAFETY CATCH: ICNT = ',ICNT
      IF (F .LT. SELF%BEST_SCORE) THEN
        SELF%BEST_SCORE = F
        SELF%BEST_VALUES = X
      ENDIF
      RETURN
      END SUBROUTINE VA04A
!**************************************************************************


end module ConjugateGradients
