!>  \brief  Gauss-Legendre quadrature tools, to evaluate integrals of polynomials
!!
!! http://stackoverflow.com/questions/26456455/bezier-and-b-spline-arc-length-algorithm-giving-me-problems-solved
!! https://pomax.github.io/bezierinfo/legendre-gauss.html
module GaussLegendreQuadratures

    use Globals

    implicit none

    private

    public :: GaussLegendreQuadrature, GaussLegendreQuadrature_UnitTests

    type GaussLegendreQuadrature
        private
        integer             ::  number_of_points    !<  The quadrature will be exacts for polynomials of order 2*n-1 or less
        real,   allocatable ::  abscissae(:)
        real,   allocatable ::  weights(:)
        contains
            procedure,  public  ::  Init
            procedure,  public  ::  Integrate
            procedure,  public  ::  IsAllocated
            procedure,  public  ::  ArcLengthIntegral
            procedure,  public  ::  Deallocate
    end type

    !>  \brief  What an integratable function should look like
    abstract interface
        pure function integratable_function(x)
            real,   intent(in)  ::  x
            real                ::  integratable_function
        end function
    end interface

#include "gauss_legendre_abscissae.f90"
#include "gauss_legendre_weights.f90"

    contains

    pure real function GL_test_function_1(x)
        real,   intent(in)  ::  x
        GL_test_function_1 = 1.5 * x - 2.0
    end function

    pure real function GL_test_function_2(x)
        real,   intent(in)  ::  x
        GL_test_function_2 = 3.14 * x**2 - 2.3 * x + 2201.44
    end function

    pure real function GL_test_function_3(x)
        real,   intent(in)  ::  x
        GL_test_function_3 = sqrt(0.9912*x**4 - 0.300211*x**3 + 1.0002*x**2 - 2.3*x + 1.44)
    end function

    !>  \brief  Unit tests for Gauss-Legendre quadrature
    subroutine GaussLegendreQuadrature_UnitTests()
        type(GaussLegendreQuadrature)   ::  self
        real    ::  test_result
        real    ::  expected_result
        ! Start work

        ! Test # 1
        call self%Init(1)
        test_result = self%Integrate(GL_test_function_1,0.0,1.0)
        expected_result = -1.25
        print *, test_result, expected_result
        if (abs(test_result - expected_result) .gt. abs(expected_result/10000.0)) then
            call this_program%TerminateWithFatalError('GaussLegendreQuadrature::UnitTests','Failed test # 1')
        endif

        ! Test # 2
        call self%Init(2)
        test_result = self%Integrate(GL_test_function_2,3.2,43.0)
        expected_result = 168686.0
        print *, test_result, expected_result
        if (abs(test_result - expected_result) .gt. abs(expected_result/10000.0)) then
            call this_program%TerminateWithFatalError('GaussLegendreQuadrature::UnitTests','Failed test # 2')
        endif

        ! Test # 3
        call self%Init(3)
        test_result = self%Integrate(GL_test_function_3,3.2,4.0)
        expected_result = 10.1169
        print *, test_result, expected_result
        if (abs(test_result - expected_result) .gt. abs(expected_result/10000.0)) then
            call this_program%TerminateWithFatalError('GaussLegendreQuadrature::UnitTests','Failed test # 3')
        endif

    end subroutine

    !>  \brief  Setup Gauss-Legendre quadrature
    subroutine Init(self,number_of_points)
        class(GaussLegendreQuadrature), intent(inout)   ::  self
        integer,                        intent(in)      ::  number_of_points
        !
        self%number_of_points = number_of_points
        !
        select case (self%number_of_points)
            case (1)
                self%abscissae = abscissae_1
                self%weights = weights_1
            case (2)
                self%abscissae = abscissae_2
                self%weights = weights_2
            case (3)
                self%abscissae = abscissae_3
                self%weights = weights_3
            case (4)
                self%abscissae = abscissae_4
                self%weights = weights_4
            case (5)
                self%abscissae = abscissae_5
                self%weights = weights_5
            case (6)
                self%abscissae = abscissae_6
                self%weights = weights_6
            case (7)
                self%abscissae = abscissae_7
                self%weights = weights_7
            case (8)
                self%abscissae = abscissae_8
                self%weights = weights_8
            case (9)
                self%abscissae = abscissae_9
                self%weights = weights_9
            case (10)
                self%abscissae = abscissae_10
                self%weights = weights_10
            case (11)
                self%abscissae = abscissae_11
                self%weights = weights_11
            case (12)
                self%abscissae = abscissae_12
                self%weights = weights_12
            case (13)
                self%abscissae = abscissae_13
                self%weights = weights_13
            case (14)
                self%abscissae = abscissae_14
                self%weights = weights_14
            case (15)
                self%abscissae = abscissae_15
                self%weights = weights_15
            case (16)
                self%abscissae = abscissae_16
                self%weights = weights_16
            case (17)
                self%abscissae = abscissae_17
                self%weights = weights_17
            case (18)
                self%abscissae = abscissae_18
                self%weights = weights_18
            case (19)
                self%abscissae = abscissae_19
                self%weights = weights_19
            case (20)
                self%abscissae = abscissae_20
                self%weights = weights_20
            case default
                call this_program%TerminateWithFatalError('GaussLegendreQuadrature','Number of points not implemented')
        end select
    end subroutine Init

    subroutine Deallocate(self)
        class(GaussLegendreQuadrature), intent(inout)   ::  self
        if (allocated(self%abscissae)) deallocate(self%abscissae)
        if (allocated(self%weights)) deallocate(self%weights)
        self%number_of_points = 0
    end subroutine

    pure logical function IsAllocated(self)
        class(GaussLegendreQuadrature), intent(in)      ::  self
        IsAllocated = self%number_of_points .gt. 0 .and. allocated(self%abscissae) .and. allocated(self%weights)
    end function

    !>  \brief  Integrate an arbitrary function
    pure real function Integrate(self,function_to_integrate,lower_bound,upper_bound)
        class(GaussLegendreQuadrature), intent(in)  ::  self
        procedure(integratable_function)            ::  function_to_integrate
        real,                           intent(in)  ::  lower_bound
        real,                           intent(in)  ::  upper_bound
        ! private variables
        integer ::  counter
        real    ::  t
        ! start work
        Integrate = 0.0
        do counter=1,self%number_of_points
            t = 0.5*(upper_bound+lower_bound+self%abscissae(counter)*(upper_bound-lower_bound))
            Integrate = Integrate + self%weights(counter)*function_to_integrate(t)
        enddo
        Integrate = 0.5 * (upper_bound-lower_bound) * Integrate
    end function Integrate

    !>  \brief  Compute the arc length integral of a cubic polynomial space curve, given the first derivatives of the polynomials in each dimension
    !!
    !!  The first derivatives of the x, y and z cubic polynomials have the form :
    !!  x = a_x t^2 + b_x t + c_x
    !!  y = a_y t^2 + b_y t + c_y
    !!  z = a_z t^2 + b_z t + c_z
    !!
    !!  It is assumed that the GaussLegendreQuadrature was setup with at least 3 points
    !!
    !!  \todo   Currently, we have mixed double and single-precision arithmetics in there. Either call with single-precision, or convert GL to double
    pure real function ArcLengthIntegral(self,a_x,b_x,c_x,a_y,b_y,c_y,a_z,b_z,c_z,lower_bound,upper_bound)
        class(GaussLegendreQuadrature), intent(in)  ::  self
        real(kind=8),                   intent(in)  ::  a_x,b_x,c_x !<  The coefficient for the derivative of the x cubic polynomial
        real(kind=8),                   intent(in)  ::  a_y,b_y,c_y
        real(kind=8),                   intent(in)  ::  a_z,b_z,c_z
        real(kind=8),                   intent(in)  ::  lower_bound !<  Lower bound for the integration
        real(kind=8),                   intent(in)  ::  upper_bound !<  Upper bound for the integration
        !
        integer ::  counter
        real    :: t
        !
        ArcLengthIntegral = 0.0
        do counter=1,self%number_of_points
            t = 0.5*(upper_bound+lower_bound+self%abscissae(counter)*(upper_bound-lower_bound))
            ArcLengthIntegral = ArcLengthIntegral + self%weights(counter) &
                                *sqrt((a_x*t**2+b_x*t+c_x)**2+(a_y*t**2+b_y*t+c_y)**2+(a_z*t**2+b_z*t+c_z)**2)
        enddo
        ArcLengthIntegral = 0.5 * (upper_bound-lower_bound) * ArcLengthIntegral
    end function ArcLengthIntegral



end module GaussLegendreQuadratures
