!>  \brief  Methods for the image class that add noise and depend on the GSL
module ImageNoise

    contains

        !>  \brief  replace each value in an image by a value randomly selected from a poisson distribution
    !!          with parameter equal to the input value. output therefore has integer values.
    subroutine image_poissonify(self)
        use fgsl
        use Images
        ! arguments
        type(image),    intent(inout)   ::  self
        ! private variables
        type(fgsl_rng)          ::  rng
        type(fgsl_rng_type)     ::  default_rng
        integer                 ::  i,j,k
        ! start work

        ! create a generator (the algorithm is chosen by the environment variable gsl_rng_type
        default_rng = fgsl_rng_env_setup()
        rng         = fgsl_rng_alloc(default_rng)

        ! loop over the voxels
        do k=1,self%GetLogicalDimension(3)
            do j=1,self%GetLogicalDimension(2)
                do i=1,self%GetLogicalDimension(1)
                    self%real_values(i,j,k) = real(fgsl_ran_poisson(rng,real(self%real_values(i,j,k),fgsl_double)))
                enddo
            enddo
        enddo
    end subroutine image_poissonify

    !>  \brief  for each count in input image, add a value selected from a landau distribution
    subroutine image_landauify(self)
        use fgsl
        use Images
        ! arguments
        type(image),    intent(inout)   ::  self
        ! private variables
        type(fgsl_rng)          ::  rng
        type(fgsl_rng_type)     ::  default_rng
        integer                 ::  i,j,k,e
        integer                 ::  n_electrons
        ! start work

        ! create a generator (the algorithm is chosen by the environment variable gsl_rng_type
        default_rng = fgsl_rng_env_setup()
        rng         = fgsl_rng_alloc(default_rng)

        ! loop over the voxels
        do k=1,self%GetLogicalDimension(3)
            do j=1,self%GetLogicalDimension(2)
                do i=1,self%GetLogicalDimension(1)
                    n_electrons = int(self%real_values(i,j,k))
                    self%real_values(i,j,k) = 0.0
                    do e=1,n_electrons
                        self%real_values(i,j,k) = self%real_values(i,j,k) + fgsl_ran_landau(rng)
                    enddo
                enddo
            enddo
        enddo
    end subroutine image_landauify

    !>  \brief  for each count in input image, add a value selected from a landau distribution
    subroutine image_landauify_with_thresholding(self,threshold)
        use fgsl
        use Images
        ! arguments
        type(image),    intent(inout)   ::  self
        real,           intent(in)      ::  threshold   !<  For every simulated electron event, only those above this threshold will be registered
        ! private variables
        type(fgsl_rng)          ::  rng
        type(fgsl_rng_type)     ::  default_rng
        integer                 ::  i,j,k,e
        integer                 ::  n_electrons
        ! start work

        ! create a generator (the algorithm is chosen by the environment variable gsl_rng_type
        default_rng = fgsl_rng_env_setup()
        rng         = fgsl_rng_alloc(default_rng)

        ! loop over the voxels
        do k=1,self%GetLogicalDimension(3)
            do j=1,self%GetLogicalDimension(2)
                do i=1,self%GetLogicalDimension(1)
                    n_electrons = int(self%real_values(i,j,k))
                    self%real_values(i,j,k) = 0.0
                    do e=1,n_electrons
                        if (fgsl_ran_landau(rng) .gt. threshold) then
                            self%real_values(i,j,k) = self%real_values(i,j,k) + 1.0
                        endif
                    enddo
                enddo
            enddo
        enddo
    end subroutine image_landauify_with_thresholding

    !>  \brief adds gaussian noise to an image. stdev_noise is a real.
    subroutine AddGaussianNoise(img, stdev_noise,seed)
        use fgsl
        use Images
        ! arguments
        class(image) ,          intent(inout)   ::  img
        real        ,           intent(in)      ::  stdev_noise
        integer,    optional,   intent(in)      ::  seed        !<  if none supplied, default to 0
        ! private variables
        integer                         :: i, j, k
        type(fgsl_rng)          ::  rng
        type(fgsl_rng_type)     ::  default_rng
        real(fgsl_double)       ::  stdev_noise_fgsl
        integer(kind=8)         ::  sseed

        sseed = 0
        if (present(seed)) sseed = seed

        !if you inputted snr, this is how noise variance would be calculated.
        !grey_var = imgin%ReturnVariance()
        !noise_var=grey_var/snr

        stdev_noise_fgsl=stdev_noise
        default_rng= fgsl_rng_env_setup()
        rng=fgsl_rng_alloc(default_rng)
        call fgsl_rng_set(rng,sseed)

        do i=1,img%GetLogicalDimension(1)
            do j=1,img%GetLogicalDimension(2)
                do k=1,img%GetLogicalDimension(3)
                    img%real_values(i,j,k)=img%real_values(i,j,k)+fgsl_ran_gaussian(rng, stdev_noise_fgsl)
                enddo
            enddo
        enddo

        call fgsl_rng_free(rng)
    end subroutine AddGaussianNoise

end module ImageNoise
