
# Facilitate running of mode FILAMENT_REFINE in parallel
class FilamentRefineParallel
  NAME="Parallel runner for mode FILAMENT_REFINE"

  # Stuff we need
  require_relative "FrealixRunner.rb"
  require_relative "rubylix.rb"
  require 'fileutils'

  # Initialise
  def initialize( options = {} )
    # Default values for optional arguments
    defaults = {
      :dir                =>  FrealixRefinementDirectory.new(),
      :round_number 	    => 	FrealixRefinementDirectory.new().currentRound(),
      :num_nodes		      => 	1,
      :num_nodes_rec      =>  1,
      :num_threads 	      => 	1,
      :params			        =>	FrealixParameters.new(),
      :hosts              =>  ["cluster"],
      :mem_per_thread_gb	=>	1,
      :prefix             => "flx",
      :gpu_nodes_only     =>  false,
      :resume             =>  false,
      :num_classes        =>  1,
      :extra_qsub_options =>  ""}
    options = defaults.merge(options)
    # Instance variables
    @dir		            = options[:dir]
    @round              = options[:round_number]
    @num_nodes          = options[:num_nodes]
    @num_nodes_rec      = options[:num_nodes_rec]
    @num_threads        = options[:num_threads]
    @mem_free	          = options[:mem_per_thread_gb]	# Free memory required per thread, in GB
    @params	            = options[:params]
    @avail_hosts        = options[:hosts]
    @prefix             = options[:prefix]
    @gpu_nodes_only     = options[:gpu_nodes_only]
    @resume             = options[:resume]
    @num_classes        = options[:num_classes]
    @extra_qsub_options = options[:extra_qsub_options]
    if (@params.rtp["SKIP_REFINEMENT"] == "T")
      @we_are_refining = false
    else
      @we_are_refining = true
    end
    $LOG.debug("FilamentRefineParalle init, avail_hosts = #{@avail_hosts}")
  end

  # Do the processing for the round
  def parallelRun()
    $LOG.info("parallelRun") {"Round #{@round}: Processing with #{@num_nodes} nodes, each with #{@num_threads} threads"}
    # The subdirectory for the current & previous rounds
    round_dir = @dir.currentRoundDir()
    previous_round_dir = @dir.previousRoundDir()
    #
    if (@we_are_refining)
      if (!@resume) # we are not resuming
        # Check the previous directory has the expected number of classes
        if (@round > 0)
          number_of_classes_in_previous_round = @dir.numberOfClassesInRound(@round - 1 )
          if ( number_of_classes_in_previous_round != @num_classes)
            if (number_of_classes_in_previous_round == 0 && @num_classes == 1 )
              # This means that the previous round was a single-class round (before multi-ref was enabled). This can be done as long as we're only doing 1 class refinement
              FileUtils.cp previous_round_dir+"/waypoints.ali", previous_round_dir+"/waypoints_r001.ali"
              FileUtils.cp previous_round_dir+"/3drec.mrc", previous_round_dir+"/3drec_r001.mrc"
            else
              $LOG.fatal("parallelRun") {"Round #{@round-1} has #{number_of_classes_in_previous_round} classes, but you requested #{@num_classes} for this round. Changing number of classes during refinement is not yet implemented. Sorry."}
              exit 1
            end
          end
        end
        # Split the .ali file
        $LOG.info("parallelRun") {"Round #{@round}: Splitting the input alignments file into #{@num_nodes}"}
        if (@round == 0)
          round_dir_to_use = round_dir
        else
          round_dir_to_use = previous_round_dir
        end
        if ($enable_multi_ref)
          num_nodes_per_class = self.nodesPerClass(@num_nodes)
          1.upto(@num_classes) do |class_counter|
            self.splitAli(round_dir_to_use+"/waypoints_r"+"%03d"%class_counter+".ali",num_nodes_per_class[class_counter-1],@params.rtp['preserve_order_of_filaments'])
          end
        else
          self.splitAli(round_dir_to_use+"/waypoints.ali",@num_nodes,@params.rtp['preserve_order_of_filaments'])
        end
      else # we are resuming
        $LOG.info("parallelRun") {"Round #{@round}: Resuming interrupted refinement. Will re-use split waypoints files."}
        # Let's check we have the correct number of
        # input ali files
        if (@num_nodes != countSplitAli(previous_round_dir+"/waypoints.ali"))
          $LOG.fatal("parallelRun") {"Round #{@round}: Was expecting #{@num_nodes} input ali files, but found #{countSplitAli(previous_round_dir+"/waypoints.ali")} in #{previous_round_dir}. Will not be able to resume processing."}
          exit 1
        end
      end
      # Generate the array of parameter & log files
      p = Array.new(@num_nodes,FrealixParameters)
      l = Array.new(@num_nodes,FrealixOutput)
      if ($enable_multi_ref)
        #
        num_nodes_per_class = self.nodesPerClass(@num_nodes)
        i = -1
        0.upto(@num_classes-1) do |class_counter|
          class_number = "%03i"%(class_counter+1)
          0.upto(num_nodes_per_class[class_counter]-1) do |node_counter|
            node_number = "%03i"%(node_counter+1)
            i += 1
            # Prepare parameters for this node
            p[i]= @params.dclone
            p[i].filename = round_dir+"/frealix_parameters_r"+class_number+"_"+node_number+".flx"
            if (@round == 0)
              p[i].rtp["INPUT_PAR_FILENAME"]  = round_dir+"/waypoints_r"+class_number+"_"+node_number+".ali"
            else
              p[i].rtp["INPUT_PAR_FILENAME"]  = previous_round_dir+"/waypoints_r"+class_number+"_"+node_number+".ali"
            end
            p[i].rtp["MODEL_3D_FILENAME"]  = previous_round_dir+"/3drec_r"+class_number+".mrc"
            p[i].rtp["MODEL_3D_FSC_FILENAME"] = previous_round_dir+"/3drec_r"+class_number+"_fsc.dat"
            p[i].rtp["OUTPUT_PAR_FILENAME"] = round_dir+"/waypoints_r"+class_number+"_"+node_number+".ali"
            p[i].rtp["OUTPUT_3D_FILENAME"]  = round_dir+"/3drec_r"+class_number+"_"+node_number+".mrc"
            p[i].rtp["FINALISE_3D_RECONSTRUCTION"] = "F"
            p[i].rtp["CALC_3D_RECONSTRUCTION"] = "F"
            # Prepare output log file
            l[i] = FrealixOutput.new(round_dir+"/frealix_output_r"+class_number+"_"+node_number+".log")
          end
        end
      else
        0.upto(@num_nodes-1) do |node_counter|
          num = sprintf("%03i",node_counter+1)
          # Prepare parameters for this node
          p[node_counter]= @params.dclone
          p[node_counter].filename = round_dir+"/frealix_parameters_"+num+".flx"
          if (@round == 0)
            p[node_counter].rtp["INPUT_PAR_FILENAME"]  = round_dir+"/waypoints_"+num+".ali"
          else
            p[node_counter].rtp["INPUT_PAR_FILENAME"]  = previous_round_dir+"/waypoints_"+num+".ali"
          end
          p[node_counter].rtp["MODEL_3D_FILENAME"]  = previous_round_dir+"/3drec.mrc"
          p[node_counter].rtp["MODEL_3D_FSC_FILENAME"] = previous_round_dir+"/3drec_fsc.dat"
          p[node_counter].rtp["OUTPUT_PAR_FILENAME"] = round_dir+"/waypoints_"+num+".ali"
          p[node_counter].rtp["OUTPUT_3D_FILENAME"]  = round_dir+"/3drec_"+num+".mrc"
          p[node_counter].rtp["FINALISE_3D_RECONSTRUCTION"] = "F"
          p[node_counter].rtp["CALC_3D_RECONSTRUCTION"] = "F"
          # Prepare output log file
          l[node_counter] = FrealixOutput.new(round_dir+"/frealix_output_"+num+".log")
        end
      end
      # Prepare the frealix runner
      $LOG.debug("parallelRun") {"Starting runner with #{@num_threads} threads, each requesting #{@mem_free} GB"}
      flx_runner = FrealixRunner.new(
      :num_jobs=>@num_nodes,
      :params=>p,
      :logs=>l,
      :hosts=>@avail_hosts,
      :threads_per_job=>@num_threads,
      :mem_per_thread_gb=>@mem_free,
      :prefix=>@prefix,
      :gpu_nodes_only=>@gpu_nodes_only,
      :extra_qsub_options=>@extra_qsub_options,
      :resume=>@resume)
      # Run frealix
      $LOG.info("parallelRun") {"Round #{@round}: Starting refinement on #{@num_nodes} nodes"}
      run_success = flx_runner.run()
      if (run_success)
        # Now that all jobs are done, we need to compile the results & cleanup
        $LOG.info("parallelRun") {"Round #{@round}: Now that all refinement jobs are done, we need to compile the results & cleanup"}
        # We merge the .ali files
        $LOG.info("parallelRun") {"Round #{@round}: Merging the .ali files..."}
        if ($enable_multi_ref)
          0.upto(@num_classes-1) do |class_counter|
            class_number = "%03i"%(class_counter+1)
            self.mergeAli(round_dir+"/waypoints_r"+class_number+".ali",num_nodes_per_class[class_counter])
          end
        else
          self.mergeAli(round_dir+"/waypoints.ali",@num_nodes)
        end
      else
        #Not all refinement jobs succeeded. Let's crash out
        $LOG.fatal("parallelRun") { "Not all refinement jobs in round #{@round} were succesful. See output above for detail."}
        exit( -1)
      end
    else
      # we are not refining, so let's just copy the waypoints.ali file
      $LOG.info("parallelRun") { "No refinement, skipping to 3D reconstruction" }
      if (@round > 0)
        if $enable_multi_ref
          0.upto(@num_classes-1) do |class_counter|
            class_number = "%03i"%(class_counter+1)
            FileUtils.cp previous_round_dir+"/waypoints_r"+class_number+".ali", round_dir+"/waypoints_r"+class_number+".ali"
          end
        else
          FileUtils.cp previous_round_dir+"/waypoints.ali", round_dir+"/waypoints.ali"
        end
      end
    end # end of test for @we_are_refining
    #
    #
    # Now refinement has been done, we need to do the reconstruction
    #
    #
    # Split the .ali file
    if $enable_multi_ref
      1.upto(@num_classes) do |class_counter|
        class_number = "%03i"%(class_counter)
        FileUtils.cp round_dir+"/waypoints_r"+class_number+".ali", round_dir+"/waypoints_rec_r"+class_number+".ali"
      end
    else
      FileUtils.cp round_dir+"/waypoints.ali", round_dir+"/waypoints_rec.ali"
    end
    @num_nodes_rec = 1 if (@num_nodes_rec < 1)
    $LOG.info("parallelRun") {"Round #{@round}: Splitting the output alignment file into #{@num_nodes_rec} for 3D reconstruction"}
    if $enable_multi_ref
      num_nodes_per_class = self.nodesPerClass(@num_nodes_rec)
      1.upto(@num_classes) do |class_counter|
        self.splitAli(round_dir+"/waypoints_rec_r"+"%03d"%class_counter+".ali",num_nodes_per_class[class_counter-1],@params.rtp['preserve_order_of_filaments'])
      end
    else
      self.splitAli(round_dir+"/waypoints_rec.ali",@num_nodes_rec,@params.rtp['preserve_order_of_filaments'])
    end
    p = Array.new(@num_nodes_rec,FrealixParameters)
    l = Array.new(@num_nodes_rec,FrealixOutput)
    if ($enable_multi_ref)
      num_nodes_per_class = self.nodesPerClass(@num_nodes_rec)
      i = 0
      0.upto(@num_classes-1) do |class_counter|
        class_number = "%03i"%(class_counter+1)
        0.upto(num_nodes_per_class[class_counter]-1) do |node_counter|
          node_number = "%03i"%(node_counter+1)
          # Prepare parameters for this node
          p[i]= @params.dclone
          p[i].filename = round_dir+"/frealix_parameters_rec_r"+class_number+"_"+node_number+".flx"
          p[i].rtp["INPUT_PAR_FILENAME"] = round_dir+"/waypoints_rec_r"+class_number+"_"+node_number+".ali"
          p[i].rtp["OUTPUT_PAR_FILENAME"] = round_dir+"/waypoints_rec_out_r"+class_number+"_"+node_number+".ali"
          p[i].rtp["OUTPUT_3D_FILENAME"]  = round_dir+"/3drec_r"+class_number+"_"+node_number+".mrc"
          p[i].rtp["FINALISE_3D_RECONSTRUCTION"] = "F"
          p[i].rtp["CALC_3D_RECONSTRUCTION"] = "T"
          p[i].rtp["MODEL_3D_AVAILABLE"]  = "F"
          p[i].rtp["SKIP_REFINEMENT"] = "T"
          p[i].rtp["recompute_subframe_alignments"] = "f"
          # Prepare output log file
          l[i] = FrealixOutput.new(round_dir+"/frealix_output_rec_r"+class_number+"_"+node_number+".log")
          i += 1
        end
      end
    else
      0.upto(@num_nodes_rec-1) do |node_counter|
        num = sprintf("%03i",node_counter+1)
        # Prepare parameters for this node
        p[node_counter]= @params.dclone
        p[node_counter].filename = round_dir+"/frealix_parameters_rec_"+num+".flx"
        p[node_counter].rtp["INPUT_PAR_FILENAME"]  = round_dir+"/waypoints_rec_"+num+".ali"
        p[node_counter].rtp["OUTPUT_PAR_FILENAME"] = round_dir+"/waypoints_rec_out_"+num+".ali"
        p[node_counter].rtp["OUTPUT_3D_FILENAME"]  = round_dir+"/3drec_"+num+".mrc"
        p[node_counter].rtp["FINALISE_3D_RECONSTRUCTION"] = "F"
        p[node_counter].rtp["MODEL_3D_AVAILABLE"]  = "F"
        p[node_counter].rtp["SKIP_REFINEMENT"] = "T"
        p[node_counter].rtp["CALC_3D_RECONSTRUCTION"] = "T"
        p[node_counter].rtp["recompute_subframe_alignments"] = "f"
        # Prepare output log file
        l[node_counter] = FrealixOutput.new(round_dir+"/frealix_output_rec_"+num+".log")
      end
    end
    # Prepare the frealix runner
    $LOG.debug("parallelRun") {"Starting runner with #{@num_threads_rec} threads, each requesting #{@mem_free} GB"}
    flx_runner = FrealixRunner.new(
    :num_jobs=>@num_nodes_rec,
    :params=>p,
    :logs=>l,
    :hosts=>@avail_hosts,
    :threads_per_job=>@num_threads,
    :mem_per_thread_gb=>@mem_free,
    :prefix=>@prefix,
    :gpu_nodes_only=>@gpu_nodes_only,
    :extra_qsub_options=>@extra_qsub_options,
    :resume=>@resume)
    # Run frealix
    $LOG.info("parallelRun") {"Round #{@round}: Starting reconstruction on #{@num_nodes_rec} nodes"}
    run_success = flx_runner.run()
    if (run_success)
      # Now that all jobs are done, we need to compile the results & cleanup
      $LOG.info("parallelRun") {"Round #{@round}: Now that all reconstruction jobs are done, we need to compile the results & cleanup"}
      if ($enable_multi_ref)
        p = Array.new(@num_classes,FrealixParameters)
        l = Array.new(@num_classes,FrealixOutput)
        0.upto(@num_classes-1) do |class_counter|
          class_number = "%03i"%(class_counter+1)
          # First, we merge the .ali files
          $LOG.info("parallelRun") {"Round #{@round}: Merging the .ali files for class #{class_counter}..."}
          self.mergeAli(round_dir+"/waypoints_rec_r"+class_number+".ali",num_nodes_per_class[class_counter])

          p[class_counter] = @params.dclone
          p[class_counter].filename = round_dir+"/frealix_parametrers_r"+class_number+"_finalise.flx"
          p[class_counter].rtp["FREALIX_MODE"] = "FINALISE_3DS"
          p[class_counter].rtp["INPUT_THREED_FILES"] = round_dir+"/3drec_r"+class_number+"_???.mrc"
          p[class_counter].rtp["OUTPUT_THREED_FILENAME"] = round_dir+"/3drec_r"+class_number+".mrc"
          p[class_counter].rtp["INPUT_THREED_SELECTION"] = "1-#{num_nodes_per_class[class_counter]}"
          p[class_counter].rtp["HELICAL_MASKING"] = "T"
          p[class_counter].rtp["CALC_FSC"] = "T"
          l[class_counter] = FrealixOutput.new(round_dir+"/frealix_output_r"+class_number+"_finalise.log")
        end
        $LOG.info("parallelRun") {"Round #{@round}: Finalising the 3D reconstructions..."}
        flx_runner = FrealixRunner.new( :num_jobs=>@num_classes,:params=>p,
                                        :logs=>l,
                                        :hosts=>@avail_hosts,:prefix=>@prefix)
      else
        # First, we merge the .ali files
        $LOG.info("parallelRun") {"Round #{@round}: Merging the .ali files..."}
        self.mergeAli(round_dir+"/waypoints_rec.ali",@num_nodes_rec) # We don't actually need this file - leaving here for debugging
        # Now, we need to finalise the 3D
        $LOG.info("parallelRun") {"Round #{@round}: Finalising the 3D reconstruction..."}
        p = Array.new(1,FrealixParameters)
        l = Array.new(1,FrealixOutput)
        l[0] = FrealixOutput.new(round_dir+"/frealix_output_finalise.log")
        p[0] = @params.dclone
        p[0].filename = round_dir+"/frealix_parameters_finalise.flx"
        p[0].rtp["FREALIX_MODE"] = "FINALISE_3DS"
        p[0].rtp["INPUT_THREED_FILES"] = round_dir+"/3drec_???.mrc"
        p[0].rtp["OUTPUT_THREED_FILENAME"] = round_dir+"/3drec.mrc"
        p[0].rtp["INPUT_THREED_SELECTION"] = "1-#{@num_nodes_rec}"
        p[0].rtp["HELICAL_MASKING"] = "T"
        p[0].rtp["CALC_FSC"] = "T"
        # Prepare the frealix runner
        flx_runner = FrealixRunner.new( :num_jobs=>1,:params=>p,
        :logs=>l,
        :hosts=>@avail_hosts,:prefix=>@prefix)
      end
      # Run frealix
      run_success = flx_runner.run()
      if (run_success)
        # Clean up the directory
        $LOG.info("parallelRun") { "Round #{@round}: Cleaning up..."}
        subdirCleanup(round_dir)
        subdirCleanup(previous_round_dir)
      else
        $LOG.fatal("parallelRun") { "Finalisation job in round #{@round} was not succesful. See output above for detail."}
        exit (-1)
      end
      # Let's compute an FSC between the reconstruction and the previous reconstruction
      if ($enable_multi_ref)
        num_nodes_per_class = self.nodesPerClass(@num_nodes_rec)
        i = 0
        p = Array.new(@num_classes,FrealixParameters)
        l = Array.new(@num_classes,FrealixOutput)
        0.upto(@num_classes-1) do |class_counter|
          class_number = "%03i"%(class_counter+1)
          if (File.exists?(previous_round_dir+"/3drec_r"+class_number+"_hm.mrc") && File.exists?(round_dir+"/3drec_r"+class_number+"_hm.mrc"))
            $LOG.info("parllelRun") { "Round #{@round}: Computing FSCs to previous round." }
            class_number = "%03i"%(class_counter+1)
            l[class_counter] = FrealixOutput.new(round_dir+"/frealix_output_r"+class_number+"_fsc.log")
            p[class_counter] = @params.dclone
            p[class_counter].filename = round_dir+"/frealix_parameters_r"+class_number+"_fsc.flx"
            p[class_counter].rtp["FREALIX_MODE"] = "CALC_FSC"
            p[class_counter].rtp["INPUT_VOLUME_1"] = previous_round_dir+"/3drec_r"+class_number+"_hm.mrc"
            p[class_counter].rtp["INPUT_VOLUME_2"] = round_dir+"/3drec_r"+class_number+"_hm.mrc"
          else
            $LOG.debug("parallelRun") { "Round #{@round}: Not doing FSC to previous round for class #{class_number} because could not find 3D"}
          end
        end # end of loop over classes

        # Prepare the frealix runner
        flx_runner = FrealixRunner.new( :num_jobs=>1,:params=>p,
                                        :logs=>l,
                                        :hosts=>@avail_hosts,:prefix=>@prefix)
        # Run frealix
        run_success = flx_runner.run()
        if (run_success)
            # Parse the FSCs
            0.upto(@num_classes-1) do |class_counter|
              fsc_values = l[class_counter].getFSCValues()
              ave_fsc = fsc_values.reduce(:+) / fsc_values.length.to_f
              $LOG.info("parallelRun") {"Round #{@round}: average value of FSC to previous round for class #{class_counter} = #{ave_fsc}"}
              $LOG.info("parallelRun") {"Round #{@round}: it looks like refinement of class #{class_counter} has converged"} if (ave_fsc > 0.85)
            end  # end of loop over classes
        else
            $LOG.fatal("parallelRun") { "Round #{@round}: FSC to previous round failed. See #{l[0].filename} to #{l[@num_classes-1].filename}" }
            exit(-1)
        end
      else # No multi-reference refinement
        if (File.exists?(previous_round_dir+"/3drec_hm.mrc") && File.exists?(round_dir+"/3drec_hm.mrc"))
          $LOG.info("parllelRun") { "Round #{@round}: Computing FSC to previous round"}
          p = Array.new(1,FrealixParameters)
          l = Array.new(1,FrealixOutput)
          l[0] = FrealixOutput.new(round_dir+"/frealix_output_fsc.log")
          p[0] = @params.dclone
          p[0].filename = round_dir+"/frealix_parameters_fsc.flx"
          p[0].rtp["FREALIX_MODE"] = "CALC_FSC"
          p[0].rtp["INPUT_VOLUME_1"] = previous_round_dir+"/3drec_hm.mrc"
          p[0].rtp["INPUT_VOLUME_2"] = round_dir+"/3drec_hm.mrc"
          # Prepare the frealix runner
          flx_runner = FrealixRunner.new( :num_jobs=>1,:params=>p,
          :logs=>l,
          :hosts=>@avail_hosts,:prefix=>@prefix)
          # Run frealix
          run_success = flx_runner.run()
          if (run_success)
            # Parse the FSC
            fsc_values = l[0].getFSCValues()
            ave_fsc = fsc_values.reduce(:+) / fsc_values.length.to_f
            $LOG.info("parallelRun") {"Round #{@round}: average value of FSC to previous round = #{ave_fsc}"}
            $LOG.info("parallelRun") {"Round #{@round}: it looks like refinement has converged"} if (ave_fsc > 0.85)
          else
            $LOG.error("parallelRun") { "Round #{@round}: FSC to previous round failed. See #{l[0].filename}" }
          end
        else
          $LOG.debug("parallelRun") { "Round #{@round}: Not doing FSC to previous round"}
        end
      end
      # All done
      $LOG.info("parallelRun") {"Round #{@round}: All done."}
    else
      $LOG.fatal("parallelRun") { "Not all jobs in round #{@round} were succesful. See output above for detail."}
      exit (-1)
    end
  end

  # Work out how many nodes will be assigned to processing of each class
  def nodesPerClass(total_num_nodes)
    nodes_per_class = Array.new(@num_classes)
    1.upto(@num_classes) do |class_number|
      nodes_per_class[class_number-1] = (total_num_nodes / @num_classes).floor
    end
    nodes_per_class[-1] += total_num_nodes - ((total_num_nodes / @num_classes).floor * @num_classes)
    $LOG.debug("nodesPerClass") {"Number of nodes per class: #{nodes_per_class}"}
    return nodes_per_class
  end

  # Cleanup a round subdirectory after parallel filament refinement
  def subdirCleanup(path)
    if ($enable_multi_ref)
      f = Dir.glob(["#{path}/3drec_r???_???_{unweighted,ctfweights,fsc,ampl}*.mrc","#{path}/3drec_r???_???_stats*.dat","#{path}/3drec_r???_???_fsc.dat","#{path}/3drec_r???_???_ssnr.dat","#{path}/waypoints_r???_???.ali","#{path}/waypoints_rec_out_r???_???.ali","#{path}/waypoints_rec_r???_???.ali"])
    else
      f = Dir.glob(["#{path}/3drec_???_{unweighted,ctfweights,fsc,ampl}*.mrc","#{path}/3drec_???_stats*.dat","#{path}/3drec_???_fsc.dat","#{path}/3drec_???_ssnr.dat"])
    end
    f.each do |x|
      File.delete(x)
    end
  end

  # Create copies of the ali file, with randomized occupancies. Useful when starting a classification
  def self.createClassesFromAli(ali_filename,num_classes)
    # Check the ali file actually exists
    if (!File.exists?(ali_filename))
      $LOG.fatal("splitAli") {"File does not exist: #{ali_filename}"}
      puts caller(start=0)
      exit 1
    end
    # The different parts of the filename are dir+"/"+base+ext
    dir  = File.dirname (ali_filename)
    ext  = File.extname (ali_filename)
    base = File.basename(ali_filename,ext)
    # Set up the parameters for frealix
    params = FrealixParameters.new(filename=dir+"/frealix_parameters_create_random_classes.flx")
    params.rtp = {
      "FREALIX_MODE"          =>  "create_random_classes",
      "INPUT_FILENAME"        =>  ali_filename,
      "OUTPUT_ROOT_FILENAME"  =>  dir+"/"+base+"_r???"+ext,
      "NUMBER_OF_GROUPS"      =>  num_classes
    }
    # Set up the output log file
    log = FrealixOutput.new(filename=dir+"/frealix_output_create_random_classes.log")
    # Set up the frealix run
    f = FrealixRunner.new(:params=>[params],:logs=>[log],:hosts=>@avail_hosts)
    # Run frealix
    success = f.run()
    exit (-1) if (!success)
  end

  # Split ali file
  def splitAli(ali_filename,num_output_files,param_preserve_order_of_filaments)
    # Check the ali file actually exists
    if (!File.exists?(ali_filename))
      $LOG.fatal("splitAli") {"File does not exist: #{ali_filename}"}
      puts caller(start=0)
      exit 1
    end
    if (param_preserve_order_of_filaments == "none") then
      preserve_order = "f"
    else
      preserve_order = param_preserver_order_of_filaments
    end
    # The different parts of the filename are dir+"/"+base+ext
    dir  = File.dirname (ali_filename)
    ext  = File.extname (ali_filename)
    base = File.basename(ali_filename,ext)
    # Set up the parameters for frealix
    params = FrealixParameters.new(filename=dir+"/frealix_parameters_split.flx")
    params.rtp = {
      "FREALIX_MODE"          =>  "SPLIT_PAR",
      "INPUT_FILENAME"        =>  ali_filename,
      "OUTPUT_ROOT_FILENAME"  =>  dir+"/"+base+"_???"+ext,
      "NUMBER_OF_GROUPS"      =>  num_output_files,
      "preserve_order_of_filaments" => preserve_order
    }
    # Set up the output log file
    log = FrealixOutput.new(filename=dir+"/frealix_output_split.log")
    # Set up the frealix run
    f = FrealixRunner.new(:params=>[params],:logs=>[log],:hosts=>@avail_hosts)
    # Run frealix
    success = f.run()
    # Give the filesystem a breather (on NFS, sometimes the first job after splitting starts so quickly that the files are not yet readable)
    sleep 3
    exit (-1) if (!success)
  end

  # Count the number of split ali files
  def countSplitAli(ali_filename)
    # The different parts of the filename are dir+"/"+base+ext
    dir  = File.dirname (ali_filename)
    ext  = File.extname (ali_filename)
    base = File.basename(ali_filename,ext)
    if ($enable_multi_ref)
      Dir.glob("#{dir}/#{base}_r[0-9][0-9][0-9]_[0-9][0-9][0-9]#{ext}").count
    else
      Dir.glob("#{dir}/#{base}_[0-9][0-9][0-9]#{ext}").count
    end
  end

  # Merge ali files together
  def mergeAli(ali_filename,num_input_files)
    # The different parts of the filename are dir+"/"+base+ext
    dir  = File.dirname (ali_filename)
    ext  = File.extname (ali_filename)
    base = File.basename(ali_filename,ext)
    # Open the output file for reading (make sure it is overwritten)
    f = File.new(ali_filename,"w")
    # Read all the files in
    1.upto(num_input_files) do |i|
      num = sprintf("%03i",i)
      input_filename = dir+"/"+base+"_"+num+ext
      str = File.read(input_filename)
      # Strip the first line, unless we're on the first file
      if (i >1)
        a = str.split("\n")
        a.delete_at(0)
        str = a.join("\n")
        str = str+"\n"
      end
      # Append to file
      f.print(str)
    end
    f.close()
  end

end
