# Dealing with frealix output
class FrealixOutput

  NAME = "Frealix output"

  attr_reader :str

  # Initialiser method with default parameter
  def initialize(filename="frealix_output.log")
    @filename = filename
    @file_last_read = Time.new(1900)	# We pretend the file was last read in 1900
    if  File.exist?(@filename)
      readFromDisk()
      findMode
    end
  end

  # Did the frealix run complete?
  def completed?()
    self.readFromDiskIfNecessary()
    self.successful? || self.error?
  end

  # Did the frealix run complete successfully?
  def successful?()
    self.readFromDiskIfNecessary()
    (@str =~ /goodbye/)?true:false
  end

  # Was there an error?
  def error?()
    self.readFromDiskIfNecessary()
    error_regex = /error while loading shared libraries|forrtl: severe|symbol lookup error|Abort|Fatal|Segmentation fault|Program terminated by fatal error|Killed|severe|Program aborted/
    (@str =~ error_regex)?true:false
  end


  # Read contents from disk if necessary
  def readFromDiskIfNecessary()
    # We will read from disk if:
    # - the file was modified since we last read it
    # OR
    # - the file exists and we never read it in
    if (File.exists?(@filename))
      if (@str == nil || @str == "")
        self.readFromDisk()
      else
        if ((@file_last_read <=> File.mtime(@filename)) == -1)
          self.readFromDisk()
        end
      end
    end
  end

  # Read contents from disk
  def readFromDisk()
    # Check file exists
    if File.exist?(@filename)
      # Read whole file's content into memory, since
      # we'll probably have to parse it several times anyway
      @file_last_read = Time.now()
      @str = File.read(@filename)
      @str.encode!('UTF-8', 'UTF-8', :invalid => :replace)
      @lines = @str.split(/\n/)
    else
      #puts "File does not exist: #{@filename}"
      #puts caller(start=0)
      #exit 1
    end
  end


  # Print out content
  def print()
    puts @str
  end

  # Instance variable with automatic getter method
  attr_reader :mode
  attr_accessor :filename

  # Grab the total change of direction
  def getChangeOfDirection(fil_nums=nil)
    fil_nums = getFilamentNumbers if (!fil_nums)
    changeOfDirection = []
    fil_nums.each do |fil_num|
      regex_final = /\*\*info\(mode_filament_refine\): final state of filament #{fil_num}:[\n\s]*Filament info for filament #{fil_num}:(\s\s.*\n)+?\s\stotal change of direction \(rad\) = (?<change_dir>[\d\.]*)\n/
      regex_init  = /^Extra info for filament #{fil_num}:[a-zA-Z\n\s,\(\)\d:\.\/=\-\+]*^\s*total change of direction \(rad\) = \s*(?<change_dir>[\d\.]*)/
      if (m = @str.match(regex_final))
        changeOfDirection << m["change_dir"].to_f
      elsif (m = @str.match(regex_init))
        changeOfDirection << m["change_dir"].to_f
      else
        $LOG.fatal('getChangeOfDirection') {"Could not find total change of direction for filament #{fil_num} in #{@filename}"}
        puts caller
        exit 1
      end
    end
    changeOfDirection
  end

  # In an array, return the total curvature for each filament
  def getTotalCurvature()
  	fil_nums = getFilamentNumbers
  	total_curvatures = []
  	fil_nums.each do |fil_num|
  		regex_final = /\*\*info\(mode_filament_refine\): final state of filament #{fil_num}:[\n\s]*Filament info for filament #{fil_num}:(\s\s.*\n)+?\s\stotal curvature \(1\/pixels\) = (?<tot_curv>[\d\.]*)\n/
  		regex_init  = /^Extra info for filament #{fil_num}:[a-zA-Z\n\s,\(\)\d:\.\/=\-]*^\s*total curvature \(1\/pixels\) = \s*(?<tot_curv>[\d\.]*)/
  		if (m = @str.match(regex_final))
  		  total_curvatures << m["tot_curv"]
  		elsif (m = @str.match(regex_init))
  		  total_curvatures << m["tot_curv"]
  		end
  	end
  	total_curvatures
  end

  # In an array of arrays, return the total axial curvature, helical overtwist and stretch for each filament
  def getTotalCurvatureOvertwistStretch(fil_nums=nil)
    fil_nums = getFilamentNumbers if (!fil_nums)
    curvatures = []
    overtwists = []
    stretches  = []

    fil_nums.each do |fil_num|
regex_int = /\*\*info\(mode_filament_refine\): final state of filament #{fil_num}:[\n\s]*\
Filament info for filament #{fil_num}:\n\
(\s\s.*\n)+?\
(Averg.*\nStdDv.*\n)\
(.*\n.*\n)\
Normal mode amplitudes for filament #{fil_num}:\n\
(\s\s.*\n)+?\
(Averg.*\nStdDv.*\n)\
Linear fit residuals\n\
(\s\s.*\n)+?\
(Averg.*\nStdDv.*\n)\
(\s\s.*\n)+?\
\s\stotal curvature, overtwist, stretch \(1\/pixels\) :\s*(?<curvature>[\d\.E-]*)\s*(?<overtwist>[\d\.E-]*)\s*(?<stretch>[\d\.E-]*)\n/
regex_int = /\*\*info\(mode_filament_refine\): final state of filament #{fil_num}:[\n\s]*\
Filament info for filament #{fil_num}:\n\
(\s\s.*\n)+?\
(Averg.*\nStdDv.*\n)\
(.*\n.*\n)\
Normal mode amplitudes for filament #{fil_num}\s*\n\
(\s\s.*\n)+?\
(Averg.*\nStdDv.*\n)\
Linear fit residuals\s*\n\
(\s\s.*\n)+?\
(Averg.*\nStdDv.*\n)\
(\s\s.*\n)+?\
\s\stotal curvature, overtwist, stretch \(1\/pixels\) :\s* (?<curvature>[\d\.E\-+]*)\s*(?<overtwist>[\d\.E\-+]*)\s*(?<stretch>[\d\.E\-+]*)\n/
regex_old = /\*\*info\(mode_filament_refine\): final state of filament #{fil_num}:[\n\s]*\
Filament info for filament #{fil_num}:\n\
(\s\s.*\n)+?\
\s\stotal curvature, overtwist, stretch \(1\/pixels\) :\ (?<curvature>[\d\.E\-+]*)\s*(?<overtwist>[\d\.E\-+]*)\s*(?<stretch>[\d\.E\-+]*)\n/
regex_new = /\*\*info\(mode_filament_refine\): final statistics for filament #{fil_num}:[\n\s]*\
Extra info for filament #{fil_num}:\n\
(\s\s.*\n)*\
\s\stotal curvature, overtwist, stretch \(1\/pixels\) :\s*(?<curvature>[\d\.E\-+]*)\s*(?<overtwist>[\d\.E\-+]*)\s*(?<stretch>[\d\.E\-+]*)\s*\n/
regex_new_init = /^Extra info for filament #{fil_num}:[a-zA-Z\n\s,\(\)\d:\.\/=\-]*^\s*total curvature, overtwist, stretch \(1\/pixels\) : \s*(?<curvature>[\d\.E\-+]*)\s*(?<overtwist>[\d\.E\-+]*)\s*(?<stretch>[\d\.E\-+]*)/
      if (m = @str.match(regex_old))
        #puts "Match against old regex"
        #p m
        curvatures << m["curvature"].to_f
        overtwists << m["overtwist"].to_f
        stretches  << m["stretch"].to_f
      elsif (m = @str.match(regex_new))
      	#puts "Match against new regex "+fil_num.to_s
      	curvatures << m["curvature"].to_f
        overtwists << m["overtwist"].to_f
        stretches  << m["stretch"].to_f
      elsif (m = @str.match(regex_new_init))
        #puts "Match against new regex "+fil_num.to_s
        curvatures << m["curvature"].to_f
        overtwists << m["overtwist"].to_f
        stretches  << m["stretch"].to_f
      elsif (m = @str.match(regex_int))
        #puts "Match against intermediate regex"
        #p m
        curvatures << m["curvature"].to_f
        overtwists << m["overtwist"].to_f
        stretches  << m["stretch"].to_f
      else
      	$LOG.info('getTotalCurvatureOvertwistStretch') {"Did not find  curvature, overtwist and stretch totals for filament #{fil_num} in "+self.filename}
      	#$LOG.fatal('getTotalCurvatureOvertwistStretch') {"No match for regex"}
      	#puts caller
      	#exit 1
      end #if
    end #do
    return curvatures, overtwists, stretches
  end #def

  # Find the film number
  def getFilmNum(fil_nums=nil)
    fil_nums = getFilamentNumbers if (!fil_nums)
    film_nums = []
    fil_nums.each {|fil_num|
regex = /Filament info for filament #{fil_num}:\n\
(\s\s.*\n)*\
\s\son micrograph number (?<film_num>\d+)\s*\n/
regex_140409 = /Filament info for filament #{fil_num}:\n\
(\s\s.*\n)*\
\s\son micrograph number \(location in file\) (?<film_num>\d+) \((?<loc_num>\d+)\)\s*\n/
     if (m = @str.match(regex))
       film_nums << m[:film_num]
     elsif (m = @str.match(regex_140409))
       film_nums << m[:film_num]
     else
       $LOG.fatal('getFilmNum') {"No match for regex in: #{@filename}"}
       puts caller(start = 0)
       exit 1
     end
    }
    return film_nums
  end

  # Update the filament object with information from the Frealix output
  def updateFilament(fil)
  	# Pixel size
  	pixel_size = getPixelSize
  	# The score waypoints
  	fil.score_wps = getFinalScoreWaypoints(fil.num)
	  # Curvature, overtwist, stretch
  	(c,o,s) = getTotalCurvatureOvertwistStretch([fil.num])
  	fil.curvature = c[0] / pixel_size if (c[0])
  	fil.overtwist = o[0] / pixel_size if (o[0])
  	fil.stretch   = s[0] / pixel_size if (s[0])
  	fil.changeOfDirection = getChangeOfDirection([fil.num]).at(0)
  	if (!c[0])
  		#$LOG.error(updateFilament) {"Did not find curvature"}
  	end
  	fil.wps_changes = getChangesWaypoints(fil.num)
  	fil.polarity_flipped = true if (getPolarity(fil.num) == 2)
	  if (d=getDetailledScoringInfo(fil.num)) then
		  fil.score_no_restraints = d[1].to_f
		else
		  fil.score_no_restraints = 0.0
		end
	  fil.film_num = getFilmNum([fil.num])
  	fil.num_subunits = getNumberOfSubunits(fil.num)
  	fil.polarity_dscore = getPolarityDScore(fil.num)
  	tmp = fil.updateWpToWpDistStats()
  end

  # Grab the difference in score between the two polarities, if they were tested. Return 0.0 otherwise
  def getPolarityDScore(filnum)
    pat=/\*\*info\(mode_filament_refine\): for filament #{filnum} in polarity (?:1|2), phi polarity (?:1|2) was best, with a score of ([.\d\-]*)/
    if (m = @str.scan(pat))
      if (m.count == 2) then
        (m[0][0].to_f-m[1][0].to_f).abs.round(5)
      else
        0.0
      end
    else
      0.0
    end
  end

  # Return the detailled scoring information of the filament
  # at the end of refinement
  def getDetailledScoringInfo(filnum)
    $LOG.debug("getDetailledScoringInfo") {"Processing filament #{filnum} in #{@filename}"}
regex_noscore=/not doing the final scoring/
regex =/^\*\*info\(MODE_FILAMENT_REFINE\): final state of filament #{filnum}:.*\
^\*\*info\(FMC_EVAL_SLAVE\): detailled score information along path of filament #{filnum}:\s*\n\
^(?:\(region from T = [\d\.]* to T = [\d\.]* used to compute score\)\n)?\
^\s+WP\#\s+X\s+Y\s+Z\s+Path_length\s+Phi\s+Theta\s+Psi\s+Subunit\s+Rise\s+Twist\s+Curvature\s+Refine\?\s+Score\n\
^([ \d\.\-E\nTF]*)\
^Averg\s+[ \d\.\-E\nTF]*\n\
^StdDv\s+[ \d\.\-E\nTF]*\n\
^\*\*info\(FMC_EVAL_SLAVE\): score = ([\d\.]*)/m
regex_131109=/^\*\*info\(MODE_FILAMENT_REFINE\): detailled scoring information for final state of filament #{filnum} follows:.*\
^\*\*info\(FMC_EVAL_SLAVE\): detailled score information along path of filament #{filnum}:\s*\n\
^(?:\(region from T = [\d\.]* to T = [\d\.]* used to compute score\)\n)?\
^\s+WP\#\s+X\s+Y\s+Z\s+Path_length\s+Phi\s+Theta\s+Psi\s+Subunit\s+Rise\s+Twist\s+Curvature[\w\(\)\?\s]*\n\
^([ \d\.\-E\nTF\*]*)\
^Averg\s+[ \d\.\-E\nTF]*\n\
^StdDv\s+[ \d\.\-E\nTF]*\n\
^\*\*info\(FMC_EVAL_SLAVE\): score = ([\d\.]*)/m
regex_131109=/^\*\*info\(MODE_FILAMENT_REFINE\): detailled scoring information for final state of filament #{filnum} follows:.*\
^\*\*info\(FMC_EVAL_SLAVE\): detailled score information along path of filament #{filnum}:\s*\n\
^(?:\(region from T = [\d\.]* to T = [\d\.]* used to compute score\)\n)?\
^\s+WP\#\s+X\s+Y\s+Z\s+Path_length\s+Phi\s+Theta\s+Psi\s+Subunit\s+Rise\s+Twist\s+Curvature[\w\(\)\?\s]*\n\
^([ \d\.\-E\nTF\*]*)\
^Averg\s+[ \d\.\-E\nTF]*\n\
^StdDv\s+[ \d\.\-E\nTF]*\n\
^\*\*info\(FMC_EVAL_SLAVE\): score = ([\d\.]*)/m
regex_131215=/^\*\*info\(MODE_FILAMENT_REFINE\): detailled scoring information for final state of filament #{filnum} follows:.*\
^\*\*info\(FMC_EVAL_SLAVE\): detailled score information along path of filament #{filnum}:\s*\n\
^(?:\(region from T = [\d\.]* to T = [\d\.]* used to compute score\)\n)?\
^\s+WP\#\s+X\s+Y\s+Z\s+Path_length\s+Phi\s+Theta\s+Psi\s+Subunit\s+Rise\s+Twist\s+Curvature[\w\(\)\?\s]*\n\
^([ \d\.\-E\nTF\*]*)\
^\s+Averg\s+[ \d\.\-E\nTF]*\n\
^\s+StdDv\s+[ \d\.\-E\nTF]*\n\
^\*\*info\(FMC_EVAL_SLAVE\): score = ([\d\.]*)/m
regex_140409=/^\*\*info\(MODE_FILAMENT_REFINE\): detailled scoring information for final state of filament #{filnum} follows:.*\
^\*\*info\(FMC_EVAL_SLAVE\): detailled score information along path of filament #{filnum}:\s*\n\
^(?:\(region from T = [\d\.]* to T = [\d\.]* used to compute score, in steps of [\d\.]*\)\n)?\
^\s+WP\#\s+X\s+Y\s+Z\s+Path_length\s+Phi\s+Theta\s+Psi\s+Subunit\s+Rise\s+Twist\s+Curvature[\w\(\)\?\s]*\n\
^([ \d\.\-E\nTF\*]*)\
^\s+Averg\s+[ \d\.\-E\nTF]*\n\
^\s+StdDv\s+[ \d\.\-E\nTF]*\n\
^\*\*info\(FMC_EVAL_SLAVE\): score = ([\d\.]*)/m
regex_140625=/^\*\*info\(MODE_FILAMENT_REFINE\): detailled scoring information for final state of filament #{filnum} follows:.*\
^\*\*info\(FMC_EVAL_SLAVE\): detailled score information along path of filament #{filnum}:\s*\n\
^(?:\(region from T = [\d\.]* to T = [\d\.]* used to compute score, in steps of [\d\.]*\)\n)?\
^\s+WP\#\s+X\s+Y\s+Z\s+Path_length\s+Phi\s+Theta\s+Psi\s+Subunit\s+Rise\s+Twist\s+Curvature[\w\(\)\?\s\.]*\n\
^([ \d\.\-E\nTF\*]*)\
^Averg\s+[ \d\.\-E\nTF]*\n\
^StdDv\s+[ \d\.\-E\nTF]*\n\
^\*\*info\(FMC_EVAL_SLAVE\): score = ([\d\.]*)/m
regex_140626=/^\*\*info\(mode_filament_refine\): detailled scoring information for final state of filament #{filnum} follows:.*\
^\*\*info\(fmc_eval_slave\): detailled score information along path of filament #{filnum}:\s*\n\
^(?:\(region from t = [\d\.]* to t = [\d\.]* used to compute score, in steps of [\d\.]*\)\n)?\
^\s+WP\#\s+X\s+Y\s+Z\s+Path_length\s+Phi\s+Theta\s+Psi\s+Subunit\s+Rise\s+Twist\s+Curvature[\w\(\)\?\s\.]*\n\
^([ \d\.\-E\nTF\*]*)\
^Averg\s+[ \d\.\-E\nTF]*\n\
^StdDv\s+[ \d\.\-E\nTF]*\n\
^\*\*info\(fmc_eval_slave\): score = ([\d\.]*)/m
	if (@str.match(regex_noscore))
		$LOG.info("getDetailledScoringInfo") {"No scoring info available for filament #{filnum}"}
		return nil
	else
		if (m = @str.match(regex))
		  $LOG.debug("getDetailledScoringInfo") {"Matched with old regex"}
		  return m[1],m[2]
    elsif (n = @str.match(regex_131215))
          $LOG.debug("getDetailledScoringInfo") {"Matched with new regex. Num lines = #{n[1].split("\n").count}"}
          return n[1],n[2]
    elsif (n = @str.match(regex_131109))
      $LOG.debug("getDetailledScoringInfo") {"Matched with new regex. Num lines = #{n[1].split("\n").count}"}
		  return n[1],n[2]
		elsif (n = @str.match(regex_140409))
      $LOG.debug("getDetailledScoringInfo") {"Matched with new regex. Num lines = #{n[1].split("\n").count}"}
      return n[1],n[2]
    elsif (n = @str.match(regex_140625))
      $LOG.debug("getDetailledScoringInfo") {"Matched with new regex. Num lines = #{n[1].split("\n").count}"}
      return n[1],n[2]
    elsif (n = @str.match(regex_140626))
      $LOG.debug("getDetailledScoringInfo") {"Matched with new regex. Num lines = #{n[1].split("\n").count}"}
      return n[1],n[2]
    else
      if (getModel3DAvailable == 1)
        $LOG.fatal("getDetailledScoringInfo") {"No match found for regexp in #{@filename}"}
        puts caller(start=0)
        exit 1
      else
        $LOG.debug("getDetailledScoringInfo") {"No match found for regexp in #{@filename}.That's OK because 3D model was not available for scoring"}
      end
		end
	end
  end

  # Return waypoints of the filament after refinement
  def getFinalWaypoints(filnum)
regex =/^\*\*info\(MODE_FILAMENT_REFINE\): final state of filament #{filnum}:.*?\
^\s+score = [\d\.\-]*\n\
(^\s+rigid parameters: [XYZ \:\-\.\d]*\n)?\
^\s+WP\#\s+X\s+Y\s+Z\s+Path_length\s+Phi\s+Theta\s+Psi\s+Subunit\s+Rise\s+Twist\s+Curvature\s+Refine\?\s+Score\n\
^(?<wpstr>[ \d\.\-E\nTF]*)\
^Averg/m
regex_131109 =/^\*\*info\(MODE_FILAMENT_REFINE\): final state of filament #{filnum}:.*?\
^\s+score = [\d\.\-]*\n\
(^\s+rigid parameters: [XYZ \:\-\.\d]*\n)?\
^\s+WP\#\s+X\s+Y\s+Z\s+Path_length\s+Phi\s+Theta\s+Psi\s+Subunit\s+Rise\s+Twist\s+Curvature\s+Refine\?\s+Score\n\
^(?<wpstr>[ \d\.\-E\nTF]*)\
^Averg/m
regex_140626 =/^\*\*info\(mode_filament_refine\): final state of filament #{filnum}:.*?\
^\s+score = [\d\.\-]*\n\
(^\s+rigid parameters: [XYZ \:\-\.\d]*\n)?\
^\s+WP\#\s+X\s+Y\s+Z\s+Path_length\s+Phi\s+Theta\s+Psi\s+Subunit\s+Rise\s+Twist\s+Curvature\s+Refine\?\s+Score\n\
^(?<wpstr>[ \d\.\-E\nTF]*)\
^Averg/m

    wps = []
    film_num = getFilmNum([filnum])
    if (m=@str.match(regex_140626))
      m[:wpstr].split("\n").each_with_index{ |str,i|
        w = str.split
        wp = FrealixWaypoint.new( :fil_num  => filnum.to_i,
                        :num    => i+1,
                        :x      => w[1].to_f,
                        :y      => w[2].to_f,
                        :z      => w[3].to_f,
                        :path_length=> w[4].to_f,
                        :phi    => w[5].to_f,
                        :theta    => w[6].to_f,
                        :psi    => w[7].to_f,
                        :subunit_num=> w[8].to_f,
                        :curvature  => w[9].to_f,
                        :score      => w[13].to_f,
                        :film_num   => film_num[0]
                        )
        wps << wp
      }
    else
      $LOG.error("getFinalWaypoints") {"No match found for regexp"}
    end
    return wps
  end

  # Return waypoints for the final call to the scoring function
  def getFinalScoreWaypoints(filnum)
  	require_relative "FrealixWaypoints.rb"
  	s = getDetailledScoringInfo(filnum)
  	wps = []
	if (s)
		s[0].split("\n").each_with_index { |str,i|
		w  = str.split
		wp = FrealixWaypoint.new( :fil_num 	=> filnum.to_i,
									:num		=> i+1,
									:x			=> w[1].to_f,
									:y			=> w[2].to_f,
									:z			=> w[3].to_f,
									:path_length=> w[4].to_f,
									:phi		=> w[5].to_f,
									:theta 		=> w[6].to_f,
									:psi		=> w[7].to_f,
									:subunit_num=> w[8].to_f,
									:curvature	=> w[11].to_f,
									:score      => w[13].to_f
									)
		wps << wp
		}
	end
	return wps
  end

  # Find the number of subunits that went into the 3D reconstruction
  def getNumberOfSubunits(fil_num)
    pat=/\*\*info\(filament_calc_threed\): will reconstruct filament #{fil_num} with (\d*) subunits/
    if (m = @str.match(pat)) then
      m[1].to_i
    else
      $LOG.warn("getNumberOfSubunits") {"No match found for regexp"}
      0
    end
  end

  # Find the average and std dev of crossover-to-crossover distances
  def getCrossoverDistancesAveStd(fil_num)
regex_new = /\*\*info\(mode_filament_refine\): final statistics for filament #{fil_num}:[\n\s]*\
Extra info for filament #{fil_num}:\n\
(\s\s.*\n)*\
\s\saverage \(stdev\) crossover-to-crossover distance \(pixels\) :\s*(?<ave>[\d\.E\-+]*)\s*\(\s*(?<stdev>[\d\.E\-+]*)\s*\)\s*\n/
    if (m = @str.match(regex_new))
      ave << m["ave"].to_f
      std << m["stdev"].to_f
    else
      ave = 0.0
      std = 0.0
      $LOG.info('getCrossoverDistancesAveStd') {"Did not find crossover-to-crossover distances for filament #{fil_num} in "+self.filename}
    end
    return ave,std
  end

  # Parse output to find out which polarity the filament refined to
  def getPolarity(fil_num)
    regex = /^\*\*info\(mode_filament_refine\): for filament #{fil_num}, polarity ([12]) was best, with a score of/
    if (m = @str.match(regex)) then
      m[1].to_i
    else
      1
    end
  end

  # How many threads did we use?
  def getNumThreads()
    regex = /Omp_num_threads env\. var\.: (\d*)/
    if (m = @str.match(regex))
      m[1].to_i
    else
      1
    end
  end

  # Return waypoints containing changes during refinements
  def getChangesWaypoints(fil_num)
    require_relative "FrealixWaypoints.rb"
    wps = []
    regex = /^Changes in filament #{fil_num}\s*\n.*\n([\s\d\-\.\n]*)/
    m = @str.scan(regex)
    # The last printout of changes is the one we're interested in
    n = m[-1]
    m[-1][0].split("\n").each_with_index {|str,i|
      w = str.split
      wp = FrealixWaypoint.new( :fil_num  => fil_num.to_i,
                      :num    => i+1,
                      :x      => w[1].to_f,
                      :y      => w[2].to_f,
                      :z      => w[3].to_f,
                      :path_length=> w[5].to_f,
                      :phi    => w[6].to_f,
                      :theta    => w[7].to_f,
                      :psi    => w[8].to_f,
                      :subunit_num=> w[9].to_f
                      )
      wps << wp
    }
    return wps
  end

  # Return the numbers of the filaments which were processed in an array
  def getFilamentNumbers()
	regex = /Filament info for filament (\d*):/
  	fil_nums = @str.scan(regex).flatten.uniq
  	if (fil_nums.size != getNumberOfFilaments)
  		puts "Unexpected number of filaments found"
		$LOG.error("getFilamentNumbers") {"Unexpected number of filaments found"}
  	end
  	fil_nums.collect!{|x| x.to_i}
  	return fil_nums
  end

  # The number of filaments that were processed
  def getNumberOfFilaments()
	if (@mode == 'FILAMENT_REFINE' || @mode == 'filament_refine')
		regex = /\*\*info\(MODE_FILAMENT_REFINE\): (\d*) filaments were initialised/
    regex = /\*\*info\(mode_filament_refine\): (\d*) filaments were initialised/
	else
		$LOG.error("getNumberOfFilaments") {"Cannot process this mode's (#{@mode}) output"}
	end
	m = @str.match(regex)
	m[1].to_i
  end

  # Method to find out the frealix mode that was run
  def findMode()
    # Look for the string that frealix prints out at the beginning to announce the mode
    regex = /Frealix will run its [A-Za-z_]*/
    match = (@str =~ regex)?true:false
    if (match)
      @lines.each do |line|
        @mode = line.match(regex).string.split[-2] if (line.match(regex))
      end
    else
        @mode = nil
#      $LOG.fatal("findMode") {"Could not find frealix mode in log file #{@filename}"}
#      puts caller(start = 0)
#      exit 1
    end
  end

  # Return the pixel size in Angstroms
  def getPixelSize
  	regex = /Pixel\/voxel size *= *(\d*\.\d*) A/
  	regex_rtp = /^\s*PIXEL_SIZE\s*(\d*\.\d*)\s*[TF]/
  	if (m = @str.match(regex_rtp))
  		return m[1].to_f
  	elsif (m = @str.match(regex))
  	  return m[1].to_f
    else
  		return 0.0
  	end
  end

  # Return 0 if the model 3D was not availble, 1 if it was
  def getModel3DAvailable
    regex = /\s*MODEL_3D_AVAILABLE\s*T/
    if (m = @str.match(regex))
      return 1
    else
      return 0
    end
  end

  # Return FSC values as an array
  def getFSCValues(options={})
  	defaults = {	:prefix_string => "",
  				}
  	options = defaults.merge(options)
  	regex = /#{options[:prefix_string]}.*\s*Shell.*\n((?:.*[\d]+.*[\d\-\.]+.*\n)*)/
    regex_new = /#{options[:prefix_string]}.*\s*\!Shell.*\n((?:\s*[\d]+\s*[\d\-\.]+.*\n)*)/
    regex_new_150303 = /#{options[:prefix_string]}.*\s*Spatial frequency *FSC\s*\n((?:\s*[\d]+\s*[\d\-\.]+.*\n)*)/
    m = @str.match(regex_new_150303)
    if (m)
      tmp = m[1]
      tmp = tmp.split("\n")
      fsc = []
      tmp.each { |l|
        fsc << l.split[1].to_f
      }
      return fsc
    else
      m = @str.match(regex_new)
    	if (m)
        tmp = m[1]
    		tmp = tmp.split("\n")
    		fsc = []
    		tmp.each { |l|
    			fsc << l.split[2].to_f
    		}
    		return fsc
    	else
    	  m = @str.match(regex)
    	  if (m)
          tmp = m[1]
          tmp = tmp.split("\n")
          fsc = []
          tmp.each { |l|
            fsc << l.split[1].to_f
          }
          return fsc
    	  else
    	    $LOG.error("getFSCValues") {"Did not find match"}
    	    return nil
    	  end
      end
    end
  end

  # Return the duration of the Frealix run, in seconds
  def getDuration()
	require_relative "rubylix.rb"
	pattern = /^\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d: frealix says (goodbye|sorry) \(total execution time: (.*)\)/
	m = @str.match(pattern)
	if (m)
		duration = m[2]
		return duration_for(duration)
	else
		return nil
	end
  end

  # Return the step size used by the scoring function (in pixels)
  def getScoringStepSizes()
    regex = /^\s*Step size \(pixels\)\s*([\d\.]*)/
    lengths = []
    @str.scan(regex){|m|
      lengths << m[0].to_f
    }
    return lengths
  end

  # Return array of strings, each string being a unique warning found in the log file
  def getWarnings()
    regex = /warning|WARNING/
    warnings = []
    @str.split("\n").grep(regex){|m|
      #p m.gsub(/\d/,'#')
      warnings << m.gsub(/\d/,'#')
    }
    return warnings
  end
end
