# Dealing with frealix runtime
class FrealixRunner
  NAME = "Frealix runner"

  # Stuff we need
  require_relative "Sge.rb"
  require_relative "LinuxHost.rb"
  require_relative "FrealixOutput.rb"
  require_relative "FrealixParameters.rb"
  require_relative "FrealixRefinementDirectory.rb"
  require_relative "rubylix.rb"

  # Initialiser method with default parameter
  def initialize( options = {})
    # Defaults values for the options
    defaults = {  :num_jobs => 1,
                  :params => [FrealixParameters.new()],
                  :logs => [FrealixOutput.new()],
                  :num_threads => [1],
                  :dir => FrealixRefinementDirectory.new(),
                  :hosts => [],
                  :threads_per_job => 1,
                  :mem_per_thread_gb=>1,
                  :prefix=>"flx",
                  :gpu_nodes_only=>false,
                  :resume=>false,
                  :extra_qsub_options=>""}
    options = defaults.merge(options){  |key,defval,optval|  (optval.nil?) ? defval:optval}

    # Checks on arguments
    # params & logs need to be arrays
    if (options[:params].class != Array)
      puts "Params argument should be an array, not #{options[:params].class}"
      puts caller(start=0)
      exit 1
    end
    if (options[:logs].class != Array)
      puts "Logs argument should be an array, not #{options[:logs].class}"
      puts caller(start=0)
      exit 1
    end

    # Copy arguments into instance variables
    @num_jobs = options[:num_jobs]
    @params = options[:params]
    @log = options[:logs]
    @num_threads = options[:num_threads]
    @dir = options[:dir]
    @hosts = options[:hosts]
    @threads_per_job = options[:threads_per_job]
    @mem_free_per_thread_gb = options[:mem_per_thread_gb]
    @prefix = options[:prefix]
    @gpu_nodes_only = options[:gpu_nodes_only]
    @resume = options[:resume]
    @extra_qsub_otions = options[:extra_qsub_options]


    # Check whether qsub is available
    @qsub_available = which("qsub")
    if (@qsub_available)
      # Prepare interactions with SGE
      @sge = Sge.new(:extra_qsub_options=>@extra_qsub_options)
      if (@sge.total_num_slots == 0)
        @qsub_available= false
      end
    else
      if (@hosts.include?("cluster"))
        puts "Oops, you specified you wanted to use the cluster, but you don't appear to be the headnode: can't find qsub"
        puts caller(start=0)
        exit 1
      end
    end

    # Check whether the available hosts are reachable
    $LOG.debug("FrealixRunner init remove unreachable hosts from this list: #{@hosts}")
    remove_unreachable_hosts()

    # Deal with the case where no hosts are listed
  	if (@hosts == [])
  		if (@qsub_available)
  			@hosts = ["cluster"]
  		else
  			@hosts = ["localhost"]
  		end
  	end

    # How many cores do we want to leave alone on hosts?
    @num_cores_preserve = 2

    # We will wait to submit jobs to the queue until there are max
    # this number of qw jobs
    # If set to 999999, do not impose a limit
    @cluster_qw_threshold = 999999


    # Set up a few arrays
    @hosts_num_cores = Array.new(@hosts.count,0)
    @hosts_mem_free = Array.new(@hosts.count,0)
    @pid = Array.new(@num_jobs,0) # Holds either the pid or the sge job_id
    @status = Array.new(@num_jobs) # Holts a status object
    @hostname = Array.new(@num_jobs,"") # Where that job should be run
    # Values for job_status array:
    # 0: nothing's happened;
    # 1: the spawned process was started succesfully (or: the job was submitted to the queue succesfully);
    # 2: the spawned process exited (or: the job is not in the queue anymore)
    # 3: Frealix reported error-free termination
    # -1: something went wrong when spawning the process, or the job was not submitted to the queue succesfully
    # -2: the spawned process stopped (or: the queued job is in an error state)
    # -3: Frealix reported an error or crashed out
    @job_status = Array.new(@num_jobs,0)

    # Work out the number of cores available on each host
    @hosts.each_with_index do |host,i|
      if (host != "cluster")
        h = LinuxHost.new(:hostname=>host)
  	    @hosts_num_cores[i] = h.numCoresAvailable()
  	    @hosts_mem_free[i] = h.mem_free_gb
  	    $LOG.debug("FrealixRunner init") {"#{host} has #{@hosts_num_cores[i]} cores"}
      end
    end

    # Check whether frealix is available
    if (File.executable?("./frealix")) then
      @frealix_exec = "./frealix"
    elsif (which("frealix")) then
      @frealix_exec = "frealix"
    else
      @frealix_exec = nil
      $LOG.fatal("FrealixRunner init") {"Frealix executable not found in path or current directory"}
      exit 1
    end


    # Check the number of parameters & log files is as expected
    if (@params.count != @num_jobs || @log.count != @num_jobs)
      $LOG.error("FrealixRunner init") {"Expected #{@num_jobs} parameters and log files, but got #{@params.count} and #{@log.count}"}
      puts caller(start=0)
      exit 1
    end

  end # initialize

  # Return the number of jobs currently running on the given hostname
  def num_jobs_running_on_host(hostname)
  	num_jobs = 0
  	# Loop over all jobs, and count the times that a job
  	# with status == 1 has the hostname assigned to it
  	0.upto(@num_jobs-1) do |i|
  	  num_jobs += 1 if (@job_status[i] ==1 && @hostname[i] == hostname)
  	end
  	return num_jobs
  end

  # Check hosts given exist / are reachable
  def all_hosts_reachable?()
    all_good = true
    if (!@hosts.empty?)
      #0.upto(@hosts.count-1) do |i|
      @hosts.each { |host|
          sleep 0.5
		if (! host_reachable?(host))
		  all_good = false
		  $LOG.fatal("#{host} is unreachable")
		  exit 1
		else
	      #$LOG.debug("#{host} is reachable")
		end
	  }
      @end # do loop
    end
    return all_good
  end

  def remove_unreachable_hosts()
    new_hosts = []
    if (!@hosts.empty?)
      #0.upto(@hosts.count-1) do |i|
      @hosts.each { |host|
        sleep 0.5
        new_hosts << "localhost" if (host == "localhost")
        new_hosts << "cluster" if (host == "cluster")
        if (host_reachable?(host))
			new_hosts << host
		else
			$LOG.info("remove_unreachable_hosts") {"#{host} was unreachable and has been removed from the list"}
        end
	  }
    end
    @hosts = []
    @hosts = new_hosts
  end

  # Check whether a host is reachable
  def host_reachable?(hostname)
    ok = true
	if (hostname != "localhost" && hostname != "cluster")
	  command = "ping -q -c 1 #{hostname}"
	  p = Process.spawn(command,[:err,:out]=>"/dev/null")
	  pid, status = Process.wait2(pid=p)
	  ok = (status.exitstatus == 0)?true:false
	  #puts "Host #{hostname} pingable? #{ok}"
	  if (! ok)
		# That host is not available
		$LOG.warn("host_reachable?") {"#{hostname} is not reachable by ping. Command was: #{command}"}
		#puts exit 1
	  end
	end
	return ok
  end

  # Check whether a host is available for new job
  def host_available?(hostname)
    case hostname
	  when "cluster"
  		# Check that there are at most @cluster_qw_threshold jobs
  		# waiting in the queue
  		$LOG.debug("host_available?") {"Will check number of jobs waiting in queue"}
  		cluster_available = true
      if (@cluster_qw_threshold != 999999) then
    		cluster_available = false if (@sge.numJobsWaiting > @cluster_qw_threshold)
    		if (@sge.numUsers > 1) then
    		  cluster_available = false if (@sge.numSlotsRunning.to_f / @sge.total_num_slots.to_f > 0.85)
    		end
      end
  		return cluster_available
	  else
  		# Check that the host has @threads_per_job+1 cores free
  		h = LinuxHost.new(:hostname=>hostname)
  		$LOG.debug("host_available?") {"#{hostname} has #{h.numCoresAvailable} available cores"}
  		return (h.numCoresAvailable >= @threads_per_job && h.mem_free_gb >= @threads_per_job * @mem_free_per_thread_gb)?true:false
    end
  end

  # Assign a host to a job
  def assign_host_to_job(i)
    #update_all_jobs_status()
    $LOG.debug('assign_host_to_job') {"Looking for host for job #{i}. Hosts #{@hosts} Qsub available? #{@qstat_available}"}
    if (@hosts.empty?)
      if (@qsub_available)
        # We're on the head node
    		if (host_available?("cluster"))
    		  @hostname[i] = "cluster"
    		else
    		  @hostname[i] = ""
    		end
      else
        # We're not on the head node, but we don't know any other hosts,
        # so we'll run locally
        $LOG.debug("assign_host_to_job") {"About to check if localhost is available: #{host_available?("localhost")}"}
        if (host_available?("localhost"))
          @hostname[i] = "localhost"
        else
          @hostname[i] = ""
        end
      end # qsub_available
    else # hosts was not empty
      $LOG.debug('assign_host_to_job') {"Hosts array was not empty"}
      @hostname[i] = ""
      # We were given a list of available hosts
      # Let's look for a host which hasn't had all its available cores
      # assigned yet
      @hosts.each_with_index do |h,hi|
        $LOG.debug('assign_host_to_job') {"Checking #{h}"}
        c = (h =~ /cluster/)?true:false
        if (c)
    			$LOG.debug('assign_host_to_job') {"cluster host_available? #{host_available?(h)}"}
    			if (host_available?(h))
    				@hostname[i] = h
    				break
    			end
        else #cluster not in list of hosts
          $LOG.debug("assign_host_to_job") {"Checking whether #{h} is available"}
    			# Threads currently running on that host
    			tmp = (num_jobs_running_on_host(h) * @threads_per_job)
    			$LOG.debug('assign_host_to_job') {"#{h} has #{tmp} threads running on it (total cores: #{@hosts_num_cores[hi]})"}
    			# Threads available on that host
    			tmp = @hosts_num_cores[hi] - tmp
    			$LOG.debug('assign_host_to_job') {"#{h} has #{tmp} threads available (total cores: #{@hosts_num_cores[hi]})"}
    			# same kind of thing with memory
    			tmp_mem = (num_jobs_running_on_host(h) * @threads_per_job * @mem_free_per_thread_gb)
    			tmp_mem = @hosts_mem_free[hi]*0.75 - tmp_mem
    			$LOG.debug('assign_host_to_job') {"#{h} has #{tmp_mem} GB mem available (total mem: #{@hosts_mem_free[hi]})"}
    			# If there are enough threads, we can assign this host
    			# always leave a couple of cores alone
    			if (tmp >= @threads_per_job + @num_cores_preserve && tmp_mem > @threads_per_job * @mem_free_per_thread_gb)
            $LOG.debug("assign_host_to_job") {"Assinging #{h}"}
    			  @hostname[i] = h
    			  break
    			end
        end # test if cluster in list of hosts
      end # loop over hosts
    end # test if hosts was empty
  end #assign_host_to_job

  # Start a job
  def start_run_serial(i)
	  string = "Running job #{i} (#{i+1} of #{@num_jobs}) "
	  case @hostname[i]
	  when "localhost"
		$LOG.info("start_run_serial") {"#{string} on localhost"}
		start_run_serial_local(i=i)
	  when "cluster"
		$LOG.info("start_run_serial") {"#{string} on cluster"}
		start_run_serial_qsub(i=i)
	  else
		$LOG.info("start_run_serial") {"#{string} on #{@hostname[i]}"}
		start_run_serial_ssh(i=i)
	  end
  end #start_run_serial

  # Run the frealix job(s) and wait for it (them) to finish
  def run()
    # We need to check the status of every job, in case we are resuming and some of the job were already completed
    # Do this in the loop so we don't repeat ourselves too much
    #update_all_jobs_status
    # Spawn/start/submit all jobs
    0.upto(@num_jobs-1) do |i|
      $LOG.debug("run") {"Will now try to start/spawn/submit job #{i}"}
      update_job_status(i)
      #sleep 1
      if (@job_status[i] == 3)
        # That job was already completed
        $LOG.info("run") {"Job #{i} was already completed, skipping"}
      else
        assign_host_to_job(i)
    	  while(@hostname[i] == "")
    	    sleep 1 # 1 second
          0.upto(i-1) do |j|
            update_job_status(j)
          end
    	    assign_host_to_job(i)
    	  end
    	  start_run_serial(i)
    	  $LOG.debug("run") {"Job #{i} has started"}
      end
    end

    # Wait for them to finish
    until(all_jobs_finished?())
      sleep 2.0 #sleep for 2 seconds
    end
    return all_jobs_successful?
  end #run

  # Have all jobs been started?
  def all_jobs_started?
  	r = true
  	@job_status.each { |s|
  	  r = false if (s==0)
  	}
  end

  # Is a job finished? (be it with error or succesfully)
  def job_finished?(i=0)
    update_job_status(i)
    (@job_status[i] == 3 || @job_status[i] < 0)?true:false
  end

  # Are all jobs finished?
  def all_jobs_finished?
    update_all_jobs_status()
    (@job_status.include?(0) || @job_status.include?(1) || @job_status.include?(2))?false:true
  end

  # Were all jobs successful?
  def all_jobs_successful?
  	update_all_jobs_status()
  	r = true
  	@job_status.each_with_index do |val,i|
  	  r = false if (val != 3)
  	end
  	return r
  end


  # Update status of all jobs
  def update_all_jobs_status
    0.upto(@num_jobs-1) do |i|
      update_job_status(i)
    end
  end

  # Update status of a job
  def update_job_status(i=0)
    $LOG.debug("update_job_status") {"Checking on job #{i}, hostname #{@hostname[i]}, current status #{@job_status[i]}"}
    case (@job_status[i])
    when 0
	    # This job hasn't been submitted yet
      if (@resume)
        # If we are resuming, it could be that the job was completed previously
        if (frealix_completed?(i))
          if (frealix_successful?(i))
            @job_status[i] = 3
          end
        end
      end
    when 1
      # We know the job was submitted succesfully.
      # If qsub, check whether it is still listed by qstat, and with which status
      # Else check the status of the pid, did it exit or was it stopped?
      case @hostname[i]
      when "cluster"
        # Check what qstat can tell us about the job.
        # If it is listed as recently finished, set to 2
        # If it is still listed with an E, go to -2
        # If it still listed with r or qw, do nothing
        q_job_status = @sge.getJobStatus(@pid[i])
        if (q_job_status == "finished")
          @job_status[i] = 2
        elsif (q_job_status == "r" || q_job_status == "qw")
          # Do nothing
        elsif (q_job_status =~ /E/)
          # Error state
          @job_status[i] = -2
          $LOG.error("update_job_status") {"Job #{i}: error state (sge job id #{@pid[i]}"}
        elsif (frealix_completed?(i))
          # If for some reason we lost track of it in the queueing system, we may still
          # find out it finished by checking the log file
          @job_status[i] = 2
        end
      else # not cluster
        # Check whether the process has exited or stopped, and if so whether it was successful
        # If it exited, go to 2
        # If it stopped, go to -2
        if ((Process.getpgid(@pid[i]) rescue nil))
          status = nil
          pid, status = Process.wait2(pid=@pid[i],Process::WNOHANG)
          if (status)
      			if (status.exited? || status.stopped?)
      			  if (status.success?)
        				@job_status[i] = 2
        				#puts "Job #{i}: success (pid #{@pid[i]})"
      			  else
        				$LOG.error("update_job_status") {"Job #{i}: error (pid #{@pid[i]}). See #{@log[i].filename}"}
        				@job_status[i] = -2
      			  end
      			else
      				$LOG.debug("update_job_status") {"Job #{i}: not finished yet"}
      			end
          end
        else
          $LOG.error("update_job_status") {"Job #{i}: error (pid #{@pid[i]}). See #{@log[i].filename}"}
          @job_status[i] = -2
        end
      end
    when 2
      # We need to check whether Frealix completed - With or without error?
      # If frealix didn't complete, nothing to do
      # If frealix finished in error, set to -3
      # If frealix finished normally, set to 3
      if (frealix_error?(i))
        @job_status[i] = -3
        $LOG.error("update_job_status") {"Job #{i}: error when running frealix. See #{@log[i].filename}"}
        # Frealix hasn't finished yet, nothing to do
      elsif (frealix_completed?(i))
        if (frealix_successful?(i))
          @job_status[i] = 3
          $LOG.info("update_job_status") {"Job #{i}: frealix finished normally"}
        else
          @job_status[i] = -3
          $LOG.error("update_job_status") {"Job #{i}: frealix reported an error. See #{@log[i].filename}"}
        end
      end
    when -1, -2, -3, 3
      # Nothing to do - the job finished in error state or finished successfully
    else
      # This shouldn't happen
      $LOG.fatal("update_job_status") {"Strange job status value: #{@job_status[i]}"}
      puts caller(start = 0)
      exit 1
    end # case job_status
  end # update_job_status


  # Check whether frealix completed
  def frealix_completed?(i=0)
    @log[i].completed?
  end
  def frealix_successful?(i=0)
    @log[i].successful?
  end
  def frealix_error?(i=0)
    @log[i].error?
  end

  # Start a frealix run on a remote machine, via SGE's qsub
  def start_run_serial_qsub(i=0)
    # Dump the parameters to disk
    @params[i].writeToDisk()
    # Prepare the command
    # Submit the job to the queue
    $LOG.debug("start_run_serial_qsub") {"About to @sge.submitJob"}
    @pid[i] = @sge.submitJob(executable=@frealix_exec,
                            arguments="#{@dir.directory}/#{@params[i].filename}",
                            out="#{@dir.directory}/#{@log[i].filename}",
                            err="#{@dir.directory}/#{@log[i].filename}",
                            job_name="#{@prefix}#{@dir.currentRound()}_#{i}",
                            num_threads=@threads_per_job,
                            mem_free_gb=@mem_free_per_thread_gb
                            )
    if (@pid[i] == 0)
      puts "Something went wrong when submitting job #{i} to qsub"
      @job_status[i] = -1
    else
      @job_status[i] = 1
    end
  end


  # Start a frealix run on a remote machine, via SSH
  def start_run_serial_ssh(i=0)
    # Dump the parameters to disk
    @params[i].writeToDisk()
    # The command we'll be running on the remote host. Note
    # that this is written assuming we'll be running under tcsh
    # in the remote host. Tcsh has its own version of nice.
    # Tcsh: nice +15 command
    # GNU: nice -n 15 command
    command = "/usr/bin/ssh -nqx #{@hostname[i]} \"cd #{@dir.directory}; export OMP_NUM_THREADS=#{@threads_per_job}; nice -n 15 #{@frealix_exec} #{@params[i].filename} >& #{@log[i].filename}\" "
    #puts "SSH command will be: "
    #puts command
    #$LOG.debug("start_run_serial_ssh") {"About to dummy spawn"}
    #pid = Process.spawn("sleep 4")
    #$LOG.debug("start_run_serial_ssh") {"Dummy spawn done, let's wait for it to finish"}
    #stat = nil
    #until (stat && (stat.stopped? || stat.exited? || stat.success?))
	#	pidd,stat = Process.wait2(pid=pid,Process::WNOHANG)
	#	$LOG.debug("start_run_serial_ssh") {"Still waiting for dummy spawn"}
		sleep 0.25
    #end
    #$LOG.debug("start_run_serial_ssh") {"Dummy spawn finished"}

    $LOG.debug("start_run_serial_ssh") {"About to spawn ssh process: #{command}"}
    #$LOG.debug("start_run_serial_ssh") {"pid array (#{@pid.count} elements) = #{@pid}"}
    #$LOG.debug("start_run_serial_ssh") {"job_status array (#{@job_status.count} elements) = #{@job_status}"}
    @pid[i] = Process.spawn(command,[:err,:out]=>[".ssh.#{i}.log","w"] )
    $LOG.debug("start_run_serial_ssh") {"SSH spawned, pid = #{@pid[i]}"}

    if (@pid[i] == 0)
      $LOG.error("start_run_serial_ssh") {"Something went wrong when submitting job #{i} via ssh"}
      @job_status[i] = -1
    else
      @job_status[i] = 1
    end
  end


  # Start a frealix run on a single machine
  def start_run_serial_local(i=0)
    # Dump the parameters to disk
    @params[i].writeToDisk()
    # Run frealix - we spawn a process
    @pid[i] = Process.spawn({"OMP_NUM_THREADS"=>@threads_per_job.to_s},"nice -n 15 #{@frealix_exec} #{@params[i].filename}",[:err,:out]=>[@log[i].filename,"w"] )
    if (@pid[i] == 0)
      puts "Something went wrong when submitting job #{i} to qsub"
      @job_status[i] = -1
    else
      @job_status[i] = 1
    end
  end

end
