#  FrealixWaypoints.rb
#
#  Copyright 2014 Howard Hughes Medical Institute.
#  All rights reserved.
#  Use is subject to Janelia Farm Research Campus Software Copyright 1.1
#  license terms ( http://license.janelia.org/license/jfrc_copyright_1_1.html )
#
#
#
class FrealixWaypoint
	NAME = "Frealix waypoint"

	require_relative "rubylix.rb"

	attr_reader :fil_num, :num
	attr_accessor :fil_num, :num, :refine, :film_num
	attr_accessor :psi, :theta, :phi
	attr_accessor :x,:y,:z
	attr_accessor :subunit_num
	attr_accessor :path_length, :score
	attr_accessor :twist, :rise
	attr_accessor :subframe_shifts, :number_of_frames

	def initialize(options = {})
		# Default values for the options
		defaults = {	:fil_num		=>	0,
						:num			=>	0,
						:film_num		=>	0,
						:x				=>	0.0,
						:y				=>	0.0,
						:z				=>	0.0,
						:phi			=>	0.0,
						:theta			=>	0.0,
						:psi			=>	0.0,
						:path_length	=>	0.0,
						:subunit_num	=>	0.0,
						:score			=> 	0.0,
						:logp       => 0.0,
						:sigma      => 0.0,
						:occ        => 0.0,
						:string			=>	"",
						:refine			=>	"T",
						:curvature		=>	0.0,
						:torsion		=>	0.0,
						:stretch		=>	0.0,
						:twist      =>  0.0,
						:rise       =>  0.0,
						:subframe_shifts => "0.0 0.0 0.0",
						:number_of_frames => 1,
						:par_version => 5
					}

		# Apply defaults (?)
		@refine = defaults[:refine]
		@subframe_shifts = defaults[:subframe_shifts]

		# If string was supplied, we parse it assuming it was
		# from a waypoints.ali file
		if (options[:string] != defaults[:string] && options[:string])
			if (options[:par_version].to_i == 5)
        populateFromString(options[:string])
			elsif (options[:par_version].to_i == 6)
				populateFromStringV6(options[:string])
			elsif (options[:par_version].to_i == 4)
			  populateFromStringOld(options[:string])
			end
		end

		# The other optional arguments override the string
		@fil_num 					= options[:fil_num]		if (options[:fil_num] 		)
		@film_num 				= options[:film_num] 	if (options[:film_num] 		)
		@num 							= options[:num] 		if (options[:num] 			)
		@x 								= options[:x] 			if (options[:x] 			)
		@y 								= options[:y] 			if (options[:y] 			)
		@z 								= options[:z] 			if (options[:z] 			)
		@path_length 			= options[:path_length]	if (options[:path_length]	)
		@phi 							= options[:phi] 		if (options[:phi] 			)
		@theta 						= options[:theta] 		if (options[:theta] 		)
		@psi							= options[:psi] 		if (options[:psi] 			)
		@subunit_num			= options[:subunit_num]	if (options[:subunit_num] 	)
    @rise      				= options[:rise]     if (options[:rise]       )
    @twist      			= options[:twist]     if (options[:twist]       )
		@score 						= options[:score] 		if (options[:score] 		)
		@logp       			= options[:logp]      if (options[:logp])
		@sigma      			= options[:sigma]     if (options[:sigma])
		@occ        			= options[:occ]       if (options[:occ])

		@refine 					= options[:refine] 		if (options[:refine] 		)
		@curvature 				= options[:curvature] 	if (options[:curvature] 	)
		@torsion 					= options[:torsion] 	if (options[:torsion] 		)
		@stretch 					= options[:stretch] 	if (options[:stretch] 		)
		@number_of_frames	= options[:number_of_frames]   if (options[:number_of_frames]     )
    @subframe_shifts  = options[:subframe_shifts]   if (options[:subframe_shifts]     )
	end

	def score
		@score
	end

	# Everything is re-initialised except XY, number, fil number etc
	def keepXYOnly!
    #@fil_num    =
    #@film_num        =
    #@num        =
    #@x        =
    #@y        =
    @z        = 0.0
    @path_length  = 0.0
    @phi      = 0.0
    @theta      = 0.0
    @psi      = 0.0
    @subunit_num  = 0.0
    @score      = 0.0
		@logp       = 0.0
		@sigma      = 0.0
		@occ        = 0.0
    #@refine     =
    @curvature    = 0.0
    @torsion    = 0.0
    @stretch    = 0.0
	end

	# Print out waypoint information to string, formatted like a line from an ali file
	def toString
	  zeroIfNotSet
		case @default_ali_format
		when 5
			sprintf("%6i %6i %6i %11.3f %11.3f %11.3f %11.3f %12.3f %12.3f %12.3f %11.3f %12.3f %12.3f %8.3f %7s %43s %s\n",
		          @fil_num,@num,@film_num,@x,@y,@z,@path_length,@phi,@theta,@psi,@subunit_num,@twist,@rise,@score,@refine," ",@subframe_shifts)
	  when 6
			sprintf("%6i %6i %6i %11.3f %11.3f %11.3f %11.3f %12.3f %12.3f %12.3f %11.3f %12.3f %12.3f %8.3f %6i %6.4f %6.2f %7s %43s %s\n",
		          @fil_num,@num,@film_num,@x,@y,@z,@path_length,@phi,@theta,@psi,@subunit_num,@twist,@rise,@score,@logp,@sigma,@occ,@refine," ",@subframe_shifts)
		when 4
			sprintf("%6i %6i %6i %11.3f %11.3f %11.3f %11.3f %12.3f %12.3f %12.3f %11.3f %8.3f %7s \n",
	            @fil_num,@num,@film_num,@x,@y,@z,@path_length,@phi,@theta,@psi,@subunit_num,@score,@refine)
		else
			$LOG.fatal("FrealixWaypoint::toString") {"Unknown ali format: #{@default_ali_format}"}
		end
	end

  def toStringOld
    zeroIfNotSet
    sprintf("%6i %6i %6i %11.3f %11.3f %11.3f %11.3f %12.3f %12.3f %12.3f %11.3f %8.3f %7s \n",
            @fil_num,@num,@film_num,@x,@y,@z,@path_length,@phi,@theta,@psi,@subunit_num,@score,@refine)
  end

	def zeroIfNotSet
    @fil_num        = 0   unless (@fil_num)
    @num            = 0   unless (@num)
    @film_num       = 0   unless (@film_num)
    @x              = 0.0 unless (@x)
    @y              = 0.0 unless (@y)
    @z              = 0.0 unless (@z)
    @phi            = 0.0 unless (@phi)
    @theta          = 0.0 unless (@theta)
    @psi            = 0.0 unless (@psi)
    @path_length    = 0.0 unless (@path_length)
    @subunit_num    = 0.0 unless (@subunit_num)
    @rise           = 0.0 unless (@rise)
    @twist          = 0.0 unless (@twist)
    @score          = 0.0 unless (@score)
		@logp           = 0.0 unless (@logp)
		@sigma          = 0.0 unless (@sigma)
		@occ            = 0.0 unless (@occ)
    @string         = ""  unless (@string)
    @refine         = "T" unless (@refine)
    @curvature      = 0.0 unless (@curvature)
    @torsion        = 0.0 unless (@torsion)
    @stretch        = 0.0 unless (@stretch)
    @subframe_shifts= "0.0 0.0 0.0" unless (@subframe_shifts)
	end


	# Populate a waypoint with values taken from a string from a waypoints.ali file
	def populateFromStringV6(string)
		(@fil_num,@num,@film_num,@x,@y,@z,@path_length,@phi,@theta,@psi,@subunit_num,@twist,@rise,@score,@logp,@sigma,@occ,@refine,*subframe_shifts)  = string.split
    # Work out number of frames
    @number_of_frames = subframe_shifts.length / 3
		#
    @subframe_shifts = subframe_shifts.join(" ")
		#
		@fil_num = @fil_num.to_i
		@num = @num.to_i
		@x = @x.to_f
		@y = @y.to_f
		@z = @z.to_f
		@path_length = @path_length.to_f
		@phi = @phi.to_f
		@theta = @theta.to_f
		@psi = @psi.to_f
		@subunit_num = @subunit_num.to_f
		@twist = @twist.to_f
		@rise = @rise.to_f
		@score = @score.to_f
		@logp = @logp.to_f
		@sigma = @sigma.to_f
		@occ = @occ.to_f
		$LOG.debug() {"New waypoint has score #{@score} (5)"}
	end

	# Populate a waypoint with values taken from a string from a waypoints.ali file
	def populateFromString(string)
		(@fil_num,@num,@film_num,@x,@y,@z,@path_length,@phi,@theta,@psi,@subunit_num,@twist,@rise,@score,@refine,*subframe_shifts)  = string.split
    # Work out number of frames
    @number_of_frames = subframe_shifts.length / 3
		#
    @subframe_shifts = subframe_shifts.join(" ")
		#
		@fil_num = @fil_num.to_i
		@num = @num.to_i
		@x = @x.to_f
		@y = @y.to_f
		@z = @z.to_f
		@path_length = @path_length.to_f
		@phi = @phi.to_f
		@theta = @theta.to_f
		@psi = @psi.to_f
		@subunit_num = @subunit_num.to_f
		@twist = @twist.to_f
		@rise = @rise.to_f
		@score = @score.to_f
		$LOG.debug() {"New waypoint has score #{@score} (5)"}
	end

  def populateFromStringOld(string)
      (@fil_num,@num,@film_num,@x,@y,@z,@path_length,
      @phi,@theta,@psi,@subunit_num,@score,@refine) = string.split
      #
      @fil_num = @fil_num.to_i
      @num = @num.to_i
      @x = @x.to_f
      @y = @y.to_f
      @z = @z.to_f
      @path_length = @path_length.to_f
      @phi = @phi.to_f
      @theta = @theta.to_f
      @psi = @psi.to_f
      @subunit_num = @subunit_num.to_f
      @score = @score.to_f
      @number_of_frames = 1
      @subframe_shifts = "0.0 0.0 0.0"
      $LOG.debug() {"New waypoint has score #{@score} (4)"}
  end

	def convertPixToAng(pixel_size)
	  @x = (@x-1.0) * pixel_size
	  @y = (@y-1.0) * pixel_size
	  @z = (@z-1.0) * pixel_size
	  @path_length = @path_length * pixel_size
	  @rise = @rise * pixel_size
	end

	def convertAngToPix(pixel_size)
    @x = (@x / pixel_size) + 1.0
    @y = (@y / pixel_size) + 1.0
    @z = (@z / pixel_size) + 1.0
    @path_length = (@path_length / pixel_size)
    @rise = (@rise / pixel_size)
	end

	def correctMagDistortion!(scale_x,scale_y,azimuth,micrograph_center_x,micrograph_center_y)
		# Relative to center
		temp_x = @x - micrograph_center_x
		temp_y = @y - micrograph_center_y
		# in radians
		angle_in_radians = azimuth * Math::PI / 180.0
		# Rotate
    new_x =  temp_y*Math.sin(-angle_in_radians) + temp_x*Math.cos(-angle_in_radians)
		new_y =  temp_y*Math.cos(-angle_in_radians) - temp_x*Math.sin(-angle_in_radians)
		# Scale
		new_x = new_x * scale_x
		new_y = new_y * scale_y
		# Rotate back
		temp_x =  new_y*Math.sin(angle_in_radians) + new_x*Math.cos(angle_in_radians)
    temp_y =  new_y*Math.cos(angle_in_radians) - new_x*Math.sin(angle_in_radians)
		# Relative to corner origin
		@x = temp_x + micrograph_center_x
	  @y = temp_y + micrograph_center_y
	end

	def addToPhi!(val)
	  @phi += val
	end

	def self.arcLength(one,two)
	  two.path_length - one.path_length
	end
	def self.chordLength(one,two)
	  Math.sqrt((two.x-one.x)**2+(two.y-one.y)**2+(two.z-one.z)**2)
	end
end
