# Dealing with the sun grid engine
class Sge
  NAME = "SGE"

  # Initialiser method with default parameter
  def initialize(options = {})
    guessLocation
    defaults = {
      :location => @loc,
      :gpu_nodes_only => false,
      :extra_qsub_options => ' '
    }
    options = defaults.merge(options)
    @user = ENV['USER']
    @loc  = options[:location] if options[:location]
    @qsub_exec = 'qsub'
    @qstat_exec = 'qstat'
    @qstat_interval = 20.0 #in seconds (was 60.0 at Brandeis)
    @gpu_nodes_only = options[:gpu_nodes_only]
    @extra_qsub_options = options[:extra_qsub_options]
    @extra_qsub_options = "-l " + @extra_qsub_options if @extra_qsub_options && @extra_qsub_options !~ /\A[[:space:]]*\z/
    #puts "debug(Sge.initialize): gpu_nodes_only = #{gpu_nodes_only}, #{@gpu_nodes_only}"
    # Location-dependent settings
    case @loc
    when "brandeis"
      # The Brandeis cluster uses a (fairly old) Sun Grid Engine
      @qsub_exec = '/opt/gridengine/bin/lx26-amd64/qsub'
      @qstat_exec = '/opt/gridengine/bin/lx26-amd64/qstat'
      @qsub_options = {'-hard'=>'','-cwd'=>'','-r'=>'yes','-V'=>'','-b'=>'yes',
                       '-pe'=> 'multi 1','-v'=>'OMP_NUM_THREADS 1',
                       '-l'=>'mem_free=1G'}
    when "janelia"
      # The main janelia cluster uses Univa Grid Engine
      @qsub_options = { '-cwd'=>'','-r'=>'yes','-V'=>'','-b'=>'yes','-v'=>'OMP_NUM_THREADS 1','-pe'=>'batch 1' }
      if (@gpu_nodes_only) then
        @qsub_options << { '-l'=>'gpu=true'}
        puts "Janelia: will use only GPU nodes"
        p @qsub_options
      end
    when "grigorieff_workstations"
      # The Grigorieff workstations run a Torque PBS engine
      @qsub_options = { '-j'=>'oe','-V'=>'','-v'=>'OMP_NUM_THREADS=1','-l'=>'nodes=1:ppn=1','-q'=>'batch', '-d'=>ENV['PWD'] }
    when "ucsf"
      # UCSF run some flavour of Sun Grid Engine. Not sure which exactly.
      @qsub_options = { '-cwd'=>'','-r'=>'yes','-V'=>'','-b'=>'yes','-v'=>'OMP_NUM_THREADS 1','-pe'=>'cpu 1', '-q'=>'all.q' }
    when "nyu"
      @qsub_options = { '-cwd'=>'','-r'=>'yes','-V'=>'','-b'=>'yes','-v'=>'OMP_NUM_THREADS 1','-pe'=>'openmpi 1', '-q'=>'all.q' }

      #@qsub_exec = '/cm/shared/apps/sge/2011.11p1/bin/linux-x64/qsub' #'/opt/gridengine/bin/lx26-amd64/qsub'
      #@qstat_exec = '/cm/shared/apps/sge/2011.11p1/bin/linux-x64/qstat' #'/opt/gridengine/bin/lx26-amd64/qstat'
      #@qsub_options = {'-hard'=>'','-cwd'=>'','-r'=>'yes','-V'=>'','-b'=>'yes',
      #                 '-pe'=> 'openmpi 1',#'-v'=>'OMP_NUM_THREADS 1',
      #                 '-l'=>'mem_free=1G'}
    else
      puts "Unknown location: #{@loc}"
      puts caller(start=0)
      exit 1
    end
    @jobid = []
    @job_stat = {} # Parsed output from qstat
    @job_stat_finished = {} # qstat-reported status for recently finished jobs

    @qstat_time_last = nil
    @qstat_force_refresh = false
    @num_jobs_waiting = 0	# Number of jobs queued waiting (any user)

    # Array of temporary text files
    @tmp_files = []


    #
    self.loadQstatInfo
    self.getTotalNumSlots
  end

  attr_reader :total_num_slots

  # Guess what cluster we're on
  def guessLocation()
    require 'socket'
    hostname = Socket.gethostbyname(Socket.gethostname)[0]
    $LOG.debug("guessLocation") {"hostname = #{hostname}"}
    if (hostname.match("int.janelia.org"))
      @loc = "janelia"
    elsif (hostname.match("janelia.priv"))
      @loc = "grigorieff_workstations"
    elsif (hostname.match("ucsf"))
      @loc = "ucsf"
    elsif (hostname.match("brandeis"))
      @loc = "brandeis"
    elsif (hostname.match("phoenix"))
      @loc = "nyu"
    end
  end

  # Return the string status of specified job, using qstat
  def getJobStatus(jobid)
    self.loadQstatInfo()
    if (@job_stat.has_key?(jobid))
      getJobStatus = @job_stat[jobid]
    elsif (@job_stat_finished.has_key?(jobid))
      getJobStatus = "finished"
    else
      getJobStatus = nil
    end
    return getJobStatus
  end

  # Return the number of jobs found waiting in the queue
  # (i.e. with status "qw")
  def numJobsWaiting()
    loadQstatInfo()
    $LOG.debug("numJobsWaiting") {"#{@num_jobs_waiting} jobs are waiting in the queue"}
    @num_jobs_waiting
  end

  def numSlotsRunning()
    loadQstatInfo()
    @num_slots_running
  end

  def numUsers()
    loadQstatInfo()
    @cluster_users.count
  end

  # Call qstat if it hasn't been called for more than a certain duration
  # Parse the results and load them into a hash
  def loadQstatInfo()
    command = "#{@qstat_exec} -u #{@user}"
    case @loc
    when "grigorieff_workstations"
      command_finished = "#{@qstat_exec} -u #{@user}"
      header_line_pattern = "Job ID"
    else
      command_finished = "#{@qstat_exec} -s z -u #{@user}" # Command to list recently-finished jobs
      header_line_pattern = "job-ID"
    end

    #$LOG.debug("loadQstatInfo") {"Entering. Qstat command = #{command}"}

    if (!which("qstat"))
      puts "qstat not available"
      puts caller(start=0)
      exit 1
    end

    # Only run the qstat command if it's been more than
    # @qstat_interval since the last time we did so
    if (!@qstat_time_last || (Time.now()-@qstat_time_last) > @qstat_interval || @qstat_force_refresh)
      $LOG.debug("loadQstatInfo") {"About to run qstat. Time last, current: #{@qstat_time_last} #{Time.now()}. Command: #{command}"}
      # Run the qstat command and read its output
      q = IO.popen(command,"r")
      qstat_out = q.readlines
      q.close
      # Let's parse the output
      @job_stat = {}

      @num_jobs_waiting = 0
      @num_slots_running = 0  # Number of slots used by jobs currently running
      @cluster_users = [] # All unique users of the cluster
      case @loc
      when "grigorieff_workstations"
        qstat_out.each do |line|
          next if (line.match("Job ID") || line.match("-----") || line.match(/^\n/) || line.match("london.janelia.priv:") || line.match("Req'd"))
          words = line.split
          job_id = words[0].split(".")[0].to_i
          @job_stat[job_id] = words[9] if (words[1] == "#{@user}")
          @num_jobs_waiting += 1 if (words[9] == 'Q')
          @num_slots_running += words[5].to_i * words[6].to_i if (words[9] == 'R')
          @job_stat_finished[job_id] = 'C' if (words[9] == 'C')
          @cluster_users << words[1] if (!@cluster_users.include?(words[1]))
          $LOG.debug("loadQstatInfo") {"JobID, user, status, qw?: #{words[0]} #{words[1]} #{words[9]} #{(words[9] == 'Q')}"}
        end
      else
        qstat_out.each do |line|
          next if (line.match(header_line_pattern) || line.match("-----"))
          words = line.split
          @job_stat[words[0].to_i] = words[4] if (words[3] == "#{@user}")
          @num_jobs_waiting += 1 if (words[4] == 'qw')
          @num_slots_running += words[8].to_i if (words[4] == 'r')
          @cluster_users << words[3] if (!@cluster_users.include?(words[3]))
          $LOG.debug("loadQstatInfo") {"JobID, user, status, qw?: #{words[0]} #{words[3]} #{words[4]} #{(words[4] == 'qw')}"}
        end
        # Now run the command to find recently-finished jobs
        q = IO.popen(command_finished,"r")
        qstat_out = q.readlines
        q.close
        # Let's parse the output
        qstat_out.each do |line|
          next if (line.match("job-ID") || line.match("-----"))
          words = line.split
          @job_stat_finished[words[0].to_i] = words[4]  if words[3] == "#{@user}"
        end
        # Remember
        @qstat_time_last = Time.now()
        @qstat_force_refresh = false
      end # end of test for location
    end # end of test for whether we should run qstat again
  end

  # Submit a job to the queue, return the sge job id or 0 if something went wrong
  def submitJob(executable,arguments,out,err,job_name,num_threads=1,mem_free_gb=1)
    require 'tempfile'
    require 'fileutils' # debug only
    @qsub_options['-v'] = "OMP_NUM_THREADS=#{num_threads}"
    case @loc
    when "brandeis"
      @qsub_options['-pe'] = "multi #{num_threads}"
      @qsub_options['-l'] = "mem_free=#{mem_free_gb}G"
    when "janelia"
      @qsub_options['-pe'] = "batch #{num_threads}"
      if (mem_free_gb > 7.5)
        puts "Janelia cluster only has 7.5 GB per slot"
        puts caller(start=0)
        exit 1
      end
    when "grigorieff_workstations"
      @qsub_options['-l'] = "nodes=1:ppn=#{num_threads}"
      @qsub_options['-v'] = "OMP_NUM_THREADS=#{num_threads}"
    when "ucsf"
      @qsub_options['-pe'] = "cpu #{num_threads}"
    when "nyu"
      @qsub_options['-pe'] = "openmpi #{num_threads}"
      @qsub_options['-l'] = "mem_free=#{mem_free_gb}G"
    end
    qsub_options_string = @qsub_options.collect {|k,v| "#{k} #{v}"}.join(" ")
    File.delete(out) if File.exists?(out)
    File.delete(err) if File.exists?(err)
    case @loc
    when "grigorieff_workstations"
      # Torque PBS does not allow to qsub a compiled binary. Have to make an ascii script
      @tmp_files.push(Tempfile.new(['.frealix_','txt'],tmpdir=ENV['PWD']))
      @tmp_files[-1].write("\#!/bin/bash\n#{executable} #{arguments}\n")
      @tmp_files[-1].close()
      FileUtils.cp(@tmp_files[-1].path,"test.txt")
      command = "#{@qsub_exec} #{qsub_options_string} #{@extra_qsub_options} -N #{job_name} -o #{out} #{@tmp_files[-1].path}"
    else
      command = "#{@qsub_exec} #{qsub_options_string} #{@extra_qsub_options} -N #{job_name} -e #{err} -o #{out} #{executable} #{arguments} "
    end
    $LOG.debug("submitJob") {"About to @sge.submitJob"}
    $LOG.debug("submitJob") {"SGE command is: #{command}"}
    q = IO.popen(command,"r")
    output = q.gets
    q.close
    case @loc
    when "grigorieff_workstations"
      # For Torque, we just check the exit status of the qsub command
      success = $?.success?
      jobid = output.split(".")[0]
    else
      # For (Sun/Univa) grid engines, we check stdout
      success = (output =~ /Your job .* has been submitted/)?true:false
      jobid = output.split[2].to_i
    end
    if (success)
      @qstat_force_refresh = true
      return jobid
    else
      return 0
    end
  end

  # Find the total number of cores available on the cluster
  def getTotalNumSlots()
    case @loc
    when "grigorieff_workstations"
      command = "#{@qstat_exec} -Q"
    else
      command = "#{@qstat_exec} -g c"
    end
    # This will output something like:
    # CLUSTER QUEUE                   CQLOAD   USED  AVAIL  TOTAL aoACDS  cdsuE
    # -------------------------------------------------------------------------------
    # all.q                             0.00      0    368    372      0      4
    # guest.q                           0.00      0     99    103      0      4
    q = IO.popen(command,"r")
    output = q.gets(sep=nil)
    q.close
    case @loc
    when "brandeis"
      regex = /^all\.q\s*[\d\.]*\s*[\d\.]*\s*([\d]*).*$/
    when "janelia"
      regex = /^allnew\.q\s*[\d\.]*\s*[\d\.]*\s*[\d\.]*\s*([\d]*).*$/
    when "grigorieff_workstations"
      regex = /^batch\s*(\d*)/
    when "ucsf"
      regex = /^all\.q\s*[\d\.]*\s*[\d\.]*\s*([\d]*).*$/
    when "nyu"
      regex = /^all\.q\s*[\d\.]*\s*[\d\.]*\s*([\d]*).*$/
    end
    if (output)
      m = output.match(regex)
      @total_num_slots = m[1]
    else
      @total_num_slots = 0
    end
  end

end
