#!/usr/bin/env ruby
#  analyse_filament.rb
#
#  Analyse a single filament after refinement with Frealix
#  
#  Copyright 2014 Howard Hughes Medical Institute.
#  All rights reserved.
#  Use is subject to Janelia Farm Research Campus Software Copyright 1.1
#  license terms ( http://license.janelia.org/license/jfrc_copyright_1_1.html )
#  

require_relative "rubylix.rb"
require_relative "FrealixOutput.rb"
require_relative "FrealixFilament.rb"
require_relative "FrealixWaypoints.rb"
require 'gnuplot'

f_output = ARGV[0]
fil_num  = ARGV[1].to_i
@output_dir = File.dirname(f_output)

$LOG.info() {"Will be analysing filament #{fil_num} from #{f_output}"}

o = FrealixOutput.new(f_output)
f = FrealixFilament.filamentsFromWaypoints(o.getFinalScoreWaypoints(fil_num))
f2= FrealixFilament.filamentsFromWaypoints(o.getFinalWaypoints(fil_num))

pixel_size = o.getPixelSize
pl = f[0].getWaypointPathLengths
pl.collect! { |t| t*pixel_size}	
s  = f[0].getWaypointScores
mean, std = s.sample_stats
s_mean = Array.new(s.count,mean)
s_std  = Array.new(s.count,mean-std)
psi = f[0].getWaypointPsis
theta = f[0].getWaypointThetas
psi_ave, psi_std = psi.sample_stats
theta_ave, theta_std = theta.sample_stats
psi.collect! {|p| p-psi_ave}
theta.collect! {|t| t-theta_ave}
z = f[0].getWaypointZs

wp_path_lengths = f2[0].getWaypointPathLengths
wp_z = f2[0].getWaypointZs
wp_path_lengths.collect! {|t| t*pixel_size}
wp_path_lengths_y = wp_path_lengths.collect{|f| f}
wp_path_lengths_y.fill(0.0)


#
# PLOTTING
#
Gnuplot.open do |gp|
	Gnuplot::Plot.new(gp) do |plot|
		# Some basic formatting
		plot.set "border linewidth 1.5"
		plot.style "line 1 lc rgb '#0060ad' lt 1 lw 2 pt 7 ps 1.5"   # --- blue
		plot.style "line 2 lc rgb '#dd181f' lt 1 lw 2 pt 5 ps 1.5"   # --- red
		plot.style "line 3 lc rgb 'grey30'  lt 1 lw 2 pt 5 ps 1.5"		
		plot.terminal "pngcairo" # size 1024,800"	#"pngcairo enhanced font 'Verdana,10'"
		plot.output "#{@output_dir}/filament_#{fil_num}.png"
		#
		# Plot score along filament, as well as deviations in Euler angles
		#
		plot.title "Filament #{fil_num}"
		plot.set "key font ',10' "
		plot.xlabel "Path length (Angstroms)"
		plot.ylabel "Score"
		plot.yrange "[0.0:0.7]"		
		plot.y2label "Angle difference (degrees)"
		plot.ytics	"nomirror"
		plot.y2tics "-5,1."
		plot.data << Gnuplot::DataSet.new([pl,s]) do |ds|
			ds.using = "1:2"
			ds.with  = "line ls 1 lw 3"
			ds.title = "Score"
		end
		plot.data << Gnuplot::DataSet.new([pl,s_mean]) do |ds|
      ds.using = "1:2"
      ds.with  = "line ls 1 lw 2"
      ds.title = "Score mean"
    end
    plot.data << Gnuplot::DataSet.new([pl,s_std]) do |ds|
      ds.using = "1:2"
      ds.with  = "line ls 1 lw 1"
      ds.title = "Score mean - sigma"
    end
		plot.data << Gnuplot::DataSet.new([pl,psi]) do |ds|
			ds.using = "1:2"
			ds.axes  = "x1y2"
			ds.with  = "line ls 2 lw 3"
			ds.title = "Psi deviation from average"
		end
		plot.data << Gnuplot::DataSet.new([pl,theta]) do |ds|
			ds.using = "1:2"
			ds.axes  = "x1y2"
			ds.with  = "line ls 3 lw 3"
			ds.title = "Theta deviation from average"
		end
		plot.data << Gnuplot::DataSet.new([wp_path_lengths,wp_path_lengths_y]) do |ds|
		  ds.using = "1:2"
		  ds.axes  = "x1y2"
		  ds.with  = "points ls 1 pointsize 0.5"
		  ds.title = "Refinement waypoints"
    end
    
	end
	Gnuplot::Plot.new(gp) do |plot|
    # Some basic formatting
    plot.set "border linewidth 1.5"
    plot.style "line 1 lc rgb '#0060ad' lt 1 lw 2 pt 7 ps 1.5"   # --- blue
    plot.style "line 2 lc rgb '#dd181f' lt 1 lw 2 pt 5 ps 1.5"   # --- red
    plot.style "line 3 lc rgb 'grey30'  lt 1 lw 2 pt 5 ps 1.5"    
    plot.terminal "pngcairo"
    plot.title "Filament #{fil_num}"
    plot.set "key font ',10' "
    plot.xlabel "Path length (Angstroms)"
    plot.ylabel "Z (Angstroms)"
    plot.set "autoscale x"
    plot.yrange "[-28500:-26500]"
    plot.y2label " "
    plot.ytics  "nomirror"
    plot.unset "y2tics"
    plot.unset "key"
	  plot.output "#{@output_dir}/filament_#{fil_num}_z.png"
    plot.data << Gnuplot::DataSet.new([pl,z]) do |ds|
      ds.using = "1:2"
      ds.axes  = "x1y1"
      ds.with  = "line ls 3 lw 3"
      ds.title = "Z height"
    end
    plot.data << Gnuplot::DataSet.new([wp_path_lengths,wp_z]) do |ds|
      ds.using = "1:2"
      ds.axes  = "x1y1"
      ds.with  = "points ls 1 pointsize 2.0"
      ds.title = "Refinement waypoints"
    end
  end
end
	
