#!/usr/bin/env ruby
#
# Wrapper script for frealix
#
# Alexis Rohou, March 2012
#
# Last update: November 2015
#
#  Copyright 2014 Howard Hughes Medical Institute.
#  All rights reserved.
#  Use is subject to Janelia Farm Research Campus Software Copyright 1.1
#  license terms ( http://license.janelia.org/license/jfrc_copyright_1_1.html )
#
#


# We need rubylix, which defines classes to interact with frealix
require_relative "rubylix.rb"
require_relative "FrealixOutput.rb"
require_relative "FrealixParameters.rb"
require_relative "FrealixRefinementDirectory.rb"
require_relative "FrealixRunner.rb"
require_relative "FilamentRefineParallel.rb"


#
# Some options are only enabled for me
#
if (ENV["USER"] == "rohoua")
  $developer_mode = true
	$enable_multi_ref = true
	$default_ali_format = 6
end

#
# Set up command line options
#
require 'fileutils'
require 'optparse'

ARGV << '--help' if ARGV.empty?

options = {}
defaults = {}
optparse = OptionParser.new do |opts|
  opts.banner = "Usage: flx_wrap.rb [options] parameters.flx"

  # Help message
  opts.on_tail("-h","--help","Show this message") do
    puts opts
    exit
  end

  # Number of rounds
  opts.on("-r","--rounds [R]",Integer,"Number of refinement rounds") do |r|
    options[:num_rounds] = r
  end
  defaults[:num_rounds] = 1

  # Maximum resolution to use during refinement
  opts.on("-h","--high_resolution_cutoff [HIGH_RESOLUTION_CUTOFF]",Float,"Highest resolution (Angstroms) used by the scoring function during refinement. Keep this well below the resolution of the model 3D to avoid noise bias.") do |s|
    options[:high_resolution_cutoff] = s
  end
  defaults[:high_resolution_cutoff] = 20.0

  # Target resolution
  #opts.on("-t","--target_resolution [TARGET_RES]",Float,"Target resolution (Angstroms) you hope to achieve in the final 3D reconstruction") do |t|
  #  options[:target_res] = t
  #end
  #defaults[:target_res] = 5.0

  # Number of classes
  if ($developer_mode)
    opts.on("-c","--classes [NUM_CLASSES]",Integer,"Number of classes to refine / reconstruct") do |n|
      options[:num_classes] = n
    end
  end
  defaults[:num_classes] = 1

  # Number of nodes
  opts.on("-n","--nodes [N]",Integer,"Number of nodes") do |n|
    options[:num_nodes] = n
    options[:num_nodes_rec] = n
  end
  defaults[:num_nodes] = 1

  # Number of nodes for reconstruction
  opts.on("-m","--nodes_for_reconstruction [N]",Integer,"Number of nodes for 3D reconstruction") do |n|
    options[:num_nodes_rec] = n
  end
  defaults[:num_nodes_rec] = 0

  # Use GPU nodes only (Janelia-specific)
  #opts.on("-g","--gpu_nodes_only","Use GPU nodes only (Janelia-specific)") do |n|
  #  options[:gpu_nodes_only] = n
  #end
  #defaults[:gpu_nodes_only] = false

  # Starting model
  opts.on("-m","--model_3d [MODEL]","Starting 3D model") do |m|
    options[:starting_model_3d] = m
  end

  # Input alignment file
  opts.on("-a","--input_alignments [ALI]","Starting alignments") do |a|
    options[:starting_alignments_file] = a
  end

  # Prefix for the queue jobs
  opts.on("-p", "--prefix [P]","Prefix for SGE jobs") do |p|
    options[:prefix] = p
  end
  defaults[:prefix] = "flx"

  # Number of threads
  opts.on("--threads [T]",Integer,"Number of threads") do |t|
    options[:threads] = t
  end
  defaults[:threads] = 1

  # Extra qsub options
  opts.on("--extra_qsub_options [OPT]","Extra options to the qsub command (do not use - broken)") do |a|
    options[:extra_qsub_options] = a
  end

  # Free memory Gb
  opts.on("--free_memory_gb [N]",Float,"Free memory required per thread (GB)  ") do |fm|
    options[:mem] = fm
  end
  defaults[:mem] = 0.0

  # Additional hosts available
  #if ($developer_mode)
  opts.on("--hostsfile [extra_hosts.txt]","Flat text list of hosts available for processing. localhost and cluster are recognised.") do |h|
    options[:hostsfile] = h
  #end
  end


  # No cleanup not implemented

  # Apply defaults
  options = defaults.merge(options)


end.parse! # End of command-line options

# Parse the command-line. Remember there are two forms
# of the parse method. The 'parse' method simply parses
# ARGV, while the 'parse!' method parses ARGV and removes
# any options found there, as well as any parameters for
# the options. What's left is frealix parameter file.
optparse.parse!


#
# Finished getting options and parameters from the command line, let's start work
#
$LOG.info("flx_wrap") {"Thanks for using Frealix! Let's get to work..."}

# Load extra available hosts into array
available_hosts = []
if (options.has_key?(:hostsfile))
  if (File.exists?(options[:hostsfile]))
    str = File.read(options[:hostsfile])
    available_hosts = str.split("\n")
  else
    $LOG.fatal("Error: file does not exist: #{options[:hostsfile]}")
    exit 1
  end
end
$LOG.debug("Available hosts = #{available_hosts}")

# Initialise the current directory as a Frealix refinement directory
flxdir = FrealixRefinementDirectory.new(:number_of_classes=>options[:number_of_classes])

# What round will we be starting on?
starting_round = flxdir.lastCompletedRound() + 1
$LOG.info("We will be starting on round #{starting_round} and doing #{options[:num_rounds]} refinement rounds using parameters from #{ARGV[0]}")
finishing_round = starting_round + options[:num_rounds] -1
finishing_round += 1 if (starting_round == 0)

# Load the base parameters
base_params = FrealixParameters.new(filename=ARGV[0])
base_params.readFromDisk()

# Define a list of required RTPs
required_rtps = ["CTFTILT_FILES","CTFFIND_FILES","FILM_FILES",#"FILM_SELECTION",
                 "TWIST_PER_SUBUNIT","RISE_PER_SUBUNIT","PIXEL_SIZE",
                 "HELIX_MAX_RADIUS"]
required_rtps_startup = required_rtps + [] #["INPUT_BOX_FILES"]
required_rtps_refine  = required_rtps + []

# Check the base parameters have the required RTPs
#if (! base_params.checkForParameters?(required_rtps))
  #$LOG.error("flx_wrap") {"The following parameters are missing from #{base_params.filename}: "}
  #$LOG.error("flx_wrap") {base_params.findMissingParameters(required_rtps)}
  #$LOG.error("flx_wrap") {"Would normally crash out now!"}
  #exend

# We will only use provided starting 3Ds and waypoints files when starting on round 0
if (starting_round > 0)
  if (options[:starting_alignments_file] != nil)
    $LOG.warn("Starting ali file will be ignored, since refinement has already started in this directory")
  end
  if (options[:starting_model_3d])
    $LOG.warn("Starting model 3D will be ignored, since refinement has already started in this directory")
  end
end

#
# For round 0, we may need to run frealix mode FILAMENT_STARTUP
#
if (starting_round == 0)

  # If the directory already exists, it probably means that round 0
  # was started but aborted
  if (Dir.exists?(flxdir.currentRoundDir()))
    $LOG.fatal("flx_wrap") {"Round 0 was started previously, but not completed. Please delete the folder and start again."}
    exit 1
  else
    # Make the directory
    current_round_dir = flxdir.mkdirCurrentRound()
  end

  # Clone the runtime parameters
  rtps_startup = base_params.dclone
  rtps_startup.filename = current_round_dir+"frealix_parameters.flx"


  # First, create an ali file if needed
  if (options[:starting_alignments_file] == nil)
    rtps_startup.rtp["FREALIX_MODE"] = "FILAMENT_STARTUP"
    log = FrealixOutput.new(current_round_dir+"/frealix_output.log")
    flx_runner = FrealixRunner.new( :num_nodes=>1,:params=>[rtps_startup],
                                    :logs=>[log],:hosts=>available_hosts,
                                    :threads_per_job=>options[:threads],
                                    :mem_per_thread_gb=>options[:mem],
                                    :prefix=>options[:prefix],
                                    :gpu_nodes_only=>options[:gpu_nodes_only])

    flx_runner.run()
  else
    if ($enable_multi_ref)
      1.upto(options[:num_classes]) do |class_counter|
        class_number = "%03i"%(class_counter)
        FileUtils.cp options[:starting_alignments_file], current_round_dir+"waypoints_r"+class_number+File.extname(options[:starting_alignments_file])
      end
    else
      FileUtils.cp options[:starting_alignments_file], current_round_dir+"waypoints.ali"
    end
  end

  # Now compute a starting 3D reconstruction if needed
  if (options[:starting_model_3d] == nil)
    rtps_startup.rtp["FREALIX_MODE"]        = "FILAMENT_REFINE"
    rtps_startup.rtp["SKIP_REFINEMENT"]     = "T"
    rtps_startup.rtp["MODEL_3D_AVAILABLE"]  = "F"
    #rtps_startup.rtp["OUTPUT_3D_FILENAME"]  = current_round_dir+"3drec.mrc"
    if ($enable_multi_ref)
      FilamentRefineParallel.createClassesFromAli(current_round_dir+"waypoints.ali",options[:num_classes])
    end
    #rtps_startup.rtp["INPUT_PAR_FILENAME"]  = current_round_dir+"waypoints.ali"
    rtps_startup.rtp["FINALISE_3D_RECONSTRUCTION"] = 'T'
    options[:starting_alignments_file]      = current_round_dir+"waypoints.ali"
    rtps_refine = rtps_startup
    flx_refine_para = FilamentRefineParallel.new(
                :num_nodes=>options[:num_nodes],
                :num_nodes_rec=>options[:num_nodes_rec],
                :num_threads=>options[:threads],
                :params=>rtps_startup,
                :hosts=>available_hosts,
                :mem_per_thread_gb=>options[:mem],
                :prefix=>options[:prefix],
                :gpu_nodes_only=>options[:gpu_nodes_only])
    flx_refine_para.parallelRun()
  else
    starting_threed = options[:starting_model_3d]
    if ($enable_multi_ref)
      1.upto(options[:num_classes]) do |class_counter|
        class_number = "%03i"%(class_counter)
        FileUtils.cp starting_threed, current_round_dir+"3drec_r"+class_number+File.extname(starting_threed)
      end
    else
      FileUtils.cp starting_threed, current_round_dir+"3drec"+File.extname(starting_threed)
    end
  end



  starting_round = 1
end # round 0

#
# For rounds 1+, we need to run frealix mode FILAMENT_REFINE
#
starting_round.upto(finishing_round) do |round|
  # The directories
  previous_round_dir = flxdir.previousRoundDir()
  current_round_dir  = flxdir.currentRoundDir()
  resume_processing = false
  if (Dir.exists?(current_round_dir))
    if (round == starting_round)
      $LOG.warn("flx_wrap") {"Will attempt to resume processing at round #{round}"}
      resume_processing = true
    else
      $LOG.fatal("flx_wrap") {"It looks like round #{round} was started previously. Please delete the folder and start again"}
      exit 1
    end
  else
    flxdir.mkdirCurrentRound()
  end
  # Prepare the runtime parameters
  rtps_refine = base_params.dclone
  rtps_refine.rtp["FREALIX_MODE"] = "FILAMENT_REFINE"

  if (rtps_refine.rtp["CALC_3D_RECONSTRUCTION"] == 'F')
    $LOG.warn("flx_wrap") {"You specified CALC_3D_RECONSTRUCTION = F, so no 3D will be computed"}
  else
    rtps_refine.rtp["CALC_3D_RECONSTRUCTION"]      = 'T'
  end

  rtps_refine.rtp["HIGH_RESOLUTION_CUTOFF"] = options[:high_resolution_cutoff]
  $LOG.info("flx_wrap") { "Round #{round}: Cutoff resolution = #{rtps_refine.rtp["HIGH_RESOLUTION_CUTOFF"]} A" }


  # Do the actual refinement for this round
  flx_refine_para = FilamentRefineParallel.new(
								:num_nodes=>options[:num_nodes],
								:num_nodes_rec=>options[:num_nodes_rec],
								:num_threads=>options[:threads],
								:params=>rtps_refine,
								:hosts=>available_hosts,
								:mem_per_thread_gb=>options[:mem],
								:prefix=>options[:prefix],
								:gpu_nodes_only=>options[:gpu_nodes_only],
								:resume=>resume_processing)
  flx_refine_para.parallelRun()


end
