#!/usr/bin/ruby
#
# Given a set of coordinates, alignment parametrs and CTF estimations,
# generate an initial waypoints.ali file for Frealix
#
# Alexis Rohou, 21-Apr-2014
#
# Last modified 7-May-2014
#
require_relative "FrealixWaypoints.rb"
require_relative "FrealixAli.rb"
require_relative "FrealixFilament.rb"
require_relative "rubylix.rb"

#
# Old (not 2D sorted) input file locations
#
input_files_root="/scratch/rohoua/data_elena"
@ctf_summary_fn="#{input_files_root}/misc/ctfpar_micros"
@particle_summary_fn="#{input_files_root}/misc/particle_coord.ali"
@particle_selection_fn="#{input_files_root}/misc/goodpartlist034polar.spi"
@particle_alignment_fn="#{input_files_root}/misc/fparamzds+034.spi"

#
# New (Relion 2D sorted) input file locations
#
input_files_root="/nobackup/grigorieff/rohoua/140514_PhuZGTP/files_from_elena_fixed"
@ctf_summary_fn="#{input_files_root}/ctfpar_micros"
#@particle_summary_fn="#{input_files_root}/particle_coord_2dsort.ali"
@particle_summary_fn="#{input_files_root}/40olap_particles_2dsort.spi"
@particle_selection_fn="#{input_files_root}/goodpartlist030polar.spi"
@particle_alignment_fn="#{input_files_root}/fparamzds+030.spi"
@particle_pre_alignment_fn="#{input_files_root}/fparamzds_4x.spi"


#
# Output ali file location
#
@output_ali_fn="frealix_round_000/waypoints.ali"


#
# Experimental variables
#
@pixel_size = 1.2156
@spherical_aberration = 2.0
@acceleration_voltage = 300.0
@amplitude_contrast = 0.07
@rise_per_subunit = 14.4
@twist_per_subunit = 118.4
@pre_alignment_binning_factor = 4

#
# Micrograph class
#
class Micrograph

	attr_reader :num, :defocus_1, :defocus_2, :astigmatism_angle

	def initialize(options = {})
		# Default values for the options
		defaults = {	:num => 0,
			   			:defocus_1 => 0.0,
						:defocus_2 => 0.0,
						:astigmatism_angle => 0.0
		}
		@num = options[:num] if (options[:num])
		@defocus_1 = options[:defocus_1] if (options[:defocus_1])
		@defocus_2 = options[:defocus_2] if (options[:defocus_2])
		@astigmatism_angle = options[:astigmatism_angle] if (options[:astigmatism_angle])
	end
end


#
# Read in the files
#

# CTF files
@micrographs = []
ctf_summary_file = File.open(@ctf_summary_fn,'r')
ctf_summary_file.each_line do |line|
	if (line.split[0] != 'Num') then
		@micrographs << Micrograph.new(:num=>line.split[0].to_i,
									   :defocus_1=>line.split[1].to_f,
									   :defocus_2=>line.split[2].to_f,
									   :astigmatism_angle=>line.split[4].to_f)
	end
end
puts "Number of micrographs = #{@micrographs.length}"


# Dump fake CTFFIND results to disk
puts "Generating fake CTFFIND output files..."
ctf_results_dir="/groups/grigorieff/home/rohoua/data/PhuZ_Apr2014/ctffind_results"
Dir.mkdir(ctf_results_dir) if (! File.directory?(ctf_results_dir))
@micrographs.each do |micrograph|
	filename = ctf_results_dir+ "/"+micrograph.num.to_s.rjust(5,"0")+".log"
	file = File.new(filename,'w')
	file.printf(" CS[mm], HT[kV], AmpCnst, XMAG, DStep[um]\n %3.1f   %5.1f   %4.2f\n\n",@spherical_aberration,@acceleration_voltage,@amplitude_contrast)
	file.printf(" %11.2f %11.2f %11.2f %11.5f  Final Values\n",micrograph.defocus_1,micrograph.defocus_2,micrograph.astigmatism_angle,0.10)
	file.close
end


puts "Processing particles, converting to waypoints..."

# Picked particles
@waypoints = []
current_waypoint_number  = 0
current_filament_number  = 0
previous_filament_number = 0
particle_summary_file = File.open(@particle_summary_fn,'r')
particle_summary_file.each_line do |line|
	if (line.split[0] !~ /ZZZ/) then
		current_filament_number  = line.split[0].to_i
		if (current_filament_number != previous_filament_number) then
			current_waypoint_number  = 1
		else
			current_waypoint_number += 1
		end
		previous_filament_number = current_filament_number
		@waypoints << FrealixWaypoint.new(:fil_num=>line.split[3].to_i,
								   	      :num=>current_waypoint_number,
								          :film_num=>line.split[2].to_i,
								          :x=>line.split[4].to_f,
								          :y=>line.split[5].to_f,
										  :psi=>90.0-line.split[6].to_f)
	end
end
puts "Number of picked particles = #{@waypoints.length}"

# Pre-alignment parameters
#
# In Elena's procedure, a few rounds of refinement were done with decimated
# boxed images. Then the Y-shifts only (those orthogonal to the helical axis)
# were applied to the non-decimated (bin1) stack of boxes to "pre-center" them for 
# further alignments.
dx_pre_arr = []
dy_pre_arr = []
pre_alignment_file = File.open(@particle_pre_alignment_fn,'r')
pre_alignment_file.each_line do |line|
	waypoint_number = pre_alignment_file.lineno - 2
	if (waypoint_number >= 0) then
		in_plane = line.split[13].to_f / 180.0 * Math::PI
		shx      = line.split[14].to_f * @pre_alignment_binning_factor.to_f
		shy      = line.split[15].to_f * @pre_alignment_binning_factor.to_f

		#
		dx_pre_arr  << (shy * Math.sin(in_plane)).to_i
		dy_pre_arr  << -(shy * Math.cos(in_plane)).to_i
	end
end

#
# Alignment parameters
# (see project update 6-Mar-2013, modified 23-Apr-2013
#
particle_alignment_file = File.open(@particle_alignment_fn,'r')
particle_alignment_file.each_line do |line|
	waypoint_number = particle_alignment_file.lineno - 2
	if (waypoint_number >= 0) then
		in_plane                           = line.split[13].to_f
		@waypoints[waypoint_number].psi    = -in_plane
		@waypoints[waypoint_number].theta  =  line.split[3].to_f
		@waypoints[waypoint_number].phi    =  line.split[4].to_f
		dx                                 = line.split[14].to_f
		dy                                 = line.split[15].to_f
		x_pick                             = @waypoints[waypoint_number].x
		y_pick                             = @waypoints[waypoint_number].y
		dx_pre                             = dx_pre_arr[waypoint_number]
		dy_pre                             = dy_pre_arr[waypoint_number]
		sin_psi                            = Math.sin(@waypoints[waypoint_number].psi / 180.0 * Math::PI)
		cos_psi                            = Math.cos(@waypoints[waypoint_number].psi / 180.0 * Math::PI)
		#
		@waypoints[waypoint_number].x = x_pick + ( dx_pre - sin_psi*dy - cos_psi*dx )
		@waypoints[waypoint_number].y = y_pick + ( dy_pre - cos_psi*dy + sin_psi*dx ) 
		# Convert from pixels to Angstroms
        @waypoints[waypoint_number].x = (@waypoints[waypoint_number].x - 1.0) * @pixel_size
        @waypoints[waypoint_number].y = (@waypoints[waypoint_number].y - 1.0) * @pixel_size
		# Force theta > 0.0
		if (@waypoints[waypoint_number].theta < 0.0) then
			#@waypoints[waypoint_number].phi   +=  180.0
			#@waypoints[waypoint_number].theta *= -1.0
			#@waypoints[waypoint_number].psi   -=  180.0
			@waypoints[waypoint_number].theta += 360.0 * (-(@waypoints[waypoint_number].theta / 360.0).to_i + 1)

		end
	end
end

#
# Remove particles (waypoints) that were not selected for 3D with Spider
#
@selected_waypoints = []
particle_selection_file = File.open(@particle_selection_fn,'r')
particle_selection_file.each_line do |line|
	output_waypoint_number = particle_selection_file.lineno - 2
	if (output_waypoint_number >= 0) then
		input_waypoint_number = line.split[2].to_i - 1
		@selected_waypoints << @waypoints[input_waypoint_number]
	end
end
puts "Number of particles selected for 3D = #{@selected_waypoints.length}"

#
# Convert array of waypoints to array of filaments
#
puts "Generating filament objects..."
@filaments = FrealixFilament.filamentsFromWaypoints(@selected_waypoints)

#
# Make phi monotonic increasing
#
#puts "Fixing phi..."
twist_per_pixel = @twist_per_subunit / @rise_per_subunit * @pixel_size
twist_per_angstrom = @twist_per_subunit / @rise_per_subunit
#@filaments.each { |fil| fil.makePhiMonotonicIncreasing!(twist_per_angstrom) }

#
# Remove filaments that are too short
#
puts "Removing short filaments..."
@filaments.delete_if { |fil| fil.wps.length < 3 }

#
# Fix theta & psi
#
#puts "Fixing theta & psi..."
#@filaments.each { |fil| fil.replacePsiAndThetaOutliersWithMedians! }

#
# Write the waypoints to disk
#
puts "Writing to disk..."
FrealixFilament.filamentsToAliFile(@filaments,@output_ali_fn)
#@output_ali = FrealixAli.new(:filename=>'test.ali',:units=>'Angstroms')
#@output_ali.putWaypoints(@waypoints)
#@output_ali.writeToDisk()
