#!/usr/bin/env ruby
#  parse_flx.rb
#
#  A script to parse the results of frealix refinement
#
#  Alexis Rohou, March 2012
#
#  Copyright 2014 Howard Hughes Medical Institute.
#  All rights reserved.
#  Use is subject to Janelia Farm Research Campus Software Copyright 1.1
#  license terms ( http://license.janelia.org/license/jfrc_copyright_1_1.html )
#  
#

# What we need
require_relative "FrealixOutput.rb"
require_relative "FrealixRefinementDirectory.rb"
require_relative "FrealixAli.rb"
require_relative "FrealixFilament.rb"
require_relative "rubylix.rb"

$LOG.level = Logger::INFO

# From the standard library
require 'csv'

# External dependencies
require 'histogram/array'
require 'gnuplot'


# Initialise Frealix Refinement Directory
rootdir = FrealixRefinementDirectory.new

# The first argument specifies which round we want to analyse
#roundnum = ARGV[0] || rootdir.lastCompletedRound

# The subdirectory in which the processing was done for this round
#d = rootdir.roundSubDir(roundnum)
d = ARGV[0] || rootdir.roundSubDir(rootdir.lastCompletedRound)
roundnum = FrealixRefinementDirectory.roundNumFromSubDir(d)

# Summary_text will accumulate any extra information we want in the final pdf
wd = Dir::pwd
summary_text = []
summary_text << "Summary for round #{roundnum} of Frealix processing"
summary_text << "Working directory: #{wd}"


#
# Now let's get at the waypoints
#
ali = FrealixAli.new(:filename=>d+"/waypoints.ali")
wps = ali.getWaypoints
$LOG.debug() {"First wp has score of #{wps[0].score}"}
# Parse waypoints into an array of filaments
fils = FrealixFilament.filamentsFromWaypoints(wps)
# Get the scores into array, and get a histogram
scores = []
fils.each { |fil|
	scores << fil.score
}
summary_text << "Number of filaments = #{scores.size}"
average_score = scores.inject(0.0){|sum, el| sum + el} / scores.size
lowest_score = scores.min
highest_score = scores.max
summary_text << sprintf("Average filament score = %.4f",average_score)
summary_text << sprintf("Lowest, highest filament score = %.4f ; %.4f",lowest_score,highest_score)
if (! scores.all_same?)
	(score_bins,score_freqs) = scores.histogram(:middle,	:tp => :min)
	score_bin_titles = []
	score_bins.each_with_index { |bin,i|
		score_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,score_bins[i+1]||scores.max)
	}
end
$LOG.info() {"Waypoints.ali file has been parsed, will now parse the log files"}

#
# Let's find all the log files in there and parse them all
#
logs = []
Dir.glob(d+"/frealix_output_[0-9][0-9][0-9].log") { |log_filename|
	logs << FrealixOutput.new(log_filename)
}
$LOG.debug() {"Initialised "+logs.size.to_s+" log files"}

#
# For each filament, work out which log file it was processed in
# and grab all the information we can from the log file
#
# Cycle through all the log files until we find the one
# which deals with the current filament
warnings = []
logs.each{ |log|
	fil_nums = log.getFilamentNumbers
	# Cycle through the filaments
	fils.each{ |fil|
  	if (fil_nums.index(fil.num))
  		fil.log_filename = log.filename
  		$LOG.debug() {"Updating filament "+fil.num.to_s+" with log file "+fil.log_filename}
  		log.updateFilament(fil)
  	end
	}
	warnings += log.getWarnings
}
$LOG.info() {"All filaments have been updated with info from log files"}
# Sort the filament array according to filament number
fils.sort!{|x,y| x.num <=> y.num}
  
# Generate a single warnings string for printout to pdf later
warnings.uniq!
warnings_summary = warnings.join("\n")

#
# Let's turn the filaments to their "smooth" Euler angle description
#
fils.each{|fil| fil.psiThetaSmooth}


log_table = [["Filament #","Log filename","Score"]]
fils.each{|fil| log_table << [fil.num,fil.log_filename,fil.score]}


# Total number of subunits that went into the reconstuction
total_num_subunits = 0
fils.each{|fil| total_num_subunits += fil.num_subunits}
summary_text << sprintf("Total number of asymmetric subunits which contributed to the 3D reconstruction = %d",total_num_subunits)
  
#
# Get the duration for each Frealix run
#
dollars_per_slot_per_hour = 0.05
durations = []
logs.each { |log|
	secs = log.getDuration()
	$LOG.debug() {log.filename+": "+secs.to_s if (secs)}
	durations << (secs.to_f / 60.0 / 60.0) # Seconds -> hours
}
duration_max = durations.max
summary_text << sprintf("Maximum frealix run time = %.1f hours",duration_max)
num_threads = logs[0].getNumThreads()
cost = durations.reduce(:+) * dollars_per_slot_per_hour * num_threads
summary_text << sprintf("Total cost on Janelia cluster would have been $%.2f",cost)


# We now need to obtain a histogram of the durations
# by default, uses Freedman-Diaconis method to calculate optimal number of bins
# and the bin values are midpoints between the bin edges (we override that)
if (durations.count(scores[0]) != durations.size)
	$LOG.debug() {"About to compute histogram"}
	if (durations.size > 1)
	  (duration_bins,duration_freqs) =  durations.histogram(:middle, :tp => :min)
	else
	  duration_bins = []
	  duration_freqs = []
	end
	# Come up with useful bin titles
	duration_bin_titles = []
	duration_bins.each_with_index { |bin,i|
		duration_bin_titles << sprintf("\"%0.1f-%0.1f hrs\"",bin,duration_bins[i+1]||duration_max)
	}
end

#
# Pixel size
#
pixel_size = 0.0
logs.each do |log|
  pixel_size = log.getPixelSize
  break if (pixel_size)
end
pixel_size = 0.0 if (!pixel_size)
summary_text << sprintf("Pixel size = %.7g A",pixel_size)

#
# Step size / segment length
# in scoring function
#
scoring_step_sizes = []
logs.each { |log|
  # Get step sizes in pixels
  scoring_step_sizes += log.getScoringStepSizes
}
scoring_step_sizes.map!{|s| s*pixel_size} # Convert to Angstroms
(mean,sd)   = scoring_step_sizes.sample_stats
summary_text << sprintf("Scoring function step size, average & standard deviation: %.7g (%.7g) A",mean,sd) if (mean && sd)
if (! scoring_step_sizes.all_same?)
  (scoring_step_sizes_bins,scoring_step_sizes_freqs) = scoring_step_sizes.histogram(:scott,:tp=>:min)
  scoring_step_sizes_bin_titles = []
  scoring_step_sizes_bins.each_with_index{|bin,i|
    scoring_step_sizes_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,scoring_step_sizes_bins[i+1]||scoring_step_sizes.max)
  }  
end

#
# Number of waypoints
#
num_wps = []
fils.each {|fil|
  num_wps << fil.wps.size
}
(mean,sd) = num_wps.sample_stats
summary_text << sprintf("Min, max, mean & std number of waypoints: %d, %d, %d, %d",num_wps.min,num_wps.max,mean,sd)

#
# Get statistics of waypoint-to-waypoint distances
#
distances = []
distances_nofal = []  # Without FAL WPs
fils.each {|fil|
  fil.wps.each_index{|i|
    if (i < fil.wps.size-1)
      dist = fil.wps[i+1].path_length - fil.wps[i].path_length
      distances << dist
      if (i > 0 && i < fil.wps.size-2)
        distances_nofal << dist
      end
    end

  }
}
(mean,sd)   = distances.sample_stats
summary_text << sprintf("Waypoint-to-waypoint distance, average & standard deviation: %.7g (%.7g) A",mean,sd)
(mean,sd)   = distances_nofal.sample_stats
summary_text << sprintf("Waypoint-to-waypoint distance, average & standard deviation (excluding FAL WPs): %.7g (%.7g) A",mean,sd)


#
# How many filaments changed polarity?
#
num_flips = 0
fils.each_index{|i|
  num_flips += 1 if fils[i].polarity_flipped
}
summary_text << sprintf("Number of filaments with flipped polarity: #{num_flips}")

#
# If polarity was checked, it's interesting to know how different the polarities scored
polarity_dscores = []
fils.each{|fil| polarity_dscores << fil.polarity_dscore}
if (polarity_dscores.all_zeroes?) then
  polarity_dscore_bins = [0.0]
  polarity_dscore_freqs = [polarity_dscores.size.to_f]
else
  (polarity_dscore_bins,polarity_dscore_freqs) = polarity_dscores.histogram(:middle,:tp=>:min)
end
polarity_dscore_bin_titles = []
polarity_dscore_bins.each_with_index{|bin,i|
  polarity_dscore_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,polarity_dscore_bins[i+1]||polarity_dscores.max)
}  
(mean,sd)   = polarity_dscores.sample_stats
summary_text << sprintf("Difference in score between polarities, average & standard deviation: %.7g (%.7g) A",mean,sd)



#
# Changes in X,Y,Z,Phi,Theta,Psi,subunit_num
#
dxs = []
dys = []
dzs = []
dphis = []
dthetas = []
thetas = []
dpsis = []
dsubs = []
fils.each {|fil|
  fil.wps_changes.each { |wp|
    dxs << wp.x * pixel_size
    dys << wp.y * pixel_size
    dzs << wp.z * pixel_size
    dphis << wp.phi
    dthetas << wp.theta
    dpsis << wp.psi
    dsubs << wp.subunit_num
  }
  fil.wps.each { |wp|
    if (wp.theta < 0.0) then
      thetas << wp.theta + 90.0
    else
      thetas << wp.theta - 90.0
    end
  }
}
if ( dxs.all_same?)
  dx_bins = [0.0]
  dx_freqs = [dxs.size.to_f]
else
  (dx_bins,dx_freqs) = dxs.histogram(:scott,:tp=>:min)
end
dx_bin_titles = []
dx_bins.each_with_index{|bin,i|
  dx_bin_titles << sprintf("\"%0.3f-%0.3f A\"",bin,dx_bins[i+1]||dxs.max)
}
if ( dys.all_same?) then
  dy_bins = [0.0]
  dy_freqs = [dys.size.to_f]
else
  (dy_bins,dy_freqs) = dys.histogram(:scott,:tp=>:min)
end
dy_bin_titles = []
dy_bins.each_with_index{|bin,i|
  dy_bin_titles << sprintf("\"%0.3f-%0.3f A\"",bin,dy_bins[i+1]||dys.max)
}
if ( dzs.all_same? ) then
  dz_bins = [0.0]
  dz_freqs = [dzs.size.to_f]
else
  (dz_bins,dz_freqs) = dzs.histogram(:scott,:tp=>:min)
end
dz_bin_titles = []
dz_bins.each_with_index{|bin,i|
  dz_bin_titles << sprintf("\"%0.3f-%0.3f A\"",bin,dz_bins[i+1]||dzs.max)
}
if ( dphis.all_same?) then
  dphi_bins = [0.0]
  dphi_freqs = [dphis.size.to_f]
else
  (dphi_bins,dphi_freqs) = dphis.histogram(:scott,:tp=>:min)
end
dphi_bin_titles = []
dphi_bins.each_with_index{|bin,i|
  dphi_bin_titles << sprintf("\"%0.3f-%0.3f deg\"",bin,dphi_bins[i+1]||dphis.max)
}
if (dthetas.all_same?) then
  dtheta_bins = [0.0]
  dtheta_freqs = [dthetas.size.to_f]
elsif (dthetas.all_zeroes_or_180s?)
  (dtheta_bins,dtheta_freqs) = dthetas.histogram(:middle,:tp=>:min)
else
  (dtheta_bins,dtheta_freqs) = dthetas.histogram(:scott,:tp=>:min)
end
dtheta_bin_titles = []
dtheta_bins.each_with_index{|bin,i|
  dtheta_bin_titles << sprintf("\"%0.3f-%0.3f deg\"",bin,dtheta_bins[i+1]||dthetas.max)
}
if (thetas.all_zeroes?) then
  theta_bins = [0.0]
  theta_freqs = [thetas.size.to_f]
elsif (thetas.all_zeroes_or_180s?)
  (theta_bins,theta_freqs) = thetas.histogram(:middle,:tp=>:min)
else
  (theta_bins,theta_freqs) = thetas.histogram(:middle,:tp=>:min)
end
theta_bin_titles = []
theta_bins.each_with_index{|bin,i|
  theta_bin_titles << sprintf("\"%0.3f-%0.3f deg\"",bin,theta_bins[i+1]||thetas.max)
}
if (dpsis.all_same?) then
  dpsi_bins = [0.0]
  dpsi_freqs = [dpsis.size.to_f]
else
  (dpsi_bins,dpsi_freqs) = dpsis.histogram(:scott,:tp=>:min)
end
dpsi_bin_titles = []
dpsi_bins.each_with_index{|bin,i|
  dpsi_bin_titles << sprintf("\"%0.3f-%0.3f deg\"",bin,dpsi_bins[i+1]||dpsis.max)
}
if (dsubs.all_same?) then
  dsub_bins = [0.0]
  dsub_freqs = [dsubs.size.to_f]
else
  (dsub_bins,dsub_freqs) = dsubs.histogram(:scott,:tp=>:min)
end
dsub_bin_titles = []
dsub_bins.each_with_index{|bin,i|
  dsub_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,dsub_bins[i+1]||dpsis.max)
}
(mean,sd)   = dxs.abs.sample_stats
summary_text << sprintf("Waypoint X shift, average & standard deviation: %.7g (%.7g) A",mean,sd)
(mean,sd)   = dys.abs.sample_stats
summary_text << sprintf("Waypoint Y shift, average & standard deviation: %.7g (%.7g) A",mean,sd)
(mean,sd)   = dzs.abs.sample_stats
summary_text << sprintf("Waypoint Z shift, average & standard deviation: %.7g (%.7g) A",mean,sd)
(mean,sd)   = dphis.abs.sample_stats
summary_text << sprintf("Waypoint Phi shift, average & standard deviation: %.7g (%.7g) deg",mean,sd)
(mean,sd)   = dthetas.abs.sample_stats
summary_text << sprintf("Waypoint Theta shift, average & standard deviation: %.7g (%.7g) deg",mean,sd)
(mean,sd)   = dpsis.abs.sample_stats
summary_text << sprintf("Waypoint Psi shift, average & standard deviation: %.7g (%.7g) deg",mean,sd)
(mean,sd)   = dsubs.abs.sample_stats
summary_text << sprintf("Waypoint subunit_num shift, average & standard deviation: %.7g (%.7g)",mean,sd)
(mean,sd)   = thetas.abs.sample_stats
summary_text << sprintf("Waypoint out-of-plane tilt, average & standard deviation: %.7g (%.7g) deg",mean,sd)

#
# Prepare data of score vs. total length and score vs. film_num and z
#
scores = []
lengths = []
film_numbers = []
fil_zs = []
fils.each {|fil| scores << fil.score; lengths << fil.length; film_numbers << fil.film_num; fil_zs << fil.average_z}

#
# Prepare score (without mechanical restraints)
#
scores_no_restraints = []
fils.each {|fil| scores_no_restraints << fil.score_no_restraints}
if (! scores_no_restraints.all_same?)
	(score_nr_bins,score_nr_freqs) = scores_no_restraints.histogram(:middle,  :tp => :min)
	score_nr_bin_titles = []
	score_nr_bins.each_with_index { |bin,i|
	score_nr_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,score_nr_bins[i+1]||scores_no_restraints.max)
	}
end

#
# Prepare scoring segment score v distance to nearest waypoint
#
segment_scores = []
segment_dist_to_wp = []
# Loop over filaments
fils.each{ |fil|
  # Loop over score waypoints
  fil.score_wps.each { |score_wp|
    # Look for the nearest waypoint
    shortest_dist = 9999999999.9
    fil.wps.each { |wp|
      # I believe the score waypoints have path lengths in pixels
      # whereas the filament waypoints have path lengths in Angstroms
      dist = (score_wp.path_length * pixel_size - wp.path_length).abs
      shortest_dist = dist if (dist < shortest_dist)
    }
    segment_scores << score_wp.score
    segment_dist_to_wp << shortest_dist
  }
}

#
# Get arc/chord length ratio
#
ac_ratios_nofal = []
ac_ratios = []
fils.each {|fil|
          ac_ratios_nofal << fil.getArcChordLengthRatioNoFAL
          ac_ratios       << fil.getArcChordLengthRatioNoFAL
          }
(ac_ratio_mean, ac_ratio_sd) = ac_ratios.sample_stats
(ac_ratio_nofal_mean,ac_ratio_nofal_sd) = ac_ratios_nofal.sample_stats
summary_text << sprintf("Filament arc/chord length ratio, average & standard deviation: %.7g (%.7g) /A",ac_ratio_mean,ac_ratio_sd)
summary_text << sprintf("Filament arc/chord length ratio (no FAL waypoints), average & standard deviation: %.7g (%.7g) /A",ac_ratio_nofal_mean,ac_ratio_nofal_sd)

#
# Prepare arc/chord length ratio histograms
#
if (! ac_ratios.all_same? && ! ac_ratios.any_nan?)
(ac_ratio_bins,ac_ratio_freqs) = ac_ratios.histogram(:scott, :tp=> :min)
ac_ratio_bin_titles = []
ac_ratio_bins.each_with_index { |bin,i|
  ac_ratio_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,ac_ratio_bins[i+1]||ac_ratios.max)
}
end
if (! ac_ratios_nofal.all_same? && ! ac_ratios_nofal.any_nan?)
(ac_ratio_nofal_bins,ac_ratio_nofal_freqs) = ac_ratios_nofal.histogram(:scott, :tp=> :min)
ac_ratio_nofal_bin_titles = []
ac_ratio_nofal_bins.each_with_index { |bin,i|
  ac_ratio_nofal_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,ac_ratio_nofal_bins[i+1]||ac_ratios_nofal.max)
}
end


#
# Get the total curvature, overtwist & stretch for each filament
#
curvatures = []
overtwists = []
stretches  = []
fils.each {|fil|
	curvatures << fil.curvature if (fil.curvature)
	overtwists << fil.overtwist if (fil.curvature)
	stretches << fil.stretch if (fil.curvature)
}
(curvature_mean, curvature_sd) = curvatures.sample_stats
(overtwist_mean, overtwist_sd) = overtwists.sample_stats
(stretch_mean,   stretch_sd)   = stretches.sample_stats
summary_text << sprintf("Filament curvature, average & standard deviation: %.7g (%.7g) /A",curvature_mean,curvature_sd) if (curvature_mean && curvature_sd)
summary_text << sprintf("Filament overtwist, average & standard deviation: %.7g (%.7g) /A",overtwist_mean,overtwist_sd) if (overtwist_mean && overtwist_sd)
summary_text << sprintf("Filament stretch, average & standard deviation: %.7g (%.7g) /A",stretch_mean,stretch_sd)       if (stretch_mean   && stretch_sd)

# Are the scores & curvatures correlated?
require 'statsample'
summary_text << sprintf("Correlation between scores (no restraints) and curvatures:%.7g", Statsample::Bivariate.pearson(scores_no_restraints.to_scale,curvatures.to_scale)) unless (curvatures.index(nil) ||
scores_no_restraints.index(nil))


#
# Prepare curvature, overtwist & stretch histograms
#
if (! curvatures.all_same?)
(curvature_bins,curvature_freqs) = curvatures.histogram(:scott, :tp=> :min)
curvature_bin_titles = []
curvature_bins.each_with_index { |bin,i|
  curvature_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,curvature_bins[i+1]||curvatures.max)
}
end
if (! overtwists.all_same?)
(overtwist_bins,overtwist_freqs) = overtwists.histogram(:scott, :tp=> :min)
overtwist_bin_titles = []
overtwist_bins.each_with_index { |bin,i|
  overtwist_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,overtwist_bins[i+1]||overtwists.max)
}
end
if (! stretches.all_same?)
(stretch_bins,stretch_freqs) = stretches.histogram(:scott, :tp=> :min)
stretch_bin_titles = []
stretch_bins.each_with_index { |bin,i|
  stretch_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,stretch_bins[i+1]||stretches.max)
}
end

#
# Prepare curvature, overtwist & stretch histograms
# without outliers
#
std_dev_thresh = 4.0
curvatures_trim = []
overtwists_trim = []
stretches_trim  = []
curvatures.each {|curv| curvatures_trim << curv if (curv <= std_dev_thresh*curvature_sd)}
overtwists.each {|over| overtwists_trim << over if (over <= std_dev_thresh*overtwist_sd)}
stretches.each  {|stre| stretches_trim  << stre if (stre <= std_dev_thresh*stretch_sd)}
summary_text << sprintf("Trimmed curvature, overtwist & stretch histograms exclude filament beyond %.2fx std. dev.",std_dev_thresh)
if (! curvatures_trim.all_same?)
(curvature_trim_bins,curvature_trim_freqs) = curvatures_trim.histogram(:middle, :tp=> :min)
curvature_trim_bin_titles = []
curvature_trim_bins.each_with_index { |bin,i|
  curvature_trim_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,curvature_trim_bins[i+1]||curvatures_trim.max)
}
end
if (! overtwists_trim.all_same?) 
(overtwist_trim_bins,overtwist_trim_freqs) = overtwists_trim.histogram(:middle, :tp=> :min)
overtwist_trim_bin_titles = []
overtwist_trim_bins.each_with_index { |bin,i|
  overtwist_trim_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,overtwist_trim_bins[i+1]||overtwists_trim.max)
}
end
if (! stretches_trim.all_same?)
(stretch_trim_bins,stretch_trim_freqs) = stretches_trim.histogram(:middle, :tp=> :min)
stretch_trim_bin_titles = []
stretch_trim_bins.each_with_index { |bin,i|
  stretch_trim_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,stretch_trim_bins[i+1]||stretches_trim.max)
}
end

#
# Estimate persistence length
#
# Bending
ave_sq_curv = 0.0
curvatures_trim.each {|curv| ave_sq_curv += curv**2  }
ave_sq_curv = ave_sq_curv / curvatures_trim.size
pers_length_curv = Math.sqrt(2.0 / ave_sq_curv)
summary_text << sprintf("Estimated bending persistence length = %.3f A",pers_length_curv)
# Stretch
ave_sq_stretch = 0.0
stretches_trim.each {|stretch| ave_sq_stretch += stretch**2  }
ave_sq_stretch = ave_sq_stretch / stretches.size
pers_length_stretch = Math.sqrt(2.0 / ave_sq_stretch)
summary_text << sprintf("Estimated stretching persistence length = %.3f A",pers_length_stretch)
# Overtwist
ave_sq_over = 0.0
overtwists_trim.each {|over| ave_sq_over += over**2  }
ave_sq_over = ave_sq_over / overtwists.size
pers_length_over = Math.sqrt(2.0 / ave_sq_over)
summary_text << sprintf("Estimated overtwisting persistence length = %.3f A",pers_length_over)

#
# Sort filaments by descending curvature, overtwist, stretch
#
fils_curv = fils.sort{|x,y| y.curvature <=> x.curvature}
fils_twist = fils.sort{|x,y| y.overtwist <=> x.overtwist}
fils_stretch = fils.sort{|x,y| y.stretch <=> x.stretch}
curv_table = [["#","Filament #","Curvature\n(1/A)","Curvature\n(equivalent angle, deg.)","Score"]]
twist_table = [["#","Filament #","Overtwist\n(1/A)","Score"]]
stretch_table = [["#","Filament #","Stretch\n(1/A)","Score"]]
fils_curv.each_with_index {|fil,i| curv_table << [i+1,fil.num,fil.curvature,(fil.curvature*fil.length*pixel_size)/Math::PI*180.0,fil.score] }
fils_twist.each_with_index {|fil,i| twist_table << [i+1,fil.num,fil.overtwist,fil.score] }
fils_stretch.each_with_index {|fil,i| stretch_table << [i+1,fil.num,fil.stretch,fil.score] }

#
# Sort filaments by ascending score
#
fils_score = fils.sort{|x,y| x.score <=> y.score}
score_table = [["#","Filament #","Score"]]
fils_score.each_with_index{|fil,i| score_table << [i+1,fil.num,fil.score]}

#
# Sort filaments by descending score std dev
#
fils_score_std = fils.sort{|x,y| y.score_std <=> x.score_std}
score_std_table = [["#","Filament #","Length","Score","Sigma(score)","Log filename"]]
fils_score_std.each_with_index{|fil,i| score_std_table << [i+1,fil.num,fil.length,fil.score,fil.score_std,fil.log_filename]}
# Prepare histogram of score std dev
score_stds = fils.collect{|fil| fil.score_std}
if (! score_stds.all_same?) 
(score_std_bins,score_std_freqs) = score_stds.histogram(:scott, :tp=> :min)
score_std_bin_titles = []
score_std_bins.each_with_index { |bin,i|
  score_std_bin_titles << sprintf("\"%0.3f-%0.3f\"",bin,score_std_bins[i+1]||score_stds.max)
}
end

# Correlation between score std dev and filament length
require 'statsample'
fl = fils.collect{|fil| fil.length}
summary_text << sprintf("Correlation between score sigma along filament and filament length: %.7g",Statsample::Bivariate.pearson(score_stds.to_scale,fl.to_scale))



#
# FSC / FRC
#
log = FrealixOutput.new(d+"/frealix_output_finalise.log")
fsc = log.getFSCValues(:prefix_string => "FSC between half-dataset reconstructions")
frc = log.getFSCValues(:prefix_string => "FRC between central slices")
fsc = [1.0,0.0,0.0,0.0] if (!fsc)
frc = [1.0,0.0,0.0,0.0]	if (!frc)
shell = 0.upto(fsc.size-1).entries
threshold = []
shell.each_index { |i|
	shell[i] = i.to_f / shell.size.to_f / 2.0 / pixel_size
	threshold[i] = 0.143
}
# Find where the FSC crosses 0.143
fsc_below_threshold_at = []
fsc_below_threshold = []
fsc.each_index do |i|
	if (fsc[i].to_f <= 0.143)
		fsc_below_threshold[0]    = fsc[i-1]
		fsc_below_threshold_at[0] = shell[i-1]
		break
	end
	if (i == fsc.size-1)
	  fsc_below_threshold[0] = fsc[i]
	  fsc_below_threshold_at[0] = shell[i]
  end
end
resolution = 1/fsc_below_threshold_at[0]
summary_text << sprintf("FSC is above 0.143 threshold until %.2f A",resolution)

#
# Print out info to the terminal
#
summary_text.each {|t|
	$LOG.info() {t}
}

#
# Dump CSV with as much info as possible
#
CSV.open("#{d}/filaments.csv","wb",:headers=>:first_row) {|csv|
  csv << ["Filament #","Score","Score (no restraints)","Sigma(score)","Polarity dScore","Log filename",
          "Polarity flipped?","Length (A)", "Film #","Average Z (unit?)",
          "Arc/chord length ratio","Arc/chord length ratio (no FAL)",
          "Curvature","Overtwist","Stretch","Theta (mean)","Theta (SD)", "Psi (mean)", "Psi (SD)",
          "WP to WP distance average (A)", "WP to WP distance average (no FAL) (A)",
          "WP to WP distance std dev (A)", "WP to WP distance std dev (no FAL) (A)",
          "Arc length (A) (no FAL)","Change of direction (rad)"]
  fils.each {|fil|
    csv << [fil.num,fil.score,fil.score_no_restraints,fil.score_std,fil.polarity_dscore,fil.log_filename,
            fil.polarity_flipped,
            fil.length,fil.film_num,fil.average_z,
            fil.getArcChordLengthRatio,fil.getArcChordLengthRatioNoFAL,
            fil.curvature,fil.overtwist,fil.stretch,fil.thetaMeanSD[0],fil.thetaMeanSD[1],fil.psiMeanSD[0],fil.psiMeanSD[1],
            fil.wpdist_mean,fil.wpdist_mean_nofal,fil.wpdist_sd,fil.wpdist_sd_nofal,fil.lengthNoFal,fil.changeOfDirection]
  }
}

#
# PLOTTING (to multipage PDF file)
#
output_pdf_filename = "#{d}/summary.pdf"
output_pdf_filename_tmp_gnuplot = "#{d}/summary_tmp_gnuplot.pdf"
output_pdf_filename_tmp_text = "#{d}/summary_tmp_text.pdf"

Gnuplot.open do |gp|
	Gnuplot::Plot.new(gp) do |plot|
		# Some basic formatting
		plot.set "border linewidth 1.5"
		plot.style "line 1 lc rgb '#0060ad' lt 1 lw 2 pt 7 ps 1.5"   # --- blue
		plot.style "line 2 lc rgb '#dd181f' lt 1 lw 2 pt 5 ps 1.5"   # --- red
		plot.style "line 3 lc rgb 'grey30'"
		# Formatting for the histogram
		plot.style "data histograms"
		plot.set "key off"
		plot.style "fill solid 0.5 border"
		plot.xtics "nomirror rotate by -45"
		plot.terminal "pdf"	#"pngcairo enhanced font 'Verdana,10'"
		plot.output output_pdf_filename_tmp_gnuplot
		#
		# Plot histogram of run times
		#
		plot.title "Frealix run time (hours)"
		plot.ylabel "Number of jobs"
		plot.data << Gnuplot::DataSet.new([duration_bin_titles,duration_freqs]) do |ds|
			ds.using = "2:xtic(1)"
			ds.with  = "boxes ls 1"
			ds.title = "data 1"
		end
	end

	if (score_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of scores
		#
		plot.title ""
		plot.title "Filament scores"
		plot.ylabel "Number of filaments"
		plot.data << Gnuplot::DataSet.new([score_bin_titles,score_freqs]) do |ds|
			ds.using = "2:xtic(1)"
			ds.with  = "boxes ls 1"
			ds.title = "data 1"
		end
	end
	end
	
	if (score_std_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of score std dev along filaments
		#
		plot.title ""
		plot.title "Sigmas of score along filament"
		plot.ylabel "Number of filaments"
		plot.data << Gnuplot::DataSet.new([score_std_bin_titles,score_std_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
	end
	end
  end

	if (score_nr_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of scores (without mechanical restraints / penalties)
		#
		plot.title ""
		plot.title "Filament scores (without mechanical penalties)"
		plot.ylabel "Number of filaments"
		plot.data << Gnuplot::DataSet.new([score_nr_bin_titles,score_nr_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end
	


  if (scoring_step_sizes_freqs)
  Gnuplot::Plot.new(gp) do |plot|
    #
    # Plot histogram of changes in polarity scores
    #
    plot.title ""
    plot.title "Scoring function step size (A)"
    plot.ylabel "Number of filaments"
    plot.data << Gnuplot::DataSet.new([scoring_step_sizes_bin_titles,scoring_step_sizes_freqs]) do |ds|
    ds.using = "2:xtic(1)"
    ds.with  = "boxes ls 1"
    ds.title = "data 1"
    end
  end
  end
  
	if (polarity_dscore_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of changes in polarity scores
		#
		plot.title ""
		plot.title "Difference in score between polarities"
		plot.ylabel "Number of filaments"
		plot.data << Gnuplot::DataSet.new([polarity_dscore_bin_titles,polarity_dscore_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end

	if (dx_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of dx
		#
		plot.title ""
		plot.title "Waypoint changes in X"
		plot.ylabel "Number of waypoints"
		plot.data << Gnuplot::DataSet.new([dx_bin_titles,dx_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end

	if (dy_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of dy
		#
		plot.title ""
		plot.title "Waypoint changes in Y"
		plot.ylabel "Number of waypoints"
		plot.data << Gnuplot::DataSet.new([dy_bin_titles,dy_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end

	if (dz_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of dz
		#
		plot.title ""
		plot.title "Waypoint changes in Z"
		plot.ylabel "Number of waypoints"
		plot.data << Gnuplot::DataSet.new([dz_bin_titles,dz_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end

	if (dphi_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of dphi
		#
		plot.title ""
		plot.title "Waypoint changes in Phi"
		plot.ylabel "Number of waypoints"
		plot.data << Gnuplot::DataSet.new([dphi_bin_titles,dphi_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end

	if (dtheta_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of dtheta
		#
		plot.title ""
		plot.title "Waypoint changes in Theta"
		plot.ylabel "Number of waypoints"
		plot.data << Gnuplot::DataSet.new([dtheta_bin_titles,dtheta_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end

	if (dpsi_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of dpsi
		#
		plot.title   ""
		plot.title "Waypoint changes in Psi"
		plot.ylabel "Number of waypoints"
		plot.data << Gnuplot::DataSet.new([dpsi_bin_titles,dpsi_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end

	if (dsub_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of dsub
		#
		plot.title   ""
		plot.title "Waypoint changes in subunit_num"
		plot.ylabel "Number of waypoints"
		plot.data << Gnuplot::DataSet.new([dsub_bin_titles,dsub_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end
  
	if (theta_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of theta values
		#
		plot.title ""
		plot.title "Out-of-plane tilt at waypoints"
		plot.ylabel "Number of waypoints"
		plot.data << Gnuplot::DataSet.new([theta_bin_titles,theta_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end
	
	if (ac_ratio_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of curvatures
		#
		plot.title ""
		plot.title "Filament arc/chord length ratios"
		plot.ylabel "Number of filaments"
		plot.data << Gnuplot::DataSet.new([ac_ratio_bin_titles,ac_ratio_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end
	
	
	if (ac_ratio_nofal_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of curvatures
		#
		plot.title ""
		plot.title "Filament arc/chord length ratios (excluding FAL waypoints)"
		plot.ylabel "Number of filaments"
		plot.data << Gnuplot::DataSet.new([ac_ratio_nofal_bin_titles,ac_ratio_nofal_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end
	

	if (curvature_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of curvatures
		#
		plot.title ""
		plot.title "Filament curvatures (1/A)"
		plot.ylabel "Number of filaments"
		plot.data << Gnuplot::DataSet.new([curvature_bin_titles,curvature_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end

	if (overtwist_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of overtwists
		#
		plot.title ""
		plot.title "Filament overtwists (1/A)"
		plot.ylabel "Number of filaments"
		plot.data << Gnuplot::DataSet.new([overtwist_bin_titles,overtwist_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end

	if (stretch_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of stretches
		#
		plot.title ""
		plot.title "Filament stretches (1/A)"
		plot.ylabel "Number of filaments"
		plot.data << Gnuplot::DataSet.new([stretch_bin_titles,stretch_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end

	if (curvature_trim_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of curvatures (trimmed)
		#
		plot.title ""
		plot.title "Filament curvatures (1/A) (trimmed)"
		plot.ylabel "Number of filaments"
		plot.data << Gnuplot::DataSet.new([curvature_trim_bin_titles,curvature_trim_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end

	if (overtwist_trim_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of overtwists
		#
		plot.title ""
		plot.title "Filament overtwists (1/A) (trimmed)"
		plot.ylabel "Number of filaments"
		plot.data << Gnuplot::DataSet.new([overtwist_trim_bin_titles,overtwist_trim_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end

	if (stretch_trim_freqs)
	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot histogram of stretches
		#
		plot.title ""
		plot.title "Filament stretches (1/A) (trimmed)"
		plot.ylabel "Number of filaments"
		plot.data << Gnuplot::DataSet.new([stretch_trim_bin_titles,stretch_trim_freqs]) do |ds|
		ds.using = "2:xtic(1)"
		ds.with  = "boxes ls 1"
		ds.title = "data 1"
		end
	end
	end


	#
	# Scatter plot of score v length
	#
	if (lengths && scores)
	Gnuplot::Plot.new(gp) do |plot|
		plot.title ""
		plot.title "Filament score vs total length"
		plot.xlabel "Filament length (A)"
		plot.ylabel "Score"
		plot.data << Gnuplot::DataSet.new([lengths,scores]) do |ds|
		ds.using = "1:2"
		ds.with = "dots"
		end
	end
	end

	#
	# Scatter plot of score v curvature
	#
	if (curvatures && scores)
	Gnuplot::Plot.new(gp) do |plot|
		plot.title ""
		plot.title "Filament score vs curvature"
		plot.xlabel "Filament curvature (1/A)"
		plot.ylabel "Score"
		plot.data << Gnuplot::DataSet.new([curvatures,scores]) do |ds|
		ds.using = "1:2"
		ds.with = "dots"
		end
	end
	end

	#
	# Scatter plot of score (no mechanical penalties) v curvature
	#
	if (curvatures && scores_no_restraints)
	Gnuplot::Plot.new(gp) do |plot|
		plot.title ""
		plot.title "Filament score (no mechanical penalties) vs curvature"
		plot.xlabel "Filament curvature (1/A)"
		plot.ylabel "Score"
		plot.data << Gnuplot::DataSet.new([curvatures,scores_no_restraints]) do |ds|
		ds.using = "1:2"
		ds.with = "dots"
		end
	end
	end

	#
	# Scatter plot of score v film number
	#
	if (film_numbers && scores_no_restraints)
	Gnuplot::Plot.new(gp) do |plot|
		plot.title ""
		plot.title "Filament score (no mechanical penalties) vs film  number"
		plot.xlabel "Filament film number"
		plot.ylabel "Score"
		plot.data << Gnuplot::DataSet.new([film_numbers,scores_no_restraints]) do |ds|
		ds.using = "1:2"
		ds.with = "dots"
		end
	end
	end

	#
	# Scatter plot of score v defocus
	#
	if (fil_zs && scores_no_restraints)
	Gnuplot::Plot.new(gp) do |plot|
		plot.title ""
		plot.title "Filament score (no mechanical penalties) vs defocus"
		plot.xlabel "Filament defocus (A)"
		plot.ylabel "Score"
		plot.data << Gnuplot::DataSet.new([fil_zs,scores_no_restraints]) do |ds|
		ds.using = "1:2"
		ds.with = "dots"
		end
	end
	end

	#
	# Scatter plot of segment score v distance from nearest waypoint
	#
	if (segment_dist_to_wp && segment_scores)
	Gnuplot::Plot.new(gp) do |plot|
		plot.title ""
		plot.title "Segment score vs distance to nearest waypoint"
		plot.xlabel "Distance to nearest waypoint (A)"
		plot.ylabel "Segment score"
		plot.data << Gnuplot::DataSet.new([segment_dist_to_wp,segment_scores]) do |ds|
		ds.using = "1:2"
		ds.with = "dots"
		end
	end
	end


	Gnuplot::Plot.new(gp) do |plot|
		#
		# Plot FSC / FRC
		#
		plot.title ""
		plot.title "FSC between half-dataset reconstructions\\nand FRC between their central (XY) slices"
		plot.set "key"
		plot.set "yrange [0.:1.]"
		plot.set "xrange [0:#{shell.max}]"
		plot.xlabel "Spatial frequency (1/A)"
		plot.ylabel "Normalised cross-correlation"
		plot.data << Gnuplot::DataSet.new([shell,fsc,frc]) do |ds|
			ds.using = "1:2"
			ds.with  = "line ls 2 lw 3"
			ds.title = "FSC"
		end

		plot.data << Gnuplot::DataSet.new([shell,fsc,frc]) do |ds|
			ds.using = "1:3"
			ds.with  = "line ls 1 lw 2"
			ds.title = "FRC"
		end
		plot.data << Gnuplot::DataSet.new([[fsc_below_threshold_at],[fsc_below_threshold]]) do |ds|
			ds.using = "1:2"
			ds.with  = "impulses ls 3 lw 1"
			ds.title = "notitle"
		end
		plot.data << Gnuplot::DataSet.new([shell,threshold]) do |ds|
			ds.using = "1:2"
			ds.with  = "line ls 3 lw 1"
			ds.title = "notitle"
		end
	end

end

#
# Edit PDF file to append summary information
#
# Update (June 2012): Perhaps because I now use statsample, which requires prawn 0.8.4, the merging of PDF files
# using the template feature of prawn is broken. Therefore I "manually" merge the pdf files using pdftk.
#gem 'prawn', '>=0.12.0'
require "prawn"
require "prawn/layout"
Prawn::Document.generate(output_pdf_filename_tmp_text#,
#						:template=>output_pdf_filename_tmp_gnuplot,
						) do |pdf|
#	pdf.go_to_page(pdf.page_count)
#	pdf.start_new_page(:size=>"LETTER",
#						:layout=>:portrait)
	summary_text.each {|st|
		pdf.text st
	}
	# Warnings
	pdf.start_new_page
	pdf.text "Summary of warning messages (digits replaced by # symbols)"
	pdf.text warnings_summary
	# Tables of filaments sorted by..
	pdf.start_new_page
	pdf.text "Filaments sorted by curvature"
	pdf.table(curv_table)
	pdf.start_new_page
	pdf.text "Filaments sorted by overtwist"
	pdf.table(twist_table)
	pdf.start_new_page
	pdf.text "Filaments sorted by stretch"
	pdf.table(stretch_table)
  pdf.start_new_page
  pdf.text "Filaments sorted by score"
  pdf.table(score_table)
  pdf.start_new_page
  pdf.text "Filaments sorted by standard deviation of score along path"
  pdf.table(score_std_table)
	pdf.start_new_page
	pdf.text "Log file for each filament"
	pdf.table(log_table)
end

#
# Merge PDFs using pdftk
#
if (which("pdftk"))
  # First merge
  cmd = "pdftk #{output_pdf_filename_tmp_gnuplot} #{output_pdf_filename_tmp_text} cat output temp.pdf"
  pdftk = IO.popen(cmd,'r')
  output = pdftk.gets
  pdftk.close
  File.delete(output_pdf_filename_tmp_gnuplot,output_pdf_filename_tmp_text)
  # Then edit title
  f = File.new("temp.txt",mode="w")
  f.write("InfoKey: Title\nInfoValue: summary for #{d}\n")
  f.close
  cmd = "pdftk temp.pdf update_info temp.txt output #{output_pdf_filename}"
  pdftk = IO.popen(cmd,'r')
  output = pdftk.gets
  pdftk.close
  File.delete("temp.pdf","temp.txt")
else
  $LOG.fatal() {"Command not found: pdftk"}
end
