#!/usr/bin/env ruby
#
# Parse output from finalise3d frealix run and plot FSC
#
#  Copyright 2014 Howard Hughes Medical Institute.
#  All rights reserved.
#  Use is subject to Janelia Farm Research Campus Software Copyright 1.1
#  license terms ( http://license.janelia.org/license/jfrc_copyright_1_1.html )
#  

require_relative "rubylix.rb"
require_relative "FrealixOutput.rb"

require 'optparse'

options = {}
defaults = {}

optparse = OptionParser.new do |opts|
  opts.banner = "Usage: plot_fsc.rb [options] log_file output_png_file"

  # Help message
  opts.on_tail("-h","--help","Show this message") do
    puts opts
    exit
  end

  options[:pixel_size] = 1.0
  opts.on("-p PSIZE", "--pixel-size PSIZE", Float, "Pixel size in Angstroms") do |b|
    options[:pixel_size] = b
  end

end

optparse.parse!

# Check command-line arguments
if (ARGV.count < 2) 
	puts "Usage: plot_fsc.rb [options] log_file output_png_file"
	exit 1
end

# Parse arguments
log_file = ARGV[0]
output_file = ARGV[1]

# Initialise log file & find out the frealix mode that was run
log = FrealixOutput.new(log_file)
log.findMode()
puts log.mode

# Useful parsing functions
class String
  def is_i?
    !!(self =~ /^[-+]?[0-9]+$/)
  end
end
class String
  def is_f?
    !!(self =~ /^[+-]?[0-9]+[.]?[0-9]*$/)
  end
end


# There are three cases. We are parsing the output from either the CALC_FSC, the FILAMENT_REFINE or the finalise 3Ds mode
if (log.mode == "CALC_FSC")
  fsc_values = []
  fsc_values = log.getFSCValues()
  # Print FSC values to a temporary file 
  File.open("test.dat","w") do |f|
    fsc_values.each_with_index do |val,index|
      freq = index.to_f / fsc_values.count.to_f / 2.0 / options[:pixel_size].to_f
      f.write "#{freq} #{val} 0.143\n"
    end
  end

  # The GNUplot commands
  commands = %Q(
  set terminal pngcairo
  set output "#{output_file}"
  set title "FSC, parsed from #{log_file}"
  set xlabel "Frequency"
  set xrange [0:0.5/#{options[:pixel_size].to_f}]
  set ylabel "FSC"
  set yrange [0.:1.]
  set grid
  set key
  plot "test.dat" using 1:2 title "FSC" with line linetype 3 linewidth 3,"test.dat" using 1:3 title "0.143" with line linetype 2 linewidth 1
  exit
  )
elsif (log.mode == "FILAMENT_REFINE" || log.mode == "FINALISE_3DS")
  # Here are the strings we'll be looking for
  fsc_start = "FSC between half-dataset reconstructions follows"
  frc_start = "FRC between central slices of two half-dataset reconstructions follows"
  fsc_header = "Shell"
  
  fsc_values = log.getFSCValues(:prefix_string => fsc_start)
  frc_values = log.getFSCValues(:prefix_string => frc_start)

  # Print FSC values to a temporary file 
  File.open("test.dat","w") do |f|
    fsc_values.each_with_index do |val,index|
      f.write "#{index} #{val} #{frc_values[index]} 0.143\n"
    end
  end

  # The GNUplot commands
  commands = %Q(
  set terminal png
  set output "#{output_file}"
  set title "FSC, parsed from #{log_file}"
  set xlabel "Shell #"
  set xrange [0:#{fsc_values.count}]
  set ylabel "FSC"
  set yrange [0.:1.]
  set grid
  set key
  plot "test.dat" using 1:2 title "FSC" with line linetype 3 linewidth 3, "test.dat" using 1:3 title "FRC" with line linetype 4 linewidth 2, "test.dat" using 1:4 title "0.143" with line linetype 2 linewidth 1
  exit
  )
end

# Define pipe to GNUplot
def gnuplot(commands)
	IO.popen("gnuplot", "w") { |io| io.puts commands }
end

# Run GNUplot
gnuplot(commands)
