#!/usr/bin/env ruby
#  select_filaments.rb
#
#  Select and modify filaments from a waypoints.ali file
#
#  Copyright 2014 Howard Hughes Medical Institute.
#  All rights reserved.
#  Use is subject to Janelia Farm Research Campus Software Copyright 1.1
#  license terms ( http://license.janelia.org/license/jfrc_copyright_1_1.html )
#
#

require_relative "rubylix.rb"
require_relative "FrealixFilament.rb"
require_relative "FrealixAli.rb"

require 'optparse'

options = {}
defaults = {}

optparse = OptionParser.new do |opts|
  opts.banner = "Usage: select_filaments.rb [options] input output"

  # Help message
  opts.on_tail("-h","--help","Show this message") do
    puts opts
    exit
  end

  options[:score_below] = 999.9
  opts.on("--score-below SCORE", Float, "Select filaments with scores lower than this") do |b|
    options[:score_below] = b
  end

  options[:score_above] = -999.9
  opts.on("--score-above SCORE", Float, "Select filaments with scores higher than this") do |b|
    options[:score_above] = b
  end

  options[:twist_below] = 360.0
  opts.on("--twist-below ANGLE", Float, "Select waypoints with twists per subunit lower than this") do |b|
    options[:twist_below] = b
  end

  options[:twist_above] = -360.0
  opts.on("--twist-above TWIST", Float, "Select waypoints with twists per subunit higher than this") do |b|
    options[:twist_above] = b
  end

  options[:rise_below] = 999.9
  opts.on("--rise-below RISE", Float, "Select waypoints with rise per subunit lower than this") do |b|
    options[:rise_below] = b
  end

  options[:rise_above] = -999.9
  opts.on("--rise-above RISE", Float, "Select waypoints with rise per subunit higher than this") do |b|
    options[:rise_above] = b
  end

  options[:include_list_file] = nil
  opts.on("--include-list-file FILE","Select filaments with IDs listed in FILE") do |file|
    options[:include_list_file] = file
  end

  options[:exclude_list_file] = nil
  opts.on("--exclude-list-file FILE","Exclude filaments with IDs listed in FILE") do |file|
    options[:exclude_list_file] = file
  end

  options[:include_pos_list_file] = nil
  opts.on("--include-pos-list-file FILE","Select filaments with positions listed in FILE (starting at 1)") do |file|
    options[:include_pos_list_file] = file
  end

  options[:exclude_pos_list_file] = nil
  opts.on("--exclude-pos-list-file FILE","Exclude filaments with positions listed in FILE (starting at 1)") do |file|
    options[:exclude_pos_list_file] = file
  end

  options[:refine_to_true] = nil
  opts.on("--refine-to-true","Set the refine flags for selected filaments to T") do |r|
    options[:refine_to_true] = r
  end

  options[:refine_to_false] = nil
  opts.on("--refine-to-false","Set the refine flags for selected filaments to F") do |r|
    options[:refine_to_false] = r
  end

  options[:fal_refine_to_true] = nil
    opts.on("--fal-refine-to-true","Set the refine flags for first and last waypoints of selected filaments to T") do |r|
 	options[:fal_refine_to_true] = r
  end

  options[:fal_refine_to_false] = nil
    opts.on("--fal-refine-to-false","Set the refine flags for first and last waypoints of selected filaments to F") do |r|
    options[:fal_refine_to_false] = r
  end


  options[:subunit_num_to_zero] = nil
  opts.on("--subunit-num-to-zero","Set all subunit_num values to 0.0 for selected filaments") do |r|
    options[:subunit_num_to_zero] = r
  end

  options[:z_to_zero] = nil
  opts.on("--z-to-zero","Set all Z values to 0.0 for selected filaments") do |r|
    options[:z_to_zero] = r
  end

  options[:make_phi_monotonic_increasing] = nil
  opts.on("--make-phi-monotonic-increasing","Add multiples of 360 to phi values so that the phi function is monotonic increasing") do |r|
    options[:make_phi_monotonic_increasing] = r
  end

  options[:flip_polarity] = nil
  opts.on("--flip-polarity","Flip the polarity of selected filaments") do |r|
    options[:flip_polarity] = r
  end

  options[:x_shift] = nil
  opts.on("--x-shift SHIFT","Shift all waypoints in X by this (Angstroms) for selected filaments") do |r|
	  options[:x_shift] = r
  end

  options[:y_shift] = nil
  opts.on("--y-shift SHIFT","Shift all waypoints in Y by this (Angstroms) for selected filaments") do |r|
	  options[:y_shift] = r
  end

  options[:phi_shift] = nil
  opts.on("--phi-shift SHIFT","Shift all waypoints in phi by this angle (degrees) for selected filaments") do |r|
	  options[:phi_shift] = r
  end

  options[:axial_shift] = nil
  opts.on("--axial-shift SHIFT","Shift all waypoints along XY by this (Angstrom) along the helical axis (in 3D) for selected filaments") do |r|
    options[:axial_shift] = r
  end

  options[:xy_scale] = nil
  opts.on("--xy-scale SCALE","Scale all waypoints XY coordinates for selected filaments") do |r|
    options[:xy_scale] = r
  end

  options[:exclude_short_filaments] = nil
  opts.on("--exclude-short-filaments","Remove filaments with less than 3 waypoints") do |r|
    options[:exclude_short_filaments] = r
  end

  options[:keep_all_filaments] = nil
  opts.on("--keep-all-filaments","All filaments are kept. Those that were not selected will be left unchanged and copied over") do |r|
    options[:keep_all_filaments] = r
  end

end

optparse.parse!

@input_waypoints_filename = ARGV[0]
@output_waypoints_filename = ARGV[1]
@select_filaments_with_score_below = options[:score_below]
@select_filaments_with_score_above = options[:score_above]


#
# Get filaments from the waypoints file
#
ali_in = FrealixAli.new(:filename=>@input_waypoints_filename)
fils = FrealixFilament.filamentsFromWaypoints(ali_in.getWaypoints)
# Sort filaments by ID, so that their positions are meaningful
fils.sort!{|x,y| x.num <=> y.num}

$LOG.info() {"There are #{fils.size} filaments in input waypoints file"}


#
# If we were given a file containing a list, let's parse it
#
include_list = []
if (options[:include_list_file])
  str = File.read(options[:include_list_file])
  str.split("\n").each {|n| include_list << n.to_i}
  puts "Will select only filaments with IDs: #{include_list}"
else
  fils.each {|fil| include_list << fil.num}
end

exclude_list = []
if (options[:exclude_list_file])
  str = File.read(options[:exclude_list_file])
  str.split("\n").each {|n| exclude_list << n.to_i}
  puts "Will exclude filaments with IDs: #{exclude_list}"
end

#
# If we were given a file containing a list of positions, let's parse it
#
include_pos_list = []
if (options[:include_pos_list_file])
  str = File.read(options[:include_pos_list_file])
  str.split("\n").each {|n| include_pos_list << n.to_i}
  puts "Will select only filaments with positions: #{include_pos_list}"
else
  fils.each_with_index{|fil,i| include_pos_list << i+1}
end
#
exclude_pos_list = []
if (options[:exclude_pos_list_file])
  str = File.read(options[:exclude_pos_list_file])
  str.split("\n").each {|n| exclude_pos_list << n.to_i}
  puts "Will exclude filaments with positions: #{exclude_pos_list}"
end


#
# Select filaments
#
minimum_number_of_waypoints = 0
minimum_number_of_waypoints = 3 if (options[:exclude_short_filaments])
fils_out = []
fils_not_selected = []
fils.each_with_index { |fil,i|
	selected = (	fil.score < @select_filaments_with_score_below 	&&
			fil.score >= @select_filaments_with_score_above &&
			include_list.include?(fil.num) 			&&
			!exclude_list.include?(fil.num)			&&
			include_pos_list.include?(i+1) 			&&
			!exclude_pos_list.include?(i+1)     &&
			fil.wps.length >= minimum_number_of_waypoints)
  if (selected)
		fils_out << fil
	else
	  fils_not_selected << fil
	end
}

#
# Select waypoints
#
new_wps = []
fils_out.each {|fil|
	fil.wps.each{|wp|
    if (wp.twist && wp.rise)
  		if (wp.twist > options[:twist_above] &&
  			wp.twist < options[:twist_below] &&
  			wp.rise  > options[:rise_above] &&
  			wp.rise  < options[:rise_below])
  			#wp.twist = 118.62
  			#wp.rise  = 14.59
  			new_wps << wp
  			#puts wp.twist
  		end
    else
      new_wps << wp
    end
	}
}
fils_out = []
fils_out = FrealixFilament.filamentsFromWaypoints(new_wps)


#
# Zero values if requested
#
fils_out.each { |fil|
	if (options[:subunit_num_to_zero])
		fil.wps.each{|wp| wp.subunit_num = 0.0}
	end
  if (options[:z_to_zero])
    fil.wps.each{|wp| wp.z = 0.0}
  end
	if (options[:refine_to_true])
		fil.wps.each{|wp| wp.refine = "T"}
	end
	if (options[:refine_to_false])
		fil.wps.each{|wp| wp.refine = "F"}
	end
	if (options[:fal_refine_to_true])
    fil.wps[0].refine  = "T"
	  fil.wps[-1].refine = "T"
  end
  if (options[:fal_refine_to_false])
    fil.wps[0].refine  = "F"
	  fil.wps[-1].refine = "F"
	end
	if (options[:flip_polarity])
	  fil.flipPolarity!
	end
	# X and Y shifts
	fil.wps.each{|wp|
		wp.x   += options[:x_shift].to_f if (options[:x_shift])
		wp.y   += options[:y_shift].to_f if (options[:y_shift])
		wp.phi += options[:phi_shift].to_f if (options[:phi_shift])
    if (options[:axial_shift])
      dx = -Math.sin(wp.theta*Math.PI/180.0)*Math.cos(wp.psi*Math.PI/180.0)
      dy =  Math.sin(wp.theta*Math.PI/180.0)*Math.sin(wp.psi*Math.PI/180.0)
      wp.x += options[:axial_shift] * dx
      wp.y += options[:axial_shift] * dy
    end
	}
	fil.changePixelSize(1.0,options[:xy_scale].to_f) if (options[:xy_scale])
  #
	if (options[:make_phi_monotonic_increasing])
	  puts "make phi monotonic"
	  fil.makePhiMonotonicIncreasing!
	end
}

$LOG.info() {"#{fils_out.size} filaments were selected from the input ali file"}

#
# Maybe we want to add the non-selected / non-modified filaments back to the output array of filaments
#
if (options[:keep_all_filaments])
  fils_out.concat(fils_not_selected)
end

# Sort filaments by ID
fils_out.sort!{|x,y| x.num <=> y.num}

#
# Write out the waypoints file
#
FrealixFilament::filamentsToAliFile(fils_out,@output_waypoints_filename)

$LOG.info() {"#{fils_out.size} filaments are included in the output ali file"}
