#!/usr/bin/env ruby
#  split_long_filaments.rb
#
#  Split long filaments described in input waypoints.ali file
#
#  Copyright 2014 Howard Hughes Medical Institute.
#  All rights reserved.
#  Use is subject to Janelia Farm Research Campus Software Copyright 1.1
#  license terms ( http://license.janelia.org/license/jfrc_copyright_1_1.html )
#
#
#

require_relative "rubylix.rb"
require_relative "FrealixFilament.rb"
require_relative "FrealixAli.rb"

require 'optparse'

options = {}
defaults = {}

optparse = OptionParser.new do |opts|
  opts.banner = "Usage: split_long_filaments.rb [options] input output"

  # Help message
  opts.on_tail("-h","--help","Show this message") do
    puts opts
    exit
  end

  options[:max_num_wps] = 1000
  opts.on("-m","--max_num_wps NUM",Integer, "Filaments with more than NUM waypoints will be split up") do |b|
    options[:max_num_wps] = b
  end

  options[:ignore_fal_wps] = nil
  opts.on("-i","--ignore_fal_wps", "Processing ignores the first and last waypoints of each waypoint") do |b|
    options[:ignore_fal_wps] = b
  end

end

optparse.parse!

@input_waypoints_filename = ARGV[0]
@output_waypoints_filename = ARGV[1]
@max_num_wps = options[:max_num_wps]
@ignore_fal_wps = options[:ignore_fal_wps]


#
# Get filaments from the waypoints file
#
ali_in = FrealixAli.new(:filename=>@input_waypoints_filename)
fils = FrealixFilament.filamentsFromWaypoints(ali_in.getWaypoints)
# Sort filaments by ID, so that their positions are meaningful
fils.sort!{|x,y| x.num <=> y.num}


# Minimum number of waypoints in output filaments
@min_num_wps = 3
@min_num_wps = 5 if (@ignore_fal_wps)

# Loop over filaments to find the largest filament ID
max_fil_num = 0
fils.each { |fil|
  max_fil_num = fil.num if (fil.num > max_fil_num)
}
new_fil_num = 10**Math.log10(max_fil_num).ceil

#
# Loop over filaments and split the ones we need to split
#
fils_out = []
fils.each_with_index { |fil,i|
  #puts "Checking filament #{fil.num} with score #{fil.score}"
  if (fil.wps.count > @max_num_wps)
    # How many output filaments?
    num_output_fils = (fil.wps.count.to_f / (@max_num_wps.to_f - 2.0)).ceil
    $LOG.info() {"Filament #{fil.num} will be split into #{num_output_fils}, because it has #{fil.wps.count} waypoints"}
    # Loop over the output filaments
    first_wp = 0
    last_wp = 1
    1.upto(num_output_fils) { |o|
      first_wp = last_wp
      last_wp = first_wp + fil.wps.count/num_output_fils - 1
      if (@ignore_fal_wps) then
        first_wp -= 2
        last_wp += 1
      end
      first_wp = 1 if (o == 1)
      last_wp = fil.wps.count if (o == num_output_fils)
      new_fil_num += 1
      $LOG.info() {"First, last waypoints for output filament #{new_fil_num} : #{first_wp} - #{last_wp}"}
      if (last_wp-first_wp+1 < @min_num_wps) then
        $LOG.error() {"Output filament #{new_fil_num} will be too short!"}
      end
      fils_out << FrealixFilament.new(:wps=>fil.wps[first_wp..last_wp],:num=>new_fil_num)
    }
  else
    fils_out << fil
  end
}


#
# Write out the waypoints file
#
FrealixFilament::filamentsToAliFile(fils_out,@output_waypoints_filename)

$LOG.info() {"#{fils_out.size} filaments are included in the output ali file"}

puts " "
puts " Check carefully the output above for any ERRORs. The filament-splitting algorithm is not very robust."
puts " "
