!>  \brief  Routines to deal with coordinates (for single particles, for xovers, for points on images), including
!!          to read from / write to disk
!!
!!          To read a boxer file in:
!!              type(boxer_db)  ::  db
!!              db%read_from_disk(filename)
!!          The coordinates (center of boxes) are then accessed like this:
!!              db%point(i)%coo(1) = x
!!              db%point(i)%coo(2) = y
!!          These center coordinates can then be modified to will, and when a .box file needs to be written:
!!              box_size(1:3) = [128,128,1]
!!              call db%set_box_size(box_size)
!!              call db%write_to_disk(filename,[LUN])
!!
!!          To create a set of boxer coordinates from scratch:
!!              type(boxer_db)  ::  db
!!              call db%allocate(num_of_coordinates)
!!              do I=1,num_of_coordinates
!!                  db%point(i)%coo(1) = x(i)
!!                  db%point(i)%coo(2) = y(i)
!!              enddo
!!
!!
!!  The proper way to implement coordinates would be with an abstract class with descendents points, boxer_point, waypoint etc.
!!  Unfortunately, as of Feb-2014, the Intel compiler is buggy in that regard and so i have re-implemented all of these as separate types
!!  See the following URLs for a description of the problem and compiler updates:
!!  http://software.intel.com/en-us/forums/topic/506527
!!  http://software.intel.com/forums/topic/488776
!!  The intel problem report id is dpd200249796
!!
module coos

    use euleranglesmod
    use helices

    ! Some useful variables
    character(len=22),  private ::  boxer_file_format = '(2(I6,1x),2(i6,1x),i0)'


    !>  This is the basic data structure for representing a point in space
    type   ::  point
        real(kind=8)    ::  coo(3)  =   [0.0d0,0.0d0,0.0d0] !<  x, y, z coordinates
    end type point

    !>  \brief  a peak is a set of coordinates with an associated value (see module mod_peaks)
    !type ::  peak
    !    real(kind=8)    ::  coo(3)  =   [0.0d0,0.0d0,0.0d0] !<  x, y, z coordinates
    !    real                ::  rvalue      !<  value of the peak / peak height
    !    real                ::  nsigma      !<  number of times above the standard deviation
    !end type

    !>  \brief  Boxer (from the eman package) needs to know the box-size and to work out the top-left coordinates
    type :: boxer_point
        integer         ::  filament_number = 0             !< Boxer files can store coordinates for several filaments. For each boxer point, we remember which filament it belonged to.
        real(kind=8)    ::  coo(3)  =   [0.0d0,0.0d0,0.0d0] !<  x, y, z coordinates
        integer,private ::  box_dim(3)  = [1,1,1]               !<  box size (pixels)
        real,   private ::  coo_tl(3)   = [.0,.0,.0]            !<  top-left coordinates of the box
    end type

    !>  \brief  imod points file list coordinates in integer or float format, with or without extra information about the object
    !!          or contour the coordinates belong to
    type ::  imod_point
        real(kind=8)    ::  coo(3)  =   [0.0d0,0.0d0,0.0d0] !<  x, y, z coordinates
        integer         ::  contour_number
        integer         ::  object_number
        integer         ::  section_number
    end type

    !>  \brief  a waypoint is a point which a filament goes through
    !!
    !!  Waypoints lie exactly on the helical path, but do not necessarily coincide with helical subunits. In fact,
    !!  by convention, the only certainty is that the waypoint which has euler%phi = 0.0 will have an integer value
    !!  for SUBUNIT_NUM.
    type ::  waypoint
        real(kind=8)    ::  coo(3)  =   [0.0d0,0.0d0,0.0d0]                     !<  x, y, z coordinates
        type(euler)                     ::  euler                               !<  The orientation of the filament at that waypoint. Note that the angle "phi" stored here
                                                                                !!  is not necessarily actually an Euler angle. Rather it represents the twist of the helix at that point.
                                                                                !!  The two are not necessarily the same. By convention, we set phi=twist when theta>0.0 and
                                                                                !!  phi=twist+-180.0 when theta<0.0. Also by convention, we enforce theta > 0.0, to keep thing simple.
        type(helical_symmetry)          ::  helix                               !<  The helical symmetry of the filament at that waypoint
        logical                         ::  refine      =   .true.              !<  Whether this way point should be refined
        real(kind=8)                    ::  path_length =   0.0d0               !<  In pixels, how far along the path of the filament the waypoint is. the first waypoint has path_length = 0.0.
        real(kind=8)                    ::  subunit_num =   0.0d0               !<  Helical subunit number at this waypoint (if there are axial and/or perpendicular symmetries,
                                                                                !!  or several starts, symmetry mates are not counted)
        real                            ::  logp        =   0.0                 !<  Local log probability
        real                            ::  noise_sigma =   0.0                 !<  Standard deviation of noise
        real                            ::  occupancy   =   100.0               !<  Local occupancy
        real                            ::  score       =   0.0                 !<  In some cases, a local score can be associated with a waypoint
        real(kind=8)                    ::  dcoo(3)     =   [0.0d0,0.0d0,0.0d0] !<  the derivative of the space curve
        real                            ::  curvature   =   0.0                 !<  Local curvature (1/pixels)
        real                            ::  curvature_a =   0.0                 !<  Local curvature (1/pixels) along the first principal axis
        real                            ::  curvature_b =   0.0                 !<  Local curvature (1/pixels) along the second principal axis
        real                            ::  overtwist   =   0.0                 !<  Local overtwist (1/pixels)
        real                            ::  stretch     =   0.0                 !<  Local stretch (1/pixels)
        ! Movie processing
        real,   allocatable             ::  local_subframe_shift(:,:)           !<  The local subframe shift in dimension specified in first index (1=X,2=Y,3=Z) for movie frame # specified in second index
    end type

    type boxer_db
        integer                         ::  n           = 0             !<  number of boxes in the "database"
        type(boxer_point),  allocatable ::  point(:)                    !<  array of boxes
        logical                         ::  in_memory   = .false.       !<  whether memory is currently allocated to allocatable array(s)
    end type

    interface operator (+)
        module procedure waypoint_plus_waypoint
        module procedure point_plus_point
    end interface

    interface operator (-)
        module procedure waypoint_minus_waypoint
    end interface

    interface operator (*)
        module procedure waypoint_times_waypoint
    end interface

    interface operator (/)
        module procedure waypoint_divide_by_real
        module procedure point_divide_by_real
    end interface

    interface sqrt
        module procedure waypoint_sqrt
    end interface

    interface assignment(=)
       module procedure assign_waypoint_to_point
       module procedure assign_boxer_point_to_point
       module procedure assign_point_to_waypoint
       module procedure waypoint_assign
       module procedure assign_waypoint_to_boxer_point
    end interface

    interface interpolate
        module procedure waypoint_interpolate
        module procedure point_interpolate
    end interface

    contains


    !>  \brief  Unit tests for the coordinates types
    subroutine coos_unit_tests()
        implicit none
        call coos_unit_test_1()
    end subroutine coos_unit_tests

    !>  \brief  Test the assignment operation between various types of coordinates
    subroutine coos_unit_test_1()
        use runtime_parameters, only : terminate
        implicit none
        ! Private variables
        type(point),        allocatable ::  test_points(:)
        type(waypoint),     allocatable ::  test_waypoints(:)
        type(point)                     ::  test_point
        type(waypoint)                  ::  test_waypoint
        ! Start work

        allocate(test_waypoints(10))
        call waypoint_init(test_waypoints)
        test_waypoints(2)%COO(1) = 12.3
        call waypoint_init(test_waypoint)
        test_waypoint%coo(1) = 12.3

        allocate(test_points(size(test_waypoints)))
!#ifndef __INTEL_COMPILER
        test_points = test_waypoints
        test_point = test_waypoint
!#endif

        if (test_point%coo(1) .ne. 12.3) then
            write(*,'(a,F0.4)') '**ERROR(coos_unit_test_1): bad value, should be 12.3, is ', test_point%coo(1)
            call terminate('Failed COOS_UNIT_TEST_1')
        endif
        if (test_points(2)%coo(1) .ne. 12.3) then
            write(*,'(a,F0.4)') '**ERROR(coos_unit_test_1): bad value, should be 12.3, is ', test_points(2)%coo(1)
            call terminate('Failed COOS_UNIT_TEST_1')
        endif

        write(*,'(a)') '**info(coos_unit_test_1): passed'

    end subroutine coos_unit_test_1

    !>  \brief  Reset values for the waypoint
    pure elemental subroutine waypoint_init(self)
        use euleranglesmod, only : euler_init
        use helices, only : helical_symmetry_reset
        implicit none
        ! Arguments
        type(waypoint),                 intent(inout)   ::  self
        ! Start work
        self%coo = 0.0D0
        call euler_init(self%euler)
        call helical_symmetry_reset(self%helix)
        self%refine      =   .true.
        self%path_length =   0.0D0
        self%subunit_num =   0.0D0
        self%score       =   0.0
        self%logp        =   0.0
        self%noise_sigma =   0.0
        self%occupancy   =   100.0
        self%dcoo        =   [0.0D0,0.0D0,0.0D0]
        self%curvature   =   0.0
        self%curvature_a =   0.0
        self%curvature_b =   0.0
        self%overtwist   =   0.0
        self%stretch     =   0.0
        if (allocated(self%local_subframe_shift)) self%local_subframe_shift = 0.0e0
    end subroutine waypoint_init

    !>  \brief  Deallocate memory dynamically allocated to the waypoints
    subroutine waypoint_deallocate(self)
        type(waypoint), intent(inout)   ::  self
        if (allocated(self%local_subframe_shift)) deallocate(self%local_subframe_shift)
    end subroutine

    !>  \brief  Print nicely waypoint info. Argument must be an array of way points
    subroutine waypoints_print_info(self,phi_mod_360)
        implicit none
        ! Arguments
        type(waypoint),                 intent(in)  ::  self(:)
        logical,            optional,   intent(in)  ::  phi_mod_360
        ! Private variables
        integer ::  i
        type(waypoint)  ::  ave_std(2)
        ! Start work
        write(*,'(a5,1x,4(a11,1x),1(a12,1x),2(a10,1x),1(a10,1x),2(a10,1x),5(a11,1x),1(a7,1x),4(a6,1x))') &
                'WP#','X','Y','Z','Path_length','Phi','Theta','Psi','Subunit','Rise','Twist',   &
                'Curvature','Curv. (1)','Curv. (2)',                                    &
                'Overtwist','Stretch','Refine?','Score','LogP','Sigma','Occupancy'
        do i=1,size(self)
            write(*,'(i5,1x)',advance='NO') i
            call wp_print_info(self(i),phi_mod_360=phi_mod_360)
        enddo
        if (size(self) .ge. 1) then
            call waypoints_ave_std(self,ave_std)
            write(*,'(a5,1x)',advance='NO') 'Averg'
            call wp_print_info(ave_std(1))
            write(*,'(a5,1x)',advance='NO') 'StdDv'
            call wp_print_info(ave_std(2))
        endif
    end subroutine waypoints_print_info

    !>  \brief  Print nicely peak info
    !subroutine peak_print_info(self)
    !    implicit none
    !    ! Arguments
    !    type(peak),             intent(in)  ::  self
    !    ! Private variables
    !
    !    ! Start work
    !    write(*,'(3(F11.3,1x),2(G15.2,1x))') self%coo, self%rvalue, self%nsigma
    !end subroutine peak_print_info

    !>  \brief  Print nicely waypoint info.
    subroutine wp_print_info(self,phi_mod_360)
        implicit none
        ! argument
        type(waypoint),                 intent(in)  ::  self
        logical,            optional,   intent(in)  ::  phi_mod_360
        ! private variables
        real    ::  phi
        ! start work
        phi = self%euler%phi
        if (present(phi_mod_360)) then
            if (phi_mod_360) phi = mod(phi,360.)
        endif
        write(*,'(4(f11.3,1x),1(f12.2,1x),2(f10.2,1x),1(f10.2,1x),2(f10.3,1x),5(g11.3,1x),'//&
                '1(l7,1x),1(f6.3,1x),1(i6,1x),1(f6.4,1x),1(f6.2,1x))')&
                                self%coo(1:3),self%path_length,                                         &
                                phi,self%euler%theta,self%euler%psi,self%subunit_num,                   &
                                self%helix%rise,self%helix%twist,                                       &
                                self%curvature, self%curvature_a, self%curvature_b,                     &
                                self%overtwist, self%stretch,                                           &
                                self%refine,self%score,                                                 &
                                nint(self%logp),self%noise_sigma,self%occupancy
    end subroutine wp_print_info

    !>  \brief  Scale the coordinates of waypoints such that they can fit on a resampled micrograph (where the resampling was done by cropping the Fourier transform)
    subroutine waypoints_rescale_coos(self,original_micrograph_dimensions,new_micrograph_dimensions,scaling_factor)
        !
        type(waypoint),     intent(inout)   ::  self(:)                             !<  Waypoints with coordinates in pixels, corresponding to the micrograph before rescaling
        integer,            intent(in)      ::  original_micrograph_dimensions(3)   !<  The dimensions of the micrograph before rescaling
        integer,            intent(in)      ::  new_micrograph_dimensions(3)        !<  The dimensions of the micrograph after rescaling
        real,               intent(in)      ::  scaling_factor                      !<  Input coordinates will be multiplied by this factor
        ! private variables
        integer ::  i
        ! start work

        do i=1,size(self)
            call waypoint_rescale_coos(self(i),original_micrograph_dimensions,new_micrograph_dimensions,scaling_factor)
        enddo
    end subroutine waypoints_rescale_coos

    !>  \brief  Scale the coordinates of waypoints such that they can fit on a resampled micrograph (where the resampling was done by cropping the Fourier transform)
    subroutine waypoint_rescale_coos(self,original_micrograph_dimensions,new_micrograph_dimensions,scaling_factor)
        use Images, only : IndexOfCentralPixelGivenLogicalDimension
        !
        type(waypoint),     intent(inout)   ::  self                                !<  Waypoint with coordinates in pixels, corresponding to the micrograph before rescaling
        integer,            intent(in)      ::  original_micrograph_dimensions(3)   !<  The dimensions of the micrograph before rescaling
        integer,            intent(in)      ::  new_micrograph_dimensions(3)        !<  The dimensions of the micrograph after rescaling
        real,               intent(in)      ::  scaling_factor                      !<  Input coordinates will be multiplied by this factor
        ! private variables
        !real    ::  scaling_factor(3)
        integer ::  original_physical_address_of_center(3)
        integer ::  new_physical_address_of_center(3)
        integer ::  i
        ! start work

        ! The center of the micrograph before and after rescaling
        do i=1,2
            original_physical_address_of_center(i)  =   IndexOfCentralPixelGivenLogicalDimension(original_micrograph_dimensions(i))
            new_physical_address_of_center(i)       =   IndexOfCentralPixelGivenLogicalDimension(new_micrograph_dimensions(i))
        enddo
        original_physical_address_of_center(3) = 1
        new_physical_address_of_center(3) = 1

        ! Express coordinates relative to the center of the old micrograph
        do i=1,2
            self%coo(i) = self%coo(i) - original_physical_address_of_center(i)
        enddo

        ! Scale the coordinates
        do i=1,3
            self%coo(i) = self%coo(i) * scaling_factor
        enddo
        self%path_length = self%path_length * scaling_factor

        ! Scale the subframe alignments
        if (allocated(self%local_subframe_shift)) then
            self%local_subframe_shift = self%local_subframe_shift * scaling_factor
        endif

        ! Express coordinates relative to the top-left corner of the new micrograph
        do i=1,2
            self%coo(i) = self%coo(i) + new_physical_address_of_center(i)
        enddo

        ! Scale helix
        self%helix = helix_scale_distances(self%helix,scaling_factor)
    end subroutine waypoint_rescale_coos


    !>  \brief  Extend the array of waypoints by adding waypoints at the end of it
    subroutine waypoints_grow_array(self,increment)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(waypoint), allocatable,    intent(inout)   ::  self(:)
        integer,        optional,       intent(in)      ::  increment       !<  How many waypoints should be added to the end of the array. Defaults to 1.
        ! Private variables
        integer ::  old_size, new_size, iincrement
        type(waypoint) :: padding_wp
        type(waypoint), allocatable  ::  temp_waypoints(:)
        ! Start work

        ! Increment
        if (present(increment)) then
            iincrement = increment
        else
            iincrement = 1
        endif

        ! Work out old size
        if (allocated(self)) then
            old_size = size(self)
        else
            old_size = 0
        endif

        ! New size
        new_size = old_size + iincrement

        if (allocated(self)) then
            ! Reshape the array
            allocate(temp_waypoints(new_size))
            temp_waypoints = reshape(self,[new_size],PAD=[padding_wp])
            deallocate(self)
            allocate(self(new_size))
            self = temp_waypoints
        else
            allocate(self(new_size))
        endif


    end subroutine waypoints_grow_array

    !>  \brief  Extend the array of waypoints by adding waypoints at the end of it
    subroutine points_grow_array(self,increment)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(point), allocatable,    intent(inout)      ::  self(:)
        integer,        optional,       intent(in)      ::  increment       !<  How many waypoints should be added to the end of the array. Defaults to 1.
        ! Private variables
        integer ::  old_size, new_size, iincrement
        type(point) :: padding_wp
        type(point), allocatable  ::  temp_points(:)
        ! Start work

        ! Increment
        if (present(increment)) then
            iincrement = increment
        else
            iincrement = 1
        endif

        ! Work out old size
        if (allocated(self)) then
            old_size = size(self)
        else
            old_size = 0
        endif

        ! New size
        new_size = old_size + iincrement

        if (allocated(self)) then
            ! Reshape the array
            allocate(temp_points(new_size))
            temp_points = reshape(self,[new_size],PAD=[padding_wp])
            deallocate(self)
            allocate(self(new_size))
            self = temp_points
        else
            allocate(self(new_size))
        endif


    end subroutine points_grow_array

    !>  \brief  Extend the array of waypoints by adding waypoints at the end of it
    !!  \todo   Use intrinsic reshape function to grow the array of waypoints.
    function grow_path_old(old_path,increment) result(new_path)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(waypoint), allocatable,    intent(in)  ::  old_path(:)
        integer,        optional,       intent(in)  ::  increment       !<  How many waypoints should be added to the end of the array. Defaults to 1.
        ! Result
        type(waypoint), allocatable ::  new_path(:)
        ! Private variables
        integer ::  old_size, new_size, iincrement
        integer ::  ierr
        ! Start work

        ! Increment
        if (present(increment)) then
            iincrement = increment
        else
            iincrement = 1
        endif

        ! Work out old size
        if (allocated(old_path)) then
            old_size = size(old_path)
        else
            old_size = 0
        endif

        ! New size
        new_size = old_size + iincrement

        ! Allocate result array
        allocate(new_path(new_size),stat=IERR)
        if (ierr .ne. 0) then
            write(*,'(a)') '**ERROR(grow_path): memory allocation for new path failed with error ', ierr
            call terminate('Fatal error in GROW_PATH')
        endif

        ! Copy contents
        if (old_size .ge. 1) then
            new_path(1:old_size) = old_path
        endif

    end function grow_path_old

    !>  \brief  Given an array of waypoints, add Gaussian disorder
    subroutine waypoints_add_disorder(self,cumulative_disorder_std,non_cumulative_disorder_std)
        use runtime_parameters, only : terminate
        use fgsl
        implicit none
        ! Arguments
        type(waypoint),     allocatable,    intent(inout)   ::  self(:)                         !<  Array of waypoints. Modified in place.
        real,                               intent(in)      ::  cumulative_disorder_std(9)      !<  For each parameter (x,y,z,phi,theta,psi,subunit_num), the standard deviation of the cumulative disorder. In units of pixels, degrees and subunits.
        real,                               intent(in)      ::  non_cumulative_disorder_std(9)  !<  For each parameter, the standard deviation of the non-cumulative disorder
        ! Private variables
        real                ::  cumul_shift(9), non_cumul_shift(9)
        type(fgsl_rng)      ::  rng
        real(fgsl_double)   ::  cumul_std(9), non_cumul_std(9)
        real                ::  rn
        integer             ::  current_wp, current_dim
        integer(fgsl_long)  ::  seed
        ! Start work

        ! Check for errors in the inputs
        if (.not. allocated(self) .or. size(self) .lt. 1) then
            write(*,'(a,1x,i0,1x,l1)') '**ERROR(waypoints_add_disorder): waypoints array not allocated or too small: ', &
                                        allocated(self), size(self)
            call terminate('Fatal error in WAYPOINTS_ADD_DISORDER')
        endif
        if (any(cumulative_disorder_std .lt. 0.0) .or. any(non_cumulative_disorder_std .lt. 0.0)) then
            write(*,'(a,1x,2(7(F0.3,1x),1x))') '**ERROR(waypoints_add_disorder): standard deviations must be positive: ', &
                                               cumulative_disorder_std, non_cumulative_disorder_std
            call terminate('Fatal error in WAYPOINTS_ADD_DISORDER')
        endif

		! We assume (for now?) that it only makes sense to add disorder to waypoints that are going to be refined
		if (any(cumulative_disorder_std .ne. 0.0) .and. any(.not. self%refine)) then
			write(*,'(2a)') '**ERROR(waypoints_add_disorder): cannot add cumulative disorder unless all',    &
			                ' waypoints are flagged for refinement'
			call terminate('Fatal error in WAYPOINTS_ADD_DISORDER')
		endif

        ! If all std dev values are 0.0, we don't need to do anything
        if (all(cumulative_disorder_std .eq. 0.0) .and. all(non_cumulative_disorder_std .eq. 0.0)) then
            ! Nothing to do
        else
            ! Prepare random number generator
            cumul_std     = cumulative_disorder_std
            non_cumul_std = non_cumulative_disorder_std
            rng           = fgsl_rng_alloc(fgsl_rng_mt19937)
            call system_clock(COUNT=seed)
            call fgsl_rng_set(rng,seed)
            ! Initialise variables
            rn = 0.0
            ! Loop over waypoints
            do current_wp=1,size(self)
                cumul_shift = 0.0
                non_cumul_shift = 0.0
                ! Loop over the dimensions to prepare the shifts
                do current_dim=1,9
                    ! Cumulative
                    if (cumulative_disorder_std(current_dim) .gt. 0.0) then
                        rn = fgsl_ran_gaussian(rng,cumul_std(current_dim))
                        cumul_shift(current_dim) = rn
                    endif
                    ! Non-cumulative
                    if (non_cumulative_disorder_std(current_dim) .gt. 0.0) then
                        rn = fgsl_ran_gaussian(rng,non_cumul_std(current_dim))
                        non_cumul_shift(current_dim) = rn
                    endif
                enddo
                ! Apply the shifts
                call waypoints_apply_shift(self(current_wp:size(self)),cumul_shift)
				if (self(current_wp)%refine) call waypoint_apply_shift (self(current_wp)           ,non_cumul_shift)
            enddo
        endif

    end subroutine waypoints_add_disorder

    !>  \brief  Apply a shift to an array of waypoints
    subroutine waypoints_apply_shift(self,shift)
        implicit none
        ! Arguments
        type(waypoint),                 intent(inout)   ::  self(:)
        real,                           intent(in)      ::  shift(9)
        ! Private variables
        integer ::  i
        !
        do i=1,size(self)
            call waypoint_apply_shift(self(i),shift)
        enddo
    end subroutine waypoints_apply_shift

    !>  \brief  Apply a shift to a waypoint
    subroutine waypoint_apply_shift(self,shift)
        implicit none
        ! Arguments
        type(waypoint), intent(inout)   ::  self
        real,           intent(in)      ::  shift(9)    !<  Shift to be applied (x,y,z,phi,theta,psi,subunit_num)
        ! Private variables

        ! Start work
        self%coo(1:3)       =   shift(1:3) + self%coo(1:3)
        self%euler%phi      =   shift(4)   + self%euler%phi
        self%euler%theta    =   shift(5)   + self%euler%theta
        self%euler%psi      =   shift(6)   + self%euler%psi
        self%subunit_num    =   shift(7)   + self%subunit_num
        self%helix%rise     =   shift(8)   + self%helix%rise
        self%helix%twist    =   shift(9)   + self%helix%twist
    end subroutine waypoint_apply_shift

    !>  \brief  Apply a symmetry operator to a waypoint
    subroutine waypoint_apply_symmetry(self,axial_sym_mate,axial_sym_total,perp_sym_mate,perp_sym_total)
        use Globals
        implicit none
        ! Arguments
        type(waypoint), intent(inout)   ::  self
        integer,        intent(in)      ::  axial_sym_mate  !<  Which symmetry mate are we on?
        integer,        intent(in)      ::  axial_sym_total !<  What is the axial symmetry?
        integer,        intent(in)      ::  perp_sym_mate   !<  Which symmetry mate are we on?
        integer,        intent(in)      ::  perp_sym_total  !<  What is the perpendicular symmetry? For filaments with a side dyad, this should be 2. Else, should be 1.
        !
        if (axial_sym_mate .gt. 1 .or. perp_sym_mate .gt. 1) then
            call this_program%TerminateWithFatalError('waypoint_apply_symmetry','This is deprecated')
        endif
        self%euler%phi = self%euler%phi + real(axial_sym_mate-1,kind=8)*360.0d0/real(axial_sym_total,kind=8)
        if (perp_sym_total .eq. 2 .and. perp_sym_mate .eq. 2) then
            !self%euler%theta = 180.0 - self%euler%theta
            !self%euler%phi   = 180.0d0 + self%euler%phi
            !self%euler%psi   = 180.0d0 + self%euler%psi
            self%euler%theta    =   180.0d0 + self%euler%theta
            self%euler%phi      =   180.0d0 + self%euler%phi
        endif

    end subroutine waypoint_apply_symmetry

    !>  \brief  Remove duplicate waypoints from an array of waypoints
    subroutine waypoints_remove_duplicates(self)
        use float_comparisons, only : doubles_are_equal
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(waypoint),   allocatable,    intent(inout)   ::  self(:) !<  Array of points
        ! Private variables
        type(waypoint), allocatable ::  new_arr(:)
        integer ::  current_wp
        logical, parameter :: debug = .false.
        ! Start work
        allocate(new_arr(1))
        new_arr(1) = self(1)

        !
        if (size(self) .ge. 2) then
            do current_wp=2,size(self)
                if (        doubles_are_equal(self(current_wp)%coo(1),self(current_wp-1)%coo(1))  &
                    .and.   doubles_are_equal(self(current_wp)%coo(2),self(current_wp-1)%coo(2))  &
                    ) then
                    ! This waypoint is too close to the previous waypoint (most likely a duplicate), we will not include it
                    write(*,'(2(A,i0),a)') '**info(waypoints_remove_duplicates): waypoint ', &
                                            current_wp, ' of ', size(self), &
                                            ' removed because it was too close to the previous waypoint'
                else
                    call waypoints_grow_array(new_arr)
                    new_arr(size(new_arr)) = self(current_wp)
                endif

            enddo
        endif

        if (debug) call waypoints_print_info(new_arr)
        ! Done
        self = new_arr
    end subroutine waypoints_remove_duplicates

    !>  \brief  Remove duplicate waypoints from an array of waypoints
    subroutine points_remove_duplicates(self)
        use float_comparisons, only : doubles_are_equal
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(point),   allocatable,    intent(inout)   ::  self(:) !<  Array of points
        ! Private variables
        type(point), allocatable ::  new_arr(:)
        integer ::  current_wp
        logical, parameter :: debug = .false.
        ! Start work
        allocate(new_arr(1))
        new_arr(1) = self(1)

        !
        if (size(self) .ge. 2) then
            do current_wp=2,size(self)
                if (        doubles_are_equal(self(current_wp)%coo(1),self(current_wp-1)%coo(1))  &
                    .and.   doubles_are_equal(self(current_wp)%coo(2),self(current_wp-1)%coo(2))  &
                    ) then
                    ! This waypoint is too close to the previous waypoint (most likely a duplicate), we will not include it
                    write(*,'(2(A,i0),a)') '**info(points_remove_duplicates): waypoint ', &
                                            current_wp, ' of ', size(self), &
                                            ' removed because it was too close to the previous waypoint'
                else
                    call points_grow_array(new_arr)
                    new_arr(size(new_arr)) = self(current_wp)
                endif

            enddo
        endif

        ! Done
        self = new_arr
    end subroutine points_remove_duplicates

    !>  \brief  Add intermediate waypoints by linear interpolation
    subroutine waypoints_add_by_linear_interpolation(self)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(waypoint), allocatable,    intent(inout)   ::  self(:) !<  Array of points
        ! Private variables
        type(waypoint), allocatable ::  new_arr(:)
        integer ::  current_wp
        ! Start work
        allocate(new_arr(1))
        new_arr(1) = self(1)

        ! Loop over input waypoints
        if (size(self) .ge. 2) then
            do current_wp=2,size(self)
                ! Add interpolated waypoint
                call waypoints_grow_array(new_arr)
                new_arr(size(new_arr)) = interpolate(self(current_wp-1),self(current_wp))

                ! Add waypoint itself
                call waypoints_grow_array(new_arr)
                new_arr(size(new_arr)) = self(current_wp)
            enddo
        else
            write(*,'(a)') '**ERROR(waypoints_add_by_linear_interpolation): need at least 2 waypoints'
            call terminate('Fatal error in WAYPOINTS_ADD_BY_LINEAR_INTERPOLATION')
        endif
        ! Done
        deallocate(self)
        allocate(self(size(new_arr)))
        self = new_arr
    end subroutine waypoints_add_by_linear_interpolation


    !>  \brief  Add intermediate waypoints by linear interpolation
    subroutine points_add_by_linear_interpolation(self)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(point), allocatable,    intent(inout)   ::  self(:) !<  Array of points
        ! Private variables
        type(point), allocatable ::  new_arr(:)
        integer ::  current_wp
        ! Start work
        allocate(new_arr(1))
        new_arr(1) = self(1)

        ! Loop over input waypoints
        if (size(self) .ge. 2) then
            do current_wp=2,size(self)
                ! Add interpolated point
                call points_grow_array(new_arr)
                new_arr(size(new_arr)) = interpolate(self(current_wp-1),self(current_wp))

                ! Add point itself
                call points_grow_array(new_arr)
                new_arr(size(new_arr)) = self(current_wp)
            enddo
        else
            write(*,'(a)') '**ERROR(waypoints_add_by_linear_interpolation): need at least 2 waypoints'
            call terminate('Fatal error in WAYPOINTS_ADD_BY_LINEAR_INTERPOLATION')
        endif
        ! Done
        deallocate(self)
        allocate(self(size(new_arr)))
        self = new_arr
    end subroutine points_add_by_linear_interpolation

    type(waypoint) function waypoint_interpolate(wp1,wp2)
        implicit none
        ! Arguments
        type(waypoint), intent(in)  ::  wp1,wp2
        ! Start work
        waypoint_interpolate = (wp1+wp2)/2.0
    end function waypoint_interpolate

    type(point) function point_interpolate(p1,p2)
        implicit none
        ! Arguments
        type(point), intent(in)  :: p1,p2
        ! Start work
        point_interpolate = (p1+p2)/2.0
    end function point_interpolate


    !>  \brief  Given an array of waypoints, remove waypoints which are too close to their neighbors
    subroutine waypoints_remove_too_close(self)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(waypoint),     allocatable, intent(inout)  ::  self(:) !<  Array of waypoints
        ! Private variables
        real(kind=8)    ::  ds_prev
        type(waypoint), allocatable ::  new_arr(:)
        integer ::  current_wp
        logical,    parameter :: debug = .false.
        ! Start work

        if (.not. allocated(self)) then
            write(*,'(a)') '**ERROR(waypoints_remove_too_close): input waypoint array not allocated'
            call terminate('Fatal error in WAYPOINTS_REMOVE_TOO_CLOSE')
        endif

        if (debug) then
			write(*,'(a)') '**debug(waypoints_remove_too_close): on entry:'
			call waypoints_print_info(self)
		endif

        allocate(new_arr(1))
        new_arr(1) = self(1)

        ! This routine only makes sense if we have enough waypoints
		if (size(self) .gt. 3 .and. any(self%path_length .ne. 0.0D0)) then
			do current_wp=2,size(self)-1
                ! Arc length from previous
                ds_prev = self(current_wp)%path_length - new_arr(size(new_arr))%path_length
				if (ds_prev .lt. max(self(current_wp)%helix%rise,1.0D0)) then
					! The current waypoint is too close to the previous one.
                    ! Let's remove it
                    write(*,'(2(a,i0),a,f0.2,1x,f0.2,a)') '**info(waypoints_remove_too_close): waypoint ', &
                                            current_wp, ' of ', size(self), &
                                            ' removed because it was too close (subunit_num: ', &
                                            new_arr(size(new_arr))%subunit_num, &
                                            self(current_wp)%subunit_num, &
                                            ')to the previous waypoint'
                    !if (debug) then
                    !    write(*,'(a)') '**debug(waypoints_remove_too_close): waypoints info on entry:'
                    !    call waypoints_print_info(self)
                    !    call terminate('Debug in WAYPOINTS_REMOVE_TOO_CLOSE')
                    !endif
                else
                    call waypoints_grow_array(new_arr)
                    new_arr(size(new_arr)) = self(current_wp)
                endif
            enddo
            ! If the last waypoint is too close to the penultimate one, let's remove the penultimate one
            ds_prev = self(current_wp)%path_length - new_arr(size(new_arr))%path_length
            if (ds_prev .lt. max(self(current_wp)%helix%rise,1.0D0)) then
				!new_arr = self(1:size(self)-2)
                !call waypoints_grow_array(new_arr)
                new_arr(size(new_arr)) = self(size(self))
                write(*,'(2(A,i0),a)') '**info(waypoints_remove_too_close): waypoint ', &
                                        size(self)-1, ' of ', size(self), &
                                        ' removed because it was too close to the last waypoint'
            else
                call waypoints_grow_array(new_arr)
                new_arr(size(new_arr)) = self(size(self))
            endif
        else
            if (.not. any(self%path_length .ne. 0.0d0)) then
                if (debug) write(*,'(2a)') '**info(waypoints_remove_too_close): not doing anything. Path_length information ',  &
                            'not available'
            else
                if (debug) write(*,'(a)') '**info(waypoints_remove_too_close): not doing anything. Filament too short '
            endif

            if (allocated(new_arr)) deallocate(new_arr)
            allocate(new_arr(size(self)))
            new_arr = self
        endif

        if (debug) then
            write (*,'(a,i0)') '**debug(waypoints_remove_too_close): number of waypoints left in the end: ', size(new_arr)
        endif

		if (size(new_arr) .lt. 3 .and. debug) then
			write(*,'(a)') '**WARNING(waypoints_remove_too_close): output array is small, which may lead to problems later'
			write(*,'(a)') '**info(waypoints_remove_too_close): array on input was: '
			call waypoints_print_info(self)
			write(*,'(a)') '**info(waypoints_remove_too_close): array on output was: '
            call waypoints_print_info(new_arr)
            write(*,'(a)') '**inf(waypoints_remove_too_close): '
		endif

        ! Done
        deallocate(self)
        allocate(self(size(new_arr)))
        self = new_arr
    end subroutine waypoints_remove_too_close

    !>  \brief  Are all the waypoints to be refined contiguous?
    pure logical function waypoints_to_be_refined_are_contiguous(self)
        implicit none
        ! Arguments
        type(waypoint),     intent(in)  ::  self(:)
        ! Private variables
        integer :: current_wp
        logical :: in_segment_to_be_refined
        logical ::  left_segment_to_be_refined
        ! Start work
        in_segment_to_be_refined = .false.
        left_segment_to_be_refined = .false.
        waypoints_to_be_refined_are_contiguous = .true.
        do current_wp = 1,SIZE(self)
            if (in_segment_to_be_refined) then
                ! We have already encountered a segment to be refined
                if (.not. self(current_wp)%refine) then
                    ! We've left the segment to be refined
                    in_segment_to_be_refined = .false.
                    left_segment_to_be_refined = .true.
                endif
            else
                if (self(current_wp)%refine) then
                    if (left_segment_to_be_refined) then
                        ! We're entering a stretch of waypoints to be refined, but it's not the first time
                        waypoints_to_be_refined_are_contiguous = .false.
                    else
                        ! We are entering a stretch of waypoints to be refined for the first time
                        in_segment_to_be_refined = .true.
                    endif
                endif
            endif
        enddo

    end function waypoints_to_be_refined_are_contiguous


    !>  \brief  Check whether waypoint is a crossover
    pure logical function waypoint_is_crossover(self)
        implicit none
        ! Arguments
        type(waypoint),     intent(in)  ::  self
        ! Private variables
        real,   parameter   ::  phi_tolerance   =   0.001   !<  Tolerance in degrees
        ! Start work
        waypoint_is_crossover = abs(mod(self%euler%phi,180.0D0)) .le. phi_tolerance
    end function waypoint_is_crossover

    !>  \brief  Copy values to the waypoint
    subroutine value_to_waypoint(self,value,param_index)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(waypoint),             intent(inout)   ::  self
        real(kind=8),               intent(in)      ::  value
        integer ,                   intent(in)      ::  param_index !<  Parameter index. 1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta, 6.Psi, 7.Subunit_num, 8. Rise, 9. Twist
        ! Private variables

        ! Start work
        select case(param_index)
            case (1:3)
                self%coo(param_index) = value
            case (4)
                self%euler%phi = value
            case (5)
                self%euler%theta = value
            case (6)
                self%euler%psi = value
            case (7)
                self%subunit_num = value
            case (8)
                self%helix%rise = value
            case (9)
                self%helix%twist = value
            case default
                write(*,'(a,i0)') '**ERROR(value_to_waypoint): unknown parameter index: ', param_index
                call terminate('Fatal error in VALUE_TO_WAYPOINT')
        end select
    end subroutine value_to_waypoint

    !>  \brief  Get values from the waypoint
    subroutine value_from_waypoint(self,value,param_index)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(waypoint),             intent(in)  ::  self
        real(kind=8),               intent(out) ::  value
        integer ,                   intent(in)  ::  param_index !<  Parameter index. 1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta, 6.Psi, 7.Subunit_num, 8. Rise, 9. Twist
        ! Private variables

        ! Start work
        select case(param_index)
            case (1:3)
                value = self%coo(param_index)
            case (4)
                value = self%euler%phi
            case (5)
                value = self%euler%theta
            case (6)
                value = self%euler%psi
            case (7)
                value = self%subunit_num
            case (8)
                value = self%helix%rise
            case (9)
                value = self%helix%twist
            case default
                write(*,'(a,i0)') '**ERROR(fmc_value_to_wp): unknown parameter index: ', param_index
                call terminate('Fatal error in FMC_VALUE_TO_WP')
        end select
    end subroutine value_from_waypoint

    !>  \brief  Returns .TRUE. if it looks like the waypoint has just been initialised, and no changes have been made to it.
    pure logical function waypoint_has_default_values(self,ignore_helix)
        implicit none
        ! Arguments
        type(waypoint),             intent(in)  ::  self
        logical,        optional,   intent(in)  ::  ignore_helix        !<  Ignore the helical parameters.
        ! Private variables
        logical ::  iignore_helix
        ! Start work

        ! Will we be ignoring the helical parameters?
        if (present(ignore_helix)) then
            iignore_helix = ignore_helix
        else
            iignore_helix = .false.
        endif

        waypoint_has_default_values =    all  (self%coo         .eq. 0.0)   &
                                        .and. (self%euler%phi   .eq. 0.0)   &
                                        .and. (self%euler%theta .eq. 0.0)   &
                                        .and. (self%euler%psi   .eq. 0.0)   &
                                        .and. (self%path_length .eq. 0.0)   &
                                        .and. (self%subunit_num .eq. 0.0)   &
                                        .and. all(self%dcoo     .eq. 0.0)   &
                                        .and. (self%curvature   .eq. 0.0)   &
                                        .and. (self%curvature_a .eq. 0.0)   &
                                        .and. (self%curvature_b .eq. 0.0)   &
                                        .and. (self%overtwist   .eq. 0.0)   &
                                        .and. (self%stretch     .eq. 0.0)   &
                                        .and. (self%score       .eq. 0.0)   &
                                        .and. (self%logp        .eq. 0.0)   &
                                        .and. (self%noise_sigma .eq. 0.0)   &
                                        .and. (self%occupancy   .eq. 0.0)

        if (.not. iignore_helix) then
            waypoint_has_default_values =   waypoint_has_default_values         &
                                            .and. (self%helix%rise .eq. 0.0)    &
                                            .and. (self%helix%twist .eq. 0.0)
        endif

        if (allocated(self%local_subframe_shift)) then
            waypoint_has_default_values = waypoint_has_default_values .and. &
                                        all(self%local_subframe_shift .eq. 0.0e0)
        endif
    end function waypoint_has_default_values

    !>  \brief  Given an array of waypoints, return two waypoints: one with the average waypoint parameters, the other with the standard deviation
    subroutine waypoints_ave_std(self,wps_out)
        use runtime_parameters, only    :   terminate, safe_for_sqrt
        implicit none
        ! Arguments
        type(waypoint),     intent(in)  ::  self(:)
        type(waypoint),     intent(out) ::  wps_out(2)
        ! Private variables
        integer ::  i
        real    ::  real_size_self
        ! Start work
        call waypoint_init(wps_out)
        wps_out(1:2)%occupancy = 0.0
        do i=1,size(self)
            wps_out(1) = wps_out(1) + self(i)   ! sum
            wps_out(2) = wps_out(2) + (self(i) * self(i))   ! sum of squares
        enddo
        real_size_self = real(size(self))
        do i=1,2
            wps_out(i) = wps_out(i) / real_size_self ! Average and average of squares
        enddo
        wps_out(2) = wps_out(2) - (wps_out(1) * wps_out(1)) ! population variance
        ! Variance is always positive, but it could be that by rounding error, we end up with a small negative number rather than 0
        ! Maybe move these checks within the waypoint sqrt routine?
        call safe_for_sqrt(wps_out(2)%coo(1),       negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%coo(2),       negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%coo(3),       negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%dcoo(1),      negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%dcoo(2),      negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%dcoo(3),      negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%helix%rise,   negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%helix%twist,  negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%subunit_num,  negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%curvature,    negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%curvature_a,  negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%curvature_b,  negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%overtwist,    negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%stretch,      negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%euler%phi,    negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%euler%theta,  negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%euler%psi,    negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%score,        negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%logp,         negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%noise_sigma,  negative_to_zero=.true.)
        call safe_for_sqrt(wps_out(2)%occupancy,    negative_to_zero=.true.)
        if (allocated(wps_out(2)%local_subframe_shift)) then
            call safe_for_sqrt(wps_out(2)%local_subframe_shift,negative_to_zero=.true.)
        endif
        wps_out(2) = sqrt(wps_out(2)) ! std dev
    end subroutine waypoints_ave_std

    pure elemental subroutine assign_waypoint_to_point(lhs,rhs)
        implicit none
        type(point),    intent(inout)   ::  lhs
        type(waypoint), intent(in)      ::  rhs
        lhs%coo = rhs%coo
    end subroutine assign_waypoint_to_point

    pure elemental subroutine assign_boxer_point_to_point(lhs,rhs)
        implicit none
        type(point),    intent(inout)   ::  lhs
        type(boxer_point), intent(in)   ::  rhs
        lhs%coo = rhs%coo
    end subroutine assign_boxer_point_to_point

    pure elemental subroutine assign_point_to_waypoint(lhs,rhs)
        implicit none
        type(waypoint),     intent(inout)   ::  lhs
        type(point),        intent(in)      ::  rhs
        call waypoint_init(lhs)
        lhs%coo = rhs%coo
        if (.not. allocated(lhs%local_subframe_shift)) then
            allocate(lhs%local_subframe_shift(3,1))
            lhs%local_subframe_shift = 0.0
        endif
    end subroutine assign_point_to_waypoint


    pure elemental subroutine assign_waypoint_to_boxer_point(lhs,rhs)
        implicit none
        type(boxer_point),    intent(inout)   ::  lhs
        type(waypoint), intent(in)      ::  rhs
        lhs%coo = rhs%coo
    end subroutine assign_waypoint_to_boxer_point


    !>  \brief  Assignment operator overloading
    pure elemental subroutine waypoint_assign(wp_lhs,wp_rhs)
        implicit none
        ! Arguments
        type(waypoint), intent(inout)   ::  wp_lhs
        type(waypoint), intent(in)      ::  wp_rhs
        ! Private variables

        ! Start work
        wp_lhs%coo          =   wp_rhs%coo
        wp_lhs%dcoo         =   wp_rhs%dcoo
        wp_lhs%euler        =   wp_rhs%euler
        wp_lhs%helix        =   wp_rhs%helix
        wp_lhs%path_length  =   wp_rhs%path_length
        wp_lhs%subunit_num  =   wp_rhs%subunit_num
        wp_lhs%curvature    =   wp_rhs%curvature
        wp_lhs%curvature_a  =   wp_rhs%curvature_a
        wp_lhs%curvature_b  =   wp_rhs%curvature_b
        wp_lhs%overtwist    =   wp_rhs%overtwist
        wp_lhs%stretch      =   wp_rhs%stretch
        wp_lhs%refine       =   wp_rhs%refine
        wp_lhs%score        =   wp_rhs%score
        wp_lhs%logp         =   wp_rhs%logp
        wp_lhs%noise_sigma  =   wp_rhs%noise_sigma
        wp_lhs%occupancy    =   wp_rhs%occupancy
        if (allocated(wp_rhs%local_subframe_shift)) then
            wp_lhs%local_subframe_shift = wp_rhs%local_subframe_shift
        endif
    end subroutine waypoint_assign

    !>  \brief  Add two waypoint (for operator overloading)
    pure elemental function waypoint_plus_waypoint(wp1,wp2) result(wp_out)
        implicit none
        ! Arguments
        type(waypoint), intent(in)  ::  wp1, wp2
        ! Result
        type(waypoint)  ::  wp_out
        ! Start work
        wp_out%coo          =   wp1%coo + wp2%coo
        wp_out%dcoo         =   wp1%dcoo + wp2%dcoo
        wp_out%euler        =   wp1%euler + wp2%euler
        wp_out%helix        =   wp1%helix + wp2%helix
        wp_out%path_length  =   wp1%path_length + wp2%path_length
        wp_out%subunit_num  =   wp1%subunit_num + wp2%subunit_num
        wp_out%curvature    =   wp1%curvature + wp2%curvature
        wp_out%curvature_a  =   wp1%curvature_a + wp2%curvature_a
        wp_out%curvature_b  =   wp1%curvature_b + wp2%curvature_b
        wp_out%overtwist    =   wp1%overtwist + wp2%overtwist
        wp_out%stretch      =   wp1%stretch + wp2%stretch
        wp_out%score        =   wp1%score + wp2%score
        wp_out%logp         =   wp1%logp + wp2%logp
        wp_out%noise_sigma  =   wp1%noise_sigma + wp2%noise_sigma
        wp_out%occupancy    =   wp1%occupancy + wp2%occupancy
        if (allocated(wp1%local_subframe_shift) .and. allocated(wp2%local_subframe_shift)) then
            wp_out%local_subframe_shift = wp1%local_subframe_shift + wp2%local_subframe_shift
        else if (allocated(wp1%local_subframe_shift)) then
            wp_out%local_subframe_shift = wp1%local_subframe_shift
        else if (allocated(wp2%local_subframe_shift)) then
            wp_out%local_subframe_shift = wp2%local_subframe_shift
        endif
    end function waypoint_plus_waypoint

    type(point) pure elemental function point_plus_point(p1,p2) result(p_out)
        implicit none
        type(point),    intent(in)  ::  p1,p2
        p_out%coo = p1%coo + p2%coo
    end function point_plus_point

    !>  \brief  Subtract waypoints (for operator overloading)
    pure elemental function waypoint_minus_waypoint(wp1,wp2) result(wp_out)
        implicit none
        ! Arguments
        type(waypoint), intent(in)  ::  wp1, wp2
        ! Result
        type(waypoint)  ::  wp_out
        ! Start work
        wp_out%coo          =   wp1%coo - wp2%coo
        wp_out%dcoo         =   wp1%dcoo - wp2%dcoo
        wp_out%euler        =   wp1%euler - wp2%euler
        wp_out%helix        =   wp1%helix - wp2%helix
        wp_out%path_length  =   wp1%path_length - wp2%path_length
        wp_out%subunit_num  =   wp1%subunit_num - wp2%subunit_num
        wp_out%curvature    =   wp1%curvature   - wp2%curvature
        wp_out%curvature_a  =   wp1%curvature_a   - wp2%curvature_a
        wp_out%curvature_b  =   wp1%curvature_b   - wp2%curvature_b
        wp_out%overtwist    =   wp1%overtwist - wp2%overtwist
        wp_out%stretch      =   wp1%stretch - wp2%stretch
        wp_out%score        =   wp1%score - wp2%score
        wp_out%logp         =   wp1%logp - wp2%logp
        wp_out%noise_sigma  =   wp1%noise_sigma - wp2%noise_sigma
        wp_out%occupancy    =   wp1%occupancy - wp2%occupancy
        if (allocated(wp1%local_subframe_shift) .and. allocated(wp2%local_subframe_shift)) then
            wp_out%local_subframe_shift = wp1%local_subframe_shift - wp2%local_subframe_shift
        else if (allocated(wp1%local_subframe_shift)) then
            wp_out%local_subframe_shift = wp1%local_subframe_shift
        else if (allocated(wp2%local_subframe_shift)) then
            wp_out%local_subframe_shift = - wp2%local_subframe_shift
        endif
    end function waypoint_minus_waypoint

    !>  \brief  Multiply two waypoint (for operator overloading)
    pure elemental function waypoint_times_waypoint(wp1,wp2) result(wp_out)
        implicit none
        ! Arguments
        type(waypoint), intent(in)  ::  wp1, wp2
        ! Result
        type(waypoint)  ::  wp_out
        ! Start work
        wp_out%coo          =   wp1%coo * wp2%coo
        wp_out%dcoo         =   wp1%dcoo * wp2%dcoo
        wp_out%euler        =   wp1%euler * wp2%euler
        wp_out%helix        =   wp1%helix * wp2%helix
        wp_out%path_length  =   wp1%path_length * wp2%path_length
        wp_out%subunit_num  =   wp1%subunit_num * wp2%subunit_num
        wp_out%curvature    =   wp1%curvature   * wp2%curvature
        wp_out%curvature_a  =   wp1%curvature_a * wp2%curvature_a
        wp_out%curvature_b  =   wp1%curvature_b * wp2%curvature_b
        wp_out%overtwist    =   wp1%overtwist * wp2%overtwist
        wp_out%stretch      =   wp1%stretch * wp2%stretch
        wp_out%score        =   wp1%score * wp2%score
        wp_out%logp         =   wp1%logp * wp2%logp
        wp_out%noise_sigma  =   wp1%noise_sigma * wp2%noise_sigma
        wp_out%occupancy    =   wp1%occupancy * wp2%occupancy
        if (allocated(wp1%local_subframe_shift) .and. allocated(wp2%local_subframe_shift)) then
            wp_out%local_subframe_shift = wp1%local_subframe_shift + wp2%local_subframe_shift
        else if (allocated(wp1%local_subframe_shift)) then
            wp_out%local_subframe_shift = wp1%local_subframe_shift * 0.0e0
        else if (allocated(wp2%local_subframe_shift)) then
            wp_out%local_subframe_shift = wp2%local_subframe_shift * 0.0e0
        endif
    end function waypoint_times_waypoint

    !>  \brief  Return the square root of a waypoint
    pure elemental function waypoint_sqrt(wp) result(wp_sqrt)
        implicit none
        ! Arguments
        type(waypoint), intent(in)  ::  wp
        ! Result
        type(waypoint)  ::  wp_sqrt
        !
        integer :: i,j
        ! Start work
        wp_sqrt%coo         = sqrt(wp%coo)
        wp_sqrt%dcoo        = sqrt(wp%dcoo)
        wp_sqrt%euler       = sqrt(wp%euler)
        wp_sqrt%helix       = sqrt(wp%helix)
        wp_sqrt%path_length = sqrt(wp%path_length)
        wp_sqrt%subunit_num = sqrt(wp%subunit_num)
        wp_sqrt%curvature   = sqrt(wp%curvature)
        wp_sqrt%curvature_a = sqrt(wp%curvature_a)
        wp_sqrt%curvature_b = sqrt(wp%curvature_b)
        wp_sqrt%overtwist   = sqrt(wp%overtwist)
        wp_sqrt%stretch     = sqrt(wp%stretch)
        wp_sqrt%score       = sqrt(wp%score)
        wp_sqrt%logp        = sqrt(wp%logp)
        wp_sqrt%noise_sigma = sqrt(wp%noise_sigma)
        wp_sqrt%occupancy   = sqrt(wp%occupancy)
        if (allocated(wp%local_subframe_shift)) then
            !wp_sqrt%local_subframe_shift = sqrt(wp%local_subframe_shift)
            ! bug in gfortran 5.1 forces us to do this assignment manually
            allocate(wp_sqrt%local_subframe_shift(size(wp%local_subframe_shift,1),size(wp%local_subframe_shift,2)))
            do j=1,size(wp%local_subframe_shift,2)
                do i=1,size(wp%local_subframe_shift,1)
                    wp_sqrt%local_subframe_shift(i,j) = sqrt(wp%local_subframe_shift(i,j))
                enddo
            enddo
        endif
    end function waypoint_sqrt

    pure elemental function waypoint_divide_by_real(wp,val) result(wp_out)
        implicit none
        !
        type(waypoint), intent(in)  ::  wp
        real,           intent(in)  ::  val
        !
        type(waypoint)  ::  wp_out
        !
        wp_out%coo          =   wp%coo / val
        wp_out%dcoo         =   wp%dcoo / val
        wp_out%euler        =   wp%euler / val
        wp_out%helix        =   wp%helix / val
        wp_out%path_length  =   wp%path_length / val
        wp_out%subunit_num  =   wp%subunit_num  / val
        wp_out%curvature    =   wp%curvature / val
        wp_out%curvature_a  =   wp%curvature_a / val
        wp_out%curvature_b  =   wp%curvature_b / val
        wp_out%overtwist    =   wp%overtwist / val
        wp_out%stretch      =   wp%stretch / val
        wp_out%score        =   wp%score / val
        wp_out%logp         =   wp%logp / val
        wp_out%noise_sigma  =   wp%noise_sigma / val
        wp_out%occupancy    =   wp%occupancy / val
        if (allocated(wp%local_subframe_shift)) then
            wp_out%local_subframe_shift = wp%local_subframe_shift / val
        endif
    end function waypoint_divide_by_real

    pure elemental function point_divide_by_real(p,val) result(p_out)
        implicit none
        !
        type(point),    intent(in)  ::  p
        real,           intent(in)  ::  val
        !
        type(point)                 ::  p_out
        !
        p_out%coo          =   p%coo / val
    end function point_divide_by_real

    !>  \brief  Return the Euclidian distance between points 1 and 2
    pure real function waypoint_waypoint_distance(point1,point2)
        implicit none
        ! Arguments
        type(waypoint),   intent(in)  ::  point1
        type(waypoint),   intent(in)  ::  point2
        ! Private variables
        integer ::  i
        ! Start work
        waypoint_waypoint_distance = 0.0
        do i=1,3
            waypoint_waypoint_distance = waypoint_waypoint_distance + (point2%coo(i)-point1%coo(i))**2
        enddo
        waypoint_waypoint_distance = sqrt(waypoint_waypoint_distance)
    end function

    !>  \brief  Write an array of coordinates to disk in a csv format
    subroutine points_write_to_disk_csv(self,filename,lun)
        use runtime_parameters
        implicit none
        ! Arguments
        type(point),    allocatable,            intent(in)      ::  self(:)     !<  Array of point class objects
        character(len=*),                       intent(in)      ::  filename    !<  File name to write to
        integer,                    optional,   intent(in)      ::  lun         !<  file number to write to
        ! Private variables
        integer ::  file_num
        character(len=line_len) ::  csv_string
        integer ::  i
        ! Start work
        ! Did user specify LUN?
        if (present(lun)) then
            file_num = lun
        else
            file_num = available_lun()  ! Might be an idea to have a globally-defined number for temporary read-writes?
        endif
        ! Open file
        open(unit=file_num,file=filename,action='WRITE')

        ! Write lines
        do i=1,size(self)
            csv_string = point_to_csv(self(i))
            write(file_num,'(a)') csv_string
        enddo

        ! Close file
        close(unit=file_num)
        call lun_is_now_available(file_num)

    end subroutine points_write_to_disk_csv

    !>  \brief  Convert the coordinates of a point to a csv (comma-separated value) character string
    function point_to_csv(self) result(csv)
        use runtime_parameters, only: line_len
        implicit none
        ! Arguments
        type(point),        intent(in)  ::  self    !<  Point class object
        ! Result
        character(len=line_len) ::  csv !<  csv character string
        ! Private variables

        ! Start work
        write(csv,'(3(F15.6,a1))') self%coo(1), ',', self%coo(2), ',', self%coo(3)
    end function point_to_csv

    !>  \brief  Print out the boxer coordinates as they would appear in a .box file from eman/boxer
    subroutine print_boxer_point(bp,lun)
        implicit none
        ! Arguments
        type(boxer_point),  intent(inout)   ::  bp
        integer, optional,  intent(in)      ::  lun !<  file number to write to
        ! Private variables
        character(len=256)  ::  line
        ! Start work
        ! First, ensure the top left coordinates are updated with the current center coordinates and box size
        bp%coo_tl(1:3) = bp%coo(1:3) - bp%box_dim(1:3)/2
        ! Now do the actual printing
        write(line,boxer_file_format) nint(bp%coo_tl(1)), nint(bp%coo_tl(2)), bp%box_dim(1), bp%box_dim(2), -1
        if (present(lun)) then
            write(lun,'(a)') adjustl(trim(line))
        else
            write(*,'(a)') adjustl(trim(line))
        endif
    end subroutine print_boxer_point

    !>  \brief  Write a .box db file to disk
    subroutine boxer_db_write_to_disk(db,filename,lun,append)
        use runtime_parameters
        implicit none
        ! Arguments
        type(boxer_db),     intent(inout)   ::  db
        character(len=*),       intent(in)      ::  filename    !<  Filename to write to
        integer,    optional,   intent(in)      ::  lun         !<  Logical unit number for the file handle
        logical,    optional,   intent(in)      ::  append      !<  Whether to append to the file, rather than overwrite
        ! Private variables
        integer ::  file_num, i
        type(boxer_db)  ::  db_tmp
        logical ::  aappend
        ! Start work
        ! Did use specify LUN?
        if (present(lun)) then
            file_num = lun
        else
            file_num = available_lun()
        endif

        ! Will we be appending?
        if (present(append)) then
            aappend = append
        else
            aappend = .false.
        endif

        ! If we're appending, we first need to read from the file
        if (aappend) then
            call boxer_db_read_from_disk(db_tmp,filename)
        endif

        ! Open file
        open(unit=file_num,file=filename,action='WRITE')
        ! Write to file
        if (aappend) then
            do i=1,db_tmp%n
                call print_boxer_point(db_tmp%point(i),file_num)
            enddo
        endif
        do i=1,db%n
            !call db%point(i)%print_boxer(file_num)
            call print_boxer_point(db%point(i),file_num)
        enddo
        ! Close file
        close(unit=file_num)
        call lun_is_now_available(file_num)
    end subroutine boxer_db_write_to_disk

    !>  \brief  Set the box size for a set of boxer coordinate
    !!
    !!  The center coordinates are always assumed to be the most up-to-date
    subroutine boxer_db_set_box_size(db,box_size)
        implicit none
        ! Arguments
        type(boxer_db), intent(inout)   ::  db
        integer,            intent(in)      ::  box_size(3)
        ! Private variables
        integer ::  i
        ! Start work
        forall (i=1:db%n)
            db%point(i)%box_dim(1:3) = box_size(1:3)
        end forall
        call boxer_db_update_coo_tl(db)
    end subroutine boxer_db_set_box_size

    !>  \brief  Update the top-left coordinates of a boxer database. The center coordinates and box sizes are taken as correct
    !!  \todo   Double check arithmetic
    subroutine boxer_db_update_coo_tl(db)
        implicit none
        ! Arguments
        type(boxer_db),     intent(inout)   ::  db  !<  The boxer database (set of coordinates) to be updated
        ! Private variables
        integer ::  i
        ! Start work
        forall (i=1:db%n)
            db%point(i)%coo_tl(1:2) = int( db%point(i)%coo(1:2) - db%point(i)%box_dim(1:2)/2.)      ! Not sure this is exactly correct
        end forall
    end subroutine boxer_db_update_coo_tl

    !>  \brief  Bin or unbin a set of boxer coordinates
    subroutine boxer_db_bin(dbin,dbout,bin_factor,unbin)
        implicit none
        ! Arguments
        type(boxer_db),     intent(in)  ::  dbin        !<  Input set of boxer coordinates
        type(boxer_db),     intent(out) ::  dbout       !<  Output set of boxer coordinates
        integer,                intent(in)  ::  bin_factor  !<  Binning factor
        logical,    optional,   intent(in)  ::  unbin       !<  If .TRUE., unbinning happens. Otherwise (default), binning is done.
        ! Private variables
        logical ::  uunbin
        real    ::  bbin_factor
        integer ::  new_box_size
        ! Start work

        ! Are we unbinning?
        if (present(unbin)) then
            uunbin = unbin
        else
            uunbin = .false.
        endif

        ! Allocate memory for the output
        call boxer_db_allocate(dbout,dbin%n)

        ! The bin factor as a real number
        bbin_factor = real(bin_factor)

        ! Do the work
        if (uunbin) then
            print *, '**debug(boxer_db_bin): unbinning by ', bbin_factor
            dbout%point%coo(1)    = dbin%point%coo(1) * bbin_factor + bbin_factor/2.0
            dbout%point%coo(2)    = dbin%point%coo(2) * bbin_factor + bbin_factor/2.0
            new_box_size    =   max(nint(dbin%point(1)%box_dim(1) * bbin_factor), 32)
        else
            print *, '**debug(boxer_db_bin): binning by ', bbin_factor
            dbout%point%coo(1)    = dbin%point%coo(1) / bbin_factor
            dbout%point%coo(2)    = dbin%point%coo(2) / bbin_factor
            new_box_size    =   max(nint(dbin%point(1)%box_dim(1) / bbin_factor), 10)
        endif
        call boxer_db_set_box_size(dbout,[new_box_size,new_box_size,1])
        call boxer_db_update_coo_tl(dbout)
    end subroutine boxer_db_bin

    !>  \brief  Read a .box db file from disk (created by Boxer, for example) into a boxer_db object
    subroutine boxer_db_read_from_disk(db,filename)
        use parsers
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(boxer_db)      ,   intent(inout)   ::  db          !<  Data structure to read into
        character(len=*)    ,   intent(in)      ::  filename    !<  Filename to read from
        ! Private variables
        integer,    allocatable ::  itmp(:)
        real,       allocatable ::  tmp(:)
        integer ::  ierr, i
        integer :: filament_counter
        ! Start work
        ! Check whether file exists
        if (.not. file_exists(filename)) then
            write(*,'(a,a)') '**ERROR(boxer_db_read_from_disk): file does not exist: ', trim(filename)
            call terminate('Fatal error in BOXER_DB_READ_FROM_DISK')
        endif
        ! Allocate memory for the boxer_db object
        !call db%allocate(number_of_lines_in_file(filename))
        call boxer_db_allocate(db,number_of_lines_in_file(filename))
        ! Allocate memory for temporary arrays
        allocate(tmp(db%n),itmp(db%n),stat=IERR)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(boxer_db_read_from_disk): memory allocation for temporary failed with error ', ierr
            call terminate('Fatal error in BOXER_DB_READ_FROM_DISK')
        endif

        ! Grab the x, y topleft coordinate
        call get_reals_from_column(filename,1,tmp)
        db%point%coo_tl(1) = tmp
        call get_reals_from_column(filename,2,tmp)
        db%point%coo_tl(2) = tmp
        db%point%coo_tl(3) = 0.0
        ! Grab the box dimensions
        call get_ints_from_column(filename,3,itmp)
        db%point%box_dim(1) = itmp(1:db%n)
        call get_ints_from_column(filename,4,itmp)
        db%point%box_dim(2) = itmp(1:db%n)
        db%point%box_dim(3) = 1
        ! Grab the counter index, which tells us when we are at the beginning or end of a filament
        call get_ints_from_column(filename,5,itmp)
        filament_counter = 0
        do i=1,db%n
            if(itmp(i) .eq. -1) filament_counter = filament_counter + 1
            db%point(i)%filament_number = filament_counter
        enddo

        ! Work out the actual center coordinates
        forall(i=1:db%n)
            db%point(i)%coo(:) = db%point(i)%coo_tl(:) + db%point(i)%box_dim(:)/2
        end forall
    end subroutine boxer_db_read_from_disk

    !>  \brief  Allocate memory to a boxer_db object
    subroutine boxer_db_allocate(db,n)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(boxer_db), intent(inout)   ::  db
        integer,            intent(in)      ::  n
        ! Private variables
        integer ::  ierr
        ! Start work
        ! Deallocate if already allocated
        if (db%in_memory) then
            deallocate(db%point)
        endif
        ! Allocate
        db%n = n
        allocate(db%point(db%n),stat=IERR)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(boxer_db_allocate): memory allocation failed. ierr = ', ierr
            call terminate('Fatal error in BOXER_DB_ALLOCATE')
        endif
        db%in_memory = .true.
    end subroutine boxer_db_allocate

    subroutine imod_point_assign_to_point(point_lhs,imod_point_rhs)
        implicit none
        ! Arguments
        type(imod_point),   intent(in)      ::  imod_point_rhs
        type(point),        intent(inout)   ::  point_lhs
        !
        point_lhs%coo(1:2) = imod_point_rhs%coo(1:2)
    end subroutine imod_point_assign_to_point

    !>  \brief  Write a .box db file to disk
    subroutine imod_points_write_to_disk(points,filename,lun,append)
        use runtime_parameters
        implicit none
        ! Arguments
        type(imod_point),   allocatable,    intent(inout)   ::  points(:)
        character(len=*),                   intent(in)      ::  filename    !<  Filename to write to
        integer,            optional,       intent(in)      ::  lun         !<  Logical unit number for the file handle
        logical,            optional,       intent(in)      ::  append      !<  Whether to append to the file, rather than overwrite
        ! Private variables
        integer ::  file_num, i
        type(imod_point), allocatable  ::  points_tmp(:)
        logical ::  aappend
        ! Start work
        ! Did use specify LUN?
        if (present(lun)) then
            file_num = lun
        else
            file_num = available_lun()
        endif

        ! Will we be appending?
        aappend = .false.
        if (present(append)) aappend = append

        ! If we're appending, we first need to read from the file
        if (aappend) call imod_points_read_from_disk(points_tmp,filename)

        ! Open file
        open(unit=file_num,file=filename,action='WRITE')
        ! Write to file
        if (aappend) then
            do i=1,size(points_tmp)
                call imod_point_print_info(points(i),file_num)
            enddo
        endif
        do i=1,size(points)
            call imod_point_print_info(points(i),file_num)
        enddo
        ! Close file
        close(unit=file_num)
        call lun_is_now_available(file_num)
    end subroutine imod_points_write_to_disk

    !>  \brief  Print out the imod point coordinates as they would appear in a file convert by imod from a model file
    subroutine imod_point_print_info(ipoint,lun)
        implicit none
        ! Arguments
        type(imod_point),   intent(in)      ::  ipoint
        integer, optional,  intent(in)      ::  lun !<  file number to write to
        ! Private variables
        character(len=39)  ::  line
        ! Start work

        !
        write(line,'(i6,1x,2(F12.4,1x),i6)') ipoint%contour_number, ipoint%coo(1) - 1.0 , ipoint%coo(2) - 1.0, ipoint%section_number
        if (present(lun)) then
            write(lun,'(a)') line
        else
            write(*,'(a)') line
        endif
    end subroutine imod_point_print_info

    !>  Given a text file on disk, populate an array of imod points
    subroutine imod_points_read_from_disk(points,filename)
        use parsers, only : number_of_lines_in_file, get_ints_from_column, get_reals_from_column, file_exists
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(imod_point),   allocatable,    intent(inout)   ::  points(:)
        character(len=*),                   intent(in)      ::  filename
        ! Private variables
        integer,    allocatable ::  itmp(:)
        ! Start work

        ! Check whether file exists
        if (.not. file_exists(filename)) then
            write(*,'(a,a)') '**ERROR(imod_points_read_from_disk): file does not exist: ', trim(filename)
            call terminate('Fatal error in IMOD_POINTS_READ_FROM_DISK')
        endif

        ! Allocate memory for the points
        if (allocated(points)) deallocate(points)
        allocate(points(number_of_lines_in_file(filename)))

        !
        call get_ints_from_column(filename,1,itmp)
        points%contour_number = itmp
        call get_ints_from_column(filename,2,itmp)
        points%coo(1) = itmp + 1
        call get_ints_from_column(filename,3,itmp)
        points%coo(2) = itmp + 1
        call get_ints_from_column(filename,4,itmp)
        points%section_number = itmp
    end subroutine imod_points_read_from_disk
end module coos
