!>  \brief  ctf functions etc.
module ctfs

    use ContrastTransferFunctions

    implicit none

    !private

    public  :: ctf_parameters, ctf_is_not_equal

    ! Need to make these private or type-bound
    public  ::  akv_to_wl, ctf_parameters_init, eval_ctf
    public  ::  wl_to_akv, ctf_pix_to_ang, ctf_ang_to_pix
    public  ::  z_from_ctf, ctf_change_pixel_size

    type ctf_parameters
        !private
        logical ::  pix_units   =   .false. !< Whether ctf parameters are currently stored in metric or pixel units
        real    ::  cs          =   0.0     !< Spherical aberration (mm or pix)
        real    ::  wl          =   0.0     !< Wavelength (Angstrom or pix)
        real    ::  ampl_cont   =   0.0     !< Fraction of amplitude contrast (from 0. to 1., usually btw .07 and .14, see Mindell 2003)
        real    ::  dfmid1      =   0.0     !< Underfocus along first axis (Angstrom or pix) (positive values for underfocus; larger value = more underfocus)
        real    ::  dfmid2      =   0.0     !< Underfocus along second axis (Angstrom or pix)
        real    ::  angast      =   0.0     !< Azimuth of first axis. 0.0 means axis is at 3 o'clock. (degrees)
        contains
            procedure       ::  ctf_plus_ctf
            procedure       ::  ctf_minus_ctf
            procedure       ::  ctf_is_not_equal
            generic, public ::  operator(+) => ctf_plus_ctf
            generic, public ::  operator(-) => ctf_minus_ctf
            generic, public ::  operator(.ne.) => ctf_is_not_equal
    end type ctf_parameters


    interface ang_to_pix
        module procedure ctf_ang_to_pix
    end interface ang_to_pix

    interface pix_to_ang
        module procedure ctf_pix_to_ang
    end interface pix_to_ang

    interface operator (.eq.)
        module procedure is_equal
    end interface

    !interface operator (.ne.)
    !    module procedure is_not_equal
    !end interface

    !interface operator (-)
    !    module procedure ctf_minus_ctf
    !end interface

    !interface operator (+)
    !    module procedure ctf_plus_ctf
    !end interface

    interface operator (.lt.)
        module procedure ctf_lt_ctf
    end interface

    interface abs
        module procedure ctf_abs
    end interface abs

contains

    subroutine ConvertToModernCTFClass(self,new_ctf,pixel_size)
        type(ctf_parameters),           intent(inout)   ::  self
        type(ContrastTransferFunction), intent(inout)   ::  new_ctf
        real,                           intent(in)      ::  pixel_size !< Angstroms
        !
        call ctf_pix_to_ang(self,pixel_size)
        call new_ctf%Init(wl_to_akv(self%wl),self%cs,self%ampl_cont,self%dfmid1/10000.0,self%dfmid2/10000.0,self%angast, &
                           1.0/30.0,1.0/4.0,-100.0,pixel_size,0.0)
    end subroutine


    subroutine ctf_unit_tests()
        call ctf_unit_test_1()
        call ctf_unit_test_2()
    end subroutine ctf_unit_tests

    subroutine ctf_unit_test_1()
        use runtime_parameters, only : terminate
        implicit none
        !
        type(ctf_parameters)    ::  self
        real,   allocatable     ::  zeroes(:)
        integer                 ::  i
        real                    ::  current_ctf
        real,   parameter       ::  pixel_size = 2.4
        real,   parameter       ::  tol = 0.001
        !
        write(*,'(a)') '**info(ctf_unit_test_1): entering'
        call ctf_parameters_init(self,akv=200.0,cs=2.0,ampl_cont=0.07,dfmid1=3.2,dfmid2=3.2,angast=0.0)
        call ctf_print_info(self)
        call ctf_ang_to_pix(self,pixel_size=pixel_size)
        call ctf_zeroes(self,10,zeroes)
        do i=1,size(zeroes)
            current_ctf =  eval_ctf(self,box_size=1000,rad=1000.*zeroes(i),ang=0.0)
            print *, i, pixel_size/zeroes(i),current_ctf
            if (abs(current_ctf) .gt. tol) then
                call terminate('Failed unit test')
            endif
        enddo
        write(*,'(a)') '**info(ctf_unit_test_1): passed'
    end subroutine ctf_unit_test_1

    subroutine ctf_unit_test_2()
        use runtime_parameters, only : terminate
        implicit none
        !
        type(ctf_parameters)    ::  self
        real,   allocatable     ::  defs(:)
        integer                 ::  i
        real                    ::  current_ctf
        real,   parameter       ::  pixel_size = 1.1
        real,   parameter       ::  tol = 0.99
        real,   parameter       ::  target_spacing = 4.8
        integer,parameter       ::  box_size = 10000
        real                    ::  target_sf
        !
        write(*,'(a)') '**info(ctf_unit_test_2): entering'
        write(*,'(a,F0.3)') '**info(ctf_unit_test_2): target spacing (a) = ', target_spacing
        write(*,'(a,F0.3)') '**info(ctf_unit_test_2): pixel size (a) = ', pixel_size
        target_sf = pixel_size/target_spacing
        write(*,'(a,F0.3)') '**info(ctf_unit_test_2): target sf (1/pixels) = ', target_sf

        call ctf_parameters_init(self,akv=300.0,cs=2.0,ampl_cont=0.07,dfmid1=3.2,dfmid2=3.2,angast=0.0)
        call ctf_print_info(self)
        call ctf_ang_to_pix(self,pixel_size=pixel_size)
        call ctf_find_best_defs_for_sf(self,target_sf,20,defs)
        do i=1,size(defs)
            self%dfmid1 = defs(i)
            self%dfmid2 = defs(i)
            current_ctf =  eval_ctf(self,box_size=box_size,rad=box_size*target_sf,ang=0.0)
            print '(i6,F16.1,F10.2)', i, defs(i)*pixel_size,current_ctf
            if (abs(current_ctf) .lt. tol) then
                call terminate('Failed unit test')
            endif
        enddo
        write(*,'(a)') '**info(ctf_unit_test_2): passed'
    end subroutine ctf_unit_test_2

    !>  \brief  If ctf parameters are currently stored in pixels, change them to reflect a change in pixel size
    pure subroutine ctf_change_pixel_size(self,scaling_factor)
        type(ctf_parameters),   intent(inout)   ::  self
        real,                   intent(in)      ::  scaling_factor  !<  pixel values will be mulitplied by this factor
        !
        if (self%pix_units) then
            self%dfmid1 = self%dfmid1 * scaling_factor
            self%dfmid2 = self%dfmid2 * scaling_factor
            self%cs     = self%cs * scaling_factor
            self%wl     = self%wl * scaling_factor
        endif
    end subroutine ctf_change_pixel_size


    !>  \brief  Return an array with the ctf value at each radius
    subroutine ctf_to_curve(ctf,curve,num_points,pixel_size)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(ctf_parameters),                   intent(inout)   ::  ctf         !< ctf object specifying ctf parameters
        real,                   allocatable,    intent(inout)   ::  curve(:)
        integer,                                intent(in)      ::  num_points
        real,                   optional,       intent(in)      ::  pixel_size  !< Pixel size (required if the ctf is supplied in Angstroms)
        ! Result

        ! Private variables
        integer ::  ierr, i
        ! Start work

        ! Allocate output
        if (allocated(curve)) deallocate(curve)
        allocate(curve(num_points),stat=ierr)
        if (ierr .ne. 0) then
            call terminate('Memory allocation failed in CTF_TO_CURVE')
        endif

        ! If we need the pixel size, check the user gave it
        if ( .not. ctf%pix_units) then
            if (present(pixel_size)) then
                call ang_to_pix(ctf,pixel_size)
            else
                write(*,'(a)') '**ERROR(ctf_to_curve): ctf parameters in Angstrom but no pixel size given'
                call terminate('Fatal error in CTF_TO_CURVE')
            endif
        endif

        ! Loop over points in the curve
        do i=1,size(curve)
            curve(i) = eval_ctf(ctf,size(curve)*2,real(i-1),0.)
        enddo

    end subroutine ctf_to_curve

    !>  \brief  Generate an array of ctf objects from a numeric text file as produced by CTFFind >3.6.0
    subroutine ctf_parameters_from_numeric_text_file(self,filename,acceleration_voltage,spherical_aberration,amplitude_contrast)
        use numerictextfiles
        ! arguments
        type(ctf_parameters),   allocatable,    intent(inout)   ::  self(:)
        character(len=*),                       intent(in)      ::  filename
        real,                                   intent(in)      ::  acceleration_voltage    !<  kV
        real,                                   intent(in)      ::  spherical_aberration    !<  mm
        real,                                   intent(in)      ::  amplitude_contrast      !<  fraction (e.g. 0.07)
        ! private variables
        type(numerictextfile)   ::  my_file
        real                    ::  data_from_file(6)
        integer                 ::  current_data_line
        ! start work
        call my_file%init(filename,open_to_read)

        ! Allocate output
        if (allocated(self)) deallocate(self)
        allocate(self(my_file%number_of_data_lines))

        do current_data_line=1,my_file%number_of_data_lines
            call my_file%readnextdataline(data_from_file)
            call ctf_parameters_init(self(current_data_line),acceleration_voltage,spherical_aberration,amplitude_contrast, &
                                     data_from_file(2)/10000.0,data_from_file(3)/10000.0,data_from_file(4))
        enddo

    end subroutine ctf_parameters_from_numeric_text_file

    !>  \brief  Initialise a ctf_parameters object
    subroutine ctf_parameters_init(self,akv,cs,ampl_cont,dfmid1,dfmid2,angast)
        ! Arguments
        type(ctf_parameters),               intent(inout)   ::  self        !<  ctf parameters object
        real,                               intent(in)      ::  akv         !<  Acceleration voltage (kV)
        real,                               intent(in)      ::  cs          !<  Spherical aberration (mm)
        real,                               intent(in)      ::  ampl_cont   !<  Amplitude contrast (fraction, e.g. 0.07 for 7%)
        real,                               intent(in)      ::  dfmid1      !<  Defocus (on major axis if astigmatism) (um)
        real,                   optional,   intent(in)      ::  dfmid2      !<  Defocus on minor axis
        real,                   optional,   intent(in)      ::  angast      !<  Azimuth of major axis. 0.0 means axis is at 3 o'clock. (degrees)
        ! Private variables
        real,   parameter   ::  um_to_a =   10000.
        ! Start work

        ! We initialise the parameters not in pixel units
        self%pix_units  =   .false.
        self%wl         =   akv_to_wl(akv)
        self%cs         =   cs
        self%ampl_cont  =   ampl_cont
        self%dfmid1     =   dfmid1 * um_to_a
        if (present(dfmid2)) then
            self%dfmid2 =   dfmid2 * um_to_a
        else
            self%dfmid2 =   dfmid1 * um_to_a
        endif
        if (present(angast)) then
            self%angast =   angast
        else
            self%angast =   0.0
        endif
    end subroutine ctf_parameters_init

    !>  \brief Test whether two CTFs are different
    elemental pure logical function ctf_is_not_equal(ctf1,ctf2)
        implicit none
        ! Arguments
        class(ctf_parameters),  intent(in)  ::  ctf1
        type(ctf_parameters),   intent(in)  ::  ctf2
        ! Start work
        ctf_is_not_equal =  .not. is_equal(ctf1,ctf2)
    end function ctf_is_not_equal

    !>  \brief Test whether two CTFs are equal
    elemental pure logical function is_equal(ctf1,ctf2)
        implicit none
        ! Arguments
        type(ctf_parameters),   intent(in)  ::  ctf1
        type(ctf_parameters),   intent(in)  ::  ctf2
        ! Start work
        is_equal =  ctf1%cs         .eq.    ctf2%cs         .and.   &
                    ctf1%wl         .eq.    ctf2%wl         .and.   &
                    ctf1%ampl_cont  .eq.    ctf2%ampl_cont  .and.   &
                    ctf1%dfmid1     .eq.    ctf2%dfmid1     .and.   &
                    ctf1%dfmid2     .eq.    ctf2%dfmid2     .and.   &
                    ctf1%angast     .eq.    ctf2%angast
    end function is_equal

    !>  \brief  Return the absolute of a set of ctf parameters
    elemental pure function ctf_abs(ctf)
        implicit none
        ! Arguments
        type(ctf_parameters),   intent(in)  ::  ctf
        ! Result
        type(ctf_parameters)    ::  ctf_abs
        ! Start work
        ctf_abs%pix_units   =   ctf%pix_units
        ctf_abs%cs          =   abs(ctf%cs)
        ctf_abs%wl          =   abs(ctf%wl)
        ctf_abs%ampl_cont   =   abs(ctf%ampl_cont)
        ctf_abs%dfmid1      =   abs(ctf%dfmid1)
        ctf_abs%dfmid2      =   abs(ctf%dfmid2)
        ctf_abs%angast      =   abs(ctf%angast)
    end function ctf_abs

    !>  \brief  Test with ctf1 is less than ctf2
    elemental pure logical function ctf_lt_ctf(ctf1,ctf2) result(is_lt)
        implicit none
        ! Arguments
        type(ctf_parameters),   intent(in)  ::  ctf1
        type(ctf_parameters),   intent(in)  ::  ctf2
        ! Start work
        is_lt    =  ctf1%cs         .lt.    ctf2%cs         .and.   &
                    ctf1%wl         .lt.    ctf2%wl         .and.   &
                    ctf1%ampl_cont  .lt.    ctf2%ampl_cont  .and.   &
                    ctf1%dfmid1     .lt.    ctf2%dfmid1     .and.   &
                    ctf1%dfmid2     .lt.    ctf2%dfmid2     .and.   &
                    ctf1%angast     .lt.    ctf2%angast
    end function ctf_lt_ctf

    !>  \brief  Subtract one set of ctf parameters from another
    !!
    !!  Note: we assume, but do not check, that both sets of parameters are in the same units!
    elemental pure function ctf_minus_ctf(ctf1,ctf2)
        use euleranglesmod, only : angle_difference
        implicit none
        ! Arguments
        class(ctf_parameters),  intent(in)  ::  ctf1
        type(ctf_parameters),   intent(in)  ::  ctf2
        ! Result
        type(ctf_parameters)    ::  ctf_minus_ctf
        ! Start work
        ctf_minus_ctf%pix_units =   ctf1%pix_units
        ctf_minus_ctf%cs        =   ctf1%cs - ctf2%cs
        ctf_minus_ctf%wl        =   ctf1%wl - ctf2%wl
        ctf_minus_ctf%ampl_cont =   ctf1%ampl_cont - ctf2%ampl_cont
        ctf_minus_ctf%dfmid1    =   ctf1%dfmid1 - ctf2%dfmid1
        ctf_minus_ctf%dfmid2    =   ctf1%dfmid2 - ctf2%dfmid2
        ctf_minus_ctf%angast    =   angle_difference(dble(ctf1%angast),dble(ctf2%angast))
    end function ctf_minus_ctf

    !>  \brief  Add one set of ctf parameters to another
    !!
    !!  Note: we assume, but do not check, that both sets of parameters are in the same units!
    elemental pure function ctf_plus_ctf(ctf1,ctf2)
        use euleranglesmod, only : angle_btwn_0_and_360
        implicit none
        ! Arguments
        class(ctf_parameters),  intent(in)  ::  ctf1
        type(ctf_parameters),   intent(in)  ::  ctf2
        ! Result
        type(ctf_parameters)    ::  ctf_plus_ctf
        ! Start work
        ctf_plus_ctf%pix_units =   ctf1%pix_units
        ctf_plus_ctf%cs        =   ctf1%cs + ctf2%cs
        ctf_plus_ctf%wl        =   ctf1%wl + ctf2%wl
        ctf_plus_ctf%ampl_cont =   ctf1%ampl_cont + ctf2%ampl_cont
        ctf_plus_ctf%dfmid1    =   ctf1%dfmid1 + ctf2%dfmid1
        ctf_plus_ctf%dfmid2    =   ctf1%dfmid2 + ctf2%dfmid2
        ctf_plus_ctf%angast    =   angle_btwn_0_and_360(dble(ctf1%angast) + dble(ctf2%angast))
    end function ctf_plus_ctf

    !>  \brief  Print out information describing the ctf
    subroutine ctf_print_info(self)
        implicit none
        ! Arguments
        type(ctf_parameters),   intent(in)  ::  self    !<  ctf object
        ! Private variables
        character(len=2),       parameter   ::  text_width  =   '25'
        ! Start work
        write(*,'(A'//text_width//',1x,l1)')        'Pixel units:',         self%pix_units
        write(*,'(A'//text_width//',1x,F12.2)')     'Cs:',                  self%cs
        write(*,'(A'//text_width//',1x,ES12.4)')     'Wavelength:',         self%wl
        write(*,'(A'//text_width//',1x,F0.2)')      'Amplitude contrast:',  self%ampl_cont
        write(*,'(A'//text_width//',1x,F0.2)')      'Defocus 1:',           self%dfmid1
        write(*,'(A'//text_width//',1x,F0.2)')      'Defocus 2:',           self%dfmid2
        write(*,'(A'//text_width//',1x,F0.2)')      'Astigmatism azimuth:', self%angast
    end subroutine ctf_print_info

    !>  \brief  Modify the ctf object so that its average defocus is specified by z
    subroutine ctf_at_z(ctf,z,pixel_size)
        use runtime_parameters, only : terminate, z_axis_direction
        implicit none
        ! Arguments
        type(ctf_parameters),               intent(inout)   ::  ctf     !<  ctf object to be modified
        real(kind=8),                       intent(in)      ::  z       !<  Z-height (in pixels)
        real,                   optional,   intent(in)      ::  pixel_size  !<  Pixel size (in Angstroms). Only needed if the ctf is not supplied in pixels.
        ! Private variables
        real    ::  astig1, astig2, av_def
        ! Start work

        ! Make sure the ctf is in pixel units
        if (.not. ctf%pix_units) then
            if (present(pixel_size)) then
                call ctf_ang_to_pix(ctf,pixel_size)
            else
                write(*,'(a)') '**ERROR(ctf_at_z): ctf not supplied in pixel units and pixel size not given'
                call terminate('Fatal error in CTF_AT_Z')
            endif
        endif

        !
        av_def = (ctf%dfmid1+ctf%dfmid2)/2.
        astig1 = ctf%dfmid1 - av_def
        astig2 = ctf%dfmid2 - av_def

        ! z_axis_direction defines whether the z axis goes up or down the microscope column. See comments in runtime_parmaters.f90 for more details.
        ctf%dfmid1 = z_axis_direction * z + astig1
        ctf%dfmid2 = z_axis_direction * z + astig2
    end subroutine ctf_at_z

    !>  \brief  Return the z coordinate of a point, given its ctf information
    real function z_from_ctf(ctf)
        use runtime_parameters, only : terminate, z_axis_direction
        implicit none
        ! Arguments
        type(ctf_parameters),   intent(in)  ::  ctf
        ! Private variables

        ! Start work

        ! Make sure the ctf is in pixel units
        if (.not. ctf%pix_units) then
            write(*,'(a)') '**ERROR(z_from_ctf): ctf not supplied in pixel units'
            call terminate('Fatal error in Z_FROM_CTF')
        endif

        z_from_ctf = z_axis_direction * (ctf%dfmid1+ctf%dfmid2)/2.
    end function z_from_ctf

    !>  \brief Convert ctf from Angstrom & degrees to pixels & radians
    subroutine ctf_ang_to_pix(ctf1,pixel_size)
        !use runtime_parameters, only: terminate
        use rotation_matrices, only : radians
        implicit none
        type(ctf_parameters),   intent(inout)   ::  ctf1            !< ctf object to convert
        real,                   intent(in)      ::  pixel_size  !< Pixel size (Angstrom)
        ! Private variables

        ! Start work
        if (ctf1%pix_units) then
            !write(*,'(a)') '**ERROR(ctf_ang_to_pix): trying to convert ctf parameters already converted'
            !call terminate('Fatal error in CTF_ANG_TO_PIX')
        else
            ! Do actual conversions
            ctf1%cs     =   ctf1%cs * 10.0**7 / pixel_size      ! mm to a to pix
            ctf1%wl     =   ctf1%wl / pixel_size                ! a to pix
            ctf1%dfmid1 =   ctf1%dfmid1 / pixel_size            ! a to pix
            ctf1%dfmid2 =   ctf1%dfmid2 / pixel_size            ! a to pix
            ctf1%angast =   radians(ctf1%angast)            ! deg to rad
            ctf1%pix_units = .true.
        endif
    end subroutine ctf_ang_to_pix

    !>  \brief Convert ctf from pixels & radians to Angstrom & degrees
    subroutine ctf_pix_to_ang(ctf1,pixel_size)
        use runtime_parameters, only: terminate
        use rotation_matrices, only : degrees
        implicit none
        type(ctf_parameters),   intent(inout)   ::  ctf1            !< ctf object to convert
        real,       intent(in)      ::  pixel_size  !< Pixel size (Angstrom)
        ! Private variables

        ! Start work
        if (ctf1%pix_units) then
            ! Do actual conversions
            ctf1%cs     =   ctf1%cs / 10.0e7 * pixel_size   ! pix to mm
            ctf1%wl     =   ctf1%wl * pixel_size                ! pix to a
            ctf1%dfmid1 =   ctf1%dfmid1 * pixel_size            ! pix to a
            ctf1%dfmid2 =   ctf1%dfmid2 * pixel_size            ! pix to a
            ctf1%angast =   degrees(ctf1%angast)            ! rad to deg
            ctf1%pix_units = .false.
        endif
    end subroutine ctf_pix_to_ang


    !>  \brief  Find the defocus values for which the ctf is maximal at the given target spatial frequency
    subroutine ctf_find_best_defs_for_sf(self,target_sf,number_of_defs,defs)
        implicit none
        ! Arguments
        type(ctf_parameters),   intent(in)      ::  self                !<  ctf parameters (defocus will be ignored)
        real,                   intent(in)      ::  target_sf           !<  Target spatial frequency (1/pixels)
        integer,                intent(in)      ::  number_of_defs      !<  Number of defocus values you want
        real,   allocatable,    intent(inout)   ::  defs(:)             !<  Defocus values (pixels)
        ! Private variables
        real,   allocatable                     :: defs_pos(:), defs_neg(:)
        integer ::  i,j
        ! Start work

        ! Allocate memory for results
        if (allocated(defs)) deallocate(defs)
        allocate(defs(number_of_defs))
        allocate(defs_pos(number_of_defs),defs_neg(number_of_defs))

        ! Find the position of the first few times the ctf gets to 1.0 and -1.0
        call ctf_solve_for_arg(self%ampl_cont,defs_pos,ctf_value= 1.0)
        call ctf_solve_for_arg(self%ampl_cont,defs_neg,ctf_value=-1.0)

        ! Convert the arguments to defocus values
        call ctf_df_from_arg(self%cs,self%wl,defs_neg(1),target_sf,defs(1))
        do i=2,number_of_defs-2,2
            j = (i-1)+1
            call ctf_df_from_arg(self%cs,self%wl,defs_pos(i+1),target_sf,defs(j))
            j = (i-1)+2
            call ctf_df_from_arg(self%cs,self%wl,defs_neg(i),target_sf,defs(j))
        enddo
        call ctf_df_from_arg(self%cs,self%wl,defs_pos(number_of_defs),target_sf,defs(number_of_defs))


    end subroutine ctf_find_best_defs_for_sf

    !>  \brief  Find the first few frequencies where the ctf = 0.0
    !!
    !!  We find the zeroes halfway between the two axes of the elipse
    !!
    subroutine ctf_zeroes(self,number_of_zeroes,zeroes)
        implicit none
        ! Arguments
        type(ctf_parameters),   intent(in)      ::  self                !<  ctf object
        integer,                intent(in)      ::  number_of_zeroes    !<  This many zeroes will be returned
        ! Result
        real,   allocatable,    intent(inout)   ::  zeroes(:)           !<  The position of the zeroes (in 1/pixels; 0.5 is Nyquist)
        ! Private variables
        integer ::  i
		real	::	tmp
        ! Start work

        ! Allocate memory for results
        if (allocated(zeroes)) deallocate(zeroes)
        allocate(zeroes(number_of_zeroes))

        ! Find the position of the zeroes in terms of the argument to the cosine and sine in the ctf expression
        call ctf_solve_for_arg(self%ampl_cont,zeroes)

        ! Now convert the arguments to spatial frequencies
        do i=1,size(zeroes)
            call ctf_sf_from_arg(self%cs,self%wl,self%dfmid1,self%dfmid2,self%angast,zeroes(i),self%angast+45.0,tmp)
        	zeroes(i) = tmp
		enddo
    end subroutine ctf_zeroes

    !>  \brief Wrapper around eval_ctf which takes a ctf object as argument
#ifdef skip_runtime_checks
    pure &
#endif
    real function eval_ctf(ctf1,box_size,rad,ang,sign_only)
        use runtime_parameters, only : terminate
        implicit none
        type(ctf_parameters),   intent(in)  ::  ctf1        !< ctf object containing all relevant ctf parameters
        integer,                intent(in)  ::  box_size    !< Box dimension (pixels)
        real,                   intent(in)  ::  rad         !< Radius at which to compute the ctf (pixels)
        real,                   intent(in)  ::  ang         !< Angle at which to compute the ctf (radians)
        logical,    optional,   intent(in)  ::  sign_only   !< Return the sign of the ctf rather than its value
#ifndef skip_runtime_checks
        ! Check the ctf parameters are in pixels
        if (.not. ctf1%pix_units) then
            write(*,'(a)') '**ERROR(eval_ctf_wrap): ctf parameters not supplied in appropriate units'
            call terminate('Fatal error in EVAL_CTF_WRAP')
        endif
#endif
        ! Call eval_ctf
        eval_ctf = eval_ctf_slave(ctf1%cs,ctf1%wl,ctf1%ampl_cont,ctf1%dfmid1,ctf1%dfmid2,ctf1%angast,box_size,rad,ang,sign_only)
    end function eval_ctf

    !>  \brief Returns the ctf, based on ctffind3 subroutine (see Mindell 2003)
    pure real function eval_ctf_slave(cs,wl,ampl_cont,dfmid1,dfmid2,angast,box_size,rad,ang,sign_only)
        use runtime_parameters, only: pi ! we only need pi
        implicit none
        ! Arguments
        real,                   intent(in)  ::  cs          !< Spherical aberation (pixels)
        real,                   intent(in)  ::  wl          !< Electron wavelength (pixels)
        real,                   intent(in)  ::  ampl_cont   !< Fraction of amplitude contrast (from .07 to .14, see Mindell 2003)
        real,                   intent(in)  ::  dfmid1      !< Defocus along first axis (pixels)
        real,                   intent(in)  ::  dfmid2      !< Defocus along second axis (for astigmatic ctf, dfmid1 .NE. dfmid2) (pixels)
        real,                   intent(in)  ::  angast      !< Azimuth of first axis. 0.0 means axis is at 3 o'clock. (radians)
        integer,                intent(in)  ::  box_size    !< Box dimension (pixels)
        real,                   intent(in)  ::  rad         !< Radius at which to compute the ctf (pixels)
        real,                   intent(in)  ::  ang         !< Angle at which to compute the ctf (radians)
        logical,    optional,   intent(in)  ::  sign_only   !<  Return 1.0 or -1.0
        ! Private variables
        real    ::  wgh1    ! Phase contrast
        real    ::  wgh2    ! Amplitude contrast
        real    ::  ctf_arg
        ! Start work

        ! Compute phase and amplitude contrast
        wgh1    =   sqrt(1-ampl_cont**2)
        wgh2    =   ampl_cont
        if (rad .eq. 0.0) then
            eval_ctf_slave = -wgh2
        else
            ctf_arg = eval_ctf_arg(cs,wl,ampl_cont,dfmid1,dfmid2,angast,box_size,rad,ang)
            eval_ctf_slave = - wgh1 * sin(ctf_arg) - wgh2 * cos(ctf_arg)
        endif
        if (present(sign_only)) then
            if (sign_only) eval_ctf_slave = sign(1.0,eval_ctf_slave)
        endif
    end function eval_ctf_slave

    !>  \brief Returns the argument (radians) to the sine and cosine terms of the ctf
    pure real function eval_ctf_arg(cs,wl,ampl_cont,dfmid1,dfmid2,angast,box_size,rad,ang) result(ctf_arg)
        use runtime_parameters, only: pi ! we only need pi
        implicit none
        ! Arguments
        real,                   intent(in)  ::  cs          !< Spherical aberation (pixels)
        real,                   intent(in)  ::  wl          !< Electron wavelength (pixels)
        real,                   intent(in)  ::  ampl_cont   !< Fraction of amplitude contrast (from .07 to .14, see Mindell 2003)
        real,                   intent(in)  ::  dfmid1      !< Defocus along first axis (pixels)
        real,                   intent(in)  ::  dfmid2      !< Defocus along second axis (for astigmatic ctf, dfmid1 .NE. dfmid2) (pixels)
        real,                   intent(in)  ::  angast      !< Azimuth of first axis. 0.0 means axis is at 3 o'clock. (radians)
        integer,                intent(in)  ::  box_size    !< Box dimension (pixels)
        real,                   intent(in)  ::  rad         !< Radius at which to compute the ctf (pixels)
        real,                   intent(in)  ::  ang         !< Angle at which to compute the ctf (radians)
        ! Private variables
        real    ::  sf      ! Spatial frequency (pix**-1)
        real    ::  df      ! Defocus at point at which we're evaluating the ctf
        real    ::  ccos
        ! Start work

        if (rad .eq. 0.0) then
            ctf_arg = 0.0
        else
            ! Convert the radius to spatial frequency (pix**-1)
            sf = rad / real(box_size)
            ! Compute the defocus
            ccos = cos(2.0*(ang-angast))
            df = 0.5*(dfmid1+dfmid2+ccos*(dfmid1-dfmid2))
            ! Compute the ctf argument
            ctf_arg = pi * wl * sf**2 * (df - 0.5 * wl**2 * sf**2 * cs)
        endif
    end function eval_ctf_arg

    !>  \brief  Given the argument to the ctf, return the spatial frequency
    !!
    !!  According to Maxima, there are two positive solutions
    pure subroutine ctf_sf_from_arg(cs,wl,dfmid1,dfmid2,angast,arg,ang,sf)
        use runtime_parameters, only: pi ! we only need pi
        implicit none
        ! Arguments
        real,                   intent(in)  ::  cs          !< Spherical aberation (pixels)
        real,                   intent(in)  ::  wl          !< Electron wavelength (pixels)
        real,                   intent(in)  ::  dfmid1      !< Defocus along first axis (pixels)
        real,                   intent(in)  ::  dfmid2      !< Defocus along second axis (for astigmatic ctf, dfmid1 .NE. dfmid2) (pixels)
        real,                   intent(in)  ::  angast      !< Azimuth of first axis. 0.0 means axis is at 3 o'clock. (radians)
        real,                   intent(in)  ::  arg         !< Radius at which to compute the ctf (pixels)
        real,                   intent(in)  ::  ang         !< Angle at which to compute the ctf (radians)
        real,                   intent(out) ::  sf          !< Spatial frequency
        ! Private variables
        real    ::  df      ! Defocus
        real    ::  ccos
        real    ::  tmp
        ! Start work

        ! Compute the defocus
		ccos = cos(2.0*(ang-angast))
		df = 0.5e0*(dfmid1+dfmid2+ccos*(dfmid1-dfmid2))
        tmp = pi*df**2 - 2*wl*arg*cs
        if (tmp .gt. 0.0e0) then
            sf = sqrt(pi)*df*wl - sqrt(wl**2*(tmp) )
        else
            sf = 0.0e0
        endif
        sf = sf / (sqrt(pi)*wl**3*cs)
        sf = sqrt(sf)
    end subroutine ctf_sf_from_arg

    !>  \brief  Given the argument to the ctf, return the defocus
    !!
    pure subroutine ctf_df_from_arg(cs,wl,arg,sf,df)
        use runtime_parameters, only: pi ! we only need pi
        implicit none
        ! Arguments
        real,                   intent(in)  ::  cs          !< Spherical aberation (pixels)
        real,                   intent(in)  ::  wl          !< Electron wavelength (pixels)
        real,                   intent(in)  ::  arg         !< ctf argument (radians)
        real,                   intent(in)  ::  sf          !< Spatial frequency (1/pixels)
        real,                   intent(out) ::  df          !< Defocus (pixels)
        ! Private variables

        ! Start work

        ! Compute the defocus
        df = (0.5 * cs * pi * sf**4 * wl**3 + arg) / (pi * sf**2 * wl)
    end subroutine ctf_df_from_arg

    !>  \brief  Find the ctf argument values at which CTF=0
    !!
    !!  According to Wolfram Alpha, the solutions to a*cos(t) + b*sin(t) = c are:
    !!  1)  t = 2 (atan((b-sqrt(a^2+b^2-c^2))/(a+c))+pi*n),   with n an integer
    !!  2)  t = 2 (atan((b+sqrt(a^2+b^2-c^2))/(a+c))+pi*n),   with n an integer
    !!
    !!  In our case, a = - ampl_const
    !!           and b = - sqrt(1-AMPL_CONT**2)
    !!  and t is the "argument" to the ctf, which can be "converted" to a spatial frequency
    subroutine ctf_solve_for_arg(ampl_cont,sols,ctf_value)
        use runtime_parameters, only : terminate
        use runtime_parameters, only : pi
        implicit none
        ! Arguments
        real,                   intent(in)      ::  ampl_cont
        real,   allocatable,    intent(inout)   ::  sols(:)     !<  Solutions
        real,   optional,       intent(in)      ::  ctf_value   !<  Value of the ctf at the solutions
        ! Private variables
        integer ::  i,j
        real    ::  a,b
        real    ::  c
        ! Start work
        if (.not. allocated(sols)) then
            write(*,'(a)') '**ERROR(ctf_solve_for_arg): array of solutions is not allocated'
            call terminate('Fatal error in CTF_SOLVE_FOR_ARG')
        endif
        if (size(sols) .lt. 1) then
            write(*,'(a,i0)') '**ERROR(ctf_solve_for_arg): array of solutions has unexpected size ', size(sols)
            call terminate('Fatal error in CTF_SOLVE_FOR_ARG')
        endif
        c = 0.0e0
        if (present(ctf_value)) c = ctf_value
        !
        a = -ampl_cont
        b = -sqrt(1-ampl_cont**2)
        ! Loop over the zeroes
        do i=1,size(sols)/2
            j = (i-1)*2+1
            sols(j) = 2.0*(atan((b-sqrt(a**2+b**2-c**2))/(a+c))+pi*real(i-1))
            j = (i-1)*2+2
            sols(j) = 2.0*(atan((b+sqrt(a**2+b**2-c**2))/(a+c))+pi*real(i))
        enddo
    end subroutine ctf_solve_for_arg

    !>  \brief  Convert acceleration voltage in kV into electron wavelength in Angstroms
    pure real function akv_to_wl(akv)
        implicit none
        real, intent(in) :: akv
        akv_to_wl = 12.26/sqrt(1000.0*akv+0.9784*(1000.0*akv)**2/(10.0**6.0))
    end function akv_to_wl

    !>  \brief  Convert electron wavelength in Angstroms into acceleration voltage in kV
    !!  \todo   Check carefully; not 100% sure results are correct
    pure real function wl_to_akv(wl)
        implicit none
        ! Argument
        real, intent(in) :: wl
        ! Private variables
        real    ::  tmp
        ! Start work
        tmp = 0.9784 * 10.0**(-6.0)
        wl_to_akv = (-1. + sqrt(1.+4.*tmp*12.26**2/wl**2))/(2.*tmp)
        wl_to_akv = wl_to_akv / 1000.
    end function wl_to_akv

end module ctfs
