!>  \brief  Classes and methods to represent and import/export information
!!          normally generated by ctftilt
!!
!!  To read ctftilt info from ctffind results:
!!      type(ctftilt)   ::  ctft
!!      call ctft%read_from_ctffind(filename)
!!
!!
!!
!!  The following documentation is copied from the output of a run of ctftilt:
!!
!!     equation for calculating defocus dfl1,dfl2 at location nx,ny:
!!
!!          dfl1  = dfmid1 +DF
!!          dfl2  = dfmid2 +DF
!!          df    = (N1*DX+N2*DY)*PSIZE*TAN(tangle)
!!          dx    = CX-NX
!!          dy    = CY-NY
!!          cx    = center_x =         2906
!!          cy    = center_y =         3575
!!          psize = pixel size [A] =       2.3400
!!          n1,n2 = tilt axis normal:
!!             n1 = -SIN(tltaxis) =    -0.944092
!!             n2 =  cos(tltaxis) =    -0.329682
!!
!!
!!    31837.41,    31613.90    <--(dfmid1,dfmid2)-->       33875.71,    33652.21
!!           1,           1    <------(nx,ny)------>           5812,           1
!!          +----------------------------------------------------------+
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                   33294.16,    33070.66                  |
!!          |                       2906,        3575                  |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          |                                                          |
!!          +----------------------------------------------------------+
!!           1,        7150    <------(nx,ny)------>           5812,        7150
!!    32713.09,    32489.58    <--(dfmid1,dfmid2)-->       34751.39,    34527.89


!!
!! The following formula is from ctftilt v1.7 output:
!!     equation for calculating defocus dfl1,dfl2 at location nx,ny:
!!
!!          dfl1  = dfmid1 +DF
!!          dfl2  = dfmid2 +DF
!!          df    = (N1*DX+N2*DY)*PSIZE*TAN(tangle)
!!          dx    = CX-NX
!!          dy    = CY-NY
!!          cx    = center_x =         2440
!!          cy    = center_y =         3360
!!          psize = pixel size [A] =       2.3729
!!          n1,n2 = tilt axis normal:
!!             n1 =  sin(tltaxis) =     0.625416
!!             n2 = -COS(tltaxis) =    -0.780292
!!
!!



!! The following is copied from in-source documentation written by Chuck Sindelar for his chuff project:
!!
!!   The tilting issue: what is the relationship between out of plane
!!    tilt of a helix and defocus?  I.e., if you progress along a filament
!!    as specified by a given set of Euler angles, how does the defocus
!!    change?
!!
!!   To begin, frealix/spider conventions (identical) for euler angles are: (1) phi
!!   is clockwise rotation around z; (2) theta is clockwise rotation around y;
!!   (3) psi is final clockwise rotation around z
!!
!!   note: it's easy to picture the x-y plane inconsistently:
!!      * when windowing, etc., spider and other programs depict images
!!           with the origin in the upper left hand corner, +x to the right, +y down.
!!          -> in this "SPIDER" view point psi=90 actually points "down"!!
!!                    (when using cos(psi) and sin(psi) the usual way)
!!
!!      * when performing in-plane rotations with spider, etc. +psi goes
!!           counterclockwise when viewing the spider depictions.  This is because
!!           the z axis goes into the screen in these depictions, as opposed to the
!!           "math class" view where y goes to the top of the screen, and x goes right,
!!           which leads to the z axis coming out of the screen.  In the "math class"
!!           view, as in the "SPIDER view," the spider rotations are consistent with
!!           the Euler angles as defined: clockwise rotation around z.
!!
!!   axis conventions
!!    The focal plane:  (from Niko's em reference book) Underfocus corresponds to the focal
!!     plane being "under" the sample, i.e. farther from the gun than the sample is.  The
!!     way to think about this would be that with more lens current, the focal plane moves
!!     up towards the gun (and the lenses), which seems to make sense.
!!    The z axis: Niko's guess is that increasing z goes from the gun towards the sample,
!!     i.e. the coordinate system is upside-down.
!!
!!    These above two observations (if both are correct) would uniquely specify the
!!     relationship between defocus and helical out of plane tilt, as follows:
!!
!!    Increasing z would move a particle away from the gun, thus towards the focal plane.
!!      therefore, increasing z decreases the magnitude of the defocus/underfocus.
!!      (i.e. making the focus value less negative)
!!
!!  ###########################################
!!   analysis for frealign, chumps
!!
!!   Note that in reconstruction, Euler angle rotations are executed in reverse order when
!!   inserting images into a fourier volume...
!!
!!  ###########################################
!!   in chumps (according to the "math class" view):
!!   volume begins as:      +
!!
!!   after theta = 90:              +-
!!      theta < 90 means + end is tilting up, theta > 90 means + end is tilting down
!!          "UP" means the filament moves in the positive z direction.
!!
!!      after psi=90:       +
!!                  -
!!          + end (which is "UP" theta < 90) is pointing in +y direction, which is
!!                  in the direction of travel for chumps.  Therefore:
!!                  psi=90, theta < 90 means increasing z in the direction of travel.
!!
!!  ##########################################
!!   in frealign/frealix (according to the "math class" view):
!!
!!   volume begins as:   +
!!
!!   after theta = 90:       +-
!!      psi = 0 would extend this to the right, in the "MINUS END" direction of the filament!
!!      theta < 90 would tilt the - end (which is the direction of travel) down
!!                 "DOWN" means the filament progresses in the negative z direction.
!!                 Therefore: theta < 90 always means decreasing z in the direction of travel.
!!                 Thus: given the logic above,
!!                       theta < 90 means increasing the defocus magnitude
!!                       theta > 90 means decreasing the defocus magnitude
!!  ###########################################
!!
!!
!! Additional note (alr - 12-Jan-2010): Boxer displays (x,y) = (0,0) at bottom left of window,
!!                                      so that the z axis comes out from the screen, and psi appears clockwise
!!                 (alr - 05-Nov-2010): So does v2, EMAN1's display program


module ctftilts

    use ctfs

    type ctftilt
        type(ctf_parameters)    ::  ctf
        real    ::  tilt_angle = 0.0    !<  Angle (degrees or radians) of tilt
        real    ::  tilt_axis  = 0.0    !<  Azimuth of tilt axis (degrees or radians)
        !contains
        !   procedure   ::  read_from_ctffind   =>  ctftilt_read_from_ctffind
        !   procedure   ::  read_from_ctftilt   =>  ctftilt_read_from_ctftilt
    end type ctftilt

    interface ang_to_pix
        module procedure ctftilt_ang_to_pix
    end interface ang_to_pix

    interface pix_to_ang
        module procedure ctftilt_pix_to_ang
    end interface pix_to_ang

    contains

    !>  \brief  Print out information describing the ctftilt
    subroutine ctftilt_print_info(self)
        implicit none
        ! Arguments
        type(ctftilt),          intent(in)  ::  self    !<  ctftilt object
        ! Private variables
        character(len=2),       parameter   ::  text_width  =   '25'
        ! Start work
        call ctf_print_info(self%ctf)
        write(*,'(A'//text_width//',1x,F6.3)')       'Tilt angle:',          self%tilt_angle
        write(*,'(A'//text_width//',1x,F6.3)')       'Tilt axis:',           self%tilt_axis

    end subroutine ctftilt_print_info

    !>  \brief  Convert ctftilt object from Angstrom & degrees to pixels & radians
    subroutine ctftilt_ang_to_pix(self,pixel_size)
        use runtime_parameters, only : terminate
        use rotation_matrices, only : radians
        implicit none
        ! Arguments
        type(ctftilt),  intent(inout)   ::  self        !<  ctftilt object to convert
        real,           intent(in)      ::  pixel_size  !<  Pixel size (Angstroems)
        ! Private variables

        ! Start work

        ! Check we're not already converted
        if (self%ctf%pix_units) then
            write(*,'(a)') '**ERROR(ctftilt_ang_to_pix): already in pixels and radians'
            call terminate('Fatal error in CTFTILT_ANG_TO_PIX')
        endif
        ! Do the conversion
        self%tilt_angle =   radians(self%tilt_angle)
        self%tilt_axis  =   radians(self%tilt_axis)
        call ctf_ang_to_pix(self%ctf,pixel_size)
    end subroutine ctftilt_ang_to_pix

    !>  \brief  Convert ctftilt object from pixels & radians to Angstrom & degrees
    subroutine ctftilt_pix_to_ang(self,pixel_size)
        use runtime_parameters, only : terminate
        use rotation_matrices, only : degrees
        implicit none
        ! Arguments
        type(ctftilt),  intent(inout)   ::  self        !<  ctftilt object to convert
        real,           intent(in)      ::  pixel_size  !<  Pixel size (Angstroems)
        ! Private variables

        ! Start work

        ! Check we're not already converted
        if (self%ctf%pix_units) then
            write(*,'(a)') '**ERROR(ctftilt_pix_to_ang): already in pixels and radians'
            call terminate('Fatal error in CTFTILT_PIX_TO_ANG')
        endif
        ! Do the conversion
        self%tilt_angle =   degrees(self%tilt_angle)
        self%tilt_axis  =   degrees(self%tilt_axis)
        call ctf_pix_to_ang(self%ctf,pixel_size)
    end subroutine ctftilt_pix_to_ang

    !>  \brief  Get ctftilt data from the output of a ctffind run. Tilt is assumed to be zero.
    !!
    !!  note: cs, wl, ampl_cont are not parsed from the ctffind results
    subroutine ctftilt_read_from_ctffind(self,ctffind_filename,ctffind4_micrograph_id)
        use runtime_parameters, only : filename_len, word_len, line_len, terminate
        use parsers
        use ctfs, only : akv_to_wl
        use StringManipulations, only : StringsAreEqual
        use NumericTextFiles
        use ctffind_routines
        implicit none
        ! Arguments
        type(ctftilt),                  intent(inout)   ::  self        !<  ctftilt object to be populated with data
        character(len=filename_len),    intent(in)      ::  ctffind_filename    !<  output from ctffind - will be parsed to get relevant values
        integer,    optional,           intent(in)      ::  ctffind4_micrograph_id
        ! private variables
        character(len=line_len) ::  line
        character(len=word_len), allocatable    ::  words(:)
        real    ::  ht ! high tension
        type(NumericTextFile)   ::  my_file
        real                    ::  data_from_file(7)
        integer                 ::  current_data_line
        real                    ::  acceleration_voltage, spherical_aberration, amplitude_contrast
        ! Start work
        line = grep(ctffind_filename,'Final Values')
        if (StringsAreEqual(' ',line)) then
            ! Not a valid ctffind3 output, we'll assume it's ctffind4
            if (present(ctffind4_micrograph_id)) then
                call GetParameterValueFromCTFFindResultsFile(ctffind_filename,'voltage:',acceleration_voltage)
                call GetParameterValueFromCTFFindResultsFile(ctffind_filename,'aberration:',spherical_aberration)
                call GetParameterValueFromCTFFindResultsFile(ctffind_filename,'contrast:',amplitude_contrast)
                ! start work
                call my_file%init(ctffind_filename,open_to_read)
                do current_data_line=1,my_file%number_of_data_lines
                    call my_file%readnextdataline(data_from_file)
                    if (data_from_file(1) .eq. ctffind4_micrograph_id .or. my_file%GetNumberOfDataLines() .eq. 1) then
                        call ctf_parameters_init(self%ctf,acceleration_voltage,spherical_aberration,amplitude_contrast, &
                                             data_from_file(2)/10000.0,data_from_file(3)/10000.0,data_from_file(4))
                    endif
                enddo

            else
                call terminate('Fatal error in ctffindtilt_read_from_ctffind: need micrograph id')
            endif
        else
            words = split(line)
            if (size(words) .ne. 6) then
                write(*,'(a,i2,a,a)')   '**ERROR(ctftilt_read_from_ctffind): was expecting 6 words, but found ', size(words), &
                                        ' in the results line of the ctffind file: ', trim(line)
                call terminate('Fatal error in CTFTILT_READ_FROM_CTFFIND')
            endif
            read(words(1),'(F10.2)') self%ctf%dfmid1
            read(words(2),'(F10.2)') self%ctf%dfmid2
            read(words(3),'(F10.2)') self%ctf%angast
            ! Now find the cs, wl and ampl_cont
            line = grep(ctffind_filename,'CS[mm]',SHIFT=1) ! the numbers we want are in the line following the one with 'Cs[mm]'
            if (string_equal(line,' ')) then
                write(*,'(3a)') '**ERROR(ctftilt_read_from_ctffind): ''CS[mm]'' not found in ', trim(adjustl(ctffind_filename)), &
                                '. Are you sure it really was output by CTFFIND?'
                call terminate('Fatal error in CTFTILT_READ_FROM_CTFFIND')
            endif
            words = split(line)
            !print *, '**debug(ctftilt_read_from_ctftilt): found cs words = ', words
            read(words(1),'(F10.2)') self%ctf%cs
            read(words(2),'(F10.2)') ht
            self%ctf%wl = akv_to_wl(ht)
            read(words(3),'(F10.2)') self%ctf%ampl_cont
            self%ctf%pix_units = .false.
            !call ctftilt_print_info(self)
        endif
    end subroutine ctftilt_read_from_ctffind

    !>  \brief  Get ctftilt data from the output of a ctftilt run.
    !!
    !!  note: cs, wl, ampl_cont are not parsed from the ctftilt results
    subroutine ctftilt_read_from_ctftilt(self,ctftilt_filename)
        use runtime_parameters
        use parsers
        use ctfs, only : akv_to_wl
        implicit none
        ! Arguments
        type(ctftilt),              intent(inout)   ::  self        !<  ctftilt object to be populated with data
        character(len=*),           intent(in)      ::  ctftilt_filename    !<  Output from ctffind - will be parsed to get relevant values
        ! Private variables
        character(len=LINE_LEN) ::  line
        character(len=WORD_LEN), allocatable    ::  words(:)
        real    ::  ht ! high tension
        real    ::  ctftilt_version
        ! Start work

        ! Results from ctftilt versions 1.7 and above have inverted the sign of the tilt angle compared
        ! to earlier version
        ctftilt_version = ctftilt_get_version(ctftilt_filename)
        if (ctftilt_version .ge. 1.7) then
            !write(*,'(a,F0.1,a)') '**ERROR(ctftilt_read_from_ctftilt): results from ctftilt version ', ctftilt_version, &
            !                       ' are not yet supported. Sorry.'
            !call terminate('Fatal error in CTFTILT_READ_FROM_CTFTILT')
            write(*,'(a,F0.1,a)') '**warning(ctftilt_read_from_ctftilt): support for ctftilt version ', ctftilt_version, &
                            ' was only added recently. Please check results carefully.'
        endif

        line = grep(ctftilt_filename,'Final Values')
        words = split(line)
        if (size(words) .ne. 8) then
            write(*,'(a,i2,a,a)')   '**ERROR(ctftilt_read_from_ctftilt): was expecting 8 words, but found ', size(words), &
                                    ' in the results line of the ctftilt file: ', trim(line)
            call terminate('Fatal error in CTFTILT_READ_FROM_CTFTILT')
        endif
        read(words(1),'(F10.2)') self%ctf%dfmid1
        read(words(2),'(F10.2)') self%ctf%dfmid2
        read(words(3),'(F10.2)') self%ctf%angast
        read(words(4),'(F10.2)') self%tilt_axis
        if (ctftilt_version .ge. 1.7) self%tilt_axis = - self%tilt_axis
        read(words(5),'(F10.2)') self%tilt_angle
        ! Now find the cs, wl and ampl_cont
        line = grep(ctftilt_filename,'CS[mm]',SHIFT=1) ! the numbers we want are in the line following the one with 'Cs[mm]'
        words = split(line)
        !print *, '**debug(ctftilt_read_from_ctftilt): found cs words = ', words
        read(words(1),'(F10.2)') self%ctf%cs
        !print *, '**debug(ctftilt_read_from_ctftilt): cs = ', self%ctf%cs
        read(words(2),'(F10.2)') ht
        self%ctf%wl = akv_to_wl(ht)
        read(words(3),'(F10.2)') self%ctf%ampl_cont
        self%ctf%pix_units = .false.
        !call ctftilt_print_info(self)
    end subroutine ctftilt_read_from_ctftilt

    !>  \brief  Get ctftilt version from its output
    real function ctftilt_get_version(ctftilt_filename) result (version)
        use runtime_parameters, only : terminate, line_len, word_len
        use parsers
        implicit none
        ! Arguments
        character(len=*),           intent(in)      ::  ctftilt_filename    !<  Output from ctffind - will be parsed to get relevant values
        ! Private variables
        character(len=LINE_LEN) ::  line
        character(len=WORD_LEN), allocatable    ::  words(:)
        ! Start work
        line = grep(ctftilt_filename,'CTF TILT DETERMINATION')
        words = split(line)
        if (size(words) .ne. 5) then
            write(*,'(a,i2,a,a)')   '**ERROR(ctftilt_get_version): was expecting 5 words, but found ', size(words), &
                                    ' in the greeting line of the ctftilt file: ', trim(line)
            call terminate('Fatal error in CTFTILT_GET_VERSION')
        endif
        ! The 4th word will be something like V1.6 or V1.7
        words(4) = adjustl(words(4))
        if (len_trim(words(4)) .ne. 4) then
            write(*,'(2a)') '**ERROR(ctftilt_get_version): unexpected length of version string (should be something like V1.6): ', &
                            trim(words(4))
            call terminate('Fatal error in CTFTILT_GET_VERSION')
        endif
        version = 0.0
        read(words(4)(2:4),'(F3.1)') version
    end function ctftilt_get_version

end module ctftilts
