!>  \brief  Classes and methods implementing differential evolution
!!
!!  20-Feb-2012:    it turns out Hans Elmlund pointed me to de a few months ago, but i only realised this in the last few days,
!!                  when i decided de might be a good solution to my optimisation problem. i intend eventually to implement
!!                  Hans' improvements on the methods, but for now will start with a basic version based on the original paper
!!                  (﻿Storn, R., & Price, K. (1996). Minimizing the real functions of the ICEC’96 contest by Differential Evolution.
!!                  in: ieee conference on evolutionary computation, 842 - 844)
!!
!!  Another useful resource, which includes rules-of-thumbs for setting de parameters is:
!!                  ﻿Storn, R. (1996). On the Usage of Differential Evolution for Function Optimization.
!!                  in NAFIPS’96, 519 - 523. Retrieved from http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.71.2432
module differential_evolution_mod

    use fgsl
    use, intrinsic :: iso_c_binding
    use runtime_parameters, only : terminate

    private

    public :: de_minimiser, de_init, de_run, de_best_values

    type de_individual
        real(kind=8),   allocatable ::  val(:)
        real                        ::  f       =   0.7             !<  Perturbation scaling
        real                        ::  cr      =   0.3             !<  Crossover probability
        real                        ::  score   =   0.0
    end type

    type de_population
        type(de_individual),    allocatable ::  members(:)          !<  Number of members in the population is noted np in Storn & Price (1996)
    end type

    !>  \brief   Definition of a minimisable function.
    abstract interface
        function scoring_function(values,params) bind(c)
            import :: c_double, c_ptr   ! Because we're in an abstract interface, we normally don't have access to all the stuff from the module, so we need to import
            type(c_ptr), value  ::  values
            type(c_ptr), value  ::  params
            real(kind=C_DOUBLE) ::  scoring_function
        end function
    end interface

    type de_minimiser
        type(de_population)     ::  population
        type(fgsl_rng_type)     ::  rng_type
        type(fgsl_rng)          ::  rng
        !
        type(de_individual)     ::  best_member
        !
        character(len=50)       ::  algorithm               =   'DE_BEST_2'     !<  See Storn & Price papers for algorithms
        logical                 ::  first_gen_is_gaussian   =   .false.         !<  Whether the first generation should be Gaussian-, rather than uniformly-, distributed around the starting point
        logical                 ::  evolve_f                =   .false.         !<  If true, f (perturbation scaling) is per-individual and is one of the parameters under evolutionary pressure
        logical                 ::  evolve_cr               =   .false.         !<  If true, cr (crossover probability) is per-individual and is one of the parameters under evolutionary pressur
        real                    ::  f_mutation_probability  =   0.1             !<  Probabiltiy that f  will be mutated (only applies if evolve_f)
        real                    ::  cr_mutation_probability =   0.1             !<  Probability that cr will be mutated (only applies if evolve_cr)
        real                    ::  perturbation_scaling    =   0.7             !<  Scaling factor for the perturbation/mutation. f in Storn & Price (1996). If evolve_f, this is only the starting value.
        real                    ::  crossover_probability   =   0.3             !<  Probability of mutation. Must in range [0.,1.]. cr in Storn & Price (1996). If evolve_cr, this is only the starting value.
        logical,    allocatable ::  integer_changes_only(:)                     !<  For each dimension, restrict changes from the starting point to integer increments
        !
        integer                 ::  current_generation      =   0               !<  Current generation
        integer                 ::  max_num_generations     =   100             !<  Maximum number of generations
        integer                 ::  num_function_calls      =   0               !<  Keep track of the total number of function calls
        real                    ::  average_score           =   0.0             !<  Keep track of the average score of population members
        !
        type(c_ptr)             ::  params_for_scoring      !<  a pointer to an object containing all the information needed by the scoring function (other than the function's arguments)
        procedure(scoring_function), pointer, nopass :: function => null()  !<  a pointer to the function the minimiser aims to minimise.
                                                                            !!  no pass because there an assumption by default that type-bound procedures have an object of the type to which they are bound as first argument.
                                                                            !!  This default assumption is overriden by no PASS.
        contains
            procedure   ::  num_dim         =>  de_num_dim
            procedure   ::  pop_size        =>  de_population_size
            procedure   ::  population_size =>  de_population_size
    end type


contains

        !>  \brief  Compute the score for a particular individual
        subroutine de_score_individual(self,ind)
            implicit none
            ! Arguments
            type(de_minimiser),             intent(inout)   ::  self
            type(de_individual),    target, intent(inout)   ::  ind
            ! Private variables
            type(c_ptr) ::  vals_c
            logical,    parameter   ::  debug   =   .false.
            ! Start work

            ! Make c pointer to the individual's values
            vals_c = c_loc(ind%val)

            ! Call the scoring function
            ind%score = self%function(vals_c,self%params_for_scoring)
            self%num_function_calls = self%num_function_calls + 1

            if (debug) then
                !$omp   critical (omp_crit_de_score_ind)
                write(*,'(a)',advance='NO') '**debug(de_score_individual): individual information: '
                call de_individual_print_info(ind)
                !$omp   end critical (omp_crit_de_score_ind)
            endif
        end subroutine de_score_individual

        !>  \brief  Iterate to next generation
        subroutine de_new_generation(self)
            use omp_lib
            implicit none
            ! Arguments
            type(de_minimiser), intent(inout)   ::  self
            ! Private variables
            integer ::  current_member
            logical,    parameter   ::  debug   =   .true.
            real(kind=8)    ::  new_average_score
            type(de_individual) ::  perturbed_member
            ! Start work

            ! Check the caller didn't forget to call de_first_generation
            if (self%current_generation .eq. 0) then
                write(*,'(a)') '**ERROR(de_new_generation): first generation has not been generated yet.'
                call terminate('Fatal error in DE_NEW_GENERATION')
            endif

            !
            write(*,'(a,i0)') '**info(de_new_generation): working on generation ', self%CURRENT_GENERATION+1



            !$omp   parallel        &
            !$omp   default (none)  &
            !$omp   shared(new_average_score,self)        &
            !$omp   private(current_member,perturbed_member)

            ! We will compute the new average score
            new_average_score = 0.0D0

            ! Loop through the population members
            !$omp   do  reduction(+:new_average_score)
            do CURRENT_MEMBER=1,size(self%population%members)
                !if (debug .AND. omp_get_num_threads() .GT. 1) THEN
                !    write(*,'(a,i0,a,i0,a,i0)') '**debug(de_new_generation): thread ', omp_get_thread_num()+1, ' of ', &
                !                                omp_get_num_threads(), ' working on member ', current_member
                !endif
                ! Come up with a new challenger individual
                call de_make_new_perturbed_member(self,current_member,perturbed_member)
                ! Check whether the challenger beats the current individual
                call de_score_individual(self,perturbed_member)
                ! If the challenger does score better, it will replace the current individual
                !$omp   critical (de_new_gen)
                if (perturbed_member%score .lt. self%population%members(current_member)%score) then
                    self%population%members(current_member) = perturbed_member
                    if (debug) then
                        write(*,'(a,i0,a)', advance='NO') '**info(de_new_generation): member ', current_member, &
                                                            ' replaced by new individual: '
                        call de_individual_print_info(perturbed_member)
                    endif
                    ! Is the challenger so good that it beats the previous best member?
                    if (perturbed_member%score .lt. self%best_member%score) then
                        self%best_member = perturbed_member
                    endif
                endif
                !$omp   end critical (de_new_gen)
                new_average_score = new_average_score + self%population%members(current_member)%score
            enddo
            !$omp   enddo

            !$omp   end parallel

            ! The average score
            new_average_score = new_average_score / real(size(self%population%members))
            self%average_score = new_average_score

            ! Increment the generation counter
            self%current_generation = self%current_generation + 1

            ! Output information
            write(*,'(a,i0,a,F0.4)') '**info(de_new_generation): average score for generation ', self%current_generation, ' = ', &
                                        self%average_score

        end subroutine de_new_generation

        !>  \brief  Generate the first generation
        subroutine de_first_generation(self,starting_values,search_half_range)
            use ui, only : interval_print
            implicit none
            ! Arguments
            type(de_minimiser), intent(inout)   ::  self
            real(kind=8),       intent(in)      ::  starting_values(:)
            real(kind=8),       intent(in)      ::  search_half_range(:)
            ! Private variables
            integer ::  current_member, current_dim
            real    ::  rnd
            ! Start work

            ! Check the generation number
            if (self%current_generation .ne. 0) then
                write(*,'(a,i0)') '**ERROR(de_first_generation): minimiser already appears to be on generation ', &
                                    self%current_generation
                call terminate('Fatal error in DE_FIRST_GENERATION')
            endif

            self%average_score = 0.0

            !$omp   parallel    &
            !$omp   default(none)   &
            !$omp   shared(self,starting_values,search_half_range)    &
            !$omp   private(current_member,current_dim,rnd)

            ! Loop over the members of the population
            !$omp   do
            do CURRENT_MEMBER=1,size(self%population%members)
                if (interval_print(current_member,size(self%population%members))) then
                    write(*,'(a,i0,a,i0)') '**info(de_first_generation): working on individual ',   &
                                    current_member, ' of ', size(self%population%members)
                endif
                ! For each dimension, we add a uniformly distributed random variable
                ! to the starting value
                do CURRENT_DIM=1,size(starting_values)
                    if (self%first_gen_is_gaussian) then
                        ! We use half of the search half range as the std dev
                        ! so that ~95% of the values should be within the search
                        ! range requested by the user
                        rnd = fgsl_ran_gaussian(self%rng,real(search_half_range(current_dim)/2.,KIND=FGSL_DOUBLE))
                    else
                        ! Generate a random float variable between 0.0 and 1.0, excluding 1.0
                        rnd = fgsl_rng_uniform(self%rng)
                        ! The random values is between -half_range and +half_range
                        rnd = RND*2.*SEARCH_HALF_RANGE(current_dim) - search_half_range(current_dim)
                    endif
                    if (self%integer_changes_only(current_dim)) rnd = anint(rnd)
                    ! Add random value to starting value
                    self%population%members(current_member)%val(current_dim) = &
                            starting_values(current_dim) + rnd
                enddo
                ! Score this member
                call de_score_individual(self,self%population%members(current_member))
                !$omp critical (de_first_gen)
                self%average_score = self%average_score + self%population%members(current_member)%score
                !$omp end critical (de_first_gen)
            enddo
            !$omp   enddo


            !$omp   end parallel
            self%average_score = self%average_score / real(size(self%population%members))
            self%current_generation = 1
        end subroutine de_first_generation

        !>  \brief  Print out all members of the population (for debugging purposes)
        subroutine de_population_print(self)
            implicit none
            ! Arguments
            type(de_minimiser),                 intent(in)      ::  self
            ! Start work
            call terminate('Not implemented yet')
        end subroutine de_population_print

        !>  \brief  Return the best values found by the minimiser
        subroutine de_best_values(self,best_values)
            implicit none
            ! Arguments
            type(de_minimiser),                 intent(in)      ::  self
            real(kind=8),       allocatable,    intent(inout)   ::  best_values(:)
            ! Private variables

            ! Start work
            best_values = self%best_member%val
        end subroutine de_best_values

        !>  \brief  Run the de minimiser
        subroutine de_run(self)
            implicit none
            ! Arguments
            type(de_minimiser),                 intent(inout)   ::  self
            ! Private variables
            logical,    parameter   ::  debug   =   .true.
            ! Start work

            if (debug) write(*,'(a)') '**debug(de_run): entry'

            ! Loop over generations
            do
                ! Generate the next generation
                call de_new_generation(self)

                ! Printout information about the best member
                write(*,'(a,i0,a)',advance='NO') '**info(de_run): the best member of generation ', self%current_generation, ': '
                call de_individual_print_info(self%best_member)

                ! Are we done
                if (de_minimisation_done(self)) then
                    if (debug) write(*,'(a)') '**debug(de_run): minimisation done'
                    exit
                endif
            enddo

            ! All done
            write(*,'(a,i0)') '**info(de_run): all done. Total number of function calls = ', self%num_function_calls
        end subroutine de_run


        !>  \brief  Print summary information about the de minimiser
        subroutine de_print_info(self)
            implicit none
            ! Arguments
            type(de_minimiser), intent(in)  ::  self
            ! Private variables

            ! Start work
            write(*,'(/a)')         'Differential evolution minimiser, summary info:'
            write(*,'(a,i0)')       'Population size                         = ', self%pop_size()
            write(*,'(a,i0)')       'Number of dimensions                    = ', self%num_dim()
            write(*,'(a,i0)')       'Current generation                      = ', self%current_generation
            write(*,'(a,i0)')       'Maximum # of generations                = ', self%max_num_generations
            write(*,'(a,i0)')       'Number of function calls                = ', self%num_function_calls
            write(*,'(a,l1)')       'First generation is gaussian-distributed? ', self%first_gen_is_gaussian
            write(*,'(a,2(l1,1x))') 'Evolve F? CR?                             ', self%evolve_f, self%evolve_cr
            write(*,'(a)')      ' '
        end subroutine de_print_info

        !>  \brief  Initialise a differential evolution minimiser
        subroutine de_init(self,num_dim,function_to_minimise,parameters,   &
                            starting_values,search_half_range,pop_size,crossover_probability,   &
                            evolve_f,evolve_cr, &
                            integer_changes_only)
            implicit none
            ! Arguments
            type(de_minimiser),                     intent(inout)   ::  self
            integer,                                intent(in)      ::  num_dim                 !<  Number of dimensions to minimise
            procedure(scoring_function)                             ::  function_to_minimise    !<  The function to minimise
            type(c_ptr),                            intent(in)      ::  parameters              !<  Parameters needed by the scoring function (a c pointer to an object containing...)
            real(kind=8),           allocatable,    intent(in)      ::  starting_values(:)      !<  The starting point for the minimiser
            real(kind=8),           allocatable,    intent(in)      ::  search_half_range(:)    !<  For each search dimension, the half-range which should be covered by the minimiser
            integer,    optional,                   intent(in)      ::  pop_size
            real,       optional,                   intent(in)      ::  crossover_probability   !<  Must be in [0.,1.] range
            logical,    optional,                   intent(in)      ::  evolve_f                !<  False by default. If true, f (perturbation scaling) is a per-individual parameter subject to evolution
            logical,    optional,                   intent(in)      ::  evolve_cr               !<  False by default. If true, cr (crossover probability) is a per-individual parameter subject to evolution
            logical,    optional,   allocatable,    intent(in)      ::  integer_changes_only(:) !<  For each dimension, whether changes from the starting point should be restricted to integer increments
            ! Private variables
            integer ::  i
            integer ::  ppop_size
            ! Start work

            ! Population size
            ppop_size = de_suggested_population_size(self,NUM_DIM=NUM_DIM)
            if (present(pop_size)) ppop_size = pop_size

            ! Allocate memory for the population
            allocate(self%population%members(ppop_size))


            ! Allocate memory for each population member
            do I=1,size(self%population%members)
                allocate(self%population%members(i)%val(num_dim))
            enddo

            ! f, cr under evolution?
            if (present(evolve_f)) self%evolve_f = evolve_f
            if (present(evolve_cr)) self%evolve_cr = evolve_cr


            ! Allocate memory for the perturbed member
            !allocate(self%perturbed_member%val(num_dim))

            ! Initialise random number generator (rng)
            self%rng_type = fgsl_rng_env_setup()
            self%rng = fgsl_rng_alloc(self%rng_type)

            ! Check the range of the rng
            if (fgsl_rng_max(self%rng) .lt. size(self%population%members)) then
                write(*,'(a)')  '**ERROR(de_init): random number generator goes to ', fgsl_rng_max(self%rng), &
                                ' which is less than the population size: ', size(self%population%members)
                call terminate('Fatal error in DE_INIT')
            endif

            ! The function we'll be minimising
            self%function => function_to_minimise

            ! Parameters needed by the function
            self%params_for_scoring = parameters

            ! The starting values are stored as the "best member"
            self%best_member%val = starting_values

            ! The score for the starting values
            call de_score_individual(self,self%best_member)
            write(*,'(a,F0.5)') '**info(de_init): starting score = ', self%best_member%score

            ! Crossover probability
            if (present(crossover_probability)) self%crossover_probability = crossover_probability

            ! Give each individual its value of f & cr
            do I=1,size(self%population%members)
                self%population%members(i)%f = self%perturbation_scaling
                self%population%members(i)%cr = self%crossover_probability
            enddo

            !
            if (allocated(self%integer_changes_only)) deallocate(self%integer_changes_only)
            allocate(self%integer_changes_only(num_dim))
            if (present(integer_changes_only)) then
                if (size(integer_changes_only) .ne. num_dim) then
                    write(*,'(a,2(i0,1x))') '**ERROR(de_init): unexpected number of integer_changes_only logicals: ', &
                                            size(integer_changes_only), num_dim
                    call terminate('Fatal error in DE_INIT')
                endif
                self%integer_changes_only = integer_changes_only
            else
                self%integer_changes_only = .false.
            endif

            ! Set up the first generation
            if (allocated(starting_values) .and. size(starting_values) .eq. num_dim .and.   &
                allocated(search_half_range) .and. size(search_half_range) .eq. num_dim ) then
                write(*,'(a,i0)') '**info(de_init): setting up the first generation. Number of individuals = ', ppop_size
                call de_first_generation(self,starting_values,search_half_range)
            else
                write(*,'(a)') '**ERROR(de_init): starting values and/or search_half_range arrays not allocated or not correct size'
                write(*,'(2(L1,1x,i0,1x))') allocated(starting_values), size(starting_values),    &
                                            allocated(search_half_range), size(search_half_range)
                call terminate('Fatal error in DE_INIT')
            endif

            !
            write(*,'(a)') '**info(de_init): differential evolution minimiser has been initialised. Summary:'
            call de_print_info(self)
        end subroutine de_init

        !>  \brief  Free all memory assigned to the de minimiser
        subroutine de_free(self)
            implicit none
            ! Arguments
            type(de_minimiser), intent(inout)   ::  self
            ! Private variables
            integer ::  i
            ! Start work

            ! Free the rng
            call fgsl_rng_free(self%rng)

            ! Deallocate memory for member values
            do I=1,size(self%population%members)
                if (allocated(self%population%members(i)%val)) deallocate(self%population%members(i)%val)
            enddo

            ! Deallocate memory for perturbed member
            !if (allocated(self%perturbed_member%val)) deallocate(self%perturbed_member%val)

            ! Deallocate memory for population members
            if (allocated(self%population%members)) deallocate(self%population%members)

            ! Function
            nullify(self%function)
        end subroutine de_free

        !>  \brief  Generate a new perturbed/mutated population member
        subroutine de_make_new_perturbed_member(self,member_to_challenge,perturbed_member)
            use parsers, only : string_equal
            implicit none
            ! Arguments
            type(de_minimiser),     intent(inout)   ::  self
            integer,                intent(in)      ::  member_to_challenge !<  The index value for the population memeber we will be challenging with the perturbed/mutated member
            type(de_individual),    intent(inout)   ::  perturbed_member    !<  a mutated individual
            ! Private variables
            integer,    allocatable ::  r(:)    !<  Random indices
            integer ::  current_dim
            real    ::  f
            ! Start work

            perturbed_member%score = 0.0
            if (allocated(perturbed_member%val)) then
                if (size(perturbed_member%val) .ne. self%num_dim()) then
                    deallocate(perturbed_member%val)
                endif
            endif
            if (.not. allocated(perturbed_member%val)) allocate(perturbed_member%val(self%num_dim()))

            ! Set the new cr and f values for the perturbed member
            perturbed_member%f  =  self%population%members(member_to_challenge)%f
            perturbed_member%cr =  self%population%members(member_to_challenge)%cr
            if (self%evolve_f)  call individual_mutate_f(self,perturbed_member)
            if (self%evolve_cr) call individual_mutate_cr(self,perturbed_member)

            if (string_equal(self%algorithm,'DE_RAND_1')) then
                !
                ! Algorithm de/rand/1
                ! (Storn & Price, 1996)
                !
                call de_get_mulitple_random_member_indices(self,NUM_TO_GET=3,RANDOM_INDICES=R,INDEX_TO_AVOID=MEMBER_TO_CHALLENGE)
                do CURRENT_DIM=1,size(self%best_member%val)
                    f = perturbed_member%f
                    if (self%integer_changes_only(current_dim)) f = anint(f)
                    perturbed_member%val(current_dim) =   &
                                                    self%population%members(r(1))%val(current_dim) +           &
                                                    f *                                                   &
                                                (   self%population%members(r(2))%val(current_dim) -      &
                                                    self%population%members(r(3))%val(current_dim) )
                enddo
            else if (string_equal(self%algorithm,'DE_BEST_1')) then
                !
                ! Algorithm de/best/1
                ! (Storn & Price, 1996)
                !
                call de_get_mulitple_random_member_indices(self,NUM_TO_GET=2,RANDOM_INDICES=R,INDEX_TO_AVOID=MEMBER_TO_CHALLENGE)
                do CURRENT_DIM=1,size(self%best_member%val)
                    f = perturbed_member%f
                    if (self%integer_changes_only(current_dim)) f = anint(f)
                    perturbed_member%val(current_dim) =   self%best_member%val(current_dim) +   &
                                                  f *           &
                                                ( self%population%members(r(1))%val(current_dim) -   &
                                                  self%population%members(r(2))%val(current_dim) )
                enddo
            else if (string_equal(self%algorithm,'DE_BEST_2')) then
                !
                ! Algorithm de/best/2
                ! (Storn & Price, 1996)
                !
                call de_get_mulitple_random_member_indices(self,NUM_TO_GET=4,RANDOM_INDICES=R,INDEX_TO_AVOID=MEMBER_TO_CHALLENGE)
                do CURRENT_DIM=1,size(self%best_member%val)
                    f = perturbed_member%f
                    if (self%integer_changes_only(current_dim)) f = anint(f)
                    perturbed_member%val(current_dim) =   self%best_member%val(current_dim) +   &
                                                  f *           &
                                                ( self%population%members(r(1))%val(current_dim) +   &
                                                  self%population%members(r(2))%val(current_dim) -   &
                                                  self%population%members(r(3))%val(current_dim) -   &
                                                  self%population%members(r(4))%val(current_dim) )
              enddo
            else
                write(*,'(a)') '**ERROR(de_make_new_perturbed_member): name of algorithm not recognised: ', &
                                trim(adjustl(self%algorithm))
                call terminate('Fatal error in DE_MAKE_NEW_PERTURBED_MEMBER')
            endif

            ! Do the crossover, so that only some of the perturbed value are kept (those in the crossover region)
            ! Note that here we think of the perturbed member as being modified by the individual we are challenging,
            ! whereas in Storn & Price (1996; Fig 2), the crossover is described in the other direction.
            call de_individuals_crossover(self,perturbed_member,self%population%members(member_to_challenge),CR=PERTURBED_MEMBER%cr)

            ! All done
        end subroutine de_make_new_perturbed_member

        !>  \brief  Mutate the f value (perturbation scaling) for an individual, as in Brest et al (2006)
        subroutine individual_mutate_f(self,ind)
            implicit none
            ! Arguments
            type(de_minimiser),     intent(in)      ::  self
            type(de_individual),    intent(inout)   ::  ind
            ! Private variables
            real        ::  rnd
            ! Start work
            rnd = fgsl_rng_uniform(self%rng)    ! Random variable, uniform distribution in [0.,1.) range
            if (rnd .lt. self%f_mutation_probability) then
                rnd = fgsl_rng_uniform(self%rng)
                ind%f = 0.1 + rnd * 0.9    ! The new value is in [0.1,1.0] range
            endif
        end subroutine individual_mutate_f

        !>  \brief  Mutate the cr value (crossover probability) for an individual, as in Brest et al (2006)
        subroutine individual_mutate_cr(self,ind)
            implicit none
            ! Arguments
            type(de_minimiser),     intent(in)      ::  self
            type(de_individual),    intent(inout)   ::  ind
            ! Private variables
            real        ::  rnd
            ! Start work
            rnd = fgsl_rng_uniform(self%rng)    ! Random variable, uniform distribution in [0.,1.) range
            if (rnd .lt. self%cr_mutation_probability) then
                rnd = fgsl_rng_uniform(self%rng)
                ind%cr = rnd
            endif
        end subroutine individual_mutate_cr

        !>  \brief  Apply crossover between two individuals
        subroutine de_individuals_crossover(self,member1,member2,cr)
            implicit none
            ! Arguments
            type(de_minimiser),     intent(in)      ::  self
            type(de_individual),    intent(inout)   ::  member1     !<  This individual will be modified by the crossover
            type(de_individual),    intent(in)      ::  member2     !<  Values from this individual will be crossed into the other individual
            real,                   intent(in)      ::  cr          !<  Crossover probability
            ! Private variables
            integer ::  crossover_start
            integer ::  crossover_length
            integer ::  current_dimension
            integer ::  num_dim
            integer ::  i
            real    ::  rnd
            ! Start work

            num_dim = size(member1%val)

            ! Random crossover characteristics
            crossover_start = de_get_random_dimension_index(self)
            crossover_length = 0
            do
                rnd = fgsl_rng_uniform(self%rng)    ! Random variable, uniform distribution in [0.,1.) range
                ! The greater the crossover_probability, the more likely the crossover is to end. This is because
                ! we define the crossover as being in the opposite direction as in Storn & Price.
                ! If the crossover_probability is high, then the crossover from member 2 (the population member)
                ! to member 1 (the perturbed member) will tend to be shorter. In other words, if crossover_probability
                ! is high, the perturbed member is more different from the population member.
                if (crossover_length .ge. num_dim .or. rnd .ge. (1.0-CR)) exit
                crossover_length = crossover_length + 1
            enddo

            ! Do the crossover
            current_dimension = crossover_start
            do I=CROSSOVER_START,CROSSOVER_START+CROSSOVER_LENGTH
                current_dimension = mod(i,num_dim)
                if (current_dimension .eq. 0) CURRENT_DIMENSION = num_dim
                member1%val(current_dimension) = member2%val(current_dimension)
            enddo

        end subroutine de_individuals_crossover

        !>  \brief  Check whether the minimisation is done
        logical function de_minimisation_done(self)
            implicit none
            ! Arguments
            type(de_minimiser), intent(in)  ::  self
            ! Private variables
            real, parameter ::  min_separation_btw_best_and_average_score = 0.002
            ! Start work
            de_minimisation_done = (self%current_generation .ge. self%max_num_generations) .or. &
                                   (self%average_score - self%best_member%score) .le. min_separation_btw_best_and_average_score
        end function de_minimisation_done

        !>  \brief  Come up with a default value for the number of members in the population
        !!
        !!  We use the rule of thumb in ﻿Storn, R. (1996). On the Usage of Differential Evolution for Function Optimization. in NAFIPS’96, 519 - 523.
        integer function de_suggested_population_size(self,num_dim)
            implicit none
            ! Arguments
            type(de_minimiser), intent(in)  ::  self
            integer,            intent(in)  ::  num_dim !<  Number of dimensions of the minimisation problem
            ! Private variables

            ! Start work
            de_suggested_population_size = num_dim * 10
        end function de_suggested_population_size

        !>  \brief  Return the population size
        pure integer function de_num_dim(self)
            implicit none
            ! Arguments
            class(de_minimiser), intent(in)  ::  self
            !
            de_num_dim = size(self%best_member%val)
        end function de_num_dim

        !>  \brief  Return the number of dimensions
        pure integer function de_population_size(self)
            implicit none
            ! Arguments
            class(de_minimiser), intent(in)  ::  self
            !
            de_population_size = size(self%population%members)
        end function de_population_size


        !>  \brief  Pick several's members' indices at random while optionally avoiding the index value given as argument
        subroutine de_get_mulitple_random_member_indices(self,num_to_get,random_indices,index_to_avoid)
            implicit none
            ! Arguments
            type(de_minimiser),         intent(in)  ::  self
            integer,                    intent(in)  ::  num_to_get                  !<  Desired number of indices
            integer,    allocatable,    intent(out) ::  random_indices(:)           !<  Resulting indices
            integer,                    intent(in)  ::  index_to_avoid              !<  Not allowed to return this index
            ! Private variables
            integer                 ::  current_output
            logical                 ::  check_one, check_two
            logical,    parameter   ::  debug = .false.
            integer                 ::  i
            ! Start work

            ! Allocate memory for the result
            allocate(random_indices(num_to_get))

            ! We loop until all conditions have been met
            current_output = 1
            do
                random_indices(current_output) = de_get_random_member_index(self)
                check_one = .true.
                check_two = .true.
                ! Check whether the current value is different from the previous ones
                if (current_output .gt. 1) then
                    do I=1,CURRENT_OUTPUT-1
                        if (random_indices(current_output) .eq. random_indices(i)) check_one = .false.
                    enddo
                endif
                ! Check it's different from the index we've been told to avoid
                if (random_indices(current_output) .eq. index_to_avoid) check_two = .false.
                ! If we were succesful, we move on to the next one
                if (check_one .and. check_two) then
                    if (current_output .eq. num_to_get) then
                        ! We're done
                        exit
                    else
                        current_output = current_output + 1
                    endif
                else
                    ! For some reason, we were not successful. Let's try again.
                endif
            enddo

            ! Debug
            if (debug) then
                write(*,'(a,i0)') '**debug(de_get_multiple_random_member_indices): was told to avoid this index: ', index_to_avoid
                write(*,'(a)') '**debug(de_get_multiple_random_member_indices): the following indices were selected: '
                write(*,*)  random_indices
            endif
        end subroutine de_get_mulitple_random_member_indices


        !>  \brief  Pick an individual's index at random
        integer function de_get_random_member_index(self) result(member_index)
            implicit none
            ! Arguments
            type(de_minimiser),    intent(in)  ::  self
            ! Private variables

            ! Start work

            ! Get an integer (uniformly distributed between 0 and the integer specified in second argument)
            member_index = fgsl_rng_uniform_int(self%rng,int(size(self%population%members)-1,c_long)) + 1


        end function de_get_random_member_index

        !>  \brief  Pick an dimension's index at random
        integer function de_get_random_dimension_index(self) result(dimension_index)
            implicit none
            ! Arguments
            type(de_minimiser),    intent(in)  ::  self
            ! Private variables

            ! Start work

            ! Get an integer (uniformly distributed between 0 and the integer specified in second argument)
            dimension_index = fgsl_rng_uniform_int(self%rng,int(size(self%population%members(1)%val)-1,c_long)) + 1


        end function de_get_random_dimension_index

        !>  \brief  Assignment operation from one individual to another
        subroutine de_individual_assignment(ind1,ind2)
            implicit none
            ! Arguments
            type(de_individual),    intent(inout)   ::  ind1
            type(de_individual),    intent(in)      ::  ind2
            ! Private variables

            ! Start work

            ! The lhs will be overwritten
            if (allocated(ind1%val)) deallocate(ind1%val)

            ! Copy data over
            if (allocated(ind2%val)) then
                allocate(ind1%val(size(ind2%val)))
                ind1%val = ind2%val
            endif
            ind1%score = ind2%score
            ind1%f     = ind2%f
            ind1%cr    = ind2%cr
        end subroutine de_individual_assignment

        !>  \brief  Print summary information about an individual
        subroutine de_individual_print_info(self)
            implicit none
            ! Arguments
            type(de_individual),    intent(in)  ::  self
            ! Private variables
            logical,    parameter   ::  debug   =   .false.
            character(len=50) :: fmt
            ! Start work
            write(fmt,'(a,i0,a,a)') '(a,F0.4,x,2(a,F0.2,x),a,',size(self%val),'(G0.2,x)',')'
            if (debug) write(*,'(2a)') '**debug(de_individual_print_info): printout format = ', trim(adjustl(fmt))
            write(*,fmt) 'Score= ', self%score, '; F= ', self%f, '; CR= ', self%cr, '; Values= ', self%val
        end subroutine de_individual_print_info

end module differential_evolution_mod
