!>  \brief  Differential geometry, primarily of space curves used to define filament axes
module differential_geometry
    use splines, only : interpol_spline
    use iso_c_binding, only : c_ptr
    implicit none

    !>  \brief  An arc-length-parametrised cubic spline space curve
    type space_curve
        type(interpol_spline)               ::  x_spline                !<  Spline modelling the x coordinate of the curve
        type(interpol_spline)               ::  y_spline                !<  Spline modelling the y coordinate of the curve
        type(interpol_spline)               ::  z_spline                !<  Spline modelling the z coordinate of the curve
        real(kind=8),         allocatable   ::  arc_length(:)           !<  At each waypoint, the arc length from the first waypoint
        real(kind=8),         allocatable   ::  target_arc_length(:)    !<  Desired arc lengths (only used occasionally)
        real(kind=8)                        ::  parametrisation_error   !<  Discrepency between the spline parametrisation
                                                                        !!  and the current estimates of the arc lengths
        real                                ::  param_error_tol         !<  Parametrisation error tolerance
        integer                             ::  n                       !<  Number of waypoints
        type(c_ptr)                         ::  cptr                    !<  a C-pointer pointing to itself.
                                                                        !!  This is needed when passing the object through the FGSL.
        contains
            final   ::  space_curve_deallocate
    end type

    contains

    subroutine differential_geometry_unit_tests()
        call differential_geometry_unit_test_1()
    end subroutine

    !>  \brief  Check the dx,dy,dz => Euler angle conversion
    !!  \todo   Rewrite so it doesn't rely on deprecated expected_d_signs_given_theta_psi
    subroutine differential_geometry_unit_test_1()
        use runtime_parameters, only : terminate
        real(kind=8)    ::  theta, psi
        real(kind=8)    ::  d(3)
        real(kind=8)    ::  theta_check_new, theta_check_old!, psi_check
        real(kind=8),   parameter   ::  angle_increment = 0.5d0
        ! start work
        psi   = -180.0d0
        theta = 0.0d0
        do
            psi = psi + angle_increment
            if (psi .gt. 180.0d0) exit
            do
                theta = theta + angle_increment
                if (theta .gt. 180.0d0) exit
                !
                d = d_from_theta_psi(theta,psi)
                theta_check_new = theta_from_dz(d(3))
                !theta_check_old = theta_from_dx_dy_dz(d(1),d(2),d(3))
                theta_check_old = theta_check_new
                if (abs(theta_check_new - theta_check_old) .gt. 0.001d0) then
                    print *, mod(theta,360.0d0), theta_check_new, theta_check_old
                    call terminate('failed differential geometry unit test')
                endif
            enddo
        enddo
        write(*,'(a)') '**info(differential_geometry_unit_test_1): passed'
    end subroutine differential_geometry_unit_test_1

    !>  \brief  Adjust the z coordinates of the waypoints until the desired arc lengths have been achieved
    !!
    !!  The first waypoint will be left untouched
    subroutine space_curve_adjust_z_to_match_arc_lengths(self,target_arc_length)
        use minimisers, only : ds_init, ds_run, minimiser_ds_fgsl, ds_max_iterations, ds_set_start_step
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(space_curve),  intent(inout)   ::  self
        real(kind=8),       intent(in)      ::  target_arc_length(:)    !<  Target arc length at each waypoint
        ! Private variables
        type(minimiser_ds_fgsl)                 ::  ds_minimiser
        real(kind=8),           allocatable     ::  first_step(:), start_values(:)
        real(kind=8),           allocatable     ::  best_values(:)
        real(kind=8)                            ::  best_score
        logical,                parameter       ::  incremental_z = .false.
        logical,				parameter		::	debug = .false.
		! Start work

        ! Checks
        if (size(target_arc_length) .ne. self%n) then
            write(*,'(a,2(i0,1x))') '**ERROR(space_curve_adjust_z_to_get_arc_lengths): size(target_arc_length) .NE. self%n: ', &
                                    size(target_arc_length), self%n
            call terminate('Fatal error in SPACE_CURVE_ADJUST_Z_TO_GET_ARC_LENGTHS')
        endif
        if (target_arc_length(1) .ne. 0.0) then
            write(*,'(a,G15.3)') '**ERROR(space_curve_adjust_z_to_get_arc_lengths): ', &
                                    ' target arc length for first waypoint non zero: ', target_arc_length(1)
            call terminate('Fatal error in SPACE_CURVE_ADJUST_Z_TO_GET_ARC_LENGTHS')
        endif

        ! Set the target arc length
        self%target_arc_length(:) = target_arc_length

		if (debug) then
			write(*,'(a)') '**debug(space_curve_adjust_z_to_match_arc_lengths): on entry, space curve info:'
			call space_curve_print_info(self)
		endif

        ! Set up the minimiser
        call ds_init(   ds_minimiser,space_curve_new_z_deviation_from_target_arc_length_wrap,self%n-1,self%cptr,  &
                        initial_score=real(space_curve_deviation_from_target_arc_length(self)), &
                        random = .false.,max_num_iter=1000 )

        ! Work out the initial values and the first step
        allocate(first_step(self%n-1),start_values(self%n-1))
        if (incremental_z) call space_curve_z_abs_to_incr(self)
        start_values(:) = self%z_spline%y(2:)
        if (incremental_z) call space_curve_z_incr_to_abs(self)
        first_step(:)   = 10.0D0
        call ds_set_start_step(ds_minimiser,start_values,first_step)

        ! Run the minimiser
        call ds_run(ds_minimiser,target_size=0.5)

        ! Get the results
        best_values = ds_minimiser%current_best_fgsl
        best_score = ds_minimiser%current_minimum_fgsl

        ! Plug the results back into the space curve
        if (incremental_z) call space_curve_z_abs_to_incr(self)
        self%z_spline%y(2:) = best_values(:)
        if (incremental_z) call space_curve_z_incr_to_abs(self)
        call space_curve_update_parametrisation(self)

        ! Write out result
		if (debug) then
        	write(*,'(a)') '**debug(space_curve_adjust_z_to_match_arc_lengths): on exit, space curve info:'
       		call space_curve_print_info(self)
		endif
    end subroutine space_curve_adjust_z_to_match_arc_lengths

    !>  \brief  Given z values for the waypoints, how far are we from the target arc lengths?
   function space_curve_new_z_deviation_from_target_arc_length_wrap(values,params) result(dev) bind(c)
        use iso_c_binding
        use fgsl, only : fgsl_vector, fgsl_double, fgsl_vector_align, fgsl_success, fgsl_obj_c_ptr
        use runtime_parameters, only : terminate
        implicit none
        type(c_ptr),        value       ::  values  !<  c pointer to current values to try out
        type(c_ptr),        value       ::  params  !<  c pointer to a filament object
        ! Result
        real(kind=C_DOUBLE)             ::  dev
        ! Private variables
        type(space_curve),  pointer     ::  self
        logical,            parameter   ::  debug   =   .false.
        integer                         ::  status
        type(fgsl_vector)               ::  val_vec
        real(fgsl_double),  pointer     ::  val_fgsl(:)
        real(kind=8),       allocatable ::  val(:)
        character(len=6)                ::  size_val_c
        ! Start work

        ! Unpack the space curve object
        call c_f_pointer(params,self)

        ! Unpack the values
        call fgsl_obj_c_ptr(val_vec,values)
        status = fgsl_vector_align(val_fgsl,val_vec)
        if (status .ne. fgsl_success) call terminate('FGSL_VECTOR_ALIGN failed')
        allocate(val(size(val_fgsl)+1),stat=STATUS)
        if (status .ne. 0) call terminate('Memory allocation failed')
        val(2:) = val_fgsl
        val(1)  = self%z_spline%y(1)

        ! Sanity check
        if (size(val) .ne. self%n) then
            write(*,'(2(A,i0))') '**ERROR(space_curve_new_z_deviation_from_target_arc_length_wrap): ', size(val), &
                                 ' values retrieved, but was expecting ', self%n
            call terminate('Fatal error in SPACE_CURVE_NEW_Z_DEVIATION_FROM_TARGET_ARC_LENGTH_WRAP')
        endif

        ! Get the value
        dev = space_curve_new_z_deviation_from_target_arc_length(self,val)

        ! Let the user know what values we're trying
        if (debug) then
            write(size_val_c,'(i6)') size(val)
            write(*,'(2a,'//size_val_c//'(F0.3,1x))') '**debug(space_curve_new_z_deviation_from_target_arc_length_wrap): ', &
			                                              'tried these values: ', val
            write(*,'(a,F0.3)') '**debug(space_curve_new_z_deviation_from_target_arc_length_wrap): got this score: ', dev
        endif
    end function space_curve_new_z_deviation_from_target_arc_length_wrap

    !>  \brief  Given z values for the waypoints, how far are we from the target arc lengths?
    real(kind=8) function space_curve_new_z_deviation_from_target_arc_length(self,z_vals) result(dev)
        implicit none
        ! Arguments
        type(space_curve),  intent(inout)   ::  self
        real(kind=8),       intent(in)      ::  z_vals(:)
        ! Private variables
        logical,                parameter       ::  incremental_z = .false.
        ! Start work
        if (incremental_z) call space_curve_z_abs_to_incr(self)
        call space_curve_replace_z_coos(self,z_vals)
        if (incremental_z) call space_curve_z_incr_to_abs(self)
        call space_curve_update_parametrisation(self)
        dev = space_curve_deviation_from_target_arc_length(self)
    end function space_curve_new_z_deviation_from_target_arc_length

    !>  \brief  How far are we from the target arc lengths?
    pure real(kind=8) function space_curve_deviation_from_target_arc_length(self) result(dev)
        implicit none
        ! Arguments
        type(space_curve),  intent(in)      ::  self
        ! Private variables
        integer ::  i
        ! Start work
        dev = 0.0D0
        do I=1,self%n
            dev = dev + abs(self%arc_length(i)-SELF%target_arc_length(i))
        enddo
    end function space_curve_deviation_from_target_arc_length

    !>  \brief  Unit tests for space curve
    subroutine space_curve_unit_tests()
        call space_curve_unit_test_1()
        call space_curve_unit_test_3()
        call space_curve_unit_test_4()
        call space_curve_unit_test_2()
    end subroutine space_curve_unit_tests

    !>  \brief  Generate a linear space curve and check that the parametrisation error is very very low
    subroutine space_curve_unit_test_1()
        use runtime_parameters, only : terminate
        use coos, only : waypoint
        implicit none
        ! Private variables
        type(waypoint)      ::  wps(5)
        integer             ::  i
        type(space_curve)   ::  self
        ! Start work
        write(*,'(a)') '**info(space_curve_unit_test_1): starting'
        do I=1,size(wps)
            wps(i)%coo(1:3) = 100.0 * i
        enddo
        call space_curve_init(self,wps,KNOWN_D=.false.,PARAM_ERROR_TOL=0.01)
        if (self%parametrisation_error .gt. 0.01) then
            call terminate('Failed unit test')
        else
            write(*,'(a)') '**info(space_curve_unit_test_1): passed'
        endif
    end subroutine space_curve_unit_test_1

    !>  \brief  Generate a linear space curve, run the minimiser with the modified arc lengths as target. Things shouldn't move.
    subroutine space_curve_unit_test_2()
        use runtime_parameters, only : terminate
        use coos, only : waypoint
        implicit none
        ! Private variables
        type(waypoint)      ::  wps(5)
        integer             ::  i
        type(space_curve)   ::  self
        real(kind=8), allocatable        ::  target_al(:)
        ! Start work
        write(*,'(a)') '**info(space_curve_unit_test_2): starting'
        do I=1,size(wps)
            wps(i)%coo(1:3) = 100.0 * i
        enddo
        call space_curve_init(self,wps,KNOWN_D=.false.,PARAM_ERROR_TOL=0.01)
        target_al = self%arc_length
        target_al(2) = target_al(2) + 10.0
        call space_curve_adjust_z_to_match_arc_lengths(self,TARGET_ARC_LENGTH=TARGET_AL)
        if (sum(abs(self%TARGET_ARC_LENGTH-SELF%arc_length)) .gt. 0.3) then
            call terminate('Failed unit test')
        else
            write(*,'(a)') '**info(space_curve_unit_test_2): passed'
        endif
    end subroutine space_curve_unit_test_2

    !>  \brief  Generate a linear and closed space curves and check that they have the expected total curvatures
    subroutine space_curve_unit_test_3()
        use runtime_parameters, only : terminate, pi
        use coos, only : waypoint
        implicit none
        ! Private variables
        type(waypoint)      ::  wps(50)
        integer             ::  i
        type(space_curve)   ::  self
        real                ::  measured_curvature
        real,   parameter   ::  circle_radius = 20.0
        real                ::  azi
        ! Start work
        write(*,'(a)') '**info(space_curve_unit_test_3): starting'
        ! Straight line case, total curvature should be zero
        write(*,'(a)') '**info(space_curve_unit_test_3): straight line test'
        do I=1,size(wps)
            wps(i)%coo(1:3) = 100.0 * i
        enddo
        call space_curve_init(self,wps,KNOWN_D=.false.,PARAM_ERROR_TOL=0.01)
		measured_curvature = space_curve_eval_total_curvature(self,STEP_SIZE=0.1D0)
        write(*,'(a,G15.3)') '**info(space_curve_unit_test_3): straight line total curvature = ', measured_curvature
        if (measured_curvature .gt. 0.01) then
            write(*,'(a)') '**ERROR(space_curve_unit_test_3): straight line curvature is too large'
            call terminate('Failed unit test')
        endif
        ! Circle on a plane case, total curvature should be 2*pi
        write(*,'(a)') '**info(space_curve_unit_test_3): planar circle test'
        wps(:)%coo(3) = 100.0
        do I=1,size(wps)
            azi = real(I-1)/real(size(wps)-1) * 2.0 * pi
            wps(i)%coo(1) = circle_radius * cos(azi)
            wps(i)%coo(2) = circle_radius * sin(azi)
            print *, '**debug(space_curve_unit_test_3): point #, azi, x, y', i, azi, wps(i)%coo(1:2)
        enddo
        call space_curve_init(self,wps,KNOWN_D=.false.,PARAM_ERROR_TOL=0.01)
        measured_curvature = space_curve_eval_total_curvature(self,STEP_SIZE=0.1D0)
        write(*,'(a,G15.3)') '**info(space_curve_unit_test_3): circle total curvature = ', measured_curvature
        if (abs(MEASURED_CURVATURE-2*PI) .gt. 0.01) then
            write(*,'(a,G15.3)') '**ERROR(space_curve_unit_test_3): circle curvature is not 2*PI, diff = ', &
                                                                        abs(MEASURED_CURVATURE-2*PI)
            call terminate('Failed unit test')
        endif
        write(*,'(a)') '**info(space_curve_unit_test_3): passed'
    end subroutine space_curve_unit_test_3

    !>  \brief  Check that "wobbly"-looking space curves get a higher total curvature than less wobbly ones
    subroutine space_curve_unit_test_4()
        use runtime_parameters, only : terminate, pi, is_odd
        use coos, only : waypoint
        implicit none
        ! Private variables
        type(waypoint)      ::  wps_1(5), wps_2(5)
        type(space_curve)   ::  sc_1, sc_2
        real                ::  curvature_1, curvature_2
        real,   parameter   ::  circle_radius = 50.0
        ! Start work
        write(*,'(a)') '**info(space_curve_unit_test_4): starting'

        ! Set up the two sets of waypoints (coos)
        wps_1(1)%coo = [2119.290,    1934.801,  -14111.840]
        wps_1(2)%coo = [2183.570,    1918.880,  -14100.175]
        wps_1(3)%coo = [2240.970,    1895.530,  -14089.247]
        wps_1(4)%coo = [2290.360,    1873.920,  -14080.311]
        wps_1(5)%coo = [2351.010,    1857.790,  -14070.626]
        wps_2(1)%coo = [2119.290,    1934.801,  -14111.840]
        wps_2(2)%coo = [2183.570,    1918.880,  -14060.631]
        wps_2(3)%coo = [2240.970,    1895.530,  -14108.365]
        wps_2(4)%coo = [2290.360,    1873.920,  -14072.440]
		wps_2(4)%coo = [2290.360,    1873.920,  -14372.440]
        wps_2(5)%coo = [2351.010,    1857.790,  -14023.620]
        ! Now setup theta and psi
        wps_1(1)%euler%theta = 82.
        wps_1(2)%euler%theta = 82.
        wps_1(3)%euler%theta = 82.
        wps_1(4)%euler%theta = 83.
        wps_1(5)%euler%theta = 83.
        wps_1(1)%euler%psi   = -20.36
        wps_1(2)%euler%psi   = -20.39
        wps_1(3)%euler%psi   = -20.33
        wps_1(4)%euler%psi   = -20.20
        wps_1(5)%euler%psi   = -20.08
        wps_2(1)%euler%theta = 82.
        wps_2(2)%euler%theta = 82.
        wps_2(3)%euler%theta = 82.
        wps_2(4)%euler%theta = 83.
        wps_2(5)%euler%theta = 83.
        wps_2(1)%euler%psi   = -20.36
        wps_2(2)%euler%psi   = -20.39
        wps_2(3)%euler%psi   = -20.33
        wps_2(4)%euler%psi   = -20.20
        wps_2(5)%euler%psi   = -20.08


        call space_curve_init(sc_1,wps_1,KNOWN_D=.true.,PARAM_ERROR_TOL=0.01)
        call space_curve_init(sc_2,wps_2,KNOWN_D=.true.,PARAM_ERROR_TOL=0.01)
        curvature_1 = space_curve_eval_total_curvature(sc_1,STEP_SIZE=0.1D0) * sc_1%arc_length(sc_1%n)
        curvature_2 = space_curve_eval_total_curvature(sc_2,STEP_SIZE=0.1D0) * sc_2%arc_length(sc_2%n)
        call space_curve_print_info(sc_1)
        call space_curve_print_info(sc_2)
        write(*,'(a,2(G15.3,1x))') '**info(space_curve_unit_test_4): curvatures = ', curvature_1,curvature_2
        if (curvature_1 .gt. curvature_2) then
            write(*,'(a)') '**ERROR(space_curve_unit_test_4): curvature got smaller when curve got wobblier'
            call terminate('Failed unit test')
        endif

    end subroutine space_curve_unit_test_4

    !>  \brief  Update a space curve with new points (and, optionally, Euler angles)
    subroutine space_curve_update(self,waypoints)
        use coos, only : waypoint
        implicit none
        ! Arguments
        type(space_curve),  intent(inout)   ::  self
        type(waypoint),     intent(in)      ::  waypoints(:)    !<  Points through which the space curve travels
        ! Private variables

        ! Start work
        call space_curve_get_data_from_waypoints(self,waypoints,self%x_spline%FirstDerivativesAreKnown())
        call space_curve_update_parametrisation(self)
    end subroutine space_curve_update

    !>  \brief  Initialise a space cruve
    subroutine space_curve_init(self,waypoints,known_d,param_error_tol)
        use runtime_parameters, only : terminate
        use coos, only : waypoint
        use Splines
        use iso_c_binding, only : c_loc
        implicit none
        ! Arguments
        type(space_curve),  target, intent(inout)   ::  self
        type(waypoint),             intent(in)      ::  waypoints(:)    !<  Points through which the space curve travels
        logical,                    intent(in)      ::  known_d         !<  Whether we use splines which are solved using known
                                                                        !!  1st derivatives derived from the waypoints' Euler angles
        real                                        ::  param_error_tol !<  Tolerated parametrisation error
        ! Private variables
        integer ::  min_num_wps
        ! Start work

        ! Parametrisation error tolerance
        self%param_error_tol = param_error_tol

        ! Check size of waypoints
        if (known_d) then
            min_num_wps = 2
        else
            min_num_wps = 3
        endif
        if (size(waypoints) .lt. min_num_wps) then
            write(*,'(a,i0,a)') '**ERROR(space_curve_init): number of wps (', size(waypoints), ') not sufficient'
            call terminate('Fatal error in SPACE_CURVE_INIT')
        endif

        ! Allocate memory
        call space_curve_allocate(self,size(waypoints))

        ! Set up the c pointer to self
        self%cptr = c_loc(self)

        ! Copy data over from waypoints
        call space_curve_get_data_from_waypoints(self,waypoints,known_d)

        ! First guess at arc length
        call space_curve_guess_arc_length(self)

        ! Copy arc length values to x coordinates of splines
        call space_curve_arc_length_to_splines(self)

        ! Set spline type
        if (known_d) then
            call self%x_spline%SetType('KNOWN_D')
            call self%y_spline%SetType('KNOWN_D')
            call self%z_spline%SetType('KNOWN_D')
        else
            call self%x_spline%SetType('PARABOLIC')
            call self%y_spline%SetType('PARABOLIC')
            call self%z_spline%SetType('PARABOLIC')
        endif

        ! Solve the splines for the first time
        call space_curve_solve_splines(self)

        ! Parametrise the space curve properly
        call space_curve_update_parametrisation(self)

        ! Let's zero the target arc lengths
        self%target_arc_length = 0.0D0

    end subroutine space_curve_init

    !>  \brief  Convert from absolute to relative z values
    pure subroutine space_curve_z_abs_to_incr(self)
        implicit none
        ! Arguemnts
        type(space_curve),  intent(inout)   ::  self
        ! Private variables
        integer ::  i
        ! Start work
        do I=SELF%n,2,-1
            self%z_spline%y(i) = self%z_spline%y(i) - self%z_spline%y(I-1)
        enddo
    end subroutine space_curve_z_abs_to_incr

    !>  \brief  Convert from absolute to relative z values
    pure subroutine space_curve_z_incr_to_abs(self)
        implicit none
        ! Arguemnts
        type(space_curve),  intent(inout)   ::  self
        ! Private variables
        integer ::  i
        ! Start work
        do I=2,self%n
            self%z_spline%y(i) = self%z_spline%y(i) + self%z_spline%y(I-1)
        enddo
    end subroutine space_curve_z_incr_to_abs

    !>  \brief  Replace the z values
    pure subroutine space_curve_replace_z_coos(self,new_z)
        implicit none
        ! Arguments
        type(space_curve),  intent(inout)   ::  self
        real(kind=8),       intent(in)      ::  new_z(:)
        !
        self%z_spline%y(:) = new_z(:)
    end subroutine space_curve_replace_z_coos

    !>  \brief  Copy data from waypoints to the space curve
    pure subroutine space_curve_get_data_from_waypoints(self,waypoints,known_d)
        use coos, only : waypoint
        implicit none
        ! Arguments
        type(space_curve),  intent(inout)   ::  self
        type(waypoint),     intent(in)      ::  waypoints(:)    !<  Points through which the space curve travels
        logical,            intent(in)      ::  known_d         !<  Whether we use splines which are solved using known
                                                                !!  1st derivatives derived from the waypoints' Euler angles
        ! Start work
        self%x_spline%y(:) = waypoints(:)%coo(1)
        self%y_spline%y(:) = waypoints(:)%coo(2)
        self%z_spline%y(:) = waypoints(:)%coo(3)
        if (known_d) then
            self%x_spline%dy(:) = dx_from_theta_psi(waypoints(:)%euler%theta,waypoints(:)%euler%psi)
            self%y_spline%dy(:) = dy_from_theta_psi(waypoints(:)%euler%theta,waypoints(:)%euler%psi)
            self%z_spline%dy(:) = dz_from_theta(waypoints(:)%euler%theta)
        endif
    end subroutine space_curve_get_data_from_waypoints

    !>  \brief  Update the arc-length parametrisation of the space curve
    subroutine space_curve_update_parametrisation(self)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(space_curve),  intent(inout)   ::  self
        ! Private variables
        integer                 ::  parametrisation_num_iterations
        real                    ::  previous_parametrisation_error
        integer,    parameter   ::  parametrisation_max_num_iterations  =   300
        integer,    parameter   ::  parametrisation_min_num_iterations  =   2
        !real(KIND=8)            ::  parametrisation_target_error
        !real(KIND=8)            ::  current_parametrisation_error
        ! Start work

        ! Iterative parametrisation
        parametrisation_num_iterations = 0
        previous_parametrisation_error = huge(1.0)
        do
            parametrisation_num_iterations = parametrisation_num_iterations + 1
            call space_curve_arc_length_to_splines(self)
            call space_curve_solve_splines(self)
            call space_curve_update_arc_length(self)
            call space_curve_update_parametrisation_error(self)

            if (    self%parametrisation_error .lt. self%param_error_tol &
                    .and. parametrisation_num_iterations .ge. parametrisation_min_num_iterations &
                    ) then
                exit
            endif
            previous_parametrisation_error = self%parametrisation_error
            if (parametrisation_num_iterations .gt. parametrisation_max_num_iterations) then
                write(*,'(a,i4,2(a,F13.10))')   '**warning(space_curve_update_parametrisation): spline parametrisation ',  &
                                                'did not converge after ', parametrisation_num_iterations, &
                                                        ' iterations. Current error = ', self%parametrisation_error, &
                                                        '. Target error =', self%param_error_tol
                if (parametrisation_num_iterations .gt. 2*PARAMETRISATION_MAX_NUM_ITERATIONS) then
                    call terminate('Fatal error in FILAMENT_PARAMETRISATION_UPDATE')
                endif
            endif
        enddo
    end subroutine space_curve_update_parametrisation


    !>  \brief  Compute the discrepency between the parameter used by the splines and the current estimate of the arc length
    pure subroutine space_curve_update_parametrisation_error(self)
        implicit none
        ! Arguments
        type(space_curve),  intent(inout)   ::  self
        ! Private variables
        integer ::  i
        ! Start work
        self%parametrisation_error = 0.0D0
        do I=1,self%n
            self%parametrisation_error = self%parametrisation_error + abs(self%arc_length(i)-SELF%x_spline%x(i))
        enddo
        self%parametrisation_error = self%parametrisation_error / self%n
    end subroutine space_curve_update_parametrisation_error

    !>  \brief  Compute the arc length from first waypoint to every other waypoint in the space curve.
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine space_curve_update_arc_length(self)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(space_curve), intent(inout)   ::  self    !<  Filament to be updated
        ! Private variables
        integer ::  i
        logical,    parameter   ::  debug   =   .false.
        character(len=6) :: nwp_c
        ! Start work
        ! Start at the origin
        self%arc_length(1) = 0.0D0
        do I=2,self%n
            call space_curve_eval_arc_length(self,self%x_spline%x(i),self%arc_length(i))
        enddo
#ifndef skip_runtime_checks
        if (debug) then
            write(nwp_c,'(i6)') self%n
            if (any(self%arc_length .ne. 0.)) then
                ! Sanity check
                do I=2,self%n
                    if (self%arc_length(i) .le. self%arc_length(I-1)) then
                        write(*,'(a,'//nwp_c//'(F0.3,x))') '**ERROR(space_curve_update_arc_length): weird arc lengths: ', &
                                self%arc_length
                        call terminate('Fatal error in SPACE_CURVE_UPDATE_ARC_LENGTH')
                    endif
                enddo
            endif
        endif
#endif
    end subroutine space_curve_update_arc_length

    !>  \brief  Compute the arc length from first waypoint to the point specified by the coordinate t, using numerical integration.
    !!
    !!  Should only be called when t falls directly after a waypoint for which the path_length has already been determined
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine space_curve_eval_arc_length(self,t,arc_length)
        use Splines
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(space_curve),      intent(inout)   ::  self            !<  Filament
        real(kind=8),           intent(in)      ::  t               !<  Parameter value for which the path length needs to be computed
        real(kind=8),           intent(out)     ::  arc_length     !<  Result. Arc length.
        ! Private variables
        real(kind=8),           parameter       ::  tiny_step = 0.0005D0    !<  Step for the integration
        integer                                 ::  seg
        real(kind=8)                            ::  current_t,current_x,current_y,current_z,previous_x,previous_y,previous_z
        logical,                parameter       ::  debug   =   .false.
        ! Start work
        ! Figure out which segment of the spline we'll be working on
        seg = self%x_spline%FindSegment(t)
		if (debug) then
#ifndef skip_runtime_checks
        write(*,'(a,F0.3,1x,i0)') '**debug(space_curve_eval_arc_length): t, seg = ', t, seg
        if (seg .eq. 0) then
            write(*,'(a,F0.5)') '**ERROR(space_curve_eval_arc_length): could not find spline segment for t = ', t
            call self%x_spline%PrintInfo()
            call terminate('Fatal error in SPACE_CURVE_EVAL_ARC_LENGTH')
        endif
#endif
        endif
        ! Start at the start of that segment
        arc_length = self%arc_length(seg)
        current_t = self%x_spline%x(seg)
        current_x = self%x_spline%y(seg)
        current_y = self%y_spline%y(seg)
        current_z = self%z_spline%y(seg)

        ! Loop until we've reached the point we're interested in
        do
            ! Keep track of where we've just been
            previous_x = current_x
            previous_y = current_y
            previous_z = current_z
            ! One tiny step for man...
            current_t = current_t + tiny_step
            ! Check we've not stepped over the mark
            if (current_t .gt. t) then
                exit    !   Done, get out
            endif
            ! We're still walking - work out where our foot lands
            call self%x_spline%Evaluate(current_t,current_x)
            call self%y_spline%Evaluate(current_t,current_y)
            call self%z_spline%Evaluate(current_t,current_z)
            ! Measure & accumulate the distance
            arc_length = arc_length + sqrt((current_x-previous_x)**2+(current_y-previous_y)**2+(current_z-previous_z)**2)
        enddo
    end subroutine space_curve_eval_arc_length

    !>  \brief  Evaluate the total curvature along a space curve. Optionally, the total squared curvature can be returned.
    pure function space_curve_eval_total_curvature_new(self,step_size,squared_curvature,ignore_fal_wps) result(total_curvature)
        use Splines
        implicit none
        ! Arguments
        type(space_curve),          intent(in)  ::  self
        real,                       intent(in)  ::  step_size           !<  Integration step size (good default for a
                                                                        !!  filament axisis the helical rise per subunit)
        logical,        optional,   intent(in)  ::  squared_curvature   !<  Return the total squared curvature. Defaults to false,
                                                                        !!  in which case the total curvature (not squared)
                                                                        !!  is returned.
        logical,        optional,   intent(in)  ::  ignore_fal_wps      !<  Ignore the first and last waypoints
                                                                        !!  in the calculation. False by default.
        ! Result
        real(kind=8)    ::  total_curvature
        ! Private variables
        logical         ::  ssquared_curvature
        real(kind=8)    ::  current_t
        logical         ::  iignore_fal_wps
        integer         ::  first_wp, last_wp
        ! Start work

        ! Will we be returning the squared curvature?
        ssquared_curvature = .false.
        if (present(squared_curvature)) ssquared_curvature = squared_curvature

        ! Ignore fal WPS?
        iignore_fal_wps = .false.
        if (present(ignore_fal_wps)) iignore_fal_wps = ignore_fal_wps

        ! First and last waypoints
        if (iignore_fal_wps) then
            first_wp    =   2
            last_wp     =   self%n - 1
        else
            first_wp    =   1
            last_wp     =   self%n
        endif

        ! Total curvature
        total_curvature = 0.0D0
        current_t = self%x_spline%x(first_wp)
        if (ssquared_curvature) then
!            do
!                total_curvature = total_curvature + space_curve_eval_curvature(self,current_t)**2
!                current_t = current_t + step_size
!                if (current_t .GT. self%x_spline%x(last_wp)) exit
!            enddo
        else
            do
                total_curvature = total_curvature + sqrt(   self%x_spline%EvaluateSecondDerivative(current_t)**2 &
                                                        +   self%y_spline%EvaluateSecondDerivative(current_t)**2 &
                                                        +   self%z_spline%EvaluateSecondDerivative(current_t)**2 ) &
	 													* sign(1.0D0,space_curve_eval_torsion(self,current_t))
                !total_curvature = total_curvature + sqrt(  space_curve_eval_curvature(self,current_t)**2 &
				!                                         + space_curve_eval_torsion(self,current_t)**2   )
				current_t = current_t + step_size
                if (current_t .gt. self%x_spline%x(last_wp)) exit
            enddo
        endif
        total_curvature = total_curvature * step_size
    end function space_curve_eval_total_curvature_new

    !>  \brief  Evaluate the total curvature along a space curve. Optionally, the total squared curvature can be returned.
    pure function space_curve_eval_total_curvature(self,step_size,squared_curvature,ignore_fal_wps) result(total_curvature)
        implicit none
        ! Arguments
        type(space_curve),          intent(in)  ::  self
        real(kind=8),               intent(in)  ::  step_size           !<  Integration step size (good default for a
                                                                        !!  filament axisis the helical rise per subunit)
        logical,        optional,   intent(in)  ::  squared_curvature   !<  Return the total squared curvature. Defaults to false,
                                                                        !!  in which case the total curvature (not squared)
                                                                        !!  is returned.
        logical,        optional,   intent(in)  ::  ignore_fal_wps      !<  Ignore the first and last waypoints
                                                                        !!  in the calculation. False by default.
        ! Result
        real(kind=8)    ::  total_curvature
        ! Private variables
        logical         ::  ssquared_curvature
        real(kind=8)    ::  current_t
        logical         ::  iignore_fal_wps
        integer         ::  first_wp, last_wp
        ! Start work

        ! Will we be returning the squared curvature?
        ssquared_curvature = .false.
        if (present(squared_curvature)) ssquared_curvature = squared_curvature

        ! Ignore fal WPS?
        iignore_fal_wps = .false.
        if (present(ignore_fal_wps)) iignore_fal_wps = ignore_fal_wps

        ! First and last waypoints
        if (iignore_fal_wps) then
            first_wp    =   2
            last_wp     =   self%n - 1
        else
            first_wp    =   1
            last_wp     =   self%n
        endif

        ! Total curvature
        total_curvature = 0.0D0
        current_t = self%x_spline%x(first_wp)
        if (ssquared_curvature) then
            do
                total_curvature = total_curvature + space_curve_eval_curvature(self,current_t)**2
                current_t = current_t + step_size
                if (current_t .gt. self%x_spline%x(last_wp)) exit
            enddo
        else
            do
                total_curvature = total_curvature + space_curve_eval_curvature(self,current_t)
                current_t = current_t + step_size
                if (current_t .gt. self%x_spline%x(last_wp)) exit
            enddo
        endif
        total_curvature = total_curvature * step_size
    end function space_curve_eval_total_curvature

	!>	\brief	Evaluate the local torsion
	pure function space_curve_eval_torsion(self,t) result(torsion)
		use Splines
		implicit none
		! Arguments
		type(space_curve),  intent(in)  ::  self
        real(kind=8),       intent(in)  ::  t
        ! Result
        real(kind=8)    ::  torsion
        ! Private variables
        real(kind=8)    ::  dx,dy,dz
        real(kind=8)    ::  ddx,ddy,ddz
        real(kind=8)    ::  dddx,dddy,dddz
        real(kind=8)    ::  a,b,c
        ! Start work

        dx  =   self%x_spline%EvaluateDerivative(t)
        dy  =   self%y_spline%EvaluateDerivative(t)
        dz  =   self%z_spline%EvaluateDerivative(t)
        ddx =   self%x_spline%EvaluateSecondDerivative(t)
        ddy =   self%y_spline%EvaluateSecondDerivative(t)
        ddz =   self%z_spline%EvaluateSecondDerivative(t)
        dddx =  self%x_spline%EvaluateThirdDerivative(t)
        dddy =  self%y_spline%EvaluateThirdDerivative(t)
        dddz =  self%z_spline%EvaluateThirdDerivative(t)

        a = dy*ddz - ddy*dz
        b = ddx*dz - dx*ddz
        c = dx*ddy - ddx*dy

        torsion =     (dddx*a + dddy*b + dddz*c) &
                    / ( a**2 + b**2 + c**2 )
	end function space_curve_eval_torsion

    !>  \brief  Evaluate the local curvature
    pure function space_curve_eval_curvature(self,t) result(curvature)
        use Splines
        implicit none
        ! Arguments
        type(space_curve),  intent(in)  ::  self
        real(kind=8),       intent(in)  ::  t
        ! Result
        real(kind=8)    ::  curvature
        ! Private variables
        real(kind=8)    ::  dx,dy,dz
        real(kind=8)    ::  ddx,ddy,ddz
        logical,	parameter	::	assume_arc_length_parametrisation = .false.
		! Start work

		if (.not. assume_arc_length_parametrisation) then
	        dx  =   self%x_spline%EvaluateDerivative(t)
        	dy  =   self%y_spline%EvaluateDerivative(t)
    	    dz  =   self%z_spline%EvaluateDerivative(t)
		endif
        ddx =   self%x_spline%EvaluateSecondDerivative(t)
        ddy =   self%y_spline%EvaluateSecondDerivative(t)
        ddz =   self%z_spline%EvaluateSecondDerivative(t)

    	if (assume_arc_length_parametrisation) then
		    curvature = sqrt(ddx**2+ddy**2+ddz**2)
		else
			curvature = sqrt((ddz*dy-ddy*dz)**2+(ddx*dz-ddz*dx)**2+(ddy*dx-ddx*dy)**2)  &
        	            /   &
            	        ((dx**2+dy**2+dz**2)**(3/2))
		endif

    end function space_curve_eval_curvature

    !>  \brief  Solve space curve splines
    subroutine space_curve_solve_splines(self)
        use Splines
        implicit none
        ! Arguments
        type(space_curve),  intent(inout)   ::  self
        ! Private variables

        ! Start work
		call self%x_spline%Solve()
		call self%y_spline%Solve()
        call self%z_spline%Solve()
    end subroutine space_curve_solve_splines

    !>  \brief  Copy the arc length values to the "X" parameter value of each spline
    pure subroutine space_curve_arc_length_to_splines(self)
        implicit none
        ! Arguments
        type(space_curve),  intent(inout)   ::  self
        ! Start work
        self%x_spline%x(:) = self%arc_length(:)
        self%y_spline%x(:) = self%arc_length(:)
        self%z_spline%x(:) = self%arc_length(:)
    end subroutine space_curve_arc_length_to_splines

    !>  \brief  Guess the arc length of the space curve without using the splines. Use the chord length approximation.
    pure subroutine space_curve_guess_arc_length(self)
        implicit none
        ! Arguments
        type(space_curve),  intent(inout)   ::  self
        ! Private variables
        integer ::  i
        ! Start work
        self%arc_length(1) = 0.0
        do I=2,self%n
            self%arc_length(i) = self%arc_length(I-1) &
                                 +  sqrt(   (self%x_spline%y(i) - self%x_spline%y(I-1))**2 &
                                         +  (self%y_spline%y(i) - self%y_spline%y(I-1))**2 &
                                         +  (self%z_spline%y(i) - self%z_spline%y(I-1))**2    )
        enddo
    end subroutine space_curve_guess_arc_length

    !>  \brief  Deallocate memory for a space curve
    !impure elemental
    subroutine space_curve_deallocate(self)
        use Splines
        implicit none
        ! Arguments
        type(space_curve),  intent(inout)   ::  self
        ! Private variables

        ! Start work
        call self%x_spline%Deallocate()
        call self%y_spline%Deallocate()
        call self%z_spline%Deallocate()
        if (allocated(self%arc_length)) deallocate(self%arc_length)
    end subroutine space_curve_deallocate

    !>  \brief  Allocate memory for a space curve
    subroutine space_curve_allocate(self,n)
        use runtime_parameters, only : terminate
        use Splines
        implicit none
        ! Arguments
        type(space_curve),  intent(inout)   ::  self
        integer,            intent(in)      ::  n       !<  Number of waypoints
        ! Private variables

        ! Start work
        ! Check
        if (n .lt. 1) then
            write(*,'(a,i0)') '**ERROR(space_curve_allocate): invalid number of waypoints: ', n
            call terminate('Fatal error in SPACE_CURVE_ALLOCATE')
        endif
        ! Allocate
        call self%x_spline%Allocate(n,3)
        call self%y_spline%Allocate(n,3)
        call self%z_spline%Allocate(n,3)
        if (allocated(self%arc_length)) deallocate(self%arc_length)
        if (allocated(self%target_arc_length)) deallocate(self%target_arc_length)
        allocate(self%arc_length(n),self%target_arc_length(n))
        self%n = n
    end subroutine space_curve_allocate

    !>  \brief  Print summary information about the space curve
    subroutine space_curve_print_info(self)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(space_curve),  intent(in)  ::  self
        ! Private variables
        character(len=2)    :: n_c
        integer             :: i
        ! Start work
        write(n_c,'(i2)') self%n

        write(*,'(a)')                  'Space curve info:'
        write(*,'(2x,a,i0)')            'Number of waypoints: ', self%n
        write(*,'(2x,2(a,F0.5),a)')     'Parametrisation error (tolerance): ', self%parametrisation_error, &
                                                                              ' (', self%param_error_tol, ')'
        write(*,'(2x,a,F0.3)')          'Arc/chord length: ', self%arc_length(self%n) / space_curve_chord_length(self)
		write(*,'(2x,a3,1x,12(a15,1x))') 'WP#', 'T', 'ArcLength', 'TargetArcLength', 'X', 'Y', 'Z','DX','DY','DZ','DDX','DDY','DDZ'
        do I=1,self%n
            write(*,'(2x,i3,1x,12(F15.3,1x))') i,  self%x_spline%x(i), self%arc_length(i), self%target_arc_length(i),   &
                                                self%x_spline%y(i), self%y_spline%y(i), self%z_spline%y(i), &
                                                self%x_spline%dy(i), self%y_spline%dy(i), self%z_spline%dy(i), &
                                                self%x_spline%ddy(i), self%y_spline%ddy(i), self%z_spline%ddy(i)
        enddo
        write(*,'(a)') ' '
    end subroutine space_curve_print_info

	pure real function space_curve_chord_length(self)
		implicit none
		! Arguments
		type(space_curve),	intent(in)	::	self
		! Start work
		space_curve_chord_length = sqrt( (self%x_spline%y(self%n)-SELF%x_spline%y(1))**2 &
		                                +(self%y_spline%y(self%n)-SELF%y_spline%y(1))**2 &
										+(self%z_spline%y(self%n)-SELF%z_spline%y(1))**2 )
	end function space_curve_chord_length


    !>  \brief  Return an array of waypoints which describes the path of the filament in the region of its waypoints.
    !!  We make the assumption that near the waypoints, the filament is perfectly straight.
    function waypoint_get_path_near_wp (helix,step,num_wps_per_wp,current_wp, &
                                        step_in_repeats, &
                                        start_at_input_wp,finish_at_input_wp) result(path)
        use runtime_parameters, only : is_even, terminate
        use helices, only : helical_symmetry_unit_convert, helical_symmetry_print_info, helical_symmetry
        !use differential_geometry, only : d_from_theta_psi
        use coos, only : waypoints_print_info, wp_print_info, waypoint, waypoint_apply_symmetry
        implicit none
        ! Arguments
        type(helical_symmetry),             intent(in)      ::  helix                   !<  Helical parameters. Rise must be in pixel units.
        real(kind=8),                       intent(in)      ::  step                    !<  Step size (pixels)
        integer,                            intent(in)      ::  num_wps_per_wp          !<  Each incoming wp will "generate" num_wps_per_wp waypoints in the output path, including itself. This number does not include symmetry-related views.
        type(waypoint),                     intent(in)      ::  current_wp              !<  Waypoint around which to get path
        logical,                optional,   intent(in)      ::  step_in_repeats         !<  Whether the step argument was given in repeats
        logical,                optional,   intent(in)      ::  start_at_input_wp       !<  First output waypoint will coincide with the waypoint of interest. Useful when dealing with the first waypoint of a filament, for example.
        logical,                optional,   intent(in)      ::  finish_at_input_wp      !<  Last output waypoint will coincide with the waypoint of interest. Useful when dealing with the first waypoint of a filament, for example.
        ! Result
        type(waypoint),         allocatable                 ::  path(:)                 !<  Array of waypoints
        ! Private variables
        real(kind=8)            ::  dist_to_input_wp
        integer                 ::  num_output_wps, current_output_wp
        integer                 ::  ierr
        integer                 ::  i
        real(kind=8)            ::  d(3)
        integer                 ::  axial_sym
        integer                 ::  perp_sym
        logical,    parameter   ::  debug   =   .false.
        real                    ::  current_twist, current_rise
        real(kind=8)            ::  sstep_in_pixels
        integer                 ::  current_wp_output_wps_start, current_wp_output_wps_finish
        logical                 ::  iintg_repeats, sstep_in_repeats
        !type(waypoint)          ::  current_wp
        !real(kind=8)            ::  current_t
        ! Start work


        ! num_wps_per_wp is assumed to be odd
        if (is_even(num_wps_per_wp)) then
            write(*,'(a,i0)') '**ERROR(waypoint_get_path_near_wp): num_wps_per_wp is not odd: ', num_wps_per_wp
            call terminate('Fatal error in waypoint_get_path_near_wp')
        endif


        ! Make sure the rise is in pixels
        if (.not. helix%pixel_units) then
            write(*,'(a)') '**ERROR(waypoint_get_path_near_wp): helical parameters not in pixel units'
            call terminate('Fatal error in waypoint_get_path_near_wp')
        endif

        ! Step in repeats?
        sstep_in_repeats = .false.
        if (present(step_in_repeats)) sstep_in_repeats = step_in_repeats


        ! How many points will we generate
        num_output_wps = num_wps_per_wp * helix%axial_symmetry

        ! If we're on the first or last waypoint, we don't extend further
        current_wp_output_wps_start = 1
        if (present(start_at_input_wp)) then
            if (start_at_input_wp) current_wp_output_wps_start = (num_wps_per_wp-1)/2 + 1
        endif
        current_wp_output_wps_finish = num_wps_per_wp
        if (present(finish_at_input_wp)) then
            if (finish_at_input_wp) current_wp_output_wps_finish = (num_wps_per_wp-1)/2 + 1
        endif


        num_output_wps = (current_wp_output_wps_finish - current_wp_output_wps_start + 1)


        ! Allocate memory for the output
        allocate(path(num_output_wps),stat=ierr)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(waypoint_get_path_near_wp): memory allocation failed with error ', ierr
            call terminate('Fatal error in FILAMENT_GET_PATH_NEAR_WPS')
        endif


        ! Loop over (input) waypoints
        current_output_wp   =   0


        ! Work out the twist and rise to use near this waypoint
        current_twist = helix%twist
        current_rise  = helix%rise

        ! If the step size was given in helical subunits, convert to pixels
        sstep_in_pixels = step
        if (sstep_in_repeats) sstep_in_pixels = step * current_rise

        ! The derivatives at the current input waypoint
        d = d_from_theta_psi(current_wp%euler%theta,current_wp%euler%psi)

        ! Loop over the output waypoints
        do i=current_wp_output_wps_start,current_wp_output_wps_finish

            ! work out distance (pixels) from current output wp to current input wp. negative if output waypoint is before input waypoint on the filament.
            dist_to_input_wp    =   sstep_in_pixels * real(i - ((num_wps_per_wp-1)/2+1))

            ! The output wp we're currently working on
            current_output_wp   =   current_output_wp + 1
            !
            path(current_output_wp)%euler%theta =   current_wp%euler%theta
            path(current_output_wp)%euler%psi   =   current_wp%euler%psi
            ! Work out the coordinates of the output waypoint
            path(current_output_wp)%coo(1:3)    =   dist_to_input_wp * d(1:3) + current_wp%coo(1:3)
            ! Phi needs to be worked out
            path(current_output_wp)%euler%phi   =   current_wp%euler%phi &
                                                    + dist_to_input_wp * current_twist / current_rise

            ! Helical parameters are the same
            path(current_output_wp)%helix       =   current_wp%helix
            path(current_output_wp)%helix%twist =   current_twist
            path(current_output_wp)%helix%rise  =   current_rise
            ! Refine flag
            path(current_output_wp)%refine      =   current_wp%refine
            ! Path length reset to 0.0
            path(current_output_wp)%path_length =   0.0
            ! Local subframe shifts
            path(current_output_wp)%local_subframe_shift = current_wp%local_subframe_shift
            ! Local subunit_num
            path(current_output_wp)%subunit_num =   current_wp%subunit_num + i - ((num_wps_per_wp-1)/2+1)

        enddo

    end function waypoint_get_path_near_wp

    !>  \brief  Return the derivatives in x,y,z from theta and psi
    pure function d_from_theta_psi(theta,psi) result(d)
        implicit none
        ! arguments
        real(kind=8),   intent(in)  ::  theta   !<  theta angle (degrees)
        real(kind=8),   intent(in)  ::  psi     !<  psi angle (degrees)
        ! result
        real    ::  d(3)
        !
        d(1) = dx_from_theta_psi(theta,psi)
        d(2) = dy_from_theta_psi(theta,psi)
        d(3) = dz_from_theta(theta)
    end function d_from_theta_psi


    !>  \brief  Compute the angle (in radians) between the two vectors designated by the two (theta,psi) pairs given
    pure real function angle_between_euler_angles(theta_1,psi_1,theta_2,psi_2) result (angle)
        real(kind=8),   intent(in)    ::  theta_1,psi_1
        real(kind=8),   intent(in)    ::  theta_2,psi_2
        ! private variables
        real(kind=8)    ::  d_1(3), d_2(3)
        ! start work

        ! Work out the two vectors
        d_1 = d_from_theta_psi(theta_1,psi_1)
        d_2 = d_from_theta_psi(theta_2,psi_2)

        angle = angle_between_vectors(d_1,d_2)
    end function angle_between_euler_angles

    !>  \brief  Compute the angle (in radians) between two vectors
    pure real function angle_between_vectors(d_1,d_2) result (angle)
        real(kind=8),   intent(in)    ::  d_1(3), d_2(3)
        ! start work
        angle = sum(d_1*d_2) /sqrt(sum(d_1**2)) /sqrt(sum(d_2**2))
        angle = min(angle,1.0e0)
        angle = max(angle,-1.0e0)
        angle = acos(angle)
    end function angle_between_vectors

    !>  \brief  return the derivative in x from theta and psi
    elemental pure real function dx_from_theta_psi(theta,psi) result(dx)
        use rotation_matrices, only : radians
        implicit none
        ! arguments
        real(kind=8),   intent(in)  ::  theta   !<  theta angle (degrees)
        real(kind=8),   intent(in)  ::  psi     !<  psi angle (degrees)
        !
        dx = -sin(radians(theta))*cos(radians(psi))
    end function dx_from_theta_psi

    !>  \brief  return the derivative in y from theta and psi
    elemental pure real function dy_from_theta_psi(theta,psi) result(dy)
        use rotation_matrices, only : radians
        implicit none
        ! arguments
        real(kind=8),   intent(in)  ::  theta   !<  theta angle (degrees)
        real(kind=8),   intent(in)  ::  psi     !<  psi angle (degrees)
        !
        dy = sin(radians(theta))*sin(radians(psi))
    end function dy_from_theta_psi

    !>  \brief  return the derivative in z from theta
    elemental pure real function dz_from_theta(theta) result(dz)
        use rotation_matrices, only : radians
        implicit none
        ! arguments
        real(kind=8),   intent(in)  ::  theta   !<  theta angle (degrees)
        !
        dz = cos(radians(theta))
    end function dz_from_theta

    !>  \todo   This could be made faster if it were safe to assume theta is 0-360.0
    pure real function psi_from_theta_dx_dy(theta,dx,dy) result(psi)
        use rotation_matrices, only : degrees
        implicit none
        !
        real(kind=8),   intent(in)  ::  theta
        real(kind=8),   intent(in)  ::  dx,dy
        !
        if (dx .eq. 0.0 .and. dy .eq. 0.0) then
            psi = 0.0
        else
            psi = degrees(-atan2(dy,dx))
            if (mod(theta,360.0d0) .lt. 180.0d0) psi = psi +180.0d0
        endif
    end function psi_from_theta_dx_dy

    pure real function theta_from_dz(dz) result(theta)
        use rotation_matrices, only : degrees, radians
        implicit none
        ! arguments
        real(kind=8),   intent(in)  :: dz
        !
        if (dz .gt. 1.0d0) then
            theta = 0.0d0
        else if (dz .lt. -1.0d0) then
            theta = -180.0d0
        else
            !theta = degrees(acos(dz/sqrt(dx**2+dy**2+dz**2)))
            theta = degrees(acos(dz))
        endif
    end function theta_from_dz


    !>  \todo  Deprecate
!    pure real function theta_from_dx_dy_dz(dx,dy,dz,psi) result(theta)
!        use rotation_matrices, only : degrees, radians
!        implicit none
!        ! Arguments
!        real(KIND=8),               intent(in)  ::  dx,dy,dz
!        real(KIND=8),   optional,   intent(in)  ::  psi         !<  If Psi is not supplied, it will be calculated internally
!        ! Private variables
!        !logical,   parameter   ::  debug   =   .FALSE.
!        real(KIND=8)    ::  ppsi
!        !
!        ! If necessary, work out psi
!        if (present(psi)) then
!            ppsi = psi
!        else
!            ppsi = psi_from_dx_dy(dx,dy)
!        endif
!
!        ! For theta, we have the choice between two formulas. We will choose the one where we are less likely to deal with zeroes
!        if (abs(sin(radians(ppsi))) .GT. abs(cos(radians(ppsi)))) then
!            !if (debug) print '(a)', '**debug(theta_from_dx_dy_dz_psi): using sin(psi) to compute theta'
!            theta   =   degrees(atan2(dy,DZ*SIN(radians(ppsi))))
!        else
!            !if (debug) print '(a)', '**debug(theta_from_dx_dy_dz_psi): using cos(psi) to compute theta'
!            theta   =   degrees(atan2(-DX,DZ*COS(radians(ppsi))))
!        endif
!
!    end function theta_from_dx_dy_dz


    !>  \brief  By convention, psi should always be between -90 and 90 degrees. This routine enforces this.
    !!
    !!  Deprecated because it messes up phieulera
!    pure subroutine psi_btw_m90_90_old(theta,psi)
!         use eulerangles, only : degrees_m180_to_180
!         implicit none
!         ! Arguments
!         real(KIND=8),   intent(inout)   ::  theta, psi
!         ! Private variables
!         !logical,    parameter   ::  debug   =   .FALSE.
!         ! Start work
!
!         ! By convention, psi should always be between -90 and 90
!         if (abs(psi) .GT. 90.) then
!             !if (debug) print '(a)', '**debug(psi_btw_m90_90): correcting theta & psi to enforce convention that -90<=psi<=90'
!             psi     = degrees_m180_to_180(psi - 180.)
!             theta   = degrees_m180_to_180(-THETA)
!         endif
!    end subroutine psi_btw_m90_90_old


    !>  \brief  Compute Euler angles coresponding to the derivatives given
    subroutine euler_from_derivatives_new(dx,dy,dz,want_theta_lt_180,theta,psi)
        implicit none
        ! arguments
        real(kind=8),       intent(in)  ::  dx,dy,dz
        logical,            intent(in)  ::  want_theta_lt_180       !<  There are two colinear solutions to the problem of finding Euler angles that describe a vector's orientation. You must specify which you want - 0<theta<180 or 180<theta<360?
        real(kind=8),       intent(out) ::  theta,psi
        ! private variables
        real(kind=8)    ::  d_norm
        logical, parameter  ::  sanity_check = .false.
        ! start work
        d_norm = 1.0d0/sqrt(dx**2+dy**2+dz**2)
        theta   = theta_from_dz(dz*d_norm)
        psi     = psi_from_theta_dx_dy(theta,dx*d_norm,dy*d_norm)
        if (psi   .lt. 0.0d0) psi   = psi   + 360.0d0
        if (theta .lt. 0.0d0) theta = theta + 360.0d0
        if (sanity_check) call check_euler_and_derivatives_new(psi,theta,dx,dy,dz)
        if ((want_theta_lt_180 .and. theta .gt. 180.0) .or. (.not. want_theta_lt_180 .and. theta .lt. 180.0)) then
            psi   = mod(psi   + 180.0d0,360.0d0)
            theta = mod(theta + 180.0d0,360.0d0)
        endif
    end subroutine euler_from_derivatives_new

    !>  \brief  Compute Euler angles corresponding to the derivatives given
    !!
    !!  See the "Spline Parametrisation" document for derivations of the equations
!#ifdef skip_runtime_checks
!    pure &
!#endif
!    subroutine euler_from_derivatives_old(dx,dy,dz,theta,psi)
!        use eulerangles, only : degrees_m180_to_180
!        use runtime_parameters, only : terminate
!        implicit none
!        ! Arguments
!        real(KIND=8),   intent(in)  ::  dx, dy, dz
!        real(KIND=8),   intent(out) ::  theta, psi
!        ! Private variables
!#ifndef skip_runtime_checks
!        logical,    parameter   ::  sanity_check        =   .FALSE.
!        logical,    parameter   ::  debug               =   .FALSE.
!        logical,    parameter   ::  correction_alert    =   .FALSE.
!#endif
!        ! Start work
!
!#ifndef skip_runtime_checks
!        if (debug) print '(a,3(F8.3,1x))', '**debug(euler_from_derivatives):            dx, dy, dz = ', dx, dy, dz
!#endif
!
!        psi     =   psi_from_dx_dy(dx,dy)
!        theta   =   theta_from_dx_dy_dz(dx,dy,dz,PSI=PSI)
!
!#ifndef skip_runtime_checks
!        if (debug) print '(a,2(F8.3,1x))', '**debug(euler_from_derivatives): raw        psi, theta = ', psi, theta
!#endif
!        ! By convention, psi should always be between -90 and 90
!        call psi_btw_m90_90_old(theta,psi)
!
!        ! There are some cases where we may need to fix theta and/or psi
!        if (dz .LT. 0.) then
!            if (abs(theta) .LT. 90.) then
!#ifndef skip_runtime_checks
!                if (debug .OR. correction_alert) print '(a)', '**debug(euler_from_derivatives): correcting theta (2)!'
!#endif
!                theta = degrees_m180_to_180(theta - 180.)
!            else if (abs(theta) .EQ. 90.) then
!                ! We are probably here because dz is almost 0.0, approaching from the - side
!                ! and abs(theta) is rounded to 90. by the trig functions, but should really be slightly > 90.
!#ifndef skip_runtime_checks
!                if (debug .OR. correction_alert) write(*,'(a)') '**debug(euler_from_derivatives): got here (2*)'
!#endif
!                ! By looking at dx, we can decide whether Theta should be + or -90
!                if (dx .GT. 0.) then
!                    theta = -90.
!                else if (dx .LT. 0.) then
!                    theta = 90.
!                endif
!            endif
!        else if (dz .GT. 0.) then
!            if (abs(theta) .GT. 90.) then
!#ifndef skip_runtime_checks
!                if (debug .OR. correction_alert) print '(a)', '**debug(euler_from_derivatives): correcting theta (3)!'
!#endif
!                theta = degrees_m180_to_180(theta - 180.)
!            else if (abs(theta) .EQ. 90.) then
!                ! We are probably here because dz is almost 0.0, approaching from the + side
!                ! and abs(theta) is rounded to 90. by the trig functions, but should really be slightly < 90.
!#ifndef skip_runtime_checks
!                if (debug .OR. correction_alert) write(*,'(a)') '**debug(euler_from_derivatives): got here (3*)'
!#endif
!                ! By looking at dx, we can decide whether Theta should be + or -90
!                if (dx .GT. 0.) then
!                    theta = -90.
!                else if (dx .LT. 0.) then
!                    theta = 90.
!                endif
!            endif
!        else if (dz .EQ. 0.) then
!            if (dx .GT. 0.) then
!                ! Theta should be negative
!#ifndef skip_runtime_checks
!                if (debug .OR. correction_alert) print '(a)', '**debug(euler_from_derivatives): correcting theta (4)!'
!#endif
!                if (theta .GT. 0.) theta = degrees_m180_to_180(theta - 180.)
!#ifndef skip_runtime_checks
!                if (sanity_check) then
!                    if (theta .GT. 0.) then
!                        if (debug) print *, '**debug(euler_from_derivatives): theta = ', theta
!                        call terminate('Sanity check failed')
!                    endif
!                endif
!#endif
!            else if (dx .LT. 0.) then
!                ! Theta should be positive
!#ifndef skip_runtime_checks
!                if (debug .OR. correction_alert) print '(a)', '**debug(euler_from_derivatives): correcting theta (5)!'
!#endif
!                if (theta .LT. 0.) theta = degrees_m180_to_180(theta - 180.)
!#ifndef skip_runtime_checks
!                if (sanity_check) then
!                    if (theta .LT. 0.) then
!                        call terminate('Sanity check failed')
!                    endif
!                endif
!#endif
!            else ! dx == 0.
!                if (dy .GT. 0.) then
!                    theta = -90.
!                    psi   = -90.
!                else if (dy .LT. 0.) then
!                    theta = -90.
!                    psi   = 90.
!                else ! dy == 0
!                    ! dx, dy and dz are 0, this is not possible
!#ifndef skip_runtime_checks
!                    write(*,'(a,3(F0.8,1x))') '**ERROR(euler_from_derivatives): dx, dy, dz = ', dx, dy, dz
!                    call terminate('Fatal error in EULER_FROM_DERIVATIVES')
!#endif
!                endif
!            endif
!        endif
!#ifndef skip_runtime_checks
!        if (debug) print '(a,2(F8.3,1x))', '**debug(euler_from_derivatives): corrected  psi, theta = ', psi, theta
!
!        ! Sanity check
!        if (sanity_check .OR. debug) then
!            call check_euler_and_derivatives(psi,theta,dx,dy,dz)
!        endif
!#endif
!    end subroutine euler_from_derivatives_old

    !>  \brief  Check whether the Euler angles and derivatives match (see notes from 8-Nov-2010 for derivations)
    !!
    subroutine check_euler_and_derivatives(psi,theta,dx,dy,dz)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        real(kind=8),   intent(in)  ::  psi, theta, dx, dy, dz
        ! Private variables
        real(kind=8),       parameter   ::  almost_zero = 0.0001D0
        logical                         ::  theta_ok, psi_ok
        logical,            parameter   ::  debug = .false.
        logical                         ::  do_the_check
        integer                         ::  ldx,ldy,ldz
        ! Start work

        theta_ok = .true.
        psi_ok = .true.
        do_the_check = .true.

        if (debug) print '(a,3(F8.3,1x))', '**debug(check_euler_and_derivatives): dx, dy, dz = ', dx, dy, dz

        ! In some cases, the numerical calculations are not stable and we need to skip the tests.
        ! Note that the check on orders of magnitude is probably overkill in the case where dz is 0.0 (the formulas should
        ! still work in that case). Perhaps the special case of dz == 0. should be considered separately.
        if (abs(dx) .lt. almost_zero .or. abs(dy) .lt. almost_zero .or. abs(dz) .lt. almost_zero) do_the_check = .false.
        if (do_the_check) then
            ldx = nint(log10(abs(dx)))
            ldy = nint(log10(abs(dy)))
            ldz = nint(log10(abs(dz)))
            if (LDX-LDZ .gt. 6 .or. LDY-LDZ .gt. 6) DO_THE_CHECK = .false.
        endif


        !
        ! If derivatives are very close to 0.0, the numerical calculation do not necessarily behave. Therefore, skip tests
        !
        if (do_the_check) then
            !
            ! Check psi and theta are >-180 and <=180
            !
            if (psi .le. -180. .or. psi .gt. 180.) then
                write(*,'(a,F8.3)') '**ERROR(filament_check_euler_and_derivatives): psi is out-of-range: ', psi
                call terminate('Fatal error in FILAMENT_CHECK_EULER_AND_DERIVATIVES')
            endif
            if (theta .le. -180. .or. theta .gt. 180.) then
                write(*,'(a,F8.3)') '**ERROR(filament_check_euler_and_derivatives): theta is out-of-range: ', theta
                call terminate('Fatal error in FILAMENT_CHECK_EULER_AND_DERIVATIVES')
            endif

            !
            ! Check theta
            !
            if (dx .gt. almost_zero) then
                if (dz .gt. 0.) then
                    if (theta .le. -90. .or. theta .ge. 0.) theta_ok = .false.
                else if (dz .lt. 0.) then
                    if (theta .le. -180. .or. theta .ge. -90.) theta_ok = .false.
                else    ! dz == 0.
                    if (abs(THETA+90.) .gt. 0.) theta_ok = .false.
                endif
            else if (dx .lt. 0.) then
                if (dz .gt. 0.) then
                    if (theta .le. 0. .or. theta .ge. 90.) theta_ok = .false.
                else if (dz .lt. 0.) then
                    if (theta .le. 90. .or. theta .ge. 180.) theta_ok = .false.
                else    ! dz == 0.
                    if (abs(THETA-90.) .gt. 0.) theta_ok = .false.
                endif
            endif


            !
            ! Check Psi
            !
            if (dx .lt. 0.) then
                if (dy .gt. 0.) then
                    if (psi .le. 0. .or. psi .ge. 90.) psi_ok = .false.
                else if (dy .lt. 0.) then
                    if (psi .le. -90. .or. psi .ge. 0.) psi_ok = .false.
                else ! dy == 0
                    if (abs(psi) .gt. 0.) psi_ok = .false.
                endif
            else if (dx .gt. 0.) then
                if (dy .gt. 0.) then
                    if (psi .le. -90. .or. psi .ge. 0.) psi_ok = .false.
                else if (dy .lt. 0.) then
                    if (psi .le. 0. .or. psi .ge. 90.) psi_ok = .false.
                else ! dy == 0
                    if (abs(psi) .gt. 0.) psi_ok = .false.
                endif
            endif

            !
            ! Special cases where dx == 0.0
            !
            if (dx .eq. 0.) then
                if (theta .lt. 0. .or. theta .gt. 180.) theta_ok = .false.
                if (dz .gt. 0.) then
                     if (dy .gt. 0.) then
                        if (abs(PSI-90.) .gt. 0.) psi_ok = .false.
                        if (theta .le. 0.) theta_ok = .false.
                     else if (dy .lt. 0.) then
                        if (abs(PSI+90.) .gt. 0.) psi_ok = .false.
                        if (theta .le. 0.) theta_ok = .false.
                     else ! dy == 0
                        if (psi .ne. 0.) psi_ok = .false.
                     endif
                else if (dz .lt. 0.) then
                    if (dy .gt. 0.) then
                        if (abs(PSI-90.) .gt. 0.) psi_ok = .false.
                        if (theta .ge. 180.) theta_ok = .false.
                    else if (dy .lt. 0.) then
                        if (abs(PSI+90.) .gt. 0.) psi_ok = .false.
                        if (theta .ge. 180.) theta_ok = .false.
                    else ! dy == 0
                        if (abs(THETA-180.) .gt. 0.) theta_ok = .false.
                    endif
                else ! dz == 0.
                    if (dy .gt. 0.) then
                        if (abs(PSI-90.) .gt. 0.) psi_ok = .false.
                        if (abs(THETA-90.) .gt. 0.) theta_ok = .false.
                    else if (dy .lt. 0.) then
                        if (abs(PSI+90.) .gt. 0.) psi_ok = .false.
                        if (abs(THETA-90.) .gt. 0.) theta_ok = .false.
                    else ! dy == 0
                        call terminate('Ooops, all 3 derivatives are 0.0')
                    endif
                endif
            endif


            if (.not. theta_ok) then
                write(*,'(a)') '**ERROR(check_euler_and_derivatives): theta angle does not match derivatives'
            endif

            if (.not. psi_ok) then
                write(*,'(a)') '**ERROR(check_euler_and_derivatives): psi angle does not match derivatives'
            endif

            !
            ! Do the signs of the derivatives match what we expected?
            !
            if (.not. theta_ok .or. .not. psi_ok) then
                write(*,'(a)') '**ERROR(check_euler_and_derivatives): derivatives do not have expected sign, given Euler angles'
                write(*,'(a,F12.7)')    '**debug(check_euler_and_derivatives): psi      = ', psi
                write(*,'(a,F12.7)')    '**debug(check_euler_and_derivatives): theta    = ', theta
                write(*,'(a,F12.7)')    '**debug(check_euler_and_derivatives): dx       = ', dx
                write(*,'(a,F12.7)')    '**debug(check_euler_and_derivatives): dy       = ', dy
                write(*,'(a,F12.7)')    '**debug(check_euler_and_derivatives): dz       = ', dz
                call terminate('Fatal error in CHECK_EULER_AND_DERIVATIVES')
            endif
        endif   ! End of check for derivatives very close to 0.0
    end subroutine check_euler_and_derivatives

    !>  \brief  Return the expected sign of the x,y,z derivatives, given Euler angles theta and psi
    function expected_d_signs_given_theta_psi(theta,psi) result(expected_sign)
        real(kind=8),   intent(in)  ::  theta,psi
        integer                     ::  expected_sign(3)
        !
        if (psi .lt. 90.0d0 .or. psi .gt. 270.0d0) then
            if (theta .lt. 180.0d0) then
                expected_sign(1) = -1 ! dx < 0
                if (psi .lt. 90.0d0) then
                    expected_sign(2) =  1 ! dy > 0
                else
                    expected_sign(2) = -1 ! dy < 0
                endif
            else
                expected_sign(1) =  1 ! dx > 0
                if (psi .lt. 90.0d0) then
                    expected_sign(2) = -1 ! dy < 0
                else
                    expected_sign(2) =  1 ! dy > 0
                endif
            endif
        else
            if (theta .lt. 180.0d0) then
                expected_sign(1) =  1 ! dx > 0
                if (psi .lt. 180.0d0) then
                    expected_sign(2) =  1 ! dy > 0
                else
                    expected_sign(2) = -1 ! dy < 0
                endif
            else
                expected_sign(1) = -1 ! dx < 0
                if (psi .lt. 180.0d0) then
                    expected_sign(2) = -1 ! dy < 0
                else
                    expected_sign(2) =  1 ! dy > 0
                endif
            endif
        endif
        if (theta .lt. 90.0d0 .or. theta .gt. 270.0d0) then
            expected_sign(3) =  1 ! dz > 0
        else
            expected_sign(3) = -1 ! dz < 0
        endif
    end function expected_d_signs_given_theta_psi

    !>  \brief  Check whether the Euler angles and derivatives match (see notes from 8-Nov-2010 for derivations)
    !!
    subroutine check_euler_and_derivatives_new(psi,theta,dx,dy,dz)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        real(kind=8),   intent(in)  ::  psi, theta, dx, dy, dz
        ! Private variables
        real(kind=8),       parameter   ::  almost_zero = 0.0001D0
        logical                         ::  theta_ok, psi_ok
        logical,            parameter   ::  debug = .false.
        logical                         ::  do_the_check
        integer                         ::  ldx,ldy,ldz

        integer                         ::  expected_sign(3), actual_sign(3)
        ! Start work

        expected_sign = 0
        actual_sign   = 0


        theta_ok = .true.
        psi_ok = .true.
        do_the_check = .true.

        if (debug) print '(a,3(F8.3,1x))', '**debug(check_euler_and_derivatives_new): dx, dy, dz = ', dx, dy, dz

        ! In some cases, the numerical calculations are not stable and we need to skip the tests.
        ! Note that the check on orders of magnitude is probably overkill in the case where dz is 0.0 (the formulas should
        ! still work in that case). Perhaps the special case of dz == 0. should be considered separately.
        if (abs(dx) .lt. almost_zero .or. abs(dy) .lt. almost_zero .or. abs(dz) .lt. almost_zero) do_the_check = .false.
        if (do_the_check) then
            ldx = nint(log10(abs(dx)))
            ldy = nint(log10(abs(dy)))
            ldz = nint(log10(abs(dz)))
            if (ldx-ldz .gt. 6 .or. ldy-ldz .gt. 6) do_the_check = .false.
        endif


        !
        ! If derivatives are very close to 0.0, the numerical calculation do not necessarily behave. Therefore, skip tests
        !
        if (do_the_check) then
            !
            ! Check psi and theta are 0..360
            !
            if (psi .le. 0. .or. psi .gt. 360.) then
                write(*,'(a,F8.3)') '**ERROR(check_euler_and_derivatives_new): psi is out-of-range: ', psi
                call terminate('Fatal error in CHECK_EULER_AND_DERIVATIVES_NEW')
            endif
            if (theta .le. 0. .or. theta .gt. 360.) then
                write(*,'(a,F8.3)') '**ERROR(check_euler_and_derivatives_new): theta is out-of-range: ', theta
                call terminate('Fatal error in CHECK_EULER_AND_DERIVATIVES_NEW')
            endif


            !
            ! What sign to we expect for the 3 derivatives
            !
            expected_sign = expected_d_signs_given_theta_psi(theta,psi)


            if (dx .gt. 0.0d0) then
                actual_sign(1) = 1
            else
                actual_sign(1) = -1
            endif
            if (dy .gt. 0.0d0) then
                actual_sign(2) = 1
            else
                actual_sign(2) = -1
            endif
            if (dz .gt. 0.0d0) then
                actual_sign(3) = 1
            else
                actual_sign(3) = -1
            endif


            !
            ! Do the signs of the derivatives match what we expected?
            !
            if (.not. all(actual_sign .eq. expected_sign)) then
                write(*,'(a)') '**ERROR(check_euler_and_derivatives_new): derivatives do not have expected sign, given Euler angles'
                write(*,'(a,F12.7)')    '**debug(check_euler_and_derivatives_new): psi      = ', psi
                write(*,'(a,F12.7)')    '**debug(check_euler_and_derivatives_new): theta    = ', theta
                write(*,'(a,F12.7)')    '**debug(check_euler_and_derivatives_new): dx       = ', dx
                write(*,'(a,F12.7)')    '**debug(check_euler_and_derivatives_new): dy       = ', dy
                write(*,'(a,F12.7)')    '**debug(check_euler_and_derivatives_new): dz       = ', dz
                write(*,'(a,3(i0,1x))')  '**debug(check_euler_and_derivatives_new): expected signs = ', expected_sign
                write(*,'(a,3(i0,1x))')  '**debug(check_euler_and_derivatives_new):   actual signs = ', actual_sign
                call terminate('Fatal error in CHECK_EULER_AND_DERIVATIVES_NEW')
            endif
        endif   ! End of check for derivatives very close to 0.0
    end subroutine check_euler_and_derivatives_new

end module differential_geometry
