!>  \brief  Methods to compute eigenvectors and values
module eigen

    !use eigen_nr
    use fgsl

    contains

    !>  \brief  Compute the eigenvectors and eigenvalues for a 2d matrix
    subroutine matrix_eigen_analysis(self,eigenvectors,eigenvalues)
        use runtime_parameters, only: terminate
        implicit none
        ! Arguments
        real(kind=8),                target, intent(in)      ::  self(:,:)
        real(kind=8),   allocatable, target, intent(inout)   ::  eigenvectors(:,:)
        real(kind=8),   allocatable, target, intent(inout)   ::  eigenvalues(:)
        ! Private variables
        type(fgsl_eigen_symmv_workspace)    ::  ws
        type(fgsl_matrix)                   ::  matrix, evec
        type(fgsl_vector)                   ::  eval
        integer(fgsl_int)                   ::  status
        integer(fgsl_size_t)                ::  n_fgsl
        integer                             ::  n
        logical,    parameter               ::  debug   =   .true.
        ! Start work

        n = size(self,1)
        n_fgsl = n

        ! Check allocation
        if (.not. allocated(eigenvectors)) allocate(eigenvectors(n,n))
        if (.not. allocated(eigenvalues))  allocate(eigenvalues(n))

        ! Check dimensions
        if (     n .ne. size(self,2)         .or. n .ne. size(eigenvectors,1) &
            .or. n .ne. size(eigenvectors,2) .or. n .ne. size(eigenvalues)) then
            write(*,'(a,5(i0,1x))') '**ERROR(matrix_eigen_analysis): bad input dimensions: ', size(self,1),size(self,2), &
                                    size(eigenvectors,1),size(eigenvectors,2), size(eigenvalues)
            call terminate('Fatal error in MATRIX_EIGEN_ANALYSIS')
        endif


        ! Copy data over to fgsl format
        matrix = fgsl_matrix_init(TYPE=1.0_FGSL_DOUBLE)
        evec   = fgsl_matrix_init(TYPE=1.0_FGSL_DOUBLE)
        eval   = fgsl_vector_init(TYPE=1.0_FGSL_DOUBLE)
        status = fgsl_matrix_align(self,n_fgsl,n_fgsl,n_fgsl,matrix)
        status = fgsl_matrix_align(eigenvectors,n_fgsl,n_fgsl,n_fgsl,evec)
        status = fgsl_vector_align(eigenvalues,n_fgsl,eval,n_fgsl,0_fgsl_size_t,1_fgsl_size_t)

        ! Allocate gsl workspace
        ws = fgsl_eigen_symmv_alloc(int(n,fgsl_size_t))

        !
        status = fgsl_eigen_symmv(matrix, eval, evec, ws)
        status = fgsl_eigen_symmv_sort(eval,evec,fgsl_eigen_sort_abs_asc)

        ! Copy results back
        !eigenvectors = evec
        !eigenvalues = eval

        if (debug) then
            !nothing
        endif

        ! Deallocate memory
        call fgsl_eigen_symmv_free(ws)
        call fgsl_matrix_free(matrix)
        call fgsl_matrix_free(evec)
        call fgsl_vector_free(eval)
    end subroutine matrix_eigen_analysis



end module eigen
