!>  \brief  Classes and methods dealing with Euler angles
!!          By default, we will use the Spider/Frealign convention
!!
!!  \todo   Add functions to convert between various conventions
module euleranglesmod

    implicit none

    type euler
        real(kind=8)    ::  phi     =   0.0D0   !<  Phi angle (Spider/Frealign convention) - twist  (equiv. to Imagic's Gamma)
        real(kind=8)    ::  theta   =   0.0D0   !<  Theta angle (Spider/Frealign convention) - out-of-plane (equiv. to Imagic's Beta)
        real(kind=8)    ::  psi     =   0.0D0   !<  Psi angle, (Spider/Frealign convention) - in-plane (equiv. to Imagic's Alpha)
    end type

    interface operator (.eq.)
        module procedure is_equal
    end interface

    interface operator (.ne.)
        module procedure is_not_equal
    end interface

    interface operator (.lt.)
        module procedure euler_is_lt
    end interface

    interface operator (+)
        module procedure euler_plus_euler
    end interface

    interface operator (-)
        module procedure euler_minus_euler
    end interface

    interface operator (*)
        module procedure euler_times_euler
    end interface

    interface operator (/)
        module procedure euler_divide_by_real
    end interface

    interface sqrt
        module procedure euler_sqrt
    end interface sqrt

    interface abs
        module procedure euler_abs
    end interface abs

    contains

    pure subroutine euler_init(self)
        implicit none
        ! Arguments
        type(euler),    intent(inout)  ::  self
        ! Start work
        self%phi   = 0.0D0
        self%theta = 0.0D0
        self%psi   = 0.0D0
    end subroutine euler_init

    elemental pure function euler_divide_by_real(angles_in,val) result(angles_out)
        implicit none
        ! Arguments
        type(euler),    intent(in)  ::  angles_in
        real,           intent(in)  ::  val
        ! Result
        type(euler) ::  angles_out
        ! Start work
        angles_out%phi      =   angles_in%phi / val
        angles_out%theta    =   angles_in%theta / val
        angles_out%psi      =   angles_in%psi / val
    end function euler_divide_by_real

    elemental pure function euler_plus_euler(angles1,angles2)
        implicit none
        ! Arguments
        type(euler),    intent(in)  ::  angles1, angles2
        ! Result
        type(euler)     ::  euler_plus_euler
        ! Start work
        euler_plus_euler%phi    =   angles1%phi + angles2%phi
        euler_plus_euler%theta  =   angles1%theta + angles2%theta
        euler_plus_euler%psi    =   angles1%psi + angles2%psi
    end function euler_plus_euler

    elemental pure function euler_minus_euler(angles1,angles2)
        implicit none
        ! Arguments
        type(euler),    intent(in)  ::  angles1, angles2
        ! Result
        type(euler)     ::  euler_minus_euler
        ! Start work
        euler_minus_euler%phi   =   angles1%phi - angles2%phi
        euler_minus_euler%theta =   angles1%theta - angles2%theta
        euler_minus_euler%psi   =   angles1%psi - angles2%psi
    end function euler_minus_euler

    elemental pure function euler_times_euler(angles1,angles2)
        implicit none
        ! Arguments
        type(euler),    intent(in)  ::  angles1, angles2
        ! Result
        type(euler)     ::  euler_times_euler
        ! Start work
        euler_times_euler%phi   =   angles1%phi * angles2%phi
        euler_times_euler%theta =   angles1%theta * angles2%theta
        euler_times_euler%psi   =   angles1%psi * angles2%psi
    end function euler_times_euler

    elemental pure function euler_sqrt(angles)
        implicit none
        ! Arguments
        type(euler),    intent(in)  ::  angles
        ! Result
        type(euler)     ::  euler_sqrt
        ! Start work
        euler_sqrt%phi      =   sqrt(angles%phi)
        euler_sqrt%theta    =   sqrt(angles%theta)
        euler_sqrt%psi      =   sqrt(angles%psi)
    end function euler_sqrt

    elemental pure function euler_abs(angles)
        implicit none
        ! Argumnets
        type(euler),    intent(in)  ::  angles
        ! Result
        type(euler)     ::  euler_abs
        ! Start work
        euler_abs%phi       =   abs(angles%phi)
        euler_abs%theta     =   abs(angles%theta)
        euler_abs%psi       =   abs(angles%psi)
    end function euler_abs

    elemental pure logical function is_not_equal(angles1,angles2)
        implicit none
        ! Arguments
        type(euler),    intent(in)  ::  angles1
        type(euler),    intent(in)  ::  angles2
        ! Start work
        is_not_equal = .not. is_equal(angles1,angles2)
    end function is_not_equal

    !>  \brief  Test whether two sets of angles are equal
    elemental pure logical function is_equal(angles1,angles2)
        implicit none
        ! Arguments
        type(euler),    intent(in)  ::  angles1
        type(euler),    intent(in)  ::  angles2
        ! Start work
        is_equal =  angles1%phi .eq. angles2%phi .and. angles1%theta .eq. angles2%theta .and. angles1%psi .eq. angles2%psi
    end function is_equal

    elemental pure logical function euler_is_lt(angles1,angles2)
        implicit none
        ! Arguments
        type(euler),    intent(in)  ::  angles1
        type(euler),    intent(in)  ::  angles2
        ! Start work
        euler_is_lt =   angles1%phi .lt. angles2%phi .and. &
                        angles1%theta .lt. angles2%theta .and. &
                        angles1%psi .lt. angles2%psi
    end function euler_is_lt

    elemental pure function degrees_m180_to_180(degrees)
        implicit none
        ! Argument
        real(kind=8),   intent(in)  ::  degrees
        ! Result
        real(kind=8)    ::  degrees_m180_to_180
        ! Private variables
        ! Start work
        degrees_m180_to_180 = mod(degrees,360.D0)
        if (degrees_m180_to_180 .gt. 180.D0) then
            degrees_m180_to_180 = degrees_m180_to_180 - 360.D0
        else if (degrees_m180_to_180 .lt. -180.D0) then
            degrees_m180_to_180 = degrees_m180_to_180 + 360.D0
        else
            ! Nothing happens
        endif
    end function degrees_m180_to_180

    !>  \brief  Returns the difference between two sets of Euler angles (angles1-angles2). The difference lies between -180 and +180
    pure function euler_difference(angles1,angles2)
        implicit none
        ! Arguments
        type(euler),    intent(in)  ::  angles1
        type(euler),    intent(in)  ::  angles2
        ! Result
        type(euler) ::  euler_difference
        ! Private variables

        ! Start work
        euler_difference%phi    =   angle_difference(angles1%phi,   angles2%phi)
        euler_difference%theta  =   angle_difference(angles1%theta, angles2%theta)
        euler_difference%psi    =   angle_difference(angles1%psi,   angles2%psi)
    end function euler_difference

    !>  \brief  Returns the difference between two angles (angle1-angle2). The difference lies between -180 and +180
    pure elemental function angle_difference(angle1,angle2)
        implicit none
        ! Arguments
        real(kind=8),   intent(in)  ::  angle1  !<  Angle1 (in degrees)
        real(kind=8),   intent(in)  ::  angle2  !<  Angle2 (in degrees)
        ! Result
        real(kind=8)                ::  angle_difference
        ! Private variables
        real(kind=8)    ::  angle1_tmp,angle2_tmp
        ! Start work
        !print '(a,2F9.2)', '**debug(angle_difference): angle1, angle2 = ', angle1, angle2

        ! Bring everything back to the 0.0 - 360.0 interval
        angle1_tmp = angle_btwn_0_and_360(angle1)
        angle2_tmp = angle_btwn_0_and_360(angle2)

        !print '(a,2F9.2)', '**debug(angle_difference): angle1, angle2 (0-360) = ', angle1_tmp, angle2_tmp

        ! Calculate the difference
        angle_difference = angle1_tmp - angle2_tmp

        !print '(a,F7.2)', '**debug(angle_difference): angle diff = ', angle_difference

        ! If it would have been shorter to go over the 0/360 boundary,
        if (angle_difference .gt. 180.0) then
            angle_difference = angle_difference - 360.0
        else if (angle_difference .lt. -180.0) then
            angle_difference = angle_difference + 360.0
        endif

        !print '(a,F7.2)', '**debug(angle_difference): angle diff (-180,180) = ', angle_difference
    end function angle_difference

    !>  \brief  Return angles between 0 and 360.0
    pure function euler_btwn_0_and_360(angles) result(angles_out)
        implicit none
        ! Arguments
        type(euler),    intent(in)  ::  angles
        ! Result
        type(euler) ::  angles_out
        ! Private variables

        ! Start work
        angles_out%phi      =   angle_btwn_0_and_360(angles%phi)
        angles_out%theta    =   angle_btwn_0_and_360(angles%theta)
        angles_out%psi      =   angle_btwn_0_and_360(angles%psi)
    end function euler_btwn_0_and_360

    !>  \brief  Add or substract 360 until the angle is .GE. 0.0 and .LT. 360.0
    pure function angle_btwn_0_and_360(angle)
        implicit none
        ! Arguments
        real(kind=8),   intent(in)  ::  angle
        ! Result
        real(kind=8)                ::  angle_btwn_0_and_360
        ! Private variables

        ! Start work
        angle_btwn_0_and_360 = mod(angle,360.0D0)
        if (angle_btwn_0_and_360 .lt. 0.0D0) angle_btwn_0_and_360 = angle_btwn_0_and_360 + 360.0D0
    end function angle_btwn_0_and_360

end module euleranglesmod
