!>  \brief  Classes and methods related to filament objects
!!
!!  For details on spline 3d modelling of filaments, parametrisation and related topics, see the documentation.
!!
module filaments

    use filaments_base

    implicit none


    contains

    !>  \brief  Get a set of values from a filament's waypoints.
    !!
    !!  The convention for the order of the values is that they are ordered first by parameter index
    !!  (1-3: x-z, 4-6: phi,theta,psi, 7: subunit_num,8.HRise,9.HTwist) and then by waypoint #.
    subroutine values_from_filament(self,values,params_mask,params_mask_hold_average,incremental_values)
        use runtime_parameters, only : terminate
        use filaments_geometry, only : filament_params_absolute_to_incremental, filament_params_incremental_to_absolute
        use filaments_normal_modes, only : filament_wps_to_nm_amplitudes
        use coos, only : value_from_waypoint
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  self
        real(kind=8),       allocatable,    intent(inout)   ::  values(:)
        logical,                            intent(in)      ::  params_mask(9)  !<  For each of the 9 parameters, whether they are included in the values array
        logical,            optional,       intent(in)      ::  params_mask_hold_average(9) !<  For each of the 9 parameters whether the average value should be left alone
        logical,            optional,       intent(in)      ::  incremental_values(9)   !<  For each parameters, whether incremental (relative to previous waypoint) rather than absolute values should be returned
        ! Private variables
        integer ::  num_expected_values
        integer ::  count_dim
        integer ::  current_param
        integer ::  current_wp
        logical ::  pparams_mask_hold_average(9)
        logical ::  iincremental_values(9)
        ! Start work

        ! Check whether we'll need to hold any parameter's average constant
        if (present(params_mask_hold_average)) then
            pparams_mask_hold_average = params_mask_hold_average
        else
            pparams_mask_hold_average = .false.
        endif

        ! Check incremental_values flags
        iincremental_values = .false.
        if (present(incremental_values)) iincremental_values = incremental_values

        ! Allocate array
        num_expected_values = filament_expected_number_of_values(self,params_mask,params_mask_hold_average)
        if (allocated(values)) deallocate(values)
        allocate(values(num_expected_values))


        ! If we need to get incremental values, let's do the conversion now
        if (any(iincremental_values)) then
            call filament_params_absolute_to_incremental(self,iincremental_values)
        endif


        ! If we need to give the nm amplitudes, we need to work them out
        if (any(self%refine_normal_modes)) then
            call filament_wps_to_nm_amplitudes(self)
        endif

        count_dim = 0
        ! Loop over the waypoints
        do current_wp=1,self%nwp
            ! Loop over the parameters
            do current_param=1,9
                ! Is this parameter being refined?
                if (params_mask(current_param)) then
                    ! Is this a per-wp parameter, or a rigid parameter that applies to the filament as a whole
                    if (self%rigid_params(current_param)) then
                        if (current_wp .eq. 1) then
                            ! This is a parameter that has a single value for the whole filament
                            count_dim = count_dim + 1
                            if (self%refine_normal_modes(current_param)) then
                                call terminate('Normal mode refinement not implemented with rigid parameters')
                            else
                                call value_from_waypoint(self%rigid_wp,values(count_dim),current_param)
                            endif
                        endif
                    else
                        ! This is a per-wp parameter; is it being refined?
                        if (self%wp(current_wp)%refine) then
                            if (self%refine_normal_modes(current_param) .and. current_wp .eq. 1 &
                                .and. pparams_mask_hold_average(current_param)) then
                                cycle
                            endif
                            count_dim = count_dim + 1
                            if (self%refine_normal_modes(current_param)) then
                                call value_from_waypoint(self%wp_nm(current_wp),values(count_dim),current_param)
                            else
                                call value_from_waypoint(self%wp(current_wp),values(count_dim),current_param)
                            endif
                        endif
                    endif
                endif
            enddo
        enddo

        ! If we were getting incremental values, let's do the conversion back to the normal (absolute) representation
        if (any(iincremental_values)) then
            call filament_params_incremental_to_absolute(self,iincremental_values)
        endif

        ! Sanity check
        if (count_dim .ne. num_expected_values) then
            write(*,'(2(A,i0))') '**ERROR(values_from_filament): was expecting ', num_expected_values, ' but got ', count_dim
            call terminate('Sanity check failed')
        endif
    end subroutine values_from_filament


    !>  \brief  Copy a set of values to a filament's waypoints.
    !!
    !!  The convention for the order of the values is that they are ordered first by parameter index
    !!  (1-3: x-z, 4-6: phi,theta,psi, 7: subunit_num,8.HRise,9.HTwist) and then by waypoint #.
    subroutine values_to_filament(self,values,params_mask,params_mask_hold_average,incremental_values)
        use runtime_parameters, only : terminate
        use coos, only : value_to_waypoint
        use filaments_normal_modes, only : filament_nm_amplitudes_to_wps
        use filaments_geometry, only : filament_params_incremental_to_absolute, filament_params_absolute_to_incremental
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  self
        real(kind=8),       allocatable,    intent(in)      ::  values(:)
        logical,                            intent(in)      ::  params_mask(9)  !<  For each of the 9 parameters, whether they are included in the values array
        logical,            optional,       intent(in)      ::  params_mask_hold_average(9) !<  For each of the 9 parameters whether the average value should be left alone
        logical,            optional,       intent(in)      ::  incremental_values(9)   !<  For each parameter, whether incremental (relative to previous waypoint) rather than absolute values are given
        ! Private variables
        integer ::  num_expected_values
        integer ::  count_dim
        integer ::  current_param
        integer ::  current_wp
        logical,    parameter   ::  debug   =   .false.
        logical ::  pparams_mask_hold_average(9)
        logical ::  iincremental_values(9)
        ! Start work

        ! Check whether we'll need to hold any parameter's average constant
        if (present(params_mask_hold_average)) then
            pparams_mask_hold_average = params_mask_hold_average
        else
            pparams_mask_hold_average = .false.
        endif

        ! Check incremental_values flags
        iincremental_values = .false.
        if (present(incremental_values)) iincremental_values = incremental_values

        ! i don't think i can support incremental values and normal modes simultaneously.
        ! Or if i can, i need to double check the sequence of events after the values have
        ! been plugged in to the filament (see below)
        if (any(self%refine_normal_modes) .and. any(iincremental_values)) then
            write(*,'(a)') '**ERROR(values_to_filament): incremental values are not supported in conjunction with normal modes'
            call terminate('Fatal error in VALUES_TO_FILAMENT')
        endif

        ! Holding a parameter's average value constant is only feasible when refining normal modes
        if (any(pparams_mask_hold_average .and. params_mask .and. .not. self%refine_normal_modes)) then
            write(*,'(a)') '**ERROR(values_to_filament): cannot hold the average value of a parameter constant '//&
                                                        'unless normal modes are being refined'
            call terminate('Fatal error in VALUES_TO_FILAMENT')
        endif

        ! Check we got the expected number of values
        num_expected_values = filament_expected_number_of_values(self,params_mask,params_mask_hold_average)
        if (size(values) .ne. num_expected_values) then
            write(*,'(2(A,i0),a)')    '**ERROR(values_to_filament): was expecting ', num_expected_values, &
                                    ' but got ', size(values), ' instead.'
            call terminate('Fatal error in VALUES_TO_FILAMENT')
        endif

        ! Convert to incremental if necessary
        if (any(iincremental_values)) then
            call filament_params_absolute_to_incremental(self,iincremental_values)
        endif

        count_dim = 0
        ! Loop over the waypoints
        do current_wp=1,self%nwp
            ! Loop over the parameters
            do current_param=1,9
                ! Is this parameter being refined?
                if (params_mask(current_param)) then
                    ! Is this a per-wp parameter, or a rigid parameter that applies to the filament as a whole
                    if (self%rigid_params(current_param)) then
                        if (current_wp .eq. 1) then
                            ! This is a parameter that has a single value for the whole filament
                            count_dim = count_dim + 1
                            if (self%refine_normal_modes(current_param)) then
                                call terminate('Normal mode refinement not implemented with rigid parameters')
                            else
                                call value_to_waypoint(self%rigid_wp,values(count_dim),current_param)
                            endif
                        endif
                    else
                        ! This is a per-wp parameter; is it being refined?
                        if (self%wp(current_wp)%refine) then
                            if (self%refine_normal_modes(current_param) .and. current_wp .eq. 1 &
                                .and. pparams_mask_hold_average(current_param)) then
                                cycle
                            endif
                            count_dim = count_dim + 1
                            if (self%refine_normal_modes(current_param)) then
                                call value_to_waypoint(self%wp_nm(current_wp),values(count_dim),current_param)
                            else
                                call value_to_waypoint(self%wp(current_wp),values(count_dim),current_param)
                            endif
                        endif
                    endif
                endif
            enddo
        enddo

        ! If we were dealing with incremental values, we can now convert to the normal representation
        !!
        !! Should this happen before or after num_amplitudes_to_wps ?
        if (any(iincremental_values)) then
            call filament_params_incremental_to_absolute(self,iincremental_values)
        endif

        ! Now that we've transferred values to the WP_NMs, we need to update the WPs
        if (any(self%refine_normal_modes)) then
            !if (debug) then
            !    write(*,'(a)') '**debug(values_to_filament): filament info before NM_AMPLITUDES_TO_WPS'
            !    call filament_print_info(self)
            !endif
            call filament_nm_amplitudes_to_wps(self)
            !if (debug) then
            !    write(*,'(a)') '**debug(values_to_filament): filament info after NM_AMPLITUDES_TO_WPS'
            !    call filament_print_info(self)
            !endif
        endif

        ! Sanity check
        if (count_dim .ne. num_expected_values) then
            write(*,'(2(A,i0))') '**ERROR(values_to_filament): counted ', count_dim, ' dimensions, but was expecting ', &
                                num_expected_values
            call terminate('Sanity check failed')
        endif
    end subroutine values_to_filament


    !>  \brief  Given a (filament) waypoint, create a single particle object
    !!  \todo   Consolidate with fmc_box_out_from_micrograph
    subroutine wp_to_sp(wp,sp,ctf,sp_image,segment_length,micrograph_frames,pixel_size,apply_shift_to_image, &
                        apply_ctf_to_image,ctf_flip_only,ignore_subunit_num,&
                        rotate_sp_image,unbend,fil,apply_rect_mask,&
                        too_close_to_edge,square_image, &
                        first_frame_to_sum,last_frame_to_sum,apply_drift_filter, &
                        apply_dose_filter,electron_dose,exposure_per_frame)
        use runtime_parameters, only : terminate, psf_radius_pixels, is_odd
        use Images
        use images_arithm, only  : image_float_norm_mask
        use ctftilts
        use images_ctf, only : image_apply_ctf
        use single_particles
        use InterpolatingWindows
        use rotation_matrices, only : rotmat
        use interpolation, only : default_interpol_scheme, interpolation_scheme, interpolation_scheme_init
        use ElectronDoses
        implicit none
        ! Arguments
        type(waypoint),                     intent(in)      ::  wp
        type(single_particle),              intent(out)     ::  sp
        type(ctf_parameters),               intent(inout)   ::  ctf                     !<  ctf parameters (the actual defocus value will be computed from the waypoint's z coordinate)
        type(image),            optional,   intent(inout)   ::  sp_image                !<  Image of the single particle extracted from the supplied film image
        integer,                optional,   intent(in)      ::  segment_length          !<  Length of filament segment to cut out, in pixels
        type(image),            optional,   intent(inout)   ::  micrograph_frames(:)    !<  Micrograph frame images from which to box out the sp image. Give array of 1 if not movie.
        real,                   optional,   intent(in)      ::  pixel_size              !<  Pixel size in Angstroms
        logical,                optional,   intent(in)      ::  apply_shift_to_image    !<  Apply non-integer shift to image
        logical,                optional,   intent(in)      ::  apply_ctf_to_image      !<  Apply local ctf to the image before masking
        logical,                optional,   intent(in)      ::  ctf_flip_only           !<  Multiply by sign(ctf) rather than the ctf itself
        logical,                optional,   intent(in)      ::  ignore_subunit_num      !<  Ignore the WP's subunit_num value
        logical,                optional,   intent(in)      ::  rotate_sp_image         !<  Rotate the sp image to align so that psi is zeroed
        logical,                optional,   intent(in)      ::  unbend                  !<  Unbend the image
        type(filament),         optional,   intent(inout)   ::  fil                     !<  Filament the waypoint belongs to. Needed only when doing unbending.
        logical,                optional,   intent(in)      ::  apply_rect_mask         !<  Apply a soft-edge rectangular mask
        logical,                optional,   intent(out)     ::  too_close_to_edge       !<  Indicate whether single particle is too close to the edge of the micrograph for processing
        logical,                optional,   intent(in)      ::  square_image            !<  Ensure the sp_image is square
        integer,                optional,   intent(in)      ::  first_frame_to_sum      !<  First micrograph frame to use when boxing out the particle image
        integer,                optional,   intent(in)      ::  last_frame_to_sum       !<
        logical,                optional,   intent(in)      ::  apply_drift_filter
        logical,                optional,   intent(in)      ::  apply_dose_filter
        type(ElectronDose),     optional,   intent(in)      ::  electron_dose
        real,                   optional,   intent(in)      ::  exposure_per_frame      !<   electrons per square angstroms

        ! Private variables
        integer                     ::  icoo(3)
        real                        ::  rcoo(3)
        type(ctf_parameters)        ::  ctf_local
        logical                     ::  aapply_shift_to_image, aapply_ctf_to_image, iignore_subunit_num
        logical                     ::  aapply_rect_mask
        logical                     ::  rrotate_sp_image, uunbend
        real(kind=8)                ::  tmp_subunit_num
        real                        ::  rot_tmp, rotation_mat(4,4)
        type(image)                 ::  sp_image_rot
        integer                     ::  sp_image_dim(2), sp_image_dim_large(2)
        real                        ::  azimuth
        logical,    parameter       ::  debug   =   .false.
        real                        ::  rect_mask_dim(2)
        logical                     ::  ttoo_close_to_edge, ssquare_image
        type(interpolation_scheme)  ::  interpol
        type(window)                ::  interpol_window
        integer                     ::  current_frame
        real                        ::  coordinates_on_current_frame(3)
        type(Image)                 ::  temp_cutout_image
        integer                     ::  ffirst_frame_to_sum
        integer                     ::  llast_frame_to_sum
        logical                     ::  aapply_dose_filter, aapply_drift_filter
        real                        ::  shift_during_current_frame(3)
        ! Start work

        ttoo_close_to_edge = .false.

        ! If we want an output sp image, we'll need the
        ! dimension and the film image
        if (present(sp_image)) then
            if (.not. (present(micrograph_frames) .and. present(segment_length))) then
                write(*,'(a)') '**ERROR(wp_to_sp): need film image and segment length as arguments to be able to output sp image'
                call terminate('Fatal error in WP_TO_SP')
            endif
        endif

        !
        aapply_shift_to_image = .false.
        if (present(apply_shift_to_image))  aapply_shift_to_image = apply_shift_to_image

        !
        aapply_ctf_to_image = .false.
        if (present(apply_ctf_to_image)) aapply_ctf_to_image = apply_ctf_to_image

        !
        aapply_rect_mask = .true.
        if (present(apply_rect_mask)) aapply_rect_mask = apply_rect_mask

        !
        iignore_subunit_num = .false.
        if (present(ignore_subunit_num)) iignore_subunit_num = ignore_subunit_num

        !
        rrotate_sp_image = .false.
        if (present(rotate_sp_image))   rrotate_sp_image = rotate_sp_image

        !
        ssquare_image = .false.
        if (present(square_image)) ssquare_image = square_image

        ! Unbending option
        uunbend = .false.
        if (present(unbend)) uunbend = unbend
        if (uunbend .and. rrotate_sp_image) then
            write(*,'(a)') '**info(wp_to_sp): rotate_sp_image flag is redundant when unbending'
            rrotate_sp_image = .false.
        endif
        if (uunbend .and. .not. present(fil)) then
            write(*,'(a)') '**ERROR(wp_to_sp): unbending requires that you supply the filament object as an argument'
            call terminate('Fatal error in WP_TO_SP')
        endif
        if (uunbend .and. aapply_ctf_to_image) then
            ! For now, don't have to implement ctf correction before unbending happens
            write(*,'(a)') '**warning(wp_to_sp): ctf will be applied after unbending, which is not accurate'
        endif

        ! Filters
        aapply_dose_filter  = .false.
        if (present(apply_dose_filter)) aapply_dose_filter = apply_dose_filter
        if (aapply_dose_filter .and. (.not. present(electron_dose) .or. .not. present(exposure_per_frame))) then
            call terminate('Need electron-dose object and exposure per frame to apply a dose filter')
        endif
        aapply_drift_filter = .false.
        if (present(apply_drift_filter)) aapply_drift_filter = apply_drift_filter

        ! Which frames are we summing?
        ffirst_frame_to_sum = 1
        llast_frame_to_sum = size(micrograph_frames)
        if (present(first_frame_to_sum)) ffirst_frame_to_sum = first_frame_to_sum
        if (present(last_frame_to_sum)) llast_frame_to_sum = last_frame_to_sum

        ! Work out output image dimensions
        if (present(sp_image)) then
            if (aapply_ctf_to_image .and. psf_radius_pixels .eq. 0.0) then
                write(*,'(a)') '**ERROR(wp_to_sp): psf_radius_pixels has not been set. Needed to compute box size.'
                call terminate('Fatal error in WPS_TO_SP')
            endif
            if (.not. present(fil)) then
                write(*,'(a)') '**ERROR(wp_to_sp): need filament object to be given as argument, because need to know max radius'
                call terminate('Fatal error in WP_TO_SP')
            endif
            !sp_image_dim(1) = filament_suggested_box_width(fil)
            sp_image_dim(1:2) = segment_length
            if (is_odd(sp_image_dim(1))) sp_image_dim(1) = sp_image_dim(1) + 1
            if (is_odd(sp_image_dim(2))) sp_image_dim(2) = sp_image_dim(2) + 1
            if (ssquare_image) sp_image_dim(1:2) = max(sp_image_dim(1),sp_image_dim(2))
            sp_image_dim_large(1:2) = max(segment_length,fil%SuggestedBoxWidth() + ceiling(2.*psf_radius_pixels))
            if (is_odd(sp_image_dim_large(1))) sp_image_dim_large(1) = sp_image_dim_large(1) + 1
            if (is_odd(sp_image_dim_large(2))) sp_image_dim_large(2) = sp_image_dim_large(2) + 1
            if (ssquare_image) sp_image_dim_large(1:2) = max(sp_image_dim_large(1),sp_image_dim_large(2))
            if (debug) then
                write(*,'(a,2(i0,1x))') '**debug(wp_to_sp): image dim       = ', sp_image_dim
                write(*,'(a,i0)') '**debug(wp_to_sp): filament suggested box width = ', fil%SuggestedBoxWidth()
                if (aapply_ctf_to_image) write(*,'(a,2(i0,1x))') '**debug(wp_to_sp): image dim large = ', sp_image_dim_large
            endif
        endif

        ! Work out rectangular mask dimensions
        if (aapply_rect_mask) then
            rect_mask_dim = (/ segment_length-20., 2.*fil%max_radius+10./)
            if (debug) then
                write(*,'(a,i0)') '**debug(wp_to_sp): segment length = ', segment_length
                write(*,'(a,F0.3)') '**debug(wp_to_sp): fil max rad = ', fil%max_radius
                write(*,'(a,2(F0.3,1x))') '**debug(wp_to_sp): rect_mask_dim = ', rect_mask_dim
            endif
        endif

        ! Will we be too close to the edge of the micrograph to avoid artefacts?
        ! If we're doing ctf correction, we need to be at least sp_image_dim_large / 2 away from edge of micrograph
        ! If we're doing rectangular mask, we need to be at least rect_mask_dim(2)/2 away from the edge
        if (        any(wp%coo(1:2) .lt. (real(sp_image_dim_large(1:2)) / 2. + 1.5))   &
            .or.    any(wp%coo(1:2) .gt. (micrograph_frames(1)%logical_dimensions(1:2) &
                                       - (real(sp_image_dim_large(1:2)) / 2. + 1.5))) ) then
            if (debug) then
                write(*,'(a,2(F0.3,1x))') '**debug(wp_to_sp): too close to edge for ctf: ', wp%coo(1:2)
            endif
            ttoo_close_to_edge = .true.
        endif
        if (aapply_rect_mask) then
            if (        any(wp%coo(1:2) .lt. (                       rect_mask_dim(2) / 2. + 1.5))   &
                .or.    any(wp%coo(1:2) .gt. (micrograph_frames(1)%logical_dimensions(1:2) - (rect_mask_dim(2) / 2. + 1.5))) ) then
                if (debug) then
                    write(*,'(a,2(F0.3,1x))') '**debug(wp_to_sp): too close to edge for ctf: ', wp%coo(1:2)
                endif
                ttoo_close_to_edge = .true.
            endif
        endif

        ! Work out the image coordinates and non-integer shifts
        icoo = nint(wp%coo)
        icoo(3) = 1
        rcoo = wp%coo - icoo

        ! Cut out the image
        if (uunbend) then
            if (size(micrograph_frames) .gt. 1) then
                write(*,'(a)') '**ERROR(wp_to_sp): sorry, filament unbending not yet implemented with movie frames'
                call terminate('Feature not yet implemented')
            endif
            call filament_unbend_segment(   fil,wp%path_length,    &
                                            segment_length=sp_image_dim(2), &
                                            segment_width=sp_image_dim(1), &
                                            micrograph_image=micrograph_frames(1), & ! Eventually will have to pass the stack of frames
                                            unbent_segment=sp_image)
        else
            if (sp_image%IsAllocated()) call sp_image%Deallocate()
            temp_cutout_image%logical_dimensions = (/sp_image_dim_large(1),sp_image_dim_large(2),1/)
            ! Loop over frames
            do current_frame=ffirst_frame_to_sum,llast_frame_to_sum
                coordinates_on_current_frame = wp%coo - wp%local_subframe_shift(:,current_frame)
                icoo(1:2) = nint(coordinates_on_current_frame(1:2))
                icoo(3)   = 1
                rcoo      = coordinates_on_current_frame - icoo
                call micrograph_frames(current_frame)%ClipInto(temp_cutout_image,icoo)
                if (debug) then
                    write(*,'(a,i0,a,2(i0,1x))') '**debug(wp_to_sp): boxed out from frame ', current_frame, &
                                                    ' at coos: ', icoo(1:2)
                    write(*,'(a,2(f0.2,1x))') '**debug(wp_to_sp): wp%coo = ', wp%coo(1:2)
                    write(*,'(a,2(f0.2,1x))') '**debug(wp_to_sp): wp%local_subframe_shifts = ', &
                                                wp%local_subframe_shift(1:2,current_frame)
                endif
                if (aapply_shift_to_image) then
                    if (debug) write(*,'(a)') '**debug(wp_to_sp): applying phase shift'
                    call temp_cutout_image%ForwardFFT()
                    call temp_cutout_image%PhaseShift(-rcoo(1),-rcoo(2),0.0e0)
                endif
                ! Apply dose filter
                if (aapply_dose_filter) then
                    if (debug) write(*,'(a)') '**debug(wp_to_sp): applying dose filter'
                    if (temp_cutout_image%IsInRealSpace()) call temp_cutout_image%ForwardFFT()
                    call electron_dose%ApplyDoseFilterToImage(temp_cutout_image, &
                                                    dose_start=(current_frame-1)*exposure_per_frame, &
                                                    dose_finish=current_frame*exposure_per_frame, &
                                                    pixel_size=pixel_size)
                endif
                ! Apply drift filter
                if (aapply_drift_filter) then
                    if (debug) write(*,'(a)') '**debug(wp_to_sp): applying drift filter'
                    if (current_frame .eq. 1) then
                        shift_during_current_frame =  wp%local_subframe_shift(:,current_frame+1) &
                                                    - wp%local_subframe_shift(:,current_frame  )
                    else if (current_frame .eq. size(wp%local_subframe_shift,2)) then
                        shift_during_current_frame =  wp%local_subframe_shift(:,current_frame  ) &
                                                    - wp%local_subframe_shift(:,current_frame-1)
                    else
                        shift_during_current_frame = (wp%local_subframe_shift(:,current_frame+1) &
                                                    - wp%local_subframe_shift(:,current_frame-1))*0.5
                    endif
                    call temp_cutout_image%ApplyDriftFilter(shift_during_current_frame(1),shift_during_current_frame(2))
                endif
                ! Add image to running sum
                if (debug) call temp_cutout_image%WriteToDisk('dbg_temp_cutout_image.mrc')
                if (current_frame .eq. 1) then
                    sp_image = temp_cutout_image
                else
                    call sp_image%AddImage(temp_cutout_image)
                endif
            enddo
        endif

        if (debug) call sp_image%WriteToDisk('dbg_sp_image_1.mrc')

        ! Find the ctf for this waypoint
        ctf_local = ctf
        if (.not. ctf_local%pix_units) then
            if (.not. present(pixel_size)) then
                write(*,'(a)') '**ERROR(wp_to_sp): ctf is not in pixel units'
                call terminate('Fatal error in WP_TO_SP')
            else
                call ctf_ang_to_pix(ctf_local,pixel_size)
            endif
        endif
        call ctf_at_z(ctf_local,wp%coo(3))

        ! Apply the ctf
        if (aapply_ctf_to_image) then
            if (sp_image%IsInRealSpace()) call sp_image%ForwardFFT()
            call image_apply_ctf(sp_image,ctf_local,flip_only=ctf_flip_only)
        endif

        ! Apply shift to image
        if (aapply_shift_to_image) then

            ! the actual shift has already been done already done, therefore the single particle
            ! object will have no remaining shift
            rcoo = 0.0
        endif

        ! Rotate the image
        if (rrotate_sp_image) then
            ! Initialise interpolation
            interpol = default_interpol_scheme()
            call interpol_window%AllocateWindow(interpol%window_width,comp=.false.,threed=.false.)
            if (sp_image%IsInFourierSpace()) call sp_image%BackwardFFT()
            rot_tmp = wp%euler%psi - 90. !! i think i needed -90 to align with the y axis rather than the x axis. ... are my angles off? what about astig?
            if (wp%euler%theta .gt. 180.) then
                rot_tmp = rot_tmp + 180.    ! Some filaments have oop near 270 (rather than 90) because they have reverse polarity
            endif
            rotation_mat = rotmat(-rot_tmp,0.0,0.0)
            call sp_image%ApplyAffineTransformation(sp_image_rot,rotation_mat)
            sp_image = sp_image_rot
        endif

        if (debug) call sp_image%WriteToDisk('dbg_sp_image_5.mrc')

        ! Let's crop the image now that we're done with ctf
        !if (aapply_ctf_to_image) then
            if (sp_image%IsInFourierSpace()) call sp_image%BackwardFFT()
            call sp_image%Resize((/sp_image_dim(1),sp_image_dim(2),1/))
        !endif

        if (debug) call sp_image%WriteToDisk('dbg_sp_image_6.mrc')

        ! Normalise the image
        ! \todo     More accurate normalisation by normalising based on pixels outside helical rectangular mask, rather than outside circular mask
        if (sp_image%IsInFourierSpace()) call sp_image%BackwardFFT()
        if (debug) write(*,'(a,f0.3)') '**debug(wp_to_sp): about to norm with max rad = ', fil%max_radius
        call image_float_norm_mask(sp_image,object_radius=fil%max_radius,apply_mask=.false.,normalise_solvent=.true.)

        if (debug) call sp_image%WriteToDisk('dbg_sp_image_7.mrc')

        ! We'll apply a simple soft rectangular mask
        if (aapply_rect_mask) then
            azimuth = wp%euler%psi
            if (uunbend .or. rrotate_sp_image) azimuth=90.
            if (.not. present(fil)) call terminate('Need filament object as argument')

            call sp_image%ApplySoftRectangularMask(   &
                                rect_dim=rect_mask_dim,  &
                                azimuth=azimuth,falloff=5.0,wipe_input=.false.,mask_with_average=.true.)
        endif

        if (debug) call sp_image%WriteToDisk('dbg_sp_image_10.mrc')

        ! Initialise the single particle
        tmp_subunit_num = wp%subunit_num-nint(wp%subunit_num)
        if (debug) then
            write(*,'(a,F0.6)') '**debug(wp_to_sp): original subunit_num = ', wp%subunit_num
            write(*,'(a,F0.6)') '**debug(wp_to_sp): sp       subunit_num = ', tmp_subunit_num
        endif
        if (iignore_subunit_num) tmp_subunit_num = 0.0
        call sp_init(   sp,ctf_local,angles=wp%euler,shift_x=dble(rcoo(1)),shift_y=dble(rcoo(2)),    &
                        subunit_num=tmp_subunit_num,   &
                        helix=wp%helix)

        ! Were we too close to the edge?
        if (present(too_close_to_edge)) too_close_to_edge = ttoo_close_to_edge


        if (debug) call sp_image%WriteToDisk('dbg_sp_image_wp_to_sp_exit.mrc')
    end subroutine wp_to_sp




    !>  \brief  Calculate an unbent image of a segment of the filament centered around the specified point along its path
    subroutine filament_unbend_segment(self,center_t,segment_length,segment_width,micrograph_image,unbent_segment,interpol)
        use filaments_path
        use Splines
        use UsefulFunctions, only : IsOdd
        use Images
        use InterpolatingWindows
        use interpolation, only : interpolation_scheme, interpolation_scheme_init
        use UsefulFunctions, only : IsOdd
        implicit none
        ! Arguments
        type(filament),                         intent(inout)   ::  self            !<  Filament object
        real(kind=8),                           intent(in)      ::  center_t        !<  The parametrisation of the point along the filament which will be the center of the output unbent image
        integer,                                intent(in)      ::  segment_length  !<  Segment length, in pixels
        integer,                                intent(in)      ::  segment_width   !<  Segment width, in pixels
        type(image),                            intent(in)      ::  micrograph_image!<  Film image
        type(image),                            intent(inout)   ::  unbent_segment  !<  Output, unbent image
        type(interpolation_scheme), optional,   intent(in)      ::  interpol        !<  Interpolation scheme
        ! Private variables
        integer             ::  output_box_dim(2)
        integer             ::  i,j
        integer             ::  addr_origin(3), micrograph_origin(3)
        real                ::  coos_in_output(2), coos_in_micrograph(2)
        real(kind=8)        ::  t_normal, normal_vector(2), tangent_vector(2)
        type(waypoint)      ::  wp_normal
        type(window)        ::  win
        real                ::  rvalue
        integer             ::  addr_min(3), addr_max(3)
        type(interpolation_scheme)  ::  iinterpol
        ! Start work

        ! Interpolation scheme
        if (present(interpol)) then
            iinterpol = interpol
        else
            ! This needs to be replaced with some globally-available default interpolation scheme
            call interpolation_scheme_init(iinterpol,'LINEAR','RECT',2)
        endif

        ! Check filament is set up properly
        if (.not. self%IsAllocated()) then
            write(*,'(a)') '**ERROR(filament_unbend_segment): input filament not allocated'
            call terminate('Fatal error in FILAMENT_UNBEND_SEGMENT')
        endif

        ! Check micrograph image is set up properly
        if (.not. micrograph_image%IsAllocated()) then
            write(*,'(a)') '**ERROR(filament_unbend_segment): input micrograph image not allocated'
            call terminate('Fatal error in FILAMENT_UNBEND_SEGMENT')
        endif

        ! Decide on dimensions for output (take into account psf safety radius)
        output_box_dim(2) = segment_length
        output_box_dim(1) = segment_width
        if (IsOdd(output_box_dim(1))) output_box_dim(1) = output_box_dim(1) + 1
        if (IsOdd(output_box_dim(2))) output_box_dim(2) = output_box_dim(2) + 1
        write(*,'(a,2(i0,1x))') '**info(filament_unbend_segment): unbent segment box dimensions = ', output_box_dim

        ! Check t parameter is within range
        if (center_t - output_box_dim(2)/2 .lt. self%t(1) .or. center_t + output_box_dim(2)/2 .gt. self%t(self%nwp)) then
            write(*,'(a,i0,a,F0.12,a,i0,a,F0.12,a,F0.12)') '**ERROR(filament_unbend_segment): the segment (length ', &
                                                            output_box_dim(2) , ') centered around t (', center_t,  &
                                                            ') falls beyond the ends of filament ', self%id, &
                                                            ', which goes from ', self%t(1), ' to ', self%t(self%nwp)
            call terminate('Fatal error in FILAMENT_UNBEND_SEGMENT')
        endif

        ! Check the box size is not larger than the micrograph
        if (maxval(output_box_dim) .gt. minval(micrograph_image%logical_dimensions(1:2))) then
            write(*,'(a,2(i0,1x),a,i0,a,i0,a)') '**ERROR(filament_unbend_segment): output box size (', &
                                                output_box_dim, ') larger than input micrograph (', &
                                                micrograph_image%logical_dimensions(1),',', &
                                                micrograph_image%logical_dimensions(2),')'
            call terminate('Fatal error in FILAMENT_UNBEND_SEGMENT')
        endif

        ! Allocate memory for output
        unbent_segment%logical_dimensions = [OUTPUT_BOX_DIM(1),OUTPUT_BOX_DIM(2),1]
        call unbent_segment%Allocate()

        ! Address of origin
        if (unbent_segment%is_in_real_space .and. unbent_segment%object_is_centered_in_box) then
            addr_origin = unbent_segment%physical_address_of_box_center
        else
            addr_origin = [1,1,1]
        endif
        if (micrograph_image%is_in_real_space .and. micrograph_image%object_is_centered_in_box) then
            micrograph_origin = micrograph_image%physical_address_of_box_center
        else
            micrograph_origin = [1,1,1]
        endif

        ! Prepare the interpolating window
        call win%AllocateWindow(iinterpol%window_width,micrograph_image%IsInFourierSpace(),micrograph_image%IsAVolume())


        addr_min = micrograph_image%logical_lower_bound_real
        addr_max = micrograph_image%logical_upper_bound_real

        ! Loop over all pixels of output image
        do j=1,unbent_segment%logical_dimensions(2)
            coos_in_output(2) = real(j - addr_origin(2))
            ! t_normal = 2nd coordinate + center_t defines a point along the filament. The point we want to interpolate from lies somewhere on the normal to the filament at point t_normal
            t_normal = center_t + coos_in_output(2)

            ! Compute the coordinates of the micrograph point corresponding to t_normal
            wp_normal = waypoint_from_t(self,t_normal)

            ! The tangent vector
            tangent_vector(1) = self%x_spline%EvaluateDerivative(t_normal)
            tangent_vector(2) = self%y_spline%EvaluateDerivative(t_normal)

            ! The normal vector (there are two normal vectors - I'm not sure which one is correct; if the wrong one is used, we'll probably end up mirroring around the y axis)
            ! Option 1 of 2
            !normal_vector(1) = - tangent_vector(2)
            !normal_vector(2) =   tangent_vector(1)
            ! Option 2 of 2
            normal_vector(1) =   tangent_vector(2)
            normal_vector(2) = - tangent_vector(1)

            do i=1,unbent_segment%logical_dimensions(1)
                ! Compute coordinates of current point relative to center of output image
                coos_in_output(1) = real(i - addr_origin(1))

                ! The 1st coordinate defines the distance from the micrograph point to point t_normal along this normal.

                ! Compute the coordinates of the micrograph point we want to interpolate from
                coos_in_micrograph = (wp_normal%coo(1:2) - real(micrograph_origin(1:2))) + coos_in_output(1)* normal_vector

                ! Interpolate a value
                win%centered_on(1:2) = coos_in_micrograph
                win%centered_on(3) = 0.0
                call eval_window_bounds(win)
                if (any(win%lbound(1:3) .gt. addr_max) .or. any(win%ubound(1:3) .lt. addr_min)) then
                    cycle
                endif
                call window_from_image_real(micrograph_image,win)
                call get_real_from_window(rvalue,win,iinterpol%interpolation_function,iinterpol%windowing_function)

                !write(*,'(a,F0.3)') '**debug(filament_unbend_segment): rvalue = ', rvalue
                ! Stick that value into the output image
                unbent_segment%real_values(i,j,1) = rvalue
            enddo
        enddo
    end subroutine filament_unbend_segment


    !>  \brief  Print out a summary of all changes from filament 1 to filament 2
    !!  \todo   Use waypoint arithmetics (via operator overloading)
    subroutine filament_print_diff(fil1,fil2)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament), intent(in)      ::  fil1    !<  Filament, starting state
        type(filament), intent(in)      ::  fil2    !<  Filament, final state
        ! Private variables
        integer                         ::  i
        real                            ::  dX,dY,dZ,dR,dPhi,dTheta,dPsi,dPath,dS
        real                            ::  AdX,AdY,AdZ,AdR,AdPhi,AdTheta,AdPsi,AdPath,AdS,AdScore
        character(len=5),   parameter   ::  param_label(8)  =   ['X    ','Y    ','Z    ','Phi  ','Theta','Psi  ','Rise ','SubNm']
        ! Start work

        ! Check the two filaments have same number of waypoints
        if (fil1%nwp .ne. fil2%nwp) then
            write(*,'(a,2i3)') '**ERROR(filament_print_diff): the two filaments do not have the same number of waypoints: ', &
                                fil1%nwp, fil2%nwp
            write(*,'(a)') '**ERROR(filament_print_diff): info for filaments 1 and 2: '
            call fil1%PrintInfo()
            call fil2%PrintInfo()
            call terminate('Fatal error in FILAMENT_PRINT_DIFF')
        endif

        ! Loop over way points
        write(*,'(/a,i0)') 'Changes in filament ', fil1%id
        write(*,'(a7,1x,5(a11,1x),3(a11,1x),(a11,1x),(a7,1x))') &
                                            'WP#','dX','dY','dZ','dR','dPathLgth','dPhi','dTheta','dPsi','dS','dScore'
        AdX     = 0.
        AdY     = 0.
        AdZ     = 0.
        AdR     = 0.
        AdPhi   = 0.
        AdTheta = 0.
        AdPsi   = 0.
        AdPath  = 0.
        AdS     = 0.
        do I=1,fil1%nwp
            dX      = fil2%wp(i)%coo(1) - fil1%wp(i)%coo(1)
            dY      = fil2%wp(i)%coo(2) - fil1%wp(i)%coo(2)
            dZ      = fil2%wp(i)%coo(3) - fil1%wp(i)%coo(3)
            dR      = sqrt(dX**2+dY**2+dZ**2)
            dPath   = fil2%wp(i)%path_length - fil1%wp(i)%path_length
            dPhi    = fil2%wp(i)%euler%phi - fil1%wp(i)%euler%phi
            dTheta  = fil2%wp(i)%euler%theta - fil1%wp(i)%euler%theta
            dPsi    = fil2%wp(i)%euler%psi - fil1%wp(i)%euler%psi
            dS      = fil2%wp(i)%subunit_num - fil1%wp(i)%subunit_num
            write(*,'(i7,1x,5(F11.4,1x),3(F11.4,1x),1(F11.4,1x),1(a7,1x))')  i,dX,dY,dZ,dR,dPath,dPhi,dTheta,dPsi,dS,' '
            AdX     = AdX + abs(dX)
            AdY     = AdY + abs(dY)
            AdZ     = AdZ + abs(dZ)
            AdR     = AdR + dR
            AdPath  = AdPath + dPath
            AdPhi   = AdPhi + abs(dPhi)
            AdTheta = AdTheta + abs(dTheta)
            AdPsi   = AdPsi + abs(dPsi)
            AdS     = AdS + abs(dS)
        enddo
        AdX     = AdX / fil1%nwp
        AdY     = AdY / fil1%nwp
        AdZ     = AdZ / fil1%nwp
        AdR     = AdR / fil1%nwp
        AdPath  = AdPath / fil1%nwp
        AdPhi   = AdPhi / fil1%nwp
        AdTheta = AdTheta / fil1%nwp
        AdPsi   = AdPsi / fil1%nwp
        AdS     = AdS / fil1%nwp
        AdScore = fil2%score - fil1%score

        ! Print out averages
        write(*,'(a7,1x,5(F11.4,1x),3(F11.4,1x),1(F11.4,1x),1(F7.4)/)') &
                                                                'Average',AdX,AdY,AdZ,AdR,AdPath,AdPhi,AdTheta,AdPsi,AdS,AdScore

        ! Rigid parameters
        if (any(fil1%rigid_params)) then
            write(*,'(a)',advance='NO') &
                                    'Changes in rigid parameters: '
            do i=1,9
                if (fil1%rigid_params(i)) then
                    write(*,'(2a)',advance='NO') param_label(i), ': '
                    select case (i)
                        case (1:3)
                            write(*,'(G15.3)',advance='NO') fil2%rigid_wp%coo(i) - fil1%rigid_wp%coo(i)
                        case (4)
                            write(*,'(G15.3)',advance='NO') fil2%rigid_wp%euler%phi - fil1%rigid_wp%euler%phi
                        case (5)
                            write(*,'(G15.3)',advance='NO') fil2%rigid_wp%euler%theta - fil1%rigid_wp%euler%theta
                        case (6)
                            write(*,'(G15.3)',advance='NO') fil2%rigid_wp%euler%psi - fil1%rigid_wp%euler%psi
                        case (7)
                            write(*,'(G15.3)',advance='NO') fil2%rigid_wp%subunit_num - fil1%rigid_wp%subunit_num
                        case (8)
                            write(*,'(G15.3)',advance='NO') fil2%rigid_wp%helix%rise - fil1%rigid_wp%helix%rise
                        case (9)
                            write(*,'(G15.3)',advance='NO') fil2%rigid_wp%helix%twist - fil1%rigid_wp%helix%twist
                    end select
                    write(*,'(a)',advance='NO') '  '
                endif
            enddo
            write(*,'(a)') ' '
        endif

    end subroutine filament_print_diff



    !>  \brief  Write out a per-subunit trace to a csv file on disk
    subroutine filament_trace_to_disk_csv(self,filename)
        use filaments_path
        use coos
        implicit none
        ! Arguments
        type(filament),     intent(inout)   ::  self        !<  Filament object
        character(len=*),   intent(in)      ::  filename    !<  Filename to write to
        ! Private variables
        type(waypoint), allocatable     ::  wp(:)
        type(point),    allocatable     ::  coors(:)
        real(kind=8),   parameter       ::  trace_step  =   1.0D0
        ! Start work
        call filament_get_path(self,wp,trace_step,STEP_IN_REPEATS=.true.)
        allocate(coors(size(wp)))
!#ifdef __INTEL_COMPILER
!        call point_assign(coors,wp)
!#else
        coors = wp
!#endif
        call points_write_to_disk_csv(coors,filename)    ! i don't understand why i can't give wp as the argument here (see http://software.intel.com/en-us/forums/showthread.php?t=73564)
    end subroutine filament_trace_to_disk_csv

    !>  \brief  Write a boxer db containing the way points as boxes, and a trace as tiny boxes
    subroutine filament_boxer_db_to_disk(self,filename,with_trace)
        use filaments_path
        use coos
        implicit none
        ! Arguments
        type(filament),     intent(inout)   ::  self        !<  Filament object
        character(len=*),   intent(in)      ::  filename    !<  Filename to write to
        logical,    optional,   intent(in)  ::  with_trace  !<  Include the trace as tiny boxes? (by default, .TRUE.)
        ! Private variables
        type(waypoint), allocatable     ::  wp(:)
        type(boxer_db)                  ::  db
        real(kind=8),   parameter       ::  trace_step      =   1.0D0
        integer,        parameter       ::  trace_box_size  =   0
        integer,        parameter       ::  wp_box_size     =   10
        integer ::  i
        logical ::  wwith_trace
        ! Start work

        if (present(with_trace)) then
            wwith_trace = with_trace
        else
            wwith_trace = .true.
        endif

        !
        ! First, do the waypoints
        !
        call boxer_db_allocate(db,self%nwp)
!#ifdef __INTEL_COMPILER
!        do concurrent (I=1:self%nwp)
!            call point_assign(db%point(i),self%wp(i))
!        enddo
!#else
        forall(I=1:self%nwp) db%point(i) = self%wp(i)
!#endif
        call boxer_db_set_box_size(db,[WP_BOX_SIZE,wp_box_size,WP_BOX_SIZE])
        call boxer_db_write_to_disk(db,filename)

        !
        ! Now do the trace, if asked to do so
        !
        if (wwith_trace) then
            call filament_get_path(self,wp,trace_step,STEP_IN_REPEATS=.true.,IGNORE_SYMMETRY_MATES=.true.)
            call boxer_db_allocate(db,size(wp))
!#ifdef __INTEL_COMPILER
!            do concurrent (I=1:size(wp))
!                call point_assign(db%point(i),wp(i))
!            enddo
!#else
            forall (I=1:size(wp)) db%point(i) = wp(i)
!#endif
            call boxer_db_set_box_size(db,[TRACE_BOX_SIZE,trace_box_size,TRACE_BOX_SIZE])
            call boxer_db_write_to_disk(db,filename,APPEND=.true.)
        endif
    end subroutine filament_boxer_db_to_disk

    !>  \brief  return the b factor that should be applied to an image before it is inserted into a 3d reconstruction, given its score after refinement, and the average score of the dataset
    real function bfactor_from_score(score,average_score) result(bfactor)
        use runtime_parameters, only : pbc
        use rotation_matrices, only : degrees
        implicit none
        ! arguments
        real,       intent(in)      ::  score           !<  score of the image after refinement
        real,       intent(in)      ::  average_score   !<  average score of all images after refinement
        ! private variables
        logical,    parameter       ::  debug   =   .false.

        ! start work
        if (pbc .eq. 0.0) then
            bfactor = 0.0
        else
            if (debug) write(*,'(a,3(f0.4,1x))') '**debug(bfactor_from_score): average score, score, pbc: ', &
                                                degrees(acos(average_score)), degrees(acos(score)), pbc
            ! we want to return a negative b factor if the image has a better-than-average score
            bfactor = (degrees(acos(score)) - degrees(acos(average_score))) / pbc
            bfactor = max(bfactor,-2.0) ! a la frealign, we do not allow very negative b factors
            bfactor = bfactor * 16. ! this factor is there to match frealign. see notes of 29-apr-2012.
        endif

        if (debug) then
            write(*,'(a,3(f0.4,1x))') '**debug(bfactor_from_score): average score, score, bfactor: ', average_score, score, bfactor
        endif

    end function bfactor_from_score

    !>  \brief  Set the filament subframe shifts to the desired values (all waypoints are given the same shifts)
    subroutine filament_SetSubframeShifts(self,new_subframe_shifts)
        !
        class(Filament),    intent(inout)   ::  self
        real,               intent(in)      ::  new_subframe_shifts(:,:)    !<  First dimension: 1,2,3 for x,y,z; Second dimension: frame number
        !
        integer ::  current_wp
        !
        do current_wp=1,self%nwp
            self%wp(current_wp)%local_subframe_shift = new_subframe_shifts
        enddo
    end subroutine filament_SetSubframeShifts

    !>  \brief Check whether subframe shifts have been set
    pure logical function filament_SubframeShiftsAreSet(self)
        class(Filament),    intent(in)  ::  self
        ! private variables
        integer ::  current_wp
        ! start work
        filament_SubframeShiftsAreSet = .false.
        do current_wp=1,self%nwp
            if (allocated(self%wp(current_wp)%local_subframe_shift)) then
            if (any(self%wp(current_wp)%local_subframe_shift .ne. 0.0e0)) then
                filament_SubframeShiftsAreSet = .true.
            endif
            endif
        enddo
    end function filament_SubframeShiftsAreSet


    !>  \brief  Refine the local subframe shifts by doing local iterative realignment of regions around each waypoint
    subroutine filament_RefineSubframeShifts(self,box_size,apply_dose_filter)
        use Images
        use unblur_functions, only : unblur_refine_alignment
        use ctfs, only : ctf_parameters, ctf_ang_to_pix, ctf_at_z
        use images_ctf, only : ctf_to_img
        use UsefulFunctions, only : IsEven
        !
        class(Filament),    intent(inout)   ::  self
        integer,            intent(in)      ::  box_size            !<  Size of box to cut out around each waypoints to determine its local per-frame shifts
        logical,            intent(in)      ::  apply_dose_filter   !<  Apply dose filter to the frames before refining alignments
        ! private variables
        integer                     ::  current_wp
        integer                     ::  current_frame
        integer                     ::  integer_coordinates(3)
        integer                     ::  minimum_distance_from_edge
        real                        ::  remainder_coordinates(3)
        real,           allocatable ::  additional_x_shifts(:)
        real,           allocatable ::  additional_y_shifts(:)
        type(Image),    allocatable ::  current_set_of_images(:)
        type(Image),    allocatable ::  all_frames(:)
        logical,        parameter   ::  apply_ctf = .false.
        type(ctf_parameters)        ::  local_ctf
        type(image)                 ::  local_ctf_image
        ! Debug
        logical,    parameter       ::  debug = .false.
        type(Image)                 ::  frame_sum
        ! start work


        ! When we box things out of the micrograph, the center of the boxes should be at least this far away from the edges
        minimum_distance_from_edge = box_size / 2 + 1

        ! Allocate memory for the images
        allocate(current_set_of_images(self%film%GetNumberOfFrames()))
        allocate(all_frames(self%film%GetNumberOfFrames()))

        ! Debug
        if (apply_dose_filter) write(*,'(a)') '**debug(filament::RefineSubframeShifts): apply dose filter'

        ! Let's read all the frames in and hold them in memory
        do current_frame=1,self%film%GetNumberOfFrames()
            ! Read from disk
            call all_frames(current_frame)%ReadFromDisk(self%film%GetFilename(),self%film%GetFrameLocationInFile(current_frame))
            if (debug) then
                if (current_frame .eq. 1) then
                    frame_sum = all_frames(current_frame)
                else
                    call frame_sum%AddImage(all_frames(current_frame))
                endif
            endif
            if (apply_dose_filter) then
                if (all_frames(current_frame)%IsInRealSpace()) call all_frames(current_frame)%ForwardFFT()
                call self%film%electron_dose%ApplyDoseFilterToImage( &
                                                all_frames(current_frame), &
                                                dose_start=self%film%GetExposurePerFrame()*(current_frame-1), &
                                                dose_finish=self%film%GetExposurePerFrame()*current_frame, &
                                                pixel_size=self%film%GetPixelSize())
            endif
        enddo ! end of loop over frames

        if (debug) then
            call frame_sum%WriteToDisk('dbg_frame_sum.mrc')
        endif


        ! Loop over waypoints
        do current_wp=1,self%nwp
            write(*,'(a,i0,a,i0)') '**info(filament::RefineSubframeShifts): refining subframe shifts of waypoint ', current_wp, &
                                    ' of ', self%nwp

            ! We may need to know the CTF
            if (apply_ctf) then
                local_ctf = self%film%ctftilt%ctf
                call ctf_ang_to_pix(local_ctf,self%film%GetPixelSize())
                call ctf_at_z(local_ctf,self%wp(current_wp)%coo(3))
                if (IsEven(box_size)) then
                    call local_ctf_image%Allocate(dims=[box_size/2+1,box_size,1])
                else
                    call local_ctf_image%Allocate(dims=[(box_size-1)/2+1,box_size,1])
                endif
                call ctf_to_img(local_ctf,local_ctf_image,self%film%GetPixelSize(),one_at_origin=.true.)
            endif

            ! Loop over frames
            do current_frame=1,self%film%GetNumberOfFrames()
                ! Integer coordinates at which to box out
                integer_coordinates         = max(nint(self%wp(current_wp)%coo),minimum_distance_from_edge)
                integer_coordinates(1:2)    = min(integer_coordinates(1:2), &
                                                  self%film%GetLogicalDimensions()-minimum_distance_from_edge)
                integer_coordinates(3)      = 1
                ! Remainder, non-integer, coordinates
                remainder_coordinates = self%wp(current_wp)%coo - nint(self%wp(current_wp)%coo)
                ! Box out
                if (current_wp .eq. 1) then
                    call current_set_of_images(current_frame)%Allocate(dims=[box_size,box_size,1],in_real_space=.true.)
                endif
                if (debug) then
                    write(*,*) '**debug(filament::RefineSubframeShifts): wp coos = ', self%wp(current_wp)%coo
                    write(*,*) '**debug(filament::RefineSubframeShifts): integer = ', integer_coordinates
                    write(*,'(a,3(f0.1,1x))') '**debug(filament::RefineSubframeShifts): subframe shifts = ', &
                                                 self%wp(current_wp)%local_subframe_shift(:,current_frame)
                endif
                if (all_frames(current_frame)%IsInFourierSpace()) call all_frames(current_frame)%BackwardFFT()
                call all_frames(current_frame)%ClipInto(current_set_of_images(current_frame),integer_coordinates)

                if (current_set_of_images(current_frame)%IsConstant()) then
                    write(*,'(a)') '**warning(Filaments::RefineSubframeShifts): blank image boxed from frame'
                endif

                call current_set_of_images(current_frame)%ForwardFFT()

                ! Apply the CTF
                if (apply_ctf) then
                    current_set_of_images(current_frame)%complex_values(:,:,:) = &
                    current_set_of_images(current_frame)%complex_values(:,:,:) &
                    * local_ctf_image%real_values(1:local_ctf_image%logical_dimensions(1),:,:)
                endif


                ! Apply a phase shift consisting of:
                ! - the remaining non-integer part of coordinates
                ! + the subframe shift currently stored in the waypoint
                call current_set_of_images(current_frame)%PhaseShift( &
                                     -remainder_coordinates(1)+self%wp(current_wp)%local_subframe_shift(1,current_frame), &
                                     -remainder_coordinates(2)+self%wp(current_wp)%local_subframe_shift(2,current_frame), &
                                     0.0e0)

            enddo ! end of loop over frames
            ! See if the alignment can be improved further
            call unblur_refine_alignment( current_set_of_images, &
                                          20, & ! max iter
                                          1500.0, & ! b factor
                                          .true., 3,3, & ! central cross
                                          .false.,0.0, & ! circular mask
                                          0.0,real(box_size)*0.4, & ! min max radius for peak search
                                          0.1, &! convergence
                                          additional_x_shifts, additional_y_shifts, &
                                          self%film%GetPixelSize(), &
                                          indicate_progress=.false.  )
            ! Add the new shifts to the existing shifts
            self%wp(current_wp)%local_subframe_shift(1,:) =     self%wp(current_wp)%local_subframe_shift(1,:) &
                                                            +   additional_x_shifts(:)
            self%wp(current_wp)%local_subframe_shift(2,:) =     self%wp(current_wp)%local_subframe_shift(2,:) &
                                                            +   additional_y_shifts(:)
        enddo ! end of loop over waypoints
    end subroutine filament_RefineSubframeShifts


    !>  \brief  Calcuate a 3d reconstruction from a filament
    subroutine filament_calc_threed(self,micrograph_frames,rec3d,safe_box_size,finalise,finalised_vol,wipe,    &
                                    sp_mode,sp_segment_length,sp_helix, &
                                    apply_a_bfactor,average_score,bfactor_per_subunit,apply_drift_filter,   &
                                    first_frame_for_reconstruction,last_frame_for_reconstruction, &
                                    apply_dose_filter)
        use threeds
        use filaments_geometry, only : filament_average_wp_to_wp_distance
        use filaments_path, only : filament_get_path
        use filaments_model_modify, only : filament_convert_to_one_start
        use ctfs
        use ui, only : interval_print
        use runtime_parameters, only : is_even
        use omp_lib
        use coos, only : waypoints_print_info
        use Micrographs
        use rotation_matrices, only : euler_to_rotmat
        use images_ctf, only : image_apply_ctf
        use interpolation, only : default_interpol_scheme
        use helices, only : helical_symmetry_print_info
        use DatesAndTimes
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  self                !<  Filament
        type(image),                        intent(inout)   ::  micrograph_frames(:)!<  Frames of the micorgraph on which the filament is. Will not be modified, but need intent inout for window_img routine.
        type(threed),                       intent(inout)   ::  rec3d               !<  3d reconstruction, which will be computed by inserting every helical subunit
        integer,                optional,   intent(in)      ::  safe_box_size       !<  Minimum safe box size (pixels) for boxing out segments, ctf flipping them etc. This can be larger than the box size used for the 3d reconstruction itself.
        logical,                optional,   intent(in)      ::  finalise            !<  If .TRUE., a "finalised" 3d reconstruction will be returned (i.e. one that has been corrected for interpolation scheme, and where the ctf is corrected for)
        type(image),            optional,   intent(inout)   ::  finalised_vol       !<  The finalised 3d reconstruction volume (only needed if finalise = .TRUE.)
        logical,                optional,   intent(in)      ::  wipe                !<  Whether the 3d reconstruction should be wiped at beginning
        !logical,               optional,   intent(in)      ::  ignore_fal_wps      !<  .FALSE. by default. If .TRUE., only data between the second and second-to-last waypoints are used for the reconstruction.
        logical,                optional,   intent(in)      ::  sp_mode             !<  Single particle mode. In single particle mode, the reconstruction is calculated from points near the filament's waypoint,
                                                                                    !!  with the assumption that the filament is straight near each of the waypoints.
        real,                   optional,   intent(in)      ::  sp_segment_length   !<  Single particle segment length (pixels)
        type(helical_symmetry), optional,   intent(in)      ::  sp_helix            !<  Helical parameters to use for reconstruction in single-particle mode. If none given, the helical parameters found in
                                                                                    !!  the filament object or its waypoints will be used, which may not be desirable since in sp_mode, the full-filament model does not hold.
        logical,                optional,   intent(in)      ::  apply_a_bfactor     !<  Whether a B factor should be applied to the images before they are added to the 3d reconstruction. Defaults to false.
        real,                   optional,   intent(in)      ::  average_score       !<  Average score of filaments within the dataset. This will be used to work out what b factor to apply to the images. Defaults to 0.0.
        logical,                optional,   intent(in)      ::  bfactor_per_subunit !<  Apply the B factor filter per subunit, rather than per filament
        logical,                optional,   intent(in)      ::  apply_drift_filter  !<  Apply a drift filter per-frame and per-subunit
        integer,                optional,   intent(in)      ::  first_frame_for_reconstruction  !<  First frame of a dose fractionation movie to include in the 3D reconstruction
        integer,                optional,   intent(in)      ::  last_frame_for_reconstruction   !<  Last frame of a dose fractionation movie to include in the 3D reconstruction
        logical,                optional,   intent(in)      ::  apply_dose_filter               !<  Apply a filter to reflect expected radiation damage
        ! Private variables
        type(waypoint), allocatable     ::  subunits(:)
        logical                         ::  ffinalise
        integer                         ::  i
        integer                         ::  icoo(3)
        real                            ::  rcoo(3)
        type(image)                     ::  box!, box_tmp
        type(ctf_parameters)            ::  ctf
        logical                         ::  wwipe
        !logical                         ::  iignore_fal_wps
        integer                         ::  nsr ! Non-symmetry redundancy
        logical                         ::  ssp_mode
        integer                         ::  num_wps_per_wp  ! only used in sp_mode
        logical,    parameter           ::  threading_allowed   =   .true.      !<  Allow OpenMP threading?
        logical                         ::  tthreading
        logical                         ::  aapply_bfactor
        logical                         ::  aapply_drift_filter
        real                            ::  aaverage_score
        real                            ::  bfactor
        type(filament)                  ::  self_onestart
        logical                         ::  use_inclusion_threshold
        real                            ::  inclusion_threshold
        logical                         ::  include_current_subunit
        real                            ::  current_subunit_score
        integer                         ::  nearest_wp
        integer                         ::  num_subunits_included
        integer                         ::  ssafe_box_size
        integer                         ::  current_frame
        real                            ::  coordinates_for_current_wp_on_current_frame(3)
        real                            ::  shift_during_current_frame(3)
        !real                            ::  matrix(4,4)
        !type(interpolation_scheme)      ::  interpol
        logical,    parameter           ::  use_3d_spline   =   .true.  !<  The filament's z info is used for working out the defocus, rather than the ctftilt info in conjuction with (x,y)
        logical,    parameter           ::  use_nsr =   .true.  !< Whether to use non-symmetry-redundancy factor during reconstruction.
        logical                         ::  bbfactor_per_subunit
        logical,    parameter           ::  real_space_2d_shift_rot = .false.
        logical,    parameter           ::  debug = .false.
        type(DateAndTime)               ::  myTime
        integer                         ::  first_frame, last_frame
        type(Image)                     ::  dose_filter_for_current_frame, dose_filter_for_current_frame_swapped
        logical                         ::  aapply_dose_filter

        ! Start work

        ! Debug
        if (debug) print '(a,l1)', '**debug(filament_calc_threed): use_3d_spline = ', use_3d_spline

        ! Check filament is fully allocated
        if (.not. self%IsAllocated()) then
            write(*,'(a)') '**ERROR(filament_calc_threed): filament is not fully allocated (filament information follows)'
            call self%PrintInfo()
            call terminate('Fatal error in FILAMENT_CALC_THREED')
        endif

        ! Will we use inclusion threshold?
        use_inclusion_threshold = allocated(self%wp_score)
        if (debug) write(*,'(a)') '**debug(filament_calc_threed): overriding use_inclusion_threshold to F'
        use_inclusion_threshold = .false.

        ! Output a finalised 3d reconstruction?
        ffinalise = .false.
        if (present(finalise)) ffinalise = finalise

        ! Wipe?
        wwipe = .false.
        if (present(wipe)) wwipe = wipe


        ! Finalised 3d volume?
        if (ffinalise) then
            if (.not. present(finalised_vol)) then
                write(*,'(a)') '**ERROR(filament_calc_threed): finalised_vol argument not supplied, but finalise = .TRUE.'
                call terminate('Fatal error in FILAMENT_CALC_THREED')
            endif
        endif

        ! Ignore first and last (fal) way points (wps)
        !if (present(ignore_fal_wps)) then
        !   iignore_fal_wps = ignore_fal_wps
        !else
        !   iignore_fal_wps = .FALSE.
        !endif

        ! Single-particle mode
        ssp_mode = .false.
        if (present(sp_mode)) ssp_mode = sp_mode


        ! Apply b factor?
        aapply_bfactor = .false.
        if (present(apply_a_bfactor)) aapply_bfactor = apply_a_bfactor


        ! b factor per subunit?
        bbfactor_per_subunit = .false.
        if (present(bfactor_per_subunit)) bbfactor_per_subunit = bfactor_per_subunit

        ! drift filter
        aapply_drift_filter = .false.
        if (present(apply_drift_filter)) aapply_drift_filter = apply_drift_filter
        if (aapply_drift_filter) write(*,'(a)') '**info(filament_calc_threed): will apply drift filter to frames'

        ! Dose filter
        aapply_dose_filter = .false.
        if (present(apply_dose_filter)) aapply_dose_filter = apply_dose_filter
        if (aapply_dose_filter) write(*,'(a)') '**info(filament_calc_threed): will apply dose filter to frames'

        ! Which frames
        first_frame = 1
        last_frame = size(micrograph_frames)
        if (present(first_frame_for_reconstruction)) first_frame = first_frame_for_reconstruction
        if (present(first_frame_for_reconstruction)) last_frame = last_frame_for_reconstruction

        ! Average filament score ?
        if (present(average_score)) then
            aaverage_score = average_score
        else
            if (aapply_bfactor) then
                write(*,'(2a)') '**warning(filament_calc_threed): you asked to apply a b factor, but did not supply ',   &
                                'an average score. Will assume an average score of 0.0.'
            endif
            aaverage_score = 0.0
        endif

        ! b factor to be used during 3d reconstruction
        if (aapply_bfactor ) then
            if (bbfactor_per_subunit) then
                write(*,'(a,i0,a)') '**info(filament_calc_threed): the subunits in filament ', self%id, &
                                    ' will be added with b factors proportional to their local score'
            else
                bfactor = bfactor_from_score(self%score,aaverage_score)
                write(*,'(a,i0,a,F0.4,a,F0.2,a)') '**info(filament_calc_threed): filament ', self%id, ', with a score of ', &
                                                    self%score, ', will be added to the 3d reconstruction with a b factor of ', &
                                                    bfactor, ' pixels^2'
            endif
        else
            bfactor = 0.0
        endif

        ! Let the user know if we're doing single-particle mode
        if (ssp_mode) then
            write(*,'(3a)') '**info(filament_calc_threed): single-particle mode. Only subunits near filament waypoints will',    &
                            ' be included in the reconstruction, and their parameters will be computed by ',    &
                            'linear extrapolation from each waypoint.'
        else
            write(*,'(2a)') '**info(filament_calc_threed): normal filament mode. All subunits of the filament will be included', &
                            ' in the reconstruction.'
        endif

        ! Let the user know if we're ignoring first and last wps
        if (.not. self%fal_wps_in_threed) then
            write(*,'(2a)') '**info(filament_calc_threed): only data between second and second-to-last waypoints will be',   &
                            ' included in the 3d reconstruction'
        endif

        ! Check the threed object was initialised
        if (.not. is_allocated(rec3d)) then
            write(*,'(a)') '**ERROR(filament_calc_threed): 3d reconstruction object was not initialised'
            call terminate('Fatal error in FILAMENT_CALC_THREED')
        endif

        ! Wipe the threed
        if (wwipe) then
            call threed_wipe(rec3d)
        endif

        ! What will be the non-symmetry redundancy (nsr) caused by the insertion of the repeats?
        ! We approximate this as the number of repeats that fit in the reconstruction volume
        if (use_nsr) then
            if (self%helix%pixel_units) then
                nsr = nint(real(rec3d%logical_dimensions(1)) / self%helix%rise)
            else
                nsr = nint(real(rec3d%logical_dimensions(1)) / (self%helix%rise/self%film%GetPixelSize()))
            endif
            !call threed_set_nsr(rec3d,nsr)
            if (debug) print '(a,i0)', '**debug(filament_calc_threed): non-symmetry redundancy set to ', nsr
        endif

        ! We have the option to only include in the reconstruction those subunits in parts of the filament
        ! where the local score is above a threshold
        if (use_inclusion_threshold) then
            inclusion_threshold = filament_threshold_for_inclusion_into_threed(self)
            write(*,'(a,F0.3)') '**info(filament_calc_threed): threshold for inclusion in the 3d set to ', inclusion_threshold
        else
            inclusion_threshold = -1.0
        endif

        ! Obtain the coordinates and orientation of all helical subunits
        if (ssp_mode) then
            ! This option does not work with microtubules yet
            if (self%microtubule) then
                write(*,'(2a)') '**ERROR(filament_calc_threed): single-particle 3d reconstruction not enabled yet',  &
                                ' from microtubule filaments'
                call terminate('Fatal error from FILAMENT_CALC_THREED')
            endif
            if (.not. self%helix%pixel_units) then
                call terminate('Need helical parameters in pixels')
            endif
            ! Work out a suitable number of subunits for reconstruction from each waypoint
            if (.not. present(sp_segment_length)) then
                num_wps_per_wp = nint(    filament_average_wp_to_wp_distance(self)                   /self%helix%rise)
            else if (filament_average_wp_to_wp_distance(self) .eq. 0.0) then
                num_wps_per_wp = sp_segment_length/self%helix%rise
            else
                num_wps_per_wp = nint(min(filament_average_wp_to_wp_distance(self),sp_segment_length)/self%helix%rise)
            endif
            if (is_even(num_wps_per_wp)) num_wps_per_wp = num_wps_per_wp + 1
            write(*,'(a,i0,a)')         '**info(filament_calc_threed): ', num_wps_per_wp, &
                                        ' subunits per filament waypoint will be used for the 3d reconstruction'
            write(*,'(a,2(f0.2,a))')    '**info(filament_calc_threed): this number of subunits is equivalent to ', &
                                        num_wps_per_wp * self%helix%rise, ' pixels, or ', &
                                        num_wps_per_wp * self%helix%rise * self%film%GetPixelSize(), ' Angstroms.'
        endif
        self_onestart = self
        if (self%helix%num_starts .ne. 1) then
            write(*,'(a,i0,2a)') '**info(filament_calc_threed): filament is currently described by a ', self%helix%num_starts, &
                                '-start helical operator. Will be converted to 1-start helical operator so',    &
                                ' that subunits are explicitely descibed.'
            call terminate('filament_calc_threed: no conversion to onestart is available')
            !call filament_convert_to_one_start(self_onestart)
        endif

        if (debug) then
            write(*,'(a)') '**info(filament_calc_threed): filament info before get_path'
            call self_onestart%PrintInfo()
        endif

        ! Get array of waypoints, one per subunit
        call filament_get_path(         self_onestart, subunits, &
                                        step=1.0d0, &
                                        step_in_repeats=.true.,ignore_fal_wps=.not. self_onestart%fal_wps_in_threed,    &
                                        sp_mode=sp_mode,num_wps_per_wp=num_wps_per_wp,helix=sp_helix)


        ! Debug: only get a few positions to save time for testing
        !print *, '**debug(filament_calc_threed): not doing every subunit - testing'
        !subunits = filament_get_path(self,100.0,STEP_IN_PIXELS=.TRUE.)


        ! Safe box size?
        if (present(safe_box_size)) then
            ssafe_box_size = safe_box_size
        else
            ssafe_box_size = rec3d%logical_dimensions(1)
        endif
        if (debug) write(*,'(a,(i0,1x))') '**debug(filament_calc_threed): box dim = ', ssafe_box_size

        ! Debug: printout detailled path info
        write(*,'(3a)') '**info(filament_calc_threed): below is a list of the subunits ', &
                        'which will be included in the 3D reconstruction. ', &
                        'Note that views related by point-group symmetries (axial, perpendicular) are not explicitly listed.'
        call waypoints_print_info(subunits,phi_mod_360=.true.)

        ! Avoid nested parallelisation
#ifdef _OPENMP
        if (omp_get_active_level() .gt. 0 .and. threading_allowed) then
            tthreading = .false.
        else
            tthreading = .true.
        endif
#else
        tthreading = .false.
#endif

        ! Count how many subunits are actually added to the 3d reconstruction
        num_subunits_included = 0

        print '(a,i0,a,i0,a)', '**info(filament_calc_threed): will reconstruct filament ', self%id ,&
                                ' with ', size(subunits), ' subunits'

        !$omp   parallel    if(tthreading)              &
        !$omp               default(shared)               &
        !$omp               private(i,icoo,rcoo,ctf)    &
        !$omp               private(box)                &
        !$omp               firstprivate(bfactor)       &
        !$omp               private(current_subunit_score,include_current_subunit,nearest_wp,bbfactor_per_subunit) &
        !$omp               private(coordinates_for_current_wp_on_current_frame,shift_during_current_frame) &
        !$omp               reduction(+:num_subunits_included)

        ! Check that the number of threads is reasonable - good speed up only occurs with interpolation schemes which use large supports
#ifdef _OPENMP
        if (tthreading) then
            !$omp   single
            if (omp_get_num_threads() .gt. threed_get_window_width(rec3d) / 2) then
                write(*,'(a,i0,a,i0,a)') '**info(filament_calc_threed): threading with ', omp_get_num_threads(), &
               ' threads may not offer great speed-ups when using an interpolation scheme with kernel support (window width) of ', &
                threed_get_window_width(rec3d), ' pixels. In extreme cases, you may actually experience a slow-down.'
            endif
            !$omp   end single  nowait
        endif
#endif

        ! OpenMP private does not set up derived type objects properly, so let's do it explicitely here
        call box%Reset()
        !call box_tmp%Reset()
        box%is_in_memory     = .false.
        box%planned          = .false.
        box%logical_dimensions(3)           = 1
        !box_tmp%is_in_memory = .false.
        !box_tmp%planned      = .false.
        !box_tmp%logical_dimensions(3)       = 1

        box%logical_dimensions(1:2)         = ssafe_box_size
        !box_tmp%logical_dimensions(1:2)     = ssafe_box_size


        ! Outer loop over dose-fractionation movie frames
        do current_frame=first_frame,last_frame

            ! Progress for the user
            !$omp single
            if (size(micrograph_frames) .gt. 1) then
                write(*,'(a,i0,a,i0)') '**info(filament_calc_threed): on frame ', current_frame, ' of ', size(micrograph_frames)
            endif
            !$omp end single nowait

            ! Work out the dose filter for this frame
            !$omp critical (threed_generate_dose_filter)
            call dose_filter_for_current_frame%Allocate(box,in_real_space=.true.)
            dose_filter_for_current_frame = 1.0e0
            if (aapply_dose_filter) then
                call self%film%electron_dose%ApplyDoseFilterToImage(dose_filter_for_current_frame, &
                                                                    dose_start=(current_frame-1)*self%film%GetExposurePerFrame(), &
                                                                    dose_finish=current_frame*self%film%GetExposurePerFrame(), &
                                                                    pixel_size=self%film%GetPixelSize())
            endif
            dose_filter_for_current_frame_swapped = dose_filter_for_current_frame
            call dose_filter_for_current_frame_swapped%SwapRealSpaceQuadrants()
            !$omp end critical (threed_generate_dose_filter)

            ! This variable keeps track of the nearest wp, which is worked out by the function to find the local score
            nearest_wp = 1


            ! Loop over the helical subunits
            !$omp   do
            do i=1,size(subunits)
                if (interval_print(i,size(subunits))) then
                    write(*,'(3a,i4,a,i4)',advance='NO') '**info(filament_calc_threed): ', myTime%DateAndTimeAsString(), &
                                                        ': working on subunit ', i, ' of ', size(subunits)
#ifdef _OPENMP
                    if (tthreading) then
                        write(*,'(a,i0,a,i0)') ' on thread ', omp_get_thread_num()+1, ' of ', omp_get_num_threads()
                    else
                        write(*,'(a)') ' '
                    endif
#else
                    write(*,'(a)') ' '
#endif
                endif
                ! Find the score of the current subunit, if needed
                if (use_inclusion_threshold .or. (aapply_bfactor .and. bbfactor_per_subunit)) then
                    ! Find the local score
                    current_subunit_score = filament_find_local_score(self,subunits(i)%path_length,nearest_wp)
                endif
                ! Will we include the current subunit (based on its score)?
                if (use_inclusion_threshold) then
                    include_current_subunit = current_subunit_score .gt. inclusion_threshold
                    if (debug) then
                    write(*,'(a,i0,a,F0.2,a,F0.3,a,l1)') &
                                    '**debug(filament_calc_threed): subunit ', i, ' at path length ', subunits(i)%path_length,   &
                                    ' has a score of ', current_subunit_score, '. Is this above threshold? ', &
                                    include_current_subunit
                    endif
                else
                    include_current_subunit = .true.
                endif
                ! Will we include the current subunit (based on its helical parameters)?
                include_current_subunit =       include_current_subunit &
                                        .and.   subunits(i)%helix%rise  .ge. self%helix_min_for_3drec%rise &
                                        .and.   subunits(i)%helix%twist .ge. self%helix_min_for_3drec%twist &
                                        .and.   subunits(i)%helix%rise  .le. self%helix_max_for_3drec%rise &
                                        .and.   subunits(i)%helix%twist .le. self%helix_max_for_3drec%twist

                if (debug) then
                    call helical_symmetry_print_info(subunits(i)%helix)
                    call helical_symmetry_print_info(self%helix_min_for_3drec)
                    call helical_symmetry_print_info(self%helix_max_for_3drec)
                    write(*,'(a,i0,a,l1)') '**debug(filament_calc_threed): subunit ', i, &
                                            ' included based on helical parameters? ', &
                                                subunits(i)%helix%rise  .ge. self%helix_min_for_3drec%rise &
                                        .and.   subunits(i)%helix%twist .ge. self%helix_min_for_3drec%twist &
                                        .and.   subunits(i)%helix%rise  .le. self%helix_max_for_3drec%rise &
                                        .and.   subunits(i)%helix%twist .le. self%helix_max_for_3drec%twist
                endif


                if (include_current_subunit) then
                    num_subunits_included = num_subunits_included + 1
                    ! Calculate the appropriate ctf parameters
                    if (use_3d_spline) then
                        ! The average ctf for the film
                        ctf = self%film%ctftilt%ctf
                        ! Convert ctf parameters to pixel units
                        if (.not. ctf%pix_units) call ang_to_pix(ctf,self%film%GetPixelSize())
                        ! The ctf at z (but with astigmatism)
                        call ctf_at_z(ctf,subunits(i)%coo(3))
                    else
                        ctf = self%film%ReturnCTFAtCoos(subunits(i)%coo(1),subunits(i)%coo(2))
                        ! Convert ctf parameters to pixel units
                        if (.not. ctf%pix_units) call ang_to_pix(ctf,self%film%GetPixelSize())
                    endif
                    ! Take into account local subframe shifts
                    coordinates_for_current_wp_on_current_frame = subunits(i)%coo &
                                                                - subunits(i)%local_subframe_shift(:,current_frame)
                    ! Get integer coordinates & remainders
                    icoo = nint(coordinates_for_current_wp_on_current_frame)
                    icoo(3) = 1
                    rcoo = coordinates_for_current_wp_on_current_frame - icoo
                    ! Window image
                    box%is_in_real_space = .true.
                    box%object_is_centered_in_box = .true.
                    if (debug) then
                    write(*,'(2(a,i0),3(a,2(f0.2,1x)),a,2(i0,1x))') &
                                                        '**debug(filament_calc_threed): subunit ', i, ' on frame ', current_frame, &
                                                        ' subunit coo: ', subunits(i)%coo(1:2), &
                                                        ' subframe shifts: ', subunits(i)%local_subframe_shift(1:2,current_frame), &
                                                        ' rcoo: ', rcoo(1:2), &
                                                        ' icoo: ', icoo(1:2)
                    endif
                    call micrograph_frames(current_frame)%ClipInto(box,icoo)
                    if (box%IsConstant()) write(*,'(a)') '**warning(filament_calc_threed): blank image boxed from micrograph'
                    ! Do the translation shift in Fourier space, don't do in-plane rotatoin
                    call box%ForwardFFT()
                    call box%PhaseShift(-rcoo(1),-rcoo(2),0.0)

                    ! Compute the subunit's b factor
                    if (aapply_bfactor .and. bbfactor_per_subunit) then
                        bfactor = bfactor_from_score(current_subunit_score,aaverage_score)
                        if (bfactor .lt. -500.) then
                            bfactor = -500.
                            write(*,'(a,i0,a,i0,a,F0.3,a)') '**warning(filament_calc_threed): subunit ', &
                                                            i, ' of ', size(subunits), &
                                                            ' had a high score and its b factor was limited to ', bfactor, ' A^2'
                        endif
                        write(*,'(a,i0,a,i0,a,F0.3,a)') '**debug(filament_calc_threed): subunit ', i, ' of ', size(subunits), &
                                                        ' will be added with a b factor of ', bfactor, ' A^2'
                    endif
                    ! Add slice to the 3d reconstruction
                    if (aapply_drift_filter) then
                        if (current_frame .eq. 1) then
                            shift_during_current_frame =  subunits(i)%local_subframe_shift(:,current_frame+1) &
                                                        - subunits(i)%local_subframe_shift(:,current_frame  )
                        else if (current_frame .eq. size(micrograph_frames)) then
                            shift_during_current_frame =  subunits(i)%local_subframe_shift(:,current_frame  ) &
                                                        - subunits(i)%local_subframe_shift(:,current_frame-1)
                        else
                            shift_during_current_frame = (subunits(i)%local_subframe_shift(:,current_frame+1) &
                                                        - subunits(i)%local_subframe_shift(:,current_frame-1))*0.5
                        endif
                    else
                        shift_during_current_frame = 0.0
                    endif
                    call threed_add_slice(  rec3d,box,ctf,subunits(i)%euler,bfactor=bfactor/self%film%GetPixelSize()**2,  &
                                            apply_ctf_to_slice=.true.,&
                                            add_to_threed_num=1+int(2.*real(i-1)/real(size(subunits))), &
                                            drift_x=shift_during_current_frame(1), &
                                            drift_y=shift_during_current_frame(2), &
                                            apply_dose_filter=aapply_dose_filter, &
                                            dose_filter_image=dose_filter_for_current_frame, &
                                            dose_filter_image_swapped=dose_filter_for_current_frame_swapped &
                                            )

                    ! Debug: check for NaNs
                    if (debug) then
                        if (threed_has_nan(rec3d)) then
                            write(*,'(a,i0,a,i0)') '**debug(filament_calc_threed): 3d reconstruction has at least '//&
                                                   'one NaN value after inserting slice ', i , ' of ', size(subunits)
                        endif
                    endif
                endif
            enddo ! end of loop over subunits
            !$omp   end do

        enddo ! end of loop over movie frames
        if (debug) write(*,'(a)') '**debug(filament_calc_threed): about to deallocate images'
        call box%Deallocate()
        !call box_tmp%Deallocate()
        if (debug) write(*,'(a)') '**debug(filament_calc_threed): about to leave parallel region'
        !$omp   end parallel

        write(*,'(3a,i0,a,i0,a)') '**info(filament_calc_threed): ', myTime%DateAndTimeAsString(), ': ', &
                                    num_subunits_included, ' of ', size(subunits),  &
                                    ' subunits have been added to the 3d reconstruction'

        ! If required, compute a weighted / corrected 3d reconstruction volume
        if (ffinalise) then
            call threed_eval_reconstruction(rec3d,finalised_vol)
        endif

    end subroutine filament_calc_threed

    !>  \brief  Find the local score at a point along the filament specified by its path length
    real function filament_find_local_score(self,path_length,nearest_wp_last_time)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament), intent(in)      ::  self                    !<  Filament
        real(kind=8),   intent(in)      ::  path_length             !<  Ordinate at which to evaluate the score
        integer,        intent(inout)   ::  nearest_wp_last_time    !<  The index of the nearest wapoint the last time
                                                                    !!  this function was called. Set this to 1 the first
                                                                    !!  time you call this function.
        ! Private variables
        integer         ::  nearest_wp, nearest_previous_wp
        integer         ::  current_wp
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        ! Check the score was actually computed
        if (.not. allocated(self%wp_score)) then
            write(*,'(a)') '**ERROR(filament_find_local_sc0re): score has not yet been computed'
            call terminate('Fatal error in FILAMENT_FIND_LOCAL_SCORE')
        endif

        if (debug) then
            write(*,'(a,i0)') '**debug(filament_find_local_sc0re): nearest wp last time = ',  nearest_wp_last_time
        endif

        nearest_wp = 0
        nearest_previous_wp = 0
        ! Iterate from previous guess to end of filament
        do CURRENT_WP=MAX(1,NEAREST_WP_LAST_TIME-1),size(self%wp_score) - 1
            if (    path_length .ge. self%wp_score(current_wp)%path_length .and. &
                    path_length .le. self%wp_score(CURRENT_WP+1)%path_length ) then
                nearest_previous_wp = current_wp
            endif
        enddo
        ! If we still haven't found it, look from start of filament to previous guess
        if (nearest_previous_wp .eq. 0) then
            if (debug) write(*,'(a)') '**debug(filament_find_local_sc0re): searching from start of filament'
            do CURRENT_WP=1,max(1,NEAREST_WP_LAST_TIME-2)
                if (    path_length .ge. self%wp_score(current_wp)%path_length .and. &
                        path_length .le. self%wp_score(CURRENT_WP+1)%path_length ) then
                    nearest_previous_wp = current_wp
                endif
            enddo
        endif
        ! If we still haven't found it, look after the last wp_score, but before the last wp
        if (nearest_previous_wp .eq. 0) then
            if (debug) write(*,'(a)') '**debug(filament_find_local_sc0re): searching at end of filament'
            if (    path_length .ge. self%wp_score(size(self%wp_score))%path_length .and. &
                    path_length .le. self%wp(self%nwp)%path_length) then
                nearest_previous_wp = size(self%wp_score)
            endif
        endif
        ! Now we know which interval our point is in, find out which wp it's nearest to
        if (nearest_previous_wp .eq. 0) then
            write(*,'(a,F0.3,a)') '**ERROR(filament_find_local_score): point with path_length ', path_length, &
                            ' does not appear to lie on filament '
            call self%PrintInfo()
            call terminate('Fatal error in FILAMENT_FIND_LOCAL_SCORE')
        else
            if (    path_length - self%wp_score(nearest_previous_wp)%path_length .le. &
                    self%wp_score(NEAREST_PREVIOUS_WP+1)%path_length - path_length ) then
                nearest_wp = nearest_previous_wp
            else
                nearest_wp = NEAREST_PREVIOUS_WP+1
            endif
        endif

        if (nearest_wp .eq. 0) then
            ! If we haven't found the nearest wp by now and we haven't already crashed out, something
            ! strange is happening.
            call terminate('Programming error in FILAMENT_FIND_LOCAL_SCORE')
        else
            nearest_wp_last_time = nearest_wp    ! Save this for the next time the function is called
            filament_find_local_score = self%wp_score(nearest_wp)%score
        endif
    end function filament_find_local_score


    !>  \brief  Compute a thresold for inclusion into the 3d reconstruction
    real function filament_threshold_for_inclusion_into_threed(self) result(threshold)
        implicit none
        ! Arguments
        type(filament), intent(in)  ::  self    !<  Filament
        ! Private variables
        real,   parameter   ::  num_stdev = 1.0
        real(kind=8)    ::  ave,std
        ! Start work

        ! Get statistics for filament score
        if (allocated(self%wp_score)) then
            call filament_segment_score_stats(self,ave,std)
            threshold = ave - NUM_STDEV*STD
        else
            ! This case arises when the scoring function has never been called,
            ! for example during filament_startup
            threshold = -1.0
        endif
    end function filament_threshold_for_inclusion_into_threed





    !>  \brief  Compute average and std dev of segment scores
    subroutine filament_segment_score_stats(self,ave,std)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament), intent(in)  ::  self    !<  Filament
        real(kind=8),   intent(out) ::  ave     !<  Average of scores
        real(kind=8),   intent(out) ::  std     !<  Standard deviation of scores
        ! Private variables
        integer ::  current_wp
        ! Start work
        ave = 0.0D0
        std = 0.0D0
        if (allocated(self%wp_score)) then
            do CURRENT_WP=1,size(self%wp_score)
                ave = ave + self%wp_score(current_wp)%score
                std = std + self%wp_score(current_wp)%SCORE**2
            enddo
            std = (std - (AVE**2/size(self%wp_score)))/size(self%wp_score)
            std = sqrt(std)
            ave = ave / size(self%wp_score)
        else
            write(*,'(a)') '**ERROR(filament_segment_score_stats): wp_score array not allocated. Scoring probably not done.'
            call terminate('Fatal error in FILAMENT_SEGMENT_SCORE_STATS')
        endif
    end subroutine filament_segment_score_stats


    !>  \brief  Shift waypoints of a filament so that when reconstructed, the desired phi and Z shifts will have been applied to the 3D
    !!
    subroutine filament_shift_wps_to_obtain_3d_z_phi_shift(self,wanted_3d_z_shift,wanted_3d_phi_shift)
        use coos, only : waypoint_apply_shift
        use filaments_model_update, only : filament_update_model
        ! Arguments
        type(filament),                 intent(inout)   ::  self
        real,                           intent(in)      ::  wanted_3d_z_shift
        real,                           intent(in)      ::  wanted_3d_phi_shift
        ! Private variables
        integer                 ::  wp_counter
        real                    ::  shift_array(9)
        logical,    parameter   ::  debug   =   .true.
        !
        shift_array = 0.0
        do wp_counter=1,self%nwp
            ! Subunit_num
            shift_array(7) = -wanted_3d_z_shift / self%wp(wp_counter)%helix%rise
            ! Phi
            shift_array(4) = shift_array(7) * self%wp(wp_counter)%helix%twist + wanted_3d_phi_shift
            if (debug) then
                write(*,'(a,i0,a,f0.2,a,f0.2,a)') &
                                '**debug(filament_shift_wps_to_obtain_3d_z_phi_shift): shifting wp ', wp_counter, &
                                ' by ', shift_array(7), ' subunits and ', shift_array(4), ' degrees'
            endif
            call waypoint_apply_shift(self%wp(wp_counter),shift_array)
        enddo
        call filament_update_model(self)
    end subroutine filament_shift_wps_to_obtain_3d_z_phi_shift



    !>  \brief  Waypoints with non-integer subunit_num cannot be compared to the 3d model, which always has subunit_num=0.0 at origin.
    !!          This subroutine works out how the model would need to be rotated and shifted to be compared to that waypoint.
    !!
    !!  See notes of December 7th, 2011, for details
    !!
    subroutine filament_model_phasing_from_wp(self,wp,model_z_shift,local_h_sym,model3d_oversampling)
        use rotation_matrices
        use filaments_model_base, only : wp_compute_subunit_num_phasing
        use helices, only : helical_symmetry_unit_convert
        implicit none
        ! Arguments
        type(filament),                 intent(inout)   ::  self
        type(waypoint),                 intent(inout)   ::  wp                      !<  Waypoint. Its phi value will be modified to account for the non-integer subunit_num value.
        real(kind=8),                   intent(out)     ::  model_z_shift           !<  This z shift (pixels) should be applied to the model before calculating a projection
        logical,                        intent(in)      ::  local_h_sym             !<  Whether the model 3d should be phased (shifted & rotated) using local helical paramters. Defaults to true.
        integer,                        intent(in)      ::  model3d_oversampling    !<  Oversampling factor (1 if no overscaling) which has been applied to the 3d reference volume
        ! Private variables
        real                    ::  subunit_num_shift
        real                    ::  subunit_num_phi
        logical,    parameter   ::  debug   =   .false.
        ! Start work


        ! Check helical symmetry is in pixel units
        if (.not. self%helix%pixel_units) call helical_symmetry_unit_convert(self%helix,self%film%GetPixelSize())

        ! Work out additional shift due to model phasing by subunit_num
        subunit_num_shift = wp_compute_subunit_num_phasing(wp) ! In units of subunit_num

        ! Working out the phi rotation
        if (local_h_sym) then
            subunit_num_phi = subunit_num_shift * wp%helix%twist
        else
            subunit_num_phi = subunit_num_shift * self%helix%twist
        endif

        ! Convert from subunit number to axial distance
        if (local_h_sym) then
            subunit_num_shift = subunit_num_shift * wp%helix%rise * model3d_oversampling
        else
            subunit_num_shift = subunit_num_shift * self%helix%rise * model3d_oversampling
        endif

        ! This is the amount we would need to translate the model along z
        model_z_shift   = - subunit_num_shift

        ! This is the amount we would need to rotate the model around z
        subunit_num_phi = - subunit_num_phi

        ! Rather than rotate the model itself, we'll just extract the slice at a different angle
        wp%euler%phi = wp%euler%phi + subunit_num_phi

        ! 150820: not sure why this wasn't done here.. warning: bug potential
        ! 151021: looks like this made one of the tests fail. commenting out
        ! 151025: not sure anymore - leaving uncommented - running tests...
        wp%subunit_num = nint(wp%subunit_num)

     end subroutine filament_model_phasing_from_wp



    !>  \brief  Modify a waypoint's phi and (x,y) coordinates so that it can be compared
    !!          directly to the 3d model with canonical helical symmetry (i.e. it will coincide
    !!          exactly with a helical subunit)
    !!
    !!  See also documentation for FILAMENT_COMPUTE_WP_PHASING.
    !!
    !!  Note that with high curvature, it might be better to adjust the t value for the waypoint rather than do a xy shift as in here, so that the curvature might be better accounted for.
    !!  Because this may require more major re-writes, and because i anticipate the phase shifts to be small, i am implementing things like this for now (May/June 2011).
    subroutine filament_apply_phasing_to_wp_old(self,wp,local_h_sym,neg)
        use rotation_matrices
        use filaments_model_base, only : wp_compute_subunit_num_phasing
        use runtime_parameters, only : terminate
        use helices, only : helical_symmetry_unit_convert
		implicit none
        ! Arguments
        type(filament),                 intent(inout)   ::  self
        type(waypoint),                 intent(inout)   ::  wp
        logical,            optional,   intent(in)      ::  local_h_sym !<  Whether the model 3d we are phasing against is h-symmetrised for every waypoint. Defaults to false.
        logical,            optional,   intent(in)      ::  neg
        ! Private variables
        real                    ::  subunit_num_shift
        real                    ::  subunit_num_phi
        real                    ::  subunit_num_shift_xy(2)
        real                    ::  mat(4,4)
        logical,    parameter   ::  debug   =   .false.
        integer                 ::  calc_neg
        logical                 ::  llocal_h_sym
        logical,    parameter   ::  sanity_check    =   .false.
        real                    ::  expected_shift_xy(2)
        ! Start work

        ! Is the model symmetrised using the local helical symmetry?
        if (present(local_h_sym)) then
            llocal_h_sym = local_h_sym
        else
            llocal_h_sym = .false.
        endif

        ! Are we calculating the negative?
        calc_neg = 1
        if (present(neg)) then
            if (neg) then
                calc_neg = -1
            else
                calc_neg = 1
            endif
        endif

        ! Check helical symmetry is in pixel units
        if (.not. self%helix%pixel_units) call helical_symmetry_unit_convert(self%helix,self%film%GetPixelSize())

        ! Work out additional shift due to model phasing by subunit_num
        subunit_num_shift = wp_compute_subunit_num_phasing(wp) * real(calc_neg) ! In units of subunit_num

        ! Working out the phi rotation
        if (llocal_h_sym) then
            subunit_num_phi = subunit_num_shift * wp%helix%twist
        else
            subunit_num_phi = subunit_num_shift * self%helix%twist
        endif

        ! Convert from subunit number to axial distance
        if (llocal_h_sym) then
            subunit_num_shift = subunit_num_shift * wp%helix%rise
        else
            subunit_num_shift = subunit_num_shift * self%helix%rise
        endif

        ! Convert axial distance to xy in the projection
        ! The safest way to do this might be with an actual rotation matrix used by the projection routines.
        ! We think of the shift as a point on the z axis, and we multiply it by the rotation matrix
        mat = euler_to_rotmat(wp%euler%phi,wp%euler%theta,wp%euler%psi,reverse=.false.)
        subunit_num_shift_xy(1:2) = mat(1:2,3) * subunit_num_shift
        if (debug) then
            write(*,'(a,4(G15.3,1x))') '**debug(filament_apply_phasing_to_wp): subunit_num_shift, subunit_num_shift_x, '//&
                                    'subunit_num_shift_y, subunit_num_phi = ', &
                                    subunit_num_shift / wp%helix%rise, subunit_num_shift_xy(1:2), subunit_num_phi
        endif

        ! Sanity check (see notes 1-Dec-2011 for derivation) - minus sign is because clockwise positive convention
        ! 23-Jan: doesn't appear to work correctly
        if (sanity_check) then
            expected_shift_xy(1) = cos(-radians(wp%euler%psi)) * cos(-radians(90.-wp%euler%theta)) * subunit_num_shift
            expected_shift_xy(2) = sin(-radians(wp%euler%psi)) * cos(-radians(90.-wp%euler%theta)) * subunit_num_shift
            ! Check the shifts are as expected
            if (abs(subunit_num_shift_xy(1) - expected_shift_xy(1)) .gt. 0.001 .or. &
                abs(subunit_num_shift_xy(2) - expected_shift_xy(2)) .gt. 0.001) then
                write(*,'(a)') '**ERROR(filament_apply_phasing_to_wp): Sanity check failed.'
                write(*,'(a,1(F0.5,1x))') '**ERROR(filament_apply_phasing_to_wp):        subunit_num = ', wp%subunit_num
                write(*,'(a,2(F0.5,1x))') '**ERROR(filament_apply_phasing_to_wp): expected x y shift = ', expected_shift_xy
                write(*,'(a,2(F0.5,1x))') '**ERROR(filament_apply_phasing_to_wp):   actual x y shift = ', subunit_num_shift_xy
                call terminate('Failed sanity check in FILAMENT_APPLY_PHASING_TO_WP')
            endif
        endif

        ! Apply the shift and the rotation
        wp%coo(1:2) = wp%coo(1:2) - subunit_num_shift_xy(1:2)
        wp%euler%phi = wp%euler%phi - subunit_num_phi

        !
        wp%subunit_num = nint(wp%subunit_num)
    end subroutine filament_apply_phasing_to_wp_old


    !>  \brief  Rescale the coordinates of the filament such that it matches a resampled micrograph (where the resampling was done by cropping the Fourier transform)
    subroutine filament_change_pixel_size(self,original_micrograph_dimensions,new_micrograph_dimensions,scaling_factor)
        use coos, only : waypoints_rescale_coos, waypoint_rescale_coos
        use ctfs, only : ctf_change_pixel_size
        use Helices, only : helix_scale_distances
        ! arguments
        type(filament), intent(inout)   ::  self
        integer,        intent(in)      ::  original_micrograph_dimensions(3)
        integer,        intent(in)      ::  new_micrograph_dimensions(3)
        real,           intent(in)      ::  scaling_factor  !<  The coordinates will be multiplied by this factor
        !
        type(waypoint), allocatable     ::  tmp_wps(:)
        type(filament)                  ::  tmp_fil
        integer                         ::  wp_counter

        ! Change the coordinates
        tmp_wps = self%wp
        tmp_fil = self
        call waypoint_rescale_coos(tmp_fil%rigid_wp,original_micrograph_dimensions,new_micrograph_dimensions,scaling_factor)
        call waypoints_rescale_coos(tmp_wps,original_micrograph_dimensions,new_micrograph_dimensions,scaling_factor)

        ! Don't forget to change the pixel size of the film object also
        call self%film%SetPixelSize(self%film%GetPixelSize() / scaling_factor)
        if (self%film%ctftilt%ctf%pix_units) then
            !print *, '**debug(filament_change_pixel_size): fixing filament ctf'
            call ctf_change_pixel_size(self%film%ctftilt%ctf,scaling_factor)
        else
            !print *, '**debug(filament_change_pixel_size): not doing anything to filament ctf'
        endif

        ! As well as the film's dimenions
        call self%film%SetLogicalDimensions(new_micrograph_dimensions)


        !! The helical parameters won't have been updated by filament_init if wp_as_sps
        !if (self%wps_as_sps .or. self%fragmented_subunit_num) then
            !tmp_fil%helix%rise = self%helix%rise * scaling_factor
            !tmp_wps%helix%rise = self%wp%helix%rise * scaling_factor
            tmp_fil%helix = helix_scale_distances(self%helix,scaling_factor)
            do wp_counter=1,size(tmp_wps)
                tmp_wps(wp_counter)%helix = helix_scale_distances(self%wp(wp_counter)%helix,scaling_factor)
            enddo
        !endif
        call filament_init(self,waypoints_in=tmp_wps,mould=tmp_fil,film_in=self%film)


        ! Don't forget about persistence lengths
        self%persistence_length = self%persistence_length * scaling_factor
        self%torsion_persistence_length = self%torsion_persistence_length * scaling_factor
        self%stretch_persistence_length = self%stretch_persistence_length * scaling_factor



        ! Maximum radius
        self%max_radius = self%max_radius * scaling_factor
    end subroutine filament_change_pixel_size




    !>  \brief  Resample waypoints along the path of a filament
    subroutine filament_resample_waypoints(self,step_size,step_in_repeats,step_in_phi)
        use filaments_path, only : filament_get_path_subunit_steps, filament_get_path_pixel_steps, filament_get_path_crossovers
        use coos, only : waypoints_print_info, point, waypoint, waypoints_remove_too_close
        use filaments_model_update, only : filament_update_model
		implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self            !<  Filament object
        real(kind=8),               intent(in)      ::  step_size       !<  Desired distance from waypoint to waypoint, in pixels
        logical,        optional,   intent(in)      ::  step_in_repeats !<  The step size was given in numbers of repeats
        logical,        optional,   intent(in)      ::  step_in_phi     !<  The step size was given in phi degrees
        ! Private variables
        type(filament)              ::  fil_tmp
        type(waypoint), allocatable ::  wps_tmp(:)
        integer                     ::  min_num_wps
        logical                     ::  ignore_fal_segments, sstep_in_repeats, sstep_in_phi
        ! Start work

        ! Will we ignore the first and last segments?
        ignore_fal_segments = .not. self%fal_wps_in_score .or. .not. self%fal_wps_in_threed
        if (ignore_fal_segments) then
            min_num_wps = 5
        else
            min_num_wps = 3
        endif

        sstep_in_repeats = .false.
        if (present(step_in_repeats)) sstep_in_repeats = step_in_repeats
        sstep_in_phi     = .false.
        if (present(step_in_phi)) sstep_in_phi = step_in_phi

        ! Info
        if (sstep_in_repeats) then
            write(*,'(a,i0,a,F0.2,a)') '**info(filament_resample_waypoints): will resample waypoints along the path of filament ', &
                                        self%id, ' so that waypoints are ', step_size, ' repeats apart'
        else if (sstep_in_phi) then
            write(*,'(a,i0,a,F0.2,a)') '**info(filament_resample_waypoints): will resample waypoints along the path of filament ', &
                                        self%id, ' so that waypoints are ', step_size, ' degrees apart in phi (twist)'
        else
            write(*,'(a,i0,a,F0.2,a)') '**info(filament_resample_waypoints): will resample waypoints along the path of filament ', &
                                        self%id, ' so that waypoints are ', step_size, ' pixels apart'
        endif


        ! Get new waypoints
        if (sstep_in_repeats) then
            call    filament_get_path_subunit_steps(self,wps_tmp,int(step_size),ignore_fal_wps=ignore_fal_segments)
        else if (sstep_in_phi) then
            ! Turn ignore_fal_wps off because often the 2nd wp has phi=3 degrees (or some small value)
            ! and we end up losing the first crossover
            call    filament_get_path_crossovers(   self,wps_tmp,step=1,ignore_fal_wps=ignore_fal_segments .and. .false.,   &
                                                    ignore_symmetry_mates=.true., &
                                                    phi_step_size=step_size, &
                                                    include_fal_wps=ignore_fal_segments)
        else
            call    filament_get_path_pixel_steps(  self,step_size,wps_tmp,   &
                                                    ignore_fal_wps=ignore_fal_segments, &
                                                    include_finish_wp=.true.,   &
                                                    include_fal_wps=ignore_fal_segments)
        endif

        call waypoints_remove_too_close(wps_tmp)


        if (size(wps_tmp) .ge. min_num_wps) then
            ! Initialise new filament
            call filament_init( fil_tmp,waypoints_in=wps_tmp,                                   &
                                film_in                     =   self%film,                      &
                                mould                       =   self)


            ! Deallocate the original filament
            call self%Deallocate()

            ! Copy the new filament over
            fil_tmp%score = self%score
            self = fil_tmp

            ! Make sure the model is up-to-date
            call filament_update_model(self)

            ! Now that the model (and therefore the arc lengths) have been updated, we need to checks that the waypoints
            ! are not too close to each other
            call filament_remove_waypoints_too_close(self)
        else
            write(*,'(a,i0,a)') '**warning(filament_resample_waypoints): resampling would have led to less than ',min_num_wps,&
                            ' waypoints. Will do nothing.'
        endif

    end subroutine filament_resample_waypoints

    !>  \brief  Remove waypoints which are too close to each other
    subroutine filament_remove_waypoints_too_close(self)
        use coos, only : waypoints_remove_too_close
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        ! Private variables
        type(waypoint), allocatable ::  wps_tmp(:)
        type(filament)              ::  self_tmp
        ! Start work
        wps_tmp = self%wp
        call filament_init(self_tmp,waypoints_in=self%wp,mould=self,called_by_filament_remove_waypoints_too_close=.true.)
        call waypoints_remove_too_close(wps_tmp)
        call filament_init(self,waypoints_in=wps_tmp,mould=self_tmp,called_by_filament_remove_waypoints_too_close=.true.)
    end subroutine filament_remove_waypoints_too_close


    !>  \brief  Fix obvious inconsistencies in the waypoints parameters, as related to filament polarity. This is useful
    !!          when working with parameters determined by single-particle analysis. With full-filament refinement (or any
    !!          refinement within Frealix), there should never be occation to use this, since the polarity state should always
    !!          be consistent.
    subroutine filament_fix_polarity_inconsistencies(self,allow_polarity_flip,no_init)
        use runtime_parameters, only : terminate
        use differential_geometry, only : expected_d_signs_given_theta_psi
        use differential_geometry, only : angle_between_euler_angles
        use filaments_model_update, only : filament_update_model
        use filaments_model_modify, only : filament_waypoint_alt_theta, filament_reverse_order_of_waypoints
        use filaments_model_modify, only : filament_flip_waypoint_polarity
        use filaments_geometry, only : filament_wp_euler_are_consistent_with_general_direction
        use filaments_geometry, only : filament_wp_euler_are_consistent_with_differential_geometry
        use filaments_geometry, only : filament_relax_helix
        use filaments_model_modify, only : filament_reverse_order_of_waypoint_pairs
        use Globals, only : pi
        !
        type(filament), intent(inout)   ::  self
        logical,        intent(in)      ::  allow_polarity_flip
        logical,    optional,   intent(in)  ::  no_init
        ! private variables
        logical                         ::  theta_mostly_near_90
        integer                         ::  current_wp
        logical                         ::  euler_and_d_agree(self%nwp)
        logical,            parameter   ::  debug   =   .false.
        type(waypoint),     allocatable ::  tmp_wps(:)
        type(filament)                  ::  tmp_fil
        logical                         ::  do_init
        logical                         ::  euler_angles_agree_with_first_wp(self%nwp)
        logical                         ::  first_wp_euler_angles_are_correct
        real                            ::  angle_with_previous
        ! start work

        do_init = .true.
        if (present(no_init)) do_init = .not. no_init

        if (debug) then
            write(*,'(a)') '**debug(filament_fix_polarity_inconsistencies): filament info on entry'
            call self%PrintInfo()
        endif

        if (self%nwp .gt. 1) then
            !
            ! There may be segments which have the opposite polarity to the rest of the filament
            ! so let's flip their Euler angles around
            !
            ! To do this, we will assume that the polarity is basically whether theta ~90.0 or theta ~270.0 and
            ! work out which is the dominant case.
            !
            ! Note that this steps just manipulates the Euler angle without changing their "meaning", i.e. without
            ! changing the resulting projection. It's just cosmetic.
            !
            !
            theta_mostly_near_90 = (real(count(self%wp%euler%theta .lt. 180.0d0)) / real(self%nwp)) .gt. 0.5
            do current_wp=1,self%nwp
                if (      (theta_mostly_near_90 .and. self%wp(current_wp)%euler%theta .gt. 180.0d0) &
                    .or. &
                    (.not. theta_mostly_near_90 .and. self%wp(current_wp)%euler%theta .lt. 180.0d0)) then
                    if (debug) write(*,'(a,i0)') '**debug(filament_fix_polarity_inconsistencies): alt theta for waypoint ', &
                                                    current_wp
                    call filament_waypoint_alt_theta(self,current_wp)
                endif
            enddo


            !
            ! The order of the waypoints may need to be reversed.
            !
            do current_wp=1,self%nwp
                euler_and_d_agree(current_wp) = filament_wp_euler_are_consistent_with_general_direction(self,current_wp)
            enddo
            if (debug) print *, '**debug(filament_fix_polarity_inconsistencies): euler_and_d_agree = ', euler_and_d_agree

            !
            ! If most of the Euler angles disagree with the derivatives, we definitely need to reverse the order of waypoints
            if (count(euler_and_d_agree) .lt. self%nwp/2) then
                write(*,'(a)') '**info(filament_fix_polarity_inconsistencies): most or all waypoints'' euler '//&
                                'angles contradict local x,y,z derivatives. Reversing order of waypoints'
                call filament_reverse_order_of_waypoints(self)
                ! The phi angles may now be off - for example they may be increasing whereas the twist in negative.
                ! We shouldn't really care right now because we're in wps_as_sps mode, but we might as well make the sign
                ! of the twist agree.
                call filament_relax_helix(self)
            endif

            !
            ! Sometimes, pairs of waypoints had their order reversed by an earlier bug. Let's try to fix this
            !
            call filament_reverse_order_of_waypoint_pairs(self)

            !
            ! Let's ensure that Euler angles agree with the diff geometry and conventions of frealix so that we
            ! can safely go to and from Euler angles and tangent vectors
            !
            do current_wp=1,self%nwp
                euler_and_d_agree(current_wp) = filament_wp_euler_are_consistent_with_differential_geometry(self,current_wp)
            enddo
            if (count(euler_and_d_agree) .lt. self%nwp/2) then
                write(*,'(a)') '**info(filament_fix_polarity_inconsistencies): most or all waypoints'' euler '//&
                                'angles contradict Frealix convention. Switching to alternative set of Euler angles.'
                do current_wp=1,self%nwp
                    call filament_waypoint_alt_theta(self,current_wp)
                enddo
            endif


            !
            ! Now let's make sure the filament is up-to-date by re-initialising it
            !
            if (debug) then
                write(*,'(2a)') '**debug(filament_fix_polarity_inconsistencies): filament info',&
                                    ' after cosmetic changes, before first init'
                call self%PrintInfo()
            endif

            if (do_init) then
                tmp_wps = self%wp
                tmp_fil = self
                call filament_init(self,waypoints_in=tmp_wps,mould=tmp_fil,called_by_filament_fix_polarity_inconsistencies=.true.)
            endif

            if (debug) then
                write(*,'(a)') '**debug(filament_fix_polarity_inconsistencies): filament info after first init'
                call self%PrintInfo()
            endif

            !
            ! Now let's work out which waypoints contradict the rest of the filament by measuring
            ! the angle between the (theta,psi) vectors for subsequent waypoints
            !
            if (allow_polarity_flip) then
                euler_angles_agree_with_first_wp(1) = .true. ! by definition
                do current_wp=2,self%nwp
                    angle_with_previous = angle_between_euler_angles(&
                                                self%wp(current_wp-1)%euler%theta,self%wp(current_wp-1)%euler%psi,&
                                                self%wp(current_wp  )%euler%theta,self%wp(current_wp  )%euler%psi)
                    euler_angles_agree_with_first_wp(current_wp) = (euler_angles_agree_with_first_wp(current_wp-1) &
                                                                    .and. angle_with_previous .lt. pi*0.5) .or. &
                                                                   (.not. euler_angles_agree_with_first_wp(current_wp-1) &
                                                                    .and. angle_with_previous .ge. pi*0.5)
                enddo
                if (debug) then
                    write(*,'(a)',advance='no') '**debug(filament_fix_polarity_inconsistencies): agree with first: '
                    write(*,*) euler_angles_agree_with_first_wp
                endif

                ! Is the first waypoint in the majority?
                first_wp_euler_angles_are_correct = count(euler_angles_agree_with_first_wp) .ge. self%nwp/2

                ! Now let's flip the minority waypoints into submission
                do current_wp=1,self%nwp
                    if (     (first_wp_euler_angles_are_correct .and. .not. euler_angles_agree_with_first_wp(current_wp)) &
                        .or. (.not. first_wp_euler_angles_are_correct .and. euler_angles_agree_with_first_wp(current_wp))) then
                        write(*,'(a,i0)') '**info(filament_fix_polarity_inconsistencies): flipping polarity of waypoint ', &
                                                    current_wp
                        call filament_flip_waypoint_polarity(self,current_wp)
                    endif
                enddo

                ! The phi angles may now be off - for example they may be increasing whereas the twist in negative.
                ! We shouldn't really care right now because we're in wps_as_sps mode, but we might as well make the sign
                ! of the twist agree.
                if (.not. all(euler_and_d_agree)) call filament_relax_helix(self)

                ! Re-initialise the filament
                tmp_wps = self%wp
                tmp_fil = self
                call filament_init(self,waypoints_in=tmp_wps,mould=tmp_fil,called_by_filament_fix_polarity_inconsistencies=.true.)

                ! Debug
                if (debug) then
                    write(*,'(a)') '**debug(filament_fix_polarity_inconsistencies): filament info after polarity flipping'
                    call self%PrintInfo()
                endif
            endif

        endif ! Test for nwp > 1


        if (debug) then
            write(*,'(a)') '**debug(filament_fix_polarity_inconsistencies): filament info on exit'
            call self%PrintInfo()
        endif

    end subroutine filament_fix_polarity_inconsistencies



    !>  \brief  Initialise a filament by giving it a set of coordinates
    !!
    !!
    !!  As of 9-Jun-2010, option add_extra_fal_wps does not appear to work on wobbly filaments and will probably not be refined/tested further. As of 20-Jul-2010, this is even more so,
    !!  because i do not intend to fully test the 3d spline upgrade in combination with this option. Sep-2011: significant fixes to the spline extrapolation routines may mean guessing
    !!  fal waypoins works again, but i have not tested this. Jul-2012: add_extra_fal_wps should work.
    !!
    !!  \todo   Simplify this subroutine or fragment it. See for example mode_estimate_pl_slave for 4 lines to initialise a simple filament without helix.
    recursive subroutine filament_init(   self, &
                                points_in, waypoints_in, &
                                film_in,crossovers,helix, &
                                helix_min_for_3drec, helix_max_for_3drec, &
                                id, &
                                add_extra_fal_wps, &
                                fal_wps_are_crossovers,max_radius,    &
                                microtubule,mt_pf_number,mt_rise,refine_params,rigid_params,rigid_wp,   &
                                constant_rise_per_subunit,constant_twist_per_length,    &
                                wps_as_sps,persistence_length,torsion_persistence_length,stretch_persistence_length,        &
                                fal_wps_in_score,fal_wps_in_threed,   &
                                refine_normal_modes,average_z_to_ctftilt_plane,use_hermite_splines, &
                                fragmented_phi, fragmented_subunit_num, amyloid_beta,mould,&
                                called_by_filament_remove_waypoints_too_close, &
                                called_by_filament_fix_polarity_inconsistencies, &
                                recompute_subframe_shifts_from_unblur_results, &
                                frame_shifts_from_unblur,subframe_alignment_box_size, &
                                refresh_fal_wps,apply_dose_filter)
		use coos
		use Micrographs
		use filaments_model_update, only : filament_update_model
		use filaments_normal_modes, only : filament_wps_to_nm_amplitudes, filament_update_nms
		use filaments_geometry, only : filament_update_wp_res, filament_fit_splines_linear, filament_eval_total_curvature
		use filaments_geometry, only : filament_splines_are_monotonic
		use filaments_geometry, only : filament_relax_twist_per_subunit, filament_relax_rise_per_subunit
		use filaments_model_base, only : filament_check_subunit_num, filament_guess_wps_phi
		use filaments_model_base, only : filament_check_subunit_num_increases
		use filaments_model_update, only : filament_update_rigid_wp, filament_update_average_helix_twist, filament_update_wp_helix
		use filaments_model_update, only : filament_update_average_helix_rise
		use filaments_model_modify, only : filament_shift_average_z_to_ctftilt_plane
		use Splines
		use Globals
		use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament),         target,     intent(inout)   ::  self                                !<  Filament object to populate
        type(point),            optional,   intent(in)      ::  points_in(:)                        !<  Array of points
        type(waypoint),         optional,   intent(in)      ::  waypoints_in(:)                     !<  Array of waypoints
        type(Micrograph),       optional,   intent(in)      ::  film_in                             !<  a film object (already initiated) from which the film properties will be inherited
        logical,                optional,   intent(in)      ::  crossovers                          !<  Whether the points given are crossovers (defaults to .FALSE.). If so and no angles are given, they will be set to 0/180/360/...
        type(helical_symmetry), optional,   intent(in)      ::  helix                               !<  (Average) helical parameters for the filament.
        type(helical_symmetry), optional,   intent(in)      ::  helix_min_for_3drec                 !<  Smallest values of the helical parameters when selecting which waypoints to allow into the 3D reconstruction
        type(helical_symmetry), optional,   intent(in)      ::  helix_max_for_3drec                 !<  Largest values of the helical parameters when selecting which waypoints to allow into the 3D reconstruction
        integer,                optional,   intent(in)      ::  id
        logical,                optional,   intent(in)      ::  add_extra_fal_wps                   !<  Add extra waypoints before/after the first-and-last (fal) provided waypoints (wps).
        logical,                optional,   intent(in)      ::  fal_wps_are_crossovers              !<  First and last (fal) waypoints (wps) are also crossovers, i.e. their twist angles are known.
                                                                                                    !!  By default, all wps are crossovers (if crossovers was .TRUE. of course), so this option is true.
        real,                   optional,   intent(in)      ::  max_radius                          !<  Maximum filament radius (pixels)
        logical,                optional,   intent(in)      ::  microtubule                         !<  Whether the filament is a microtubule
        integer,                optional,   intent(in)      ::  mt_pf_number                        !<  In the case of microtubules, the number of protofilaments
        integer,                optional,   intent(in)      ::  mt_rise                             !<  How much further along the filament (distance measured in tubulin subunits) the end of a MT_PF_NUMBER-element "washer"
                                                                                                    !!  is compared to its beginning. The 13-pf microtobules have a value of 3.
        logical,                optional,   intent(in)      ::  refine_params(9)                    !<  Which of the 7 degrees of freedom to refine per way point (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi,7.Subunit_num,8.HRise,9.HTwist)
        logical,                optional,   intent(in)      ::  rigid_params(9)                     !<  Whether to keep the following parameters fixed throughout the filament: 1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta, 6.Psi, 7-H.Rise, 8-H.Twist.
                                                                                                    !!  If all are set to .TRUE., the filament is a rigid rod, treated as a single particle.
        type(waypoint),         optional,   intent(in)      ::  rigid_wp                            !<  Rigid waypoint
        logical,                optional,   intent(in)      ::  constant_rise_per_subunit           !<  Whether we'll assume the rise-per-subunit to be constant (i.e. no cumulative axial disorder)
        logical,                optional,   intent(in)      ::  constant_twist_per_length           !<  Wether we should assume a constant twist per unit length along the filament
        logical,                optional,   intent(in)      ::  wps_as_sps                          !<  Whether waypoints are treated as independent single particles
        real,                   optional,   intent(in)      ::  persistence_length                  !<  Persistence length (pixels)
        real,                   optional,   intent(in)      ::  torsion_persistence_length          !<  Torsional persistence length (pixels)
        real,                   optional,   intent(in)      ::  stretch_persistence_length          !<  Stretching persistence length (pixels)
        logical,                optional,   intent(in)      ::  fal_wps_in_score                    !<  If false, the parts of the filament between the first and the second as well as between the penultimate and
                                                                                                    !!  last waypoints will not be included in the scoring function
        logical,                optional,   intent(in)      ::  fal_wps_in_threed                   !<  If false, the parts of the filament between the first and the second as well as between the penultimate and
                                                                                                    !!  last waypoints will not be included in the 3d reconstruction
        !logical,                optional,   intent(in)      ::  refine_in_fourier_space             !<  Whether the filament model or its Fourier transform should be refined
        logical,                optional,   intent(in)      ::  refine_normal_modes(9)              !<  Whether normal mode amplitudes should be refined, instead of the real-space filament model
        logical,                optional,   intent(in)      ::  average_z_to_ctftilt_plane          !<  Shift the filament in z so that its average lies on the plane described by the ctftilt plane
        logical,                optional,   intent(in)      ::  use_hermite_splines                 !<  Use Hermite splines (i.e. use theta,psi to constrain the space curve)
        logical,                optional,   intent(in)      ::  fragmented_phi                      !<  There are discontinuities in the phi function between waypoints
        logical,                optional,   intent(in)      ::  fragmented_subunit_num              !<  There are discontinuities in the subunit_num function between waypoints
        logical,                optional,   intent(in)      ::  amyloid_beta                        !<  The filament is an amyloid beta fibril
        type(filament),         optional,   intent(in)      ::  mould                               !<  Another filament object from which properties will be copied unless they are specified by other arguments
                                                                                                    !!  (allocatable components, i.e. waypoints, splines and films, are not copied over). Cannot be the same as SELF.
        logical,                optional,   intent(in)      ::  called_by_filament_remove_waypoints_too_close !<  If we are being called by filament_remove_waypoints_too_close, we need to know this is the case to avoid a recursive call
        logical,                optional,   intent(in)      ::  called_by_filament_fix_polarity_inconsistencies
        ! Movie processing
        logical,                optional,   intent(in)      ::  recompute_subframe_shifts_from_unblur_results   !<  If true, caller needs to provide unblur results as next argument
        real,                   optional,   intent(in)      ::  frame_shifts_from_unblur(:,:)       !<  Should have extent of 3 in first dimension (1,2,3 for x,y,z) and number_of_frames for second dimension
        integer,                optional,   intent(in)      ::  subframe_alignment_box_size         !<  Pixels. Size of the box to cut out around waypoints when working out subframe alignments.
        logical,                optional,   intent(in)      ::  refresh_fal_wps                     !<  The first and last waypoints should be removed and then added again using extrapolation from the other waypoints
        logical,                optional,   intent(in)      ::  apply_dose_filter                   !<  Apply dose (radiation damage) filter to micrographs, in particular when recomputing subframe shifts
        ! Private variables
        integer                 ::  i
        logical                 ::  aadd_extra_fal_wps
        logical                 ::  estimate_phi
        integer                 ::  min_num_points
        logical                 ::  z_unknown
        integer                 ::  twist_sign
        !logical                 ::  wwps_as_sps
        logical                 ::  fal_wps_phi_tbd !<  Whether the first-and-last waypoints' phi angles are to be determined
        logical,    parameter   ::  debug           = .false.
        logical,    parameter   ::  rrelax_helix    = .true.
        logical,    parameter   ::  refresh_fal_wps_if_not_in_score_or_threed = .true.
        logical                 ::  rrefresh_fal_wps
        integer                 ::  first_xover_wp,last_xover_wp
        logical                 ::  known_psi_theta
        logical                 ::  aaverage_z_to_ctftilt_plane
        type(waypoint), allocatable :: wp_bak(:)
        character(len=6)        ::  nwp_c
        type(point),    allocatable ::  points(:)
        type(waypoint), allocatable ::  waypoints(:)
        real                    ::  total_curvature
        integer                 ::  id_copy
        logical                 ::  ccalled_by_filament_remove_waypoints_too_close
        logical                 ::  ccalled_by_filament_fix_polarity_inconsistencies
        if (present(id)) id_copy = id


        ! Start work

        !
        ! By default
        !
        estimate_phi = .false.
        fal_wps_phi_tbd = .false.
        aaverage_z_to_ctftilt_plane = .false.
        ccalled_by_filament_remove_waypoints_too_close = .false.
        ccalled_by_filament_fix_polarity_inconsistencies = .false.
        if(present(called_by_filament_remove_waypoints_too_close)) then
            ccalled_by_filament_remove_waypoints_too_close = called_by_filament_remove_waypoints_too_close
        endif
        if (present(called_by_filament_fix_polarity_inconsistencies)) then
            ccalled_by_filament_fix_polarity_inconsistencies = called_by_filament_fix_polarity_inconsistencies
        endif
        if (present(average_z_to_ctftilt_plane)) aaverage_z_to_ctftilt_plane = average_z_to_ctftilt_plane
        call self%ResetProperties()


        !
        ! Do we have a mould?
        !
        if (present(mould)) then
            call self%CopyPropertiesFrom(mould)
            if (.not. present(film_in)) then
                self%film = mould%film
            endif
        endif


        !
        ! If optional parameters were given, heed them
        !
        if (present(id))                        self%id                         = id_copy
        if (present(wps_as_sps))                self%wps_as_sps                 = wps_as_sps
        if (present(crossovers))                self%wps_are_crossovers         = crossovers
        if (present(fal_wps_are_crossovers)) then
                                                self%fal_wps_are_crossovers     = fal_wps_are_crossovers
        else
            if (.not. present(mould))           self%fal_wps_are_crossovers     = self%wps_are_crossovers
        endif
        if (present(fal_wps_in_score))          self%fal_wps_in_score           = fal_wps_in_score
        if (present(fal_wps_in_threed))         self%fal_wps_in_threed          = fal_wps_in_threed
        if (present(microtubule))               self%microtubule                = microtubule
        if (present(amyloid_beta))              self%amyloid_beta               = amyloid_beta
        if (present(refine_params))             self%refine_params              = refine_params
        if (present(rigid_params)) then
                                                self%rigid_params               = rigid_params
        else
            if (.not. present(mould)) then
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(2a)') '**info(filament_init): no rigid parameters supplied, overriding with flexible space curve ',&
                                    '& flexible helical parameters'
                endif
                self%rigid_params = .false.
            endif
        endif
        if (present(rigid_wp))                  self%rigid_wp = rigid_wp
        if (present(constant_rise_per_subunit)) self%constant_rise_per_subunit  = constant_rise_per_subunit
        if (present(constant_twist_per_length)) self%constant_twist_per_length  = constant_twist_per_length
        if (present(film_in))                   self%film                       = film_in
        if (present(persistence_length))        self%persistence_length         = persistence_length
        if (present(torsion_persistence_length))self%torsion_persistence_length = torsion_persistence_length
        if (present(stretch_persistence_length))self%stretch_persistence_length = stretch_persistence_length
        if (present(max_radius)) then
                                                self%max_radius                 = max_radius
        else
            if (.not. present(mould)) then
                write(*,'(a)') '**warning(filament_init): maximum filament radius not given'
                !call terminate('Debug filament_init, no max radius given')
            endif
        endif
        if (present(refine_normal_modes))       self%refine_normal_modes        = refine_normal_modes
        if (present(use_hermite_splines))       self%use_hermite_splines        = use_hermite_splines
        if (present(fragmented_phi))            self%fragmented_phi             = fragmented_phi
        if (present(fragmented_subunit_num))    self%fragmented_subunit_num     = fragmented_subunit_num

        if (debug) then
            write(*,'(a,9l1)') '**debug(filament_init): rigid parameters: ', self%rigid_params
            write(*,'(a,2(i0,1x))') '**debug(filament_init): filament, film id = ', self%id, self%film%GetID()
        endif

        ! Are we adding extra waypoints?
        aadd_extra_fal_wps = .false.
        if (present(add_extra_fal_wps)) aadd_extra_fal_wps = add_extra_fal_wps

        ! Are we refreshing the first and last waypoint (i.e. computing them from scratch)
        rrefresh_fal_wps = .false.
        if (present(refresh_fal_wps)) then
            rrefresh_fal_wps = refresh_fal_wps
        else if (refresh_fal_wps_if_not_in_score_or_threed .and. .not. aadd_extra_fal_wps) then
            if (.not. self%fal_wps_in_score .and. .not. self%fal_wps_in_threed) then
                rrefresh_fal_wps = .true.
            endif
        endif
        if (rrefresh_fal_wps) then
            write(*,'(a)') '**info(filament_init): the first and last waypoints will be re-calculated using extrapolation'
        endif

        ! How many points do we need to define a filament?
        if (aadd_extra_fal_wps) then
            if (.not. self%fal_wps_in_score .and. self%fal_wps_in_threed) then
                min_num_points = 3
            else
                min_num_points = 3
            endif
        else
            if (.not. (self%fal_wps_in_score .and. self%fal_wps_in_threed)) then
                min_num_points = 4 ! If the first and last waypoints are not crossovers, later down the line we'll want to reconstruct from the 2nd waypoint to the 2nd-to-last waypoint
            else
                min_num_points = 3
            endif
        endif

        ! Override, because it may be that we'll be adding fal wps after the filament has been initialised
        min_num_points = 3
        if (self%wps_as_sps) min_num_points = 1

        !
        if (present(points_in)) then

            ! Copy input points
            allocate(points(size(points_in)))
            points = points_in
            ! Remove duplicates
            call points_remove_duplicates(points)
            ! Check we have the required number of points to define the filament. If we do, remember.
            if (size(points) .eq. 1) then
                write(*,'(a,i0,a,i0,a,i0,a)') '**error(filament_init): insufficient number of unique input points (',&
                                                size(points), &
                                                ') for filament ', self%id, '. ',min_num_points,'+ points needed'
                call terminate('fatal error in filament_init')
            else if (size(points) .lt. min_num_points) then
                write(*,'(a,i0,a,i0,a)') '**warning(filament_init): insufficient number of unique input points (', &
                                                size(points), &
                                                ') for filament ', self%id, '. adding points by linear interpolation.'
                call points_add_by_linear_interpolation(points)
            endif
            ! Remember how many waypoints we have in the filament
            self%nwp = size(points)

        else if (present(waypoints_in)) then

            if (debug) then
                print *, '**debug(): h info for waypoints_in'
                do i=1,self%nwp
                    call helical_symmetry_print_info(waypoints_in(i)%helix)
                enddo
            endif

            ! Copy input waypoints
            if (rrefresh_fal_wps) then
                ! Remove the FAL waypoints and make a note to add new ones later
                allocate(waypoints(size(waypoints_in)-2))
                waypoints = waypoints_in(2:size(waypoints_in)-1)
                aadd_extra_fal_wps = .true.
            else
                allocate(waypoints(size(waypoints_in)))
                waypoints = waypoints_in
            endif
            ! Remove duplicates
            call waypoints_remove_duplicates(waypoints)
            ! Check we have the required number of points to define the filament. If we do, remember.
            !if (size(waypoints) .eq. 1) then
            !    write(*,'(a,i0,a,i0,a,i0,a)') '**error(filament_init): insufficient number of unique input points (',&
            !                                    size(waypoints), &
            !                                    ') for filament ', self%id, '. ',min_num_points,'+ points needed'
            !    call terminate('fatal error in filament_init')
            !else
            if (size(waypoints) .lt. min_num_points) then
                write(*,'(a,i0,a,i0,a)') '**warning(filament_init): insufficient number of unique input waypoints (', &
                                                size(waypoints),&
                                                ') for filament ', self%id, '. adding points by linear interpolation.'
                call waypoints_add_by_linear_interpolation(waypoints)
            endif
            ! Remember how many waypoints we have in the filament
            self%nwp = size(waypoints)

            if (debug) then
                print *, '**debug(): h info for waypoints'
                do i=1,self%nwp
                    call helical_symmetry_print_info(waypoints(i)%helix)
                enddo
            endif
        else
            write(*,'(a)') '**ERROR(filament_init): need to supply either points or waypoints'
            call terminate('Fatal error in FILAMENT_INIT')
        endif ! end of test for presence of points_in or waypoints_in

        if (debug) then
            write(*,'(a)') '**debug(filament_init): filament info after processing (way)points array'
            call self%PrintInfo(verbose=.true.)
        endif

        ! We need this for some write format statements
        write(nwp_c,'(i6)') self%nwp

        ! If we're not dealing with crossovers, but the user specified fal_wps_not_xovers, something is wrong.
        if (.not. self%wps_are_crossovers) then
            if (self%fal_wps_are_crossovers) then
                write(*,'(a)') '**WARNING(filament_init): fal waypoints can be crossovers only if the others are too. Ignoring.'
                self%fal_wps_are_crossovers = .false.
            endif
        endif


        !
        ! Microtubule-specific stuff
        !
        if (self%microtubule) then
            if (present(mt_pf_number)) then
                self%mt_pf_number = mt_pf_number
            endif
            if (.not. present(mt_pf_number) .and. self%mt_pf_number == 0) then
                write(*,'(a)') '**ERROR(filament_init): microtubule number of protofilaments not supplied'
                call terminate('Fatal error in FILAMENT_INIT')
            endif
            if (present(mt_rise)) then
                self%mt_rise = mt_rise
            endif
            if (.not. present(mt_rise) .and. self%mt_rise == 0) then
                write(*,'(a)') '**ERROR(filament_init): microtubule rise not supplied'
                call terminate('Fatal error in FILAMENT_INIT')
            endif
        endif


        !
        ! If we were given helical parameters, copy them over
        !
        if (present(helix)) then
            self%helix = helix
!            self%wp%helix = helix   ! Is it correct to do this here? May be better off computing this later...
!                                    ! E.g. in FILAMENT_UPDATE_MODEL?
            if (self%rigid_params(7)) then
                self%rigid_wp%helix = helix ! Is it correct to do this here?
                                            ! It might be more genrenal to do this in filament_update_model by taking
                                            ! the average of the waypoint helical parameters, or some sort of average
                                            ! over the whole filament?
            endif
            ! The outer radius will not have been setup by the ali file parser
            if (allocated(self%wp)) self%wp%helix%outer_radius = helix%outer_radius
            waypoints%helix%outer_radius = helix%outer_radius
        else
            if (.not. present(mould)) then
                write(*,'(a)') '**info(filament_init): helical parameters not supplied'
                call terminate('Debug stop - no helical parameters?')
            endif
        endif

        if (present(helix_min_for_3drec)) then
            self%helix_min_for_3drec = helix_min_for_3drec
        else if (.not. present(mould)) then
            call helical_symmetry_init( self%helix_min_for_3drec,0.0,-179.99, &
                                        self%helix%outer_radius, &
                                        self%helix%axial_symmetry,self%helix%perpendicular_symmetry, &
                                        self%helix%num_starts,self%helix%stack, &
                                        do_not_setup_unit_cell=.true.)
        endif

        if (present(helix_max_for_3drec)) then
            self%helix_max_for_3drec = helix_max_for_3drec
        else if (.not. present(mould)) then
            call helical_symmetry_init( self%helix_max_for_3drec,huge(1.0)*0.001,179.99, &
                                        self%helix%outer_radius, &
                                        self%helix%axial_symmetry,self%helix%perpendicular_symmetry, &
                                        self%helix%num_starts,self%helix%stack, &
                                        do_not_setup_unit_cell=.true.)
        endif

        if (debug) then
            write(*,'(a)') '**debug(filament_init): min, max helices'
            call helical_symmetry_print_info(self%helix_min_for_3drec)
            call helical_symmetry_print_info(self%helix_max_for_3drec)
        endif

        ! If crossovers, we can't assume constant twist per subunit!
        if (self%wps_are_crossovers) then
            self%constant_twist_per_length = .false.
        endif

        !
        ! What is the sign of the twist?
        !
        if (mod(self%helix%twist,180.D0) .gt. 0.D0) then
            twist_sign = 1
        else
            twist_sign = -1
        endif
        if (debug) write(*,'(a,f8.2,1x,i2)') '**debug(filament_init): twist, twist_sign = ', self%helix%twist, twist_sign


        !
        ! Check whether we're given points (just coordinates) or waypoints (coordinates & angles)
        !
        if (present(waypoints_in)) then
            !
            ! Waypoints were supplied, which contain angle as well as x,y and z coordinates etc.
            !

            if (debug) then
                write(*,'(a)') '**debug(filament_init): waypoints before remove_too_close:'
                call waypoints_print_info(waypoints)
            endif

            ! Remove waypoints which are too close together (in terms of subunit num)
            allocate(wp_bak(size(waypoints)),source=waypoints)
            wp_bak = waypoints
            call waypoints_remove_too_close(wp_bak)
            if (debug) then
                write(*,'(a)') '**debug(filament_init): waypoints after remove_too_close:'
                call waypoints_print_info(wp_bak)
            endif

            ! Allocate memory for the filament
            if (aadd_extra_fal_wps) then
                ! Allocate memory
                call self%Allocate(size(wp_bak)+2)
                ! Copy coordinates over
                forall(i=1:size(wp_bak))self%wp(i+1) = wp_bak(i)
            else
                ! Allocate memory
                call self%Allocate(size(wp_bak))
                ! Copy coordinates over
                forall(i=1:self%nwp) self%wp(i) = wp_bak(i)
            endif


            ! Check that the supplied waypoints are indeed crossovers, if that's what the arguments suggest
            if (self%wps_are_crossovers) then
                if (self%fal_wps_are_crossovers) then
                    first_xover_wp = 1
                    last_xover_wp = self%nwp
                else
                    first_xover_wp = 2
                    last_xover_wp = self%nwp - 1
                endif
                do i=first_xover_wp,last_xover_wp
                    if (.not. waypoint_is_crossover(waypoints(i))) then
                        write(*,'(a,i0,a,F0.5,a)')    '**warning(filament_init): you supplied waypoint ', i, &
                                                    ' saying it was a crossover, but it has a phi angle of ', &
                                                     waypoints(i)%euler%phi, '. Setting wps_are_crossovers to false.'
                        self%wps_are_crossovers = .false.
                    endif
                enddo
            endif

            ! Carry all information from waypoints to filament
            if (aadd_extra_fal_wps) then
                do i=1,size(waypoints)
                    self%wp(i+1) = wp_bak(i)
                enddo
            else
                self%wp = wp_bak
            endif

            ! Make sure helices are stored in pixels rather than Angstroms
            do i=1,self%nwp
                if (debug) call helical_symmetry_print_info(self%wp(i)%helix)
                if (.not. self%wp(i)%helix%pixel_units) then
                    call helical_symmetry_unit_convert(self%wp(i)%helix,self%film%GetPixelSize())
                endif
                if (debug) call helical_symmetry_print_info(self%wp(i)%helix)
            enddo
            if (.not. self%helix%pixel_units) call helical_symmetry_unit_convert(self%helix,self%film%GetPixelSize())

            ! Overwrite wp%helix objects ?????
            if (.not. self%wps_as_sps .or. &
                     (self%wps_as_sps .and. all(self%wp%helix%rise  .eq. 0.0) &
                                      .and. all(self%wp%helix%twist .eq. 0.0)) ) then
                self%wp(:)%helix = self%helix  ! This is definitely not correct when wps_as_sps and h refinement happened.
                                               ! It may even not be correct in other cases...
            endif

            ! Make sure helix objects in each waypoints were probably not properly initialised
            do i=1,self%nwp
                call helical_symmetry_init( self%wp(i)%helix,real(self%wp(i)%helix%rise),real(self%wp(i)%helix%twist), &
                                            self%wp(i)%helix%outer_radius, &
                                            self%helix%axial_symmetry,self%helix%perpendicular_symmetry, &
                                            num_starts=self%helix%num_starts, &
                                            stack=self%helix%stack )
            enddo


            !
            ! Do we know any of the z coordinates? (we expect that we know z coordinates,
            ! but if we don't, we'll guess them from the film information)
            !
            if (any(self%wp%coo(3) .ne. 0.0)) then
                z_unknown = .false.
            else
                write(*,'(3a)') '**info(filament_init): supplied waypoints had z coordinates = 0.0. ', &
                                    'z coordinates will be guessed for each waypoint using its x,y ', &
                                    'coordinates and the ctftilt/ctffind information'
                z_unknown = .true.
            endif
            !
            ! If necessary, work out z coordinates from film properties (the film may have ctftilt information, for example)
            !
            if (z_unknown) then
                if (debug) write(*,'(a)') '**debug(filament_init): assigning z coordinates to waypoints:'
                if (present(film_in)) then
                    if (any(self%film%GetLogicalDimensions() .ne. 0)) then
                        if (debug) then
                            write(*,'(a)') '**debug(filament_init): about to compute z coordinates using this film info:'
                            call self%film%PrintInfo()
                        endif
                        do i=1,self%nwp
                            self%wp(i)%coo(3) = self%film%ReturnZFromXY(self%wp(i)%coo(1),self%wp(i)%coo(2))
                        enddo
                        if (debug) write(*,'(a,'//nwp_c//'(F15.3,1x))') 'Z values: ', self%wp(:)%coo(3)
                        ! Update the rigid waypoint
                        call filament_update_rigid_wp(self,update_rigid_params=.true.)
                    else
                        write(*,'(2a)') '**ERROR(filament_init): cannot calculate z coordinates ', &
                                        'of points with unknown film dimensions'
                        call terminate('Fatal error in FILAMENT_INIT')
                    endif
                else
                    write(*,'(a)') '**ERROR(filament_init): cannot calculate z coordinates of points without a film object'
                    call terminate('Fatal error in FILAMENT_INIT')
                endif
            endif

            ! Update the rigid waypoint
            call filament_update_rigid_wp(self,update_rigid_params=.true.)


            if (all(self%t .eq. 0.0d0) .and. .not. all(self%wp%path_length .eq. 0.0d0)) then
                self%t = self%wp%path_length
            endif
            if (.not. ccalled_by_filament_fix_polarity_inconsistencies) then
            if (any(self%t .ne. 0.0d0) .and. .not. aadd_extra_fal_wps ) then
                ! When wps_as_sps, polarity flipping can be done elsewhere (under the control of sp_to_ff)
                call filament_fix_polarity_inconsistencies(self,allow_polarity_flip=.not. self%wps_as_sps,no_init=.true.)
            endif
            endif

            if (debug) then
                write(*,'(a)') '**debug(filament_init): before calling update model: '
                call self%PrintInfo(verbose=.false.)
            endif

            !
            ! Trigger a spline fit (known_d) (if we're adding extra first and last wps, this will also calculate their coordinates)
            !
            known_psi_theta = all(self%wp%euler%psi .eq. 0.0) .and. all(self%wp%euler%theta .eq. 0.0)
            known_psi_theta = .not. known_psi_theta
            call filament_update_model(self,guess_fal_wp_params=aadd_extra_fal_wps,wps_as_sps=self%wps_as_sps)

            if (debug) then
                write(*,'(a)') '**debug(filament_init): after calling update model: '
                call self%PrintInfo()
            endif

        else if (present(points_in)) then
            !
            ! In the default case, only coordinates (no angles) were given
            !
            if (debug) write(*,'(a)') '**debug(filament_init): coordinates (not waypoints) were supplied'


            !
            ! If the user wants extra first/last wps, they need to give us helical paramters. Otherwise,
            ! we can't guess what the Phi angles are for the waypoints we're adding.
            !
            if (aadd_extra_fal_wps .or. .not. self%wps_are_crossovers) then
                if (.not. present(helix)) then
                    write(*,'(2a)') '**ERROR(filament_init): cannot guess first/last waypoints ',&
                                    'parameters without helical parameters'
                    call terminate('Fatal error in FILAMENT_INIT')
                endif
            endif


            !
            ! Allocate memory for the filament object
            !
            if (aadd_extra_fal_wps) then

                ! Allocate memory
                call self%Allocate(size(points)+2)
                ! Copy coordinates over
                do  i=1,size(points)
                    self%wp(i+1) = points(i)
                enddo
            else
                ! Allocate memory
                call self%Allocate(size(points))
                ! Copy coordinates over
                do  i=1,size(points)
                    self%wp(i) = points(i)
                enddo
            endif


            ! Do we know any of the z coordinates? (since users pick waypoints on 2d micrographs we don't expect so, but let's check)
            if (any(self%wp(:)%coo(3) .ne. 0.0)) then
                z_unknown = .false.
                write(*,'(a)') '**WARNING(filament_init): unexpectedly, supplied points (which were not waypoints) '//&
                                'had non-zero z coordinates. Programming error?'
                write(*,'(a,'//nwp_c//'(F0.3,1x))') 'Z values: ', self%wp(:)%coo(3)
            else
                z_unknown = .true.
            endif

            ! If necessary, work out z coordinates from film properties (the film may have ctftilt information, for example)
            if (z_unknown) then
                !print *, '**debug(filament_init): assigning z coordinates to waypoints'
                if (present(film_in)) then
                    do i=1,self%nwp
                        self%wp(i)%coo(3) = self%film%ReturnZFromXY(self%wp(i)%coo(1),self%wp(i)%coo(2))
                    enddo
                else
                    write(*,'(a)') '**ERROR(filament_init): cannot calculate z coordinates of points without a film object'
                    call terminate('Fatal error in FILAMENT_INIT')
                endif
            endif

            !
            ! Work out Euler angle phi (twist) for the waypoints
            !
            if (aadd_extra_fal_wps) then
                write(*,'(2a)') '**warning(filament_init): option to add extra first-and-last waypoints',    &
                                ' is experimental and seldom tested'
                ! Phi (twist) angle: if crossovers, give first way-point 0 then add/subtract 180 to each subsequent way point
                if (self%wps_are_crossovers) then
                    self%wp(2)%euler%phi = 0.0
                    do i=3,self%nwp-1
                        self%wp(i)%euler%phi = self%wp(i-1)%euler%phi + 180.0 * twist_sign
                    enddo
                else
                    self%wp(1:self%nwp)%euler%phi = 0.0
                    if (.not. present(helix)) then
                        write(*,'(a)') '**WARNING(filament_init): helical parameters not provided, not crossovers. ',   &
                                        'Will not be able to assign Phi angles.'
                        call terminate('Unable to assign phi angles')
                    else
                        estimate_phi = .true.
                    endif
                endif
            else ! We're not adding extra points
                if (self%wps_are_crossovers) then
                    if (twist_sign .eq. -1) then
                        print '(a)', '**info(filament_init): assuming right-handed helix with crossovers, '//&
                                        'so that first xover has phi(1)=0.0, and each subsequent xover has phi(i+1)'//&
                                        ' = phi(i)-180. If this is not the case, results will be wrong.'
                    else
                        print '(a)', '**info(filament_init): assuming left-handed helix with crossovers, '//&
                                        'so that first xover has phi(1)=0.0, and each subsequent xover has phi(i+1)'//&
                                        ' = phi(i)+180. If this is not the case, results will be wrong.'
                    endif
                    if (self%fal_wps_are_crossovers) then
                        write(*,'(a)') '**info(filament_init): all waypoints are crossovers'
                        ! Phi (twist) angle: if crossovers, give first way-point 0 then add (or subtract, if left-handed) 180 to each subsequent way point
                        self%wp(1)%euler%phi = 0.0
                        do i=2,self%nwp
                            self%wp(i)%euler%phi = self%wp(i-1)%euler%phi + 180.0 * twist_sign
                        enddo
                    else
                        write(*,'(a)') '**info(filament_init): first and last waypoints are not crossovers. All other are.'
                        self%wp(2)%euler%phi = 0.0
                        do i=3,self%nwp-1
                            self%wp(i)%euler%phi = self%wp(i-1)%euler%phi + 180.0 * twist_sign
                        enddo
                        if (debug) write(*,'(2a)') '**debug(filament_init): crossover waypoints have phi already, but the',  &
                                                 ' fal wps phi values will have to be determined later by filament_update_model'
                        fal_wps_phi_tbd = .true.
                    endif
                else
                    self%wp(1:self%nwp)%euler%phi = 0.0
                    if (.not. present(helix)) then
                        write(*,'(2a)') '**WARNING(filament_init): helical parameters not provided, not crossovers. ',   &
                                        'Will not be able to assign Phi angles.'
                        call terminate('Unable to assign phi angles')
                    else
                        estimate_phi = .true.
                    endif
                endif
            endif

            ! Update the filament model
            call filament_update_model( self,guess_fal_wp_params=aadd_extra_fal_wps, &
                                        wps_as_sps=self%wps_as_sps,guess_fal_wp_euler=fal_wps_phi_tbd)

            ! In case we haven't been able to work out phi yet, do so now (note that if Phi is one of the rigid parameters, filamnet_update_model will have taken care of it
            if (estimate_phi .and. .not. self%rigid_params(4)) then
                ! Debug
                if (debug) then
                    write (*,'(a)') '**debug(filament_init): filament info before guessing WPS_PHI'
                    call self%PrintInfo()
                endif
                ! Now that we have the splines, we can make a guess at the phi angles
                call filament_guess_wps_phi(self)
            endif

            !
            ! Since we now have a first estimate for psi & theta, we'll fix it and we can refine it later
            !
            call self%x_spline%SetWhetherFirstDerivativesAreKnown(.true.)
            call self%y_spline%SetWhetherFirstDerivativesAreKnown(.true.)
            call self%z_spline%SetWhetherFirstDerivativesAreKnown(.true.)

        else
            call terminate('Fatal error in filament_init: need either points or waypoints')
        endif  ! End of selection as to whether user supplied waypoints or regular points


        !
        ! Movie processing
        !
        if (present(recompute_subframe_shifts_from_unblur_results)) then
            if(recompute_subframe_shifts_from_unblur_results) then
                if (present(frame_shifts_from_unblur) .and. present(subframe_alignment_box_size)) then
                    call filament_SetSubframeShifts(self,frame_shifts_from_unblur)
                    write(*,'(a)') '**debug(filament_init): subframe shifts before refinement'
                    call filament_print_subframe_shifts(self)
                    call filament_RefineSubframeShifts(self,subframe_alignment_box_size, &
                                                       apply_dose_filter=apply_dose_filter)
                    write(*,'(a)') '**debug(filament_init): subframe shifts after  refinement'
                    call filament_print_subframe_shifts(self)
                else
                    call terminate('Fatal error in filament_init - no results from unblur were supplied')
                endif
            endif
        endif

        if (present(frame_shifts_from_unblur) .and.  .not. filament_SubframeShiftsAreSet(self)) then
            call filament_SetSubframeShifts(self,frame_shifts_from_unblur)
        endif


        !
        if (aaverage_z_to_ctftilt_plane) then
            call filament_shift_average_z_to_ctftilt_plane(self)
        endif


        ! Now that we have proper splines and path length information, let's refine our helical parameters per wp,
        ! unless the helical parameters and/or phi are to be kept rigid
        if (present(helix) .or. present(mould)) then
            if (.not. (self%rigid_params(4) .or. self%rigid_params(7)) .and. .not. self%wps_as_sps) then
                ! Work out the rise per subunit for each segment.
                call filament_update_wp_helix(self)
                ! Work out the average twist per subunit. This is useful later on when a reference with appropriate twist needs to be generated.
                call filament_update_average_helix_twist(self)
            endif
        endif

        ! Let's do a final check for waypoints that may be too close to each other ! May 2014: commented out because of recursion
        !if (.NOT. ccalled_by_filament_remove_waypoints_too_close) then
        !    call filament_remove_waypoints_too_close(self)
        !endif


        if (debug) then
            write(*,'(a,'//nwp_c//'L1)') '**debug(filament_init): final self%wp(:)%helix%pixel_units = ', &
                                        self%wp(:)%helix%pixel_units
        endif

        ! Update the rigid waypoint
        if (.not. self%wps_as_sps) call filament_update_rigid_wp(self)

        ! Check the subunit_num values
        if (.not. self%fragmented_subunit_num) then
            call filament_check_subunit_num(self)
        endif

        ! Fit the splines & compute residuals
        call filament_fit_splines_linear(self)
        call filament_update_wp_res(self)

        ! Compute the normal modes
        call filament_update_nms(self)
        call filament_wps_to_nm_amplitudes(self)

        ! Check the total curvature - it should have a reasonable value
        total_curvature = filament_eval_total_curvature(self,squared_curvature=.true.)
        if (debug) write(*,'(a,G15.4)') '**debug(filament_init): total curvature = ', total_curvature
        if (total_curvature .gt. 5.0) then
            write(*,'(a,G15.4)') '**WARNING(filament_init): filament appears to have extreme curvature: ', total_curvature
            write(*,'(a)') '**WARNING(filament_init): maybe there is something funny about the input coordinates?'
            !call filament_print_info(self)
            !call terminate('Fatal error in FILAMENT_INIT')
        endif


        ! Iteratively relax the helix and update the average helical parameters for the filament
        do i=1,2
            if (debug) then
                write(*,'(a,i0,a)') '**debug(filament_init): relaxing helix, iteration ', i, ' of 2, filament info: '
                call self%PrintInfo()
            endif
            if (rrelax_helix) then
                !write(*,'(a)') '**info(filament_init): relaxing helix (experimental)'
                !call filament_relax_helix(self)
                if (.not. (self%fal_wps_in_threed .or. self%fal_wps_in_score)) then
                    write(*,'(2a)') '**info(filament_init): relaxing rise & twist for first and last segments,', &
                                    ' since they are not used in refinement or 3d reconstruction'
                    call filament_relax_rise_per_subunit(self,fal_only=.true.)
                    call filament_relax_twist_per_subunit(self,fal_only=.true.,allow_any_change=.true.)
                endif
                if (self%fragmented_subunit_num) then
                    if (this_program%IsVerbose() .or. debug) then
                        write(*,'(a)') '**info(filament_init): relaxing rise per subunit (experimental)'
                    endif
                    call filament_relax_rise_per_subunit(self)
                endif
                if (self%fragmented_phi) then
                    if (this_program%IsVerbose() .or. debug) then
                        write(*,'(a)') '**info(filament_init): relaxing twist per subunit (experimental)'
                    endif
                    call filament_relax_twist_per_subunit(self)
                endif
            endif

            if (.not. self%rigid_params(7) .and. .not. self%wps_as_sps .and. .not. self%fragmented_subunit_num) then
                if (debug) write(*,'(a)') '**debug(filament_init): updating helix rise'
                call filament_update_average_helix_rise(self)
            endif

            if (.not. self%rigid_params(4) .and. .not. self%wps_as_sps .and. .not. self%fragmented_phi) then
                if (debug) write(*,'(a)') '**debug(filament_init): updating average helix twist'
                call filament_update_average_helix_twist(self)
            endif
            call filament_update_wp_helix(self)
        enddo

        ! todo here: do the same (update filament average) for the rise per subunit?

        ! Check the subunit_num values, they should increase wp-to-wp
        call filament_check_subunit_num_increases(self)

        ! Check that the relevant splines are monotonic
        if (.not. self%wps_as_sps) then
            if (.not. self%fragmented_phi) then
                if (.not. self%twist_spline%IsMonotonic()) then
                    write(*,'(a)') '**ERROR(filament_init): phi spline is not monotonic'
                    call self%PrintInfo()
                    call terminate('Fatal error in FILAMENT_INIT')
                endif
            endif
            if (.not. self%fragmented_subunit_num) then
                if (.not. self%s_spline%IsMonotonic()) then
                    write(*,'(a)') '**ERROR(filament_init): subunit_num spline is not monotonic'
                    call self%PrintInfo()
                    call terminate('Fatal error in FILAMENT_INIT')
                endif
            endif
        endif

    end subroutine filament_init

end module filaments
