!>  \brief  Filament class and basic house-keeping routines
module filaments_base

    use Globals

    use coos, only : waypoint
    use splines, only : interpol_spline
    use Micrographs, only : Micrograph
    use helices, only : helical_symmetry
    use GaussLegendreQuadratures
    use iso_c_binding, only : c_ptr

    implicit none

    !>  \brief  Filament class
    type Filament
        integer                     ::  nwp                                     !<  Number of waypoints defined for the filament
        type(waypoint), allocatable ::  wp(:)                                   !<  Array of waypoints
        type(Micrograph)            ::  film                                    !<  Film on which the filament is imaged (this is where we'll get all the ctf info from)

        ! Space curve model of the filament
        real(kind=8),   allocatable ::  t(:)                                    !<  For each waypoint, the corresponding value of the t parameter. The t parameter should be a good approximation of the path length.
        type(interpol_spline)       ::  twist_spline                            !<  Spline modelling the twist angle along the filament, as a function of t
        type(interpol_spline)       ::  x_spline                                !<  Spline modelling the x position of the filament, as a function of t
        type(interpol_spline)       ::  y_spline                                !<  Spline modelling the y position of the filament, as a function of t
        type(interpol_spline)       ::  z_spline                                !<  Spline modelling the z position of the filament, as a function of t
        type(interpol_spline)       ::  s_spline                                !<  Spline modelling the helical subunit number along the filament, as a function of t
        logical                     ::  refine_params(9)            =   .false. !<  Which of the 9 degrees of freedom to refine per way point (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi,7.SubunitNumber,8.HRise,9.HTwist)
        logical                     ::  use_hermite_splines         =   .true.  !<  use SPLINESs where the first derivative is used as a constraint (aka known_d in parts of Frealix).
                                                                                !!  In other words, psi & theta are not used to solve the splines, they are computed from them.
        type(GaussLegendreQuadrature):: gl_quadrature                           !<  Gauss-Legendre quadrature object. Will be used to compute arc lengths along the space curve

        ! Linear fit of the splines and their residuals
        real(kind=8),   allocatable ::  t_linear(:)                             !<  The t values (ordinates) that were used when computing the linear fit.
        real(kind=8)                ::  x_linear(2)                 =   0.0D0   !<  Coefficients of the linear fit of the spline. The first number is multiplicative (slope), the second additive (offset).
        real(kind=8)                ::  y_linear(2)                 =   0.0D0
        real(kind=8)                ::  z_linear(2)                 =   0.0D0
        real(kind=8)                ::  s_linear(2)                 =   0.0D0
        real(kind=8)                ::  phi_linear(2)               =   0.0D0
        type(waypoint), allocatable ::  wp_res(:)                               !<  These waypoints  store the residuals between the linear fits to the splines and the actual wp values

        ! Rigid-body modelling of the filament
        logical                     ::  rigid_params(9)             =   .false. !<  Whether to keep the following parameters fixed throughout the filament: 1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta, 6.Psi, 7-H.Rise.,8.HRise,9.HTwist
                                                                                !!  If all are set to .TRUE., the filament is a rigid rod, treated as a single particle.
        type(waypoint)              ::  rigid_wp                                !<  a special waypoint object storing values for the parameters which are to be kept rigid.

        ! Normal modes of the filament
        logical                     ::  refine_normal_modes(9)      =   .false. !<  Whether normal mode amplitudes, rather than the model itself, should be refined
        real(kind=8),   allocatable ::  nm_eigenvectors(:,:)                    !<  Eigenvectors. The elements of each vector are in the first dimension and the vectors are indexed in the second dimension.
        real(kind=8),   allocatable ::  nm_eigenvalues(:)                       !<  Eigenvalues
        type(waypoint), allocatable ::  wp_nm(:)                                !<  Array of waypoints storing normal mode (nm) amplitudes

        ! Subunit_num-related flags
        logical                     ::  constant_rise_per_subunit   =   .false.  !<  Whether to assume that the rise-per-subunit is constant.
        logical                     ::  constant_twist_per_length   =   .false.  !<  Whether to assume that the twist per unit length is constant.

        ! Book-keeping
        logical,        private     ::  in_memory                   =   .false. !<  Whether memory is already allocated
        integer                     ::  id                                      !<  id number, from the parameter file

        ! Description of the helix
        type(helical_symmetry)      ::  helix                                   !<  a descriptor for the helical symmetry the filament obeys on average.
        real                        ::  max_radius                  =   0.0     !<  Maximum radius of the filament (pixels)
        logical                     ::  fragmented_subunit_num      =   .false. !<  Introduce breaks/discontinuities in the subunit_num function between waypoints.
        logical                     ::  fragmented_phi              =   .false. !<  Introduce breaks/discontinuities in the phi function between waypoints.
        logical                     ::  wps_as_sps                  =   .false. !<  Waypoints are treated as independent single particles

        ! Min & max helical parameters for inclusion into the 3D reconstruction
        type(helical_symmetry)      ::  helix_min_for_3drec
        type(helical_symmetry)      ::  helix_max_for_3drec

        ! Amyloid-beta-specific components
        logical                     ::  amyloid_beta                =   .false. !<  Special flag for amyloid beta filaments
        logical                     ::  wps_are_crossovers          =   .false. !<  Waypoints are crossovers
        logical                     ::  fal_wps_are_crossovers      =   .false. !<  If waypoints are crossoves, this flag can be used to indicate that the first and last waypoints are in fact not crossovers.

        ! Flags concerning the treatment of the first and last waypoints (fal = first and last; wps = waypoints)
        logical                     ::  fal_wps_in_score            =   .true.  !<  If false, the parts of the filament between the first and the second as well as between the penultimate and last waypoints will not be included in the scoring function
        logical                     ::  fal_wps_in_threed           =   .true.  !<  If false, the parts of the filament between the first and the second as well as between the penultimate and last waypoints will not be included in the 3d reconstruction

        ! Microtubule-specific components
        logical                     ::  microtubule                 =   .false. !<  Special flag for microtubule filaments
        integer                     ::  mt_pf_number                =   0       !<  For microtubules only: number of protofilaments (e.g. 13)
        integer                     ::  mt_rise                     =   0       !<  For microtubules only: how much further along the filament (distance measured in tubulin subunits) the end
                                                                                !!  of a MT_PF_NUMBER-element "washer" is compared to its beginning. The 13-pf microtobules have a value of 3.

        ! Mechanical properties
        real                        ::  persistence_length          =   0.0     !<  The persistence length of the filament, in pixels
        real                        ::  torsion_persistence_length  =   0.0     !<  The torsional persistence length of the filament, in pixels
        real                        ::  stretch_persistence_length  =   0.0     !<  The stretching/compression persistence length of the filament, in pixels.

        ! Scoring
        real                        ::  score                       =   0.0     !<  Score for the filament to see ways to calculate the score, check the filament_model_comparison module.
        type(waypoint), allocatable ::  wp_score(:)                             !<  The array of waypoints used by the scoring function

        !
        type(c_ptr)                 ::  cptr                                    !<  a C-pointer pointing to itself. This is needed when passing the object through the FGSL.
        !
        contains
            ! Useful for debug
            procedure               ::  PrintInfo
            ! Space curve parametrisation
            procedure               ::  SetTFromPathLength
            ! Setup, initialisation
            procedure               ::  PopulateWithWaypoints
            procedure               ::  ResetProperties
            procedure               ::  CopyPropertiesFrom
            ! Check angle properties
            procedure               ::  PsiAndThetaAreNonZero
            procedure               ::  ThetaAreNonZero
            ! Helper functions for refinements
            procedure               ::  SuggestedBoxWidth
            ! Memory management
            final                   ::  Finalise!, filament_array_deallocate
            procedure               ::  Deallocate
            procedure               ::  Assign
            procedure               ::  Allocate
            procedure               ::  IsAllocated
            ! Operator overloading
            generic,    public      ::  assignment(=)       =>  Assign
    end type

    contains

    !>  \brief  Return reasonable value for box width, in pixels
    pure function SuggestedBoxWidth(self) result(width)
        use UsefulFunctions, only : IsOdd
        ! Arguments
        class(filament), intent(in)  ::  self
        ! Result
        integer ::  width
        ! Start work
        width = 3.*self%max_radius
        if (IsOdd(width)) width = width+1
    end function SuggestedBoxWidth

    !>  \brief  Split an array of filament objects into an array of smaller arrays of filament objects. This is useful when
    !!          distributing processing of filaments over nodes etc.
    !!
    !!          Each filament is assigned to a group. Group #1 should be processed first.
    subroutine SplitArrayOfFilaments(filaments_in,num_output_groups,group_membership,members_per_group)
        ! Arguments
        type(filament),                     intent(in)      ::  filaments_in(:)         !<  Array of filaments
        integer,                            intent(in)      ::  num_output_groups       !<  Number of groups of filaments to be defined
        integer,        allocatable,        intent(inout)   ::  group_membership(:)     !<  Output array assigning a group to each filament
        integer,        allocatable,        intent(inout)   ::  members_per_group(:)    !<  Number of filaments in each group
        ! Private variables
        integer ::  num_input_fils
        integer ::  ierr
        integer ::  fils_per_group, fils_in_current_group, current_group, groups_with_more_fils
        integer ::  i
        integer ::  nwp_array(size(filaments_in))
        logical ::  to_be_refined(size(filaments_in))
        integer ::  group_capacity(num_output_groups)
        integer ::  nwp_per_group(num_output_groups)
        integer ::  length_of_longest_fil, index_of_longest_fil
        integer ::  length_of_shortest_fil, index_of_shortest_fil
        ! Start work

        ! The number of filaments in the array
        num_input_fils = size(filaments_in)
        write(*,'(a,i0,a)') '**info(SplitArrayOfFilaments): there are ', num_input_fils, ' input filaments'

        !  Cannot have more output groups than input filaments
        if (num_output_groups .gt. num_input_fils) then
            write(*,'(a,i0,a,i0,a)') '**ERROR(SplitArrayOfFilaments): number of output groups (', num_output_groups, &
                                    ') greater than number of input filaments (', num_input_fils, ')'
            call this_program%TerminateWithFatalError('SplitArrayOfFilaments','Fatal error')
        endif

        ! The number of waypoints in each filament
        nwp_array(1:num_input_fils) = filaments_in(1:num_input_fils)%nwp

        ! Whether filaments are set to be refined
        forall(i=1:num_input_fils)
            to_be_refined(i) = any(filaments_in(i)%wp%refine)
        end forall

        ! Allocate memory to group membership array
        allocate(group_membership(num_input_fils),members_per_group(num_output_groups),stat=IERR)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(SplitArrayOfFilaments): memory allocation error ', ierr
            call this_program%TerminateWithFatalError('SplitArrayOfFilaments','Fatal error')
        endif

        ! Initialise group membership array
        group_membership = 0

        ! Work out a number of filaments to go into each group.
        fils_per_group = int(num_input_fils/num_output_groups)
        groups_with_more_fils = mod(num_input_fils,num_output_groups)

        ! The "group capacity" array keeps track for each group of how many members we're looking for
        group_capacity(:) = fils_per_group
        if (groups_with_more_fils .gt. 0) then
            !group_capacity(1:groups_with_more_fils) = fils_per_group + 1
            group_capacity(num_output_groups-groups_with_more_fils+1:num_output_groups) = fils_per_group + 1
            write(*,'(4(A,i0),a)') '**info(SplitArrayOfFilaments): splitting into ', groups_with_more_fils, &
                                    ' groups of ', fils_per_group + 1, ' and ', num_output_groups - groups_with_more_fils, &
                                    ' groups of ', fils_per_group, ' filaments'
        else
            write(*,'(2(A,i0),a)') '**info(SplitArrayOfFilaments): splitting into ', num_output_groups, ' groups of ', &
                                    fils_per_group, ' filaments'
        endif


        ! Loop until we have assigned membership to every filament
        current_group = 1
        fils_in_current_group = 0
        nwp_per_group = 0
        do while (any(group_membership .eq. 0))

            ! If the current group is full, move on to the next one
            if (fils_in_current_group .eq. group_capacity(current_group)) then
                current_group = current_group + 1
                fils_in_current_group = 0
            endif

            ! Find the longest filament still available, which has at least one waypoint set to be refined
            if (any(group_membership .eq. 0 .and. to_be_refined)) then
                length_of_longest_fil   =   maxval(nwp_array,       mask = (group_membership .eq. 0 .and. to_be_refined))
                index_of_longest_fil    =   maxloc(nwp_array, 1,    mask = (group_membership .eq. 0 .and. to_be_refined))
            else
                length_of_longest_fil   =   maxval(nwp_array,       mask =  group_membership .eq. 0)
                index_of_longest_fil    =   maxloc(nwp_array, 1,    mask =  group_membership .eq. 0)
            endif

            ! Assign it to the current group
            group_membership(index_of_longest_fil) = current_group
            fils_in_current_group = FILS_IN_CURRENT_GROUP+1
            nwp_per_group(current_group) = nwp_per_group(current_group) + filaments_in(index_of_longest_fil)%nwp

            ! If the current group is full, move on to the next one
            if (fils_in_current_group .eq. group_capacity(current_group)) then
                current_group = current_group + 1
                fils_in_current_group = 0
                cycle
            endif

            ! Fill the group with small filaments
            do while (fils_in_current_group .lt. group_capacity(current_group))
                ! Find the shortest filament still available. If possible use a filament that's not going to be refined
                if (any(group_membership .eq. 0 .and. .not. to_be_refined)) then
                    length_of_shortest_fil  =   minval(nwp_array,       mask = (group_membership .eq. 0 .and. .not. to_be_refined))
                    index_of_shortest_fil   =   minloc(nwp_array, 1,    mask = (group_membership .eq. 0 .and. .not. to_be_refined))
                else
                    length_of_shortest_fil  =   minval(nwp_array,       mask =  group_membership .eq. 0)
                    index_of_shortest_fil   =   minloc(nwp_array, 1,    mask =  group_membership .eq. 0)
                endif

                ! Assign it to the current group
                group_membership(index_of_shortest_fil) = current_group
                fils_in_current_group = fils_in_current_group + 1
                nwp_per_group(current_group) = nwp_per_group(current_group) + filaments_in(index_of_shortest_fil)%nwp
            enddo

        enddo

        !
        ! Diagnostics
        !
        write(*,'(3(A20,1x))') 'Group #', '# of filaments', 'total # of waypoints'
        do I=1,num_output_groups
            write(*,'(3(I20,1x))') i, group_capacity(i),nwp_per_group(i)
        enddo

        members_per_group = group_capacity

    end subroutine SplitArrayOfFilaments


    !>  \brief  Given an array of waypoints, populate a filament object
    subroutine PopulateWithWaypoints(self,wps,id,film_id,pixel_size,score,number_of_frames_per_micrograph)
        ! Arguments
        class(filament),                intent(inout)   ::  self
        type(waypoint), allocatable,    intent(in)      ::  wps(:)      !<  Array of waypoints
        integer,        optional,       intent(in)      ::  id          !<  Filament id
        integer,        optional,       intent(in)      ::  film_id     !<  Film id
        real,           optional,       intent(in)      ::  pixel_size  !<  Film pixel size, in Angstroms
        real,           optional,       intent(in)      ::  score       !<  Filament score
        integer,        optional,       intent(in)      ::  number_of_frames_per_micrograph
        ! Private variables
        integer ::  i
        ! Start work

        if (.not. allocated(wps)) then
            write(*,'(a)') '**ERROR(Filament::PopulateWithWaypoints): input array of waypoints not allocated'
            call this_program%TerminateWithFatalError('Filament::PopulateWithWaypoints','Fatal error')
        endif

        ! Copy the waypoints over
        if (allocated(self%wp)) deallocate(self%wp)
        allocate(self%wp(size(wps)))
        self%wp = wps
        self%nwp = size(wps)

        ! By convention, we only allow positive theta values
        do i=1,self%nwp
            if (self%wp(i)%euler%theta .lt. 0.0d0) self%wp(i)%euler%theta = mod(self%wp(i)%euler%theta,360.0d0) + 360.0d0
        enddo

        ! Other properties
        if (present(id))            self%id = id
        if (present(film_id))       call self%film%SetID(film_id)
        if (present(pixel_size))    call self%film%SetPixelSize(pixel_size)
        if (present(score))         self%score = score
        if (present(number_of_frames_per_micrograph)) call self%film%SetNumberOfFrames(number_of_frames_per_micrograph)

        self%in_memory = .true.

    end subroutine PopulateWithWaypoints

    !>  \brief  Subroutine which overloads the assignment operator (=)
    !!
    !!  This subroutine does not work for "in-place" arguments, e.g. a = a
    subroutine Assign(fil_lhs,fil_rhs)
        ! Arguments
        type(filament),    intent(in)      ::  fil_rhs     !<  Filament object on the right-hand-side of the assignment
        class(filament),   intent(inout)   ::  fil_lhs     !<  Filament object on the left-hand-side of the assignment
        ! Private variables
        logical,    parameter   ::  debug   =   .false.
        ! Start work
        ! Allocate memory
        if (fil_lhs%IsAllocated() .and. (fil_lhs%nwp .ne. fil_rhs%nwp)) call fil_lhs%Deallocate()
        if (.not. fil_lhs%IsAllocated()) call fil_lhs%Allocate(fil_rhs%nwp)
        ! Copy contents over
        call fil_lhs%CopyPropertiesFrom(fil_rhs)
        if (allocated(fil_rhs%wp)) then
            fil_lhs%wp                      =   fil_rhs%wp
        else
            write(*,'(2a)') '**ERROR(Filament::Assign): attempt to copy non-allocated array ',    &
                            'of waypoints from rhs to lhs of assignment'
            call this_program%TerminateWithFatalError('Filament::Assign','Fatal error')
        endif

        fil_lhs%film                        =   fil_rhs%film
        if (allocated(fil_rhs%t)) then
            fil_lhs%t                       =   fil_rhs%t
        else
            if (debug) write(*,'(2a)') '**debug(Filament::Assign): attempt to copy non-allocated array of ', &
                                        'T values from rhs to lhs of assignment'
        endif
        fil_lhs%twist_spline                =   fil_rhs%twist_spline
        fil_lhs%x_spline                    =   fil_rhs%x_spline
        fil_lhs%y_spline                    =   fil_rhs%y_spline
        fil_lhs%z_spline                    =   fil_rhs%z_spline
        fil_lhs%s_spline                    =   fil_rhs%s_spline
        if (allocated(fil_rhs%t_linear)) then
            fil_lhs%t_linear                =   fil_rhs%t_linear
        endif
        fil_lhs%x_linear                    =   fil_rhs%x_linear
        fil_lhs%y_linear                    =   fil_rhs%y_linear
        fil_lhs%z_linear                    =   fil_rhs%z_linear
        fil_lhs%s_linear                    =   fil_rhs%s_linear
        fil_lhs%phi_linear                  =   fil_rhs%phi_linear
        if (allocated(fil_rhs%wp_res)) then
            fil_lhs%wp_res                  =   fil_rhs%wp_res
        endif

        fil_lhs%in_memory                   =   fil_rhs%in_memory

        if (allocated(fil_rhs%wp_score)) then
            !if (.not. allocated(fil_lhs%wp_score)) allocate(fil_lhs%wp_score(size(fil_rhs%wp_score)))
            fil_lhs%wp_score                =   fil_rhs%wp_score
        endif

        ! Deal with normal modes

        if (allocated(fil_rhs%nm_eigenvectors)) then
            fil_lhs%nm_eigenvectors         =   fil_rhs%nm_eigenvectors
        endif
        if (allocated(fil_rhs%nm_eigenvalues)) then
            fil_lhs%nm_eigenvalues          =   fil_rhs%nm_eigenvalues
        endif
        if (allocated(fil_rhs%wp_nm)) then
            fil_lhs%wp_nm                   =   fil_rhs%wp_nm
        endif
    end subroutine Assign

    !>  \brief  Copy proerpties of the filament that do not relate to the waypoints themselves
    !!          or other allocatable components. In other words, this can be used between filaments
    !!          with different numbers of waypoints
    pure subroutine CopyPropertiesFrom(fil_lhs,fil_rhs)
        ! Arguments
        type(filament),     intent(in)      ::  fil_rhs     !<  Filament object on the right-hand-side of the assignment
        class(filament),    intent(inout)   ::  fil_lhs     !<  Filament object on the left-hand-side of the assignment
        ! Start work
        fil_lhs%nwp                         =   fil_rhs%nwp
        fil_lhs%refine_params               =   fil_rhs%refine_params
        fil_lhs%use_hermite_splines         =   fil_rhs%use_hermite_splines
        fil_lhs%gl_quadrature               =   fil_rhs%gl_quadrature
        fil_lhs%x_linear                    =   fil_rhs%x_linear
        fil_lhs%y_linear                    =   fil_rhs%y_linear
        fil_lhs%z_linear                    =   fil_rhs%z_linear
        fil_lhs%s_linear                    =   fil_rhs%s_linear
        fil_lhs%phi_linear                  =   fil_rhs%phi_linear
        fil_lhs%rigid_params                =   fil_rhs%rigid_params
        fil_lhs%rigid_wp                    =   fil_rhs%rigid_wp
        fil_lhs%refine_normal_modes         =   fil_rhs%refine_normal_modes
        fil_lhs%constant_rise_per_subunit   =   fil_rhs%constant_rise_per_subunit
        fil_lhs%constant_twist_per_length   =   fil_rhs%constant_twist_per_length
        fil_lhs%id                          =   fil_rhs%id
        fil_lhs%helix                       =   fil_rhs%helix
        fil_lhs%helix_max_for_3drec         =   fil_rhs%helix_max_for_3drec
        fil_lhs%helix_min_for_3drec         =   fil_rhs%helix_min_for_3drec
        fil_lhs%max_radius                  =   fil_rhs%max_radius
        fil_lhs%fragmented_subunit_num      =   fil_rhs%fragmented_subunit_num
        fil_lhs%fragmented_phi              =   fil_rhs%fragmented_phi
        fil_lhs%wps_as_sps                  =   fil_rhs%wps_as_sps
        fil_lhs%amyloid_beta                =   fil_rhs%amyloid_beta
        fil_lhs%wps_are_crossovers          =   fil_rhs%wps_are_crossovers
        fil_lhs%fal_wps_are_crossovers      =   fil_rhs%fal_wps_are_crossovers
        fil_lhs%fal_wps_in_score            =   fil_rhs%fal_wps_in_score
        fil_lhs%fal_wps_in_threed           =   fil_rhs%fal_wps_in_threed
        fil_lhs%microtubule                 =   fil_rhs%microtubule
        fil_lhs%mt_pf_number                =   fil_rhs%mt_pf_number
        fil_lhs%mt_rise                     =   fil_rhs%mt_rise
        fil_lhs%persistence_length          =   fil_rhs%persistence_length
        fil_lhs%torsion_persistence_length  =   fil_rhs%torsion_persistence_length
        fil_lhs%stretch_persistence_length  =   fil_rhs%stretch_persistence_length
        fil_lhs%score                       =   fil_rhs%score
    end subroutine CopyPropertiesFrom

    subroutine ResetProperties(self)
        ! Arguments
        class(filament), intent(inout)      ::  self
        ! Start work
        self%refine_params               =  .false.
        self%use_hermite_splines         =  .true.
        self%rigid_params                =  .false.
        !self%rigid_wp                    =
        self%constant_rise_per_subunit   =  .false.
        self%constant_twist_per_length   =  .false.
        self%score                       =  0.0
        self%id                          =  0
        !self%helix                       =
        self%max_radius                  =  0.0
        self%fragmented_subunit_num      =  .false.
        self%fragmented_phi              =  .false.
        self%wps_as_sps                  =  .false.
        self%amyloid_beta                =  .false.
        self%wps_are_crossovers          =  .false.
        self%fal_wps_are_crossovers      =  .false.
        self%microtubule                 =  .false.
        self%mt_pf_number                =  0
        self%mt_rise                     =  0
        self%persistence_length          =  0.0
        self%torsion_persistence_length  =  0.0
        self%stretch_persistence_length  =  0.0
        self%fal_wps_in_score            =  .true.
        self%fal_wps_in_threed           =  .true.
        self%refine_normal_modes         =  .false.
        self%x_linear                    =  0.0
        self%y_linear                    =  0.0
        self%z_linear                    =  0.0
        self%s_linear                    =  0.0
        self%phi_linear                  =  0.0
    end subroutine ResetProperties

    subroutine Allocate(self,n)
        class(Filament),    intent(inout)   ::  self
        integer,            intent(in)      ::  n
        call AllocateFilament(self,n)
    end subroutine Allocate


    !>  \brief  Allocate memory for filament - need to know the number of way points
    subroutine AllocateFilament(self,n)
        use Splines
        use Globals
        use, intrinsic  ::  iso_c_binding
        implicit none
        ! Arguments
        type(filament), target, intent(inout)   ::  self
        integer,                intent(in)      ::  n       !<  Number of way points in the filament
        ! Private variables
        integer ::  ierr
        logical,    parameter   ::  debug    =  .false.
        ! Start work
        if (self%in_memory) then
            if (this_program%IsVerbose() .or. debug) then
                print '(a)', '**warning(AllocateFilament): already allocated, will deallocate first'
            endif
            call self%Deallocate()
        endif
        self%nwp = n
        if (allocated(self%wp)) then
            if (debug) write(*,'(a)') '**debug(AllocateFilament): filament was not allocated, yet had to deallocate waypoints'
            deallocate(self%wp)
        endif
        allocate(self%wp(n),stat=ierr)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(AllocateFilament): allocation of waypoints failed. n = ', n
            call this_program%TerminateWithFatalError('AllocateFilament','Fatal error')
        endif
        ! Allocate & initialise the t array for spline parameter
        allocate(self%t(n),self%t_linear(n),stat=ierr)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(AllocateFilament): allocation of parameter array failed. n = ', n
            call this_program%TerminateWithFatalError('AllocateFilament','Fatal error')
        endif
        self%t = 0.0D0
        self%t_linear = 0.0D0
        ! Allocate the splines
        call self%twist_spline%Allocate(n,3)
        call self%x_spline%Allocate(n,3)
        call self%y_spline%Allocate(n,3)
        call self%z_spline%Allocate(n,3)
        call self%s_spline%Allocate(n,3)
        call self%gl_quadrature%Init(3)
        self%in_memory = .true.
        ! Allocate nm data
        if (allocated(self%nm_eigenvectors)) deallocate(self%nm_eigenvectors)
        if (allocated(self%nm_eigenvalues)) deallocate(self%nm_eigenvalues)
        if (allocated(self%wp_nm)) deallocate(self%wp_nm)
        if (allocated(self%wp_res)) deallocate(self%wp_res)
        allocate(self%nm_eigenvectors(n,n),self%nm_eigenvalues(n))
        allocate(self%wp_nm(n))
        allocate(self%wp_res(n))
        self%nm_eigenvectors = 0.0D0
        self%nm_eigenvalues  = 0.0D0
        ! The c pointer is setup
        self%cptr = c_loc(self)
    end subroutine AllocateFilament

    subroutine filament_array_deallocate(self)
        type(filament), intent(inout)       ::  self(:)
        !
        integer ::  current_fil
        !
        do current_fil=1,size(self)
            call self(current_fil)%Deallocate()
        enddo
    end subroutine filament_array_deallocate

    subroutine Finalise(self)
        type(filament), intent(inout)   ::  self
        call self%Deallocate()
    end subroutine

    !>  \brief  Deallocate memory from filament
    !impure elemental
    subroutine Deallocate(self)
        use Splines
        use coos, only : waypoint_deallocate
        ! Arguments
        class(filament), intent(inout)   ::  self
        ! Private variables
        integer ::  wp_counter
        ! Start work
        if (self%in_memory) then
            if (allocated(self%t)) deallocate(self%t)
            if (allocated(self%t_linear)) deallocate(self%t_linear)
            ! Deallocate splines
            call self%twist_spline%Deallocate()
            call self%x_spline%Deallocate()
            call self%y_spline%Deallocate()
            call self%z_spline%Deallocate()
            call self%s_spline%Deallocate()
            call self%gl_quadrature%Deallocate()
            ! Deallocate waypoints
            do wp_counter=1,size(self%wp)
                call waypoint_deallocate(self%wp(wp_counter))
            enddo
            deallocate(self%wp)
            !
            if (allocated(self%wp_nm)) deallocate(self%wp_nm)
            if (allocated(self%wp_res)) deallocate(self%wp_res)
            if (allocated(self%nm_eigenvectors)) deallocate(self%nm_eigenvectors)
            if (allocated(self%nm_eigenvalues)) deallocate(self%nm_eigenvalues)
            !
            self%in_memory = .false.
        endif
    end subroutine Deallocate

    pure logical function PsiAndThetaAreNonZero(self)
        class(filament), intent(in)  ::  self
        PsiAndThetaAreNonZero = any(self%wp%euler%psi .ne. 0.0) .or. any(self%wp%euler%theta .ne. 0.0)
    end function PsiAndThetaAreNonZero

    pure logical function ThetaAreNonZero(self)
        class(filament), intent(in)  ::  self
        ThetaAreNonZero = any(self%wp%euler%theta .ne. 0.0)
    end function ThetaAreNonZero

    pure logical function IsAllocated(self) result(filament_allocated)
        ! Arguments
        class(filament), intent(in)  ::  self
        ! Start work
        filament_allocated =    self%in_memory .and. &
                                self%x_spline%IsAllocated() .and. &
                                self%y_spline%IsAllocated() .and. &
                                self%z_spline%IsAllocated() .and. &
                                self%twist_spline%IsAllocated() .and. &
                                self%s_spline%IsAllocated() .and. &
                                self%gl_quadrature%IsAllocated() .and. &
                                allocated(self%wp) .and. &
                                allocated(self%wp_res) .and. &
                                allocated(self%t) .and. &
                                allocated(self%t_linear) .and. &
                                allocated(self%nm_eigenvectors) .and. &
                                allocated(self%nm_eigenvalues) .and. &
                                allocated(self%wp_nm)
    end function IsAllocated

    !>  \brief  Check whether the space curve is being refined via its normal modes
    pure logical function filament_refine_axis_nms(self)
        ! Arguments
        type(filament), intent(in)  ::  self
        ! Start work
        filament_refine_axis_nms = all(self%refine_normal_modes(1:3)) .and. all(self%refine_normal_modes(5:6))
    end function filament_refine_axis_nms

    !>  \brief  Work out the total number of dimensions of the search space
    integer function filament_expected_number_of_values(self,params_mask,params_mask_hold_average) result(num_dim_total)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament),                     intent(in)  ::  self
        logical,                            intent(in)  ::  params_mask(9)  !<  For each of the 9 parameters, whether they are included in the values array
        logical,            optional,       intent(in)  ::  params_mask_hold_average(9) !<  For each of the 9 parameters whether the average value should be left alone
        ! Private variables
        integer ::  num_dim_per_wp, num_wp, num_dim_rigid
        ! Start work
        num_dim_per_wp  =   count(params_mask .and. .not. self%rigid_params)    ! Number of parameters to refine per waypoint
        num_wp          =   count(self%wp%refine)                                           ! Number of way points we're refining
        num_dim_rigid   =   count(params_mask .and. self%rigid_params)
        num_dim_total   =   num_dim_per_wp * num_wp + num_dim_rigid
        if (present(params_mask_hold_average)) then
            num_dim_total = num_dim_total - count(params_mask_hold_average .and. params_mask)
        endif
    end function filament_expected_number_of_values

    !>  \brief  Print out local subframe shift alignments
    subroutine filament_print_subframe_shifts(self)
        use StringManipulations, only : IntegerToString
        !
        class(filament),    intent(in)  ::  self
        ! private variables
        integer ::  current_wp
        integer ::  number_of_frames
        ! start work
        number_of_frames = size(self%wp(1)%local_subframe_shift,2)
        do current_wp=1,self%nwp
            write(*,'(a,1x,i0,1x,a,1x,'//IntegerToString(number_of_frames)//'(f0.1,1x))') 'Waypoint', current_wp, 'X:', &
                                                                            self%wp(current_wp)%local_subframe_shift(1,:)
            write(*,'(a,1x,i0,1x,a,1x,'//IntegerToString(number_of_frames)//'(f0.1,1x))') 'Waypoint', current_wp, 'Y:', &
                                                                            self%wp(current_wp)%local_subframe_shift(2,:)
        enddo

    end subroutine filament_print_subframe_shifts

    !>  \brief  Print nicely all data contained in the Filament object
    subroutine PrintInfo(self,verbose)
        use coos, only : waypoints_print_info
        use helices, only : helical_symmetry_print_info
        ! Arguments
        class(filament),            intent(in)      ::  self
        logical,        optional,   intent(in)      ::  verbose !<  Print out more details
        ! Private variables
        integer ::  i
        character(len=5),   parameter   ::  param_label(7)  =   (/  'X    ',  &
                                                                    'Y    ',  &
                                                                    'Z    ',  &
                                                                    'Phi  ',  &
                                                                    'Theta',  &
                                                                    'Psi  ',  &
                                                                    'SubNm'/)
        logical ::  vverbose
        ! Start work

        vverbose = .false.
        if (present(verbose)) vverbose = verbose


        !
        write(*,'(/a,i0,a)')                    'Filament info for filament ', self%id, ':'
        if (vverbose) write(*,'(a,l1)')         '  allocated? ', self%IsAllocated()
        if (vverbose) write(*,'(a,i0)')         '  number of way points = ', self%nwp
        write(*,'(a,i0,a,i0,a)')                '  on micrograph number (location in file) ', self%film%GetID(), &
                                                                                        ' (',self%film%GetLocationInFile(),')'
        if (vverbose) write(*,'(a,4(l1,1x))')   '  known_d (x,y,z,Phi)? ',  self%x_spline%FirstDerivativesAreKnown(), &
                                                                            self%y_spline%FirstDerivativesAreKnown(), &
                                                                            self%z_spline%FirstDerivativesAreKnown(), &
                                                                            self%twist_spline%FirstDerivativesAreKnown()
        write(*,'(a,l1)')                       '  amyloid beta? ', self%amyloid_beta
        if (vverbose) write(*,'(a,2l2)')        '  WPs, fal WPs are crossovers? ', self%wps_are_crossovers, &
                                                                                    self%fal_wps_are_crossovers
        write(*,'(a,2l2)')                      '  fal WPs in scoring function? 3d rec? ', self%fal_wps_in_score, &
                                                                                           self%fal_wps_in_threed
        if (vverbose) write(*,'(a,9(l1,1x))')   '  refine normal mode amplitudes? ', self%refine_normal_modes
        if (vverbose) write(*,'(a,l1)')         '  constant rise per subunit? ', self%constant_rise_per_subunit
        if (vverbose) write(*,'(a,l1)')         '  constant twist per length? ', self%constant_twist_per_length
        if (vverbose) write(*,'(a,9(l1,1x))')   '  rigid parameters? ', self%rigid_params
        if (vverbose) write(*,'(a,l1)')         '  use Hermite splines for the space curve? ', self%use_hermite_splines
        if (vverbose) write(*,'(a,2(l1,1x))')   '  fragmented phi? subunit_num? ', self%fragmented_phi, self%fragmented_subunit_num
        write(*,'(a)',advance='NO')             '  average helical parameters: '
        call helical_symmetry_print_info(self%helix)
        if (vverbose) then
        write(*,'(a)',advance='NO')             '  min helical parameters for inclusion in 3d rec: '
        call helical_symmetry_print_info(self%helix_min_for_3drec)
        write(*,'(a)',advance='NO')             '  max helical parameters for inclusion in 3d rec: '
        call helical_symmetry_print_info(self%helix_max_for_3drec)
        endif

        if (vverbose) write(*,'(a,F0.3)')                     '  max radius (pixels) = ', self%max_radius
        write(*,'(a,3(F0.3,1x))')               '  persistence lengths (curvature torsion stretch) (pixels) = ', &
                                                    self%persistence_length, self%torsion_persistence_length, &
                                                    self%stretch_persistence_length
        write(*,'(a,F0.4)')                     '  score = ', self%score
        if (self%microtubule) then
            write(*,'(a,i0,1x,i0)')             '  microtubule pf number; rise ', self%mt_pf_number, self%mt_rise
        endif
        if (any(self%rigid_params)) then
            if (vverbose) write(*,'(a)',advance='NO') &
                                                '  rigid parameters: '
            do I=1,7
                if (self%rigid_params(i)) then
                    if (vverbose)  write(*,'(2a)',advance='NO') param_label(i), ': '
                    select case (i)
                        case (1:3)
                            if (vverbose) write(*,'(F0.3)',advance='NO') self%rigid_wp%coo(i)
                        case (4)
                            if (vverbose) write(*,'(F0.3)',advance='NO') self%rigid_wp%euler%phi
                        case (5)
                            if (vverbose) write(*,'(F0.3)',advance='NO') self%rigid_wp%euler%theta
                        case (6)
                            if (vverbose) write(*,'(F0.3)',advance='NO') self%rigid_wp%euler%psi
                        case (7)
                            if (vverbose) write(*,'(F0.3)',advance='NO') self%rigid_wp%subunit_num
                    end select
                    if (vverbose) write(*,'(a)',advance='NO') '  '
                endif
            enddo
            if (vverbose) write(*,'(a)') ' '
        endif
        call waypoints_print_info(self%wp)
        write(*,'(a/)')             ' '
        if (allocated(self%wp_nm)) then
            if (vverbose) write(*,'(a,i0)') 'Normal mode amplitudes for filament ', self%id
            if (vverbose) call waypoints_print_info(self%wp_nm)
        else
            if (vverbose) write(*,'(a)') 'Normal modes not set up'
        endif
        !nm_amp_sum = filament_nm_amplitudes_weighted_sum(self)
        !write(*,'(a,F0.3)') 'Weighted sum of nm amplitudes = ', nm_amp_sum
        !write(*,'(a,2(F0.3,1x))') 'Linear fit coefficients (x): ', self%x_linear
        !write(*,'(a,2(F0.3,1x))') 'Linear fit coefficients (y): ', self%y_linear
        !write(*,'(a,2(F0.3,1x))') 'Linear fit coefficients (z): ', self%z_linear
        !write(*,'(a,2(F0.3,1x))') 'Linear fit coefficients (Phi): ', self%phi_linear
        !write(*,'(a,2(F0.3,1x))') 'Linear fit coefficients (Subunit_num): ', self%s_linear
        if (allocated(self%wp_res)) then
            if (vverbose) write(*,'(a)') 'Linear fit residuals'
            if (vverbose) call waypoints_print_info(self%wp_res)
        else
            if (vverbose) write(*,'(a)') 'Residuals not set up'
        endif

        ! Print out spline information
        if (vverbose) then
            if (self%x_spline%IsAllocated()) then
                write(*,'(a)') 'Spline X'
                call self%x_spline%PrintInfo()
            endif
            if (self%y_spline%IsAllocated()) then
                write(*,'(a)') 'Spline Y'
                call self%y_spline%PrintInfo()
            endif
            if (self%z_spline%IsAllocated()) then
                write(*,'(a)') 'Spline Z'
                call self%z_spline%PrintInfo()
            endif
            if (self%twist_spline%IsAllocated()) then
                write(*,'(a)') 'Spline Twist'
                call self%twist_spline%PrintInfo()
            endif
            if (self%s_spline%IsAllocated()) then
                write(*,'(a)') 'Spline S'
                call self%s_spline%PrintInfo()
            endif
            write(*,'(a//)') ' '
        endif

        ! Print out subframe shifts
        if (vverbose) then
            call filament_print_subframe_shifts(self)
        endif

    end subroutine PrintInfo

    !>  \brief  The t parameter is set to the current estimate of the arc length, plus the appropriate offset
    pure subroutine SetTFromPathLength(self)
        ! Arguments
        class(filament), intent(inout)   ::  self
        ! Start work
        self%t = self%wp%path_length
    end subroutine SetTFromPathLength

end module filaments_base
