!>  \brief  Methods for the filament class which deal with geometry, euler angles, measurements etc.
!!
!!          Generally, these are methods which don't have to do with image processing.
#define use_quadrature
module filaments_geometry
    use filaments_base, only : filament
    implicit none

    contains

    !>  \brief  Change the subunit_num parameters from the usual values to incremental values
    !!
    !!  The "normal" representation has each waypoint's subunit_num value indicate directly the
    !!  value of the subunit_num function at that point of the space curve.
    !!  Alternatively, in the "incremental" representation, each waypoint's subunit_num value
    !!  is the difference in subunit_num between that waypoint and its preceding waypoint.
    !!  The filament's first waypoint is the only one which is unchanged by the transformation.
    !!  \todo deprecate and consolidate with the more general "params" subroutine
    pure subroutine filament_subunit_num_absolute_to_incremental(self)
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        ! Private variables
        integer ::  current_wp
        ! Start work

        ! Loop over waypoints
        do current_wp=self%nwp,2,-1
            self%wp(current_wp)%subunit_num = self%wp(current_wp)%subunit_num - &
                                            self%wp(current_wp-1)%subunit_num
        enddo
    end subroutine filament_subunit_num_absolute_to_incremental

    !>  \brief  Change the subunit_num representation from incremental to absolute
    pure subroutine filament_subunit_num_incremental_to_absolute(self)
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        ! Private variables
        integer ::  current_wp
        ! Start work

        ! Loop over waypoints
        do current_wp=2,self%nwp
            self%wp(current_wp)%subunit_num = self%wp(current_wp)%subunit_num + &
                                            self%wp(current_wp-1)%subunit_num
        enddo
    end subroutine filament_subunit_num_incremental_to_absolute

    !>  \brief  Change the filament parameter representation from incremental to absolute
    pure subroutine filament_params_incremental_to_absolute(self,param_mask)
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        logical,        intent(in)      ::  param_mask(7)
        ! Private variables
        integer ::  current_wp
        ! Start work

        ! Loop over waypoints
        do current_wp=2,self%nwp
            if (param_mask(1)) then
                self%wp(current_wp)%coo(1) = self%wp(current_wp)%coo(1) + &
                                             self%wp(current_wp-1)%coo(1)
            endif
            if (param_mask(2)) then
                self%wp(current_wp)%coo(2) = self%wp(current_wp)%coo(2) + &
                                             self%wp(current_wp-1)%coo(2)
            endif
            if (param_mask(3)) then
                self%wp(current_wp)%coo(3) = self%wp(current_wp)%coo(3) + &
                                             self%wp(current_wp-1)%coo(3)
            endif
            if (param_mask(4)) then
                self%wp(current_wp)%euler%phi = self%wp(current_wp)%euler%phi + &
                                                self%wp(current_wp-1)%euler%phi
            endif
            if (param_mask(5)) then
                self%wp(current_wp)%euler%theta = self%wp(current_wp)%euler%theta + &
                                                  self%wp(current_wp-1)%euler%theta
            endif
            if (param_mask(6)) then
                self%wp(current_wp)%euler%psi = self%wp(current_wp)%euler%psi + &
                                                self%wp(current_wp-1)%euler%psi
            endif
            if (param_mask(7)) then
                self%wp(current_wp)%subunit_num = self%wp(current_wp)%subunit_num + &
                                                  self%wp(current_wp-1)%subunit_num
            endif
        enddo
    end subroutine filament_params_incremental_to_absolute

    pure subroutine filament_params_absolute_to_incremental(self,param_mask)
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        logical,        intent(in)      ::  param_mask(7)   !<  Which parameters should be switched
        ! Private variables
        integer ::  current_wp
        ! Start work

        ! Loop over waypoints
        do current_wp=self%nwp,2,-1
            if (param_mask(1)) then
                self%wp(current_wp)%coo(1) = self%wp(current_wp)%coo(1) - &
                                             self%wp(current_wp-1)%coo(1)
            endif
            if (param_mask(2)) then
                self%wp(current_wp)%coo(2) = self%wp(current_wp)%coo(2) - &
                                             self%wp(current_wp-1)%coo(2)
            endif
            if (param_mask(3)) then
                self%wp(current_wp)%coo(3) = self%wp(current_wp)%coo(3) - &
                                             self%wp(current_wp-1)%coo(3)
            endif
            if (param_mask(4)) then
                self%wp(current_wp)%euler%phi = self%wp(current_wp)%euler%phi - &
                                                self%wp(current_wp-1)%euler%phi
            endif
            if (param_mask(5)) then
                self%wp(current_wp)%euler%theta = self%wp(current_wp)%euler%theta - &
                                                self%wp(current_wp-1)%euler%theta
            endif
            if (param_mask(6)) then
                self%wp(current_wp)%euler%psi = self%wp(current_wp)%euler%psi - &
                                                self%wp(current_wp-1)%euler%psi
            endif
            if (param_mask(7)) then
                self%wp(current_wp)%subunit_num = self%wp(current_wp)%subunit_num - &
                                                  self%wp(current_wp-1)%subunit_num
            endif
        enddo
    end subroutine filament_params_absolute_to_incremental


    !>  \brief  Return the average (x,y,z) coordinates of the filament
    function filament_average_wp(self) result(wp_ave)
        use coos, only : waypoints_ave_std, waypoint
        implicit none
        ! Argument
        type(filament), intent(in)  ::  self
        ! Result
        type(waypoint)  ::  wp_ave
        ! Private variables
        type(waypoint)  ::  wps(2)
        ! Start work
        call waypoints_ave_std(self%wp,wps)
        wp_ave = wps(1)
    end function filament_average_wp

    !>  \brief  For each spline, compute the residuals
    subroutine filament_update_wp_res(self)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        ! Private variables
        real(kind=8)    ::  theta_ave, psi_ave
        ! Start work

        if (self%nwp .ge. 3) then
            ! Safety check
            if (all (self%t_linear .eq. 0.0D0)) then
                write(*,'(a)') '**ERROR(filament_update_wp_res): t_linear not setup'
                call terminate('Fatal error in FILAMENT_UDPATE_WP_RES. Programming error.')
            endif

            self%wp_res%coo(1)      = self%wp%coo(1)      - self%x_linear(1)   * self%t_linear   - self%x_linear(2)
            self%wp_res%coo(2)      = self%wp%coo(2)      - self%y_linear(1)   * self%t_linear   - self%y_linear(2)
            self%wp_res%coo(3)      = self%wp%coo(3)      - self%z_linear(1)   * self%t_linear   - self%z_linear(2)
            self%wp_res%subunit_num = self%wp%subunit_num - self%s_linear(1)   * self%t_linear   - self%s_linear(2)
            self%wp_res%euler%phi   = self%wp%euler%phi   - self%phi_linear(1) * self%t_linear   - self%phi_linear(2)

            ! They're not strictly residuals (since there is no spline), but we can fill the values for theta & psi with deviations
            ! from the mean
            psi_ave   = sum(self%wp%euler%psi)   / real(self%nwp,kind=8)
            theta_ave = sum(self%wp%euler%theta) / real(self%nwp,kind=8)
            self%wp_res%euler%theta = self%wp%euler%theta - theta_ave
            self%wp_res%euler%psi   = self%wp%euler%psi   - psi_ave
        else
            self%wp_res%coo(1) = 0.0d0
            self%wp_res%coo(2) = 0.0d0
            self%wp_res%coo(3) = 0.0d0
            self%wp_res%subunit_num = 0.0d0
            self%wp_res%euler%phi = 0.0d0
            self%wp_res%euler%theta = 0.0d0
            self%wp_res%euler%psi = 0.0d0
        endif
    end subroutine filament_update_wp_res

    !>  \brief  Re-compute waypoint values from the residuals
    subroutine filament_update_wps_from_wp_res(self,axis,phi,subunit_num)
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self
        logical,        optional,   intent(in)      ::  axis        !<  Update axis properties (x,y,z)
        logical,        optional,   intent(in)      ::  phi         !<  Update phi
        logical,        optional,   intent(in)      ::  subunit_num !<  Update subunit_num
        ! Private variables
        logical ::  aaxis, pphi, ssubunit_num
        ! Start work
        aaxis = .true.
        if (present(axis)) aaxis = axis
        pphi = .true.
        if (present(phi)) pphi = phi
        ssubunit_num = .true.
        if (present(subunit_num)) ssubunit_num = subunit_num
        !
        if (aaxis) then
            self%wp%coo(1)      = self%wp_res%coo(1)      + self%x_linear(1)   * self%t_linear   + self%x_linear(2)
            self%wp%coo(2)      = self%wp_res%coo(2)      + self%y_linear(1)   * self%t_linear   + self%y_linear(2)
            self%wp%coo(3)      = self%wp_res%coo(3)      + self%z_linear(1)   * self%t_linear   + self%z_linear(2)
        endif
        if (ssubunit_num) then
            self%wp%subunit_num = self%wp_res%subunit_num + self%s_linear(1)   * self%t_linear   + self%s_linear(2)
        endif
        if (pphi) then
            self%wp%euler%phi   = self%wp_res%euler%phi   + self%phi_linear(1) * self%t_linear   + self%phi_linear(2)
        endif
    end subroutine filament_update_wps_from_wp_res


    !>  \brief  Compute a linear fit of the splines
    subroutine filament_fit_splines_linear(self)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament),             intent(inout)  ::  self
        ! Private variables
        integer ::  i
        real(kind=8)    ::  s, st, stt, sx, sy, sz, ss, sphi
        real(kind=8)    ::  stx, sty, stz, sts, stphi
        real(kind=8)    ::  d
        logical         ::  all_spline_ordinates_match
        ! Start work

        ! Need at least 2 (or maybe 3?) wps
        if (self%nwp .ge. 3) then

            ! Keep a record of the ordinates
            self%t_linear = self%t  !self%x_spline%x

            ! Sanity check. The ordinates should be the same for all the splines
            all_spline_ordinates_match =    all(self%t_linear .eq. self%x_spline%x) .and. &
                                            all(self%t_linear .eq. self%y_spline%x) .and. &
                                            all(self%t_linear .eq. self%z_spline%x) .and. &
                                            all(self%t_linear .eq. self%s_spline%x) .and. &
                                            all(self%t_linear .eq. self%twist_spline%x )
            if (.not. all_spline_ordinates_match) then
                write(*,'(a)') '**ERROR(filament_fit_splines_linear): not all splines have the same parameter values'
                call terminate('Fatal error in FILAMENT_FIT_SPLINES_LINEAR. Programming error.')
            endif

            ! See eq 15.2.4 and onwards of Numerical recipes
            s        = real(self%x_spline%NumberOfKnots(),KIND=8)
            st       = 0.0D0
            stt      = 0.0D0
            sx       = 0.0D0
            sy       = 0.0D0
            sz       = 0.0D0
            ss       = 0.0D0
            sphi     = 0.0D0
            stx      = 0.0D0
            sty      = 0.0D0
            stz      = 0.0D0
            sts      = 0.0D0
            stphi    = 0.0D0
            do i=1,self%x_spline%NumberOfKnots()
                ! Accumulate sums
                st       =   st      + self%t_linear(i)
                stt      =   stt     + self%t_linear(i)**2
                sx       =   sx      + self%x_spline%y(i)
                sy       =   sy      + self%y_spline%y(i)
                sz       =   sz      + self%z_spline%y(i)
                ss       =   ss      + self%s_spline%y(i)
                sphi     =   sphi    + self%twist_spline%y(i)
                stx      =   stx     + self%x_spline%x(i) *   self%x_spline%y(i)
                sty      =   sty     + self%x_spline%x(i) *   self%y_spline%y(i)
                stz      =   stz     + self%x_spline%x(i) *   self%z_spline%y(i)
                sts      =   sts     + self%x_spline%x(i) *   self%s_spline%y(i)
                stphi    =   stphi   + self%x_spline%x(i) *   self%twist_spline%y(i)
            enddo

            d = s*stt - st**2
            !
            self%x_linear(1)   = ( s*stx    - st*sx    ) / d
            self%x_linear(2)   = ( stt*sx   - st*stx   ) / d

            self%y_linear(1)   = ( s*sty    - st*sy    ) / d
            self%y_linear(2)   = ( stt*sy   - st*sty   ) / d

            self%z_linear(1)   = ( s*stz    - st*sz    ) / d
            self%z_linear(2)   = ( stt*sz   - st*stz   ) / d

            self%s_linear(1)   = ( s*sts    - st*ss    ) / d
            self%s_linear(2)   = ( stt*ss   - st*sts   ) / d

            self%phi_linear(1) = ( s*stphi  - st*sphi  ) / d
            self%phi_linear(2) = ( stt*sphi - st*stphi ) / d

        endif

    end subroutine filament_fit_splines_linear

    !>  \brief  Find the maximal normal distance for the filament
    subroutine filament_maximal_normal_distance(self,maximal_normal_distance)
        use coos, only : waypoint_waypoint_distance
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        real,           intent(out)     ::  maximal_normal_distance
        ! Private variables
        integer         ::  start_wp, finish_wp
        real            ::  chord_length, chord_vector(3)
        real(kind=8)    ::  t
        real            ::  normal_distance
        integer         ::  i
        ! Start work

        ! Initialise the output
        maximal_normal_distance = 0.0

        ! Define the chord axis
        if (self%fal_wps_in_threed .and. self%fal_wps_in_score) then
            start_wp = 1
            finish_wp = self%nwp
        else
            start_wp = 2
            finish_wp = self%nwp - 1
        endif

        ! Chord length
        chord_vector = self%wp(finish_wp)%coo - self%wp(start_wp)%coo
        chord_length = sqrt(chord_vector(1)**2 + chord_vector(2)**2 + chord_vector(3)**2)

        ! Loop along the filament (1 pixel per step) and look for the longest normal distance
        do I=1,int(self%t(finish_wp)-SELF%t(start_wp))+1
            t = self%t(start_wp) + dble(i)
            normal_distance =  filament_eval_normal_distance(self,self%wp(start_wp)%coo,chord_vector,chord_length,t)
            if (normal_distance .gt. maximal_normal_distance) then
                maximal_normal_distance = normal_distance
            endif
        enddo



    end subroutine filament_maximal_normal_distance

    !>  \brief  Evaluate the mid-point normal distance for the filament, in 3d and in 2d projection
    subroutine filament_mid_point_normal_distance(self,mid_point_normal_dist,mid_point_normal_dist_2d)
        implicit none
        ! Arguments
        type(filament),     intent(inout)   ::  self
        real,               intent(out)     ::  mid_point_normal_dist
        real,               intent(out)     ::  mid_point_normal_dist_2d
        ! Private variables
        integer ::  start_wp, finish_wp
        real    ::  chord_length, chord_vector(3)
        ! Start work

        ! Define the first and last waypoints we're considering
        if (self%fal_wps_in_threed .and. self%fal_wps_in_score) then
            start_wp = 1
            finish_wp = self%nwp
        else
            start_wp = 2
            finish_wp = self%nwp - 1
        endif

        !
        chord_vector = self%wp(finish_wp)%coo - self%wp(start_wp)%coo
        chord_length = sqrt(chord_vector(1)**2+CHORD_VECTOR(2)**2+CHORD_VECTOR(3)**2)

        !
        mid_point_normal_dist    = filament_eval_normal_distance(self,self%wp(start_wp)%coo,chord_vector,chord_length,          &
                                                                 T=(self%t(finish_wp)-SELF%t(start_wp))/2+SELF%t(start_wp))
        mid_point_normal_dist_2d = filament_eval_normal_distance(self,self%wp(start_wp)%coo,chord_vector,chord_length,          &
                                                                 T=(self%t(finish_wp)-SELF%t(start_wp))/2+SELF%t(start_wp),     &
                                                                 TWO_D=.true.)

    end subroutine filament_mid_point_normal_distance

    real function filament_eval_normal_distance(self,start_coo,chord_vector,chord_length,t,two_d) result(normal_distance)
        use coos, only  :   waypoint
        use Splines
        ! Arguments
        type(filament),         intent(inout)   ::  self                !<  Filament object
        real(kind=8),           intent(in)      ::  start_coo(3)        !<  Coordinates of the first waypoint considered
        real,                   intent(in)      ::  chord_vector(3)     !<  Chord vector
        real,                   intent(in)      ::  chord_length        !<  Chord length
        real(kind=8),           intent(in)      ::  t
        logical,    optional,   intent(in)      ::  two_d               !<  Return the 2d normal distance
        ! Private variables
        real(kind=8)                    ::  coo(3)
        real                            ::  wp_vector(3)
        real                            ::  cross_product(3)
        logical                         ::  ttwo_d
        ! Start work

        ttwo_d = .false.
        if (present(two_d)) ttwo_d = two_d

        if (self%nwp .ge. 3) then
            !
            call self%x_spline%Evaluate(t,coo(1))
            call self%y_spline%Evaluate(t,coo(2))
            call self%z_spline%Evaluate(t,coo(3))

            wp_vector = coo - start_coo

            ! Compute the cross product
            ! u = chord
            ! v = wp
            cross_product(1) =  chord_vector(2) * wp_vector(3) - chord_vector(3) * wp_vector(2)
            cross_product(2) =  chord_vector(3) * wp_vector(1) - chord_vector(1) * wp_vector(3)
            cross_product(3) =  chord_vector(1) * wp_vector(2) - chord_vector(2) * wp_vector(1)

            !
            if (ttwo_d) then
                normal_distance = sqrt(cross_product(1)**2 + cross_product(2)**2)  &
                                        / chord_length
            else
                normal_distance = sqrt(cross_product(1)**2 + cross_product(2)**2 + cross_product(3)**2)  &
                                        / chord_length
            endif
        else
            normal_distance = 0.0
        endif
    end function filament_eval_normal_distance


    !>  \brief  Return the chord length of the filament
    subroutine filament_chord_length(self,chord_length_3d,chord_length_2d)
        use coos, only : waypoint_waypoint_distance
        implicit none
        ! Arguments
        type(filament), intent(in)  ::  self
        real,           intent(out) ::  chord_length_3d
        real,           intent(out) ::  chord_length_2d
        ! Private variables
        integer ::  start_wp, finish_wp
        ! Start work
        ! Define the chord axisSELF%nwp
        if (self%fal_wps_in_threed .and. self%fal_wps_in_score) then
            start_wp = 1
            finish_wp = self%nwp
        else
            start_wp = 2
            finish_wp = self%nwp - 1
        endif
        chord_length_3d = waypoint_waypoint_distance(self%wp(start_wp),self%wp(finish_wp))
        chord_length_2d = sqrt(     (self%wp(finish_wp)%coo(1)-self%wp(start_wp)%coo(1))**2 &
                                +   (self%wp(finish_wp)%coo(2)-self%wp(start_wp)%coo(2))**2)
    end subroutine filament_chord_length

    !>  \brief  Return the arc length of the filament (i.e. the distance along the helical axis)
    pure real function filament_arc_length(self,ignore_fal_wps)
        implicit none
        ! Arguments
        type(filament),             intent(in)  ::  self
        logical,        optional,   intent(in)  ::  ignore_fal_wps      !<  Ignore the first and last waypoints
        ! Private variables
        logical ::  iignore_fal_wps
        ! Start work
        iignore_fal_wps = .false.
        if (present(ignore_fal_wps)) iignore_fal_wps = ignore_fal_wps

        if (iignore_fal_wps) then
            filament_arc_length = self%wp(self%nwp-1)%path_length - self%wp(2)%path_length
        else
            filament_arc_length = self%wp(self%nwp  )%path_length - self%wp(1)%path_length
        endif
    end function filament_arc_length

    !>  \broef  Find the average changes in phi and subunit_num per unit length
    subroutine filament_eval_average_dphi_ds(self,average_dphi,average_ds,ignore_fal_wps)
        use Splines
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self
        real(kind=8),               intent(out)     ::  average_dphi
        real(kind=8),               intent(out)     ::  average_ds
        logical,        optional,   intent(in)      ::  ignore_fal_wps
        ! Private variables
        logical         ::  iignore_fal_wps
        integer         ::  first_wp,last_wp
        real(kind=8)    ::  step_size, current_t
        integer         ::  num_steps
        ! Ignore FAL?
        iignore_fal_wps = .not. self%fal_wps_in_score
        if (present(ignore_fal_wps)) iignore_fal_wps = ignore_fal_wps

        if (        (iignore_fal_wps .and. self%nwp .ge. 3) &
            .or.    (.not. iignore_fal_wps .and. self%nwp .ge. 2) ) then
            ! First and last waypoint
            if (iignore_fal_wps) then
                first_wp = 2
                last_wp  = self%nwp -1
            else
                first_wp = 1
                last_wp  = self%nwp
            endif

            ! Step size for search
            if (self%helix%pixel_units) then
                step_size = self%helix%rise
            else
                step_size = self%helix%rise / self%film%GetPixelSize()
            endif

            ! Search the filament
            current_t = self%t(first_wp)
            average_dphi = 0.0d0
            average_ds   = 0.0d0
            num_steps    = 0
            do
                num_steps = num_steps + 1
                !
                average_dphi = average_dphi + self%twist_spline%EvaluateDerivative(current_t)
                average_ds   = average_ds   + self%s_spline%EvaluateDerivative(current_t)
                !
                current_t = current_t + step_size
                if (current_t .gt. self%t(last_wp)) exit
            enddo
            average_dphi = average_dphi / real(num_steps,kind=8)
            average_ds   = average_ds   / real(num_steps,kind=8)
        endif
    end subroutine filament_eval_average_dphi_ds


    !>  \brief  Find the largest local curvature of a filament.
    subroutine filament_eval_maximum_curvature_overtwist_stretch(self,max_curvature,max_overtwist,max_stretch,ignore_fal_wps)
        ! Arguments
        type(filament),             intent(inout)   ::  self
        real(kind=8),               intent(out)     ::  max_curvature
        real(kind=8),               intent(out)     ::  max_overtwist
        real(kind=8),               intent(out)     ::  max_stretch
        logical,        optional,   intent(in)      ::  ignore_fal_wps      !<  ignore the first and last segments
        ! Private variables
        logical         ::  iignore_fal_wps
        integer         ::  first_wp,last_wp
        real(kind=8)    ::  step_size, current_t, current_c, current_o, current_s
        real(kind=8)    ::  average_dphi, average_ds
        logical, parameter  ::  debug = .false.
        ! Start work

        if (debug) write(*,'(a)') '**debug(filament_eval_maximum_curvature_overtwist_stretch): entering'

        ! Ignore FAL?
        iignore_fal_wps = .not. self%fal_wps_in_score
        if (present(ignore_fal_wps)) iignore_fal_wps = ignore_fal_wps

        ! First and last waypoint
        if (iignore_fal_wps) then
            first_wp = 2
            last_wp  = self%nwp -1
        else
            first_wp = 1
            last_wp  = self%nwp
        endif

        ! Step size for search
        if (self%helix%pixel_units) then
            step_size = self%helix%rise
        else
            step_size = self%helix%rise / self%film%GetPixelSize()
        endif

        if (debug) write(*,'(a)') '**debug(filament_eval_maximum_curvature_overtwist_stretch): about to get average dphi ds'

        !
        call filament_eval_average_dphi_ds(self,average_dphi,average_ds,ignore_fal_wps=iignore_fal_wps)

        if (debug) write(*,'(a)') '**debug(filament_eval_maximum_curvature_overtwist_stretch): about to loop to find max'

        ! Search the filament
        current_t = self%t(first_wp)
        max_curvature = 0.0d0
        max_overtwist = 0.0d0
        max_stretch   = 0.0d0
        do
            current_c = filament_eval_curvature(self,current_t)
            current_o = abs(filament_eval_overtwist(self,current_t,average_dphi))
            current_s = abs(filament_eval_stretch(self,current_t,average_ds))
            if (current_c .gt. max_curvature) max_curvature = current_c
            if (current_o .gt. max_overtwist) max_overtwist = current_o
            if (current_s .gt. max_stretch  ) max_stretch   = current_s
            current_t = current_t + step_size
            if (current_t .gt. self%t(last_wp)) exit
        enddo

        if (debug) write(*,'(a)') '**debug(filament_eval_maximum_curvature_overtwist_stretch): exiting'
    end subroutine filament_eval_maximum_curvature_overtwist_stretch


    !>  \brief  Find the largest local curvature of a filament.
    subroutine filament_eval_average_curvature_overtwist_stretch(self,ave_curvature,ave_overtwist,ave_stretch,ignore_fal_wps)
        ! Arguments
        type(filament),             intent(inout)   ::  self
        real(kind=8),               intent(out)     ::  ave_curvature
        real(kind=8),               intent(out)     ::  ave_overtwist
        real(kind=8),               intent(out)     ::  ave_stretch
        logical,        optional,   intent(in)      ::  ignore_fal_wps      !<  ignore the first and last segments
        ! Private variables
        logical         ::  iignore_fal_wps
        integer         ::  first_wp,last_wp
        real(kind=8)    ::  step_size, current_t
        real(kind=8)    ::  average_dphi, average_ds
        integer         ::  num_steps
        ! Start work

        ! Ignore FAL?
        iignore_fal_wps = .not. self%fal_wps_in_score
        if (present(ignore_fal_wps)) iignore_fal_wps = ignore_fal_wps

        ! First and last waypoint
        if (iignore_fal_wps) then
            first_wp = 2
            last_wp  = self%nwp -1
        else
            first_wp = 1
            last_wp  = self%nwp
        endif

        ! Step size for search
        if (self%helix%pixel_units) then
            step_size = self%helix%rise
        else
            step_size = self%helix%rise / self%film%GetPixelSize()
        endif

        !
        call filament_eval_average_dphi_ds(self,average_dphi,average_ds,ignore_fal_wps=iignore_fal_wps)

        ! Search the filament
        current_t = self%t(first_wp)
        ave_curvature = 0.0d0
        ave_overtwist = 0.0d0
        ave_stretch   = 0.0d0
        num_steps     = 0
        do
            num_steps = num_steps + 1

            ave_curvature = ave_curvature + filament_eval_curvature(self,current_t)
            ave_overtwist = ave_overtwist + abs(filament_eval_overtwist(self,current_t,average_dphi))
            ave_stretch   = ave_stretch   + abs(filament_eval_stretch(self,current_t,average_ds))

            current_t = current_t + step_size
            if (current_t .gt. self%t(last_wp)) exit
        enddo
        ave_curvature = ave_curvature / real(num_steps,kind=8)
        ave_overtwist = ave_overtwist / real(num_steps,kind=8)
        ave_stretch   = ave_stretch   / real(num_steps,kind=8)
    end subroutine filament_eval_average_curvature_overtwist_stretch

    !>  \brief  Evaluate the total curvature along a filament. Optionally, the total squared curvature can be returned.
    function filament_eval_average_curvature(self,ignore_fal_wps) result(average_curvature)
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self
        logical,        optional,   intent(in)      ::  ignore_fal_wps      !<  Whether to ignore the first and last waypoints in the calculation
        ! Result
        real(kind=8)    ::  average_curvature
        ! Private variables
        real(kind=8)    ::  step_size
        real(kind=8)    ::  current_t
        logical         ::  iignore_fal_wps
        integer         ::  first_wp, last_wp
        integer         ::  count_steps
        ! Start work

        ! Ignore fal WPS?
        if (present(ignore_fal_wps)) then
            iignore_fal_wps = ignore_fal_wps
        else
            iignore_fal_wps = .not. self%fal_wps_in_score
        endif

        ! First and last waypoints
        if (iignore_fal_wps) then
            first_wp    =   2
            last_wp     =   self%nwp - 1
        else
            first_wp    =   1
            last_wp     =   self%nwp
        endif

        ! The step size for the integration
        if (self%helix%pixel_units) then
            step_size = self%helix%rise
        else
            step_size = self%helix%rise/self%film%GetPixelSize()
        endif

        ! Total curvature
        average_curvature = 0.0D0
        count_steps = 0
        current_t = self%t(first_wp)
        do
            average_curvature = average_curvature + filament_eval_curvature(self,current_t)
            current_t = current_t + step_size
            count_steps = count_steps + 1
            if (current_t .gt. self%t(last_wp)) exit
        enddo
        average_curvature = average_curvature / count_steps
    end function filament_eval_average_curvature

    !>  \brief  Evaluate the total change of direction of the filament in radians, i.e. as though all the curvature was "flattened" into a single arc of circle
    real function filament_eval_change_of_direction(self)
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self
        ! Private variables
        ! Start work
        filament_eval_change_of_direction = filament_eval_total_curvature(self,SQUARED_CURVATURE=.false.)
    end function filament_eval_change_of_direction


    !>  \brief  Evaluate the total curvature along a filament. Optionally, the total squared curvature can be returned.
    function filament_eval_total_curvature(self,squared_curvature,ignore_fal_wps) result(total_curvature)
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self
        logical,        optional,   intent(in)      ::  squared_curvature   !<  If .TRUE., the total squared curvature will be returned. Defaults to false, in which case the total curvature (not squared) is returned.
        logical,        optional,   intent(in)      ::  ignore_fal_wps      !<  Whether to ignore the first and last waypoints in the calculation. False by default.
        ! Result
        real(kind=8)    ::  total_curvature
        ! Private variables
        logical         ::  ssquared_curvature
        real(kind=8)    ::  step_size
        real(kind=8)    ::  current_t
        logical         ::  iignore_fal_wps
        integer         ::  first_wp, last_wp
        ! Start work
        ! Will we be returning the squared curvature?
        if (present(squared_curvature)) then
            ssquared_curvature = squared_curvature
        else
            ssquared_curvature = .true.
        endif

        ! Ignore fal WPS?
        if (present(ignore_fal_wps)) then
            iignore_fal_wps = ignore_fal_wps
        else
            iignore_fal_wps = .not. self%fal_wps_in_score
        endif

        ! First and last waypoints
        if (iignore_fal_wps) then
            first_wp    =   2
            last_wp     =   self%nwp - 1
        else
            first_wp    =   1
            last_wp     =   self%nwp
        endif

        ! The step size for the integration
        if (self%helix%pixel_units) then
            step_size = self%helix%rise
        else
            step_size = self%helix%rise/self%film%GetPixelSize()
        endif

        ! Total curvature
        total_curvature = 0.0D0
        current_t = self%t(first_wp)
        if (ssquared_curvature) then
            do
                total_curvature = total_curvature + filament_eval_curvature(self,current_t)**2
                current_t = current_t + step_size
                if (current_t .gt. self%t(last_wp)) exit
            enddo
        else
            do
                total_curvature = total_curvature + filament_eval_curvature(self,current_t)
                current_t = current_t + step_size
                if (current_t .gt. self%t(last_wp)) exit
            enddo
        endif
        total_curvature = total_curvature * step_size
    end function filament_eval_total_curvature

    !>  \brief  Evaluate the local curvature
    function filament_eval_curvature(self,t) result(curvature)
        use Splines
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        real(kind=8),   intent(in)      ::  t
        ! Result
        real(kind=8)    ::  curvature
        ! Private variables
        real(kind=8)    ::  x,y,z
        real(kind=8)    ::  dx,dy,dz
        real(kind=8)    ::  ddx,ddy,ddz
		logical,	parameter	::	assume_arc_length_parametrisation = .false.
        ! Start work

        if (self%nwp .ge. 3) then

            call self%x_spline%HornerEvaluateAll(t,x,dx,ddx)
            call self%y_spline%HornerEvaluateAll(t,y,dy,ddy)
            call self%z_spline%HornerEvaluateAll(t,z,dz,ddz)

            if (assume_arc_length_parametrisation) then
                curvature = sqrt(ddx**2+ddy**2+ddz**2)
            else
                curvature = sqrt((ddz*dy-ddy*dz)**2+(ddx*dz-ddz*dx)**2+(ddy*dx-ddx*dy)**2)  &
                            /   &
                            !(dx**2+dy**2+dz**2)**(3/2)
                            sqrt((dx**2+dy**2+dz**2))**3
            endif
		else
            curvature = 0.0d0
		endif

    end function filament_eval_curvature

    !>  \brief  Evaluate the curvature vector, which should be normal to the tangent vector.
    !!
    !!  With an ideal arc-length parametrised space curve, this would just be the second derivative vector.
    !!  However, small errors creep in when the space curve is not exactly arc-length parametrised. This aims to correct for those errors.
    !!  See notes 131107 for derivation
    !!
    !!  \todo make more efficient
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine filament_eval_curvature_vector(self,t,dx,dy,dz,ddx,ddy,ddz,kx,ky,kz)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        real(kind=8),   intent(in)      ::  t
        real(kind=8),   intent(in)      ::  dx,dy,dz    !<  Tangent (first derivative) vector
        real(kind=8),   intent(in)      ::  ddx,ddy,ddz !<  Second derivative vector
        real(kind=8),   intent(out)     ::  kx,ky,kz    !<  Curvature vector
        ! Private variables
        real(kind=8)    ::  t_norm
        real(kind=8)    ::  txdt_x,txdt_y,txdt_z,txdt_norm
        !real(KIND=8)    ::  tdt
        real(kind=8)    ::  curv
        logical,parameter   ::  sanity_check = .false.
        ! Start work

        if (self%nwp .ge. 3) then

            ! The norm of the first and second derivatives
            t_norm    = dsqrt(dx**2 +dy**2 +dz**2)
            t_norm    = 1.0d0 / t_norm**3 ! we will need to divide by t_norm
            !dt_norm   = sqrt(DDX**2+DDY**2+DDZ**2)

            ! Cross product of first and second derivative vectors
            txdt_x = dy*ddz-dz*ddy
            txdt_y = dz*ddx-dx*ddz
            txdt_z = dx*ddy-dy*ddx
            if (sanity_check) txdt_norm = dsqrt(txdt_x**2 + txdt_y**2 + txdt_z**2)

            ! Curvature vector (alt)
            kx = txdt_x * t_norm
            ky = txdt_y * t_norm
            kz = txdt_z * t_norm

#ifndef skip_runtime_checks
            ! Sanity check - |K| should be equal to curvature
            if (sanity_check) then
                curv      = filament_eval_curvature(self,t)
                if (abs(dsqrt(kx**2+ky**2+kz**2)-curv) .gt. 0.0001) then
                    write(*,'(a)') '**ERROR(filament_eval_curvature_vector): norm of output not as expected'
                    print *, 'TXDT_X,y,z              = ', txdt_x, txdt_y, txdt_z
                    !print *, 'T_NORM                  = ', t_norm
                    print *, '1/t_norm**3             = ', t_norm**3
                    print *, 'sqrt(kx**2+ky**2+kz**2) = ', dsqrt(kx**2+ky**2+kz**2)
                    print *, 'sqrt(txdt_x**2+txdt_y**2+txdt_z**2) / t_norm**3'
                    print *, '                        = ', dsqrt(txdt_x**2+txdt_y**2+txdt_z**2) / t_norm**3
                    print *, 'straight from derivs    = ', dsqrt((dy*ddz-dz*ddy)**2+(dz*ddx-dx*ddz)**2+(dx*ddy-dy*ddx)**2) &
                                                            / dsqrt(dx**2 +dy**2 +dz**2)**3
                    print *, 'curv                    = ', curv
                    print *, 'diff                    = ', dsqrt(kx**2+ky**2+kz**2) - curv
                    call terminate('Failed sanity check in FILAMENT_EVAL_CURVATURE_VECTOR')
                endif
            endif
#endif
        else
            kx = 0.0d0
            ky = 0.0d0
            kz = 0.0d0
        endif

    end subroutine filament_eval_curvature_vector


    !>  \brief  Evaluate the two principal curvatures
    !!
    !!  Note that arc-length parametrisation is assumed
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine filament_eval_principal_curvatures(self,t,curvature_a,curvature_b)
        use Splines
        use rotation_matrices
        use differential_geometry, only : euler_from_derivatives_new
        use runtime_parameters, only : terminate
        use float_comparisons, only : doubles_are_equal
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self        !<  Filament object
        real(kind=8),   intent(in)      ::  t           !<  Arc length at which to evaluate the curvature
        real(kind=8),   intent(out)     ::  curvature_a !<  Curvature along the first principal axis
        real(kind=8),   intent(out)     ::  curvature_b !<  Curvature along the second principal axis
        ! Private variables
        logical,    parameter   ::  sanity_check   = .false.
        logical,    parameter   ::  debug = .false.
        real(kind=8)            ::  x,y,z, dx,dy,dz, ddx,ddy,ddz, phi,theta,psi
!        real(KIND=8)            ::  cdx,cdy,cdz,cddx,cddy,cddz
        type(rotation_matrix)   ::  matrix
        real(kind=8)            ::  coos_in_rec(4), coos_in_lab(4)
        real(kind=8)            ::  curv
        real(kind=8)            ::  d_dd_dot_product, d_dd_dot_product_xy
        real(kind=8)            ::  kx,ky,kz
        ! Start work

!        if (self%fragmented_phi) then
!            write(*,'(a)') '**ERROR(filament_eval_principal_curvatures): only continuous description implemented for now.'
!            call terminate('Fatal error FILAMENT_EVAL_PRINCIPAL_CURVATURES. Feature not yet implemented.')
!        endif

        if (self%nwp .ge. 3) then

            ! Let's get the derivatives
            call self%x_spline%HornerEvaluateAll(t,x,dx,ddx)
            call self%y_spline%HornerEvaluateAll(t,y,dy,ddy)
            call self%z_spline%HornerEvaluateAll(t,z,dz,ddz)


            ! Let's get the curvature vector k
            call filament_eval_curvature_vector(self,t,dx,dy,dz,ddx,ddy,ddz,kx,ky,kz)


#ifndef skip_runtime_checks
            ! First sanity check
            ! Let's check that the k and d vectors are orthgonal
            if (sanity_check) then
                d_dd_dot_product    = (dx*kx + dy*ky + dz*kz)
                d_dd_dot_product_xy = (dx*kx + dy*ky        )
                if (abs(d_dd_dot_product) .gt. 0.00001) then
                    write(*,'(2a,G10.3)') '**ERROR(filament_eval_principal_curvatures): d and k vectors are not orthogonal. ', &
                                            'Dot product = ', d_dd_dot_product
                    write(*,'(a,G10.3)')  '**ERROR(filament_eval_principal_curvatures): angle = ', &
                                            degrees(acos(d_dd_dot_product/curv))
                    write(*,'(a,G10.3)')  '**ERROR(filament_eval_principal_curvatures): angle (xy only) = ', &
                                            degrees(acos(d_dd_dot_product_xy/curv))
                endif
            endif
#endif

            ! Let's get the rotation matrix to go from the lab frame to the filament reconstruction frame
            call euler_from_derivatives_new(dx,dy,dz,self%wp(2)%euler%theta .lt. 180.0D0,theta,psi)
            call self%twist_spline%Evaluate(t,phi)
            matrix%rmatrix  =   euler_to_rotmat(phi,theta,psi,reverse=.true.)

            ! Affine-transform the curvature (second derivative) vector to the filament frame of reference
            coos_in_lab     =   (/kx,ky,kz,1.0D0/)
            coos_in_rec     =   matmul(matrix%rmatrix,coos_in_lab)

            ! Now we have our result
            curvature_a     =   coos_in_rec(1)
            curvature_b     =   coos_in_rec(2)


#ifndef skip_runtime_checks
            if (sanity_check) then
                curv = filament_eval_curvature(self,t)
                if (debug) print '(a,G10.3)', '**debug(filament_eval_principal_curvatures): curvature = ', curv
                if (abs(coos_in_rec(3)) .gt. 0.1D0 * curv) then
                    write(*,'(a,G10.3)') '**ERROR(filament_eval_principal_curvatures): coos_in_rec(3) should be 0.0, but it is ', &
                                        coos_in_rec(3)
                    call terminate('Failed sanity check in FIALMENT_EVAL_PRINCIPAL_CURVATURES')
                endif
                if (abs(sqrt(curvature_a**2+curvature_b**2) - curv) .gt. 0.01*curv) then
                !if (.NOT. doubles_are_equal(sqrt(CURVATURE_A**2+CURVATURE_B**2),curv)) then
                    write(*,'(a)') '**ERROR(filament_eval_principal_curvatures): failed sanity check:'
                    write(*,'(a,G10.3)') 'curvature_a                         = ', curvature_a
                    write(*,'(a,G10.3)') 'curvature_b                         = ', curvature_b
                    write(*,'(a,G10.3)') 'curvature_c (should be 0.0)         = ', coos_in_rec(3)
                    write(*,'(a,G10.3)') 'sqrt(curvature_a**2+curvature_b**2) = ', sqrt(curvature_a**2+curvature_b**2)
                    write(*,'(a,g10.3)') 'curvature                           = ', curv
                    call terminate('Fatal error in FILAMENT_EVAL_PRINCIPAL_CURVATURES')
                endif
            endif
#endif
        else
            curvature_a = 0.0d0
            curvature_b = 0.0d0
        endif

    end subroutine filament_eval_principal_curvatures


    !>  \brief  Evaluate the total (integral) squared axial curvature, helical overtwist and stretching along a filament,
    !!          in reciprocal pixels.
    subroutine filament_eval_total_curvature_overtwist_stretch(self,curvature,overtwist,stretch,ignore_fal_wps)
        use runtime_parameters, only : dpi
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self
        real(kind=8),               intent(out)     ::  curvature           !<  Total squared curvature integral (1/pixel)
        real(kind=8),               intent(out)     ::  overtwist           !<  Total squared overtwist integral (1/pixel)
        real(kind=8),               intent(out)     ::  stretch             !<  Total squared stretch integral (1/pixel)
        logical,        optional,   intent(in)      ::  ignore_fal_wps      !<  Whether to ignore the first and last waypoints in the calculation. False by default.
        ! Result
        ! Private variables
        real(kind=8)    ::  step_size
        real(kind=8)    ::  current_t
        logical         ::  iignore_fal_wps
        integer         ::  first_wp, last_wp
        real(kind=8)    ::  average_dphi    !<  Average change in phi per unit path length
        real(kind=8)    ::  average_ds      !<  Average change in subunit_num per unit path length
        ! Start work


        ! Ignore fal WPS?
        if (present(ignore_fal_wps)) then
            iignore_fal_wps = ignore_fal_wps
        else
            iignore_fal_wps = .not. self%fal_wps_in_score
        endif

        ! First and last waypoints
        if (iignore_fal_wps) then
            first_wp    =   2
            last_wp     =   self%nwp - 1
        else
            first_wp    =   1
            last_wp     =   self%nwp
        endif

        if (last_wp .le. first_wp) then
            curvature = 0.0D0
            overtwist = 0.0D0
            stretch   = 0.0D0
            write(*,'(a)') '**warning(filament_eval_total_curvature_overtwist_stretch): cannot compute. filament too short? '
        else
        ! The step size for the integration
        if (self%helix%pixel_units) then
            step_size = max(self%helix%rise,1.0D0)
        else
            step_size = max(self%helix%rise/self%film%GetPixelSize(),1.0D0)
        endif

        ! The average change in phi per pixel along the filament
        !average_dphi = (self%wp(last_wp)%euler%phi - self%wp(first_wp)%euler%phi) / (self%t(last_wp) - self%t(first_wp))

        ! The average change in subunit_num per pixel along the filament
        !average_ds = (self%wp(last_wp)%subunit_num - self%wp(first_wp)%subunit_num) / (self%t(last_wp) - self%t(first_wp))

        call filament_eval_average_dphi_ds(self,average_dphi,average_ds,ignore_fal_wps=iignore_fal_wps)

        ! Total squared torsion
        curvature = 0.0D0
        overtwist = 0.0D0
        stretch   = 0.0D0

        current_t = self%t(first_wp)
        do
            overtwist = overtwist + filament_eval_overtwist(self,current_t,average_dphi)**2
            curvature = curvature + filament_eval_curvature(self,current_t)**2
            stretch   = stretch   + filament_eval_stretch(self,current_t,average_ds)**2
            current_t = current_t + step_size
            if (current_t .gt. self%t(last_wp)) exit
        enddo
        ! Convert to radians
        overtwist = overtwist / 180.0D0**2 * DPI**2
        ! The integral
        curvature = curvature * step_size
        overtwist = overtwist * step_size
        stretch   = stretch   * step_size
        endif
    end subroutine filament_eval_total_curvature_overtwist_stretch

    !>  \brief  Evaluate the local torsion at some point along the filament
    !!
    !!  Actually, we're evaluating the over/undertwist of the helical assembly (not the axial space curve's torsion).
    !!  The result is returned in degrees.
    pure function filament_eval_overtwist(self,t,average_twist_per_distance) result(torsion)
        use Splines
        ! Arguments
        type(filament), intent(in)  ::  self
        real(kind=8),   intent(in)  ::  t
        real(kind=8),   intent(in)  ::  average_twist_per_distance  !<  Average change in phi per pixel along the filament
        ! Result
        real(kind=8)    ::  torsion
        ! Private variables

        ! Start work
        if (self%nwp .ge. 3) then
            torsion = self%twist_spline%EvaluateDerivative(t) - average_twist_per_distance
        else
            torsion = 0.0d0
        endif
    end function filament_eval_overtwist

    pure function filament_eval_stretch(self,t,average_ds_per_distance) result(stretch)
        use Splines
        ! Arguments
        type(filament), intent(in)  ::  self
        real(kind=8),   intent(in)  ::  t
        real(kind=8),   intent(in)  ::  average_ds_per_distance     !<  Average change in subunit_num per unit distance along filament
        ! Result
        real(kind=8)    ::  stretch
        ! Private variables

        ! Start work
        if (self%nwp .ge. 3) then
            stretch = self%s_spline%EvaluateDerivative(t) - average_ds_per_distance
        else
            stretch = 0.0d0
        endif
    end function filament_eval_stretch


    !>  \brief  Evaluate the local helical twist at a point along the filament
    pure real function filament_eval_helical_twist(self,t,rise) result(twist)
        use Splines
        ! Arguments
        type(filament),             intent(in)  ::  self
        real(kind=8),               intent(in)  ::  t
        real(kind=8),   optional,   intent(in)  ::  rise
        ! Private variables

        ! Start work
        if (self%twist_spline%IsAllocated()) then
            if (present(rise)) then
                twist = self%twist_spline%EvaluateDerivative(t) * rise
            else
                twist = self%twist_spline%EvaluateDerivative(t) / self%s_spline%EvaluateDerivative(t)
            endif
        else
            twist = 0.0d0
        endif
    end function filament_eval_helical_twist

    !> \brief   Evaluate the local helical rise at a point along the filament
    pure real function filament_eval_helical_rise(self,t) result(rise)
        use Splines
        implicit none
        ! Arguments
        type(filament), intent(in)  ::  self
        real(kind=8),   intent(in)  ::  t
        ! Private variables

        ! Start work
        if (self%s_spline%IsAllocated()) then
            rise = 1. / self%s_spline%EvaluateDerivative(t)
        else
            rise = 0.0d0
        endif
    end function filament_eval_helical_rise

    !>  \brief  Compute the path length from first waypoint to the point specified by the coordinate T.
    !!
    !!  Should only be called when t falls directly after a waypoint for which the path_length has already been determined
#if defined (skip_runtime_checks)
    pure &
#endif
    subroutine filament_eval_path_length(self,t,path_length)
        use Splines
        use runtime_parameters, only : terminate
#ifndef __GFORTRAN__
        use ieee_arithmetic
#endif
        implicit none
        ! Arguments
        type(filament),         intent(inout)   ::  self            !<  Filament
        real(kind=8),           intent(in)      ::  t               !<  Parameter value for which the path length needs to be computed
        real(kind=8),           intent(inout)   ::  path_length     !<  Result. Arc length in pixels.
        ! Private variables
#ifndef use_quadrature
        real(kind=8),           parameter       ::  tiny_step = 0.0005D0    !<  Step for the integration (pixels)
        integer                                 ::  seg
        real(kind=8)                            ::  current_t,current_x,current_y,current_z,previous_x,previous_y,previous_z
        logical,                parameter   ::  debug   =   .false.
#endif
        ! Start work
        if (self%nwp .ge. 3) then
#ifdef use_quadrature
            !
            call filament_eval_arc_length_by_quadrature(self,t,path_length)
#else
            ! Figure out which segment of the spline we'll be working on
            seg = self%x_spline%FindSegment(t)
#ifndef skip_runtime_checks
            if (debug) then
            if (seg .eq. 0) then
                write(*,'(a,F0.5)') '**ERROR(filament_eval_path_length): could not find spline segment for t = ', t
                call self%x_spline%PrintInfo()
                call terminate('Fatal error in FILAMENT_EVAL_PATH_LENGTH')
            endif
#endif
            endif
            ! Start at the start of that segment
            path_length = self%wp(seg)%path_length
            current_t = self%t(seg)
            current_x = self%x_spline%y(seg)
            current_y = self%y_spline%y(seg)
            current_z = self%z_spline%y(seg)

            ! Loop until we've reached the point we're interested in
            do
                ! Keep track of where we've just been
                previous_x = current_x
                previous_y = current_y
                previous_z = current_z
                ! One tiny step for man...
                current_t = current_t + tiny_step
                ! Check we've not stepped over the mark
                if (current_t .gt. t) then
                    exit    !   Done, get out
                endif
                ! We're still walking - work out where our foot lands
                call self%x_spline%Evaluate(current_t,current_x)   ! May 2014: why can't we pass the seg to the eval routine? Wouldn't that save a lot of time?
                call self%y_spline%Evaluate(current_t,current_y)
                call self%z_spline%Evaluate(current_t,current_z)
                ! Measure & accumulate the distance
                path_length = path_length + sqrt((current_x-previous_x)**2+(current_y-previous_y)**2+(current_z-previous_z)**2)
            enddo
#endif
        else if (self%nwp .eq. 2) then
            path_length = self%wp(1)%path_length + (t-self%t(1))/(self%t(2)-self%t(1))&
                                                    *sqrt(    (self%wp(2)%coo(1)-self%wp(1)%coo(1))**2 &
                                                            + (self%wp(2)%coo(2)-self%wp(1)%coo(2))**2 &
                                                            + (self%wp(2)%coo(3)-self%wp(1)%coo(3))**2 )
        else
            ! Only one waypoint
            path_length = self%wp(1)%path_length + (t-self%t(1))
#ifndef skip_runtime_checks
            call terminate('Filament_eval_path_length was called on a filament with 1 wp')
#endif
        endif
    end subroutine filament_eval_path_length

    !>  \brief  Compute arc length from first waypoint to the point specified by parameter t
    !!  Should only be called when t falls directly after a waypoint for which the arc length has already been computed
    pure subroutine filament_eval_arc_length_by_quadrature(self,t,arc_length)
        type(filament),                     intent(in)      ::  self
        real(kind=8),                       intent(in)      ::  t
        real(kind=8),                       intent(out)     ::  arc_length
        !
        integer ::  seg
        !
        seg = self%x_spline%FindSegment(t)
        arc_length = self%gl_quadrature%ArcLengthIntegral( &
                        3.0*self%x_spline%w(seg,1),2.0*self%x_spline%w(seg,2),self%x_spline%w(seg,3), &
                        3.0*self%y_spline%w(seg,1),2.0*self%y_spline%w(seg,2),self%y_spline%w(seg,3), &
                        3.0*self%z_spline%w(seg,1),2.0*self%z_spline%w(seg,2),self%z_spline%w(seg,3), &
                        0.0d0,t-self%t(seg))

        arc_length = arc_length + self%wp(seg)%path_length
    end subroutine

    !>  \brief  Return the nearest waypoint for the input arc length
    pure subroutine nearest_wp_from_t(self,t,nearest_wp,dist)
        implicit none
        ! Arguments
        type(filament),                     intent(in)      ::  self
        real(kind=8),                       intent(in)      ::  t
        integer,                            intent(out)     ::  nearest_wp  !<  Index of nearest waypoint
        real(kind=8),                       intent(out)     ::  dist        !<  (Signed) distance to nearest waypoint
        ! Private variables
        integer ::  itmp(1)
        real(kind=8) :: distances(self%nwp)
        ! Start work
        distances = t - self%t
        itmp = minloc(abs(distances))
        nearest_wp = itmp(1)
        dist = t - self%t(nearest_wp)
    end subroutine nearest_wp_from_t


    !>  \brief  Return the nearest previous waypoint for the input path length
    integer function nearest_previous_wp_from_t(self,t) result(previous_wp)
        use Splines
        use Globals
        implicit none
        ! Arguments
        type(filament),                     intent(in)      ::  self
        real(kind=8),                       intent(in)      ::  t
        ! Private variables

        ! Start work
        if (self%nwp .ge. 3) then
            previous_wp = self%x_spline%FindSegment(t)
            if(previous_wp .eq. 0) then
                if (t .lt. self%x_spline%x(1) .and. abs(t-self%x_spline%x(1)) .lt. 0.01) then
                    previous_wp = 1
                else
                    write(*,'(a,F0.3,a)') '**WARNING(nearest_previous_wp_from_t): could not find '//&
                                            'segment within which input t value (', &
                                            t, ') falls. Maybe the splines are not set up properly?'
                    call self%x_spline%PrintInfo()
                    !call this_program%TerminateWithFatalError('nearest_previous_wp_from_t','debug terminate')
                endif
            endif
        else
            previous_wp = 1
        endif
    end function nearest_previous_wp_from_t

    !>  \brief  Check that a waypoint has a set of Euler angles which are consistent with the general direction of the filament, as defined
    !!          by the waypoint-to-waypoint path
    logical function filament_wp_euler_are_consistent_with_general_direction(self,wp_number) result(they_are_consistent)
        use differential_geometry, only : d_from_theta_psi, euler_from_derivatives_new, check_euler_and_derivatives_new
        use UsefulFunctions, only : DegreesBetween0And360
        use runtime_parameters, only : pi
        ! arguments
        type(filament),     intent(in)  ::  self
        integer,            intent(in)  ::  wp_number
        ! private variables
        real(kind=8)        ::  d_according_to_euler(3) ! tangent vector at that point
        !real        ::  d_according_to_euler_check(3)
        !real        ::  d_according_to_euler_sign(3)
        !real        ::  d_according_to_euler_check_sign(3)
        !real(kind=8)::  theta_tmp, psi_tmp
        real        ::  d_according_to_coos(3) ! approximate direction of filament
        integer     ::  wp_start, wp_finish
        real        ::  angle
        ! start work

        ! The tangent at that waypoint
        d_according_to_euler = d_from_theta_psi(self%wp(wp_number)%euler%theta,self%wp(wp_number)%euler%psi)

        call check_euler_and_derivatives_new(DegreesBetween0And360(self%wp(wp_number)%euler%psi), &
                                             DegreesBetween0And360(self%wp(wp_number)%euler%theta), &
                                             d_according_to_euler(1),d_according_to_euler(2),d_according_to_euler(3))

        ! The approximate direction of the filament
        if (wp_number .lt. self%nwp) then
            wp_start  = wp_number
            wp_finish = wp_number + 1
        else if (wp_number .eq. self%nwp) then
            wp_start  = self%nwp -1
            wp_finish = self%nwp
        endif
        d_according_to_coos  = self%wp(wp_finish)%coo - self%wp(wp_start)%coo


        ! The angle between these two vectors should be <90.0
        angle = sum(d_according_to_coos * d_according_to_euler) &
                         /sqrt(sum(d_according_to_coos**2)) &
                         /sqrt(sum(d_according_to_euler**2))
        angle = min(angle,  1.0e0)
        angle = max(angle, -1.0e0)
        angle = acos(angle)
        they_are_consistent = abs(angle) .lt. pi*0.5

    end function filament_wp_euler_are_consistent_with_general_direction


    !>  \brief  Check that a waypoint has a set of Euler angles which are consistent with differential_geometry routines & conventions of Frealix
    logical function filament_wp_euler_are_consistent_with_differential_geometry(self,wp_number) result(they_are_consistent)
        use differential_geometry, only : d_from_theta_psi, euler_from_derivatives_new, check_euler_and_derivatives_new
        use UsefulFunctions, only : DegreesBetween0And360
        use runtime_parameters, only : pi
        ! arguments
        type(filament),     intent(in)  ::  self
        integer,            intent(in)  ::  wp_number
        ! private variables
        real(kind=8)        ::  d_according_to_euler(3) ! tangent vector at that point
        real        ::  d_according_to_euler_check(3)
        real        ::  d_according_to_euler_sign(3)
        real        ::  d_according_to_euler_check_sign(3)
        real(kind=8)::  theta_tmp, psi_tmp
        ! start work

        ! The tangent at that waypoint
        d_according_to_euler = d_from_theta_psi(self%wp(wp_number)%euler%theta,self%wp(wp_number)%euler%psi)

        ! we would also like to be able to go back and forth between Euler angles and derivatives without trouble
        call euler_from_derivatives_new(d_according_to_euler(1),d_according_to_euler(2),d_according_to_euler(3), &
                                        self%wp(wp_number)%euler%theta .lt. 180.0d0,theta_tmp,psi_tmp)
        d_according_to_euler_check = d_from_theta_psi(theta_tmp,psi_tmp)

        ! Check the signs of the derivatives
        d_according_to_euler_sign       = sign(1.0d0,d_according_to_euler)
        d_according_to_euler_check_sign = sign(1.0e0,d_according_to_euler_check)


        !
        they_are_consistent = sum(d_according_to_euler_sign*d_according_to_euler_check_sign) .gt. 2.9

    end function filament_wp_euler_are_consistent_with_differential_geometry



    !>  \brief  Returns the straight-line distance in pixels from waypoint wp_num to the nearest waypoint
    real function filament_wp_distance_to_nearest_wp(fil,wp_num)
        use coos, only : waypoint_waypoint_distance
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament),     intent(in)  ::  fil     !<  Filament
        integer,            intent(in)  ::  wp_num  !<  The number of the waypoint we're interested in
        ! Private variables
        integer ::  wp_before, wp_after, i
        real    ::  tmp_dist
        real,   parameter   ::  large_value = 999999.99
        ! Start work

        ! Define the two neighbouring waypoints
        if (wp_num .eq. 1) then
            wp_before = 2
            wp_after = 2
        else if (wp_num .eq. fil%nwp) then
            wp_before = fil%nwp - 1
            wp_after = fil%nwp - 1
        else if (wp_num .gt. 1 .and. wp_num .lt. fil%nwp) then
            wp_before = WP_NUM-1
            wp_after = WP_NUM+1
        else
            write(*,'(3(A,i3),a)') '**ERROR(filament_wp_distance_to_nearest_wp): waypoint ', wp_num, &
                                    ' does not exist (filament ', fil%id, ' has ', fil%nwp, ' waypoints).'
            call terminate('Fatal error in FILAMENT_WP_DISTANCE_TO_NEAREST_WP')
        endif

        ! Loop over the neighbours to find the nearest one
        filament_wp_distance_to_nearest_wp = large_value
        do i = wp_before,wp_after,2
            tmp_dist = waypoint_waypoint_distance(fil%wp(wp_num),fil%wp(i))
            if (tmp_dist .lt. filament_wp_distance_to_nearest_wp) then
                filament_wp_distance_to_nearest_wp = tmp_dist
            endif
        enddo

        ! If we didn't find a value, crash out
        if (filament_wp_distance_to_nearest_wp .eq. large_value) then
            write(*,'(a,F12.2)') '**ERROR(filament_wo_distance_to_nearest_wp): did not find a distance-to-neighbour less than ', &
                                    large_value
            call terminate('Fatal error in FILAMENT_WP_DISTANCE_TO_NEAREST_WP')
        endif
    end function filament_wp_distance_to_nearest_wp

    !>  \brief  Compute the average defocus of the filament
    pure real function filament_average_defocus(self,ignore_fal_wps) result(average_defocus)
        implicit none
        ! Arguments
        type(filament),             intent(in)  ::  self            !<  Filament object
        logical,       optional,    intent(in)  ::  ignore_fal_wps  !<  Whether to ignore the first and last waypoints in the calculation. False by default.
        ! Private variables
        logical ::  iignore_fal_wps
        integer ::  first_wp, last_wp, current_wp
        ! Start work

        ! Ignore fal WPS?
        iignore_fal_wps = .false.
        if (present(ignore_fal_wps)) iignore_fal_wps = ignore_fal_wps

        ! First and last waypoints
        if (iignore_fal_wps) then
            first_wp    =   2
            last_wp     =   self%nwp - 1
        else
            first_wp    =   1
            last_wp     =   self%nwp
        endif

        average_defocus = 0.0
        do current_wp=first_wp,last_wp
            average_defocus = average_defocus + self%wp(current_wp)%coo(3)
        enddo

        average_defocus = average_defocus / real(last_wp-first_wp+1)

    end function filament_average_defocus

    !>  \brief  Return the difference between the min and max z (defocus) of waypoints in the filament
    pure real function filament_defocus_range(self,ignore_fal_wps) result(defocus_range)
        implicit none
        ! Arguments
        type(filament),             intent(in)  ::  self            !<  Filament object
        logical,       optional,    intent(in)  ::  ignore_fal_wps  !<  Whether to ignore the first and last waypoints in the calculation. False by default.
        ! Private variables
        real    ::  defocus_min, defocus_max
        logical ::  iignore_fal_wps
        integer ::  first_wp, last_wp, current_wp
        ! Start work

        ! Ignore fal WPS?
        if (present(ignore_fal_wps)) then
            iignore_fal_wps = ignore_fal_wps
        else
            iignore_fal_wps = .false.
        endif

        ! First and last waypoints
        if (iignore_fal_wps) then
            first_wp    =   2
            last_wp     =   self%nwp - 1
        else
            first_wp    =   1
            last_wp     =   self%nwp
        endif

        !
        defocus_min =  9999999.99
        defocus_max = -9999999.99
        do CURRENT_WP=FIRST_WP,last_wp
            if (self%wp(current_wp)%coo(3) .gt. defocus_max) defocus_max = self%wp(current_wp)%coo(3)
            if (self%wp(current_wp)%coo(3) .lt. defocus_min) defocus_min = self%wp(current_wp)%coo(3)
        enddo
        defocus_range = abs(DEFOCUS_MAX-DEFOCUS_MIN)
    end function filament_defocus_range

    !>  \brief  Return the average waypoint-to-waypoint distance (along the path of the filament)
    function filament_average_wp_to_wp_distance(self,ignore_fal_wps) result(dist_ave)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament),             intent(in)  ::  self            !<  Filament object
        logical,        optional,   intent(in)  ::  ignore_fal_wps  !<  Whether to ignore the first and last waypoints in the calculation. False by default.
        ! Result
        real    ::  dist_ave    !<  Average distance from one waypoint to the next (pixels)
        ! Private variables
        logical ::  iignore_fal_wps
        integer ::  first_wp, last_wp
        ! Start work

        ! Ignore fal WPS?
        if (present(ignore_fal_wps)) then
            iignore_fal_wps = ignore_fal_wps
        else
            iignore_fal_wps = .false.
        endif

        ! First and last waypoints
        if (iignore_fal_wps) then
            first_wp    =   2
            last_wp     =   self%nwp - 1
        else
            first_wp    =   1
            last_wp     =   self%nwp
        endif

        ! Compute the average distance
        if ((iignore_fal_wps .and. self%nwp .ge. 4) .or. (.not. iignore_fal_wps .and. self%nwp .ge. 2)) then
            dist_ave = self%wp(last_wp)%path_length - self%wp(first_wp)%path_length
            dist_ave = dist_ave / (last_wp-first_wp)
        else
            dist_ave = 0.0
        endif
    end function filament_average_wp_to_wp_distance

        !>  \brief  Given a filament with Theta and Psi Euler angles set for each waypoint, compute the spline derivatives at each point
    pure subroutine filament_derivatives_from_euler(self)
        use differential_geometry, only : dx_from_theta_psi, dy_from_theta_psi, dz_from_theta
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self    !<  Filament object
        ! Private variables
        integer ::  i
        ! Start work
        forall(i=1:self%nwp)
            self%x_spline%dy(i) = dx_from_theta_psi(self%wp(i)%euler%theta,self%wp(i)%euler%psi)
            self%y_spline%dy(i) = dy_from_theta_psi(self%wp(i)%euler%theta,self%wp(i)%euler%psi)
            self%z_spline%dy(i) = dz_from_theta(self%wp(i)%euler%theta)
        end forall
    end subroutine filament_derivatives_from_euler



    !>  \brief  Given a filament with solved x,y and z splines, compute the Theta and Psi Euler angles at each waypoint
    !!
    !!  See the "Spline Parametrisation" document for derivations of the equations
    subroutine filament_euler_from_derivatives(self,known_psi,known_theta)
        use runtime_parameters, only : terminate
        use differential_geometry, only : euler_from_derivatives_new
        implicit none
        ! Arguments
        type(filament),         intent(inout)   ::  self        !<  Filament object
        logical,    optional,   intent(in)      ::  known_psi   !<  If true, we won't recalculate psi
        logical,    optional,   intent(in)      ::  known_theta !<  If true, we won't recalculate theta
        ! Private variables
        logical ::  kknown_psi, kknown_theta
        integer ::  i
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        ! Defaults are false
        if (present(known_theta)) then
            kknown_theta = known_theta
        else
            kknown_theta = .false.
        endif
        if (present(known_psi)) then
            kknown_psi = known_psi
        else
            kknown_psi = .false.
        endif

        ! If we already know the angles, why are we even calling this subroutine?
        if (kknown_psi .and. kknown_theta) then
            write(*,'(2a)') '**ERROR(filament_euler_from_derivatives): psi and theta are already known, ',   &
                            'no need to call this subroutine'
            call terminate('Programming error')
        endif

        if (kknown_psi .or. kknown_theta) then
            call terminate('Not coded yet - is this needed?')
        endif

        !
        ! Loop over waypoints
        !
        do I=1,self%nwp
            call euler_from_derivatives_new(    self%x_spline%dy(i),        &
                                                self%y_spline%dy(i),        &
                                                self%z_spline%dy(i),        &
                                                self%wp(2)%euler%theta .lt. 180.0D0, &
                                                self%wp(i)%euler%theta,     &
                                                self%wp(i)%euler%psi    )
        enddo

    end subroutine filament_euler_from_derivatives

    !>  \brief  Ensure that psi and theta are as smooth as possible
    !!          We assume that theta is always positive
    pure subroutine filament_psi_theta_smooth(self)
        type(filament), intent(inout)   ::  self
        call waypoints_psi_theta_smooth(self%wp)
    end subroutine filament_psi_theta_smooth

    !>  \brief  Ensure that psi and theta are as smooth as possible
    !!          We assume that theta is always positive
    pure subroutine waypoints_psi_theta_smooth(self)
        use coos, only : waypoint
        type(waypoint), intent(inout)   ::  self(:)
        !
        logical         ::  first_theta_near_90
        logical         ::  current_theta_near_90
        integer         ::  current_wp
        real(kind=8)    ::  sign_of_theta_to_add, sign_of_psi_to_add, sign_of_phi_to_add
        logical, parameter  ::  allowed_to_flip_polarity = .false.
        ! Start work

        if (allowed_to_flip_polarity) then
            ! are we near 90 or near 270?
            first_theta_near_90 = mod(self(1)%euler%theta,360.0d0) .lt. 180.0d0
            do current_wp=2,size(self)
                current_theta_near_90 = mod(self(current_wp)%euler%theta,360.0d0) .lt. 180.0d0
                if ((first_theta_near_90 .and. .not. current_theta_near_90) .or. &
                    (.not. first_theta_near_90 .and. current_theta_near_90)) then
                    if (first_theta_near_90) then
                        sign_of_theta_to_add = -1.0d0
                    else
                        sign_of_theta_to_add = 1.0d0
                    endif
                    if (self(current_wp)%euler%psi .gt. self(current_wp-1)%euler%psi) then
                        sign_of_psi_to_add = -1.0d0
                    else
                        sign_of_psi_to_add = 1.0d0
                    endif
                    if (self(current_wp)%euler%phi .gt. self(current_wp-1)%euler%phi) then
                        sign_of_phi_to_add = -1.0d0
                    else
                        sign_of_phi_to_add = 1.0d0
                    endif
                    self(current_wp)%euler%theta = self(current_wp)%euler%theta + 180.0d0 * sign_of_theta_to_add
                    self(current_wp)%euler%psi   = self(current_wp)%euler%psi + 180.0d0 * sign_of_psi_to_add
                    self(current_wp)%euler%phi   = self(current_wp)%euler%phi + 180.0d0 * sign_of_phi_to_add
                endif
            enddo
        else
            ! we're not allowed to flip polarity
            ! the best we can do is add/subtract i*360.0
            do current_wp=2,size(self)
                ! not implemented yet
                self(current_wp)%euler%theta = self(current_wp)%euler%theta - &
                                    360.0d0*int((self(current_wp)%euler%theta - self(current_wp-1)%euler%theta)/360.0d0)
                self(current_wp)%euler%psi   = self(current_wp)%euler%psi - &
                                    360.0d0*int((self(current_wp)%euler%psi - self(current_wp-1)%euler%psi)/360.0d0)
            enddo
        endif
    end subroutine waypoints_psi_theta_smooth

    !>  \brief  Ensure that psi and theta are as smooth as possible, even if it
    !!          means breaking the convention of -90<=psi<=90
    pure subroutine filament_psi_theta_smooth_old(self)
        use euleranglesmod, only : degrees_m180_to_180
        implicit none
        ! Arguments
        type(filament), intent(inout)      ::  self
        ! Private variables
        integer ::  current_wp
        logical ::  keep_theta_positive, theta_positive
        ! Start work

        ! We will keep the sign of theta constant
        keep_theta_positive = (self%wp(1)%euler%theta .gt. 0.0)

        do CURRENT_WP=1,self%nwp
            theta_positive = (self%wp(current_wp)%euler%theta .gt. 0.0)
            if (    (theta_positive .and. .not. keep_theta_positive) &
                .or.(.not. theta_positive .and. keep_theta_positive) ) then
                ! We need to switch the sign of Theta
                self%wp(current_wp)%euler%theta = - self%wp(current_wp)%euler%theta
                self%wp(current_wp)%euler%psi   = degrees_m180_to_180(self%wp(current_wp)%euler%psi + 180.)
                self%wp(current_wp)%euler%phi   = self%wp(current_wp)%euler%phi + 180.
            endif
        enddo
    end subroutine filament_psi_theta_smooth_old

    !>  \brief  Ensure that psi and theta are as smooth as possible
    pure subroutine waypoints_psi_theta_smooth_old(wps)
        use euleranglesmod, only : degrees_m180_to_180
        use coos, only : waypoint
        implicit none
        ! Arguments
        type(waypoint), intent(inout)      ::  wps(:)
        ! Private variables
        integer ::  current_wp
        logical ::  keep_theta_positive, theta_positive
        ! Start work

        ! We will keep the sign of theta constant
        keep_theta_positive = (wps(1)%euler%theta .gt. 0.0)

        do CURRENT_WP=1,size(wps)
            theta_positive = (wps(current_wp)%euler%theta .gt. 0.0)
            if (    (theta_positive .and. .not. keep_theta_positive) &
                .or.(.not. theta_positive .and. keep_theta_positive) ) then
                ! We need to switch the sign of Theta
                wps(current_wp)%euler%theta = - wps(current_wp)%euler%theta
                wps(current_wp)%euler%psi   = degrees_m180_to_180(wps(current_wp)%euler%psi + 180.0d0)
                wps(current_wp)%euler%phi   = wps(current_wp)%euler%phi + 180.0 ! This is because we keep the twist angle (rather than the actual Euler angle) in "phi"
            endif
        enddo
    end subroutine waypoints_psi_theta_smooth_old

    !>  \brief  Force the theta angle to be positive & adjust the other Euler angles to be consistent
    !pure subroutine waypoints_theta_positive(self)
    !    use coos, only : waypoint
    !    type(waypoint), intent(inout)   ::  self(:)
    !    ! private variables
    !    integer     ::  current_wp
    !    ! start work
    !    do current_wp=1,size(self)
    !        if (self(current_wp)%euler%theta .lt. 0.0) then
    !            self(current_wp)%euler%theta = - self(current_wp)%euler%theta
    !            self(current_wp)%euler%psi   =   mod(self(current_wp)%euler%psi+180.0d0,360.0d0)
    !            self(current_wp)%euler%phi   =   self(current_wp)%euler%phi + 180.0d0
    !        endif
    !    enddo
    !end subroutine waypoints_theta_positive

    !>  \brief  Enforce convention of -90<=psi<=90 on all waypoints of the filament
    !!  \todo   Deprecate
    !pure subroutine filament_psi_btw_m90_90(self)
    !    use differential_geometry, only : psi_btw_m90_90_old
    !    use runtime_parameters, only : terminate
    !    implicit none
    !    ! Arguments
    !    type(filament), intent(inout)   ::  self
    !    ! Private variables
    !    integer ::  current_wp
    !    ! Start work
    !    do current_wp = 1, SELF%nwp
    !        call psi_btw_m90_90_old(self%wp(current_wp)%euler%theta,self%wp(current_wp)%euler%psi)
    !    enddo
    !end subroutine filament_psi_btw_m90_90





    !>  \brief  Solve the filament axis' splines, based on the current values for t
    subroutine filament_solve_axis_splines(self)
        use splines
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self
        ! Private variables
        logical,    parameter   ::  debug = .false.
        logical ::  derivatives_from_euler
        ! Start work

        ! Initialise the "X" values for the splines
        self%x_spline%x     = self%t
        self%y_spline%x     = self%t
        self%z_spline%x     = self%t

        ! Initialise the "Y" values for the splines
        self%x_spline%y     = self%wp%coo(1)
        self%y_spline%y     = self%wp%coo(2)
        self%z_spline%y     = self%wp%coo(3)

        ! If we know the Euler angles at waypoints, they can force the (x,y,z) splines' derivatives.
        derivatives_from_euler = self%use_hermite_splines .and. self%PsiAndThetaAreNonZero()
        if (derivatives_from_euler) then
            call filament_derivatives_from_euler(self)
            call self%x_spline%SetType('KNOWN_D')
            call self%y_spline%SetType('KNOWN_D')
            call self%z_spline%SetType('KNOWN_D')
        else
            call self%x_spline%SetType('PARABOLIC')
            call self%y_spline%SetType('PARABOLIC')
            call self%z_spline%SetType('PARABOLIC')
        endif

        ! Now we can actually solve the splines
        call self%x_spline%Solve()
        call self%y_spline%Solve()
        call self%z_spline%Solve()

        ! Update the Euler angles if needed
        if (.not. derivatives_from_euler) call filament_euler_from_derivatives(self)

        if (debug) then
            write(*,'(a)') '**debug(filament_solve_splines): x spline on exit'
            call self%x_spline%PrintInfo()
        endif
    end subroutine filament_solve_axis_splines

    !>  \brief  Solve the filament's helix's splines, based on the current values for t
    subroutine filament_solve_helix_splines(self)
        use splines
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self
        ! Private variables

        ! Start work


        ! Initialise the "X" values for the splines
        self%s_spline%x     = self%t
        self%twist_spline%x = self%t

        ! Initialise the "Y" values for the splines
        self%s_spline%y     = self%wp%subunit_num
        self%twist_spline%y = self%wp%euler%phi

        ! The phi & s splines are always natural, since we don't have a good way to measure local helical parameters
        ! Until 20130909, parabolic splines were used here
        call self%twist_spline%SetType('NATURAL')
        call self%s_spline%SetType('NATURAL')

        ! Now we can actually solve the splines
        call self%s_spline%Solve()
        call self%twist_spline%Solve()

    end subroutine filament_solve_helix_splines


    !>  \brief  Solve the filament's splines, based on the current values for t
    subroutine filament_solve_splines(self)
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self
        ! Private variables

        ! Start work

        ! Update the helical axis splines
        call filament_solve_axis_splines(self)

        ! Update the helical splines
        call filament_solve_helix_splines(self)

    end subroutine filament_solve_splines

    !>  \brief Check whether the phi & subunit_num spline segments are monotonic
    logical function filament_splines_are_monotonic(self)
        use Splines
        ! Arguments
        type(filament), intent(in)  ::  self
        ! Private variable

        ! Start work
        filament_splines_are_monotonic = .true.
        if (.not. self%fragmented_phi) then
            filament_splines_are_monotonic = filament_splines_are_monotonic .and. self%twist_spline%IsMonotonic()
        endif
        if (.not. self%fragmented_subunit_num) then
            filament_splines_are_monotonic = filament_splines_are_monotonic .and. self%s_spline%IsMonotonic()
        endif
    end function filament_splines_are_monotonic

    !>  \brief  Check whether waypoints' path lengths increase along the filament
    pure logical function filament_wp_path_lengths_increase(self)
        ! Arguments
        type(filament), intent(in)  ::  self
        ! Private variables
        integer ::  current_wp
        ! Start work
        filament_wp_path_lengths_increase = .true.
        do CURRENT_WP=2,self%nwp
            if (self%wp(current_wp)%path_length .le. self%wp(CURRENT_WP-1)%path_length) filament_wp_path_lengths_increase = .false.
        enddo
    end function filament_wp_path_lengths_increase

    !>  \brief  Evaluate deviations from canonical helical parameters
    subroutine filament_deviations_from_canonical_helix(self)
        ! Arguments
        type(filament), intent(inout)  ::  self
        ! Private variables
        integer ::  current_wp
        real    ::  dist_to_next_wp
        real    ::  expected_change_in_phi, actual_change_in_phi
        real    ::  expected_change_in_subunit_num, actual_change_in_subunit_num
        ! Start work

        ! Measure the total curvature - high curvature probably means something is wrong (probably in the z direction)
        write(*,'(a,G15.4)') '**info(filament_deviations_from_canonical_helix): total curvature = ', &
                             filament_eval_total_curvature(self)

        !
        ! Since we have a path length estimate for each waypoint, we can see whether the phi angles agree with each other
        ! (assuming constant canonical helical parameters)
        !
        if (self%helix%twist .ne. 0.) then
            do CURRENT_WP=1,self%NWP-1
                ! Work out the distance to the next waypoint
                dist_to_next_wp = self%wp(CURRENT_WP+1)%path_length - self%wp(current_wp)%path_length
                ! Work out what change in phi this should correspond to
                expected_change_in_phi = dist_to_next_wp * self%helix%twist / self%helix%rise
                ! What change in phi do we actually have?
                actual_change_in_phi = self%wp(CURRENT_WP+1)%euler%phi - self%wp(current_wp)%euler%phi
                ! Expected change in subunit_num
                expected_change_in_subunit_num = dist_to_next_wp / self%helix%rise
                ! Actual change in subunit_num
                actual_change_in_subunit_num = self%wp(CURRENT_WP+1)%subunit_num - self%wp(current_wp)%subunit_num
                ! How far off are we?
                if (.not. self%fragmented_subunit_num .and. .not. self%wps_as_sps) then
                    write(*,'(a,i0,a,i0,a,F0.2,a)') &
                                    '**info(filament_deviations_from_canonical_helix): subunit_num change between waypoints ', &
                                    current_wp, ' and ', CURRENT_WP+1, &
                                    ' is ', actual_change_in_subunit_num - expected_change_in_subunit_num, &
                                    ' subunits off compared to expected value'
                endif
                if (.not. self%fragmented_phi .and. .not. self%wps_as_sps) then
                    write(*,'(2(A,i0),2(a,F0.4),a)',advance='NO')    &
                                                '**info(filament_deviations_from_canonical_helix): phi change between waypoints ', &
                                                current_wp, ' and ', CURRENT_WP+1, &
                                                ' is ', actual_change_in_phi/expected_change_in_phi, &  ! Watch out - division by zero if helix%twist = 0.0
                                                ' times the expected change, corresponding to a difference of ', &
                                                (ACTUAL_CHANGE_IN_PHI-EXPECTED_CHANGE_IN_PHI), ' degrees'
                    if (self%helix%twist .ne. 0.0) then
                        write(*,'(a,F0.4,a)') ' or ', (ACTUAL_CHANGE_IN_PHI-EXPECTED_CHANGE_IN_PHI)/self%helix%twist, ' subunits'
                    else
                        write(*,'(a)') ' '
                    endif
                endif
            enddo
        else
            ! Helical twist is 0.0, therefore we can't really check phi angles
        endif

        !
        ! When a filament was refined in sp mode against a reference with canonical helical symmetry, we can infer the subunit number
        ! of each waypoint from its phi angle. We can then see if the s_spline looks linear. If it doesn't then there are inconsistencies.
        !
    end subroutine filament_deviations_from_canonical_helix

    !>  \brief
    real function filament_discrepency_btw_wps_phi_and_canonical_helix(self) result(total_error)
        implicit none
        ! Arguments
        type(filament),     intent(in)  ::  self
        ! Private variables
        integer ::  current_wp
        ! Start work
        total_error = 0.0
        do CURRENT_WP=1,self%NWP-1
            total_error = total_error + filament_wp_error_in_change_in_phi(self,current_wp)
        enddo
    end function filament_discrepency_btw_wps_phi_and_canonical_helix

    !>  \brief  Work out how far off the following waypoint's phi is from what is predicted by the current waypoint's phi and the arc length distance to the next wp
    real function filament_wp_error_in_change_in_phi(self,wp_num) result(error_in_change_in_phi)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament),     intent(in)  ::  self
        integer,            intent(in)  ::  wp_num      !<  The error will be computed between this waypoint and the next
        ! Private variables
        real    ::  dist_to_next_wp, expected_change_in_phi, actual_change_in_phi
        ! Start work
        if (wp_num .ge. self%nwp .or. wp_num .lt. 1) then
            write(*,'(a)') '**ERROR(filament_wp_error_in_change_in_phi): waypoint number ', wp_num, &
                            ' either does not exist or is the last waypoint in the filament'
            call terminate('Fatal programming error')
        endif
        ! Work out the distance to the next waypoint
        dist_to_next_wp = self%wp(wp_num+1)%path_length - self%wp(wp_num)%path_length
        ! Work out what change in phi this should correspond to
        expected_change_in_phi = dist_to_next_wp * self%helix%twist / self%helix%rise
        ! What change in phi do we actually have?
        actual_change_in_phi = self%wp(wp_num+1)%euler%phi - self%wp(wp_num)%euler%phi
        ! How far off are we?
        error_in_change_in_phi = (actual_change_in_phi-expected_change_in_phi)
    end function filament_wp_error_in_change_in_phi

    !>  \brief  Numerically tweak the filament model to get as close as possible to a relaxed helix
    subroutine filament_relax_helix(self,force_intg_subunit_num,ignore_subunit_phasing,forbid_undertwist,forbid_stretching)
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self                    !<  Filament
        logical,        optional,   intent(in)      ::  force_intg_subunit_num  !<  Subunit_num values can be forced to integers
        logical,        optional,   intent(in)      ::  ignore_subunit_phasing  !<  Do not preserve subunit_num phasing along the model
        logical,        optional,   intent(in)      ::  forbid_undertwist       !<  Do not allow less phi change between waypoints than the canonical helix would give
        logical,        optional,   intent(in)      ::  forbid_stretching       !<  Do not allow less subunits between waypoints than the canonical helix would give
        ! Private variables

        ! Start work

        ! First, sort out the subunit_num values
        call filament_relax_rise_per_subunit(self,force_intg_subunit_num,ignore_subunit_phasing,forbid_stretching)

        ! Now we can modify phi so that the twist per subunit is relaxed
        call filament_relax_twist_per_subunit(self,ignore_subunit_phasing,forbid_undertwist)


    end subroutine filament_relax_helix

    !>  \brief  Numerically tweak subunit_num values to get the rise per subunit as close to canonical as possible
    subroutine filament_relax_rise_per_subunit(self,force_intg_subunit_num,ignore_subunit_num_phasing,forbid_stretching,fal_only)
        use runtime_parameters, only : terminate
        use Splines
        ! Arguments
        type(filament),             intent(inout)   ::  self
        logical,        optional,   intent(in)      ::  force_intg_subunit_num      !<  Subunit_num values can be forced to integers
        logical,        optional,   intent(in)      ::  ignore_subunit_num_phasing  !<  Don't preserve subunit_num phasing along the filament
        logical,        optional,   intent(in)      ::  forbid_stretching           !<  Do not allow less subunits between waypoints than the canonical helix would give
        logical,        optional,   intent(in)      ::  fal_only                    !<  First and last segments only will be relaxed
        ! Private variables
        logical,    parameter   ::  fforce_intg_subunit_num_default = .false.
        logical                 ::  fforce_intg_subunit_num
        integer                 ::  current_wp
        real(kind=8)            ::  dist_to_next_wp
        real(kind=8)            ::  expected_change_in_subunit_num, actual_change_in_subunit_num
        real(kind=8)            ::  subunit_num_to_add
        logical,    parameter   ::  debug   =   .false.
        logical ::  iignore_subunit_num_phasing
        logical ::  fforbid_stretching
        logical ::  ffal_only
        integer ::  wp_incr

        ! Start work
        iignore_subunit_num_phasing = .false.
        if (present(ignore_subunit_num_phasing)) iignore_subunit_num_phasing = ignore_subunit_num_phasing

        fforce_intg_subunit_num = fforce_intg_subunit_num_default
        if (present(force_intg_subunit_num)) fforce_intg_subunit_num = force_intg_subunit_num

        fforbid_stretching = .false.
        if (present(forbid_stretching)) fforbid_stretching = forbid_stretching

        ffal_only = .false.
        if (present(fal_only)) ffal_only = fal_only

        wp_incr = 1
        if (ffal_only) wp_incr = self%NWP-2

        if (fforbid_stretching .and. iignore_subunit_num_phasing) then
            write(*,'(a)') '**WARNING(filament_relax_rise_per_subunit): non-sensical options'
            write(*,'(2a)') '**WARNING(filament_relax_rise_per_subunit): with ignore_subunit_num_phasing, there can be no', &
                            ' stretching or compression, so the forbid_stretching flag is unnecessary or meaningless'
            call terminate('Fatal error in FILAMENT_RELAX_RISE_PER_SUBUNIT. Programming error.') ! This is to catch programming error - the conditions isn't really fatal
        endif

        !
        do current_wp = 1, self%nwp-1, wp_incr
            ! Distance to next wp
            dist_to_next_wp = self%wp(current_wp+1)%path_length - self%wp(current_wp)%path_length
            ! Expected change in subunit_num
            expected_change_in_subunit_num = dist_to_next_wp / self%helix%rise
            ! Actual change in subunit_num
            actual_change_in_subunit_num = self%wp(current_wp+1)%subunit_num - self%wp(current_wp)%subunit_num
            !
            if (debug) then
                write(*,'(2(A,i0),a,2(F0.3,1x))')    '**debug(filament_relax_rise_per_subunit): between waypoints ', &
                                                    current_wp, ' and ', current_wp+1, &
                                                    ' expected, actual change in subunit_num = ', &
                                                    expected_change_in_subunit_num, actual_change_in_subunit_num
            endif
            !
            if (iignore_subunit_num_phasing) then
                subunit_num_to_add =      expected_change_in_subunit_num - actual_change_in_subunit_num
            else
                if (fforbid_stretching) then
                    subunit_num_to_add = ceiling(expected_change_in_subunit_num - actual_change_in_subunit_num)
                else
                    subunit_num_to_add = nint(expected_change_in_subunit_num - actual_change_in_subunit_num)
                endif
            endif
            if (current_wp .gt. self%nwp/2) then
                ! Debug
                if (debug .and. subunit_num_to_add .ne. 0.) then
                    write(*,'(a,F0.5,a,i0,a,i0)') '**debug(filament_relax_rise_per_subunit): adding ', &
                                    subunit_num_to_add, ' subunits to waypoints ', current_wp+1, ' to ', self%nwp
                endif
                ! Correct values
                self%wp(current_wp+1:self%nwp)%subunit_num      =   self%wp(current_wp+1:self%nwp)%subunit_num + &
                                                                    subunit_num_to_add
            else
                ! Debug
                if (debug .and. subunit_num_to_add .ne. 0.) then
                    write(*,'(a,F0.5,a,i0,a,i0)') '**debug(filament_relax_rise_per_subunit): subtracting ', &
                                    subunit_num_to_add, ' subunits to waypoints ', 1, ' to ', current_wp
                endif
                ! Correct values
                self%wp(1:current_wp)%subunit_num      =   self%wp(1:current_wp)%subunit_num - &
                                                                    subunit_num_to_add
            endif
        end do

        !
        ! Testing.
        ! If we are doing FAL only, we should first and foremost make sure that we keep the s monotonic
        !
        if (ffal_only) then
            self%wp(1)%subunit_num = self%wp(2)%subunit_num &
                                - (self%wp(3)%subunit_num-self%wp(2)%subunit_num) &
                                /(self%t(3)-self%t(2))&
                                *(self%t(2)-self%t(1))
            self%wp(self%nwp)%subunit_num = self%wp(self%nwp-1)%subunit_num &
                                + (self%wp(self%nwp-1)%subunit_num-self%wp(self%nwp-2)%subunit_num) &
                                /(self%t(self%nwp-1)-self%t(self%nwp-2)) &
                                *(self%t(self%nwp)-self%t(self%nwp-1))
        endif



        if (fforce_intg_subunit_num) then
            self%wp%subunit_num = nint(self%wp%subunit_num)
        endif

        !
        if (debug) write(*,'(a)') '**debug(filament_relax_rise_per_subunit): making sure subunit_num values not equal'
        call filament_ensure_subsequent_subunit_num_values_not_equal(self)

        ! Update the relevant splines
        call filament_solve_helix_splines(self)

        ! The spline should be monotonic
        if (debug) then
        if (.not. self%s_spline%IsMonotonic()) then
            write(*,'(a)') '**ERROR(filament_relax_rise_per_subunit): subunit_num spline is not monotonic'
            call self%PrintInfo()
            call terminate('Fatal error in FILAMENT_RELAX_RISE_PER_SUBUNIT')
        endif
        endif

    end subroutine filament_relax_rise_per_subunit

    !>  \brief  Check against two subsquent values of subunit_num being equal and add integer numbers of subunit_nums if necessary.
    subroutine filament_ensure_subsequent_subunit_num_values_not_equal(self)
        type(filament), intent(inout)   ::  self
        ! private variables
        integer ::  current_wp
        ! start work
        do current_wp=1,self%nwp-1
                if (self%wp(current_wp)%subunit_num .eq. self%wp(current_wp+1)%subunit_num) then
                    ! See if we can move it one back
                    if (current_wp .eq. 1) then
                        self%wp(current_wp)%subunit_num = self%wp(current_wp)%subunit_num - 1.0
                    else if (self%wp(current_wp-1)%subunit_num .lt. self%wp(current_wp)%subunit_num - 1.5) then
                        self%wp(current_wp)%subunit_num = self%wp(current_wp)%subunit_num - 1.0
                    ! See if we can move the next one forward
                    else if (current_wp .eq. self%nwp - 1) then
                        self%wp(self%nwp)%subunit_num = self%wp(self%nwp)%subunit_num + 1.0
                    else if (self%wp(current_wp+2)%subunit_num .gt. self%wp(current_wp+1)%subunit_num + 1.5) then
                        self%wp(current_wp+1)%subunit_num = self%wp(current_wp+1)%subunit_num + 1
                    else
                        write(*,'(a,i0,a,i0,a,F0.3)') '**warning(filament_guess_wps_subunit_num): waypoints ',     &
                                                        current_wp, ' and ',  current_wp+1, &
                                                        ' have the same subunit_num value: ', self%wp(current_wp)%subunit_num
                    endif
                endif
            enddo
    end subroutine filament_ensure_subsequent_subunit_num_values_not_equal

    !>  \brief  Add or subtract helical turns to get as close as possible to a canonical twist per subunit
    subroutine filament_relax_twist_per_subunit(self,ignore_subunit_num_phasing,forbid_undertwist,fal_only,allow_any_change)
        use runtime_parameters, only : terminate
        use Splines
        ! Arguments
        type(filament),             intent(inout)   ::  self                        !< Filament
        logical,        optional,   intent(in)      ::  ignore_subunit_num_phasing  !<  Do not preserve phasing of subunits
        logical,        optional,   intent(in)      ::  forbid_undertwist           !<  Do not allow less phi change between waypoints than the canonical helix would give
        logical,        optional,   intent(in)      ::  fal_only                    !<  First and last segments only will be relaxed
        logical,        optional,   intent(in)      ::  allow_any_change            !<  Allow a change of any value in phi (by default only 360.0/axial_sym is allowed)
        ! Private variables
        integer ::  current_wp
        real(kind=8)    ::  subunits_to_next_wp
        real(kind=8)    ::  expected_change_in_phi, actual_change_in_phi
        real(kind=8)    ::  dphi
        real(kind=8)    ::  phi_to_add
        real(kind=8)    ::  allowed_change
        logical,    parameter   ::  debug   =   .false.
        logical ::  iignore_subunit_num_phasing
        logical ::  fforbid_undertwist
        logical ::  ffal_only
        integer ::  wp_incr
        logical ::  aallow_any_change
        logical ::  safe_to_add
        ! Start work

        !if (debug) then
        !    write(*,'(a)') '**debug(filament_relax_twist_per_subunit): filament info on entry'
        !    call filament_print_info(self)
        !endif

        iignore_subunit_num_phasing = .false.
        if (present(ignore_subunit_num_phasing)) iignore_subunit_num_phasing = ignore_subunit_num_phasing

        fforbid_undertwist = .false.
        if (present(forbid_undertwist)) fforbid_undertwist = forbid_undertwist

        if (fforbid_undertwist .and. iignore_subunit_num_phasing) then
            write(*,'(a)') '**WARNING(filament_relax_twist_per_subunit): non-sensical options'
            write(*,'(2a)') '**WARNING(filament_relax_twist_per_subunit): with ignore_subunit_num_phasing, there can be no', &
                            ' stretching or compression, so the fforbid_undertwist flag is unnecessary or meaningless'
            call terminate('Fatal error in FILAMENT_RELAX_TWIST_PER_SUBUNIT. Programming error.') ! This is to catch programming error - the conditions isn't really fatal
        endif

        ffal_only = .false.
        if (present(fal_only)) ffal_only = fal_only

        wp_incr = 1
        if (ffal_only) wp_incr = self%nwp-2

        aallow_any_change = .false.
        if (present(allow_any_change)) aallow_any_change = allow_any_change

        ! Work out the allowed change
        if (aallow_any_change) then
            allowed_change = 0.001D0
        else
            allowed_change = 360.0D0 / self%helix%axial_symmetry
        endif


        ! Loop over waypoints
        do current_wp = 1, self%nwp-1, wp_incr
            ! number of subunits to the next waypoint
            subunits_to_next_wp             = self%wp(current_wp+1)%subunit_num - self%wp(current_wp)%subunit_num
            ! expected change
            expected_change_in_phi          = subunits_to_next_wp * self%helix%twist
            ! actual change
            actual_change_in_phi            = self%wp(current_wp+1)%euler%phi - self%wp(current_wp)%euler%phi
            !
            if (debug) then
                write(*,'(2a,i0,a,i0,a,2(F0.2,1x))')    '**debug(filament_relax_twist_per_subunit): expected, actual ',&
                                                        'change between waypoints ', &
                                                        current_wp, ' and ', CURRENT_WP+1, ': ', &
                                                        expected_change_in_phi, actual_change_in_phi
            endif
            !
            dphi = actual_change_in_phi-expected_change_in_phi
            if (iignore_subunit_num_phasing) then
                phi_to_add = - dphi
            else
                if (fforbid_undertwist) then
                    phi_to_add = -allowed_change*(ceiling(dphi/allowed_change))
                else
                    phi_to_add = -allowed_change*(nint(dphi/allowed_change))
                endif
            endif



            ! Correct values
            if (current_wp .gt. self%nwp/2) then
                safe_to_add =   (self%helix%twist .gt. 0.0  .and.                                                   &
                                (self%wp(current_wp+1)%euler%phi + phi_to_add .gt. self%wp(current_wp)%euler%phi))  &
                                .or.                                                                                &
                                (self%helix%twist .lt. 0.0  .and.                                                   &
                                (self%wp(current_wp+1)%euler%phi + phi_to_add .lt. self%wp(current_wp)%euler%phi))
                if ( safe_to_add ) then
                    ! Debug
                    if (debug .and. phi_to_add .ne. 0.) then
                        write(*,'(a,F0.2,a,i0,a,i0)') '**debug(filament_relax_twist_per_subunit): adding ', &
                                        phi_to_add, ' twist degrees to waypoints ', current_wp+1, ' to ', self%nwp
                    endif
                    self%wp(current_wp+1:self%nwp)%euler%phi    =   self%wp(current_wp+1:self%nwp)%euler%phi + &
                                                                    phi_to_add
                endif
            else
                safe_to_add =   (self%helix%twist .gt. 0.0  .and.                                                   &
                                (self%wp(current_wp)%euler%phi - phi_to_add .lt. self%wp(current_wp+1)%euler%phi))  &
                                .or.                                                                                &
                                (self%helix%twist .lt. 0.0  .and.                                                   &
                                (self%wp(current_wp)%euler%phi - phi_to_add .gt. self%wp(current_wp+1)%euler%phi))
                if ( safe_to_add ) then
                    ! Debug
                    if (debug .and. phi_to_add .ne. 0.) then
                        write(*,'(a,F0.2,a,i0,a,i0)') '**debug(filament_relax_twist_per_subunit):  subtracting ', &
                                        phi_to_add, ' twist degrees to waypoints ', 1, ' to ', current_wp
                    endif
                    self%wp(1:current_wp)%euler%phi             =   self%wp(1:current_wp)%euler%phi - &
                                                                    phi_to_add
                endif
            endif
        end do

        !
        ! Testing.
        ! If we are doing FAL only, we should first and foremost make sure that we keep the phi monotonic
        !
        if (ffal_only) then
            self%wp(1)%euler%phi = self%wp(2)%euler%phi &
                                - (self%wp(3)%euler%phi-self%wp(2)%euler%phi) &
                                /(self%t(3)-self%t(2))&
                                *(self%t(2)-self%t(1))
            self%wp(self%nwp)%euler%phi = self%wp(self%nwp-1)%euler%phi &
                                + (self%wp(self%nwp-1)%euler%phi-self%wp(self%nwp-2)%euler%phi) &
                                /(self%t(self%nwp-1)-self%t(self%nwp-2)) &
                                *(self%t(self%nwp)-self%t(self%nwp-1))
        endif

        ! Update the relevant splines
        call filament_solve_helix_splines(self)

        ! The spline should be monotonic
        if (debug) then
            if (.not. (self%fragmented_phi .or. self%fragmented_subunit_num .or. self%wps_as_sps)) then
                if (.not. filament_splines_are_monotonic(self)) then
                    write(*,'(a)') '**ERROR(filament_relax_twist_per_subunit): phi & subunit_num splines are not both monotonic'
                    call self%PrintInfo()
                    call terminate('Fatal error in FILAMENT_RELAX_TWIST_PER_SUBUNIT')
                endif
            endif
        endif

    end subroutine filament_relax_twist_per_subunit

    !>  \brief  Add or subtract helical turns to get closer to a relaxed helix
    subroutine filament_add_helical_turns_to_relax(self)
        implicit none
        ! Arguments
        type(filament), intent(inout)  ::  self    !< Filament
        ! Private variables
        integer ::  current_wp
        real    ::  dist_to_next_wp
        real    ::  expected_change_in_subunit_num, actual_change_in_subunit_num
        real    ::  expected_change_in_phi, actual_change_in_phi
        real    ::  dphi
        real    ::  subunit_num_to_add, phi_to_add
        logical,    parameter   ::  debug   =   .true.
        ! Start work

        ! Loop over waypoints
        do current_wp = 1, SELF%NWP-1
            ! Distance to next wp
            dist_to_next_wp = self%wp(CURRENT_WP+1)%path_length - self%wp(current_wp)%path_length
            ! Expected change
            expected_change_in_subunit_num = dist_to_next_wp / self%helix%rise
            expected_change_in_phi         = expected_change_in_subunit_num * self%helix%twist
            ! Actual change
            actual_change_in_subunit_num = self%wp(CURRENT_WP+1)%subunit_num - self%wp(current_wp)%subunit_num
            actual_change_in_phi         = self%wp(CURRENT_WP+1)%euler%phi - self%wp(current_wp)%euler%phi
            !
            dphi = ACTUAL_CHANGE_IN_PHI-EXPECTED_CHANGE_IN_PHI
            phi_to_add = 360.*(nint(dphi/360.))
            subunit_num_to_add = nint(phi_to_add / self%helix%twist)      ! i would have thought a nint here would make sense
                                                                    ! but testing so far does not confirm this (29-feb-2012)

            ! Debug
            if (debug .and. phi_to_add .ne. 0.) then
!                write(*,'(a,F0.2,a,F0.1,a,i0,a,i0)') '**debug(filament_add_helical_turns_to_relax): adding ', &
!                                subunit_num_to_add, ' subunits and ', &
!                                phi_to_add, ' twist degrees to waypoints ', CURRENT_WP+1, ' to ', self%nwp
                write(*,'(a,F0.1,a,i0,a,i0)') '**debug(filament_add_helical_turns_to_relax): adding ', &
                                phi_to_add, ' twist degrees to waypoints ', CURRENT_WP+1, ' to ', self%nwp
            endif

            ! Correct values
!            self%wp(CURRENT_WP+1:self%nwp)%subunit_num  =   self%wp(CURRENT_WP+1:self%nwp)%subunit_num + &
!                                                            subunit_num_to_add
            self%wp(CURRENT_WP+1:self%nwp)%euler%phi    =   self%wp(CURRENT_WP+1:self%nwp)%euler%phi + &
                                                            phi_to_add
        end do
    end subroutine filament_add_helical_turns_to_relax

    !>  \brief  Add or subtract integer multiples of 360 degrees to waypoints' phi values to minimise discrepencies with canonical helical symmetry
    !!
    !!  It is assumed the filament model is up-to-date (in particular the path lengths)
    subroutine filament_wps_add_sub_phi_360(self)
        implicit none
        ! Arguments
        type(filament),     intent(inout)   ::  self    !<  Filament object. Only WPS' phi values will be changed
        ! Private variables
        integer ::  current_wp
        real    ::  error_in_change_in_phi
        ! Start work

        !
        ! Since we have a path length estimate for each waypoint, we can see whether the phi angles agree with each other
        ! (assuming constant canonical helical parameters). If there is a disagreement which is over 180 degrees, add/subtract
        ! n x 360deg
        !
        do CURRENT_WP=1,self%NWP-1
            ! How far off are we?
            error_in_change_in_phi = filament_wp_error_in_change_in_phi(self,current_wp)
            ! If we're more than 180 degrees off, then we can improve things by adding/subtracting 360
            if (abs(error_in_change_in_phi) .gt. 180.) then
                self%wp(CURRENT_WP+1)%euler%phi = self%wp(CURRENT_WP+1)%euler%phi &
                                                - 360. * sign(1.,error_in_change_in_phi) * nint(abs(error_in_change_in_phi)/360.)
            endif
        enddo

    end subroutine filament_wps_add_sub_phi_360

    !>  \brief  Compute the ideal path_lengths for the waypoints, taking their phi values to be correct
    !!          and aiming for canonical helical symmetry
    function filament_ideal_path_lengths_for_canonical_helix(self) result(path_lengths)
        implicit none
        ! Arguments
        type(filament), intent(in)  ::  self                !<  Filament object
        ! Result
        real,           allocatable ::  path_lengths(:)
        ! Private varibles
        integer ::  current_wp
        ! Start work

        ! Allocate & initialise results
        allocate(path_lengths(self%nwp))
        path_lengths(1) = 0.0

        ! Loop over waypoints
        if (self%helix%twist .ne. 0.0) then
            do CURRENT_WP=2,self%nwp
                path_lengths(current_wp) =  path_lengths(CURRENT_WP-1) + &
                                            ( self%wp(current_wp)%euler%phi - self%wp(CURRENT_WP-1)%euler%phi ) * &
                                            self%helix%rise / self%helix%twist
            enddo
        else
            path_lengths = 0.0
        endif
    end function filament_ideal_path_lengths_for_canonical_helix

    !>  \brief  The discrepency between the actual path lengths and the path lengths which would accomodate the canonical helical
    !!          symmetry given the waypoints' phi values
    real function filament_deviation_from_ideal_path_lengths(self,ideal_path_lengths) result(dev)
        implicit none
        ! Arguments
        type(filament),                     intent(in)  ::  self                    !<  Filament object
        real,   allocatable,    optional,   intent(in)  ::  ideal_path_lengths(:)   !<  Desired path length values for the filament's waypoints
        ! Private variables
        integer         ::  current_wp
        real(kind=8)    ::  ddev
        real,   allocatable ::  iideal_path_lengths(:)
        ! Start work

        ! Ideal path lengths
        if (present(ideal_path_lengths)) then
            iideal_path_lengths = ideal_path_lengths
        else
            iideal_path_lengths = filament_ideal_path_lengths_for_canonical_helix(self)
        endif

        ! Accumulate sum of deviations
        ddev = 0.0D0
        do CURRENT_WP=2,self%nwp
            ddev = ddev + abs(self%wp(current_wp)%path_length - iideal_path_lengths(current_wp))
        enddo
        dev = ddev
    end function filament_deviation_from_ideal_path_lengths

end module filaments_geometry
