module filaments_microtubules
    use filaments_base
    implicit none

    contains

    !>  \brief  Check whether the microtubule is expected to have a seam
    !!  \todo   Implement logic to check arbitrary microtubule geometry
    logical function microtubule_has_seam(self)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament),     intent(in)  ::  self    !<  Filament object. Must be a microtubule
        ! Private variables

        ! Start work
        if (.not. self%microtubule) then
            write(*,'(a)') '**ERROR(microtubule_has_seam): filament is not a microtubule'
            call terminate('Fatal error in MICROTUBULE_HAS_SEAM')
        endif

        ! i do not have time to implement all possible geometries for now,
        ! so I'm just implementing the particular case i need during development.
        if (self%mt_pf_number .eq. 13 .and. self%mt_rise .eq. 3) then
            microtubule_has_seam = .true.
        else
            write(*,'(a)') '**ERROR(microtubule_has_seam): feature not implemented yet.'
            call terminate('Fatal error in MICROTUBULE_HAS_SEAM')
        endif
    end function microtubule_has_seam

    !>  \brief  Shift the waypoints of a microtubule filament so that protofilament 1 has been moved pf_shift protofilaments over
    subroutine microtubule_shift_pf(self,pf_shift)
        use runtime_parameters, only : terminate
        use filaments_geometry, only : filament_solve_splines
        implicit none
        ! Arguments
        type(filament),     intent(inout)   ::  self        !<  Microtubule filament
        integer,            intent(in)      ::  pf_shift    !<  By how many protofilaments should this microtubule be shifted
        ! Private variables
        integer         ::  current_wp
        real            ::  change_in_phi!, change_in_t
        !type(waypoint)  ::  wp_tmp
        ! Start work

        ! Check filament is a microtubule
        if (.not. self%microtubule) then
            write(*,'(a)') '**ERROR(microtubule_shift_pf): filament is not a microtubule'
            call terminate('Fatal error in MICROTUBULE_SHIFT_PF')
        endif

        ! Check the pf shift value makes sense
        if (pf_shift .gt. self%mt_pf_number) then
            write(*,'(a,2(i0,a))') '**ERROR(microtubule_shift_pf): cannot shift a ', self%mt_pf_number, '-pf microtubules by ', &
                                    pf_shift, ' protofilaments'
            call terminate('Fatal error in MICROTUBULE_SHIFT_PF')
        endif

        ! Loop over waypoints
        do CURRENT_WP=1,self%nwp

            ! Work out changes in t and Phi
            !change_in_t = microtubule_local_washer_rise_per_dimer(self,self%t(current_wp)) * real(pf_shift)
            change_in_phi = microtubule_local_washer_twist_per_dimer(self,self%t(current_wp)) * real(pf_shift)

            ! Apply change in Phi
            self%wp(current_wp)%euler%phi   = self%wp(current_wp)%euler%phi - change_in_phi

            ! Apply change in subunit_num (integer subunit_num indicates dimer of pf1)
            self%wp(current_wp)%subunit_num = self%wp(current_wp)%subunit_num - real(pf_shift) / real(self%mt_pf_number)

        enddo

        ! Update splines
        call filament_solve_splines(self)   !,USE_PSI_THETA_CONSTRAINTS=.TRUE.) ! Commented out 29-Nov, with introduction of use_hermite_splines property of filaments
    end subroutine microtubule_shift_pf

    !>  \brief  Compute the angular distance from one protofilament (tubulin dimer) to the next within a washer at the given
    !!          abscissa along the microtubule filament
    real function microtubule_local_washer_twist_per_dimer(self,t) result(twist_per_dimer)
        use filaments_geometry, only : filament_eval_helical_twist
        implicit none
        ! Arguments
        type(filament),     intent(in)      ::  self        !<  Microtubule filament
        real(kind=8),       intent(in)      ::  t           !<  The abscissa along the filament at which the twist will be computed
        ! Private variables
        real    ::  pf1_to_pf1_twist
        ! Start work

        ! What's the twist from pf1 of one washer to pf1 of the next?
        pf1_to_pf1_twist = filament_eval_helical_twist(self,t) + 360.0

        !
        twist_per_dimer = pf1_to_pf1_twist/real(self%mt_pf_number)
    end function microtubule_local_washer_twist_per_dimer

    !>  \brief  Compute the distance from one protofilament (tubulin dimer) to the next within a washer at the given
    !!          abscissa along the microtubule filament
    real function microtubule_local_washer_rise_per_dimer(self,t) result(rise_per_dimer)
        use filaments_geometry, only : filament_eval_helical_rise
        implicit none
        ! Arguments
        type(filament),     intent(in)      ::  self        !<  Microtubule filament
        real(kind=8),       intent(in)      ::  t           !<  The abscissa along the filament at which the rise will be computed
        ! Private variables
        real    ::  local_rise
        ! Start work

        ! Work out local rise from pf1 of one washer to pf1 of the next
        local_rise = filament_eval_helical_rise(self,t)

        !
        rise_per_dimer = local_rise/real(self%mt_pf_number)
    end function microtubule_local_washer_rise_per_dimer

    !>  \brief  Given a microtubule symmetry, return the number of the protofilament which can be reconstructed correctly
    integer function microtubule_get_good_pf(self) result(good_pf)
        use runtime_parameters, only : terminate, is_even
        implicit none
        ! Arguments
        type(filament),     intent(in)  ::  self    !<  Filament object. Must be a microtubule.
        ! Private variables

        ! Start work

        ! Check the filament is a microtubule
        if (.not. self%microtubule) then
            write(*,'(a)') '**ERROR(microtubule_get_good_pf): filament is not a microtubule'
            call terminate('Fatal error in MICROTUBULE_GET_GOOD_PF')
        endif

        ! Work out which protofilaments we will aim for
        if (is_even(self%mt_pf_number)) then
            ! Even number of protofilaments
            good_pf = self%mt_pf_number/2
        else
            ! Odd number of protofilaments
            good_pf = (self%MT_PF_NUMBER-1)/2 + 1
        endif
    end function microtubule_get_good_pf

    !>  \brief  Impose microtubule symmetry in real space. This is necessary when calculating
    !!          a reconstruction of a microtubule with a seam, in which case only one protofilament is reconstructed correctly
    subroutine image_impose_microtubule_symmetry(self,vol)
        use euleranglesmod
        use interpolation
        use rotation_matrices, only : radians
        use runtime_parameters, only : terminate
        use InterpolatingWindows
        use Images
        implicit none
        ! Arguments
        type(filament),                     intent(in)      ::  self
        type(image),                        intent(inout)   ::  vol             !<  Volume to be symmetrised
        ! Private variables
        integer                 ::  good_pf
        integer                 ::  i,j,k, ii,jj,kk, ii2,jj2,kk2
        real                    ::  max_rad_sq
        integer                 ::  current_pf
        real                    ::  segment_phi
        integer                 ::  origin(3)
        real                    ::  segment_z
        real                    ::  z_shift
        logical,    parameter   ::  sanity_check = .false.
        logical,    parameter   ::  debug = .false.
        real                    ::  phi_rot
        real                    ::  coo(3)
        real                    ::  dist_from_edge_of_pf
        ! Private variables to do with interpolation etc
        type(window)            ::  win
        type(interpolation_scheme)  ::  interpol_scheme
        ! Start work

        ! Check the volume is in real space
        if (.not. vol%is_in_real_space) then
            write(*,'(a)') '**info(image_impose_microtubule_symmetry): volume will be back-Fourier-transformed to real space'
            call vol%BackwardFFT()
        endif

        ! Prepare the window which will be used for interpolation
        call win%AllocateWindow(2,VOL%IsInFourierSpace(),vol%IsAVolume())

        ! Prepare an interpolation scheme
        call interpolation_scheme_init(interpol_scheme,'LINEAR','RECT',2)

        ! Get address of origin (centre) voxel
        if (vol%is_in_real_space .and. vol%object_is_centered_in_box) then
            origin = vol%physical_address_of_box_center
        else
            origin = [1,1,1]
        endif

        ! Square the maximum filament radius
        max_rad_sq = self%MAX_RADIUS**2
        if (debug) print '(a,F0.4)', '**debug(image_impose_microtubule_symmetry): self%max_radius = ', self%max_radius

        ! The phi extent of each segment
        segment_phi = 360./real(self%mt_pf_number)

        ! The Z-shift from one protofilament to the next
        segment_z = 1. / real(self%mt_pf_number) * real(self%mt_rise) / 2. * self%helix%rise

        ! Which is the good PF?
        good_pf = microtubule_get_good_pf(self)
        if (debug) print '(a,i0)', '**debug(image_impose_microtubule_symmetry): good_pf = ', good_pf

        ! Loop over output voxels
        k = origin(3)
        kk = 0
        do J=1,vol%logical_dimensions(2)
            jj = (J-ORIGIN(2))
            jj2 = JJ**2
            do I=1,vol%logical_dimensions(1)
                ii = (I-ORIGIN(1))
                ii2 = II**2
                ! If beyond the maximum cylinder radius of the filament, or if at x,y origin, skip
                if (II2+JJ2 .gt. max_rad_sq .or. II2+JJ2 .eq. 0) cycle
                ! Which pf are we on currently?
                if (debug) print '(a,3(F0.3,1x))', '**debug(image_impose_microtubule_symmetry): coo = ', real(ii),real(jj),real(kk)
                current_pf = microtubule_pf_from_coos(self,[REAL(ii),real(jj),real(kk)],segment_phi)
                if (debug) print '(a,i0)', '**debug(image_impose_microtubule_symmetry): current_pf = ', current_pf
                ! Now we know which protofilament we're in, we can copy the appropriate value to the output volume
                if (current_pf .eq. good_pf) then
                    ! unchanged
                    cycle
                else
                    ! we need to copy values from the good protofilament
                    phi_rot = - segment_phi * real(CURRENT_PF-GOOD_PF)  ! + or -?
                    z_shift = - segment_z   * real(CURRENT_PF-GOOD_PF)  ! + or -?
                    if (debug) then
                        print '(a,i0,1x,F0.3)', '**debug(image_impose_microtubule_symmetry): got here with current_pf, phi_rot = ',&
                                             current_pf, phi_rot
                    endif
                    ! The coordinates from which we're copying
                    coo(1) =  cos(radians(phi_rot)) * real(ii) + sin(radians(phi_rot)) * real(jj)
                    coo(2) = -SIN(radians(phi_rot)) * real(ii) + cos(radians(phi_rot)) * real(jj)
                    ! Sanity check - we should now be within the good pf
                    if (sanity_check) then
                        current_pf = microtubule_pf_from_coos(self,coo,segment_phi)
                        if (current_pf .ne. good_pf) then
                            ! It appears that we are copying from the wrong PF. But let's first check whether we're actually very close
                            ! to the edge of the good pf, in which case we might just be victim of rounding errors.
                            dist_from_edge_of_pf = microtubule_pf_from_coos_distance(self,coo,segment_phi)
                            if (debug) then
                                print '(a,F0.2)', '**debug(image_impose_microtubule_symmetry): distance from edge = ', &
                                                    dist_from_edge_of_pf
                            endif
                            if (abs(dist_from_edge_of_pf) .gt. 0.0005 .or. abs(CURRENT_PF-GOOD_PF) .gt. 1) then
                                write(*,'(2a,i0,a)') '**ERROR(image_impose_microtubule_symmetry): trying to copy ',  &
                                                    'from wrong pf segment (', current_pf, ')'
                                call terminate('Fatal error in IMAGE_IMPOSE_MICROTUBULE_SYMMETRY')
                            endif
                        endif
                    endif
                    ! Loop over that column of voxels
                    do K=1,vol%logical_dimensions(3)
                        kk = (K-ORIGIN(3))
                        kk2 = KK**2
                        ! Check whether the z shift takes us beyond the edge of the volume
                        if (k + z_shift .lt. 1 .or. k + z_shift .gt. vol%logical_dimensions(3)) then   ! + or - ?
                            cycle
                        endif
                        ! The coordinates from which we're copying
                        coo(3) =  real(kk) + z_shift
                        ! Center the window around the coordinates
                        win%centered_on = coo
                        ! Define the bounds of the window over the volume
                        call eval_window_bounds(win)
                        ! Copy image data to the window, get interpolated value from window
                        call window_from_image_real(vol,win)
                        call get_real_from_window(vol%real_values(i,j,k),win,&
                                                    interpol_scheme%interpolation_function,&
                                                    interpol_scheme%windowing_function)
                    enddo
                endif
            enddo
        enddo
    end subroutine image_impose_microtubule_symmetry

    !>  \brief  Find out which protofilament pie slice the coordinates lie in
    pure integer function microtubule_pf_from_coos(self,coos,segment_phi) result(pf)
        use rotation_matrices, only : degrees
        use runtime_parameters, only : mt_phi_first_pf
        use euleranglesmod, only : angle_btwn_0_and_360
        implicit none
        ! Arguments
        type(filament),     intent(in)  ::  self        !<  Microtubule filament
        real,               intent(in)  ::  coos(3)     !<  Coordinates (relative to origin at center of volume)
        real,               intent(in)  ::  segment_phi !<  The Phi extent of each segment
        ! Private variables
        real(kind=8)                    ::  phi
        real(kind=8)                    ::  phi_diff
        !logical,   parameter           ::  debug = .TRUE.
        ! Start work
        !if (debug) print '(a,4(F0.3,x))', '**debug(microtubule_pf_from_coos): coos, segment_phi  =', coos, segment_phi
        ! Which Phi are we at right now?
        phi = degrees(-ATAN2(coos(2),coos(1)))
        ! How far is that from the center of the pf1 slice of the pie? (come up with a positive difference angle between 0 and 360.)
        phi_diff = angle_btwn_0_and_360(phi - mt_phi_first_pf)
        !if (debug) print '(a,2(F0.3,x))', '**debug(microtubule_pf_from_coos): phi, phi_diff = ', phi, phi_diff
        !  Now we can work out which protofilament pie slice we're in
        if (phi_diff .le. segment_phi/2. .or. phi_diff .ge. (real(self%mt_pf_number)-0.5)*SEGMENT_PHI) then
            pf = 1
        else
            pf = 2 + int(phi_diff/SEGMENT_PHI-0.5)
        endif
        !if (debug) print '(a,i0,x,F0.6,x,F0.6)', '**debug(microtubule_pf_from_coos): pf = ', pf, phi_diff/segment_phi, mod(phi_diff,segment_phi)-0.5*SEGMENT_PHI
    end function microtubule_pf_from_coos

    !>  \brief  Return the angular distance (degrees) of the coordinates from the assigned pf segment
    pure real function microtubule_pf_from_coos_distance(self,coos,segment_phi)
        use rotation_matrices, only : degrees
        use runtime_parameters, only : mt_phi_first_pf
        use euleranglesmod, only : angle_btwn_0_and_360
        implicit none
        ! Arguments
        type(filament),     intent(in)  ::  self        !<  Microtubule filament
        real,               intent(in)  ::  coos(3)     !<  Coordinates (relative to origin at center of volume)
        real,               intent(in)  ::  segment_phi !<  The Phi extent of each segment
        ! Private variables
        real(kind=8)                    ::  phi
        real                            ::  phi_diff
        logical,            parameter   ::  debug = .true.
        ! Start work

        ! Which Phi are we at right now?
        phi = degrees(-ATAN2(coos(2),coos(1)))
        ! How far is that from the center of the pf1 slice of the pie? (come up with a positive difference angle between 0 and 360.)
        phi_diff = angle_btwn_0_and_360(phi - mt_phi_first_pf)

        ! Result
        microtubule_pf_from_coos_distance = mod(phi_diff,segment_phi)-0.5*SEGMENT_PHI
    end function microtubule_pf_from_coos_distance

end module filaments_microtubules



