!>  \brief  Minimisation methods for the filament class
module filaments_minimisation
    use filaments_base
    implicit none

    contains

    !>  \brief  Minimise deviations from canonical helix by moving waypoints in z direction
    subroutine filament_minimise_deviations_from_canonical_helix(self)
        use filaments_geometry, only : filament_relax_helix, filament_relax_rise_per_subunit
        use filaments_geometry, only : filament_ideal_path_lengths_for_canonical_helix
        use filaments_model_update, only : filament_update_wps_z_from_wps_theta_and_rigid_z, filament_update_model
        use filaments_model_base, only : filament_print_extra_info
        use filaments, only : filament_print_diff
        use differential_geometry, only : space_curve, space_curve_init, space_curve_adjust_z_to_match_arc_lengths
        use differential_geometry, only : space_curve_eval_total_curvature_new, space_curve_print_info
		use filaments_geometry, only : filament_eval_total_curvature_overtwist_stretch, filament_deviations_from_canonical_helix
		use filaments_geometry, only : filament_eval_total_curvature
		use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  self            !<  Filament object
        ! Private variables
        logical,    parameter   ::  debug                  = .true.
        logical,    parameter   ::  all_wps_simultaneously = .false.
        logical,    parameter   ::  use_space_curves       = .true.
        integer,    parameter   ::  refine_wps_group_size  = 3
        integer,    parameter   ::  refine_wps_group_olap  = 2
        integer,    parameter   ::  number_of_iterations   = 2
        integer                 ::  refine_wps_first, refine_wps_last
        type(filament)          ::  fil_bak
        integer                 ::  current_iteration
        type(space_curve)       ::  current_space_curve
        real(kind=8), allocatable :: ideal_path_lengths(:)
        real(kind=8)            ::  curvature, overtwist, stretch
        real(kind=8)            ::  curvature_start, curvature_finish
        ! Start work

        write(*,'(a)') '**info(filament_minimise_deviations_from_canonical_helix): attempting z coo adjustments to minimise '// &
                                                                                'discrepency from canonical helical symmetry'
        ! Keep track of our starting point
        fil_bak = self
        call space_curve_init(  current_space_curve,self%wp,known_d=.true.,param_error_tol=0.01)
        call space_curve_print_info(current_space_curve)
        curvature_start = space_curve_eval_total_curvature_new(current_space_curve,step_size=0.1,squared_curvature=.false.)
        if (debug) then
            call self%PrintInfo()
            call filament_print_extra_info(self)
        endif
        ! We will only modify the z coordinates
        self%refine_params = .false.
        self%refine_params(3) = .true.
        ! We move the z coordinates so as to minimise vertical curvature. We can do this since we know the
        ! out-of-plane angle (theta) at each waypoint and the avereage z of the filament (rigid_wp%coo(3))
        if (debug) then
            print *, '**debug(filament_minimise_deviations_from_canonical_helix): rigid wp z = ', self%rigid_wp%coo(3)
            call filament_eval_total_curvature_overtwist_stretch(self,curvature,overtwist,stretch,  &
                                                     ignore_fal_wps=.not. (self%fal_wps_in_score .and. self%fal_wps_in_threed))
            write(*,'(3a,3(G15.4,1x))')  '**debug(filament_minimise_deviations_from_canonical_helix): ',   &
                                        'before minimising z curvature, ', &
                                        'total curvature, overtwist, stretch (1/pixels) : ', curvature, overtwist, stretch

        endif
        call filament_update_wps_z_from_wps_theta_and_rigid_z(self)
        call filament_update_model(self)
        if (debug) then
            print *, '**debug(filament_minimise_deviations_from_canonical_helix): rigid wp z = ', self%rigid_wp%coo(3)
            call filament_eval_total_curvature_overtwist_stretch(self,curvature,overtwist,stretch,  &
                                                     ignore_fal_wps=.not. (self%fal_wps_in_score .and. self%fal_wps_in_threed))
            write(*,'(3a,3(G15.4,1x))')  '**debug(filament_minimise_deviations_from_canonical_helix): ',   &
                                        'after minimising z curvature, ', &
                                        'total curvature, overtwist, stretch (1/pixels) : ', curvature, overtwist, stretch

            call self%PrintInfo()
        endif

        ! No we can relax the helix again, but only by adding subunits or turns between waypoints, so that if anything it
        ! will have overtwist and compression.
        ! That way, the minimiser will later be able to add arc length between waypoints to "accomodate" these extra
        ! subunits or turns and relieve the overtwist and compression.
        ! On the other hand, if we let the helix relaxer subtract subunits or turns between waypoints,
        ! the minimiser will not be able to do anything about it, since we already have the waypoints as close
        ! together as possible
        do current_iteration=1,number_of_iterations
            if (current_iteration==1) then
                call filament_relax_helix(self,forbid_undertwist=.true.,forbid_stretching=.true.)
            else
                call filament_relax_helix(self,forbid_undertwist=.false.,forbid_stretching=.false.)
            endif
            call filament_update_model(self)
            if (debug) then
                call self%PrintInfo()
                call filament_print_extra_info(self)
            endif
            ! What are our ideal path lengths?
            ideal_path_lengths = filament_ideal_path_lengths_for_canonical_helix(self)
            print *, '**debug(filament_minimise_deviations_from_canonical_helix): ideal path lengths = ', ideal_path_lengths
			if (any(ideal_path_lengths .lt. 0.0)) then
				write(*,'(a)') '**ERROR(filament_minimise_deviations_from_canonical_helix): weird values for ideal path lengths'
				print *, '**debug(filament_minimise_deviations_from_canonical_helix): ideal path lengths = ', ideal_path_lengths
				call self%PrintInfo()
				call terminate('Fatal error in FILAMENT_MINIMISE_DEVIATIONS_FROM_CANONICAL_HELIX')
			endif


            ! We will refine waypoints' z coordinates.
            ! Optionally, we do this by groups of waypoints to take it easy on the minimiser and
            ! hopefully save time too
            refine_wps_first = 0
            refine_wps_last  = 0
            do
                if (all_wps_simultaneously) then
                    refine_wps_first = 1
                    refine_wps_last  = self%nwp
                else
                    ! The first and last waypoints which will be refined this time
                    refine_wps_first = refine_wps_first + refine_wps_group_size - refine_wps_group_olap
                    refine_wps_last  = refine_wps_first + refine_wps_group_size - 1
                    ! Take care of edge conditions
                    if (refine_wps_last .gt. self%nwp) refine_wps_last = self%nwp
                endif
                write(*,'(2a,i0,a,i0)') '**info(filament_minimise_deviations_from_canonical_helix): tweaking z coordinates for ', &
                                        'waypoints ', refine_wps_first, ' to ', refine_wps_last
                if (use_space_curves) then
                    call space_curve_init(  current_space_curve,self%wp(refine_wps_first:refine_wps_last),    &
                                            known_d=.true.,param_error_tol=0.01)

                    call space_curve_adjust_z_to_match_arc_lengths(current_space_curve,             &
                       target_arc_length=    ideal_path_lengths(refine_wps_first:refine_wps_last)   &
                                            -ideal_path_lengths(refine_wps_first))
                    ! Plug the values back in to the filament
                    self%wp(refine_wps_first:refine_wps_last)%coo(3) = current_space_curve%z_spline%y(:)
                    call filament_update_model(self)
                else
                    ! Set the refine flags
                    self%wp%refine = .false.
                    self%wp(refine_wps_first:refine_wps_last)%refine = .true.
                    ! Work on the current group of waypoints
                    call filament_minimise( self,         &
                                            minimise_curvature=.false.,     &
                                            minimise_deviations_from_ideal_path_lengths=.true.  )
                endif
                ! Check whether this was the last group of waypoints we're to process
                if (refine_wps_last .eq. self%nwp) then
                    exit
                endif
             enddo
            call filament_relax_rise_per_subunit(self,force_intg_subunit_num=.true.)
        enddo
        call filament_update_model(self)
        if (debug) then
            write(*,'(a)') '**debug(filament_minimise_deviations_from_canonical_helix): before relax helix, filament info:'
            call self%PrintInfo()
            call filament_print_extra_info(self)
        endif

        call filament_relax_helix(self,FORBID_UNDERTWIST=.false.,FORBID_STRETCHING=.false.)
        call filament_update_model(self)
        call space_curve_init(  current_space_curve,self%wp,KNOWN_D=.true.,PARAM_ERROR_TOL=0.01)
        call space_curve_print_info(current_space_curve)
        curvature_finish = space_curve_eval_total_curvature_new(current_space_curve,STEP_SIZE=0.1,SQUARED_CURVATURE=.false.)
        !if (debug) then
            write(*,'(a)') '**debug(filament_minimise_deviations_from_canonical_helix): on exit, filament info:'
            call self%PrintInfo()
            call filament_print_extra_info(self)
            write(*,'(a)') '**debug(filament_minimise_deviations_from_canonical_helix): on exit, filament changes:'
            call filament_print_diff(fil_bak,self)
            write(*,'(a,2(F0.2,1x))') '**debug(filament_minimise_deviations_from_canonical_helix): curvature start, finish: ', &
                                    curvature_start, curvature_finish
        !endif

    end subroutine filament_minimise_deviations_from_canonical_helix

    !>  \brief  Minimise physical characteristics of the filament, e.g. curvature, torsion etc.
    !!  \todo   Use case construct and character string argument to specify what to minimise
    subroutine filament_minimise(self,  minimise_curvature,minimise_deviations_from_canonical_helix,  &
                                        minimise_deviations_from_ideal_path_lengths)
        use minimisers, only : ds_init, ds_run, minimiser_ds_fgsl, ds_max_iterations, ds_set_start_step
        use runtime_parameters, only : terminate
        use filaments, only : values_to_filament, values_from_filament
        use filaments_model_update, only : filament_update_model
        use filaments_geometry, only : filament_discrepency_btw_wps_phi_and_canonical_helix
        use filaments_geometry, only : filament_deviation_from_ideal_path_lengths
        use filaments_geometry, only : filament_eval_total_curvature
        use ConjugateGradients
        use, intrinsic :: iso_c_binding
        implicit none
        ! Arguments
        type(filament),                     intent(inout)       ::  self                !<  Filament object
        logical,            optional,       intent(in)          ::  minimise_curvature  !<  Whether the curvature should be minimised. Defaults to true.
        logical,            optional,       intent(in)          ::  minimise_deviations_from_canonical_helix
        logical,            optional,       intent(in)          ::  minimise_deviations_from_ideal_path_lengths
        ! Private variables
        logical                                 ::  mminimise_curvature
        logical                                 ::  mminimise_deviations_from_canonical_helix
        logical                                 ::  mminimise_deviations_from_ideal_path_lengths
        character(len=2)                        ::  num_dim_c
        ! Private variables for the minimiser
        type(minimiser_ds_fgsl)                 ::  ds_minimiser
        type(ConjugateGradient)                 ::  cg_minim
        integer                                 ::  num_dim, ierr
        real(kind=8),           allocatable     ::  best_values(:)
        real(kind=8)                            ::  best_score
        real                                    ::  ds_target_size                  !<  Target size of the simplex.
        logical,                parameter       ::  debug   =   .false.
        real(kind=8),           allocatable     ::  first_step(:), start_values(:), accuracy(:)
        logical,				parameter		::	path_lengths_cg = .false.
		! Start work


        ! What will we be minimising?
        mminimise_curvature = .true.
        if (present(minimise_curvature)) mminimise_curvature = minimise_curvature
        mminimise_deviations_from_canonical_helix = .false.
        if (present(minimise_deviations_from_canonical_helix)) then
            mminimise_deviations_from_canonical_helix = minimise_deviations_from_canonical_helix
        endif
        mminimise_deviations_from_ideal_path_lengths = .false.
        if (present(minimise_deviations_from_ideal_path_lengths)) then
            mminimise_deviations_from_ideal_path_lengths = minimise_deviations_from_ideal_path_lengths
        endif

        if (count([ mminimise_curvature,mminimise_deviations_from_canonical_helix,   &
                    MMINIMISE_DEVIATIONS_FROM_IDEAL_PATH_LENGTHS]) .gt. 1               ) then
            write(*,'(a)') '**ERROR(filament_minimise): can only minimise one property at a time!'
            call terminate('Fatal programming error')
        endif

        ! How many dimensions will the minimiser work with?
        num_dim = count(self%refine_params) * count(self%wp%refine)

!        ! Sanity check
!        if (mminimise_deviations_from_ideal_path_lengths) then
!            if (num_dim .NE. self%nwp) then
!                write(*,'(a,i0,1x,i0)') '**ERROR(filament_minimise): num_dim .ne. self%nwp: ', num_dim, self%nwp
!                write(*,'(a,7(l1,1x))') '**ERROR(filament_minimise): refine_params: ', self%refine_params
!                call terminate('Failed sanity check in FILAMENT_MINIMISE')
!            endif
!        endif

        ! Allocate memory for the result and the start / first step values
        allocate(best_values(num_dim),start_values(num_dim),first_step(num_dim),accuracy(num_dim),stat=IERR)
        if (ierr .ne. 0) call terminate('Memory allocation failed')

        ! Set the starting position and the first step for the ds
        if (mminimise_deviations_from_ideal_path_lengths) then
            !start_values = self%wp%coo(3)
            call values_from_filament(self,start_values,self%refine_params, &
                                        incremental_values=[.false.,.false.,.true.,.false.,.false.,.false.,.false.,.false.,.false.])
            first_step = 20.0
            accuracy=0.5
        else
            call terminate('Not implemented properly')
            start_values = self%wp(1:3)%coo(3)
            first_step = 10.0
            accuracy=0.5
        endif

        ! Work out a reasonable value for the ds target size
        write(*,'(a)') '**warning(filament_minimise): setting the ds target size to arbitrary value of 0.1'
        ds_target_size = 0.5

        ! Initialise the downhill simplex minimiser
        if (mminimise_curvature) then
            call ds_init(   ds_minimiser,filament_eval_total_curvature_wrap,num_dim,self%cptr,  &
                            initial_score=REAL(filament_eval_total_curvature(self)) )
        else if (mminimise_deviations_from_ideal_path_lengths) then
			if (.not. path_lengths_cg) then
            	call ds_init(   ds_minimiser,fil_dev_from_canon_path_lengths_wrap,num_dim,self%cptr,  &
                	            initial_score=-filament_deviation_from_ideal_path_lengths(self), &
							   random = .false.	)
			else
            !print *, 'START_VALUES  = ', start_values
            !print *, 'ACCURACY      = ', accuracy
                call cg_minim%Init(num_dim,fil_dev_from_canon_path_lengths_wrap,self%cptr, &
                                starting_values=start_values, &
            	                accuracy=accuracy )
			endif
        else if (mminimise_deviations_from_canonical_helix) then
            call ds_init(   ds_minimiser,fil_dev_from_canon_helix_wrap,num_dim,self%cptr,   &
                            initial_score=filament_discrepency_btw_wps_phi_and_canonical_helix(self) )
        endif



        ! debug
        !print *, '**debug(filament_minimise): ds_max_iterations = ', ds_max_iterations
        !print *, '**debug(filament_minimise): self pointer associated? = ', c_associated(self%cptr)
        !print *, '**debug(filament_minimise): self refine_params  = ', self%refine_params


        if (mminimise_deviations_from_ideal_path_lengths .and. path_lengths_cg) then
            call cg_minim%Run()
            best_values = self%wp%coo(3)
            best_score = filament_deviation_from_ideal_path_lengths(self)
        else
            !print *, '**debug(filament_minimise): self refine_params 2a= ', self%refine_params
            !call c_f_pointer(self%cptr,tmpfil)
            !print *, '**debug(filament_minimise): self refine_params 2b=', tmpfil%refine_params
            call ds_set_start_step(ds_minimiser,start_values,first_step)
            !print *, '**debug(filament_minimise): self refine_params 3= ', self%refine_params

            ! Run the minimiser
            call ds_run(ds_minimiser,ds_target_size)
            ! Remember the best values
            best_values =   ds_minimiser%current_best_fgsl
            best_score  =   ds_minimiser%current_minimum_fgsl
        endif



        if (debug) then
           write(num_dim_c,'(i2)') num_dim
           write(*,'(a,'//num_dim_c//'(F0.3,1x))') '**debug(filament_minimise): best values = ', best_values
           write(*,'(a,F0.3)') '**debug(filament_minimise): best score = ', best_score
        endif

        ! Set the filament to the best values
        call values_to_filament(self,best_values,self%refine_params,    &
                                incremental_values=[.false.,.false.,.true.,.false.,.false.,.false.,.false.,.false.,.false.])
        call filament_update_model(self)

        if (debug) then
           write(*,'(a)') '**debug(filament_minimise): filament info at end:'
           call self%PrintInfo()
        endif
    end subroutine filament_minimise

    function fil_dev_from_canon_path_lengths_wrap(values,params) bind(c)
        use, intrinsic  ::  iso_c_binding
        use fgsl, only : fgsl_vector, fgsl_double, fgsl_vector_align, fgsl_success, fgsl_obj_c_ptr
        use filaments_geometry, only : filament_deviation_from_ideal_path_lengths
        use filaments_model_update, only : filament_update_model
        use runtime_parameters, only : terminate
        use filaments, only : values_to_filament
        implicit none
        ! Arguments
        type(c_ptr),        value       ::  values  !<  c pointer to current values to try out
        type(c_ptr),        value       ::  params  !<  c pointer to a filament object
        ! Result
        real(kind=C_DOUBLE)             ::  fil_dev_from_canon_path_lengths_wrap
        ! Private variables
        integer                         ::  status
        type(fgsl_vector)               ::  val_vec
        real(fgsl_double),  pointer     ::  val_fgsl(:)
        real(kind=8),       allocatable ::  val(:)
        real(kind=8),       pointer     ::  val_ptr(:)
        type(filament),     pointer     ::  self
        logical,            parameter   ::  debug   =   .false.
        character(len=6)                ::  size_val_c
        logical,            parameter   ::  input_vals_are_fgsl = .true.
        integer                         ::  num_dim
        ! Start work

        ! "Unpack" the filament object
        call c_f_pointer(params,self)
        if (debug) print *, '**debug(fil_dev_from_canon_path_lengths_wrap): refine_params = ', self%refine_params
        num_dim = count(self%refine_params)*SELF%nwp

        ! Unpack the values to try
        if (input_vals_are_fgsl) then
            call fgsl_obj_c_ptr(val_vec,values)
            status = fgsl_vector_align(val_fgsl,val_vec)
            if (status .ne. fgsl_success) call terminate('FGSL_VECTOR_ALIGN failed')
            allocate(val(size(val_fgsl)),stat=STATUS)
            if (status .ne. 0) call terminate('Memory allocation failed')
            val = val_fgsl
        else
            call c_f_pointer(values,val_ptr,SHAPE=[NUM_DIM])
            allocate(val(num_dim))
            val = val_ptr
        endif

        ! Let the user know what values we're trying
        if (debug) then
            write(size_val_c,'(i6)') size(val)
            write(*,'(a,'//size_val_c//'(F0.3,1x))') '**debug(fil_dev_from_canon_path_lengths_wrap): trying these values: ', val
        endif

        ! Assign those values to the waypoints
        call values_to_filament(self,val,self%refine_params,    &
                                incremental_values=[.false.,.false.,.true.,.false.,.false.,.false.,.false.,.false.,.false.])

        if (debug) then
            write(*,'(a)') '**debug(fil_dev_from_canon_path_lengths_wrap): filament info after values_to_filament:'
            call self%PrintInfo()
            write(*,'(a)') '**debug(fil_dev_from_canon_path_lengths_wrap): about to call filament_update_model'
        endif

        ! Update the filament model
        call filament_update_model(self)

        if (debug) then
            write(*,'(a)') '**debug(fil_dev_from_canon_path_lengths_wrap): filament info after update model: '
            call self%PrintInfo()
        endif

        !
        fil_dev_from_canon_path_lengths_wrap = filament_deviation_from_ideal_path_lengths(self)

        if (debug) then
            write(*,'(a,(G15.6))') '**debug(fil_dev_from_canon_path_lengths_wrap): got this score: ', &
                                    fil_dev_from_canon_path_lengths_wrap
        endif
    end function fil_dev_from_canon_path_lengths_wrap



    function fil_dev_from_canon_helix_wrap(values,params) bind(c)
        use, intrinsic  ::  iso_c_binding
        use fgsl, only : fgsl_vector, fgsl_double, fgsl_vector_align, fgsl_success, fgsl_obj_c_ptr
		use filaments_geometry, only : filament_discrepency_btw_wps_phi_and_canonical_helix
		use filaments_model_update, only : filament_update_model
		use runtime_parameters, only : terminate
		use filaments, only : values_to_filament
        implicit none
        ! Arguments
        type(c_ptr),        value       ::  values  !<  c pointer to current values to try out
        type(c_ptr),        value       ::  params  !<  c pointer to a filament object
        ! Result
        real(kind=C_DOUBLE)             ::  fil_dev_from_canon_helix_wrap
        ! Private variables
        integer                         ::  status
        type(fgsl_vector)               ::  val_vec
        real(fgsl_double),  pointer     ::  val_fgsl(:)
        real(kind=8),       allocatable ::  val(:)
        type(filament),     pointer     ::  self
        logical,            parameter   ::  debug   =   .false.
        character(len=6)                ::  size_val_c
        ! Start work

        ! "Unpack" the filament object
        call c_f_pointer(params,self)

        ! Unpack the values to try
        call fgsl_obj_c_ptr(val_vec,values)
        status = fgsl_vector_align(val_fgsl,val_vec)
        if (status .ne. fgsl_success) call terminate('FGSL_VECTOR_ALIGN failed')
        allocate(val(size(val_fgsl)),stat=STATUS)
        if (status .ne. 0) call terminate('Memory allocation failed')
        val = val_fgsl

        ! Let the user know what values we're trying
        if (debug) then
            write(size_val_c,'(i6)') size(val)
            write(*,'(a,'//size_val_c//'(F0.3,x))') '**debug(fil_dev_from_canon_helix_wrap): trying these values: ', val
        endif

        ! Assign those values to the waypoints
        call values_to_filament(self,val,self%refine_params)

        ! Update the filament model
        call filament_update_model(self)

        !
        fil_dev_from_canon_helix_wrap = filament_discrepency_btw_wps_phi_and_canonical_helix(self)

        if (debug) then
            write(*,'(a,(G15.6))') '**debug(fil_dev_from_canon_helix_wrap): got this score: ', fil_dev_from_canon_helix_wrap
        endif
    end function fil_dev_from_canon_helix_wrap

!   function filament_eval_phasing_stdev_wrap(values,params) bind(c)
!       use, intrinsic  ::  iso_c_binding
!       use runtime_parameters
!       implicit none
!       ! Arguments
!       type(c_ptr),        value       ::  values  !<  c pointer to current values to try out
!       type(c_ptr),        value       ::  params  !<  c pointer to a filament object
!       ! Result
!       real(KIND=C_DOUBLE)             ::  filament_eval_phasing_stdev_wrap
!       ! Private variables
!       integer                         ::  status
!       type(fgsl_vector)               ::  val_vec
!       real(fgsl_double),  pointer     ::  val_fgsl(:)
!       real(KIND=8),       allocatable ::  val(:)
!       type(filament),     pointer     ::  self
!       real(KIND=8)                    ::  ave_std_phasing(2)
!       logical,            parameter   ::  debug   =   .TRUE.
!       ! Start work
!
!       ! "Unpack" the filament object
!       call c_f_pointer(params,self)
!
!       ! Unpack the values to try
!       call fgsl_obj_c_ptr(val_vec,values)
!       status = fgsl_vector_align(val_fgsl,val_vec)
!       if (status .NE. fgsl_success) call terminate('FGSL_VECTOR_ALIGN failed')
!       allocate(val(size(val_fgsl)),STAT=STATUS)
!       if (status .NE. 0) CALL terminate('Memory allocation failed')
!       val = val_fgsl
!
!       ! Let the user know what values we're trying
!       if (debug) then
!           write(*,'(a,<SIZE(val)>(F0.3,x))') '**debug(filament_eval_phasing_stdev_wrap): trying these values: ', val
!       endif
!
!       ! Assign those values to the waypoints
!       call values_to_filament(self,val,self%refine_params)
!
!       ! Update the filament model
!       call filament_update_model(self)
!
!       ! Compute the average and standard deviation phasing
!       ave_std_phasing = filament_compute_ave_std_wp_phasing(self)
!       filament_eval_phasing_stdev_wrap = ave_std_phasing(2)
!
!       if (debug) then
!           write(*,'(a,(G0.6))') '**debug(filament_eval_phasing_stdev_wrap): got this score: ', filament_eval_phasing_stdev_wrap
!       endif
!   end function filament_eval_phasing_stdev_wrap


    !>  \brief  Wrapper to be called by the ds minimiser
    function filament_eval_total_curvature_wrap(values,params) bind(c)
        use, intrinsic  ::  iso_c_binding
        use fgsl, only : fgsl_vector, fgsl_double, fgsl_obj_c_ptr, fgsl_vector_align, fgsl_success
		use filaments_geometry, only : filament_eval_total_curvature
		use runtime_parameters, only : terminate
		use filaments_model_update, only : filament_update_model
		use filaments, only : values_to_filament
        implicit none
        ! Arguments
        type(c_ptr),        value       ::  values  !<  c pointer to current values to try out
        type(c_ptr),        value       ::  params  !<  c pointer to a filament object
        ! Result
        real(kind=C_DOUBLE)             ::  filament_eval_total_curvature_wrap
        ! Private variables
        integer                         ::  status
        type(fgsl_vector)               ::  val_vec
        real(fgsl_double),  pointer     ::  val_fgsl(:)
        real(kind=8),       allocatable ::  val(:)
        type(filament),     pointer     ::  self
        logical,            parameter   ::  debug   =   .true.
        character(len=6)                ::  size_val_c
        ! Start work

        ! "Unpack" the filament object
        call c_f_pointer(params,self)

        ! Unpack the values to try
        call fgsl_obj_c_ptr(val_vec,values)
        status = fgsl_vector_align(val_fgsl,val_vec)
        if (status .ne. fgsl_success) call terminate('FGSL_VECTOR_ALIGN failed')
        allocate(val(size(val_fgsl)),stat=STATUS)
        if (status .ne. 0) call terminate('Memory allocation failed')

        ! Let the user know what values we're trying
        if (debug) then
            write(size_val_c,'(i6)') size(val)
            write(*,'(a,'//size_val_c//'(F0.3,x))') '**debug(filament_eval_total_curvature_wrap): trying these values: ', val
        endif

        ! Assign those values to the waypoints
        call values_to_filament(self,val,self%refine_params)

        ! Update the filament model
        call filament_update_model(self)

        ! Compute the total curvature
        filament_eval_total_curvature_wrap = filament_eval_total_curvature(self)

        if (debug) then
            write(*,'(a,(G15.6))') '**debug(filament_eval_total_curvature_wrap): got this score: ', &
                                    filament_eval_total_curvature_wrap
        endif
    end function filament_eval_total_curvature_wrap

end module filaments_minimisation
