!>  \brief  Methods dealing with the internal representation of the physical model for a filament
module filaments_model_base
    use filaments_base
    use runtime_parameters, only : terminate
    implicit none

    contains



    !>  \brief  Print out extra information about the filament
    subroutine filament_print_extra_info(self)
        use filaments_geometry, only : filament_eval_total_curvature, filament_eval_change_of_direction
        use filaments_geometry, only : filament_eval_total_curvature_overtwist_stretch, filament_deviations_from_canonical_helix
        use filaments_geometry, only : filament_ideal_path_lengths_for_canonical_helix, filament_deviation_from_ideal_path_lengths
        use filaments_parametrisation, only : filament_parametrisation_error
        use filaments_geometry, only : filament_maximal_normal_distance, filament_chord_length, filament_mid_point_normal_distance
        implicit none
        ! Arguments
        type(filament), intent(inout)  ::  self
        ! Private variables
        real             ::  helical_stats(4), param_error, rigid_t, xover_stats(2)
        real             ::  max_normal_dist, chord_length_3d,chord_length_2d
        real             ::  mid_point_normal_dist_3d, mid_point_normal_dist_2d
        real(kind=8)     ::  curvature, overtwist,stretch
        character(len=3) ::  nwpc
        ! Start work
        write(nwpc,'(i3)') self%nwp
        helical_stats = filament_eval_helical_parameters_stats(self)
        param_error = filament_parametrisation_error(self)
        rigid_t = filament_get_t_for_rigid_wp(self)
        write(*,'(/a,i0,a)')                 'Extra info for filament ', self%id, ':'
        write(*,'(a,F0.6,1x,F0.3,1x,F0.3)')  '  parametrisation error, t(1), t(rigid): ', param_error, self%t(1), rigid_t
        write(*,'(a,G15.4)')                 '  total curvature (1/pixels) = ', filament_eval_total_curvature(self)
        write(*,'(a,G15.4)')                 '  total change of direction (rad) = ', filament_eval_change_of_direction(self)
        write(*,'(2(A,G15.4),a)')            '  average (stdev) twist per subunit (degrees): ', helical_stats(1), ' (', &
                                                                                                helical_stats(2), ')'
        write(*,'(2(A,G15.4),a)')            '  average (stdev) rise per subunit (pixels): ', helical_stats(3), ' (',    &
                                                                                            helical_stats(4), ')'
        write(*,'(a,'//nwpc//'(F0.3,1x))')   '  ideal path lengths for canonical helix: ', &
                                                            filament_ideal_path_lengths_for_canonical_helix(self)
        write(*,'(a,F0.3)')                  '  deviation from ideal path lenghts for canonical helix: ', &
                                                            filament_deviation_from_ideal_path_lengths(self)
        if (self%amyloid_beta) then
            xover_stats = filament_crossover_stats(self)
            write(*,'(2(A,G15.4),a)')        '  average (stdev) crossover-to-crossover distance (pixels): ', xover_stats(1), &
                                                                                                    ' (', xover_stats(2), ')'
        endif
        call filament_eval_total_curvature_overtwist_stretch(self,curvature,overtwist,stretch,  &
                                                     IGNORE_FAL_WPS=.not. (self%fal_wps_in_score .and. self%fal_wps_in_threed))
        write(*,'(a,3(G15.4,1x))')            '  total curvature, overtwist, stretch (1/pixels) : ', curvature, overtwist, stretch
        call filament_maximal_normal_distance(self,max_normal_dist)
        call filament_mid_point_normal_distance(self,mid_point_normal_dist_3d,mid_point_normal_dist_2d)
        call filament_chord_length(self,chord_length_3d,chord_length_2d)
        write(*,'(a,2(F0.3,1x))')             '  chord length 3d, 2d (pixels): ', chord_length_3d, chord_length_2d
        write(*,'(a,2(F0.3,1x))')             '  mid-point normal distance 3d, 2d (pixels): ',  mid_point_normal_dist_3d, &
                                                                                                mid_point_normal_dist_2d
        write(*,'(a,F0.3)')                   '  maximal normal distance (pixels): ', max_normal_dist
        call filament_deviations_from_canonical_helix(self)
    end subroutine filament_print_extra_info

    !>  \brief  Evaluate the average helical twist and rise per subunit and the standard deviation
    function filament_eval_helical_parameters_stats(self,ignore_fal_wps) result(stats)
        use runtime_parameters, only : terminate, safe_for_sqrt
        use Splines
        use coos, only : waypoint
        use filaments_path, only : filament_get_path_subunit_steps, filament_get_path_subunit_steps_frag_helix
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self
        logical,        optional,   intent(in)      ::  ignore_fal_wps
        ! Result
        real(kind=8)    ::  stats(4)    !<  The average and standard deviation are returned first and second respectively, with 1&2 for the twist and 3&4 for the rise
        ! Private variables
        real(kind=8)    ::  twist_sum, twist_sum2, rise_sum, rise_sum2
        type(waypoint), allocatable ::  s_path(:)
        integer         ::  i
        ! Start work

        if (self%nwp .eq. 1) then
            stats(1) = self%wp(1)%helix%twist
            stats(3) = self%wp(1)%helix%rise
            stats(2) = 0.0d0
            stats(4) = 0.0d0
        else
            if (self%fragmented_phi .or. self%fragmented_subunit_num .or. self%wps_as_sps) then
                call filament_get_path_subunit_steps_frag_helix(self,1,self%nwp,s_path)
            else
                call filament_get_path_subunit_steps(self,s_path,step=1,ignore_fal_wps=ignore_fal_wps)
            endif

            twist_sum = 0.0D0
            twist_sum2 = 0.0D0
            rise_sum = 0.0D0
            rise_sum2 = 0.0D0

            do i=1,size(s_path)
                ! Keep running sum and sum of squares
                twist_sum = twist_sum + s_path(i)%helix%twist
                twist_sum2 = twist_sum2 + s_path(i)%helix%TWIST**2
                rise_sum = rise_sum + s_path(i)%helix%rise
                rise_sum2 = rise_sum2 + s_path(i)%helix%RISE**2
            enddo

            ! Averages
            stats(1) = twist_sum / dble(size(s_path))
            twist_sum2 = twist_sum2 / dble(size(s_path))
            stats(3) = rise_sum / dble(size(s_path))
            rise_sum2 = rise_sum2 / dble(size(s_path))
            ! Variance
            stats(2) = twist_sum2 - stats(1)**2
            stats(4) = rise_sum2 - stats(3)**2
            ! Standard deviation
            call safe_for_sqrt(stats(2),negative_to_zero=.true.)
            call safe_for_sqrt(stats(4),negative_to_zero=.true.)
            stats(2) = sqrt(stats(2))
            stats(4) = sqrt(stats(4))
        endif

    end function filament_eval_helical_parameters_stats


    !>  \brief  Compute the average and standard deviation of crossover-to-crossover distances
    function filament_crossover_stats(self) result(stats)
        use runtime_parameters, only : terminate
        use filaments_path, only : filament_get_path_crossovers
        use coos, only : waypoints_print_info
        implicit none
        ! Arguments
        type(filament),     intent(inout)   ::  self    !<  Filament object
        ! Result
        real    ::  stats(2)
        ! Private variables
        real, allocatable                   ::  distances(:)
        real(kind=8)                        ::  sum_distances, sum_sq_distances
        type(waypoint),     allocatable     ::  xovers(:)
        integer                             ::  current_xover
        integer                             ::  ierr
        logical,            parameter       ::  debug = .false.
        ! Start work

        ! Get the crossovers
        if (self%wps_are_crossovers) then
            if (self%fal_wps_are_crossovers) then
                xovers = self%wp
            else
                xovers = self%wp(2:self%NWP-1)
            endif
        else
            call filament_get_path_crossovers(self,xovers,STEP=1)
        endif

        if (debug) then
            write(*,'(a)') '**debug(filament_crossover_stats): xover info:'
            call waypoints_print_info(xovers)
        endif

        ! Allocate memory for xover-to-xover distances
        if (allocated(distances)) deallocate(distances)
        allocate(distances(size(xovers)-1),stat=IERR)
        if (ierr .ne. 0) call terminate('Failed memory allocation in FILAMENT_CROSSOVER_STATS')

        ! Compute distances
        sum_distances = 0.0D0
        sum_sq_distances = 0.0D0
        do CURRENT_XOVER=1,size(xovers)-1
            distances(current_xover)= xovers(CURRENT_XOVER+1)%path_length - xovers(current_xover)%path_length
            sum_distances = sum_distances + distances(current_xover)
            sum_sq_distances = sum_sq_distances + distances(current_xover)**2
        enddo

        ! Average
        if (size(distances) .ne. 0) then
            stats(1) = sum_distances / size(distances)
            stats(2) = sum_sq_distances / size(distances) - stats(1)**2
            if (stats(2) .ge. 0.0D0) then
                stats(2) = sqrt( stats(2) )
            else
                stats(2) = 0.0D0
            endif
        else
            write(*,'(a,i0)') '**WARNING(filament_crossover_stats): not enough crossovers. Number of intervals = ', size(distances)
            !call terminate('Fatal error in FILAMENT_CROSSOVER_STATS')
        endif
    end function filament_crossover_stats






    !>  \brief  When the helical operator describes an n-start helix with m-fold symmetry,
    !           the number of subunits should be a multiple of n*m. This subroutine checks this is the case.
    subroutine filament_check_subunit_num(self)
        use helices, only : helical_symmetry_print_info
        use Globals
        implicit none
        ! Arguments
        type(filament),     intent(in)  ::  self
        ! Private variables
        integer ::  current_wp
        real    ::  error_relative_to_path_length
        real    ::  error_relative_to_phi
        real    ::  average_error
        real    ::  expected_from_path_length
        real    ::  expected_from_phi
        real    ::  subunit_num_from_start
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(filament_check_subunit_num): subunit_num values on entry:'
            print *, self%wp%subunit_num
        endif

        if (self%nwp .ge. 2) then

            ! Loop over the waypoints
            error_relative_to_path_length   =   0.0
            error_relative_to_phi           =   0.0
            do current_wp=2,self%nwp
                ! Number of subunits from 1st to current waypoint
                subunit_num_from_start = self%wp(current_wp)%subunit_num - self%wp(1)%subunit_num
                ! According to path_length, expected number of subunits (i.e. assuming constant rise per subunit)
                expected_from_path_length =     wp_subunit_num_from_path_length(self%wp(current_wp),self%helix) &
                                            -   wp_subunit_num_from_path_length(self%wp(1),self%helix)
                ! According to phi, expected number of subunits (i.e. assuming constant twist per subunit)
                if (self%helix%twist .eq. 0.0) then
                    ! If the twist is 0.0 (can happen, for example when talking about pf1 of a 13-pf mt), we
                    ! just can't estimate subunit_num
                    expected_from_phi = subunit_num_from_start
                else
                    expected_from_phi =     wp_subunit_num_from_phi(self%wp(current_wp),self%helix) &
                                        -   wp_subunit_num_from_phi(self%wp(1),self%helix)
                endif
                ! Debug output
                if (debug) then
                    write(*,'(a,i0,a,i0,a,i0,a,a,3(F0.3,1x))')  '**debug(filament_check_subunit_num): waypoint ', &
                                                        current_wp, ' of ', self%nwp, &
                                                        ' (filament ', self%id, ') ', &
                                                        ': subunit_num, expected from path length, expected from phi: ', &
                                                        subunit_num_from_start, expected_from_path_length,  expected_from_phi
                endif
                ! Accumulate average error
                error_relative_to_path_length   =   error_relative_to_path_length + expected_from_path_length &
                                                                                    / subunit_num_from_start
                error_relative_to_phi           =   error_relative_to_phi         + expected_from_phi / subunit_num_from_start
            enddo
            ! Average -fold error
            error_relative_to_path_length   =   error_relative_to_path_length / (self%nwp-1) - 1.
            if (self%fragmented_phi) then
                error_relative_to_phi       =   error_relative_to_path_length
            else
                error_relative_to_phi       =   error_relative_to_phi         / (self%nwp-1) - 1.
            endif

            ! Average of the two average errors
            average_error = (error_relative_to_path_length + error_relative_to_phi) / 2.

            if (debug) write(*,'(a,F0.2,a)') '**debug(filament_check_subunit_num): average error = ', average_error, '-fold'

            if (nint(average_error) .ne. 0) then
                write(*,'(a,f0.2,a)') '**info(filament_check_subunit_num): it looks like the subunit_num values are ',  &
                                            average_error, '-fold out compared to expected values based on helical operator'
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(2a)') '**info(filament_check_subunit_num): perhaps the number of starts and/or symmetries in ',    &
                                    'the helical operator were not taken into account when estimating subunit_num values'
                    call helical_symmetry_print_info(self%helix)
                    call self%PrintInfo()
                endif
            endif

        endif


    end subroutine filament_check_subunit_num

    !>  \brief  The subunit_num should increase from waypoint to waypoint, if not crash out
    subroutine filament_check_subunit_num_increases(self)
        implicit none
        type(filament), intent(inout)  ::  self    !<  Filament
        ! Private variables
        integer ::  current_wp
        ! Start work
        do current_wp = 2, self%nwp
            if ( self%wp(current_wp)%subunit_num .lt. self%wp(current_wp-1)%subunit_num) then
                if (self%fragmented_subunit_num) then
                    ! Because fragmented, we can add integer numbers of subunits
                    self%wp(current_wp)%subunit_num = ceiling(self%wp(current_wp-1)%subunit_num) + 1.0
                else
                    write(*, '(a)') '**ERROR(filament_check_subunit_num_increases): subunit_num values should increase'
                    call self%PrintInfo()
                    call terminate('Fatal error in FILAMENT_CHECK_SUBUNIT_NUM_INCREASES')
                endif
            endif
        end do
    end subroutine filament_check_subunit_num_increases


    !>  \brief  Based on the waypoint's phi, return the subunit_num predicted by the helical parameters, assuming
    !!          constant twist per subunit.
    !!
    !!  Note that by convention, waypoints' "subunit_num" value is the number of subunits along the 1-start helix, without
    !!  counting symmetry mates.
    !!
    !!  The helical parameters given could either be the canonical or local ones
    pure real function wp_subunit_num_from_phi(self,helix) result(subunit_num)
        implicit none
        ! Arguments
        type(waypoint),         intent(in)  ::  self
        type(helical_symmetry), intent(in)  ::  helix
        ! Private variables

        ! Start work
        subunit_num = self%euler%phi / helix%twist
    end function wp_subunit_num_from_phi

    !>  \brief  Based on the waypoint's path length, return the subunit_num predicted by the helical parameters,
    !!          assuming constant rise per subunit.
    !!
    !!  The helical parameters given could either be the canonical or local ones
    pure real function wp_subunit_num_from_path_length(self,helix) result(subunit_num)
        implicit none
        ! Arguments
        type(waypoint),         intent(in)  ::  self
        type(helical_symmetry), intent(in)  ::  helix
        ! Private variables

        ! Start work
        subunit_num = self%path_length / helix%rise
    end function wp_subunit_num_from_path_length

    !>  \brief  Compute the subunit_num shift to the nearest integer subunit number for the given waypoint.
    !!
    !!  By convention, the reference 3d model has the origin subunit at phi=0.0, so that s=0.0. Any rotation around the helical axis (phi angle)
    !!  still leaves a subunit exactly at the origin of a projection image. Therefore, if a waypoint as a non-integer subunit_num value, the
    !!  reference volume will need to be "phase shifted" to match it, i.e. a screw operation will need to be applied to the reference.
    !!  To work out the smallest possible screw operation that would need to be performed, we first need to work out the shift in subunit_num
    !!  value to the nereast integer subunit.
    pure real function wp_compute_subunit_num_phasing(wp) result(subunit_num_shift)
        implicit none
        ! Arguments
        type(waypoint),                 intent(in)      ::  wp
        ! Private variables
        ! Start work
        subunit_num_shift = wp%subunit_num - nint(wp%subunit_num)
    end function wp_compute_subunit_num_phasing



    !>  \brief  Return the t parameter at which we conceptually place the fil_as_sp_wp
    pure function filament_get_t_for_rigid_wp(self) result(t)
        implicit none
        ! Arguments
        type(filament),             intent(in)  ::  self
        ! Result
        real(kind=8)                            ::  t
        ! Private variables

        ! Start work

        ! We chose to "position" the virtual rigid waypoint at the midpoint of the space curve
        t = self%t(1) + (self%t(self%nwp) - self%t(1)) / 2.
    end function filament_get_t_for_rigid_wp

    !>  \brief  Estimate the Phi angles of waypoints in the absence of a model, but given helical parameters and t parameters for each waypoints
    !!
    subroutine filament_guess_wps_phi(self)
        use helices, only : helical_symmetry_unit_convert
        implicit none
        ! Arguments
        type(filament),     intent(inout)       ::  self    !<  Filament whose waypoints are to be updated with new phi values
        ! Private variables
        logical,    parameter   ::  debug   =   .false.
        integer                 ::  twist_sign, i
        ! Start work

        ! Check helical parameters are in pixel units
        if (.not. self%helix%pixel_units) then
            !write(*,'(a)') '**ERROR(filament_estimate_wps_phi): filament helical parameters not given in pixel units!'
            !call terminate('Fatal error in FILAMENT_GUESS_WPS_PHI')
            call helical_symmetry_unit_convert(self%helix,self%film%GetPixelSize())
        endif

        ! Check waypoints have t values
        if (all(self%t .eq. 0.0)) then
            write(*,'(a)') '**ERROR(filament_estimate_wps_phi): filament was not parametrised - all t values are 0'
            call terminate('Fatal error in FILAMENT_GUESS_WPS_PHI')
        endif

        !
        ! What is the sign of the twist?
        !
        if (mod(self%helix%twist,180.D0) .gt. 0.D0) then
            twist_sign = 1
        else
            twist_sign = -1
        endif
        if (debug) write(*,'(a,F8.2,1x,i2)') '**debug(filament_guess_wps_phi): twist, twist_sign = ', self%helix%twist, twist_sign

        if (debug) write(*,'(a)') '**debug(filament_guess_wps_phi): guessing phi values for waypoints'
        if (self%wps_are_crossovers) then
            if (debug) write(*,'(a)') '**debug(filament_guess_wps_phi): waypoints are crossovers'
            if (self%fal_wps_are_crossovers) then
                self%wp(1)%euler%phi = 0.
                do I=2,self%nwp
                    self%wp(i)%euler%phi = self%wp(I-1)%euler%phi + 180.0 * twist_sign
                enddo
            else
                if (debug) write(*,'(a)') '**debug(filament_guess_wps_phi): first and last waypoints are not crossovers'
                self%wp(2)%euler%phi = 0.
                do I=3,self%NWP-1
                    self%wp(i)%euler%phi = self%wp(I-1)%euler%phi + 180.0 * twist_sign
                enddo
                call filament_guess_fal_wps(self,guess_fal_euler_only=.true.)!,KNOWN_PSI=SELF%x_spline%known_d .AND. self%y_spline%known_d,KNOWN_THETA=SELF%z_spline%known_d)
            endif
        else
            self%wp%euler%phi   =   self%t / self%helix%rise * self%helix%twist
        endif

    end subroutine filament_guess_wps_phi


    !>  \brief  Find the integer subunit_num value nearest to the given waypoint, making sure that subunit is within the filament
    integer function filament_find_nearest_subunit_to_waypoint(self,wp) result(nearest_subunit_num)
        implicit none
        ! Arguments
        type(filament),                 intent(in)      ::  self
        type(waypoint),                 intent(in)      ::  wp
        ! Private variables

        ! Start work
        nearest_subunit_num = nint(wp%subunit_num)
        nearest_subunit_num = max(nearest_subunit_num,ceiling(self%wp(1)%subunit_num))
        nearest_subunit_num = min(nearest_subunit_num,floor(self%wp(self%nwp)%subunit_num))
    end function filament_find_nearest_subunit_to_waypoint

    !>  \brief  For each waypoint in a filament, "guess"/estimate the subunit number
    !!
    !!  The keep_rigid_subunit_num option allows the subunit number values of each waypoints to be recalculated assuming constant rise per subunit while
    !!  (for example) refining the rigid subunit_num value.
    subroutine filament_guess_wps_subunit_num(  self,from_rigid_subunit_num,  &
                                                constant_rise_per_subunit,constant_twist_per_subunit,intg_subunit_num   )
        use helices, only : helical_symmetry_unit_convert
        use filaments_geometry, only : filament_ensure_subsequent_subunit_num_values_not_equal
        implicit none
        ! Arguments
        type(filament),                 intent(inout)   ::  self                        !<  Filament whose waypoints are to be updated with new subunit_num values
        logical,            optional,   intent(in)      ::  from_rigid_subunit_num      !<  If true, the subunit_num values will be recalculated for each waypoint while keeping the rigid waypoint's subunit_num value unchanged. Defaults to true.
        logical,            optional,   intent(in)      ::  constant_rise_per_subunit   !<  If true, we assume the rise per subunit is constant. Defaults to true.
        logical,            optional,   intent(in)      ::  constant_twist_per_subunit  !<  If true, we assume the twist per subunit is constant. Defaults to ???.
        logical,            optional,   intent(in)      ::  intg_subunit_num            !<  If true, the subunit_num value for each waypoint will necessarily be integer
        ! Private variables
        logical                 ::  ffrom_rigid_subunit_num
        logical                 ::  cconstant_rise_per_subunit
        logical                 ::  cconstant_twist_per_subunit
        logical                 ::  iintg_subunit_num
        integer                 ::  current_wp
        logical,    parameter   ::  debug   =   .false.
        !real                    ::  dist_to_next_wp, expected_change_in_phi, actual_change_in_phi
        ! Start work

        ! Will we be leaving the rigid subunit_num value unchanged?
        if (present(from_rigid_subunit_num)) then
            ffrom_rigid_subunit_num = from_rigid_subunit_num
        else
            ffrom_rigid_subunit_num = .true.
        endif

        ! Constant rise per subunit?
        if (present(constant_rise_per_subunit)) then
            cconstant_rise_per_subunit = constant_rise_per_subunit
        else
            cconstant_rise_per_subunit = self%constant_rise_per_subunit
        endif

        ! Constant twist per subunit
        if (present(constant_twist_per_subunit)) then
            cconstant_twist_per_subunit = constant_twist_per_subunit
        else
            cconstant_twist_per_subunit = self%constant_rise_per_subunit .and. self%constant_twist_per_length
        endif

        ! Will we assign integer values for subunit_num?
        ! This can be useful for example when assigning subunit_num values to waypoints which were previously refined as single particles
        if (present(intg_subunit_num)) then
            iintg_subunit_num = intg_subunit_num
        else
            iintg_subunit_num = .false.
        endif


        ! It is not possible to garantee constant rise per subunit and integer subunit_num values together, since this subroutine cannot move waypoints
        if (cconstant_rise_per_subunit .and.  iintg_subunit_num) then
            write(*,'(3a)') '**warning(filament_guess_wps_subunit_num): it is not possible to find integer subunit_num', &
                            ' values while forcing constant rise per subunit. Integer subunit_num values will be',  &
                            ' returned while keeping as close as possible to a constant rise per subunit.'
        endif

        ! Check the helix is in pixel units
        if (.not. self%helix%pixel_units) call helical_symmetry_unit_convert(self%helix,self%film%GetPixelSize())


        ! Start the actual work

        if (cconstant_twist_per_subunit .and. self%helix%twist .ne. 0.) then
            if (debug) write(*,'(2a)') '**debug(filament_guess_wps_subunit_num): calculating subunit_num values for all',   &
                                        ' waypoints assuming constant twist per subunit'
            do current_wp=1,self%nwp
                self%wp(current_wp)%subunit_num = wp_subunit_num_from_phi(self%wp(current_wp),self%helix)
            enddo
        else if (cconstant_rise_per_subunit .or. all(self%wp%subunit_num .eq. 0.0)) then
            if (debug) write(*,'(a)') '**debug(filament_guess_wps_subunit_num): calculating subunit_num values for all',    &
                                    ' waypoints assuming constant rise per subunit'
            do current_wp=1,self%nwp
                self%wp(current_wp)%subunit_num = wp_subunit_num_from_path_length(self%wp(current_wp),self%helix)
            enddo
        else
            !write(*,'(a)') '**ERROR(filament_update_wps_subunit_num): neither constant rise nor constant twist per subunit was specified'
            !call terminate('Fatal error in FILAMENT_UPDATE_WPS_SUBUNIT_NUM')
            !do CURRENT_WP=1,self%NWP-1
            !    ! Work out the distance to the next waypoint
            !    dist_to_next_wp = self%wp(CURRENT_WP+1)%path_length - self%wp(current_wp)%path_length
            !    ! Work out what change in phi this should correspond to
            !    expected_change_in_phi = dist_to_next_wp * self%helix%twist / self%helix%rise
            !    ! What change in phi do we actually have?
            !    actual_change_in_phi = self%wp(CURRENT_WP+1)%euler%phi - self%wp(current_wp)%euler%phi
            !
            !    !
            !    self%wp(CURRENT_WP+1:self%nwp)%subunit_num =  -NINT((ACTUAL_CHANGE_IN_PHI-EXPECTED_CHANGE_IN_PHI)/self%helix%twist) + &
            !                                        self%wp(current_wp)%subunit_num
            !    print *, '**DEBUG(): change in subunit_num = ', -NINT((ACTUAL_CHANGE_IN_PHI-EXPECTED_CHANGE_IN_PHI)/self%helix%twist)
            !enddo
        endif

        if (iintg_subunit_num) then
            if (ffrom_rigid_subunit_num) then
                write(*,'(a)') '**warning(filament_guess_wps_subunit_num): rouding of all subunit_num values will mean',    &
                                ' that the rigid subunit_num value is not necessarily preserved'
            endif
            if (debug) write(*,'(a)') '**debug(filament_guess_wps_subunit_num): rounding all subunit_num values to nearest integer'
            self%wp%subunit_num = anint(self%wp%subunit_num)
            ! Check against two subsequent equal values
            call filament_ensure_subsequent_subunit_num_values_not_equal(self)
        endif

        if (ffrom_rigid_subunit_num) then
            if (debug) write(*,'(3a,F0.3,a)') '**debug(filament_guess_wps_subunit_num): recalculating subunit_num values',   &
                                                ' for all waypoints to take into account subunit_num value',    &
                                                ' of rigid waypoint (', self%rigid_wp%subunit_num, ')'
            self%wp%subunit_num = self%wp%subunit_num - (sum(self%wp%subunit_num) / real(self%nwp)) + self%rigid_wp%subunit_num
        endif

        ! It's nicer to have positive values throughout
        if (self%wp(1)%subunit_num .lt. 0.0001) then
            self%wp%subunit_num = self%wp%subunit_num - floor(self%wp(1)%subunit_num)
        endif

        if (debug) then
            write(*,'(a)') '**debug(filament_guess_wps_subunit_num): output subunit_num values: '
            print *, self%wp%subunit_num
        endif

    end subroutine filament_guess_wps_subunit_num



    !>  \brief  Guess the coordinates and/or angles of the first and last waypoints of a filament.
    subroutine filament_guess_fal_wps(self,guess_fal_euler_only)
        use splines
        use filaments_geometry, only : filament_solve_splines
        use differential_geometry, only : euler_from_derivatives_new
        use Globals
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self                    !<  Filament object
        logical,        intent(in)      ::  guess_fal_euler_only    !<  The coordinates for the fal WPs are already known, we only need to guess their Euler angles. By default .FALSE.
        ! Private variables
        type(interpol_spline)   ::  tmp_x_spline, tmp_y_spline, tmp_z_spline, tmp_s_spline, tmp_twist_spline
        integer                 ::  i
        real(kind=8)            ::  theta_tmp, psi_tmp
        logical,        parameter   ::  debug   =   .false.
		logical						::	hermite_bak
		type(waypoint)              ::  wps_on_entry(self%nwp)
		integer                     ::  lbounds(2), ubounds(2)
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(filament_guess_fal_wps): filament info on entry'
            call self%PrintInfo()
        endif


        ! There is a special case when we just need to work out Euler angles and we have less than 5 waypoints
        if (guess_fal_euler_only .and. self%nwp .eq. 4) then
            self%wp(1)%euler = self%wp(2)%euler
            self%wp(1)%euler%phi = self%wp(2)%euler%phi &
                                - (self%wp(2)%path_length - self%wp(1)%path_length) &
                                * (self%wp(3)%euler%phi - self%wp(2)%euler%phi) &
                                / (self%wp(3)%path_length - self%wp(2)%path_length)
            !
            self%wp(4)%euler = self%wp(3)%euler
            self%wp(4)%euler%phi = self%wp(3)%euler%phi &
                                + (self%wp(4)%path_length - self%wp(3)%path_length) &
                                * (self%wp(3)%euler%phi - self%wp(2)%euler%phi) &
                                / (self%wp(3)%path_length - self%wp(2)%path_length)
        else if (guess_fal_euler_only .and. self%nwp .eq. 3) then
            self%wp(1)%euler = self%wp(2)%euler
            self%wp(1)%euler%phi = self%wp(2)%euler%phi &
                                - (self%wp(2)%path_length - self%wp(1)%path_length) &
                                * self%helix%twist / self%helix%rise
            !
            self%wp(3)%euler = self%wp(2)%euler
            self%wp(3)%euler%phi = self%wp(2)%euler%phi &
                                + (self%wp(3)%path_length - self%wp(2)%path_length) &
                                * self%helix%twist / self%helix%rise
        else

        ! This will only work with >= 5 waypoints
        if (self%nwp .lt. 5) then
           write(*,'(a,i0)') '**ERROR(filament_guess_fal_wps): need at least 5 waypoints, but only have ', self%nwp
           call terminate('Fatal error in FILAMENT_GUESS_FAL_WPS')
        endif

        ! Back up the waypoints on entry
        wps_on_entry = self%wp

        ! Initialise the "X" values for the splines
        self%x_spline%x     = self%t
        self%y_spline%x     = self%t
        self%z_spline%x     = self%t
        self%s_spline%x     = self%t
        self%twist_spline%x = self%t

        ! We assume that we haven't solved the splines yet and therefore don't have weights to calculate the x,y,z values
        ! One solution is to create temporary spline objects without the extra first/last waypoints, solve those,
        ! and then work out the coos for the extra first/last.
        if (this_program%IsVerbose() .or. debug) then
            write(*,'(2a)') '**warning(filament_guess_fal_wps): options to guess fal wp parameters are experimental',    &
                            ' - please check results carefully'
        endif
        ! Create temporary spline objects
        if (.not. guess_fal_euler_only) then
            call tmp_x_spline%Allocate(   self%nwp-2,   self%x_spline%Degree())
            call tmp_y_spline%Allocate(   self%nwp-2,   self%y_spline%Degree())
            call tmp_z_spline%Allocate(   self%nwp-2,   self%z_spline%Degree())
            call tmp_s_spline%Allocate(   self%nwp-2,   self%s_spline%Degree())
        endif
        call tmp_twist_spline%Allocate( self%nwp-2,   self%twist_spline%Degree())
        ! Copy contents of filament splines to temporary splines (except of course first and last)
        if (.not. guess_fal_euler_only) then
            tmp_x_spline%x(1:self%nwp-2)     = self%t(2:self%nwp-1)
            tmp_y_spline%x(1:self%nwp-2)     = self%t(2:self%nwp-1)
            tmp_z_spline%x(1:self%nwp-2)     = self%t(2:self%nwp-1)
            tmp_s_spline%x(1:self%nwp-2)     = self%t(2:self%nwp-1)
            tmp_x_spline%y(1:self%nwp-2)     = self%wp(2:self%nwp-1)%coo(1)
            tmp_y_spline%y(1:self%nwp-2)     = self%wp(2:self%nwp-1)%coo(2)
            tmp_z_spline%y(1:self%nwp-2)     = self%wp(2:self%nwp-1)%coo(3)
            tmp_s_spline%y(1:self%nwp-2)     = self%wp(2:self%nwp-1)%subunit_num
        endif
        tmp_twist_spline%x(1:self%nwp-2)   = self%t(2:self%nwp-1)
        tmp_twist_spline%y(1:self%nwp-2)   = self%wp(2:self%nwp-1)%euler%phi
        ! Set type of temporary splines
        if (.not. guess_fal_euler_only) then
            call tmp_x_spline%SetType('PARABOLIC')
            call tmp_y_spline%SetType('PARABOLIC')
            call tmp_z_spline%SetType('PARABOLIC')
            call tmp_s_spline%SetType('PARABOLIC')
        endif
        call tmp_twist_spline%SetType('NATURAL')
        ! Solve temporary splines
        if (.not. guess_fal_euler_only) then
            call tmp_x_spline%Solve()
            call tmp_y_spline%Solve()
            call tmp_z_spline%Solve()
            call tmp_s_spline%Solve()
        endif
        if (debug) then
            print *, 'tmp_twist_spine info before solving'
            call tmp_twist_spline%PrintInfo()
        endif
        call tmp_twist_spline%Solve()
        if (debug) then
            print *, 'tmp_twist_spine info'
            call tmp_twist_spline%PrintInfo()
        endif
        ! Copy the results back to the filament's splines
        !do I=2,self%x_spline%N-1
        do i=2,self%nwp-1
            if (.not. guess_fal_euler_only) then
                ! x spline
                self%x_spline%x(i)                          =   tmp_x_spline%x(i-1)
                self%x_spline%y(i)                          =   tmp_x_spline%y(i-1)
                self%x_spline%dy(i)                         =   tmp_x_spline%dy(i-1)
                self%x_spline%ddy(i)                        =   tmp_x_spline%ddy(i-1)
                self%x_spline%w(i,1:self%x_spline%Degree()+1)      =   tmp_x_spline%w(i-1,1:self%x_spline%Degree()+1)
                ! y spline
                self%y_spline%x(i)                          =   tmp_y_spline%x(i-1)
                self%y_spline%y(i)                          =   tmp_y_spline%y(i-1)
                self%y_spline%dy(i)                         =   tmp_y_spline%dy(i-1)
                self%y_spline%ddy(i)                        =   tmp_y_spline%ddy(i-1)
                self%y_spline%w(i,1:self%y_spline%Degree()+1)      =   tmp_y_spline%w(i-1,1:self%y_spline%Degree()+1)
                ! z spline
                self%z_spline%x(i)                          =   tmp_z_spline%x(i-1)
                self%z_spline%y(i)                          =   tmp_z_spline%y(i-1)
                self%z_spline%dy(i)                         =   tmp_z_spline%dy(i-1)
                self%z_spline%ddy(i)                        =   tmp_z_spline%ddy(i-1)
                self%z_spline%w(i,1:self%z_spline%Degree()+1)      =   tmp_z_spline%w(i-1,1:self%z_spline%Degree()+1)
                ! s spline
                self%s_spline%x(i)                          =   tmp_s_spline%x(i-1)
                self%s_spline%y(i)                          =   tmp_s_spline%y(i-1)
                self%s_spline%dy(i)                         =   tmp_s_spline%dy(i-1)
                self%s_spline%ddy(i)                        =   tmp_s_spline%ddy(i-1)
                self%s_spline%w(i,1:self%z_spline%Degree()+1)      =   tmp_s_spline%w(i-1,1:self%z_spline%Degree()+1)
            endif
            ! phi spline
            self%twist_spline%x(i)                          =   tmp_twist_spline%x(i-1)
            self%twist_spline%y(i)                          =   tmp_twist_spline%y(i-1)
            self%twist_spline%dy(i)                         =   tmp_twist_spline%dy(i-1)
            self%twist_spline%ddy(i)                        =   tmp_twist_spline%ddy(i-1)
            self%twist_spline%w(i,1:self%twist_spline%Degree()+1)  =   tmp_twist_spline%w(i-1,1:self%twist_spline%Degree()+1)
        enddo
        ! Extrapolate fal waypoint coordinates
        if (.not. guess_fal_euler_only) then
            lbounds = 1 !nint(1.0*self%max_radius)
            ubounds = self%film%GetLogicalDimensions() !nint(self%film%GetLogicalDimensions() - 1.0*self%max_radius)
            if (debug) then
                write(*,'(a,2(i0,1x))') '**debug(filament_guess_fal_wps): lbounds = ', lbounds
                write(*,'(a,2(i0,1x))') '**debug(filament_guess_fal_wps): ubounds = ', ubounds
            endif
            ! x, y (first waypoint)
            do
                call self%x_spline%Evaluate(self%t(1),self%wp(1)%coo(1), first_and_last_unknown=.true.)
                call self%y_spline%Evaluate(self%t(1),self%wp(1)%coo(2), first_and_last_unknown=.true.)
                if (debug) then
                    write(*,'(a,3(f0.2,1x))') '**debug(filament_guess_fal_wps): t, x, y = ', self%t(1), self%wp(1)%coo(1:2)
                endif
                if (any(self%wp(1)%coo(1:2) .lt. lbounds(1:2) .or. self%wp(1)%coo(1:2) .gt. ubounds(1:2))) then
                    !self%t(1) = self%t(1) + 0.1*self%max_radius
                    self%t(1) = self%t(1) + 0.5 * (self%t(2)-self%t(1))
                    if (self%t(1) .gt. self%t(2)) then
                        write(*,'(2a)') '**error(filament_guess_fal_wps): oops, had to move the first waypoint ', &
                                        'to avoid edge of the micrograph, but ended up jamming into the second waypoint'
                        call terminate('fatal error in filament_guess_fal_wps')
                    endif
                else
                    exit
                endif
            enddo
            ! x, y (last waypoint)
            do
                call self%x_spline%Evaluate(self%t(self%nwp),self%wp(self%nwp)%coo(1),first_and_last_unknown=.true.)
                call self%y_spline%Evaluate(self%t(self%nwp),self%wp(self%nwp)%coo(2),first_and_last_unknown=.true.)
                if (any(self%wp(self%nwp)%coo(1:2) .lt. lbounds(1:2) .or. self%wp(self%nwp)%coo(1:2) .gt. ubounds(1:2))) then
                    !self%t(self%nwp) = self%t(self%nwp) - 0.1*self%max_radius
                    self%t(self%nwp) = self%t(self%nwp) - 0.5 * (self%t(self%nwp)-self%t(self%nwp-1))
                    if (self%t(self%nwp) .lt. self%t(self%nwp-1)) then
                        ! should not happen!
                        write(*,'(3a)') '**error(filament_guess_fal_wps): oops, had to move the last waypoint ', &
                                        'to avoid edge of the micrograph, but ended up jamming into the second',&
                                        ' to last waypoint'
                        call terminate('fatal error in filament_guess_fal_wps')
                    endif
                else
                    exit
                endif
            enddo
            ! z
            call self%z_spline%Evaluate(self%t(1)       ,self%wp(1)%coo(3), first_and_last_unknown=.true.)
            call self%z_spline%Evaluate(self%t(self%nwp),self%wp(self%nwp)%coo(3),first_and_last_unknown=.true.)
            ! s
            call self%s_spline%Evaluate(self%t(1)       ,self%wp(1)%subunit_num, first_and_last_unknown=.true.)
            call self%s_spline%Evaluate(self%t(self%nwp),self%wp(self%nwp)%subunit_num,&
                                                                                                    first_and_last_unknown=.true.)
        endif
        ! phi
        if (debug) then
            write(*,'(a)') '**debug(filament_guess_fal_wps): about to guess phi values. here is the spline'
            call self%twist_spline%PrintInfo()
        endif
        call self%twist_spline%Evaluate(self%t(1)       ,self%wp(1)%euler%phi, first_and_last_unknown=.true.)
        call self%twist_spline%Evaluate(self%t(self%nwp),self%wp(self%nwp)%euler%phi,&
                                                                                                first_and_last_unknown=.true.)

        if (debug) write(*,'(a,2(f0.3,1x))') '**debug(filament_guess_fal_wps): assigned phi to first, last wp: ', &
                                    self%wp(1)%euler%phi, self%wp(self%nwp)%euler%phi
        ! Re-solve all the splines
        ! Have to make them parabolic, otherwise the derivatives at the new first and last
        ! points will not be computed
        !call spline_set_type(self%x_spline,  'PARABOLIC')
        !call spline_set_type(self%y_spline,  'PARABOLIC')
        !call spline_set_type(self%z_spline,  'PARABOLIC')
        hermite_bak = self%use_hermite_splines
		self%use_hermite_splines = .false.
		call filament_solve_splines(self)
		self%use_hermite_splines = hermite_bak
        if (debug) then
            write(*,'(a)') '**debug(filament_guess_fal_wps): spline info before guess theta & psi'
            call self%x_spline%PrintInfo()
            call self%y_spline%PrintInfo()
            call self%z_spline%PrintInfo()
        endif

        ! If we know the Psi and Theta angles for the other waypoints, we probably mean to estimate them for the fal WPs too
        if (self%PsiAndThetaAreNonZero()) then
			if (debug) write(*,'(a)') '**debug(filament_guess_fal_wps): will estimate psi & theta for new fal WPs'
            do i=1,self%nwp,self%nwp-1
                call euler_from_derivatives_new(self%x_spline%dy(i),self%y_spline%dy(i),self%z_spline%dy(i), &
                                                self%wp(2)%euler%theta .lt. 180.0D0, &
                                                theta_tmp,psi_tmp)
                !if (known_theta)    self%wp(i)%euler%theta  =   theta_tmp
                !if (known_psi)      self%wp(i)%euler%psi    =   psi_tmp
                self%wp(i)%euler%theta  =   theta_tmp
                self%wp(i)%euler%psi    =   psi_tmp
            enddo
			hermite_bak = self%use_hermite_splines
			self%use_hermite_splines = .false.
            call filament_solve_splines(self)
			self%use_hermite_splines = hermite_bak
            !call spline_set_type(self%x_spline,  'KNOWN_D')
            !call spline_set_type(self%y_spline,  'KNOWN_D')
            !call spline_set_type(self%z_spline,  'KNOWN_D')
		else
            if (debug) then
                write(*,'(2a)') '**debug(filament_guess_fal_wps): psi & theta are not known for any point, ',    &
                                'will not estimate them for new fal WPs'
            endif
		endif

		! Now that we're all done, we can plug the original waypoints back in - after all, the purpose of this subroutine is to guess fal WPs, not mess with the other ones
		self%wp(2:self%nwp-1) = wps_on_entry(2:self%nwp-1)

		! Check the subframe alignments are OK
		if (allocated(self%wp(2)%local_subframe_shift)) then
            self%wp(1)%local_subframe_shift = self%wp(2)%local_subframe_shift
        endif
        if (allocated(self%wp(self%nwp-1)%local_subframe_shift)) then
            self%wp(self%nwp)%local_subframe_shift = self%wp(self%nwp-1)%local_subframe_shift
        endif

        if (this_program%IsVerbose() .or. debug) then
            write(*,'(a,f0.3)') '**debug(filament_guess_fal_wps): on exit, this phi to last wp: ', self%wp(self%nwp)%euler%phi
        endif

        endif ! end of test for Euler-only and nwp < 5


    end subroutine filament_guess_fal_wps



end module filaments_model_base
