!>  \brief   Routines to modify the filament's model
module filaments_model_modify
    use filaments_base
    implicit none

    contains

    !>  \brief  Look for pairs of subsequent waypoints whose order was reversed and flip them
    subroutine filament_reverse_order_of_waypoint_pairs(self)
        use differential_geometry, only : angle_between_vectors
        use Globals, only : pi
        !
        type(filament), intent(inout)   ::  self
        ! private variables
        integer         ::  current_wp
        real(kind=8)    ::  current_d(3), previous_d(3)
        real(kind=8)    ::  current_angle
        logical         ::  same_direction_as_previous(self%nwp)
        logical         ::  need_to_swap(self%nwp)
        integer         ::  wp_to_be_swapped
        type(waypoint)  ::  temp_wp
        !real(kind=8)    ::  temp_path_length
        real(kind=8)    ::  temp_subunit_num, temp_t
        real(kind=8)    ::  last_path_length, last_t, last_subunit_num
        ! start work

        same_direction_as_previous(1) = .true.
        previous_d = self%wp(2)%coo - self%wp(1)%coo

        ! loop over waypoints and check whether the path is still pointing in the same direction
        do current_wp=2,self%nwp
            current_d = self%wp(current_wp)%coo - self%wp(current_wp-1)%coo
            ! compute angle between current_d and previous_d
            current_angle = angle_between_vectors(current_d,previous_d)
            ! if angle less than 120.0 (0.66*pi), we're roughly in the same direction
            same_direction_as_previous(current_wp) = current_angle .lt. 0.66*pi
            ! remember current angle as previous angle for next iteration
            previous_d = current_d
        enddo

        ! Now if we find two waypoints which don't have "same direction as previous"
        ! but are surrounded by waypoints that do, we will need to flip the first of those waypoints with its
        ! predecessor
        need_to_swap(1) = .false.
        need_to_swap(self%nwp) = .false.
        do current_wp=2,self%nwp-1
            need_to_swap(current_wp) =              same_direction_as_previous(current_wp-1) &
                                        .and. .not. same_direction_as_previous(current_wp) &
                                        .and. .not. same_direction_as_previous(current_wp+1)
            if (need_to_swap(current_wp)) wp_to_be_swapped = current_wp
        enddo

        ! Now we swap a pair of waypoints
        if (count(need_to_swap) .eq. 1) then
            write(*,'(a,i0,a,i0)') '**info(filament_reverse_order_of_waypoint_pairs): will swap waypoints ', &
                                    wp_to_be_swapped, ' and ', wp_to_be_swapped-1

            last_path_length = self%wp(self%nwp)%path_length
            last_t           = self%t(self%nwp)
            last_subunit_num = self%wp(self%nwp)%subunit_num

            temp_wp = self%wp(wp_to_be_swapped-1)
            temp_wp%path_length = last_path_length - self%wp(wp_to_be_swapped-1)%path_length
            temp_t = last_t - self%t(wp_to_be_swapped-1)
            temp_subunit_num = ceiling(last_subunit_num) - self%wp(wp_to_be_swapped-1)%subunit_num

            self%wp(wp_to_be_swapped-1) = self%wp(wp_to_be_swapped)
            self%wp(wp_to_be_swapped-1)%path_length = last_path_length - self%wp(wp_to_be_swapped)%path_length
            self%wp(wp_to_be_swapped-1)%subunit_num = ceiling(last_subunit_num) - self%wp(wp_to_be_swapped)%subunit_num
            self%t(wp_to_be_swapped-1) = last_t - self%t(wp_to_be_swapped)

            self%wp(wp_to_be_swapped) = temp_wp
            self%t(wp_to_be_swapped) = temp_t
            self%wp(wp_to_be_swapped)%subunit_num = temp_subunit_num

        else if (count(need_to_swap) .gt. 2) then
            write(*,'(a)') '**warning(filament_reverse_order_of_waypoint_pairs): waypoints appear to be in a disordered mess'
            print *, need_to_swap
        endif

    end subroutine filament_reverse_order_of_waypoint_pairs


    !>  \brief  Reverse the order of waypoints in a filament without touching their Euler angles. Model is not updated
    subroutine filament_reverse_order_of_waypoints(self)
        use UsefulFunctions, only : IsEven
        !
        type(filament), intent(inout)   ::  self
        ! private variables
        integer         ::  i
        integer         ::  mid_wp_index
        type(waypoint)  ::  temp_wp
        real(kind=8)    ::  last_path_length, last_t, temp_t, last_subunit_num, temp_subunit_num
        ! start work
        last_path_length = self%wp(self%nwp)%path_length
        last_t           = self%t(self%nwp)
        last_subunit_num = self%wp(self%nwp)%subunit_num
        if (IsEven(self%nwp)) then
            mid_wp_index = self%nwp/2
        else
            mid_wp_index = (self%nwp-1)/2 + 1
        endif
        do i=1,mid_wp_index
            temp_wp = self%wp(i)
            temp_wp%path_length = last_path_length - self%wp(i)%path_length
            temp_t  = last_t - self%t(i)
            temp_subunit_num = ceiling(last_subunit_num) - self%wp(i)%subunit_num
            !
            self%wp(i) = self%wp(self%nwp-i+1)
            self%wp(i)%path_length = last_path_length - self%wp(self%nwp-i+1)%path_length
            self%wp(i)%subunit_num = ceiling(last_subunit_num) - self%wp(self%nwp-i+1)%subunit_num
            self%t(i)  = last_t - self%t(self%nwp-i+1)
            !
            self%wp(self%nwp-i+1) = temp_wp
            self%t(self%nwp-i+1) = temp_t
            self%wp(self%nwp-i+1)%subunit_num = temp_subunit_num
        enddo


    end subroutine filament_reverse_order_of_waypoints



    !>  \brief  Reverse the polarity of a filament (in-place)
    subroutine filament_flip_polarity(self)
        use euleranglesmod
        use filaments_model_update, only : filament_update_model
        use filaments_normal_modes, only : filament_update_nms, filament_wps_to_nm_amplitudes
        use filaments_geometry, only : filament_update_wp_res, filament_fit_splines_linear
        use runtime_parameters, only : terminate
        use Splines

        implicit none
        ! Arguments
        type(filament),         intent(inout)   ::  self    !<  Filament object
        ! Private variables
        type(filament)          ::  tmpfil
        integer                 ::  i, nwp
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        if (debug) then
            write(*,'(/a)') '**debug(filament_flip_polarity): filament before flipping:'
            call self%PrintInfo()
        endif

        ! number of waypoints
        nwp = self%nwp

        if (debug) then
            write(*,*) '**debug(): here 5. t values = ', self%t
        endif

        ! allocate memory for tempfil
        call tmpfil%Allocate(self%nwp)

        ! copy the filament info over
        tmpfil = self

        if (debug) then
            write(*,*) '**debug(): here 4. t values = ', tmpfil%t
        endif

        ! deal with the order of the waypoints
        call filament_flip_polarity_reverse_order(tmpfil)

        ! loop over the input filament's waypoints and flip their polarity
        do i=1,nwp
            call filament_flip_waypoint_polarity(tmpfil,i)

        enddo


        ! copy the flipped filament over
        self = tmpfil


        ! update splines
        call filament_update_model(self,force_update_rigid_wp=.true.)

        ! sanity check
        if (debug) then
            if (self%wp(1)%path_length .ne. 0.) then
                write(*,'(a,f0.3)') '**error(filament_flip_polarity): path_length of first waypoint is not 0.0: ', &
                                    self%wp(1)%path_length
                call terminate('fatal error in filament_flip_polarity')
            endif
        endif

        ! we may need to update the normal modes & residuals
        call filament_fit_splines_linear(self)
        call filament_update_wp_res(self)
        call filament_update_nms(self)
        call filament_wps_to_nm_amplitudes(self)
        call filament_update_model(self)

        !
        ! diagnostics
        !
        if (debug) then
            write(*,'(a)') '**debug(filament_flip_polarity): filament after flipping:'
            call self%PrintInfo()
            write(*,'(a)') '**debug(filament_flip_polarity): splines after flipping:'
            call self%x_spline%PrintInfo()
            call self%y_spline%PrintInfo()
            call self%z_spline%PrintInfo()
            call self%twist_spline%PrintInfo()
            call self%s_spline%PrintInfo()
        endif

    end subroutine filament_flip_polarity


    !>  \brief  Reverse the order of waypoints in a filament (this is part of the filament polarity flipping)
    subroutine filament_flip_polarity_reverse_order(self)
        use UsefulFunctions, only : IsEven
        !
        type(filament), intent(inout)   ::  self
        ! private variables
        integer         ::  i
        integer         ::  mid_wp_index
        type(waypoint)  ::  temp_wp
        real(kind=8)    ::  last_path_length, last_t, temp_t, last_subunit_num, temp_subunit_num, last_phi, temp_phi
        ! start work
        last_path_length = self%wp(self%nwp)%path_length
        last_t           = self%t(self%nwp)
        last_subunit_num = self%wp(self%nwp)%subunit_num
        ! we will treat phi like we do subunit_num, so we first need to have as a function of 360.0
        self%wp%euler%phi= self%wp%euler%phi / 360.0
        last_phi         = self%wp(self%nwp)%euler%phi
        if (IsEven(self%nwp)) then
            mid_wp_index = self%nwp/2
        else
            mid_wp_index = (self%nwp-1)/2 + 1
        endif
        do i=1,mid_wp_index
            temp_wp             = self%wp(i)
            temp_wp%path_length = last_path_length - self%wp(i)%path_length
            temp_t              = last_t           - self%t(i)
            temp_subunit_num    = ceiling(last_subunit_num) - self%wp(i)%subunit_num
            temp_phi            = ceiling(last_phi)         - self%wp(i)%euler%phi
            !
            self%wp(i)             = self%wp(self%nwp-i+1)
            self%wp(i)%path_length = last_path_length          - self%wp(self%nwp-i+1)%path_length
            self%t(i)              = last_t                    - self%t(self%nwp-i+1)
            self%wp(i)%subunit_num = ceiling(last_subunit_num) - self%wp(self%nwp-i+1)%subunit_num
            self%wp(i)%euler%phi   = ceiling(last_phi)         - self%wp(self%nwp-i+1)%euler%phi
            !
            self%wp(self%nwp-i+1)             = temp_wp
            self%t(self%nwp-i+1)              = temp_t
            self%wp(self%nwp-i+1)%subunit_num = temp_subunit_num
            self%wp(self%nwp-i+1)%euler%phi   = temp_phi
        enddo

        ! convert phi back to actual angles rather than fractions of 360.0
        self%wp%euler%phi = self%wp%euler%phi * 360.0


    end subroutine filament_flip_polarity_reverse_order

    !>  \brief  Flip the polarity of a single waypoint in the filament, but do not change its position. Use this when flipping polarity
    subroutine filament_flip_waypoint_polarity(self,waypoint_number)
        use UsefulFunctions, only : DegreesBetween0And360
        !
        type(filament), intent(inout)   ::  self
        integer,        intent(in)      ::  waypoint_number     !<  index of the waypoint to be modified
        ! private variables
        ! start work

        ! deal with theta (note convention that theta > 0.0 by convention)
        self%wp(waypoint_number)%euler%theta  =   DegreesBetween0And360(self%wp(waypoint_number)%euler%theta + 180.0d0)

    end subroutine filament_flip_waypoint_polarity

    !>  \brief  Change a waypoints' Euler angle so that they give the same view, but with theta ~ 180.0 away
    subroutine filament_waypoint_alt_theta(self,waypoint_number)
        use UsefulFunctions, only : DegreesBetween0And360
        !
        type(filament), intent(inout)   ::  self
        integer,        intent(in)      ::  waypoint_number     !<  index of the waypoint to be modified
        ! private variables
        logical ::  add_to_phi
        ! start work

        ! deal with phi
        ! to avoid going to crazy values after many rounds of flipping polarity, we
        ! have an arbitrary convention as to whether to add or subtract 180 degrees
        add_to_phi = (waypoint_number .eq. 1 .and. self%wp(1)%euler%theta .gt. 180.0d0) .or. &
                     (waypoint_number .gt. 1 .and. self%wp(1)%euler%theta .lt. 180.0d0)
        if (add_to_phi) then
            self%wp(waypoint_number)%euler%phi    =   self%wp(waypoint_number)%euler%phi + 180.0d0
        else
            self%wp(waypoint_number)%euler%phi    =   self%wp(waypoint_number)%euler%phi - 180.0d0
        endif

        ! deal with theta (note convention that theta > 0.0 by convention)
        self%wp(waypoint_number)%euler%theta  =   DegreesBetween0And360(-self%wp(waypoint_number)%euler%theta)

        ! deal with psi
        self%wp(waypoint_number)%euler%psi    =   DegreesBetween0And360(self%wp(waypoint_number)%euler%psi + 180.0d0)

        ! deal with subunit_num - not - this is done by filament_reverse_order_of_waypoints
        !self%wp(waypoint_number)%subunit_num  =   ceiling(self%wp(self%nwp)%subunit_num) - self%wp(waypoint_number)%subunit_num
    end subroutine filament_waypoint_alt_theta

    !>  \brief  Set up the filament so that it follows the 1-start operator
    !!
    !!  This is useful when one wants to make sure the positions of each helical subunits are explicitely described,
    !!  for example before starting 3d reconstruciton
    subroutine filament_convert_to_one_start(self)
        use helices, only : helix_to_one_start
        use filaments_path, only : waypoint_from_t
        use filaments_geometry, only : filament_eval_helical_twist, filament_solve_splines
        use Splines
        use filaments_model_update, only : filament_update_model
        implicit none
        ! Arguments
        type(filament),     intent(inout)   ::  self    !<  Input filament
        ! Private variables
        integer                     ::  current_wp
        type(filament)              ::  fil
        logical,        parameter   ::  debug       =   .false.
        real                        ::  tmp, twist_per_sub_at_wp1
        type(waypoint)              ::  first_subunit_after_wp1
        real(kind=8)                ::  t_first_subunit_after_wp1
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(filament_convert_to_one_start): filament info on entry: '
            call self%PrintInfo()
        endif

        ! First, copy the filament over
        fil = self

        ! Adjust phi values
        do current_wp = 2, FIL%nwp
            fil%wp(current_wp)%euler%phi = fil%wp(current_wp)%euler%phi + 360. * real((fil%wp(current_wp)%subunit_num &
                                                                        - fil%wp(1)%subunit_num)/ fil%helix%num_starts)
            !fil%wp(current_wp)%euler%phi = fil%wp(current_wp)%euler%phi + 360. * int((fil%wp(current_wp)%euler%phi - fil%wp(CURRENT_WP-1)%euler%phi) / fil%helix%twist)
        enddo
        ! Find nearest subunit after the first waypoint
        t_first_subunit_after_wp1 = fil%s_spline%FindRoot(dble(ceiling(fil%wp(1)%subunit_num)))
        first_subunit_after_wp1 = waypoint_from_t(fil,t_first_subunit_after_wp1)
        ! In subunits, how far the first waypoint is to the nearest subunit
        tmp = fil%wp(1)%subunit_num - first_subunit_after_wp1%subunit_num
        ! Local twist per subunit at first waypoint
        twist_per_sub_at_wp1 = filament_eval_helical_twist(fil,fil%t(1))
        ! Convert the twist per subunit into 1-start equivalent
        twist_per_sub_at_wp1 = (360. + twist_per_sub_at_wp1) / real(fil%helix%num_starts)
        !
        fil%wp%euler%phi = fil%wp%euler%phi - fil%wp(1)%euler%phi + first_subunit_after_wp1%euler%phi + tmp * twist_per_sub_at_wp1

        ! Diagnostics
        !if (debug) then
        !    write(*,'(a)') '**debug(filament_convert_to_one_start): filament after adjusting phi values: '
        !    call filament_print_info(fil)
        !endif


        ! Update phi spline
        call filament_solve_splines(fil)

        ! Convert the helical operator to its 1-start equivalent
        call helix_to_one_start(fil%helix)

        ! Update model
        call filament_update_model(fil)

        ! Diagnostics
        !if (debug) then
        !    write(*,'(a)') '**debug(filament_convert_to_one_start): filament after updating model: '
        !    call filament_print_info(fil)
        !endif

        ! Copy the filament back
        self = fil

        ! Diagnostics
        if (debug) then
            write(*,'(a)') '**debug(filament_convert_to_one_start): filament info on exit: '
            call self%PrintInfo()
        endif

    end subroutine filament_convert_to_one_start

    !>  \brief  Shift a filament in the z direction so that it coincides with the ctftilt plane
    subroutine filament_shift_average_z_to_ctftilt_plane(self)
        use filaments_geometry, only : filament_average_wp
        use Micrographs, only : Micrograph
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self            !<  Filament object
        ! Private variables
        type(waypoint)  ::  ave_wp
        real(kind=4)    ::  shift(6)
        real            ::  new_ave_z
        logical,        parameter   ::  debug   =   .true.
        ! Start work

        ! Compute the average coordinates of the filament
        ave_wp =  filament_average_wp(self)

        ! Compute the desired new average z from the ctftilt information
        new_ave_z = self%film%ReturnZFromXY(ave_wp%coo(1),ave_wp%coo(2))

        ! Apply the shift
        shift = (/ 0.0,0.0,NEW_AVE_Z-REAL(ave_wp%coo(3),KIND=4),0.0,0.0,0.0  /)
        if (debug) write(*,'(a,F0.3)') '**debug(filament_shift_average_z_to_ctftilt_plane): shifting filament in z by ', shift(3)
        call filament_shift_all_wps(self,shift)
    end subroutine filament_shift_average_z_to_ctftilt_plane


    !>  \brief  Straighten filament to a straigt line with constant helical parameters
    subroutine filament_straighten(self)
        use filaments_model_update, only : filament_update_model
        use runtime_parameters, only : terminate
        use filaments_geometry, only : filament_update_wps_from_wp_res, filament_relax_helix
        use coos, only : waypoints_print_info
        implicit none
        ! Arguments
        type(filament),     intent(inout)   ::  self
        ! Private variables
        logical,        parameter   ::  debug    = .false.
        integer                     ::  i
        logical,        parameter   ::  sanity_check = .true.
        real(kind=8),   parameter   ::  sanity_tol = 0.001D0
        integer                     ::  iter
        integer,        parameter   ::  num_iter = 10
        logical                     ::  hermite_backup
        ! Start work



        ! First, straighten the axis
        do ITER=1,num_iter
        self%wp_res%coo(1)      = 0.0D0
        self%wp_res%coo(2)      = 0.0D0
        self%wp_res%coo(3)      = 0.0D0
        self%wp_res%subunit_num = 0.0D0
        self%wp_res%euler%phi   = 0.0D0
        call filament_update_wps_from_wp_res(self)
        hermite_backup = self%use_hermite_splines
        self%use_hermite_splines = .false.
        call filament_update_model(self,REDO_LINEAR_FIT=.true.) !,KNOWN_PSI=.FALSE.,KNOWN_THETA=.FALSE.
        self%use_hermite_splines = hermite_backup
        enddo

        if (sanity_check) then
            do I=1,self%nwp
                ! All the linear residuals should be 0.0
                if (any (abs(self%wp_res(i)%coo)          .gt. sanity_tol)) then
                    write(*,'(a,i0,a,3(l1))') '**ERROR(filament_straighten): axis sanity check failed for waypoint ', i, &
                                                '. Residuals not 0 (xyz): ', abs(self%wp_res(i)%coo) .gt. sanity_tol
                    write(*,'(3(F0.8,1x))') self%wp_res(i)%coo
                    call waypoints_print_info(self%wp_res)
                    call terminate('Failed sanity check in FILAMENT_STRAIGHTEN')
                endif
            enddo
        endif

        ! Second, relax the helix
        call filament_relax_helix(self,IGNORE_SUBUNIT_PHASING=.true.)
        call filament_update_model(self,REDO_LINEAR_FIT=.true.)!,KNOWN_PSI=.FALSE.,KNOWN_THETA=.FALSE.)
        self%wp_res%subunit_num = 0.0D0
        self%wp_res%euler%phi   = 0.0D0
        call filament_update_wps_from_wp_res(self)
        call filament_update_model(self,REDO_LINEAR_FIT=.true.)!,KNOWN_PSI=.FALSE.,KNOWN_THETA=.FALSE.)


        if (sanity_check) then
            do I=1,self%nwp
                ! All the linear residuals should be 0.0
                if (any (abs(self%wp_res(i)%coo)          .gt. sanity_tol) .or. &
                         abs(self%wp_res(i)%subunit_num)  .gt. sanity_tol .or. &
                         abs(self%wp_res(i)%euler%phi)    .gt. sanity_tol .or. &
                         abs(self%wp_res(i)%euler%theta)  .gt. sanity_tol .or. &
                         abs(self%wp_res(i)%euler%psi)    .gt. sanity_tol) then
                    write(*,'(a,i0,a)') '**ERROR(filament_straighten): helix sanity check failed for waypoint ', &
                                        i, '. Residuals not 0: '
                    call waypoints_print_info(self%wp_res)
                    call terminate('Failed sanity check in FILAMENT_STRAIGHTEN')
                endif
            enddo
        endif

        ! Diagnostic
        if (debug) then
            write(*,'(a)') '**debug(filament_straighten): straightened filament info'
            call self%PrintInfo()
        endif

    end subroutine filament_straighten


    !>  \brief  Apply an out-of-plane tilt to the whole filament as a rigid body about its "midpoint". Waypoints keep the same x,y coordinates.
    !!
    !!  Note: for extreme simplification during inital coding: 1. the "midpoint" will be the median waypoint of the filament, 2. the filament
    !!  will be assumed to be completely straight.
    !!
    !!  \todo   Add a check for how straight the filament really is?
    !!
    subroutine filament_apply_tilt(self,tilt,wps_as_sps)
        use rotation_matrices, only : radians
        use runtime_parameters, only : is_even
        use filaments_model_update, only : filament_update_model
        implicit none
        ! Arguments
        type(filament),                 intent(inout)   ::  self                !<  Filament object
        real,                           intent(in)      ::  tilt                !<  Tilt angle to be applied. In degrees
        logical,            optional,   intent(in)      ::  wps_as_sps          !<  Whether waypoints are being treated as single particles. False by default
        ! Private variables
        integer                 ::  wp_midpoint
        integer                 ::  i
        real                    ::  dist_to_midpoint, z_add
        logical                 ::  wwps_as_sps
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        ! Waypoints as single particles?
        if (present(wps_as_sps)) then
            wwps_as_sps = wps_as_sps
        else
            wwps_as_sps = .false.
        endif

        ! Diagnostics
        if (debug) then
            write(*,'(/a)') '**debug(filament_apply_tilt): filament state before tilt is applied:'
            call self%PrintInfo()
        endif

        if (self%rigid_params(5)) then
            self%rigid_wp%euler%theta = self%rigid_wp%euler%theta + tilt
            call filament_update_model(self)
        else
            !
            ! Find the "midpoint" about which the rotation will happen
            ! Different logic depending on whether there is an odd or even number of waypoints in the filament
            !
            if (is_even(self%nwp)) then
                ! Even number of waypoints
                wp_midpoint = self%nwp/2 + 1
            else
                ! Odd number of waypoints
                wp_midpoint = (self%NWP-1)/2 + 1
            endif

            !
            ! Loop over the waypoints
            !
            do I=1,self%nwp
                if (i .ne. wp_midpoint) then
                    ! Compute straight-line distance from current waypoint to mid waypoint
                    dist_to_midpoint = sqrt(    (self%wp(i)%coo(1)-SELF%wp(wp_midpoint)%coo(1))**2 &
                                              + (self%wp(i)%coo(2)-SELF%wp(wp_midpoint)%coo(2))**2 &
                                              + (self%wp(i)%coo(3)-SELF%wp(wp_midpoint)%coo(3))**2 )
                    ! Compute the additional z
                    ! The tilt is applied clockwise-positive around the y axis (following Spider, Frealign convention, bah)
                    ! The x axis is the direction of the filament, and the coordinate system is right-handed
                    z_add = tan(radians(tilt)) * dist_to_midpoint
                    if ( i .lt. wp_midpoint) then
                        z_add = - z_add
                    endif
                    ! Apply the change in z
                    call filament_shift_wp(self,i,[0.,0.,z_add,0.,0.,0.],UPDATE_SPLINES=.false.)
                endif
                ! In the single-particle case, we update the theta (out-of-plane) Euler angle "manually"
                if (wwps_as_sps) then
                    self%wp(i)%euler%theta = self%wp(i)%euler%theta + tilt
                endif
            enddo

            ! The Euler angles will be re-calculated
            self%wp%euler%theta = 0.0
            self%wp%euler%psi   = 0.0
            call self%x_spline%SetWhetherFirstDerivativesAreKnown(.false.)
            call self%y_spline%SetWhetherFirstDerivativesAreKnown(.false.)
            call self%z_spline%SetWhetherFirstDerivativesAreKnown(.false.)

            ! Update the model
            call filament_update_model(self)
            call self%x_spline%SetWhetherFirstDerivativesAreKnown(.true.)
            call self%y_spline%SetWhetherFirstDerivativesAreKnown(.true.)
            call self%z_spline%SetWhetherFirstDerivativesAreKnown(.true.)
        endif

        ! Diagnostics
        if (debug) then
            write(*,'(a)') '**debug(filament_apply_tilt): filament state after tilt is applied:'
            call self%PrintInfo()
            write(*,'(a)') ' '
        endif

    end subroutine filament_apply_tilt



    !>  \brief  Apply the same shift (in x,y,z,phi,theta and psi) to all waypoints of a filament
    !!
    !!  \todo   Update to work with subunit_num
    subroutine filament_shift_all_wps(self,shift,filament_coordinates,update_model)
        use filaments_model_update, only : filament_update_model
        use filaments_geometry, only : filament_solve_helix_splines, filament_fit_splines_linear, filament_update_wp_res
        use filaments_normal_modes, only : filament_wps_to_nm_amplitudes
        implicit none
        ! Arguments
        type(filament),         intent(inout)   ::  self                    !<  Filament object
        real,                   intent(in)      ::  shift(6)                !<  Shift to be applied in 1.X, 2.Y, 3.Z/defocus, 4.Phi, 5.Theta, 6.Psi
        logical,    optional,   intent(in)      ::  filament_coordinates    !<  If .TRUE., x refers to filament axis, y refers to perpendicular. If .FALSE., which is the default, x and y refer to shifts in the film's coordinate system.
        logical,    optional,   intent(in)      ::  update_model            !<  If .TRUE., update splines after having shifted the waypoints. Defaults to .TRUE.
        ! Private variables
        logical ::  uupdate_model
        integer ::  i
        ! Start work

        ! Will we update splines once we've shifted all the waypoints?
        if (present(update_model)) then
            uupdate_model = update_model
        else
            uupdate_model = .true.
        endif

        ! Shift the waypoints
        do i=1,self%nwp
            call filament_shift_wp(self,i,shift,filament_coordinates,update_splines=.false.)
        end do

        ! Update splines if we are meant to
        if (uupdate_model) then
            call filament_update_model(self,force_update_rigid_wp=.true.)
            call filament_fit_splines_linear(self)
            call filament_update_wp_res(self)
            call filament_wps_to_nm_amplitudes(self)
            call filament_update_model(self,force_update_rigid_wp=.true.)
        endif
    end subroutine filament_shift_all_wps

    !>  \brief  Apply a shift to a filament's way point. The shift is in x,y,z/defocus,phi,theta and psi. Translation can be applied in regular coordinate system,
    !!          or with x being along the filament axis and y perpendicular to it.
    !!
    !!  \todo   Update to work with subunit_num
    subroutine filament_shift_wp(self,wp_num,shift,filament_coordinates,update_splines)
        use rotation_matrices, only : radians
        use filaments_model_update, only : filament_update_model
        use runtime_parameters, only : terminate
        implicit none
        ! Argument
        type(filament),         intent(inout)   ::  self                    !<  Filament
        integer,                intent(in)      ::  wp_num                  !<  Number of the way point to be shifted
        real,                   intent(in)      ::  shift(6)                !<  Shift to be applied in 1.X, 2.Y, 3.Z/defocus, 4.Phi, 5.Theta, 6.Psi
        logical,    optional,   intent(in)      ::  filament_coordinates    !<  If .TRUE., x refers to filament axis, y refers to perpendicular. If .FALSE., which is the default, x and y refer to shifts in the film's coordinate system.
        logical,    optional,   intent(in)      ::  update_splines          !<  If .TRUE., update splines after having shifted the waypoint. Defaults to .TRUE.
        ! Private variables
        logical ::  ffilament_coordinates, uupdate_splines
        real    ::  sshift(6), ipa
        ! Start work
        ! Check the way point number
        if (wp_num .gt. self%nwp) then
            write(*,'(a,i0,a,i0)') '**ERROR(filament_shift_wp): waypoint ', wp_num, &
                                    ' does not exist in this filament - number of way points = ', self%nwp
            call terminate('Fatal error in FILAMENT_SHIFT_WP')
        endif
        ! Are we thinking relative to film or filament?
        if (present(filament_coordinates)) then
            ffilament_coordinates = filament_coordinates
        else
            ffilament_coordinates = .false.
        endif
        ! Will we update splines once we've shifted the waypoint?
        if (present(update_splines)) then
            uupdate_splines = update_splines
        else
            uupdate_splines = .true.
        endif

        ! If necessary, convert the shifts to film coordinates.
        ! This can be done by looking up the psi (in-plane) angle at the relevant way point.
        if (ffilament_coordinates) then
            ipa = radians(self%wp(wp_num)%euler%psi)
            sshift(1) = cos(ipa)*shift(1) + sin(ipa)*shift(2)
            sshift(2) = sin(ipa)*shift(1) + cos(ipa)*shift(2)
        else
            sshift(1) = shift(1)
            sshift(2) = shift(2)
        endif
        ! The z shift is applied directly
        sshift(3) = shift(3)
        ! The angular shifts will be applied directly
        sshift(4:6) = shift(4:6)
        ! Apply the shifts
        self%wp(wp_num)%coo(1:3) = self%wp(wp_num)%coo(1:3) + sshift(1:3)
        self%wp(wp_num)%euler%phi = self%wp(wp_num)%euler%phi + sshift(4)
        self%wp(wp_num)%euler%theta = self%wp(wp_num)%euler%theta + sshift(5)
        self%wp(wp_num)%euler%psi = self%wp(wp_num)%euler%psi + sshift(6)
        ! Diagnostics
        !print '(a,6F12.3)', '**debug(filament_shift_wp): applying shifts: ', sshift

        !
        ! Update splines if we're meant to
        !
        if (uupdate_splines) then
            call filament_update_model(self)
        endif
    end subroutine filament_shift_wp

   !>  \brief  Move a filament's waypoints to the nearest subunit position
    subroutine filament_move_wps_to_nearest_subunit(self)
        use filaments_path, only : waypoint_from_t
        use filaments_model_base, only : filament_find_nearest_subunit_to_waypoint
        use Splines
        use coos, only : wp_print_info
        use filaments_geometry, only : filament_solve_splines
        implicit none
        ! Arguments
        type(filament),                 intent(inout)   ::  self                        !<  Filament whose waypoints will be modified
        ! Private variables
        integer                 ::  current_wp, nearest_subunit
        real(kind=8)            ::  nearest_subunit_t
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        do current_wp=1,self%nwp
            if (debug) then
                write(*,'(a,i0,a,i0,a)') '**debug(filament_move_wps_to_nearest_subunit): waypoint ', current_wp, ' of ', &
                                            self%nwp, ' before change: '
                call wp_print_info(self%wp(current_wp))
            endif
            ! Find the t parameter for the nearest subunit
            nearest_subunit   = filament_find_nearest_subunit_to_waypoint(self,self%wp(current_wp))
            nearest_subunit_t = self%s_spline%FindRoot(dble(nearest_subunit))
            ! Replace the waypoint
            self%wp(current_wp) = waypoint_from_t(self,nearest_subunit_t)
            if (debug) then
                write(*,'(a,i0,a,i0,a)') '**debug(filament_move_wps_to_nearest_subunit): waypoint ', current_wp, ' of ', &
                                            self%nwp, ' after change: '
                call wp_print_info(self%wp(current_wp))
                write(*,'(a,i0,1x,f0.2)') '**debug(filament_move_wps_to_nearest_subunit): nearest subunit, t = ',  &
                                                                                      nearest_subunit, nearest_subunit_t
            endif
        enddo

        ! Update all splines
        call filament_solve_splines(self)

    end subroutine filament_move_wps_to_nearest_subunit

    !>  \brief  Add disorder to a filament's waypoints
    subroutine filament_add_disorder(self,cumulative_disorder_std,non_cumulative_disorder_std)
        use filaments_model_update, only : filament_update_model
        use coos, only : waypoints_add_disorder
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        real,                               intent(in)      ::  cumulative_disorder_std(9)      !<  For each parameter (x,y,z,phi,theta,psi,subunit_num), the standard deviation of the cumulative disorder. In units of pixels, degrees and subunits.
        real,                               intent(in)      ::  non_cumulative_disorder_std(9)  !<  For each parameter, the standard deviation of the non-cumulative disorder
        ! Private variables

        ! Start work
        call waypoints_add_disorder(self%wp,cumulative_disorder_std,non_cumulative_disorder_std)
        call filament_update_model(self)
    end subroutine filament_add_disorder



    !>  \brief  Add new waypoints to the ends of the filament
    subroutine filament_add_fal_wps(self)
        use filaments_model_base, only : filament_guess_fal_wps
        use filaments_model_update, only : filament_update_model
        use filaments_normal_modes, only : filament_update_nms, filament_wps_to_nm_amplitudes
        use filaments_geometry
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        ! Private variables
        type(filament) :: tmp_fil
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(filament_add_fal_wps): filament info on entry'
            call self%PrintInfo()
        endif

        ! Backup the filament
        tmp_fil = self

        ! Deallocate & rallocate filament memory
        call self%Deallocate()
        call self%Allocate(tmp_fil%nwp+2)

        ! Restore WPs
        self%wp(2:self%nwp-1) = tmp_fil%wp

        ! Give path length values for the first and last waypoints
        write(*,'(a)') '**info(filament_add_fal_wps): adding waypoints at beginning and end of the filament'
        self%wp(1)%path_length        = self%wp(2)%path_length - &
                                        0.5*(self%wp(3)%path_length - self%wp(2)%path_length)
        self%wp(self%nwp)%path_length = self%wp(self%nwp-1)%path_length +   &
                                        0.5*(self%wp(self%nwp-1)%path_length - self%wp(self%nwp-2)%path_length)

        ! Ensure the first waypoint has path_length = 0.0
        self%wp%path_length = self%wp%path_length - self%wp(1)%path_length

        ! Deal with the t array
        call self%SetTFromPathLength()

        if (debug) then
            write(*,'(a)') '**debug(filament_add_fal_wps): filament before guess fal'
            call self%PrintInfo()
        endif

        ! Get values for the fal waypoints
        call filament_guess_fal_wps(self,guess_fal_euler_only=.false.)!,KNOWN_PSI=.TRUE.,KNOWN_THETA=.TRUE.)

        if (debug) then
            write(*,'(a)') '**debug(filament_add_fal_wps): filament before solve splines'
            call self%PrintInfo()
        endif

        ! The full splines can now be solved
        call filament_solve_splines(self)!,KNOWN_PSI_THETA=.TRUE.) ! Commented out 29-Nov-2012 with introduction of use_hermite_splines flag

        ! Update the full model
        call filament_update_model(self)

        if (debug) then
            write(*,'(a)') '**debug(filament_add_fal_wps): filament before linear &res'
            call self%PrintInfo()
        endif

        ! Update the normal modes & residuals
        ! Fit the splines & compute residuals
        call filament_fit_splines_linear(self)
        call filament_update_wp_res(self)

        if (debug) then
            write(*,'(a)') '**debug(filament_add_fal_wps): filament info before NMs'
            call self%PrintInfo()
        endif

        ! Compute the normal modes
        call filament_update_nms(self)
        call filament_wps_to_nm_amplitudes(self)

        if (debug) then
            write(*,'(a)') '**debug(filament_add_fal_wps): filament info on exit'
            call self%PrintInfo()
        endif

    end subroutine filament_add_fal_wps



end module filaments_model_modify
