!>  \brief   Routine(s) to ensure the filament's model is kept up-to-date
module filaments_model_update
    use filaments_base
    implicit none

    contains

    !>  \brief  Update the continous model of the filament using the current way-point coordinates & parameters
    !!
    !!          This routine should be called any time the waypoint parameters (x,y,z,Euler angles, etc.) have been changed, to
    !!          ensure that the splines are recalculated, that the waypoints have updated PATH_LENGTHs, and generally that the
    !!          filament object is kept coherently up-to-date.
    !!
    !!          If known_psi and/or known_theta are true, then these known Euler angles are used as "constraints" for the splines.
    !!          If we are guessing first-and-last waypoint parameters, but the Euler angles are known for the other waypoints, then
    !!          Euler angles are guessed for the new fal waypoints.
    !!
    !!          Note that guess_fal_wp_params implies guess_fal_wp_euler, but not vice-versa.
    !!
    !!
    subroutine filament_update_model(   self,&
                                        !known_psi,known_theta,&
                                        guess_fal_wp_params,guess_fal_wp_euler, &
                                        force_update_rigid_wp,wps_as_sps,reset_parametrisation,relax_helix,redo_linear_fit)
        use runtime_parameters, only : terminate
        use ctfs
        use float_comparisons, only : reals_are_different
        use filaments_geometry, only : filament_wp_path_lengths_increase, filament_update_wp_res
        use filaments_geometry, only : filament_relax_twist_per_subunit, filament_relax_helix, filament_wps_add_sub_phi_360
        use filaments_geometry, only : filament_relax_rise_per_subunit
        use filaments_geometry, only : filament_solve_splines, filament_fit_splines_linear
        use filaments_parametrisation, only : filament_parametrisation_initial_guess, filament_parametrisation_update
        use filaments_model_base, only : filament_guess_fal_wps, filament_guess_wps_subunit_num
        use filaments_model_base, only : filament_guess_wps_phi
        use filaments_normal_modes, only : filament_wps_to_nm_amplitudes
        use coos, only : waypoint_has_default_values
        use Globals
        implicit none
        !Argument
        type(filament)                      ::  self
        !logical,    optional,   intent(in)  ::  known_psi               !<  Whether we know the in-plane rotation of way points already (and therefore they should be used as constraints on the splines)
        !logical,    optional,   intent(in)  ::  known_theta             !<  Whether we know the out-of-plane rotation of way points already (and therefore they should be used as constraints on the splines)
        logical,    optional,   intent(in)  ::  guess_fal_wp_params     !<  Extra waypoints before/after the first and last (fal) waypoints have been inserted (i.e. allocated memory), but this subroutine
                                                                        !!  needs to fill all the details in (xyz coordinates and Euler angles)
        logical,    optional,   intent(in)  ::  guess_fal_wp_euler      !<  The x,y,z coordinates of the first and last waypoints are known, but not their their Euler angles
        logical,    optional,   intent(in)  ::  force_update_rigid_wp   !<  If .TRUE., the rigid waypoint parameters will be updated with waypoint values from the filament. Defaults to .FALSE.
        logical,    optional,   intent(in)  ::  wps_as_sps              !<  If .TRUE., the waypoints are treated as independent single particles. For example this means that rigid parameters have no effect on them.
        logical,    optional,   intent(in)  ::  reset_parametrisation   !<  If .TRUE., the path_length and t values will be reset.
        logical,    optional,   intent(in)  ::  relax_helix             !<  If .TRUE., the helix will be "relaxed" so that it is as near to canonical as possible (subunit_num and phi values for waypoints may be changed)
        logical,    optional,   intent(in)  ::  redo_linear_fit         !<  Recalculate the linear fit of the model
        ! Private variables
        !logical                 ::  kknown_psi, kknown_theta
        logical                 ::  gguess_fal_wp_params, gguess_fal_wp_euler
        logical                 ::  initial_parametrisation ! Whether we'll need to do the initial parametrisation for the splines
        logical                 ::  subunit_num_unknown
        logical,    parameter   ::  debug   =   .false.
        logical                 ::  phi_unknown
        logical                 ::  rigid_wp_not_set
        logical                 ::  fforce_update_rigid_wp
        logical                 ::  wwps_as_sps
        real                    ::  rigid_wp_z_old
        !logical                ::  kknown_subunit_num                  !<  Whether the WPS' subunit numbers are to be taken as constraints because they are known
        logical                 ::  rreset_parametrisation
        logical,    parameter   ::  relax_helix_default =   .false.  ! experimental!
        logical                 ::  rrelax_helix
        logical                 ::  rredo_linear_fit
        character(len=6)        ::  nwp_c
        ! Start work

        ! Check memory allocation was done
        if (.not. self%IsAllocated()) then
            write(*,'(a)') '**ERROR(filament_update_model): filament not allocated'
            call terminate('Fatal error in FILAMENT_UPDATE_MODEL')
        endif

        if (self%nwp .ge. 2) then
            ! Do we know out-of-plane rotations at way points already?
            !kknown_theta = self%z_spline%known_d
            !if (present(known_theta)) kknown_theta = known_theta


            ! Do we know in-plane rotations at way points already?
            ! To work out the psi angle, we need to know the derivatives of both the x and y splines (or the derivative of the z spline along with one of the others, but we'll assume that's not a possibility we need to cater for)
            !kknown_psi = self%x_spline%known_d .AND. self%y_spline%known_d
            !if (present(known_psi)) kknown_psi = known_psi


            ! Let's keep things simple: you either know Euler angles for the waypoints, or you don't.
            !if (kknown_psi .AND. .NOT. kknown_theta) then
            !    write(*,'(a)') '**ERROR(filament_update_model): both psi and theta need to be known to impose constraints on the x and y splines, but you only supplied psi'
            !    call terminate('Fatal error in FILAMENT_UPDATE_MODEL')
            !endif


            if (debug) then
                !write(*,'(a,2l)') '**debug(filament_update_model): known psi, theta = ', kknown_psi, kknown_theta
                write(*,'(a)') '**debug(filament_update_model): filament info on entry'
                call self%PrintInfo()
            endif

            ! Have the phi angles already been set?
            phi_unknown = all(self%wp%euler%phi .eq. 0.0)

            ! Have values for rigid parameters been set?
            ! (we ignore the helical parameters, since they will probably have been set during filament_init)
            rigid_wp_not_set =  waypoint_has_default_values(self%rigid_wp,ignore_helix=.true.)

            ! Do we have to guess new first-and-last waypoints (i.e. guess their coordinates & angles)?
            gguess_fal_wp_params = .false.
            if (present(guess_fal_wp_params)) gguess_fal_wp_params = guess_fal_wp_params

            ! Do we have to guess the z coordinate and Euler angles of the first-and-last waypoints?
            gguess_fal_wp_euler = .false.
            if (present(guess_fal_wp_euler)) gguess_fal_wp_euler = guess_fal_wp_euler


            ! Check the logic. If gguess_fal_wp_params, gguess_fal_wp_z_euler is already implied!
            if (gguess_fal_wp_params .and. gguess_fal_wp_euler) then
                write(*,'(2a)') '**ERROR(filament_update_model): both gguess_fal_wp_params and gguess_fal_wp_euler are true. ',  &
                                'Programming error.'
                call terminate('Fatal error in FILAMENT_UPDATE_MODEL')
            endif

            ! These options are not tested often enough to be "safe"
            if (gguess_fal_wp_params .or. gguess_fal_wp_euler) then
                write(*,'(2a)') '**warning(filament_update_model): guessing of first-and-last waypoint parameters may not',  &
                                ' be reliable. Please check results carefully.'
            endif

            ! Are we forcing an update of the rigid waypoint?
            fforce_update_rigid_wp = .false.
            if (present(force_update_rigid_wp)) fforce_update_rigid_wp = force_update_rigid_wp


            ! Are the waypoints treated as single particles?
            wwps_as_sps = .false.
            if (present(wps_as_sps)) wwps_as_sps = wps_as_sps


            ! Are we forcing a reset of the parametrisation?
            rreset_parametrisation = .false.
            if (present(reset_parametrisation)) rreset_parametrisation = reset_parametrisation
            if (debug) write(*,'(a,l1)') '**debug(filament_update_model): reset_parametrisation = ', rreset_parametrisation


            ! Will we re-do the linear fits
            rredo_linear_fit = .false.
            if (present(redo_linear_fit)) rredo_linear_fit = redo_linear_fit
            if (debug) write(*,'(a)') '**debug(filament_update_model): will redo linear fits'

            ! Will we be relaxing the helix?
            rrelax_helix = relax_helix_default
            if (present(relax_helix)) rrelax_helix = relax_helix
            if (debug .and. rrelax_helix) write(*,'(a)') '**debug(filament_update_model): will relax helix'

            ! Will we need to do an initial parametrisation, or do we already have a guess of the path length at each waypoint?
            if (    (any(self%wp%path_length .ne. 0.0) &
                .and. filament_wp_path_lengths_increase(self)) .and. .not. rreset_parametrisation ) then
                !if (gguess_fal_wp_params .or. gguess_fal_wp_euler) then
                !    write(*,'(2a)') '**ERROR(filament_update_model): can only guess fal wp parameters if the filaments', &
                !                    ' are being initialised for the first time, but you supplied path_lengths'
                !    call terminate('Fatal error in FILAMENT_UPDATE_MODEL. Feature not yet implemented')
                !else
                    initial_parametrisation = .false.
                !endif
            else
                initial_parametrisation = .true.
                ! Debug
                if (filament_wp_path_lengths_increase(self)) then
                    write(*,'(2a)') '**debug(filament_update_model): waypoint path lengths do not increase. ',   &
                                    'Will re-initialise the parametrisation.'
                endif
                self%wp%path_length = 0.0D0
                self%t = 0.0D0
            endif

            if (debug) then
                write(*,'(a)') '**debug(filament_update_model): filament info after psi between -90 and 90'
                call self%PrintInfo()
            endif

            ! If necessary, perform initial parametrisation. If we already have path_length estimates, use those as parameters.
            if (initial_parametrisation) then
                ! Initial parametrisation, if we don't yet know the fal wp parameters,
                ! they will be assigned parameter values with a rough guess
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a,i0)') '**info(filament_update_model): triggering initial parametrisation for filament ', self%id
                endif
                call filament_parametrisation_initial_guess(self,guess_fal=gguess_fal_wp_params)
                self%wp%path_length = self%t
            endif

            ! Maybe we know the path length for each waypoint, but not the t parameter
            if (debug) then
                write(nwp_c,'(i6)') self%nwp
                print '(a,'//nwp_c//'(G15.3,1x))', '**debug(filament_update_model): self%t = ', self%t
                print '(a,'//nwp_c//'(F0.2,1x))', '**debug(filament_update_model): self%wp%path_length  =', self%wp%path_length
                print '(a,2l1)', '**debug(filament_update_model): all(self%t .EQ. 0.0), any(self%wp%path_length .NE. 0.0): ',    &
                                all(self%t .eq. 0.0), any(self%wp%path_length .ne. 0.0)
            endif
            if (all(self%t .eq. 0.0) .and. any(self%wp%path_length .ne. 0.0)) then
                if (debug) write(*,'(a)') '**debug(filament_update_model): setting t from path length'
                call self%SetTFromPathLength()
            endif

            ! Do we know the waypoints' subunit numbers?
            subunit_num_unknown = all(self%wp%subunit_num .eq. 0.0)

            ! If we need to, guess (extrapolate) first and last waypoint positions and / or angles
            if (gguess_fal_wp_params .or. gguess_fal_wp_euler) then
                call filament_guess_fal_wps(self,guess_fal_euler_only=gguess_fal_wp_euler .and. .not. gguess_fal_wp_params)
                call filament_solve_splines(self)
            endif

            ! Here, update waypoints with values for rigid parameters that influence the space curve (x,y,z,Theta,Psi)
            if (any(self%rigid_params(1:2)) .or. self%rigid_params(6)) then
                write(*,'(a)') '**ERROR(filament_update_model): the following rigid parameters are not implemented yet: x, y, Psi'
                call terminate('Not implemented yet')
            endif

            ! Here we deal with the cases where theta is being refined per-wp but z is either not refined or refined per-filament only
            ! Note that we assume that if theta is being refined per-wp and z is not being refined at all, the rigid_param flag for z will have been set to .TRUE. elsewhere.
            ! If this is not the case, theta will be refined per wp without any adjustments to the z heights of the waypoints, which could lead to outlandish space curves.
            if ( (.not. self%rigid_params(5)) .and. self%rigid_params(3)) then
                if (self%rigid_wp%coo(3) .ne. 0.0) then
                    ! This only makes sense if the theta angles are known - otherwise how can we update the z values of the waypoints based on their theta?
                    if (self%ThetaAreNonZero()) then
                        ! Since the theta angles might have changed, we have to update the z values of the waypoints so that:
                        ! 1. the rigid waypoint's z height is kept
                        ! 2. the curvature in z is minimised
                        ! Note that the rigid waypoint's z should be up-to-date already, since
                        ! either it is being refined and so it has been updated with the new value by the minimiser
                        ! or it is not being refined and it was updated from wps z values at the end of the last call
                        ! to filament_update_model, so that after we've done all this, the rigid waypoint's z should not have
                        ! changed at all
                        if (debug) then
                            rigid_wp_z_old = self%rigid_wp%coo(3)
                        endif

                        ! If we're being called by filament_init, we may not yet have the splines solved
                        if (all (self%z_spline%x .eq. 0.0D0)) then
                            call filament_solve_splines(self)!,USE_PSI_THETA_CONSTRAINTS=.TRUE.)    Commented out 29-Nov-2012 with introduction of use_hermite_splines flag
                            call filament_parametrisation_update(self)  !,KNOWN_PSI_THETA=.TRUE.) ! Commented out 29-Nov-2012 with introduction of use_hermite_splines flag
                        endif

                        call filament_update_wps_z_from_wps_theta_and_rigid_z(self)
                        ! So now the z coordinates for all waypoints have been updated.
                        ! The problem is this may change the path_length and thus the t values significantly.
                        ! Therefore, we re-parametrise
                        call filament_solve_splines(self)!,USE_PSI_THETA_CONSTRAINTS=.TRUE.) Commented out 29-Nov-2012 with introduction of use_hermite_splines flag
                        call filament_parametrisation_update(self)!,KNOWN_PSI_THETA=.TRUE.)  !Commented out 29-Nov-2012 with introduction of use_hermite_splines flag
                        ! We now re-update the z coordinates, but this time with the near-certainty that the change will be small
                        ! in t / path length
                        call filament_update_wps_z_from_wps_theta_and_rigid_z(self)
                        ! The z coordinates have been updated - we leave the spline solving for later

                        if (debug) then
                            if (reals_are_different(real(self%rigid_wp%coo(3)),rigid_wp_z_old)) then
                                write(*,'(2a,2(G16.9,1x))') '**debug(filament_update_model): sanity check failed: ',  &
                                                            'old rigid z, new rigid z = ', rigid_wp_z_old, self%rigid_wp%coo(3)
                                call terminate('Fatal error in FILAMENT_UPDATE_MODEL')
                            endif
                        endif
                    else
                        ! This can occur when we're called by the filament_apply_tilt routine
                        if (this_program%IsVerbose() .or. debug) then
                            write(*,'(2a)') '**info(filament_update_model): theta unknown, yet was going to ',   &
                                            'update wp Zs using their theta. Skipping.'
                            !call terminate('Fatal error in FILAMENT_UPDATE_MODEL')
                        endif
                    endif   ! end of test for kknown_theta
                else
                    if (debug) write(*,'(2a)') '**debug(filament_update_model): it looks like the rigid waypoint z value',   &
                                                ' is not set, so will not update other waypoint z values based on their theta'
                endif
            endif

            if (debug) then
                write(*,'(a)') '**debug(filament_update_model): filament info before solve spline'
                call self%PrintInfo()
            endif

            ! Solve the interpolation splines, to get an updated continuous model for our filament
            call filament_solve_splines(self)!,KNOWN_PSI_THETA=KKNOWN_PSI) ! Commented out 29-Nov-2012 with introduction of use_hermite_splines parameter

            if (debug) then
                write(*,'(a)') '**debug(filament_update_model): filament info after solving splines, before param update'
                call self%PrintInfo()
            endif

            ! Update t values to ensure that we achieve arc length parametrisation
            ! (this re-solves the splines with the updated t, and also updates the waypoints' path lengths)
            call filament_parametrisation_update(self)  !,KNOWN_PSI_THETA=KKNOWN_PSI) ! Commented out 29-Nov-2012 with introduction of use_hermite_splines flag

            if (debug) then
                write(*,'(a)') '**debug(filament_update_model): filament info after parametrisation update'
                call self%PrintInfo()
            endif

            ! If necessary, guess phi for the waypoints
            if (phi_unknown) then
                call filament_guess_wps_phi(self)
                call filament_solve_splines(self)! ,KNOWN_PSI_THETA=KKNOWN_PSI) ! Commented out 29-Nov-2012 with introduction of use_hermite_splines parameter
            endif

            ! If necessary, compute subunit_num values for waypoints
            if (subunit_num_unknown) then ! .AND. .NOT. wwps_as_sps) then
                if (debug .or. this_program%IsVerbose()) then
                    write(*,'(a)') '**info(filament_update_model): supplied waypoints did not specify subunit numbers. '//&
                                    'Will make educated guess based on path length and average helical rise.'
                endif
                !call filament_wps_add_sub_phi_360(self) ! Commented out 25-jan-2013
                ! Of the 3 options below, the 3rd seems to work better with tmv transition from sp to FF.
                if (debug) then
                    write(*,'(a,F0.4)') '**debug(filament_update_model): about to call UPDATE_WPS_SUBUNIT_NUM. Twist = ', &
                                        self%helix%twist
                    write(*,'(a,l1)')   '**debug(filament_update_model): about to call UPDATE_WPS_SUBUNIT_NUM. wwps_as_sps = ', &
                                        wwps_as_sps
                endif
                !if (self%amyloid_beta .OR. self%helix%twist .EQ. 0.0) then
                    call filament_guess_wps_subunit_num(   self,from_rigid_subunit_num=.false.,     &
                                                            constant_rise_per_subunit=.true.,       &
                                                            constant_twist_per_subunit=.false.,     &
                                                            intg_subunit_num=wwps_as_sps)
                !else !if (wwps_as_sps) then
                !    call filament_guess_wps_subunit_num(   self,FROM_RIGID_SUBUNIT_NUM=.FALSE.,&
                !                                            CONSTANT_RISE_PER_SUBUNIT=.TRUE., &
                !                                            INTG_SUBUNIT_NUM=.TRUE. )
                !else
                !    call filament_guess_wps_subunit_num(   self,FROM_RIGID_SUBUNIT_NUM=.FALSE.,    &
                !                                            CONSTANT_RISE_PER_SUBUNIT=.FALSE.,      &
                !                                            CONSTANT_TWIST_PER_SUBUNIT=.TRUE.,      &
                !                                            INTG_SUBUNIT_NUM=.FALSE.)
                !endif
                call filament_solve_splines(self) !,KNOWN_PSI_THETA=KKNOWN_PSI)  ! Commented out 29-Nov-2012 with introduction of use_hermite_splines parameter
            endif


            if (debug) then
                write(*,'(a)') '**debug(filament_update_model): filament info before helix relaxation'
                call self%PrintInfo()
            endif

            ! When refining with wps_as_sps, we want to find a helix that interpolates exactly
            ! between the waypoints (ie without changing their values) and which is as closed
            ! to canonical as possible. We can add/subtract integer values from subunit_num
            ! and we can add/subtract integer * 360.0 degrees.
            ! When refining in full-filament mode, the value of the refining waypoints are no
            ! more reliable than at any other point along the splines.
            ! In sp mode, we also want to be dealing with waypoints which are centered
            ! exactly on subunits.

            if (wwps_as_sps .or. self%fragmented_subunit_num) then
                if (debug) write(*,'(a)') '**debug(filament_update_model): relaxing rise per subunit'
                call filament_update_average_helix_rise(self) ! added Sep 2014 to account for cases where single-particle refinement of h parameters was done, so that filament h parameters need to be updated
                call filament_relax_rise_per_subunit(self)
            endif

            if (wwps_as_sps .or. self%fragmented_phi) then
                if (debug) write(*,'(a)') '**debug(filament_update_model): relaxing twist per subunit'
                call filament_update_average_helix_twist(self)! added Sep 2014 to account for cases where single-particle refinement of h parameters was done, so that filament h parameters need to be updated
                call filament_relax_twist_per_subunit(self)
            endif

            !
            ! Experimental options to relax the helix
            !
            if (rrelax_helix) then
                write(*,'(a)') '**debug(filament_update_model): experimental - relaxing helix'
                call filament_relax_helix(self)
            endif

            if (debug) then
                write(*,'(a)') '**debug(filament_update_model): filament info after helix relaxation'
                call self%PrintInfo()
            endif

            ! If we didn't know psi and/or theta, we now have estimates
            ! 29-Nov-2012: moved this to a lower level (in solve_splines)
    !        if (.NOT. kknown_psi .OR. .NOT. kknown_theta) then
    !            call filament_euler_from_derivatives(self,KNOWN_PSI=KKNOWN_PSI,KNOWN_THETA=KKNOWN_THETA)
    !        endif

            ! We're about to deal with rigid parameters
            if (debug) then
                write(*,'(a)') '**debug(filament_update_model): filament info before updating with rigid parameters'
                call self%PrintInfo()
            endif

            ! We only deal with rigid waypoint parameters if we're in full-filament mode, not single particles
            if (any(self%rigid_params) .and. .not. self%wps_as_sps) then
                if (debug) write(*,'(a)') '**debug(filament_update_model): will deal with rigid parameters'
                if (rigid_wp_not_set .or. fforce_update_rigid_wp) then
                    ! If there are rigid parameters, and this is the first time we update the filament model (maybe we're being called by filament_init),
                    ! we need to find out the current values of the rigid parameters
                    call filament_update_rigid_wp(self,update_rigid_params=.true.)
                    ! We can then update other waypoints from the rigid waypoint (in most cases this doesn't do anything, but it is needed when initialising a filament with no subunit_num values)
                    call filament_update_wps_with_rigid_parameters(self)
                else
                    ! Now that we have achieved arc-length parametrisation, we may need to update waypoint parameters
                    ! which are kept rigid (e.g. phi, when the helical parameters are kept rigid)
                    call filament_update_wps_with_rigid_parameters(self)

                    ! We can also update the virtual rigid waypoint with values for the parameters which are not actually rigid
                    call filament_update_rigid_wp(self)
                endif
            endif

            if (debug) then
                write(*,'(a)') '**debug(filament_update_model): filament info before updating wp helix'
                call self%PrintInfo()
            endif


            ! Update the helical parameters for each waypoint
            call filament_update_wp_helix(self)
            call filament_update_wp_dcoo(self)

            if (debug) then
                write(*,'(a)') '**debug(filament_update_model): filament info after updating wp helix'
                call self%PrintInfo()
            endif

            ! Update the filament's (average) helical parameters
            call filament_update_average_helix_twist(self)
            call filament_update_average_helix_rise(self)

            if (debug) then
                write(*,'(a)') '**debug(filament_update_model): filament info after updating with rigid parameters'
                call self%PrintInfo()
            endif

            ! Make sure the nm amplitudes are updated
            call filament_wps_to_nm_amplitudes(self)

            ! If requested, re-do the linear fits
            if (rredo_linear_fit) then
                call filament_fit_splines_linear(self)
                call filament_update_wp_res(self)
            endif

            ! And make sure the residuals are updated also, provided we've already done the linear fits
            if (any(self%t_linear .ne. 0.0D0)) then
                call filament_update_wp_res(self)
            endif
        else
            ! 1 waypoint, nothing to do
            if (debug) write(*,'(a)') '**debug(filament_update_model): 1 waypoint, just have to check theta is set'
            if (self%wp(1)%euler%theta .eq. 0.0d0) then
                write(*,'(a)') '**debug(filament_update_model): single waypoint does not have theta value. ', &
                                'Will guess one based on ctffind/ctftilt information.'
                self%wp(1)%euler%theta = 90.0
                if (self%film%ctftilt%tilt_angle .ne. 0.0) then
                    write(*,'(a)') '**warning(filament_update_model): sorry, ctftilt support not yet implemented', &
                                    ' for single-waypoint filaments. Leaving theta at 90.0.'
                endif
            endif
        endif

        if (debug) then
            write(*,'(a)') '**debug(filament_update_model): filament info on exit'
            call self%PrintInfo()
        endif

    end subroutine filament_update_model



    !>  \brief  Update the waypoint parameters with new values for the rigid parameters
    subroutine filament_update_wps_with_rigid_parameters(self)
        use filaments_model_base, only : filament_guess_wps_subunit_num
        use filaments_geometry, only : filament_solve_splines
        use differential_geometry, only : euler_from_derivatives_new
        use filaments_parametrisation, only : filament_parametrisation_update
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self
        ! Private variables
        !real(KIND=8)               ::  rigid_t
        logical,        parameter   ::  debug                   =   .false.
        real(kind=8)                ::  theta_current, theta_new
        real(kind=8),   parameter   ::  theta_new_tolerance     = .1D0
        real(kind=8)                ::  current_error, previous_error
        integer                     ::  num_iter
        integer,        parameter   ::  num_iter_max            =   100
        real,           parameter   ::  dampen_factor           =   .5
        real(kind=8)                ::  psi_current
        integer                     ::  current_wp
        real                        ::  angle_to_add
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(filament_update_wps_with_rigid_parameters): filament info at beginning'
            call self%PrintInfo()
        endif

        if ( any(self%rigid_params(1:2)) .or. self%rigid_params(6) ) then
            write(*,'(a,8l1)') '**ERROR(filament_update_wps_with_rigid_parameters): unsupported set of rigid parameter flags: ', &
                                self%rigid_params
            call terminate('Fatal error in FILAMENT_UPDATE_WPS_WITH_RIGID_PARAMETERS')
        endif

        !
        ! First, we'll deal with anything that affects the space curve. Once that's done, we'll deal with helical & twisting parameters
        !

        ! If theta is rigid.
        ! Rigid theta is defined as the out-of-(xy)-plane angles between the first and last waypoints.
        ! When it is changed, rather than actually rotate the the whole filament, i think it is more desirable
        ! to leave waypoints at their current (x,y,psi) and only change their (z,theta).
        ! i don't know the correct way to apply such a change in one step so instead i do it iteratively.
        ! Note that we assume that the rigid z height has already been updated and that the
        ! rigid waypoint's z height already matches that of the wps.
        if (self%rigid_params(5)) then
            num_iter = 0
            ! We want the new rigid theta to be
            theta_new = self%rigid_wp%euler%theta
            ! Starting conditions for the iteration
            call euler_from_derivatives_new(self%wp(self%nwp)%coo(1) - self%wp(1)%coo(1),   &
                                            self%wp(self%nwp)%coo(2) - self%wp(1)%coo(2),   &
                                            self%wp(self%nwp)%coo(3) - self%wp(1)%coo(3),   &
                                            self%wp(2)%euler%theta .lt. 180.0D0,            &
                                            theta_current,                                  &
                                            psi_current)

            ! test
            if (theta_new * theta_current .lt. 0.) then
                write(*,'(a)') '**WARNING(filament_update_wps_with_rigid_parameters): untested!!! check carefully'
                theta_new = -1.0D0 * theta_new
            endif
            current_error = theta_current - theta_new
            if (debug) write(*,'(a,F0.6,1x,G15.3)') '**debug(filament_update_wps_with_rigid_parameters): starting theta, error = ',&
                                                theta_current, current_error
            previous_error = 0.0D0
            do
                num_iter = num_iter + 1
                previous_error = current_error
                if (num_iter .gt. 1 .and. previous_error * current_error .lt. 0.0D0) then
                    ! If the error changed sign, it probably means we overshot, so let's dampen
                    ! To avoid oscillations in the iterations, we dampen - not sure I'm doing the right thing by making the damping stronger and stronger, but i haven't yet found a case where it doesn't work.
                    angle_to_add = -current_error * dampen_factor!**(num_iter-num_iter_start_dampen+1)
                else
                    angle_to_add = -current_error
                endif
                ! We add theta_new - theta_current to all waypoints' theta
                do current_wp = 1, self%nwp
                    !if ((self%wp(current_wp)%euler%theta-180.0) * (theta_new-180.0) .gt. 0.0) then
                    !    self%wp(current_wp)%euler%theta = self%wp(current_wp)%euler%theta + angle_to_add
                    !else
                    !    self%wp(current_wp)%euler%theta = self%wp(current_wp)%euler%theta - angle_to_add
                    !endif
                    self%wp(current_wp)%euler%theta = self%wp(current_wp)%euler%theta + angle_to_add
                enddo
                ! Now that we've done that, we can update the waypoints' z coordinate, because we
                ! assume that the rigid_wp z coordinate is already in synch with the wps' z coordinates
                call filament_update_wps_z_from_wps_theta_and_rigid_z(self)
                ! So now we have updated the z and theta for all the filament's waypoints, let's solve the splines and update parametrisation
                call filament_solve_splines(self)
                call filament_parametrisation_update(self)
                ! ok, let's check if we achieved our target theta_new
                call euler_from_derivatives_new(self%wp(self%nwp)%coo(1) - self%wp(1)%coo(1),   &
                                                self%wp(self%nwp)%coo(2) - self%wp(1)%coo(2),   &
                                                self%wp(self%nwp)%coo(3) - self%wp(1)%coo(3),   &
                                                self%wp(2)%euler%theta .lt. 180.0D0,            &
                                                theta_current,                                  &
                                                psi_current)
                current_error = theta_current - theta_new
                ! Sanity check  / debug
                !if (abs(current_error) .gt. abs(previous_error)*1.01) then
                !    write(*,'(a,2(f0.1,1x))') '**error(filament_update_wps_with_rigid_parameters): error got worse: ', previous_error, current_error
                !    call terminate('Fatal error in filament_update_wps_with_rigid_parameters')
                !endif
                if (debug) then
                    write(*,'(a,i0,a,ES11.3,a,F0.6,a,F0.6,a)') '**debug(filament_update_wps_with_rigid_parameters): on iteration ',&
                                                                num_iter, ', error is ', current_error, '. Target: ',&
                                                                 theta_new, '. Current: ', theta_current, '.'
                endif
                !if (debug) write(*,'(3(A,F0.5),a,G0.3)') '**debug(filament_update_wps_with_rigid_parameters): new theta (', theta_current, '), expected value of (', theta_new, '). Starting value was: ', theta_current, '. Difference is: ', current_error
                if (abs(current_error) .le. theta_new_tolerance ) then
                    ! If the curent error is within our tolerance, we can stop iterating
                    exit
                else
                    ! If the error is still too large, we'll keep iterating unless we've reached the maximum number of iterations
                    if (num_iter .gt. num_iter_max) then
                        write(*,'(2a,i0,a,G15.3,a,G15.3)')   '**ERROR(filament_update_wps_with_rigid_parameters): ',    &
                                                            'did not manage to impose rigid theta onto filament after ', &
                                                            num_iter_max, ' iterations. Remaining error = ', &
                                                            theta_current - theta_new, '. Tolerance = ', theta_new_tolerance
                        write(*,'(2a,G17.10,1x,G17.10,1x,G17.10)')    '**ERROR(filament_update_wps_with_rigid_parameters): ', &
                                                                    'rigid theta, theta_new_check, theta_new = ', &
                                                                    self%rigid_wp%euler%theta, theta_current, theta_new
                        call terminate('Fatal error in FILAMENT_UPDATE_WPS_WITH_RIGID_PARAMETERS')
                    endif
                endif
            enddo
        endif

        !
        ! Now that the space curve is done, we deal with phi and subunits numbers first and then with helical parameters
        !

        if (self%rigid_params(4)) then
            ! Update the phi values of the waypoints
            !!rigid_t = filament_get_t_for_rigid_wp(self)
            !!self%wp%euler%phi = self%rigid_wp%euler%phi + (self%t - rigid_t) / self%helix%rise * self%helix%twist ! This formula assumes constant helical parameters
            !self%wp%euler%phi = self%wp%euler%phi - sum(self%wp%euler%phi)/real(self%nwp) + self%rigid_wp%euler%phi
            call filament_update_wps_phi(self,from_rigid_phi=.true.,constant_twist=self%constant_twist_per_length)
        endif

        if (self%rigid_params(7)) then
            ! Update the subunit numbers of the waypoints
            !self%wp%subunit_num = self%wp%subunit_num - sum(self%wp%subunit_num)/real(self%nwp) + self%rigid_wp%subunit_num
            call filament_guess_wps_subunit_num(self,from_rigid_subunit_num=.true., &
                                                constant_rise_per_subunit=self%constant_rise_per_subunit)
        endif

        if (self%rigid_params(4) .or. self%rigid_params(7)) then
            ! We need to solve the relevant splines (phi_spline and s_spline)
            call filament_solve_splines(self)
        endif

        if (debug) then
            write(*,'(a)') '**debug(filament_update_wps_with_rigid_parameters): filament info at end'
            call self%PrintInfo()
        endif

    end subroutine filament_update_wps_with_rigid_parameters

    !>  \brief  For each waypoint in a filament, estimate the phi angle
    subroutine filament_update_wps_phi(self,from_rigid_phi,constant_twist)
        use runtime_parameters, only : terminate
        use helices, only : helical_symmetry_unit_convert
        implicit none
        ! Arguments
        type(filament),                 intent(inout)   ::  self                    !<  Filament whose waypoints are to be updated with new phi values
        logical,            optional,   intent(in)      ::  from_rigid_phi          !<  If true, the phi values will be recalculated for each waypoint which keeping the rigid waypoint's phi value unchanged. Defaults to true.
        logical,            optional,   intent(in)      ::  constant_twist          !<  If true, we assume the twist per unit length is constant. Defaults to true.
        ! Private variables
        logical                 ::  ffrom_rigid_phi, cconstant_twist
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        ! Will we be leaving the rigid subunit_num value unchanged?
        if (present(from_rigid_phi)) then
            ffrom_rigid_phi = from_rigid_phi
        else
            ffrom_rigid_phi = .true.
        endif

        ! Constant rise?
        if (present(constant_twist)) then
            cconstant_twist = constant_twist
        else
            cconstant_twist = .true.
        endif

        ! If constant twist is on, but from_rigid isn't, we'll probably get meaningless values
        if (cconstant_twist .and. .not. ffrom_rigid_phi) then
            write(*,'(2a)') '**ERROR(filament_update_wps_phi): it may be unsafe to assume constant twist per unit length',   &
                            ' without imposing the rigid_wp phi value'
            call terminate('Unsupported settings. Fatal error in FILAMENT_UPDATE_WPS_PHI')
        endif

        ! Check the helix is in pixel units
        if (.not. self%helix%pixel_units) call helical_symmetry_unit_convert(self%helix,self%film%GetPixelSize())

        ! Start the actual work
        if (cconstant_twist) then
            ! We guess the phi values based on the assumption of a fixed helical rise per subunit
            if (debug) write(*,'(2a)') '**debug(filament_update_wps_phi): calculating phi values for all waypoints assuming',    &
                                        ' constant twist per unit length'
            self%wp%euler%phi = self%t / self%helix%rise * self%helix%twist
        endif

        if (ffrom_rigid_phi) then
            ! Add a constant to all subunit_num values such that the new average is equal to the rigid waypoint's subunit_num value
            if (debug) write(*,'(2a,F0.3,a)') '**debug(filament_update_wps_phi): recalculating phi values for all waypoints',    &
                                            ' to take into account phi value of rigid waypoint (', self%rigid_wp%euler%phi, ')'
            self%wp%euler%phi = self%wp%euler%phi - (sum(self%wp%euler%phi) / real(self%nwp)) + self%rigid_wp%euler%phi
        endif
    end subroutine filament_update_wps_phi



    !>  \brief  Update the virtual rigid waypoint with average values from the waypoints for the parameters which are not being held rigid
    subroutine filament_update_rigid_wp(self,update_rigid_params)
        use filaments_model_base, only : filament_get_t_for_rigid_wp
        use Splines
        use differential_geometry, only : euler_from_derivatives_new
        implicit none
        ! Arguments
        type(filament),                 intent(inout)   ::  self
        logical,            optional,   intent(in)      ::  update_rigid_params !<  Whether rigid parameters should also be updated. False by default.
        ! Private variables
        real(kind=8)            ::  rigid_t
        logical                 ::  uupdate_rigid_params
        logical,    parameter   ::  debug   =   .false.
        real(kind=8)            ::  dummy
        ! Start work

        ! Will we be updating the rigid parameters too?
        if (present(update_rigid_params)) then
            uupdate_rigid_params = update_rigid_params
        else
            uupdate_rigid_params = .false.
        endif

        ! x
        if (.not. self%rigid_params(1) .or. uupdate_rigid_params) then
            self%rigid_wp%coo(1) = sum(self%wp%coo(1)) / real(self%nwp)
        endif

        ! y
        if (.not. self%rigid_params(2) .or. uupdate_rigid_params) then
            self%rigid_wp%coo(2) = sum(self%wp%coo(2)) / real(self%nwp)
        endif

        ! z
        if (.not. self%rigid_params(3) .or. uupdate_rigid_params) then
            if (all(self%t .eq. 0.0)) then
                self%rigid_wp%coo(3) = sum(self%wp%coo(3)) / real(self%nwp)
            else
                rigid_t = filament_get_t_for_rigid_wp(self)
                call self%z_spline%Evaluate(rigid_t,self%rigid_wp%coo(3))
            endif
        endif

        ! Phi
        if (.not. self%rigid_params(4) .or. uupdate_rigid_params) then
            !rigid_t = filament_get_t_for_rigid_wp(self)
            !self%rigid_wp%euler%phi = interpol_spline_eval(self%phi_spline,rigid_t)
            !print '(a,g0.3)', '**debug(filament_update_rigid_wp): rigid phi = ', self%rigid_wp%euler%phi
            self%rigid_wp%euler%phi = sum(self%wp%euler%phi) / real(self%nwp)
            if (debug) print '(a,g15.3)', '**debug(filament_update_rigid_wp): rigid phi (average) = ', self%rigid_wp%euler%phi
        endif

        ! Theta
        if (.not. self%rigid_params(5) .or. uupdate_rigid_params) then
            !self%rigid_wp%euler%theta = sum(self%wp%euler%theta) / real(self%nwp)
            if (self%nwp .eq. 1) then
                self%rigid_wp%euler%theta = self%wp(1)%euler%theta
            else
                call euler_from_derivatives_new(self%wp(self%nwp)%coo(1) - self%wp(1)%coo(1),   &
                                                self%wp(self%nwp)%coo(2) - self%wp(1)%coo(2),   &
                                                self%wp(self%nwp)%coo(3) - self%wp(1)%coo(3),   &
                                                self%wp(2)%euler%theta .lt. 180.0D0,            &
                                                self%rigid_wp%euler%theta,                      &
                                                dummy)
            endif
            if (debug) print '(a,F0.3)', '**debug(filament_update_rigid_wp): rigid waypoint theta  = ', self%rigid_wp%euler%theta
        endif

        ! Psi
        if (.not. self%rigid_params(6) .or. uupdate_rigid_params) then
            if (self%nwp .eq. 1) then
                self%rigid_wp%euler%psi = self%wp(1)%euler%psi
            else
                call euler_from_derivatives_new(self%wp(self%nwp)%coo(1) - self%wp(1)%coo(1),   &
                                                self%wp(self%nwp)%coo(2) - self%wp(1)%coo(2),   &
                                                self%wp(self%nwp)%coo(3) - self%wp(1)%coo(3),   &
                                                self%wp(2)%euler%theta .lt. 180.0D0,            &
                                                dummy,                      &
                                                self%rigid_wp%euler%psi)
            endif
            if (debug) print '(a,F0.3)', '**debug(filament_update_rigid_wp): rigid waypoint psi  = ', self%rigid_wp%euler%psi
        endif

        ! Subunit number
        if (.not. self%rigid_params(7) .or. uupdate_rigid_params) then
            self%rigid_wp%subunit_num = sum(self%wp%subunit_num) / real(self%nwp)
        endif

        ! Helical rise
        if (.not. self%rigid_params(7) .or. uupdate_rigid_params) then
            self%rigid_wp%helix%rise = sum(self%wp%helix%rise) / real(self%nwp)
        endif

        ! Helical twist
        if (.not. self%rigid_params(4) .or. uupdate_rigid_params) then
            self%rigid_wp%helix%twist = sum(self%wp%helix%twist) / real(self%nwp)
        endif

    end subroutine filament_update_rigid_wp

    !>  \brief  For each waypoints, update the z value so that the curvature in z is minimised, given the waypoints' theta values
    !!
    !!  Call this subroutine when the theta values are being refined, but the waypoint z values are not being refined.
    !!  We assume that the rigid waypoint's z has already been updated (if the rigid waypoint's z is not being refined, we
    !!  assume it has the average z height of the filament)
    !!
    !!  \todo   Move the math into the spline module so it can be reused later
    !!  \todo   Move to filaments_model_modify module
    subroutine filament_update_wps_z_from_wps_theta_and_rigid_z(self)
        use differential_geometry, only : dz_from_theta
        use filaments_geometry, only : nearest_previous_wp_from_t
        use filaments_model_base, only : filament_get_t_for_rigid_wp
        use Splines
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament),             intent(inout)   ::  self
        ! Private variables
        integer                     ::  wp_before_rigid_wp
        integer                     ::  current_wp
        real(kind=8)                ::  y_q, y_r, dy_qp, dy_q, x_r, x_q,h_q ! See notation in "spline interpolation" document. r is for rigid, q is the index just before the rigid waypoint.
        real(kind=8)                ::  h_i
        real(kind=8)                ::  rigid_t
        logical,        parameter   ::  debug   =   .false.
        character(len=6)            ::  nwp_c
        real(kind=8)                ::  sign_multiplier
        ! Start work

        if (debug) then
            write(nwp_c,'(i6)') self%nwp
            write(*,'(a)') '**debug(filament_update_wps_z_from_wps_theta_and_rigid_z): filament info on entry:'
            call self%PrintInfo()
            write(*,'(a,'//nwp_c//'(F0.3,1x))') '**debug(filament_update_wps_z_from_wps_theta_and_rigid_z): z values on entry = ', &
                                            self%wp%coo(3)
            call self%z_spline%PrintInfo()
        endif

        ! Get the z derivatives from the theta values
        do current_wp=1,size(self%wp)
            self%z_spline%dy(current_wp) = dz_from_theta(self%wp(current_wp)%euler%theta)
        enddo

        ! The rigid waypoint's parameter
        rigid_t = filament_get_t_for_rigid_wp(self)

        if (debug) then
            write(*,'(a,F0.3)') '**debug(filament_update_wps_z_from_wps_theta_and_rigid_z): rigid t = ', rigid_t
            write(*,'(a,2(F0.3,1x))') '**debug(filament_update_wps_z_from_wps_theta_and_rigid_z): self%t(1), self%t(self%nwp) = ',&
                                    self%t(1), self%t(self%nwp)
        endif

        ! Find out which waypoint is the one just before the rigid waypoint
        wp_before_rigid_wp = nearest_previous_wp_from_t(self,rigid_t)
        if (wp_before_rigid_wp .eq. 0) then
            write(*,'(2a)') '**ERROR(filament_update_wps_z_from_wps_theta_and_rigid_z): could not find ',    &
                            'nearest previous wp from rigid_wp'
            call terminate('Fatal error in FILAMENT_UPDATE_WPS_Z_FROM_WPS_THETA_AND_RIGID_Z')
        endif

        ! Initially, this routine was written with assumption that theta would always be near 90.0, 150921 I have an ad hoc fix for this other case
        if (self%rigid_wp%euler%theta .gt. 180.0d0) then
            sign_multiplier = -1.0d0
        else
            sign_multiplier = 1.0d0
        endif

        ! We can compute that waypoint's z value (see "spline interpolation" document for derivation)

        ! Set up short variable names to follow notations in my derivation
        h_q = self%t(wp_before_rigid_wp+1) - self%t(wp_before_rigid_wp)
        x_r = rigid_t
        y_r = self%rigid_wp%coo(3)
        dy_q = self%z_spline%dy(wp_before_rigid_wp)
        dy_qp = self%z_spline%dy(wp_before_rigid_wp+1)
        x_q = self%t(wp_before_rigid_wp)
        ! The actual work ! 150921 - switch sign
        y_q = y_r - ((dy_qp - dy_q)/(2*h_q) * (x_r-x_q)**2) - (dy_q*(x_r-x_q)) * sign_multiplier
        ! The result is plugged back in to the spline and filament objects
        self%z_spline%y(wp_before_rigid_wp) = y_q
        self%wp(wp_before_rigid_wp)%coo(3) = y_q

        ! Now work out the z value for the other waypoints.
        do current_wp = wp_before_rigid_wp+1, self%nwp
            h_i = self%t(current_wp) - self%t(current_wp-1)
            ! current_wp = i+1 in eq 29 of spline document
            self%wp(current_wp)%coo(3) = self%wp(current_wp-1)%coo(3)   + sign_multiplier * h_i * (self%z_spline%dy(current_wp) &
                                                                        + self%z_spline%dy(current_wp-1))/2.
        enddo
        do current_wp = wp_before_rigid_wp-1, 1, -1
            h_i = self%t(current_wp+1) - self%t(current_wp)
            ! current_wp = i in eq 29 of spline document
            self%wp(current_wp)%coo(3) = self%wp(current_wp+1)%coo(3)   - sign_multiplier * h_i * (self%z_spline%dy(current_wp+1) &
                                                                        + self%z_spline%dy(current_wp))/2.
        enddo

        if (debug) then
            write(*,'(a,'//nwp_c//'(F0.3,1x))') '**debug(filament_update_wps_z_from_wps_theta_and_rigid_z): z values on exit = ', &
                                            self%wp%coo(3)
        endif

    end subroutine filament_update_wps_z_from_wps_theta_and_rigid_z


    !>  \brief  Calculate an average twist per subunit for the filament, based on the arc length distances between waypoints
    subroutine filament_update_average_helix_rise(self)
        use euleranglesmod, only : angle_difference
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self    !< Filament
        ! Private variables
        integer                         ::  i
        real(kind=8)                    ::  num_repeats, path_length
        logical,        parameter       ::  debug   =   .false.
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(filament_update_average_helix_rise): filament info on entry'
            call self%PrintInfo()
        endif


        ! If we're in single-particle or fragmented_phi, it makes no sense
        ! to compute an average twist, except perhaps if there was single-particle refinement
        ! of helical parameters
        if (self%fragmented_subunit_num .or. self%wps_as_sps) then
            if (self%wps_as_sps) then
                if (self%fal_wps_in_score) then
                    self%helix%rise = sum(self%wp%helix%rise) / self%nwp
                else
                    self%helix%rise = sum(self%wp(2:self%nwp-1)%helix%rise) / (self%nwp-2)
                endif
            endif
            ! Not doing anything
        else

            !
            ! Loop over the waypoints. Work out the average change in phi per repeat
            !
            num_repeats = 0.0D0
            path_length = 0.0D0
            do i=2,self%nwp
                num_repeats = num_repeats + self%wp(i)%subunit_num - self%wp(i-1)%subunit_num
                path_length = path_length + self%wp(i)%path_length - self%wp(i-1)%path_length
            enddo

            self%helix%rise = path_length /  num_repeats
        endif

        if (debug) then
            write(*,'(a)') '**debug(filament_update_average_helix_rise): filament info on exit'
            call self%PrintInfo()
        endif

    end subroutine filament_update_average_helix_rise

    !>  \brief  Calculate an average twist per subunit for the filament, based on the arc length distances between waypoints
    subroutine filament_update_average_helix_twist(self)
        use euleranglesmod, only : angle_difference
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self    !< Filament
        ! Private variables
        integer                         ::  i
        real(kind=8)                    ::  num_repeats
        real(kind=8)                    ::  change_in_phi, twist_per_subunit, tmp
        logical,        parameter       ::  debug   =   .false.
		integer							::	start_wp,finish_wp
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(filament_update_average_helix_twist): filament info on entry'
            call self%PrintInfo()
        endif


        ! If we're in single-particle or fragmented_phi, it makes no sense
        ! to compute an average twist, except perhaps if there was single-particle refinement
        ! of helical parameters
        if (self%fragmented_phi .or. self%wps_as_sps) then
            ! Not doing anything
            if (self%wps_as_sps) then
                if (self%fal_wps_in_score) then
                    self%helix%twist = sum(self%wp%helix%twist) / self%nwp
                else
                    self%helix%twist = sum(self%wp(2:self%nwp-1)%helix%twist) / (self%nwp-2)
                endif
            endif
        else

			if (self%fal_wps_in_score) then
				start_wp = 1
				finish_wp = self%nwp
			else
				start_wp = 2
				finish_wp = self%nwp - 1
			endif

            !
            ! Loop over the waypoints. Work out the average change in phi per repeat
            !
            num_repeats = 0.0
            change_in_phi = 0.0
            do i=start_wp+1,finish_wp
                num_repeats = num_repeats + self%wp(i)%subunit_num - self%wp(i-1)%subunit_num
                change_in_phi = change_in_phi + (self%wp(i)%euler%phi - self%wp(i-1)%euler%phi)
                if (debug) write(*,'(2a,3(f0.1,1x))') '**debug(filament_average_twist_per_subunit): num_repeats, ', &
                                                        'change_in_phi, phi/repeats = ', &
                                                        (self%wp(i)%subunit_num - self%wp(i-1)%subunit_num), &
                                                        (self%wp(i)%euler%phi - self%wp(i-1)%euler%phi), &
                                                        (self%wp(i)%subunit_num - self%wp(i-1)%subunit_num)/ &
                                                        (self%wp(i)%euler%phi - self%wp(i-1)%euler%phi)
            enddo
            twist_per_subunit = change_in_phi / num_repeats

            !
            ! If the filament has reversed polarity, we still want the original twist, not its negative
            !
    !        if (self%wp(1)%euler%theta .GT. 180.) then
    !            twist_per_subunit = - twist_per_subunit
    !        endif

            !
            ! When the helical parameter twist is close to 180 (e.g. 179.70), we don't want to be calculating the 'slow' version of the twist (-0.3)
            !
            tmp = abs(angle_difference(self%helix%twist,twist_per_subunit))
            if ( tmp .gt. 90.0) then
                if (debug) write(*,'(a)') '**debug(filament_average_twist_per_subunit): adding 180'
                twist_per_subunit = 180.0 + twist_per_subunit
            endif

            if (debug) print '(a,i0,a,F0.4,a)', '**debug(filament_average_twist_per_subunit): average twist for filament ', &
                                                self%id, '= ', twist_per_subunit, &
                                                '. Setting this twist value for the filament'
            self%helix%twist = twist_per_subunit
        endif
    end subroutine filament_update_average_helix_twist

    !>  \brief  Update each waypoint with the local derivative values
    subroutine filament_update_wp_dcoo(self)
        use Splines
        !
        type(filament), intent(inout)   ::  self
        ! private variables
        integer :: current_wp
        !
        do current_wp=1,self%nwp
            self%wp(current_wp)%dcoo(1) = self%x_spline%EvaluateDerivative(self%t(current_wp))
            self%wp(current_wp)%dcoo(2) = self%y_spline%EvaluateDerivative(self%t(current_wp))
            self%wp(current_wp)%dcoo(3) = self%z_spline%EvaluateDerivative(self%t(current_wp))
        enddo
    end subroutine filament_update_wp_dcoo

    !>  \brief  For each way point, calculate the local helical parameters.
    !!  \todo   Fix behaviour when fragmented helix - perhaps it makes sense to still calculated local parameters?
    subroutine filament_update_wp_helix(self,num_subunits_btwn_wps_mulitple_of)
        use filaments_geometry, only : filament_eval_helical_rise, filament_eval_helical_twist
        use runtime_parameters, only : terminate
        use helices, only : helical_symmetry_print_info
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  self                                !<  Filament
        integer,                optional,   intent(in)      ::  num_subunits_btwn_wps_mulitple_of   !<  The number of subunits between successive waypoints should be a multiple of this integer number, if given. If not given, defaults to 1.
        ! Private variables
        integer                 ::  i
        logical,    parameter   ::  debug   =   .false.

        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(filament_update_wp_helix): filament info on entry:'
            call self%PrintInfo()
        endif

        if (present(num_subunits_btwn_wps_mulitple_of)) then
            call terminate('option num_subunits_btwn_wps_mulitple_of not supported anymore')
        endif

        ! Check the helical parameters for the filament have been setup
        if (self%helix%rise .eq. 0.0) then
            write(*,'(a)') '**ERROR(filament_update_wp_helix): helical parameters for the filament not setup'
            call terminate('Fatal error in FILAMENT_UPDATE_WP_HELIX')
        endif


        ! Loop over the way points
        do i=1,self%nwp
            if (self%fragmented_subunit_num .or. self%wps_as_sps) then
                if (self%constant_rise_per_subunit) then
                    self%wp(i)%helix%rise   =   self%helix%rise ! Sep 2014: turning this off when constant_rise_per_subunit is not specified, since it could be that the helical parameters were refined per-wp
                endif
            else
                self%wp(i)%helix%rise   =   filament_eval_helical_rise(self,self%t(i))
            endif
            if (self%fragmented_phi .or. self%wps_as_sps) then
                if (self%constant_rise_per_subunit .and. self%constant_twist_per_length) then
                    self%wp(i)%helix%twist  =   self%helix%twist
                endif
            else
                if (self%fragmented_subunit_num .or. self%wps_as_sps) then
                    !self%wp(i)%helix%twist  =   filament_eval_helical_twist(self,self%t(i),RISE=SELF%wp(i)%helix%rise) ! Sep 2014: turning this off, since it could be that the helical parameters were refined per-wp
                else
                    self%wp(i)%helix%twist  =   filament_eval_helical_twist(self,self%t(i))
                endif
            endif

            ! Debug diagnostics
            if (debug) then
                write(*,'(a,i0,a)',advance='NO') '**debug(filament_update_wp_helix): setting helical parameters for wp ', &
                                                                                                                i , ' to '
                call helical_symmetry_print_info(self%wp(i)%helix)
            endif

        ! End of loop over way points
        enddo

        ! Sanity check
        if (self%constant_rise_per_subunit) then! .and. &
            !.not. (self%wps_as_sps .or. self%fragmented_phi .or. self%fragmented_subunit_num)) then
            ! Check the rise per subunit is indeed constant
            if (any(  abs(self%wp(2:self%nwp)%helix%rise-self%wp(1)%helix%rise) .gt. 0.01 )) then
                write(*,'(a)') '**ERROR(filament_update_wp_helix): rise per subunit should be constant but it isn''t: '
                write(*,*) self%wp%helix%rise
                call terminate('Fatal error in FILAMENT_UPDATE_WP_HELIX')
            endif
            if (self%constant_twist_per_length) then
                ! Check the twist per subunit is constant
                if (any(  abs(self%wp(2:self%nwp)%helix%TWIST-SELF%wp(1)%helix%twist) .gt. 0.01 )) then
                    write(*,'(a)') '**ERROR(filament_update_wp_helix): twist per subunit should be constant but it isn''t: '
                    write(*,*) self%wp%helix%twist
                    call terminate('Fatal error in FILAMENT_UPDATE_WP_HELIX')
                endif
            endif
        endif
    end subroutine filament_update_wp_helix


end module filaments_model_update
