!>  \brief  Methods to deal with normal modes
module filaments_normal_modes
    use filaments_base
    use runtime_parameters, only : terminate
    implicit none


    contains

    !>  \brief  Return a weighted sum of normal mode amplitudes. The amplitudes are weighted by the mode's eigenvalues.
    function filament_nm_amplitudes_weighted_sum(self) result(amp_sum)
        implicit none
        ! Arguments
        type(filament), intent(in)  ::  self
        ! Result
        real(kind=8)    ::  amp_sum
        ! Private variables
        integer ::  current_mode
        logical,    parameter   ::  debug   =   .true.
        ! Relative weights
        real(kind=8)   ::  rel_weight_xyz
        real(kind=8)   ::  rel_weight_psitheta
        real(kind=8)   ::  rel_weight_sphi
        real(kind=8)    ::  dummy
        ! Start work
        amp_sum = 0.0D0

        ! User-set values
        rel_weight_xyz      =   1.0D0
        rel_weight_psitheta =   3.0D0
        rel_weight_sphi     =   1.0D0

        ! Normalise so that the total power is maintained (assume equal power in each mode)
        dummy = (3.*REL_WEIGHT_XYZ + 2.*REL_WEIGHT_PSITHETA + 2.*REL_WEIGHT_SPHI) / 7.
        rel_weight_xyz = rel_weight_xyz / dummy
        rel_weight_psitheta = rel_weight_psitheta / dummy
        rel_weight_sphi = rel_weight_sphi / dummy


        ! Loop over each mode, ignore the first mode which has no associated energy
        do current_mode = 2,self%nwp
            amp_sum = amp_sum + self%nm_eigenvalues(current_mode) * (             &
                                    rel_weight_xyz      * abs(self%wp_nm(current_mode)%coo(1))      +   &
                                    rel_weight_xyz      * abs(self%wp_nm(current_mode)%coo(2))      +   &
                                    rel_weight_xyz      * abs(self%wp_nm(current_mode)%coo(3))      +   &
                                    rel_weight_sphi     * abs(self%wp_nm(current_mode)%euler%phi)   +   &
                                    rel_weight_psitheta * abs(self%wp_nm(current_mode)%euler%theta) +   &
                                    rel_weight_psitheta * abs(self%wp_nm(current_mode)%euler%psi)   +   &
                                    rel_weight_sphi     * abs(self%wp_nm(current_mode)%subunit_num)     &
                                                                                    )
        enddo

        if (debug) then
            write(*,'(a,i0,a,F0.3)') '**debug(filament_nm_amplitudes_weighted_sum): weighted sum of nm amplitudes for filament ', &
                                        self%id, ' : ', amp_sum
        endif
    end function filament_nm_amplitudes_weighted_sum

    !>  \brief  Convert nm amplitudes to waypoint data
    subroutine filament_nm_amplitudes_to_wps(self)
        use runtime_parameters, only : terminate
        use filaments_geometry, only : filament_update_wps_from_wp_res
        implicit none
        ! Aruguments
        type(filament), intent(inout)  ::  self
        ! Private variables
        integer ::  current_wp, current_mode
        ! Start work

        ! Sanity checks
        if (size(self%nm_eigenvectors,2) .ne. self%nwp) then
            call terminate('Failed sanity check')
        endif

        ! The x coordinates along the filament can be recovered
        ! by doing a sum of the eigenvectors weighted by the
        ! amplitudes stored in the x coordinates of the nm WPs
        do CURRENT_WP=1,self%nwp
            ! Reset
            self%wp_res(current_wp)%coo         = 0.0D0
            self%wp_res(current_wp)%euler%phi   = 0.0D0
            self%wp(current_wp)%euler%theta     = 0.0D0
            self%wp(current_wp)%euler%psi       = 0.0D0
            self%wp_res(current_wp)%subunit_num = 0.0D0
            do current_mode = 1, SIZE(self%nm_eigenvectors,2)
                if (filament_refine_axis_nms(self)) then
                    self%wp_res(current_wp)%coo(1:3)    = self%wp_res(current_wp)%coo(1:3) + &
                                                            self%wp_nm(current_mode)%coo(1:3) * &
                                                            self%nm_eigenvectors(current_wp,current_mode)
                    self%wp(current_wp)%euler%theta     = self%wp(current_wp)%euler%theta + &
                                                            self%wp_nm(current_mode)%euler%theta * &
                                                            self%nm_eigenvectors(current_wp,current_mode)
                    self%wp(current_wp)%euler%psi       = self%wp(current_wp)%euler%psi + &
                                                            self%wp_nm(current_mode)%euler%psi * &
                                                            self%nm_eigenvectors(current_wp,current_mode)
                endif
                ! The helical parameters
                if (self%refine_normal_modes(4)) then
                    self%wp_res(current_wp)%euler%phi   = self%wp_res(current_wp)%euler%phi + &
                                                            self%wp_nm(current_mode)%euler%phi * &
                                                            self%nm_eigenvectors(current_wp,current_mode)
                endif
                if (self%refine_normal_modes(7)) then
                    self%wp_res(current_wp)%subunit_num = self%wp_res(current_wp)%subunit_num + &
                                                            self%wp_nm(current_mode)%subunit_num * &
                                                            self%nm_eigenvectors(current_wp,current_mode)
                endif
            enddo
        enddo

        ! Now we need to convert residuals to actual waypoint data for the relevant dimensions (x,y,z,phi,s)
        call filament_update_wps_from_wp_res(self,  AXIS=FILAMENT_REFINE_AXIS_NMS(self),    &
                                                    PHI=SELF%refine_normal_modes(4),        &
                                                    SUBUNIT_NUM=SELF%refine_normal_modes(7))

    end subroutine filament_nm_amplitudes_to_wps


    !>  \brief  Convert waypoint data to nm amplitudes
    subroutine filament_wps_to_nm_amplitudes(self)
        use runtime_parameters, only : terminate
        use filaments_geometry, only : filament_psi_theta_smooth
        use coos, only : wp_print_info
        implicit none
        ! Aruguments
        type(filament), intent(inout)  ::  self
        ! Private variables
        integer ::  current_wp, current_mode
        logical,    parameter   ::  debug    = .false.
        ! Start work

        ! Sanity checks
        if (.not. allocated(self%nm_eigenvectors)) then
            write(*,'(a)') '**ERROR(filament_wps_to_nm_amplitudes): eigenvectors array not allocated'
            call terminate('Fatal error in filament_wps_to_nm_amplitudes')
        endif
        if (size(self%nm_eigenvectors,2) .ne. self%nwp) then
            write(*,'(a,3(i0,1x))') '**ERROR(filament_wps_to_nm_amplitudes): eigenvectors dim, nwp = ', &
                                                                       size(self%nm_eigenvectors,1), &
                                                                       size(self%nm_eigenvectors,2), &
                                                                       self%nwp
            call terminate('Failed sanity check')
        endif

        if (debug) then
            write(*,'(a)') '**debug(filament_wps_to_nm_amplitudes): NMs follow'
            call filament_print_nms(self)
        endif

        ! Allocate wp_nm array if necessary
        if (.not. allocated(self%wp_nm)) then
            allocate(self%wp_nm(self%nwp))
        endif

        ! Make sure psi & theta are as smooth as possible for our nm calculation
        call filament_psi_theta_smooth(self)

        !
        do current_mode=1,size(self%nm_eigenvectors,2)
            !
            self%wp_nm(current_mode)%coo = 0.0D0
            self%wp_nm(current_mode)%euler%phi = 0.0D0
            self%wp_nm(current_mode)%euler%theta = 0.0D0
            self%wp_nm(current_mode)%euler%psi = 0.0D0
            self%wp_nm(current_mode)%subunit_num = 0.0D0
            !
            do CURRENT_WP=1,self%nwp
                self%wp_nm(current_mode)%coo = self%wp_nm(current_mode)%coo + &
                                            self%wp_res(current_wp)%coo * self%nm_eigenvectors(current_wp,current_mode)
                self%wp_nm(current_mode)%euler%phi = self%wp_nm(current_mode)%euler%phi + &
                                            self%wp_res(current_wp)%euler%phi * self%nm_eigenvectors(current_wp,current_mode)
                self%wp_nm(current_mode)%euler%theta = self%wp_nm(current_mode)%euler%theta + &
                                            self%wp(current_wp)%euler%theta * self%nm_eigenvectors(current_wp,current_mode)
                self%wp_nm(current_mode)%euler%psi = self%wp_nm(current_mode)%euler%psi + &
                                            self%wp(current_wp)%euler%psi * self%nm_eigenvectors(current_wp,current_mode)
                self%wp_nm(current_mode)%subunit_num = self%wp_nm(current_mode)%subunit_num + &
                                            self%wp_res(current_wp)%subunit_num * self%nm_eigenvectors(current_wp,current_mode)
            enddo
            if (debug) then
                write(*,'(a,i0)') '**debug(filament_wps_to_nm_amplitudes): nm amplitudes for mode ', current_mode
                call wp_print_info(self%wp_nm(current_mode))
            endif
        enddo

    end subroutine filament_wps_to_nm_amplitudes


    !>  \brief  Print out normal modes for the filament
    subroutine filament_print_nms(self)
        implicit none
        ! Arguments
        type(filament), intent(in)  ::  self
        ! Private variables
        integer ::  i
        character(len=64) :: line_format
        ! Start work

        ! Set up the printout format
        line_format = ' '
        write(line_format,'(a,i0,a)') '(i4,x,g  10.3,x,', size(self%nm_eigenvectors,1), '(F8.3,x))'


        ! Do the printing
        write(*,'(a,i0)') 'Normal modes for filament ', self%id
        write(*,'(3(A,1x))') 'NM #', 'Eigenvalue', 'Eigenvector'
        do I=1,size(self%nm_eigenvalues)
            write(*,line_format) i, self%nm_eigenvalues(i), self%nm_eigenvectors(:,i)
        enddo

    end subroutine filament_print_nms

    !>  \brief  Sort eigenvectors by increasing eigenvalue (aka eigenfrequency)
    subroutine filament_sort_nms(self)
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        ! Private variables
        integer, allocatable ::  position_of_lowest_eigenvalue(:)
        integer ::  i
        real, allocatable    ::  current_eigenvector(:)
        real    ::current_eigenvalue
        ! Start work

        ! Loop over the vectors / values
        do I=1,size(self%nm_eigenvalues)-1
            ! Find the position of lowest eigenvalue
            position_of_lowest_eigenvalue = minloc(self%nm_eigenvalues(i:size(self%nm_eigenvalues))) + i - 1
            ! Move the eigenvectors and value to the current position
            ! First we copy to temp variables
            current_eigenvalue  = self%nm_eigenvalues(position_of_lowest_eigenvalue(1))
            current_eigenvector = self%nm_eigenvectors(:,position_of_lowest_eigenvalue(1))
            ! Then we clear space for them
            self%nm_eigenvalues(position_of_lowest_eigenvalue(1)) = self%nm_eigenvalues(i)
            self%nm_eigenvectors(:,position_of_lowest_eigenvalue(1)) = self%nm_eigenvectors(:,i)
            ! Now we place the vector/value in their correct, sorted position
            self%nm_eigenvalues(i) = current_eigenvalue
            self%nm_eigenvectors(:,i) = current_eigenvector
        enddo
    end subroutine filament_sort_nms

    !>  \brief  Update normal modes for the filament
    subroutine filament_update_nms(self)
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        ! Start work
        call filament_nms_from_path_lengths(self)
        !call filament_sort_nms(self) ! Not necessary, because the fsgsl eigen analysis routine sorted the vectors already
    end subroutine filament_update_nms

    !>  \brief  Compute eigenvectors & eigenvalues (i.e. normal modes) from the waypoints' path lengths
    !!
    !!  Following method suggested by Niko (11-Feb-2012)
    subroutine filament_nms_from_path_lengths(self)
        use eigen
        use runtime_parameters, only : terminate ! debug
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self    !<  Filament object
        ! Private variables
        integer ::  i,j,k
        real(kind=8)    ::  e((self%NWP-1)*2,self%nwp)
        real(kind=8)    ::  a(self%nwp,self%nwp)
        real(kind=8)    ::  x(self%NWP-1)
        !real(KIND=8)    ::  d(self%nwp), o(self%nwp)
        !real(KIND=8)    ::  z(self%nwp,self%nwp)
        real(kind=8)    ::  s
        real(kind=8),   parameter   ::  y1 = 1.0D0
        real(kind=8),   parameter   ::  overall_restoring_constant  =   0.0D0
        logical,        parameter   ::  debug   =   .false.
        logical,        parameter   ::  use_fgsl = .true.
        character(len=6)    ::  nwp_m, nwp_c
        ! Start work
        if (debug) then
            write(nwp_m,'(i6)') self%NWP-1
            write(nwp_c,'(i6)') self%nwp
        endif
        ! The x array contains the path-length distances from
        ! waypoint to waypoint
        do I=2,self%nwp
            x(I-1) = 1./(self%wp(i)%path_length - self%wp(I-1)%path_length)
        enddo

        if (debug) then
            write(*,'(a)') '**debug(filament_nms_from_path_lengths): coupling values:'
            write(*,'('//nwp_m//'(F0.2,1x))') (x(i),I=1,self%NWP-1)
        endif


        ! Set up e matrix (coupling matrix)
        e = 0.0D0
        do I=1,self%NWP-1
            e(i,i)   = -1.0D0
            e(I+SELF%NWP-1,i) = -1.0D0
            e(i,I+1) =  1.0D0
        enddo

        if (debug) then
            write(*,'(a)') '**debug(filament_nms_from_path_lengths): coupling matrix:'
            do I=1,self%NWP-1
                write(*,'('//nwp_m//'(F7.3,1x))')  (e(j,i),J=1,self%NWP-1)
            enddo
        endif

        ! Find the eigenvectors & values for the matrix
        if (use_fgsl) then
            ! Set up a matrix (eigen matrix)
            do I=1,self%nwp
                do J=1,self%nwp
                    S=0.0D0
                    do K=1,self%NWP-1
                        S=S+Y1*E(k,i)*E(k,j)*X(k)
                        S=S+OVERALL_RESTORING_CONSTANT*E(K+SELF%NWP-1,i)*E(K+SELF%NWP-1,j)
                    enddo
                    a(i,j) = s
                enddo
            enddo

            if (debug) then
                write(*,'(a)') '**debug(filament_nms_from_path_lengths): eigen matrix:'
                do I=1,self%nwp
                    write(*,'('//nwp_c//'(F7.3,1x))') (a(j,i),J=1,self%nwp)
                enddo
            endif

            call matrix_eigen_analysis(a,self%nm_eigenvectors,self%nm_eigenvalues)
            self%nm_eigenvectors = transpose(self%nm_eigenvectors)
            if (debug) then
                print *, 'SELF%nm_eigenvalues NEW', self%nm_eigenvalues
                print *, 'SELF%nm_eigenvectors NEW', self%nm_eigenvectors
            endif
        endif




        if (debug) then
            print *, '**DEBUG(filaments_normal_modes): EIGENVECTORS'
            print *, self%nm_eigenvectors
            print *, '**DEBUG(filaments_normal_modes): end EIGENVECTORS'
            !call filament_print_nms(self)
            !call filament_wps_to_nm_amplitudes(self)
            !write(*,'(a)') '**DEBUG(filament_nms_from_path_lengths): wp nm array:'
            !call waypoints_print_info(self%wp_nm)
            !call filament_nm_amplitudes_to_wps(self)
            !write(*,'(a)') '**DEBUG(filament_nms_from_path_lengths): reconstructed wp array:'
            !call waypoints_print_info(self%wp)
            !call terminate('DEBUG')
        endif

    end subroutine filament_nms_from_path_lengths
end module filaments_normal_modes
