!>  \brief  Routines to deal with the arc length parametrisation of the filament's axial space curve
module filaments_parametrisation
    use filaments_base
    implicit none

    contains

    !>  \brief  Iteratively refine the parametrisation of the splines until the target error tolerance is reached
    subroutine filament_parametrisation_update(self,tolerance)
        use runtime_parameters, only : terminate
        use filaments_geometry, only : filament_solve_splines
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  self
        !logical,            optional,       intent(in)      ::  use_psi_theta_constraints   !<  Whether the psi & theta angles are known for each waypoint and should be used as constraints when solving splines
        real,               optional,       intent(in)      ::  tolerance                   !<  Parametrisation error tolerance (pixels)
        ! Private variables
        integer                 ::  parametrisation_num_iterations
        real                    ::  previous_parametrisation_error
        integer,    parameter   ::  parametrisation_max_num_iterations  =   300
        integer,    parameter   ::  parametrisation_min_num_iterations  =   2
        real(kind=8)            ::  parametrisation_target_error
        real(kind=8)            ::  current_parametrisation_error
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(filament_parametrisation_update): filament info on entry:'
            call self%PrintInfo()
        endif

        !if (present(use_psi_theta_constraints)) call terminate('USE_PSI_THETA_CONSTAINTS is deprecated here')

        ! Work out tolerance
        if (present(tolerance)) then
            parametrisation_target_error = dble(tolerance)
        else
            if (self%helix%rise .eq. 0.0) then
                parametrisation_target_error = 0.1D0
                write(*,'(2a)') '**warning(filament_parametrisation_update): using parametrisation tolerance of 0.1 ', &
                                'because helical parameters were not supplied. This may not be sufficient in some cases'
            else
                if (self%helix%pixel_units) then
                    parametrisation_target_error = min(self%helix%rise / 100.D0,0.1D0)
                else
                    write(*,'(2a)') '**ERROR(filament_parametrisation_update): helical symmetry needs to be in ',  &
                                    'pixel units to work out tolerance for parametrisation'
                    call terminate('Fatal error in FILAMENT_PARAMETRISATION_UPDATE')
                endif
            endif
        endif

        if (.not. any(self%wp%path_length .ne. 0.0)) then
            call filament_update_path_length(self)
        endif

        ! Iterative parametrisation
        parametrisation_num_iterations = 0
        previous_parametrisation_error = huge(1.0)
        do
            parametrisation_num_iterations = parametrisation_num_iterations + 1
            call self%SetTFromPathLength()
            if (debug) then
                write(*,'(a,i0)') '**debug(filament_parametrisation_update): (before solve_splines) arc length ',   &
                                    'parametrisation, iteration #', parametrisation_num_iterations
                write(*,'(a)') '**debug(filament_parametrisation_update): filament info:'
                call self%PrintInfo()
            endif
            call filament_solve_splines(self)   !,USE_PSI_THETA_CONSTRAINTS=USE_PSI_THETA_CONSTRAINTS) ! Commented out 29-Nov-2012 with introduction of use_hermite_splines parameter
            if (debug) then
                write(*,'(a,i0)') '**debug(filament_parametrisation_update): (after solve_splines) arc ',   &
                                'length parametrisation, iteration #', parametrisation_num_iterations
                write(*,'(a)') '**debug(filament_parametrisation_update): filament info:'
                call self%PrintInfo()
            endif
            current_parametrisation_error = filament_parametrisation_error(self)
            if (debug) then
                write(*,'(a,i0,a,F0.11,a)') '**debug(filament_parametrisation_update): arc length parametrisation', &
                                            ', iteration #', parametrisation_num_iterations, '; parametrisation error = ', &
                                            current_parametrisation_error, ' pixels'
                !write(*,'(a)') '**debug(filament_parametrisation_update): filament info:'
                !call filament_print_info(self)
            endif
            if (    current_parametrisation_error .lt. parametrisation_target_error &
                    .and. parametrisation_num_iterations .ge. parametrisation_min_num_iterations &
                    !.AND. current_parametrisation_error .LE. previous_parametrisation_error    &
                    ) then
                exit
            endif
            previous_parametrisation_error = current_parametrisation_error
            if (parametrisation_num_iterations .gt. parametrisation_max_num_iterations) then
                write(*,'(2a,i0,2(a,f13.10))')   '**warning(filament_parametrisation_update): spline parametrisation ',  &
                                                'did not converge after ', parametrisation_num_iterations, &
                                                        ' iterations. Current error = ', filament_parametrisation_error(self), &
                                                        '. Target error =', parametrisation_target_error
                if (parametrisation_num_iterations .gt. 2*parametrisation_max_num_iterations) then
                    call terminate('Fatal error in FILAMENT_PARAMETRISATION_UPDATE')
                endif
            endif
        enddo

        if (debug) then
            write(*,'(a)') '**debug(filament_parametrisation_update): filament info on exit:'
            call self%PrintInfo()
        endif

    end subroutine filament_parametrisation_update



    !>  \brief  Measure the discrepency between the filament's spline parametrisation and the current estimates of the arc lengths between waypoints
    function filament_parametrisation_error(self) result(error)
        use runtime_parameters, only : terminate
#ifndef __GFORTRAN__
        use, intrinsic :: ieee_arithmetic, only : ieee_is_nan
#endif
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self
        ! Result
        real(kind=8)                    ::  error   !< Average over waypoints of the difference (pixels) between the parameter value and the path length
        ! Private variables
        integer ::  i
        logical,    parameter   ::  debug = .false.
        real(kind=8) :: dummy
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(filament_parametrisation_error): filament info on entry'
            call self%PrintInfo()
            dummy = self%wp_nm(2)%coo(1)
        endif

        ! Update the path lengths (this is done by walking in tiny steps along the parameter and
        ! integrating the path lengths)
        if (debug) then
            write(*,'(a)') '**debug(filament_parametrisation_error): about to update path length'
        endif
        call filament_update_path_length(self)

        if (debug) then
            write(*,'(a)') '**debug(filament_parametrisation_error): filament info after update path length'
            call self%PrintInfo()
            if (self%wp_nm(2)%coo(1) .ne. dummy) then
                print *, self%wp_nm(2)%coo(1), dummy
                call terminate ('BUG')
            endif
        endif

        ! Sanity check. First waypoint should have parameter and path_length of 0.0
        if (self%wp(1)%path_length .ne. 0.0) then
            write(*,'(a,F13.8)') '**ERROR(filament_parametrisation_error): non-zero path length for first waypoint: ', &
                                self%wp(1)%path_length
            call terminate('Fatal error in FILAMENT_PARAMETRISATION_ERROR')
        endif

        ! Compute error
        error = 0.0D0
        do i=1,self%nwp
            error = error + abs(self%t(i)-self%t(1)-self%wp(i)%path_length)
        enddo
        error = error / self%nwp

        !
        if (debug) then
#ifdef __GFORTRAN__
            if (isnan(error)) then
#else
            if (ieee_is_nan(error)) then
#endif
                write(*,'(a)') '**ERROR(filament_parametrisation_error): NaN will be returned'
                call self%PrintInfo()
                call terminate('Fatal error in FILAMENT_PARAMETRISATION_ERROR')
            endif
        endif

        if (debug) then
            write(*,'(a)') '**debug(filament_parametrisation_error): filament info on exit'
            call self%PrintInfo()
        endif
    end function filament_parametrisation_error

    !>  \brief  For each waypoint of the filament, compute an estimate of the parameter value for spline interpolation
    subroutine filament_parametrisation_initial_guess(self,guess_fal)
        !use filaments_deprecated, only: coos_fil_to_spline
        use filaments_geometry, only : filament_wp_path_lengths_increase
        use coos, only : waypoint_waypoint_distance, waypoints_print_info
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  self            !<  Filament
        logical,            optional,       intent(in)      ::  guess_fal       !<  The coordinates for first and last waypoints are not known. Their parameters will be "guessed".
        ! Private variables
        character(len=*),   parameter   ::  param_scheme    =   'CHORD_LENGTH'  !<  Parametrisation scheme to use. See "Spline Parametrisation" document for detailled descriptions. Possible values: minimal, chord, cumulative_chord, centripetal
        integer                         ::  i
        integer                         ::  first_wp, last_wp
        logical                         ::  gguess_fal
        logical,            parameter   ::  debug           = .false.
        character(len=6)                ::  nwp_c
        ! Start work

        ! Check: this subroutine is only really meant for filaments that don't already have path-length estimates or whose path lengths are messed up somehow
        write(nwp_c,'(i6)') self%nwp
        if (any(self%wp%path_length .ne. 0.0) .and. filament_wp_path_lengths_increase(self)) then
            write(*,'(a)') '**ERROR(filament_initial_parametrisation): filament already appears to have path-length estimates'
            if (debug) then
                write(*,'(/a4,1x,'//nwp_c//'(i8,1x))'  ) 'WP #',(i,I=1,self%nwp)
                write(*,'(a4,1x,'//nwp_c//'(F8.3,1x)/)') 'WP T',self%t(:)
            endif
            call terminate('Fatal error in FILAMENT_INITIAL_PARAMETRISATION')
        endif

        if (debug) then
            write(*,'(a)') '**debug(filament_parametrisation_initial_guess): waypoints on entry'
            call waypoints_print_info(self%wp)
        endif

        ! Are we ignoring first and last waypoints?
        if (present(guess_fal)) then
            gguess_fal = guess_fal
        else
            gguess_fal = .false.
        endif

        ! Work out which are the first and last waypoints to which we will be assigning parameter values
        if (gguess_fal) then
            first_wp = 2
            last_wp  = self%nwp -1
        else
            first_wp = 1
            last_wp  = self%nwp
        endif

        ! Parametrise
        select case (param_scheme)
            !case ('MINIMAL')
            !    do I=FIRST_WP,last_wp
            !        tmp(1:3) = coos_fil_to_spline(self,self%wp(i)%coo(1:3),IGNORE_FIRST_WP=GGUESS_FAL)
            !        self%t(i) = tmp(1)
            !    enddo
            case ('CHORD_LENGTH')
                self%t(first_wp)    =   0.0
                do I=FIRST_WP,LAST_WP-1
                    self%t(I+1) = self%t(i) + waypoint_waypoint_distance(self%wp(i),self%wp(I+1))
                enddo
            case ('CENTRIPETAL')
                self%t(first_wp)    =   0.0
                do I=FIRST_WP,LAST_WP-1
                    self%t(I+1) = self%t(i) + sqrt(waypoint_waypoint_distance(self%wp(i),self%wp(I+1)))
                enddo
            case default
                write(*,'(a)') '**ERROR(filament_initial_parametrisation): unknown parametrisation scheme: ', &
                                trim(adjustl(param_scheme))
                call terminate('Fatal error in FILAMENT_INITIAL_PARAMETRISATION')
        end select

        if (debug) then
            write(*,'(a)') '**debug(filament_parametrisation_initial_guess): before gguess_fal'
            write(*,'(/a4,1x,'//nwp_c//'(i8,1x))'  ) 'WP #',(i,I=1,self%nwp)
            write(*,'(a4,1x,'//nwp_c//'(F8.3,1x)/)') 'WP T',self%t(:)
        endif

        ! If we need to make guesses for fal wp
        if (gguess_fal) then
            self%t(1)           =   self%t(2) - (self%t(3)-SELF%t(2))*0.5
            self%t(self%nwp)    =   self%t(self%NWP-1) + (self%t(self%NWP-1)-SELF%t(self%NWP-2))*0.5
        endif

        if (debug) then
            write(*,'(a)') '**debug(filament_parametrisation_initial_guess): after gguess_fal'
            write(*,'(/a4,1x,'//nwp_c//'(i8,1x))'  ) 'WP #',(i,I=1,self%nwp)
            write(*,'(a4,1x,'//nwp_c//'(F8.3,1x)/)') 'WP T',self%t(:)
        endif


        do I=SELF%nwp,1,-1
            self%t(i) = self%t(i) - self%t(1)
        enddo

        if (debug) then
            write(*,'(a)') '**debug(filament_parametrisation_initial_guess): after reset'
            write(*,'(/a4,1x,'//nwp_c//'(i8,1x))'  ) 'WP #',(i,I=1,self%nwp)
            write(*,'(a4,1x,'//nwp_c//'(F8.3,1x)/)') 'WP T',self%t(:)
        endif

        ! Sanity checks
        if (self%t(1) .ne. 0.0) then
            write(*,'(a,F0.4)') '**ERROR(filament_initial_parametrisation): sanity check failed, self%t(1) = ', self%t(1)
            call terminate('Fatal error in FILAMENT_INITIAL_PARAMETRISATION')
        endif

        ! Diagnostic message
        if (debug) then
            write(*,'(a)') '**debug(filament_initial_parametrisation): initial parameters: '
            write(*,'(a4,1x,'//nwp_c//'(i8,1x))'  ) 'WP #',(i,I=1,self%nwp)
            write(*,'(a4,1x,'//nwp_c//'(F8.3,1x))') 'WP T',self%t(:)
        endif
    end subroutine filament_parametrisation_initial_guess

    !>  \brief  Compute the path length from first waypoint to every other waypoint in the filament.
#if defined(skip_runtime_checks)
    pure &
#endif
    subroutine filament_update_path_length(self)
        use runtime_parameters, only : terminate
        use filaments_geometry, only : filament_eval_path_length
        implicit none
        ! Arguments
        type(filament), intent(inout)   ::  self    !<  Filament to be updated
        ! Private variables
        integer ::  i
        logical,    parameter   ::  debug   =   .false.
        character(len=6) :: nwp_c
        ! Start work
        !if (debug) then
        !    write(*,'(a)') '**debug(filaments_update_path_length): filament info on entry'
        !    call filament_print_info(self)
        !endif
        ! Start at the origin
        self%wp(1)%path_length = 0.0
        do I=2,self%nwp
            if (debug) write(*,'(a,i0,1x,2(F0.5,1x))') '**debug(filament_update_path_length): about to  eval path length ', i, &
                       self%t(i), self%wp(i)%path_length
            !call filament_print_info(self)
            call filament_eval_path_length(self,self%t(i),self%wp(i)%path_length)
        enddo
        if (debug) then
#ifndef skip_runtime_checks
            write(nwp_c,'(i6)') self%nwp
            if (any(self%wp%path_length .ne. 0.)) then
                ! Sanity check
                do I=2,self%nwp
                    if (self%wp(i)%path_length .le. self%wp(I-1)%path_length) then
                        write(*,'(a,'//nwp_c//'(F0.3,1x))') '**ERROR(filament_update_path_length): weird path lengths: ', &
                                self%wp%path_length
                        call terminate('Fatal error in FILAMENT_UPDATE_PATH_LENGTH')
                    endif
                enddo
            endif
#endif
        endif
    end subroutine filament_update_path_length



end module filaments_parametrisation
