!>  \brief  Methods for the filament class related to paths
!!
!!  Paths are arrays of waypoints which lie along the helical path of a filament.
!!
module filaments_path
    use filaments_base
    use runtime_parameters, only : terminate
    implicit none

    contains

    !>  \brief  Return a waypoint (with x,y,z coordinates and Euler angles) of the filament self at t
    function waypoint_from_t(self,t) result(output_wp)
        use float_comparisons, only : dbles_are_different
        use filaments_geometry, only : nearest_previous_wp_from_t, filament_eval_helical_rise, filament_eval_helical_twist, &
                                        filament_eval_curvature, nearest_wp_from_t, &
                                        filament_eval_overtwist, filament_eval_stretch, &
                                        filament_eval_principal_curvatures
        use differential_geometry, only : euler_from_derivatives_new, d_from_theta_psi
        use Splines
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  self                        !<  Filament
        real(kind=8),                       intent(in)      ::  t                           !<  Spline parameter (assumed to be == path length + constant)
        ! Result
        type(waypoint)  ::  output_wp
        ! Private variables
        integer                 ::  wp_before, wp_nearest
        real(kind=8)            ::  ddx,ddy,ddz
        real(kind=8)            ::  tt, dist_to_nearest_wp
        logical,    parameter   ::  debug = .false.
        real(kind=8)            ::  average_twist_per_distance, average_ds_per_distance
        integer                 ::  first_wp, last_wp
        real(kind=8)            ::  c_a, c_b
        ! Start work

        if (debug) write(*,'(a,F0.6)') '**debug(waypoint_from_path_length): input parameter value = ', t

        if (self%nwp .eq. 1) then
            output_wp = self%wp(1)
!#ifndef SKIP_RUNTIME_CHECKS
            if (all(self%wp(1)%dcoo .eq. 0.0d0)) then
                output_wp%dcoo = d_from_theta_psi(output_wp%euler%theta,output_wp%euler%psi)
                !call terminate('Fatal error in waypoint_from_t. Derivatives of only waypoint are not set.')
            endif
!#endif
            output_wp%coo(:) = self%wp(1)%coo(:) + (t-self%t(1)) * self%wp(1)%dcoo(:)
            output_wp%euler%phi = self%wp(1)%euler%phi + (t-self%t(1)) * self%wp(1)%helix%twist / self%wp(1)%helix%rise
            output_wp%subunit_num = self%wp(1)%subunit_num + (t-self%t(1)) / self%wp(1)%helix%rise
        else
            ! Copy value over
            tt = t

            ! Find between which filament waypoints we are
            wp_before = nearest_previous_wp_from_t(self,tt)
            if (wp_before .eq. 0) then
                write(*,'(a,F0.12,a,i0,a,F0.12,a,F0.12)') '**WARNING(waypoint_from_t): t (', tt, &
                                                            ') falls beyond the ends of filament ', self%id, &
                                                            ', which goes from ', self%t(1), ' to ', self%t(self%nwp)
                write(*,'(a,G15.6)') '**debug(waypoint_from_t): the spacing is ', spacing(tt)
                !call terminate('Fatal error in WAYPOINT_FROM_T')

                ! It could be that we are dealing with a simple rounding error. Let's check and keep going.
                if (.not. dbles_are_different(tt,self%t(1))) then
                    tt = self%t(1)
                endif
            endif

            ! We may need to know the waypoint we are nearest to
            if (self%fragmented_phi .or. self%fragmented_subunit_num .or. self%wps_as_sps) then
                call nearest_wp_from_t(self,tt,wp_nearest,dist_to_nearest_wp)
            endif

            ! Work out X,Y,Z and their derivatives
            if (self%nwp .eq. 2) then
                if (dbles_are_different(tt,self%t(1))) then
                    output_wp%coo(:) = self%wp(1)%coo(:) + (self%wp(2)%coo(:) - self%wp(1)%coo(:))/(tt-self%t(1))
                    output_wp%dcoo(:) = self%wp(1)%dcoo(:)
                else
                    output_wp = self%wp(1)
                endif
            else
                if (self%wps_as_sps) then
                    output_wp%coo(:) = self%wp(wp_nearest)%coo(:) + dist_to_nearest_wp * self%wp(wp_nearest)%dcoo(:)
                    output_wp%dcoo(:) = self%wp(wp_nearest)%dcoo(:)
                else
                    ! Evaluate the splines and their derivatives so we can work out Euler angles
                    call self%x_spline%HornerEvaluateAll(tt,output_wp%coo(1),output_wp%dcoo(1),ddx)
                    call self%y_spline%HornerEvaluateAll(tt,output_wp%coo(2),output_wp%dcoo(2),ddy)
                    call self%z_spline%HornerEvaluateAll(tt,output_wp%coo(3),output_wp%dcoo(3),ddz)
                endif
            endif
            if (debug) then
                write(*,'(a,3(F0.3,1x))') '**debug(waypoint_from_path_length):  x, y, z = ', &
                                            output_wp%coo(1), output_wp%coo(2), output_wp%coo(3)
                write(*,'(a,3(F0.3,1x))') '**debug(waypoint_from_path_length): dx,dy,dz = ', &
                                            output_wp%dcoo(1), output_wp%dcoo(2), output_wp%dcoo(3)
            endif

            ! Work out the Euler angles
            if (self%wps_as_sps) then
                ! Let's just copy the Eule angles over
                output_wp%euler%theta = self%wp(wp_nearest)%euler%theta
                output_wp%euler%psi   = self%wp(wp_nearest)%euler%psi
            else
                call euler_from_derivatives_new(output_wp%dcoo(1),output_wp%dcoo(2),output_wp%dcoo(3), &
                                                self%wp(2)%euler%theta .lt. 180.0D0, &
                                                output_wp%euler%theta,output_wp%euler%psi)
            endif


            ! Helical parameters
            if (self%fragmented_phi .or. self%fragmented_subunit_num .or. self%wps_as_sps) then
                output_wp%helix            =   self%wp(wp_nearest)%helix
                if (self%fragmented_phi .or. self%wps_as_sps) then
                    output_wp%euler%phi     =   self%wp(wp_nearest)%euler%phi &
                                                + dist_to_nearest_wp * self%wp(wp_nearest)%helix%twist   &
                                                                    / self%wp(wp_nearest)%helix%rise
                else
                    if (self%nwp .eq. 2) then
                        output_wp%euler%phi = self%wp(1)%euler%phi + (self%wp(2)%euler%phi - self%wp(1)%euler%phi)/(tt-self%t(1))
                    else
                        call self%twist_spline%Evaluate(tt,output_wp%euler%phi)
                    endif
                endif
                if (self%fragmented_subunit_num .or. self%wps_as_sps) then
                    output_wp%subunit_num  =   self%wp(wp_nearest)%subunit_num + dist_to_nearest_wp / self%wp(wp_nearest)%helix%rise
                else
                    if (self%nwp .eq. 2) then
                        output_wp%subunit_num = self%wp(1)%subunit_num &
                                             + (self%wp(2)%subunit_num - self%wp(1)%subunit_num)/(tt-self%t(1))
                    else
                        call self%s_spline%Evaluate(tt,output_wp%subunit_num)
                    endif
                endif
            else
                if (self%nwp .eq. 2) then
                    output_wp%euler%phi   = self%wp(1)%euler%phi + (self%wp(2)%euler%phi - self%wp(1)%euler%phi)/(tt-self%t(1))
                    output_wp%subunit_num = self%wp(1)%subunit_num &
                                         + (self%wp(2)%subunit_num - self%wp(1)%subunit_num)/(tt-self%t(1))
                    output_wp%helix       = self%wp(1)%helix
                    output_wp%helix%rise  = self%wp(1)%helix%rise + (self%wp(2)%helix%rise - self%wp(1)%helix%rise)/(tt-self%t(1))
                    output_wp%helix%twist = self%wp(1)%helix%twist &
                                         + (self%wp(2)%helix%twist - self%wp(1)%helix%twist)/(tt-self%t(1))
                else
                    call self%twist_spline%Evaluate(tt,output_wp%euler%phi)
                    ! Subunit number
                    call self%s_spline%Evaluate(tt,output_wp%subunit_num)
                    !
                    output_wp%helix        =   self%wp(wp_before)%helix
                    output_wp%helix%rise   =   filament_eval_helical_rise(self,tt)
                    output_wp%helix%twist  =   filament_eval_helical_twist(self,tt)
                endif
            endif

            ! Occupancy, sigma, LogP etc
            ! 151109: for now, we assume constant value for filament
            output_wp%occupancy   = self%wp(1)%occupancy
            output_wp%logp        = self%wp(1)%logp
            output_wp%noise_sigma = self%wp(1)%noise_sigma

            ! Compute local curvature
            output_wp%curvature    =   filament_eval_curvature(self,tt)
            ! Compute the local curvatures along the major axes
            !! todo: do only this, then compute the curvature as the norm of the two
            call filament_eval_principal_curvatures(self,tt,c_a,c_b)
            output_wp%curvature_a = c_a
            output_wp%curvature_b = c_b
            ! For the overtwist and overstretch, we need the averate twist and rise per distance along the filament
            ! todo: don't do this here, it should only be done once per filament - this is a waste!
            ! Also, fix the first and last wp
            !> \todo: don't do this here
            first_wp = 1
            last_wp = self%nwp
            average_twist_per_distance  =   &
                                (self%wp(last_wp)%euler%phi - self%wp(first_wp)%euler%phi) / (self%t(last_wp) - self%t(first_wp))
            average_ds_per_distance     =   &
                                (self%wp(last_wp)%subunit_num - self%wp(first_wp)%subunit_num)&
                                     / (self%t(last_wp) - self%t(first_wp))
            output_wp%overtwist    =   filament_eval_overtwist(self,tt,average_twist_per_distance)
            output_wp%stretch      =   filament_eval_stretch(self,tt,average_ds_per_distance)

            output_wp%refine       =   self%wp(wp_before)%refine .or. self%wp(wp_before+1)%refine
            output_wp%path_length  =   tt-self%t(1)

            if (debug) then
                write(*,'(a,3(F0.3,1x))') '**debug(waypoint_from_path_length): phi, theta, psi = ', &
                                            output_wp%euler%phi, output_wp%euler%theta, output_wp%euler%psi
            endif

            if(debug) then
                write(*,'(a)', advance='no') '**debug(waypoint_from_t): local subframe shifts allocated?'
                do wp_before=1,self%nwp
                    write(*,'(l1)',advance='no') allocated(self%wp(wp_before)%local_subframe_shift)
                enddo
                print *, ' '
            endif

            ! Work out subframe shifts
            call subframe_shifts_from_t(self,t,output_wp%local_subframe_shift)

            if (debug) then
                write(*,'(a)') '**debug(waypoint_from_t): local subframe shifts: '
                print *, output_wp%local_subframe_shift
            endif
        endif ! test for nwp == 1

    end function waypoint_from_t


    !>  \brief  For a given T, compute the local subframe shifts
    subroutine subframe_shifts_from_t(self,t,subframe_shifts)
        use filaments_geometry, only : nearest_previous_wp_from_t
        !
        type(Filament),                     intent(in)      ::  self
        real(kind=8),                       intent(in)      ::  t       !<  T (arc length) parameter
        real,               allocatable,    intent(inout)   ::  subframe_shifts(:,:)
        ! private variables
        integer ::  wp_before, wp_after
        real    ::  linear_interpol_factor
        ! start work

        wp_before = nearest_previous_wp_from_t(self,t)
        wp_after  = min(wp_before+1,self%nwp)

        if (wp_after .eq. wp_before) then
            ! we must be on the first or on the last
            linear_interpol_factor = 0.0
        else
            linear_interpol_factor = (t-self%t(wp_before))/(self%t(wp_after)-self%t(wp_before))
        endif

        if (.not. allocated(subframe_shifts)) allocate(subframe_shifts(3,size(self%wp(wp_before)%local_subframe_shift,2)))

		subframe_shifts(:,:) =      self%wp(wp_before)%local_subframe_shift(:,:) &
                                +  (self%wp(wp_after) %local_subframe_shift(:,:) - self%wp(wp_before)%local_subframe_shift(:,:)) &
                                *  linear_interpol_factor

    end subroutine subframe_shifts_from_t

    !>  \brief  For each waypoint, "convert" the value stored as phi Euler angle from twist angle to actual phi angle
    !subroutine waypoints_twist_to_phi(self)
    !    ! arguments
    !    type(waypoint),     intent(inout)   ::  self(:)
    !    ! private variables
    !    integer ::  current_wp
    !    ! start work
    !    do current_wp=1,size(self)
    !        if (self(current_wp)%euler%theta .lt. 0.0) then
    !            self(current_wp)%euler%phi = self(current_wp)%euler%phi + 180.0
    !        endif
    !    enddo
    !end subroutine waypoints_twist_to_phi



    !>  \brief  Return an array of waypoints which describes the path of the filament in the region of its waypoints.
    !!  We make the assumption that near the waypoints, the filament is perfectly straight.
    function filament_get_path_near_wps(self,step,num_wps_per_wp, &
                                        ignore_fal_wps,helix,step_in_repeats,intg_repeats) result(path)
        use runtime_parameters, only : is_even
        use helices, only : helical_symmetry_unit_convert, helical_symmetry_print_info
        use differential_geometry, only : d_from_theta_psi
        use coos, only : waypoints_print_info, wp_print_info, waypoints_grow_array
        use differential_geometry, only : waypoint_get_path_near_wp
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  self                    !<  Filament
        real(kind=8),                       intent(in)      ::  step                    !<  Step size (pixels)
        integer,                            intent(in)      ::  num_wps_per_wp          !<  Each incoming wp will "generate" num_wps_per_wp waypoints in the output path, including itself. This number does not include symmetry-related views.
        logical,                optional,   intent(in)      ::  ignore_fal_wps          !<  If .TRUE., the path will extend form the second to the second-to-last waypoint. .FALSE. by default.
        type(helical_symmetry), optional,   intent(in)      ::  helix                   !<  Helical parameters. If not given as argument, the helical parameters found in the filament's waypoints will be used instead.
        logical,                optional,   intent(in)      ::  step_in_repeats         !<  Whether the step argument was given in repeats
        logical,                optional,   intent(in)      ::  intg_repeats            !<  Return waypoints that coincide exactly with helical repeats
        ! Result
        type(waypoint),         allocatable                 ::  path(:)                 !<  Array of waypoints
        ! Private variables
        type(waypoint), allocatable ::  path_for_current_wp(:)
        logical                     ::  iignore_fal_wps
        integer                     ::  current_wp_num, first_wp, last_wp
        integer                     ::  i
        type(waypoint)              ::  current_wp
        real(kind=8)                ::  current_t
        logical                     ::  sstep_in_repeats, iintg_repeats
        type(helical_symmetry)      ::  helix_to_use

        ! Start work

        ! Not coded for microtubules yet
        if (self%microtubule) then
            call terminate('Not coded for microtubules yet')
        endif

        ! We need the input waypoints' helical parameters to be in pixel units
        do i=1,self%nwp
            if (.not. self%wp(i)%helix%pixel_units) then
                call helical_symmetry_unit_convert(self%wp(i)%helix,pixel_size=self%film%GetPixelSize())
            endif
        enddo

        ! Step in repeats?
        sstep_in_repeats = .false.
        if (present(step_in_repeats)) sstep_in_repeats = step_in_repeats

        ! Return waypoints that coincide with subunits?
        iintg_repeats = .false.
        if (present(intg_repeats)) iintg_repeats = intg_repeats
        if (iintg_repeats .and. .not. sstep_in_repeats) then
            write(*,'(2a)') '**ERROR(filament_get_path_near_wps): option to return waypoints on ',  &
                            'repeats only available when step size given in repeats'
            call terminate('Feature not yet implemented')
        endif



        ! Ignore first and last waypoints?
        iignore_fal_wps = .false.
        if (present(ignore_fal_wps)) iignore_fal_wps = ignore_fal_wps

        ! Work out the start and finish waypoints
        if (iignore_fal_wps) then
            first_wp    = 2
            last_wp     = self%nwp - 1
        else
            first_wp    = 1
            last_wp     = self%nwp
        endif

        ! Loop over (input) waypoints
        do current_wp_num = first_wp,last_wp

            ! We may need to consider only waypoints that coincide precisely with subunits
            if (iintg_repeats) then
                if (self%wps_as_sps) then
                    if (abs(self%wp(current_wp_num)%subunit_num - nint(self%wp(current_wp_num)%subunit_num)) .gt. 0.001) then
                        call this_program%TerminateWithFatalError('filament_get_path_near_wps', &
                                                        'in SP mode, finding near whole subunit is not yet implemented')
                    else
                        current_wp = self%wp(current_wp_num)
                    endif
                else
                    current_t = self%t(current_wp_num)
                    current_t = filament_find_t_of_nearest_subunit(self,current_t)
                    current_wp = waypoint_from_t(self,current_t)
                endif
            else
                current_wp = self%wp(current_wp_num)
            endif

            if (present(helix)) then
                helix_to_use = helix
            else
                helix_to_use = current_wp%helix
            endif

            ! Get path near the current waypoint
            path_for_current_wp = waypoint_get_path_near_wp(helix_to_use,step,num_wps_per_wp,current_wp, &
                                                            step_in_repeats, &
                                            start_at_input_wp=.not. self%wps_as_sps .and. current_wp_num .eq. first_wp, &
                                            finish_at_input_wp=.not. self%wps_as_sps .and. current_wp_num .eq. last_wp)


            ! Append path to the output path
            call waypoints_grow_array(path,increment=size(path_for_current_wp))
            path(size(path)-size(path_for_current_wp)+1:) = path_for_current_wp

        enddo

    end function filament_get_path_near_wps



    !>  \brief  Return an array of waypoints along the filament's axial space curve, with specified step size in pixels
    !!  \todo   Rename ignore_fal_wps to ignore_fal_segments
    subroutine filament_get_path_pixel_steps(   self,step,path,    &
                                                ignore_fal_wps,include_finish_wp,   &
                                                start_wp,finish_wp,include_extra_wp,include_fal_wps, &
                                                start_t,finish_t,strict_start_finish_t, &
                                                integer_subunit_num)
        use coos, only : waypoints_grow_array, waypoints_print_info
        use Splines
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  self                    !<  Filament object
        real(kind=8),                       intent(in)      ::  step                    !<  Step size (pixels)
        type(waypoint),         allocatable,intent(out)     ::  path(:)                 !<  Resulting array of waypoints
        logical,                optional,   intent(in)      ::  ignore_fal_wps          !<  Whether to ignore segments between the first and second and penultimate and last waypoints
        logical,                optional,   intent(in)      ::  include_finish_wp       !<  Whether the last "active" waypoint should be included in the output path. Defaults to false.
        integer,                optional,   intent(in)      ::  start_wp                !<  Return waypoints only between this and the finish_wp waypoints
        integer,                optional,   intent(in)      ::  finish_wp               !<  Return waypoints only between start_wp and this waypoint
        logical,                optional,   intent(in)      ::  include_extra_wp        !<  Add an extra waypoint at the end, which may be beyond the finish_wp
        logical,                optional,   intent(in)      ::  include_fal_wps         !<  Even if the first and last segments were ignored, the fal wps can be included in the output path. Defaults to FALSE.
        real(kind=8),           optional,   intent(in)      ::  start_t                 !<  Overrides start_wp. Defines the starting point of the path.
        real(kind=8),           optional,   intent(in)      ::  finish_t                !<  Overrides finish_wp. Defines the finishing point of the path.
        logical,                optional,   intent(in)      ::  strict_start_finish_t   !<  Ensure that no waypoint will be allowed beyond those t values. By default, there is some flexibility there.
        logical,                optional,   intent(in)      ::  integer_subunit_num     !<  Return waypoints which coincide exactly with subunits

        ! Private variables
        logical,        parameter   ::  debug   =   .false.
        real(kind=8)                ::  current_t, sstart_t, ffinish_t
        integer                     ::  current_output_wp
        integer                     ::  sstart_wp, ffinish_wp
        logical                     ::  iignore_fal_wps
        logical                     ::  iinclude_finish_wp
        logical                     ::  last_output_wp
        logical                     ::  iinclude_extra_wp
        logical                     ::  iinclude_fal_wps
        type(waypoint), allocatable ::  path_tmp(:)
        logical                     ::  round_to_nearest_subunit
        integer                     ::  predicted_num_output_wps
        logical                     ::  sstrict_start_finish_t
        ! Start work

        ! Ignore first and last waypoints?
        iignore_fal_wps = .false.
        if (present(ignore_fal_wps)) iignore_fal_wps = ignore_fal_wps


        ! Will we force the finish wp to be included in the output path?
        iinclude_finish_wp = .false.
        if (present(include_finish_wp)) iinclude_finish_wp = include_finish_wp

        ! Will we add one output waypoint past the finish_wp
        iinclude_extra_wp = .false.
        if (present(include_extra_wp)) iinclude_extra_wp = include_extra_wp

        ! Include first and last waypoints (even if the first and last segments are ignored)
        iinclude_fal_wps = .false.
        if (present(include_fal_wps)) iinclude_fal_wps = include_fal_wps

        ! Work out the start and finish waypoints for the path
        if (iignore_fal_wps) then
            sstart_wp   = 2
            ffinish_wp  = self%nwp - 1
        else
            sstart_wp   = 1
            ffinish_wp  = self%nwp
        endif

        ! Integer subunit_num?
        round_to_nearest_subunit = .false.
        if (present(integer_subunit_num)) round_to_nearest_subunit = integer_subunit_num

        !if (debug) then
        !   write(*,'(a)') '**debug(filament_get_path_pixel_steps): filament info: '
        !   call filament_print_info(self)
        !endif

        ! The user may have supplied their own start and finish WPs, which will override
        if (present(start_wp)) then
            if (start_wp .eq. 1 .and. iignore_fal_wps) then
                write(*,'(a)') '**warning(filament_get_path_pixel_steps): you asked for path starting at waypoint 1 '//&
                                'but specified that fal WPs should be ignored. Will start at waypoint 2.'
                sstart_wp = 2
            else
                sstart_wp = start_wp
            endif
        endif
        if (present(finish_wp)) then
            if (finish_wp .eq. self%nwp .and. iignore_fal_wps) then
                write(*,'(a)') '**warning(filament_get_path_pixel_steps): you asked for path finishing at last waypoint '//&
                                'but specified that fal WPs should be ignored. Will finish at penultimate waypoint.'
                ffinish_wp = self%nwp-1
            else
                ffinish_wp = finish_wp
            endif
        endif

        ! The user may have supplied their own start and finish Ts, which will override start and finish wp
        sstart_t = self%t(sstart_wp)
        if (present(start_t)) sstart_t = start_t
        ffinish_t = self%t(ffinish_wp)
        if (present(finish_t)) ffinish_t = finish_t

        ! Start, finish t strictly enforced?
        sstrict_start_finish_t = .false.
        if (present(strict_start_finish_t)) sstrict_start_finish_t = strict_start_finish_t

        ! Sanity check: finish t must be greater than or equal to start t
        if (sstart_t .gt. ffinish_t) then
            write(*,'(a,2(F0.6,1x))') '**ERROR(filament_get_path_pixel_steps): sstart_t > ffinish_t: ', sstart_t, ffinish_t
            call self%PrintInfo()
            call terminate('Fatal error in FILAMENT_GET_PATH_PIXEL_STEPS')
        endif


        ! Allocate memory for the path
        if (allocated(path)) deallocate(path)

        if (debug) then
            write(*,'(a,3(F0.10,1x))') '**debug(filament_get_path_pixel_steps): start_t, finish_t, step: ', &
                                                sstart_t, ffinish_t, step
            ! We can predict how many waypoints we expect (ignoring for now fal waypoints we may have been asked to add anyway)
            predicted_num_output_wps = int((FFINISH_T-SSTART_T)/step) + 1
            write(*,'(a,i0)') '**debug(filament_get_path_pixel_steps): predicted number of output waypoints = ', &
                                    predicted_num_output_wps
        endif

        current_t           =   sstart_t
        ! We may need to explicitely add the first waypoint
        if (iignore_fal_wps .and. iinclude_fal_wps) then
            call waypoints_grow_array(path,increment=1)
            if (allocated(path_tmp)) deallocate(path_tmp)
            allocate(path_tmp(1))
            path_tmp(1) = self%wp(1)
            path = path_tmp
            current_output_wp   =   1
            if (debug) predicted_num_output_wps = predicted_num_output_wps + 1
        else
            current_output_wp   =   0
        endif

        ! Start looping
        last_output_wp      =   .false.
        do
            if (current_t .gt. ffinish_t) then
                if (sstrict_start_finish_t) exit
                if (iinclude_finish_wp) then
                    current_t = ffinish_t
                    last_output_wp = .true.
                else if (iinclude_extra_wp) then
                    if (current_t .lt. self%t(self%nwp)) then
                        last_output_wp = .true.
                    else
                        exit
                    endif
                else if (current_t-ffinish_t .lt. step*0.5d0 .and. current_t .lt. self%t(self%nwp)) then
                    last_output_wp = .true.
                else
                    ! If we're past the last waypoint and we've not been asked to include the last waypoint in the path, exit without adding
                    exit
                endif
            else if (current_t .eq. ffinish_t) then
                last_output_wp = .true.
                exit
            endif
            if (round_to_nearest_subunit) then
                current_t = filament_find_t_of_nearest_subunit(self,current_t)
            endif
            ! Add a waypoint to the end of the output path, without symmetry related views
            call add_waypoint_to_path(path,self,current_t)
            current_output_wp = current_output_wp + 1
            if (last_output_wp) exit
            current_t = current_t + step
        enddo

        ! We may need to explicitely add the last waypoint
        if (iignore_fal_wps .and. iinclude_fal_wps) then
            path_tmp = path
            call waypoints_grow_array(path_tmp,INCREMENT=1)
            path_tmp(size(path_tmp)) = self%wp(size(self%wp))
            path = path_tmp
            current_output_wp = current_output_wp + 1
        endif

        ! There are cases where the two last waypoints may be in the same place, which is wasteful and can cause problems elsewhere
        if (size(path) .gt. 1) then
            if (path(size(path))%path_length - path(size(path)-1)%path_length .le. 0.1D0) then
                ! Remove the penultimate point in the path
                path_tmp = path
                deallocate(path)
                allocate(path(size(path_tmp)-1))
                do current_output_wp=1,size(path_tmp)-2
                    path(current_output_wp) = path_tmp(current_output_wp)
                enddo
                ! The last point
                path(size(path)) = path_tmp(size(path_tmp))
            endif
        endif

        if (debug) then
            write(*,'(a,i0,a)') '**debug(filament_get_path_pixel_steps): full filament path with ', size(path), ' waypoints'
            call waypoints_print_info(path)
            write(*,'(a)') '**debug(filament_get_path_pixel_steps): all done'
        endif

        ! Sanity checks
        if (size(path) .ne. current_output_wp) then
            write (*,'(a,2(i0,1x))') '**ERROR(filament_get_path_pixel_steps): failed sanity check: ', size(path), current_output_wp
            call terminate('Fatal error in FILAMENT_GET_PATH_PIXEL_STEPS')
        endif
        if (round_to_nearest_subunit) then
            if (any (abs(path%subunit_num-nint(path%subunit_num)) .gt. 0.05)) then
                write(*,'(a)') '**ERROR(filament_get_path_pixel_steps): not all subunit_num values are intg: '
                print *, path%subunit_num
                call terminate('Fatal error in FILAMENT_GET_PATH_PIXEL_STEPS')
            endif
        endif

    end subroutine filament_get_path_pixel_steps

    !>  \brief  Find the position of the nearest subunit to the point at t along the filament
    function filament_find_t_of_nearest_subunit(self,t) result(t_of_nearest_subunit)
        use Splines
        ! Arguments
        type(filament), intent(inout)   ::  self    !< Filament object
        real(kind=8),   intent(in)      ::  t
        ! Result
        real(kind=8)                    ::  t_of_nearest_subunit
        ! Private variables
        real(kind=8)                    ::  subunit_num, nearest_subunit
        logical,        parameter       ::  debug = .false.
        ! Start work

        ! What subunit_num are we on?
        call self%s_spline%Evaluate(t,subunit_num)

        ! What's the nearest subunit?
        nearest_subunit = anint(subunit_num,kind=8)

        ! We don't allow ourselves to go beyond the first or last subunit
        nearest_subunit = max(nearest_subunit,dble(ceiling(self%s_spline%y(1))))
        nearest_subunit = min(nearest_subunit,dble(floor(self%s_spline%y(self%nwp))))

        ! We look for the t corresponding to the nearest integer of subunit_num
        if (self%nwp .ge. 3) then
            t_of_nearest_subunit = self%s_spline%FindRoot(nearest_subunit)
        else
            t_of_nearest_subunit = self%t(1) + (nearest_subunit - self%wp(1)%subunit_num) * self%wp(1)%helix%rise
        endif

        ! Debug
        if (debug) then
            call self%s_spline%Evaluate(t_of_nearest_subunit,subunit_num)
            write(*,'(a,f0.3)') '**debug(filament_find_t_of_nearest_subunit): subunit_num = ', subunit_num
        endif
    end function filament_find_t_of_nearest_subunit

    !>  \brief  Return an array of waypoints along the filament's axial space curve, with the step size specified in subunits. By default, all symmetry-related subunits are returned
    !!  \todo   This subroutine could be made much shorter by calling the newer interpol_spline_find_roots_of_multiple, as is done in filament_get_path_crossovers
    subroutine filament_get_path_subunit_steps(self,path,step,ignore_fal_wps,ignore_symmetry_mates,start_wp,finish_wp)
        use procfs, only : print_memory_usage_stats
        use coos, only : waypoints_print_info
        use Splines
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  self                    !<  Filament object
        type(waypoint),      allocatable,   intent(inout)   ::  path(:)                 !<  Array of waypoints
        integer,                            intent(in)      ::  step                    !<  Step size (subunits)
        logical,                optional,   intent(in)      ::  ignore_fal_wps          !<  Whether to ignore segments between the first and second and penultimate and last waypoints
        logical,                optional,   intent(in)      ::  ignore_symmetry_mates   !<  False by default. If true, symmetry-related subunits (and in the microtubule case, tubulin dimers from the other protofilaments) will not be included
        integer,                optional,   intent(in)      ::  start_wp,finish_wp
        ! Private variables
        logical,        parameter   ::  debug   =   .false.
        real(kind=8)                ::  current_t, sstart_t, ffinish_t, next_t, previous_t
        real(kind=8)                ::  current_s, start_s, finish_s, next_s, previous_s
        integer                     ::  current_output_wp
        logical                     ::  iignore_fal_wps
        integer                     ::  sstart_wp,ffinish_wp
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(filament_get_path_subunit_steps): entry'
        endif

        ! Ignore first and last waypoints?
        if (present(ignore_fal_wps)) then
            iignore_fal_wps = ignore_fal_wps
        else
            iignore_fal_wps = .false.
        endif

        ! Work out the start and finish waypoints for the path
        if (iignore_fal_wps) then
            sstart_wp    = 2
            ffinish_wp   = self%nwp - 1
        else
            sstart_wp    = 1
            ffinish_wp   = self%nwp
        endif

        if (present(start_wp)) sstart_wp = start_wp
        if (present(finish_wp)) ffinish_wp = finish_wp

        !
        if (present(ignore_symmetry_mates)) then
            if (ignore_symmetry_mates) then
                write(*,'(a)') '**warning(filament_get_path_subunit_steps): option to ignore symmetry mates is not implemented yet'
                !call terminate('Not implemented')
            endif
        endif


        ! Allocate memory for the path
        if (allocated(path)) deallocate(path)


        ! We start from the first waypoint, find the nearest subunit, then start the walk from there, along the full filament till the last waypoint
        sstart_t            =   self%t(sstart_wp)
        start_s             =   ceiling(self%wp(sstart_wp)%subunit_num)
        sstart_t            =   self%s_spline%FindRootGivenSegment(start_s,sstart_wp)
        if (sstart_t .lt. self%t(sstart_wp)) then
            write(*,'(a,G15.6)') '**warning(filament_get_path_subunit_steps): start_t is before t for first waypoint, by ', &
                                    self%t(sstart_wp) - sstart_t
            if (self%t(sstart_wp) - sstart_t .lt. 1.0) then
                write(*,'(2a)') '**warning(filament_get_path_subunit_steps): this is probably due to rounding errors ',  &
                                '(check the difference above is very small). Overriding.'
                sstart_t = self%t(sstart_wp)
            else
                write(*,'(a)') '**ERROR(filament_get_path_subunit_steps): this is not a tolerable error. Crashing out.'
                call terminate('Fatal error in FILAMENT_GET_PATH_SUBUNIT_STEPS')
            endif
        endif

        ffinish_t       =   self%t(ffinish_wp)
        call self%s_spline%Evaluate(ffinish_t,finish_s)

        if (debug) then
            write(*,'(a,2(F0.3,1x))') '**debug(filament_get_path_subunit_steps): start, finish t: ', sstart_t, ffinish_t
            write(*,'(a,2(F0.3,1x))') '**debug(filament_get_path_subunit_steps): start, finish s: ', start_s, finish_s
            write(*,'(a)') '**debug(filament_get_path_subunit_steps): s spline info:'
            call self%s_spline%PrintInfo()
        endif

        ! If the helix is fragmented (i.e. with discontinuities between waypoints, we need to call another function
        if (self%fragmented_phi .or. self%fragmented_subunit_num) then
            call filament_get_path_subunit_steps_frag_helix(self,sstart_wp,ffinish_wp,path)
        else
            ! The helix is not fragmented
            ! Initialise the search
            current_t           =   sstart_t
            current_s           =   start_s
            next_s              =   current_s + dble(step)
            next_t              =   self%s_spline%FindRoot(next_s)
            previous_s          =   current_s - dble(step)
            current_output_wp   =   0

            ! Memory usage before the search starts
            if (debug) then
                write(*,'(a)') '**debug(filament_get_path_subunit_steps): memory usage before search'
                call print_memory_usage_stats()
            endif

            ! Start the search
            do
                if (debug) write(*,'(a,F0.3)') '**debug(filament_get_path_subunit_steps): s is = ', current_s

                ! Update previous, current and next s values
                previous_s = current_s
                current_s = next_s
                next_s = next_s + dble(step)

                ! Check whether the s values are still within range
                if (next_s .gt. finish_s ) then
                    exit
                else if (previous_s .lt. start_s) then
                    write(*,'(a)') '**ERROR(filament_get_path_subunit_steps): should not reach this point'
                    call terminate('Programming error')
                    cycle
                endif

                if (debug) write(*,'(a)') '**debug(filament_get_path_subunit_steps): here'
                ! Update previous, current and next t values
                previous_t = current_t
                current_t = next_t
                if (debug) write(*,'(a)') '**debug(filament_get_path_subunit_steps): finding next_t'
                next_t = self%s_spline%FindRoot(next_s)
                if (debug) write(*,'(a)') '**debug(filament_get_path_subunit_steps): found next_t'
                ! Check whether t values are still within range
                if (debug) write(*,'(a,i0,a,2(F0.3,1x))')    '**debug(filament_get_path_subunit_steps): output waypoint ', &
                                                            CURRENT_OUTPUT_WP+1, ' s, t = ', current_s, current_t
                if (current_t .gt. ffinish_t .or. (self%microtubule .and. next_t .gt. ffinish_t)) then
                    if (finish_s - current_s .gt. 0.01) then
                        write(*,'(a,G15.4,2(A,F0.3),a)') &
                                        '**WARNING(filament_get_path_subunit_steps): current_t is beyond finish_t by ', &
                                            CURRENT_T-FFINISH_T, ' but current_s (', current_s, &
                                            ') is less than finish_s (', finish_s, ')'
                    endif
                    exit
                else if (current_t .lt. sstart_t) then  ! This can happen with rounding errors
                    current_t = sstart_t
                endif
                ! Keep track of how many waypoints we'll be adding to path (this only works well with non-microtubules)
                if (self%microtubule) then
                    current_output_wp = current_output_wp + self%mt_pf_number
                else
                    current_output_wp = current_output_wp + self%helix%axial_symmetry * self%helix%perpendicular_symmetry
                endif
                ! Add waypoints to path
                if (debug) write(*,'(a)') '**debug(filament_get_path_subunit_steps): calling add_waypoint_to_path'
                call add_waypoint_to_path(    path,self,current_t,symmetry_mates=.true.,  &
                                                t_for_next_new_wp=next_t,t_for_previous_wp=previous_t)
                if (debug) write(*,'(a)') '**debug(filament_get_path_subunit_steps): add_waypoint_to_path returned'


                !if (debug) then
                !    write(*,'(a)') '**debug(filament_get_path_subunit_steps): only returning first subunit'
                !    exit
                !endif
            enddo

            ! Sanity check (only works for non-microtubules, because with microtubules some of the dimers in the first or last washer
            ! can be beyond the end of the filament)
            if (.not. self%microtubule) then
                if (size(path) .ne. current_output_wp ) then
                    write (*,'(a,2(i0,1x))') '**ERROR(filament_get_path_subunit_steps): failed sanity check: ', &
                                            size(path), current_output_wp
                    call terminate('Fatal error in FILAMENT_GET_PATH_SUBUNIT_STEPS')
                endif
            endif
        endif   ! End of check for fragmented helix

        if (debug) then
            write(*,'(a)') '**debug(filament_get_path_subunit_steps): returning the following set of waypoints'
            call waypoints_print_info(path)
        endif


    end subroutine filament_get_path_subunit_steps


    !>  \brief  Return an array of waypoints along the filament's axial space curve, one at each subunit.
    !!          Use when fragmented_phi or fragmented_subunit_num. All symmetry-related subunits are returned
    subroutine filament_get_path_subunit_steps_frag_helix(self,start_wp,finish_wp,path)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  self            !<  Filament object
        integer,                            intent(in)      ::  start_wp        !<  First waypoint
        integer,                            intent(in)      ::  finish_wp       !<  Last waypoint
        type(waypoint), allocatable,        intent(inout)   ::  path(:)         !<  Array of waypoints
        ! Private variables
        real(kind=8),   parameter   ::  subunit_num_tol = 0.001D0
        logical,        parameter   ::  check_subunit_num = .false.
        integer                     ::  current_origin_s, subunits_to_origin, current_wp
        real(kind=8)                ::  current_start_t, current_finish_t, current_origin_t, current_t
        ! Private variables for debugging
		integer						::	subunit_count, subunit_count_current_wp
		! Start work

		subunit_count = 0

        ! Loop over filament waypoints
        do current_wp=start_wp,finish_wp

            ! Work out the start and finish t for the current wp
            call filament_t_neighbourhood_of_wp(self,current_wp,current_start_t,current_finish_t)
            if (current_wp .eq. start_wp .or. current_wp .eq. finish_wp) current_start_t = self%t(current_wp)

            ! Work out the t value for the central subunit for the current segment
            if (current_wp .eq. start_wp) then
                current_origin_s = ceiling(self%wp(current_wp)%subunit_num)
            else if (current_wp .eq. finish_wp) then
                current_origin_s =   floor(self%wp(current_wp)%subunit_num)
            else
                current_origin_s =    nint(self%wp(current_wp)%subunit_num)
            endif
            current_origin_t =     self%t(current_wp)   &
                                + (current_origin_s-self%wp(current_wp)%subunit_num) * self%wp(current_wp)%helix%rise

            ! Work out the t value for the first subunit of the current segment
            subunits_to_origin =   - floor((current_origin_t-current_start_t)/self%wp(current_wp)%helix%rise)
            current_t          =     current_origin_t     &
                                  +  subunits_to_origin*self%wp(current_wp)%helix%rise

			! Perhaps because of floating point arithm, we may end up just the wrong side of the discontinuity
			current_t          =     max(current_t,min(current_start_t+self%wp(current_wp)%helix%rise*0.001D0,0.0001d0))
			if (current_t .gt. current_finish_t) cycle

			subunit_count_current_wp = 0

			! Loop over subunits
            do

                ! Add a waypoint at the current t
                call add_waypoint_to_path(    path,self,current_t,symmetry_mates=.true.)

				subunit_count = subunit_count + 1
				subunit_count_current_wp = subunit_count_current_wp + 1

				! Check we're on a subunit exactly. If not, print out useful debug info.
				if (check_subunit_num) then
                if ( abs(path(size(path))%subunit_num-nint(path(size(path))%subunit_num)) .gt. subunit_num_tol) then
                    write(*,'(2a,F0.5)') '**ERROR(filament_get_path_subunit_steps_frag_helix): ',   &
                                            'current output subunit_num is not integer: ', path(size(path))%subunit_num
                    write(*,'(a,i0)') '**debug(): subunit_count = ', subunit_count
					write(*,'(a,i0)') '**debug(): subunit_count_current_wp = ', subunit_count_current_wp
					write(*,'(a,F0.5)') '**debug(): current_t = ', current_t
					write(*,'(a,F0.5)') '**debug(): distance to current_wp = ', self%t(current_wp) - current_t
					write(*,'(a,2(F0.5,3x))') '**debug(): discontinuities around current_wp = ', &
																		0.5*(self%t(current_wp)-SELF%t(CURRENT_WP-1)), &
				                                                        0.5*(self%t(CURRENT_WP+1)-SELF%t(current_wp))
					write(*,'(a,2(F0.5,3x))') '**debug(): current_start_t, current_finish_t = ', &
													current_start_t, current_finish_t
					call terminate('Fatal error in FILAMENT_GET_PATH_SUBUNIT_STEPS_FRAG_HELIX. Failed sanity check')
                endif
                endif

                ! Increment t
                subunits_to_origin = subunits_to_origin + 1
                current_t = current_origin_t + subunits_to_origin*self%wp(current_wp)%helix%rise

                ! Check we are still in the neighbourhood of the current waypoint
                if (current_t .gt. current_finish_t) then
                    exit
                endif

            enddo

        enddo

    end subroutine filament_get_path_subunit_steps_frag_helix

    !>  \brief  Define the neighbourhood of the current waypoint (i.e. halfway to the previous and next waypoints
    pure subroutine filament_t_neighbourhood_of_wp(self,wp_num,start_t,finish_t)
        implicit none
        ! Arguments
        type(filament), intent(in)  ::  self
        integer,        intent(in)  ::  wp_num
        real(kind=8),   intent(out) ::  start_t
        real(kind=8),   intent(out) ::  finish_t
        ! Private variables

        ! Start work
        if (wp_num .eq. 1) then
            start_t  = self%t(1)
        else
            start_t  = self%t(wp_num) - (self%t(wp_num)-SELF%t(WP_NUM-1))*0.5D0
        endif

        if (wp_num .eq. self%nwp) then
            finish_t = self%t(self%nwp)
        else
            finish_t = self%t(wp_num) + (self%t(WP_NUM+1)-SELF%t(wp_num))*0.5D0
        endif
    end subroutine filament_t_neighbourhood_of_wp


    !>  \brief  Return an array of waypoints along the filament's axial space curve, with the step size specified in numbers of crossovers.
    subroutine filament_get_path_crossovers(self,path,step,ignore_fal_wps,ignore_symmetry_mates,phi_step_size, &
                                            include_fal_wps)
        use Splines
        ! Arguments
        type(filament),                     intent(inout)   ::  self                    !<  Filament object
        type(waypoint),         allocatable,intent(inout)   ::  path(:)                 !<  Output array of waypoints
        integer,                            intent(in)      ::  step                    !<  Step size (number of crossovers). Call with STEP=1 to get every crossover.
        logical,                optional,   intent(in)      ::  ignore_fal_wps          !<  Whether to ignore segments between the first and second and penultimate and last waypoints
        logical,                optional,   intent(in)      ::  ignore_symmetry_mates   !<  False by default. If true, symmetry-related subunits (and in the microtubule case, tubulin dimers from the other protofilaments) will not be included
        real(kind=8),           optional,   intent(in)      ::  phi_step_size           !<  Phi (twist) step size. Defaults to 180.0 for crossovers.
        logical,                optional,   intent(in)      ::  include_fal_wps         !<  Even if the first and last segments were ignored, the fal wps can be included in the output path. Defaults to FALSE.
        ! Private variables
        logical,        parameter   ::  debug   =   .false.
        logical                     ::  iignore_fal_wps, iignore_symmetry_mates
        integer                     ::  start_wp, finish_wp, wp_count
        real(kind=8),   allocatable ::  t(:)
        real(kind=8)                ::  start_t, finish_t
        real(kind=8)                ::  pphi_step_size
        logical                     ::  iinclude_fal_wps
        ! Start work

        ! Ignore first and last waypoints?
        iignore_fal_wps = .false.
        if (present(ignore_fal_wps)) iignore_fal_wps = ignore_fal_wps

        IIGNORE_SYMMETRY_MATES=.true.
        if (present(ignore_symmetry_mates)) iignore_symmetry_mates = ignore_symmetry_mates

        IINCLUDE_FAL_WPS=.false.
        if (present(include_fal_wps)) iinclude_fal_wps = include_fal_wps

        ! Phi step size
        pphi_step_size = 180.0D0
        if (present(phi_step_size)) pphi_step_size = phi_step_size

        ! Work out the start and finish waypoints for the path
        if (iignore_fal_wps) then
            start_wp    = 2
            finish_wp   = self%nwp - 1
        else
            start_wp    = 1
            finish_wp   = self%nwp
        endif

        start_t             =   self%t(start_wp)
        finish_t            =   self%t(finish_wp)

        ! Allocate memory for the path
        if (allocated(path)) deallocate(path)

        if (iinclude_fal_wps) then
            call add_waypoint_to_path(path,self,self%t(1),symmetry_mates=.not. iignore_symmetry_mates)
        endif

        !if (debug) then
        !   write(*,'(a)') '**debug(filament_get_path_crossovers): filament info upon entry'
        !   call filament_print_info(self)
        !endif

        ! Find t values at which phi is a multiple of pphi_step_size
        t = self%twist_spline%FindRootsOfMultiple(pphi_step_size)

        ! Loop over the t values and find corresponding waypoints
        do wp_count=1,size(t)
            if (t(wp_count) .ge. start_t .and. t(wp_count) .le. finish_t) then
                call add_waypoint_to_path(path,self,t(wp_count),symmetry_mates=.not. iignore_symmetry_mates)
            endif
        enddo

        if (iinclude_fal_wps) then
            call add_waypoint_to_path(path,self,self%t(self%nwp),symmetry_mates=.not. iignore_symmetry_mates)
        endif

    end subroutine filament_get_path_crossovers



    !>  \brief  Return an array of coordinates which describes the path of a filament, with specified step size (in Angstroems unless specified otherwise)
    subroutine filament_get_path(   self,path,step,step_in_pixels,step_in_repeats,ignore_fal_wps,sp_mode,num_wps_per_wp,helix,&
                                    start_wp,finish_wp,include_extra_wp,ignore_symmetry_mates, &
                                    intg_subunit_num)
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  self                    !<  Filament
        type(waypoint),         allocatable,intent(inout)   ::  path(:)                 !<  Array of waypoints
        real(kind=8),                       intent(in)      ::  step                    !<  Step size (pixels)
        logical,                optional,   intent(in)      ::  step_in_pixels          !<  Specifies (if .TRUE.) that the step is given in pixels
        logical,                optional,   intent(in)      ::  step_in_repeats         !<  .FALSE. by default. If .TRUE., the step is interpreted as being in number of helical subunits, rather than pixels or Angstroms.
                                                                                        !!  So e.g. if STEP=1.0, a path with every subunit will be returned.
        logical,                optional,   intent(in)      ::  ignore_fal_wps          !<  If .TRUE., the path will extend form the second to the second-to-last waypoint. .FALSE. by default.
        logical,                optional,   intent(in)      ::  sp_mode                 !<  Single particle mode. In single particle mode, it is assumed the waypoints were processed independently
                                                                                        !!  from each other, so we return waypoints in the output path which are near
                                                                                        !!  the input waypoints, with the assumption that the filament is straight near the waypoints.
        integer,                optional,   intent(in)      ::  num_wps_per_wp          !<  In single particle mode, the number of output waypoints per waypoint in the input filament
        type(helical_symmetry), optional,   intent(in)      ::  helix                   !<  Overriding helical parameters. If not given, helical parameters found in the filament object will be used.
        integer,                optional,   intent(in)      ::  start_wp                !<  Path will be restricted to start at this waypoint
        integer,                optional,   intent(in)      ::  finish_wp               !<  Path will be restricted to finish at/before this waypoint
        logical,                optional,   intent(in)      ::  include_extra_wp        !<  Return one last waypoint which will be past the finish input wp (but will not be beyond the end of the filament)
        logical,                optional,   intent(in)      ::  ignore_symmetry_mates   !<  If true, symmetry-related subunits (and in the microtubule case, tubulin dimers from the other protofilaments) will not be included
        logical,                optional,   intent(in)      ::  intg_subunit_num        !<  Waypoints should have integer subunit_num values
        ! Private variables
        real(kind=8)                ::  sstep_in_pixels         !   The step size, in pixels
        logical                     ::  sstep_in_repeats
        logical                     ::  iignore_fal_wps                         !   Whether to ignore the first and last (fal) waypoints (wps)
        logical,    parameter       ::  full_filament_path          =   .true.
        logical,    parameter       ::  path_from_each_waypoint     =   .false.
        logical,    parameter       ::  debug   =   .false.
        logical                     ::  ssp_mode
        !type(waypoint), allocatable ::  path_check(:) ! For debugging purposes


        !
        ! Start work
        !

        ! This subroutine does not handle perpendicular helical symmetries yet
        if (self%helix%perpendicular_symmetry .gt. 1) then
            write(*,'(a)') '**warning(filament_get_path): helical symmetries with perpendicular symmetries not ',   &
                            'tested yet. Please check results carefully & remove this warning when done.'
        endif

        ! Is the steps in number of repeats?
        if (present(step_in_repeats)) then
            sstep_in_repeats = step_in_repeats
        else
            sstep_in_repeats = .false.
        endif

        ! Ignore first and last waypoints?
        if (present(ignore_fal_wps)) then
            iignore_fal_wps = ignore_fal_wps
        else
            iignore_fal_wps = .false.
        endif


        !
        ! Find the step size in pixels. If the user gave STEP_IN_PIXELS=.TRUE., then we don't need to do anything.
        !
        if (sstep_in_repeats) then
            ! If the step was given in number of repeats, we first need to update the helical rise information for each segment of the filament
            !call filament_update_wp_helix(self) ! Not needed anymore, with the new s spline
            if (present(step_in_pixels)) then
                if (step_in_pixels) then
                    write(*,'(a)') '**ERROR(filament_get_path): step can be given either in number of ',    &
                                    'repeats, or in Angstroms or pixels, not both'
                    call terminate('Fatal error in FILAMENT_GET_PATH')
                endif
            endif
        else
            sstep_in_pixels = step * self%film%GetPixelSize()
            if (present(step_in_pixels)) then
                if (step_in_pixels) then
                    sstep_in_pixels = step
                endif
            endif
        endif

        ! start_wp and finish_wp are only implemented with pixel step sizes
        if (present(start_wp) .and. sstep_in_repeats) then
            write(*,'(a)') '**warning(filament_get_path): start_wp optional argument only supported for pixel step sizes. '//&
                            'Will be ignored.'
        endif
        if (present(finish_wp) .and. sstep_in_repeats) then
            write(*,'(a)') '**warning(filament_get_path): finish_wp optional argument only supported for pixel step sizes. '//&
                            'Will be ignored.'
        endif

        !
        if (present(include_extra_wp) .and. sstep_in_repeats) then
            write(*,'(a)') '**warning(filament_get_path): include_extra_waypoint option is only supported for pixel step sizes. '//&
                            'Will be ignored.'
        endif

        ! Are we in single particle mode?
        ssp_mode = .false.
        if (present(sp_mode))  ssp_mode = sp_mode


        ! If we're in single particle mode, we need to know how many output WPs we want per input wp
        if (ssp_mode) then
            if (.not. present(num_wps_per_wp)) then
                write(*,'(a)') '**ERROR(filament_get_path): in sp mode, need to know how many output wps you want per input WP'
                call terminate('Fatal error in filament_get_path - programming error')
            endif
        endif

        if (ssp_mode) then
            if (sstep_in_repeats) then
                path = filament_get_path_near_wps(  self,step,num_wps_per_wp,ignore_fal_wps=iignore_fal_wps,  &
                                                    helix=helix,step_in_repeats=sstep_in_repeats,intg_repeats=.true.)
            else
                path = filament_get_path_near_wps(  self,sstep_in_pixels,num_wps_per_wp,ignore_fal_wps=iignore_fal_wps, &
                                                    helix=helix,step_in_repeats=sstep_in_repeats,intg_repeats=intg_subunit_num)
            endif
        else
            if (present(helix)) then
                write(*,'(a)') '**ERROR(filament_get_path): helical parameters override is only intended for single-particle mode'
                call terminate('Fatal error in FILAMENT_GET_PATH')
            endif

            if (full_filament_path) then
                ! In this mode, we start from the first waypoint, and walk along the full filament till the last waypoint
                if (sstep_in_repeats) then
                    call filament_get_path_subunit_steps(   self,path,nint(step),ignore_fal_wps=iignore_fal_wps,  &
                                                            start_wp=start_wp,finish_wp=finish_wp, &
                                                            ignore_symmetry_mates=ignore_symmetry_mates)
                else
                    call filament_get_path_pixel_steps(     self,sstep_in_pixels,path,iignore_fal_wps,  &
                                                            start_wp=start_wp,finish_wp=finish_wp,  &
                                                            include_extra_wp=include_extra_wp, &
                                                            integer_subunit_num=intg_subunit_num)
                endif
            else if (path_from_each_waypoint) then
                ! In this mode, we start from each waypoint and travel along the path until
                ! we're halfway to the next/previous waypoint
                call terminate('Not implemented yet')
            else
                write(*,'(a)') '**ERROR(filament_get_path): oops, programming error'
                call terminate('Fatal error in FILAMENT_GET_PATH')
            endif
        endif   ! End of test for sp_mode

    end subroutine filament_get_path

    !>  \brief  Add waypoint (and optionally its symmetry-mates) to the end of a path
    subroutine add_waypoint_to_path(old_path,self,t_for_new_wp,symmetry_mates,t_for_next_new_wp,t_for_previous_wp)
        use filaments_microtubules, only : microtubule_get_good_pf
        use filaments_geometry, only : filament_eval_helical_rise
        use coos, only : waypoints_grow_array, wp_print_info, waypoint_apply_symmetry
        implicit none
        ! Arguments
        type(waypoint), allocatable,    intent(inout)   ::  old_path(:)
        type(filament),                 intent(inout)   ::  self                !<  Filament object
        real(kind=8),                   intent(in)      ::  t_for_new_wp        !<  The arc length parameter (t) for the waypoint(s) to be added
        logical,        optional,       intent(in)      ::  symmetry_mates      !<  If .TRUE., symmetry-mates (tubulin dimers in the microtubule case) will also be added to the path. False by default.
        real(kind=8),   optional,       intent(in)      ::  t_for_next_new_wp   !<  The arc length parameter (t) for the waypoint which will be added next (needed only in the mt case, where it helps determine the end of the "washer")
        real(kind=8),   optional,       intent(in)      ::  t_for_previous_wp   !<  The t parameter for the waypoints which was last added (needed for the mt case)
        ! Private variables
        logical,    parameter   ::  debug                   =   .false.
        logical                 ::  ssymmetry_mates
        integer                 ::  current_output_wp
        integer                 ::  perp_sym, axial_sym
        ! Microtubule-specific private variables
        logical,    parameter   ::  ignore_seam_dimers      =   .false.
        real                    ::  washer_total_phi
        real                    ::  additional_t, additional_phi
        type(waypoint)          ::  pf1_wp,next_pf1_wp, wp
        real(kind=8)            ::  pf1_t,next_pf1_t
        integer                 ::  current_pf,first_pf,last_pf,good_pf
        integer                 ::  num_dimers_beyond_fal_wps
        real                    ::  local_helical_rise
        type(waypoint), allocatable ::  new_path(:)
        ! Start work

        ! Will we be adding symmetry mates too?
        if (present(symmetry_mates)) then
            ssymmetry_mates = symmetry_mates
        else
            ssymmetry_mates = .false.
        endif

        ! 160501: the symmetry mates will be added to the reconstruction by threed_add_slice
        ssymmetry_mates = .false.

        ! If we're dealing with microtubules, we need to know the t parameter for where the next washer starts
        if (self%microtubule .and. ssymmetry_mates) then
            if (.not. present(t_for_next_new_wp)) then
                write(*,'(a)') '**ERROR(add_waypoint_to_path): the filament is a microtubule, so you need to specify ', &
                                'the t parameter for the following waypoint'
                call terminate('Fatal error in ADD_WAYPOINT_TO_PATH')
            endif
        endif

        ! How many waypoints already in the old path?
        if (allocated(old_path)) then
            current_output_wp = size(old_path)
        else
            current_output_wp = 0
        endif

        if (debug) write(*,'(a,i0)') '**debug(add_waypoint_to_path): on entry, current_output_wp = ', current_output_wp

        !
        if (ssymmetry_mates) then
            if (self%microtubule) then
                ! Check we have the required info
                if (.not. present(t_for_next_new_wp) .or. .not. present(t_for_previous_wp)) then
                    write(*,'(a)') '**ERROR(add_waypoint_to_path): when dealing with microtubule, need to know the position'// &
                                    ' preceding and following waypoints'
                    call terminate('Fatal error in ADD_WAYPOINT_TO_PATH. Programming error.')
                endif
                ! In the case of microtubules, the symmetry-mates are defined by the helical symmetry operator of the "washer"
                good_pf     =   microtubule_get_good_pf(self)   ! For 13-pf mts, this is pf 7
                next_pf1_t  =   t_for_new_wp+(t_for_next_new_wp-t_for_new_wp)/real(self%mt_pf_number)*real(good_pf)
                next_pf1_wp =   waypoint_from_t(self,next_pf1_t)
                pf1_t       =   t_for_previous_wp+(t_for_new_wp-t_for_previous_wp)/real(self%mt_pf_number)*real(good_pf)
                pf1_wp      =   waypoint_from_t(self,pf1_t)
                ! We may need to ignore the dimers either side of the seam
                if (ignore_seam_dimers) then
                    first_pf = 2
                    last_pf = self%mt_pf_number - 1
                else
                    first_pf = 1
                    last_pf = self%mt_pf_number
                endif
                num_dimers_beyond_fal_wps = 0
                if (allocated(old_path)) new_path = old_path
                ! Total phi angle change from the start of the current washer to the start of the next washer
                washer_total_phi = 360. + next_pf1_wp%euler%phi - pf1_wp%euler%phi
                ! Local helical rise
                local_helical_rise = filament_eval_helical_rise(self,t_for_new_wp)
                ! Loop over the dimers / protofilaments
                do current_pf=first_pf,last_pf
                    ! Additional t, to place origin at level of desired pf dimer
                    additional_t = -real(current_pf-good_pf)/real(self%mt_pf_number) * real(self%mt_rise) / 2.0 * local_helical_rise
                    ! Sanity check
                    if (abs(additional_t) .gt. abs(local_helical_rise * real(self%mt_rise)/2.0)) then   ! abs - not sure
                        write(*,'(a,F0.6,a,F0.6,a)') '**ERROR(add_waypoint_to_path): additional_t (', additional_t, &
                                                    ') > local washer rise (', LOCAL_HELICAL_RISE* real(self%mt_rise)/2.0, ')'
                        call terminate('Fatal error in ADD_WAYPOINT_TO_PATH')
                    endif
                    ! Additional twist/phi, to bring current pf to correct phi
                    ! i suspect washer_total_phi here should be replaced by 360, since we're getting the waypoint from t anyway, which means we're already compensating for the any twist of the microtubule
                    additional_phi = -real(current_pf-good_pf) * washer_total_phi/real(self%mt_pf_number)
                    ! Get waypoint at t
                    if (t_for_new_wp+additional_t .gt. self%t(self%nwp) .or. t_for_new_wp+additional_t .lt. self%t(1)) then
                        num_dimers_beyond_fal_wps = num_dimers_beyond_fal_wps + 1
                        if (debug) write(*,'(a)') '**debug(add_waypoint_to_path): dimer beyond ends of filament'
                    else
                        if (debug) print '(a,f8.3,1x,f8.3)', '**debug(add_waypoint_to_path): additional_t, additional_phi = ', &
                                                            additional_t, additional_phi
                        wp = waypoint_from_t(self,t_for_new_wp+additional_t)
                        ! add appropriate phi angle to waypoint
                        wp%euler%phi = pf1_wp%euler%phi + additional_phi
                        ! add waypoint to output path
                        current_output_wp = current_output_wp + 1
                        if (debug) then
                            write(*,'(a)',advance='no') '**debug(add_waypoint_to_path): adding this wp: '
                            call wp_print_info(wp)
                        endif
                        call waypoints_grow_array(new_path,increment=1)
                        new_path(current_output_wp) = wp
                    endif
                enddo
                if (debug) then
                    write(*,'(a,i0,a)') '**debug(add_waypoint_to_path): ', num_dimers_beyond_fal_wps ,&
                                        ' dimers fell beyond the ends of the filament'
                endif
            else
                ! Symmetry mates, but not microtubule
                if (allocated(old_path)) then
                    if (allocated(new_path)) deallocate(new_path)
                    allocate(new_path(size(old_path)))
                    new_path = old_path
                endif
                call waypoints_grow_array(new_path,increment=self%helix%perpendicular_symmetry*self%helix%axial_symmetry)
                if (debug) write(*,'(a,i0)') '**debug(add_waypoint_to_path): size(new_path) after grow_array: ', size(new_path)
                do perp_sym=1,self%helix%perpendicular_symmetry
                    do axial_sym=1,self%helix%axial_symmetry
                        current_output_wp                       =   current_output_wp + 1
                        new_path(current_output_wp)             =   waypoint_from_t(self,t_for_new_wp)
                        call waypoint_apply_symmetry(new_path(current_output_wp),axial_sym,self%helix%axial_symmetry, &
                                                                                 perp_sym,self%helix%perpendicular_symmetry)
                        if (debug) then
                            write(*,'(a)',advance='no') '**debug(add_waypoint_to_path): adding this wp: '
                            call wp_print_info(new_path(current_output_wp))
                        endif
                    enddo
                enddo
            endif   ! End of test for microtubule
        else
            ! No symmetry-mates to be added
            if (allocated(old_path)) then
                if (allocated(new_path)) deallocate(new_path)
                allocate(new_path(size(old_path)))
                new_path = old_path
            endif
            call waypoints_grow_array(new_path)
            new_path(size(new_path))    =   waypoint_from_t(self,t_for_new_wp)
        endif   ! End of test for symmetry mates

        !
        if (allocated(old_path)) deallocate(old_path)
        allocate(old_path(size(new_path)))
        old_path = new_path
    end subroutine add_waypoint_to_path

    !>  \brief  Define the range of parameter values in the neighbourhood of waypoints to be refined. This range can
    !!          then be used to define the region the scoring function should examine.
    subroutine filament_t_interval_for_scoring_function(self,start_wp,finish_wp)
        use coos, only : waypoints_to_be_refined_are_contiguous, waypoints_print_info
        implicit none
        ! Arguments
        type(filament), intent(in)  ::  self
        integer,        intent(out) ::  start_wp
        integer,        intent(out) ::  finish_wp
        ! Private variables
        integer ::  i
        integer ::  first_refine_wp, last_refine_wp
        logical,    parameter   ::  halfway_to_next_wp = .false.
        ! Start work
        ! Check that waypoints to be refined are contiguous
        if (.not. waypoints_to_be_refined_are_contiguous(self%wp)) then
            write(*,'(a)') '**ERROR(filament_t_interval_for_scoring_function): waypoints to be refined are not contiguous'
            call waypoints_print_info(self%wp)
            call terminate('Fatal error in FILAMENT_T_INTERVAL_FOR_SCORING_FUNCTION')
        endif

        !print *, '**DEBUG(): fal_wps_in_score = ', self%fal_wps_in_score

        ! Work out the first and last waypoints which are being refined
        first_refine_wp = 0
        if (self%wp(1)%refine) first_refine_wp = 1
        last_refine_wp  = 0
        if (self%wp(self%nwp)%refine) last_refine_wp = self%nwp
        do i=2,self%nwp
            if (self%wp(i)%refine .and. .not. self%wp(i-1)%refine .and. first_refine_wp .eq. 0) first_refine_wp = i
        enddo
        do i=1,self%nwp-1
            if (self%wp(i)%refine .and. .not. self%wp(i+1)%refine .and. last_refine_wp .eq. 0) last_refine_wp = i
        enddo

		! In case all waypoints are set to f, we probably want to score the whole filament
		first_refine_wp = 1
		last_refine_wp  = self%nwp

        !print *, '**DEBUG(): first, last refine = ', first_refine_wp, last_refine_wp

        if (halfway_to_next_wp) then
!            ! We go halfway to the previous/next waypoints
!            ! Work out the first and last waypoints for the scoring function
!            start_wp = 0
!            finish_wp = 0
!            if (self%fal_wps_in_score) then
!                start_wp = first_refine_wp
!                finish_wp = last_refine_wp
!            else
!                start_wp = max(first_refine_wp,2)
!                start_wp = min(start_wp,self%NWP-1)
!                finish_wp = max(last_refine_wp,2)
!                finish_wp = min(finish_wp,self%NWP-1)
!            endif
!
!            !print *, '**DEBUG(): start, finish wp = ', start_wp, finish_wp
!
!            ! Work out the t parameters
!            start_t = 0.0D0
!            finish_t = 0.0D0
!
!            if (start_wp .EQ. 1) THEN
!                start_t = self%t(1)
!            else
!                start_t = self%t(start_wp) - (self%t(start_wp)-SELF%t(START_WP-1)) * 0.5D0
!            endif
!
!            if (finish_wp .EQ. self%nwp) then
!                finish_t = self%t(self%nwp)
!            else
!                finish_t = self%t(finish_wp) + (self%t(FINISH_WP+1)-SELF%t(finish_wp)) * 0.5D0
!            endif
!
!            !print *, '**DEBUG(): start_t, finish_t = ', start_t, finish_t
!
!            ! If FAL_WPs are not in scoring function, make sure t is limited to that also
!            if (.NOT. self%fal_wps_in_score) then
!                start_t = max(start_t,self%t(2))
!                finish_t = min(finish_t,self%t(self%NWP-1))
!            endif
!
!            ! When we have fragment_helix, being exactly half-way between waypoints can cause numerical instabilities,
!            ! because of the discontinuity in the helix that happens at this point
!            start_t = start_t + 0.1
        else
            ! We go all the way to the previous & next waypoints
            start_wp = 0
            finish_wp = 0
            if (self%fal_wps_in_score) then
                start_wp = max(FIRST_REFINE_WP-1,1)
                finish_wp = min(LAST_REFINE_WP+1,self%nwp)
            else
                start_wp = max(FIRST_REFINE_WP-1,2)
                finish_wp = min(LAST_REFINE_WP+1,self%NWP-1)
            endif
!            start_t = self%t(start_wp)
!            finish_t = self%t(finish_wp)
        endif

        !print *, '**DEBUG(): start_t, finish_t = ', start_t, finish_t

    end subroutine filament_t_interval_for_scoring_function

end module filaments_path
