!>  \brief  Classes and methods to deal with em films / negatives
module Micrographs

    use runtime_parameters, only    :   filename_len
    use ctftilts, only : ctftilt
    use ElectronDoses, only : ElectronDose

    private

    public :: Micrograph

    type Micrograph
        private
        integer                         ::  id                                  !<  Micrograph id
        character(len=filename_len)     ::  filename                            !<  Filename
        integer                         ::  location_in_file = 1                !<  Location of the first frame image within the image file
        type(ctftilt), public           ::  ctftilt                             !<  object containing ctf info for this film
        integer                         ::  logical_dimensions(2)               !<  dimensions (pixels)
        real                            ::  pixel_size  =   0.0                 !<  pixel size (angstroems)
        ! Movie properties
        integer                         ::  number_of_frames = 1                !<  Number of frames in the movie
        real,   allocatable             ::  relative_frame_shift_x(:)           !<  The relative frame shift in X
        real,   allocatable             ::  relative_frame_shift_y(:)           !<  The relative frame shift in X
        real                            ::  exposure_per_frame = 1.0e0          !<  Exposure per frame, in electrons per square Angstroms
        type(ElectronDose), public      ::  electron_dose                       !<  An electron dose object, used to compute dose filters
        contains
            procedure           ::  MicrographInit
            procedure,  public  ::  Init => MicrographInit
            procedure,  public  ::  PrintInfo
            procedure,  public  ::  ReturnZFromXY
            procedure,  public  ::  ReturnCTFAtCoos
            ! Get/set
            procedure,  public  ::  GetID
            procedure,  public  ::  SetID
            procedure,  public  ::  GetPixelSize
            procedure,  public  ::  SetPixelSize
            procedure,  public  ::  GetFrameLocationInFile
            procedure,  public  ::  GetLocationInFile
            procedure,  public  ::  SetLocationInFile
            procedure,  public  ::  GetLogicalDimensions
            procedure,  public  ::  GetLogicalDimension
            procedure,  public  ::  SetLogicalDimension
            procedure,  public  ::  SetLogicalDimensions
            procedure,  public  ::  GetFilename
            procedure,  public  ::  GetNumberOfFrames
            procedure,  public  ::  SetNumberOfFrames
            procedure,  public  ::  GetExposurePerFrame
            ! Operators
            procedure           ::  film_assignment
            generic,    public  ::  assignment(=) => film_assignment
            ! Destructor
            final               ::  Destroy
    end type Micrograph

    contains

    pure integer function GetNumberOfFrames(self)
        class(Micrograph),  intent(in)      ::  self
        GetNumberOfFrames = self%number_of_frames
    end function GetNumberOfFrames

    pure real function GetExposurePerFrame(self)
        class(Micrograph),  intent(in)      ::  self
        GetExposurePerFrame = self%exposure_per_frame
    end function GetExposurePerFrame

    pure subroutine SetNumberOfFrames(self,number_of_frames)
        class(Micrograph),  intent(inout)   ::  self
        integer,            intent(in)      ::  number_of_frames
        self%number_of_frames = number_of_frames
    end subroutine SetNumberOfFrames


    pure function GetFilename(self)
        class(Micrograph),  intent(in)      ::  self
        character(len=filename_len)         ::  GetFilename
        GetFilename = self%filename
    end function GetFilename


    pure integer function GetID(self)
        class(Micrograph),  intent(in)      ::  self
        GetID = self%id
    end function GetID

    subroutine SetID(self,new_id)
        class(Micrograph),  intent(inout)   ::  self
        integer,            intent(in)      ::  new_id
        self%id = new_id
    end subroutine

    pure real function GetPixelSize(self)
        class(Micrograph),  intent(in)      ::  self
        GetPixelSize = self%pixel_size
    end function GetPixelSize

    subroutine SetPixelSize(self,new_pixel_size)
        class(Micrograph),  intent(inout)   ::  self
        real,               intent(in)      ::  new_pixel_size
        self%pixel_size = new_pixel_size
    end subroutine SetPixelSize

    pure integer function GetLocationInFile(self)
        class(Micrograph),  intent(in)  ::  self
        GetLocationInFile = self%location_in_file
    end function GetLocationInFile

    pure integer function GetFrameLocationInFile(self,frame_number)
        class(Micrograph),  intent(in)  ::  self
        integer,            intent(in)  ::  frame_number
        GetFrameLocationInFile = self%location_in_file + frame_number - 1
    end function GetFrameLocationInFile

    subroutine SetLocationInFile(self,location)
        class(Micrograph),  intent(inout)   ::  self
        integer,            intent(in)      ::  location
        self%location_in_file = location
    end subroutine SetLocationInFile

    pure function GetLogicalDimensions(self) result(ldims)
        class(Micrograph),  intent(in)  ::  self
        integer                         ::  ldims(2)
        ldims = self%logical_dimensions
    end function GetLogicalDimensions

    pure integer function GetLogicalDimension(self,which_dimension)
        class(Micrograph),  intent(in)      ::  self
        integer,            intent(in)      ::  which_dimension
        GetLogicalDimension = self%logical_dimensions(which_dimension)
    end function GetLogicalDimension

    subroutine SetLogicalDimension(self,which_dimension,new_value)
        class(Micrograph),  intent(inout)   ::  self
        integer,            intent(in)      ::  which_dimension
        integer,            intent(in)      ::  new_value
        self%logical_dimensions(which_dimension) = new_value
    end subroutine SetLogicalDimension

    subroutine SetLogicalDimensions(self,new_dimensions)
        class(Micrograph),  intent(inout)   ::  self
        integer,            intent(in)      ::  new_dimensions(2)
        self%logical_dimensions = new_dimensions
    end subroutine SetLogicalDimensions

    subroutine film_assignment(self_lhs,self_rhs)
        ! argument
        class(Micrograph),                  intent(inout)   ::  self_lhs
        type(Micrograph),                   intent(in)      ::  self_rhs
        !
        self_lhs%id = self_rhs%id
        self_lhs%filename = self_rhs%filename
        self_lhs%location_in_file = self_rhs%location_in_file
        self_lhs%ctftilt = self_rhs%ctftilt
        self_lhs%logical_dimensions = self_rhs%logical_dimensions
        self_lhs%pixel_size = self_rhs%pixel_size
        self_lhs%number_of_frames = self_rhs%number_of_frames
        if (allocated(self_rhs%relative_frame_shift_x)) then
            self_lhs%relative_frame_shift_x = self_rhs%relative_frame_shift_x
        endif
        if (allocated(self_rhs%relative_frame_shift_y)) then
            self_lhs%relative_frame_shift_y = self_rhs%relative_frame_shift_y
        endif
        self_lhs%electron_dose = self_rhs%electron_dose
    end subroutine film_assignment

    !>  \brief  Print out summary information about a film object
    subroutine PrintInfo(self)
        use ctftilts, only : ctftilt_print_info
        implicit none
        ! Arguments
        class(Micrograph),                  intent(in)      ::  self
        ! Private variables

        ! Start work
        write(*,'(a,i0,a)')     'Information for film ', self%id, ' follows:'
        write(*,'(2a)')         'File name             : ', trim(adjustl(self%filename))
        write(*,'(a,i0)')       'Location in file      : ', self%location_in_file
        write(*,'(a,2(i0,1x))') 'Simensions (pixels)   : ', self%logical_dimensions
        write(*,'(a,F0.2)')     'Pixel size (Angstroms): ', self%pixel_size
        write(*,'(a,i0)')       'Number of frames      : ', self%number_of_frames
        write(*,'(a,f0.2)')     'Expos./frame (e/A^2)  : ', self%exposure_per_frame
        write(*,'(a)')          'CTF tilt info         : '
        call ctftilt_print_info(self%ctftilt)
    end subroutine PrintInfo

    !>  \brief  Initialise a film object - the dimensions are worked out by opening the image file
    !!  \todo   Add warning if the pixel size given as input does not match the pixel size specified in the ctftilt/ctffind output
    subroutine MicrographInit(self,id,filename,pixel_size,ctftilt_filename,ctffind_filename,film_image_available, &
                         location_within_file,ctf_params, &
                         number_of_frames, relative_frame_shifts_x,relative_frame_shifts_y, &
                         exposure_per_frame)
        use ImageFiles
        use parsers, only : file_exists
        use ctfs, only : ctf_parameters, ctf_pix_to_ang
        use runtime_parameters, only : terminate
        use ctftilts, only : ctftilt_read_from_ctffind, ctftilt_read_from_ctftilt
        use ctfs, only : wl_to_akv
        use Globals
        use ContrastTransferFunctions
        implicit none
        ! Arguments
        class(Micrograph),                  intent(inout)   ::  self
        integer,                            intent(in)      ::  id                          !<  Film id
        character(len=*),                   intent(in)      ::  filename                    !<  Film filename
        real,                               intent(in)      ::  pixel_size                  !<  Pixel size (Angstroms)
        character(len=*),       optional,   intent(in)      ::  ctffind_filename            !<  Output from ctffind
        character(len=*),       optional,   intent(in)      ::  ctftilt_filename            !<  Output from ctftilt
        logical,                optional,   intent(in)      ::  film_image_available        !<  Whether the actual film image is available somewhere disk
        integer,                optional,   intent(in)      ::  location_within_file        !<  Location of the film image within the file
        type(ctf_parameters),   optional,   intent(in)      ::  ctf_params                  !<  ctf parameters for that micrograph.
        integer,                optional,   intent(in)      ::  number_of_frames            !<  If the micrograph is a movie, number of frames
        real,                   optional,   intent(in)      ::  relative_frame_shifts_x(:)  !<  Relative frame shifts in X
        real,                   optional,   intent(in)      ::  relative_frame_shifts_y(:)  !<  Relative frame shifts in Y
        real,                   optional,   intent(in)      ::  exposure_per_frame          !<  Number of electrons per square Angstroms, per frame
        ! Private variables
        type(imagefile) ::  imgfile
        integer ::  idummy(3)
        logical ::  ffilm_image_available
        logical, parameter  ::  debug = .false.

        ! Start work

        ! We assume the film image is available
        ffilm_image_available = .true.
        if (present(film_image_available)) ffilm_image_available = film_image_available

        !
        ! Remember the filename & the id
        !
        if (id .le. 0) then
            write(*,'(a,i0)') '**error(MicrographInit): micrograph id must be > 0, but it is ', id
            call this_program%TerminateWithFatalError('Micrograph::Init','Bad micrograph id')
        endif
        self%id = id
        self%filename = trim(adjustl(filename))

        self%number_of_frames = 1
        if (present(number_of_frames)) self%number_of_frames = number_of_frames
        if (self%number_of_frames .gt. 1) then
            if (present(relative_frame_shifts_x)) self%relative_frame_shift_x = relative_frame_shifts_x
            if (present(relative_frame_shifts_y)) self%relative_frame_shift_y = relative_frame_shifts_y
        endif

        ! Guess the location in the file
        self%location_in_file = (self%id-1) * self%number_of_frames + 1
        if (present(location_within_file)) self%location_in_file = location_within_file

        !
        ! Initialise a file object so that we can check the file exists, and get its dimensions
        !
        if (ffilm_image_available) then
            call imgfile%init(self%filename,read_only=.true.)
            if (debug) call imgfile%PrintInfo()
            idummy = imgfile%GetDimensions()
            self%logical_dimensions = idummy(1:2)
            if (self%location_in_file .gt. imgfile%GetStackSize()) then
                !write(*,'(2(a,i0),3a)') '**ERROR(Micrograph::Init): location in file (', self%location_in_file, &
                !                        ') > number of locations in the file (', imgfile%GetStackSize(), &
                !                        '; ', trim(imgfile%GetFilename()), ')'
                !call terminate('Fatal error in Micrograph::Init')

                ! It could be that we are dealing with a set of input micrographs which were not in a stack
                self%location_in_file = 1
            endif
            call imgfile%close()
        else
            if (file_exists(self%filename)) then
                write(*,'(2a)') '**warning(Micrograph::Init): was told this file did not exist, but it does: ', &
                                trim(adjustl(self%filename))
            else
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a)') '**info(Micrograph::Init): film image not available'
                endif
            endif
            self%logical_dimensions = 0
        endif


        !
        ! Set the pixel size
        !
        self%pixel_size = pixel_size

        !
        ! Get the ctffind / ctftilt information.
        ! If both are present, the ctftilt information is used.
        !
        if (present(ctf_params)) then
            self%ctftilt%ctf = ctf_params
            self%ctftilt%tilt_angle = 0.0
            self%ctftilt%tilt_axis  = 0.0
        else
            if (present(ctffind_filename)) then
                if (present(ctftilt_filename)) then
                    write(*,'(a)') '**info(Micrograph::Init): both ctffind and ctftilt info was given. '//&
                                    'ctftilt will take precedence.'
                endif
                call ctftilt_read_from_ctffind(self%ctftilt,ctffind_filename,self%id)
            endif
            if (present(ctftilt_filename)) then
                call ctftilt_read_from_ctftilt(self%ctftilt,ctftilt_filename)
            endif
        endif

        !
        ! Movie parameters
        !
        if (self%ctftilt%ctf%pix_units) then
            !call terminate('Need wavelength in Angstroms')
            call ctf_pix_to_ang(self%ctftilt%ctf,pixel_size)
        endif
        if (self%ctftilt%ctf%wl .gt. 0.0) then
            call self%electron_dose%Init(wl_to_akv(self%ctftilt%ctf%wl))
        else
            write(*,'(a)')'**info(Micrograph::Init): electron wavelength is not known. '//&
                            'Assuming 300 keV for purpose of electron dose initialisation.'
            call self%electron_dose%Init(300.0)
        endif

        !
        ! Error checking
        !
        if (self%location_in_file .lt. 1) then
            write(*,'(a,i0)') '**error(MicrographInit): bad location in file: ', self%location_in_file
            call self%PrintInfo()
            call this_program%TerminateWithFatalError('Micrograph::Init','Bad location in file')
        endif

        !
        ! Done
        !
        if (debug) then
            write(*,'(a)') '**debug(MicrographInit): micrograph info on exit'
            call self%PrintInfo()
        endif
    end subroutine MicrographInit

    !>  \brief  Destructor
    !impure elemental
    subroutine Destroy(self)
        type(Micrograph),   intent(inout)   ::  self
        !
        if (allocated(self%relative_frame_shift_x)) deallocate(self%relative_frame_shift_x)
        if (allocated(self%relative_frame_shift_y)) deallocate(self%relative_frame_shift_y)
    end subroutine Destroy

    !>  \brief  Return the ctf (underfocus) parameters at a point on a film
    function ReturnCTFAtCoos(self,x,y) result(ctf)
        use rotation_matrices, only : radians,degrees   ! Conversion routines
        use runtime_parameters, only : terminate
        use ctfs, only : ctf_parameters
        implicit none
        ! Arguments
        class(Micrograph),  intent(in)  ::  self    ! This used to be class, but that caused problems with openmp
        real(kind=8),       intent(in)  ::  x,y
        ! Result
        type(ctf_parameters)        ::  ctf
        ! Private variables
        real    ::  n1,n2   ! Tilt axis normal
        real    ::  cx,cy   ! Film center coordinates
        real    ::  dx,dy   ! Distance from coos to the center
        real    ::  df      ! Distance (in defocus axis) between center of film and coos
        logical,    parameter   ::  debug = .false.
        ! Start work

        ! We assume the ctf is in Angstroms, not pixels
        if (self%ctftilt%ctf%pix_units) then
            write(*,'(a)') '**ERROR(Microgaph::ReturnCTFAtCoos): ctf should be in Angstroms/degrees, not pixels/radians'
            call terminate('Fatal error in Microgaph::ReturnCTFAtCoos')
        endif

        if (.not. any(self%logical_dimensions .ne. 0)) then
            write(*,'(a,2(i0,1x))') '**ERROR(Microgaph::ReturnCTFAtCoos): unknown dimensions for film: ', self%logical_dimensions
            call terminate('Fatal error in Microgaph::ReturnCTFAtCoos')
        endif


        !print *, '**debug(find_ctf_at_coos): central ctf = ', self%ctftilt%ctf
        if (debug) then
            write(*,'(a)') '**debug(film_find_ctf_at_coos): film info on entry:'
            call self%PrintInfo()
        endif
        ctf = self%ctftilt%ctf  ! This copies the central ctf parameters
        ! Do the calculation according to ctftilt documentation
        n1 = -SIN(radians(self%ctftilt%tilt_axis))
        n2 =  cos(radians(self%ctftilt%tilt_axis))
        cx =  self%logical_dimensions(1) / 2.0
        cy =  self%logical_dimensions(2) / 2.0
        dx =  cx - x
        dy =  cy - y
        df = (N1*DX+N2*DY)*SELF%PIXEL_SIZE*TAN(radians(self%ctftilt%tilt_angle))
        if (debug) then
            print '(a,2(F8.4,1x))', '**debug(Microgaph::ReturnCTFAtCoos): n1, n2 = ', n1, n2
            print '(a,2(F9.2,1x))', '**debug(Microgaph::ReturnCTFAtCoos): dx, dy = ', dx, dy
            print '(a,F9.2)', '**debug(film_find_ctf_at_coos): df = ', df
        endif
        ctf%dfmid1 = self%ctftilt%ctf%dfmid1 + df
        ctf%dfmid2 = self%ctftilt%ctf%dfmid2 + df
    end function ReturnCTFAtCoos

    !>  \brief  Given x,y coordinates, return the z coordinate (defocus, in pixels)
    function ReturnZFromXY(self,x,y) result(z)
        use runtime_parameters, only : terminate, z_axis_direction
        use ctfs, only : ctf_parameters
        use ctftilts, only : ctftilt_print_info
        implicit none
        ! Arguments
        class(Micrograph),  intent(in)  ::  self    !<  film object
        real(kind=8),       intent(in)  ::  x       !<  x coordinates (in pixels)
        real(kind=8),       intent(in)  ::  y
        ! Result
        real                            ::  z
        ! Private variables
        type(ctf_parameters)            ::  ctf
        logical,    parameter           ::  debug = .false.
        ! Start work

        ! We assume the ctf is in Angstroms, not pixels
        if (self%ctftilt%ctf%pix_units) then
            write(*,'(a)') '**ERROR(Micrograph::ReturnZFromXY): ctf should be in Angstroms, not pixels'
            call terminate('Fatal error in Micrograph::ReturnZFromXY')
        endif

        !ctf = film_find_ctf_at_coos(self,x,y)
        ctf = self%ReturnCTFAtCoos(x,y)
        if (debug) then
            write(*,'(a,2(F0.2,1x))') '**debug(Micrograph::ReturnZFromXY): central ctf: '
            call ctftilt_print_info(self%ctftilt)
            write(*,'(a,2(F0.2,1x))') '**debug(Micrograph::ReturnZFromXY): x, y = ', x, y
            print '(a,F0.2)', '**debug(Micrograph::ReturnZFromXY): ctf%dfmid1 = ', ctf%dfmid1
        endif
        z = (ctf%dfmid1 + ctf%dfmid2)/2.0 ! In Angstroms
        ! z_axis_direction defines whether the z axis goes up or down the microscope column. See comments in runtime_parmaters.f90 for more details.
        z =  z_axis_direction * z / self%pixel_size
    end function ReturnZFromXY

end module Micrographs
