!>  \brief  Classes and methods to do with comparing filaments and models and optimising their similarity. fmc stands for Filament-Model Comparison
module fmcs

    use fmcs_base
    use ConjugateGradients

    implicit none

    private
    public :: fmc_maximise, fmc_threed_size, fmc_bf_tilt, fmc_bf_subunit_num_intg_refine, fmc_bf_find_mt_seam
    public :: fmc_unit_tests

    contains

    !>  \brief  Unit tests
    subroutine fmc_unit_tests()
        call fmc_unit_test_1()
    end subroutine fmc_unit_tests

    !>  \brief  Test the OpenMP threading of FMCs
    subroutine fmc_unit_test_1()
        use interpolation, only : default_interpol_scheme
        use images_helices, only : image_impose_helical_symmetry_new
        use images_arithm, only : cyl_mask
        use filaments, only : filament_init
        use Micrographs, only : Micrograph
        use fmcs_evaluation, only : fmc_eval, fmc_model3d_prep
        use ctfs
        use ImageNoise
        use Images
        implicit none
        ! Private variables
        type(filament_model_comparison),    target      ::  self, fmc_local
        type(waypoint)                                  ::  wps(3)
        type(filament)                                  ::  fil
        type(image),                        target      ::  film_image(1)
        type(image),                        target      ::  model3dvol
        type(helical_symmetry)                          ::  helix
        type(interpolation_scheme)                      ::  interpol
        type(Micrograph)                                ::  my_micrograph
        logical,                            parameter   ::  tthreading          =   .false.
        real,                               parameter   ::  pixel_size          =   2.3
        real,                               parameter   ::  target_resolution   =   20.0
        real,                               parameter   ::  high_res_cutoff     =   40.0
        integer                                         ::  number_of_tests     =   2
        integer                                         ::  test_num
        real,                               allocatable ::  final_score(:)
        integer                                         ::  current_wp
        ! Start work

        write(*,'(a)') '**info(fmc_unit_test_1): entering'

        ! Initialise a helix
        call helical_symmetry_init(helix,rise=1.408/pixel_size,twist=-22.03,outer_radius=90.0/pixel_size)

        ! Initialise an interpolation scheme
        call interpolation_scheme_init(interpol,'LINEAR','RECT',2)

        ! Create a film image
        film_image(1)%logical_dimensions = [512,512,1]
        call film_image(1)%Allocate()
        film_image(1)%real_values = 0.0e0
        call AddGaussianNoise(film_image(1),stdev_noise=1.0)
        call my_micrograph%Init(id=1,filename='film.mrc',pixel_size=pixel_size,film_image_available=.false.)
        ! Manually set-up the ctf information for the micrograph
        call ctf_parameters_init(my_micrograph%ctftilt%ctf,akv=200.0,cs=2.0,ampl_cont=0.07,dfmid1=-1.0)

        ! Create  a 3d model
        model3dvol%logical_dimensions = [100,100,100]
        call model3dvol%Allocate()
        model3dvol = 0.0e0
        call AddGaussianNoise(model3dvol,stdev_noise=1.0)
        call image_impose_helical_symmetry_new(model3dvol,helix,interpol_scheme=interpol)

        call cyl_mask(model3dvol,rad=30.0,soft=3,rad_in=5.0)

        ! Create an array of waypoints
        wps%helix=helix
        wps(1)%coo = [100.0, 90.0,-5000.0]
        wps(2)%coo = [250.0,300.0,-5000.0]
        wps(3)%coo = [400.0,390.0,-5000.0]

        do current_wp=1,3
            allocate(wps(current_wp)%local_subframe_shift(3,1))
            wps(current_wp)%local_subframe_shift = 0.0e0
        enddo


        ! Create a filament
        call filament_init( fil,waypoints_in=wps,helix=helix,id=123,max_radius=28.0,film_in=my_micrograph,  &
                            refine_params=[.true.,.false.,.false.,.false.,.false.,.false.,.false.,.false.,.false.],  &
                            fragmented_phi=.true.,fragmented_subunit_num=.true.)

        ! Initialise a fmc object
        call fmc_init(      self,fil,film_image,model3dvol,helix,                       &
                            target_resolution,                                          &
                            interpol=interpol,  &
                            asym_unit_mw=17000.0,  &
                            use_conjugate_gradient=.true.,&
                            high_resolution_cutoff=high_res_cutoff, &
							num_minim_runs=1,minim_random_start_pos=.false.)
        call fmc_print_info(self)
        call fmc_model3d_prep(self)

        ! Call the scoring function several times
        allocate(final_score(number_of_tests))
        !$omp   parallel    if(tthreading)  &
        !$omp               default(none)   &
        !$omp               private(test_num,fmc_local) &
        !$omp               shared(number_of_tests,fil,final_score,self)
        call fmc_assignment(fmc_local,self)
        !$omp do
        do test_num=1,number_of_tests
            call fmc_new_filament(fmc_local,fil)
            call fmc_maximise(fmc_local)
            final_score(test_num) = fmc_eval(fmc_local,self%fil)
        enddo
        !$omp enddo
        ! Deallocate
        call fmc_deallocate(fmc_local)
        !$omp   end parallel

        ! Check all the results were the same
        if (any(abs(final_score(2:size(final_score)) - final_score(1)) .gt. 0.0001)) then
            write(*,'(a)') '**ERROR(fmc_unit_test_1): not all scores were identical:'
            print *, final_score
            call terminate('Failed unit test FMC_UNIT_TEST_1')
        else
            write(*,'(a)') '**info(fmc_unit_test_1): as expected, all scores were identical'
        endif

        call fmc_deallocate(self)

        write(*,'(a)') '**info(fmc_unit_test_1): all done'

    end subroutine fmc_unit_test_1

    !>  \brief  Maximise the agreement between a filament's model and the 3d reference
    subroutine fmc_maximise(self)
        use fmcs_evaluation, only : fmc_eval, fmc_model3d_prep
        use filaments_model_modify, only : filament_add_disorder
        use filaments_model_base, only :  filament_print_extra_info
        use images_comparison, only : score_got_worse
        use omp_lib
		implicit none
        ! Arguments
        type(filament_model_comparison),    target,     intent(inout)   ::  self
        ! private variables
        integer                                         ::  num_random_starts, num_dim_total, current_minim_run
        integer                                         ::  ierr, best_minim_run(1)
        real(kind=8)                                    ::  score_on_entry, score_on_exit
        type(filament)                                  ::  fil_on_entry
        type(filament),                     allocatable ::  fil_tmp(:)
        type(filament_model_comparison),    allocatable ::  self_private(:)
        real(kind=8),                       allocatable ::  final_values(:,:), final_score(:)
        real(kind=8),                       allocatable ::  start_values(:,:), start_score(:), best_final_values(:)
        real(kind=8),                       allocatable ::  values_on_entry(:)
        real(kind=8),                       allocatable ::  vals(:), step(:)
        real                                            ::  random_std(9)
        real,                               parameter   ::  random_std_cumul(9) = (/0.,0.,0.,0.,0.,0.,0.,0.,0./)
        character(len=6)                                ::  num_dim_c, num_random_starts_c
#ifdef SKIP_RUNTIME_CHECKS
        logical,                            parameter   ::  sanity_checks = .false.
#else
        logical,                            parameter   ::  sanity_checks = .true.
#endif
        type(filament),                     allocatable ::  final_fil(:)
        logical,                            parameter   ::  debug = .false.
        integer                                         ::  current_thread_num
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(fmc_maximise): on entry, helix extra info '
            call helical_symmetry_print_extra_info(self%helix)
            call helical_symmetry_print_extra_info(self%fil%helix)
            call helical_symmetry_print_extra_info(self%fil%wp(1)%helix)
        endif

        ! Remember the starting point
        fil_on_entry = self%fil
        call fmc_start_step_values(self,values_on_entry,step,values_to_fil=.false.)

        ! Make sure the model 3d is ready
        call fmc_model3d_prep(self)

        ! Score on entry
        score_on_entry = fmc_eval(self,self%fil,detailled_score_info=.false.)
        write(*,'(a,f0.4)') '**info(fmc_maximise): score on entry: ', score_on_entry

        ! How many randomised starts of the minimiser will we do?
        num_random_starts = 1
        if (self%minim_random_start_pos) num_random_starts = self%num_minim_runs
        write(num_random_starts_c,'(i6)') num_random_starts

        ! Allocate memory for the results of the minimiser runs
        num_dim_total  = fmc_num_dim_total(self) ! number of dimensions the simplex is going to deal with
        write(num_dim_c,'(i6)') num_dim_total
        allocate(   final_values(num_dim_total,num_random_starts), final_score(num_random_starts),   &
                    start_values(num_dim_total,num_random_starts), start_score(num_random_starts),   &
                    final_fil(num_random_starts), &
                    best_final_values(num_dim_total),   &
                    stat=ierr )
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**error(fmc_ds_maximise): memory allocation failed with error ', ierr
            call terminate('Fatal error in fmc_ds_maximise')
        endif

        ! Standard deviations for the filament randomisation later on
        random_std = 0.0
        if (self%minim_random_start_pos) then
            where (self%refine_params) random_std = self%search_half_range/2.0
			if (debug) then
			    write(*,*) '**debug(fmc_maximise): random_std = ', random_std
			    write(*,*) '**debug(fmc_maximise): refine_params = ', self%refine_params
			    write(*,*) '**debug(fmc_maximise): half_range = ', self%search_half_range
			endif
        endif

        ! Loop over the number of runs of the maximiser
        ! ifort 13.1.3.192 has unexpected behaviour here, see
        ! http://software.intel.com/en-us/forums/topic/429902
        ! In fact, OpenMP does not support allocatable components, so i think
        ! i will resort to keeping everything shared.
        if (debug) write(*,'(a)') '**debug(fmc_maximise): about to enter parallel region'
        if (allocated(vals)) deallocate(vals) ! we deallocate these because they are private in the OpenMP region
        if (allocated(step)) deallocate(step)
        !$omp   parallel    if(num_random_starts .gt. 1) &
        !$omp               default(shared)                       &
        !$omp               private(current_minim_run,vals,step,current_thread_num)

#ifdef _OPENMP
        current_thread_num = omp_get_thread_num() + 1
#else
        current_thread_num = 1
#endif

        !$omp   single
        if (debug) write(*,'(a)') '**debug(fmc_maximise): about to allocate fil_tmp and self_private'
#ifdef _OPENMP
        allocate(fil_tmp(omp_get_num_threads()),self_private(omp_get_num_threads()))
#else
        allocate(fil_tmp(1),self_private(1))
#endif
        !$omp   end single


        !$omp critical
        if (debug) write(*,'(a)') '**debug(fmc_maximise): about to assign fmc object'
        call fmc_assignment(self_private(current_thread_num),self)
        if (debug) write(*,'(a)') '**debug(fmc_maximise): assigned fmc object'
        !$omp end critical

        !$omp   do
        do current_minim_run = 1, num_random_starts
            if (debug) write(*,'(a,i0)') '**debug(fmc_maximise): on ', current_minim_run
            !
            if (self_private(current_thread_num)%minim_random_start_pos) then
                write(*,'(2(a,i0))') '**info(fmc_maximise): on random start ', current_minim_run, ' of ', num_random_starts
            endif
			! randomise the filament
			if (debug) write(*,'(a)') '**debug(fmc_maximise): about to assign filament'
            fil_tmp(current_thread_num) = self_private(current_thread_num)%fil
			if (current_minim_run .gt. 1) then
				call filament_add_disorder( fil_tmp(current_thread_num),              &
    	                                    cumulative_disorder_std=random_std_cumul,   &
        	                                non_cumulative_disorder_std=random_std )
			endif
			if (self_private(current_thread_num)%minim_random_start_pos) then
                !$omp	critical
				write(*,'(//2(a,i0),a)') '**info(fmc_maximise): on random start ', current_minim_run, ' of ', num_random_starts, &
									'. below are random changes to the filament: '
				call filament_print_diff(self_private(current_thread_num)%fil,fil_tmp(current_thread_num))
				!$omp	end critical
			endif
			! It may make sense to turn off refinement of fal waypoints
			! Don't do this here because it causes problems with the number of dimensions being refined. Instead, maybe introduce a runtime flag
			! called refine_fal_wps.
			!if ((.not. any(fil_tmp(current_thread_num)%refine_normal_modes)) .and. &
			!    (.not. fil_tmp(current_thread_num)%fal_wps_in_score)  ) then
            !    write(*,'(a)') '**debug(fmc_maximise): will not refine first and last waypoints'
            !    fil_tmp(current_thread_num)%wp(1)%refine = .false.
            !    fil_tmp(current_thread_num)%wp(fil_tmp(current_thread_num)%nwp)%refine = .false.
            !endif
			!
			call fmc_new_filament(self_private(current_thread_num),fil_tmp(current_thread_num))
            ! remember the starting point
            call fmc_start_step_values(self_private(current_thread_num),vals,step,values_to_fil=.false.)
            !print *, '**debug(fmc_maximise): fmc_num_dim_total = ', fmc_num_dim_total(self_private(current_thread_num))
            !print *, '**debug(fmc_maximise): size(vals), size(start_values,1) = ', size(vals), size(start_values,1)
            !if (size(vals) .ne. size(start_values,1)) call terminate('Fatal error in fmc_maximise. Wrong number of values.') ! Comment this out in a few weeks (141125)
            start_values(:,current_minim_run) = vals
            start_score(current_minim_run) = fmc_eval(  self_private(current_thread_num),     &
                                                        self_private(current_thread_num)%fil, &
                                                        detailled_score_info=.false.)
            ! call the maximiser routine
            if (self_private(current_thread_num)%refine_wps_simultaneously) then
                call fmc_maximise_simultaneous(self_private(current_thread_num))
            else
                call fmc_maximise_sequential(self_private(current_thread_num))
            endif
            ! remember the finishing point
            call fmc_start_step_values(self_private(current_thread_num),vals,step,values_to_fil=.false.)
            final_values(:,current_minim_run) = vals
            !$omp critical
            write(*,'(//2(a,i0),a,2(f0.4,1x))') '**info(fmc_maximise): final score for start ',  &
													current_minim_run, ' of ', num_random_starts
            final_score(current_minim_run) = fmc_eval(  self_private(current_thread_num),     &
                                                        self_private(current_thread_num)%fil, &
                                                        detailled_score_info=.false.)
            final_fil(current_minim_run) = self_private(current_thread_num)%fil
            !call filament_print_info(final_fil(current_minim_run),verbose=.true.)
            !call filament_print_extra_info(final_fil(current_minim_run))
            !call filament_print_info(self_private(current_thread_num)%fil,verbose=.true.)
            !$omp end critical
            ! sanity check
			if (sanity_checks .and. score_got_worse(start_score(current_minim_run),final_score(current_minim_run)) &
                    .and. .not. self%wps_as_sps) then
                !$omp critical
                write(*,'(//2(a,i0),a,2(f0.4,1x))') '**error(fmc_maximise): on start ',  &
													current_minim_run, ' of ', num_random_starts, &
												', score at start > score at finish: ', &
                                                start_score(current_minim_run), final_score(current_minim_run)
                if (any(self%refine_params(8:9))) then
                    write(*,'(a)') '**info(fmc_maximise): ignoring warning above since we are refining helical parameters'
                else
                    final_score(current_minim_run) = fmc_eval(  self_private(current_thread_num),     &
                                                                self_private(current_thread_num)%fil, &
                                                                detailled_score_info=.true.)
                    call self_private(current_thread_num)%fil%PrintInfo(verbose=.true.)
                    call terminate('failed sanity check in fmc_maximise')
				endif
				!$omp end critical
            endif
        enddo
        !$omp   enddo
        call fil_tmp(current_thread_num)%Deallocate()
        call fmc_deallocate(self_private(current_thread_num))

        !$omp   barrier
        !$omp   single
        deallocate(fil_tmp,self_private)
        !$omp   end single

        if (allocated(vals)) deallocate(vals)
        if (allocated(step)) deallocate(step)

        !$omp   end parallel

        ! how many of the random starts gave improved score?
        if (self%minim_random_start_pos) then
            !$omp critical
            write(*,'(//a,f0.4)') '**info(fmc_maximise): score on entry was ', score_on_entry
            write(*,'(a,'//num_random_starts_c//'(f0.4,1x))') '**info(fmc_maximise): final scores: ', final_score
            write(*,'(a,2(i0,a))') '**info(fmc_maximise): ', count(final_score .gt. score_on_entry), ' of ', &
                                        num_random_starts, ' maximisation runs gave improved score'
            !$omp end critical
        endif

        ! decide which run of the minimiser gave the best result
        best_minim_run = maxloc(final_score)
        write(*,'(a,i0,a,f0.5)') '**info(fmc_maximise): maximisation run ', best_minim_run(1), &
                                ' gave the best results, with score: ', &
                                 final_score(best_minim_run(1))
        best_final_values = final_values(:,best_minim_run(1))

        ! the best final score should be better than the starting score, except if we strarted from randomised
        ! positions, in which case we may be worse
		if (score_got_worse(score_on_entry,final_score(best_minim_run(1)))) then
            if (.not. self%minim_random_start_pos) then
                write(*,'(a,2(f0.5,1x))') '**error(fmc_maximise): best final score less than starting score: ', &
                                        final_score(best_minim_run(1)), score_on_entry
                if (any(self%refine_params(8:9))) then
                    write(*,'(a)') '**info(fmc_maximise): ignoring warning above since we are refining helical parameters'
                else
                    if (sanity_checks) then
                        call this_program%TerminateWithFatalError('fmc_maximise','best final score less than starting score')
                    endif
                endif
            else
                write(*,'(a)') '**info(fmc_maximise): none of the ds runs gave better results than on entry, ',  &
                                'so the filament will be left unchanged'
                best_final_values = values_on_entry
				final_fil(best_minim_run(1)) = fil_on_entry
            endif
        endif

        ! Stick best filament into fmc object
        call fmc_new_filament(self,final_fil(best_minim_run(1)))!,update_step_and_box_size=.false.)
        ! plug those best values back into the filament     (the ds routine does not necessarily finish on the minimum)
        !call fmc_start_step_values(self,best_final_values,step,values_to_fil=.true.) ! note that in this call, nothing will happen to first_step

        ! We run the scoring routine one last time, because this calls filament_update_model with the proper parameters, so that the best parameters are propagated through the filament model
        write(*,'(a)') '**info(fmc_maximise): detailed score of best-scoring filament model:'
        score_on_exit = fmc_eval(self,self%fil,detailled_score_info=.true.)

        if (sanity_checks) then
            if (score_got_worse(final_score(best_minim_run(1)),score_on_exit) &
                .and. score_on_exit .gt. 0.0 .and. final_score(best_minim_run(1)) .gt. 0.0) then
				write(*,'(a,2(f0.7,1x))') '**warning(fmc_maximise): score on exit less than score after minim: ', &
                                            score_on_exit, final_score(best_minim_run(1))
                !if (score_on_exit .lt. final_score(best_minim_run(1))-0.05) then
                if (      .not. final_fil(best_minim_run(1))%fragmented_phi &
                    .and. .not. final_fil(best_minim_run(1))%fragmented_subunit_num &
                    .and. .not. any(self%refine_params(8:9))) then
                    ! When fragmented phi or subunit_num, one of the waypoints can easily land on the other side of the divide
                    ! and give a significantly different score
                    ! Also, when the helical parameters have been refined, the reference will be resymmetrised after refinement, and this
                    ! could lead to worse scores.
                    write(*,'(a)') '**error(fmc_maximise): discrepency in scores is too large, is there a bug?'
                    score_on_exit = fmc_eval(self,final_fil(best_minim_run(1)),detailled_score_info=.true.)
                    call final_fil(best_minim_run(1))%PrintInfo(verbose=.true.)
                    call filament_print_extra_info(final_fil(best_minim_run(1)))
                    call self%fil%PrintInfo(verbose=.true.)
                    call terminate('fatal error in fmc_maximise')
                else
                    if (any(self%refine_params(8:9))) then
                        write(*,'(a)') '**warning(fmc_maximise): this could be symptomatic of inaccuracies '//&
                                        'when doing helical parameter refinements. That feature is still experimental!'
                    endif
                endif
            endif
			if (score_got_worse(score_on_entry,score_on_exit)) then
				write(*,'(a,2(f0.7,1x))') '**error(fmc_maximise): score on exit less than on entry: ', &
                                            score_on_exit, score_on_entry
                call terminate('fatal error in fmc_maximise')
            endif
        endif

        write(*,'(a)') '**info(fmc_maximise): filament info on exit:'
        call self%fil%PrintInfo()

        write(*,'(a,f0.5)') '**info(fmc_maximise): score on exit: ', score_on_exit

    end subroutine fmc_maximise

    !>  \brief  Maximise the agreement between a filament's model and the 3d model density by letting
    !!          a minimiser tweak all the waypoints' parameters simultaneously
    subroutine fmc_maximise_simultaneous(self)
        implicit none
        ! Arguments
        type(filament_model_comparison),    target,     intent(inout)   ::  self
        ! Private variables

        ! Start work

        if (self%wps_as_sps) then
            call fmc_sp_maximise(self)
        else if (self%type_of_minimiser .eq. downhill_simplex) then
            call fmc_ds_maximise(self)
        else if (self%type_of_minimiser .eq. conjugate_gradient) then
            call fmc_cg_maximise(self)
        else if (self%type_of_minimiser .eq. differential_evolution) then
            call fmc_de_maximise(self)
        else if (self%type_of_minimiser .eq. brute_force) then
            call fmc_cg_maximise(self,bf_instead=.true.)
        else
            write(*,'(a)') '**ERROR(fmc_maximise): don''t know which minimiser to use'
            call terminate('Fatal error in FMC_DS_MAXIMISE')
        endif
    end subroutine fmc_maximise_simultaneous

    !>  \brief  Treat the filament waypoints as single particles, and refine the waypoints independently from each other
    subroutine fmc_sp_maximise(self)
        use single_particles, only : single_particle, sp_print_info
		use filaments, only : wp_to_sp
        use sp_model_comparisons, only : sp_model_comparison, spmc_init, spmc_get_sp, spmc_print_info, spmc_ds_maximise
        use sp_model_comparisons, only : spmc_deallocate_work_memory, spmc_bf_maximise
        use imagefiles, only : imagefile
        use omp_lib
		use filaments_model_update, only : filament_update_model
        use fmcs_evaluation, only : fmc_model3d_prep, fmc_eval, fmc_eval_wrap
        use ctfs, only : z_from_ctf, ctf_ang_to_pix
        use images_helices, only : image_impose_helical_symmetry_new
        use runtime_parameters, only : psf_radius_pixels, filename_len
        use minimisers
        use Globals
        ! debug
        use coos, only : wp_print_info
!#ifdef __INTEL_COMPILER
!        use coos, only : point_assign
!#endif
		implicit none
        ! Arguments
        type(filament_model_comparison),    target,     intent(inout)   ::  self                !<  Filament model comparison object
        ! Private variables
        integer                     ::  current_wp, first_wp, last_wp
        logical,    parameter       ::  debug   =   .false.
        integer                     ::  sp_box_size
        logical                     ::  tthreading
        character(len=filename_len) ::  dbg_boxes_filename, dbg_projs_filename
        type(imagefile)             ::  dbg_boxes_file,dbg_projs_file
        type(waypoint),             allocatable     ::  wp_phased_private(:)
        type(image),                allocatable     ::  sp_image_private(:)
        type(image),                allocatable     ::  model3d_private(:)
        type(sp_model_comparison),  allocatable     ::  spmc_private(:)
        type(single_particle),      allocatable     ::  sp_private(:)
        integer                                     ::  current_thread, number_of_threads
        logical,                    parameter       ::  hextend_per_sp = .false. ! 150604: warning, this is buggy
        logical,                    parameter       ::  write_matching_projections_to_disk = .false.
        ! Start work

        if (debug) write(*,'(a)') '**debug(fmc_sp_maximise): entry'


        ! There are cases where no work needs to be done
        if (.not. any(self%fil%wp%refine)) then
            write(*,'(a)') '**WARNING(fmc_sp_maximise): nothing to do, since no way points marked for refinement'
            return
        endif

        ! Work out what box size we will use
        sp_box_size = fmc_box_size(self%interpol,self%sp_segment_length/self%fil%film%GetPixelSize()/2.)
        if (debug) then
            write(*,'(a,i0,a,F0.3,a,F0.3,a)')   '**debug(fmc_sp_maximise): will use box size of ', sp_box_size, &
                                                ' pixels. sp segment length is ',   &
                                                self%sp_segment_length/self%fil%film%GetPixelSize(), &
                                                ' pixels and the psf radius is ',                  &
                                                psf_radius_pixels, ' pixels'
        endif

        ! Debug
        if (write_matching_projections_to_disk) then
            dbg_boxes_filename = 'dbg_sp_max_boxes.mrc'
            dbg_projs_filename = 'dbg_sp_max_projs.mrc'
            call dbg_boxes_file%init(dbg_boxes_filename)
            call dbg_projs_file%init(dbg_projs_filename)
        endif

        ! Make sure the model 3d is ready
        call fmc_model3d_prep(self,threed_size=sp_box_size)

        ! Work out the first and last waypoints we'll be processing
        if (.not. self%fil%fal_wps_in_score) then
            first_wp = 2
            last_wp = self%fil%nwp - 1
        else
            first_wp = 1
            last_wp = self%fil%nwp
        endif

        if (debug) print '(a,i0)', '**debug(fmc_sp_maximise): using large box size: ', sp_box_size

        ! Avoid nested parallelisation
#ifdef _OPENMP
        if (omp_get_active_level() .gt. 0 .or. debug) then
            tthreading = .false.
        else
            tthreading = .true.
        endif
#else
        tthreading = .false.
#endif

        tthreading = .false.
        write(*,'(a)') '**info(fmc_sp_maximise): disabling OpenMP threading because buggy'

        !$omp   parallel    if(tthreading)  &
        !$omp               default(shared)   &
        !$omp               private(current_thread)

#ifdef _OPENMP
        current_thread      = omp_get_thread_num()+1
#else
        current_thread      = 1
#endif

        ! Allocate objects that will be pseudo-thread-private
        !$omp single
#ifdef _OPENMP
        number_of_threads   = omp_get_num_threads()
#else
        number_of_threads   = 1
#endif
        allocate(wp_phased_private(number_of_threads),sp_image_private(number_of_threads), &
                sp_private(number_of_threads),spmc_private(number_of_threads), &
                model3d_private(number_of_threads))
        !$omp end single

        call model3d_private(current_thread)%Reset()
        call sp_image_private(current_thread)%Reset()

        ! Loop over the waypoints. Each in turn will be "converted" to a single particle and refined.
        !$omp   do
        do current_wp = first_wp, last_wp
            !$omp   critical (fmc_sp_printout_omp_crit)
            write(*,'(a,i0,a,i0)',advance='NO') '**info(fmc_sp_maximise): working on waypoint ', current_wp, &
                                                ' of filament ', self%fil%id
#ifdef _OPENMP
            if (tthreading) then
                write(*,'(a,i0,a,i0)') ' on thread ', omp_get_thread_num()+1, ' of ', omp_get_num_threads()
            else
                write(*,'(a)') ' '
            endif
#else
            write(*,'(a)') ' '
#endif
            !$omp   end critical (fmc_sp_printout_omp_crit)

            ! Create a thread-private copy of the current waypoint
            wp_phased_private(current_thread) = self%fil%wp(current_wp)
            if(debug) then
                write(*,'(a)') '**debug(fmc_sp_maximise): helix info of fil%wp and wp_phased_private'
                call helical_symmetry_print_extra_info(self%fil%wp(current_wp)%helix)
                call helical_symmetry_print_extra_info(wp_phased_private(current_thread)%helix)
            endif

            ! Impose helical symmetry
            model3d_private(current_thread) = self%model3dvol_private
            if (hextend_per_sp) then
                write(*,'(a)',advance='no') '**debug(fmc_sp_maximise): about to impose h sym: '
                call helical_symmetry_print_info(wp_phased_private(current_thread)%helix)
                call image_impose_helical_symmetry_new(model3d_private(current_thread),wp_phased_private(current_thread)%helix, &
                                                        self%interpol, max_radius=self%fil%max_radius)
            endif

            !

            ! Convert the waypoint to a single particle object
            ! and box the image from the film image
            call wp_to_sp(wp_phased_private(current_thread),sp_private(current_thread), &
                            self%fil%film%ctftilt%ctf,   &
                            sp_image = sp_image_private(current_thread),    &
                            segment_length = sp_box_size, &
                            micrograph_frames = self%micrograph_frames,&
                            pixel_size=self%fil%film%GetPixelSize(), &
                            fil=self%fil,square_image=.true., &
                            apply_shift_to_image=.true., &
                            first_frame_to_sum=self%first_frame_for_refinement, &
                            last_frame_to_sum=self%last_frame_for_refinement, &
                            apply_drift_filter=self%apply_drift_filter, &
                            apply_dose_filter=self%apply_dose_filter, &
                            electron_dose=self%fil%film%electron_dose, &
                            exposure_per_frame=self%fil%film%GetExposurePerFrame())


            ! Detailled debug info
            if (debug) then
                write(*,'(a,i0,a)',advance='NO') '**debug(fmc_sp_maximise): waypoint #', current_wp, ' info: '
                call wp_print_info(self%fil%wp(current_wp))
                call wp_print_info(wp_phased_private(current_thread))
                write(*,'(a,i0,a)',advance='NO') '**debug(fmc_sp_maximise): waypoint #', current_wp, ' into single particle: '
                call sp_print_info(sp_private(current_thread))
            endif

            ! Initialise a single particle - model comparison object
            if (self%minim_random_start_pos) then
                write(*,'(a)') '**ERROR(fmc_sp_maximise): randomised starting positions not yet implemented with SPs'
                call terminate('Fatal error in FMC_SP_MAXIMISE')
            endif
            call spmc_init(     spmc_private(current_thread),sp_private(current_thread),        &
                                model3d_private(current_thread),sp_image_private(current_thread),   &
                                refine_params=self%refine_params,                               &
                                pixel_size=self%fil%film%GetPixelSize(),                            &
                                target_resolution=self%target_resolution,                       &
                                low_resolution_cuton=self%fil%film%GetPixelSize()/self%freq_min,    &
                                high_resolution_cutoff=self%fil%film%GetPixelSize()/self%freq_max,  &
                                obj_max_rad=max(self%sp_segment_length/2./self%fil%film%GetPixelSize(),self%fil%max_radius), &
                                interpol=self%interpol,                                         &
                                real_space_proj=self%real_space_proj,                           &
                                project_individual_subunits=self%project_helical_subunits,      &
                                search_half_range_override=self%search_half_range,              &
                                num_minim_runs=self%num_minim_runs,                             &
                                ds_random_start=self%ds_random_start_conf,                      &
                                obj_max_rad_cyl=self%fil%max_radius,                            &
                                boost_low_frequency_contrast=self%boost_low_frequency_contrast, &
                                model_helix=wp_phased_private(current_thread)%helix ,         &
                                model3d_oversampling=self%model3d_oversampling, &
                                minimiser_algorithm=self%type_of_minimiser, &
                                search_step_size=self%search_sampling &
                                )

            ! Detailled debug info
            if (debug .or. current_wp .eq. first_wp) then
                call spmc_print_info(spmc_private(current_thread))
            endif

            select case (self%type_of_minimiser)
                case (downhill_simplex)
                    call spmc_ds_maximise(spmc_private(current_thread))
                case (brute_force)
                    call spmc_bf_maximise(spmc_private(current_thread))
                case (conjugate_gradient)
                    if (this_program%IsVerbose()) then
                        write(*,'(a)') '**info(fmc_sp_maximise): single-particle conjugate gradient not yet implement,'//&
                                       ' using downhill simplex instead'
                    endif
                    call spmc_ds_maximise(spmc_private(current_thread))
                case default
                    write(*,'(a,i0)') '**error(fmc_sp_maximise): oops, unknown minimiser: ', self%type_of_minimiser
                    call this_program%TerminateWithFatalError('fmc_sp_maximise','Unknown minimiser')
            end select


            ! Get the result out, and plug back into the filament's waypoint
            sp_private(current_thread) = spmc_get_sp(spmc_private(current_thread))
            call ctf_ang_to_pix(sp_private(current_thread)%ctf,self%fil%film%GetPixelSize())

            if (debug) then
                write(*,'(a)') '**debug(fmc_sp_maximise): sp_private after refinement: '
                call sp_print_info(sp_private(current_thread))
            endif

!#ifdef __INTEL_COMPILER
!            call point_assign(self%fil%wp(current_wp),wp_phased)
!#else
            self%fil%wp(current_wp)             =   wp_phased_private(current_thread)
!#endif
            !self%fil%wp(current_wp)%coo(1)      =   nint(wp_phased_private(current_thread)%coo(1)) + &
            !                                        sp_private(current_thread)%shift_x
            !self%fil%wp(current_wp)%coo(2)      =   nint(wp_phased_private(current_thread)%coo(2)) + &
            !                                        sp_private(current_thread)%shift_y
            self%fil%wp(current_wp)%coo(1)      =   (wp_phased_private(current_thread)%coo(1)) + &
                                                    sp_private(current_thread)%shift_x
            self%fil%wp(current_wp)%coo(2)      =   (wp_phased_private(current_thread)%coo(2)) + &
                                                    sp_private(current_thread)%shift_y
            self%fil%wp(current_wp)%coo(3)      =   z_from_ctf(sp_private(current_thread)%ctf)
            self%fil%wp(current_wp)%euler       =   sp_private(current_thread)%angles
            self%fil%wp(current_wp)%subunit_num =   nint(self%fil%wp(current_wp)%subunit_num) + &
                                                    sp_private(current_thread)%subunit_num
            self%fil%wp(current_wp)%score       =   sp_private(current_thread)%score
            self%fil%wp(current_wp)%helix       =   sp_private(current_thread)%helix

            ! Write to disk
            if (write_matching_projections_to_disk) then
                call spmc_private(current_thread)%sp_image_ready%WriteToImageFile(dbg_boxes_file,current_wp)
                call spmc_private(current_thread)%proj_ready%WriteToImageFile(dbg_projs_file,current_wp)
            endif

        enddo ! End of loop over waypoints
        !$omp   end do

        call spmc_deallocate_work_memory(spmc_private(current_thread))
        call sp_image_private(current_thread)%Deallocate()
        !$omp   barrier

        !$omp   single
        deallocate(wp_phased_private,sp_image_private,sp_private,spmc_private,model3d_private)
        !$omp   end single

        !$omp   end parallel

        self%fil%rigid_wp%subunit_num = 0.0

        if (debug) then
            write(*,'(a)') '**debug(fmc_sp_maximise): after refinement, but before updating model, here is filament info:'
            call self%fil%PrintInfo()
        endif

        if (write_matching_projections_to_disk) then
            call dbg_boxes_file%close()
            call dbg_projs_file%close()
        endif

        ! Update splines
        call filament_update_model( self%fil,wps_as_sps=.true.,force_update_rigid_wp=.true.)

        if (debug) then
            write(*,'(a)') '**debug(fmc_sp_maximise): after refinement, and after updating model, here is filament info:'
            call self%fil%PrintInfo()
        endif

        ! We may have altered the size of the model3dvol, so make sure it's ready for its next use
        call fmc_model3d_prep(self)

        ! All done
        if (debug) write(*,'(a)') '**debug(fmc_sp_maximise): all done'
    end subroutine fmc_sp_maximise

    !>  \brief  Maximimse the scoring function by varying subunit_num values by integer amounts
    subroutine fmc_bf_subunit_num_intg_refine(self)
        use fmcs_evaluation, only : fmc_eval
        use filaments_geometry, only : filament_update_wp_res
        use filaments_normal_modes, only : filament_wps_to_nm_amplitudes
		implicit none
        ! Arguments
        type(filament_model_comparison),    intent(inout)   ::  self    !<  Filament-model comparison object
        ! Private variables
        logical,                            parameter       ::  debug   =   .false.
        integer                                             ::  subunit_num_half_range
        integer,                            parameter       ::  num_iterations = 1
        integer                                             ::  current_iteration
        integer                                             ::  current_wp
        integer                                             ::  i
        integer                                             ::  current_subunit_num_shift
        real                                                ::  current_score
        real                                                ::  best_score_for_current_wp
        integer                                             ::  best_subunit_num_shift_for_current_wp
        integer,                            allocatable     ::  wp_order(:)
        logical,                            parameter       ::  maximise = .true.
        integer,                            parameter       ::  minimum_search_half_range = 2
        type(filament)                                      ::  refined_fil, starting_fil, tmp_fil
        type(filament),                     allocatable     ::  refined_fils(:)
        logical,                            parameter       ::  concomitant_phi_change = .false.
        ! Start work

        ! Diagnostics
        write(*,'(a)') '**info(fmc_bf_subunit_num_intg_refine): filament info on entry:'
        call self%fil%PrintInfo()
        if (debug) then
            write(*,'(a)') '**info(fmc_bf_subunit_num_intg_refine): detailled score on entry'
            current_score = fmc_eval(self,self%fil,DETAILLED_SCORE_INFO=.true.)
        endif

        ! Search half range for subunit num
        subunit_num_half_range = max(nint(self%search_half_range(7)),minimum_search_half_range)
        !subunit_num_half_range = 7 ! override
        write(*,'(a,i0)') '**info(fmc_bf_subunit_num_intg_refine): search half-range = ', subunit_num_half_range
        if (concomitant_phi_change) then
            write(*,'(a)') '**info(fmc_bf_subunit_num_intg_refine): phi will be changed along with subunit_num'
        endif

        ! For each subunit_num shift, remember the refined filament configuration by storing it
        allocate(refined_fils(-SUBUNIT_NUM_HALF_RANGE:subunit_num_half_range))

        ! What order will we be processing the waypoints?
        wp_order = fmc_get_wp_order(self)

        ! We make sure we keep track of where we started from
        starting_fil = self%fil
        refined_fil = starting_fil

        ! Loop over waypoints
        do current_iteration = 1, NUM_ITERATIONS
        do I=1,self%fil%NWP+1
            current_wp = wp_order(i)
            if (debug) write(*,'(a,i0)') '**debug(fmc_bf_subunit_num_intg_refine): working on waypoint ', current_wp
            self%fil = refined_fil
            self%fil_original = refined_fil
            best_score_for_current_wp = -999.
            ! Loop over subunit_num shift values
            do CURRENT_SUBUNIT_NUM_SHIFT=-SUBUNIT_NUM_HALF_RANGE,subunit_num_half_range
                tmp_fil = refined_fil
                write(*,'(a,i0,a,i0)') '**info(fmc_bf_subunit_num_intg_refine): on waypoint ', current_wp,  &
                                        ', trying subunit_num shift of ', current_subunit_num_shift
                if (current_wp .ge. wp_order(1)) then
                    ! Apply subunit shift to current waypoint and all subsequent ones
                    tmp_fil%wp(current_wp:)%subunit_num = refined_fil%wp(current_wp:)%subunit_num + real(current_subunit_num_shift)
                    ! Apply phi shift if asked for it
                    if (concomitant_phi_change) then
                        tmp_fil%wp(current_wp:)%euler%phi = refined_fil%wp(current_wp:)%euler%phi + &
                                                               refined_fil%wp(current_wp:)%helix%twist * &
                                                               real(current_subunit_num_shift)
                    endif
                else
                    ! In this case, we're moving from the first wp towards the first one, so let's shift waypoints from the beginning of the filament to the current wp
                    tmp_fil%wp(1:current_wp)%subunit_num = refined_fil%wp(1:current_wp)%subunit_num + &
                                                            real(current_subunit_num_shift)
                    if (concomitant_phi_change) then
                        tmp_fil%wp(1:current_wp)%euler%phi = refined_fil%wp(1:current_wp)%euler%phi + &
                                                                refined_fil%wp(1:current_wp)%helix%twist * &
                                                                real(current_subunit_num_shift)
                    endif
                endif
                ! Plug the modified filament into the fmc object
                call fmc_new_filament(self,tmp_fil)
                ! Maximise the score starting from the current subunit_num value
                if (maximise) then
                    current_score = fmc_eval(self,self%fil)
                    write(*,'(a,i0,a,i0,a,F0.4,a)') '**info(fmc_bf_subunit_num_intg_refine): on waypoint ', current_wp, &
                                                    ', subunit_num shift of ', current_subunit_num_shift, &
                                                    ' gave a score of ', current_score, ' (before ds maximisation)'
                    ! Remember which waypoints we are refining
                    !refine_wp_bak = self%fil%wp%refine
                    !self%fil%wp%refine = .FALSE.
                    !self%fil%wp(current_wp)%refine = .TRUE.
                    !self%fil_original%wp%refine = self%fil%wp%refine
                    call fmc_maximise(self) ! Could possibly check that subunit_num search half range is ~+-0.5 before starting the DS?
                    self%fil_original = self%fil
                    !self%fil%wp%refine = refine_wp_bak
                    !self%fil_original%wp%refine = self%fil%wp%refine
                endif
                ! Get score
                current_score = fmc_eval(self,self%fil)
                write(*,'(a,i0,a,i0,a,F0.4)') '**info(fmc_bf_subunit_num_intg_refine): on waypoint ', current_wp, &
                                                ', subunit_num shift of ', current_subunit_num_shift, &
                                                ' gave a score of ', current_score
                refined_fils(current_subunit_num_shift) = self%fil
                ! Is this the new best score?
                if (current_score .gt. best_score_for_current_wp) then
                    best_score_for_current_wp = current_score
                    best_subunit_num_shift_for_current_wp = current_subunit_num_shift
                endif
            enddo
            ! Now we know the best subunit_num shift for the current waypoint, apply it before we move on to the next waypoint
            write(*,'(a,i0,a,i0,a,F0.4)') '**info(fmc_bf_subunit_num_intg_refine): best shift for waypoint ', current_wp, &
                                            ' was ', best_subunit_num_shift_for_current_wp, ' with score of ', &
                                            best_score_for_current_wp
            refined_fil = refined_fils(best_subunit_num_shift_for_current_wp)
            ! Diagnostics
            write(*,'(a,i0,a)') '**info(fmc_bf_subunit_num_intg_refine): filament info after processing waypoint ', current_wp, ':'
            call refined_fil%PrintInfo()
            if (debug) then
                write(*,'(a,i0,a)') '**debug(fmc_bf_subunit_num_intg_refine): detailled scoring info after processing waypoint ',&
                                     current_wp, ':'
                self%fil = refined_fil
                self%fil_original = refined_fil
                current_score =  fmc_eval(self,self%fil,DETAILLED_SCORE_INFO=.true.)
            endif
        enddo
        enddo

        ! We're done

        ! We should update normal modes and residuals
        call filament_update_wp_res(refined_fil)
        call filament_wps_to_nm_amplitudes(refined_fil)

        ! Copy the results back into the FMC's filament
        call fmc_new_filament(self,refined_fil)

        ! Print out information
        write(*,'(a)') '**info(fmc_bf_subunit_num_intg_refine): filament state after bf subunit_num search: '
        call refined_fil%PrintInfo()
        write(*,'(a)') '**info(fmc_bf_subunit_num_intg_refine): all done'
    end subroutine fmc_bf_subunit_num_intg_refine


    !>  \brief  Maximise the correlation between filament and model, one waypoint at a time
    subroutine fmc_maximise_sequential(self)
        use fmcs_evaluation, only : fmc_eval
		use filaments_model_update, only : filament_update_model
!#ifdef __INTEL_COMPILER
!        use coos, only : point_assign
!#endif
		implicit none
        ! Arguments
        type(filament_model_comparison),    intent(inout)   ::  self    !<  Filament-model comparison object
        ! Private variables
        logical                 ::  refine_wp_bak(self%fil%nwp)
        integer                 ::  current_wp
        integer,    allocatable ::  wp_order(:)
        integer,    parameter   ::  num_iterations      =   1
        integer                 ::  current_iteration
        type(waypoint)          ::  first_wp
        integer                 ::  i
        real                    ::  dummy
        logical,    parameter   ::  redo_first_wp_at_the_end = .false.
        integer                 ::  number_of_waypoints_to_refine
        ! Start work

        ! Do we have anything to do
        if (.not. any(self%fil%wp%refine)) then
            write(*,'(a)') '**WARNING(fmc_maximise_sequential): nothing to do, since no way points marked for refinement'
            return
        endif

        ! Call scoring function on entry to be sure we know where we're starting from
        write(*,'(a,i0,a)') '**info(fmc_maximise_sequential): entering. Detailled score information for filament ', &
                            self%fil%id, ' follows'
        dummy = fmc_eval(self,self%fil,DETAILLED_SCORE_INFO=.true.)

        ! Perhaps we should skip the first and last waypoint
        if (.not. (any(self%refine_params(1:3)) .or. any(self%refine_params(5:6)))) then
            ! We are not refining anything to do with the space curve
            if (self%fil%fragmented_phi .and. self%fil%fragmented_subunit_num) then
                ! We are refining only things to do with the helix (phi, subunit_num), but the helix is fragmented
                if (.not. self%fil%fal_wps_in_score) then
                    self%fil%wp(1)%refine = .false.
                    self%fil%wp(self%fil%nwp)%refine = .false.
                endif
            endif
        endif


        ! Remember which waypoints we are refining
        refine_wp_bak = self%fil%wp%refine

        ! What order are we going to do waypoints?
        wp_order = fmc_get_wp_order(self)

        !
        if (redo_first_wp_at_the_end) then
            number_of_waypoints_to_refine = self%fil%nwp + 1
        else
            number_of_waypoints_to_refine = self%fil%nwp
        endif

        ! Loop over iterations
        do current_iteration = 1, num_iterations
            ! Remember the curent position of the first waypoint.
            first_wp = self%fil%wp(wp_order(1))

            ! Loop over waypoints
            do i=1,number_of_waypoints_to_refine
                current_wp = wp_order(i)
                if (self%fil%wp(current_wp)%refine) then
                    if (i .eq. self%fil%nwp+1) then
                        ! We're giving the first waypoint that was refined a second chance
                        write(*,'(a,i0,a)') '**info(fmc_maximise_sequential): giving the first waypoint which was refined (', &
                                            current_wp, ') a second chance'
                        self%fil%wp(current_wp) = first_wp
                        call filament_update_model(self%fil,reset_parametrisation=.true.)
                        self%fil_original = self%fil
                    else
                        write(*,'(2(A,i0))') '**info(fmc_maximise_sequential): going to be refining waypoint ', current_wp, &
                                            ' of filament ', self%fil%id
                    endif

                    ! All the other waypoints are switched off
                    self%fil%wp%refine = .false.
                    self%fil%wp(current_wp)%refine = .true.
                    self%fil_original%wp%refine = self%fil%wp%refine
                    !
                    write(*,'(a,i0,a)') '**info(fmc_maximise_sequential): before refining waypoint ', current_wp, &
                                            ' here is the detailled score: '
                    dummy = fmc_eval(self,self%fil,detailled_score_info=.true.)
                    ! Refine that waypoint
                    call fmc_maximise_simultaneous(self)
                    ! To make sure the next waypoint refinement starts with the previous waypoint in its minimised position, we copy over to original state of filament
                    self%fil_original = self%fil
                endif
                self%fil%wp%refine = refine_wp_bak
                self%fil_original%wp%refine = self%fil%wp%refine
            enddo
        enddo
        write(*,'(a)') '**info(fmc_maximise_sequential): all waypoints have been refined'
    end subroutine fmc_maximise_sequential

    !>  \brief  Use a downhill simplex to maximise the correlation between filament and model. All waypoints are refined simultaneously.
    !!  \todo   When fragment_helix and only refining phi and/or subunit_num, should switch refine flags to false for first and last waypoint, as is done in fmc_ds_maximise_sequential
    subroutine fmc_ds_maximise(self)
        ! Modules
        use minimisers, only : ds_init, ds_run, minimiser_ds_fgsl, ds_free, ds_print_progress_summary, ds_set_start_step
        use fmcs_evaluation, only : fmc_eval, fmc_eval_wrap, fmc_model3d_prep
		use filaments_model_modify, only : filament_add_disorder
		use images_comparison, only : score_got_worse
		implicit none
        ! Arguments
        type(filament_model_comparison),            intent(inout)   ::  self            !<  Filament model comparison (fmc) object
        ! Private variables
        type(minimiser_ds_fgsl)     ::  ds_minimiser    !<  a minimiser object (we use a fgsl minimiser - this is all we have as of Jan2010)
        real(kind=8),   allocatable ::  first_step(:), start_values(:), values_on_entry(:)
        real(kind=8),   allocatable ::  final_values(:,:), final_score(:)
        real(kind=8),   allocatable ::  best_final_values(:)
        integer                     ::  num_dim_total
        integer                     ::  ierr
        real                        ::  ds_target_size                  !<  Target size of the simplex. Used 0.05 for Ab(40) to 7a and Ab(42) to 6A.
        real                        ::  score_on_entry, score_on_exit
        real                        ::  score_at_run_start, score_at_run_finish
        logical,    parameter       ::  sanity_checks = .true.
        logical,    parameter       ::  debug   =   .false.
        integer                     ::  current_minim_run
        integer                     ::  best_minim_run(1)
        type(filament)              ::  fil_on_entry
        integer, parameter          ::  num_sequential_runs  = 4
        integer                     ::  current_sequential_run
        character(len=6)            ::  num_dim_c
        real                        ::  dummy
        integer                     ::  num_random_conf
        ! Start work

        if (.not. any(self%fil%wp%refine)) then
            write(*,'(a)') '**WARNING(fmc_ds_maximise): nothing to do, since no way points marked for refinement'
            return
        endif

        ! How many random starting conformations?
        num_random_conf = 1
        if (self%ds_random_start_conf .and. .not. self%minim_random_start_pos) then
            num_random_conf = self%num_minim_runs
        endif
        if (self%ds_random_start_conf .and. self%minim_random_start_pos) then
            write(*,'(a)') '**WARNING(fmc_ds_maximise): minim_random_start_pos is overriding DS_RANDOM_START_CONF'
        endif

        ! Filament on entry is backed-up
        fil_on_entry = self%fil
        call fmc_start_step_values(self,values_on_entry,first_step,values_to_fil=.false.)
        print '(a)', '**info(fmc_ds_maximise): filament info on entry:'
        call self%fil%PrintInfo()

        ! Make sure the model 3d is ready
        call fmc_model3d_prep(self)

        ! What's the score on entry?
        score_on_entry = fmc_eval(self,self%fil,detailled_score_info=.false.)
        write(*,'(a,F0.5)') '**info(fmc_ds_maximise): score on entry: ', score_on_entry

        ! Allocate memory for the results of the minimiser runs
        num_dim_total  = fmc_num_dim_total(self) ! number of dimensions the simplex is going to deal with
		write(num_dim_c,'(i6)') num_dim_total
		allocate(   final_values(num_dim_total,num_random_conf),   &
		            final_score(num_random_conf),   &
		            best_final_values(num_dim_total),   &
		            stat=ierr )
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(fmc_ds_maximise): memory allocation failed with error ', ierr
            call terminate('Fatal error in FMC_DS_MAXIMISE')
        endif

        do current_minim_run = 1, num_random_conf

            if (current_minim_run .eq. 1) then
                ! Initialise the downhill simplex (ds) minimiser
                ds_target_size = fmc_ds_target_size(self) ! target size of the simplex
                call ds_init(   ds_minimiser,fmc_eval_wrap,num_dim_total,self%cptr,     &
                                initial_score=score_on_entry,                           &
                                random=self%ds_random_start_conf,                       &
                                max_num_iter=self%ds_max_num_iter                       &
                                )
            endif


            ! Work out the starting position & configuration of the simplex
            call fmc_start_step_values(self,start_values,first_step,values_to_fil=.false.)
            if (debug) then
                write(*,'(a,'//num_dim_c//'(F0.3,1x))') '**debug(fmc_ds_maximise): start values: ', start_values
                write(*,'(a,'//num_dim_c//'(F0.3,1x))') '**debug(fmc_ds_maximise): step values: ', first_step
            endif


            ! Remember the score at start of run
            score_at_run_start = fmc_eval(self,self%fil,detailled_score_info=.false.)
            if (debug) write(*,'(a,F0.4)') '**debug(fmc_ds_maximise): score at run start = ', score_at_run_start

            ! Run the ds
            print '(a)', '**info(fmc_ds_maximise): will now optimise waypoint parameters'
            call ds_set_start_step(ds_minimiser,start_values,first_step) ! Set the starting position and the first step for the ds
                                                                         ! (the starting configuration is randomised if that was specified to ds_init
            call ds_run(ds_minimiser,ds_target_size)
            print '(a)', '**info(fmc_ds_maximise): optimisation done.'

            ! Start a second run from the end of the first
            if (num_sequential_runs .gt. 1) then
                do current_sequential_run=2,num_sequential_runs
                    print '(2(A,i0))', '**info(fmc_ds_maximise): starting the optimisation #', current_sequential_run, &
                                        ' from the end optimisation #', current_sequential_run-1
                    call ds_set_start_step(ds_minimiser,ds_minimiser%current_best_fgsl,first_step)
                    call ds_run(ds_minimiser,ds_target_size)
                    print '(a,i0,a)', '**info(fmc_ds_maximise): optimisation #', current_sequential_run, ' done.'
                enddo
            endif

            ! Remember the score at end of run
            score_at_run_finish = - ds_minimiser%current_minimum_fgsl

            ! debug
            if (debug) then
                print '(a,i0,a)', '**debug(fmc_ds_maximise): detailed score of filament after maxim run ',current_minim_run ,':'
                dummy = fmc_eval(self,self%fil,detailled_score_info=.true.)
            endif

            if (sanity_checks) then
				if (score_got_worse(score_at_run_start,score_at_run_finish)) then
                    write(*,'(2(A,i0),a,2(F0.5,1x))') '**ERROR(fmc_ds_maximise): run ', &
                                                    current_minim_run,  ' of ', self%num_minim_runs, &
                                                    ' had higher score at start than at finish: ', &
                                                    score_at_run_start, score_at_run_finish
                    call terminate('Fatal error in FMC_DS_MAXIMISE')
                endif
            endif

            ! Get the last values from the minimiser
            final_values(:,current_minim_run) = ds_minimiser%current_best_fgsl(:)
            final_score(current_minim_run) = - ds_minimiser%current_minimum_fgsl

            if (debug) then
                print '(a,i0,a,i0,a,'//num_dim_c//'(F0.5,1x))', '**debug(fmc_ds_maximise): final score for run ', &
                                                                current_minim_run, ' of ' , self%num_minim_runs, ' = ', &
                                                                final_score(current_minim_run)
                print '(a,i0,a,i0,a,'//num_dim_c//'(F0.5,1x))', '**debug(fmc_ds_maximise): final values for run ', &
                                                                current_minim_run, ' of ' , self%num_minim_runs, ' = ', &
                                                                final_values(:,current_minim_run)
            endif

            ! Print out information on progress of minimisation
            write(*,'(a,i0,a,i0)') '**info(fmc_ds_maximise): Summary of minimisation run ', &
                                    current_minim_run, ' of ' , self%num_minim_runs
            call ds_print_progress_summary(ds_minimiser)
            write(*,'(a,i0)') '**info(fmc_ds_maximise): number of calls to the scoring function: ', self%calls_from_minimiser
            write(*,'(a)') '**info(fmc_ds_maximise): end of minimisation summary'
        enddo ! End of loop over runs of the ds


        ! How many of the runs gave improvements on the score?
        if (num_random_conf .gt. 1) then
            write(*,'(a,2(i0,a))') '**info(fmc_ds_maximise): ', count(final_score .gt. score_on_entry), ' of ', &
                                            self%num_minim_runs, ' ds runs gave improved score'
        endif

        ! Decide which run of the minimiser gave the best results
        best_minim_run = maxloc(final_score)
        if (num_random_conf .gt. 1) then
            write(*,'(a,i0,a,F0.5)') '**info(fmc_ds_maximise): maximisation run ', best_minim_run(1), &
                                    ' gave the best results, with score: ', &
                                     final_score(best_minim_run)
        endif
        best_final_values = final_values(:,best_minim_run(1))

        ! The best final score should be better than the starting score, except if we strarted from randomised
        ! positions, in which case we may be worse
		if (score_got_worse(score_on_entry,real(final_score(best_minim_run(1))))) then
                write(*,'(a,2(F0.5,1x))') '**ERROR(fmc_ds_maximise): best final score less than starting score: ', &
                                        final_score(best_minim_run), score_on_entry
                call terminate('Fatal error in FMC_DS_MAXIMISE')
        endif

        ! Restore filament to original state
        self%fil = fil_on_entry
        ! Plug those last values back into the filament     (the ds routine does not necessarily finish on the minimum)
        call fmc_start_step_values(self,best_final_values,first_step,values_to_fil=.true.) ! Note that in this call, nothing will happen to first_step

        ! We run the scoring routine one last time, because this calls filament_update_model with the proper parameters, so that the best parameters are propagated through the filament model
        score_on_exit = fmc_eval(self,self%fil,detailled_score_info=.false.)

        if (sanity_checks) then
            if (score_got_worse(score_on_entry,score_on_exit) &!.and. number_of_score_wps_on_entry .eq. number_of_score_wps_on_exit &
		    .and. score_on_entry .gt. 0.0 .and. score_on_exit .gt. 0.0) then
                write(*,'(a,2(F0.7,1x))') '**ERROR(fmc_ds_maximise): score on exit less than on entry: ', &
                                            score_on_exit, score_on_entry
                call terminate('Fatal error in FMC_DS_MAXIMISE')
            endif
        endif

        print '(a)', '**info(fmc_ds_maximise): filament info on exit:'
        call self%fil%PrintInfo()

        write(*,'(a,F0.5)') '**info(fmc_ds_maximise): score on exit: ', score_on_exit

        ! Now we can free the minimiser
        call ds_free(ds_minimiser)

        ! Print out some diagnostics regarding projection caching
        !if (self%cache_projections) then
        !    call projection_cache_print_info(self%proj_cache)
        !endif
    end subroutine fmc_ds_maximise



    !>  \brief  Compute the target size for the downhill simplex
    !!
    !!  Note that the size of the simplex is defined as "the rms distance of each vertex from the center", when using
    !!  the gsl algorithm gsl_multimin_fminimizer_nmsimplex2. Otherwise, it's the mean distance.
    real function fmc_ds_target_size(self) result (ds_target_size)
        use minimisers, only : search_accuracy_to_ds_size_ratio
        implicit none
        ! Arguments
        type(filament_model_comparison),    intent(in)  ::  self
        ! Private variables
        integer ::  num_wp, num_dim_total
        integer ::  i, num_dim_for_current_param
        logical,    parameter   ::  use_rms = .true.
        real    ::  x
        ! Start work

        num_wp          =   count(self%fil%wp%refine)
        num_dim_total   =   fmc_num_dim_total(self)

        ! Loop over the possible refinement parameters (1 to 7 for x to subunit_num)
        ds_target_size = 0.0
        do I=1,7
            if (self%refine_params(i)) then
                if (self%fil%rigid_params(i)) then
                    num_dim_for_current_param = 1
                else
                    num_dim_for_current_param = num_wp
                    if (self%fil%refine_normal_modes(i)) then
                        if (self%hold_param_average(i)) num_dim_for_current_param = num_dim_for_current_param - 1
                    endif
                endif
                !
                x = real(num_dim_for_current_param) * (self%search_accuracy(i)/self%param_scale(i))  &
                                                      / search_accuracy_to_ds_size_ratio &
                                                      * self%nm_param_scale(i)
                if (use_rms) then
                    ds_target_size = ds_target_size + X**2
                else
                    ds_target_size = ds_target_size + x
                endif
            endif
        enddo
        ds_target_size = ds_target_size / num_dim_total
        if (use_rms) ds_target_size = sqrt(ds_target_size)
        ! Diagnostics
        write(*,'(a,F0.4,a,i0,a)') '**info(fmc_ds_target_size): target size for simplex = ', ds_target_size, &
                                    ' (total of ', num_dim_total, ' search dimensions)'
    end function fmc_ds_target_size


    !>  \brief  maximise the fmc scoring function using conjugate gradient
    subroutine fmc_cg_maximise(self,bf_instead)
        use minimisers, only : minimiser_bf, bf_init, bf_run
		use fmcs_evaluation, only : fmc_eval_wrap, fmc_eval, fmc_model3d_prep
        use images_comparison, only : score_got_worse
        use images
        use omp_lib
        use Globals
		implicit none
        ! arguments
        type(filament_model_comparison),    target, intent(inout)   ::  self
        logical,                        optional,   intent(in)      ::  bf_instead !< do a brute force search
        ! private variables
        logical,        parameter   ::  debug   =   .false.
        type(ConjugateGradient)     ::  minimizer_cg
        type(minimiser_bf)          ::  minimizer_bf
        integer                     ::  num_dim_total
        real(kind=8),   allocatable ::  first_step(:), final_values(:)
        real(kind=8),   allocatable, target :: start_values(:)
        real(kind=8),   allocatable ::  search_half_range(:), search_step_size(:)
        real                        ::  score_on_entry, best_score
        logical,        allocatable ::  integer_changes_only(:)
        type(filament)              ::  fil_on_entry, fil_on_exit
        logical                     ::  bbf_instead
        logical                     ::  spawn_threads
        type(filament_model_comparison),    target, allocatable     ::  self_copied(:)
        type(c_ptr),                                allocatable     ::  param_cptrs(:)
        integer                                                     ::  current_thread
        integer                     ::  number_of_score_wps_on_entry
        integer                     ::  number_of_score_wps_on_exit
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(fmc_cg_maximise): filament info on entry'
            call self%fil%PrintInfo()
        endif

        bbf_instead = .false.
        if (present(bf_instead)) bbf_instead = bf_instead

        ! Work out the number of dimensions the minimiser is going to deal with
        num_dim_total = fmc_num_dim_total(self)

        ! Debug
        if (debug) then
            write(*,'(a)') '**debug(fmc_cg_maximise): detailled score on entry, before model3d_prep: '
            score_on_entry = fmc_eval(self,self%fil,detailled_score_info=debug)
            if (.not. self%model3dvol_private%is_in_real_space) call self%model3dvol_private%BackwardFFT()
            call self%model3dvol_private%WriteToDisk('dbg_cg_model3d1.mrc')
            call self%model3dvol_private%ForwardFFT()
        endif

        ! Make sure the model 3d is ready
        call fmc_model3d_prep(self)

        ! Score on entry
        if (debug) write(*,'(a)') '**debug(fmc_cg_maximise): detailled score on entry: '
        fil_on_entry = self%fil
        call fmc_new_filament(self,fil_on_entry)
        call fmc_start_step_values(self,start_values,first_step,values_to_fil=.false.,dim_is_subunit_num=integer_changes_only, &
                                        search_half_range=search_half_range,search_step_size=search_step_size)
        score_on_entry = fmc_eval(self,self%fil,detailled_score_info=debug)
        number_of_score_wps_on_entry = size(self%fil%wp_score)


        if (debug) then
            write(*,'(a)',advance='NO') '**debug(fmc_cg_maximise): starting values: '
            print *, start_values
            if (.not. self%model3dvol_private%is_in_real_space) call self%model3dvol_private%BackwardFFT()
            call self%model3dvol_private%WriteToDisk('dbg_cg_model3d2.mrc')
            call self%model3dvol_private%ForwardFFT()
        endif


        ! Minimizer-specific code from here
        ! start_values contains starting values for refinement (real)
        ! first_step can be used as a measure for the step size to use (real)
        ! num_dim_total is the total number of parameters to refine (integer)
        !
        if (bbf_instead) then
            ! Will we be doing multi-threaded brute force search?
            spawn_threads = .false.
#ifdef _OPENMP
            if (.not. omp_in_parallel()) then
                if (omp_get_max_threads() .gt. 1) spawn_threads = .true.
            endif
#else
            spawn_threads = .false.
#endif
            ! if we'll be threading we need several copies of the fmc object
            if (spawn_threads) then
#ifdef _OPENMP
                allocate(self_copied(omp_get_max_threads()),param_cptrs(omp_get_max_threads()))
#else
                allocate(self_copied(1),param_cptrs(1))
#endif
#ifdef _OPENMP
                do current_thread=1,omp_get_max_threads()
#else
                do current_thread=1,1
#endif
                    call fmc_assignment(self_copied(current_thread),self)
                    param_cptrs(current_thread) = self_copied(current_thread)%cptr
                enddo
            else
                allocate(self_copied(1),param_cptrs(1))
                call fmc_assignment(self_copied(1),self)
                param_cptrs(1) = self%cptr
            endif
            !
            call bf_init(   minimizer_bf, &
                            function_to_minimise=fmc_eval_wrap, &
                            parameters=param_cptrs, &
                            num_dim=num_dim_total, &
                            starting_value=start_values, &
                            half_range=search_half_range, &
                            step_size=search_step_size &

                        )
            call bf_run(minimizer_bf,score_on_entry)
            final_values = minimizer_bf%best_value
            if (allocated(self_copied)) then
                do current_thread=1,size(self_copied)
                    call fmc_deallocate(self_copied(current_thread))
                enddo
                deallocate(self_copied)
            endif
        else
            call minimizer_cg%Init(    num_dim=num_dim_total,              &
                                    function_to_minimize=fmc_eval_wrap, &
                                    parameters=self%cptr,               &
                                    starting_values=start_values,       &
                                    accuracy=first_step, &
                                    print_progress=this_program%IsVerbose() )
            call minimizer_cg%Run()

            if (debug) then
                write(*,*) '**debug(fmc_cg_maximise): final values = ', final_values
                write(*,*) '**debug(fmc_cg_maximise): best values  = ', minimizer_cg%GetBestValues()
            endif

            ! Record the final values and apply them to the filament
            final_values = minimizer_cg%GetBestValues()
        endif

        self%fil = self%fil_original
        call fmc_start_step_values(self,final_values,first_step,values_to_fil=.true.)
        fil_on_exit = self%fil

        !call fmc_new_filament(self,fil_on_exit,REPLACE_ORIGINAL=.FALSE.)


        ! Print out best solution in detail
        print '(a)', '**info(fmc_cg_maximise): detailed score of best-scoring filament model:'
        best_score = fmc_eval(self,self%fil,detailled_score_info=.true.)
        number_of_score_wps_on_exit = size(self%fil%wp_score)

        if (debug) then
            write(*,'(a)') '**debug(fmc_cg_maximise): filament info on exit'
            call self%fil%PrintInfo()
        endif

        ! Sanity check
		if (score_got_worse(score_on_entry,best_score) .and. number_of_score_wps_on_entry .eq. number_of_score_wps_on_exit &
		    .and. score_on_entry .gt. 0.0 .and. best_score .gt. 0.0) then
            write(*,'(a,2(F0.4,1x))') '**ERROR(fmc_cg_maximise): score on entry > on exit: ', score_on_entry, best_score
            call terminate('Failed sanity check in FMC_CG_MAXIMISE')
        endif

		if (debug) write(*,'(a)') '**debug(fmc_cg_maximise): exiting'
    end subroutine fmc_cg_maximise


    !>  \brief  Maximise the fmc scoring function using differential evolution
    subroutine fmc_de_maximise(self)
        use differential_evolution_mod
		use fmcs_evaluation, only : fmc_eval, fmc_eval_wrap, fmc_model3d_prep
        implicit none
        ! Arguments
        type(filament_model_comparison),    intent(inout)   ::  self
        ! Private variables
        logical,    parameter   ::  debug   =   .true.
        type(de_minimiser)      ::  minimiser
        integer                 ::  num_dim_total
        real(kind=8),   allocatable ::  first_step(:), start_values(:), best_values(:)
        real                    ::  dummy
        logical,    allocatable ::  integer_changes_only(:)
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(fmc_de_maximise): filament info on entry'
            call self%fil%PrintInfo()
        endif

        ! Work out the number of dimensions the simplex is going to deal with
        num_dim_total = fmc_num_dim_total(self)

        ! Make sure the model 3d is ready
        call fmc_model3d_prep(self)

        ! Work out the start values and the first step
        call fmc_start_step_values(self,start_values,first_step,VALUES_TO_FIL=.false.,DIM_IS_SUBUNIT_NUM=INTEGER_CHANGES_ONLY)

        !
        if (.not. self%integer_changes_only(7)) integer_changes_only = .false.

        if (debug) write(*,'(a,7(l1,1x))') '**debug(fmc_de_maximise): integer_changes_only = ', integer_changes_only

        ! Initialise the de minimiser
        call de_init(minimiser, NUM_DIM=NUM_DIM_TOTAL,   &
                                FUNCTION_TO_MINIMISE=FMC_EVAL_WRAP,   &
                                PARAMETERS=SELF%cptr,    &
                                STARTING_VALUES=START_VALUES,   &
                                SEARCH_HALF_RANGE=FIRST_STEP,  &
                                EVOLVE_F=.true.,                &
                                EVOLVE_CR=.true.,            &
                                INTEGER_CHANGES_ONLY=INTEGER_CHANGES_ONLY &
                                )

        ! Run the de minimiser
        call de_run(minimiser)

        ! Plug best values back into filament
        call de_best_values(minimiser, best_values)
        self%fil = self%fil_original
        call fmc_start_step_values(self,best_values,first_step,VALUES_TO_FIL=.true.)

        ! Print out best solution in detail
        print '(a)', '**info(fmc_ds_maximise): detailed score of best-scoring filament model:'
        dummy = fmc_eval(self,self%fil,DETAILLED_SCORE_INFO=.true.)

        if (debug) then
            write(*,'(a)') '**debug(fmc_de_maximise): filament info on exit'
            call self%fil%PrintInfo()
        endif
    end subroutine fmc_de_maximise


    !>  \brief  Align way points by calculating cross-correlation functions
    !!  \todo   Add filtering step according to rad_max in fmc object
    !!  \todo   important: add step to measure height of peaks in real space to make comparable!
!    subroutine wp_ccf_align(fmc)
!        ! Modules
!        use images_ctf
!        use images_comparison
!        use images_threed
!		use Micrographs
!		use filaments_geometry, only : filament_wp_distance_to_nearest_wp
!		use fmcs_evaluation, only : fmc_model3d_prep
!		use eulerangles
!		use omp_lib
!        implicit none
!        ! Arguments
!        type(filament_model_comparison),    intent(inout)   ::  fmc     !<  Filament model comparison (fmc) object, must be already initialised
!        ! Private variables
!        integer                 ::  i
!        integer                 ::  icoo(3)
!        real                    ::  rcoo(3)
!        type(image)             ::  box, box_ctf,mask, proj
!        real                    ::  matrix(4,4)
!        type(ctf_parameters)    ::  ctf
!        type(euler)             ::  angles, angles_search_half_range, angles_search_sampling
!        real                    ::  shift(3)
!        integer                 ::  first_wp, last_wp
!        real                    ::  max_rad_shift
!        ! Start work
!
!        write(*,'(a)') '**ERROR(wp_ccf_align): deprecated. Use wps_as_sps with brute-force search instead'
!        call terminate('Deprecated routine')
!
!#if _OPENMP
!        if (omp_in_parallel()) call terminate('OpenMP not supported by WP_CCF_ALIGN')
!#endif
!
!        if (.not. any(fmc%fil%wp%refine)) then
!            write(*,'(a)') '**WARNING(wp_ccf_align): nothing to do, since no way points marked for refinement'
!            return
!        endif
!
!        ! Make sure the model 3d is ready
!        call fmc_model3d_prep(fmc)
!
!
!        ! Allocate memory for the box
!        box%logical_dimensions(1:2) = fmc%box_size
!        box%logical_dimensions(3) = 1
!        call box%Allocate()
!        call box_ctf%Allocate(box)
!
!        ! Prepare a 2d soft mask (we use quite a small mask because the multiplication by the ctf -> delocalisation)
!        call mask%Allocate(box)
!        call mask%ApplySoftCircularMask(fmc%box_size/2.0*.7,fmc%box_size/2.0*.1)
!
!        ! Initialise a rotation/shift matrix
!        matrix = 0.0
!        do I=1,4
!            matrix(i,i) = 1.0
!        enddo
!
!        ! Work out first and last waypoints to be aligned
!        if (.not. fmc%fil%fal_wps_in_score) then
!            first_wp = 2
!            last_wp = fmc%fil%NWP-1
!        else
!            first_wp = 1
!            last_wp = fmc%fil%nwp
!        endif
!
!        ! Loop over the way points
!        do i=first_wp,last_wp
!            !print '(a,i2,a,i2)', '**debug(wp_ccf_align): on way point ', i, ' of ', fmc%fil%nwp
!            ! We only tweak the way points that are set to be refined
!            if (fmc%fil%wp(i)%refine) then
!                ! Get integer coordinates & remainders
!                icoo = nint(fmc%fil%wp(i)%coo)
!                rcoo = fmc%fil%wp(i)%coo - icoo
!                ! The box needs to be real and centered at origin for the windowing
!                box%is_in_real_space = .true.
!                box%object_is_centered_in_box = .true.
!                box%real_values = 0.0
!
!                ! Window image (maybe this should be windowed bigger initially to recover information spread by psf/CTF? (this should be ok if using small-ish mask of, e.g. 0.7)
!                !print *, '**debug(wp_ccf_align): windowing at coos ', icoo
!                call fmc%film_image%ClipInto(box,icoo)
!
!                ! Shift image to center on the non-integer coordinates defined by the path
!                matrix(1,4) = rcoo(1)
!                matrix(2,4) = rcoo(2)
!                !print *, '**debug(wp_ccf_align): non-integer shift'
!                call box%ApplyAffineTransformation(box_ctf,matrix)  ! We use box_ctf as a temporary image
!                box%real_values = box_ctf%real_values
!                box_ctf%real_values = 0.0
!
!                ! Apply ctf
!                !print *, '**debug(wp_ccf_align): applying CTF'
!                call box%ForwardFFT()
!                ctf = fmc%fil%film%ReturnCTFAtCoos(fmc%fil%wp(i)%coo(1),fmc%fil%wp(i)%coo(2))
!                call ctf_to_img(ctf,box_ctf,fmc%fil%film%GetPixelSize())
!                box%complex_values = box%complex_values * box_ctf%real_values
!
!                ! Get matching projection
!                !print '(a,3(F8.3,x))', '**debug(wp_ccf_align): calculating projection with angles', fmc%fil%wp(i)%euler
!                !call fmc%model3dvol%WriteToDisk('dbg_3d.mrc')
!                call volume_project_lean(   fmc%model3dvol_private,proj,    &
!                                            fmc%fil%wp(i)%euler%phi,fmc%fil%wp(i)%euler%theta,fmc%fil%wp(i)%euler%psi,  &
!                                            fmc%interpol,fmc%interp_win_3d(1) )
!                !print *, '**debug(wp_ccf_align): finished projection'
!                !call proj%WriteToDisk('dbg_proj.mrc')
!                !stop 'Testing volume_project_lean'
!
!                ! Impose CTF^2 on projection (since the image also had it twice - once in the microscope, and once in correction)
!                call proj%ForwardFFT()
!                proj%complex_values = proj%complex_values * box_ctf%real_values**2
!
!                ! Here, add filter of both according to rad_max found in the fmc object
!                call proj%ApplyBandPassFilter(0.0,fmc%freq_max,0.025)
!                !call box%ApplyBandPassFilter(0.0,fmc%freq_max,0.025)
!
!                ! Back to real space
!                call proj%BackwardFFT()
!                call box%BackwardFFT()
!
!                ! Debug dump
!                !print *, '**debug(wp_ccf_align): dumping proj & box to disk'
!                !call write_to_disk(box,'dbg_ccf_box',LOC=I,STACK=FMC%fil%nwp)
!                !call write_to_disk(proj,'dbg_ccf_proj',LOC=I,STACK=FMC%fil%nwp)
!
!                ! Compute the maximum allowable radial shift during alignment. It will be the distance to the nearest waypoint or 1.5*the helical radius, which ever is smaller
!                max_rad_shift = min(filament_wp_distance_to_nearest_wp(fmc%fil,i)-4.0,fmc%fil%MAX_RADIUS*1.5)
!                if (max_rad_shift .lt. 0.5) max_rad_shift = 0.5
!                print '(a,i3,a,i4,a,F12.3)', '**debug(wp_ccf_align): waypoint ', i, ' of filament ', fmc%fil%id, &
!                                            '- maximum allowable shift = ', max_rad_shift
!
!                ! Align using ccf - the proj is aligned onto the box. Rotation first.
!                angles_search_half_range%phi   = fmc%search_half_range(4)
!                angles_search_half_range%theta = fmc%search_half_range(5)
!                angles_search_half_range%psi   = fmc%search_half_range(6)
!                angles_search_sampling%phi   = fmc%search_sampling(4)
!                angles_search_sampling%theta = fmc%search_sampling(5)
!                angles_search_sampling%psi   = fmc%search_sampling(6)
!                !print *, '**debug(wp_ccf_align): calling FIND_SHIFT_EULER'
!                call find_shift_euler(  proj,box,shift,angles,  &
!                                        ROTATION_SAMPLING=ANGLES_SEARCH_SAMPLING,   &
!                                        ROTATION_HALF_RANGE=ANGLES_SEARCH_HALF_RANGE,   &
!                                        MAX_RAD_SHIFT=MAX_RAD_SHIFT)
!
!                ! Diagnostics
!                print '(a,i2,a,i5,a,2(F8.2,1x),a,F6.2,a,F8.2)', '**info(wp_ccf_align): way point ', i, ' of filament ', &
!                                                                fmc%fil%id, '; shift = ', shift(1:2),   &
!                                                                '; rot = ', angles%psi, '; max. shift = ', max_rad_shift
!
!                ! Apply shift rot
!                fmc%fil%wp(i)%euler%psi = fmc%fil%wp(i)%euler%psi + angles%psi
!                fmc%fil%wp(i)%coo = fmc%fil%wp(i)%coo + shift
!            endif
!        enddo
!
!    end subroutine wp_ccf_align

    !>  \brief  Improve the score by tilting the whole filament as a rigid body. Search over the tilting angle range using brute force.
    subroutine fmc_bf_tilt(fmc,tilt_half_range,tilt_sampling,plot_filename_root)
        use plots
        use parsers
        use fmcs_evaluation, only : fmc_eval
		use filaments_model_modify, only : filament_apply_tilt
		implicit none
        ! Arguments
        type(filament_model_comparison),    intent(inout)   ::  fmc                 !<  Filament-model comparison object
        real,                   optional,   intent(in)      ::  tilt_half_range     !<  Half-range of the search over tilt angle
        real,                   optional,   intent(in)      ::  tilt_sampling       !<  The step-size of the brute force search for the best tilt angle
        character(len=*),       optional,   intent(in)      ::  plot_filename_root  !<  Root for filename for plot of score v oop tilt. Should include filament number and correct relative path already.
        ! Private variables
        real                            ::  ttilt_half_range, ttilt_sampling
        type(filament)                  ::  orig_fil
        logical,            parameter   ::  debug   =   .true.
        real                            ::  current_tilt, current_score
        real                            ::  best_tilt, best_tilt_smooth
        real                            ::  best_score, best_score_smooth
        real,               allocatable ::  tilt_angle(:), score(:), score_smooth(:)
        type(filament),     allocatable ::  tilted_filament(:)
        type(filament)                  ::  best_filament
        type(filament)                  ::  tmp_fil
        integer                         ::  ierr, i, num_tilt_angles
        integer                         ::  j_min, j_max
        integer,            allocatable ::  locs_best_score(:), locs_best_score_smooth(:)
        logical,            parameter   ::  use_smooth_score = .true.
        integer                         ::  which_maximum
        real                            ::  theta_half_range_bak
        type(filament_model_comparison) ::  fmc_private
        ! Private variables for plotting
        logical,            parameter   ::  wwrite_plots_to_disk = .true.
        character(len=FILENAME_LEN)     ::  plot_filename
        character(len=128)              ::  plot_title
        real,               allocatable ::  scores(:,:)
        logical,            parameter   ::  maximise_for_each_tilt = .true.
        real                            ::  starting_score
        ! Start work

        ! If we're not given a tilt range, use the search half-range for Theta
        if (present(tilt_half_range)) then
            ttilt_half_range = tilt_half_range
        else
            ttilt_half_range = fmc%search_half_range(5)
        endif
        if (present(tilt_sampling)) then
            ttilt_sampling = tilt_sampling
        else
            ttilt_sampling = max(fmc%search_sampling(5),2.0)
        endif

        if (ttilt_sampling .eq. 0.0) then
            ttilt_sampling = max(ttilt_half_range / 10.,2.0)
        endif

        ttilt_sampling = max(ttilt_sampling,1.0)
        if (debug) write (*,'(2a,2(F0.3,1x))') '**info(fmc_bf_tilt): will search out-of-plane tilt with half-range', &
                                                ' & sampling for search = ', ttilt_half_range, ttilt_sampling
        if (debug) then
            if (maximise_for_each_tilt) then
                write(*,'(a)') '**debug(fmc_bf_tilt): will do minisation for each tilt in the search'
            else
                write(*,'(a)') '**debug(fmc_bf_tilt): will not do minisation for each tilt in the search'
            endif
        endif

        if (ttilt_sampling .eq. 0. .or. ttilt_half_range .eq. 0.) then
            write(*,'(a)') '**ERROR(fmc_bf_tilt): tilt search sampling or half-range is 0.0'
            call terminate('Fatal error in FMC_BF_TILT')
        endif

        ! "Backup" the original state of the filament
        orig_fil = fmc%fil

        ! Initialise values
        best_tilt       =   0.0
        current_tilt    =   0.0
        starting_score  =   fmc_eval(fmc,fmc%fil,DETAILLED_SCORE_INFO=DEBUG)
        current_score   =   starting_score
        best_score      =   -1.0    !current_score  ! We do not leave the current score in the race, because it has an unfair advantage (it's been maximised by the ds)

        print '(a,F0.4)', '**info(fmc_bf_tilt): starting score = ', starting_score


        ! Work out how many tilt angles we will try
        current_tilt = ttilt_half_range
        num_tilt_angles = 0
        do
            if (current_tilt .gt. ttilt_half_range) current_tilt = ttilt_half_range
            if (current_tilt .lt. -TTILT_HALF_RANGE) current_tilt = -TTILT_HALF_RANGE
            !
            num_tilt_angles = num_tilt_angles + 1
            ! Increment tilt
            current_tilt = current_tilt - ttilt_sampling
            ! Check whether we're done. If so, get out of the loop
            if (current_tilt .gt. ttilt_half_range .or. current_tilt .lt. -TTILT_HALF_RANGE) exit
        enddo

        ! Allocate memory for result arrays
        allocate(   tilt_angle(num_tilt_angles), score(num_tilt_angles), score_smooth(num_tilt_angles), &
                    scores(num_tilt_angles,2),tilted_filament(num_tilt_angles),stat=IERR)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(fmc_bf_tilt): memory allocation for output arrays failed with error ', ierr
            call terminate('Fatal error in FMC_BF_TILT')
        endif

        ! Set up the array of tilt angles
        current_tilt = ttilt_half_range
        i = 0
        do
            if (current_tilt .gt. ttilt_half_range) current_tilt = ttilt_half_range
            if (current_tilt .lt. -TTILT_HALF_RANGE) current_tilt = -TTILT_HALF_RANGE
            i = i + 1
            tilt_angle(i) = current_tilt
            current_tilt = current_tilt - ttilt_sampling
            if (current_tilt .gt. ttilt_half_range .or. current_tilt .lt. -TTILT_HALF_RANGE) exit
        enddo

!        !$omp   parallel    default(shared)                                               &
!        !$omp               private(i,current_tilt,tmp_fil,fmc_private,current_score)   &
!        !$omp               private(theta_half_range_bak)
!		call fmc_assignment(fmc_private,fmc)
        !
        ! Main loop over the search range
        !
!        !$omp do
        do I=1,size(tilt_angle)
            current_tilt = tilt_angle(i)
            ! Apply tilt
            tmp_fil = orig_fil
            call filament_apply_tilt(tmp_fil,current_tilt,WPS_AS_SPS=FMC_PRIVATE%wps_as_sps)
            call fmc_new_filament(fmc_private,tmp_fil)
            ! Find local maximum
            if (maximise_for_each_tilt) then
                theta_half_range_bak = fmc_private%search_half_range(5)
                fmc_private%search_half_range(5) = ttilt_sampling
                call fmc_maximise(fmc_private)
                fmc_private%search_half_range(5) = theta_half_range_bak
            else
                !$omp   critical        (fmc_bf_tilt_print_fil_info)
                write(*,'(a,F0.4)') '**info(fmc_bf_tilt): filament info at tilt ', current_tilt
                call fmc_private%fil%PrintInfo()
                !$omp   end critical    (fmc_bf_tilt_print_fil_info)
            endif
            ! Remember current state of tilted filament
            tilted_filament(i) = fmc_private%fil
            ! Score
            current_score = fmc_eval(fmc_private,fmc_private%fil,DETAILLED_SCORE_INFO=DEBUG)
            score(i) = current_score
            print '(a,2(F0.4,1x))', '**info(fmc_bf_tilt): current tilt, score = ', current_tilt, current_score
        enddo
!        !$omp   enddo
!        call fmc_deallocate(fmc_private)
!        !$omp   end parallel

        ! Running average (smooth)
        do I=1,num_tilt_angles
            j_min = max(I-1,1)
            j_max = min(I+1,num_tilt_angles)
            score_smooth(i) = sum(score(j_min:j_max))/real(J_MAX-J_MIN+1)
        enddo

        ! Write plots to disk if requested
        if (wwrite_plots_to_disk) then
#if defined(have_plplot)
            ! Work out filename
            if (present(plot_filename_root)) then
                if (is_blank(plot_filename_root)) then
                    write(*,'(a)') '**ERROR(fmc_bf_tilt): supplied plot filename root is blank!'
                    call terminate('Fatal error in FMC_BF_TILT')
                else
                    if (string_equal(extension_from_filename(plot_filename_root),'   ')) then
                        plot_filename = trim(adjustl(plot_filename_root))//'.png'
                    else
                        plot_filename = trim(adjustl(plot_filename_root))
                    endif
                endif
            else
                write(plot_filename,'(a,i0,a)') 'theta_search_',fmc%fil%id,'.png'
            endif
            ! Plot title
            write(plot_title,'(a,i0)') 'Out-of-plane tilt (theta) search for filament ', fmc%fil%id
            ! Plot series to png
            scores(:,1) = score
            scores(:,2) = score_smooth
            call series_to_png(plot_filename,   YS=SCORES,XS=TILT_ANGLE,                    &
                                                CURVE_TITLES=['score','score (smooth)'],    &
                                                PLOT_TITLE=PLOT_TITLE,                      &
                                                X_TITLE='Theta',                            &
                                                Y_TITLE='Score')
            write(*,'(2a)') '**info(fmc_bf_tilt): score as a function of tilt angle plotted in file: ', plot_filename

#else
            write(*,'(a)') '**info(fmc_bf_tilt): plotting library not available - will not plot score as a function of tilt angle'
#endif
        endif

        ! Find maxima
        locs_best_score = maxloc(score)
        locs_best_score_smooth = maxloc(score_smooth)
        ! Decide which maximum we'll use (not implemented yet - we always use the first maximum
        if (use_smooth_score) then
            if (size(locs_best_score) .eq. 1) then
                which_maximum = 1
            else
                which_maximum = 1
            endif
        else
            if (size(locs_best_score_smooth) .eq. 1) then
                which_maximum = 1
            else
                which_maximum = 1
            endif
        endif

        ! Remember the results of the search
        best_tilt = tilt_angle(locs_best_score(which_maximum))
        best_tilt_smooth = tilt_angle(locs_best_score_smooth(which_maximum))
        best_score = score(locs_best_score(which_maximum))
        best_score_smooth = score(locs_best_score_smooth(which_maximum))
        if (.not. use_smooth_score) then
            if (debug) write(*,'(a,i0)') '**debug(fmc_bf_tilt): here 1: ', locs_best_score(which_maximum)
            best_filament = tilted_filament(locs_best_score(which_maximum))
        else
            if (debug) write(*,'(a,i0)') '**debug(fmc_bf_tilt): here 2: ', locs_best_score_smooth(which_maximum)
            best_filament = tilted_filament(locs_best_score_smooth(which_maximum))
        endif

        ! Printout info
        write(*,'(a,3(F0.3,a))') '**info(fmc_bf_tilt): best tilt was ', best_tilt, '. (searched +/- ', &
                                                    ttilt_half_range ,' in ', ttilt_sampling , ' steps)'
        write(*,'(a,F0.3)') '**info(fmc_bf_tilt): best tilt according to smoothed score curve was ', best_tilt_smooth

        ! Get best filament state into fmc
        if ((maximise_for_each_tilt .and. best_score .gt. starting_score) .or. (.not. maximise_for_each_tilt)) then
            write(*,'(a)') '**info(fmc_bf_tilt): recording state of filament which gave the best score'
            ! If we were doing ds per tilt, we improved on the score
            call fmc_new_filament(fmc,best_filament)
        else
            write(*,'(a,F0.3,a)') '**info(fmc_bf_tilt): since the filament score on entry (', starting_score, &
                                    ') was not improved upon despite each tilt being minimised, the filament will not be altered'
            call fmc_new_filament(fmc,orig_fil)
        endif

        write(*,'(a)') '**info(fmc_bf_tilt): detailled scoring info for best tilt'
        current_score = fmc_eval(fmc,fmc%fil,DETAILLED_SCORE_INFO=.true.)

        ! If we haven't previously done ds_maximise, let's do ds_maximise now
        if (.not. maximise_for_each_tilt) then
            write(*,'(2a)') '**info(fmc_bf_tilt): since we had not done a ds maximisation for each tilt, ', &
                            'we will do one now for the best tilt.'
            call fmc_maximise(fmc)
            ! Score the final state of the filament (print out details too)
            write(*,'(a)') '**info(fmc_bf_tilt): detailled scoring info for final state of the filament'
            current_score = fmc_eval(fmc,fmc%fil,DETAILLED_SCORE_INFO=.true.)
        endif

    end subroutine fmc_bf_tilt

    !>  \brief  Look for the seam by trying all possible positions and scoring them. On output, the filament's waypoints
    !!          should correspond to protofilament 1.
    subroutine fmc_bf_find_mt_seam(self)
        use filaments_microtubules, only : microtubule_has_seam, microtubule_shift_pf
        use fmcs_evaluation, only : fmc_eval
        use UsefulFunctions, only : IsOdd
		implicit none
        ! Arguments
        type(filament_model_comparison),    intent(inout)      ::  self        !<  Filament model comparison object
        ! Private variables
        integer                 ::  pf_shift, pf_shift_start, pf_shift_finish, best_pf_shift
        type(filament)          ::  fil_start_backup, fil_tmp
        real                    ::  current_score, best_score
        logical,    parameter   ::  debug   =   .true.
        ! Start work

        ! Check we're dealing with a microtubule ...
        if (.not. self%fil%microtubule) then
            write(*,'(a)') '**ERROR(fmc_bf_find_mt_seam): filament is not a microtubule'
            call terminate('Fatal error in FMC_BF_FIND_MT_SEAM')
        endif

        ! ... in which we expect a seam
        if (.not. microtubule_has_seam(self%fil)) then
            write(*,'(a)') '**ERROR(fmc_bf_find_mt_seam): microtubule geometry does not have a seam'
            call terminate('Fatal error in FMC_BF_FIND_MT_SEAM')
        endif

        ! Backup the starting point of the filament
        fil_start_backup    = self%fil

        ! We have to find where pf 1 really is.
        ! There are self%fil%mt_pf_number possibilities, including our starting point
        ! We will loop over all of them
        if (IsOdd(self%fil%mt_pf_number)) then
            pf_shift_start  = -(self%fil%mt_pf_number-1) / 2
            pf_shift_finish =  (self%fil%mt_pf_number-1) / 2
        else
            pf_shift_start  = -(self%fil%mt_pf_number/2)
            pf_shift_finish =  (self%fil%mt_pf_number/2) - 1
        endif

        ! Diagnostics
        if (debug) then
            write(*,'(a,i0,a,i0)') '**debug(fmc_bf_find_mt_seam): will search for seam by shifting pf number from ', &
                                    pf_shift_start, ' to ', pf_shift_finish
        endif

        ! Loop over the search range, look for the best score
        best_score = -1.0
        do pf_shift=pf_shift_start,pf_shift_finish
            ! Diagnostics
            if (debug) then
                write(*,'(a,i0)') '**debug(fmc_bf_find_mt_seam): trying to shift protofilaments by ', pf_shift
            endif
            ! Shift protofilament 1
            fil_tmp = fil_start_backup
            call microtubule_shift_pf(fil_tmp,pf_shift)
            call fmc_new_filament(self,fil_tmp)
            current_score =  fmc_eval(self,self%fil)
            ! Diagnostics
            if (debug) then
                write(*,'(a,i0,a,F0.4)') '**debug(fmc_bf_find_mt_seam): protofilament shift of ', pf_shift, &
                                            ' gave a score ', current_score
            endif
            ! Check whether we've improved on the previous best
            if (current_score .gt. best_score) then
                best_pf_shift = pf_shift
                best_score = current_score
            endif
        enddo

        ! Output for the user
        write(*,'(2(A,i0),a,F0.4,a,i0)') '**info(fmc_bf_find_mt_seam): tried shifting protofilaments from ', &
                                            pf_shift_start, ' to ', pf_shift_finish, '. Best score (', best_score, &
                                            ') was obtained with shift of ', best_pf_shift

        ! Leave fmc with best filament state we found
        fil_tmp = fil_start_backup
        call microtubule_shift_pf(fil_tmp,best_pf_shift)
        call fmc_new_filament(self,fil_tmp)

    end subroutine fmc_bf_find_mt_seam

    !>  \brief  Improve the score by moving all the waypoints simultaneously. Akin to a "rigid body" search.
    subroutine fmc_bf_rigid(fmc)
        use fmcs_evaluation, only : fmc_eval
        use filaments_model_modify, only : filament_shift_wp, filament_shift_all_wps
		implicit none
        ! Arguments
        type(filament_model_comparison),    intent(inout)   ::  fmc         !<  Filament-model-comparison object
        ! Private variables
        real            ::  current_shift(6), best_shift(6)
        type(filament)  ::  orig_fil
        integer         ::  i
        real            ::  current_score, best_score
        ! Start work

        ! Subunit numbers are not implemeneted
        ! Anyway, maybe this subroutine could be superseeded / replaced by a more general bf routine?
        call terminate('Subunit number not implemented')

        ! "Backup" the original state of the filament
        orig_fil = fmc%fil

        ! Initialise values
        best_shift      =   0.0
        current_shift   =   0.0
        current_score   =   fmc_eval(fmc,fmc%fil)
        best_score      =   current_score

        ! Loop over all the dimensions
        ! Loop over Psi
        current_shift(6) = - fmc%search_half_range(6)
        do
            if (current_shift(6) .gt. fmc%search_half_range(6)) current_shift(6) = fmc%search_half_range(6)
            ! Loop over Theta
            current_shift(5) = -fmc%search_half_range(5)
            do
                if (current_shift(5) .gt. fmc%search_half_range(5)) current_shift(5) = fmc%search_half_range(5)
                ! Loop over Phi
                current_shift(4) = -fmc%search_half_range(4)
                do
                    if (current_shift(4) .gt. fmc%search_half_range(4)) current_shift(4) = fmc%search_half_range(4)
                    ! Loop over z
                    current_shift(3) = -fmc%search_half_range(3)
                    do
                        if (current_shift(3) .gt. fmc%search_half_range(3)) current_shift(3) = fmc%search_half_range(3)
                        ! Loop over y
                        current_shift(2) = -fmc%search_half_range(2)
                        do
                            if (current_shift(2) .gt. fmc%search_half_range(2)) current_shift(2) = fmc%search_half_range(2)
                            ! Loop over x
                            current_shift(1) = -fmc%search_half_range(1)
                            do
                                if (current_shift(1) .gt. fmc%search_half_range(1)) current_shift(1) = fmc%search_half_range(1)

                                ! Apply shifts to all waypoints
                                fmc%fil = orig_fil
                                call filament_shift_all_wps(fmc%fil,current_shift)
                                print '(a,6(F12.2,1x))', '**debug(fmc_bf_rigid): trying shifts: ', current_shift
                                ! Now score the current state
                                current_score = fmc_eval(fmc,fmc%fil)
                                ! If the current score beats our current maximum, save it and the corresponding shifts
                                if (current_score .gt. best_score) then
                                    best_score = current_score
                                    best_shift = current_shift
                                endif

                                ! Increment x
                                current_shift(1) = current_shift(1) + fmc%search_sampling(1)
                                ! Check whether we're done
                                if (current_shift(1) .ge. fmc%search_half_range(1) + fmc%search_sampling(1)) exit
                            enddo ! End of loop over x
                            ! Increment y
                            current_shift(2) = current_shift(2) + fmc%search_sampling(2)
                            ! Check whether we're done
                            if (current_shift(2) .ge. fmc%search_half_range(2) + fmc%search_sampling(2)) exit
                        enddo ! End of loop over y
                        ! Increment z
                        current_shift(3) = current_shift(3) + fmc%search_sampling(3)
                        ! Check whether we're done
                        if (current_shift(3) .ge. fmc%search_half_range(3) + fmc%search_sampling(3)) exit
                    enddo ! End of loop over z
                    ! Increment Phi
                    current_shift(4) = current_shift(4) + fmc%search_sampling(4)
                    ! Check whether we're done
                    if (current_shift(4) .ge. fmc%search_half_range(4) + fmc%search_sampling(4)) exit
                enddo ! End of loop over Phi
                ! Increment Theta
                current_shift(5) = current_shift(5) + fmc%search_sampling(5)
                ! Check whether we're done
                if (current_shift(5) .ge. fmc%search_half_range(5) + fmc%search_sampling(5)) exit
            enddo ! End of loop over Theta
            ! Increment Psi
            current_shift(6) = current_shift(6) + fmc%search_sampling(6)
            ! Check whether we're done
            if (current_shift(6) .ge. fmc%search_half_range(6) + fmc%search_sampling(6)) exit
        enddo ! End of loop over Psi

        ! Debug output
        write(*,'(a,6(F7.2,1x))') '**debug(fmc_bf_rigid): best shifts = ', best_shift

        ! Apply the best shift
        fmc%fil = orig_fil
        do I=1,fmc%fil%nwp
            call filament_shift_wp(fmc%fil,i,best_shift)
        enddo
        current_score = fmc_eval(fmc,fmc%fil)

    end subroutine fmc_bf_rigid


    !>  \brief  Improve the positioning / in-plane angle of way-points by brute-force search, one way-point at a time.
    !!  \todo   Option to search along/perp to filament axis
    !!  \todo   Integrate search parameters with fmc components
    !!  \todo   Integrate wp selection and other newer fmc features
    subroutine filament_bf_wp_by_wp(fmc,x_half_range,y_half_range,a_half_range,x_step_size,y_step_size,a_step_size,filament_coos)
        use fmcs_evaluation, only : fmc_eval
		implicit none
        ! Arguments
        type(filament_model_comparison),    intent(inout)   ::  fmc             !<  Filament-model comparison object, must be already initialised
        real,                               intent(in)      ::  x_half_range    !<  Half-range for exhaustive search in x coordinate
        real,                               intent(in)      ::  y_half_range    !<  Half-range for exhaustive search in y coordinate
        real,                               intent(in)      ::  a_half_range    !<  Half-range for exhaustive search in in-plane rotation angle
        real,                               intent(in)      ::  x_step_size     !<  Step size for exhaustive search in x coordinate
        real,                               intent(in)      ::  y_step_size     !<  Step size for exhaustive search in y coordinate
        real,                               intent(in)      ::  a_step_size     !<  Step size for exhaustive search in in-plane rotation angle
        logical,    optional,               intent(in)      ::  filament_coos   !<  If .TRUE., x is a shift along the filament axis, y is perpendicular. Otherwise, use film's coordinate system.
        ! Private variables
        logical ::  ffilament_coos
        real    ::  score_entry, score_exit
        real    ::  score_start, score_best, score_current
        real    ::  x_best, y_best, a_best
        real    ::  x_current, y_current, a_current
        real    ::  x_start, y_start, a_start
        integer ::  wp_counter, x_counter, y_counter, a_counter
        integer ::  a_num_steps, x_num_steps, y_num_steps
        ! Start work

        ! Check whether we're using the film or filament coordinate system
        if (present(filament_coos)) then
            ffilament_coos = filament_coos
        else
            ffilament_coos = .false.
        endif

        ! Option to shift way points along/perpendicular to filaments is not implemented yet!
        if (ffilament_coos) then
            write (*,'(a)') '**WARNING(filament_bf_wp_by_wp): option to shift along/perp to filament not implemented yet!'
            write (*,'(a)') '**WARNING(filament_bf_wp_by_wp): switching to shifts along x/Y'
            ffilament_coos = .false.
        endif

        ! Get the score upon entry to the subroutine
        score_entry = fmc_eval(fmc,fmc%fil,WRITE_STACKS=.false.)
        write(*,'(a,F8.3)') '**info(filament_bf_wp_by_wp): score upon entry to subroutine = ', score_entry

        ! Loop over the way points
        do wp_counter = 1, FMC%fil%nwp

            ! Record the starting point
            a_start     = fmc%fil%wp(wp_counter)%euler%psi
            x_start     = fmc%fil%wp(wp_counter)%coo(1)
            y_start     = fmc%fil%wp(wp_counter)%coo(2)
            score_start = fmc_eval( fmc,fmc%fil,WRITE_STACKS=.false.)
            write(*,'(a,i2,a,F8.3)') '**info(filament_bf_wp_by_wp): score before searching waypoint ', wp_counter, '= ', score_start

            ! Reset the best for the current wp
            a_best = a_start
            x_best = x_start
            y_best = y_start
            score_best = score_start


            ! Loop over the in-plane rotations. We start at -half range, go every step size until just before +half range, and then do the starting point and +half range
            a_num_steps = floor((2.*A_HALF_RANGE)/a_step_size)+2
            if (a_half_range .eq. 0. .or. a_step_size .eq. 0.) a_num_steps = 1
            do a_counter = 1, A_NUM_STEPS
                if (a_counter .le. A_NUM_STEPS-2) then
                    a_current = a_start - a_half_range + A_STEP_SIZE*(A_COUNTER-1)
                else if (a_counter .eq. A_NUM_STEPS-1) then
                    a_current = a_start
                else if (a_counter .eq. a_num_steps) then
                    a_current = a_start + a_half_range
                endif

                ! Check current value falls within range
                if (abs(A_CURRENT-A_START) .gt. A_HALF_RANGE+0.0001) then
                    write(*,'(a,3F8.2)') '**ERROR(filament_bf_wp_by_wp): a_current outside range: ', &
                                        a_start, a_current, a_half_range
                    call terminate('Fatal error in FILAMENT_BF_WP_BY_WP')
                endif

                ! Loop over X-shifts
                x_num_steps = floor((2.*X_HALF_RANGE)/x_step_size)+2
                if (x_half_range .eq. 0. .or. x_step_size .eq. 0.) x_num_steps = 1
                do x_counter = 1, X_NUM_STEPS
                    if (x_counter .le. X_NUM_STEPS-2) then
                        x_current = x_start - x_half_range + X_STEP_SIZE*(X_COUNTER-1)
                    else if (x_counter .eq. X_NUM_STEPS-1) then
                        x_current = x_start
                    else if (x_counter .eq. x_num_steps) then
                        x_current = x_start + x_half_range
                    endif

                    ! Check current value falls within range
                    if (abs(X_CURRENT-X_START) .gt. X_HALF_RANGE+0.0001) then
                        write(*,'(a,3F8.2)') '**ERROR(filament_bf_wp_by_wp): x_current outside range: ', &
                                            x_start, x_current, x_half_range
                        call terminate('Fatal error in FILAMENT_BF_WP_BY_WP')
                    endif

                    ! Loop over Y-shifts
                    y_num_steps = floor((2.*Y_HALF_RANGE)/y_step_size)+2
                    if (y_half_range .eq. 0. .or. y_step_size .eq. 0.) y_num_steps = 1
                    do y_counter = 1, Y_NUM_STEPS
                        if (y_counter .le. Y_NUM_STEPS-2) then
                            y_current = y_start - y_half_range + Y_STEP_SIZE*(Y_COUNTER-1)
                        else if (y_counter .eq. Y_NUM_STEPS-1) then
                            y_current = y_start
                        else if (y_counter .eq. y_num_steps) then
                            y_current = y_start + y_half_range
                        endif

                        ! Check current value falls within range
                        if (abs(Y_CURRENT-Y_START) .gt. Y_HALF_RANGE+0.0001) then
                            write(*,'(a,3F8.2)') '**ERROR(filament_bf_wp_by_wp): y_current outside range: ', &
                                                y_start, y_current, y_half_range
                            call terminate('Fatal error in FILAMENT_BF_WP_BY_WP')
                        endif

                        ! Move the way point currently being searched to its current position
                        fmc%fil%wp(wp_counter)%euler%psi    = a_current
                        fmc%fil%wp(wp_counter)%coo(1)       = x_current
                        fmc%fil%wp(wp_counter)%coo(2)       = y_current

                        ! Score the current way point position
                        score_current = fmc_eval(fmc,fmc%fil,WRITE_STACKS=.false.)
                        print '(a,5F10.5)', '**debug(filament_bf_wp_by_wp): a,x,y,score,score_best =', &
                                                    a_current,x_current,y_current,score_current,score_best

                        ! If the score beats our previous best, keep track
                        if (score_current .gt. score_best) then
                            score_best = score_current
                            a_best = a_current
                            x_best = x_current
                            y_best = y_current
                        endif

                    ! End of loop over Y-shifts
                    enddo
                ! End of loop over X-shifts
                enddo
            ! End of loop over in-plane rotations
            enddo

            ! Diagnostics
            print *
            print '(a,i2,a,F8.2,a,F8.2)', '**debug(filament_bf_wp_by_wp): waypoint ', wp_counter, ' went from ',    &
                                                                                     score_start, ' to ', score_best
            print '(a,3F8.2)'          , '**debug(filament_bf_wp_by_wp):    with x,y,a (old): ', x_start, y_start, a_start
            print '(a,3F8.2)'          , '**debug(filament_bf_wp_by_wp):    and  x,y,a (new): ', x_best,  y_best,  a_best
            print *

            ! We've now got the best position for the waypoint, so let's keep it at that before we move on to the next waypoint
            fmc%fil%wp(wp_counter)%euler%psi    = a_best
            fmc%fil%wp(wp_counter)%coo(1)       = x_best
            fmc%fil%wp(wp_counter)%coo(2)       = y_best

            ! Debug
            print *, '**debug(filament_bf_wp_by_wp): only doing 1'
            call terminate('Debug')

        ! End of loop over waypoints
        enddo

        ! Get the final best score
        score_exit = fmc_eval(fmc,fmc%fil,WRITE_STACKS=.false.)
        write(*,'(a,F8.3)') '**info(filament_bf_wp_by_wp): score on exit from subroutine = ', score_exit

    end subroutine filament_bf_wp_by_wp


end module fmcs
