!>  \brief  Functions to evaluate the simiarity between a filament and a model
module fmcs_evaluation

    use fmcs_base

    contains


    !>  \brief  Prepare the helical model so that it can be used for the comparison routines. This includes cropping/padding to appropriate dimensions, imposing helical symmetry, degridding and masking.
    !!          Note that this is only done when necessary.
    subroutine fmc_model3d_prep(fmc,threed_size,mask_rad)
        use images_helices
        use images_arithm
        use masks
        use helices
        use procfs, only : print_memory_usage_stats
        use Globals
        implicit none
        ! Arguments
        type(filament_model_comparison),    target,     intent(inout)   ::  fmc         !<  Filament model comparison (fmc) object
        integer,                            optional,   intent(in)      ::  threed_size !<  If present, this argument will force the model3D to have this size.
        real,                               optional,   intent(in)      ::  mask_rad    !<  Radius of spherical mask to impose on the 3d model. Pixels.
        ! Private variables
        !integer                        ::  num_central_slices
        logical                     ::  different_helix
        logical,    parameter       ::  debug   =   .false.
        integer                     ::  tthreed_size
        logical                     ::  subtract_fil_from_ref
        real                        ::  mmask_rad
        logical                     ::  new_mask
        logical, parameter          ::  skip_hextend =  .false.  ! For debug purposes, to save time
        logical, parameter          ::  use_new_hextend = .true.
        real                        ::  rise_tol, twist_tol
        type(image)                 ::  soft_mask
        type(image)                 ::  temp_img
        type(helical_symmetry)      ::  helix_oversampled
        integer                     ::  new_threed_dimensions(3)
        ! Start work

        ! Memory footprint diag
        !if (debug) then
        !    write(*,'(a)') '**debug(fmc_model3d_prep): memory usage on entry'
        !    call print_memory_usage_stats()
        !endif

        ! Check the model is cubic
        if (.not. fmc%model3dvol_private%IsSquare()) then
            write(*,'(a,3(i0,1x))') '**ERROR(fmc_model3d_prep): sorry, 3d model must be in a cubic box. Image supplied is ', &
                                    fmc%model3dvol_private%logical_dimensions
            call terminate('Fatal error in FMC_MODEL3D_PREP')
        endif

        ! What size are we looking for
        if (present(threed_size)) then
            tthreed_size = threed_size
        else
            tthreed_size = fmc%box_size
            !tthreed_size = fmc%threed_size  ! 1-SEP-2011 test
        endif

        !
        ! Are the filament's helical parameters significantly different from the FMC's helical parameters?
        !
        rise_tol            =   fmc_helical_rise_tolerance(fmc)
        twist_tol           =   fmc_helical_twist_tolerance(fmc)
        different_helix     =   abs(fmc%helix%twist - fmc%fil%helix%twist) .gt. twist_tol .or. &
                                abs(fmc%helix%rise  - fmc%fil%helix%rise)  .gt. rise_tol


        ! Will we need to subtract a new filament from the reference 3d reconstruction?
        subtract_fil_from_ref = .false.
        if (fmc%subtract_fil_from_ref) then
            if (fmc%id_of_subtracted_fil .eq. 0 .or. fmc%id_of_subtracted_fil .ne. fmc%fil%id) then
                subtract_fil_from_ref = .true.
            endif
        endif

        !
        ! What mask radius will we use?
        !
        if (present(mask_rad)) then
            mmask_rad = mask_rad
        else
            if (fmc%wps_as_sps) then
                mmask_rad = fmc%sp_segment_length / 2.0 / fmc%fil%film%GetPixelSize()
            else
                mmask_rad = fmc%fil%max_radius
            endif
        endif

        !
        ! Are we using a new mask radius?
        !
        new_mask = .false.
        if (abs(mmask_rad - fmc%model3dvol_mask_rad) .gt. 0.5) then
            new_mask = .true.
        endif

        !
        ! Do we need to do anything?
        !
        if ( fmc%model3dvol_ready .and. tthreed_size .eq. fmc%model3dvol_private%logical_dimensions(1) .and. &
            .not. different_helix .and. .not. subtract_fil_from_ref .and. .not. new_mask) then
            ! Everything is ready
            return
        else
            ! Let the user know what we're doing
            if (this_program%IsVerbose() .or. debug) then
                write(*,'(2a)') '**info(fmc_model3d_prep): preparing the model 3d volume for use as a reference ',   &
                                'in scoring function (resizing, imposing helical symmetry, masking etc.)'
            endif
            if (debug) then
                write(*,'(a,5l1)') '**debug(fmc_model3d_prep): logicals = ', fmc%model3dvol_ready, &
                        tthreed_size .eq. fmc%model3dvol_private%logical_dimensions(1), .not. different_helix, &
                         .not. subtract_fil_from_ref, .not. new_mask
            endif
            !print '(a,i4)', '**debug(fmc_model3d_prep): preparing model 3d (pad/crop, helical sym, masking), box size = ', fmc%box_size
            fmc%model3dvol_private = fmc%model3dvol_original

            ! Reset the projection cache
            !if (fmc%cache_projections) then
            !    call projection_cache_clean(fmc%proj_cache)
            !endif

            ! Print info regarding the helical symmetrisation
            if (different_helix) then
                if (this_program%IsVerbose() .or. debug) then
                    print '(4(A,F0.4))', '**info(fmc_model3d_prep): using the filament-specific helical parameters of ', &
                                        fmc%fil%helix%twist, ' and ', fmc%fil%helix%rise, &
                                            ' rather than ', fmc%helix%twist, ' and ', fmc%helix%rise
                endif
                fmc%helix = fmc%fil%helix
            else
                if (this_program%IsVerbose() .or. debug) then
                    print '(5(A,F0.4))', '**info(fmc_model3d_prep): the fmc helical parameters (',fmc%helix%twist,   &
                                        ' and ', fmc%helix%rise, &
                                       ') are close enough to this filament''s helical parameters (',fmc%fil%helix%twist,    &
                                       ' and ', fmc%fil%helix%rise, ') and is left unchanged at ', fmc%helix%twist
                endif
            endif

            ! Memory footprint diag
            !if (debug) then
            !    write(*,'(a)') '**debug(fmc_model3d_prep): memory usage before crop/pad'
            !    call print_memory_usage_stats()
            !endif

            ! Debug
            if (debug) then
                call fmc%model3dvol_private%WriteToDisk('dbg_model3dprep_onentry.mrc')
            endif

            !
            ! Check whether padding or cropping necessary
            !
            !print *, '**debug(): before crop/pad fmc%model3dvol%type = ', fmc%model3dvol%type
            !print *, '**debug(): before crop/pad fmc%model3dvol_original%type = ', fmc%model3dvol_original%type
            if (fmc%box_size .lt. fmc%model3dvol_private%logical_dimensions(1)) then
                if (this_program%IsVerbose() .or. debug) then
                    print '(a,i3)', '**info(fmc_model3d_prep): cropping input 3d to ', tthreed_size
                endif
                call fmc%model3dvol_private%Resize([tthreed_size,tthreed_size,tthreed_size])
            else if (fmc%box_size .gt. fmc%model3dvol_private%logical_dimensions(1)) then
                if (this_program%IsVerbose() .or. debug) then
                    print '(2(A,i0))', '**info(fmc_model3d_prep): padding input 3d from ', &
                                    fmc%model3dvol_private%logical_dimensions(1) , &
                                    ' to ', tthreed_size
                endif
                new_threed_dimensions = tthreed_size
                if (debug) then
                    write(*,*) '**debug(fmc_model3d_prep): about to resize from, to ', &
                                fmc%model3dvol_private%logical_dimensions, new_threed_dimensions
                endif
                call fmc%model3dvol_private%Resize(new_threed_dimensions, &
                                                   padding_value_real=fmc%model3dvol_private%GetAverageOfValuesOnEdges())
            endif

            ! Debug
            if (debug) then
                call fmc%model3dvol_private%WriteToDisk('dbg_model3dprep_resize.mrc')
            endif

            !
            ! Subtract the filament from the model 3d reconstruction
            !
            if (subtract_fil_from_ref) then
                if (fmc%subtract_fil_from_ref) then
                    if (this_program%IsVerbose() .or. debug) then
                        write(*,'(a,i0,a)') '**info(fmc_model3d_prep): subtracting current state of filament ', fmc%fil%id, &
                                            ' from the reference 3d reconstruction'
                        ! Subtract the filament reconstruction from the model 3d reconstruction
                        write(*,'(a)') '**info(fmc_model3d_prep): subtracting current reconstruction from reference 3D'
                    endif
                    call threed_add(fmc%threed_original,fmc%fil_threed,fmc%fil_threed,SUBTRACT=.true.)
                    ! Finalise this, and store as model3dvol
                    !call fmc%model3dvol%WriteToDisk('dbg_model3d.mrc')
                    !call threed_write_to_disk(fmc%threed_original,'dbg_threed.mrc')
                    call threed_eval_reconstruction(fmc%fil_threed,fmc%model3dvol_private)
                    !call fmc%model3dvol_private%WriteToDisk('dbg_model3d_sub.mrc')
                    !call threed_write_to_disk(fmc%fil_threed,'dbg_filthreed.mrc')
                    ! Remember that we subtracted from this filament
                    fmc%id_of_subtracted_fil = fmc%fil%id
                    !if (debug) call fmc%model3dvol_private%WriteToDisk('dbg_model3dprep_aftersubtract.mrc')
                else
                    write(*,'(a)') '**ERROR(fmc_model3d_prep): bug - it should be impossible for subtract_fil_from_ref '//&
                                    'to be true without fmc%subtract_fil_from_ref!'
                    call terminate('Fatal error in FMC_MODEL3D_PREP')
                endif
            endif

            ! Use the current filament's 3d reconstruction as a reference for refinement
            if (fmc%single_filament_refinement) then
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a)') '**info(fmc_model3d_prep): evaluating single filament reconstruction'
                endif
                call threed_eval_reconstruction(fmc%fil_threed,fmc%model3dvol_private)
                if (debug) call fmc%model3dvol_private%WriteToDisk('dbg_model3dprep_fil3d.mrc')
            endif



            ! Memory footprint diag
            !if (debug) then
            !    write(*,'(a)') '**debug(fmc_model3d_prep): memory usage before HEXTEND'
            !    call print_memory_usage_stats()
            !endif

            if (fmc%model_ave_cyl) then
                if (debug) call fmc%model3dvol_private%WriteToDisk('dbg_model3dprep_beforecyl.mrc')
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a)') '**info(fmc_model3d_prep): imposing cylindrical averaging on volume'
                endif
                call fmc%model3dvol_private%ApplyCylindricalAverage()
                if (debug) call fmc%model3dvol_private%WriteToDisk('dbg_model3dprep_cyl.mrc')
            else
                !
                ! Impose helical symmetry
                !
                if (debug) call fmc%model3dvol_private%WriteToDisk('dbg_model3dprep_nomask_nohextend.mrc')
                if (skip_hextend) then
                    write(*,'(a)') '**DEBUG(fmc_model3d_prep): not imposing h SYMMETRY'
                else
                    if (this_program%IsVerbose() .or. debug) then
                        write(*,'(a)',advance='NO') '**info(fmc_model3d_prep): imposing helical symmetry on model 3d: '
                    endif
                    if (use_new_hextend) then
                        call image_impose_helical_symmetry_new(fmc%model3dvol_private,fmc%helix,fmc%interpol,  &
                                                                max_radius=fmc%fil%max_radius)
                    else
                        call image_impose_helical_symmetry( fmc%model3dvol_private,fmc%helix,fmc%interpol,  &
                                                            max_radius=fmc%fil%max_radius,progress_report=.true.)
                    endif
                endif

                if (debug) call fmc%model3dvol_private%WriteToDisk('dbg_model3dprep_nomask.mrc')
                !if (debug) call terminate('DEBUG FMC_MODEL3D_PREP')

                ! Memory footprint diag
                !if (debug) then
                !    write(*,'(a)') '**debug(fmc_model3d_prep): memory usage after HEXTEND'
                !    call print_memory_usage_stats()
                !endif

                !
                ! Apply b factor
                !
                if (fmc%model3dvol_bfactor .ne. 0.0) then
                    !write(*,'(a,F0.3,a)') '**info(fmc_model3d_prep): applying b factor (', fmc%model3dvol_bfactor,') to model 3D'
                    if (fmc%model3dvol_private%is_in_real_space) call fmc%model3dvol_private%ForwardFFT()
                    ! We want the falloff to start at the target resolution, and end 5 pixels later
                    if (this_program%IsVerbose() .or. debug) then
                        write(*,'(a,F0.2,a,2(F0.2,1x))') '**info(fmc_model3d_prep): applying b factor of ', &
                                                fmc%model3dvol_bfactor, &
                                                ' Angstroems^2, with cosine-edged low-pas filter with parameters: ', &
                                                min(0.5,fmc%freq_max + (5.0/real(fmc%threed_size))), (5.0/real(fmc%threed_size))
                    endif
                    call fmc%model3dvol_private%ApplyBFactor( fmc%model3dvol_bfactor/fmc%fil%film%GetPixelSize()**2,   &
                                        lp_cosine=(fmc%freq_max .lt. 0.5),                                     &
                                        lp_cosine_freq=min(0.5,fmc%freq_max + (5.0/real(fmc%threed_size))),    &
                                        lp_cosine_falloff=(5.0/real(fmc%threed_size)))
                    if (.not. fmc%model3dvol_private%is_in_real_space) call fmc%model3dvol_private%BackwardFFT()
                endif

                !
                ! Masking (envelope of the filament)
                !
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a)') '**info(fmc_model3d_prep): calculating and applying mask to model 3D'
                endif
                if (debug) call fmc%model3dvol_private%WriteToDisk('dbg_model3dprep_beforehmask.mrc')
                call image_helix_mask(fmc%model3dvol_private,fmc%helix,fmc%asym_unit_mw, &
                                      fmc%fil%film%GetPixelSize(),fmc%fil%max_radius, &
                                      mask_resolution=fmc%fil%film%GetPixelSize()/fmc%estimated_resolution_of_model, &
                                      soft_mask=soft_mask)

                ! Memory footprint diag
                !if (debug) then
                !    write(*,'(a)') '**debug(fmc_model3d_prep): memory usage after masking'
                !    call print_memory_usage_stats()
                !endif

            endif

            if (debug) call fmc%model3dvol_private%WriteToDisk('dbg_before_resampling.mrc')

            !
            ! Oversample
            !
            if (fmc%model3d_oversampling .gt. 1) then
                if (fmc%model3dvol_private%IsInRealSpace()) call fmc%model3dvol_private%ForwardFFT()
                call temp_img%Allocate(dims=fmc%model3dvol_private%GetLogicalDimensions()*fmc%model3d_oversampling)
                call fmc%model3dvol_private%ClipInto(temp_img)
                fmc%model3dvol_private = temp_img
                call temp_img%Deallocate()
            endif

            if (debug) call fmc%model3dvol_private%WriteToDisk('dbg_before_correct3d.mrc')

            !
            ! Degridding
            !
            ! If the projection will happen in real space, the correction needs to happen in Fourier
            ! space, and vice versa.
            !
            if (fmc%real_space_proj) then
                if (fmc%model3dvol_private%is_in_real_space) call fmc%model3dvol_private%ForwardFFT()
            else
                if (.not. fmc%model3dvol_private%is_in_real_space) call fmc%model3dvol_private%BackwardFFT()
            endif

            if (this_program%IsVerbose() .or. debug) then
                write(*,'(a)') '**info(fmc_model3d_prep): dividing by reciprocal of interpolation function (aka (de)gridding)'
            endif
            call correct_3d(    fmc%model3dvol_private, &
                                fmc%interpol%interpolation_instr_function,  &
                                fmc%interpol%windowing_instr_function,      &
                                fmc%interpol%window_width,                  &
                                reverse=.false. )
            if (.not. fmc%model3dvol_private%IsInRealSpace()) call fmc%model3dvol_private%BackwardFFT()
            if (soft_mask%IsAllocated()) then
                if (fmc%model3d_oversampling .gt. 1) then
                    call temp_img%Allocate(dims=soft_mask%GetLogicalDimensions()*fmc%model3d_oversampling)
                    if (soft_mask%IsInRealSpace()) call soft_mask%ForwardFFT()
                    call soft_mask%ClipInto(temp_img)
                    soft_mask = temp_img
                    call temp_img%Deallocate()
                    if (soft_mask%IsInFourierSpace()) call soft_mask%BackwardFFT()
                endif
                call multiply_images(fmc%model3dvol_private,soft_mask,fmc%model3dvol_private)
            endif
            if (debug) call fmc%model3dvol_private%WriteToDisk('dbg_after_correct3d.mrc')


            !
            ! Debug dump
            !
            if (debug) then
                !print *, '**debug(fmc_model3d_prep): may dump prep 3d to disk ', fmc%fil%id
                !if (fmc%fil%id .EQ. 1) THEN
                    print '(a)', '**debug(fmc_model3d_prep): dumping prepared 3d to disk'
                    call fmc%model3dvol_private%WriteToDisk('dbg_model3dprep.mrc')
                    !call terminate ('debug')
                !endif
            endif

            !
            ! We may need to mask the helical subunit
            !
            if (fmc%project_helical_subunits) then
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a)') '**debug(fmc_model3d_prep): masking central helical subunit (soft)'
                endif
                if (debug) then
                    call fmc%model3dvol_private%WriteToDisk('dbg_model3d_prep_before_hunit_mask.mrc')
                endif
                call image_mask_helical_subunit(fmc%model3dvol_private, &
                                                helix_scale_distances(fmc%helix,real(fmc%model3d_oversampling)), &
                                                fmc%fil%max_radius*real(fmc%model3d_oversampling),soften_the_mask=.true.)
                if (debug) then
                    call fmc%model3dvol_private%WriteToDisk('dbg_model3d_prep_after_hunit_mask.mrc')
                endif
            endif

            !
            ! We may need to go to Fourier space and/or apply a phase shift
            !
            if (fmc%real_space_proj) then
                if (.not. fmc%model3dvol_private%ObjectIsCenteredInBox()) call fmc%model3dvol_private%SwapRealSpaceQuadrants()
                if (fmc%model3dvol_private%IsInFourierSpace()) call fmc%model3dvol_private%BackwardFFT()
            else
                if (fmc%model3dvol_private%is_in_real_space) call fmc%model3dvol_private%ForwardFFT()
                if (fmc%model3dvol_private%object_is_centered_in_box) call fmc%model3dvol_private%SwapRealSpaceQuadrants()
            endif

            !
            ! The model is now ready
            !
            fmc%model3dvol_ready = .true.
            if (this_program%IsVerbose() .or. debug) then
                write(*,'(a)') '**info(fmc_model3d_prep): all done'
            endif
        endif ! end of test for whether we need to do anything
    end subroutine fmc_model3d_prep



    !>  \brief  Energy function to be called by the fgsl sa minimiser
    function fmc_eval_sa_energy(fmc_ptr) bind(c)
        use, intrinsic  ::  iso_c_binding
        implicit none
        ! Arguments
        type(c_ptr),    value   ::  fmc_ptr !<  Pointer to filament model comparison object
        ! Result
        real(kind=c_double)     ::  fmc_eval_sa_energy
        ! Private variables
        type(filament_model_comparison),    pointer ::  fmc
        ! Start work
        call c_f_pointer(fmc_ptr,fmc)
        fmc_eval_sa_energy = - fmc_eval(fmc,fmc%fil,DETAILLED_SCORE_INFO=.false.)
    end function fmc_eval_sa_energy


    !>  \brief  Score the comparison between the filament and the model using projections of individual subunits
    real function fmc_eval_with_subunits(self,fil)
        use images_helices, only : image_project_many_h_subunits_with_h_distortion
        use filaments_path, only : filament_get_path_near_wps
        ! Arguments
        type(filament_model_comparison),    intent(inout)   ::  self
        type(filament),                     intent(inout)   ::  fil
        ! Private variables
        logical,            parameter   ::  debug   =   .false.
        integer                         ::  current_wp
        type(waypoint),     allocatable ::  nearby_subunits(:)
        type(Image)                     ::  proj

        ! Start work

        ! Loop over waypoints
        do current_wp=1,fil%nwp
            if (fil%wp(current_wp)%refine) then
                ! Box out relevant area of micrograph
                call fmc_box_out_from_micrograph(self,fil%wp(current_wp),1)
                ! Compute position & parameters for helical subunits in vicinity

                ! Generate projection
                call image_project_many_h_subunits_with_h_distortion(self%model3dvol_private,self%helix, &
                                                                     proj,nearby_subunits,fil%max_radius)

            endif ! refine current waypoint?
        enddo ! loop over waypoints

    end function fmc_eval_with_subunits

    !>  \brief  Box out relevant region of micrograph
    subroutine fmc_box_out_from_micrograph(self,wp_phased,local_thread_num)
        implicit none
        ! Arguments
        type(filament_model_comparison),    intent(inout)   ::  self
        type(waypoint),                     intent(in)      ::  wp_phased           !<  Waypoint at which to box out from micrograph
        integer,                            intent(in)      ::  local_thread_num    !<  (OpenMP) thread from which is we are being called
        ! Private variables
        logical,        parameter   ::  debug   =   .false.
        type(Image)                 ::  temp_cutout_from_frame
        integer                     ::  current_frame
        real                        ::  coordinates_for_current_wp_on_current_frame(3)
        integer                     ::  icoo(3)
        real                        ::  rcoo(3)
        real                        ::  shift_during_current_frame(3)


        ! Start work

        ! Prepare a temporary image
        call temp_cutout_from_frame%Reset()
        temp_cutout_from_frame%is_in_real_space = .false.
        temp_cutout_from_frame%object_is_centered_in_box = .true.
        temp_cutout_from_frame%logical_dimensions = self%box(local_thread_num)%logical_dimensions

        ! Loop over micrograph (movie) frames
        do current_frame=self%first_frame_for_refinement,self%last_frame_for_refinement
            if (debug) then
                write(*,'(a,i0,a,3(f0.2,1x))') '**debug(fmc_box_out_from_micrograph): on frame ', current_frame, &
                                                    ', subframe shifts: ', wp_phased%local_subframe_shift(:,current_frame)
            endif
            coordinates_for_current_wp_on_current_frame = wp_phased%coo - wp_phased%local_subframe_shift(:,current_frame)
            ! Get integer coordinates & remainders
            icoo(1:2) = nint(coordinates_for_current_wp_on_current_frame(1:2))
            icoo(3) = 1
            rcoo = coordinates_for_current_wp_on_current_frame - icoo
            ! Window image
            ! (maybe this should be windowed bigger initially to recover information spread by psf/CTF?)
            ! (no, we're imposing a small-ish mask later on instead)
            call self%micrograph_frames(current_frame)%ClipInto(temp_cutout_from_frame,icoo)
            ! Apply phase shift
            call temp_cutout_from_frame%ForwardFFT()
            call temp_cutout_from_frame%PhaseShift(-rcoo(1),-rcoo(2),0.0e0)
            ! Apply dose filter
            if (self%apply_dose_filter) then
                call self%fil%film%electron_dose%ApplyDoseFilterToImage( temp_cutout_from_frame, &
                                                            dose_start=(current_frame-1)*self%fil%film%GetExposurePerFrame(),&
                                                            dose_finish=current_frame*self%fil%film%GetExposurePerFrame(), &
                                                            pixel_size=self%fil%film%GetPixelSize())
            endif

            ! Apply drift filter
            if (self%apply_drift_filter) then
                if (current_frame .eq. 1) then
                    shift_during_current_frame =  wp_phased%local_subframe_shift(:,current_frame+1) &
                                                - wp_phased%local_subframe_shift(:,current_frame  )
                else if (current_frame .eq. self%fil%film%GetNumberOfFrames()) then
                    shift_during_current_frame =  wp_phased%local_subframe_shift(:,current_frame  ) &
                                                - wp_phased%local_subframe_shift(:,current_frame-1)
                else
                    shift_during_current_frame = (wp_phased%local_subframe_shift(:,current_frame+1) &
                                                - wp_phased%local_subframe_shift(:,current_frame-1))*0.5
                endif
                call temp_cutout_from_frame%ApplyDriftFilter(shift_during_current_frame(1),shift_during_current_frame(2))
            endif

            ! Add to sum
            if (current_frame .eq. self%first_frame_for_refinement) then
                self%box(local_thread_num) = temp_cutout_from_frame
            else
                call self%box(local_thread_num)%AddImage(temp_cutout_from_frame)
            endif
        enddo
        call temp_cutout_from_frame%Deallocate()
    end subroutine fmc_box_out_from_micrograph


    !>  \brief  Wrapper to be called by the fgsl ds minimiser
    !!
    !!  The values and parameters are passed as c pointers so that this subroutine can be called by the (f)GSL.
    !!  The parameter passed is in fact a filament_model_comparison object, which contains all the information needed
    !!  to do the comparison.
    function fmc_eval_wrap(values,params) bind(c)
        use, intrinsic :: iso_c_binding
        use fgsl, only : fgsl_double, fgsl_vector, fgsl_int, fgsl_obj_c_ptr, fgsl_vector_align, fgsl_success
		use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(c_ptr), value  ::  values  !<  c pointer to current values to try out
        type(c_ptr), value  ::  params  !<  c pointer to a fmc object
        ! Result
        real(kind=c_double) ::  fmc_eval_wrap   !<  Result of the function
        ! Private variables
        type(filament_model_comparison),    pointer     ::  fmc
        real(fgsl_double),                  pointer     ::  val_fgsl(:)
        real(kind=8),                       allocatable ::  val(:)
        real(kind=8),                       pointer     ::  val_ptr(:)
        type(fgsl_vector)                               ::  val_vec
        integer                                         ::  expected_num_values
        integer(fgsl_int)                               ::  status
        integer                                         ::  num_dim_total, ierr
        logical,                            parameter   ::  debug           =   .false.
        logical,                            parameter   ::  sanity_check    =   .false.
        real                                            ::  dummy
        type(filament)                                  ::  fil
        logical,                            parameter   ::  skip_scaling    =   .false. ! For debugging purposes
        character(len=6)                                ::  size_val_c
        ! Start work


        call c_f_pointer(params,fmc)

        ! Sanity check (if something went wrong the c pointer, we can end up with the wrong area of memory for the fmc)
        if (fmc%fil%nwp .ne. fmc%fil_original%nwp .or. fmc%box_size .lt. 0 .or. fmc%num_minim_runs .lt. 0) then
            write(*,'(a)') '**ERROR(fmc_eval_wrap): something strange with incoming FMC'
            call terminate('Fatal error in FMC_EVAL_WRAP')
        endif

        ! Work out how many dimensions
        num_dim_total = fmc_num_dim_total(fmc)
        expected_num_values = num_dim_total


        if (fmc%type_of_minimiser .eq. downhill_simplex) then
            ! Get fgsl_double reals from the incoming c pointer
            call fgsl_obj_c_ptr(val_vec,values)
            status = fgsl_vector_align(val_fgsl,val_vec)
            if (status .ne. fgsl_success) call terminate('FGSL_VECTOR_ALIGN failed in FMC_EVAL_WRAP')
            ! Check we got the expected number of variables
            if (size(val_fgsl) .ne. expected_num_values) then
                write(*,'(a,2(i0,1x))') '**ERROR(fmc_eval_wrap): did not get expected number of variables ', &
                                    size(val_fgsl), expected_num_values
                call terminate('Fatal error in FMC_EVAL_WRAP')
            endif

            ! Convert from fgsl to real
            allocate(val(expected_num_values),stat=IERR)
            if (ierr .ne. 0) then
                write(*,'(a,i0)') '**ERROR(fmc_eval_wrap): allocation of real values array failed with error', ierr
                call terminate('Fatal error in FMC_EVAL_WRAP')
            endif
            val = val_fgsl
        else
            call c_f_pointer(values,val_ptr,SHAPE=(/expected_num_values/))
            val = val_ptr
            if (size(val) .ne. expected_num_values) then
                write(*,'(a,2(i0,1x))') '**ERROR(fmc_eval_wrap): did not get expected number of variables ', &
                                size(val), expected_num_values
                call terminate('Fatal error in FMC_EVAL_WRAP')
            endif

        endif

        !if (debug) then
        !   write(*,*) '**debug(fmc_eval_wrap): trying these values: ', val, val_ptr
        !   write(*,'(a,<SIZE(val)>(G0.5,x))') '**debug(fmc_eval_wrap): trying these values: ', val
        !endif

        !if (debug) then
        !   write(*,'(a)') '**debug(fmc_eval_wrap): filament info before values_to_fil (in fmc)'
        !   call filament_print_info(fmc%fil)
        !endif

        ! Restore filament to original state
        fil = fmc%fil_original

        !if (debug) then
        !   write(*,'(a)') '**debug(fmc_eval_wrap): filament info before values_to_fil (copied)'
        !   call filament_print_info(fil)
        !endif

        ! Assign the new positions & angles to the filament's waypoints
        if (skip_scaling) then
            write(*,'(a)') '**WARNING(fmc_eval_wrap): not scaling values'
        else
            call fmc_scale_values(fmc,val,values_to_fil=.true.)
        endif
        call values_to_filament(fil,val,fmc%refine_params,fmc%hold_param_average,incremental_values=fmc%refine_param_increments)


        if (debug) then
           write(*,'(a)') '**debug(fmc_eval_wrap): filament info after values_to_fil, before scoring routine'
           call filament_print_diff(fmc%fil_original,fil)
           call fil%PrintInfo()
        endif

        ! We can now call the actual function that does the scoring
        fmc_eval_wrap = - fmc_eval(fmc,fil,detailled_score_info=.false., &
                                    check_each_wp_has_a_scoring_wp=.false.)

        ! Remember we called the scoring function
        fmc%calls_from_minimiser = fmc%calls_from_minimiser + 1

        if (debug) then
            write(size_val_c,'(i6)') size(val)
            write(*,'(a,i0,a,'//size_val_c//'(F0.3,1x),a,F0.6,a)') '**debug(fmc_eval_wrap): call # ', fmc%calls_from_minimiser, &
                                                             ' , values (score): ', val , '( ', -FMC_EVAL_WRAP, ' )'
        endif

        if (sanity_check) then
            dummy = - fmc_eval(fmc,fil,detailled_score_info=.false.)
            if (debug) write(*,'(a,F0.6)') '**debug(fmc_eval_wrap): got this score (2nd time, no threading): ', dummy
            if ((dummy - fmc_eval_wrap) .gt. 0.000001) then
                write(*,'(a,F0.6,a,F0.6,a,G15.3)') '**ERROR(fmc_eval_wrap): non-theaded score (', dummy, &
                                                    ') different from threaded score (', fmc_eval_wrap, &
                                                    '). Difference = ', dummy - fmc_eval_wrap
                call terminate('Fatal error in FMC_EVAL_WRAP')
            endif
        endif

        !print *, '**debug(filament_eval_model_agreement_wrap): done'
    end function fmc_eval_wrap

    !>  \brief  Score the comparison between the filament and the model
    real function fmc_eval(self,fil,write_stacks,stacks_rootname,detailled_score_info, &
                            check_each_wp_has_a_scoring_wp,threading, &
                            compute_logp)
        ! Arguments
        type(filament_model_comparison),    intent(inout)   ::  self
        type(filament),                     intent(inout)   ::  fil
        logical,                optional,   intent(in)      ::  write_stacks                    !<  If .TRUE., write stacks to disk of segmented filament and matching projections
        character(len=*),       optional,   intent(in)      ::  stacks_rootname                 !<  Root file name to use when writing stacks out.
        logical,                optional,   intent(in)      ::  detailled_score_info            !<  Whether to print out detailled info on the score at each point of path along the filament. Defaults to false.
        logical,                optional,   intent(in)      ::  check_each_wp_has_a_scoring_wp  !<  True by default
        logical,                optional,   intent(in)      ::  threading                       !<  Can be used to force threading off
        logical,                optional,   intent(in)      ::  compute_logp                    !<  Return the log probability rather than the score
        ! Private variables
        logical,    parameter   ::  debug   =   .false.
        integer                 ::  number_of_calls, current_call
        real                    ::  score_check
        type(filament)          ::  fil_check(2)
        ! Start work

        ! Are we going to call the scoring routine twice, for testing purposes?
        if (debug) then
            number_of_calls = 2
        else
            number_of_calls = 1
        endif
        ! Call the scoring subroutine
        do current_call = 1, number_of_calls
            if (debug) then
                write(*,'(a,i0,a,i0)') '**debug(fmc_eval): filament info before call ', current_call, ' of ', number_of_calls
                call self%fil%PrintInfo()
            endif
            fmc_eval = fmc_eval_slave(             self,                                        &
                                                   fil,                                         &
                                                   write_stacks=write_stacks,                   &
                                                   stacks_rootname=stacks_rootname,             &
                                                   detailled_score_info=detailled_score_info,   &
                                                   check_each_wp_has_a_scoring_wp=check_each_wp_has_a_scoring_wp, &
                                                   threading=threading, &
                                                   compute_logp=compute_logp)
            if (number_of_calls .gt. 1 .and. current_call .eq. 1) score_check = fmc_eval
            if (debug) then
                write(*,'(a,i0,a,i0)') '**debug(fmc_eval): filament info after call ', current_call, ' of ', number_of_calls
                call self%fil%PrintInfo()
                fil_check(current_call) = self%fil
            endif
        enddo

        ! Sanity check
        if (debug) then
            if (abs(score_check - fmc_eval) .gt. .00001) then
                write(*,'(a,2(G15.6,1x))') '**ERROR(fmc_eval): the scoring routine was called twice, with different results: ', &
                                        fmc_eval, score_check
                call filament_print_diff(fil_check(1),fil_check(2))
                call terminate('Failed sanity check in FMC_EVAL')
            endif
        endif
    end function fmc_eval


    !>  \brief  Measure how well a model matches a filament by measuring the total correlation of the model's projection
    !!          along the filament's path
    !!  \todo   Move masks generation to outside this subroutine?
    function fmc_eval_slave(    self,fil,       &
                                write_stacks,stacks_rootname,detailled_score_info,                  &
                                check_each_wp_has_a_scoring_wp,threading, compute_logp)         &
                                    result(score)
        use images_comparison, only : image_prepare, projection_prepare, cclin_num_denoms_rings, cc3_finalise, cclin_num_denoms
		use images_comparison, only : cc3, cclin, cclin_num_denoms_from_rings
        use images_ctf, only : ctf_to_img
        use omp_lib
        use rotation_matrices, only : identity_rotmat
        use fscs, only : compute_fsc
        use parsers, only : string_equal, extension_from_filename, is_blank
        use filaments_model_update, only : filament_update_model
		use filaments_geometry, only : filament_defocus_range, filament_eval_total_curvature_overtwist_stretch
		use filaments_geometry, only : filament_splines_are_monotonic
		use filaments_path, only : filament_get_path, filament_t_interval_for_scoring_function, add_waypoint_to_path
		use filaments_normal_modes, only : filament_nm_amplitudes_weighted_sum
		use coos, only : waypoints_print_info, wp_print_info
		use Micrographs
		use masks, only : fourier_mask_init_amyloid
		use images
		use runtime_parameters, only : default_falloff_pixels, safe_for_sqrt, filename_len
		use ImageFiles
		use images_helices, only : image_impose_helical_symmetry_new
		use Globals
#ifndef __GFORTRAN__
		use, intrinsic :: ieee_arithmetic, only : ieee_is_nan
#endif
		use filaments_geometry, only : nearest_wp_from_t
		use ctfs, only : ctf_parameters, ctf_ang_to_pix, ctf_at_z, ctf_to_curve
		implicit none
        ! Arguments
        type(filament_model_comparison),intent(inout)   ::  self                            !<  Filament-model comparison object
        type(filament),                 intent(inout)   ::  fil                             !<  Filament object
        logical,            optional,   intent(in)      ::  write_stacks                    !<  If .TRUE., write stacks to disk of segmented filament and matching projections
        character(len=*),   optional,   intent(in)      ::  stacks_rootname                 !<  Root file name to use when writing stacks out.
        logical,            optional,   intent(in)      ::  detailled_score_info            !<  Whether to print out detailled info on the score at each point of path along the filament. Defaults to false.
        logical,            optional,   intent(in)      ::  check_each_wp_has_a_scoring_wp  !<  True by default
        logical,            optional,   intent(in)      ::  threading                       !<  Use this to forcefully turn off threading
        logical,            optional,   intent(in)      ::  compute_logp                    !<  Return the log probability rather than the score
        ! Result
        real                                            ::  score
        ! Private variables
        real                            ::  score_local                             ! Local, normalised score
        real(kind=8)                    ::  score_nd_local(3), score_nd_whole(3)    ! nd stands for numerator denominator
        real(kind=8)                    ::  score_nd_whole_private(3)
        real(kind=8),       allocatable ::  cc3_nd_rings_local(:,:), cc3_nd_rings_whole(:,:)    ! nd stands for numerator denominator
        real(kind=8),       allocatable ::  cc3_nd_rings_whole_private(:,:)
        integer                         ::  scoring_waypoint_counter, j
        integer                         ::  icoo(3)                                             ! integer coordinates
        real                            ::  rcoo(3)                                             ! remainder coordinates
        type(ctf_parameters)            ::  ctf
        logical                         ::  wwrite_stacks, ddetailled_score_info
        character(len=filename_len)     ::  boxes_filename, projs_filename, ampls_filename, ampl_masks_filename
        real                            ::  matrix(4,4)
        integer                         ::  num_local_threads                                   ! Number of openmp threads
        integer                         ::  local_thread_num                                    ! Current thread number
        integer                         ::  loc_count
        integer                         ::  number_of_scoring_waypoints
        character(len=6)                ::  number_of_scoring_waypoints_c                                           ! Used in format specs
        logical                         ::  tthreading                                          ! Whether we will thread this scoring function
        character(len=filename_len)     ::  tmp_string
        integer                         ::  minim_nwp
        real,               parameter   ::  soft_edge               =   default_falloff_pixels  !<  The width of the soft-edge falloff for masking (as a fraction of maximum inner radius)
        logical                         ::  wwhole_filament_norm                                !<  If true, the score is the correlation is calculated over the whole filament, rather than an average of normalised cc of segments
        logical,            parameter   ::  get_ctf_from_z                          =   .true.  !<  If true, the z coordinate of a waypoint is used to get the required CTF.
                                                                                                !!  If .FALSE., the x,y coordinates, in conjunction with the ctftilt information is used instead.
        logical,            parameter   ::  debug                                   =   .false.
        logical,            parameter   ::  normalised_scoring_function             =   .true.  !<  Whether a normalised cc should be used as a scoring function
        logical,            parameter   ::  single_ctf_per_filament_if_reasonable   =   .false. !<  To save time, one can use a single ctf function for the whole filament. This can be
                                                                                                !!  safely done if the filament's z does not vary more than desired defocus accuracy.
		logical,            parameter   ::  return_minus_one_if_not_monotonic       =   .true.  !<  If the phi & subunit_num splines are not monotonic, there is something funny going, so let's return a bad score. (this may give a performance hit)
		logical,            parameter   ::  return_minus_one_if_any_negatives_coos  =   .true.  !<  If any of the filament's waypoints has negative coordinates, return -1.0
		logical,            parameter   ::  return_minus_one_if_coos_beyond_edges   =   .true.  !<  If any of the filament's waypoints has coordinates that lie beyond the edges of the micrograph, return -1.0
		type(ImageFile)                 ::  boxes_file
        type(ImageFile)                 ::  projs_file
        logical                         ::  box_is_blank, return_minus_one
        real                            ::  obj_max_rad
        logical                         ::  first_time                                      ! helps keep track when doing cc3 calculation
        real                            ::  rect_dim(2)
        logical                         ::  oone_ctf_for_whole_filament
        real                            ::  proj_max_rad
        logical,        allocatable     ::  wp_has_score_wp_nearby(:)
        real(kind=8)                    ::  dummy
        logical                         ::  ccheck_each_wp_has_a_scoring_wp
        ! Variables for phase-shifting due to subunit_num
        type(waypoint)                  ::  wp_phased
        real(kind=8)                    ::  model_z_shift
        ! Curvature, torsion & stretch penalties, maximum-likelihood-like
        logical                         ::  estimate_noise_variance
        real                            ::  number_of_pixels_in_scoring_function
        real                            ::  estimated_noise_variance   ! see notes of 9-Jul-2011 and Sigworth 1998
        real                            ::  estimated_signal_variance   ! see notes of 9-Jul-2011 and Sigworth 1998
        real(kind=8)                    ::  curvature, torsion, stretch
        real                            ::  curvature_penalty
        real                            ::  penalty_factor
        real                            ::  torsion_penalty
        real                            ::  stretch_penalty
        integer                         ::  start_wp, finish_wp ! If only a few waypoints are being refined, it may not make sense to keep
        real(kind=8)                    ::  start_t, finish_t
        logical,        parameter       ::  oonly_score_refining_neighbourhood = .true.
        logical,        parameter       ::  hextend_per_wp = .false.
        !
        logical,        parameter       ::  make_synth_micrograph   =   .false.
        character(len=11), parameter    ::  synth_micrograph_fn     =   'dbg_mic.mrc'
        type(image)                     ::  synth_micrograph
        !
        logical                         ::  ccompute_logp


        !
        ! Start work
        !


        if (debug) then
            write(*,'(a)') '**debug(fmc_eval_slave): filament info on entry:'
            call fil%PrintInfo()
        endif

        ! Are we computing the LogP instead of score?
        ccompute_logp = .false.
        if (present(compute_logp)) ccompute_logp = compute_logp

        ! Will we print out scoring info along the path?
        ddetailled_score_info = .false.
        if (present(detailled_score_info)) ddetailled_score_info = detailled_score_info
        if (ddetailled_score_info .and. debug) write(*,'(a)') '**debug(fmc_eval_slave): detailled scoring information follows'

        ! Will we be writing stacks to disk?
        wwrite_stacks = .false.
        if (present(write_stacks)) wwrite_stacks = write_stacks


        ! If we're writing stacks to disk, prepare filenames & files
        if (wwrite_stacks) then
            if (present(stacks_rootname)) then
                tmp_string = trim(adjustl(stacks_rootname))
                if (string_equal(extension_from_filename(tmp_string),'   ')) then
                    boxes_filename = trim(adjustl(tmp_string))//'_boxes.mrc'
                    projs_filename = trim(adjustl(tmp_string))//'_projs.mrc'
					ampls_filename = trim(adjustl(tmp_string))//'_boxes_ampl.mrc'
					ampl_masks_filename = trim(adjustl(tmp_string))//'_boxes_ampl_mask.mrc'
                else
                    boxes_filename = tmp_string(1:len_trim(tmp_string)-4)//'_boxes'//tmp_string(len_trim(tmp_string)-3:)
                    projs_filename = tmp_string(1:len_trim(tmp_string)-4)//'_projs'//tmp_string(len_trim(tmp_string)-3:)
					ampls_filename = tmp_string(1:len_trim(tmp_string)-4)//'_boxes_ampl'//tmp_string(len_trim(tmp_string)-3:)
               ampl_masks_filename = tmp_string(1:len_trim(tmp_string)-4)//'_boxes_ampl_mask'//tmp_string(len_trim(tmp_string)-3:)
                endif
            else
                boxes_filename = 'boxes.mrc'
                projs_filename = 'projs.mrc'
				ampls_filename = 'boxes_ampl.mrc'
				ampl_masks_filename = 'boxes_ampl_mask.mrc'
            endif

            ! Initialise the files
            call boxes_file%Init(boxes_filename,dim_x=self%box_size,dim_y=self%box_size,dim_z=1)
            call projs_file%Init(projs_filename,dim_x=self%box_size,dim_y=self%box_size,dim_z=1)
		endif

        ! If waypoints are independent single particles, it makes no sense to do whole-filament normalisation of the score
        wwhole_filament_norm = .not. self%wps_as_sps

        ! Will we make sure thate every filament waypoint has at least one scoring waypoint?
        ccheck_each_wp_has_a_scoring_wp = .true.
        if (present(check_each_wp_has_a_scoring_wp)) ccheck_each_wp_has_a_scoring_wp = check_each_wp_has_a_scoring_wp

        ! Are we estimating the noise variance? (we'll only do so if we need to)
        estimate_noise_variance = .false.

        ! If we want to apply the curvature penalty, ...
        if (self%curvature_restraint) then
            ! ... we need to have an estimate of the persistence length
            if (fil%persistence_length == 0.0) then
                write(*,'(a)') '**ERROR(fmc_eval_slave): curvature penalties cannot be applied without'//&
                                                        ' an estimate of the filament persistence length'
                call terminate('Fatal error in FMC_EVAL_SLAVE')
            endif
            ! ... we need to be using linear cross-correlation, since we don't know how to implement mll restraints a la Sigworth with weighted correlation coefficients
            if (self%type_of_scoring_function .ne. linear_correlation) then
                write(*,'(a)') '**ERROR(fmc_eval_slave): ML-like curvature restraints are only implemented '//&
                                'in conjunction with linear cross correlation'
                call terminate('Fatal error in FMC_EVAL_SLAVE. Feature not implemented.')
            endif
            ! ... we need to estimate the noise variance
            estimate_noise_variance = .true.
        endif

        ! We only know how to estimate noise variance if we're doing whole-filament normalisation
        if (estimate_noise_variance .and. .not. wwhole_filament_norm) then
             write(*,'(a)') '**ERROR(fmc_eval_slave): have not implemented noise variance estimator in case where '//&
                            'whole-filament normalisation is not activated'
             call terminate('Feature not yet implemented in FMC_EVAL_SLAVE')
        endif

        ! Check the filament has enough waypoints
        if (.not. fil%fal_wps_in_score) then
            minim_nwp = 3
        else
            minim_nwp = 1
        endif

        if (fil%nwp .lt. minim_nwp) then
            write(*,'(3(A,i0),a)') '**ERROR(fmc_eval_slave): filament ', fil%id, ' only has ', fil%nwp, &
                                    ' waypoints. a minimum of ', minim_nwp, ' are needed.'
            call terminate('Fatal error in FMC_EVAL_SLAVE')
        endif

        ! Diagnostics
        if (wwrite_stacks) then
            print '(4a)', '**info(fmc_eval_slave): dumping stacks ', trim(adjustl(boxes_filename)), &
                                                            ' and ', trim(adjustl(projs_filename))
        endif

        if (debug) then
            write(*,'(a)') '**debug(fmc_eval_slave): filament info before updating model:'
            call fil%PrintInfo()
        endif



        !
        ! Make sure our filament model is up-to-date
        !
        call filament_update_model(fil,wps_as_sps=self%wps_as_sps)

        !
        ! Now that the filament model has been updated, we know the curvature, overtwist etc, so we can update the step size
        ! for the scoring function
        !
        call fmc_update_step_size(self,fil)


        if (debug) then
            write(*,'(a)') '**debug(fmc_eval_slave): filament info after model was updated:'
            call fil%PrintInfo()
        endif



        ! Will we return minus one?
        return_minus_one = .false.
        if (return_minus_one_if_not_monotonic .and. .not. self%wps_as_sps) then
            if (.not. filament_splines_are_monotonic(fil)) then
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a)') '**WARNING(fmc_eval_slave): non-monotonic phi or subunit_num splines. Will return -1.0.'
                endif
                return_minus_one = .true.
            endif
        endif
        if (return_minus_one_if_any_negatives_coos) then
            if (        any(fil%wp%coo(1) .lt. 0.0d0) &
                .or.    any(fil%wp%coo(2) .lt. 0.0d0)) then
                return_minus_one = .true.
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a)') '**warning(fmc_eval_slave): negative coordinates. Will return -1.0.'
                endif
            endif
        endif
        if (return_minus_one_if_coos_beyond_edges) then
            if (    any(fil%wp%coo(1) .gt. real(fil%film%GetLogicalDimension(1))) &
                .or.any(fil%wp%coo(2) .gt. real(fil%film%GetLogicalDimension(2))) ) then
                return_minus_one = .true.
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a)') '**warning(fmc_eval_slave): coordinates beyond edges of micrograph. Will return -1.0.'
                endif
            endif
        endif

        ! debug
        if (ddetailled_score_info .and. (this_program%IsVerbose() .or. debug)) then
            write(*,'(a)') '**info(fmc_eval_slave): changes relative to original filament'
            call filament_print_diff(self%fil_original,fil)
        endif

        ! Will we use a single ctf for the whole filament, or one per segment?
        if (single_ctf_per_filament_if_reasonable) then
            if (filament_defocus_range(fil,ignore_fal_wps=.not.fil%fal_wps_in_score) .lt. self%search_accuracy(6)) then
                !if (debug) then
                    write(*,'(a)') '**debug(fmc_eval_slave): defocus variations along filament less than desired z accuracy. '//&
                                    'Will compute only one ctf for the whole filament.'
                !endif
                oone_ctf_for_whole_filament = .true.
            else
                oone_ctf_for_whole_filament = .false.
            endif
        else
            oone_ctf_for_whole_filament = .false.
        endif


        ! If only one or a few waypoints are to be refined, we should only get a path between neighbouring waypoints
        ! Note that we assume the waypoints which are to be refined are contiguous
        if (oonly_score_refining_neighbourhood) then
            call filament_t_interval_for_scoring_function(fil,start_wp,finish_wp)
        else
            if (.not. fil%fal_wps_in_score) then
                start_wp = 2
                finish_wp = fil%nwp-1
            else
                start_wp = 1
                finish_wp = fil%nwp
            endif
        endif
        ! Convert to t values
        start_t = fil%t(start_wp)
        finish_t = fil%t(finish_wp)

		if (debug) then
			write(*,'(a,2(F0.4,1x))') '**debug(fmc_eval_slave): scoring only in refinement neighbourhood. Start, finish t = ', &
                                                                                                        start_t, finish_t
		endif

        ! Sanity check
        if (start_t .gt. finish_t) then
            write(*,'(a,2(F0.6,1x))') '**ERROR(fmc_eval_slave): sanity check failed: start_t is greater than finish_t: ', &
                                    start_t, finish_t
            call terminate('Failed sanity check in FMC_EVAL_SLAVE')
        endif

        ! Get path
        if (self%wps_as_sps) then
            fil%wp_score =  fil%wp(start_wp:finish_wp)
        else
            call filament_get_path( fil,fil%wp_score,self%step,self%step_in_pixels, &
                                    start_wp=start_wp,finish_wp=finish_wp,intg_subunit_num=self%project_helical_subunits)
        endif
        number_of_scoring_waypoints = size(fil%wp_score)
        write(number_of_scoring_waypoints_c,'(i6)') number_of_scoring_waypoints

        ! If any of the parameters are not full-filament, we need to ensure there is at least one scoring waypoint per refinement waypoint.
        if (fil%fragmented_phi .or. fil%fragmented_subunit_num) then
            if (ccheck_each_wp_has_a_scoring_wp) then
                allocate(wp_has_score_wp_nearby(fil%nwp))
                wp_has_score_wp_nearby = .false.
                if (.not. fil%fal_wps_in_score) then
                    ! We don't really need the first and last waypoint
                    wp_has_score_wp_nearby(1) = .true.
                    wp_has_score_wp_nearby(fil%nwp) = .true.
                endif
                do scoring_waypoint_counter=1,size(fil%wp_score)
                    call nearest_wp_from_t(fil,fil%wp_score(scoring_waypoint_counter)%path_length,j,dummy)
                    wp_has_score_wp_nearby(j) = .true.
                enddo
                if (.not. all(wp_has_score_wp_nearby)) then
                    if ( all(wp_has_score_wp_nearby(1:fil%nwp-1) .and. fil%fal_wps_in_score)) then
                        ! Only the last waypoint doesn't have a scoring waypoint near it
                        write(*,'(a)') '**info/debug(fmc_eval_slave): adding scoring waypoint at last refinement waypoint'
                        call add_waypoint_to_path(fil%wp_score,fil,fil%t(fil%nwp))
                    else if (all(wp_has_score_wp_nearby(1:fil%nwp-2) .and. .not. fil%fal_wps_in_score)) then
                        call add_waypoint_to_path(fil%wp_score,fil,fil%t(fil%nwp-1))
                    else
                        write(*,'(a)',advance='NO') '**WARNING(fmc_eval_slave): not all refinement '//&
                                                    'waypoints have a scoring waypoint nearby'
                        write(*,*) wp_has_score_wp_nearby
                        !call terminate('Fatal error in FMC_EVAL_SLAVE')
                    endif
                endif
            endif
        endif


        ! Get the ft of the model if we're working with slice extraction rather than projection
        if (self%real_space_proj) then
            if (.not. self%model3dvol_private%ObjectIsCenteredInBox()) call self%model3dvol_private%SwapRealSpaceQuadrants()
            if (self%model3dvol_private%IsInFourierSpace()) call self%model3dvol_private%BackwardFFT()
        else
            if (self%model3dvol_private%IsInRealSpace()) call self%model3dvol_private%ForwardFFT()
            if (self%model3dvol_private%object_is_centered_in_box) call self%model3dvol_private%SwapRealSpaceQuadrants()
        endif

        ! Maximum radius of object, in pixels
        if (self%wps_as_sps) then
            obj_max_rad = max(self%sp_segment_length/fil%film%GetPixelSize()/2.0,fil%max_radius)
        else
            obj_max_rad = fil%max_radius
        endif


        ! Prepare synthetic micrograph
        if (make_synth_micrograph) then
            call synth_micrograph%Allocate(self%micrograph_frames(1))
            synth_micrograph = 0.0e0
        endif



        !
        !   Move along the path to calculate the score
        !

        ! We will be keeping track of how many pixels contributed to the scoring function
        number_of_pixels_in_scoring_function = 0.0

        ! Dimensions of rectangular mask to be imposed on boxes & projections
        rect_dim = (/ min(real(self%step,kind=4),2.*obj_max_rad)+soft_edge,(obj_max_rad+soft_edge)*2./)
        if (self%wps_as_sps) rect_dim(1) = self%sp_segment_length/self%fil%film%getpixelsize()+soft_edge
        proj_max_rad = sqrt((rect_dim(1)*.5)**2 + (rect_dim(2)*.5)**2)


        !
        ! Decide whether we're going to use threading
        !
        tthreading = .false.
        num_local_threads = 1
#ifdef fmc_eval_thread
        ! We won't do threading if we are already within a parallel region (for example if we're being
        ! called by fmc_bf_tilt)
        if ( .not. omp_in_parallel()) then
            if (omp_get_max_threads() .gt. 1) then
                tthreading = .true.
            endif
        endif
        ! If the user specified we should not do threading with the relevant argument,
        ! we definitely won't do threading
        if (tthreading .and. present(threading)) then
            if (.not. threading) tthreading = .false.
        endif

        !
        ! OpenMP parallel region
        !
        !$omp   parallel    if(tthreading)  &
        !$omp               default(shared)   &
        !$omp               private(scoring_waypoint_counter,icoo,rcoo,local_thread_num,score_local,box_is_blank,cc3_nd_rings_local)            &
        !$omp               private(matrix,ctf,wp_phased)                            &
        !$omp               private(model_z_shift) &
        !$omp               private(score_nd_local,cc3_nd_rings_whole_private,first_time,score_nd_whole_private,temp_cutout_from_frame) &
        !$omp               private(coordinates_for_current_wp_on_current_frame)
#endif


        ! See how many threads have been started (note that single has an implicit barrier on exit)
#ifdef fmc_eval_thread
        !$omp   single
#endif
#ifdef _OPENMP
        if (tthreading) num_local_threads = omp_get_num_threads()
#endif
        if (num_local_threads .gt. 1 .and. ddetailled_score_info) then
            write(*,'(a,i0,a)') '**info(fmc_eval_slave): using ', num_local_threads, ' OpenMP threads.'
        endif
        ! Initialise shared variables
        loc_count           = 1
        score               = 0.0
        score_nd_whole      = [0.0,0.0,0.0]

        ! Ensure that the fmc object has the necessary workspace memory allocated
        call fmc_set_num_work_areas(self,num_local_threads)
        if (debug) then
            write(*,'(a,i0)') '**debug(fmc_eval_slave): num local threads = ', num_local_threads
            write(*,'(a,i0)') '**debug(fmc_eval_slave): num work areas    = ', self%num_work_areas
        endif

#ifdef fmc_eval_thread
        !$omp   end single
#endif

        ! Initialise private variables
        score_local         = 0.0
        score_nd_local      = [0.0,0.0,0.0]
        matrix              = identity_rotmat()
        first_time          = .true. ! This is true for the first location encountered by a thread
        local_thread_num    = 1

        ! Make private copies of arrays that need to be "reduced"
        score_nd_whole_private = score_nd_whole

        ! Loop over the scoring waypoints
#ifdef fmc_eval_thread
        !$omp   do  reduction(+:number_of_pixels_in_scoring_function) &
        !$omp       reduction(+:score)
#endif
        do scoring_waypoint_counter=1,number_of_scoring_waypoints

#ifdef fmc_eval_thread
            local_thread_num = omp_get_thread_num()+1

            if (debug) then
                print '(4(A,i0))', '**debug(fmc_eval_slave): thread ', local_thread_num,' of ', num_local_threads,    &
                                    ' is on step ', scoring_waypoint_counter,' of ', number_of_scoring_waypoints
            endif
#endif

            ! Because the waypoint doesn't necessarily have an integer subunit_num value,
            ! we'll need to shift & rotate the reference to accomodate for the non-integer part of subunit_num
            ! Below, the phi angle of the waypoint is modified, and we also compute a z shift which will need to be
            ! applied to the 3d model before projection
            wp_phased = fil%wp_score(scoring_waypoint_counter)
            call filament_model_phasing_from_wp(fil,wp_phased,model_z_shift=model_z_shift,local_h_sym=.true., &
                                                model3d_oversampling=1) ! Model3d oversampling will be taken into account by the projection_prepare code


            !
            ! Hextend for this wp?
            !
            if (.not. self%project_helical_subunits .and. hextend_per_wp) then

                if (.not. self%model3dvol_private%ObjectIsCenteredInBox()) call self%model3dvol_private%SwapRealSpaceQuadrants()
                if (self%model3dvol_private%IsInFourierSpace()) call self%model3dvol_private%BackwardFFT()
                call self%model3dvol_private%WriteToDisk('dbg_before_hextend.mrc')
                ! 150603: looks like there's a bug with this hextend_per_wp option - the projections from this hextended volumes look completely different than they should
                call image_impose_helical_symmetry_new(self%model3dvol_private,self%fil%helix,self%interpol,self%fil%max_radius)
                call self%model3dvol_private%WriteToDisk('dbg_after_hextend.mrc')
                call self%model3dvol_private%ForwardFFT()
                call self%model3dvol_private%SwapRealSpaceQuadrants()
                !call terminate('debugging fmc_eval_slave')
            endif


            !
            ! Box out area surrounding the current waypoint
            !
            call fmc_box_out_from_micrograph(self,wp_phased,local_thread_num)
            ! Check the image isn't blank
            box_is_blank = self%box(local_thread_num)%IsConstant()
            if (box_is_blank) then
                write(*,'(2a)') '**warning(fmc_eval_slave): image boxed from micrograph was blank',    &
                                    '. Will return a score of -1.0'
                return_minus_one = .true.
            endif

            ! If necessary, deal with ctf
            if (oone_ctf_for_whole_filament) then
                call terminate('Not implemented yet')
            else
                ! Get the appropriate ctf parameters
                if (get_ctf_from_z) then
                    ctf = fil%film%ctftilt%ctf
                    call ctf_ang_to_pix(ctf,fil%film%GetPixelSize())
                    call ctf_at_z(ctf,fil%wp_score(scoring_waypoint_counter)%coo(3))
                else
                    ctf = fil%film%ReturnCTFAtCoos(fil%wp_score(scoring_waypoint_counter)%coo(1), &
                                                   fil%wp_score(scoring_waypoint_counter)%coo(2))
                endif

                ! Compute the ctf image
                call ctf_to_img(ctf,self%box_ctf(local_thread_num),fil%film%GetPixelSize(), &
                                one_at_origin=self%boost_low_frequency_contrast)
            endif


            ! Prepare a projection image
            wp_phased%coo = 0.0d0 ! Since we don't need to shift the projection away from the center of the box
            call projection_prepare(self%model3dvol_private,self%proj(local_thread_num),&
                                    self%box(local_thread_num)%logical_dimensions, &
                                    wp_phased,self%box_ctf(local_thread_num),   &
                                    pixel_size=self%fil%film%GetPixelSize(), &
                                    proj_max_rad=proj_max_rad,   &
                                    interpol=self%interpol,   &
                                    win=self%interp_win_3d(local_thread_num), &
                                    env_mask=self%mask_env(local_thread_num), &
                                    real_space_proj=self%real_space_proj,     &
                                    vol_shift_z=model_z_shift,apply_disc_mask=.false.,  &
                                    freq_max=self%freq_max, &
                                    rect_mask_width=fil%max_radius*2., &
                                    segment_with_helical_subunits=self%project_helical_subunits, &
                                    helix=self%helix, &
                                    helix_max_rad=self%fil%max_radius, &
                                    volume_oversampling=self%model3d_oversampling)

            if (self%proj(local_thread_num)%IsConstant()) write(*,'(a)') '**WARNING(fmc_eval_slave): blank projection'

            ! Prepare the image
            call image_prepare( self%box(local_thread_num),self%box_tmp(local_thread_num),self%box_ctf(local_thread_num), &
                                obj_max_rad, &
                                apply_ctf=.not. ccompute_logp, &
                                env_mask=self%mask_env(local_thread_num), &
                                real_space_shift=self%real_space_box_shift, &
                                interpol=self%interpol,interpol_window=self%interp_win_box_shift(local_thread_num),   &
                                apply_disc_mask=.false.)

            ! Apply rectangular masks to both
            if (.not. self%proj(local_thread_num)%is_in_real_space) call self%proj(local_thread_num)%BackwardFFT()
            call self%proj(local_thread_num)%ApplySoftRectangularMask( rect_dim,real(wp_phased%euler%psi), &
                                falloff=soft_edge,wipe_input=.false.,mask_with_average_edge_value=.true.)
            if (.not. self%box(local_thread_num)%is_in_real_space) call self%box(local_thread_num)%BackwardFFT()
            call self%box(local_thread_num)%ApplySoftRectangularMask(    rect_dim,real(wp_phased%euler%psi), &
                                falloff=soft_edge,wipe_input=.false.,mask_with_average_edge_value=.true.)


            ! If we are going to compute log probability, we need a separate image of the mask
            if (ccompute_logp) then
                call self%mask_env(local_thread_num)%ApplySoftRectangularMask(rect_dim,real(wp_phased%euler%psi), &
                                falloff=soft_edge,wipe_input=.false.,mask_with_average_edge_value=.true.)
            endif

            if (estimate_noise_variance) then
                ! From the projection mask, we can work out how many pixels will contribute to the scoring function
                call self%mask_env(local_thread_num)%ApplySoftRectangularMask( rect_dim,real(wp_phased%euler%psi), &
                                    falloff=soft_edge,wipe_input=.false.,mask_with_average=.false.)
                number_of_pixels_in_scoring_function =  number_of_pixels_in_scoring_function &
                                                        + self%mask_env(local_thread_num)%GetSumOfValues()
            endif

            ! Synthetic micrograph
            if (make_synth_micrograph) then
                !$omp critical (fmc_synth_micrograph)
                call self%proj(local_thread_num)%AddInto(synth_micrograph,icoo)
                !$omp end critical (fmc_synth_micrograph)
            endif

            ! The projection and the image need to be in Fourier space before we measure correlation
            if (self%proj(local_thread_num)%IsInRealSpace()) call self%proj(local_thread_num)%ForwardFFT()
            if (self%box(local_thread_num)%IsInRealSpace()) call self%box(local_thread_num)%ForwardFFT()

            ! Prepare a Fourier mask for the amyloid case
            ! todo: move this to fmc_init or somewhere similar
            if (self%cross_beta_filter) then
                call fourier_mask_init_amyloid(         self%ampl_mask(local_thread_num),                   &
                                                        self%box(local_thread_num),fil%film%GetPixelSize(),     &
                                                        fil%max_radius  *   fil%film%GetPixelSize(),            &
                                                        fil%film%GetPixelSize() /   self%freq_max,              &
                                                        wp_phased%euler)
            endif

            ! Measure correlation or log probability
            if (ccompute_logp) then
                ! We are measuring log probability
                call self%box(local_thread_num)%ComputeFrealign9Probability(self%proj(local_thread_num),   &
                                                                            self%mask_env(local_thread_num), &
                                                                            fil%wp_score(scoring_waypoint_counter)%noise_sigma, &
                                                                            fil%wp_score(scoring_waypoint_counter)%logp)
            else
                ! We are measuring cross-correlation-like score
                if (debug) write(*,'(a,F0.3,a,l1)') '**debug(fmc_eval_slave): about to calc correlation with freq_max = ', &
                                                    self%freq_max, ' and whole-filament norm = ', wwhole_filament_norm
                if (.not. normalised_scoring_function) then
                    write(*,'(a)') '**ERROR(fmc_eval_slave): unnormalised scoring function not implemented yet'
                    call terminate('Fatal error in FMC_EVAL_SLAVE')
                endif
                if (wwhole_filament_norm) then
                    select case (self%type_of_scoring_function)
                    case (cubed_correlation)
                        if (self%cross_beta_filter) then
                            cc3_nd_rings_local = cclin_num_denoms_rings(self%box(local_thread_num),self%proj(local_thread_num),  &
                                                                        fourier_mask=self%ampl_mask(local_thread_num)     )
                        else
                            cc3_nd_rings_local = cclin_num_denoms_rings(self%box(local_thread_num),self%proj(local_thread_num),  &
                                                    rad_max=real(self%box(local_thread_num)%logical_dimensions(1))*self%freq_max,&
                                                    rad_min=real(self%box(local_thread_num)%logical_dimensions(1))*self%freq_min )
                        endif
                        if (first_time) then
                            ! The first time around, cc3_nd_rings_whole hasn't yet been allocated yet, so we can't have it on the rhs of the assignment
                            cc3_nd_rings_whole_private = cc3_nd_rings_local
                        else
                            cc3_nd_rings_whole_private = cc3_nd_rings_whole_private + cc3_nd_rings_local
                        endif
                        first_time = .false.
                        score_local = cc3_finalise( cc3_nd_rings_local,  &
                                                    self%box(local_thread_num),self%proj(local_thread_num),   &
                                                    rad_max=real(self%box(local_thread_num)%logical_dimensions(1))*self%freq_max, &
                                                    rad_min=real(self%box(local_thread_num)%logical_dimensions(1))*self%freq_min)
                        if (estimate_noise_variance) then
                            score_nd_local = cclin_num_denoms_from_rings(cc3_nd_rings_local)
                            score_nd_whole_private = score_nd_whole_private + score_nd_local
                        endif
                    case (linear_correlation)
                        if (self%cross_beta_filter) then
                            call cclin_num_denoms(  self%box(local_thread_num),self%proj(local_thread_num),   &
                                                                score_nd_local, &
                                                                fourier_mask=self%ampl_mask(local_thread_num)   &
                                                                )
                        else
                            call cclin_num_denoms(  self%box(local_thread_num),self%proj(local_thread_num),   &
                                                             score_nd_local, &
                                                rad_max=real(self%box(local_thread_num)%logical_dimensions(1))*self%freq_max,   &
                                                rad_min=real(self%box(local_thread_num)%logical_dimensions(1))*self%freq_min)
                        endif
                        score_nd_whole_private = score_nd_whole_private + score_nd_local
                        if (score_nd_local(2) .le. 0.0D0 .or. score_nd_local(3) .le. 0.0D0) then
                            ! i believe this only happens when there is a blank box
                            score_local = -999.0D0
                            if (.not. box_is_blank .and. score_nd_local(2) .le. 0.0D0) then
                                ! Something must have gone wrong, let's dump some diagnostics and crash out
                                if (.not. self%box(local_thread_num)%is_in_real_space) then
                                    call self%box(local_thread_num)%BackwardFFT()
                                endif
                                if (.not. self%proj(local_thread_num)%is_in_real_space) then
                                    call self%proj(local_thread_num)%BackwardFFT()
                                endif
                                call self%box(local_thread_num)%WriteToDisk('dbg_box.mrc')
                                call self%proj(local_thread_num)%WriteToDisk('dbg_proj.mrc')
                                call self%mask_env(local_thread_num)%WriteToDisk('dbg_mask_env.mrc')
                                call wp_print_info(wp_phased)
                                write(*,'(a)') '**ERROR(fmc_eval_slave): this is strange - '
                                write(*,'(a)') '**ERROR(fmc_eval_slave): i was getting cc denominator terms <= 0.0 '//&
                                                                        'yet the box isn''t blank. '//&
                                                                        'See dbg_*mrc files dumped to current directory.'
                                call terminate('Thought box was blank')
                            else
                                write(*,'(a)') '**warning(fmc_eval_slave): strange values because of blank box'
                            endif
                        else
                            score_local = real(score_nd_local(1)/sqrt(score_nd_local(2)*score_nd_local(3)))
                        endif
                    case (ssnr_weighting)
                        ! The Frealign 9 way of scoring
                        call terminate('ssnr_weighting scoring function not implemented')
                    case default
                        write(*,'(a,i0)') '**error(fmc_eval_slave): unknown type of scoring function: ', &
                                            self%type_of_scoring_function
                        call terminate('Uknown type of scoring function')
                    end select
                else ! We're not doing whole-filament normalisation
                    select case (self%type_of_scoring_function)
                    case (cubed_correlation)
                        if (self%cross_beta_filter) then
                            score_local = cc3(  self%box(local_thread_num),self%proj(local_thread_num),   &
                                                fourier_mask=self%ampl_mask(local_thread_num)   &
                                                )
                        else
                            score_local = cc3(    self%box(local_thread_num),self%proj(local_thread_num),   &
                                                rad_max=real(self%box(local_thread_num)%logical_dimensions(1))*self%freq_max,   &
                                                rad_min=real(self%box(local_thread_num)%logical_dimensions(1))*self%freq_min)
                        endif
                    case(linear_correlation)
                        if (self%cross_beta_filter) then
                            score_local = cclin(    self%box(local_thread_num),self%proj(local_thread_num),   &
                                                    fourier_mask=self%ampl_mask(local_thread_num)   &
                                                )
                        else
                            score_local = cclin(    self%box(local_thread_num),self%proj(local_thread_num),   &
                                                rad_max=real(self%box(local_thread_num)%logical_dimensions(1))*self%freq_max,   &
                                                rad_min=real(self%box(local_thread_num)%logical_dimensions(1))*self%freq_min)
                        endif
                    case default
                        write(*,'(a)') '**error(fmc_eval_slave): unknown type of scoring function'
                        call terminate('Uknown type of scoring function')
                    end select
                    score = score + score_local
                endif

                ! Remember the local correlation score
                fil%wp_score(scoring_waypoint_counter)%score = score_local
                if (self%wps_as_sps) fil%wp(scoring_waypoint_counter)%score = score_local

                !
                ! Check for NaN
                !
#ifdef __GFORTRAN__
                if (isnan(score_local)) then
#else
                if (ieee_is_nan(score_local)) then
#endif
                    ! Check whether box is blank at this point
                    call self%box(local_thread_num)%BackwardFFT()
                    box_is_blank = self%box(local_thread_num)%IsConstant()
                    if (box_is_blank) then
                        write(*,'(2a)') '**WARNING(fmc_eval_slave): image boxed from micrograph was blank, ',    &
                                        'leading to a non-sensical score. Will return a score of -1.0'
                        return_minus_one = .true.
                    else
                        write(*,'(2a,F8.4)') '**ERROR(fmc_eval_slave): intermediate score is not a number ,',    &
                                            '(but boxed image was not blank): ', score_local
                        if (wwhole_filament_norm) then
                            write(*,'(a,3(F8.4,1x))') '**debug(fmc_eval_slave): numerator, denominator1, denominator2: ', &
                                                    score_nd_local(1:3)
                        endif
                        write(*,*) '**debug(fmc_eval_slave): ctf = ', ctf
                        write(*,'(a,3(i6,1x))') '**debug(fmc_eval_slave): icoo = ', icoo(1:3)
                        write(*,'(a,3(F7.4,1x))') '**debug(fmc_eval_slave): rcoo = ', rcoo(1:3)
                        write(*,'(a)') '**debug(fmc_eval_slave): dumping box and projection to disk as dbg_box.mrc and dbg_proj.mrc'
                        call self%box(local_thread_num)%WriteToDisk('dbg_box.mrc')
                        call self%proj(local_thread_num)%WriteToDisk('dbg_proj.mrc')
                        call self%box_ctf(local_thread_num)%WriteToDisk('dbg_box_ctf.mrc')
                        call terminate('Fatal error in FMC_EVAL_SLAVE')
                    endif
                endif ! End of test for NaN local score
            endif ! End of test for score or log probability

            ! If requested, write to stack
            if (wwrite_stacks) then
                if (.not. self%box(local_thread_num)%is_in_real_space) call self%box(local_thread_num)%BackwardFFT()
                if (.not. self%proj(local_thread_num)%is_in_real_space) call self%proj(local_thread_num)%BackwardFFT()
#ifdef fmc_eval_thread
                !$omp   critical        (write_stacks_to_disk_omp_crit)
#endif
                call self%box(local_thread_num)%WriteToDisk(boxes_file,wanted_location=loc_count)
                call self%proj(local_thread_num)%WriteToDisk(projs_file,wanted_location=loc_count)
				loc_count = loc_count + 1
#ifdef fmc_eval_thread
                !$omp   end critical    (write_stacks_to_disk_omp_crit)
#endif
            endif

        enddo   ! End loop over path waypoints
#ifdef fmc_eval_thread
        !$omp   enddo
#endif

        if (debug) then
            write(*,'(a)') '**debug(fmc_eval_slave): filament info after loop over waypoints:'
            call fil%PrintInfo()
        endif

        !
        ! Let's do our manual array reductions
        !
        if (.not. ccompute_logp) then
#ifdef fmc_eval_thread
            !$omp   critical        (fmc_eval_manual_reductions)
#endif
            if (self%type_of_scoring_function .eq. cubed_correlation) then
                cc3_nd_rings_whole = cc3_nd_rings_whole + cc3_nd_rings_whole_private
            endif
            score_nd_whole = score_nd_whole + score_nd_whole_private
#ifdef fmc_eval_thread
            !$omp   end critical    (fmc_eval_manual_reductions)
            ! Make sure we've done all the array reductions before moving on
            !$omp barrier
#endif
        endif


        !!!!! TODO: finish logp implementation
        !!!
        !!
        !  - sum LogPs from the scoring waypoints to get the joint probability for the filament
        !  - update the filament waypoints with logp values
        !  - decide what to do about noise variance, curvature penalties etc. (probably just bypass them)


        !
        ! Let's finalise the score
        !
#ifdef fmc_eval_thread
        !$omp   single
#endif
        if (estimate_noise_variance) then
            if (return_minus_one) then
                ! This only happens when boxes are blank (negative coordinates etc),
                ! i.e. there is no signal
                estimated_noise_variance  = 10000000.0e0
                estimated_signal_variance = 0.0e0
            else
                if (debug) then
                    write(*,'(a,F0.2)') '**debug(fmc_eval_slave): number of pixels in scoring function = ', &
                                        number_of_pixels_in_scoring_function
                    write(*,'(a,i0)') '**debug(fmc_eval_slave): total number of pixels = ', &
											number_of_scoring_waypoints*self%box(1)%logical_dimensions(1)*self%box(1)%logical_dimensions(2)
                endif
                estimated_noise_variance = real(score_nd_whole(2) - (score_nd_whole(1)**2/score_nd_whole(3)))  &
                                                        / number_of_pixels_in_scoring_function
                estimated_signal_variance = real(score_nd_whole(1)/score_nd_whole(3)**2)**2 ! alpha, the scaling factor for the reference projection, squared
                estimated_signal_variance = estimated_signal_variance * score_nd_whole(2)**2 / number_of_pixels_in_scoring_function
                if (estimated_noise_variance .lt. 0.0) then
                    write(*,'(a,F0.4)') '**ERROR(fmc_eval_slave): negative variance per pixel! ', estimated_noise_variance
                    call terminate('Fatal error in FMC_EVAL_SLAVE')
                endif
                if (normalised_scoring_function) then
                    ! Since we are actually using a normalised cross-correlation for the scoring function,
                    ! and not just the cross-product of the noisy image with the reference as in Sigworth 1998 Eq 21,
                    ! we modify the estimate of the variance to include the normalisation.
                    ! See notes of 8-July-2011 for derivation.
                    !estimated_noise_variance_per_pixel =   estimated_noise_variance_per_pixel / &
                    !                                       (score_nd_whole(1) * sqrt(score_nd_whole(2)/score_nd_whole(3)))
                    ! New derivation, 9-July-2011:
                    estimated_noise_variance  =   estimated_noise_variance  / (sqrt(score_nd_whole(2))*sqrt(score_nd_whole(3)))
                    estimated_signal_variance =   estimated_signal_variance / (sqrt(score_nd_whole(2))*sqrt(score_nd_whole(3)))
                endif
            endif
        endif
        if (return_minus_one) then
            score = -1.0
        else
            if (wwhole_filament_norm) then
                select case (self%type_of_scoring_function)
                    case (cubed_correlation)
                        score = cc3_finalise(   cc3_nd_rings_whole,self%box(local_thread_num),self%proj(local_thread_num),    &
                                                rad_max=real(self%box(local_thread_num)%logical_dimensions(1))*self%freq_max,   &
                                                rad_min=real(self%box(local_thread_num)%logical_dimensions(1))*self%freq_min)
                    case (linear_correlation)
                        score = real(score_nd_whole(1)/sqrt(score_nd_whole(2)*score_nd_whole(3)))
                    case default
                        write(*,'(a)') '**error(fmc_eval_slave): unknown type of scoring function'
                        call terminate('Unknown type of scoring function')
                end select
            else
                score = score/number_of_scoring_waypoints
            endif
        endif
#ifdef fmc_eval_thread
        !$omp   end single
        ! Deallocate objects which were private and allocated dynamically within the parallel region
        if (allocated(cc3_nd_rings_local)) deallocate(cc3_nd_rings_local)
        if (allocated(cc3_nd_rings_whole_private)) deallocate(cc3_nd_rings_whole_private)
        !$omp single
#endif
        ! This is actually a shared object, so we only deallocate it one of the threads
        if (allocated(cc3_nd_rings_whole)) deallocate(cc3_nd_rings_whole)
#ifdef fmc_eval_thread
        !$omp end single

        ! The end of the parallel region is here, because cc3_nd_rings_whole is only allocated within it,
        ! and is needed for the score calculate, above
        !$omp   end parallel
#endif

        ! The detailled scoring info can be printed out now
        if (ddetailled_score_info) then
            print *
            print '(a,i0,a)', '**info(fmc_eval_slave): detailled score information along path of filament ', fil%id, ':'
            if (oonly_score_refining_neighbourhood .and. .not. fil%wps_as_sps .and. .not. self%wps_as_sps) then
                write(*,'(3(A,F0.4),a)')    '(region from t = ', start_t, ' to t = ', finish_t, &
                                            ' used to compute score, in steps of ', self%step ,')'
            endif
            call waypoints_print_info(fil%wp_score)
            write(*,'(a,F0.5)') '**info(fmc_eval_slave): score = ', score
        endif


        !
        ! Check for NaN
        !
#ifdef __GFORTRAN__
        if (isnan(score)) then
#else
        if (ieee_is_nan(score)) then
#endif
            write(*,'(a,F0.4)') '**ERROR(fmc_eval_slave): final score is not a number: ', score
            call terminate('Fatal error in FMC_EVAL_SLAVE')
        endif

        if (debug) print '(a,F0.5)', '**debug(fmc_eval_slave): score = ', score


        !
        ! Apply curvature, overtwist and/or stretching restraints
        !
        if (self%curvature_restraint .or. self%torsion_restraint .or. self%stretch_restraint) then
            ! Diagnostic output
            if (ddetailled_score_info .or. debug) then
                write(*,'(a,F0.4)') '**info(fmc_eval_slave): score before penalties = ', score
                write(*,'(a,F0.4)') '**info(fmc_eval_slave): estimated snr (Bershad & Rockmore estimator) = ', &
                                      score / (1.0-SCORE) ! Bershad & Rockmore 1974
                write(*,'(2(A,G15.4),a)') '**info(fmc_eval_slave): With an estimated noise variance per pixel of ', &
                            estimated_noise_variance, ' , (estimated snr ', &
                            estimated_signal_variance / estimated_noise_variance, &
                            !score_nd_whole(1)**2 / estimated_noise_variance_per_pixel, &
                            '), the following restraints lead to score penalties: '
                write(*,'(4(A20,1x))') 'Property', 'Total along filament', 'Persistence length', 'Penalty'
            endif
            ! For testing purposes, we can either do this properly or use a fudge factor
            if (estimate_noise_variance) then
                penalty_factor = estimated_noise_variance
            else
                penalty_factor = 0.1
            endif
            ! Compute the squared curvature, overtwist and stretch integrated over the filament's length
            call filament_eval_total_curvature_overtwist_stretch(   fil,  &
                                                                    curvature,torsion,stretch,   &
                                                                    ignore_fal_wps=.not.fil%fal_wps_in_score)
            ! Compute the penalties
            curvature_penalty = curvature * 0.5 * penalty_factor * fil%persistence_length
            torsion_penalty   = torsion   * 0.5 * penalty_factor * fil%torsion_persistence_length
            stretch_penalty   = stretch   * 0.5 * penalty_factor * fil%stretch_persistence_length
            ! Debug
            if (curvature_penalty .gt. 100.0) then
                write(*,'(a,G17.10)') '**debug/warning(fmc_eval_slave): very large curvature penalty = ', curvature_penalty
            endif
            ! Apply the penalties
            score = score - curvature_penalty - torsion_penalty - stretch_penalty
            ! Diagnostic output
            if (ddetailled_score_info .or. debug) then
                write(*,'(a20,1x,3(G20.4,1x))') 'Curvature', curvature, fil%persistence_length,         curvature_penalty
                write(*,'(a20,1x,3(G20.4,1x))') 'Overtwist', torsion,   fil%torsion_persistence_length, torsion_penalty
                write(*,'(a20,1x,3(G20.4,1x))') 'Stetch',    stretch,   fil%stretch_persistence_length, stretch_penalty
                write(*,'(a)') ' '
            endif
        endif


        ! The filament keeps track of its own score
        fil%score = score
        if (.not. self%wps_as_sps) fil%wp%score = score

        ! Synth micrograph
        if (make_synth_micrograph) then
            call synth_micrograph%WriteToDisk(synth_micrograph_fn)
            call synth_micrograph%Deallocate()
        endif

        ! Close files
        if (wwrite_stacks) then
            if (debug) then
                print '(a)', '**debug(fmc_eval_slave): closing these files'
                call boxes_file%PrintInfo()
                call projs_file%PrintInfo()
            endif
            call boxes_file%close()
            call projs_file%close()
        endif

        ! Explicit deallocations
        if (allocated(cc3_nd_rings_whole)) deallocate(cc3_nd_rings_whole)
        if (allocated(cc3_nd_rings_local)) deallocate(cc3_nd_rings_local)

        if (ddetailled_score_info .and. debug) then
            write(*,'(a)') '**debug(fmc_eval_slave): all done'
        endif


    end function fmc_eval_slave




end module fmcs_evaluation
