!>  \brief  Module to interact with frealign parameter files
!!
module frealign_parameters

    contains

    !>  \brief  Write a frealign parameter file from an array of single particles
    subroutine sp_to_frealign_par(sp,filename,mag,pixel_size)
        use single_particles
        use runtime_parameters, only : available_lun, lun_is_now_available, terminate
        use rotation_matrices, only : degrees
        use parsers, only : file_exists
        implicit none
        ! Arguments
        type(single_particle),  allocatable,    intent(in)  ::  sp(:)       !<  Array of single particles
        character(len=*),                       intent(in)  ::  filename    !<  Filename of frealign parameter file to written out
        real,                                   intent(in)  ::  mag         !<  The magnification to be written to the Frealign parameter file
        real,                                   intent(in)  ::  pixel_size  !<  Pixel size in Angstroms
        ! Private variables
        integer ::  par_lun !<  Logical unit number for our parameter file
        integer ::  ios     !<  io status code
        integer ::  current_sp
        logical,    parameter :: debug = .false.
        ! Start work

        if (debug) then
            write(*,'(a)') '**debug(sp_to_frealign_par): sp info on entry: '
            call sps_print_info(sp)
        endif

        ! Open the file
        par_lun = available_lun()

        ! Check whether the file already exists, in which case we'll overwrite it
        if (file_exists(filename)) then
            write(*,'(2a)') '**WARNING(sp_to_frealign_par): parameter file already exists, will be overwritten : ', &
                                trim(adjustl(filename))
        endif
        open(unit=PAR_LUN,file=TRIM(adjustl(filename)),status='REPLACE',iostat=IOS)
        if (ios .ne. 0) then
            write(*,'(3a,i0)') '**ERROR(sp_to_frealign_par): file opening failed for file ', trim(adjustl(filename)), &
                                ' with error ', ios
            call terminate('Fatal error in SP_TO_FREALIGN_PAR')
        endif

        ! Write the header
        write(par_lun,'(a)')'C           PSI   THETA     PHI       SHX       SHY     MAG  FILM      DF1      DF2'//&
                            '  ANGAST     OCC      LogP      SIGMA   SCORE  CHANGE'

        ! Loop over the single particles
        if (debug) write(*,'(a)') '**debug(sp_to_frealign_par): about to start looping'
        do current_sp=1,size(sp)
            if (abs(sp(current_sp)%subunit_num - nint(sp(current_sp)%subunit_num)) .gt. 0.001) then
                write(*,'(a,i0,a,F0.3,a)') '**ERROR(sp_to_frealign_par): sp # ', current_sp,    &
                                ' has non-integer subunit_num of ', sp(current_sp)%subunit_num, &
                                '. Frealign parameter file cannot store this information. Retry with integer subunit_num values'
                call terminate('Fatal error in SP_TO_FREALIGN_PAR')
            endif
            if (.not. sp(current_sp)%ctf%pix_units) then
                write(*,'(a)') '**ERROR(sp_to_frealign_par): ctf needs to be given in pixel / radians units'
                call terminate('Fatal error in SP_TO_FREALIGN_PAR')
            endif
            write(par_lun,'(i7,1x,3(f7.2,1x),2(f9.2,1x),i7,1x,i5,1x,2(f8.1,1x),2(f7.2,1x),i9,1x,f10.4,1x,f7.2,1x,f7.2)') &
                                    current_sp, &
                                    sp(current_sp)%angles%psi, sp(current_sp)%angles%theta, mod(sp(current_sp)%angles%phi,360.D0), &
                                    sp(current_sp)%shift_x, sp(current_sp)%shift_y, &
                                    nint(mag), sp(current_sp)%film_num, &
                                    sp(current_sp)%ctf%dfmid1 * pixel_size, sp(current_sp)%ctf%dfmid2 * pixel_size, &
                                    degrees(sp(current_sp)%ctf%angast), &
                                    100.0, 0,0.5, &
                                    degrees(acos(sp(current_sp)%score)), 0.0
        enddo

        ! Close the file
        if (debug) write(*,'(a)') '**debug(sp_to_frealign_par): about to close output file'
        close(unit=PAR_LUN)
        call lun_is_now_available(par_lun)
        if (debug) write(*,'(a)') '**debug(sp_to_frealign_par): about to return'
    end subroutine sp_to_frealign_par

    !>  \brief  Parse a Frealign parameter file and return an array of single particles
    function frealign_par_to_sp(filename) result(sp)
        use single_particles
        use parsers
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        character(len=*),       intent(in)  ::  filename    !<  Filename of Frealign parameter file
        ! Result
        type(single_particle),  allocatable ::  sp(:)       !<  Output array of single-particle objects
        !  Private variables
        integer ::  num_sp, ierr
        real,   allocatable ::  tmp(:)
        integer, allocatable :: itmp(:)
        ! Start work
        ! Check the file exists & find out number of lines
        if (.not. file_exists(filename)) then
            write(*,'(a,a)') '**ERROR(frealign_par_to_sp): file does not exist: ', trim(filename)
            call terminate('Fatal error in FREALIGN_PAR_TO_SP')
        endif
        ! Number of particles
        num_sp = number_of_lines_in_file(filename,ignore_comments=.true.)
        ! Allocate
        allocate(sp(num_sp),tmp(num_sp),stat=ierr)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(frealign_par_to_sp): memory allocation failed with error # ', ierr
            call terminate('Fatal error in FREALIGN_PAR_TO_SP')
        endif
        ! Read data in
        call get_ints_from_column(filename,1,itmp)
        sp%loc_in_stack = itmp
        ! The Euler angles and the shifts are those that need to be applied to the 3d to obtain the single-particle,
        ! which is counter-intuitive to me, so i convert them here straight away!
        call get_reals_from_column(filename,2,tmp)
        sp%angles%psi = tmp
        call get_reals_from_column(filename,3,tmp)
        sp%angles%theta = tmp
        call get_reals_from_column(filename,4,tmp)
        sp%angles%phi = tmp
        ! Need to apply negative shifts to align particles with (unshifted) reference
        call get_reals_from_column(filename,5,tmp)
        sp%shift_x    = -1.0 * tmp
        call get_reals_from_column(filename,6,tmp)
        sp%shift_y    = -1.0 * tmp
        call get_reals_from_column(filename,8,tmp)
        sp%film_num   = tmp
        call get_reals_from_column(filename,9,tmp)
        sp%ctf%dfmid1 = tmp
        call get_reals_from_column(filename,10,tmp)
        sp%ctf%dfmid2 = tmp
        call get_reals_from_column(filename,11,tmp)
        sp%ctf%angast = tmp
        call get_reals_from_column(filename,12,tmp)
        sp%score      = tmp

    end function frealign_par_to_sp


end module frealign_parameters
