!>  \brief  Module to interact with frealix parameter files
!!
!!
!!  There are several versions of the frealix parameter files. For full details, see functions frealix_par_header, :
!!  - version 1 is from the frealign-like first incarnation of frealix, written by Grigorieff, Sindelar and Rohou
!!  - version 2 is from the modern incarnation, until 13-July-2010
!!  - version 3 is new as of July/August 2010. It incorporates a z coordinate and a path length for each waypoint as stores all coordinates in Angstroems rather than pixels
!!  - version 4 is new as of April 2011. It incorporates the subunit number, and may store coordinates in pixels or Angstroems
!!
module frealix_parameters

    ! The format of the frealix1 parameter file
    character(len=*), parameter   ::  flx1_par_format                     =   &
                                    '(i6,1x,3(F7.2,1x),2(F9.2,1x),(F7.1,1x),(i5,1x),2(F8.1,1x),'//&
                                    '2(F7.2,1x),(E12.5,1x),2(F7.2,1x),(i8,1x),F7.2)'
    character(len=6), parameter   ::  flx1_par_header(17)                 =   &
                                    ['WP#   ','AIP   ','A-OOP ','AA    ','SHX   ','SHY   ','MAG   ',&
                                    'FILM  ','DF1   ','DF2   ','ANGAST','OCC   ','LogP  ','SCORE ','CHANGE','IFIL  ','dR    ']
    character(len=*), parameter   ::  flx1_par_header_format              =   &
                                    '(a6,1x,3(a7,1x),2(a9,1x),(a7,1x),(a5,1x),2(A8.1X),2(A7.1X),(a12,1x),2(a7,1x),(a8,1x),(a7))'

    ! The format for the frealix2 parameter file
    character(len=*), parameter   ::  flx2_par_format                     =   '(3(I6,1x),2(F11.3,1x),3(F12.3,1x),(F8.3,1x),l7,1x)'
    character(len=*), parameter   ::  flx2_par_header_format              =   '(3(A6,1x),2(a11,1x),  3(a12,1x),  (a8,1x),  a7,1x)'

    ! The current latest version of the frealix2 parameter file
    integer,          parameter   ::  flx2_par_lastest_version            =   6
    ! If the parameter file does not specify which version it is, this version will be assumed
    integer,          parameter   ::  flx2_par_version_if_not_specified   =   2 ! since this was the last version which didn't specify it's own version number in the header




    contains

    !>  \brief  Return the header for the relevant frealix parameter file version
    subroutine frealix_par_header(par_header,flx_par_version,units,number_of_frames)
        use runtime_parameters, only : terminate
        use StringManipulations, only : IntegerToString
        implicit none
        ! Arguments
        integer,                            intent(in)  ::  flx_par_version
        character(*),       optional,       intent(in)  ::  units               !<  For versions 4 and above, we can be in Angstroems or pixels. The default is Angstroms, but this can be overriden here. If the first letter is p, the header will indicate pixels
        integer,            optional,       intent(in)  ::  number_of_frames    !<  For versions 5 and above, can store relative subframe shifts
        ! Result
#ifdef __GFORTRAN__
        character(len=10),  dimension(24),  intent(inout)               ::  par_header
#else
        character(len=:),   dimension(:),   intent(inout),  allocatable ::  par_header
#endif
        ! Private variables
#ifndef __GFORTRAN__
        character(len=:),   dimension(:),   allocatable ::  flx_par_headers_1,flx_par_headers_2,flx_par_headers_3, flx_par_headers_4
#endif
#ifdef __GFORTRAN__
        character(len=10),  dimension(21)               ::  flx_par_headers_5
        character(len=10),  dimension(24)               ::  flx_par_headers_6
#else
        character(len=:),   dimension(:),   allocatable ::  flx_par_headers_5, flx_par_headers_6
#endif
        character(len=10)                               ::  uunits
        character(len=10)                               ::  nnumber_of_frames
        integer                                         ::  current_par
        ! Start work

        ! Deal with units
        if (present(units)) then
            if (flx_par_version .lt. 4) then
                write(*,'(a,i0,a)') '**WARNING(frealix_par_header): specification of units will be ignored, since version ', &
                                    flx_par_version, ' is in use'
                call terminate('Debug')
            endif
            if (units(1:1) .eq. 'P') then
                uunits = 'Pixels'
            else
                uunits = 'Angstr'
            endif
        else
            uunits = 'Angstr'
        endif

#ifdef __GFORTRAN__
        if (flx_par_version .ne. 6) then
            call terminate('Fatal error in frealix_par_header. Only par version 6 is supported by gfortran build')
        endif
#endif

        ! Deal with number of frames
        nnumber_of_frames = '1'
        if (present(number_of_frames)) nnumber_of_frames = IntegerToString(number_of_frames)

        ! Here are all the frealix parameter file headers
#ifndef __GFORTRAN__
        flx_par_headers_1 = [CHARACTER(6)  ::   'WP#   ','AIP   ','A-OOP ','AA    ','SHX   ','SHY   ','MAG   ','FILM  ',&
                                                'DF1   ','DF2   ','ANGAST','OCC   ','LogP  ','SCORE ','CHANGE','IFIL  ','dR    ']
        flx_par_headers_2 = [CHARACTER(9)  ::   '!Fil.#   ','WP #     ', 'Film #   ', 'X        ','Y        ',&
                                                'Phi      ','Theta    ', 'Psi      ', 'Score    ','Refine?  ',&
                                                'par_vers.','2        ']
        flx_par_headers_3 = [CHARACTER(10) ::   '!Fil.#    ','WP #      ','Film #    ', &
                                                'X         ','Y         ','Z         ',&
                                                'PathLength','Phi       ','Theta     ',&
                                                'Psi       ','Score     ','Refine?   ',&
                                                'par_vers. ','3         ']
        flx_par_headers_4 = [CHARACTER(10) ::   '!Fil.#    ','WP #      ','Film #    ',&
                                                'X         ','Y         ','Z         ',&
                                                'PathLength','Phi       ','Theta     ',&
                                                'Psi       ','Subunit # ','Score     ',&
                                                'Refine?   ','Units     ',uunits,      &
                                                'par_vers. ','4         ']
        flx_par_headers_5 = [CHARACTER(10) ::   '!Fil.#    ','WP #      ','Film #    ',&
                                                'X         ','Y         ','Z         ',&
                                                'PathLength','Phi       ','Theta     ',&
                                                'Psi       ','Subunit # ',             &
                                                'Twist     ','Rise      ','Score     ',&
                                                'Refine?   ','Units     ',uunits,      &
                                                '# frames  ', nnumber_of_frames,       &
                                                'par_vers. ','5         ']
        flx_par_headers_6 = [CHARACTER(10) ::   '!Fil.#    ','WP #      ','Film #    ',&
                                                'X         ','Y         ','Z         ',&
                                                'PathLength','Phi       ','Theta     ',&
                                                'Psi       ','Subunit # ',             &
                                                'Twist     ','Rise      ','Score     ',&
                                                'LogP      ','Sigma     ','Occupancy ',&
                                                'Refine?   ','Units     ',uunits,      &
                                                '# frames  ', nnumber_of_frames,       &
                                                'par_vers. ','6         ']
#else
        flx_par_headers_5( 1) = '!Fil.#    '
        flx_par_headers_5( 2) = 'WP #      '
        flx_par_headers_5( 3) = 'Film #    '
        flx_par_headers_5( 4) = 'X         '
        flx_par_headers_5( 5) = 'Y         '
        flx_par_headers_5( 6) = 'Z         '
        flx_par_headers_5( 7) = 'PathLength'
        flx_par_headers_5( 8) = 'Phi       '
        flx_par_headers_5( 9) = 'Theta     '
        flx_par_headers_5(10) = 'Psi       '
        flx_par_headers_5(11) = 'Subunit # '
        flx_par_headers_5(12) = 'Twist     '
        flx_par_headers_5(13) = 'Rise      '
        flx_par_headers_5(14) = 'Score     '
        flx_par_headers_5(15) = 'Refine?   '
        flx_par_headers_5(16) = 'Units     '
        flx_par_headers_5(17) = uunits
        flx_par_headers_5(18) = '# frames  '
        flx_par_headers_5(19) = nnumber_of_frames
        flx_par_headers_5(20) = 'par_vers. '
        flx_par_headers_5(21) = '5         '

        flx_par_headers_6( 1) = '!Fil.#    '
        flx_par_headers_6( 2) = 'WP #      '
        flx_par_headers_6( 3) = 'Film #    '
        flx_par_headers_6( 4) = 'X         '
        flx_par_headers_6( 5) = 'Y         '
        flx_par_headers_6( 6) = 'Z         '
        flx_par_headers_6( 7) = 'PathLength'
        flx_par_headers_6( 8) = 'Phi       '
        flx_par_headers_6( 9) = 'Theta     '
        flx_par_headers_6(10) = 'Psi       '
        flx_par_headers_6(11) = 'Subunit # '
        flx_par_headers_6(12) = 'Twist     '
        flx_par_headers_6(13) = 'Rise      '
        flx_par_headers_6(14) = 'Score     '
        flx_par_headers_6(15) = 'LogP      '
        flx_par_headers_6(16) = 'Sigma     '
        flx_par_headers_6(17) = 'Occupancy '
        flx_par_headers_6(18) = 'Refine?   '
        flx_par_headers_6(19) = 'Units     '
        flx_par_headers_6(20) = uunits
        flx_par_headers_6(21) = '# frames  '
        flx_par_headers_6(22) = nnumber_of_frames
        flx_par_headers_6(23) = 'par_vers. '
        flx_par_headers_6(24) = '6         '
#endif


        ! Return the relevant one
        !frealix_par_header(:) = flx_par_headers(:,flx_par_version)
#ifndef __GFORTRAN__
        if (allocated(par_header)) deallocate(par_header)
        select case(flx_par_version)
            case(1)

                allocate(character(len=6) :: par_header(17))
                ! Below, used to be an implicit loop, but gfortran 4.9.1 would get an ICE over it
                do current_par=1,size(par_header)
                    par_header(current_par) = flx_par_headers_1(current_par)
                enddo
            case(2)
                allocate(character(len=7) :: par_header(12))
                ! Below, used to be an implicit loop, but gfortran 4.9.1 would get an ICE over it
                do current_par=1,size(par_header)
                    par_header(current_par) = flx_par_headers_2(current_par)
                enddo
            case(3)
                allocate(character(len=10) :: par_header(14))
                ! Below, used to be an implicit loop, but gfortran 4.9.1 would get an ICE over it
                do current_par=1,size(par_header)
                    par_header(current_par) = flx_par_headers_3(current_par)
                enddo
            case(4)
                allocate(character(len=10) :: par_header(17))
                ! Below, used to be an implicit loop, but gfortran 4.9.1 would get an ICE over it
                do current_par=1,size(par_header)
                    par_header(current_par) = flx_par_headers_4(current_par)
                enddo
            case(5)
                allocate(character(len=10) :: par_header(21))
                ! Below, used to be an implicit loop, but gfortran 4.9.1 would get an ICE over it
                do current_par=1,size(par_header)
                    par_header(current_par) = flx_par_headers_5(current_par)
                enddo
            case(6)
                allocate(character(len=10) :: par_header(24))
#endif
                ! Below, used to be an implicit loop, but gfortran 4.9.1 would get an ICE over it
                do current_par=1,size(par_header)
                    par_header(current_par) = flx_par_headers_6(current_par)
                enddo
#ifndef __GFORTRAN__
            case default
                write(*,'(a,i0)') '**ERROR(frealix_par_header): unknown parameter file format: ', flx_par_version
                call terminate('Fatal error in FREALIX_PAR_HEADER')
        end select
#endif
    end subroutine frealix_par_header

    !>  \brief  Return the header format string for the relevant frealix parameter file version
    function frealix_par_header_format(flx_par_version)
        use runtime_parameters, only : terminate, line_len
        implicit none
        ! Arguments
        integer,    intent(in)  ::  flx_par_version
        ! Result
        character(len=line_len) ::  frealix_par_header_format
        ! Private variables
        character(len=line_len) ::  flx_par_header_formats_1,flx_par_header_formats_2
        character(len=line_len) ::  flx_par_header_formats_3,flx_par_header_formats_4
        character(len=line_len) ::  flx_par_header_formats_5,flx_par_header_formats_6
        ! Start work

        ! Here are all the frealix parameter file header formats
        flx_par_header_formats_1    = &
            '(a6,1x,3(a7,1x),2(a9,1x),(a7,1x),(a5,1x),2(A8.1X),2(A7.1X),(a12,1x),2(a7,1x),(a8,1x),(a7))'

        flx_par_header_formats_2    = &                              !score              !refine
            '(3(a6,1x),2(a11,1x),  3(a12,1x),                        (a8,1x),            a7,1x,                          a11,1x,a1)'

        flx_par_header_formats_3    = &
            '(3(a6,1x),4(a11,1x),  3(a12,1x),                        (a8,1x),            a7,1x,                          a11,1x,a1)'

        flx_par_header_formats_4    = &
            '(3(a6,1x),4(a11,1x),  3(a12,1x),  1(a11,1x),            (a8,1x),            a7,1x, a5,1x,a6,1x,             a11,1x,a1)'

        flx_par_header_formats_5    = &                   !h. params
            '(3(a6,1x),4(a11,1x),  3(a12,1x),  1(a11,1x), 2(a12,1x), (a8,1x),            a7,1x, a5,1x,a6,1x,a11,1x,a4,1x,a11,1x,a1)'

        flx_par_header_formats_6    = &                                        !LogP,Sigma,OCC
            '(3(a6,1x),4(a11,1x),  3(a12,1x),  1(a11,1x), 2(a12,1x), (a8,1x),  3(a6,1x), a7,1x, a5,1x,a6,1x,a11,1x,a4,1x,a11,1x,a1)'

        ! Return the relevant one
        select case(flx_par_version)
            case(1)
                frealix_par_header_format = flx_par_header_formats_1
            case(2)
                frealix_par_header_format = flx_par_header_formats_2
            case(3)
                frealix_par_header_format = flx_par_header_formats_3
            case(4)
                frealix_par_header_format = flx_par_header_formats_4
            case(5)
                frealix_par_header_format = flx_par_header_formats_5
            case(6)
                frealix_par_header_format = flx_par_header_formats_6
            case default
                write(*,'(a,i0)') '**ERROR(frealix_par_header_format): unknown parameter file format: ', flx_par_version
                call terminate('Fatal error in FREALIX_PAR_HEADER_FORMAT')
        end select
    end function frealix_par_header_format

    !>  \brief  Return the data format string for the relevant frealix parameter file version
    function frealix_par_format(flx_par_version,number_of_frames)
        use runtime_parameters,  only : terminate, line_len
        use StringManipulations, only : IntegerToString
        implicit none
        ! Arguments
        integer,                intent(in)  ::  flx_par_version
        integer,    optional,   intent(in)  ::  number_of_frames        !< For versions 5 and above
        ! Result
        character(len=line_len) ::  frealix_par_format
        ! Private variables
        character(len=line_len) ::  flx_par_formats_5,flx_par_formats_6
        character(len=line_len) ::  flx_par_formats_1,flx_par_formats_2,flx_par_formats_3,flx_par_formats_4
        ! Start work

        ! Here are all the frealix parameter file header formats
        flx_par_formats_1   = '(i6,1x,3(F7.2,1x),2(F9.2,1x),(F7.1,1x),(i5,1x),2(F8.1,1x),'//&
                              '2(F7.2,1x),(E12.5,1x),2(F7.2,1x),(i8,1x),F7.2)'
        flx_par_formats_2   = '(3(I6,1x),2(F11.3,1x),3(F12.3,1x),            (F8.3,1x),l7,1x)'
        flx_par_formats_3   = '(3(I6,1x),4(F11.3,1x),3(F12.3,1x),            (F8.3,1x),l7,1x)'
        flx_par_formats_4   = '(3(I6,1x),4(F11.3,1x),3(F12.3,1x),1(F11.3,1x),(F8.3,1x),l7,1x)'
                                                     !euler angl !subunit #
        flx_par_formats_5   = '(3(I6,1x),4(F11.3,1x),3(F12.3,1x),1(F11.3,1x),' &
                            //'2(F12.3,1x),' & ! h params
                            //'(F8.3,1x),l7,45x,'&
                            //IntegerToString(3*number_of_frames)//'(F11.3,1x))'
        flx_par_formats_6   = '(3(I6,1x),4(F11.3,1x),3(F12.3,1x),1(F11.3,1x),' &
                            //'2(F12.3,1x),' & ! h params
                            //'(F8.3,1x),' & !score
                            //'i6,1x,f6.4,1x,f6.2,1x,' & ! LogP, Sigma,Occupancy
                            // 'l7,45x,'&
                            //IntegerToString(3*number_of_frames)//'(F11.3,1x))'

        ! Return the relevant one
        select case(flx_par_version)
            case(1)
                frealix_par_format = flx_par_formats_1
            case(2)
                frealix_par_format = flx_par_formats_2
            case(3)
                frealix_par_format = flx_par_formats_3
            case(4)
                frealix_par_format = flx_par_formats_4
            case(5)
                frealix_par_format = flx_par_formats_5
            case(6)
                frealix_par_format = flx_par_formats_6
            case default
                write(*,'(a,i0)') '**ERROR(frealix_par_format): unknown parameter file format: ', flx_par_version
                call terminate('Fatal error in FREALIX_PAR_FORMAT')
        end select
    end function frealix_par_format


    !>  \brief  Work out the number of movie frames from the header of a frealix .ali file
    integer function frealix_par_find_number_of_frames(par_filename) result(number_of_frames)
        use runtime_parameters
        use parsers
        implicit none
        ! Arguments
        character(len=*),               intent(in)  ::  par_filename            !<  Filename of parameter file
        ! Private variables
        character(len=line_len) ::  line
        character(len=word_len), allocatable    ::  words(:)
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        ! Read the first line
        call file_read_line_num(par_filename,1,line)
        if (debug) then
            print *, '**debug(frealix_par_find_number_of_frames): filename = ', trim(adjustl(par_filename))
            print *, '**debug(frealix_par_find_number_of_frames): line     = ', line
        endif

        ! It should be a comment
        if (.not. is_comment(line)) then
            write(*,'(4a)') '**ERROR(frealix_par_find_number_of_frames): first line of file ', trim(adjustl(par_filename)), &
                                ' is not a comment :'
            write(*,'(a)') trim(adjustl(line))
            call terminate('Fatal error in frealix_par_find_number_of_frames')
        endif

        ! Split the line
        words = split(line)
        !print '(2a)', '**debug(frealix_par_find_number_of_frames): line = ', trim(adjustl(line))
        !print '(a,i3)', '**debug(frealix_par_find_number_of_frames): # of words: ', size(words)

        ! The penultimate word should be par_version. If it isn't, we'll assume that the version just isn't specified.
        if (.not. string_equal(words(size(words)-3),'frames')) then
            write(*,'(2a)') '**info(frealix_par_find_number_of_frames): unexpected -3 word (not frames): ', &
                                trim(adjustl(words(size(words)-3)))
            number_of_frames = 1
            return
        endif

        !
        line = words(size(words)-2)
        if (.not. is_intg(line)) then
            write(*,'(a)') '**ERROR(frealix_par_find_number_of_frames): this word is not an integer :', trim(adjustl(line))
            call terminate('Fatal error in frealix_par_find_number_of_frames')
        endif

        ! We're almost done
        read(line,'(i4)') number_of_frames

    end function frealix_par_find_number_of_frames


    !>  \brief  Work out the version of a frealix2 parameter file
    !!
    !!  The version number is found at the end of the first line of the parameter file. The penultimate word in that line should be par_version and the last work should be an integer number.
    !!  If the version number is not specified there, this function returns the flx2_par_version_if_not_specified parameter
    integer function frealix_par_find_version(par_filename)
        use runtime_parameters
        use parsers
        implicit none
        ! Arguments
        character(len=*),               intent(in)  ::  par_filename            !<  Filename of parameter file
        ! Private variables
        character(len=line_len) ::  line
        character(len=word_len), allocatable    ::  words(:)
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        ! Read the first line
        call file_read_line_num(par_filename,1,line)
        if (debug) then
            print *, '**debug(frealix_par_find_version): filename = ', trim(adjustl(par_filename))
            print *, '**debug(frealix_par_find_version): line     = ', line
        endif

        ! It should be a comment
        if (.not. is_comment(line)) then
            write(*,'(4a)') '**ERROR(frealix_par_find_version): first line of file ', trim(adjustl(par_filename)), &
                                ' is not a comment :'
            write(*,'(a)') trim(adjustl(line))
            call terminate('Fatal error in FREALIX_PAR_FIND_VERSION')
        endif

        ! Split the line
        words = split(line)
        !print '(2a)', '**debug(frealix_par_find_version): line = ', trim(adjustl(line))
        !print '(a,i3)', '**debug(frealix_par_find_version): # of words: ', size(words)

        ! The penultimate word should be par_version. If it isn't, we'll assume that the version just isn't specified.
        if (.not. string_equal(words(size(words)-1),'par_vers.')) then
            write(*,'(3a,i2)') '**info(frealix_par_find_version): unexpected penultimate word (not par_vers.): ', &
                                trim(adjustl(words(size(words)-1))), &
                                '. Will revert to default parameter file version ', flx2_par_version_if_not_specified
            !call terminate('Fatal error in FREALIX2_PAR_FIND_VERSION')
            frealix_par_find_version = flx2_par_version_if_not_specified
            return
        endif

        ! The last word should be the version number, an integer
        line = words(size(words))
        if (.not. is_intg(line)) then
            write(*,'(a)') '**ERROR(frealix_par_find_version): last word is not an integer :', trim(adjustl(line))
            call terminate('Fatal error in FREALIX_PAR_FIND_VERSION')
        endif

        ! We're almost done
        read(line,'(i2)') frealix_par_find_version

    end function frealix_par_find_version

    !>  \brief  For versions 4 and above, frealix parameter files can store coordinates in pixels or Angstroms. Use this function to find out which is in use.
    !!
    function frealix_par_find_units(par_filename) result(units)
        use runtime_parameters
        use parsers
        implicit none
        ! Arguments
        character(len=*),               intent(in)  ::  par_filename            !<  Filename of parameter file
        ! Results
        character(len=1)    ::  units
        ! Private variables
        character(len=line_len) ::  line
        character(len=word_len), allocatable    ::  words(:)
        integer ::  par_version
        integer ::  unit_word_number
        integer ::  current_word
        ! Start work

        ! Find out the version of the parameter file first
        par_version = frealix_par_find_version(par_filename)

        if (par_version .lt. 3) then
            units = 'P'
        else if (par_version .eq. 3) then
            units = 'A'
        else
            ! For versions 4 and above, we need to check the header line

            ! Read the first line
            call file_read_line_num(par_filename,1,line)

            ! It should be a comment
            if (.not. is_comment(line)) then
                write(*,'(4a)') '**ERROR(frealix_par_find_units): first line of file ', trim(adjustl(par_filename)), &
                                    ' is not a comment :'
                write(*,'(a)') trim(adjustl(line))
                call terminate('Fatal error in FREALIX_PAR_FIND_UNITS')
            endif

            ! Split the line
            words = split(line)

            ! Loop through all the words until we find the word Unit
            unit_word_number = 0
            do current_word=1,size(words)
                if (string_equal(words(current_word),'Units')) then
                    unit_word_number = current_word + 1
                    exit
                endif
            enddo

            ! The third word to last should be either 'Pixels' or 'Angstr'
            select case (words(unit_word_number)(1:1))
                case ('P')
                    units = 'P'
                case ('A')
                    units = 'A'
                case default
                    write(*,'(2a)') '**ERROR(frealix_par_find_units): unit specification not recognised: ', &
                                    trim(adjustl(words(unit_word_number)))
                    call terminate('Fatal error in FREALIX_PAR_FIND_UNITS')
            end select
        endif ! End of check for frealix par version

    end function frealix_par_find_units


    !>  \brief  Apply a binning factor to a frealix2 parameter file
    !!  \todo   Make this work with other version of frealix parameter files
    subroutine frealix2_par_bin(input_par_filename,output_par_filename,bin_factor,unbin)
        !use filaments
        use parsers
        use runtime_parameters, only : terminate, available_lun, lun_is_now_available
		implicit none
        ! Arguments
        character(len=*),               intent(in)  ::  input_par_filename      !<  Filename for input parameter file
        character(len=*),               intent(in)  ::  output_par_filename     !<  Filename for output parameter file
        integer,                        intent(in)  ::  bin_factor              !<  Binning factor
        logical,            optional,   intent(in)  ::  unbin                   !<  If .TRUE., unbinning, rather than binning, is done
        ! Private variables
        logical ::  uunbin
        real    ::  bbin_factor
        !type(filament),    allocatable ::  fils(:)
        integer,    allocatable ::  fil_num(:),wp_num(:),film_num(:)
        real,       allocatable ::  x(:),y(:),phi(:),theta(:),psi(:),score(:)
        logical,    allocatable ::  refine(:)
        integer::   par_lun
        integer                 ::  i, ios
        ! Start work

        par_lun = available_lun()

        !
        ! This only deals with frealix2 parameter files (i.e. version 2 format) -TODO: make it work on all formats (note: nothing to do for version 3 since coordinates are stored in Angstroems)
        !
        if (frealix_par_find_version(input_par_filename) .ne. 2) then
            write(*,'(2a)') '**ERROR(frealix2_par_bin): subroutine not ready for this format of parameter file: ', &
                            frealix_par_find_version(input_par_filename)
            call terminate('Fatal error in FREALIX2_PAR_BIN')
        endif

        !
        ! Are we binning or unbinning?
        !
        if (present(unbin)) then
            uunbin = unbin
        else
            uunbin = .false.
        endif

        ! Check whether the output file already exists, in which case we'll overwrite it
        if (file_exists(output_par_filename)) then
            write(*,'(2a)') '**WARNING(frealix2_par_bin): parameter file already exists, will be overwritten : ', &
                            trim(adjustl(output_par_filename))
        endif
        open(unit=par_lun,file=trim(adjustl(output_par_filename)),status='REPLACE',iostat=ios)
        if (ios .ne. 0) then
            write(*,'(3a,i3)') '**ERROR(frealix2_par_bin): file opening failed for file ', trim(adjustl(output_par_filename)), &
                                ' with error ', ios
            !stop 'Fatal error in FREALIX2_PAR_BIN'
            call terminate()
        endif

        !
        ! Read in all the information from the input file
        !
        call get_ints_from_column       (input_par_filename,1,fil_num)
        call get_ints_from_column       (input_par_filename,2,wp_num)
        call get_ints_from_column       (input_par_filename,3,film_num)
        call get_reals_from_column      (input_par_filename,4,x)
        call get_reals_from_column      (input_par_filename,5,y)
        call get_reals_from_column      (input_par_filename,6,phi)
        call get_reals_from_column      (input_par_filename,7,theta)
        call get_reals_from_column      (input_par_filename,8,psi)
        call get_reals_from_column      (input_par_filename,9,score)
        call get_logicals_from_column   (input_par_filename,10,refine)

        !
        ! Do the binning
        !
        ! The first pixel is (1,1,1), not (0,0,0), so we have to be careful.
        !
        if (uunbin) then
            bbin_factor = 1.0/real(bin_factor)
        else
            bbin_factor = real(bin_factor)
        endif
        !print *, '**debug(frealix2_par_bin): binning by ', bbin_factor
        x = (X-1.0)/bbin_factor + 1.0
        y = (Y-1.0)/bbin_factor + 1.0

        !
        ! Write modified info to output parameter file
        !
        write(par_lun,flx2_par_header_format) '!Fil.#', 'WP #', 'Film #', 'X','Y','Phi','Theta','Psi','Score','Refine?'
        do I=1,size(wp_num)
            write(par_lun,flx2_par_format)  fil_num(i), wp_num(i), film_num(i),     &
                                            x(i), y(i),                             &
                                            phi(i), theta(i), psi(i),               &
                                            score(i),                               &
                                            refine(i)
        enddo

        close(unit=par_lun)
        call lun_is_now_available(par_lun)

    end subroutine frealix2_par_bin

    !>  \brief  Dump the current state of the filament waypoints to a frealix parameter file, using the latest version of the parameter file format
    subroutine filaments_to_frealix_par(fils,par_filename,par_version,units,subunits)
        use filaments_base
		use filaments_path, only  : filament_get_path
        use parsers
        use runtime_parameters, only: terminate, available_lun, lun_is_now_available
        implicit none
        ! Arguments
        type(filament),                     intent(inout)   ::  fils(:)             !<  Array of filaments
        character(len=*),                   intent(in)      ::  par_filename        !<  Filename for parameter file
        integer,            optional,       intent(in)      ::  par_version         !<  Version of parameter file format to use. Defaults to latest.
        character(len=1),   optional,       intent(in)      ::  units               !<  Either 'A' or 'P', for Angstroems or pixels. Defaults to Angstroems
        logical,            optional,       intent(in)      ::  subunits            !<  Output one line per helical asymetric unit, rather than one line per waypoint
        ! Private variables
        integer ::  ios ! io status
        integer ::  par_lun
        integer ::  fil_num, wp_num, film_num
        real    ::  pixel_size
        integer ::  ppar_version
        integer ::  number_of_frames
#ifdef __GFORTRAN__
        character(len=10),dimension(24)             :: ppar_header
#else
        character(len=:), dimension(:), allocatable :: ppar_header
#endif
        integer :: i
        logical, parameter :: debug = .false.
        character(len=1)    ::  uunits
        real    ::  temp_coo(2)
        logical ::  ssubunits
        type(waypoint), allocatable ::  wps(:)
        ! Start work

        ! Was a parameter file format version specified?
        if (present(par_version)) then
            ppar_version = par_version
        else
            ppar_version = flx2_par_lastest_version
        endif

        if (debug) then
            write(*,'(a,i0)') '**debug(filaments_to_frealix_par): ppar_version = ', ppar_version
        endif

        ! Were the units specified?
        if (present(units)) then
            ! Units are only availble for parameter file versions 4 and above
            if (ppar_version .lt. 4) then
                write(*,'(a)') '**ERROR(filaments_to_frealix_par): specification of units only availble for ',  &
                                'parameter file formats 4 and later'
                call terminate('Fatal error in FILAMENTS_TO_FREALIX_PAR')
            endif
            uunits = units
        else
            uunits = 'A'
        endif

        ! Subunits?
        ssubunits=.false.
        if (present(subunits)) ssubunits = subunits


        par_lun = available_lun()

        ! Check whether the file already exists, in which case we'll overwrite it
        if (file_exists(par_filename)) then
            write(*,'(2a)') '**WARNING(filaments_to_frealix_par): parameter file already exists, will be overwritten : ', &
                            trim(adjustl(par_filename))
        endif
        open(unit=par_lun,file=trim(adjustl(par_filename)),status='REPLACE',iostat=ios)
        if (ios .ne. 0) then
            write(*,'(3a,i0)') '**ERROR(filaments_to_frealix2_par): file opening failed for file ', trim(adjustl(par_filename)),&
                                 ' with error ', ios
            call terminate('Fatal error in FILAMENTS_TO_FREALIX_PAR')
        endif

        ! How many frames?
        number_of_frames = 1
        do fil_num=1,size(fils)
            if (fils(fil_num)%IsAllocated()) then
                if (allocated(fils(fil_num)%wp(1)%local_subframe_shift)) then
                    number_of_frames = size(fils(fil_num)%wp(1)%local_subframe_shift,2)
                    exit
                endif
            endif
        enddo

        ! Write header to file
        call frealix_par_header(ppar_header,ppar_version,units=uunits,number_of_frames=number_of_frames)
        write(par_lun,frealix_par_header_format(ppar_version)) (trim(ppar_header(i)),i=1,ubound(ppar_header,1))

        if (debug) then
            print *, 'header = ', ppar_header
        endif


        ! Loop over the filaments
        do fil_num=1,size(fils)
            if (debug) then
                !print '(2a)', '**debug(filaments_to_frealix_par): format = ', frealix_par_format(ppar_version)
                print '(a,i0)', '**debug(filaments_to_frealix_par): ppar_version = ', ppar_version
                write(*,'(a)') '**debug(filaments_to_frealix_par): about to add the following filament:'
                call fils(fil_num)%PrintInfo()
            endif
            if (fils(fil_num)%nwp .ge. 1) then
                film_num = fils(fil_num)%film%GetID()
                if (allocated(wps)) deallocate(wps)
                if (ssubunits) then
                    if (fils(fil_num)%helix%num_starts .ne. 1) call terminate('Conversion to one start not handled here yet')
                    call filament_get_path(     fils(fil_num),wps,  &
                                                step=1.0d0*real(fils(fil_num)%helix%axial_symmetry,kind=8)&
                                                          *real(fils(fil_num)%helix%perpendicular_symmetry,kind=8), &
                                                step_in_repeats=.true.,&
                                                ignore_fal_wps=.not. fils(fil_num)%fal_wps_in_threed)
                else
                    wps = fils(fil_num)%wp
                endif
                do wp_num=1,size(wps)
                    pixel_size = fils(fil_num)%film%GetPixelSize()
                    select case(ppar_version)
                        case (2)
                            write(par_lun,frealix_par_format(ppar_version))     fils(fil_num)%id, wp_num, film_num,    &
                                                                                wps(wp_num)%coo(1),                    &
                                                                                wps(wp_num)%coo(2),                    &
                                                                                wps(wp_num)%euler%phi,                 &
                                                                                wps(wp_num)%euler%theta,               &
                                                                                wps(wp_num)%euler%psi,                 &
                                                                                fils(fil_num)%score,                   &
                                                                                wps(wp_num)%refine
                        case (3)
                            write(par_lun,frealix_par_format(ppar_version))     fils(fil_num)%id, wp_num, film_num,    &
                                                                                (wps(wp_num)%coo(1)-1.0)*PIXEL_SIZE,   &   ! -1.0 because the origin is at (1,1) pixel or (0,0) Angstroems
                                                                                (wps(wp_num)%coo(2)-1.0)*PIXEL_SIZE,   &
                                                                                wps(wp_num)%coo(3)*PIXEL_SIZE,         &
                                                                                wps(wp_num)%PATH_LENGTH*PIXEL_SIZE,    &
                                                                                wps(wp_num)%euler%phi,                 &
                                                                                wps(wp_num)%euler%theta,               &
                                                                                wps(wp_num)%euler%psi,                 &
                                                                                fils(fil_num)%score,                   &
                                                                                wps(wp_num)%refine
                        case (4)
                            ! Need to deal with pixel / angstroem units
                            if (uunits .eq. 'P') then
                                pixel_size = 1.0
                                temp_coo(1:2) = wps(wp_num)%coo(1:2)
                            else
                                temp_coo(1:2) = wps(wp_num)%coo(1:2)-1.0       ! -1.0 because the origin is at (1,1) pixel or (0,0) Angstroems
                            endif
                            if (debug) then
                                write(*,'(3a,G15.3)')    '**debug(filaments_to_frealix_par): unit is ', uunits, &
                                                        ', therefore will use pixel size of ', pixel_size
                                print *, '**DEBUG() fmt = ', frealix_par_format(ppar_version)
                            endif
                            write(par_lun,frealix_par_format(ppar_version))     fils(fil_num)%id, wp_num, film_num,    &
                                                                                temp_coo(1)*PIXEL_SIZE,                &
                                                                                temp_coo(2)*PIXEL_SIZE,                &
                                                                                wps(wp_num)%coo(3)*PIXEL_SIZE,         &
                                                                                wps(wp_num)%PATH_LENGTH*PIXEL_SIZE,    &
                                                                                wps(wp_num)%euler%phi,                 &
                                                                                wps(wp_num)%euler%theta,               &
                                                                                wps(wp_num)%euler%psi,                 &
                                                                                wps(wp_num)%subunit_num,               &
                                                                                wps(wp_num)%score,                     &
                                                                                wps(wp_num)%refine
                        case (5)
                            ! Need to deal with pixel / angstroem units
                            if (uunits .eq. 'P') then
                                pixel_size = 1.0
                                temp_coo(1:2) = wps(wp_num)%coo(1:2)
                            else
                                temp_coo(1:2) = wps(wp_num)%coo(1:2)-1.0       ! -1.0 because the origin is at (1,1) pixel or (0,0) Angstroems
                            endif
                            if (debug) then
                                write(*,'(3a,G15.3)')    '**debug(filaments_to_frealix_par): unit is ', uunits, &
                                                        ', therefore will use pixel size of ', pixel_size
                                print *, '**DEBUG() fmt = ', trim(adjustl(frealix_par_format(ppar_version, &
                                                                number_of_frames=fils(fil_num)%film%GetNumberOfFrames())))
                            endif
                            if (allocated(wps(wp_num)%local_subframe_shift)) then
                                write(par_lun,frealix_par_format(ppar_version, &
                                                           number_of_frames=size(wps(wp_num)%local_subframe_shift,2))) &
                                                                                fils(fil_num)%id, wp_num, film_num,    &
                                                                                temp_coo(1)*pixel_size,                &
                                                                                temp_coo(2)*pixel_size,                &
                                                                                wps(wp_num)%coo(3)*pixel_size,         &
                                                                                wps(wp_num)%path_length*pixel_size,    &
                                                                                wps(wp_num)%euler%phi,                 &
                                                                                wps(wp_num)%euler%theta,               &
                                                                                wps(wp_num)%euler%psi,                 &
                                                                                wps(wp_num)%subunit_num,               &
                                                                                wps(wp_num)%helix%twist,               &
                                                                                wps(wp_num)%helix%rise*pixel_size,     &
                                                                                max(wps(wp_num)%score,-1.0),           &
                                                                                wps(wp_num)%refine,                    &
                                                                                wps(wp_num)%local_subframe_shift*pixel_size
                            else
                                write(par_lun,frealix_par_format(ppar_version, &
                                                           number_of_frames=size(wps(wp_num)%local_subframe_shift,2))) &
                                                                                fils(fil_num)%id, wp_num, film_num,    &
                                                                                temp_coo(1)*pixel_size,                &
                                                                                temp_coo(2)*pixel_size,                &
                                                                                wps(wp_num)%coo(3)*pixel_size,         &
                                                                                wps(wp_num)%path_length*pixel_size,    &
                                                                                wps(wp_num)%euler%phi,                 &
                                                                                wps(wp_num)%euler%theta,               &
                                                                                wps(wp_num)%euler%psi,                 &
                                                                                wps(wp_num)%subunit_num,               &
                                                                                wps(wp_num)%helix%twist,               &
                                                                                wps(wp_num)%helix%rise*pixel_size,     &
                                                                                max(wps(wp_num)%score,-1.0),           &
                                                                                wps(wp_num)%refine,                    &
                                                                                0.0,0.0,0.0
                            endif
                        case (6)
                            ! Need to deal with pixel / angstroem units
                            if (uunits .eq. 'P') then
                                pixel_size = 1.0
                                temp_coo(1:2) = wps(wp_num)%coo(1:2)
                            else
                                temp_coo(1:2) = wps(wp_num)%coo(1:2)-1.0       ! -1.0 because the origin is at (1,1) pixel or (0,0) Angstroems
                            endif
                            if (debug) then
                                write(*,'(3a,G15.3)')    '**debug(filaments_to_frealix_par): unit is ', uunits, &
                                                        ', therefore will use pixel size of ', pixel_size
                                print *, '**DEBUG() fmt = ', trim(adjustl(frealix_par_format(ppar_version, &
                                                                number_of_frames=fils(fil_num)%film%GetNumberOfFrames())))
                            endif
                            if (allocated(wps(wp_num)%local_subframe_shift)) then
                                write(par_lun,frealix_par_format(ppar_version, &
                                                           number_of_frames=size(wps(wp_num)%local_subframe_shift,2))) &
                                                                                fils(fil_num)%id, wp_num, film_num,    &
                                                                                temp_coo(1)*pixel_size,                &
                                                                                temp_coo(2)*pixel_size,                &
                                                                                wps(wp_num)%coo(3)*pixel_size,         &
                                                                                wps(wp_num)%path_length*pixel_size,    &
                                                                                wps(wp_num)%euler%phi,                 &
                                                                                wps(wp_num)%euler%theta,               &
                                                                                wps(wp_num)%euler%psi,                 &
                                                                                wps(wp_num)%subunit_num,               &
                                                                                wps(wp_num)%helix%twist,               &
                                                                                wps(wp_num)%helix%rise*pixel_size,     &
                                                                                max(wps(wp_num)%score,-1.0),           &
                                                                                nint(wps(wp_num)%logp),                &
                                                                                wps(wp_num)%noise_sigma,               &
                                                                                wps(wp_num)%occupancy,                 &
                                                                                wps(wp_num)%refine,                    &
                                                                                wps(wp_num)%local_subframe_shift*pixel_size
                            else
                                write(par_lun,frealix_par_format(ppar_version, &
                                                           number_of_frames=size(wps(wp_num)%local_subframe_shift,2))) &
                                                                                fils(fil_num)%id, wp_num, film_num,    &
                                                                                temp_coo(1)*pixel_size,                &
                                                                                temp_coo(2)*pixel_size,                &
                                                                                wps(wp_num)%coo(3)*pixel_size,         &
                                                                                wps(wp_num)%path_length*pixel_size,    &
                                                                                wps(wp_num)%euler%phi,                 &
                                                                                wps(wp_num)%euler%theta,               &
                                                                                wps(wp_num)%euler%psi,                 &
                                                                                wps(wp_num)%subunit_num,               &
                                                                                wps(wp_num)%helix%twist,               &
                                                                                wps(wp_num)%helix%rise*pixel_size,     &
                                                                                max(wps(wp_num)%score,-1.0),           &
                                                                                nint(wps(wp_num)%logp),                &
                                                                                wps(wp_num)%noise_sigma,               &
                                                                                wps(wp_num)%occupancy,                 &
                                                                                wps(wp_num)%refine,                    &
                                                                                0.0,0.0,0.0
                            endif
                        case default
                            write(*,'(a)') '**ERROR(filaments_to_frealix_par): unsupported parameter file format: ', ppar_version
                            call terminate('Fatal error in FILAMENTS_TO_FREALIX_PAR')
                    end select
                enddo
            endif ! end of test whether filament allocated
        enddo ! end of loop over filaments

        close(unit=par_lun)
        call lun_is_now_available(par_lun)

    end subroutine filaments_to_frealix_par

    !>  \brief  Write a frealix alignment file describing the input array of waypoints
    subroutine waypoints_to_frealix_ali(wp,par_filename,pixel_size,fil_id,film_num,score)
        use runtime_parameters, only : available_lun, lun_is_now_available, terminate
        use parsers, only : file_exists
        use coos, only  : waypoint
        implicit none
        ! Arguments
        type(waypoint),     allocatable,    intent(in)  ::  wp(:)
        character(len=*),                   intent(in)  ::  par_filename
        real,                               intent(in)  ::  pixel_size
        integer,            optional,       intent(in)  ::  fil_id
        integer,            optional,       intent(in)  ::  film_num
        real,               optional,       intent(in)  ::  score
        ! Private variables
        integer ::  par_lun, ios,i, wp_num, ffil_id, ffilm_num, ppar_version
        real    ::  sscore, temp_coo(3)
#ifdef __GFORTRAN__
        character(len=10), dimension(24) :: ppar_header
#else
        character(len=:), dimension(:), allocatable :: ppar_header
#endif
        character(len=1) :: uunits
        ! Start work

        par_lun = available_lun()

        uunits = 'A'
        ppar_version = 4
        ffil_id = 1
        ffilm_num = 1
        sscore = 0.0
        if (present(fil_id)) ffil_id = fil_id
        if (present(film_num)) ffilm_num = film_num
        if (present(score)) sscore = max(score,-1.0)

        ! Check whether the file already exists, in which case we'll overwrite it
        if (file_exists(par_filename)) then
            write(*,'(2a)') '**WARNING(waypoints_to_frealix_ali): parameter file already exists, will be overwritten : ', &
                            trim(adjustl(par_filename))
        endif
        open(unit=PAR_LUN,file=TRIM(adjustl(par_filename)),status='REPLACE',iostat=IOS)
        if (ios .ne. 0) then
            write(*,'(3a,i0)') '**ERROR(waypoints_to_frealix_ali): file opening failed for file ', &
                                trim(adjustl(par_filename)), ' with error ', ios
            call terminate('Fatal error in WAYPOINTS_TO_FREALIX_ALI')
        endif

        ! Write header to fil
        call frealix_par_header(ppar_header,ppar_version,units=uunits)
        write(par_lun,frealix_par_header_format(ppar_version)) (trim(ppar_header(i)),i=1,ubound(ppar_header,1))

        do wp_num=1,size(wp)
            ! Need to deal with pixel / angstroem units
            temp_coo(1:2) = wp(wp_num)%coo(1:2)-1.0       ! -1.0 because the origin is at (1,1) pixel or (0,0) Angstroems
            if (allocated(wp(wp_num)%local_subframe_shift)) then
                write(par_lun,frealix_par_format(ppar_version,number_of_frames=size(wp(wp_num)%local_subframe_shift,2)))     &
                                                                    ffil_id, wp_num, ffilm_num,           &
                                                                    temp_coo(1)*pixel_size,               &
                                                                    temp_coo(2)*pixel_size,               &
                                                                    wp(wp_num)%coo(3)*pixel_size,         &
                                                                    wp(wp_num)%path_length*pixel_size,    &
                                                                    wp(wp_num)%euler%phi,                 &
                                                                    wp(wp_num)%euler%theta,               &
                                                                    wp(wp_num)%euler%psi,                 &
                                                                    wp(wp_num)%subunit_num,               &
                                                                    sscore,                               &
                                                                    wp(wp_num)%refine,                    &
                                                                    wp(wp_num)%local_subframe_shift
            else
                write(par_lun,frealix_par_format(ppar_version))     ffil_id, wp_num, ffilm_num,           &
                                                                    temp_coo(1)*pixel_size,               &
                                                                    temp_coo(2)*pixel_size,               &
                                                                    wp(wp_num)%coo(3)*pixel_size,         &
                                                                    wp(wp_num)%path_length*pixel_size,    &
                                                                    wp(wp_num)%euler%phi,                 &
                                                                    wp(wp_num)%euler%theta,               &
                                                                    wp(wp_num)%euler%psi,                 &
                                                                    wp(wp_num)%subunit_num,               &
                                                                    sscore,                               &
                                                                    wp(wp_num)%refine
            endif
        enddo

        close(unit=par_lun)
        call lun_is_now_available(par_lun)
    end subroutine waypoints_to_frealix_ali


    !>  \brief  Read a frealix parameter file (specified version and return an array of filament objects, not initialised)
    subroutine filaments_from_frealix_par_noinit(fils,par_filename,pixel_size,number_of_frames)
        use filaments_base
		use runtime_parameters, only : int_unique, terminate
		use coos, only : waypoints_print_info
        use parsers
        implicit none
        ! Arguments
        type(filament),     allocatable, target,    intent(inout)   ::  fils(:)
        character(len=*),                           intent(in)      ::  par_filename                        !<  Filename for parameter file
        real,               optional,               intent(in)      ::  pixel_size                          !<  Pixel size (Angstroms). Required for frealix par format 3+.
        integer,            optional,               intent(in)      ::  number_of_frames                    !<  Number of movie frames. Needed if the waypoints don't have subframe shifts already (for example if older waypoints file is being read in)
        ! Result

        ! Private variables
        integer                     ::  flx_par_version
        integer,        allocatable ::  film_num(:), fil_num(:)
        real(kind=8),   allocatable ::  x(:), y(:), z(:), path_length(:), phi(:), theta(:), psi(:), subunit_num(:)
        real(kind=8),   allocatable ::  twist(:), rise(:)
        real(kind=4),   allocatable ::  logp(:),noise_sigma(:),occupancy(:)
        real(kind=4),   allocatable ::  score(:)
        logical,        allocatable ::  refine(:)
        integer,        allocatable ::  tmp_int(:)
        type(waypoint), allocatable ::  wp(:)
        integer                     ::  num_of_fils
        integer                     ::  i, j, j_start, ierr
        integer                     ::  num_of_wps, current_wp
        logical                     ::  par_in_angstroms
        integer                     ::  nnumber_of_frames, nnumber_of_frames_expected
        integer                     ::  current_frame
        real,           allocatable ::  local_subframe_shifts(:,:,:)
        real(kind=8),   allocatable ::  temp_subframe_shift(:)
        integer                     ::  current_dimension
        logical,        parameter   ::  debug   = .false.
        integer                     ::  col_num_refine

        ! Start work

        ! Check the parameter file exists
        if (.not. file_exists(par_filename)) then
            write(*,'(a,a)') '**ERROR(frealix_par_to_filaments): file does not exist: ', trim(adjustl(par_filename))
            call terminate('Fatal error in FREALIX_PAR_TO_FILAMENTS')
        endif

        ! Which version of frealix parameter file?
        flx_par_version = frealix_par_find_version(par_filename)

        ! Is the parameter file in Angstroms?
        par_in_angstroms = frealix_par_find_units(par_filename) .eq. 'A'

        if (debug) then
            write(*,'(a,i0,1x,l1)') '**debug(filaments_from_frealix_par_noinit): version, Angstroms? = ', &
                                    flx_par_version, par_in_angstroms
        endif

        ! If required, check the pixel size was given
        if (par_in_angstroms) then
            if (present(pixel_size)) then
                !ppixel_size  = pixel_size
            else
                write(*,'(2a)') '**ERROR(frealix_par_to_filaments_noinit): pixel size is required when reading in ', &
                                'frealix parameter files versions 3+'
                call terminate('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
            endif
        endif

        ! Number of frames (may be modified later by looking at the waypoints file)
        nnumber_of_frames = 1
        if (present(number_of_frames)) nnumber_of_frames = number_of_frames
        nnumber_of_frames_expected = nnumber_of_frames


        ! Read in data
        if (flx_par_version .eq. 1) then
            write(*,'(2a)') '**ERROR(frealix_par_to_filaments_noinit): unable to deal with frealix parameter ',  &
                            'file of version 1 yet. Sorry.'
            call terminate('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
        else if (flx_par_version .eq. 2) then
            call get_ints_from_column(par_filename,1,fil_num)
            call get_ints_from_column(par_filename,3,film_num)
            call get_dbles_from_column(par_filename,4,x)
            call get_dbles_from_column(par_filename,5,y)
            call get_dbles_from_column(par_filename,6,phi)
            call get_dbles_from_column(par_filename,7,theta)
            call get_dbles_from_column(par_filename,8,psi)
            call get_reals_from_column(par_filename,9,score)
            call get_logicals_from_column(par_filename,10,refine)
        else if (flx_par_version .ge. 3) then
            call get_ints_from_column(par_filename,1,fil_num)
            call get_ints_from_column(par_filename,3,film_num)
            call get_dbles_from_column(par_filename,4,x)
            call get_dbles_from_column(par_filename,5,y)
            call get_dbles_from_column(par_filename,6,z)
            call get_dbles_from_column(par_filename,7,path_length)
            call get_dbles_from_column(par_filename,8,phi)
            call get_dbles_from_column(par_filename,9,theta)
            call get_dbles_from_column(par_filename,10,psi)
            if (flx_par_version .eq. 3) then
                call get_reals_from_column(par_filename,11,score)
                call get_logicals_from_column(par_filename,12,refine)
            else if (flx_par_version .eq. 4) then
                call get_dbles_from_column(par_filename,11,subunit_num)
                call get_reals_from_column(par_filename,12,score)
                call get_logicals_from_column(par_filename,13,refine)
            else if (flx_par_version .ge. 5) then
                call get_dbles_from_column(par_filename,11,subunit_num)
                call get_dbles_from_column(par_filename,12,twist)
                call get_dbles_from_column(par_filename,13,rise)
                call get_reals_from_column(par_filename,14,score)
                if (flx_par_version .eq. 5) then
                    col_num_refine = 15
                elseif (flx_par_version .eq. 6) then
                    col_num_refine = 18
                endif
                call get_logicals_from_column(par_filename,col_num_refine,refine)
                if (flx_par_version .eq. 6) then
                    ! occ and logp
                    call get_reals_from_column(par_filename,15,logp)
                    call get_reals_from_column(par_filename,16,noise_sigma)
                    call get_reals_from_column(par_filename,17,occupancy)
                endif
                nnumber_of_frames = frealix_par_find_number_of_frames(par_filename)
                allocate(local_subframe_shifts(3,nnumber_of_frames,size(x)))
                do current_frame=1,nnumber_of_frames
                    do current_dimension=1,3
                        call get_dbles_from_column(par_filename, &
                                                   col_num_refine+current_dimension+(current_frame-1)*3, &
                                                   temp_subframe_shift)
                        local_subframe_shifts(current_dimension,current_frame,:) = temp_subframe_shift(:)
                    enddo
                enddo
            endif
        else
            write(*,'(a,i0)') '**ERROR(frealix_par_to_filaments_noinit): unknown frealix parameter file format: ', flx_par_version
            call terminate('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
        endif

        ! Check we have the expected number of lines etc.
        num_of_wps  =   number_of_lines_in_file(par_filename,ignore_comments=.true.)
        if (size(fil_num) .ne. num_of_wps) then
            write(*,'(2(A,i6),2a)') '**ERROR(frealix_par_to_filaments_noinit): found ', size(fil_num), &
                                    ' values in column1, but counted ', num_of_wps, ' lines in file ', trim(par_filename)
            call terminate('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
        endif

        ! Work out number of filaments
        call get_ints_from_column(par_filename,1,fil_num)
        tmp_int = int_unique(fil_num)
        num_of_fils = size(tmp_int)



        ! Allocate filament array
        if (allocated(fils)) deallocate(fils)
        allocate(fils(num_of_fils),stat=ierr)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(frealix_par_to_filaments): filament array memory allocation failed with error ', ierr
            call terminate('Fatal error in FREALIX_PAR_TO_FILAMENTS')
        endif

        ! Loop over the filaments to find out how many waypoints are in each and then copy the data over into the filament objects
        current_wp = 1
        j = 0   ! This is the current line/waypoint
        do i=1,num_of_fils
            ! At the beginning of the iteration, j points to the last waypoint of the previous filament
            j = j+1
            j_start = j

            ! For the current filament, work out how many waypoints are defined in the parameter file
            ! If we're on the last waypoint, of course we don't need to do this
            if (j .lt. num_of_wps) then
                ! Let's increment j until we reach the last waypoint in the current filament, i
                do
                    if (j .eq. num_of_wps) then
                        ! We reached the last way point
                        exit
                    else if (fil_num(j+1) .eq. fil_num(j)) then
                        ! We are not yet at last way point of current filament
                        j = j +1
                    else
                        ! We are at last waypoint of current filament
                        exit
                    endif
                enddo
            endif

            ! So now j is set to the position of the last way point of the current filament

            ! Diagnostic message
            if (debug) then
                print '(a,i0,a,i0,a)', '**debug(frealix_par_to_filaments_noinit): filament ', fil_num(j),&
                                             ' has ', j-j_start+1, ' way points'
            endif

            ! Allocate memory to an array of way points
            if (allocated(wp)) deallocate(wp)
            allocate(wp(j-j_start+1),stat=ierr)
            if (ierr .ne. 0) then
                write(*,'(a,i0)') '**ERROR(frealix_par_to_filaments_noinit): memory allocation of temporary ',   &
                                    'waypoint array failed with err ', ierr
                call terminate('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
            endif

            ! Sanity check - all waypoints should have hte same film number
            if (any(film_num(j_start+1:j) .ne. film_num(j_start))) then
                write(*,'(a,i0,a)') '**error(frealix_par_to_filament_noinit): not all waypoints for filament ', &
                                        fil_num(j_start), ' belong to the same micrograph'
            endif



            ! Set up the film - we can do this by parsing either ctftilt or ctffind
            ! If the user has given us a ctftilt file, we use that.
!           call filename_from_wild(film_filename_wild,film_num(j_start),film_filename)
!           if (ctftilt_file) then
!               call filename_from_wild(ctftilt_filename_wild,film_num(j_start),ctftilt_filename)
!               call film_init(film1,film_filename,pixel_size,CTFTILT_FILENAME=CTFTILT_FILENAME)
!           else
!               call filename_from_wild(ctffind_filename_wild,film_num(j_start),ctffind_filename)
!               call film_init(film1,film_filename,pixel_size,CTFFIND_FILENAME=CTFFIND_FILENAME)
!           endif

            ! Populate the waypoints
            select case (flx_par_version)
                case (1)
                    write(*,'(2a)') '**ERROR(frealix_par_to_filaments_noinit): unable to deal with frealix ',    &
                                    'parameter file of version 1 yet. Sorry.'
                    call terminate('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
                case (2)
                    wp(1:j-j_start+1)%coo(1)        =   x(j_start:j)
                    wp(1:j-j_start+1)%coo(2)        =   y(j_start:j)
                    wp(1:j-j_start+1)%coo(3)        =   0.0
                    wp(1:j-j_start+1)%euler%phi     =   phi(j_start:j)
                    wp(1:j-j_start+1)%euler%theta   =   theta(j_start:j)
                    wp(1:j-j_start+1)%euler%psi     =   psi(j_start:j)
                    wp(1:j-j_start+1)%refine        =   refine(j_start:j)
                case (3)
                    wp(1:j-j_start+1)%coo(1)        =   x(j_start:j)
                    wp(1:j-j_start+1)%coo(2)        =   y(j_start:j)
                    wp(1:j-j_start+1)%coo(3)        =   z(j_start:j)
                    wp(1:j-j_start+1)%path_length   =   path_length(j_start:j)/pixel_size
                    if (par_in_angstroms) then
                        wp(1:j-j_start+1)%coo(1)        =   wp(1:j-j_start+1)%coo(1)/pixel_size +1.0    ! +1.0 because the origin is at pixel (1,1) in corner
                        wp(1:j-j_start+1)%coo(2)        =   wp(1:j-j_start+1)%coo(2)/pixel_size +1.0
                        wp(1:j-j_start+1)%coo(3)        =   wp(1:j-j_start+1)%coo(3)/pixel_size
                        wp(1:j-j_start+1)%path_length   =   wp(1:j-j_start+1)%path_length/pixel_size
                    endif
                    wp(1:j-j_start+1)%euler%phi     =   phi(j_start:j)
                    wp(1:j-j_start+1)%euler%theta   =   theta(j_start:j)
                    wp(1:j-j_start+1)%euler%psi     =   psi(j_start:j)
                    wp(1:j-j_start+1)%refine        =   refine(j_start:j)
                case (4)
                    wp(1:j-j_start+1)%coo(1)        =   x(j_start:j)
                    wp(1:j-j_start+1)%coo(2)        =   y(j_start:j)
                    wp(1:j-j_start+1)%coo(3)        =   z(j_start:j)
                    wp(1:j-j_start+1)%path_length   =   path_length(j_start:j)
                    if (par_in_angstroms) then
                        wp(1:j-j_start+1)%coo(1)        =   wp(1:j-j_start+1)%coo(1)/pixel_size +1.0    ! +1.0 because the origin is at pixel (1,1) in corner
                        wp(1:j-j_start+1)%coo(2)        =   wp(1:j-j_start+1)%coo(2)/pixel_size +1.0
                        wp(1:j-j_start+1)%coo(3)        =   wp(1:j-j_start+1)%coo(3)/pixel_size
                        wp(1:j-j_start+1)%path_length   =   wp(1:j-j_start+1)%path_length/pixel_size
                    endif
                    wp(1:j-j_start+1)%euler%phi     =   phi(j_start:j)
                    wp(1:j-j_start+1)%euler%theta   =   theta(j_start:j)
                    wp(1:j-j_start+1)%euler%psi     =   psi(j_start:j)
                    wp(1:j-j_start+1)%refine        =   refine(j_start:j)
                    wp(1:j-j_start+1)%subunit_num   =   subunit_num(j_start:j)
                    wp(1:j-j_start+1)%score         =   score(j_start:j)
                case (5)
                    wp(1:j-j_start+1)%coo(1)                    =   x(j_start:j)
                    wp(1:j-j_start+1)%coo(2)                    =   y(j_start:j)
                    wp(1:j-j_start+1)%coo(3)                    =   z(j_start:j)
                    wp(1:j-j_start+1)%path_length               =   path_length(j_start:j)
                    if (par_in_angstroms) then
                        wp(1:j-j_start+1)%coo(1)                =   wp(1:j-j_start+1)%coo(1)/pixel_size +1.0    ! +1.0 because the origin is at pixel (1,1) in corner
                        wp(1:j-j_start+1)%coo(2)                =   wp(1:j-j_start+1)%coo(2)/pixel_size +1.0
                        wp(1:j-j_start+1)%coo(3)                =   wp(1:j-j_start+1)%coo(3)/pixel_size
                        wp(1:j-j_start+1)%path_length           =   wp(1:j-j_start+1)%path_length/pixel_size
                    endif
                    wp(1:j-j_start+1)%helix%twist               =   twist(j_start:j)
                    wp(1:j-j_start+1)%helix%rise                =   rise(j_start:j)
                    if (par_in_angstroms) then
                        wp(1:j-j_start+1)%helix%rise            =   wp(1:j-j_start+1)%helix%rise/pixel_size
                    endif
                    wp(1:j-j_start+1)%helix%pixel_units         =   .true.
                    wp(1:j-j_start+1)%euler%phi                 =   phi(j_start:j)
                    wp(1:j-j_start+1)%euler%theta               =   theta(j_start:j)
                    wp(1:j-j_start+1)%euler%psi                 =   psi(j_start:j)
                    wp(1:j-j_start+1)%refine                    =   refine(j_start:j)
                    wp(1:j-j_start+1)%subunit_num               =   subunit_num(j_start:j)
                    wp(1:j-j_start+1)%score                     =   score(j_start:j)
                    if (par_in_angstroms) then
                        do current_wp=1,j-j_start+1
                            wp(current_wp)%local_subframe_shift     =   local_subframe_shifts(:,:,current_wp-1+j_start)/pixel_size
                        enddo
                    endif
                case (6)
                    wp(1:j-j_start+1)%coo(1)                    =   x(j_start:j)
                    wp(1:j-j_start+1)%coo(2)                    =   y(j_start:j)
                    wp(1:j-j_start+1)%coo(3)                    =   z(j_start:j)
                    wp(1:j-j_start+1)%path_length               =   path_length(j_start:j)
                    if (par_in_angstroms) then
                        wp(1:j-j_start+1)%coo(1)                =   wp(1:j-j_start+1)%coo(1)/pixel_size +1.0    ! +1.0 because the origin is at pixel (1,1) in corner
                        wp(1:j-j_start+1)%coo(2)                =   wp(1:j-j_start+1)%coo(2)/pixel_size +1.0
                        wp(1:j-j_start+1)%coo(3)                =   wp(1:j-j_start+1)%coo(3)/pixel_size
                        wp(1:j-j_start+1)%path_length           =   wp(1:j-j_start+1)%path_length/pixel_size
                    endif
                    wp(1:j-j_start+1)%helix%twist               =   twist(j_start:j)
                    wp(1:j-j_start+1)%helix%rise                =   rise(j_start:j)
                    if (par_in_angstroms) then
                        wp(1:j-j_start+1)%helix%rise            =   wp(1:j-j_start+1)%helix%rise/pixel_size
                    endif
                    wp(1:j-j_start+1)%helix%pixel_units         =   .true.
                    wp(1:j-j_start+1)%euler%phi                 =   phi(j_start:j)
                    wp(1:j-j_start+1)%euler%theta               =   theta(j_start:j)
                    wp(1:j-j_start+1)%euler%psi                 =   psi(j_start:j)
                    wp(1:j-j_start+1)%refine                    =   refine(j_start:j)
                    wp(1:j-j_start+1)%subunit_num               =   subunit_num(j_start:j)
                    wp(1:j-j_start+1)%score                     =   score(j_start:j)
                    wp(1:j-j_start+1)%logp                      =   logp(j_start:j)
                    wp(1:j-j_start+1)%noise_sigma               =   noise_sigma(j_start:j)
                    wp(1:j-j_start+1)%occupancy                 =   occupancy(j_start:j)
                    if (par_in_angstroms) then
                        do current_wp=1,j-j_start+1
                            wp(current_wp)%local_subframe_shift     =   local_subframe_shifts(:,:,current_wp-1+j_start)/pixel_size
                        enddo
                    endif
                case default
                    write(*,'(a,i0)') '**ERROR(frealix_par_to_filaments): unknown frealix parameter file format: ', flx_par_version
                    call terminate('Fatal error in FREALIX_PAR_TO_FILAMENTS')
            end select

            ! It could be that the local subframe shifts were not known, so we set them to a reasonable value
            do current_wp=1,size(wp)
                if ( .not. allocated(wp(current_wp)%local_subframe_shift)) then
                    allocate(wp(current_wp)%local_subframe_shift(3,1))
                    wp(current_wp)%local_subframe_shift = 0.0e0
                endif
                if( size(wp(current_wp)%local_subframe_shift,2) .lt. nnumber_of_frames_expected) then
                    if (allocated(wp(current_wp)%local_subframe_shift)) deallocate(wp(current_wp)%local_subframe_shift)
                    allocate(wp(current_wp)%local_subframe_shift(3,nnumber_of_frames_expected))
                    wp(current_wp)%local_subframe_shift = 0.0e0
                endif
            enddo
            if (nnumber_of_frames_expected .ne. 1) nnumber_of_frames = nnumber_of_frames_expected

            ! Sanity check
            if (any(wp(1:j-j_start+1)%coo(1) .lt. 0.0) .or. &
                any(wp(1:j-j_start+1)%coo(2) .lt. 0.0) ) then
                write(*,'(a,i0,a)') '**WARNING(frealix_par_to_filaments_noinit): at least one of the waypoints '//&
                                    'read from file for filament ', fil_num(j_start), ' has negative coordinates: '
                call waypoints_print_info(wp(1:j-j_start+1))
                !call terminate('Fatal error in FREALIX_PAR_TO_FILAMENTS_NOINIT')
            endif

            if (debug) then
                call waypoints_print_info(wp)
            endif

            ! Now that the waypoints are ready, populate a filament object
            call fils(i)%PopulateWithWaypoints(wp,id=fil_num(j_start),film_id=film_num(j_start),&
                                                pixel_size=pixel_size,score=score(j_start), &
                                                number_of_frames_per_micrograph=nnumber_of_frames)



            ! Diagnostics
            if (debug) then
                write(*,'(a)') '**debug(filaments_from_frealix_par_noinit): filament info on return:'
                call fils(i)%PrintInfo()
            endif

        enddo ! End of loop over filaments

    end subroutine filaments_from_frealix_par_noinit

    !>  \brief  Read a frealix parameter file (specified version and return an array of initialised filament objects)
    subroutine filaments_from_frealix_par(  fils,par_filename,film_filename_wild,pixel_size,helix,              &
                                            ctftilt_filename_wild,ctffind_filename_wild,                        &
                                            wps_are_crossovers,fal_wps_are_crossovers,                          &
                                            max_radius,flx_par_version,                                         &
                                            microtubule,mt_pf_number,mt_rise,                                   &
                                            rigid_params,constant_rise_per_subunit,constant_twist_per_length,   &
                                            wps_as_sps,refine_params,persistence_length,                        &
                                            torsion_persistence_length,stretch_persistence_length,              &
                                            fal_wps_in_score,fal_wps_in_threed,                                 &
                                            refine_normal_modes,films_exist,                                    &
                                            average_z_to_ctftilt_plane,use_hermite_splines,                     &
                                            fragmented_phi,fragmented_subunit_num,amyloid_beta,                 &
                                            shift_fils_and_set_films_dim,                                       &
                                            acceleration_voltage,spherical_aberration,amplitude_contrast,       &
                                            add_fal_wps,                                                        &
                                            unblur_shifts,box_size_for_subframe_alignment,                      &
                                            recompute_subframe_alignments,                                      &
                                            number_of_frames,exposure_per_frame,apply_dose_filter,              &
                                            helix_min_for_rec,helix_max_for_rec,                                &
                                            new_film_dimensions)
        use filaments_model_base, only : filament_print_extra_info
		use filaments_base
		use filaments, only : filament_init
		use Micrographs, only : Micrograph
		use helices, only : helical_symmetry
        use parsers
        use coos, only : waypoints_print_info
        use ctfs, only : ctf_parameters_from_numeric_text_file, ctf_parameters
        use runtime_parameters, only : filename_len, terminate, is_odd
        use StringManipulations
        implicit none
        ! Arguments
        type(filament), allocatable, target,   intent(inout)::  fils(:)
        character(len=*),                      intent(in)   ::  par_filename                        !<  Filename for parameter file
        character(len=*),                      intent(in)   ::  film_filename_wild                  !<  Filename for the films, with ? wildcard
        real,                                  intent(in)   ::  pixel_size                          !<  Pixel size (in Angstroms)
        type(helical_symmetry),                intent(in)   ::  helix                               !<  (Average) helical parameters for the filaments.
        character(len=*),          optional,   intent(in)   ::  ctftilt_filename_wild               !<  Filename for the ctftilt results, with ? wildcard
        character(len=*),          optional,   intent(in)   ::  ctffind_filename_wild               !<  Filename for the ctffind results, with ? wildcard
        !integer,                  optional,   intent(in)   ::  num_subunits_btwn_wps_mulitple_of   !<  The number of subunits between successive waypoints should be a multiple of this integer number, if given. If not given, defaults to 1.
        logical,                   optional,   intent(in)   ::  wps_are_crossovers                  !<  Whether the way points in the parameter file define crossover positions
        logical,                   optional,   intent(in)   ::  fal_wps_are_crossovers              !<  Whether the first and last waypoints of each filaments are also crossovers
        real,                      optional,   intent(in)   ::  max_radius                          !<  Filament's maximum outer radius (pixels)
        integer,                   optional,   intent(in)   ::  flx_par_version                     !<  Frealix parameter file format version. Will override whatever version the parameter file may say it is.
        logical,                   optional,   intent(in)   ::  microtubule                         !<  Whether the filaments are microtubules
        integer,                   optional,   intent(in)   ::  mt_pf_number                        !<  If microtubules, how many protofilaments
        integer,                   optional,   intent(in)   ::  mt_rise                             !<  Microtubule: How much further along the filament (distance measured in tubulin subunits)
                                                                                                    !!  the end of a MT_PF_NUMBER-element "washer" is compared to its beginning. The 13-pf microtobules have a value of 3.
        logical,                   optional,   intent(in)   ::  rigid_params(9)                     !<  Flags for parameters to be held rigid (i.e. per-filament rather than per-waypoint)
        logical,                   optional,   intent(in)   ::  constant_rise_per_subunit           !<  Whether the rise per subunit is assumed to be constant
        logical,                   optional,   intent(in)   ::  constant_twist_per_length           !<  Whether the twist per unit length along the filament is assumed to be constant
        logical,                   optional,   intent(in)   ::  wps_as_sps                          !<  Whether we are treating waypoints (wps) as single particles (sps)
        logical,                   optional,   intent(in)   ::  refine_params(9)                    !<  Which parameters will be allowed to be refined
        real,                      optional,   intent(in)   ::  persistence_length                  !<  Filament persistence length (pixels)
        real,                      optional,   intent(in)   ::  torsion_persistence_length          !<  Torsional persistence length (pixels)
        real,                      optional,   intent(in)   ::  stretch_persistence_length          !<  Stretch persistence length (pixels)
        logical,                   optional,   intent(in)   ::  fal_wps_in_score                    !<  If false, the parts of the filament between the first and the second as well as between the penultimate
                                                                                                    !!  and last waypoints will not be included in the scoring function
        logical,                   optional,   intent(in)   ::  fal_wps_in_threed                   !<  If false, the parts of the filament between the first and the second as well as between the penultimate
                                                                                                    !!  and last waypoints will not be included in the 3d reconstruction
        logical,                   optional,   intent(in)   ::  refine_normal_modes(9)              !<  Whether normal mode amplitudes should be refined, rather than the filament model directly
        logical,                   optional,   intent(in)   ::  films_exist                         !<  In some cases, we want to initialise filaments without having access to the micrograph images
        logical,                   optional,   intent(in)   ::  average_z_to_ctftilt_plane          !<  Shift the filament in z so that its average lies on the plane described by the ctftilt plane
        logical,                   optional,   intent(in)   ::  use_hermite_splines                 !<  Use Hermite splines (where psi, theta and therefore the tangent at waypoints constrain the splines)
        logical,                   optional,   intent(in)   ::  fragmented_phi                      !<  Phi function is broken between waypoints, but the axial space curve is still continuous
        logical,                   optional,   intent(in)   ::  fragmented_subunit_num              !<  Subunit_num function is broken between waypoints, but the axial space curve is still continuous
        logical,                    optional,   intent(in)  ::  amyloid_beta                        !<  The filaments are amyloid beta fibrils
        logical,                    optional,   intent(in)  ::  shift_fils_and_set_films_dim        !<  Shift filaments and set film dimensions (useful when films not available, e.g. when making synthetic micrographs)
        real,                       optional,   intent(in)  ::  acceleration_voltage                !<  kV
        real,                       optional,   intent(in)  ::  spherical_aberration                !<  mm
        real,                       optional,   intent(in)  ::  amplitude_contrast                  !<  fraction
        logical,                    optional,   intent(in)  ::  add_fal_wps                         !<  Add waypoints at the beginning and the end of the filament
        real,     allocatable,      optional,   intent(in)  ::  unblur_shifts(:,:,:)                !<  In pixels. Frame shifts determined by unblur. First dimension: 1,2,3 for x,y,z; second dimension: frame number; third dimension: micrograph number
        integer,                    optional,   intent(in)  ::  box_size_for_subframe_alignment     !<  In pixels. Size of the box to cut out around waypoints when doing local frame alignments
        logical,                    optional,   intent(in)  ::  recompute_subframe_alignments       !<  On initiation of filaments, recompute local subframe shifts, starting from the whole-frame shifts determined by unblur.
        integer,                    optional,   intent(in)  ::  number_of_frames                    !<  Number of movie frames. Needed when dealing with older waypoints file which do not store subframe shift information.
        real,                       optional,   intent(in)  ::  exposure_per_frame                  !<  Number of electrons per square Angstroms, per frame
        logical,                    optional,   intent(in)  ::  apply_dose_filter                   !<  Apply dose (radiation damage) filter to micrographs, in particular when recomputing subframe shifts
        type(helical_symmetry),     optional,   intent(in)  ::  helix_min_for_rec                   !<  Smallest values of the helical parameters when selecting which waypoints to allow into the 3D reconstruction
        type(helical_symmetry),     optional,   intent(in)  ::  helix_max_for_rec                   !<  Largest values of the helical parameters when selecting which waypoints to allow into the 3D reconstruction
        integer,                    optional,   intent(in)  ::  new_film_dimensions(2)              !<  When making synthetic micrographs, can specify dimensions of  micrograph

        ! Private variables
        logical                             ::  ctftilt_file
        logical                             ::  ccrossovers
        type(waypoint),         allocatable ::  wp(:)
        integer                             ::  i,j,current_wp
        character(len=FILENAME_LEN)         ::  film_filename, ctftilt_filename, ctffind_filename
        type(Micrograph)                    ::  film1
        logical                             ::  ffilms_exist, sshift_fils_and_set_films_dim
        type(ctf_parameters),   allocatable ::  ctf_params(:)
        integer                             ::  nnumber_of_frames
        real,                   allocatable ::  unblur_shifts_for_current_micrograph(:,:)
        integer                             ::  nnew_film_dimensions(2)
        real(kind=8),           parameter   ::  micrograph_margin = 600 ! In Angstroms
        logical,                parameter   ::  debug   =   .false.
        ! Start work

        if (present(flx_par_version)) then
            write(*,'(a)') '**ERROR(filaments_from_frealix_par): optional dummy argument flx_par_version not implemented yet'
            call terminate('Feature not implemented yet')
        endif

        !  Read in the parameter file and populate the filament objects without initialising them, and without film data
        call filaments_from_frealix_par_noinit(fils,par_filename,pixel_size=pixel_size,number_of_frames=number_of_frames)

        nnumber_of_frames = size(fils(1)%wp(1)%local_subframe_shift,2)
        if (present(number_of_frames)) nnumber_of_frames = number_of_frames

        ! Check the user gave us either a ctffind or a ctftilt results file
        if (.not. present(ctftilt_filename_wild) .and. .not. present(ctffind_filename_wild)) then
            write(*,'(a)') '**ERROR(filaments_from_frealix_par): we need either ctffind or ctftilt results for each film'
            call terminate('Fatal error in FREALIX_PAR_TO_FILAMENTS')
        else if (present(ctftilt_filename_wild) .and. present(ctffind_filename_wild)) then
            if (.not. StringsAreEqual(ctftilt_filename_wild,'none')) then
                if (.not. StringsAreEqual(ctffind_filename_wild,'none')) then
                    write(*,'(2a)') '**info(filaments_from_frealix_par): both ctffind and ctftilt were given - ',  &
                                    'CTFTILT will be used, ctffind ignored'
                endif
                ctftilt_file = .true.
            else
                ctftilt_file = .false.
            endif
        else if (present(ctftilt_filename_wild)) then
            ctftilt_file = .true.
        else
            ctftilt_file = .false.
        endif

        ! Crossovers?
        ccrossovers = .false.
        if (present(wps_are_crossovers)) ccrossovers = wps_are_crossovers

        ! Films exist?
        ffilms_exist = .true.
        if (present(films_exist)) ffilms_exist = films_exist

        sshift_fils_and_set_films_dim = .false.
        if (present(shift_fils_and_set_films_dim)) sshift_fils_and_set_films_dim = shift_fils_and_set_films_dim

        nnew_film_dimensions = 0
        if (present(new_film_dimensions)) nnew_film_dimensions = new_film_dimensions

        ! Loop over the filaments
        current_wp = 1
        j = 0   ! This is the current line/waypoint
        do i=1,size(fils)

            ! Set up the film - we can do this by parsing either ctftilt or ctffind
            ! If the user has given us a ctftilt file, we use that.
            if (filename_contains_qmarks(film_filename_wild)) then
                call filename_from_wild(film_filename_wild,fils(i)%film%GetID(),film_filename)
                call fils(i)%film%SetLocationInFile(1)
            else
                film_filename = film_filename_wild
                call fils(i)%film%SetLocationInFile(fils(i)%film%GetID())
            endif
            if (ctftilt_file) then
                call filename_from_wild(ctftilt_filename_wild,fils(i)%film%GetID(),ctftilt_filename)
                call film1%Init(fils(i)%film%GetID(),film_filename,pixel_size,  &
                                ctftilt_filename=ctftilt_filename, &
                                film_image_available=ffilms_exist, &
                                number_of_frames=nnumber_of_frames, &
                                exposure_per_frame=exposure_per_frame)
            else
                if (filename_contains_qmarks(ctffind_filename_wild)) then

                    call filename_from_wild(ctffind_filename_wild,fils(i)%film%GetID(),ctffind_filename)
                    call film1%Init(fils(i)%film%GetID(),film_filename,pixel_size,  &
                                    ctffind_filename=ctffind_filename,  &
                                    film_image_available=ffilms_exist, &
                                    number_of_frames=nnumber_of_frames, &
                                    exposure_per_frame=exposure_per_frame)
                else
                    !! We were given a single file which contains new-style ctffind results
                    !! Let's check we were given acceleration voltage etc and then load the ctf results in
                    !if (.not. (present(acceleration_voltage) .and. present(spherical_aberration) &
                    !                                         .and. present(amplitude_contrast))) then
                    !    call terminate('Missing acceleration voltage, spherical aberration and/or amplitude contrast')
                    !endif
                    !call ctf_parameters_from_numeric_text_file(ctf_params,ctffind_filename_wild, &
                    !                                           acceleration_voltage,spherical_aberration,amplitude_contrast)
                    !call film1%Init(fils(i)%film%GetID(),film_filename,pixel_size, &
                    !                film_image_available=ffilms_exist, &
                    !                ctf_params=ctf_params(fils(i)%film%GetLocationInFile()), &
                    !                number_of_frames=nnumber_of_frames, &
                    !                exposure_per_frame=exposure_per_frame)
                    call film1%Init(fils(i)%film%GetID(),film_filename,pixel_size,  &
                                    ctffind_filename=ctffind_filename_wild,  &
                                    film_image_available=ffilms_exist, &
                                    number_of_frames=nnumber_of_frames, &
                                    exposure_per_frame=exposure_per_frame)
                endif
            endif

            ! Make sure the film is updated in the filament object
            fils(i)%film = film1

            if (debug) then
                write(*,'(a)') '**debug(filaments_from_frealix_par): filament info before init'
                call fils(i)%PrintInfo()
            endif

            if (ffilms_exist .or. sshift_fils_and_set_films_dim .or. all(nnew_film_dimensions .ne. 0)) then
                ! Copy the waypoint info into a separate waypoints array
                if (allocated(wp)) deallocate(wp)
                allocate(wp(size(fils(i)%wp)))
                wp = fils(i)%wp

                ! Move the filament so its min x,y are near the origin, then set the micrograph
                ! dimensions to be so that the filament just fits in (with additional margin)
                if (all(nnew_film_dimensions .ne. 0)) then
                    call film1%SetLogicalDimension(1,nnew_film_dimensions(1))
                    call film1%SetLogicalDimension(2,nnew_film_dimensions(2))
                endif
                if (sshift_fils_and_set_films_dim) then
                    ! Shift filament
                    wp%coo(1) = wp%coo(1) - minval(wp%coo(1)) + micrograph_margin/pixel_size
                    wp%coo(2) = wp%coo(2) - minval(wp%coo(2)) + micrograph_margin/pixel_size
                    ! Set micrograph dimensions
                    call film1%SetLogicalDimension(1,nint(maxval(wp%coo(1)) + micrograph_margin/pixel_size))
                    call film1%SetLogicalDimension(2,nint(maxval(wp%coo(2)) + micrograph_margin/pixel_size))
                    if (is_odd(film1%GetLogicalDimension(1))) call film1%SetLogicalDimension(1,film1%GetLogicalDimension(1) + 1)
                    if (is_odd(film1%GetLogicalDimension(2))) call film1%SetLogicalDimension(2,film1%GetLogicalDimension(2) + 1)
                    ! We have to do this temporarily to make it square, to accomodate for crappy restriction on image_apply_ctf
                    call film1%SetLogicalDimension(1,maxval(film1%GetLogicalDimensions()))
                    call film1%SetLogicalDimension(2,maxval(film1%GetLogicalDimensions()))
                endif

                if (debug) then
                    write(*,'(a)') '**debug(filaments_from_frealix_par): waypoints to be used by filament_init:'
                    call waypoints_print_info(wp)
                endif


                ! Unblur shifts?
                if (present(unblur_shifts)) then
                    if (allocated(unblur_shifts)) then
                        if (film1%GetID() .le. size(unblur_shifts,3)) then
                            unblur_shifts_for_current_micrograph = unblur_shifts(:,:,film1%GetID())
                        else
                            if (any(unblur_shifts .ne. 0.0)) then
                                write(*,'(2a,i0)') '**warning(filaments_from_frealix_par): ', &
                                                    'did not find frame shifts for micrograph ',&
                                                film1%GetID()

                            endif
                        endif
                    endif
                endif

                if (.not. allocated(unblur_shifts_for_current_micrograph)) then
                    allocate(unblur_shifts_for_current_micrograph(3,1))
                    unblur_shifts_for_current_micrograph = 0.0
                endif

                ! Initialise a filament object
                call filament_init( fils(i),waypoints_in=wp,film_in=film1,helix=helix,crossovers=ccrossovers, &
                                fal_wps_are_crossovers=fal_wps_are_crossovers,&
                                id=fils(i)%id,max_radius=max_radius,    &
                                microtubule=microtubule,mt_pf_number=mt_pf_number,mt_rise=mt_rise,rigid_params=rigid_params,    &
                                constant_rise_per_subunit=constant_rise_per_subunit,    &
                                constant_twist_per_length=constant_twist_per_length,    &
                                wps_as_sps=wps_as_sps,refine_params=refine_params,persistence_length=persistence_length,    &
                                torsion_persistence_length=torsion_persistence_length,  &
                                stretch_persistence_length=stretch_persistence_length,  &
                                fal_wps_in_score=fal_wps_in_score,fal_wps_in_threed=fal_wps_in_threed,  &
                                refine_normal_modes=refine_normal_modes,    &
                                average_z_to_ctftilt_plane=average_z_to_ctftilt_plane,use_hermite_splines=use_hermite_splines, &
                                fragmented_phi=fragmented_phi,fragmented_subunit_num=fragmented_subunit_num,    &
                                amyloid_beta=amyloid_beta, &
                                add_extra_fal_wps=add_fal_wps, &
                                recompute_subframe_shifts_from_unblur_results=recompute_subframe_alignments, &
                                frame_shifts_from_unblur=unblur_shifts_for_current_micrograph, &
                                subframe_alignment_box_size=box_size_for_subframe_alignment, &
                                apply_dose_filter=apply_dose_filter, &
                                helix_min_for_3drec=helix_min_for_rec, &
                                helix_max_for_3drec=helix_max_for_rec)
            else
                write(*,'(a)') '**debug(filaments_from_frealix_par): no film, filament not initialised'
            endif

            ! Diagnostics: print filament info
            if (debug) then
                write(*,'(a)') '**debug(filaments_from_frealix_par): filament info on return:'
                call fils(i)%PrintInfo()
                call filament_print_extra_info(fils(i))
            endif

        enddo

        if (debug) write(*,'(a)') '**debug(filaments_from_frealix_par): all done'

    end subroutine filaments_from_frealix_par


    !>  \brief  Parse a Frealix1 parameter file and return an array of filament objects
    !!  Either ctffind or ctftilt results must be available, for all films
    function frealix1_par_to_filaments( filename,film_filename_wild,pixel_size,helix,   &
                                        ctftilt_filename_wild,ctffind_filename_wild,crossovers) result(fils)
        use filaments_base
        use filaments, only : filament_init
        use parsers
        use runtime_parameters, only : filename_len, terminate
        use Micrographs, only : Micrograph
        implicit none
        ! Arguments
        character(len=*),               intent(in)  ::  filename                            !<  Filename of frealix1 parameter file
        character(len=*),               intent(in)  ::  film_filename_wild                  !<  Filename for the films, with ? wildcard
        real,                           intent(in)  ::  pixel_size                          !<  Pixel size (in Angstroms)
        type(helical_symmetry),         intent(in)  ::  helix                               !<  (Average) helical parameters for the filaments.
        character(len=*),   optional,   intent(in)  ::  ctftilt_filename_wild               !<  Filename for the ctftilt results, with ? wildcard
        character(len=*),   optional,   intent(in)  ::  ctffind_filename_wild               !<  Filename for the ctffind results, with ? wildcard
        !integer,           optional,   intent(in)  ::  num_subunits_btwn_wps_mulitple_of   !<  The number of subunits between successive waypoints should be a multiple of this integer number, if given. If not given, defaults to 1.
        logical,            optional,   intent(in)  ::  crossovers                          !<  Whether the way points in the parameter file describe crossovers

        ! Result
        type(filament), allocatable ::  fils(:) !<  Output array of filaments
        ! Private variables
        integer,        allocatable ::  fil_nums(:)
        integer                     ::  num_of_fils, num_of_wps, i, j, j_start, current_wp, ierr
        real,           allocatable ::  tmp(:)
        integer,        allocatable ::  itmp(:), ifil(:), film_num(:)
        type(waypoint), allocatable ::  wp(:)
        real,           allocatable ::  phi(:), theta(:), psi(:), shx(:), shy(:)
        character(len=FILENAME_LEN) ::  film_filename, ctftilt_filename, ctffind_filename
        logical                     ::  ctftilt_file
        type(Micrograph)                  ::  film1
        !integer                        ::  nnum_subunits_btwn_wps_mulitple_of
        logical                     ::  ccrossovers
        ! Start work

        ! Check the parameter file exists
        if (.not. file_exists(filename)) then
            write(*,'(a,a)') '**ERROR(frealix1_par_to_fil): file does not exist: ', trim(filename)
            stop 'Fatal error in FREALIX1_PAR_TO_FIL'
        endif

        ! Check the user gave us either a ctftilt or a ctftilt results file
        if (.not. present(ctftilt_filename_wild) .and. .not. present(ctffind_filename_wild)) then
            write(*,'(a)') '**ERROR(frealix1_par_to_fil): we need either ctffind or ctftilt results for each film'
            stop 'Fatal error in FREALIX1_PAR_TO_FIL'
        else if (present(ctftilt_filename_wild) .and. present(ctffind_filename_wild)) then
            write(*,'(2a)') '**INFO(frealix1_par_to_fil): both ctffind and ctftilt were given - ',   &
                            'CTFTILT will be used, ctffind ignored'
            ctftilt_file = .true.
        else if (present(ctftilt_filename_wild)) then
            ctftilt_file = .true.
        else
            ctftilt_file = .false.
        endif

        !! num_subunits_btwn_wps_mulitple_of
        !if (present(num_subunits_btwn_wps_mulitple_of)) then
        !   nnum_subunits_btwn_wps_mulitple_of = num_subunits_btwn_wps_mulitple_of
        !else
        !   nnum_subunits_btwn_wps_mulitple_of = 1
        !endif

        if (present(crossovers)) then
            ccrossovers = crossovers
        else
            ccrossovers = .false.
        endif

        ! Work out number of filaments
        call get_ints_from_column(filename,16,fil_nums)
        num_of_fils = maxval(fil_nums)

        ! Work out number of waypoints
        num_of_wps  =   number_of_lines_in_file(filename,ignore_comments=.true.)

        ! Allocate the output array and temporary arrays
        if (allocated(fils)) deallocate(fils)
        allocate(fils(num_of_fils),tmp(num_of_wps),itmp(num_of_wps),stat=IERR)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(frealix1_par_to_fil): memory allocation failed with error ', ierr
            call terminate('Fatal error in FREALIX1_PAR_TO_FIL')
        endif

        ! Read in the columns that we'll need
        call get_reals_from_column(filename,2,psi)
        call get_reals_from_column(filename,3,theta)
        call get_reals_from_column(filename,4,phi)
        call get_reals_from_column(filename,5,shx)
        call get_reals_from_column(filename,6,shy)
        call get_ints_from_column(filename,8,film_num)
        call get_ints_from_column(filename,16,ifil)

        if (size(ifil) .ne. num_of_wps) then
            write(*,'(a,i6,a,i6,a)') '**ERROR(frealix1_par_to_fil): oops, number of lines (', num_of_wps ,&
                                    ') not equal to number of ifil values read in (',size(ifil),')'
            call terminate('Fatal error in FREALIX1_PAR_TO_FIL')
        endif

        ! Loop over the filaments
        current_wp = 1
        j = 0   ! This is the current line/waypoint
        do I=1,num_of_fils
            ! At the beginning of the iteration, j points to the last waypoint of the previous filament
            j = J+1
            j_start = j

            ! For the current filament, work out how many waypoints are defined in the parameter file
            ! If we're on the last waypoint, of course we don't need to do this
            if (j .lt. num_of_wps) then
                ! Let's increment j until we reach the last waypoint in the current filament, i
                do
                    if (ifil(J+1) .eq. ifil(j)) then
                        ! We are not yet at last way point of current filament
                        j = j +1
                    else
                        ! We are at last waypoint of current filament
                        exit
                    endif
                enddo
            else
                ! We're on the last waypoint. This is strange, cos it implies the last way point is "alone"
                write(*,'(a)') '**ERROR(frealix1_par_to_fil): last way point appears to be alone'
                stop 'Fatal error in FREALIX1_PAR_TO_FIL'
            endif

            ! So now j is set to the position of the last way point of the current filament

            ! Diagnostic message
            print '(a,i3,a,i2,a)', '**debug(frealix1_par_to_fil): filament ', ifil(j), ' has ', J-J_START+1, ' way points'

            ! Allocate memory to an array of way points
            if (allocated(wp)) deallocate(wp)
            allocate(wp(J-J_START+1),stat=IERR)
            if (ierr .ne. 0) then
                write(*,'(a,i0)') '**ERROR(frealix1_par_to_fil): memory allocation of temporary waypoint array failed with err ', &
                                    ierr
                call terminate('Fatal error in FREALIX1_PAR_TO_FIL')
            endif

            ! Populate the waypoints
            wp(1:j-j_start+1)%coo(1)        =   shx(j_start:j)
            wp(1:j-j_start+1)%coo(2)        =   shy(j_start:j)
            wp(1:j-j_start+1)%euler%phi     =   phi(j_start:j)
            wp(1:j-j_start+1)%euler%theta   =   theta(j_start:j)
            wp(1:j-j_start+1)%euler%psi     =   psi(j_start:j)

            ! Set up the film - we can do this by parsing either ctftilt or ctffind
            ! If the user has given us a ctftilt file, we use that.
            call filename_from_wild(film_filename_wild,film_num(j_start),film_filename)
            if (ctftilt_file) then
                call filename_from_wild(ctftilt_filename_wild,film_num(j_start),ctftilt_filename)
                call film1%Init(film_num(j_start),film_filename,pixel_size,ctftilt_filename=ctftilt_filename)
            else
                call filename_from_wild(ctffind_filename_wild,film_num(j_start),ctffind_filename)
                call film1%init(film_num(j_start),film_filename,pixel_size,ctffind_filename=ctffind_filename)
            endif

            ! Initialise a filament object
            call filament_init(fils(i),waypoints_in=wp,film_in=film1,helix=helix,crossovers=ccrossovers)

            ! Diagnostics: print filament info
            call fils(i)%PrintInfo()

        enddo

    end function frealix1_par_to_filaments

end module frealix_parameters
