!>  \brief  Classes and methods to deal with helical symmetries, helical objects etc.
!!
module Helices

    implicit none
    private
    public  ::  helical_symmetry, helical_symmetry_init, helical_symmetry_print_info, helix_print_info,helical_symmetry_unit_convert
    public  ::  helix_to_one_start,helix_pitch
    public  ::  helix_phi_from_xy, helix_xy_from_phi_r
    public  ::  helical_symmetry_reset, helical_symmetry_print_extra_info
    public  ::  helix_unit_tests
    public	::	helix_point_is_in_asym, helix_point_to_asym, helix_point_is_near_asym
    public  ::  helix_find_layer_lines
    public  ::  operator(+), operator(-), operator(*), operator(/), sqrt
    public  ::  helix_coos_cylindrical_to_lattice, helix_coos_lattice_to_cylindrical
    public  ::  helix_point_weight_from_asym, helix_point_to_other_side_of_asym
    public  ::  helix_scale_distances


    ! Tolerances for the boundaries of the asymmetric unit
    real(kind=8),   parameter   ::  z_tol = epsilon(0.0D0) * 10.0D0
    real(kind=8),   parameter   ::  phi_tol = epsilon(0.0D0) * 10.0D0

    !>  \brief  Describes helical lattice parameters
    !!          For an explanation of how helical objects can be described as rolled-up 2d crystal lattices, see
    !!          Chapter 12 "Electron crystallography of helical structures" by David DeRosier, in
    !!          "Electron crystallography of biological macromolecules" (Eds. Glaeser, Downing, DeRosier, Chiu & Frank)
    !!
    type    ::  helical_lattice
        private
        logical ::  pixel_units = .false.   !<  Whether the lattice parameters are currently in pixel units
        real    ::  a                       !<  Unit cell dimension 1
        real    ::  b                       !<  Unit cell dimension 2
        real    ::  angle                   !<  Unit cell angle
        integer ::  circum_vector(2)        !<  Circumferential vector
    end type

    !>  \brief  Describes a helical symmetry
    !!          Again, for explanations, i recommend the DeRosier chapter in that book
    type    ::  helical_symmetry
        logical         ::  pixel_units             =   .false.     !<  Whether the rise parameter is in pixels or Angstroms
        real(kind=8)    ::  rise                    =   0.0D0       !<  The rise (in pixels/Angstroms), i.e. the translation along the helical axis to get from one subunit to the next
        real(kind=8)    ::  twist                   =   0.0D0       !<  The twist (in degrees), i.e. the rotation around the helical axis to get from one subunit to the next (positive angle implies left-handed helix)
        integer         ::  axial_symmetry          =   1           !<  Axial symmetry. If axial_symmetry = n, the helix is N-start.
        integer         ::  perpendicular_symmetry  =   1           !<  Symmetry with axis perpendicular to helix axis. This can either be 1 or 2. 2 means the helix is not polar, like dna for example.
        integer         ::  num_starts              =   1           !<  Indicate which start number the helical rise and twist refer to. It may make more sense or be more intuitive to describe helices not with the 1-start helix in some situations.
        logical         ::  stack                   =   .false.     !<  The helical assembly is a stack rather than a true helix
	        real            ::  outer_radius            =   0.0         !<  Outer radius of the helix (in pixels/Angstroms)
        ! Description of the equivalent helix with the smallest twist (i.e. the "fastest" helix)
        integer         ::  fast_helix_start        =   0           !<  The start number with the smallest twist
        real(kind=8)    ::  fast_helix_rise         =   0.0D0
        real(kind=8)    ::  fast_helix_twist        =   0.0D0
        ! the slowest helix
        integer         ::  slow_helix_start        =   0
        real(kind=8)    ::  slow_helix_rise         =   0.0D0
        real(kind=8)    ::  slow_helix_twist        =   0.0D0
        !
        real(kind=8)    ::  conv_mat(2,2)           =   0.0D0       !<  Matrix to convert (phi,z) pairs into (x,y), the lattice coordinates (distances to the fastest and 1-start helices, in units of asymetric units). See notes of 130618 and their edit on 130620/1. See also Figure 4 of Rohou & Grigorieff (2014)
        real(kind=8)    ::  inv_conv_mat(2,2)       =   0.0D0       !<  Matrix to convert from lattice to cylindrical coordinates
    end type

    interface assignment(=)
        module procedure helix_assignment
    end interface

    interface operator (+)
        module procedure helix_plus_helix
    end interface

    interface operator (-)
        module procedure helix_minus_helix
    end interface

    interface operator (*)
        module procedure helix_times_helix
        !module procedure helix_multiply_by_real
        !module procedure helix_multiply_by_integer
    end interface

    interface operator (/)
        module procedure helix_divide_by_real
    end interface

    interface sqrt
        module procedure helix_sqrt
    end interface

    interface helix_print_info
        module procedure helical_symmetry_print_info
    end interface

    interface helix_print_extra_info
        module procedure helical_symmetry_print_extra_info
    end interface

    contains

        pure subroutine helix_assignment(lhs,rhs)
            type(helical_symmetry), intent(inout)   ::  lhs
            type(helical_symmetry), intent(in)      ::  rhs
            !

            !
            lhs%pixel_units             =   rhs%pixel_units
            lhs%rise                    =   rhs%rise
            lhs%twist                   =   rhs%twist
            lhs%axial_symmetry          =   rhs%axial_symmetry
            lhs%perpendicular_symmetry  =   rhs%perpendicular_symmetry
            lhs%num_starts              =   rhs%num_starts
            lhs%stack                   =   rhs%stack
            lhs%outer_radius            =   rhs%outer_radius
            lhs%fast_helix_start        =   rhs%fast_helix_start
            lhs%fast_helix_rise         =   rhs%fast_helix_rise
            lhs%fast_helix_twist        =   rhs%fast_helix_twist
            lhs%slow_helix_start        =   rhs%slow_helix_start
            lhs%slow_helix_rise         =   rhs%slow_helix_rise
            lhs%slow_helix_twist        =   rhs%slow_helix_twist
            lhs%conv_mat                =   rhs%conv_mat
            lhs%inv_conv_mat            =   rhs%inv_conv_mat
        end subroutine helix_assignment

        subroutine helix_unit_tests()
            implicit none
            !
            call helix_unit_test_3()
            call helix_unit_test_2()
            call helix_unit_test_1()
        end subroutine helix_unit_tests

        subroutine helix_unit_test_1()
			use rotation_matrices, only : radians
			use runtime_parameters, only : terminate
            implicit none
            ! Private variables
            type(helical_symmetry), allocatable  ::  self(:)
            integer ::  i, j
            integer ::  a,b,c
            integer, parameter :: voldim = 128
            integer, parameter :: ori(3) = [51,51,51]
            real(kind=8)    ::  x,y,z, z_new
            real(kind=8)    ::  phi, phi_new
            ! Start work
            write(*,'(a)') '**info(helix_unit_test_1): entering'
            allocate(self(4))
            ! Let's pretend we're dealing with tmv
            call helical_symmetry_init(self(1),rise=1.403/1.163,twist=-22.03,outer_radius=90.0/1.163,num_starts=1,stack=.false.)
            ! let's pretend we're dealing with abeta (true helix)
            call helical_symmetry_init(self(2),rise=2.4/1.163,twist=180.385,outer_radius=50.0/1.163,num_starts=1,stack=.false.)
            call helical_symmetry_init(self(3),rise=2.4/1.163,twist=-179.615,outer_radius=50.0/1.163,num_starts=1,stack=.false.)
            ! let's pretend we're dealing with abeta (stack)
            call helical_symmetry_init(self(4),rise=4.8/1.163,twist=0.77,outer_radius=50.0/1.163,num_starts=1,stack=.true.)
            do J=1,size(self)
                write(*,'(//a,i0,a,i0)') '**info(helix_unit_test_1): testing helical symmetry ', j , ' of ', size(self)
                call helix_print_info(self(j))
                call helix_print_extra_info(self(j))
                ! Print out the conversion matrix
                write(*,'(a)') '**info(helix_unit_test_1): conversion matrix'
                print '(/2(F15.6,1x) )', self(j)%conv_mat(1,1), self(j)%conv_mat(1,2)
                print '( 2(F15.6,1x)/)', self(j)%conv_mat(2,1), self(j)%conv_mat(2,2)
                write(*,'(a)') '**info(helix_unit_test_1): all distance values below should be integer'
                ! Loop over subunits
                do i=1,40
                    z   = real(i-1,kind=8)*self(j)%rise
                    phi = real(i-1,kind=8)*self(j)%twist
                    phi = mod(phi,360.0d0)+360.0d0
                    phi = helix_phi_to_c1(self(j),phi,z)
                    call helix_coos_cylindrical_to_lattice(self(j),z,phi,y,x)
                    print '(i3,1x,4(f11.3,1x))', i-1, z, phi, y, x
                    if (abs(y-nint(y)) .gt. 0.001 .or. abs(x-nint(x)) .gt. 0.001) then
                        write(*,'(a)') '**ERROR(helix_unit_test_1): bad distance values - should be integer'
                        call terminate('Failed HELIX_UNIT_TEST_1')
                    endif
                enddo
                write(*,'(a)') '**info(helix_unit_test_1): checking helix_point_to_asym'
                ! Checking that we can loop through a volume and do symmetrisation
                do C=1,voldim
                    z = real(C-ORI(3),KIND=8)
                    do B=1,voldim
                        y = real(B-ORI(2),KIND=8)
                        do A=1,voldim
                            x = real(A-ORI(1),KIND=8)
                            !
                            phi = helix_phi_from_xy(x,y)
                            call helix_point_to_asym(self(j),z,phi,z_new,phi_new)
                        enddo
                    enddo
                enddo
            enddo   ! End of loop over symmetries

            write(*,'(a)') '**info(helix_unit_test_1): all done'
        end subroutine helix_unit_test_1


        subroutine helix_unit_test_2()
            implicit none
            ! Private variables
            type(helical_symmetry)  ::  self
            integer,    allocatable ::  n(:)
            real,       allocatable ::  z(:)
            integer                 ::  i
            real,       parameter   ::  pixel_size = 1.163
            ! Start work
            write(*,'(a)') '**info(helix_unit_test_2): entering'
            call helical_symmetry_init(self,rise=1.403/pixel_size,twist=-22.03,outer_radius=90.0/1.163,num_starts=1,stack=.false.)
            !call helical_symmetry_init(self,rise=2.4/pixel_size,twist=179.7,num_starts=1,stack=.false.)
            call helix_find_layer_lines(self,z,n)

            write(*,'(a6,1x,2(a7,1x),a9)') 'Line #', 'Z', '1/Z', 'Order (n)'
            do I=1,size(n)
                if (abs(z(i)) .le. 0.5) then
                if (z(i) .ne. 0.) then
                    write(*,'(i6,1x,2(F7.2,1x),i9)') i, z(i)/pixel_size, 1./z(i)*PIXEL_SIZE, n(i)
                else
                    write(*,'(i6,1x,2(F7.2,1x),i9)') i, z(i)/pixel_size,               999., n(i)
                endif
                endif

            enddo
            write(*,'(a)') '**info(helix_unit_test_2): all done'

        end subroutine helix_unit_test_2

        subroutine helix_unit_test_3()
            use Globals
            implicit none
            ! Private variable
            type(helical_symmetry)          ::  self
            integer                         ::  counter
            real,               parameter   ::  pixel_size = 1.163
            integer,            parameter   ::  num_syms = 6
            character(len=5)                ::  labels(num_syms)
            real,               parameter   ::  rise(num_syms) = [1.408,2.4,4.8,4.8,14.8,4.9]
            real,               parameter   ::  twist(num_syms) = [-22.03,-179.5,0.8,0.8,118.0,157.0]
            real,               parameter   ::  outer_radius(num_syms) = [105.0,90.0,90.0,90.0,50.0,320.0]
            logical,            parameter   ::  stack(num_syms) = [.false.,.false.,.true.,.true.,.false.,.false.]
            integer,            parameter   ::  sym_axial(num_syms) = [1,1,1,2,1,1]
            integer,            parameter   ::  sym_perp(num_syms)  = [1,1,1,1,1,2]
            integer,            parameter   ::  slow(num_syms) = [1, 1,0,0,1,7 ]
            integer,            parameter   ::  fast(num_syms) = [16,2,1,1,3,16]
            logical                         ::  success(num_syms)
            ! Start work
            write(*,'(a)') '**info(helix_unit_test_3): entering'

            ! Setup the labels
            labels(1) = 'tmv  ' ! TMV
            labels(2) = 'ab   ' ! Abeta staggered
            labels(3) = 'abst ' ! Abeta stack asym
            labels(4) = 'absts' ! Abeta stack 2-sym
            labels(5) = 'phuz ' ! PhuZ
            labels(6) = 'serca' ! SERCA

            ! Write header for the results
            write(*,'(/a5,1x,a5,1x,a5,1x,a8)') 'Label','slow','fast','Success?'

            ! Loop over the tests
            do counter=1,num_syms
                ! Initialise the helical symmetry
                call helical_symmetry_init(self,    rise=rise(counter)/pixel_size, &
                                                    twist=twist(counter), &
                                                    outer_radius=outer_radius(counter)/pixel_size, &
                                                    axial_sym=sym_axial(counter), &
                                                    perpendicular_sym=sym_perp(counter), &
                                                    num_starts=1, &
                                                    stack=stack(counter))
                ! Check the results
                success(counter) =          self%slow_helix_start .eq. slow(counter) &
                                    .and.   self%fast_helix_start .eq. fast(counter)
                write(*,'(a5,1x,i2,a,i2,1x,i2,a,i2,1x,l8)') labels(counter),self%slow_helix_start,'/',slow(counter), &
                                                            self%fast_helix_start,'/',fast(counter),success(counter)
                if (.not. success(counter)) then
                    write(*,'(2a)') '**error(helix_unit_test_3): failed unit test: ', labels(counter)
                call this_program%TerminateWithFatalError('helix_unit_test_3','Failed')
                endif
            enddo

            if (.not. all(success)) then
                write(*,'(a)') '**error(helix_unit_test_3): failed unit test. See above for details.'
                call this_program%TerminateWithFatalError('helix_unit_test_3','Failed')
            endif

            write(*,'(a)') '**info(helix_unit_test_3): all done'

        end subroutine helix_unit_test_3

        !>  \brief  Return the layer line positions (in z) and their order (n)
        !!
        !!  We use the DeRosier "selection rule":
        !!
        !!  rise_per_subunit*Z - n*twist_per_subunit/2pi = l
        !!
        !!  so that z = (l + n*twist_per_subunit/2pi)/rise_per_subunit
        !!
        !!  where z is in reciprocal pixels,
        !!  rise_per_subunit is in pixels,
        !!  n and l are integers,
        !!  twist_per_subunit is in radians
        subroutine helix_find_layer_lines(self,zs,ns)
            use runtime_parameters, only: pi
            implicit none
            ! Arguments
            type(helical_symmetry),     intent(in)      ::  self            !<  Helical operator
            real,       allocatable,    intent(inout)   ::  zs(:)
            integer,    allocatable,    intent(inout)   ::  ns(:)
            ! Private variables
            integer,    parameter   ::  l_max = 200
            integer,    parameter   ::  n_max = 46
            integer                 ::  i,l,n
            real                    ::  twist_rad
            ! Start work

            ! Allocate output arrays
            if (allocated(zs)) deallocate(zs)
            if (allocated(ns)) deallocate(ns)
            allocate(zs(L_MAX*N_MAX*2),ns(L_MAX*N_MAX*2))

            ! The twist per subunit in radians
            twist_rad = self%twist / 180. * pi

            ! Get values of z and n
            I=0

            do L=0,L_MAX-1
                do N=0,N_MAX-1
                    i     = I+1
                    zs(i) = (l + N*TWIST_RAD/(2.*PI))/self%rise
                    ns(i) = n
                    i     = I+1
                    zs(i) = -ZS(I-1)
                    ns(i) = -NS(I-1)
                enddo
            enddo

        end subroutine helix_find_layer_lines

        !>  \brief  Move by one subunit to land on the other side of the primitive asymetric unit
        pure subroutine helix_point_to_other_side_of_asym(self,z,phi,z_new,phi_new)
            implicit none
            ! Arguments
            type(helical_symmetry), intent(in)      ::  self            !<  Helical operator
            real(kind=8),           intent(in)      ::  z,phi           !<  Coordinates of point
            real(kind=8),           intent(out)     ::  z_new,phi_new   !<  New coordinates
            ! Private variables
            real(kind=8)    ::  y,x
            integer         ::  k,h
			! Start work

            call helix_coos_cylindrical_to_lattice(self,z,phi,y,x)
            ! Move towards the 1-start axis using the fastest helix (smallest twist)
            if (abs(y) .gt. 0.5) h = -ceiling(y-0.5)
            ! Move towards the fastest helix axis using the 1-start helix
            if (abs(x) .gt. 0.5) k = -ceiling(x-0.5)

            if (y .gt. 0.0) then
                h = -1
            else
                h = 1
            endif
            if (x .gt. 0.0) then
                k = -1
            else
                k = 1
            endif

            ! Apply the move
            z_new   = z   + h*self%fast_helix_rise  + k*self%slow_helix_rise
            phi_new = phi + h*self%fast_helix_twist + k*self%slow_helix_twist

        end subroutine helix_point_to_other_side_of_asym


        !>  \brief  Move by an integer number of subunits to land in the original asymetric unit
        !!
        !!  See notes of 130618 and their edit on 130620/1 for details, nomenclature etc
#ifdef skip_runtime_checks
        pure &
#endif
        subroutine helix_point_to_asym(self,z,phi,z_new,phi_new)
#ifndef skip_runtime_checks
            use runtime_parameters, only : terminate
#endif
            implicit none
            ! Arguments
            type(helical_symmetry), intent(in)      ::  self            !<  Helical operator
            real(kind=8),           intent(in)      ::  z,phi           !<  Coordinates of point
            real(kind=8),           intent(out)     ::  z_new,phi_new   !<  New coordinates
            ! Private variables
            real(kind=8)    ::  y,x
            integer         ::  k,h
#ifndef skip_runtime_checks
			real(kind=8)	::	yn,xn
#endif
			! Start work

            ! First, take advantage of any axial symmetry
            call helix_point_to_c1(self,z,phi,z_new,phi_new)

			h = 0
			k = 0

            call helix_coos_cylindrical_to_lattice(self,z_new,phi_new,y,x)
            ! Move towards the 1-start axis using the fastest helix (smallest twist)
            if (abs(y) .gt. 0.5) h = -ceiling(y-0.5)
            ! Move towards the fastest helix axis using the 1-start helix
            if (abs(x) .gt. 0.5) k = -ceiling(x-0.5)

            ! Apply the move
            z_new   = z_new   + h*self%fast_helix_rise  + k*self%slow_helix_rise
            phi_new = phi_new + h*self%fast_helix_twist + k*self%slow_helix_twist

            call helix_coos_cylindrical_to_lattice(self,z_new,phi_new,yn,xn)


            ! Make sure we're back in the asym range of angles (-180,180 for c1)
            phi_new =  helix_phi_to_c1(self,phi_new,z_new)


#ifndef skip_runtime_checks
            ! Sanity check
            if (.not. helix_point_is_in_asym(self,z_new,phi_new)) then
				call helix_coos_cylindrical_to_lattice(self,z_new,phi_new,yn,xn)
				if (abs(yn)-0.5D0 .gt. 0.01 .or. abs(xn)-0.5D0 .gt. 0.01) then
                write(*,'(a)') '**ERROR(helix_point_to_asym): failed sanity check'
                call helix_print_info(self)
                call helix_print_extra_info(self)
                write(*,'(a,2(F0.3,1x))') 'Z, Phi on entry: ', z, phi
                write(*,'(a,2(F0.3,1x))') 'Y, x on entry  : ', y, x
				write(*,'(a,2(F0.3,1x))') 'Z, Phi on exit : ', z_new, phi_new
				write(*,'(a,2(F0.3,1x))') 'Y, x on exit   : ', yn, xn
				write(*,'(a,2(i0,1x))')   'K, h           : ', k, h
				write(*,*) ''
				call terminate('Fatal error in HELIX_POINT_TO_ASYM')
				endif
            endif
#endif
        end subroutine helix_point_to_asym

        !>  \brief  Move a point using axial symmetry operators to the c1 region
        pure subroutine helix_point_to_c1(self,z,phi,z_new,phi_new)
            use runtime_parameters, only : terminate
            implicit none
            ! Arguments
            type(helical_symmetry), intent(in)      ::  self            !<  Helical operator
            real(kind=8),           intent(in)      ::  z,phi           !<  Coordinates of point
            real(kind=8),           intent(out)     ::  z_new,phi_new   !<  New coordinates
            ! Private variables

            ! Start work
            z_new = z
            phi_new = helix_phi_to_c1(self,phi,z_new)
        end subroutine helix_point_to_c1

        !>  \brief  Return the phi coordinate of the axial-symmetry-mate of the given point that lies near phi=0.0
        pure function helix_phi_to_c1(self,phi,z)
            implicit none
            ! Arguments
            type(helical_symmetry), intent(in)      ::  self        !<  Helical operator
            real(kind=8),           intent(in)      ::  phi         !<  Coordinate of point
            real(kind=8),           intent(in)      ::  z           !<  Coordinate of point
            ! Result
            real(kind=8)                            ::  helix_phi_to_c1
            ! Start work
            ! Private variables
            real(kind=8)    ::  c1_phi_limits(2), c1_range
            integer         ::  n
            ! Start work

            c1_range     = 360.0D0 / real(self%axial_symmetry,kind=8)
            c1_phi_limits(2) = c1_range * 0.5d0 + z * self%fast_helix_twist / self%fast_helix_rise
            c1_phi_limits(1) = c1_phi_limits(2) - c1_range
            if ( phi .lt. c1_phi_limits(1)) then
                n =   ceiling((c1_phi_limits(1)-phi)/c1_range)
                helix_phi_to_c1 = phi + n * c1_range
            else if (phi .gt. c1_phi_limits(2)) then
                n = - ceiling((phi-c1_phi_limits(2))/c1_range)
                helix_phi_to_c1 = phi + n * c1_range
            else
                helix_phi_to_c1 = phi
            endif


        end function helix_phi_to_c1


        !>  \brief  Return the weight of a point as determined by its distance from the boundary of the primitive helical unit cell
        !!  \todo   Optimise
        pure real function helix_point_weight_from_asym(self,z,phi,r) result(weight)
            use rotation_matrices, only : radians
            implicit none
            ! Arguments
            type(helical_symmetry), intent(in)      ::  self    !<  Helical operator
            real(kind=8),           intent(in)      ::  z
            real(kind=8),           intent(in)      ::  phi
            real(kind=8),           intent(in)      ::  r
            ! Private variables
            real(kind=8)    :: lattice_dist_vec(2)
            real(kind=8)    :: x,y
            real(kind=8)    :: dist_z,dist_xy, dist_phi, dist
            real            :: weights(3)
            integer         :: i
            real(kind=8),   parameter    :: width = 3.0d0
            ! Start work

            ! Compute coordinates of the point in the coordinate system defined by the helical lattice.
            ! y is distance to slow (1-start helix). x is distance to fast helix.
            call helix_coos_cylindrical_to_lattice(self,z,phi,y,x)
            ! Convert to distances from edge of asymmetric unit (still in lattice coordinates)
            x = abs(x) - 0.5d0
            y = abs(y) - 0.5d0
            ! Draw the vector from the closest edge of the asym unit to the point
            if (x .le. 0.0d0 .and. y .le. 0.0d0) then
                ! we're inside the asym unit
                if (x .ge. y) then
                    ! we are closer to the fast helix than the slow one
                    lattice_dist_vec = [x,0.0d0]
                    call helix_coos_lattice_to_cylindrical(self,lattice_dist_vec(1),lattice_dist_vec(2),dist_z,dist_phi)
                    if (abs(dist_phi) .lt. 90.0) then
                        dist_xy = sin(radians(dist_phi)) * r
                    else
                        dist_xy = max(1.0,r)
                    endif
                    if (dist_xy .gt. 0.5d0*width) then
                        weight = 0.0
                    else if (dist_xy .lt. -0.5d0*width) then
                        weight = 1.0
                    else
                        weight = (0.5*width - dist_xy)/width
                    endif
                else
                    lattice_dist_vec = [0.0d0,y]
                    call helix_coos_lattice_to_cylindrical(self,lattice_dist_vec(1),lattice_dist_vec(2),dist_z,dist_phi)
                    if (dist_z .lt. -0.5d0*width) then
                        weight = 1.0
                    else if (dist_z .gt. 0.5d0*width) then
                        weight = 0.0
                    else
                        weight = (0.5*width - dist_z)/width
                    endif
                endif
                !lattice_dist_vec = 0.0d0
                !weight = 1.0
            else if (x .gt. 0.0d0 .and. y .gt. 0.0d0) then
                ! we're outside the asym unit, and the closest point is its corner
                lattice_dist_vec = [x,y]
                weight = 0.0
            else if (x .gt. 0.0d0) then
                ! we're outside, and the closest point is on the fast helix
                lattice_dist_vec = [x,0.0d0]
                call helix_coos_lattice_to_cylindrical(self,lattice_dist_vec(1),lattice_dist_vec(2),dist_z,dist_phi)
                if (abs(dist_phi) .lt. 90.0) then
                    dist_xy = sin(radians(dist_phi)) * r
                else
                    dist_xy = max(1.0,r)
                endif
                if (dist_xy .gt. 0.5d0*width) then
                    weight = 0.0
                else if (dist_xy .lt. -0.5d0*width) then
                    weight = 1.0
                else
                    weight = (0.5*width - dist_xy)/width
                endif
            else if (y .gt. 0.0d0) then
                ! we're outside and the closest point is on the slow helix edge
                lattice_dist_vec = [0.0d0,y]
                call helix_coos_lattice_to_cylindrical(self,lattice_dist_vec(1),lattice_dist_vec(2),dist_z,dist_phi)
                if (dist_z .lt. -0.5d0*width) then
                    weight = 1.0
                else if (dist_z .gt. 0.5d0*width) then
                    weight = 0.0
                else
                    weight = (0.5*width - dist_z)/width
                endif

            endif

            ! Now that we have a distance vector, let's convert it
            ! from lattice coords to cylindrical coords
            !call helix_coos_lattice_to_cylindrical(self,lattice_dist_vec(1),lattice_dist_vec(2),dist_z,dist_phi)

            ! Convert to distance in z and xy
            !if (abs(dist_phi) .lt. 90.0) then
            !    dist_xy = sin(radians(dist_phi)) * r
            !else
            !    dist_xy = max(1.0,r)
            !endif
            ! This works for the distance to surface when on outside I think
            !dist = sin(atan2(dist_xy,dist_z)) * dist_z
            !dist = min(dist_xy,dist_z)

            !if (abs(dist) .gt. 0.5d0) then
            !    weight = 0.0
            !else
            !    weight = 1.0 - 2.0*abs(dist)
            !endif


        end function helix_point_weight_from_asym

        !>  \brief  Check whether a point is within the original asymetric unit
        pure logical function helix_point_is_in_asym(self,z,phi)
            implicit none
            ! Arguments
            type(helical_symmetry), intent(in)      ::  self    !<  Helical operator
            real(kind=8),           intent(in)      ::  z,phi   !<  Coordinates of point
            ! Private variables
            real(kind=8),   parameter   ::  tol = 0.0D0
            ! Start work
            helix_point_is_in_asym = helix_point_is_near_asym(self,z,phi,tol)
        end function helix_point_is_in_asym

        !>  \brief  Check whether a point is near or within the original asymetric unit. Specify a tolerance in number of unit cells (see Rohou & Grigorieff 2014)
        pure logical function helix_point_is_near_asym(self,z,phi,tolerance)
            implicit none
            ! Arguments
            type(helical_symmetry), intent(in)      ::  self        !<  Helical operator
            real(kind=8),           intent(in)      ::  z,phi       !<  Coordinates of point
            real(kind=8),           intent(in)      ::  tolerance   !<  Tolerance, in number of unit cells along each
            ! Private variables
            real(kind=8)    ::  y,x
            ! Start work
            call helix_coos_cylindrical_to_lattice(self,z,phi,y,x)
            helix_point_is_near_asym = (abs(y) .le. 0.5D0+tolerance) .and. (abs(x) .le. 0.5D0+tolerance)
        end function helix_point_is_near_asym

        !>  \brief  Compute the coordinates of the point defined by (z,phi) in the coordinate system defined by the helical lattice.
        !!          See Figure 4 of Rohou & Grigorieff (2014)
        !!
        !!          Here is the matrix form of what's happening
        !!          / \
        !!          |x|    | 1,1  1,2 |   | phi |
        !!          |y|  = | 2,1  2,2 | * |  z  |
        !!          \ /
        !!
        pure subroutine  helix_coos_cylindrical_to_lattice(self,z,phi,y,x)
            use rotation_matrices, only : radians
            ! Arguments
            type(helical_symmetry), intent(in)      ::  self    !<  Helical operator
            real(kind=8),           intent(in)      ::  z,phi   !<  Coordinates of point (phi in degrees)
            real(kind=8),           intent(out)     ::  y      !<  Distance to the "1-start"/slow-helix axis, in unit cells
            real(kind=8),           intent(out)     ::  x      !<  Distance to the "fastest start" axis, in unit cells

            !
            x = radians(phi)*self%conv_mat(1,1)+z*self%conv_mat(1,2)
            y = radians(phi)*self%conv_mat(2,1)+z*self%conv_mat(2,2)

        end subroutine helix_coos_cylindrical_to_lattice

        !>  \brief  Transform lattice coordinates to cylindrical coordinates. See Figure 4 of Rohou & Grigorieff (2014)
        pure subroutine helix_coos_lattice_to_cylindrical(self,x,y,z,phi)
            use rotation_matrices, only : degrees
            ! Arguments
            type(helical_symmetry), intent(in)      ::  self    !<  Helical operator
            real(kind=8),           intent(in)      ::  x,y     !<  Lattice coordinates
            real(kind=8),           intent(out)     ::  z,phi   !<  Cylindrical coordinates (phi in degrees)

            !
            phi = degrees(x*self%inv_conv_mat(1,1)+y*self%inv_conv_mat(1,2))
            z   =         x*self%inv_conv_mat(2,1)+y*self%inv_conv_mat(2,2)
        end subroutine helix_coos_lattice_to_cylindrical


        !>  \brief  Update the conversion matrix
        subroutine helix_update_conversion_matrix(self)
            use rotation_matrices, only : radians
            use runtime_parameters, only : terminate
            implicit none
            ! Arguments
            type(helical_symmetry), intent(inout)   ::  self    !<  Helical operator
            ! Private variables
            real(kind=8)	::	det
            logical,    parameter   ::  sanity_checks = .false.
            ! Start work

            !
            self%conv_mat(1,1) =  self%fast_helix_rise
            self%conv_mat(2,1) = -self%slow_helix_rise
            self%conv_mat(1,2) = -radians(self%fast_helix_twist)
            self%conv_mat(2,2) =  radians(self%slow_helix_twist)

            det = self%conv_mat(2,2)*self%conv_mat(1,1) - self%conv_mat(1,2)*self%conv_mat(2,1)

            if (det .eq. 0.0) then
                !write(*,'(a)') '**debug(helix_update_conversion_matrix): det = 0.0  (1)'
                if (self%rise .eq. 0.0) then
                    ! If the rise is 0.0, it probably means this is a helical symmetry object
                    ! which is not going to be used to impose symmetry etc (it's probably just
                    ! an object to store rise/twist values for convenience)
                    ! so we just set the conversion matrices to 0.0
                    self%conv_mat = 0.0
                    self%inv_conv_mat = 0.0
                else
                    if (sanity_checks) then
                        write(*,'(a)') '**ERROR(helix_update_conversion_matrix): singular matrix'
                        call helical_symmetry_print_info(self)
                        call helical_symmetry_print_extra_info(self)
                        write(*,'(a,2(F0.3,1x))') 'twist, rise of S-start = ', self%slow_helix_twist, self%slow_helix_rise
                        write(*,'(a,2(F0.3,1x))') 'twist, rise of F-start = ', self%fast_helix_twist, self%fast_helix_rise
                        call terminate('Fatal error in HELIX_UPDATE_CONVERSION_MATRIX')
                    endif
                endif
            else
                !if (det .eq. 0.0) then
                !   write(*,'(a)') '**debug(helix_update_conversion_matrix): det = 0.0  (2)'
                !endif
                if (det .ne. 0.0) self%conv_mat = self%conv_mat / det

				! Now compute the inverse matrix
                self%inv_conv_mat(1,1) =  self%conv_mat(2,2)
                self%inv_conv_mat(2,1) = -self%conv_mat(2,1)
                self%inv_conv_mat(1,2) = -self%conv_mat(1,2)
                self%inv_conv_mat(2,2) =  self%conv_mat(1,1)

                det = self%inv_conv_mat(2,2)*self%inv_conv_mat(1,1) - self%inv_conv_mat(1,2)*self%inv_conv_mat(2,1)

                !if (det .eq. 0.0) then
                !    write(*,'(a)') '**debug(helix_update_conversion_matrix): det = 0.0  (3)'
                !endif
                if (det .ne. 0.0) self%inv_conv_mat = self%inv_conv_mat / det
            endif

        end subroutine helix_update_conversion_matrix


        !>  \brief  Guess the basis vectors for the natural unit cell, that is the unit cell whose boundary resembles
        !!          the boundaries of the protein subunit
        !!
        !!          See notes of 6-Aug-2015 for description of algorithm
        subroutine helix_guess_natural_basis_vectors(self,a,b)
            use Globals
            use UsefulFunctions, only : DegreesToRadians, RadiansToDegrees
            use euleranglesmod, only : degrees_m180_to_180
            use runtime_parameters, only : terminate
            ! Arguments
            class(helical_symmetry),    intent(inout)       ::  self
            integer,                    intent(out)         ::  a
            integer,                    intent(out)         ::  b
            ! Private variables
            integer,    parameter   ::  largest_index_to_test = 100
            real                    ::  arc_distance(largest_index_to_test)
            real                    ::  polar_angle(largest_index_to_test)
            integer                 ::  current_index
            real                    ::  smallest_arc_distance
            logical,    parameter   ::  debug = .false.
            type(helical_symmetry)  ::  h_temp
            real                    ::  x(largest_index_to_test),y(largest_index_to_test)
            real                    ::  distance_to_a
            integer                 ::  temp
            ! Start work

            if (abs(self%rise) .gt. 10000.0 .or. abs(self%twist) .gt. 100000.0) then
                call this_program%TerminateWithFatalError('helix_guess_natural_basis_vectors','Should not have been called')
            endif

            if (debug) then
                write(*,'(a)') '**debug(helix_guess_natural_basis_vectors): on entry'
                call helical_symmetry_print_info(self)
            endif

            a = 0
            b = 0

            ! Need to know the outer radius
            if (self%outer_radius .gt. 0.0) then

                ! Need to be using the one-start helix
                !call helix_to_one_start(self)
                if (self%num_starts .ne. 1) then
                    call this_program%TerminateWithFatalError('helix_guess_natural_basis_vectors','Please give 1-start helix')
                endif

                ! Compute the distance from each lattice point to the lattice point on the origin
                ! along the surface of the cylinder, as well as its polar angle
                do current_index=1,largest_index_to_test
                    x(current_index) = self%outer_radius*DegreesToRadians(degrees_m180_to_180(current_index*self%twist))
                    y(current_index) = current_index*self%rise
                    arc_distance(current_index) = sqrt(x(current_index)**2 +y(current_index)**2)
                    polar_angle(current_index) = RadiansToDegrees(atan2(y(current_index),x(current_index)))
                    if (debug) then
                        write(*,'(a,i0,2(1x,f0.2))') '**debug(helix_guess_natural_basis_vectors): index, dist, ang = ', &
                                                    current_index, arc_distance(current_index), &
                                                    polar_angle(current_index)
                    endif
                enddo

                ! Find the lattice point closest to the origin. This will be our first basis vector.
                smallest_arc_distance = huge(1.0)
                do current_index=1,largest_index_to_test
                    if (arc_distance(current_index) .lt. smallest_arc_distance) then
                        smallest_arc_distance = arc_distance(current_index)
                        a = current_index
                    endif
                enddo

                if (debug) print *, 'a = ', a

                ! Now let's find the next closest lattice point which is
                ! at least 45 degrees away
                smallest_arc_distance = huge(1.0)
                do current_index=1,largest_index_to_test
                    !distance_to_a = sqrt((x(current_index)-x(a))**2 + (y(current_index)-y(a))**2)
                    if (        arc_distance(current_index) .lt. smallest_arc_distance &
                        .and.   current_index .ne. a &
                        .and. abs(polar_angle(a) - polar_angle(current_index)) .gt. 45.0) then
                        smallest_arc_distance = arc_distance(current_index)
                        b = current_index
                    endif
                enddo

                if (b .eq. 0 .and. .not. self%stack .and. self%axial_symmetry .eq. 1) then
                    call helical_symmetry_print_info(self)
                    call this_program%TerminateWithFatalError('helix_guess_natural_basis_vectors','unable to guess basis vectors')
                endif

                ! Now some heuristics to land back on the 1-start unless it's crazy
                if (b .ne. 0) then
                    if (        a .ne. 1 &
                        .and.   abs(polar_angle(a) - polar_angle(1)) .gt. 45.0 &
                        .and.   arc_distance(1) .lt. arc_distance(b)*2.0) then
                        b = 1
                    endif
                endif

                ! Convention: a = fast, b = slow
                if (a .lt. b) then
                    temp = a
                    a = b
                    b = temp
                endif

                ! Debug info
                if (debug) then
                    write(*,'(a, 2(i0,1x),a)',advance='no') &
                        '**debug(helix_guess_natural_basis_vectors): guessed natural basis vectors: ', a, b, 'for helix: '
                    call helix_print_info(self)
                endif

                ! Temporary
                if (self%stack) then
                    !h_temp = self
                    !call helix_change_num_starts(h_temp,1)
                    self%fast_helix_twist = self%twist
                    self%fast_helix_rise = self%rise
                else
                    h_temp = self
                    call helix_change_num_starts(h_temp,a)
                    self%fast_helix_rise = h_temp%rise
                    self%fast_helix_twist = h_temp%twist
                    self%fast_helix_start = a
                endif
                if (self%axial_symmetry .gt. 1 .or. self%stack) then
                    self%slow_helix_start = 0
                    self%slow_helix_twist = 360.0D0 / self%axial_symmetry
                    self%slow_helix_rise = 0.0D0
                else
                    h_temp = self
                    call helix_change_num_starts(h_temp,b)
                    self%slow_helix_rise = h_temp%rise
                    self%slow_helix_twist = h_temp%twist
                    self%slow_helix_start = b
                endif
            endif ! end of test for positive outer radius
            !print *, '**debug(helix_guess_natural_basis_vectors): fast, slow twist = ', self%fast_helix_twist, self%slow_helix_twist

        end subroutine helix_guess_natural_basis_vectors



        !>  \brief  Work out the helical parameters with fastest rise (slowest twist) within
        !!          the family of equivalent start helices
        !pure
        subroutine helix_update_fastest_helix(self)
            implicit none
            ! Arguments
            type(helical_symmetry), intent(inout)  ::  self
            ! Private variables
            integer ::  start_with_max_twist
            type(helical_symmetry)  ::  h_temp
            ! Start work
            if (self%stack) then
                h_temp = self
                call helix_change_num_starts(h_temp,1)
                self%fast_helix_twist = h_temp%twist
                self%fast_helix_rise = h_temp%rise
            else
                ! Find the start numbers which gives the smallest and largest twist per subunit
                call helix_find_starts_with_min_max_twist(self,self%fast_helix_start,start_with_max_twist)

                ! Convert to that start number
                h_temp = self
                call helix_change_num_starts(h_temp,self%fast_helix_start)

                self%fast_helix_twist = h_temp%twist
                self%fast_helix_rise = h_temp%rise
            endif
        end subroutine helix_update_fastest_helix

        !>  \brief  Work out the helical parameters with the slowest rise within the family of equivalent start helices
        !pure
        subroutine helix_update_slowest_helix(self)
            implicit none
            ! Arguments
            type(helical_symmetry), intent(inout)  ::  self
            ! Private variables
            type(helical_symmetry)  ::  h_temp
            ! Start work
            if (self%axial_symmetry .gt. 1 .or. self%stack) then
                self%slow_helix_start = 0
                self%slow_helix_twist = 360.0D0 / self%axial_symmetry
                self%slow_helix_rise = 0.0D0
            else
                self%slow_helix_start = 1
                h_temp = self
                call helix_change_num_starts(h_temp,1)
                self%slow_helix_twist = h_temp%twist
                self%slow_helix_rise = h_temp%rise
            endif
        end subroutine helix_update_slowest_helix

        !>  \brief  Find the number of starts such that the twist-per-subunit is minimised
        !pure
        subroutine helix_find_starts_with_min_max_twist(self,start_min,start_max)
            use euleranglesmod, only : degrees_m180_to_180
            use runtime_parameters, only : terminate
            implicit none
            ! Arguments
            type(helical_symmetry), intent(in)  ::  self            !<  Helical operator
            integer,                intent(out) ::  start_min       !<  Start with the smallest twist per subunit
            integer,                intent(out) ::  start_max       !<  Start with the largest twist per subunit
            ! Private variables
            type(helical_symmetry)  ::  h_temp_onestart, h_temp
            integer ::  max_num_starts
            integer ::  current_num_starts
            real(kind=8)    ::  min_twist_per_subunit, max_twist_per_subunit
            ! Start work
            h_temp_onestart = self
            !call helix_to_one_start(h_temp_onestart)
            if (h_temp_onestart%num_starts .ne. 1) then
                call terminate('helix_find_starts_with_min_max_twist needs 1-start helix')
            endif
            ! Work out how many starts would take the twist to > 360.
            if (h_temp_onestart%twist .ne. 0.0) then
                max_num_starts = floor(360.d0 / abs(h_temp_onestart%twist) / self%axial_symmetry)
            else
                max_num_starts = 1
            endif
            ! Loop over all possible n starts
            min_twist_per_subunit = 360.0d0
            start_min = 0
            max_twist_per_subunit = 0.0D0
            start_max = 0
            do current_num_starts=1,max_num_starts
                h_temp = h_temp_onestart
                call helix_change_num_starts(h_temp,current_num_starts)
                if (abs(h_temp%twist) .lt. min_twist_per_subunit) then
                    min_twist_per_subunit = abs(h_temp%twist)
                    start_min = current_num_starts
                endif
                if (abs(h_temp%twist) .gt. max_twist_per_subunit) then
                    max_twist_per_subunit = abs(h_temp%twist)
                    start_max = current_num_starts
                endif
            enddo
        end subroutine helix_find_starts_with_min_max_twist

        !>  \brief  Given a helical symmetry operator, modify it so that it describes the N-start equivalent
        !pure
        subroutine helix_change_num_starts(self,n)
            use runtime_parameters, only : terminate
            use euleranglesmod, only : degrees_m180_to_180
            implicit none
            ! Arguments
            type(helical_symmetry), intent(inout)   ::  self
            integer,                intent(in)      ::  n       !<  Number of starts for the helical symmetry after modification
            ! Private variables

            ! Start work
            if (self%num_starts .eq. n) then
                ! Nothing to do
            else
                if (self%num_starts .ne. 1) then
                    !print *, 'debug(change_num_starts): about to one start'
                    !call helix_to_one_start(self)
                    call terminate('cannot convert to 1 start')
                endif
                ! The operator now describes the 1-start helix. Let's go to n
                if (n .gt. 1) then
                    self%twist      =   mod(dble(n)*self%twist,360.d0)
                    self%twist      =   degrees_m180_to_180(self%twist)
                    self%rise       =   dble(n) * self%rise
                    self%num_starts =   n
                else
                    ! Nothing to do
                endif
            endif
        end subroutine helix_change_num_starts

        !>  \brief  Return the 1-start helical operator corresponding to the n-start helical operator given as input
        !pure
        subroutine helix_to_one_start(self)
            use euleranglesmod, only : degrees_m180_to_180
            use runtime_parameters, only : terminate
            implicit none
            ! Arguments
            type(helical_symmetry), intent(inout)   ::  self
            ! Private variables

            ! Start work
            call terminate('helix_to_one_start is buggy. Need to reimplement using 2 basis vectors.')
            self%twist      =   mod((360.D0 + self%twist) / dble(self%num_starts),360.D0)
            self%twist      =   degrees_m180_to_180(self%twist)
            self%rise       =   self%rise / dble(self%num_starts)
            self%num_starts =   1
        end subroutine helix_to_one_start

        !>  \brief  Compute the phi angle (in degrees) given x,y, the coordinates (from the center; in pixels) of a point
        pure function helix_phi_from_xy(x,y) result(phi)
            use rotation_matrices, only : degrees
            use runtime_parameters, only : h_asym_unit_center
            implicit none
            ! Arguments
            real(kind=8),   intent(in)  ::  x,y
            ! Result
            real(kind=8)    ::  phi
            ! Private variables

            ! Start work
            if (x .eq. 0.0 .and. y .eq. 0.0) then
                phi = 0.0
            else
                phi = degrees(  atan2(  y*(-h_asym_unit_center(1))+x*h_asym_unit_center(2), &
                                        x*( h_asym_unit_center(1))+y*h_asym_unit_center(2)))
            endif
        end function helix_phi_from_xy

        !>  \brief  Given polar coordinates (phi in degrees, radius in pixels), output the xy coordinates (in pixels from center)
        pure subroutine helix_xy_from_phi_r(phi,r,x,y)
            use rotation_matrices, only : radians
            use runtime_parameters, only : h_asym_unit_center
            implicit none
            ! Arguments
            real(kind=8),   intent(in)  ::  phi, r
            real(kind=8),   intent(out) ::  x,y
            ! Private variables
            real(kind=8)    ::  phi_rad, cos_phi, sin_phi
            !
            phi_rad = radians(phi)
            cos_phi = cos(phi_rad)
            sin_phi = sin(phi_rad)
            x =  cos_phi * h_asym_unit_center(1) * r + sin_phi * h_asym_unit_center(2) * r
            y = -sin_phi * h_asym_unit_center(1) * r + cos_phi * h_asym_unit_center(2) * r
        end subroutine helix_xy_from_phi_r



        !>  \brief  Compute the pitch of a helix, given the helical operator
        !!
        !!  The pitch of the helix is the distance along the helical axis over which the helix rotates by 360 degrees
        pure function helix_pitch(self)
            implicit none
            ! Arguments
            type(helical_symmetry), intent(in)  ::  self
            ! Result
            real(kind=8)    ::  helix_pitch
            ! Private variables

            ! Start work
            if (self%twist .eq. 0.0d0) then
                helix_pitch = abs(self%rise)
            else
                helix_pitch = abs(self%rise * 360.0D0 / self%twist)
            endif
        end function helix_pitch

        pure elemental function helix_divide_by_real(helix_in,val) result(helix_out)
            implicit none
            !
            type(helical_symmetry), intent(in)  ::  helix_in
            real,                   intent(in)  ::  val
            !
            type(helical_symmetry)  ::  helix_out
            !
            helix_out               =   helix_in
            helix_out%pixel_units   =   helix_in%pixel_units
            helix_out%rise          =   helix_in%rise / val
            helix_out%twist         =   helix_in%twist / val
            helix_out%outer_radius  =   helix_in%outer_radius / val
        end function helix_divide_by_real

        pure elemental function helix_multiply_by_real(helix_in,val) result(helix_out)
            implicit none
            !
            type(helical_symmetry), intent(in)  ::  helix_in
            real,                   intent(in)  ::  val
            !
            type(helical_symmetry)  ::  helix_out
            !
            helix_out               =   helix_in
            helix_out%pixel_units   =   helix_in%pixel_units
            helix_out%rise          =   helix_in%rise * val
            helix_out%twist         =   helix_in%twist * val
            helix_out%outer_radius  =   helix_in%outer_radius * val
        end function helix_multiply_by_real

        function helix_scale_distances(helix_in,scaling_factor) result(helix_out)
            type(helical_symmetry), intent(in)  ::  helix_in
            real,                   intent(in)  ::  scaling_factor
            type(helical_symmetry)              ::  helix_out
            call helical_symmetry_init(helix_out,real(helix_in%rise*scaling_factor),real(helix_in%twist), &
                                                helix_in%outer_radius*scaling_factor, &
                                                helix_in%axial_symmetry,helix_in%perpendicular_symmetry, &
                                                helix_in%num_starts,helix_in%stack)
            helix_out%pixel_units = helix_in%pixel_units
        end function helix_scale_distances

        pure elemental function helix_multiply_by_integer(helix_in,ival) result(helix_out)
            implicit none
            !
            type(helical_symmetry), intent(in)  ::  helix_in
            integer,                intent(in)  ::  ival
            !
            type(helical_symmetry)  ::  helix_out
            !
            helix_out = helix_multiply_by_real(helix_in,real(ival))
        end function helix_multiply_by_integer

        pure elemental function helix_plus_helix(helix1,helix2)
            implicit none
            !
            type(helical_symmetry), intent(in)  ::  helix1, helix2
            !
            type(helical_symmetry)  ::  helix_plus_helix
            !
            !call check_helix_arithmetic_is_possible(helix1,helix2)
            helix_plus_helix%pixel_units    =   helix1%pixel_units
            helix_plus_helix%rise           =   helix1%rise + helix2%rise
            helix_plus_helix%twist          =   helix1%twist + helix2%twist
        end function helix_plus_helix

        pure elemental function helix_minus_helix(helix1,helix2)
            implicit none
            !
            type(helical_symmetry), intent(in)  ::  helix1, helix2
            !
            type(helical_symmetry)  ::  helix_minus_helix
            !
            !call check_helix_arithmetic_is_possible(helix1,helix2)
            helix_minus_helix%pixel_units   =   helix1%pixel_units
            helix_minus_helix%rise          =   helix1%rise - helix2%rise
            helix_minus_helix%twist         =   helix1%twist - helix2%twist
        end function helix_minus_helix

        pure elemental function helix_times_helix(helix1,helix2)
            implicit none
            !
            type(helical_symmetry), intent(in)  ::  helix1, helix2
            !
            type(helical_symmetry)  ::  helix_times_helix
            !
            !call check_helix_arithmetic_is_possible(helix1,helix2)
            helix_times_helix%pixel_units   =   helix1%pixel_units
            helix_times_helix%rise          =   helix1%rise * helix2%rise
            helix_times_helix%twist         =   helix1%twist * helix2%twist
        end function helix_times_helix

        pure elemental function helix_sqrt(helix)
            implicit none
            !
            type(helical_symmetry), intent(in)  ::  helix
            !
            type(helical_symmetry)  ::  helix_sqrt
            !
            helix_sqrt = helix
            helix_sqrt%rise = sqrt(helix%rise)
            helix_sqrt%twist = sqrt(helix%twist)
        end function helix_sqrt


        !>  \brief  Check that two sets of helical parameters can be combined using arithmetics. If not, crash out.
        subroutine check_helix_arithmetic_is_possible(helix1,helix2)
            use runtime_parameters, only : terminate
            implicit none
            !
            type(helical_symmetry), intent(in)  ::  helix1, helix2
            !
            if (      helix1%pixel_units .eqv. helix2%pixel_units &
                .and. helix1%axial_symmetry .eq. helix2%axial_symmetry &
                .and. helix1%perpendicular_symmetry .eq. helix2%perpendicular_symmetry ) then
                ! We don't do anything
            else
                write(*,'(2a)') '**ERROR(check_helix_arithmetic_is_possible): cannot add helical parameters unless ',    &
                                'they are in the same units and have the same axial & perpendicular symmetries'
                call helical_symmetry_print_info(helix1)
                call helical_symmetry_print_info(helix2)
                call terminate('Fatal error in CHECK_HELIX_ARITHMETIC_IS_POSSIBLE')
            endif
        end subroutine check_helix_arithmetic_is_possible


        !>  \brief  Print out helical symmetry parameters to disk
        subroutine helical_symmetry_print_info(self)
            use runtime_parameters, only : word_len
            implicit none
            ! Arguments
            type(helical_symmetry), intent(in)  ::  self
            ! Private variables
            character(len=WORD_LEN)::   rise_unit
            ! Start work
            if (self%pixel_units) then
                rise_unit  = ' pixels'
            else
                rise_unit  = ' Angstr.'
            endif
            if (self%stack) then
                write(*,'(a,i0,a,F0.6,1x,a,a,F0.6,a)', advance='NO')  'Stack (', &
                                                                    self%axial_symmetry, '-fold symmetry) with rise of ', &
                                                                    self%rise, trim(adjustl(rise_unit)), &
                                                                    ' and twist of ', self%twist, ' degrees.'
            else
                write(*,'(i0,a,i0,a,F0.6,1x,a,a,F0.6,a)', advance='NO')  self%num_starts, '-start helix (', &
                                                                    self%axial_symmetry, '-fold symmetry) with rise of ', &
                                                                    self%rise, trim(adjustl(rise_unit)), &
                                                                    ' and twist of ', self%twist, ' degrees.'
            endif
            if (self%perpendicular_symmetry .eq. 2) then
                write(*,'(a)') ' Non-polar.'
            else
                write(*,'(a)') ' Polar.'
            endif
        end subroutine helical_symmetry_print_info

        !>  \brief  Print extra info about the helical operator
        subroutine helical_symmetry_print_extra_info(self)
            implicit none
            ! Arguments
            type(helical_symmetry), intent(in)  ::  self
            ! Private variables
            ! Start work
            write(*,'(a,f0.2)',advance='no') 'Outer radius = ', self%outer_radius
            if (self%pixel_units) then
                write(*,'(a)') ' pixels'
            else
                write(*,'(a)') ' A'
            endif
            write(*,'(a,2(i0,1x))') 'Starts fastest, slowest = ', self%fast_helix_start, self%slow_helix_start
            write(*,'(a,2(F0.3,1x))') 'rise, twist of fastest helix: ', self%fast_helix_rise, self%fast_helix_twist
            write(*,'(a,2(F0.3,1x))') 'rise, twist of slowest helix: ', self%slow_helix_rise, self%slow_helix_twist
            write(*,'(a,/,2(2(f0.2,1x),/))') 'conv matrix:', self%conv_mat
        end subroutine helical_symmetry_print_extra_info

        !>  \brief  Reset all values
        pure elemental subroutine helical_symmetry_reset(self)
            implicit none
            ! Arguments
            type(helical_symmetry),     intent(out)     ::  self
            ! Start work
            self%pixel_units = .false.
            self%rise = 0.0D0
            self%twist = 0.0D0
            self%axial_symmetry = 1
            self%perpendicular_symmetry = 1
            self%num_starts = 1
            !self%asym_unit_height = 0.0D0
            !self%asym_unit_width = 0.0D0
            self%stack = .false.
        end subroutine helical_symmetry_reset

        !>  \brief  Initialise a helical_symmetry object, given relevant parameters
        subroutine helical_symmetry_init(self,rise,twist,outer_radius, &
                                            axial_sym,perpendicular_sym,&
                                            num_starts,stack, &
                                            do_not_setup_unit_cell)
            use rotation_matrices, only : radians
            use runtime_parameters, only : terminate
            use euleranglesmod, only : degrees_m180_to_180
            implicit none
            ! Arguments
            type(helical_symmetry),     intent(inout)   ::  self                    !<  Helical symmetry parameters to initialise
            real,                       intent(in)      ::  rise                    !<  The rise (in pixels), i.e. the translation along the helical axis to get from one subunit to the next
            real,                       intent(in)      ::  twist                   !<  The twist (in degrees), i.e. the rotation around the helical axis to get from one subunit to the next
            real,                       intent(in)      ::  outer_radius            !<  Outer radius (pixels)
            integer,    optional,       intent(in)      ::  axial_sym               !<  Axial symmetry. If axial_symmetry = n, the helix is N-start.
            integer,    optional,       intent(in)      ::  perpendicular_sym       !<  Symmetry with axis perpendicular to helix axis. This can either be 1 or 2. 2 means the helix is not polar, like dna for example.
            integer,    optional,       intent(in)      ::  num_starts              !<  The helical parameters describes n NUM_START-helix
            logical,    optional,       intent(in)      ::  stack                   !<  This a stack rather than a true helix
            logical,    optional,       intent(in)      ::  do_not_setup_unit_cell  !<  Skip the setting up of the helical unit cell. Only set this to true if you won't use the helix for anything like symmetrisation etc.
            ! Private variables
            integer ::  aaxial_sym
            integer ::  pperpendicular_sym
            logical ::  setup_unit_cell
            logical,    parameter   ::  use_new_helical_unit_cell = .true.
            logical,    parameter   ::  debug = .false.
            ! Start work
            ! Rise must be positive
            if (rise .lt. 0.) then
                write(*,'(a,F0.6)') '**ERROR(helical_symmetry_init): rise is not positive: ', rise
                call terminate('Fatal error in HELICAL_SYMMETRY_INIT')
            endif
            ! Outer radius must be positive
            if (outer_radius .lt. 0.0) then
                write(*,'(a,F0.6)') '**ERROR(helical_symmetry_init): outer radius is not positive: ', outer_radius
                call helical_symmetry_print_info(self)
                call terminate('Fatal error in HELICAL_SYMMETRY_INIT')
            endif
            ! Deal with optional parameters
            setup_unit_cell = .true.
            if (present(do_not_setup_unit_cell)) then
                setup_unit_cell = .not. do_not_setup_unit_cell
            endif
            if (present(axial_sym)) then
                aaxial_sym = axial_sym
            else
                aaxial_sym = 1  ! if the user didn't specify a helical symmetrym we assume no symmetry, i.e. a 1-start helix
            endif
            if (present(perpendicular_sym)) then
                pperpendicular_sym = perpendicular_sym
            else
                pperpendicular_sym = 1  ! if the user didn't specify a symmetry here, we assume no symmetry, i.e. a polar filament
            endif
            ! Perpendicular symmetry can only be 1 or 2
            if (pperpendicular_sym .lt. 1 .or. pperpendicular_sym .gt. 2) then
                write(*,'(2a,i0)') '**ERROR(helical_symmetry_init): a helix cannot have this symmetry perpendicular', &
                                    ' to the helical axis : ', pperpendicular_sym
                call terminate('Fatal error in HELICAL_SYMMETRY_INIT')
            endif
            ! Number of starts
            if (present(num_starts)) then
                if (num_starts .ge. 1) then
                    self%num_starts = num_starts
                else
                    write(*,'(a)') '**ERROR(helical_symmetry_init): nonsensical number of starts given as argument: ', num_starts
                    call terminate('Fatal error in HELICAL_SYMMETRY_INIT')
                endif
            else
                self%num_starts = 1
            endif
            ! Set up the values
            self%rise           =   rise
            self%pixel_units    =   .true.
            self%twist          =   mod(twist,360.e0)
            self%twist          =   degrees_m180_to_180(self%twist)
            self%outer_radius   =   outer_radius
            self%axial_symmetry =   aaxial_sym
            self%perpendicular_symmetry = pperpendicular_sym
            self%stack = .false.
            if (present(stack)) self%stack = stack
            if (self%stack .and. self%num_starts .gt. 1) then
                write(*,'(a)') '**ERROR(helical_symmetry_init): stacks with >1 starts are not supported - do they exist?'
                call terminate('Fatal error in HELICAL_SYMMETRY_INIT')
            endif



            !
            if (setup_unit_cell) then
                if (use_new_helical_unit_cell) then
                    call helix_guess_natural_basis_vectors(self,self%fast_helix_start,self%slow_helix_start)
                else
                    call helix_update_fastest_helix(self)
                    call helix_update_slowest_helix(self)
                endif
                call helix_update_conversion_matrix(self)
            endif

            if (debug) then
                write(*,'(a)') '**debug(helical_symmetry_init): before width height'
                call helix_print_info(self)
                call helix_print_extra_info(self)
            endif


            ! Sanity check
            if (debug) then
            if (.not. (self%stack .or. self%axial_symmetry .gt. 1) .and. setup_unit_cell) then
                if (self%fast_helix_start .eq. 0 .or. self%slow_helix_start .eq. 0) then
                    write(*,'(a)') '**error(helical_symmetry_init): bad slow and/or fast helical starts'
                    call helix_print_info(self)
                    call helix_print_extra_info(self)
                    call terminate('Fatal error in helical_symmetry_init')
                endif
            endif
            endif


            ! Debug
            if (debug) then
                write(*,'(a)') '**debug(helical_symmetry_init): on exit'
                call helix_print_info(self)
                call helix_print_extra_info(self)
                write(*,'(a)') '**debug(helical_symmetry_init): done'
            endif

        end subroutine helical_symmetry_init

        !>  \brief  Given a pixel size, convert the helical parameters to/from pixel/Angstrom units
        pure subroutine helical_symmetry_unit_convert(self,pixel_size)
            implicit none
            ! Arguments
            type(helical_symmetry),     intent(inout)   ::  self
            real,                       intent(in)      ::  pixel_size  !<  Pixel size in Angstroms
            ! Private variables
            ! Start work
            if (self%pixel_units) then
                self%rise = self%rise * pixel_size
            else
                self%rise = self%rise / pixel_size
            endif
            self%pixel_units = .not. self%pixel_units
        end subroutine helical_symmetry_unit_convert


        !>  \brief  Given lattice parameters, calculate & return the corresponding helical symmetry
        !!  \todo   Write this function! i have already worked out how to do this when the lattice is assumed to be square (angle = 90 deg),
        !!          but i haven't had time to generalise to any lattice. (see /gusr/alr99/linux/notes/notes-20090510-helical-parameters.ods)
!       function helical_symmetry_from_lattice result(h_sym)
!           implicit none
!           ! Arguments
!
!           ! Result
!           type(helical_symmetry)  ::  h_sym
!           ! Private variables
!
!           ! Start work
!
!       end function helical_symmetry_from_lattice



end module Helices
