!>  \brief Module to perform arithmetic operations on images
!!
!!
module images_arithm
    use images
    use runtime_parameters, only : terminate

    !interface operator(+)
    !   module procedure add_images_func
    !end interface

    !interface operator(*)
    !   module procedure multiply_images_func
    !end interface

    !interface operator(/)
    !   module procedure divide_images_func
    !end interface


    contains

    !>  \brief  Compute the average of a stack of images
    subroutine images_average(self,ave)
        use Images
        implicit none
        ! Arguments
        type(image),    allocatable,    intent(in)      ::  self(:)
        type(image),                    intent(inout)   ::  ave
        ! Private variables
        integer ::  i
        ! Start work
        if (.not. allocated(self)) then
            write(*,'(a)') '**ERROR(images_average): input array of images is not allocated'
            call terminate('Fatal error in IMAGES_AVERAGE')
        endif

        ! Allocate memory for the average image
        if (ave%IsAllocated()) call ave%Deallocate()
        call ave%allocate(mould=self(1))

        ! Loop over the input images
        do i=1,size(self)
            if ( .not. ave%HasSameDimensionsAs(self(i)) ) then
                write(*,'(a,i0,a,2(3(i0,1x),a))') '**ERROR(images_average): dimensions of input image ', i, ' (', &
                                                  self(i)%GetLogicalDimensions(), &
                                                  ') are not compatible with the dimensions of the average image (',  &
                                                  ave%GetLogicalDimensions(), ')'
                call terminate('Fatal error in IMAGES_AVERAGE')
            endif
            ave%real_values = ave%real_values + self(i)%real_values
        enddo
        ave%real_values = ave%real_values / real(size(self))
    end subroutine images_average

    !>  \brief  Float an image average to 0.0, normalise its standard deviation to 1.0 and mask the outside to zero
    subroutine image_float_norm_mask(self,object_radius,apply_mask,normalise_solvent)
        use images
        use runtime_parameters, only : safe_for_sqrt
        implicit none
        ! Arguments
        type(image),            intent(inout)   ::  self
        real,       optional,   intent(in)      ::  object_radius           !<  In pixels. Maximum radius of the object
        logical,    optional,   intent(in)      ::  apply_mask              !<  Zero pixels beyond the object radius. If object radius  is very large or 0.0, there will be no masking.
        logical,    optional,   intent(in)      ::  normalise_solvent       !<  The densities will be modified such that the solvent is normalised to (0,1). If not, densities within the object radius are normalised.
        ! Private variables
        integer         ::  i,j,k, origin(3), x,y,z, vox_count
        real            ::  rad_tmp
        real(kind=8)    ::  sum_1, sum_2
        real            ::  rad_mask, std, ave
        logical         ::  nnormalise_solvent, aapply_mask
        ! Start work

        if (.not. self%IsInRealSpace()) then
            write(*,'(a)') '**ERROR(image_float_norm_mask): input image should be real'
            call self%PrintInfo()
            call terminate('Fatal error in IMAGE_FLOAT_NORM_MASK')
        endif

        ! The radius of the mask
        rad_mask = real(self%GetLogicalDimension(1)) * 0.5 * 0.9
        if (present(object_radius)) rad_mask = object_radius
        if (rad_mask .le. 0.001) then
            rad_mask = 1.0 + maxval(self%GetLogicalDimensions())
        endif
        rad_mask = rad_mask**2

        ! Normalise solvent?
        nnormalise_solvent = .false.
        if (present(normalise_solvent)) nnormalise_solvent = normalise_solvent
        aapply_mask = .false.
        if (present(apply_mask)) aapply_mask = apply_mask

        sum_1 = 0.0D0
        sum_2 = 0.0D0
        vox_count = 0
        if (self%IsInRealSpace() .and. self%ObjectIsCenteredInBox()) then
            origin = self%GetPhysicalAddressOfBoxCenter()
        else
            origin = [1,1,1]
        endif


        ! Loop over voxels to get stats
        do k=1,self%GetLogicalDimension(3)
            z = (k-origin(3))**2
            do j=1,self%GetLogicalDimension(2)
                y = (j-origin(2))**2
                do i=1,self%GetLogicalDimension(1)
                    x = (i-origin(1))**2
                    rad_tmp = real(x+y+z)
                    if (nnormalise_solvent) then
                        if (rad_tmp .ge. rad_mask) then
                            sum_1 = sum_1 + self%real_values(i,j,k)
                            sum_2 = sum_2 + self%real_values(i,j,k)**2
                            vox_count = vox_count + 1
                        endif
                    else
                        if (rad_tmp .lt. rad_mask) then
                            sum_1 = sum_1 + self%real_values(i,j,k)
                            sum_2 = sum_2 + self%real_values(i,j,k)**2
                            vox_count = vox_count + 1
                        endif
                    endif
                enddo
            enddo
        enddo

        std = (sum_2 - (sum_1**2/vox_count))/vox_count
        call safe_for_sqrt(std,negative_to_zero=.true.)
        !if (std .EQ. 0.0) call self%WriteToDisk('dbg_std_zero.mrc')
        std = sqrt(std)
        ave = sum_1 / vox_count

        ! Loop over voxels to modify values
        do k=1,self%GetLogicalDimension(3)
            z = (k-origin(3))**2
            do j=1,self%GetLogicalDimension(2)
                y = (j-origin(2))**2
                do i=1,self%GetLogicalDimension(1)
                    x = (i-origin(1))**2
                    rad_tmp = real(x+y+z)
                    if (rad_tmp .lt. rad_mask .or. .not. aapply_mask) then
                        if (std .ne. 0.0) self%real_values(i,j,k) = (self%real_values(i,j,k)-ave)/std
                    else
                        self%real_values(i,j,k) = 0.0
                    endif
                enddo
            enddo
        enddo


    end subroutine image_float_norm_mask

    !>  \brief  Compute the polar moment of inertia of a structure based on the input 2d image, which is assumed to be a binarised version of a slice through the structure.
    !!
    !!  The polar moment of inertia is a quantity which relates the torque applied to an object and the object's resistance to that torque. See http://en.wikipedia.org/wiki/Polar_moment_of_inertia
    !!  It is calculated here following Sachse, Grigorieff & Fandrich (2010): Nanoscale Flexibility Parameters of Alzheimer Amyloid Fibrils Determined by Electron Cryo-Microscopy, doi: 10.1002/anie.200904781
    !!
    !!  \todo   Decide whether it's a good
    real function polar_moment_of_inertia(img,pixel_size)
        use images
        implicit none
        ! Arguments
        type(image),    intent(in)  ::  img         !<  Input image. Must be 2d (i.e. dim(3) == 1) and binarised (i.e. only 1.0's and 0.0's in rvalue)
        real,           intent(in)  ::  pixel_size  !<  Pixel size (in Angstroms)
        ! Private variables
        integer ::  i,j
        real(kind=8)    ::  sum
        real(kind=8)    ::  pixel_size_2
        real(kind=8)    ::  radius
        ! Start work

        ! Check that the image is 2d
        if (.not. img%IsAVolume()) then
            write(*,'(a,i4)') '**ERROR(polar_moment_of_inertia): input image should be 2d, but dim(3) = ', &
                                img%GetLogicalDimension(3)
            call terminate('Fatal error in POLAR_MOMENT_OF_INERTIA')
        endif

        ! Check that the image contains only 0s and 1s
        if (.not. img%IsBinarised()) then
            write(*,'(a)') '**ERROR(polar_moment_of_inertia): input image should be binarised'
            call terminate('Fatal error in POLAR_MOMENT_OF_INERTIA')
        endif

        ! Initialise
        polar_moment_of_inertia = 0.0
        sum                     = 0.0D0
        pixel_size_2            = dble(pixel_size)**2

        ! Loop over the image
        do j=1,img%GetLogicalDimension(2)
            do i=1,img%GetLogicalDimension(1)
                if (img%real_values(i,j,1) .gt. 0.0) then
                    radius =    (i-img%GetPhysicalIndexOfBoxCenter(1))**2 &
                            +   (j-img%GetPhysicalIndexOfBoxCenter(2))**2
                    sum = sum + radius*pixel_size_2
                endif
            enddo
        enddo

        ! Result is ready
        polar_moment_of_inertia = real(sum)
    end function polar_moment_of_inertia

    !>  \brief  Compute the center of mass of an image
    !!
    !!  i lifted the idea for this from EMAN2.06's calc_max_location_wrap_intp routine
    subroutine image_center_of_mass(self,com_coo,lbounds,ubounds)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),                intent(in)  ::  self
        real,                       intent(out) ::  com_coo(3)  !<  Coordinates of the center of mass
        integer,        optional,   intent(in)  ::  lbounds(3)  !<  Image address lower bounds for the center of mass computation
        integer,        optional,   intent(in)  ::  ubounds(3)  !<  Image address upper bounds
        ! Private variables
        integer ::  llbounds(3), uubounds(3)
        integer ::  i,j,k
        real(kind=8)    ::  coo(3), s
        real(kind=8)    ::  val
        ! Start work

        ! Meant for real images only
        if (.not. self%IsInRealSpace()) then
            call terminate('Only implemented for real images')
        endif

        ! Bounds for the computation
        llbounds = (/1,1,1/)
        uubounds = self%GetLogicalDimensions()
        if (present(lbounds)) llbounds = lbounds
        if (present(ubounds)) uubounds = ubounds

        llbounds = max(llbounds,1)
        uubounds = min(uubounds,self%GetLogicalDimensions())

        coo = 0.0D0
        s   = 0.0D0
        do k=llbounds(3),uubounds(3)
            do j=llbounds(2),uubounds(2)
                do i=llbounds(1),uubounds(1)
                    val = dble(self%real_values(i,j,k))
                    coo(1) = coo(1) + dble(i)*val
                    coo(2) = coo(2) + dble(j)*val
                    coo(3) = coo(3) + dble(k)*val
                    s = s + val
                enddo
            enddo
        enddo

        com_coo = coo / s

    end subroutine image_center_of_mass

    !>  \brief  Return the moment of inertia along the requested dimension. See Sachse,Grigorieff&Fandrich (2010) for details.
    real function moment_of_inertia(img,dim,pixel_size)
        use images
        implicit none
        ! Arguments
        type(image),    intent(in)  ::  img             !<  Input image. Must be binarised (i.e. only 0.0's and 1.0's), real and 2D.
        integer,        intent(in)  ::  dim             !<  The dimension along which to calculate the moment of inertia. 1 or 2
        real,           intent(in)  ::  pixel_size      !<  The pixel size (Angstroms)
        ! Private variables.
        integer         ::  i,j
        real(kind=8)    ::  sum
        real(kind=8)    ::  pixel_size_2
        integer         ::  orig(3)
        ! Start work

        ! Check that the image is 2d
        if (img%GetLogicalDimension(3) .ne. 1) then
            write(*,'(a,i4)') '**ERROR(moment_of_inertia): input image should be 2d, but dim(3) = ', img%GetLogicalDimension(3)
            call terminate('Fatal error in MOMENT_OF_INERTIA')
        endif

        ! Check that the image contains only 0s and 1s
        if (.not. img%isbinarised()) then
            write(*,'(a)') '**ERROR(moment_of_inertia): input image should be binarised'
            call terminate('Fatal error in MOMENT_OF_INERTIA')
        endif

        ! Initialise
        moment_of_inertia       = 0.0
        sum                     = 0.0D0
        pixel_size_2            = dble(pixel_size)**2
        if (img%IsInRealSpace() .and. img%ObjectIsCenteredInBox()) then
            orig = img%GetPhysicalAddressOfBoxCenter()
        else
            orig = [1,1,1]
        endif

        ! Loop over the image
        do j=1,img%GetLogicalDimension(2)
            do i=1,img%GetLogicalDimension(1)
                if (img%real_values(i,j,1) .gt. 0.0) then
                    if (dim .eq. 1) then
                        sum = sum + dble((i-orig(1))**2)*pixel_size_2
                    else
                        sum = sum + dble((j-orig(2))**2)*pixel_size_2
                    endif
                endif
            enddo
        enddo

        ! Result is ready
        moment_of_inertia = real(sum)
    end function moment_of_inertia

    !>  \brief  Swap the x and y axes. This results in a mirror about the X=Y plane.
    subroutine swap_xy(self)
        implicit none
        ! Arguments
        type(image),    intent(inout)   ::  self    !<  Image object
        ! Private variables
        type(image) ::  output_image
        integer     ::  i,j,k
        ! Start work

        ! Check input image is allocated
        if (.not. self%isallocated()) then
            write(*,'(a)') '**ERROR(swap_xy): input image not allocated'
            call terminate('Fatal error in SWAP_XY')
        endif

        ! Only coding for real images, to save time
        if (.not. self%IsInRealSpace()) then
            write(*,'(a)') '**ERROR(swap_xy): this feature only implemented for real images'
            call terminate('Fatal error in SWAP_XY')
        endif

        !
        ! Set up the output image
        !

        !
        ! Allocate the output image
        !
        call output_image%allocate(mould=self)

        !
        ! Get values for output image
        !
        do k=1,self%GetLogicalDimension(3)
            do j=1,self%GetLogicalDimension(2)
                do i=1,self%GetLogicalDimension(1)
                    output_image%real_values(j,i,k)  =   self%real_values(i,j,k)
                enddo
            enddo
        enddo

        ! Reallocate input image
        call self%allocate(mould=output_image)

        ! Copy results overs
        self%real_values = output_image%real_values

        ! Deallocate temporary image object
        call output_image%Deallocate()
    end subroutine swap_xy

    !>  \brief  Zero-float an image (ie. remove average pixel value from all values in real space, or zero central pixel in reciprocal space)
    subroutine zero_float(img)
        implicit none
        ! Arguments
        type(image),    intent(inout)   ::  img
        ! Private variables
        real(kind=8)    ::  average
        integer ::  i,j,k
        ! Start work
        if (img%IsInRealSpace()) then
            ! Evaluate the average
            average = 0.0D0
            do k=1,img%GetLogicalDimension(3)
                do j=1,img%GetLogicalDimension(2)
                    do i=1,img%GetLogicalDimension(1)
                        average = average + dble(img%real_values(i,j,k))
                    enddo
                enddo
            enddo
            average = average / dble(img%GetLogicalDimension(3)*img%GetLogicalDimension(2)*img%GetLogicalDimension(1))
            ! Substract average from all pixels
            img%real_values = img%real_values - real(average)
        else if (.not. img%IsInRealSpace()) then
            img%complex_values(1,1,1) = (0.0,0.0)
        endif
    end subroutine

    !>  \brief  Binarise an image. Pixels with values above supplied threshold are set to 1.0, others are set to 0.0
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine binarise(img,threshold)
        implicit none
        ! Arguments
        type(image),    intent(inout)   ::  img         !<  Image to be binarised (overwritten)
        real,           intent(in)      ::  threshold   !<  Threshold. Pixels with values .GE. threshold are set to 1.0. Others are set to 0.0
        ! Private variables
        integer ::  i,j,k
        ! Start work

#ifndef skip_runtime_checks
        ! Only meant for real images
        if (.not. img%IsInRealSpace()) then
            write(*,'(a)') '**ERROR(binarise): image type not supported: '
            call terminate('Fatal error in BINARISE')
        endif
#endif

        ! Loop over voxel values
        do k=1,img%GetLogicalDimension(3)
            do j=1,img%GetLogicalDimension(2)
                do i=1,img%GetLogicalDimension(1)
                    if (img%real_values(i,j,k)-threshold .ge. tiny(threshold)) then
                        img%real_values(i,j,k) = 1.0
                    else
                        img%real_values(i,j,k) = 0.0
                    endif
                enddo
            enddo
        enddo

    end subroutine binarise



!    !>  \brief  Convolve two real images by mutliplying their ft
!    subroutine fast_convolution(img_in1,img_in2,img_out)
!        use images
!        implicit none
!        ! Arguments
!        type(image),    intent(inout)   ::  img_in1
!        type(image),    intent(inout)   ::  img_in2
!        type(image),    intent(inout)   ::  img_out
!        ! Private variables
!
!        ! Start work
!        ! Check both images are real
!        if (.not. (img_in1%IsInSameSpaceAs(img_in2) .and. img_in1%IsInRealSpace())) then
!            write(*,'(a)') '**ERROR(fast_convolution): invalid image types'
!            call terminate('Fatal error in FAST_CONVOLUTION')
!        endif
!        ! Check both images have same dimensions
!        if (.not. img_in1%HasSameDimsAs(img_in2)) then
!            write(*,'(a,2(3(i3,1x),a))') '**ERROR(fast_convolution): input images do not have same dimensions: ', &
!                                        img_in1%GetLogicalDimensions(), ' & ', img_in2%GetLogicalDimensions()
!            call terminate('Fatal error in FAST_CONVOLUTION')
!        endif
!        ! Allocate output image if necessary
!        if (.not. (img_in1%HasSameDimsAs(img_out) .and. img_out%IsAllocated())) then
!            call img_out%allocate(mould=img_in1)
!        endif
!        ! Prepare the output image, which will initially be complex
!        img_out%is_in_real_space = .false.
!        img_out%object_is_centered_in_box = .false.
!        ! Fourier transform both inputs
!        call img_in1%ForwardFFT(scale=.true.)
!        call img_in2%ForwardFFT(scale=.true.)
!        ! Multiply
!        img_out%complex_values = img_in1%complex_values * img_in2%complex_values / &
!                                real(img_out%logical_dimensions(1) * img_out%logical_dimensions(2) * img_out%logical_dimensions(3))
!        ! Center
!        call img_out%ShiftRealSpaceOrigin()
!        ! Reverse ft the output
!        call img_out%BackwardFFT(scale=.true.)
!    end subroutine fast_convolution



    !>  \brief  Impose of soft cylindrical mask on image
    !!  \todo   Make this subroutine more versatile, so it can deal with fractional radius better, and do softness better (see eg. mod_filters)
    subroutine cyl_mask(img,rad,soft,rad_in,mask_with_average_inside,zero_float_inside)
        use interpolation    ! We need this for the Cosine / Hann fall off function
        use images
        implicit none
        type(image),    intent(inout)   ::  img     !<  Image to be masked
        real,           intent(in)      ::  rad     !<  Radius of cylinder (pixels)
        integer,        intent(in)      ::  soft    !<  Width of soft fall-off
        real, optional, intent(in)      ::  rad_in  !<  Inner radius (pixels). If not zero, a hollow cylinder is calculated
        logical, optional, intent(in)   ::  mask_with_average_inside    !<  Mask value is average of densities within the cylinder
        logical, optional, intent(in)   ::  zero_float_inside
        ! Private variables
        integer ::  i,j,k
        real    ::  rad_tmp, rad_start, rad_in_start
        real    ::  x,y
        integer ::  origin(3)
		real	::	rrad_in
		real(kind=8)    ::  mask_value
		logical ::  mmask_with_average_inside
		logical  :: zzero_float_inside
		real(kind=8) :: voxel_count


		rrad_in = 0.0
		if (present(rad_in)) rrad_in = rad_in

        rad_start = rad-soft/2.0
        rad_in_start = rrad_in-soft/2.0

        if (img%IsInRealSpace() .and. img%ObjectIsCenteredInBox()) then
            origin = img%GetPhysicalAddressOfBoxCenter()
        else
            origin = [1,1,1]
        endif


        mmask_with_average_inside = .false.
        if (present(mask_with_average_inside)) mmask_with_average_inside = mask_with_average_inside

        zzero_float_inside = .false.
        if (present(zero_float_inside)) zzero_float_inside = zero_float_inside

        ! Determine mask value
        if (mmask_with_average_inside .or. zzero_float_inside) then
            mask_value = 0.0D0
            voxel_count = 0.0D0
            do k=1,img%GetLogicalDimension(3)
            do j=1,img%GetLogicalDimension(2)
                y = real(j-origin(2))**2
                do i=1,img%GetLogicalDimension(1)
                    x = real(i-origin(1))**2
                    rad_tmp = sqrt(x+y)
                    if (rad_tmp .ge. rad_start) then
                        mask_value = mask_value + img%real_values(i,j,k) * hann(rad_tmp-rad_start,soft)
                        voxel_count = voxel_count + hann(rad_tmp-rad_start,soft)
                    else if (rad_tmp .le. rrad_in) then
                        if (rad_tmp .ge. rad_in_start) then
                            mask_value = mask_value + img%real_values(i,j,k) * (hann(rad_tmp-rad_in_start,soft)-1.0)*(-1.0)
                            voxel_count = voxel_count + (hann(rad_tmp-rad_in_start,soft)-1.0)*(-1.0)
                        else
                            !
                        endif
                    endif
                enddo
            enddo
            enddo
            mask_value = mask_value / voxel_count
        else
            mask_value = 0.0e0
        endif

        if (zzero_float_inside) then
            img%real_values = img%real_values - mask_value
            mask_value = 0.0D0
        endif


        do j=1,img%GetLogicalDimension(2)
            y = real(j-origin(2))**2
            do i=1,img%GetLogicalDimension(1)
                x = real(i-origin(1))**2
                rad_tmp = sqrt(x+y)
                if (rad_tmp .ge. rad_start) then
                    img%real_values(i,j,1:img%GetLogicalDimension(3)) = img%real_values(i,j,1:img%GetLogicalDimension(3)) * &
                                                                            hann(rad_tmp-rad_start,soft) + &
                                                                            (hann(rad_tmp-rad_start,soft)-1.0)*(-1.0)*mask_value
                else if (rad_tmp .le. rrad_in) then
                    if (rad_tmp .ge. rad_in_start) then
                        img%real_values(i,j,1:img%GetLogicalDimension(3)) = img%real_values(i,j,1:img%GetLogicalDimension(3)) * &
                                                                            (hann(rad_tmp-rad_in_start,soft)-1.0)*(-1.0) + &
                                                                            hann(rad_tmp-rad_in_start,soft) * mask_value
                    else
                        img%real_values(i,j,1:img%GetLogicalDimension(3)) = mask_value
                    endif
                endif
            enddo
        enddo

    end subroutine cyl_mask


    function multiply_images_func(img1,img2) result(imgout)
        implicit none
        type(image), intent(in)     ::  img1, img2
        type(image) ::  imgout
        call multiply_images(img1,img2,imgout)
    end function multiply_images_func

    function divide_images_func(img1,img2) result(imgout)
        implicit none
        type(image), intent(in)     ::  img1, img2
        type(image) ::  imgout
        call divide_images(img1,img2,imgout)
    end function divide_images_func

    function add_images_func(img1,img2) result(imgout)
        implicit none
        type(image), intent(in)     ::  img1, img2
        type(image) ::  imgout
        call add_images(img1,img2,imgout)
    end function add_images_func



    !>  \brief Multiply img1 and img2, using real or complex arithmetics, depending on the type of the images.
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine multiply_images(img1,img2,imgout)
        implicit none
        type(image), intent(in)     ::  img1,img2
        type(image), intent(inout)  ::  imgout

#ifndef skip_runtime_checks
        ! Check both images have the same type
        if ( .not. img1%IsInSameSpaceAs(img2)) then
            write(*,*) '**ERROR (multiply_images): cannot multiply images of different types'
            call terminate('Fatal error in MULTIPLY_IMAGES')
        endif

        ! Check both images have the same dimensions
        if ( .not. img1%HasSameDimensionsAs(img2)) then
            write(*,'(2(A,3(I4,1x)))') '**ERROR (multiply_images): cannot multiply images of different dimensions: ', &
                                        img1%GetLogicalDimensions(), ';',img2%GetLogicalDimensions()
            call terminate('Fatal error in MULTIPLY_IMAGES')
        endif

        ! Check both input images have memory allocated
        if ( (.not. img1%IsAllocated()) .or. (.not. img2%IsAllocated())) then
            write(*,*) '**ERROR (multiply_images): one or both input images do not have allocated memory'
            call terminate('Fatal error in MULTIPLY_IMAGES')
        endif
#endif

        ! Do the multiplication
        if ( .not. img1%IsInRealSpace()) then
            imgout%complex_values = img1%complex_values * img2%complex_values
        else
            imgout%real_values = img1%real_values * img2%real_values
        endif
    end subroutine multiply_images

    !>  \brief  Apply a mask to an image, with option to level the mask so that it corresponds to the average of the image within the mask
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine image_apply_mask(self,mask,mask_with_average,mask_with_average_outside)
        use imagefiles
        implicit none
        ! Arguments
        type(image),                intent(inout)   ::  self                !<  Image to be masked
        type(image),                intent(inout)   ::  mask                !<  Mask image. Expected to go from 0. to 1.
        logical,        optional,   intent(in)      ::  mask_with_average   !<  Whether the image should be masked with a value corresponding to its average within the mask
        logical,        optional,   intent(in)      ::  mask_with_average_outside   !<  Whether the image should be masked with a value corresponding to its average outside the mask
        ! Private variables
        logical ::  mmask_with_average, mmask_with_average_outside
        integer ::  i,j,k
        real(kind=8)    ::  mask_value, num_elems
        ! Start work

#ifndef skip_runtime_checks
        ! Check the mask is kosher
        if (.not. image_suitable_as_mask(mask)) then
            write(*,'(a)') '**ERROR(image_apply_mask): mask image is not a mask'
            call mask%WriteToDisk('dbg_mask_pb.mrc')
            call mask%WriteToDisk('dbg_mask_pb_self.mrc')
            call terminate('Fatal error in IMAGE_APPLY_MASK')
        endif
#endif

        ! Will we mask with average? (rather than 0.0)
        mmask_with_average = .true.
        mmask_with_average_outside = .false.
        if (present(mask_with_average)) then
            mmask_with_average = mask_with_average
#ifndef skip_runtime_checks
            if (present(mask_with_average_outside)) then
                if (mask_with_average .and. mask_with_average_outside) then
                    call terminate('Fatal error in IMAGE_APPLY_MASK. Arguments incompatible.')
                endif
            endif
#endif
        endif
        if (present(mask_with_average_outside)) mmask_with_average_outside = mask_with_average_outside


        if (mmask_with_average) then
            ! Work out the average within the mask
            mask_value = 0.0D0
            num_elems  = 0.0D0
            do k=1,self%GetLogicalDimension(3)
                do j=1,self%GetLogicalDimension(2)
                    do i=1,self%GetLogicalDimension(1)
                        mask_value = mask_value + mask%real_values(i,j,k) * self%real_values(i,j,k)
                        num_elems  = num_elems  + mask%real_values(i,j,k)
                    enddo
                enddo
            enddo
            mask_value = mask_value / num_elems
            ! Apply the mask
            self%real_values = self%real_values * mask%real_values + (1.-mask%real_values)*mask_value
        else if (mmask_with_average_outside) then
            ! Work out the average within the mask
            mask_value = 0.0D0
            num_elems  = 0.0D0
            do k=1,self%GetLogicalDimension(3)
                do j=1,self%GetLogicalDimension(2)
                    do i=1,self%GetLogicalDimension(1)
                        mask_value = mask_value + (1.0-mask%real_values(i,j,k)) * self%real_values(i,j,k)
                        num_elems  = num_elems  + (1.0-mask%real_values(i,j,k))
                    enddo
                enddo
            enddo
            mask_value = mask_value / num_elems
            ! Apply the mask
            self%real_values = self%real_values * mask%real_values + (1.-mask%real_values)*mask_value
        else
            call multiply_images(self,mask,self)
        endif

    end subroutine image_apply_mask

    !>  \brief  Check whether the image can be used as a mask
    logical function image_suitable_as_mask(self)
        ! Arguments
        type(image),    intent(in)  ::  self    !<  Image to check
        !

        ! Start work
        image_suitable_as_mask = .true.
        ! Mask cannot be blank
        if      (self%IsConstant()) then
            image_suitable_as_mask = .false.
        endif
        ! Mask can only have values between 0. and 1.
        if      (any(self%real_values(1:self%GetLogicalDimension(1),1:self%GetLogicalDimension(2),1:self%GetLogicalDimension(3)) &
                                                                                                                    .gt. 1.0)) then
            image_suitable_as_mask = .false.
        else if (any(self%real_values(1:self%GetLogicalDimension(1),1:self%GetLogicalDimension(2),1:self%GetLogicalDimension(3)) &
                                                                                                                    .lt. 0.0)) then
            image_suitable_as_mask = .false.
        endif

    end function image_suitable_as_mask

    !>  \brief Divides img1 by img2 using appropriate arithmetics
    subroutine divide_images(img1,img2,imgout)
        implicit none
        type(image), intent(in)     ::  img1    !< Numerator
        type(image), intent(in)     ::  img2    !< Denominator
        type(image), intent(inout)  ::  imgout  !< Result

        integer ::  itemp

        ! Check both images have the same type
        if ( .not. img1%IsInSameSpaceAs(img2)) then
            write(*,*) '**ERROR (divide_images): cannot divide images of different types'
            stop 'Fatal error in DIVIDE_IMAGES'
        endif

        ! Check both images have the same dimensions
        if ( .not. img1%HasSameDimensionsAs(img2)) then
            write(*,*) '**ERROR (divide_images): cannot divide images of different dimensions'
            stop 'Fatal error in DIVIDE_IMAGES'
        endif

        ! Check both input images have memory allocated
        if ( (.not. img1%IsAllocated()) .or. (.not. img2%IsAllocated())) then
            write(*,*) '**ERROR (divide_images): one or both input images do not have allocated memory'
            stop 'Fatal error in DIVIDE_IMAGES'
        endif

        ! Count the number of zeroes in img2
        if ( .not. img2%IsInRealSpace()) then
            itemp = count(img2%complex_values .eq. (0.0,0.0))
        else
            itemp = count(img2%real_values(1:img2%GetLogicalDimension(1),:,:) .eq. 0.0)
        endif

        ! If there are any zeroes, crash out
        if (itemp .gt. 0) then
            write(*,'(a,i9,a,i3,a)')    '**ERROR: attempt to divide by an image with ', itemp, ' zeroes (',     &
                                        int(100*real(itemp)/    &
                                        (img2%GetLogicalDimension(1)/2*img2%GetLogicalDimension(2)*img2%GetLogicalDimension(3))),&
                                        '%)'
            call terminate('Fatal error in DIVIDE_IMAGES')
        endif

        ! Now we can do the actual division
        if (.not. img2%IsInRealSpace()) then
            imgout%complex_values = img1%complex_values / img2%complex_values
        else
            imgout%real_values(1:img2%GetLogicalDimension(1),:,:) = img1%real_values(1:img2%GetLogicalDimension(1),:,:) / &
                                                                   img2%real_values(1:img2%GetLogicalDimension(1),:,:)
        endif
    end subroutine divide_images

    !>  \brief  The output image contains pixel values from input image 1 or input image 2, whichever is greater. It is therefore a "greater than" image, hence the name of this subroutine.
    subroutine gt_image(img1,img2,imgout)
        implicit none
        ! Arguments
        type(image),    intent(in)      ::  img1
        type(image),    intent(in)      ::  img2
        type(image),    intent(inout)   ::  imgout
        ! Private variables

        ! Start work

        ! Input images must be of same type & real
        if (.not. img1%IsInSameSpaceAs(img2)) then
            write(*,'(a)') '**ERROR(gt_image): input images are of different types'
            call terminate('Fatal error in GT_IMAGE')
        endif
        if (.not. img1%IsInRealSpace()) then
            write(*,'(2a)') '**ERROR(gt_image): can only computer greater than ',   &
                            'image with real inputs'
            call terminate('Fatal error in GT_IMAGE')
        endif


        ! Input images must be have same dimensions
        if (.not. img1%HasSameDimensionsAs(img2)) then
            write(*,'(2(A,3(I5,1x)))') '**ERROR(gt_image): input image dimensions are not identical: ', &
                                        img1%GetLogicalDimensions(), ' and ', img2%GetLogicalDimensions()
            call terminate('Fatal error in GT_IMAGE')
        endif

        ! Output image allocation
        if (imgout%IsAllocated()) then
            if (.not. img1%HasSameDimensionsAs(imgout)) then
                write(*,'(2a)') '**WARNING(gt_image): output image was allocated with wrong dimensions. ',   &
                                'Programming error? Reallocating.'
                call imgout%allocate(img1)
            endif
        else
            call imgout%allocate(img1)
        endif

        ! Compute output image
        where(img1%real_values .gt. img2%real_values)
            imgout%real_values = img1%real_values
        elsewhere
            imgout%real_values = img2%real_values
        end where

    end subroutine gt_image

    !>  \brief Add values from img1 to those of img2, and return the result
    subroutine add_images(img1,img2,imgout)


        implicit none

        type(image), intent(in)     ::  img1
        type(image), intent(in)     ::  img2
        type(image), intent(inout)  ::  imgout

        ! Check both images have the same type
        if ( .not. img1%IsInSameSpaceAs(img2) ) then
            write(*,*) '**WARNING: adding two images of different types '
        endif

        ! Check both images have memory allocated
        if ( .not. (img1%IsAllocated() .and. img2%IsAllocated()) ) then
            write(*,*) ' **ERROR: trying to add images which do not have allocated memory'
            stop ' Fatal error in ADD_IMAGES'
        endif

        if (.not. img1%IsAllocated())       write(*,*) ' **ERROR: img1 not allocated'
        if (.not. img2%IsAllocated())       write(*,*) ' **ERROR: img2 not allocated'
        if (.not. imgout%IsAllocated())     write(*,*) ' **ERROR: imgout not allocated'



        imgout%real_values = img1%real_values + img2%real_values


    end subroutine add_images

end module images_arithm
