!>  \brief  Module with interfaces to various subroutines to compare images
module images_comparison
    !use images
    !use images_arithm, only :  binarise

    implicit none

    !>  \brief  This defines what a function which accumulates scores between images should look like
!   abstract interface
!       function score_accumulat0r(img1,img2,min_freq_rad,max_freq_rad,b_factor,absolute)
!           type(image)                 ::  img1
!           type(image)                 ::  img2
!           integer,        optional    ::  min_freq_rad, max_freq_rad, absolute
!           real,           optional    ::  b_factor
!           real(KIND=8),   allocatable ::  score_accumulat0r
!       end function score_accumulat0r
!   end interface

    !>  \brief  This defines what a function which accumulates scores between images should look like
!   abstract interface
!       function score_finaliser(num_denom_rings,img1,img2,min_freq_rad,max_freq_rad,b_factor,absolute)
!
!       end function score_finaliser
!   end interface

    !>  \brief  a similarity measure object defines all that is needed to know to compare two images (or set of images...) and obtain a quantitative assessment of their similarity
    type similarity_measure
        logical ::  per_frequency   =   .false. !<  Some similarity measures discriminate between frequency bands (i.e. they are weighted / filtered), others (like the traditional, linear cross-correlation coefficient) do not
        real    ::  min_freq, max_freq
        logical ::  arbitrary_filter    =   .false. !<  Whether some arbitrary filter should be used (for example, to up-weigh layer lines etc)

    end type similarity_measure

    type restraint

    end type restraint

	interface score_got_worse
		module procedure score_got_worse_single
		module procedure score_got_worse_double
	end interface

    contains

    !>  \brief  Do the two images have exactly the same values?
    pure logical function images_are_equal(img1,img2)
        use Images
        implicit none
        ! Arguments
        type(image),    intent(in)  ::  img1,img2
        ! Start work
        images_are_equal =  &
            all(img1%real_values(1:img1%logical_dimensions(1),1:img1%logical_dimensions(2),1:img1%logical_dimensions(3)) .eq. &
                img2%real_values(1:img2%logical_dimensions(1),1:img2%logical_dimensions(2),1:img2%logical_dimensions(3)))
        images_are_equal =  images_are_equal .and. img1%HasSameDimensionsAs(img1) .and. img1%IsInSameSpaceAs(img2)
    end function images_are_equal

    !>  \brief  Has the score become significantly worse?
    pure logical function score_got_worse_single(score_start,score_finish)
    	implicit none
		! Arguments
		real(kind=4),   intent(in)  ::  score_start, score_finish
		! Private variables
		real,   parameter   ::  tol_mul = 0.01
		real,   parameter   ::  tol_add = 0.01
		real                ::  score_start_mod
		! Start work
		score_start_mod = min(score_start-tol_add,score_start-abs(tol_mul*score_start))
		score_got_worse_single = score_finish .lt. score_start_mod
	end function score_got_worse_single

	pure logical function score_got_worse_double(score_start,score_finish)
		implicit none
		real(kind=8),	intent(in)	::	score_start,score_finish
		score_got_worse_double = score_got_worse_single(real(score_start,KIND=4),real(score_finish,KIND=4))
	end function score_got_worse_double


    !>  \brief  Prepare a 2d image for a similarity measure
    !!  \todo   Modify the envelope masking so that we mask to the average within the mask, not 0. Alternatively, make sure the projection is zero floated.
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine image_prepare(   img,img_tmp,ctf_image,obj_max_rad,    &
                                apply_ctf, &
                                shift_x,shift_y,env_mask,real_space_shift,   &
                                interpol,interpol_window,   &
                                apply_disc_mask)
        use Images
        use InterpolatingWindows
        use ctfs
        use images_arithm
        use images_ctf
        use interpolation
        use rotation_matrices
        use runtime_parameters, only : default_falloff_pixels
#ifndef skip_runtime_checks
        use runtime_parameters, only : terminate
#endif
        implicit none
        ! Arguments
        type(image),                            intent(inout)   ::  img                 !<  2d image. Will be modified (shifted, CTF-corrected, masked...)
        type(image),                            intent(inout)   ::  img_tmp             !<  2d work image. May be overwritten etc. Should be allocated to be same size as other images.
        type(image),                            intent(in)      ::  ctf_image           !<  ctf image
        real,                                   intent(in)      ::  obj_max_rad         !<  Maximum radius of the object (in pixels)
        logical,                                intent(in)      ::  apply_ctf           !<  Apply CTF to the image
        real,                       optional,   intent(in)      ::  shift_x             !<  Shift in x to be applied to 2d image
        real,                       optional,   intent(in)      ::  shift_y             !<  Shift in y to be applied to 2d image
        type(image),                optional,   intent(inout)   ::  env_mask            !<  a soft-edged, envelope mask to be applied to the image (usually obtained from a 3d projection)
        logical,                    optional,   intent(in)      ::  real_space_shift    !<  Whether the image will be shifted in real-space or in Fourier space (using a phase shift). Defaults to .FALSE.
        type(interpolation_scheme), optional,   intent(in)      ::  interpol            !<  Interpolation scheme to use for real-space shift.
        type(window),               optional,   intent(inout)   ::  interpol_window     !<  Interpolation window.
        logical,                    optional,   intent(in)      ::  apply_disc_mask     !<  Whether a soft-edged disk aperture should be applied to the projection
        ! Private variables
        real                    ::  sshift_x, sshift_y
        logical,    parameter   ::  debug   =   .false.
        logical                 ::  aapply_disc_mask
        ! Private variables below are only used when doing real space shift
        logical                 ::  rreal_space_shift
        real                    ::  matrix(4,4)
        integer					::	i,j,k
		! Start work

        ! Real-space shift?
        if (present(real_space_shift)) then
            rreal_space_shift = real_space_shift
        else
            rreal_space_shift = .false.
        endif

#ifndef skip_runtime_checks
        ! If real space shifts, check interpolation scheme was supplied
        if (rreal_space_shift) then
            if (.not. present(interpol) .or. .not. present(interpol_window)) then
                write(*,'(2a)') '**ERROR(image_prepare): real-space shift requires an',  &
                                ' interpolation scheme to be supplied as argument'
                call terminate('Fatal error in IMAGE_PREPARE')
            endif
        endif
#endif

        ! Will we apply a disk mask?
        if (present(apply_disc_mask)) then
            aapply_disc_mask = apply_disc_mask
        else
            aapply_disc_mask = .true.
        endif



        ! Apply the ctf to the image
        if (apply_ctf) then
            if (img%is_in_real_space) call img%ForwardFFT()
            forall(i=1:img%physical_upper_bound_complex(1),j=1:img%logical_dimensions(2),k=1:img%logical_dimensions(3))
                img%complex_values(i,j,k)  =   img%complex_values(i,j,k)  * ctf_image%real_values(i,j,k)
            end forall
        endif

        ! Apply mask to image to avoid artefacts during shifting
        if (.not. img%is_in_real_space) call img%BackwardFFT()
        call img%ApplySoftCircularMask(real(img%logical_dimensions(1)/2)-6.0,6.0,wipe_input=.false.,mask_with_average=.true.)

        ! Apply (x,y) shift to image
        if (present(shift_x) .or. present(shift_y)) then
            ! Work out what the shift should be
            sshift_x = 0.0
            sshift_y = 0.0
            if (present(shift_x)) sshift_x = shift_x
            if (present(shift_y)) sshift_y = shift_y
            if (sshift_x .ne. 0.0 .or. sshift_y .ne. 0.0) then
                ! Apply the shift, either in real space or Fourier space
                if (rreal_space_shift) then
                    matrix = identity_rotmat()
                    matrix(1,4) = -sshift_x
                    matrix(2,4) = -sshift_y
                    if (.not. img%is_in_real_space) call img%BackwardFFT()
                    call img%ApplyAffineTransformation(img_tmp,matrix)
                    img%real_values = img_tmp%real_values
                else
                    if (img%is_in_real_space) call img%ForwardFFT()
                    call img%PhaseShift(sshift_x,sshift_y,0.0)
                endif
            endif
        endif

        ! Apply envelope mask (calculated from the 3d) to the image, if the mask was supplied
        if (present(env_mask)) then
            if (.not. img%is_in_real_space)  call img%BackwardFFT()
#ifndef skip_runtime_checks
            if (debug) write(*,'(a)') '**debug(image_prepare): applying envelope mask'
#endif
            call image_apply_mask(img,env_mask,mask_with_average=.false.,mask_with_average_outside=.true.)
        endif

        ! Apply a circular mask
        if (aapply_disc_mask) then
#ifndef skip_runtime_checks
            if (debug) print '(a,F0.3,a,i0,a)', '**debug(image_prepare): obj_max_rad, box size = ', &
                                                 obj_max_rad, ' , ',  img%logical_dimensions(1), ' pixels'
#endif
            if (.not. img%is_in_real_space) call img%BackwardFFT()
            call img%ApplySoftCircularMask(OBJ_MAX_RAD+DEFAULT_FALLOFF_PIXELS,default_falloff_pixels,WIPE_INPUT=.false.)
        endif
    end subroutine image_prepare

    !>  \brief  Start the preparation of a projection from a 3d image. Only do the projection / slice extraction.
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine projection_prepare_raw(  vol,proj,angles,interp,win,real_space_proj,vol_shift_z,freq_max,    &
                                        shift_x,shift_y,    &
                                        helical_subunit_only,helix, &
                                        volume_oversampling)
        use euleranglesmod, only : euler
        use images_threed
        use interpolation
        use images
        use InterpolatingWindows
        use helices
        use runtime_parameters, only : terminate
        use OrientationsAndRotations, only : EulerAngles
        implicit none
        ! Arguments
        type(image),                        intent(in)      ::  vol                 !<  3d volume from which a projection will be computed
        type(image),                        intent(inout)   ::  proj                !<  Computed projection
        type(euler),                        intent(in)      ::  angles              !<  Euler angles at which the projection should be computed
        type(interpolation_scheme),         intent(in)      ::  interp              !<  Interpolation scheme
        type(window),           target,     intent(inout)   ::  win                 !<  Interpolating window
        logical,                optional,   intent(in)      ::  real_space_proj     !<  Whether the projection should be computed in real space
        real(kind=8),           optional,   intent(in)      ::  vol_shift_z         !<  The volume will be shifted along its z axis before projection by this amount (in pixels).
        real,                   optional,   intent(in)      ::  freq_max            !<  Maximum frequency (0.5 is Nyquist) of interest
        real(kind=8),           optional,   intent(in)      ::  shift_x             !<  Output projection is shifted by this amount in first dimension (pixels)
        real(kind=8),           optional,   intent(in)      ::  shift_y             !<  Output projection is shifted by this amount in second dimension (pixels)
        logical,                optional,   intent(in)      ::  helical_subunit_only!<  Only the helical subunit is projected
        type(helical_symmetry), optional,   intent(in)      ::  helix               !<  Helical operator (only needed when helical_subunit_only) (not taking oversampling into account)
        integer,                optional,   intent(in)      ::  volume_oversampling !<  If the 3D volume has been oversampled, indicate here the oversampling factor (>1). Give 1 to indicate no oversampling.
        ! Private variables
        logical,    parameter   ::  debug   =   .false.
        logical                 ::  rreal_space_proj
        real(kind=8)            ::  ffreq_max
        logical,    parameter   ::  use_new_extractor = .true.
        type(EulerAngles)       ::  euler_angles
        integer                 ::  vvolume_oversampling
        type(Image)             ::  proj_sml
        type(helical_symmetry)  ::  hhelix
        ! Start work

        ! Real space projection?
        rreal_space_proj = .false.
        if (present(real_space_proj)) rreal_space_proj = real_space_proj

        ! Was the 3d oversampled?
        vvolume_oversampling = 1
        if (present(volume_oversampling)) vvolume_oversampling = volume_oversampling

        ! Maximum frequency
        if (present(freq_max)) then
            ffreq_max = dble(freq_max)
#ifndef skip_runtime_checks
            if (rreal_space_proj) then
                write(*,'(2a)') '**warning(projection_prepare_raw): maximum frequency not implemented',  &
                                ' for real-space projection. Ignored.'
                !call terminate('Fatal error in PROJECTION_PREPARE_RAW')
            endif
#endif
        else
            ffreq_max = 0.5D0
        endif

#ifndef skip_runtime_checks
        if (debug) write(*,'(a,l1)') '**debug(projection_prepare_raw): real space projection? ', rreal_space_proj
#endif

        ! Get a projection from the volume, either via real space projection or Fourier space slice extraction
        if (rreal_space_proj) then
#ifndef skip_runtime_checks
            if (.not. vol%object_is_centered_in_box) then
                write(*,'(a)') '**ERROR(projection_prepare_raw): input volume must have origin at center'
                call terminate('Fatal error in PROJECTION_PREPARE_RAW')
            endif
            if (.not. vol%is_in_real_space) then
                write(*,'(a)') '**ERROR(projection_prepare_raw): input volume must be in real space'
                call terminate('Fatal error in PROJECTION_PREPARE_RAW')
            endif
#endif

            if (debug) write(*,*) '**debug(projection_prepare_raw): vol, proj dims = ', vol%GetLogicalDimensions(), &
                                                                                proj%GetLogicalDimensions()

            call volume_project_lean(vol,proj,angles%phi,angles%theta,angles%psi,interp,win,super_lean=.false.,   &
                                    shift_x=shift_x*vvolume_oversampling, &
                                    shift_y=shift_y*vvolume_oversampling, &
                                    z_shift=vol_shift_z*vvolume_oversampling,    &
                                    helical_subunit_only=helical_subunit_only, &
                                    helix=helix_scale_distances(helix,real(vvolume_oversampling)))
            if (debug) then
                call proj%WriteToDisk('dbg_proj_after_volume_project_lean.mrc')
            endif
        else
            ! Extract a slice from the 3d volume
#ifndef skip_runtime_checks
            if (vol%is_in_real_space) then
                write(*,'(a)') '**ERROR(projection_prepare_raw): input volume must be in Fourier space'
                call terminate('Fatal error in PROJECTION_PREPARE_RAW')
            endif

            if (vol%object_is_centered_in_box) then
                write(*,'(a)') '**ERROR(projection_prepare_raw): input volume must not have origin at center'
                call terminate('Fatal error in PROJECTION_PREPARE_RAW')
            endif
#endif

            if (use_new_extractor) then
                call euler_angles%SetAnglesInDegrees(real(angles%phi),real(angles%theta),real(angles%psi))
                call vol%ExtractComplexSlice(euler_angles,proj, &
                                            real(vol_shift_z*vvolume_oversampling), &
                                            real(shift_x*vvolume_oversampling), &
                                            real(shift_y*vvolume_oversampling), &
                                            real(ffreq_max/vvolume_oversampling))
            else
                call extract_slice( vol,proj,angles%phi,angles%theta,angles%psi,interp,win, &
                                    z_shift=vol_shift_z*vvolume_oversampling,&
                                    rad_max=dble(vol%logical_dimensions(1))*ffreq_max / vvolume_oversampling, &
                                    x_shift=shift_x * vvolume_oversampling, &
                                    y_shift=shift_y * vvolume_oversampling)
            endif
#ifndef skip_runtime_checks
            ! debug
            if (debug) then
                if (proj%IsInFourierSpace()) call proj%BackwardFFT()
                call proj%WriteToDisk('dbg_proj_beforeoriginshift.mrc')
            endif
#endif
            ! Apply shift to the projection (to move origin to center)
            call proj%SwapRealSpaceQuadrants()
#ifndef skip_runtime_checks
            ! debug
            if (debug) then
                if (proj%IsInFourierSpace()) call proj%BackwardFFT()
                call proj%WriteToDisk('dbg_proj_afteroriginshift.mrc')
            endif
#endif
        endif

        if (debug) then
            call proj%WriteToDisk('dbg_proj_before_resampling.mrc')
            write(*,'(a,i0)') '**debug(projection_prepare_raw): 3d oversampling factor = ', vvolume_oversampling
        endif

        ! Undo the effect of 3d oversampling
        if (vvolume_oversampling .gt. 1) then
            if (proj%IsInRealSpace()) call proj%ForwardFFT()
            call proj_sml%Allocate(dims=[proj%GetLogicalDimension(1)/vvolume_oversampling, &
                                         proj%GetLogicalDimension(2)/vvolume_oversampling,1])
            call proj%ClipInto(proj_sml)
            proj = proj_sml
            call proj_sml%Deallocate()
        endif

        if (debug) then
            call proj%WriteToDisk('dbg_proj_after_resampling.mrc')
        endif

    end subroutine projection_prepare_raw

    !>  \brief  Compute an envelope mask from a projection
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine env_mask_from_proj(proj,env_mask,real_space_proj,proj_max_rad,pixel_size,rect_dim,rect_azimuth,tight_mask)
        use masks
        use runtime_parameters, only : default_falloff_pixels, terminate
        use images_arithm, only : binarise
        use images
        use Globals
        implicit none
        ! Arguments
        type(image),            intent(in)      ::  proj            !<  projection image
        type(image),            intent(inout)   ::  env_mask        !<  envelope mask
        logical,                intent(in)      ::  real_space_proj !<  whether the projection was calculated in real space
        real,                   intent(in)      ::  proj_max_rad    !<  in pixels, the maximum radius of interest
        real,                   intent(in)      ::  pixel_size      !<  Pixel size
        real,       optional,   intent(in)      ::  rect_dim(2)     !<  pixels outside this rectangle will be ignored
        real,       optional,   intent(in)      ::  rect_azimuth    !<  azimuth for the rectangle
        logical,    optional,   intent(in)      ::  tight_mask      !<  The envelope should follow density tightly (true by default)
        ! Private variables
        logical,    parameter   ::  debug   =   .false.
        real                    ::  env_mask_thresh
        real,       parameter   ::  soft_edge       =   default_falloff_pixels  !<  The width of the soft-edge falloff for masking
        logical                 ::  ttight_mask
        ! Start work
        env_mask = proj

        ! Do we want a tight mask? (when generating projections during refinement, a tight mask may make sense, but not
        ! when generating synthetic micrographs, for example)
        ! Currently, we don't really have a choice...
        ttight_mask = .true.
        if (present(tight_mask)) ttight_mask = tight_mask

        if (debug) call env_mask%WriteToDisk('dbg_env_mask_entry.mrc')
        if (env_mask%is_in_real_space) call env_mask%ForwardFFT()
        call env_mask%ApplyBandPassFilter(0.0,pixel_size/25.0,width_of_cosine_falloff=0.0) ! Make the low-pass 25a always
        if (.not. env_mask%is_in_real_space) call env_mask%BackwardFFT()
#ifndef skip_runtime_checks
        if (debug) call env_mask%WriteToDisk('dbg_env_mask_lp.mrc')
#endif
        if (present(rect_dim)) then
#ifndef skip_runtime_checks
            if (.not. present(rect_azimuth)) then
                write(*,'(a)') '**ERROR(env_mask_from_proj): rect dim, but not azimuth, supplied.'
                call terminate('Fatal error in ENV_MASK_FROM_PROJ. Programming error.')
            endif
#endif
            if (env_mask%IsInFourierSpace()) call env_mask%BackwardFFT()
            call env_mask%ApplyRectangularMask(RECT_DIM=RECT_DIM,AZIMUTH=RECT_AZIMUTH,WIPE_INPUT=.false.)
#ifndef skip_runtime_checks
            if (debug) call env_mask%WriteToDisk('dbg_env_mask_lp_rectmasked.mrc')
#endif
        endif
        call env_mask%ApplyCircularMask(proj_max_rad+1.)
#ifndef skip_runtime_checks
        if (debug) call env_mask%WriteToDisk('dbg_env_mask_lp_disc.mrc')
#endif
        if (real_space_proj) then
            if (ttight_mask) then
                env_mask_thresh = sqrt(env_mask%GetVarianceOfValues())
            else
                ! This is basically 0.0 - very loose mask, which is good when making synth micrographs,
                ! but I'm not sure it's so good when making projections when doing refinements etc
                env_mask_thresh = 10.*proj%logical_dimensions(1)*tiny(1.0)
            endif

            !print *, 'old threshold = ', 10.*PROJ%logical_dimensions(1)*TINY(1.0)
            !print *, 'average = ', env_mask%GetAverageOfValues()
            !print *, 'std = ', sqrt(env_mask%GetVarianceOfValues())
        else
            if (ttight_mask) then
                env_mask_thresh = sqrt(env_mask%GetVarianceOfValues()) !! todo: shouldn't this be average + 1 sigma???
            else
                ! This is basically 0.0 - very loose mask, which is good when making synth micrographs,
                ! but I'm not sure it's so good when making projections when doing refinements etc
                env_mask_thresh = 10.*proj%logical_dimensions(1)*tiny(1.0)
            endif
        endif
        call binarise(env_mask,env_mask_thresh)
        call env_mask%ApplyCircularMask(PROJ_MAX_RAD+1.)
        call mask_add_cosine_falloff(env_mask,12.0/pixel_size) ! 12 Angstroms
#ifndef skip_runtime_checks
        if (debug) then
            write(*,'(a)') '**debug(projection_prepare): dumping envelope mask to dbg_env_mask.mrc'
            call env_mask%WriteToDisk('dbg_env_mask.mrc')
        endif
        if (env_mask%IsConstant()) then
            write(*,'(a)') '**error(env_mask_from_proj): oops, mask is not a mask'
            call env_mask%WriteToDisk('dbg_env_mask_pb.mrc')
            env_mask = proj
            call env_mask%WriteToDisk('dbg_env_mask_pb_proj.mrc')
            print *, 'low pass = ', pixel_size/25.0
            call env_mask%ApplyBandPassFilter(0.0,pixel_size/25.0,width_of_cosine_falloff=0.0)
            call env_mask%WriteToDisk('dbg_env_mask_pb_proj_lp.mrc')
        endif
#endif

    end subroutine env_mask_from_proj

    !>  \brief  Prepare a projection from a 3d image for a similarity measure
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine projection_prepare(  vol,proj,proj_dim, &
                                    wp, &
                                    ctf_image,proj_max_rad, &
                                    interpol,win,    &
                                    pixel_size, &
                                    env_mask,real_space_proj,vol_shift_z,apply_disc_mask,   &
                                    freq_max,&
                                    ctf_squared, &
                                    apply_ctf,single_helical_subunit_only,helix,rect_mask_width, &
                                    segment_with_helical_subunits,helix_max_rad, &
                                    volume_oversampling,tight_env_mask)
        use euleranglesmod
        use interpolation
        use images
        use InterpolatingWindows
        use masks
        use images_ctf
        use helices, only : helical_symmetry
        use runtime_parameters, only  : default_falloff_pixels, terminate
        use images_arithm, only : image_apply_mask
        use coos, only : waypoint, waypoints_print_info
        use UsefulFunctions, only : IsEven, DegreesToRadians
        use images_helices, only : image_project_many_h_subunits_with_h_distortion, image_project_many_helical_subunits
        use differential_geometry, only : waypoint_get_path_near_wp
        implicit none
        ! Arguments
        type(image),                        intent(inout)   ::  vol                             !<  3d volume from which a projection image will be computed
        type(image),                        intent(inout)   ::  proj                            !<  Output projection image
        integer,                            intent(in)      ::  proj_dim(3)                     !<  Desired dimensions for the projection
        type(waypoint),                     intent(in)      ::  wp                              !<  Wapoint which describes the projection we want. The X,Y coordinates of the waypoints describe the shift applied to the projection away from the projection box center.
        type(image),                        intent(in)      ::  ctf_image                       !<  ctf parameters
        real,                               intent(in)      ::  proj_max_rad                    !<  Maximum radius of the projection (pixels). Pixels beyond this radius can be set to zero, for example.
        type(interpolation_scheme),         intent(in)      ::  interpol                        !<  Interpolation scheme
        type(window),           target,     intent(inout)   ::  win                             !<  Interpolating window
        real,                               intent(in)      ::  pixel_size                      !<  In Angstroms
        type(image),            optional,   intent(inout)   ::  env_mask                        !<  If this argument is given, a envelope mask will be calculated from the projection. This can later be used to mask a noisy image.
        logical,                optional,   intent(in)      ::  real_space_proj                 !<  Whether to do real space projection. By default, this is false and Fourier space slice extraction is done instead.
        real(kind=8),           optional,   intent(in)      ::  vol_shift_z                     !<  The volume will be shifted along its z axis before projection by this amount (in pixels).
        logical,                optional,   intent(in)      ::  apply_disc_mask                 !<  Whether a soft-edged disk aperture should be applied to the projection
        real,                   optional,   intent(in)      ::  freq_max                        !<  Maximum frequency (0.5 is Nyquist) of interest
        logical,                optional,   intent(in)      ::  ctf_squared                     !<  For testing purposes, one can multiply by the ctf rather than the CTF^2
        logical,                optional,   intent(in)      ::  apply_ctf                       !<  For testing purposes, the ctf can be switched off.
        logical,                optional,   intent(in)      ::  single_helical_subunit_only     !<  Only the helical subunit is projected
        type(helical_symmetry), optional,   intent(in)      ::  helix                           !<  Helical operator corresponding to the volume to be projected (only needed when projecting helical subunits)
        real,                   optional,   intent(in)      ::  rect_mask_width                 !<  Useful when projecting helical structure. The width of the rectangle mask to be applied
        logical,                optional,   intent(in)      ::  segment_with_helical_subunits   !<  Project a whole segment, but do it by projecting many helical subunits
        real,                   optional,   intent(in)      ::  helix_max_rad                   !<  Maximum helical radius (pixels). Only needed when dealing with subunits.
        integer,                optional,   intent(in)      ::  volume_oversampling             !<  If the 3D volume has been oversampled, indicate here the oversampling factor (>1). Give 1 to indicate no oversampling.
        logical,                optional,   intent(in)      ::  tight_env_mask                  !<  Envelope mask should be tight
        ! Private variables
        real,           parameter   ::  soft_edge       =   default_falloff_pixels  !<  The width of the soft-edge falloff for masking
        logical,        parameter   ::  debug           =   .false.
        logical,        parameter   ::  distort_h_subunit = .false.
        logical                     ::  rreal_space_proj
        logical                     ::  aapply_disc_mask
        logical                     ::  cctf_squared
        logical                     ::  aapply_ctf
        integer					    ::	i,j,k
        real                        ::  rectangle_dimensions(2)
        logical                     ::  ssegment_with_helical_subunits
        logical                     ::  ssingle_helical_subunit_only
        type(waypoint), allocatable ::  subunits(:)
        integer                     ::  number_of_subunits_in_projection

		! Start work

        ! Real space projection?
        rreal_space_proj = .false.
        if (present(real_space_proj))  rreal_space_proj = real_space_proj

        ! Will we apply a disk mask?
        aapply_disc_mask = .true.
        if (present(apply_disc_mask)) aapply_disc_mask = apply_disc_mask

        ! Will we apply the ctf to the projection
        aapply_ctf = .true.
        if (present(apply_ctf)) aapply_ctf = apply_ctf

        ! Squared CTF?
        cctf_squared = .true.
        if (present(ctf_squared)) cctf_squared = ctf_squared

        ! Generate a segment by projecting many helical subunits?
        ssegment_with_helical_subunits = .false.
        if (present(segment_with_helical_subunits)) ssegment_with_helical_subunits = segment_with_helical_subunits

        ! Project a single subunit?
        ssingle_helical_subunit_only = .false.
        if (present(single_helical_subunit_only)) ssingle_helical_subunit_only = single_helical_subunit_only

        ! Check we have what we need to deal with subunits
        if (ssegment_with_helical_subunits .or. ssingle_helical_subunit_only) then
            if (.not. present(helix) .or. .not. present(helix_max_rad)) then
                write(*,'(a)') '**error(projection_prepare): need both the helical symmetry and the max helical rad'
                call terminate('Fatal error in projection_prepare')
            endif
        endif


        if (ssegment_with_helical_subunits) then
            if (.not. wp%helix%pixel_units) then
                call terminate('Fatal error in projection_prepare. Helix is not in pixel units')
            endif
            ! How many subunits will fit in the projection?
            !
            !
            ! Todo: properly determine the maximum number of subunits we could project, taking into account the max
            !       h radius
            !
            number_of_subunits_in_projection = max( &
                            ceiling(2.0*proj_max_rad / wp%helix%rise / sin(DegreesToRadians(wp%euler%theta))), &
                            nint(proj_dim(1)/wp%helix%rise / sin(DegreesToRadians(wp%euler%theta)) * 0.85) &
                            ) &
                            * 0.85 ! we assume there'll be some masking so we don't need go all the way to the edge
            if (IsEven(number_of_subunits_in_projection)) number_of_subunits_in_projection = number_of_subunits_in_projection - 1
            ! The following step will only work if we are exactly on a subunit, so let's check
            if (abs(wp%subunit_num - nint(wp%subunit_num)) .gt. 0.001) then
                write(*,'(a)') '**error(projection_prepare): please ensure the waypoint coincides with a subunit'
                write(*,'(a,f0.3)') '**debug(projection_prepare): subunit_num = ', wp%subunit_num
                call terminate('Fatal error in Frealix')
            endif
            ! We will need to get a set of waypoints (one for each helical subunit) around the input waypoint
            subunits = waypoint_get_path_near_wp(wp%helix,1.0d0,number_of_subunits_in_projection,wp,step_in_repeats=.true.)
            ! Now let's compute the projection
            !proj_larger_dims = [nint(proj%logical_dimensions(1)*1.0),nint(proj%logical_dimensions(2)*1.0),1]
            !call proj_larger%Allocate(dims=proj_larger_dims)
            !write(*,'(a,3(i0,1x))') '**debug(projection_prepare): proj_larger_dims = ', proj_larger_dims

            !
            ! Project the helical subunits. We can do this with or without distortion of the helical lattice.
            !
            if (distort_h_subunit) then
                call image_project_many_h_subunits_with_h_distortion(vol,helix,proj,subunits,helix_max_rad,&
                                                                     volume_oversampling=volume_oversampling)
            else
                call image_project_many_helical_subunits(vol,helix,proj,subunits,helix_max_rad, &
                                                         volume_oversampling=volume_oversampling, &
                                                         vol_shift_z=real(vol_shift_z))
            endif
            !call proj_larger%ClipInto(proj)
        else
            if (rreal_space_proj) then
                if (.not. vol%ObjectIsCenteredInBox()) call vol%SwapRealSpaceQuadrants()
                if (.not. vol%IsInRealSpace()) call vol%BackwardFFT()
            else
                if (vol%IsInRealSpace()) call vol%ForwardFFT()
                if (vol%ObjectIsCenteredInBox()) call vol%SwapRealSpaceQuadrants()
            endif
            call projection_prepare_raw(vol,proj,wp%euler,interpol,win,real_space_proj, &
                                        vol_shift_z=vol_shift_z,freq_max=freq_max, &
                                        shift_x=wp%coo(1),shift_y=wp%coo(2),&
                                        helical_subunit_only=single_helical_subunit_only,helix=helix, &
                                        volume_oversampling=volume_oversampling)
            if (debug) then
                call proj%WriteToDisk('dbg_proj_raw.mrc')
            endif

            ! Resize projection if necessary
            if (.not. proj%IsInRealSpace()) call proj%BackwardFFT()
            call proj%Resize(proj_dim)
        endif

        ! Check the projection is OK
#ifndef SKIP_RUNTIME_CHECKS
        if (proj%IsConstant()) then
            write(*,'(a)') '**error(projection_prepare): raw proj (before masking) is blank'
            call proj%WriteToDisk('dbg_proj_pb.mrc')
            call vol%WriteToDisk('dbg_proj_vol.mrc')
            print *, vol_shift_z
            print *, wp%euler
            print *, wp%coo(1:2)
            print *, freq_max
        endif
#endif


        ! If we're going to need it later, calculate an envelope mask from the projection,
        ! before applying the ctf
        if (present(env_mask)) then
            if (present(rect_mask_width)) then
                rectangle_dimensions = (/real(proj%logical_dimensions(1)),rect_mask_width/)
                call env_mask_from_proj(proj,env_mask,rreal_space_proj,proj_max_rad,    &
                                        pixel_size=pixel_size, &
                                        rect_dim=rectangle_dimensions,rect_azimuth=real(wp%euler%psi), &
                                        tight_mask=tight_env_mask)
            else
                call env_mask_from_proj(proj,env_mask,rreal_space_proj,proj_max_rad,pixel_size, &
                                        tight_mask=tight_env_mask)
            endif

        endif

        ! Apply a soft cosine-edged disc mask so that ctf multiplication creates less ripples
        if (.not. proj%is_in_real_space) call proj%BackwardFFT()
        call proj%ApplySoftCircularMask(real(proj%logical_dimensions(2))/2.0-2.0,proj%logical_dimensions(2)*0.2,&
                                        wipe_input=.false.,mask_with_average=.false.)


        ! Apply the ctf to the projection
        if (proj%is_in_real_space) call proj%ForwardFFT()
        if (aapply_ctf) then
            !call image_apply_ctf(proj,ctf,CTF_SQUARED=CTF_SQUARED)
            if (cctf_squared) then
                forall(i=1:proj%physical_upper_bound_complex(1),j=1:proj%logical_dimensions(2),k=1:proj%logical_dimensions(3))
                    proj%complex_values(i,j,k) =   proj%complex_values(i,j,k) * ctf_image%real_values(i,j,k)**2
                end forall
            else
                forall(i=1:proj%physical_upper_bound_complex(1),j=1:proj%logical_dimensions(2),k=1:proj%logical_dimensions(3))
                    proj%complex_values(i,j,k) =   proj%complex_values(i,j,k) * ctf_image%real_values(i,j,k)
                end forall
            endif
        endif


        ! Apply the envelope mask if calculated
        if (present(env_mask)) then
            if (.not. proj%is_in_real_space) call proj%BackwardFFT()
            call image_apply_mask(proj,env_mask,mask_with_average=.false.)
        endif


        ! Apply a circular mask
        if (aapply_disc_mask) then
            if (.not. proj%is_in_real_space) call proj%BackwardFFT()
            call proj%ApplySoftCircularMask(proj_max_rad+soft_edge,soft_edge,wipe_input=.false.,mask_with_average=.false.)
        endif

    end subroutine projection_prepare

    !>  \brief  Calculate the cross-correlation function between two images by conjugate-multiplying their FT. If they
    !!          are real, they'll be FT'ed back and forth internally
    !!
    subroutine fast_correlation(img_in1,img_in2,img_out,fourier_mask)
        use Images
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),                intent(inout)   ::  img_in1
        type(image),                intent(inout)   ::  img_in2
        type(image),                intent(inout)   ::  img_out
        type(image),    optional,   intent(in)      ::  fourier_mask    !<  Real image. The ft of the cc function is multiplied by this.
        ! Private variables
        logical ::  do_fts
        ! Start work
        ! Check both images are of same type
        if (.not. img_in1%IsInSameSpaceAs(img_in2) ) then
            call terminate('Fatal error in FAST_CORRELATION')
        endif
        ! Check both images have same dimensions
        if (.not. img_in1%HasSameDimensionsAs(img_in2)) then
            write(*,'(a,2(3(i3,1x),a))') '**ERROR(fast_correlation): input images do not have same dimensions: ', &
                                        img_in1%logical_dimensions, ' & ', img_in2%logical_dimensions
            call terminate('Fatal error in FAST_CORRELATION')
        endif
        ! Check whether we'll have to do FTs internally
        do_fts = img_in1%is_in_real_space
        ! Allocate output image if necessary
        if (.not. (img_in1%HasSameDimensionsAs(img_out) .and. img_out%IsAllocated())) then
            call img_out%Allocate(img_in1)
        endif
        ! Prepare the output image, which will initially be complex
        img_out%is_in_real_space = .false.
        img_out%object_is_centered_in_box = .false.
        ! Fourier transform both inputs if necessary
        if (do_fts) then
            call img_in1%ForwardFFT()
            call img_in2%ForwardFFT()
        endif
        ! Multiply
        img_out%complex_values = img_in1%complex_values * conjg(img_in2%complex_values)
        ! Apply Fourier mask
        if (present(fourier_mask)) then
            ! Check the dimensions are as expected
            if (        fourier_mask%logical_dimensions(1) .ne. img_out%physical_upper_bound_complex(1) &
                .or.    any(fourier_mask%logical_dimensions(2:3) .ne. img_out%logical_dimensions(2:3)) ) then
                write(*,'(a,2(3(i0,1x),1x))') '**ERROR(fast_correlation): mismatch between image and Fourier mask dimensions: ', &
                                                fourier_mask%logical_dimensions, img_out%logical_dimensions
                call terminate('Fatal error in FAST_CORRELATION')
            endif
            ! Apply the mask
            img_out%complex_values(1:img_out%physical_upper_bound_complex(1),:,:) = &
                                           img_out%complex_values(1:img_out%physical_upper_bound_complex(1),:,:) * &
                                           fourier_mask%real_values(1:FOURIER_MASK%logical_dimensions(1),:,:)
        endif
        ! Center
        call img_out%SwapRealSpaceQuadrants()
        ! Reverse ft the output
        call img_out%BackwardFFT()
        ! The scaling needs to be adjusted
        img_out%real_values = img_out%real_values * product(img_out%logical_dimensions)
        ! If necessary, return the input images to real space
        if (do_fts) then
            call img_in1%BackwardFFT()
            call img_in2%BackwardFFT()
        endif
    end subroutine fast_correlation


    !>  \brief  Find the shift and rotation Euler angles that can be applied to image1 to superimpose it onto image2
    !!  \todo   Peak lists
    subroutine find_shift_euler(img1,img2,shift,rotation,integer_shift,rotation_sampling,rotation_half_range,rotation_first,    &
                                shift_only,max_rad_shift,horizontal_only,vertical_only,fourier_mask)
        use images
        use euleranglesmod
        use rotation_matrices
        use interpolation
        use peaks
        use images_arithm
        use InterpolatingWindows
        implicit none
        ! Arguments
        type(image),                intent(inout)   ::  img1                    !<  Image 1
        type(image),                intent(inout)   ::  img2                    !<  Image 2
        real,                       intent(out)     ::  shift(3)                !<  The translational shift
        type(euler),                intent(out)     ::  rotation                !<  The Euler angles defining the rotation to be applied
        logical,        optional,   intent(in)      ::  integer_shift           !<  If .TRUE., we will look for a shift of an integer number of pixels. .TRUE. is the default.
        type(euler),    optional,   intent(in)      ::  rotation_sampling       !<  The sampling step (degrees) to be used in exhaustive search of rotations. Default is 5.0 degrees for each Euler angle.
        type(euler),    optional,   intent(in)      ::  rotation_half_range     !<  The half-range (degrees) for the exhaustive search in each Euler angle. Default is 10.0 degrees for each Euler angle.
        logical,        optional,   intent(in)      ::  rotation_first          !<  Whether the rotation should be applied to image 1 before the translation to align it to image 2. .TRUE. by default
        logical,        optional,   intent(in)      ::  shift_only              !<  If .TRUE., no rotation will be attempted
        real,           optional,   intent(in)      ::  max_rad_shift           !<  The maximum radial shift (in pixels) allowed in the alignment
        logical,        optional,   intent(in)      ::  horizontal_only         !<  Do horizontal alignment only (search along 1st dimension only)
        logical,        optional,   intent(in)      ::  vertical_only           !<  Do vertical alignment only (search along 2nd dimension only)
        type(image),    optional,   intent(in)      ::  fourier_mask            !<  Real-space image - non-zero voxels will be used in the correlation calculation
        ! Private variables
        logical     ::  iinteger_shift
        type(euler) ::  rrotation_sampling
        type(euler) ::  rrotation_half_range
        logical     ::  rrotation_first
        real        ::  current_phi, current_theta, current_psi
        type(image) ::  img_rot
        real        ::  matrix(4,4)
        type(peak)  ::  ccf_peak, ccf_peak_best
        type(image) ::  ccf
        integer     ::  origin(3)
        real        ::  current_ccc, best_ccc
        type(image) ::  mask
        real        ::  current_shift(3)
        logical, parameter     ::  real_space_rotation = .true.
        logical     ::  sshift_only
        type(interpolation_scheme)  ::  interpol
        type(window)                ::  interpol_window
        !logical        ::  hhorizontal_only,vvertical_only
        ! Check optional arguments
        if (present(integer_shift)) then
            iinteger_shift = integer_shift
        else
            iinteger_shift = .true.
        endif
        if (present(shift_only)) then
            sshift_only = shift_only
        else
            sshift_only = .false.
        endif

        ! Initialise interpolation
        interpol = default_interpol_scheme()
        call interpol_window%AllocateWindow(interpol%window_width,.not.REAL_SPACE_ROTATION,img1%IsAVolume())

        ! Set the rotational search parameters
        if (sshift_only) then
            rrotation_sampling%phi      = 1.0
            rrotation_sampling%theta    = 1.0
            rrotation_sampling%psi      = 1.0
            rrotation_half_range%phi    = 0.0
            rrotation_half_range%theta  = 0.0
            rrotation_half_range%psi    = 0.0
        else
            if (present(rotation_sampling)) then
                rrotation_sampling = rotation_sampling
            else
                rrotation_sampling%phi = 5.0
                rrotation_sampling%theta = 5.0
                rrotation_sampling%psi = 5.0
            endif
            if (present(rotation_half_range)) then
                rrotation_half_range = rotation_half_range
            else
                rrotation_half_range%phi = 10.0
                rrotation_half_range%theta = 10.0
                rrotation_half_range%psi = 10.0
            endif
        endif

        if (present(rotation_first)) then
            rrotation_first = rotation_first
        else
            rrotation_first = .true.
        endif

        ! We don't have a peak finding routine that returns non-integer peak coordinates yet
        if (.not. iinteger_shift) then
            write(*,'(2a)') '**ERROR(find_shift_euler): We dont have a peak finding routine',    &
                            ' that returns non-integer peak coordinates yet'
            call terminate('Fatal error in FIND_SHIFT_EULER')
        endif

        ! Start work

        ! For now, we only deal with 2d images, to save on coding time!
        if (img1%logical_dimensions(3) .ne. 1) then
            write(*,'(a)') '**ERROR(find_shift_euler): only 2d images allowed, for now'
            call terminate('Fatal error in FIND_SHIFT_EULER')
        endif

        ! For now, we only do rotation_first, to save on coding time!
        if (.not. rrotation_first) then
            write(*,'(a)') '**ERROR(find_shift_euler): only rotation first allowed, for now'
            call terminate('Fatal error in FIND_SHIFT_EULER')
        endif


        ! Allocate memory to mask
        call mask%Allocate(img1)
        mask%real_values = 1.0

        ! Work out the origin
        if (img1%is_in_real_space .and. img1%object_is_centered_in_box) then
            origin = img1%physical_address_of_box_center
        else
            origin = [1,1,1]
        endif

        ! Fourier transform
        if (.not. real_space_rotation) then
            call img1%ForwardFFT()
            call img2%ForwardFFT()
        endif

        ! Allocate memory to rotated image
        call img_rot%Allocate(img1)

        current_phi = 0.0
        current_theta = 0.0
        current_psi = - rrotation_half_range%psi
        ccf_peak_best%CoOrdinates = [0.0,0.0,0.0]
        ccf_peak_best%PeakHeight = -999999.9999
        best_ccc = -1.0
        do while (current_psi .le. rrotation_half_range%psi)
            ! Apply rotation (rotate_image needs the matrix that transforms output into input)
            matrix = euler_to_rotmat(dble(current_phi),dble(current_theta),dble(current_psi),REVERSE=.true.)
            if (.not. real_space_rotation) then
                call img1%SwapRealSpaceQuadrants()
            endif
            call img1%ApplyAffineTransformation(img_rot,matrix)
            if (.not. real_space_rotation) then
                call img1%SwapRealSpaceQuadrants()
                call img_rot%SwapRealSpaceQuadrants()
            endif

            !call img_rot%BackwardFFT()
            !call img_rot%WriteToDisk('debug_rot_before_cc.img')
            !call img_rot%ForwardFFT()

            ! Calculate the cross-correlation function between the two images
            call fast_correlation(img_rot,img2,ccf,FOURIER_MASK=FOURIER_MASK)

            !
            !print *, '**debug(find_shift_euler): dumping CCF'
            !call ccf%WriteToDisk('dbg_ccf.img')

            ! Find the peak
            if (iinteger_shift) then
                !print *, '**debug(find_shift_euler): going to do a search with vertical_only =', vertical_only
                if (present(vertical_only)) then
                    if (vertical_only) then
                        call terminate('not implemented')
                    endif
                endif
                if (present(horizontal_only)) then
                    if (horizontal_only) then
                        call terminate('not implemented')
                    endif
                endif
                ccf_peak = ccf%FindPeakWithIntegerCoordinates(max_radius=max_rad_shift)
            else
                ! No function for this yet
                ! 2013: there is a function for this now - just need to call it here
            endif

            ! Do a normalised cross-correlation coefficient calculation to find the peak height
            current_shift = - (ccf_peak%CoOrdinates - origin)
            !print '(a,3(i6,x))', '**debug(find_shift_euler): origin       = ', origin
            !print '(a,3(F8.3,x))', '**debug(find_shift_euler): ccf_peak%coo = ', ccf_peak%coo
            !print '(a,2(F8.3,x))', '**debug(find_shift_euler): shifting by ', current_shift(1), current_shift(2)
            if ( real_space_rotation) then
                matrix(4,1:3) = current_shift(1:3)
                call img1%ApplyAffineTransformation(img_rot,matrix)
            else
                call img_rot%PhaseShift(current_shift(1),current_shift(2),0.0)
            endif


            !call img_rot%BackwardFFT()
            !call img_rot%WriteToDisk('debug_rot_after_cc2.img')
            !call img_rot%ForwardFFT()

            current_ccc = correlation(img_rot,img2,mask,NORMALISE=.true.)
            !if (current_psi .EQ. -2.0) then
            !   if (.NOT. real_space_rotation) then
            !       call img_rot%BackwardFFT()
            !       call img2%BackwardFFT()
            !   endif
            !   call ccf%WriteToDisk('debug_ccf.img')
            !   call img_rot%WriteToDisk('debug_rot.img')
            !   call img2%WriteToDisk('debug_ref.img')
            !   if (.NOT. real_space_rotation) then
            !       call img2%ForwardFFT()
            !       call img_rot%ForwardFFT()
            !   endif
            !endif

            !print '(a,3(F6.3,x))', '**debug(find_shift_euler): current_psi, peak value, ccc: ', current_psi, ccf_peak%real_values, current_ccc

            !call terminate()

            ! Is this the best peak we've seen so far?
            !if (ccf_peak%real_values .GT. ccf_peak_best%real_values) then
            if (current_ccc .gt. best_ccc) then
                !print '(a,F6.3,a)', '**debug(find_shift_euler): current_psi of ', current_psi, ' beats current max.'
                ccf_peak_best  = ccf_peak
                rotation%phi   = current_phi
                rotation%theta = current_theta
                rotation%psi   = current_psi
                shift = - (ccf_peak_best%CoOrdinates - origin)
                best_ccc = current_ccc
            endif

            ! Increment psi for the next iteration of the search
            current_psi = current_psi + rrotation_sampling%psi
        ! End of loop over psi
        enddo

        !call terminate()

    end subroutine find_shift_euler

    !>  \brief  Find the translation shift (x,y,z) which can be applied to image1 to superimpose it onto image2.
    function find_shift(img1,img2,integer_shift,z_shift_only,min_radius) result(shift)
        !use images
        use peaks
        use images_arithm
        implicit none
        ! Arguments
        type(image),            intent(inout)   ::  img1            !<  Image 1
        type(image),            intent(inout)   ::  img2            !<  Image 2
        logical,    optional,   intent(in)      ::  integer_shift   !<  If .TRUE., we will look for a shift of an integer number of pixels. .TRUE. is the default.
        logical,    optional,   intent(in)      ::  z_shift_only    !<  Only look for shift along the 3rd axis
        real,       optional,   intent(in)      ::  min_radius      !<  Only look for peaks further or at this radius distance from origin (pixels)
        ! Result
        real        ::  shift(3)
        ! Private variables
        logical     ::  iinteger_shift
        type(image) ::  ccf ! Cross-correlation function
        type(peak)  ::  ccf_peak
        logical,    parameter   ::  debug = .false.
        integer     ::  ori(3)!, i
        ! Start work

        ! Integer shift?
        if (present(integer_shift)) then
            iinteger_shift = integer_shift
        else
            iinteger_shift = .true.
        endif

        ! We don't have a peak finding routine that returns non-integer peak coordinates yet
        if (.not. iinteger_shift) then
            write(*,'(a)') '**ERROR(find_shift): We dont have a peak finding routine that returns non-integer peak coordinates yet'
            call terminate('Fatal error in FIND_SHIFT')
        endif

        ! Maybe add filtering step here?

        ! Calculate the cross-correlation function between the two images
        call fast_correlation(img1,img2,ccf)
        if (debug) then
            call ccf%WriteToDisk('dbg_ccf.mrc')
        endif
        ! Find the peak
        if (iinteger_shift) then
            ccf_peak = ccf%FindPeakWithIntegerCoordinates(min_radius=min_radius)
        else
            ! No function for this yet
            ! 2014: there is now - call it here
        endif
        !
        ori = img1%GetPhysicalAddressOfBoxCenter()
        shift = ccf_peak%CoOrdinates - ori    ! Maybe this needs a minus?
    end function find_shift

    !>  \brief  For 6 search parameters (x,y,z,phi,theta,psi), given a target accuracy, return scaling factors to make the parameter values comparable to each other
    function param_scaling_from_accuracy(accuracy,obj_max_radius)
        use runtime_parameters, only: terminate
        use rotation_matrices, only : radians
        implicit none
        ! Arguments
        real,       intent(in)      ::  accuracy(6)     !<  The desired accuracy for each parameter
        real,       intent(in)      ::  obj_max_radius  !<  The maximum radius of the object (pixels)
        ! Result
        real    ::  param_scaling_from_accuracy(6)
        ! Private variables
        logical,    parameter       ::  debug   =   .false.
        ! Start work

        ! Check that the search accuracy is known
        if (.not. any(accuracy .ne. 0.0)) then
            write(*,'(2a)') '**ERROR(param_scaling_from_accuracy): desired search accuracy unknown', &
                            ' - cannot compute scaling factors for search parameters'
            call terminate('Fatal error in PARAM_SCALING_FROM_ACCURACY')
        endif

        ! Compute the scale factors
        param_scaling_from_accuracy(1:3)    =   accuracy(1:3) / accuracy(1)
        param_scaling_from_accuracy(4:6)    =   radians(accuracy(4:6)) * obj_max_radius / accuracy(1)

        ! Debug
        if (debug) then
            write(*,'(a,6(F8.3,1x))') '**debug(param_scaling_from_accuracy): search accuracy = ', accuracy
            write(*,'(a,6(F8.3,1x))') '**debug(param_scaling_from_accuracy): param. scaling  = ', param_scaling_from_accuracy
        endif

    end function param_scaling_from_accuracy

    !>  \brief  Return the desired accuracy in the 6 degrees of freedom (x,y,z,phi,theta,psi) to obtain the target resolution
    !! For details of derivation & meanings, see documentation.
    !! Note that these are just ballpark figures.
    !!
    !!  \todo   Thighten up definitions: what exactly are we returning here? What phase error is targeted? Maybe relate to Jensen paper.
    function accuracy_from_resolution(resolution,obj_max_rad,box_size,akv,pixel_size,helical_rise, &
                                      akv_is_wl,target_phase_error) result (accuracy)
        use ctfs, only : wl_to_akv, akv_to_wl
        use rotation_matrices, only : degrees, radians
        use Globals
        implicit none
        ! Arguments
        real,                   intent(in)  ::  resolution          !<  Target resolution (in A)
        real,                   intent(in)  ::  obj_max_rad         !<  Object maximum radius (in A)
        integer,                intent(in)  ::  box_size            !<  Reconstruction volume dimension (in pixels). Assumption that the volume is cubic and can be specified by just one value.
        real,                   intent(in)  ::  akv                 !<  Acceleration voltage, in kV, unless akv_is_wl is given as .TRUE., in which case this argument is actually the wavelength (Angstroms).
        real,                   intent(in)  ::  pixel_size          !<  Pixel size in Angstrom.
        real,                   intent(in)  ::  helical_rise        !<  Pixels
        logical,    optional,   intent(in)  ::  akv_is_wl           !<  If true, the akv argument has already been converted to wavelength.
        real,       optional,   intent(in)  ::  target_phase_error  !<  The target phase error (in degrees). Defaults to a sensible value (see below).
        ! Result
        real                                ::  accuracy(9) !<  Accuracy required to achieve the resolution given as argument. In pixels and degrees.
        ! Private variables
        !real   ::  ppixel_size
        real,   parameter       ::  object_dim_fudge = 1.0
        real                    ::  wl
        logical,    parameter   ::  debug   =   .false.
        logical,    parameter   ::  z_as_xy =   .true.
        real,       parameter   ::  default_target_phase_error = 45.0
        real                    ::  ttarget_phase_error
        ! Start work

        ! Did the user actually supply the wavelength rather than the acceleration voltage?
        if (present(akv_is_wl)) then
            if (akv_is_wl) then
                wl = akv
            else
                wl = akv_to_wl(akv)
            endif
        else
            wl = akv_to_wl(akv)
        endif

        ! What is our target phase error?
        if (present(target_phase_error)) then
            ttarget_phase_error = target_phase_error
        else
            ttarget_phase_error = default_target_phase_error
        endif

        ! x, y
        !accuracy(1:2)      = resolution / 2.
        accuracy(1:2)       = radians(ttarget_phase_error) / ( (1./resolution) * 2. * pi )
        accuracy(1:2)       = accuracy(1:2) / pixel_size

        ! z (defocus)
        if (z_as_xy) then
            if (this_program%IsVerbose() .or. debug)  then
                write(*,'(a)') '**warning(accuracy_from_resolution): z treated as x and y, not as a defocus'
            endif
            accuracy(3)         = radians(ttarget_phase_error) / ( (1./resolution) * 2. * pi )
            accuracy(3)         = accuracy(3) / pixel_size
        else
            accuracy(3)         = radians(ttarget_phase_error) / (pi*wl*(1/resolution)**2)
            accuracy(3)         = accuracy(3) / pixel_size
        endif

        ! Euler angles
        !accuracy(4:6)  = degrees(atan(resolution/(PIXEL_SIZE*2.*REAL(box_size)*OBJECT_DIM_FUDGE)))
        accuracy(4:6)   =   degrees(radians(ttarget_phase_error) / ( (1./resolution) * 2. * pi *OBJ_MAX_RAD))
        !accuracy(4:6)      = degrees(resolution/obj_max_rad)/2.

        ! Subunit_num
        accuracy(7) = 0.1 * 9.0

        ! Helical parameters
        accuracy(8) = accuracy(1) / (box_size / helical_rise)
        accuracy(9) = accuracy(4) / (box_size / helical_rise)


        ! Since each tolerance was worked out independently but they combine in effect, we need to tighten things up
        accuracy = accuracy / 9.0

        !
        ! Print out information for the user
        !
        if (this_program%IsVerbose()) then
            !$omp   single
            write(*,'(/2a)') '**info(accuracy_from_resolution): Below are important parameters used to', &
                                    ' determine the required alignment accuracies'
            write(*,'(2x,a,F8.3,1x,a)')    'Target resolution             = ', resolution, 'A'
            write(*,'(2x,a,F8.3,1x,a)')    'Target phase error            = ', ttarget_phase_error, 'degrees'
            write(*,'(2x,a,i8,  1x,a)')    'Reconstruction box size       = ', box_size, 'pixels'
            write(*,'(2x,a,F8.0,1x,a)')    'Object maximum radius         = ', obj_max_rad, 'A'
            write(*,'(2x,a,F8.0,1x,a)')    'Acceleration voltage          = ', wl_to_akv(wl), 'kV'
            write(*,'(2x,a,F8.3,1x,a)')    'Pixel/voxel size              = ', pixel_size, 'A'
            write(*,'(2x,a,F8.3,1x,a)')    'Required accuracy (x,y)       = ', accuracy(1), 'pixels'
            write(*,'(2x,a,F8.3,1x,a)')    'Required accuracy (z)         = ', accuracy(3), 'pixels'
            write(*,'(2x,a,F8.3,1x,a)')    'Required accuracy (rotations) = ', accuracy(4), 'degrees'

            write(*,'(a)') ' '
            !$omp   end single  nowait
        endif
    end function accuracy_from_resolution

    !>  \brief  Return frequency (1/A) at which fsc curve is still above given threshold
    real function fsc_curve_last_frequency_with_fsc_above_threshold(fsc_curve,threshold) result(freq)
        use Curves
        type(Curve),            intent(in)  ::  fsc_curve
        real,                   intent(in)  ::  threshold
        ! private variables
        integer  :: i
        ! start work
        do i=2,fsc_curve%GetNumberOfDataPoints()
            if (fsc_curve%GetY(i) .le. threshold) then
                freq = fsc_curve%GetX(i-1)
                exit
            endif
        enddo
    end function fsc_curve_last_frequency_with_fsc_above_threshold


    subroutine fsc_curve_write_to_disk(fsc,ssnr,filename)
        use NumericTextFiles
        use Curves
        ! arguments
        type(Curve),            intent(in)  ::  fsc                 !<  FSC curve. x values should be spatial frequencies (1/A). y values should be FSC.
        type(Curve),            intent(in)  ::  ssnr                !<  SSNR curve. x in 1/A
        character(len=*),       intent(in)  ::  filename            !<  Name of file to write to
        ! private variables
        type(NumericTextFile)   ::  my_text_file
        integer                 ::  i
        real, allocatable       ::  current_values(:)
        !
        if (ssnr%HasBeenInitialized()) then
            call my_text_file%Init(filename,open_to_write,3)
            call my_text_file%WriteCommentLine('! Frequency(1/A)         FSC           PSSNR')
            allocate(current_values(3))
            do i=1,fsc%GetNumberOfDataPoints()
                current_values(1) = fsc%GetX(i)
                current_values(2) = fsc%GetY(i)
                current_values(3) = ssnr%GetY(i)
                call my_text_file%WriteDataLineReal(current_values)
            enddo
        else
            call my_text_file%Init(filename,open_to_write,2)
            call my_text_file%WriteCommentLine('! Frequency(1/A)         FSC      ')
            allocate(current_values(2))
            do i=1,fsc%GetNumberOfDataPoints()
                current_values(1) = fsc%GetX(i)
                current_values(2) = fsc%GetY(i)
                call my_text_file%WriteDataLineReal(current_values)
            enddo
        endif

    end subroutine fsc_curve_write_to_disk

    subroutine print_fsc_curve(fsc,filename)
        use Curves
        ! arguments
        type(Curve),            intent(in)  ::  fsc                 !<  fsc curve. x values should be spatial frequencies (1/A). y values should be FSC.
        character(len=*),       intent(in)  ::  filename            !<  Name of file to write to
        ! private variables
        integer                 ::  i
        !real                    ::  current_values(2)
        !

        write(*,'(1x,a6,1x,a23,1x,a6)') '!Shell', 'Spatial frequency (1/A)','FSC'
        do i=1,fsc%GetNumberOfDataPoints()
            write(*,'(1x,i6,1x,f23.3,1x,f6.3)')   i-1, fsc%GetX(i), fsc%GetY(i)
        enddo

    end subroutine print_fsc_curve



    subroutine fsc_curve_read_from_disk(fsc,filename)
        use NumericTextFiles
        use Curves
        ! arguments
        type(Curve),            intent(inout)   ::  fsc                     !<  fsc curve. x values: spatial frequencies (1/A). y values: FSC.
        character(len=*),       intent(in)      ::  filename                !<  Name of file to read from
        ! private variables
        type(NumericTextFile)   ::  my_text_file
        integer                 ::  i
        real                    ::  current_values(2)
        !
        call fsc%ClearData()
        !
        call my_text_file%Init(filename,open_to_read,2)
        do i=1,my_text_file%number_of_data_lines
            call my_text_file%ReadNextDataLine(current_values)
            call fsc%AddPoint(current_values(1),current_values(2))
        enddo

    end subroutine fsc_curve_read_from_disk

    subroutine fsc_to_fsc_curve(fsc,fsc_curve,pixel_size)
        use Curves
        !
        real,               intent(in)      ::  fsc(:)
        type(Curve),        intent(inout)   ::  fsc_curve
        real,               intent(in)      ::  pixel_size
        ! private variables
        real    ::  spatial_frequencies(size(fsc))
        integer ::  i
        real    ::  denom
        ! start work

        ! compute the spatial frequencies
        denom = 1.0 / size(fsc) / (2.0e0*pixel_size)
        do i=1,size(fsc)
            spatial_frequencies(i) = (i-1)*denom
        enddo
        ! Create the curve object
        call fsc_curve%ClearData()
        call fsc_curve%Init(size(fsc))
        call fsc_curve%AddPoints(spatial_frequencies,fsc)
    end subroutine fsc_to_fsc_curve

    !>  \brief  print fsc to terminal
    subroutine print_fsc(fsc, fourier_bin_size)
        implicit none
        real,   allocatable,    intent(in)  ::  fsc(:)
        real,   optional,       intent(in)  ::  fourier_bin_size        !<  Size of each bin (1/A)
        ! private variables
        integer ::  i
        ! start work
        write(*,'(a)') ' '
        if (present(fourier_bin_size)) then
            write(*,'(1x,a6,1x,a23,1x,a6)') '!Shell', 'Spatial frequency (1/A)','FSC'
            do i=1,size(fsc)
                write(*,'(1x,i6,1x,f23.3,1x,f6.3)')   i-1, (i-1)*fourier_bin_size, fsc(i)
            enddo
        else
            write(*,'(a5,2x,3x,a3)') 'Shell', 'FSC'
            do i=1,size(fsc)
                write(*,'(1x,i4,2x,f6.3)')   i-1, fsc(i)
            enddo
        endif
    end subroutine print_fsc

    !>  \brief  Linear cross-correlation coefficient
    real function cclin(img1,img2,rad_min,rad_max,rbfact,fourier_mask)
        use Images
        implicit none
        ! Arguments
        type(image),                intent(in)  ::  img1            !<  Noisy image
        type(image),                intent(in)  ::  img2            !<  Reference image
        real,           optional,   intent(in)  ::  rad_min         !<  Minimum radius (in pixels) at which statistics are collected (low res limit)
        real,           optional,   intent(in)  ::  rad_max         !<  Maximum radius
        real,           optional,   intent(in)  ::  rbfact          !<  B-factor (if not given, assumed to be 0.0)
        type(image),    optional,   intent(in)  ::  fourier_mask    !<  a real image which defines a Fourier-space mask.
        ! Private variables
        real(kind=8)    ::  tmp(3)
        ! Start work
        call  cclin_num_denoms(img1,img2,tmp,rad_min,rad_max,rbfact,fourier_mask)
        ! Cross-correlation coefficient
        if (tmp(2) .eq. 0.0d0 .or. tmp(3) .eq. 0.0d0) then
            cclin = 0.0e0
        else
            cclin = real(tmp(1)/sqrt(tmp(2)*tmp(3)))
        endif
    end function cclin


    !>  \brief  Linear cross-correlation coefficient (returns numerator and denominator components separately) (denominator components are running sums of squares)
	subroutine cclin_num_denoms(img1,img2,num_denoms,rad_min,rad_max,rbfact,fourier_mask)
        use images
        implicit none
        ! Arguments
        type(Image),                intent(in)  ::  img1            !<  Noisy image
        type(Image),                intent(in)  ::  img2            !<  Reference image
        real,           optional,   intent(in)  ::  rad_min         !<  Minimum radius (in pixels) at which statistics are collected (low res limit)
        real,           optional,   intent(in)  ::  rad_max         !<  Maximum radius
        real,           optional,   intent(in)  ::  rbfact          !<  B-factor (if not given, assumed to be 0.0)
        type(image),    optional,   intent(in)  ::  fourier_mask    !<  a real image which defines a Fourier-space mask.
        ! Result
        real(kind=8)     ,intent(out)                   ::  num_denoms(3)
        ! Private variables
        real(kind=8),   allocatable ::  tmp(:,:)
        ! Start work

        tmp = cclin_num_denoms_rings(img1,img2,rad_min,rad_max,rbfact,fourier_mask)


        num_denoms    =   cclin_num_denoms_from_rings(tmp)

	end subroutine cclin_num_denoms

    function cclin_num_denoms_from_rings(rings)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        real(kind=8),   allocatable ::  rings(:,:)      !<  For each spatial frequency (indexed in the second dimension), the numerator and two denominator terms of the cross-correlation
        ! Result
        real(kind=8)                ::  cclin_num_denoms_from_rings(3)
        ! Private variables

        ! Start work
        ! Check allocation of input
        if (.not. allocated(rings)) then
            write(*,'(a)') '**ERROR(cclin_num_denoms_from_rings): input array not allocated'
            call terminate('Fatal error in CCLIN_NUM_DENOMS_FROM_RINGS')
        endif
        ! Check dimension of input
        if (size(rings,1) .ne. 3) then
            write(*,'(a,i0)') '**ERROR(cclin_num_denoms_from_rings): unexpected size of input array in first dimension: ',  &
                                size(rings,1)
            call terminate('Fatal error in CCLIN_NUM_DENOMS_FROM_RINGS')
        endif

        cclin_num_denoms_from_rings(1) = sum(rings(1,:))
        cclin_num_denoms_from_rings(2) = sum(rings(2,:))
        cclin_num_denoms_from_rings(3) = sum(rings(3,:))

    end function cclin_num_denoms_from_rings

    !>  \brief  Faster version of the linear cross-correlation function, with absolute value after 30a, and only data up to specified frequency
    !!
    !!  Note: assumes square 2d images
    pure real function cclin_with_abs_of_low_freqs(image_1,image_2,abs_above_this_freq,ignore_above_this_freq) result(cc)
        use Images
        ! arguments
        type(Image),    intent(in)  ::  image_1                     !<  Noisy image
        type(Image),    intent(in)  ::  image_2                     !<  Reference image
        real,           intent(in)  ::  abs_above_this_freq         !<  Absolute of cross-term will be taken when above this frequency (1/pixel; 0.5 is Nyquist)
        real,           intent(in)  ::  ignore_above_this_freq      !<  Do not include this frequency or above (1/pixel; 0.5 is Nyquist)
        ! private variables
        integer         ::  i,j
        integer         ::  i_max
        integer         ::  ii,jj
        integer         ::  rad_sq
        real            ::  abs_above_this_rad_sq
        real            ::  ignore_above_this_rad_sq
        real(kind=8)    ::  num, denom_1, denom_2
        ! start work

        ! Work out radius of important frequencies
        abs_above_this_rad_sq = (abs_above_this_freq * image_1%logical_dimensions(1))**2
        ignore_above_this_rad_sq = (ignore_above_this_freq * image_1%logical_dimensions(1))**2

        ! initialise results
        num = 0.0d0
        denom_1 = 0.0d0
        denom_2 = 0.0d0

        i_max = image_1%physical_upper_bound_complex(1)

        do j=1,image_1%logical_dimensions(2)
            jj = image_1%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2)**2
            do i=1,i_max
                ii = (i-1)**2
                ! We need to ignore Friedel mates
                if (i .eq. 1 .and. j .gt. 1 .and. j .lt. image_1%physical_index_of_first_negative_frequency(2)) cycle
                !
                rad_sq = ii + jj
                if (rad_sq .ge. ignore_above_this_rad_sq) cycle
                !
                if (rad_sq .ge. abs_above_this_rad_sq) then
                    num = num + abs(dble(image_1%complex_values(i,j,1)*conjg(image_2%complex_values(i,j,1))))
                else
                    num = num +     dble(image_1%complex_values(i,j,1)*conjg(image_2%complex_values(i,j,1)))
                endif
                denom_1 = denom_1 + dble(abs(image_1%complex_values(i,j,1))**2)
                denom_2 = denom_2 + dble(abs(image_2%complex_values(i,j,1))**2)
            enddo
        enddo

        cc = num / sqrt(denom_1*denom_2)

    end function cclin_with_abs_of_low_freqs


    !>  \brief  Linear cross-correlation coefficient (returns numerator and denominator components separately) (denominator components are running sums of squares)
    !!
    !!          Only those voxels which are between the min and max radius and which are non-zero in the Fourier mask
    !!          (if given) are used in the computation
    function cclin_num_denoms_rings(img1,img2,rad_min,rad_max,rbfact,fourier_mask) result (cclin_ndr)
        use runtime_parameters, only : is_even, terminate
        use Images
        implicit none
        ! Arguments
        type(image),                intent(in)  ::  img1            !<  Noisy image
        type(image),                intent(in)  ::  img2            !<  Reference image
        real,           optional,   intent(in)  ::  rad_min         !<  Minimum radius (in pixels) at which statistics are collected (low res limit)
        real,           optional,   intent(in)  ::  rad_max         !<  Maximum radius
        real,           optional,   intent(in)  ::  rbfact          !<  B-factor (if not given, assumed to be 0.0)
        type(image),    optional,   intent(in)  ::  fourier_mask    !<  a real image which defines a Fourier-space mask.
                                                                    !!  This supersedes the min/max radius. Any positions at which the mask is not 0.0
                                                                    !!  will be used in the computation.
        ! Result
        real(kind=8),   allocatable         ::  cclin_ndr(:,:)
        ! Private variables
        integer                     ::  i,j,k, irad
        real                        ::  rad, rrad_min, rrad_max
        real(kind=8)                ::  rrbfact
        logical                     ::  dnf(3)
        integer                     ::  ierr
        real,       allocatable     ::  wgt(:)
        logical,    parameter       ::  iignore_friedel_mates   =   .true.  ! Some Friedel mates are described explicelty by the fftw array, we should not count them twice!
        integer                     ::  x,y,z
        logical                     ::  include_this_voxel
        ! Start work

#ifndef skip_runtime_checks
        ! Check both images are complex
        if (.not. (img1%IsInSameSpaceAs(img2) .and. .not. img1%is_in_real_space)) then
            call terminate('Fatal error in CCLIN_NUM_DENOMS_RINGS')
        endif
        ! Check images have same size
        if (.not. img1%HasSameDimensionsAs(img2)) then
            write(*,'(a,2(3(i0,1x),1x))') '**ERROR(cclin_num_denoms_rings): images not same dimensions. ', &
                                            img1%logical_dimensions, img2%logical_dimensions
            call terminate('Fatal error in CCLIN_NUM_DENOMS_RINGS')
        endif
        ! Hardcoded for even dimension
        if (.not. img1%HasEvenDimensions()) then
            write(*,'(a,3(i0,1x))') '**ERROR(cclin_num_denoms_rings): input images dimensions are not even: ', &
                                            img1%logical_dimensions
            call terminate('Fatal error in CCLIN_NUM_DENOMS_RINGS')
        endif
        ! Hardcoded for cubic images
        if (.not. img1%IsSquare()) then
            call terminate('Hardcoded for square/cubic images')
        endif
#endif

        ! Low & high res. limits
        rrad_min = 1.0
        if (present(rad_min)) rrad_min = rad_min
        rrad_min = max(1.0,rrad_min)

        rrad_max = real(img1%logical_dimensions(1))/2.0
        if (present(rad_max)) rrad_max = rad_max
        rrad_max = min(rrad_max,real(img1%logical_dimensions(1))/2.)

        ! B-factor
        rrbfact = 0.0D0
        if (present(rbfact)) rrbfact = dble(rbfact)

        ! Allocate memory for results
        allocate(cclin_ndr(3,img1%logical_dimensions(1)/2),wgt(img1%logical_dimensions(1)/2),stat=IERR)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(cclin_num_denoms_rings): memory allocation for results failed with error ', ierr
            call terminate('Fatal error in CCLIN_NUM_DENOMS_RINGS')
        endif

        ! Initialise
        cclin_ndr = 0.0D0

        ! Calculate the B-factor values to be applied at each radius
        if (rrbfact .eq. 0.) then
            wgt = 1.0
        else
            do I=1,img1%logical_dimensions(1)/2
                wgt(i) = exp(-RRBFACT*(I-1)**2)
            enddo
        endif

        ! Loop to collect cross-correlation info
        cclin_ndr(1,:) = 0.0D0
        cclin_ndr(2,:) = 0.0D0
        cclin_ndr(3,:) = 0.0D0
        dnf = [.true.,.false.,.false.]
        do I=1,img1%physical_upper_bound_complex(1)
            x = (I-1)**2
            do J=1,img1%logical_dimensions(2)
                if (iignore_friedel_mates .and. i == 1 .and. j .gt. 1 .and. &
                    j .lt. img1%physical_index_of_first_negative_frequency(2)) cycle    ! Skip voxels whose Friedel mates are explicit in the fftw description
                y = img1%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2)**2
                do K=1,img1%logical_dimensions(3)
                    if (iignore_friedel_mates .and. i == 1 .and. j == 1 .and. k .gt. 1 .and.    &
                        k .lt. img1%physical_index_of_first_negative_frequency(3)) cycle
                    z = img1%LogicalIndexGivenPhysicalIndexInFourierSpace(k,3)**2

                    rad = sqrt(real(X+Y+Z))
                    irad = nint(rad)

                    if (present(fourier_mask)) then
                        include_this_voxel = fourier_mask%real_values(i,j,k) .ne. 0.0
                    else
                        include_this_voxel = rad .ge. rrad_min .and. rad .lt. rrad_max
                    endif

                    if (include_this_voxel) then
                        cclin_ndr(1,irad)    =   cclin_ndr(1,irad) + &
                                                 dble(img1%complex_values(i,j,k)*CONJG(img2%complex_values(i,j,k))*WGT(irad))
                        cclin_ndr(2,irad)    =   cclin_ndr(2,irad) + dble(abs(img1%complex_values(i,j,k))**2)
                        cclin_ndr(3,irad)    =   cclin_ndr(3,irad) + dble((abs(img2%complex_values(i,j,k))*WGT(irad))**2)
                    endif
                enddo
            enddo
        enddo

    end function cclin_num_denoms_rings

!   !>  \brief  Correlation coefficient calculation based on Stewart & Grigorieff 2004
!   real function cc3_old(img1,img2,rad_min,rad_max,rbfact,absolute)
!       !use images
!       implicit none
!       ! Arguments
!       type(image),            intent(in)  ::  img1        !<  Noisy image
!       type(image),            intent(in)  ::  img2        !<  Reference image
!       integer,    optional,   intent(in)  ::  rad_min     !<  Minimum radius (in pixels) at which statistics are collected (low res limit)
!       integer,    optional,   intent(in)  ::  rad_max     !<  Maximum radius
!       real,       optional,   intent(in)  ::  rbfact      !<  B-factor (if not given, assumed to be 0.0)
!       logical,    optional,   intent(in)  ::  absolute    !<  Whether to add the absolute value of each shell (turns cc3 into cc3m). Default to FALSE.
!       ! Private variables
!       integer ::  rad, i,j,k, rrad_min, rrad_max
!       logical ::  aabsolute
!       real(KIND=8)    ::  count(img1%logical_dimensions(1)/2+1) ! Count of voxels in shell
!       real(KIND=8)    ::  cp(img1%logical_dimensions(1)/2+1) ! Cross-product
!       real(KIND=8)    ::  sq1(img1%logical_dimensions(1)/2+1),sq2(img1%logical_dimensions(1)/2+1)   ! Sum of squares
!       real(KIND=8)    ::  w   ! Inverse of box dimension
!       real(KIND=8)    ::  r
!       real(KIND=8)    ::  sdpha, snpha
!       real(KIND=8)    ::  wgt ! weight
!       real(KIND=8)    ::  rrbfact ! B-factor (should be an argument)
!       logical ::  dnf(3)
!       ! Start work
!       ! Check both images are complex
!       if (.NOT. (same_type(img1,img2) .AND. img1%type == 'COMP')) then
!           write(*,'(3(A,x))') '**ERROR(cc3): images need to be complex.', img1%type, img2%type
!           stop 'Fatal error in CC3'
!       endif
!       ! Check images have same size
!       if (.NOT. same_dim(img1,img2)) then
!           write(*,'(a,2(3(i3,x),x))') '**ERROR(cc3): images not same dimensions. ', img1%logical_dimensions, img2%logical_dimensions
!           stop 'Fatal error in CC3'
!       endif
!       ! B-factor
!       if (present(rbfact)) then
!           rrbfact = dble(rbfact)
!       else
!           rrbfact = 0.0D0
!       endif
!       ! Low & high res. limits
!       if (present(rad_min)) then
!           rrad_min = rad_min
!       else
!           rrad_min = 1
!       endif
!       if (present(rad_max)) then
!           rrad_max = rad_max
!       else
!           rrad_max = img1%logical_dimensions(1)/2
!       endif
!
!       if(present(absolute)) then
!           aabsolute = absolute
!       else
!           aabsolute = .FALSE.
!       endif
!
!       ! Loop to collect cross-correlation info
!       !print *, '**debug(cc3): rrad_min, rrad_max = ', rrad_min, rrad_max
!       cp = 0.0D0
!       sq1 = 0.0D0
!       sq2 = 0.0D0
!       count = 0.0D0
!       dnf = [.TRUE.,.FALSE.,.FALSE.]
!       do K=1,img1%logical_dimensions(3)
!           do J=1,img1%logical_dimensions(2)
!               do I=1,img1%logical_dimensions(1)/2+1
!                   rad = nint(radius(i,j,k,img1,dnf))
!                   if (rad .GE. rrad_min .AND. rad .LT. rrad_max) then ! Maybe the check for the radius should be done in REALs, not INTGs?
!                       cp(rad)  =  cp(rad) + dble(img1%complex_values(i,j,k)*CONJG(img2%complex_values(i,j,k)))
!                       sq1(rad) = sq1(rad) + dble(abs(img1%complex_values(i,j,k))**2)
!                       sq2(rad) = sq2(rad) + dble(abs(img2%complex_values(i,j,k))**2)
!                       count(rad) = count(rad) + 1.0
!                   endif
!               enddo
!           enddo
!       enddo
!
!       ! Debug
!       !print *, '**debug(cc3_old): numerator, denom1, denom2 per shell before cubing:'
!       !do RAD=1,img1%logical_dimensions(1)/2
!       !   print *, rad, cp(rad), sq1(rad), sq2(rad)
!       !enddo
!
!       ! Loop over the shells
!       w = 1.0D0/dble(img1%logical_dimensions(1))     ! Asumption images are square/cube
!       sdpha = 0.0D0
!       snpha = 0.0D0
!       do RAD=2,img1%logical_dimensions(1)/2
!           if (count(rad) .NE. 0.0) then
!               !print '(a,4F12.3)', '**debug(cc3): cp, sq1, sq2 = ', cp(rad), sq1(rad), sq2(rad), count(rad)
!               cp(rad) = cp(rad)/sqrt(sq1(rad)*SQ2(rad))
!               !print *, '**debug(cc3): cp for shell ', rad, ' = ', cp(rad)
!               if (aabsolute) then
!                   r = cp(rad)**2/(W+ABS(cp(rad))) ! Eq. 16 in Stewart & Grigorieff
!               else
!                   r = cp(rad)*ABS(cp(rad))/(W+ABS(cp(rad))) ! Eq. 16 in Stewart & Grigorieff
!               endif
!               wgt = exp(-RRBFACT*(RAD-1)**2)  ! B-factor weighting
!               sdpha = sdpha + R**3 * wgt  ! signal
!               snpha = snpha + wgt         ! noise
!           endif
!       enddo
!       !print *, '**debug(cc3): count of voxels = ', sum(count), (img1%logical_dimensions(1)/2+1)*IMG1%logical_dimensions(2)*IMG1%logical_dimensions(3)
!       !
!       if (sdpha .NE. 0.0D0) then
!           sdpha = exp(log(abs(sdpha)/snpha)/3.0D0)*SDPHA/abs(sdpha) ! ?
!       endif
!       if (sdpha .GT.  1.0D0) sdpha =  1.0D0
!       if (sdpha .LT. -1.0D0) sdpha = -1.0D0
!       cc3_old = real(sdpha)
!   end function cc3_old

    !>  \brief  Correlation coefficient calculation based on Stewart & Grigorieff 2004
    real function cc3(img1,img2,rad_min,rad_max,rbfact,absolute,fourier_mask)
        use images
        implicit none
        ! Arguments
        type(image),                intent(in)  ::  img1        !<  Noisy image
        type(image),                intent(in)  ::  img2        !<  Reference image
        real,           optional,   intent(in)  ::  rad_min     !<  Minimum radius (in pixels) at which statistics are collected (low res limit)
        real,           optional,   intent(in)  ::  rad_max     !<  Maximum radius
        real,           optional,   intent(in)  ::  rbfact      !<  B-factor (if not given, assumed to be 0.0)
        logical,        optional,   intent(in)  ::  absolute    !<  Whether to add the absolute value of each shell (turns cc3 into cc3m). Default to FALSE.
        type(image),    optional,   intent(in)  ::  fourier_mask    !<  a real image which defines a Fourier-space mask.
        ! Private variables
        real(kind=8),       allocatable     ::  tmp(:,:)
        !logical,   parameter               ::  sanity_check    =   .FALSE.
        !real                               ::  cc3_check
        ! Start work

        tmp = cclin_num_denoms_rings(img1,img2,rad_min,rad_max,RBFACT=0.0,FOURIER_MASK=FOURIER_MASK)

        cc3 = cc3_finalise(tmp,img1,img2,rad_min,rad_max,rbfact,absolute)

        !if (sanity_check) then
        !   cc3_check = cc3_old(img1,img2,rad_min,rad_max,rbfact,absolute)
        !
        !   if (cc3 .NE. cc3_check) then
        !       write(*,'(a,2(F12.6,x))') '**ERROR(cc3): sanity check failed: cc3, cc3_old = ', cc3, cc3_check
        !       call terminate('Fatal error in CC3')
        !   endif
        !endif

    end function cc3

    !>  \brief  Given accumulated sums of crossproducts (numerator) and squares (denominators),
    !!          return the cc3 score
    function cc3_finalise(num_denoms_rings,img1,img2,rad_min,rad_max,rbfact,absolute) result(cc3_final)
        use Images
        implicit none
        ! Arguments
        type(image),                intent(in)  ::  img1                    !<  Noisy image
        type(image),                intent(in)  ::  img2                    !<  Reference image
        real(kind=8),allocatable,   intent(in)  ::  num_denoms_rings(:,:)
        real,       optional,       intent(in)  ::  rad_min                 !<  Minimum radius (in pixels) at which statistics are collected (low res limit)
        real,       optional,       intent(in)  ::  rad_max                 !<  Maximum radius
        real,       optional,       intent(in)  ::  rbfact                  !<  B-factor (if not given, assumed to be 0.0)
        logical,    optional,       intent(in)  ::  absolute                !<  Whether to add the absolute value of each shell (turns cc3 into cc3m). Default to FALSE.
        ! Result
        real(kind=8)                            ::  cc3_final
        ! Private variables
        real(kind=8)            ::  tmp, w, r
        real                    ::  wgt, rrbfact
        integer                 ::  rad
        real                    ::  rrad_min, rrad_max
        real(kind=8)            ::  num, denom
        logical                 ::  aabsolute
        logical,    parameter   ::  debug   =   .false.

        ! Start work

        !
        ! Deal with optional arguments
        !

        ! B-factor
        if (present(rbfact)) then
            rrbfact = dble(rbfact)
        else
            rrbfact = 0.0D0
        endif

        ! Low & high res. limits
        rrad_min = 2.0
        if (present(rad_min)) rrad_min = rad_min
        rrad_max = img1%logical_dimensions(1)/2
        if (present(rad_max)) rrad_max = rad_max

        aabsolute = .false.
        if(present(absolute)) aabsolute = absolute


        ! This factor is 1/N in the 2d case, where NxN are the dimensions of the image. See Eq. 14 in Stewart & Grigorieff (2004).
        ! It is actually the standard deviation of the correlation of a pure noise NxN image with a reference image.
        w = 1.0D0 / dble(sqrt(real(img1%logical_dimensions(1)*IMG1%logical_dimensions(2)*IMG1%logical_dimensions(3))))


        ! Initialise
        num = 0.0D0
        denom = 0.0D0

        ! Debug
        !print *, '**debug(cc3_finalise): numerator per shell before cubing:'
        !do RAD=1,img1%logical_dimensions(1)/2
        !   print *, rad, num_denoms_rings(1,rad), num_denoms_rings(2,rad), num_denoms_rings(3,rad)
        !enddo

        ! Loop over rings/shells
        do RAD=NINT(rrad_min),nint(rrad_max)-1
            wgt = exp(-RRBFACT*(RAD-1)**2)
            tmp = num_denoms_rings(1,rad)/sqrt(num_denoms_rings(2,rad)*NUM_DENOMS_RINGS(3,RAD))
            if (aabsolute) then
                r = TMP**2/(W+ABS(tmp)) ! Eq. 16 in Stewart & Grigorieff
            else
                r = TMP*ABS(tmp)/(W+ABS(tmp)) ! Eq. 16 in Stewart & Grigorieff
            endif
            if (debug) then
                write(*,'(a,i0,1x,3(G15.3,1x))') '**debug(cc3_finalise): rad, tmp, r, num_denoms_rings(3) = ', &
                                                rad, tmp, r, num_denoms_rings(3,rad)
            endif
            ! Accumulate sums for the averaging over rings/shells
            wgt = exp(-RRBFACT*(RAD-1)**2)  ! B-factor weighting
            num = num + R**3 * wgt
            denom = denom + wgt
        enddo

        ! Average
        cc3_final = num / denom

        ! Cubic root
        !cc3_final = CC3_FINAL**(1./3.) ! This is not defined for negative numbers!
        cc3_final = exp(log(abs(num)/denom)/3.0D0)*NUM/abs(num)

        ! In case we went beyond [-1.0,1.0]
        if (cc3_final .lt. -1.0)  cc3_final = -1.0
        if (cc3_final .gt. 1.0)   cc3_final = 1.0
    end function cc3_finalise

    !>  \brief  Compute the real-space correlation between two images over a mask
    real function correlation(img1,img2,mask,normalise)
        use images
        implicit none
        ! Arguments
        type(image),    intent(inout)   ::  img1, img2, mask
        logical ,       intent(in)  ::  normalise
        ! Private variables
        integer             ::  i,j,k,num_of_voxels
        double  precision   ::  cross_product, sum_of_squares1, sum_of_squares2, sum1, sum2
        logical             ::  do_fts
        ! Start work

        !
        ! Work out whether we're already in real space
        !
        do_fts = .false.
        if (.not. img1%is_in_real_space) then
            call img1%BackwardFFT()
            call img2%BackwardFFT()
            do_fts = .true.
        endif


        num_of_voxels = 0
        cross_product = 0.0
        sum_of_squares1 = 0.0
        sum_of_squares2 = 0.0
        sum1 = 0.0
        sum2 = 0.0
        do K=1,min(img1%logical_dimensions(3),IMG2%logical_dimensions(3))
            do J=1,min(img1%logical_dimensions(2),IMG2%logical_dimensions(2))
                do I=1,min(img1%logical_dimensions(1),IMG2%logical_dimensions(1))
                    if (mask%real_values(i,j,k) .ne. 0.0) then
                        num_of_voxels   = num_of_voxels     +   1
                        cross_product   = cross_product     +   img1%real_values(i,j,k) * img2%real_values(i,j,k)
                        sum_of_squares1 = sum_of_squares1   +   img1%real_values(i,j,k) **2
                        sum_of_squares2 = sum_of_squares2   +   img2%real_values(i,j,k) **2
                        sum1            = sum1              +   img1%real_values(i,j,k)
                        sum2            = sum2              +   img2%real_values(i,j,k)
                    endif
                enddo
            enddo
        enddo
        cross_product   = cross_product     /   num_of_voxels
        sum_of_squares1 = sum_of_squares1   /   num_of_voxels
        sum_of_squares2 = sum_of_squares2   /   num_of_voxels
        sum1            = sum1              /   num_of_voxels
        sum2            = sum2              /   num_of_voxels

        ! The result
        if (normalise) then
            correlation = (cross_product - SUM1*SUM2) / (sqrt(SUM_OF_SQUARES1-SUM1**2)*SQRT(SUM_OF_SQUARES2-SUM2**2))
        else
            correlation = (cross_product - SUM1*SUM2)
        endif

        ! Images back to reciprocal space if that's how they were supplied
        if (do_fts) then
            call img1%ForwardFFT()
            call img2%ForwardFFT()
        endif

    end function correlation

end module images_comparison
