!>  \brief  Methods to perform operations on images which require helical thinking
module images_helices

    use images
    use helices

    implicit none

    contains

    subroutine images_helices_unit_tests()
        implicit none
        call images_helix_unit_test_1()
        call images_helix_unit_test_2()
        call images_helix_unit_test_3()
    end subroutine images_helices_unit_tests

    subroutine images_helix_unit_test_1()
        implicit none
        ! Private variables
        type(helical_symmetry)  ::  helix
        type(image)             ::  self
        real, allocatable       ::  phi(:),z(:)
        integer                 ::  i
        ! Start work
        self%logical_dimensions = [64,64,64]
        call helical_symmetry_init(helix,1.403/1.163,-22.03,90.0/1.163)
        call volume_list_h_subunits_polar(self%logical_dimensions(3),helix,phi,z)
        do I=1,size(phi)
            write(*,'(i3,1x,2(F7.2,1x))') i, phi(i), z(i)
        enddo
    end subroutine images_helix_unit_test_1

    subroutine images_helix_unit_test_2()
        use images
        implicit none
        ! Private variables
        type(image)             ::  self, self_ampl, self_phases
        type(helical_symmetry)  ::  helix
        real,   parameter       ::  psize = 2.326
        ! Start work
        write(*,'(a)') '**info(images_helix_unit_test_2): entering'
        call helical_symmetry_init(helix,1.403/psize,-22.03,90.0/psize)
        self%logical_dimensions = [200,200,1]
        write(*,'(a)') '**info(images_helix_unit_test_2): simulating layer lines'
        call image_simulate_layer_lines(self,helix,R_IN=10.0/psize,R_OUT=105.0/psize)
        !call image_simulate_layer_lines(self,helix,R_IN=49.0,R_OUT=51.0)

        ! Dump the amplitudes of the layer lines
        call self%ComputeAmplitudeSpectrum(self_ampl)
        call self_ampl%WriteToDisk('dbg_ampl.mrc')

        ! Dump the phases
        call self%ComputePhaseSpectrum(self_phases)
        call self_phases%WriteToDisk('dbg_phases.mrc')

        ! Dump the back ft, which should look like a projection of a helix
        call self%SwapRealSpaceQuadrants()
        call self%BackwardFFT()
        call self%WriteToDisk('dbg_real_space.mrc')
        write(*,'(a)') '**info(images_helix_unit_test_2): done'
    end subroutine images_helix_unit_test_2

    subroutine images_helix_unit_test_3
        use Images
        use ImageNoise
        use runtime_parameters, only : terminate
        implicit none
        ! private variables
        type(image)     ::  self
        type(helical_symmetry)  ::  helix
        real                    ::  crossover_angle
        logical,    parameter   ::  test_with_rectangle = .true.
        ! start work
        call self%Allocate(dims=[64,64,1])
        if (test_with_rectangle) then
            self = 1.0e0
            call self%ApplySoftRectangularMask([45.0,15.0],0.0,2.0)
        else
            self = 0.0e0
            self%real_values(:,self%physical_address_of_box_center(2),1) = 1.0
        endif
        call AddGaussianNoise(self,0.2e0)
        call self%WriteToDisk('dbg_test.mrc')
        helix%axial_symmetry = 2
        call find_angle_of_crossover(self,helix,crossover_angle)
        write(*,'(a, f0.2)') '**debug(images_helix_unit_test_3): crossover_angle = ', crossover_angle
        if (abs(crossover_angle) .gt. 0.5) call terminate('Failed images_helix_unit_test_3')
    end subroutine images_helix_unit_test_3


    !>  \brief  Simulate diffraction pattern of a helical lattice
    !!  \todo   Add linear interpolation to get better representation of layer lines that fall between pixels
    subroutine image_simulate_layer_lines(self,helix,r_in,r_out)
#ifdef __INTEL_COMPILER
        use ifport
#endif
        use runtime_parameters, only : pi, terminate
        use helices, only : helix_find_layer_lines
        implicit none
        ! Arguments
        type(image),                    intent(inout)   ::  self        !<  Two-dimensional image
        type(helical_symmetry),         intent(in)      ::  helix       !<  Helical symmetry parameters
        real,                           intent(in)      ::  r_in        !<  Inner radius of filament (pixels)
        real,                           intent(in)      ::  r_out       !<  Outer radius of filament (pixels)
        ! Private variables
        integer,    allocatable ::  ns(:)
        real,       allocatable ::  bigzs(:), smallzs(:), phis(:)
        real,       parameter   ::  axial_rot   =   0.0 !<  Extra rotation of the helix around its axis
        real                    ::  r_mid
        integer                 ::  i,ii,logi(3), phys(3)
        complex                 ::  current_value, current_value_r
        real                    ::  bigpsi, bigz, bigr, smallr
        integer                 ::  current_layer_line, n, current_h_sub
        integer                 ::  z_sign
        logical,    parameter   ::  debug   =   .true.
        real,       parameter   ::  debug_psize = 2.326
        ! Start work

#ifdef NAGFOR_COMPILER
        write(*,'(a)') '**WARNING(image_simulate_layer_lines): Bessel functions are not supported by nagfor. Will return blank.'
#endif
#ifdef PG_COMPILER
        write(*,'(a)') '**WARNING(image_simulate_layer_lines): Bessel functions are not supported by pgfortran. Will return blank.'
#endif

        ! Make sure the image is set up ok
        if (self%IsAVolume()) then
            call self%Deallocate()
            self%logical_dimensions = [SELF%logical_dimensions(1),SELF%logical_dimensions(2),1]
        endif
        if (.not. self%IsAllocated()) call self%Allocate()
        self%real_values = 0.0
        self%is_in_real_space   = .false.

        ! Check the helical parameters are in pixel units
        if (.not. helix%pixel_units) then
            write(*,'(a)') '**ERROR(image_simulate_layer_lines): helical parameters not given in pixel units'
            call terminate('Fatal error in IMAGE_SIMULATE_LAYER_LINES')
        endif

        ! Get the positions and Bessel order of the layer lines
        call helix_find_layer_lines(helix,bigzs,ns)

        if (debug) then
            do I=1,size(ns)
                if (abs(bigzs(i)) .le. 0.5) then
                if (bigzs(i) .ne. 0.) then
                    write(*,'(i6,1x,2(F7.2,1x),i9)') i, bigzs(i)/debug_psize, 1./bigzs(i)*DEBUG_PSIZE, ns(i)
                else
                    write(*,'(i6,1x,2(F7.2,1x),i9)') i, bigzs(i)/debug_psize,                    999., ns(i)
                endif
                endif
            enddo
        endif

        ! Get the positions of subunits that fit into the volume
        call volume_list_h_subunits_polar(self%logical_dimensions(1),HELIX,phis,smallzs)
        ! Convert phi coordinates to radians
        do I=1,size(phis)
            phis(i) = phis(i)/180.0 * pi
        enddo

        ! Mean radius
        r_mid = (R_OUT-R_IN)*0.5

        !
        bigpsi = axial_rot/180.*PI

        ! Loop over the layer lines
        do current_layer_line=1,size(bigzs)
            bigz=bigzs(current_layer_line)
            if (abs(bigz) .lt. 0.5) then
                ! loop over the sign of bigz
                do z_sign=1,1
                    if (z_sign .eq. 2) bigz=bigz*(-1.0)
                    n=ns(current_layer_line)
                    ! loop over the reciprocal pixels along the current layer line
                    do i=1,self%physical_upper_bound_complex(1)
                        current_value=(0.0,0.0)
                        ii=i-1
                        bigr=real(ii)/real(self%logical_dimensions(1))
                        ! loop over real-space radius, small r
                        do smallr=ceiling(r_in),floor(r_out)
                            current_value_r = (0.0,0.0)
                            ! loop over helical subunits
                            do current_h_sub=1,size(phis)
                                current_value_r = current_value_r &
                                                + exp(cmplx(0.0,2.0*pi*bigz*smallzs(current_h_sub)-real(n)*phis(current_h_sub)))
                            enddo
                            !print *, n, 2.0*PI*REAL(smallr)*BIGR, smallr, current_value_r
                            !call flush(6)
! nag doesn't implement Bessel functions yet, even though it's f2008 standard
#ifdef NAGFOR_COMPILER
                            current_value_r = 0.0
#else
#ifdef PG_COMPILER
                            current_value_r = 0.0
#else
                            current_value_r = current_value_r * real(smallr) * &
#ifdef __INTEL_COMPILER
                                                BesJN(  &
#else
                                                ! The Fortran 2008 intrinsic
                                                bessel_jn( &
#endif
                                                n,2.0*pi*real(smallr)*bigr)
#endif
#endif
                            current_value = current_value + current_value_r
                        enddo ! End of loop over small r
                        current_value = current_value * exp(cmplx(0.0,real(n)*(BIGPSI+PI*0.5)))
                        ! Now that we have the value for the layer line at this point, we need to add
                        ! it to the image, but since the layer line probably falls between pixels, we need to interpolate
                        ! As an initial test, we will cheat and do nearest-neighbour
                        logi = [ii,nint(bigz*self%logical_dimensions(2)),1]
                        call self%PhysicalAddressGivenLogicalAddressInFourierSpace(logi,phys)
                        self%complex_values(phys(1),phys(2),phys(3)) = current_value
                    enddo ! End of loop over pixels along the layer line
                enddo ! End of loop over bigZ sign
            endif ! End of check for z < Nyquist
        enddo ! End of loop over layer lines

    end subroutine image_simulate_layer_lines

    !>  \brief  List the polar coordinates of the helical subunits within a volume
    subroutine volume_list_h_subunits_polar(z_dim,helix,phi,z)
        use euleranglesmod, only : degrees_m180_to_180
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        integer,                        intent(in)      ::  z_dim       !<  Number of pixels
        type(helical_symmetry),         intent(in)      ::  helix       !<  Helical symmetry operator
        real,           allocatable,    intent(inout)   ::  phi(:)      !<  List of phi values (degrees)
        real,           allocatable,    intent(inout)   ::  z(:)        !<  List of z values (pixels)
        ! Private variables
        integer ::  ori(3), num_subunits, num_subunits_above, num_subunits_below,i,current_subunit
        ! Start work

        if (.not. helix%pixel_units) call terminate('Need helix in pixel units')

        ! Work out the number of helical subunits that can fit in the volume
        ori = IndexOfCentralPixelGivenLogicalDimension(z_dim)
        num_subunits_above = floor((Z_DIM-ORI(3))/HELIX%rise)
        num_subunits_below = floor((ori(3)-1)/helix%rise)
        num_subunits = 1 + num_subunits_above + num_subunits_below
        write(*,'(a,i0)') '**debug(volume_list_h_subunits_polar): expected num subunits = ', num_subunits

        ! Allocate memory for output arrays
        if (allocated(phi)) deallocate(phi)
        if (allocated(z)) deallocate(z)
        allocate(phi(num_subunits),z(num_subunits))

        ! Loop over the subunits
        do I=-NUM_SUBUNITS_BELOW,num_subunits_above
            CURRENT_SUBUNIT=I+NUM_SUBUNITS_BELOW+1
            phi(current_subunit) = degrees_m180_to_180(real(i)*HELIX%twist)
            z(current_subunit)   =                     real(i)*HELIX%rise
        enddo

    end subroutine volume_list_h_subunits_polar

    !>  \brief  Look for the projection angle at which a crossover would be observed
    subroutine find_angle_of_crossover(self,helix,crossover_angle)
        use rotation_matrices, only : radians
        implicit none
        ! arguments
        type(image),                intent(in)      ::  self            !<  3d reconstruction of a filament
        type(helical_symmetry),     intent(in)      ::  helix           !<  helical symmetry of the filament
        real,                       intent(out)     ::  crossover_angle !<  the angle at which projecting the 3d gives the maximum grey value
        !
        integer                 ::  current_angle
        real                    ::  max_projected_value, best_max_projected_value
        integer                 ::  z_mid
        real                    ::  coo_in(3), coo_out(3)
        real                    ::  current_angle_rad
        real,       allocatable ::  projected_line(:)
        !integer                 ::  i,j
        integer                 ::  current_fwhm, minimum_fwhm
        integer                 ::  first_angle_at_minimum, last_angle_at_minimum
        logical,    parameter   ::  debug    = .true.
        ! start work

        z_mid = self%physical_address_of_box_center(3)
        coo_in = 0.0e0
        coo_out = 0.0e0

        !allocate(projected_line(-maxval(self%logical_dimensions):maxval(self%logical_dimensions)))

        best_max_projected_value = -huge(1.0e0)
        crossover_angle = 0.0e0
        minimum_fwhm = huge(1)

        ! search over possible angles
        do current_angle=0,ceiling(360.0e0/helix%axial_symmetry)
            current_angle_rad = radians(real(current_angle))
            call compute_1d_projection(self,current_angle_rad,projected_line)
            ! find the maximum value projected from the central slice of the 3d
            max_projected_value = maxval(projected_line)
            if (max_projected_value .gt. best_max_projected_value) then
                best_max_projected_value = max_projected_value
                crossover_angle = real(current_angle) - 90.0e0
            endif
        enddo

        if (debug) then
            print *, '**debug(find_angle_of_crossover): max val proj = ', best_max_projected_value
            print *, '**debug(find_angle_of_crossover): according to max val proj, crossover angle = ', crossover_angle
        endif

        ! Loop again over all possible angles. This time, characterise the width of the projection.
        first_angle_at_minimum = -1
        do current_angle=0,ceiling(360.0e0/helix%axial_symmetry)
            current_angle_rad = radians(real(current_angle))
            call compute_1d_projection(self,current_angle_rad,projected_line)
            !print *, '**debug(find_angle_of_crossover): projected_line = ', projected_line
            call compute_fwhm(projected_line,current_fwhm,max_override=best_max_projected_value,fraction_of_peak_override=0.1)!)
            print *, '**debug(find_angle_of_crossover): current_angle, current_fwhm = ', current_angle, current_fwhm
            if (current_fwhm .lt. minimum_fwhm) then
                minimum_fwhm = current_fwhm
                first_angle_at_minimum = current_angle
            else if (current_fwhm .eq. minimum_fwhm .and. first_angle_at_minimum .gt. 0) then
                last_angle_at_minimum = current_angle
            endif
        enddo

        if (debug) then
            print *, '**debug(find_angle_of_crossover): first, last angle at minimum = ', &
                                        first_angle_at_minimum, last_angle_at_minimum
        endif

        crossover_angle = real(last_angle_at_minimum - first_angle_at_minimum)*0.5e0 + real(first_angle_at_minimum) - 90.0

        if (debug) then
            print *, '**debug(find_angle_of_crossover):                      final crossover angle = ', crossover_angle
        endif

    end subroutine find_angle_of_crossover

    subroutine compute_1d_projection(self,projection_angle_rad,projected_line)
        implicit none
        ! arguments
        type(image),                intent(in)          ::  self
        real,                       intent(in)          ::  projection_angle_rad
        real,       allocatable,    intent(inout)       ::  projected_line(:)
        ! private variables
        integer     ::  z_mid, i, j
        real        ::  coo_in(3), coo_out(3)
        ! start work
        if (allocated(projected_line)) deallocate(projected_line)
        allocate(projected_line(-maxval(self%logical_dimensions):maxval(self%logical_dimensions)))
        z_mid = self%physical_address_of_box_center(3)
        projected_line = 0.0e0
        ! loop over pixels in the central (xy) slice
        do j=1,self%logical_dimensions(2)
            coo_in(2) = j - self%physical_address_of_box_center(2)
            do i=1,self%logical_dimensions(1)
                coo_in(1) = i - self%physical_address_of_box_center(1)
                !
                coo_out(1) =  coo_in(1) * cos(projection_angle_rad) + coo_in(2) * sin(projection_angle_rad)
                !coo_out(2) = -coo_in(1) * sin(projection_angle_rad) + coo_in(2) * cos(projection_angle_rad) ! we don't need this
                !
                projected_line(nint(coo_out(1))) = projected_line(nint(coo_out(1))) + self%real_values(i,j,z_mid)
            enddo
        enddo
    end subroutine compute_1d_projection

    subroutine compute_fwhm(projected_line,fwhm, max_override, fraction_of_peak_override)
        use runtime_parameters, only : terminate
        implicit none
        ! arguments
        real,               intent(in)  ::  projected_line(:)
        integer,            intent(out) ::  fwhm
        real,   optional,   intent(in)  ::  max_override       !<  Use this value instead of the max
        real,   optional,   intent(in)  ::  fraction_of_peak_override
        ! private variables
        real            ::  minimum, maximum, half_maximum
        integer         ::  current_index
        logical         ::  above_half_maximum
        integer         ::  mid_index
        logical, parameter  ::  debug   =   .false.
        ! start work
        minimum = minval(projected_line)
        maximum = maxval(projected_line)
        if (present(max_override)) maximum = max_override
        half_maximum = minimum + (maximum-minimum)*0.5e0
        if (present(fraction_of_peak_override)) half_maximum = minimum + (maximum-minimum)*fraction_of_peak_override
        if (debug) print *, '**debug(compute_fwhm): min, max, halfmax = ', minimum, maximum, half_maximum
        !
        if (maxval(projected_line) .lt. half_maximum) then
            write(*,'(a)') '**error(compute_fwhm): no value in projected line are above half max'
            call terminate('fatal error in compute_fwhm')
        endif
        fwhm = huge(1)
        above_half_maximum = .false.
        ! loop over the line from the origin of the projection towards the edge
        mid_index = (size(projected_line)-1)/2
        do current_index=mid_index,size(projected_line)
            if (debug) print *, '**debug(compute_fwhm): current value = ', projected_line(current_index)
            if (projected_line(current_index) .gt. half_maximum) above_half_maximum = .true.
            if (above_half_maximum .and. projected_line(current_index) .le. half_maximum .and. current_index .gt. mid_index+2) then
                fwhm = 2*(current_index-mid_index)
                exit
            endif
        enddo

    end subroutine compute_fwhm

    !>  \brief  Project the asymetric unit multiple times
    subroutine image_project_many_helical_subunits(self,helix,projection,waypoints,maximum_radius,volume_oversampling,vol_shift_z)
        use coos, only : waypoint, waypoints_rescale_coos
        use rotation_matrices, only : euler_to_rotmat
        ! for debug only
        !use coos, only : waypoints_print_info
        ! arguments
        type(Image),                        intent(inout)   ::  self                !<  3D volume (central helical subunit must be (soft-)masked)
        type(helical_symmetry),             intent(in)      ::  helix               !<  Helical symmetry of 3D volume (before oversampling)
        type(Image),                        intent(inout)   ::  projection          !<  Projection image
        type(waypoint),                     intent(in)      ::  waypoints(:)        !<  Array of waypoints describing the position of the subunits to be projected, relative to the central pixel
        real,                               intent(in)      ::  maximum_radius      !<  Maximum helical radius (pixels) (before oversampling)
        integer,                            intent(in)      ::  volume_oversampling !<  If the 3D volume has been oversampled, indicate here the oversampling factor (>1).  Internally we will handle oversampling of the projection, waypoints and helical parameters.
        real,                               intent(in)      ::  vol_shift_z         !<  The volume should be shifted by this many pixels before projection (after oversampling).

        ! private variables
        integer ::  h_lbounds(3), h_ubounds(3)
        integer ::  current_wp
        integer ::  i,j,k
        real    ::  input_coos(4), output_coos(4)
        real    ::  rotation_matrix(4,4)
        real    ::  current_waypoint_shifts(3)
        real    ::  radius_of_furthest_point_in_h_subunit
        type(Image) ::  projection_oversampled
        logical,    parameter   ::  check_bounds = .true.
        logical,    parameter   ::  debug   =   .false.
        type(waypoint)  ::  waypoints_scaled(size(waypoints))
        ! start work

        ! Check the 3D volume is in real space and centered
        if (.not. self%ObjectIsCenteredInBox()) call self%SwapRealSpaceQuadrants()
        if (self%IsInFourierSpace()) call self%BackwardFFT()

        ! If necessary, resacale waypoints
        waypoints_scaled = waypoints
        do current_wp=1,size(waypoints_scaled)
            waypoints_scaled(current_wp)%coo(1:2) = waypoints_scaled(current_wp)%coo(1:2) * volume_oversampling
        enddo

        ! Initialise the output
        if (.not. projection%IsAllocated()) then
            call projection%Allocate(dims=[ self%GetLogicalDimension(1)/volume_oversampling, &
                                            self%GetLogicalDimension(2)/volume_oversampling, &
                                            1], &
                                            in_real_space=.true.)
        endif
        if (projection%IsInFourierSpace()) projection%is_in_real_space = .true.
        projection = 0.0e0

        ! Prepare an oversampled version of the projection
        call projection_oversampled%Allocate(dims=[ projection%GetLogicalDimension(1)*volume_oversampling, &
                                                    projection%GetLogicalDimension(2)*volume_oversampling, &
                                                    1],in_real_space=.true.)
        projection_oversampled = 0.0e0

        ! Work out boundaries of helical subunit within the volume
        call image_compute_helical_subunit_boudaries(   self,helix_scale_distances(helix,real(volume_oversampling)), &
                                                        maximum_radius*volume_oversampling, &
                                                        h_lbounds,h_ubounds)

        if (debug) then
            !write(*,'(a)') '**debug(image_project_many_helical_subunits): waypoints:'
            !call waypoints_print_info(waypoints)
            call self%PrintInfo()
            call helix_print_info(helix)
            call helical_symmetry_print_extra_info(helix)
            write(*,'(a,3(i0,1x))') '**debug(image_project_many_helical_subunits): lbounds = ', h_lbounds
            write(*,'(a,3(i0,1x))') '**debug(image_project_many_helical_subunits): ubounds = ', h_ubounds
        endif

        radius_of_furthest_point_in_h_subunit = max( sqrt(real(sum((h_lbounds-self%physical_address_of_box_center)**2))) ,&
                                                     sqrt(real(sum((h_ubounds-self%physical_address_of_box_center)**2))))

        ! Apply mask (this is now done by model3d_prep)
        !call image_mask_helical_subunit(self,helix,maximum_radius)


        !
        input_coos(4) = 1.0
        output_coos(4) = 1.0

        ! Loop over waypoints
        do current_wp=1,size(waypoints)

            ! Check whether the subunit at this waypoint will fall within the projection image
            if (check_bounds) then
                if ( any(waypoints_scaled(current_wp)%coo(1:2) - radius_of_furthest_point_in_h_subunit &
                                                            .le. projection_oversampled%logical_lower_bound_real(1:2) ) &
                .or. any(waypoints_scaled(current_wp)%coo(1:2) + radius_of_furthest_point_in_h_subunit &
                                                            .ge. projection_oversampled%logical_upper_bound_real(1:2) ) ) then
                    if (debug) then
                        write(*,'(a,3(f0.3,1x))') '**debug(image_project_many_helical_subunits): skipping waypoint with coos ', &
                                            waypoints_scaled(current_wp)%coo(:)
                        write(*,'(a,3(i0,1x))') '**debug(image_project_many_helical_subunits): projection dimensions ', &
                                            projection_oversampled%logical_dimensions
                        print *, 'radius furthest point = ', radius_of_furthest_point_in_h_subunit
                    endif
                    cycle
                endif
            endif

            ! Compute the rotation matrix for this projection
            current_waypoint_shifts    = waypoints_scaled(current_wp)%coo
            current_waypoint_shifts(3) = 0.0

            rotation_matrix = euler_to_rotmat( waypoints_scaled(current_wp)%euler%phi, &
                                        waypoints_scaled(current_wp)%euler%theta, &
                                        waypoints_scaled(current_wp)%euler%psi, &
                                        shift=current_waypoint_shifts)
            ! Loop over voxels in the 3D volume
            do k=h_lbounds(3),h_ubounds(3)
                input_coos(3) = real(k-self%physical_address_of_box_center(3)) + vol_shift_z  ! this multiplication should be done outside the loops for efficiency
                do j=h_lbounds(2),h_ubounds(2)
                    input_coos(2) = real(j-self%physical_address_of_box_center(2))
                    do i=h_lbounds(1),h_ubounds(1)
                        input_coos(1) = real(i-self%physical_address_of_box_center(1))
                        ! Rotate to the projection coordinates
                        output_coos = matmul(rotation_matrix,input_coos)
                        ! Add the current value to the output image
                        call projection_oversampled%PutRealValueByLinearInterpolationNoBoundsCheckImage(self%real_values(i,j,k),&
                                                    output_coos(1)+projection_oversampled%physical_address_of_box_center(1), &
                                                    output_coos(2)+projection_oversampled%physical_address_of_box_center(2))
                    enddo ! end loop over i
                enddo ! end loop over j
            enddo ! end loop over k
        enddo ! end loop over waypoints

        ! Resample to compensate for volume oversampling
        if (debug) write(*,'(a,i0)') '**debug(image_project_many_helical_subunits): oversampling = ', volume_oversampling
        if (volume_oversampling .eq. 1) then
            projection = projection_oversampled
        else
            call projection_oversampled%ForwardFFT()
            call projection_oversampled%ClipInto(projection)
            call projection%BackwardFFT()
        endif

    end subroutine image_project_many_helical_subunits

    !>  \brief  Project the asymetric unit multiple times, taking into account differences between the helical
    !!          symmetry of the model and the local helical symmetry at the waypoints
    subroutine image_project_many_h_subunits_with_h_distortion(self,helix,projection,waypoints,maximum_radius,volume_oversampling)
        use coos, only : waypoint
        use rotation_matrices, only : euler_to_rotmat
        use helices, only : helix_coos_cylindrical_to_lattice
        use Globals
        ! arguments
        type(Image),            intent(inout)   ::  self            !<  3D volume (central helical subunit must be masked)
        type(helical_symmetry), intent(in)      ::  helix           !<  Helical symmetry of 3D volume
        type(Image),            intent(inout)   ::  projection      !<  Projection image
        type(waypoint),         intent(in)      ::  waypoints(:)    !<  Array of waypoints describing the position of the subunits to be projected
        real,                   intent(in)      ::  maximum_radius  !<  Maximum helical radius (pixels)
        integer,                optional,   intent(in)      ::  volume_oversampling             !<  If the 3D volume has been oversampled, indicate here the oversampling factor (>1). Give 1 to indicate no oversampling.
        ! private variables
        integer ::  h_lbounds(3), h_ubounds(3)
        integer ::  current_wp
        integer ::  i,j,k
        real(kind=8)    ::  input_coos(4)
        real(kind=8)    ::  input_coos_dist(4)
        real            ::  output_coos(4)
        real(kind=8)    ::  input_coos_sq(2)
        real(kind=8)    ::  input_phi, input_r
        real(kind=8)    ::  input_coos_lattice(2)
        real(kind=8)    ::  rotation_matrix(4,4)
        real                        ::  current_waypoint_shifts(3)
        real                        ::  radius_of_furthest_point_in_h_subunit
        integer                     ::  vvolume_oversampling
        logical,    parameter       ::  check_bounds = .true.
        logical,    parameter       ::  debug   =.false.
        ! start work

        ! Was the volume oversampled?
        vvolume_oversampling = 1
        if (present(volume_oversampling)) vvolume_oversampling = volume_oversampling
        if (vvolume_oversampling .gt. 1) then
            call this_program%TerminateWithFatalError('image_project_many_h_subunits_with_h_distortion', &
                                                        'Volume oversampling not yet supported')
        endif

        ! Check the 3D volume is in real space and centered
        if (.not. self%ObjectIsCenteredInBox()) call self%SwapRealSpaceQuadrants()
        if (self%IsInFourierSpace()) call self%BackwardFFT()

        ! Work out boundaries of helical subunit within the volume
        call image_compute_helical_subunit_boudaries(self,helix,maximum_radius,h_lbounds,h_ubounds)

        if (debug) then
            write(*,'(a,3(i0,1x))') '**debug(image_project_many_h_subunits_with_h_distortion): h_lbounds = ', h_lbounds
            write(*,'(a,3(i0,1x))') '**debug(image_project_many_h_subunits_with_h_distortion): h_ubounds = ', h_ubounds
        endif

        radius_of_furthest_point_in_h_subunit = max( sqrt(real(sum((h_lbounds-self%physical_address_of_box_center)**2))) ,&
                                                     sqrt(real(sum((h_ubounds-self%physical_address_of_box_center)**2))))


        ! Apply mask (this should be done outside)
        !call image_mask_helical_subunit(self,helix,maximum_radius,soften_the_mask=.false.)

        ! Initialise the output
        if (.not. projection%IsAllocated()) then
            call projection%Allocate(dims=[self%GetLogicalDimension(1),self%GetLogicalDimension(2),1],in_real_space=.true.)
        endif
        if (projection%IsInFourierSpace()) projection%is_in_real_space = .true.
        projection = 0.0e0

        !
        input_coos(4) = 1.0
        input_coos_dist(4) = 1.0
        output_coos(4) = 1.0

        ! Debug
        if (debug) then
            write(*,'(a)') '**debug(image_project_many_h_subunits_with_h_distortion): helix of model'
            call helical_symmetry_print_info(helix)
            write(*,'(a)') '**debug(image_project_many_h_subunits_with_h_distortion): helix of wp'
            call helical_symmetry_print_info(waypoints(1)%helix)
        endif

        ! Loop over waypoints
        do current_wp=1,size(waypoints)

            ! Check whether the subunit at this waypoint will fall within the projection image
            if (check_bounds) then
                if ( any(waypoints(current_wp)%coo(1:2) - radius_of_furthest_point_in_h_subunit &
                                                            .le. projection%logical_lower_bound_real(1:2) ) &
                .or. any(waypoints(current_wp)%coo(1:2) + radius_of_furthest_point_in_h_subunit &
                                                            .ge. projection%logical_upper_bound_real(1:2) ) ) then
                    if (debug) then
                        write(*,'(a,3(f0.3,1x))') &
                            '**debug(image_project_many_h_subunits_with_h_distortion): skipping waypoint with coos ', &
                                            waypoints(current_wp)%coo(:)
                        write(*,'(a,3(i0,1x))') &
                            '**debug(image_project_many_h_subunits_with_h_distortion): projection dimensions ', &
                                            projection%logical_dimensions
                        print *, 'radius furthest point = ', radius_of_furthest_point_in_h_subunit
                    endif
                    cycle
                endif
            endif

            ! Compute the rotation matrix for this projection
            current_waypoint_shifts    = waypoints(current_wp)%coo
            current_waypoint_shifts(3) = 0.0

            if (debug) then
                write(*,'(a,2(f0.1))') '**debug(image_project_many_h_subunits_with_h_distortion): current waypoint shifts = ', &
                                        current_waypoint_shifts(1:2)
            endif

            rotation_matrix = euler_to_rotmat( waypoints(current_wp)%euler%phi, &
                                        waypoints(current_wp)%euler%theta, &
                                        waypoints(current_wp)%euler%psi, &
                                        shift=current_waypoint_shifts)
            ! Loop over voxels in the 3D volume
            do k=h_lbounds(3),h_ubounds(3)
                input_coos(3) = real(k-self%physical_address_of_box_center(3))
                do j=h_lbounds(2),h_ubounds(2)
                    input_coos(2) = real(j-self%physical_address_of_box_center(2))
                    input_coos_sq(2) = input_coos(2)**2
                    do i=h_lbounds(1),h_ubounds(1)
                        input_coos(1) = real(i-self%physical_address_of_box_center(1))
                        input_coos_sq(1) = input_coos(1)**2

                        !if (debug) print '(a,3(f0.3,1x))', '**debug(): input_coos,  before distortion: ', input_coos(1:3)

                        input_coos_dist = input_coos

                        ! Convert to cylindrical
                        input_phi = helix_phi_from_xy(input_coos(1),input_coos(2))
                        input_r = sqrt(input_coos_sq(1)+input_coos_sq(2))

                        ! Convert to helical lattice coordinates
                        call helix_coos_cylindrical_to_lattice(helix, &
                                                               input_coos(3),input_phi, &
                                                               input_coos_lattice(2),input_coos_lattice(1))

                        ! Now convert back to cylindrical, but using the helical symmetry of the current waypoint
                        call helix_coos_lattice_to_cylindrical(waypoints(current_wp)%helix, & !helix, &
                                                               input_coos_lattice(1),input_coos_lattice(2), &
                                                               input_coos_dist(3),input_phi)

                        ! Convert back to cartesian
                        call helix_xy_from_phi_r(input_phi,input_r,input_coos_dist(1),input_coos_dist(2))

                        !if (debug) print '(a,3(f0.3,1x))', '**debug(): input_coos,  after distortion: ', input_coos_dist(1:3)

                        ! Rotate to the projection coordinates
                        output_coos = matmul(rotation_matrix,input_coos_dist)
                        !if (debug) print '(a,2(f0.3,1x))', '**debug(): output_coos                  : ', output_coos(1:2)
                        ! Add the current value to the output image
                        call projection%PutRealValueByLinearInterpolationNoBoundsCheckImage(self%real_values(i,j,k),&
                                                                    output_coos(1)+projection%physical_address_of_box_center(1), &
                                                                    output_coos(2)+projection%physical_address_of_box_center(2))
                    enddo ! end loop over i
                enddo ! end loop over j
            enddo ! end loop over k
        enddo ! end loop over waypoints

        if (debug) write(*,'(a)') '**debug(image_project_many_h_subunits_with_h_distortion): exiting'

    end subroutine image_project_many_h_subunits_with_h_distortion

    !>  \brief  Work out x,y,z boudaries of the helical subunit
    subroutine image_compute_helical_subunit_boudaries(self,helix,maximum_radius,lower_bound,upper_bound)
        type(Image),            intent(in)  ::  self
        type(helical_symmetry), intent(in)  ::  helix           !<  Helical symmetry
        real,                   intent(in)  ::  maximum_radius  !<  Pixels
        integer,                intent(out) ::  lower_bound(3)
        integer,                intent(out) ::  upper_bound(3)
        !
        integer                 ::  i,j,k
        real(kind=8)            ::  x,y,z,ysq,r, phi
        type(helical_symmetry)  ::  helix_onestart
        logical                 ::  in_asym_unit
        ! Start work


        ! Ensure the helical operator is 1-start
        helix_onestart = helix
        if (helix_onestart%num_starts .ne. 1) then
            call helix_to_one_start(helix_onestart)
        endif

        ! Initialise
        upper_bound = [1,1,1]
        lower_bound = self%logical_dimensions


        ! Loop over the image voxels
        do k=1,self%logical_dimensions(3)
            z = real(k-self%physical_address_of_box_center(3),kind=8)
            !zz = z
            do j=1,self%logical_dimensions(2)
                y = real(j-self%physical_address_of_box_center(2),kind=8)
                ysq = y**2
                do i=1,self%logical_dimensions(1)
                    x = real(i-self%physical_address_of_box_center(1),kind=8)
                    ! radius
                    r = sqrt(x**2+ysq)
                    if (r .gt. maximum_radius) then
                        cycle
                    else
                        ! Work out the longitude (phi)
                        phi = helix_phi_from_xy(x,y)
                        ! Are we in the asymetric unit?
                        in_asym_unit = helix_point_is_in_asym(helix_onestart,z,phi)
                        if (in_asym_unit) then
                            if (i .lt. lower_bound(1)) lower_bound(1) = i
                            if (j .lt. lower_bound(2)) lower_bound(2) = j
                            if (k .lt. lower_bound(3)) lower_bound(3) = k
                            if (i .gt. upper_bound(1)) upper_bound(1) = i
                            if (j .gt. upper_bound(2)) upper_bound(2) = j
                            if (k .gt. upper_bound(3)) upper_bound(3) = k
                        endif
                    endif
                enddo
            enddo
        enddo
    end subroutine image_compute_helical_subunit_boudaries

    !>  \brief  Given a small cubic helically-symmetric volume, obtain an elongated projection without extending the helix in 3d
    !!
    !!  \todo   Make this work with proper subunit phasing in the projections (see comments)
    subroutine image_project_long_helix(self,proj,helix,phi_at_origin,interpol_scheme,helix_max_rad)
        use interpolation
        use images_threed
        use images
        use UsefulFunctions, only : IsEven
        use InterpolatingWindows
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),                intent(inout)   ::  self            !<  Volume with helically-symmetric data (no Z-masking!). Expected to be cubic, with helical axis = 3rd axis.
        type(image),                intent(inout)   ::  proj            !<  Projection image. Overwritten. Must have dimensions already set.
        type(helical_symmetry),     intent(in)      ::  helix           !<  Helical symmetry operator
        real,                       intent(in)      ::  phi_at_origin   !<  Phi (twist) angle of projection at origin
        type(interpolation_scheme), intent(in)      ::  interpol_scheme !<  Interpolation scheme to be used
        real,                       intent(in)      ::  helix_max_rad   !<  Max radius of helical object (pixels)
        ! Private variables
        integer     ::  current_shift
        !integer     ::  proj_min_logi(3), proj_tmp_min_logi(3)
        integer     ::  proj_min_phys(3), proj_tmp_min_phys(3)
        !integer     ::  proj_max_logi(3), proj_tmp_max_logi(3)
        integer     ::  proj_max_phys(3), proj_tmp_max_phys(3)
        integer     ::  proj_origin(3), proj_tmp_origin(3)
        real        ::  current_phi
        type(image) ::  proj_tmp
        integer     ::  num_tmp_images, current_tmp_image
        integer     ::  segment_length
        type(window)    ::  win
        ! Start work

        ! Check input 3d is cubic
        if (.not. self%IsSquare()) then
            call terminate('Fatal error in IMAGE_PROJECT_LONG_HELIX')
        endif

        ! Check projection image has assigned dimensions and is allocated
        if (.not. proj%IsAllocated() .or. any(proj%logical_dimensions(1:2) .eq. 1)) then
            write(*,'(a,l1,1x,3(i0,1x))') '**ERROR(image_project_long_helix): projection image allocated? dim? ', &
                                        proj%IsAllocated(), proj%logical_dimensions
            call terminate('Fatal error in IMAGE_PROJECT_LONG_HELIX')
        endif

        ! Check 3d input dimension matches dim(1) of projection
        if (self%logical_dimensions(1) .ne. proj%logical_dimensions(1)) then
            write(*,'(a,2(3(i0,1x)))') '**ERROR(image_project_long_helix): incompatible dimensions: ', &
                                        self%logical_dimensions, proj%logical_dimensions
			call terminate('Fatal error in IMAGE_PROJECT_LONG_HELIX')
        endif

        ! Setup interpol window
        call win%AllocateWindow(interpol_scheme%window_width,.false.,.true.)

        ! Setup temporary projection image
        proj_tmp%logical_dimensions = self%logical_dimensions
        proj_tmp%logical_dimensions(3) = 1
        call proj_tmp%Allocate()

        ! Because the projection is only computed within the circle internal to the 3d cube, we will trim the ends
        ! of the projection so that only the central part of the projection (not affected by the circle) will
        ! be used
        segment_length = 2 * floor(sqrt(proj_tmp%logical_dimensions(2)**2/4.0 - HELIX_MAX_RAD**2))
        write(*,'(a,F0.4)') '**debug(image_project_long_helix): helix max rad = ', helix_max_rad
        write(*,'(a,i0)') '**debug(image_project_long_helix): segment length = ', segment_length
		write(*,'(a,3(3(i0,1x)))') '**debug(image_project_long_helix): dims = ', self%logical_dimensions, &
		                            proj_tmp%logical_dimensions, proj%logical_dimensions


        ! Physical address of the origins
        if (proj%is_in_real_space .and. proj%object_is_centered_in_box) then
            proj_origin = proj%physical_address_of_box_center
        else
            proj_origin = [1,1,1]
        endif
        if (proj_tmp%is_in_real_space .and. proj_tmp%object_is_centered_in_box) then
            proj_tmp_origin = proj_tmp%physical_address_of_box_center
        else
            proj_tmp_origin = [1,1,1]
        endif

        ! How many images fit into the output?
        num_tmp_images = ceiling(real(proj%logical_dimensions(2)) / real(segment_length))
        if (IsEven(num_tmp_images)) num_tmp_images = num_tmp_images + 1
        write(*,'(a,i0)') '**debug(image_project_long_helix): number of projections = ', num_tmp_images

        do current_tmp_image=1,num_tmp_images
            current_shift = segment_length * (current_tmp_image - (NUM_TMP_IMAGES+1)/2)
            !
            write(*,'(a,i0,a,i0,a,i0)') '**info(image_project_long_helix): on projection ', current_tmp_image, &
                                        ' of ', num_tmp_images, ', with shift = ', current_shift
            ! Work out phi angle - this needs to be improved
            current_phi = phi_at_origin + real(current_shift) * helix%twist / helix%rise ! Does not phase subunits correctly!
            ! Project
            call volume_project_lean(self,proj_tmp,dble(current_phi),90.0D0,90.0D0,interpol_scheme,win,max_cyl_rad=helix_max_rad)
            ! The bounds of the projection images
            !call min_max_logical_address(proj_tmp,  proj_tmp_min_logi,  proj_tmp_max_logi)
            ! We will only use the parts of proj_tmp not affected by circle mask
            ! Work out the addresses proj_tmp hovers over
            proj_min_phys = proj_origin + proj_tmp%logical_lower_bound_real
            proj_max_phys = proj_origin + proj_tmp%logical_upper_bound_real
            proj_tmp_min_phys = proj_tmp_origin + proj_tmp%logical_lower_bound_real
            proj_tmp_max_phys = proj_tmp_origin + proj_tmp%logical_upper_bound_real
            proj_min_phys(2) = proj_min_phys(2) + current_shift
            proj_max_phys(2) = proj_max_phys(2) + current_shift
            ! Maybe we're over one of the edges of the output image?
            if (proj_min_phys(2) .lt. 1) then
                proj_tmp_min_phys(2) = proj_tmp_min_phys(2) + (1 - proj_min_phys(2))
                proj_min_phys(2) = 1
            else if (proj_max_phys(2) .gt. proj%logical_dimensions(2)) then
          		proj_tmp_max_phys(2) = proj_tmp_max_phys(2) - (proj_max_phys(2) - proj%logical_dimensions(2))
                proj_max_phys(2) = proj%logical_dimensions(2)
            endif
            proj%real_values(proj_min_phys(1):proj_max_phys(1),proj_min_phys(2):proj_max_phys(2),1) = &
                        proj_tmp%real_values(proj_tmp_min_phys(1):proj_tmp_max_phys(1),proj_tmp_min_phys(2):proj_tmp_max_phys(2),1)
        enddo
        !call proj_tmp%WriteToDisk('dbg_proj_tmp.mrc')
    end subroutine image_project_long_helix

    !>  \brief  Mask a helical density, for example in preparation for use as a reference or in an fsc experiment.
    !!
    !!
    !! Masking: 3 masks are calculated & combined:
    !! 1. a cylindrical mask, safely around the helix
    !! 2. a "molecular envelope mask"
    !! 3. a Z-edge mask, to ensure no discontinuity at the edges of the box
    !!
    !! The final step is to soften this combined mask.
    !!
    subroutine image_helix_mask(self,helix,asym_unit_mw,pixel_size,helix_max_rad,f_mask,f_smooth, &
                                hard_mask,soft_mask,mask_resolution)
        use masks
        use images_arithm
        use runtime_parameters, only : default_falloff_pixels
        use Globals
        implicit none
        ! Arguments
        type(image),                        intent(inout)   ::  self            !<  Volume to be masked. Overwritten with masked version.
        type(helical_symmetry),             intent(in)      ::  helix           !<  Helical symmetry of the reconstruction
        real,                               intent(in)      ::  asym_unit_mw    !<  Expected molecular weight (Da) of the asymmetric unit
        real,                               intent(in)      ::  pixel_size      !<  Voxel size (Angstroms)
        real,                               intent(in)      ::  helix_max_rad   !<  Maximum radius of the helical assembly (pixels)
        real,                   optional,   intent(out)     ::  f_mask          !<  The fraction of voxels within the binary mask
        real,                   optional,   intent(out)     ::  f_smooth        !<  The fraction of voxels within the smooth mask (computed a la Sindelar, 2011)
        type(image),            optional,   intent(inout)   ::  hard_mask       !<  If present, the binary mask will be returned
        type(image),            optional,   intent(inout)   ::  soft_mask       !<  If present, the final mask will be returned
        real,                   optional,   intent(in)      ::  mask_resolution !<  In 1/pixel. If present, the mask will be limited to this resolution. If absent, half the Nyquist will be used as default.
        ! Private variables
        real                                        ::  total_mw, pct_in_mask
        type(image)                                 ::  mask3d
        logical,        parameter                   ::  debug   =   .false.
        real                                        ::  mmask_resolution
        real                                        ::  z_ends_mask
        real,           parameter                   ::  max_mask_resolution = 5.0  !   In Angstroems, the highest resolution for the mask
        real                                        ::  cosine_falloff_width
        ! Start work

        ! If the image is blank, no need to do anything!
        if (self%IsConstant()) then
            write(*,'(a)') '**info(image_helix_mask): input is blank, not doing anything'
        else
            ! Check the helical parameters are in pixel units
            if (.not. helix%pixel_units) then
                write(*,'(a)') '**ERROR(image_helix_mask): was expecting helical parameters in pixel units'
                call terminate('Fatal error in IMAGE_HELIX_MASK')
            endif

            ! Check the image has origin at center
            if (.not. self%object_is_centered_in_box) then
                call self%SwapRealSpaceQuadrants()
            endif

            ! Check the image is in real space
            if (.not. self%is_in_real_space) then
                call self%BackwardFFT()
            endif

            ! What resolution will the mask be at? (in reciprocal pixels; 0.5 is Nyquist)
            if (present(mask_resolution)) then
                !mmask_resolution = min(mask_resolution,0.5)
                mmask_resolution = mask_resolution
            else
                mmask_resolution = pixel_size/20.0e0
            endif
            if (pixel_size/mmask_resolution .gt. 30.0) mmask_resolution = pixel_size / 30.0
            if (pixel_size/mmask_resolution .lt. max_mask_resolution) mmask_resolution = pixel_size / max_mask_resolution
            if (this_program%IsVerbose() .or. debug) then
                write(*,'(a,F0.2,a)') '**info(image_helix_mask): will use envelope mask with resolution of ', &
                                        pixel_size/mmask_resolution, ' Angstroems'
            endif

            ! The width of the cosine falloff
            cosine_falloff_width = default_falloff_pixels
            cosine_falloff_width = 1.0e0/mmask_resolution
            cosine_falloff_width = 5.0


            ! Diagnostic dump
            if (debug) call self%WriteToDisk('dbg_before_masking.mrc')

            ! First, impose the cylindrical mask
            if (helix_max_rad .eq. 0.0) then
                call cyl_mask(self,.9,nint(cosine_falloff_width),zero_float_inside=.true.)
            else
                call cyl_mask(self,helix_max_rad,nint(cosine_falloff_width),zero_float_inside=.true.)
            endif

            ! Debug
            if (debug) then
                print *, '**debug(image_helix_mask): helix_max_rad =', helix_max_rad
                call self%WriteToDisk('dbg_cyl_mask.mrc')
            endif

            ! Mask the z edges
            !z_ends_mask = min(cosine_falloff_width,real(self%logical_dimensions(3)/2.*0.1))
            z_ends_mask = cosine_falloff_width
            z_ends_mask = max(z_ends_mask,5.0)
            if (debug) write(*,'(a,F0.3)') '**debug(image_helix_mask): z-ends mask width = ', z_ends_mask
            call image_mask_z_ends_inplace(self,nint(z_ends_mask),soft=.true.)

            ! Debug
            if (debug) then
                print *, '**debug(image_helix_mask): helix_max_rad =', helix_max_rad
                call self%WriteToDisk('dbg_cylz_mask.mrc')
            endif

            ! Masking, with an envelope this time
            if (asym_unit_mw .eq. 0.0) then
                print '(a)', '**WARNING(image_helix_mask): asymetric unit mw not given, not masking molecular envelope. ',  &
                                        'If you can estimate the mw of the asymetric unit, it is highly recommended you do so.'
                call mask3d%Allocate(self)
                mask3d%real_values = 1.0
            else
                ! First, work out the total mw of the helical assembly present in the box, taking into account
                ! the z masking
                ! If the cosine-edge falloff width of the z mask is n pixels, then this would be equivalent to
                ! zeroing n/2 slices of the volume completely, at each z boundary; i.e. n in total.
                ! The formula below assumes that this is precisely the case (double-check this is true
                ! by looking carefully at the z masking routine above)
                total_mw = asym_unit_mw * helix%axial_symmetry * helix%perpendicular_symmetry * helix%num_starts * &
                                                                    (self%logical_dimensions(3)-z_ends_mask) / helix%rise
                ! Work out the number of voxels in the mask corresponding to the mw
                pct_in_mask = total_mw / (protein_density * pixel_size**3)
                if (debug) then
                    call helix_print_info(helix)
                    print *, '**debug(): number of voxels = ', pct_in_mask
                    print *, '**debug(): box volume = ', product(self%logical_dimensions)
                endif
                ! Convert to percentage of total number of voxels
                pct_in_mask = pct_in_mask * 100. / product(self%logical_dimensions)
                if (debug) then
                    !print *, '**debug(): total_mw = ', total_mw
                    !print *, '**debug(): protein_density = ', protein_density
                    !print *, '**debug(): pixel_size = ', pixel_size
                    !print *, '**debug(): protein_density * pixel_size = ', protein_density * pixel_size
                endif

                ! Diagnostics
                if (this_program%IsVerbose() .or. debug) then
                    print '(a,F0.2,a,F0.2,a)', '**info(image_helix_mask): total mw (', total_mw, ') corresponds to ', &
                                                pct_in_mask, '% of voxels'
                ! Calculate the molecular envelope
                    write(*,'(a,F0.3,a)') '**info(image_helix_mask): computing low-pass envelope with resolution of ', &
                                            pixel_size/mmask_resolution, 'A'
                endif
                call image_low_pass_mask(self,mask3d,pct_in_mask,soften=.false.,low_pass_frequency=mmask_resolution)

                ! Debug
                if (debug) call mask3d%WriteToDisk('dbg_lp_mask.mrc')
            endif


            ! Apply Z-ends mask
            if (debug) write(*,'(a,i0)') '**debug(image_helix_mask): apply z mask ', &
                        nint(max(z_ends_mask,mask3d%logical_dimensions(3)*0.2))
            call image_mask_z_ends_inplace(mask3d,nint(max(z_ends_mask,mask3d%logical_dimensions(3)*0.2)), &
                                           soft=.false.,mask_value=0.0)

            ! If requested, compute the fraction of voxels within the binary mask
            if (present(f_mask)) then
                f_mask = mask3d%GetSumOfValuesDouble() / product(self%logical_dimensions)
            endif

            ! If requested, return the combined mask before softening
            if (present(hard_mask)) then
                !call image_assignment(hard_mask,mask3d)
                hard_mask = mask3d
            endif

            ! Soften the combined mask
            if (this_program%IsVerbose() .or. debug) then
                write(*,'(a,F0.2)') '**info(image_helix_mask): adding cosine falloff to the mask, width = ', cosine_falloff_width
            endif
            if (debug) call mask3d%WriteToDisk('dbg_mask_before_cos.mrc')
            call mask_add_cosine_falloff(mask3d,cosine_falloff_width)
            if (debug) call mask3d%WriteToDisk('dbg_mask_after_cos.mrc')

            ! If requested, compute the fraction of voxels within softened mask
            if (present(f_smooth)) f_smooth = fraction_of_voxels_within_mask(mask3d)

            if (present(soft_mask)) soft_mask = mask3d

            ! Applied the softened combined mask
            !call image_apply_mask(self,mask3d,MASK_WITH_AVERAGE_OUTSIDE=.TRUE.)
            self%real_values = self%real_values * mask3d%real_values
            if (this_program%IsVerbose() .or. debug) then
                write(*,'(a)') '**info(image_helix_mask): all done'
            endif

            ! Diagnostic dump
            if (debug) call self%WriteToDisk('dbg_masked.mrc')
        endif   ! End of test for blank image

    end subroutine image_helix_mask

    !>  \brief  Mask (zeroe) voxels that are not in the helical asymetric unit
    subroutine image_mask_helical_subunit(self,helix,max_radius,soften_the_mask)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),                        intent(inout)   ::  self
        type(helical_symmetry),             intent(in)      ::  helix
        real,                 optional,     intent(in)      ::  max_radius      !<  In pixels, the maximum radius
        logical,                optional,   intent(in)      ::  soften_the_mask
        ! Private variables
!        real,   parameter       ::  asym_unit_center(3) = (/ 1.0, 0.0, 0.0 /)   ! The unit vector that points from the center of the
!                                                                                ! volume to the origin of the helical unit cell
        integer                 ::  i,j,k, ori(3)
        real(kind=8)            ::  x,y,z,ysq,r, phi
        real                    ::  mmax_radius
        type(helical_symmetry)  ::  helix_onestart
        logical                 ::  in_asym_unit
        logical                 ::  ssoften_the_mask
        type(Image)             ::  mask
        logical,    parameter   ::  debug = .false.
        ! Start work

        !
        ssoften_the_mask = .false.
        if (present(soften_the_mask)) ssoften_the_mask = soften_the_mask

        ! Must be in real space
        if (.not. self%ObjectIsCenteredInBox()) call self%SwapRealSpaceQuadrants()
        if (self%IsInFourierSpace()) call self%BackwardFFT()

        ! The address of the origin
        ori = self%physical_address_of_box_center


        ! a maximal radius
        mmax_radius = real(self%logical_dimensions(1))/2.0
        if (present(max_radius)) then
            mmax_radius=min(mmax_radius,max_radius)
        endif

        ! Ensure the helical operator is 1-start
        helix_onestart = helix
        if (helix_onestart%num_starts .ne. 1) then
            call terminate('image_mask_helical_subunit: need 1-start helix')
        endif
        !call helix_to_one_start(helix_onestart)

        !
        if (debug) then
            call self%WriteToDisk('dbg_before_h_mask.mrc')
        endif


        !
        if (ssoften_the_mask) then
            call mask%Allocate(self)
            mask = 1.0
        endif

        ! Loop over the image voxels
        do K=1,self%logical_dimensions(3)
            z = real(k-ori(3),kind=8)
            !zz = z
            do j=1,self%logical_dimensions(2)
                y = real(j-ori(2),kind=8)
                ysq = y**2
                do i=1,self%logical_dimensions(1)
                    x = real(i-ori(1),kind=8)
                    ! radius
                    r = sqrt(x**2+ysq)
                    if (r .gt. mmax_radius) then
                        if (ssoften_the_mask) then
                            mask%real_values(i,j,k) = 0.0
                        else
                            self%real_values(i,j,k) = 0.0
                        endif
                        cycle
                    else
                        ! Work out the longitude (phi)
                        phi = helix_phi_from_xy(x,y)
                        ! Are we in the asymetric unit?
                        in_asym_unit = helix_point_is_in_asym(helix_onestart,z,phi)
                        if (.not. in_asym_unit) then
                            if (ssoften_the_mask) then
                                mask%real_values(i,j,k) = 0.0
                            else
                                self%real_values(i,j,k) = 0.0
                            endif
                        endif
                    endif
                enddo
            enddo
        enddo

        if (ssoften_the_mask) then
            if (debug) call mask%WriteToDisk('dbg_h_mask.mrc')
            call image_soften_helical_subunit_mask(mask,helix,mmax_radius)
            !call image_soften_helical_subunit_mask_old(mask)
            if (debug) call mask%WriteToDisk('dbg_soft_h_mask.mrc')
            self%real_values = self%real_values * mask%real_values
        endif

    end subroutine image_mask_helical_subunit

    !>  \brief  Apply a z-shift and a rotation to the volume that the protein density is aligned with the primitive unit cell
    subroutine image_align_to_helical_lattice(self,helix,inner_radius,outer_radius,shift_phi,shift_z)
        use OrientationsAndRotations
        use Globals
        ! Arguments
        type(Image),            intent(inout)   ::  self
        type(helical_symmetry), intent(in)      ::  helix
        real,                   intent(in)      ::  inner_radius    !<  In pixels, do not project densities that are within the cylinder of this radius
        real,                   intent(in)      ::  outer_radius    !<  Pixels
        real,                   intent(out)     ::  shift_phi       !<  Rotation (degrees) around Z applied to the 3D to align it
        real,                   intent(out)     ::  shift_z         !<  Shift (pixels) along Z applied to the 3D to align it
        ! Private variables
        type(Image)         ::  rotated_self
        type(EulerAngles)   ::  my_euler_angles
        real                ::  affine_transformation_matrix(4,4)
        logical,    parameter   ::  debug = .false.
        ! Start work

        call image_compute_alignment_to_helical_lattice(  self,helix,inner_radius,outer_radius, &
                                                        shift_phi,shift_z)

        if (debug) then
            write(*,'(a,2(f0.2,1x))') '**debug(image_align_with_helical_lattice): shifting volume by (phi,z) = ', &
                                        shift_phi, shift_z
        endif

        call my_euler_angles%Init()
        call my_euler_angles%SetAnglesInDegrees(shift_phi,0.0,0.0)
        affine_transformation_matrix = my_euler_angles%GetRotationMatrix([0.0,0.0,shift_z])
        call self%ApplyAffineTransformation(rotated_self,affine_transformation_matrix)
        self = rotated_self

    end subroutine image_align_to_helical_lattice

    !>  \brief  Project the densities from the volume onto a 2D plane with lattice coordinates
    subroutine image_compute_alignment_to_helical_lattice(self,helix,inner_radius,outer_radius,shift_phi,shift_z)
        use Globals
        use UsefulFunctions, only : DegreesToRadians
        use Peaks
        ! Arguments
        type(Image),            intent(in)      ::  self
        type(helical_symmetry), intent(in)      ::  helix
        real,                   intent(in)      ::  inner_radius    !<  In pixels, do not project densities that are within the cylinder of this radius
        real,                   intent(in)      ::  outer_radius    !<  Pixels
        real,                   intent(out)     ::  shift_phi       !<  Rotation around Z to be applied so as to center the density
        real,                   intent(out)     ::  shift_z         !<  Shift along Z to be applied so as to center the density
        ! Private variables
        integer             ::  i,j,k
        integer             ::  central_pixel_address(3)
        real(kind=8)        ::  x,y,z
        real(kind=8)        ::  x_sq,y_sq,z_sq
        real(kind=8)        ::  r,r_sq
        real(kind=8)        ::  phi
        real(kind=8)        ::  lattice_x,lattice_y
        real(kind=8)        ::  inner_radius_sq, outer_radius_sq
        real                ::  samples_per_unit_cell_edge
        real,   parameter   ::  max_lattice_distance_from_center = 1.5 ! Distance in unit cell edges
        integer             ::  projection_dims(3)
        real                ::  projection_x, projection_y
        type(Image)         ::  projection, template, ccf
        type(Peak)          ::  alignment_peak
        real                ::  shift_x, shift_y
        logical,parameter   ::  debug = .false.
        ! Start work

        samples_per_unit_cell_edge = max(   helix%fast_helix_rise, &
                                            helix%slow_helix_rise, &
                                            DegreesToRadians(helix%fast_helix_twist)*outer_radius, &
                                            DegreesToRadians(helix%slow_helix_twist)*outer_radius   )

        if (debug) then
            write(*,'(a,f0.2)') '**debug(image_compute_alignment_to_helical_lattice): samples per edge = ', &
                                samples_per_unit_cell_edge
        endif

        ! Allocate output image
        projection_dims(1:2) = ceiling(max_lattice_distance_from_center * 2.0 * samples_per_unit_cell_edge)
        projection_dims(3)   = 1
        call projection%Allocate(dims=projection_dims)
        projection = 0.0e0

        ! Useful values for later
        central_pixel_address = self%GetPhysicalAddressOfBoxCenter()
        inner_radius_sq = inner_radius**2
        outer_radius_sq = outer_radius**2



        ! Loop over volume
        do k=1,self%GetLogicalDimension(3)
            z = real(k-central_pixel_address(3))
            z_sq = z**2
            do j=1,self%GetLogicalDimension(2)
                y = real(j-central_pixel_address(2))
                y_sq = y**2
                do i=1,self%GetLogicalDimension(1)
                    x = real(i-central_pixel_address(1))
                    x_sq = x**2
                    r_sq = x_sq + y_sq
                    if (r_sq .ge. inner_radius_sq .and. r_sq .le. outer_radius_sq) then
                        phi = helix_phi_from_xy(x,y)
                        call helix_coos_cylindrical_to_lattice(helix,z,phi,lattice_y,lattice_x)
                        if (        abs(lattice_x) .lt. max_lattice_distance_from_center - 2.0/samples_per_unit_cell_edge &
                            .and.   abs(lattice_y) .lt. max_lattice_distance_from_center - 2.0/samples_per_unit_cell_edge) then
                            projection_x = lattice_x * samples_per_unit_cell_edge
                            projection_y = lattice_y * samples_per_unit_cell_edge
                            ! Add the current value to the output image
                            call projection%PutRealValueByLinearInterpolationNoBoundsCheckImage( &
                                                self%real_values(i,j,k),&
                                                projection_x+projection%GetPhysicalIndexOfBoxCenter(1), &
                                                projection_y+projection%GetPhysicalIndexOfBoxCenter(2))
                        endif
                    endif
                enddo
            enddo
        enddo


        ! Let's now generate a template to align to
        call template%Allocate(mould=projection)
        call template%ApplySoftRectangularMask([samples_per_unit_cell_edge*0.75,samples_per_unit_cell_edge*0.75], &
                                                azimuth=0.0,falloff=samples_per_unit_cell_edge*0.1, &
                                                wipe_input=.true.)

        if (debug) then
            call projection%WriteToDisk('dbg_projection.mrc')
            call template%WriteToDisk('dbg_template.mrc')
        endif

        ! Align against this template
        ccf = projection
        call ccf%CalculateCrossCorrelationImageWith(template)
        alignment_peak = ccf%FindPeakWithIntegerCoordinates(max_radius=samples_per_unit_cell_edge*sqrt(2.0)*0.5)
        if (debug) call alignment_peak%PrintInfo()
        shift_x = -alignment_peak%CoOrdinates(1)
        shift_y = -alignment_peak%CoOrdinates(2)
        call projection%PhaseShift(shift_x,shift_y,0.0)

        if (debug) then
            call ccf%WriteToDisk('dbg_ccf.mrc')
        endif

        shift_x = shift_x / samples_per_unit_cell_edge
        shift_y = shift_y / samples_per_unit_cell_edge

        ! Convert to phi / z
        shift_phi = -helix%fast_helix_twist * shift_y - helix%slow_helix_twist * shift_x
        shift_z   = -helix%fast_helix_rise  * shift_y - helix%slow_helix_rise  * shift_x

    end subroutine image_compute_alignment_to_helical_lattice


    !>  \brief  Soften the binary mask of the helical asymetric unit, such that pixels less than 1 pixel away
    !!          from the asymmetric unit boundary will have values between 0.0 and 1.0
    subroutine image_soften_helical_subunit_mask(self,helix,outer_radius)
        ! arguments
        type(Image),            intent(inout)   ::  self
        type(helical_symmetry), intent(in)      ::  helix
        real,                   intent(in)      ::  outer_radius    !<  Pixels
        ! private variables
        type(Image) ::  new_mask
        integer ::  i,j,k, ii,jj,kk
        integer ::  lower_bounds(3),upper_bounds(3)
        ! start work

        ! Allocate output
        call new_mask%Allocate(self)
        new_mask = 0.0

        ! We only need loop near the primitive unit cell
        call image_compute_helical_subunit_boudaries(self,helix,outer_radius,lower_bounds,upper_bounds)

        ! Let's add a pixel either side
        lower_bounds = max(lower_bounds-1,1)
        upper_bounds = min(upper_bounds+1,self%logical_dimensions)

        ! Now loop
        do k=lower_bounds(3),upper_bounds(3)
            do j=lower_bounds(2),upper_bounds(2)
                do i=lower_bounds(1),upper_bounds(1)
                    !if (self%real_values(i,j,k) .eq. 0.0) then
                        ! We're outside binary mask. Let's see if we're near it.
                        do kk=k-1,k+1
                            do jj=j-1,j+1
                                do ii=i-1,i+1
                                    if (self%real_values(ii,jj,kk) .ne. 0.0) then
                                        new_mask%real_values(i,j,k) = new_mask%real_values(i,j,k) + 1.0/27.0 ! There are 26 neighbours that could be filled, as well as the current voxel
                                    endif
                                enddo !ii
                            enddo !jj
                        enddo !kk
                    !endif ! outside binary mask
                enddo !i
            enddo !j
        enddo !k

        self%real_values = new_mask%real_values

        call new_mask%Deallocate()

    end subroutine image_soften_helical_subunit_mask

    !>  \brief  Soften the binary mask of the helical asymetric unit, such that pixels less than 1 pixel away
    !!          from the asymmetric unit will have values between 0.0 and 1.0
    subroutine image_soften_helical_subunit_mask_old(self)
        ! arguments
        type(Image),    intent(inout)       ::  self
        ! private variables
        type(Image) ::  new_mask
        integer ::  i,j,k, ii,jj,kk
        ! start work
        call new_mask%Allocate(self)
        new_mask = 0.0

        do k=2,self%GetLogicalDimension(3)-1
            do j=2,self%GetLogicalDimension(2)-1
                do i=2,self%GetLogicalDimension(1)-1
                    if (self%real_values(i,j,k) .eq. 0.0) then
                        ! We're outside binary mask. Let's see if we're near it.
                        do kk=k-1,k+1
                            do jj=j-1,j+1
                                do ii=i-1,i+1
                                    if (self%real_values(ii,jj,kk) .ne. 0.0) then
                                        new_mask%real_values(i,j,k) = new_mask%real_values(i,j,k) + 1.0/26.0 ! There are 26 neighbours that could be filled
                                    endif
                                enddo !ii
                            enddo !jj
                        enddo !kk
                    endif ! outside binary mask
                enddo !i
            enddo !j
        enddo !k

        self%real_values = self%real_values + new_mask%real_values

        call new_mask%Deallocate()

    end subroutine image_soften_helical_subunit_mask_old

    !>  \brief  Smear the densities within the asymetric unit according to the helical operator given
    subroutine image_smear_asymetric_unit(self,helix,smearing_helix,interpol_scheme,max_radius)
        use interpolation, only : interpolation_scheme, interpolation_scheme_init
        use helices, only : helix_phi_from_xy, helix_point_is_near_asym, helix_point_is_in_asym
        use InterpolatingWindows, only : correct_3d
        implicit none
        ! Arguments
        type(image),                        intent(inout)   ::  self                !<  Image (volume) to be symmetrised
        type(helical_symmetry),             intent(in)      ::  helix               !<  Helical symmetry operator
        type(helical_symmetry),             intent(in)      ::  smearing_helix      !<  Helical operator used for the smearing operation (for example if smearing because the helical rise is not resolved, give an operator where the rise is 1 pixel)
        type(interpolation_scheme),         intent(in)      ::  interpol_scheme     !<  Interpolation scheme. Only used for pre-emptive filtering. Should be setup as a linear interpolator, since linear interpolation is hardcoded.
        real,       optional,               intent(in)      ::  max_radius          !<  Maximum outer radius of helix (pixels)
        ! Private variables
        type(image) ::  self_output
        integer     ::  i,j,k
        real(kind=8)::  x,y,z
        real(kind=8)::  xx,yy,zz
        real        ::  ysq
        real(kind=8)::  r, phi, pphi
        integer     ::  ori(3)
        logical     ::  in_asym_unit
        real        ::  mmax_radius
        integer     ::  direction
        real(kind=8)::  rdirection
        type(Image) ::  tmpimg
        logical,    parameter   ::  preemptive_filter = .false.
        ! Start work

        write(*,'(a)') '**debug(image_smear_asymetric_unit): smearing with following operator: '
        call helical_symmetry_print_info(smearing_helix)

        ! Initialise output volume
        call self_output%Allocate(mould=self)
        self_output = 0.0e0

        ! The address of the origin
        if (self%is_in_real_space .and. self%object_is_centered_in_box) then
            ori = self%physical_address_of_box_center
        else
            ori = [1,1,1]
        endif

        ! Maximum radius
        mmax_radius = real(self%logical_dimensions(1))/2.0
        if (present(max_radius)) then
            if (max_radius .gt. 0.0) mmax_radius = max_radius
        endif

        ! High-pass filter to compensate for low-pass-filtering effect of interpolation
        tmpimg = self
        if (preemptive_filter) then
            if (tmpimg%is_in_real_space) call tmpimg%ForwardFFT()
            call correct_3d(tmpimg,interpol_scheme%interpolation_instr_function,interpol_scheme%windowing_instr_function, &
                            interpol_scheme%window_width)
            call tmpimg%BackwardFFT()
        endif

        ! Loop over input volume
        do k=1,self%logical_dimensions(3)
            z = real(k-ori(3),kind=8)
            zz = z
            do j=1,self%logical_dimensions(2)
                y = real(j-ori(2),kind=8)
                ysq = y**2
                do i=1,self%logical_dimensions(1)
                    x = real(i-ori(1),kind=8)
                    ! Radius
                    r = sqrt(x**2+ysq)
                    if (r .gt. mmax_radius) cycle
                    ! Work out the longitude (phi)
                    phi = helix_phi_from_xy(x,y)
                    ! Are we in the asymetric unit?
                    in_asym_unit = helix_point_is_in_asym(helix,z,phi)
                    if (in_asym_unit) then
                        ! We are within the asymmetric unit, so let's get the current value and smear it throughout the asymetric unit
                        self_output%real_values(i,j,k) = self_output%real_values(i,j,k) + self%real_values(i,j,k)
                        ! We will first loop upwards, then downwards
                        do direction=-1,1,2
                            pphi = phi
                            zz   = z
                            rdirection = real(direction,kind=8)
                            do
                                pphi = pphi + rdirection * smearing_helix%twist
                                zz   = zz   + rdirection * smearing_helix%rise
                                if (helix_point_is_near_asym(helix,zz,pphi,1.0d0)) then
                                    ! Convert back to Cartesian coordinates (see euler_to_rotmat)
                                    call helix_xy_from_phi_r(pphi,r,xx,yy)
                                    ! Stick this density value into the asymmetric unit
                                    call self_output%PutRealValueByLinearInterpolationNoBoundsCheckVolume( &
                                                                                tmpimg%real_values(i,j,k), &
                                                                                real(xx+ori(1)),real(yy+ori(2)),real(zz+ori(3)))
                                else
                                    ! get out of the loop over smearing directions
                                    exit
                                endif
                            enddo
                        enddo
                    endif
                enddo
            enddo
        enddo

        call self_output%WriteToDisk('dbg_central_slice_averaged.mrc')

        ! Now apply the smearing symmetry
        call image_impose_helical_symmetry_new(self_output,smearing_helix,interpol_scheme,max_radius)

        call self_output%WriteToDisk('dbg_central_slice_averaged_smeared.mrc')

        self%real_values(:,:,:) = self_output%real_values(:,:,:)

    end subroutine image_smear_asymetric_unit

    !>  \brief  Impose helical symmetry on a volume. Version where the edge of the primitive helical unit cell has a soft border.
    !!
    !!  See derivation in notes 120328 for more details
    !!  \todo   Add pre-emptive filter to correct for interpolation when imposing odd axial symmetry
    subroutine image_impose_helical_symmetry_new_soft(self,helix,interpol_scheme,max_radius,progress_report,quiet)
        use interpolation
        !use interpolation_kernels, only : triangle_instr, rect_instr
        use InterpolatingWindows, only : correct_3d
        use rotation_matrices
        use UsefulFunctions, only : IsOdd, IsEven
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),                        intent(inout)   ::  self                !<  Image (volume) to be symmetrised
        type(helical_symmetry),             intent(in)      ::  helix               !<  Helical symmetry operator
        type(interpolation_scheme),         intent(in)      ::  interpol_scheme     !<  Interpolation scheme
        real,       optional,               intent(in)      ::  max_radius          !<  Maximum outer radius of helix (pixels)
        logical,    optional,               intent(in)      ::  progress_report     !<  Give some output regularly to let the user know we're working
        logical,    optional,               intent(in)      ::  quiet               !<  Do not print anything
        ! Private variables
        type(helical_symmetry)  ::  helix_onestart
        integer                 ::  i,j,k
        integer                 ::  ori(3)
        real(kind=8)            ::  phi,pphi,ppphi,x,y,z,ysq,r
        logical                 ::  in_asym_unit
        real(kind=8)            ::  xx,yy,zz,zzz

        logical,    parameter   ::  preemptive_filter   =   .true.  ! Whether to apply a pre-emptive high-pass filter to compensate for interpolation
        type(image)             ::  tmpimg
        logical,    parameter   ::  debug   =   .true.
        real                    ::  mmax_radius
        integer                 ::  a_sym
        real                    ::  rvalue
        real                    ::  tmpval
        type(image)             ::  subunit_voxels   ! debug
        integer                 ::  new_dim(3)
        logical                 ::  qquiet
        real(kind=8)            ::  dist_from_asym(3)
        real                    ::  weight, weight_in_asym
        ! Start work

        qquiet = .false.
        if (present(quiet)) qquiet = quiet

        ! Side dyad not implemented
        if (helix%perpendicular_symmetry .gt. 1) then
            write(*,'(a)') '**ERROR(image_impose_helical_symmetry_new): perpendicular symmetry not implemented yet'
            call terminate('Feature not implemented')
        endif

        ! Assume helix is in pixel units
        if (.not. helix%pixel_units) then
            write(*,'(a)') '**ERROR(image_impose_helical_symmetry_new): helical symmetry operator must be in pixel units'
            call terminate('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW')
        endif


        ! Maximum radius
        mmax_radius = real(self%logical_dimensions(1))/2.0
        if (present(max_radius)) then
            if (max_radius .gt. 0.0) mmax_radius = max_radius
        endif


        ! With odd axial symmetries, there will be damping due to the interpolation, which can probably
        ! be corrected for, but i don't have the time to implement this now
        !if (preemptive_filter .AND. helix%axial_symmetry .GT. 1 .AND. is_odd(helix%axial_symmetry)) then
        !    write(*,'(2a)') '**ERROR(image_impose_helical_symmetry_new): pre-emptive filtering for odd', &
        !                    ' axial symmetry not yet implemented'
        !    call terminate('Feature not implemented')
        !endif

        if (debug) then
        !    failed_voxels = self
        !    failed_voxels%real_values = 0.0
            subunit_voxels = self
			subunit_voxels%real_values = 0.0
        endif

        ! Ensure the helical operator is 1-start
        helix_onestart = helix
        call helix_to_one_start(helix_onestart)
        if (.not. qquiet) then
            call helical_symmetry_print_info(helix_onestart)
            if (debug) call helical_symmetry_print_extra_info(helix_onestart)
        endif

        ! The address of the origin
        if (self%is_in_real_space .and. self%object_is_centered_in_box) then
            ori = self%physical_address_of_box_center
        else
            ori = [1,1,1]
        endif

        ! Handle the axial symmetry first - need it applied to the asymetric unit
        if (helix%axial_symmetry .gt. 1) then

            tmpimg = self

            ! High-pass filter to compensate for low-pass-filtering effect of interpolation
            if (preemptive_filter .and. IsOdd(helix%axial_symmetry)) then
                if (tmpimg%IsInRealSpace()) call tmpimg%ForwardFFT()
                call correct_3d(tmpimg,interpol_scheme%interpolation_instr_function,interpol_scheme%windowing_instr_function, &
                            interpol_scheme%window_width)
                call tmpimg%BackwardFFT()
            else
                write(*,'(a)') '**info(image_impose_helical_symmetry_new): preemptive filter turned off because even sym'
            endif

            ! Loop over voxels
            do k=1,self%logical_dimensions(3)
                z = real(k-ori(3),kind=8)
                zz = z
                do j=1,self%logical_dimensions(2)
                    y = real(j-ori(2),kind=8)
                    ysq = y**2
                    do i=1,self%logical_dimensions(1)
                        x = real(i-ori(1),kind=8)
                        ! Radius
                        r = sqrt(x**2+ysq)
                        if (r .gt. mmax_radius) cycle
                        ! Work out the longitude (phi)
                        phi = helix_phi_from_xy(x,y)
                        ! Weight of that point as a function of distance from unit cell boundary
                        weight = helix_point_weight_from_asym(helix_onestart,z,phi,r)
                        ! If we are even close to the asym unit, let's impose the axial symmetry
                        if (weight .gt. 0.0) then
                            tmpval = self%real_values(i,j,k) / real(helix%axial_symmetry)
                            ! If we're in the asymetric unit, apply the axial symmetry
                            do a_sym = 2, helix%axial_symmetry
                                pphi = phi + (a_sym-1)*(360./real(helix%axial_symmetry))
                                ! Convert back to Cartesian coordinates (see euler_to_rotmat)
                                call helix_xy_from_phi_r(phi,r,xx,yy)
                                ! Interpolate a value from those coordinates
                                call tmpimg%GetRealValueByLinearInterpolationNoBoundsCheckVolume(rvalue, &
                                                            real(xx)+tmpimg%physical_address_of_box_center(1), &
                                                            real(yy)+tmpimg%physical_address_of_box_center(2), &
                                                            real(zz)+tmpimg%physical_address_of_box_center(3))
                                tmpval = tmpval + rvalue / real(helix%axial_symmetry)
                            enddo
                            self%real_values(i,j,k) = tmpval
                        endif
                    enddo
                enddo
            enddo
        endif

        ! High-pass filter to compensate for low-pass-filtering effect of interpolation
        if (preemptive_filter) then
            tmpimg = self
            if (self%IsInRealSpace()) call self%ForwardFFT()
            call correct_3d(self,interpol_scheme%interpolation_instr_function,interpol_scheme%windowing_instr_function, &
                            interpol_scheme%window_width)
            call self%BackwardFFT()
        else
            write(*,'(a)') '**warning(image_impose_helical_symmetry_new): preemptive filter turned off'
        endif

        ! Loop over the output voxels
        !count_total_voxels = 0
        !count_successful_voxels = 0
        !total_difficulty_level = 0
        do k=1,self%logical_dimensions(3)
            z = real(k-ori(3),kind=8)
            do j=1,self%logical_dimensions(2)
                y = real(j-ori(2),kind=8)
                ysq = y**2
                do i=1,self%logical_dimensions(1)
                    x = real(i-ori(1),kind=8)
                    ! Radius
                    r = sqrt(x**2+ysq)
                    if (r .gt. mmax_radius) cycle
                    !count_total_voxels = count_total_voxels + 1
                    ! Work out the longitude (phi)
                    phi = helix_phi_from_xy(x,y)
                    ! Distance from the edge of the primitive helical unit cell
                    weight = helix_point_weight_from_asym(helix_onestart,z,phi,r)
                    if (debug) subunit_voxels%real_values(i,j,k) = weight
                    ! If we were in the asymetric unit, we're done
                    if (weight .ge. 1.0) then
                        ! Nothing to do
                        !count_successful_voxels = count_successful_voxels + 1
                        !if (debug) subunit_voxels%real_values(i,j,k) = 1.0
                    else
                        if (weight .eq. 0.0) then
                            ! If we aren't in the asymetric unit, we need to find the corresponding point within the asymetric unit
                            ! zz and phi are modified
                            call helix_point_to_asym(helix_onestart,z,phi,zz,pphi)
                            ! Convert back to Cartesian coordinates (see euler_to_rotmat)
                            call helix_xy_from_phi_r(pphi,r,xx,yy)
                        else
                            ! Weight is .ne. 0.0 and .lt. 1.0, therefore we
                            ! must be near the border; so let's first get the current point and then we'll have to get
                            ! the equivalent point on the other side of the asym unit
                            pphi = phi
                            xx = x
                            yy = y
                            zz = z
                        endif
                        ! Interpolate a value from those coordinates
                        call self%GetRealValueByLinearInterpolationNoBoundsCheckVolume(rvalue, &
                                                        real(xx)+self%physical_address_of_box_center(1), &
                                                        real(yy)+self%physical_address_of_box_center(2), &
                                                        real(zz)+self%physical_address_of_box_center(3))
                        ! What weight is this point?
                        weight_in_asym = helix_point_weight_from_asym(helix_onestart,zz,pphi,r)
                        self%real_values(i,j,k) = rvalue * weight_in_asym
                        if (weight_in_asym .lt. 1.0) then
                            ! weight less than 1.0 means we are near the edge of the asym unit
                            ! We also need to grab a value from the other edge of the asym unit
                            call helix_point_to_other_side_of_asym(helix_onestart,zz,pphi,zzz,ppphi)

                            ! Convert back to Cartesian coordinates (see euler_to_rotmat)
                            call helix_xy_from_phi_r(ppphi,r,xx,yy)
                            ! Interpolate a value from those coordinates
                            call self%GetRealValueByLinearInterpolationNoBoundsCheckVolume(rvalue, &
                                                            real(xx)+self%physical_address_of_box_center(1), &
                                                            real(yy)+self%physical_address_of_box_center(2), &
                                                            real(zzz)+self%physical_address_of_box_center(3))
                            self%real_values(i,j,k) = rvalue * (1.0-weight_in_asym) + self%real_values(i,j,k)
                        endif
                    endif
                enddo
            enddo
        enddo

        if (preemptive_filter) then
            ! We need to restore the original asymetric unit, because currently, there is a high-passed version in its place
            do k=1,self%logical_dimensions(3)
                z = real(k-ori(3),kind=8)
                do j=1,self%logical_dimensions(2)
                    y = real(j-ori(2),kind=8)
                    ysq = y**2
                    do i=1,self%logical_dimensions(1)
                        x = real(i-ori(1),kind=8)
                        ! Radius
                        r = sqrt(x**2+ysq)
                        if (r .gt. mmax_radius) cycle
                        ! Work out the longitude (phi)
                        phi = helix_phi_from_xy(x,y)
                        ! Distance from the edge of the primitive helical unit cell
                        weight = helix_point_weight_from_asym(helix_onestart,z,phi,r)
                        ! If are in the asymetric unit, restore to initial
                        if (weight .gt. 0.0) then
                            ! Restore the data
                            self%real_values(i,j,k) = tmpimg%real_values(i,j,k)
                        endif
                    enddo
                enddo
            enddo
        endif

        if (debug) then
            call subunit_voxels%WriteToDisk('dbg_subunit.mrc')
        endif
    end subroutine image_impose_helical_symmetry_new_soft




    !>  \brief  Impose helical symmetry on a volume
    !!
    !!  See derivation in notes 120328 for more details
    !!  \todo   Add pre-emptive filter to correct for interpolation when imposing odd axial symmetry
    subroutine image_impose_helical_symmetry_new(self,helix,interpol_scheme,max_radius,progress_report,quiet)
        use interpolation
        use rotation_matrices
        use UsefulFunctions, only : IsOdd, IsEven
        use runtime_parameters, only : terminate
        use InterpolatingWindows
        use Globals
        implicit none
        ! Arguments
        type(image),                        intent(inout)   ::  self                !<  Image (volume) to be symmetrised
        type(helical_symmetry),             intent(in)      ::  helix               !<  Helical symmetry operator
        type(interpolation_scheme),         intent(in)      ::  interpol_scheme     !<  Interpolation scheme
        real,       optional,               intent(in)      ::  max_radius          !<  Maximum outer radius of helix (pixels)
        logical,    optional,               intent(in)      ::  progress_report     !<  Give some output regularly to let the user know we're working
        logical,    optional,               intent(in)      ::  quiet               !<  Do not print anything

        ! Private variables
        type(helical_symmetry)  ::  helix_onestart
        type(window)            ::  win
        !real(KIND=8)            ::  pitch
        integer                 ::  i,j,k
        integer                 ::  ori(3)
        real(kind=8)            ::  phi,pphi,x,y,z,ysq,r
        logical                 ::  in_asym_unit
        real(kind=8)            ::  xx,yy,zz

        logical,    parameter   ::  preemptive_filter   =   .true.  ! Whether to apply a pre-emptive high-pass filter to compensate for interpolation
        type(image)             ::  tmpimg
        logical,    parameter   ::  debug   =   .false.
        real                    ::  mmax_radius
        integer                 ::  a_sym
        real                    ::  rvalue
        real                    ::  tmpval
        !integer                 ::  count_successful_voxels, count_total_voxels
        !logical                 ::  success
        type(image)             ::  subunit_voxels   ! debug
        !integer                 ::  difficulty_level, total_difficulty_level
        integer,    parameter   ::  fourier_pad_factor = 1
        integer                 ::  new_dim(3)
        logical                 ::  qquiet
        logical,    parameter   ::  use_new_interpolation_routine = .true.
        ! Start work

        qquiet = .false.
        if (present(quiet)) qquiet = quiet

        ! Assume helix is in pixel units
        if (.not. helix%pixel_units) then
            write(*,'(a)') '**ERROR(image_impose_helical_symmetry_new): helical symmetry operator must be in pixel units'
            call terminate('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY_NEW')
        endif

        ! Padding in Fourier space
        if (fourier_pad_factor .gt. 1) then
            write(*,'(a,i0)') '**info(image_impose_helical_symmetry_new): padding volume in Fourier space by factor of ', &
                                fourier_pad_factor
            if (self%IsInRealSpace()) call self%ForwardFFT()
            new_dim = self%logical_dimensions * fourier_pad_factor
            call self%Resize(new_dim)
            call self%BackwardFFT()
        endif


        ! Maximum radius
        mmax_radius = real(minval(self%logical_dimensions(1:2)))/2.0 - 1.0
        if (present(max_radius)) then
            if (max_radius .gt. 0.0 .and. max_radius .lt. mmax_radius) mmax_radius = max_radius
        endif
        mmax_radius = mmax_radius * real(fourier_pad_factor)
        if (debug) write(*,'(a,f0.2,a)') '**debug(image_impose_helical_symmetry_new): max rad = ', mmax_radius, ' pix'


        ! With odd axial symmetries, there will be damping due to the interpolation, which can probably
        ! be corrected for, but i don't have the time to implement this now
        !if (preemptive_filter .AND. helix%axial_symmetry .GT. 1 .AND. is_odd(helix%axial_symmetry)) then
        !    write(*,'(2a)') '**ERROR(image_impose_helical_symmetry_new): pre-emptive filtering for odd', &
        !                    ' axial symmetry not yet implemented'
        !    call terminate('Feature not implemented')
        !endif

        if (debug) then
        !    failed_voxels = self
        !    failed_voxels%real_values = 0.0
            subunit_voxels = self
			subunit_voxels%real_values = 0.0
        endif

        ! Ensure the helical operator is 1-start
        helix_onestart = helix
        !call helix_to_one_start(helix_onestart)
        if (helix_onestart%num_starts .ne. 1) then
            call terminate('image_impose_helical_symmetry_new needs a 1-start helix')
        endif
        if (fourier_pad_factor .gt. 1) helix_onestart%rise = helix_onestart%rise * fourier_pad_factor
        if (.not. qquiet .and. (this_program%IsVerbose() .or. debug)) then
            call helical_symmetry_print_info(helix_onestart)
            if (debug) call helical_symmetry_print_extra_info(helix_onestart)
        endif

        ! Compute the pitch of the helix
        !pitch = helix_pitch(helix_onestart)
        !if (debug) write(*,'(a,F0.3)') '**debug(image_impose_helical_symmetry_new): pitch of one-start helix = ', pitch

        ! The boundaries of the helical subunit
        !if (debug) then
        !    write(*,'(a,2(F0.3,1x))') '**debug(image_impose_helical_symmetry_new): phi, z boundaries of h subunit = ', &
        !                        helix_asym_unit_phi_limit(helix_onestart,pitch), helix_asym_unit_z_limit(helix_onestart,pitch)
        !    call helical_symmetry_print_extra_info(helix_onestart)
        !endif

        ! The address of the origin
        if (self%is_in_real_space .and. self%object_is_centered_in_box) then
            ori = self%physical_address_of_box_center
        else
            ori = [1,1,1]
        endif

        ! Prepare the interpolation window
        call win%AllocateWindow(interpol_scheme%window_width,self%IsInFourierSpace(),self%IsAVolume())

        ! Handle the axial symmetry first - need it applied to the asymetric unit
        if (helix%axial_symmetry .gt. 1) then

            tmpimg = self

            ! High-pass filter to compensate for low-pass-filtering effect of interpolation
            if (preemptive_filter .and. IsOdd(helix%axial_symmetry)) then
                if (tmpimg%IsInRealSpace()) call tmpimg%ForwardFFT()
                call correct_3d(tmpimg,interpol_scheme%interpolation_instr_function,interpol_scheme%windowing_instr_function, &
                                interpol_scheme%window_width)
                call tmpimg%BackwardFFT()
            else
                write(*,'(a)') '**info(image_impose_helical_symmetry_new): preemptive filter turned off because even sym'
            endif

            ! Loop over voxels
            do k=1,self%logical_dimensions(3)
                z = real(k-ori(3),kind=8)
                zz = z
                do j=1,self%logical_dimensions(2)
                    y = real(j-ori(2),kind=8)
                    ysq = y**2
                    do i=1,self%logical_dimensions(1)
                        x = real(i-ori(1),kind=8)
                        ! Radius
                        r = sqrt(x**2+ysq)
                        if (r .gt. mmax_radius) cycle
                        ! Work out the longitude (phi)
                        phi = helix_phi_from_xy(x,y)
                        ! Are we in the asymetric unit?
                        in_asym_unit = helix_point_is_in_asym(helix_onestart,z,phi)
                        if (in_asym_unit) then
                            tmpval = self%real_values(i,j,k) / real(helix%axial_symmetry)
                            ! If we're in the asymetric unit, apply the axial symmetry
                            do a_sym = 2, helix%axial_symmetry
                                pphi = phi + (a_sym-1)*(360./real(helix%axial_symmetry))
                                ! Convert back to Cartesian coordinates (see euler_to_rotmat)
                                call helix_xy_from_phi_r(phi,r,xx,yy)
                                ! Interpolate a value from those coordinates
                                if (use_new_interpolation_routine) then
                                    call tmpimg%GetRealValueByLinearInterpolationNoBoundsCheckVolume(rvalue, &
                                                                real(xx)+tmpimg%physical_address_of_box_center(1), &
                                                                real(yy)+tmpimg%physical_address_of_box_center(2), &
                                                                real(zz)+tmpimg%physical_address_of_box_center(3))
                                else
                                    win%centered_on = (/ xx,yy,zz /)
                                    call eval_window_bounds(win)
                                    call window_from_image_real(tmpimg,win)
                                    call get_real_from_window(  rvalue,win,   &
                                                                interpol_scheme%interpolation_function,&
                                                                interpol_scheme%windowing_function)
                                endif
                                tmpval = tmpval + rvalue / real(helix%axial_symmetry)
                            enddo
                            self%real_values(i,j,k) = tmpval
                        endif
                    enddo
                enddo
            enddo
        endif

        ! Also, apply the side dyad to the asymmtric unit
        ! TODO : do we need to pre-high-pass to compensate for interpolation?
        if (helix%perpendicular_symmetry .eq. 2) then
            tmpimg = self
            ! Loop over voxels
            do k=1,self%logical_dimensions(3)
                z = real(k-ori(3),kind=8)
                zz = z
                do j=1,self%logical_dimensions(2)
                    y = real(j-ori(2),kind=8)
                    ysq = y**2
                    do i=1,self%logical_dimensions(1)
                        x = real(i-ori(1),kind=8)
                        ! Radius
                        r = sqrt(x**2+ysq)
                        if (r .gt. mmax_radius) cycle
                        ! Work out the longitude (phi)
                        phi = helix_phi_from_xy(x,y)
                        ! Are we in the asymetric unit?
                        in_asym_unit = helix_point_is_in_asym(helix_onestart,z,phi)
                        if (in_asym_unit) then
                            pphi = (-1.0) * phi
                            zz   = (-1.0) * z
                            tmpval = self%real_values(i,j,k) * 0.5
                            call helix_xy_from_phi_r(pphi,r,xx,yy)
                            call tmpimg%GetRealValueByLinearInterpolationNoBoundsCheckVolume(rvalue, &
                                            real(xx)+tmpimg%physical_address_of_box_center(1), &
                                            real(yy)+tmpimg%physical_address_of_box_center(2), &
                                            real(zz)+tmpimg%physical_address_of_box_center(3))
                            tmpval = tmpval + rvalue * 0.5
                            self%real_values(i,j,k) = tmpval
                        endif
                    enddo
                enddo
            enddo
        endif

        ! High-pass filter to compensate for low-pass-filtering effect of interpolation
        if (preemptive_filter) then
            tmpimg = self
            if (self%IsInRealSpace()) call self%ForwardFFT()
            call correct_3d(self,interpol_scheme%interpolation_instr_function,interpol_scheme%windowing_instr_function, &
                            interpol_scheme%window_width)
            call self%BackwardFFT()
        else
            write(*,'(a)') '**warning(image_impose_helical_symmetry_new): preemptive filter turned off'
        endif



        !
        ! Main loop over the output voxels
        !
        do k=1,self%logical_dimensions(3)
            z = real(k-ori(3),kind=8)
            do j=1,self%logical_dimensions(2)
                y = real(j-ori(2),kind=8)
                ysq = y**2
                do i=1,self%logical_dimensions(1)
                    x = real(i-ori(1),kind=8)
                    ! Radius
                    r = sqrt(x**2+ysq)
                    if (r .gt. mmax_radius) cycle
                    !count_total_voxels = count_total_voxels + 1
                    ! Work out the longitude (phi)
                    phi = helix_phi_from_xy(x,y)
                    ! Are we in the asymetric unit?
                    in_asym_unit = helix_point_is_in_asym(helix_onestart,z,phi)
                    ! If we were in the asymetric unit, we're done
                    if (in_asym_unit) then
                        ! Nothing to do
                        !count_successful_voxels = count_successful_voxels + 1
                        if (debug) subunit_voxels%real_values(i,j,k) = 1.0
                    else
                        ! If we aren't in the asymetric unit, we need to find the corresponding point within the asymetric unit
                        ! zz and phi are modified
                        call helix_point_to_asym(helix_onestart,z,phi,zz,pphi)
                        !print *, '**debug(image_impose_helical_symmetry_new):  z,  phi = ', z, phi
                        !print *, '**debug(image_impose_helical_symmetry_new): zz, pphi = ', zz, pphi
                        ! Convert back to Cartesian coordinates (see euler_to_rotmat)
                        call helix_xy_from_phi_r(pphi,r,xx,yy)
                        ! Interpolate a value from those coordinates
                        if (use_new_interpolation_routine) then
                            call self%GetRealValueByLinearInterpolationNoBoundsCheckVolume(rvalue, &
                                                                real(xx)+self%physical_address_of_box_center(1), &
                                                                real(yy)+self%physical_address_of_box_center(2), &
                                                                real(zz)+self%physical_address_of_box_center(3))
                        else
                            win%centered_on = (/ xx,yy,zz /)
                            call eval_window_bounds(win)
                            call window_from_image_real(self,win)
                            call get_real_from_window(  rvalue,win,   &
                                                        interpol_scheme%interpolation_function,interpol_scheme%windowing_function)
                        endif
                        if (debug) then
                            !if (rvalue .eq. 0.0) then
                            !    write(*,'(a,4(F0.2,1x))') '**debug(image_impose_helical_symmetry_new): zero at xx, yy, zz, r = ', &
                            !                                    xx,yy,zz, r
                            !endif
                        endif
                        self%real_values(i,j,k) = rvalue
                    endif
                enddo
            enddo
        enddo

        if (preemptive_filter) then
            ! We need to restore the original asymetric unit, because currently, there is a high-passed version in its place
            do k=1,self%logical_dimensions(3)
                z = real(k-ori(3),kind=8)
                do j=1,self%logical_dimensions(2)
                    y = real(j-ori(2),kind=8)
                    ysq = y**2
                    do i=1,self%logical_dimensions(1)
                        x = real(i-ori(1),kind=8)
                        ! Radius
                        r = sqrt(x**2+ysq)
                        if (r .gt. mmax_radius) cycle
                        ! Work out the longitude (phi)
                        phi = helix_phi_from_xy(x,y)
                        ! Are we in the asymetric unit?
                        in_asym_unit = helix_point_is_in_asym(helix_onestart,z,phi)
                        ! If we aren't in the asymetric unit, we need to find the corresponding point within the asymetric unit
                        if (in_asym_unit) then
                            ! Restore the data
                            self%real_values(i,j,k) = tmpimg%real_values(i,j,k)
                        endif
                    enddo
                enddo
            enddo
        endif

        ! Undo the Fourier-space padding
        if (fourier_pad_factor .gt. 1) then
            write(*,'(a,i0)') '**info(image_impose_helical_symmetry_new): cropping volume in Fourier space by factor of ', &
                                fourier_pad_factor
            if (self%IsInRealSpace()) call self%ForwardFFT()
            new_dim = self%logical_dimensions / fourier_pad_factor
            call self%Resize(new_dim)
            call self%BackwardFFT()
            if (debug) then
                call subunit_voxels%ForwardFFT()
                call subunit_voxels%Resize(new_dim)
                call subunit_voxels%ForwardFFT()
            endif
        endif

        if (debug) then
            call subunit_voxels%WriteToDisk('dbg_subunit.mrc')
        endif

    end subroutine image_impose_helical_symmetry_new

    !>  \brief  Equivalent to hextend in frealix1/frealign, this subroutine symmetrises the volume using a helical symmetry operator
    !!
    !!          Warning: this subroutine has not been fully tested with helices involving axial or perpendicular symmetries
    !!
    !!  \todo   Test & debug symmetries (axial & perpendicular)
    subroutine image_impose_helical_symmetry(self,helix,interpol_scheme,num_central_slices,max_radius,progress_report)
        use rotation_matrices
        use interpolation
        use omp_lib
        use images_arithm
        use images_threed
        use imagefiles  ! debug
        use ui, only : interval_print
        use UsefulFunctions, only : IsEven
        use InterpolatingWindows
        implicit none
        ! Arguments
        type(image),                        intent(inout)   ::  self                !<  Image (volume) to be symmetrised
        type(helical_symmetry),             intent(in)      ::  helix               !<  Helical symmetry operator
        type(interpolation_scheme),         intent(in)      ::  interpol_scheme     !<  Interpolation scheme
        integer,    optional,               intent(in)      ::  num_central_slices  !<  The number of slices (starting from the central slice), which will be symmetrised into the volume. If not given, the full volume is symmetrised.
        real,       optional,               intent(in)      ::  max_radius          !<  Maximum outer radius of helix (pixels)
        logical,    optional,               intent(in)      ::  progress_report     !<  Give some output regularly to let the user know we're working
        ! Private variables
        integer                 ::  z_start, z_end, current_z   ! the first and last slices to include in the symmetrisation
        type(image)             ::  output_vol, slice
        integer                 ::  origin(3)
        integer                 ::  a_sym, p_sym, z_shift
        real(kind=8)            ::  z_shift_r1, z_shift_r2
        integer                 ::  i,j,k
        real(kind=8)            ::  rot1,rot2,rot3, matrix3(4,4)
        logical                 ::  tthreading  ! Whether we're threading
        real(kind=8)            ::  mmax_radius
        integer                 ::  current_iter,total_iter
        integer                 ::  nnum_central_slices
        integer                 ::  first_central_slice,last_central_slice
        real(kind=8)            ::  rad(2), mmax_radius2, average
        integer                 ::  voxcount
        logical,    parameter   ::  debug   =   .false.
        logical                 ::  pprogress_report
        logical,    parameter   ::  do_nothing = .false.
        logical                 ::  guess_num_central_slices
        type(window),   target  ::  win             !<  Interpolating window
        ! Start work
        ! Check the image is ready
        if (.not. self%IsAllocated()) then
            write(*,'(a)') '**ERROR(image_impose_helical_symmetry): image is not allocated'
            call terminate('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY')
        endif
        if (any(self%logical_dimensions .eq. 1)) then
            write(*,'(a,3i4)') '**ERROR(image_impose_helical_symmetry): image is not 3D. Dim = ', self%logical_dimensions
            call terminate('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY')
        endif
        ! We need the helical rise parameter in pixels, not Angstroms
        if (.not. helix%pixel_units) then
            write(*,'(a)') '**ERROR(image_impose_helical_symmetry): please supply helix rise parameter in pixels, not Angstroms.'
            call terminate('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY')
        endif
        ! Number of central slices
        guess_num_central_slices = .false.
        if (.not. present(num_central_slices)) then
            guess_num_central_slices = .true.
        else
            guess_num_central_slices = num_central_slices .le. 1
        endif
        if (present(num_central_slices) .and. .not. guess_num_central_slices) then
            ! We will assume the user asks for an odd number of central slices
            if (IsEven(num_central_slices)) then
                write(*,'(a,i0)') '**ERROR(image_impose_helical_symmetry): need an odd number of central slices. You asked for: ', &
                                    num_central_slices
                call terminate('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY')
            endif
            ! The number of central slices should be .GE. the rise, otherwise, we don't even have a full description of the central helical repeat
            if (num_central_slices .lt. helix%rise) then
                write(*,'(a,i0,a,F0.3,a)') '**ERROR(image_impose_helical_symmetry): the number of central slices (',    &
                                             num_central_slices, ') is less than the helical rise (', helix%rise, ')'
                call terminate('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY')
            endif
            nnum_central_slices = num_central_slices
        else
            !nnum_central_slices = max(ceiling(1.1*HELIX%rise),5)
            !nnum_central_slices = ceiling(helix_guess_subunit_height(helix))
            nnum_central_slices = ceiling(helix%fast_helix_rise)
            ! We cap at 1/2 of z height of volume
            nnum_central_slices = min(nnum_central_slices,self%logical_dimensions(3)/2)
            if (IsEven(nnum_central_slices)) nnum_central_slices = nnum_central_slices + 1 ! We make sure that nnum_central_slices is odd
        endif
        write(*,'(a,i0,a)') '**info(image_impose_helical_symmetry): imposing helical symmetry with ', &
                                                                nnum_central_slices, ' central slices'

        !i = ceiling(helix_guess_subunit_height(helix))
        i = ceiling(helix%fast_helix_rise)
        if (IsEven(i)) i = I+1
        if (debug) write(*,'(a,i0)') '**debug(image_impose_helical_symmetry): subunit height = ', i
        if (nnum_central_slices .lt. i) then
            write(*,'(a,i0,a,i0)') '**info(image_impose_helical_symmetry): number of central slices (',nnum_central_slices   &
                                    ,') may be too small. Perhaps try ', i
        endif

        ! Did the user ask for progress reports?
        if (present(progress_report)) then
            pprogress_report = progress_report
        else
            pprogress_report = .false.
        endif

        ! Assumption that the volume is even-dimensioned
        !if (any(is_odd(self%logical_dimensions))) then
        !   write(*,'(a,3i5)') '**ERROR(image_impose_helical_symmetry): volume dimensions need to be even: ', self%logical_dimensions
        !   call terminate('Fatal error in IMAGE_IMPOSE_HELICAL_SYMMETRY')
        !endif

        !
        ! Will we thread?
        !
        if (product(self%logical_dimensions) .ge. 64**3 .and. self%logical_dimensions(3) .gt. 1) then
            tthreading = .true.
        else
            tthreading = .false.
        endif
        ! If we're already in a threaded region, override, since ifort does not support nested parallelism
#ifdef _OPENMP
        if (omp_get_active_level() .gt. 0) then
            !write(*,'(a)') '**info(image_impose_helical_symmetry): already in a parallel scheme, so will disable local threading, since nesting not supported'
            tthreading = .false.
        endif
#else
        tthreading = .false.
#endif

        !
        ! Maximum radius
        !
        if (present(max_radius)) then
            if (debug) print '(a,F0.3)', '**debug(image_impose_helical_symmetry): optional max_radius argument was present: ', &
                                        max_radius
            if (max_radius .eq. 0.0) then
                mmax_radius = minval(self%logical_dimensions)*0.5 - 6.
            else
                mmax_radius = max_radius
            endif
        else
            mmax_radius = minval(self%logical_dimensions)*0.5 - 6.
        endif

        !
        ! Work out the index of the first and last central slices
        !
        if (self%is_in_real_space .and. self%object_is_centered_in_box) then
            origin = self%physical_address_of_box_center
        else
            origin = [1,1,1]
        endif
        first_central_slice = origin(3)-(NNUM_CENTRAL_SLICES-1)/2
        last_central_slice  = origin(3)+(NNUM_CENTRAL_SLICES-1)/2

        !
        ! Can't go beyond edges
        !
        if (first_central_slice .lt. 1) then
            write(*,'(a)') '**warning(image_impose_helical_symmetry): number of central slices to great. '//&
                                                                            'Probably programming error.'
            first_central_slice = 1
        endif
        if (last_central_slice .gt. self%logical_dimensions(3)) then
            write(*,'(a)') '**warning(image_impose_helical_symmetry): number of central slices to great. '//&
                                                                            'Probably programming error.'
            last_central_slice = self%logical_dimensions(3)
        endif

        !
        ! Work out the average value outside the cylinder in the central sections which will be used for symmetrisation.
        !
        mmax_radius2 = (mmax_radius)**2
        average = 0.0
        voxcount = 0
        do K=FIRST_CENTRAL_SLICE,last_central_slice
            do J=1,self%logical_dimensions(2)
                rad(2) = real(J-ORIGIN(2))**2
                do I=1,self%logical_dimensions(1)
                    rad(1) = real(I-ORIGIN(1))**2 + rad(2)
                    if (rad(1) .gt. mmax_radius2) then
                        average = average + self%real_values(i,j,k)
                        voxcount = voxcount + 1
                    endif
                enddo
            enddo
        enddo
        average = average / real(voxcount)

        ! Subtract this value from the volume, so that the background goes to 0.0
        if (debug) write(*,'(a,F0.6)') '**debug(image_impose_helical_symmetry): average value outside cylinder = ', average
        self%real_values = self%real_values - average

        !
        ! Zeroe voxels beyond the maximum radius
        !
        call cyl_mask(self,real(mmax_radius)+3.,6)


        if (debug) then
            call self%WriteToDisk('dbg_cyl_pre_h.mrc')
        endif

        ! Allocate output image
        call output_vol%Allocate(self)

        if (do_nothing) then
            write(*,'(a)') '**warning(image_impose_helical_symmetry): not doing anything'
            output_vol%real_values = self%real_values
        else

        ! We will have to scan the central region over the volume from bottom to top, such that every slice of the output will have
        ! "seen" nnum_central_slices additions to it. For this, the central region's center will start at z_start and will finish at
        ! z_end
        z_start = 1 - (NNUM_CENTRAL_SLICES-1)/2
        z_end = self%logical_dimensions(3) + (NNUM_CENTRAL_SLICES-1)/2
        total_iter = Z_END-Z_START+1


        ! Deal with symmetry perpendicular to the helical/z axis
        do P_SYM=1,helix%perpendicular_symmetry
            ! Rotation around an axis perpendicular to the h axis by (360/HELIX%perpendicular_symmetry)*(P_SYM-1)
            rot1 = (360/HELIX%perpendicular_symmetry)*(P_SYM-1)

            ! Deal with symmetry around the helical/z axis
            do a_sym = 1, HELIX%axial_symmetry
                ! Rotation around the h axis by (360/HELIX%axial_symmetry)*(A_SYM-1)
                rot2 = (360/HELIX%axial_symmetry)*(A_SYM-1)

                ! Setup omp threading

                !$omp   parallel    if(tthreading)  &
                !$omp               default(shared)   &
                !$omp               private(rot3,matrix3,z_shift_r1,z_shift_r2,current_z,current_iter,win)  &
                !$omp               private(slice)

#ifdef _OPENMP
                !$omp   single
                if (tthreading .and. omp_get_num_threads() .gt. 1 .and. debug) then
                    write(*,'(a,i1,a)') '**debug(image_impose_helical_symmetry): OpenMP with ', omp_get_num_threads(), ' threads'
                endif
                !$omp   end single
#endif

                call win%AllocateWindow(interpol_scheme%window_width,self%IsInFourierSpace(),self%IsAVolume())
                slice%is_in_memory = .false.
                slice%planned = .false.
                slice%logical_dimensions = self%logical_dimensions
                slice%logical_dimensions(3) = 1
                call slice%Allocate()

                ! Move the central region from bottom to top of the volume (z_shift is the z position within the output volume of its central slice, which is also the central slice of the original volume)
                !$omp do
                do z_shift = z_start, z_end
                    current_iter = z_shift - z_start + 1
                    if ((debug .or. pprogress_report) .and. interval_print(current_iter,total_iter)) then
                        !$omp   critical        (printout_progress_report)
                        write(*,'(4(A,i0))',advance='NO') '**info(image_impose_helical_symmetry): on ', current_iter, &
                                                            ' of ', total_iter, ' for axial symmetry ', a_sym, &
                                                            ' and perpendicular symmetry ', p_sym
#ifdef _OPENMP
                        if (tthreading) then
                            write(*,'(a,i0,a,i0)') ' on thread ', omp_get_thread_num()+1, ' of ', omp_get_num_threads()
                        else
                            write(*,'(a)') ' '
                        endif
#else
                        write(*,'(a)') ' '
#endif
                        !$omp   end critical    (printout_progress_report)
                    endif
                    !print '(a,i,a)', '**debug(image_impose_helical_symmetry): thread ', omp_get_thread_num(), ' working...'

                    ! The distance from the center of the image to the center of the region of interest
                    z_shift_r1 = real(Z_SHIFT-ORIGIN(3))
                    ! The additional shift we'd need to move to the nearest repeat position
                    z_shift_r2 = (nint(z_shift_r1 / helix%rise) * helix%rise) - z_shift_r1

                    ! Work out how much the image should be rotated by.
                    ! Note that the helical twist convention in frealix is such that a clockwise rotation
                    ! about the z axis (when looking from +Z towards origin) is a positive angle, giving rise
                    ! to a left-handed helix.
                    rot3 = (z_shift_r1 + z_shift_r2) * helix%twist/helix%rise

                    ! Diagnostic
                    !print *, '**debug(image_impose_helical_symmetry): ', z_shift, rot

                    ! Rotation around the helical axis (rotate_image expects the matrix that transforms from output to input, this is why REVERSE=.TRUE.)
                    matrix3 = euler_to_rotmat(ROT3+ROT2,rot1,0.0D0,REVERSE=.true.)

                    ! Add the rotated central region to the volume
                    do I=1,nnum_central_slices
                        current_z = z_shift + (I-((NNUM_CENTRAL_SLICES-1)/2+1))
                        if (current_z .lt. 1 .or. current_z .gt. self%logical_dimensions(3)) then
                            cycle
                        else
                            matrix3(3,4) = -Z_SHIFT_R2 + real(I-((NNUM_CENTRAL_SLICES-1)/2+1))
                            ! Extract 2d slice from near the center of the input volume
                            call extract_slice(self,slice,0.0D0,0.0D0,0.0D0,interpol_scheme,win,matrix3,RAD_MAX=MMAX_RADIUS)
                            !$omp critical (img_imp_h_sym)
                            output_vol%real_values(:,:,current_z) = output_vol%real_values(:,:,current_z) + slice%real_values(:,:,1)
                            !$omp end critical (img_imp_h_sym)
                        endif
                    enddo
                enddo
                !$omp   end do
                call slice%Deallocate()
                call win%DeallocateWindow()
                !$omp   end parallel
            enddo
        enddo

        endif   ! End of check for do_nothing


        if (debug) then
            write(*,'(a)') '**debug(image_impose_helical_symmetry): about to normalise densities'
            write(*,'(a,2l1)') '**debug(image_impose_helical_symmetry): self & output_vol same dim?, self allocated?: ', &
                                self%HasSameDimensionsAs(output_vol), self%IsAllocated()
            write(*,'(a,2(i0,2x), 2l1)') '**debug(image_impose_helical_symmetry): self & output_vol size, allocated: ', &
                                    size(self%real_values), size(output_vol%real_values), &
                                    self%IsAllocated(), output_vol%IsAllocated()
            write(*,'(2a,3(i0,2x))') '**debug(image_impose_helical_symmetry): nnum_central_slices, ',    &
                                    'HELIX%axial_symmetry, helix%perpendicular_symmetry =', nnum_central_slices, &
                                    helix%axial_symmetry, helix%perpendicular_symmetry

            call self%PrintInfo()
            call output_vol%PrintInfo()
        endif
        ! Normalise the densities
        self%real_values = output_vol%real_values / real(NNUM_CENTRAL_SLICES* helix%axial_symmetry * helix%perpendicular_symmetry)

        call output_vol%Deallocate()

        ! Diagnostic
        if (debug) write(*,'(a,3i4)') '**debug(image_impose_helical_symmetry): symmetrised image dimensions = ', &
                                        self%logical_dimensions
        write(*,'(a)',advance='NO') '**info(image_impose_helical_symmetry): used following helical symmetry: '
        call helical_symmetry_print_info(helix)
    end subroutine image_impose_helical_symmetry

    !>  \brief  Given an amplitude spectrum, look for meridional layer lines
    subroutine scan_for_meridonial_layer_line(  img_ampl,img_phas,plot_id,expected_frequency,expected_frequency_halfrange, &
                                                top_scores,top_frequencies,top_phase_diffs)
        use images
        use plots
        use euleranglesmod, only : angle_difference
        implicit none
        ! Arguments
        type(image),                intent(in)      ::  img_ampl                        !<  Amplitudes of image
        type(image),                intent(in)      ::  img_phas                        !<  Phases of image (must be in degrees)
        integer,                    intent(in)      ::  plot_id                         !<  Index for filename of output png plot
        real,                       intent(in)      ::  expected_frequency              !<  Frequency (1/pixel; 0.5 is Nyquist) at which layer line is expected
        real,                       intent(in)      ::  expected_frequency_halfrange    !<  Halfrange for frequency range in which layer line is expected
        real,   allocatable,        intent(inout)   ::  top_scores(:)                   !<  Will return the scores (number of std dev above local background) of the layer line candidates
        real,   allocatable,        intent(inout)   ::  top_frequencies(:)              !<  Will return the frequencies of the layer line candidates
        real,   allocatable,        intent(inout)   ::  top_phase_diffs(:)              !<  The average phase difference between pixels on the lhs and rhs of the meridian
        ! Private variables
        integer                 ::  i_logi,j_logi
        integer                 ::  i_logi_min, i_logi_max, i_phys_min, i_phys_max
        integer                 ::  j_logi_min, j_logi_max, j_phys_min, j_phys_max
        integer                 ::  lbounds(3), ubounds(3)
        integer                 ::  window_dimensions(2)
        integer                 ::  img_ampl_origin(3)
        real, allocatable       ::  average_density(:), average_density_std_series(:)
        real, allocatable       ::  average_phase_diff(:)
        real(kind=8)            ::  backg_density_ave, backg_density_std
        real, allocatable       ::  series_to_plot(:,:)
        character(len=64)       ::  plot_filename
        integer                 ::  num_in_stats
        real                    ::  j_freq
        integer                 ::  j_phys
        integer,    parameter   ::  number_of_lines =   2   ! The top number_of_lines layer lines will be returned
        integer,    parameter   ::  layer_line_width  = 21   ! In pixels
        integer,    parameter   ::  layer_line_height = 3
        real                    ::  current_score
        integer                 ::  num_top_scores_beat_by_current, current_score_position
        integer                 ::  pos
        character(len=3)        ::  size_top_scores
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        lbounds = img_ampl%logical_lower_bound_real
        ubounds = img_ampl%logical_upper_bound_real
        if (img_ampl%is_in_real_space .and. img_ampl%object_is_centered_in_box) then
            img_ampl_origin = img_ampl%physical_address_of_box_center
        else
            img_ampl_origin = [1,1,1]
        endif
        window_dimensions = [LAYER_LINE_WIDTH,LAYER_LINE_HEIGHT]
        !
        allocate(   average_density(ubounds(2)),    &
                    average_density_std_series(ubounds(2)), &
                    series_to_plot(ubounds(2),1),   &
                    average_phase_diff(ubounds(2)))
        !
        i_logi_min = -(window_dimensions(1)-1)/2
        i_logi_max = -I_LOGI_MIN
        i_phys_min = i_logi_min + img_ampl_origin(1)
        i_phys_max = i_logi_max + img_ampl_origin(1)

        ! Loop along the second axis, from origin to edge
        backg_density_ave = 0.0D0
        backg_density_std = 0.0D0
        num_in_stats = 0
        do j_logi = 0, UBOUNDS(2)
            ! The addresses over which we'll place our window
            j_logi_min = max(j_logi - (window_dimensions(2)-1)/2,0)
            j_logi_max = min(j_logi + (window_dimensions(2)-1)/2,UBOUNDS(2))
            j_phys_min = j_logi_min + img_ampl_origin(2)
            j_phys_max = j_logi_max + img_ampl_origin(2)
            ! Where we are in frequency
            j_freq = real(j_logi) / real(img_ampl%logical_dimensions(2))
            !
            if (j_logi .ge. 1) then
                ! Average amplitude at the current frequency
                average_density(j_logi) = sum(img_ampl%real_values(i_phys_min:i_phys_max,j_phys_min:j_phys_max,1))   &
                                            / product(window_dimensions)
                ! Average phase difference between pixels on lhs and pixels on rhs of meridian
                average_phase_diff(j_logi) = 0.0
                do I_LOGI=I_LOGI_MIN,-1 ! We loop over the lhs, and address the rhs in the 2nd arg to angle_difference
                    average_phase_diff(j_logi) = average_phase_diff(j_logi) + &
                        sum(angle_difference(dble(img_phas%real_values(img_ampl_origin(1)+I_LOGI,j_phys_min:j_phys_max,1)),   &
                                             dble(img_phas%real_values(img_ampl_origin(1)-I_LOGI,j_phys_min:j_phys_max,1)) )) &
                       / product(window_dimensions)
                enddo
                ! If we're near where we expect the layer line
                if (j_freq .gt. expected_frequency - 0.15 .and. j_freq .lt. expected_frequency + 0.15) then
                    if (     j_freq .lt. expected_frequency - expected_frequency_halfrange   &
                        .or. j_logi .gt. expected_frequency + expected_frequency_halfrange) then
                        num_in_stats = num_in_stats + 1
                        backg_density_ave = backg_density_ave + average_density(j_logi)
                        backg_density_std = backg_density_std + average_density(j_logi)**2
                    endif
                endif
            endif
        enddo

        ! Stats for the background near the layer line
        backg_density_ave = backg_density_ave / real(num_in_stats)
        backg_density_std = backg_density_std / real(num_in_stats)
        backg_density_std = sqrt(backg_density_std - BACKG_DENSITY_AVE**2)
        if (debug) write(*,'(a,G15.5,1x,G15.5)') '**debug(): background average, std = ', backg_density_ave, backg_density_std

        average_density_std_series = backg_density_ave + 2.5*BACKG_DENSITY_STD

        ! Plot result
        write(plot_filename,'(a,I5.5,a)') 'ave_meri_density_',plot_id,'.png'
        series_to_plot(:,1) = average_density(:)
        call series_to_png(plot_filename,series_to_plot,OVERLAY_YS=AVERAGE_DENSITY_STD_SERIES,Y_MIN=0.0,Y_MAX=3.5*.01)
        !write(*,'(i5,x,F10.3)') j_logi,average_density

        ! Look for layer line candidates
        if (allocated(top_scores)) deallocate(top_scores)
        if (allocated(top_frequencies)) deallocate(top_frequencies)
        if (allocated(top_phase_diffs)) deallocate(top_phase_diffs)
        allocate(top_scores(number_of_lines),top_frequencies(number_of_lines),top_phase_diffs(number_of_lines))
        top_scores = 0.0
        do J_LOGI=1,ubounds(2)
            j_phys = j_logi + img_ampl_origin(2)
            ! Where we are in frequency
            j_freq = real(j_logi) / real(img_ampl%logical_dimensions(2))
            ! If we are within frequency ranges where we've been told to look for layer lines
            if (      j_freq .ge. expected_frequency - expected_frequency_halfrange   &
                .and. j_freq .le. expected_frequency + expected_frequency_halfrange) then
                ! The score is the number of standard deviations above the average for the local average amplitude
                current_score = (average_density(j_logi) - backg_density_ave) / backg_density_std
                num_top_scores_beat_by_current = count(current_score .gt. top_scores)
                if (num_top_scores_beat_by_current .ge. 1) then
                    ! Current score makes it onto the list
                    current_score_position = 1 + (size(top_scores) - num_top_scores_beat_by_current)
                    ! Move all top scores and frequencies that are displaced
                    if (current_score_position .lt. size(top_scores)) then
                        do POS=SIZE(top_scores),CURRENT_SCORE_POSITION+1,-1
                            top_scores(pos) = top_scores(POS-1)
                            top_frequencies(pos) = top_frequencies(POS-1)
                            top_phase_diffs(pos) = top_phase_diffs(POS-1)
                        enddo
                    endif
                    ! Stick the current score and frequency in the list of hits
                    top_scores(current_score_position) = current_score
                    top_frequencies(current_score_position) = j_freq
                    top_phase_diffs(current_score_position) = average_phase_diff(j_logi)
                endif
            endif
        enddo

        ! Print out results
        if (debug) then
        write(size_top_scores,'(i0)') size(top_scores)
        write(*,'(a,'//size_top_scores//'(F0.3,1x))')       '**debug(scan_for_meridionial_layer_line): top scores        = ', &
                                                        top_scores
        write(*,'(a,'//size_top_scores//'(F0.3,1x))')  '**debug(scan_for_meridionial_layer_line): top frequencies   = ', &
                                                        top_frequencies
        write(*,'(a,'//size_top_scores//'(F0.3,1x))') '**debug(scan_for_meridional_layer_line): top average phase differences = ', &
                                                        top_phase_diffs
        endif

    end subroutine scan_for_meridonial_layer_line

end module images_helices
