!>  \brief  Module grouping function to go to/from 3d reconstructions and 2d images
module images_threed

    contains

    !>  \brief  Impose a missing wedge onto a 3d volume or 2d image. The wedge is
    subroutine impose_missing_wedge(img,phi,theta,psi,width)
        use images
        implicit none
        ! Arguments
        type(image),    intent(inout)   ::  img             !<  Volume
        real,           intent(in)      ::  phi, theta, psi !<  The wedge will be imposed after rotation by these Euler angles
        real,           intent(in)      ::  width           !<  The "width" of the wedge, in degrees
        ! Private variables

        !

    end subroutine impose_missing_wedge

    !>  \brief  Project a 3d volume onto a 2d slice (real-space projection) (leaner version, which does not call rotate_image
    !!  \todo   Use euler angle object
    !!  \todo   Use faster interpolation routines
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine volume_project_lean(volume,projection,phi,theta,psi,interpol_scheme,interpol_window, &
                                    super_lean,max_cyl_rad,shift_x,shift_y,z_shift,&
                                    helical_subunit_only,helix)
        use interpolation, only  : interpolation_scheme
        use Images
        use InterpolatingWindows
        use rotation_matrices
        use runtime_parameters, only : h_asym_unit_center, terminate
        use helices, only : helix_point_is_in_asym, helical_symmetry, helix_phi_from_xy
        use omp_lib
        implicit none
        ! Arguments
        type(image),                        intent(in)      ::  volume          !<  Volume to be projected
        type(image),                        intent(inout)   ::  projection      !<  Projection image
        real(kind=8),                       intent(in)      ::  phi             !<  Phi Euler angle (degrees)
        real(kind=8),                       intent(in)      ::  theta           !<  Theta Euler angle (degrees)
        real(kind=8),                       intent(in)      ::  psi             !<  Psi Euler angle (degrees)
        type(interpolation_scheme),         intent(in)      ::  interpol_scheme !<  Interpolation scheme
        type(window),                       intent(inout)   ::  interpol_window !<  Window for interpolation
        logical,                optional,   intent(in)      ::  super_lean      !<  Whether we go super-lean, which basically means skipping all safety checks etc.
        real,                   optional,   intent(in)      ::  max_cyl_rad     !<  If specified, any voxels beyond this radius will be ignored (pixels). The radius of a cylinder aligned with z axis.
        real(kind=8),           optional,   intent(in)      ::  shift_x         !<  In pixels, shift to be applied to the slice after projection
        real(kind=8),           optional,   intent(in)      ::  shift_y         !<  In pixels, shift to be applied to the slice after projection
        real(kind=8),           optional,   intent(in)      ::  z_shift         !<  In pixels, the volume will be shifted by this amount on along the z axis before slice extraction (or this is what it'll look like)
        logical,                optional,   intent(in)      ::  helical_subunit_only    !<  Whether only the helical asymetric unit should be projected
        type(helical_symmetry), optional,   intent(in)      ::  helix
        ! Private variables
        real            ::  rotation(4,4)
        integer         ::  i,j,k
        integer         ::  origin(3)
        real(kind=8)            ::  input_coos(4)
        real                    ::output_coos(4), rvalue
        real    ::  max_rad, rad(3)
        logical ::  threading
        logical ::  ssuper_lean
        real    ::  mmax_cyl_rad
        logical,    parameter   ::  debug   =   .false.
        real(kind=8)    ::  sshift_x, sshift_y
        logical ::  project_this_point
        ! Variables related to the projection of helical subunit only
        logical ::  hhelical_subunit_only
        real(kind=8)    ::  z,h_phi, zz_shift
        ! Start work

        ! Helical subunit only?
        hhelical_subunit_only = .false.
        if (present(helical_subunit_only)) hhelical_subunit_only = helical_subunit_only
#ifndef skip_runtime_checks
        if (hhelical_subunit_only) then
            if (.not. present(helix)) then
                write(*,'(2a)') '**ERROR(volume_project_lean): helical operator or pitch are missing. ', &
                                'Cannot implement helical_subunit_only option.'
                call terminate('Fatal error in VOLUME_PROJECT_LEAN. Programming error')
            endif
        endif
#endif

        ! Dimensions
        if (        projection%GetLogicalDimension(1) .ne. volume%GetLogicalDimension(1) &
            .or.    projection%GetLogicalDimension(2) .ne. volume%GetLogicalDimension(2)) then
            call projection%Allocate(dims=[volume%GetLogicalDimension(1),volume%GetLogicalDimension(2),1])
        endif

        ! Shifs in XY?
        sshift_x = 0.0D0
        sshift_y = 0.0D0
        zz_shift = 0.0D0
        if (present(shift_x)) sshift_x = shift_x
        if (present(shift_y)) sshift_y = shift_y
        if (present(z_shift)) zz_shift = z_shift

        !
        ! Are we going super lean?
        !
        ssuper_lean = .false.
        if (present(super_lean)) ssuper_lean = super_lean


        !
        ! Are we given a maximum radius?
        !
        mmax_cyl_rad = 99999999.99
        if (present(max_cyl_rad)) then
            mmax_cyl_rad = MAX_CYL_RAD**2
        endif

#ifndef skip_runtime_checks
        !
        ! We only do the nice safety check if the user didn't ask for super lean
        !
        if (.not. ssuper_lean .or. debug) then

            ! Check input volume is in memory
            if (.not. volume%is_in_memory) then
                write(*,'(a)') '**ERROR(volume_project_lean): input volume is not in memory'
                call terminate('Fatal error in VOLUME_PROJECT_LEAN')
            endif

            ! Check input volume really is a volume
            if ( any(volume%logical_dimensions .eq. 1) ) then
                write(*,'(a,3(i4,1x))') '**ERROR(volume_project_lean): input volume is not a volume, it has dimensions ', &
                                        volume%logical_dimensions
                call terminate('Fatal error in VOLUME_PROJECT_LEAN')
            endif

            ! If the slice doesn't have the correct dimensions, deallocate
            if (any(volume%logical_dimensions(1:2) .ne. projection%logical_dimensions(1:2))) then
                !write(*,'(a,3(i4,1x),a,3(i4,1x),a)') '**ERROR(volume_project_lean): dimensions of volume (', &
                !                                    volume%logical_dimensions, &
                !                                    ') do not match that of projection (', projection%logical_dimensions, ')'
                !call terminate('Fatal error in VOLUME_PROJECT_LEAN')
                call projection%Deallocate()
            endif

            ! If slice not in memory, allocate to correct dimensions.
            if ( .not. projection%is_in_memory) then
                if ( all(projection%logical_dimensions .eq. 1) ) then
                    projection%logical_dimensions = (/volume%logical_dimensions(1),VOLUME%logical_dimensions(2),1/)
                endif
                call projection%Allocate()
                projection%is_in_real_space = volume%is_in_real_space
                projection%object_is_centered_in_box = volume%object_is_centered_in_box
            endif

            ! We assume that the origin is at the center of the volume
            if (.not. volume%object_is_centered_in_box) then
                write(*,'(a)') '**ERROR(volume_project_lean): input volume does not have origin at its center'
                call terminate('Fatal error in VOLUME_PROJECT_LEAN')
                !if (volume%type .EQ. 'REAL') call volume%ForwardFFT()
                !call volume%ShiftRealSpaceOrigin()
                !if (volume%type .EQ. 'COMP') call volume%BackwardFFT()
            endif


            ! We may need to go back to real space
            if (.not. volume%is_in_real_space) then
                write(*,'(a)') '**ERROR(volume_project_lean): input volume must be in real space'
                call terminate('Fatal error in VOLUME_PROJECT_LEAN')
                !call volume%BackwardFFT()
            endif


            ! We assume even dimensions
            if (any(mod(real(volume%logical_dimensions),2.0) .ne. 0.0)) then
                write(*,'(a,3(i4,1x))') '**ERROR(volume_project_lean): input volume does not have even dimensions (',   &
                                        volume%logical_dimensions, ')'
                call terminate('Fatal error in VOLUME_PROJECT_LEAN')
            endif

        endif   !   End of check for .NOT. ssuper_lean
#endif

        ! Check type
        if (.not. projection%IsInSameSpaceAs(volume)) then
            projection%is_in_real_space = volume%is_in_real_space
        endif

        ! The projection will have origin at center, since we assume the input volume had origin at center
        projection%object_is_centered_in_box = .true.

        !
        ! End of safety checks, start real work
        !

        ! Get the appropriate rotation matrix (rotate_image needs to matrix to go from output back to input, which is why we
        ! use REVERSE=.TRUE.)
        rotation = euler_to_rotmat(phi,theta,psi,REVERSE=.true.)

        ! Initialise the projection
        projection%real_values = 0.0

        ! Get the origin address for the volume
        if (volume%is_in_real_space .and. volume%object_is_centered_in_box) then
            origin = volume%physical_address_of_box_center
        else
            origin = [1,1,1]
        endif


        ! Coordinates initialised
        input_coos = 0.0
        output_coos = 0.0
        input_coos(4) = 1.0
        output_coos(4) = 1.0

        max_rad = (real(projection%logical_dimensions(1))/2.-1.0 - real(interpol_scheme%window_width)/2. - abs(zz_shift))**2
        rad = 0.0

#ifdef low_level_threading
        threading = .true.
#else
        threading = .false.
#endif


#ifdef low_level_threading
        !$omp   parallel    if(threading)   &
        !$omp               default(none)   &
        !$omp               private(i,j,k,output_coos,input_coos,rad,rvalue,h_phi,z,project_this_point)    &
        !$omp               shared(projection,volume,rotation,origin,interpol_scheme,threading,max_rad) &
        !$omp               shared(mmax_cyl_rad,sshift_x,sshift_y,zz_shift,hhelical_subunit_only,helix)

        !$omp   single
        if (threading .and. omp_get_num_threads() .gt. 1) then
            ! write(*,'(a,i1,a)') '**debug(volume_project_lean): OpenMP threading with ', omp_get_num_threads(), ' threads'
        endif
        !$omp   end single nowait
#endif
        ! Prepare the window
        !call interpol_window%AllocateWindow(interpol_scheme%window_width,volume%IsInFourierSpace(),volume%IsAVolume())

        ! Loop over output pixels
#ifdef low_level_threading
        !$omp   do  !!schedule(static,1)
#endif
        do J=1,projection%logical_dimensions(2)
            output_coos(2) = real(j-origin(2)) - sshift_y
            rad(2) = output_coos(2)**2
            do I=1,projection%logical_dimensions(1)
                output_coos(1) = real(i-origin(1)) - sshift_x
                rad(1) = rad(2) + output_coos(1)**2
                ! Check whether we are within the radius
                if (rad(1) .le. max_rad) then
                    ! Iterate over all the voxels in the ray which lands on the current pixel
                    do K=1,volume%logical_dimensions(3)
                        output_coos(3) = real(k-origin(3))
                        rad(3) = rad(1) + output_coos(3)**2
                        ! Check we are within the radius
                        if(rad(3) .le. max_rad) then
                            ! input_coos are the coordinates before Euler angle rotation is applied
                            input_coos = matmul(rotation,output_coos)
                            input_coos(3) = input_coos(3) - zz_shift
                            ! Check the input voxel is within the cylinder we're interested in
                            if ((input_coos(1))**2+(input_coos(2))**2 .le. mmax_cyl_rad) then
                                ! Check whether the we are at a 0.0 voxel (note, this introduces slight artefacts near boundary of mask)
                                project_this_point = volume%real_values( nint(input_coos(1)+origin(1)),   &
                                                                    nint(input_coos(2)+origin(2)),  &
                                                                    nint(input_coos(3)+origin(3))) .ne. 0.0
                                project_this_point = .true.
                                if (project_this_point .and. hhelical_subunit_only) then
                                    h_phi = helix_phi_from_xy(input_coos(1),input_coos(2))
                                    z     = real(input_coos(3),kind=8)
                                    project_this_point =    project_this_point .and. &
                                                            helix_point_is_in_asym(helix,z,h_phi)
                                endif
                                if (project_this_point) then
                                    ! Center the window on the input voxel
                                    interpol_window%centered_on = input_coos(1:3)
                                    ! Define the bounds of the window over the output image
                                    ! These bounds can take any value, including 0, which means the address of the origin,
                                    ! as well as negative values.
                                    call eval_window_bounds(interpol_window)
                                    call window_from_image_real(volume,interpol_window)
                                    call get_real_from_window(rvalue,interpol_window,   interpol_scheme%interpolation_function,&
                                                                                        interpol_scheme%windowing_function)
                                    projection%real_values(i,j,1) = projection%real_values(i,j,1) + rvalue
                                endif
                            endif
                        endif
                    enddo
                endif
            enddo
        enddo
#ifdef low_level_threading
        !$omp   end do

        !$omp   end parallel
#endif
    end subroutine volume_project_lean


    !>  \brief  Extract a 2d slice from a 3d volume (equivalent to frealign's pextract, but more abstracted/generalised)
    !!  \todo   There is assumption that dimensions are even - add a check for this &/or use origin_addr calls
    !!  \todo   Use euler angles object
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine extract_slice(volume,slice,phi,theta,psi,interpol_scheme,win,rotation,rad_max,z_shift,x_shift,y_shift)
        use images, only : image
        use InterpolatingWindows
        use interpolation, only : interpolation_scheme
        use rotation_matrices, only : euler_to_rotmat
        use runtime_parameters, only : terminate, dpi
        implicit none
        ! Arguments
        type(image),    target,                 intent(in)      ::  volume          !<  Volume to extract slice from
        type(image),                            intent(inout)   ::  slice           !<  Extracted slice
        real(kind=8),                           intent(in)      ::  phi             !<  Phi Euler angle (degrees)
        real(kind=8),                           intent(in)      ::  theta           !<  Theta Euler angle (degrees)
        real(kind=8),                           intent(in)      ::  psi             !<  Psi Euler angle (degrees)
        type(interpolation_scheme),             intent(in)      ::  interpol_scheme !<  Interpolation scheme
        type(window),   target,                 intent(inout)   ::  win             !<  Interpolating window
        real(kind=8),   optional,               intent(in)      ::  rotation(4,4)   !<  Rotation matrix to position slice pixel into volume. If present, overrides Euler angles supplied in other arguments
        real(kind=8),   optional,               intent(in)      ::  rad_max         !<  In pixels. Pixels beyond this radius in the output will not be extracted.
        real(kind=8),   optional,               intent(in)      ::  z_shift         !<  In pixels, the volume will be shifted by this amount on along the z axis before slice extraction (or this is what it'll look like)
        real(kind=8),   optional,               intent(in)      ::  x_shift         !<  In pixels, the slice will be shifted by this amount along the first dimension (after slice extraction)
        real(kind=8),   optional,               intent(in)      ::  y_shift         !<  In pixels, the slice will be shifted by this amount along the second dimension (after slice extraction)
        ! Private variables
        integer ::  i,j
        integer ::  addr_slice_max(3)       !<  Address of the the last element in the slice
        integer ::  addr_slice_origin(3)    !<  Address of the origin of the slice
        !integer    ::  addr_slice_min(2)       !<  Address of the first element in the slice
        integer ::  logi_addr_in_slice(2)       !<  Logical address of current element within the slice
        real    ::  logi_addr_in_slice_r(4)
        real    ::  coos_in_volume(4)       !<  Coordinates of current element in volume
        real    ::  rrotation(4,4)          !<  Rotation matrix used to position pixel into volume
        !integer ::  volume_lbounds(3)        !<  Address of lower bound of volume relative to origin 0,0,0 at center
        !integer ::  volume_ubounds(3)
        integer ::  volume_mid_coos(3)      !<  Address of origin
        real    ::  rvalue
        complex ::  cvalue
        real    ::  rrad_max
        real(kind=8)    ::  zz_shift, phase_z
        logical ::  zzz_shift
        complex(kind=8) ::  phase_shift
        integer ::  xx,yy
        logical ::  xy_shift
        real(kind=8)    ::  xx_shift, yy_shift, phase_x, phase_y, phase
        real(kind=8)    ::  slice_phase_factor_x, slice_phase_factor_y
        real(kind=8)    ::  volume_phase_factor_z


        ! Start work
        logi_addr_in_slice_r = 0.0

#ifndef skip_runtime_checks
        ! Check input volume is in memory
        if (.not. volume%is_in_memory) then
            write(*,'(a)') '**ERROR(extract_slice): input volume is not in memory'
            call terminate('Fatal error in EXTRACT_SLICE')
        endif

        ! Check input volume really is a volume
        if ( any(volume%logical_dimensions .eq. 1) ) then
            write(*,'(a,3(i4,1x))') '**ERROR(extract_slice): input volume is not a volume, it has dimensions ', &
                                    volume%logical_dimensions
            call terminate('Fatal error in EXTRACT_SLICE')
        endif

        ! Volume is assumed to be cubic
        if (.not. volume%IsSquare()) then
            write(*,'(a)') '**ERROR(extract_slice): non-cubic volumes are not supported'
            call terminate('Fatal error in EXTRACT_SLICE')
        endif
#endif

        ! If slice not in memory, allocate to correct dimensions.
        if ( .not. slice%is_in_memory .or. slice%logical_dimensions(1) .ne. volume%logical_dimensions(1) .or. &
                                        slice%logical_dimensions(2) .ne. volume%logical_dimensions(2)) then
            slice%logical_dimensions = (/volume%logical_dimensions(1),VOLUME%logical_dimensions(2),1/)
#ifndef skip_runtime_checks
            call slice%Allocate()
#endif
            slice%is_in_real_space = volume%is_in_real_space
            slice%object_is_centered_in_box = volume%object_is_centered_in_box
        endif

        ! Check type
        if (.not. slice%IsInSameSpaceAs(volume)) then
            !write(*,'(5a)') '**WARNING(extract_slice): volume (', volume%type, ') and slice (', slice%type, ') have different types'
            !write(*,'(a)')  '**WARNING(extract_slice): changing type of slice to match that of volume'
            slice%is_in_real_space = volume%is_in_real_space
        endif
        slice%object_is_centered_in_box = volume%object_is_centered_in_box

#ifndef skip_runtime_checks
        ! If we are in Fourier space, we probably want the origin at 1,1,1 - not center
        if (.not. volume%is_in_real_space .and. volume%object_is_centered_in_box) then
            write(*,'(2a)') '**WARNING(extract_slice): working on a complex volume, ',   &
                                'yet real-space origin is in center. Are you sure?'
        endif
        if (.not. slice%is_in_real_space .and. slice%object_is_centered_in_box) then
            write(*,'(a)') '**WARNING(extract_slice): working on a complex slice, yet real-space origin is in center. Are you sure?'
        endif
#endif

        ! Evaluate the address of the slice's origin and its extent
        if (slice%is_in_real_space) then
            if (slice%object_is_centered_in_box) then
                addr_slice_origin = slice%physical_address_of_box_center
            else
                addr_slice_origin = [1,1,1]
            endif
            addr_slice_max = slice%logical_dimensions
        else
            addr_slice_origin = 1
            ! addr_slice_origin(2) is not actually determined until we know where we are, at which point we can use the nearest_origin routine
            addr_slice_max(1) = slice%physical_upper_bound_complex(1)
            addr_slice_max(2) = slice%logical_dimensions(2)
        endif

        ! Evaluate address of origin of volume
        if (volume%is_in_real_space .and. volume%object_is_centered_in_box) then
            volume_mid_coos = volume%physical_address_of_box_center
        else
            volume_mid_coos = [1,1,1]
        endif

        ! Get the appropriate rotation matrix to go from slice coordinate to 3d coordinate
        if (present(rotation)) then
            rrotation = rotation
        else
            rrotation = euler_to_rotmat(phi,theta,psi,reverse=.true.)
        endif

        ! This is used for phase shifts
        slice_phase_factor_x = 2.0d0*dpi/real(slice%logical_dimensions(1),kind=8)
        slice_phase_factor_y = 2.0d0*dpi/real(slice%logical_dimensions(2),kind=8)
        volume_phase_factor_z = 2.0d0*dpi/real(volume%logical_dimensions(3),kind=8)

        ! Additional z shift?
        if (present(z_shift)) then
#ifndef skip_runtime_checks
            if (present(rotation)) then
                write(*,'(a)') '**ERROR(extract_slice): z_shift and rotation options cannot be used together'
                call terminate('Feature not implemented')
            endif
            if (slice%is_in_real_space) then
                write(*,'(a)') '**ERROR(extract_slice): z_shift not implemented when extracting from a real volume'
                call terminate('Feature not implemented')
            endif
#endif
            zz_shift = z_shift
            zz_shift = zz_shift*volume_phase_factor_z
        else
            zz_shift = 0.0D0
        endif

        zzz_shift = zz_shift .ne. 0.0D0

        ! x, y shifts?
        xx_shift = 0.0d0
        yy_shift = 0.0d0
        if (present(x_shift)) xx_shift = x_shift*slice_phase_factor_x
        if (present(y_shift)) yy_shift = y_shift*slice_phase_factor_y
        xy_shift = xx_shift .ne. 0.0d0 .or. yy_shift .ne. 0.0d0
#ifndef skip_runtime_checks
        if (xy_shift) then
            if (slice%is_in_real_space) then
                write(*,'(a)') '**ERROR(extract_slice): sorry x,y shifts not implemented for real slice extraction'
                call terminate('Fatal error in EXTRACT_SLICE')
            endif
        endif
#endif

        ! Maximum radius
        if (present(rad_max)) then
            rrad_max = rad_max**2
        else
            rrad_max = (real(slice%logical_dimensions(1))/2.)**2
        endif



        ! Loop over pixels in the slice. We use separate loop structures for real and complex cases.
        if (slice%is_in_real_space) then
            do J=1,addr_slice_max(2)
                logi_addr_in_slice(2) = j - addr_slice_origin(2)
                do I=1,addr_slice_max(1)
                    logi_addr_in_slice(1) = i - addr_slice_origin(1)
                    ! Are we within the max radius?
                    if (real(logi_addr_in_slice(1)**2 + logi_addr_in_slice(2)**2) .gt. rrad_max) then
                        slice%real_values(i,j,1) = 0.0
                        cycle
                    endif
                    ! Work out where in the volume this pixel comes from
                    coos_in_volume = matmul(rrotation,(/real(logi_addr_in_slice(1)),real(logi_addr_in_slice(2)),0.0,1.0/))
                    ! Center the window around that point
                    win%centered_on = coos_in_volume(1:3)
                    ! Work out the bounds of the window
                    call eval_window_bounds(win)
                    ! If no part of the window falls within the volume (this shouldn't happen), the pixel is filled with 0 and we go to next
                    if (   any(win%lbound .gt. volume%logical_upper_bound_real) &
                      .or. any(win%ubound .lt. volume%logical_lower_bound_real)) then
                        slice%real_values(i,j,1) = 0.0
                        cycle
                    endif
                    ! Read relevant part of volume into window and interpolate the value we're after
                    call window_from_image_real(volume,win)
                    call get_real_from_window(rvalue,win,interpol_scheme%interpolation_function,interpol_scheme%windowing_function)
                    slice%real_values(i,j,1) = rvalue
                enddo
            enddo
        else    ! The comp case
            do J=1,addr_slice_max(2)
                logi_addr_in_slice(2) = slice%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2)
                yy = logi_addr_in_slice(2)**2
                do I=1,addr_slice_max(1)
                    logi_addr_in_slice(1) = i - addr_slice_origin(1)
                    xx = logi_addr_in_slice(1)**2
                    !print '(a,2i6)', '**debug(extract_slice): i,j = ', i,j
                    ! Are we within the maximum radius?
                    if (real(XX+YY) .gt. rrad_max) then
                        slice%complex_values(i,j,1) = (0.0,0.0)
                        cycle
                    endif
                    ! Work out where in the volume this pixel comes from
                    logi_addr_in_slice_r(1:2) = real(logi_addr_in_slice(1:2))
                    coos_in_volume = matmul(rrotation,logi_addr_in_slice_r)
                    ! Center the window around that point
                    win%centered_on = coos_in_volume(1:3)
                    ! Work out the bounds of the window
                    call win%UpdateWindowBounds()
                    ! If no part of the window falls within the volume (this shouldn't happen), the pixel is filled with 0 and we go to next
                    if (     any(win%lbound .gt. volume%logical_upper_bound_complex) &
                        .or. any(win%ubound .lt. volume%logical_lower_bound_complex)) then
                        slice%complex_values(i,j,1) = (0.0,0.0)
                        cycle
                    endif
                    ! Read relevant part of volume into window and interpolate the value we're after
                    call get_comp_from_image(cvalue,volume,interpol_scheme,win, &
                                             volume%logical_lower_bound_complex,volume%logical_upper_bound_complex)
                    slice%complex_values(i,j,1) = cvalue
                    !print *, '**debug(extract_slice): cvalue = ', cvalue, sum(win%cpointer)
                    ! If user asked for z shift of volume, apply it to the slice (i think this is equivalent)
                    if (zzz_shift) then
                        phase_z = zz_shift*coos_in_volume(3)
                        phase_shift = cmplx(cos(-phase_z),sin(-phase_z),kind=8)
                        slice%complex_values(i,j,1) = slice%complex_values(i,j,1) * phase_shift
                    endif
                    ! Apply xy shift to the slice
                    if (xy_shift) then
                        phase_x = xx_shift*real(logi_addr_in_slice(1),kind=8)
                        phase_y = yy_shift*real(logi_addr_in_slice(2),kind=8)
                        phase = - phase_x - phase_y
                        phase_shift = cmplx(cos(phase),sin(phase),kind=8)
                        slice%complex_values(i,j,1) = slice%complex_values(i,j,1) * phase_shift
                    endif
                enddo
            enddo
        endif

    end subroutine extract_slice

!    !>  \brief  Insert a 2d slice into a 3d volume.
!    !!
!    !!  Conceptually, the slice is first converted to an infinitely thin continuous function (a 2d plane), placed in a continuous
!    !!  3d volume, which is then sampled. The sampling step is analoguous to a scanner, where each voxel integrates over its volume
!    subroutine insert_slice_get2d_put3d(volume,slice,rotation,interpol_scheme,oversampling_factor)
!        use images
!        use rotation_matrices
!        use interpolation, only : interpolation_scheme
!        use eulerangles
!        use runtime_parameters, only : terminate
!        implicit none
!        ! Arguments
!        type(image),                    target,     intent(inout)           ::  volume              !<  Volume to insert slice into
!        type(image),                                intent(in)              ::  slice               !<  Input slice
!        type(rotation_matrix),                      intent(in)              ::  rotation            !<  Rotation matrix to go from slice coordinates to 3d coordinates
!        type(interpolation_scheme),                 intent(in)              ::  interpol_scheme     !<  Interpolation scheme to use
!        integer,                        optional,   intent(in)              ::  oversampling_factor !<  Oversampling factor
!        ! Private variables
!        integer                 ::  i,j
!        integer                 ::  addr_slice_max(3)       !<  Address of the the last element in the slice
!        integer                 ::  addr_slice_origin(3)    !<  Address of the origin of the slice
!        !integer                    ::  addr_slice_min(2)       !<  Address of the first element in the slice
!        integer                 ::  volume_lbound(3)        !<  Address of lower bound of window relative to origin 0,0,0 at center
!        integer                 ::  volume_ubound(3)
!        integer                 ::  slice_lbound(3), slice_ubound(3)
!        integer                 ::  volume_mid_coos(3)      !<  Address of origin
!        !real                   ::  rotation(4,4)           !<  Rotation matrix used to position pixel into volume
!        real                    ::  coos_in_slice(2)        !<  Coordinates in slice
!        real                    ::  coos_in_volume(4)
!        type(window), target    ::  slice_win,volume_win
!        real                    ::  rvalue
!        complex                 ::  cvalue
!        integer                 ::  ooversampling_factor
!
!        ! Start work
!
!        call terminate('ERROR: not updated to take into account explicit Friedel mates - do not use until updated')
!
!        ! Will we be oversampling?
!        if (present(oversampling_factor)) then
!            ooversampling_factor = oversampling_factor
!        else
!            ooversampling_factor = 1
!        endif
!
!        ! Check input volume is in memory
!        if (.not. volume%is_in_memory) then
!            write(*,'(a)') '**ERROR(insert_slice_get2d_put3d): input volume is not in memory'
!            call terminate('Fatal error in INSERT_SLICE_GET2D_PUT3D')
!        endif
!
!        ! Check input volume really is a volume
!        if ( any(volume%logical_dimensions .eq. 1) ) then
!            write(*,'(a,3(i4,1x))') '**ERROR(insert_slice_get2d_put3d): input volume is not a volume, it has dimensions ', &
!                                    volume%logical_dimensions
!            call terminate('Fatal error in INSERT_SLICE_GET2D_PUT3D')
!        endif
!
!        ! Check input slice is in memory
!        if (.not. slice%is_in_memory) then
!            write(*,'(a)') '**ERROR(insert_slice_get2d_put3d): input slice is not in memory'
!            call terminate('Fatal error in INSERT_SLICE_GET2D_PUT3D')
!        endif
!
!
!        ! Check type
!        if (.not. slice%IsInSameSpaceAs(volume)) then
!            !write(*,'(5a)') '**WARNING(insert_slice): volume (', volume%type, ') and slice (', slice%type, ') have different types'
!            !write(*,'(a)')  '**WARNING(insert_slice): changing type of slice to match that of volume'
!            !slice%type = volume%type
!            write(*,'(a)') '**ERROR(insert_slice_get2d_put3d): input slice and volume have different types'
!            call terminate('Fatal error in INSERT_SLICE_GET2D_PUT3D')
!        endif
!
!
!        ! Evaluate the address of the slice's origin and its extent
!        if (slice%is_in_real_space) then
!            if (slice%object_is_centered_in_box) then
!                addr_slice_origin = slice%physical_address_of_box_center
!            else
!                addr_slice_origin = [1,1,1]
!            endif
!            addr_slice_max = slice%logical_dimensions
!        else
!            addr_slice_origin(1) = 1
!            ! addr_slice_origin(2) is not determined until we know where we are, at which point we can use the nearest_origin routine
!            addr_slice_max(1) = slice%physical_upper_bound_complex(1)
!            addr_slice_max(2) = slice%logical_dimensions(2)
!        endif
!
!        ! Evaluate the bounds of the slice and the volume (with 0,0,0 origin at center)
!        call min_max_logical_address(slice, slice_lbound, slice_ubound )
!        call min_max_logical_address(volume,volume_lbound,volume_ubound)
!
!        ! Evaluate address of origin of volume
!        if (volume%is_in_real_space .and. volume%object_is_centered_in_box) then
!            volume_mid_coos = volume%physical_address_of_box_center
!        else
!            volume_mid_coos = [1,1,1]
!        endif
!
!        ! Prepare the windows
!        call slice_win%AllocateWindow(interpol_scheme%window_width,slice%IsInFourierSpace(),slice%IsAVolume(),OWN_MEMORY=.true.)
!        call volume_win%AllocateWindow(2,VOLUME%IsInFourierSpace(),volume%IsAVolume(),OWN_MEMORY=.true.)
!
!        ! Get the appropriate rotation matrix to go from slice coordinate to 3d coordinate
!        !rotation = euler_to_rotmat(angles%phi,angles%theta,angles%psi,REVERSE=.TRUE.)
!
!        ! Loop over the slice in 1/steps_per_pxl. We use separate loop structures for real and complex cases.
!        if (slice%is_in_real_space) then
!            do J=1,(addr_slice_max(2)-1)*OOVERSAMPLING_FACTOR+1
!                coos_in_slice(2) = 1+REAL(J-1)/real(ooversampling_factor) - real(addr_slice_origin(2))
!                do I=1,(addr_slice_max(1)-1)*OOVERSAMPLING_FACTOR+1
!                    coos_in_slice(1) = 1+REAL(I-1)/real(ooversampling_factor) - real(addr_slice_origin(1))
!                    ! Work out where in the volume this pixel goes to
!                    coos_in_volume = matmul(rotation%rmatrix,[COOS_IN_SLICE(1),COOS_IN_SLICE(2),0.0,1.0])
!                    ! Center the window around the point in the slice
!                    slice_win%centered_on(1:2) = coos_in_slice(1:2)
!                    slice_win%centered_on(3) = 0.0
!                    volume_win%centered_on = coos_in_volume(1:3)
!                    ! Work out the bounds of the windows
!                    call eval_window_bounds(slice_win)
!                    call eval_window_bounds(volume_win)
!                    ! If no part of the window falls within the volume, nothing happens to that pixel
!                    if (any(volume_win%lbound .gt. volume_ubound) .or. any(volume_win%ubound .lt. volume_lbound)) then
!                        cycle
!                    endif
!                    ! If no part of the window falls within the slice, nothing happens to that pixel    !! check this
!                    if (any(slice_win%lbound .gt. slice_ubound) .or. any(slice_win%ubound .lt. slice_lbound)) then
!                        cycle
!                    endif
!                    ! Get values from image into window
!                    call window_from_image_real(slice,slice_win)
!                    ! Interpolate value
!                    call get_real_from_window(rvalue,slice_win,interpol_scheme%interpolation_function,  &
!                                                interpol_scheme%windowing_function)
!                    ! Place it in window over volume
!                    call put_real_into_window(rvalue,volume_win,interpol_scheme%interpolation_function, &
!                                                interpol_scheme%windowing_function)
!                    ! Add value into the relevant voxel of the volume
!                    call window_to_image_real(volume,volume_win)
!                enddo
!            enddo
!        else    ! The comp case
!            do J=1,(addr_slice_max(2)-1)*OOVERSAMPLING_FACTOR+1
!                coos_in_slice(2) = 1+REAL(J-1)/real(ooversampling_factor) &
!                                    - real(index_nearest_origin(j,slice%logical_dimensions(2)))
!                do I=1,(addr_slice_max(1)-1)*OOVERSAMPLING_FACTOR+1
!                    coos_in_slice(1) = 1+REAL(I-1)/real(ooversampling_factor) - real(addr_slice_origin(1))
!                    ! Work out where in the volume this pixel goes to
!                    coos_in_volume = matmul(rotation%rmatrix,[COOS_IN_SLICE(1),COOS_IN_SLICE(2),0.0,1.0])
!                    ! Center the windows around that point
!                    slice_win%centered_on(1:2) = coos_in_slice(1:2)
!                    slice_win%centered_on(3) = 0.0
!                    volume_win%centered_on = coos_in_volume(1:3)
!                    ! Work out the bounds of the window
!                    call eval_window_bounds(slice_win)
!                    call eval_window_bounds(volume_win)
!                    ! If no part of the window falls within the volume, nothing happens to that pixel
!                    if (any(volume_win%lbound .gt. volume_ubound) .or. any(volume_win%ubound .lt. volume_lbound)) then
!                        cycle
!                    endif
!                    ! If no part of the window falls within the slice, nothing happens to that pixel    !! check this
!                    if (any(slice_win%lbound .gt. slice_ubound) .or. any(slice_win%ubound .lt. slice_lbound)) then
!                        cycle
!                    endif
!                    ! Get values from image into window
!                    call window_from_image_comp(slice,slice_win)
!                    ! Interpolate value
!                    call get_comp_from_window(cvalue,slice_win,interpol_scheme%interpolation_function,  &
!                                                interpol_scheme%windowing_function)
!                    ! Place it in window over volume
!                    call put_comp_into_window(cvalue,volume_win,interpol_scheme%interpolation_function, &
!                                                interpol_scheme%windowing_function)
!                    ! Add value into the relevant voxel of the volume
!                    call window_to_image_comp(volume,volume_win)
!                enddo
!            enddo
!        endif
!
!    end subroutine insert_slice_get2d_put3d


    !>  \brief  Insert a 2d slice into a 3d volume (equivalent to frealign's pinsert)
    subroutine insert_slice_put(volume,slice,rotation,interpol_scheme,subtract,num_contrib_elems,amplitudes_only)
        use images, only : Image
        use InterpolatingWindows
        use rotation_matrices, only : rotation_matrix
        use interpolation, only : interpolation_scheme
        !use eulerangles
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),                target,     intent(inout)   ::  volume              !<  Volume to insert slice into
        type(image),                            intent(in)      ::  slice               !<  Input slice
        type(rotation_matrix),                  intent(in)      ::  rotation            !<  Rotation matrix to go from slice coordinate to 3d coordinate
        type(interpolation_scheme),             intent(in)      ::  interpol_scheme     !<  Interpolation scheme to use
        logical,                    optional,   intent(in)      ::  subtract            !<  If .TRUE., the slice will be subtracted from, rather than added to, the volume. .FALSE. by default.
        integer,                    optional,   intent(out)     ::  num_contrib_elems   !<  Return number of contributing elements
        logical,                    optional,   intent(in)      ::  amplitudes_only     !<  The phases are thrown away - only the amplitudes are added to the volume (applies only if input image is complex)
        ! Private variables
        integer ::  i,j
        integer ::  addr_slice_max(3)       !<  Address of the the last element in the slice
        integer ::  addr_slice_origin(3)    !<  Address of the origin of the slice
        integer ::  logi_addr_in_slice(2)       !<  Address of current element within the slice
        real    ::  coos_in_volume(4)       !<  Coordinates of current element in volume
        !integer ::  volume_lbound(3)        !<  Address of lower bound of window relative to origin 0,0,0 at center
        !integer ::  volume_ubound(3)
        !integer ::  volume_mid_coos(3)      !<  Address of origin
        type(window), target    ::  win
        integer ::  nnum_contrib_elems
        logical ::  aamplitudes_only
        complex ::  cval


        ! Start work

        aamplitudes_only = .false.
        if (present(amplitudes_only)) aamplitudes_only = .true.

        ! Check input volume is in memory
        if (.not. volume%is_in_memory) then
            write(*,'(a)') '**ERROR(insert_slice): input volume is not in memory'
            call terminate('Fatal error in INSERT_SLICE')
        endif

        ! Check input volume really is a volume
        if ( any(volume%logical_dimensions .eq. 1) ) then
            write(*,'(a,3(i4,1x))') '**ERROR(insert_slice): input volume is not a volume, it has dimensions ', &
                                    volume%logical_dimensions
            call terminate('Fatal error in INSERT_SLICE')
        endif

        ! Check input slice is in memory
        if (.not. slice%is_in_memory) then
            write(*,'(a)') '**ERROR(insert_slice): input slice is not in memory'
            call terminate('Fatal error in INSERT_SLICE')
        endif


        ! Check type
        if (.not. slice%IsInSameSpaceAs(volume)) then
            !write(*,'(5a)') '**WARNING(insert_slice): volume (', volume%type, ') and slice (', slice%type, ') have different types'
            !write(*,'(a)')  '**WARNING(insert_slice): changing type of slice to match that of volume'
            !slice%type = volume%type
            write(*,'(a)') '**ERROR(insert_slice): input slice and volume have different types'
            call terminate('Fatal error in INSERT_SLICE')
        endif

        ! Evaluate the address of the slice's origin and its extent
        if (slice%is_in_real_space) then
            if (slice%object_is_centered_in_box) then
                addr_slice_origin = slice%physical_address_of_box_center
            else
                addr_slice_origin = [1,1,1]
            endif
            addr_slice_max = slice%logical_dimensions
        else
            addr_slice_origin(1) = 1
            ! addr_slice_origin(2) is not determined until we know where we are, at which point we can use the nearest_origin routine
            addr_slice_max(1) = slice%physical_upper_bound_complex(1)
            addr_slice_max(2) = slice%logical_dimensions(2)
        endif


        ! Prepare the window
        call win%AllocateWindow(interpol_scheme%window_width,volume%IsInFourierSpace(),volume%IsAVolume(),OWN_MEMORY=.true.)

        ! Start counting
        nnum_contrib_elems = 0

        ! Loop over pixels in the slice. We use separate loop structures for real and complex cases.
        if (slice%is_in_real_space) then
            if (aamplitudes_only) then
                write(*,'(a)') '**ERROR(insert_slice_put): amplitudes-only option not implemented if the input slice is real'
                call terminate('Fatal error in INSERT_SLICE_PUT')
            endif
            do J=1,addr_slice_max(2)
                logi_addr_in_slice(2) = j - addr_slice_origin(2)
                do I=1,addr_slice_max(1)
                    logi_addr_in_slice(1) = i - addr_slice_origin(1)
                    ! Count the number of elements we're "putting"
                    nnum_contrib_elems = nnum_contrib_elems + 1
                    ! Work out where in the volume this pixel goes to
                    coos_in_volume = matmul(rotation%rmatrix,(/real(logi_addr_in_slice(1)),real(logi_addr_in_slice(2)),0.0,1.0/))
                    ! Center the window around that point
                    win%centered_on = coos_in_volume(1:3)
                    ! Work out the bounds of the window
                    call eval_window_bounds(win)
                    ! If no part of the window falls within the volume (this shouldn't happen), nothing happens to that pixel
                    if (     any(win%lbound .gt. volume%logical_upper_bound_real) &
                        .or. any(win%ubound .lt. volume%logical_lower_bound_real)) then
                        cycle
                    endif
                    ! Spread value into window
                    call put_real_into_window(slice%real_values(i,j,1),win,  &
                                                interpol_scheme%interpolation_function,interpol_scheme%windowing_function)
                    ! Add values from window into volume
                    call window_to_image_real(volume,win,subtract=subtract)
                enddo
            enddo
        else    ! The comp case
            do J=1,addr_slice_max(2)
                logi_addr_in_slice(2) = slice%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2)
                do I=1,addr_slice_max(1)
                    logi_addr_in_slice(1) = i - 1

                    ! At the origin of the first dimension, we have the unusual situation that Friedel mates are described
                    ! explicitely, except for the Nyquist components, which are only described in the negative frequencies.
                    ! To avoid duplication, we therefore cycle if we are on Friedel mates.
                    if (logi_addr_in_slice(1) .eq. 0) then
                        if (logi_addr_in_slice(2) .gt. 0) then
                            cycle
                        endif
                    endif

                    nnum_contrib_elems = nnum_contrib_elems + 1

                    if (slice%complex_values(i,j,1) .eq. (0.0,0.0)) cycle

                    ! Work out where in the volume this pixel goes to
                    coos_in_volume = matmul(rotation%rmatrix,(/real(logi_addr_in_slice(1)),real(logi_addr_in_slice(2)),0.0,1.0/))
                    ! Center the window around that point
                    win%centered_on = coos_in_volume(1:3)
                    ! Work out the bounds of the window
                    call eval_window_bounds(win)
                    ! If no part of the window falls within the volume (this shouldn't happen), nothing happens to that pixel
                    if (     any(win%lbound .gt. volume%logical_upper_bound_complex) &
                        .or. any(win%ubound .lt. volume%logical_lower_bound_complex)) then
                        cycle
                    endif
                    ! Spread value into window
                    cval = slice%complex_values(i,j,1)
                    if (aamplitudes_only) cval = cabs(cval)
                    call put_comp_into_window(cval,win,interpol_scheme%interpolation_function,interpol_scheme%windowing_function)
                    ! Add values from window into volume
                    call window_to_image_comp(volume,win,subtract=subtract)
                enddo
            enddo
        endif

        if (present(num_contrib_elems)) num_contrib_elems = nnum_contrib_elems

    end subroutine insert_slice_put

end module images_threed
