!> \brief Module describing the interpolating windows
!!
!!  \todo   Deprecate
module InterpolatingWindows

    use Globals
    use, intrinsic :: iso_c_binding
    use fftw33
    use runtime_parameters, only : terminate

    private


    ! Interpolation routines
    public  ::  put_real_into_window, window_from_image_comp, get_comp_from_window, put_comp_into_window
    public  ::  get_comp_from_image, eval_window_bounds, window_from_image_real, get_real_from_window
    public  ::  window_to_image_real, window_to_image_comp

    public  ::  correct_3d

    !>  \brief  a window is a special kind of image object which windows a part of larger image.
    !!          it is not suitable for operations such as ForwardFFT, because it doesn't have the dim/2+1 trick built in.
    type, public    ::  Window
        integer                                 ::  logical_dimensions(3)               =   (/1,1,1/)   !< Logical (x,y,z) dimensions of the image. Note that this does not necessarily correspond to memory allocation (i.e. physical) dimensions.

        logical                                 ::  is_in_real_space                    =   .true.      !< The image is in real space, not Fourier space.
        logical                                 ::  object_is_centered_in_box           =   .true.      !< Whether the real-space origin of the image is at the center of the image. note that in fourier space,
                                                                                                        !! the origin is always at the first corner pixel (address 1,1,1). when an image is complex,
                                                                                                        !! the origin_at_center flag still refers to the real-space origin of the image (if it were back-ft'ed).
        ! Looping
        integer                                 ::  physical_upper_bound_complex(3)                 =   (/0,0,0/)   !<  In each dimension, the upper bound of the complex image's physical addresses
        integer                                 ::  physical_address_of_box_center(3)               =   (/0,0,0/)   !<  In each dimension, the address of the pixel at the origin
        integer                                 ::  physical_index_of_first_negative_frequency(3)   =   (/0,0,0/)   !<  In each dimension, the physical index of the first pixel which stores negative frequencies
        real                                    ::  fourier_voxel_size(3)                     =   (/0.0,0.0,0.0/) !<  Distance from Fourier voxel to Fourier voxel, expressed in reciprocal pixels
        integer                                 ::  logical_upper_bound_complex(3)                  =   (/0,0,0/)   !<  In each dimension, the upper bound of the complex image's logical addresses
        integer                                 ::  logical_lower_bound_complex(3)                  =   (/0,0,0/)   !<  In each dimension, the lower bound of the complex image's logical addresses
        integer                                 ::  logical_upper_bound_real(3)                  =   (/0,0,0/)   !<  In each dimension, the upper bound of the real image's logical addresses
        integer                                 ::  logical_lower_bound_real(3)                  =   (/0,0,0/)   !<  In each dimension, the lower bound of the real image's logical addresses



        !
        type(c_ptr)                             ::  p                                   =   c_null_ptr  !< c pointer to the image data array
        real(kind=c_float), pointer, public     ::  real_values(:,:,:)                  =>  null()      !< Real array to hold values for real images.
        complex(kind=c_float_complex), pointer, public ::  complex_values(:,:,:)        =>  null()      !< Complex array to hold values for comp images.
        logical                                 ::  is_in_memory                        =   .false.     !< Whether image values are in-memory, in other words whether the image has memory space allocated to its data array. default = .false.

        ! fftw-specific flags
        type(c_ptr)                             ::  plan_fwd                            =   c_null_ptr  !<  fftw plan for the image (fwd)
        type(c_ptr)                             ::  plan_bwd                            =   c_null_ptr  !<  fftw plan for the image (bwd)
        logical                                 ::  planned                             =   .false.     !<  Whether the plan has been setup by/for fftw

        integer                 ::  width       =   2   !<  width of the window (pixels) - window is always square/cube. for linear, width is 2.
        real                    ::  centered_on(3)      !<  coordinates within the image around which the window is centered
        real                    ::  center(3)           !<  coordinate of the center relative to the window (with origin 1,1,1 at top left corner of window)
        integer                 ::  lbound(3)           !<  address of top-left voxel in image
        integer                 ::  ubound(3)           !<  address of bottom-right voxel in image
        logical                 ::  own_memory          !<  whether the window should have its own area in memory. if not, it will just point to another area of memory.
        real,       pointer     ::  rpointer(:,:,:)     !<  real pointer used when the window does not have its own area in memory.
        complex,    pointer     ::  cpointer(:,:,:)     !<  complex pointer used when the window does not have its own area in memory.
        contains
            !
            procedure           ::  eval_window_bounds
            procedure, public   ::  UpdateWindowBounds => eval_window_bounds
            ! Memory management
            procedure           ::  allocate_window
            procedure, public   ::  AllocateWindow => allocate_window
            procedure           ::  deallocate_window
            procedure, public   ::  DeallocateWindow => deallocate_window
            ! Destruction
            final               ::  destroy_window
    end type window

    contains

    !>  \brief  a generalised routine for the compensating for the effects of the interpolation kernel used during  slice insertion
    !!
    !!  you need to supply the appropriate instrument function corresponding to the interpolation kernel used during
    !!  slice insertion. this should be the instrument function corresponding to the apodization function:
    !!  http://mathworld.wolfram.com/apodizationfunction.html
    !!  for example, if a rectangle function was used, you need to correct with a sinc
    !!  the image_interpol module should have an _instr function for each interpol/window function to make this easier for you!
    !!
#if defined(skip_runtime_checks)
    pure &
#endif
    subroutine correct_3d(img,interpol_function_instr,window_function_instr,window_width,reverse)
        use interpolation, only : interpolation_function, windowing_function, interpol_window_convolution
        use Images
#if defined (skip_runtime_checks)
#else
		use runtime_parameters, only : terminate
#endif
        implicit none
        ! arguments
        type(image),    intent(inout)               ::  img                     !<  volume to be corrected
        procedure(interpolation_function),  pointer ::  interpol_function_instr !<  the instrument function corresponding to the interpolation function used to calculate the 3d reconstruction
        procedure(windowing_function),      pointer ::  window_function_instr   !<  the instrument function corresponding to the windowing function used in calculating the 3d
        integer,    intent(in)                      ::  window_width            !<  window width used
        logical,    optional,   intent(in)          ::  reverse                 !<  unless reverse=.true., the volume will be divided by the intrument function
        ! private variables
        logical ::  rreverse
        integer ::  i,j,k
        real    ::  rad
        real    ::  boxsize(3)
        real    ::  w(3)
        integer ::  origin(3)
        integer ::  largest_dim
        integer ::  max_irad
        real,       allocatable     ::  convoluted_function(:,:)
        logical,    parameter       ::  debug   =   .false.
        ! start work
#if defined(skip_runtime_checks)
#else
        if (debug) then
            print '(a,3(i3,1x))', '**debug(correct_3d): img%logical_dimensions = ', img%logical_dimensions
        endif
        ! this is only meant with images with origin at center
        if (img%is_in_real_space .and. .not. img%object_is_centered_in_box) then
            write(*,'(a)') '**error(correct_3d): cannot deal with real images whose origin is not at the center'
            call terminate('fatal error in correct_3d')
        endif
#endif

        ! check whether we need to correct for, or apply the fall-off
        if (present(reverse)) then
            rreverse = reverse
        else
            rreverse = .false.
        endif

        ! which is the largest dimension?
        largest_dim = maxloc(img%logical_dimensions,dim=1)

        ! the box size
        boxsize = real(img%logical_dimensions)

        ! the origin
        if (img%is_in_real_space .and. img%object_is_centered_in_box) then
            origin = img%physical_address_of_box_center
        else
            origin = [1,1,1]
        endif

        ! the maximum radius (including the origin)
        max_irad = max( img%logical_dimensions(largest_dim) - origin(largest_dim) + 1, &
                        origin(largest_dim) )

#if defined(skip_runtime_checks)
#else
        if (debug) write(*,'(a,f0.2,1x,2(i0,1x))') '**debug(correct_3d): boxsize, origin, max_irad = ', boxsize, origin(1), max_irad
#endif

        ! work out values for the convolution of window_instr with interpol_instr
        allocate(convoluted_function(max_irad,3))
        do j=1,3
        do k=1,max_irad
            rad  = real(k)
            convoluted_function(k,j) = interpol_window_convolution( rad/boxsize(j),interpol_function_instr,window_function_instr, &
                                                                    window_width,img%logical_dimensions(j))

#if defined(skip_runtime_checks)
#else
            if (debug) write(*,'(a,f0.6,1x,f0.6)') '**debug(correct_3d): rad, convoluted_function(rad): ',   &
                                                    rad, convoluted_function(k,1)
#endif
        enddo
        enddo

        !
        if (img%is_in_real_space) then
            do k=1,img%logical_dimensions(3)
                w(3) = convoluted_function(abs(k-origin(3))+1,3)
                do j=1,img%logical_dimensions(2)
                    w(2) = convoluted_function(abs(j-origin(2))+1,2)
                    do i=1,img%logical_dimensions(1)
                        w(1) = convoluted_function(abs(i-origin(1))+1,1)
                        if (rreverse) then
                            img%real_values(i,j,k) = img%real_values(i,j,k) * w(1)*w(2)*w(3)
                        else
                            img%real_values(i,j,k) = img%real_values(i,j,k) / (w(1)*w(2)*w(3))
                        endif
                    enddo
                enddo
            enddo
        else
            ! in fourier space
            do k=1,img%logical_dimensions(3)
                w(3) = convoluted_function(abs(img%LogicalIndexGivenPhysicalIndexInFourierSpace(k,3))+1,3)
                do j=1,img%logical_dimensions(2)
                    w(2) = convoluted_function(abs(img%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2))+1,2)
                    do i=1,img%physical_upper_bound_complex(1)
                        w(1) = convoluted_function(i,1)
                        if (rreverse) then
                            img%complex_values(i,j,k) = img%complex_values(i,j,k) * w(1)*w(2)*w(3)
                        else
                            img%complex_values(i,j,k) = img%complex_values(i,j,k) / (w(1)*w(2)*w(3))
                        endif
                    enddo
                enddo
            enddo
        endif
    end subroutine correct_3d



    !>  \brief  get a complex value from an image by interpolation
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine get_comp_from_image(cvalue,img,interpol,win,coos_min,coos_max)
        use interpolation, only : interpolation_scheme
        use Images
        implicit none
        ! arguments
        complex,                            intent(out)         ::  cvalue      !<  value extracted from the image
        type(image),                        intent(in)          ::  img
        type(interpolation_scheme),         intent(in)          ::  interpol
        type(window),                       intent(inout)       ::  win
        integer,                            intent(in)          ::  coos_min(3) !<  the minimum logical address in the image
        integer,                            intent(in)          ::  coos_max(3) !<  the maximum logical address in the image
        ! private variables
        logical,    parameter   ::  use_hardcoded_interpol_if_available = .true.
        integer ::  i,j,k,ii,jj,kk
        real    ::  dist(3), weight(3)
        integer ::  tmp_coos(3), tmp_coos_in(3)
        ! start work
        if (use_hardcoded_interpol_if_available .and. interpol%id .eq. 'line') then
            cvalue = (0.0,0.0)

            ! we want to interpolate to the coordinates win%center(1:3) using linear interpolation
            !
            ! in one dimension, suppose we are 0.3 away from point 0 (and therefore 0.7 away from point 1)
            ! the result should be value0*(1-0.3) + value1*0.3
            ! = value0 - 0.3 value0 + 0.3 value1
            ! = value0 + 0.3 (value1-value0)
            !
            ! so we need the distance from the real coordinates to the previous pixel
            !dist = win%center - floor(win%center)

            ! we also need to the values of the nearest neighbours
            !logi_nn_000 = floor(win%center)
            !phys_nn_000 = PhysicalAddressGivenLogicalAddressInFourierSpace(img,logi_nn_000)



            ! iterate over window voxels. i,j,k is address within window, ii,jj,kk is logical address within image (with origin 0,0,0 at center),
            ! and iii,jjj,kkk is physical address within image (with origin 1,1,1) at top-left
            do k=1,2
                kk  = k + win%lbound(3) - 1
                if (kk .lt. coos_min(3) .or. kk .gt. coos_max(3)) cycle
                dist(3) = abs(k - win%center(3))
                weight(3) = interpol%interpolation_function(dist(3))
                do j=1,2
                    jj = j + win%lbound(2) - 1
                    if (jj .lt. coos_min(2) .or. jj .gt. coos_max(2)) cycle
                    dist(2) = abs(j - win%center(2))
                    weight(2) = interpol%interpolation_function(dist(2))
                    do i=1,2
                        ii = i + win%lbound(1) - 1
                        if (ii .lt. coos_min(1) .or. ii .gt. coos_max(1)) cycle
                        dist(1) = abs(i - win%center(1))
                        weight(1) = interpol%interpolation_function(dist(1))
                        ! get address within complex image which corresponds to voxel coordinate
                        tmp_coos_in = (/ii,jj,kk/)
                        call img%PhysicalAddressGivenLogicalAddressInFourierSpace(tmp_coos_in,tmp_coos)
                        ! now copy the voxel value over
                        if ( ii .ge. 0) then
                            cvalue = cvalue + img%complex_values(tmp_coos(1),tmp_coos(2),tmp_coos(3)) &
                                                                            * weight(1) * weight(2) * weight(3)
                        else
                            cvalue = cvalue + conjg(img%complex_values(tmp_coos(1),tmp_coos(2),tmp_coos(3)))    &
                                                                            * weight(1) * weight(2) * weight(3)
                        endif
                    enddo
                enddo
            enddo

        else
            call window_from_image_comp(img,win)
            call get_comp_from_window(cvalue,win,interpol%interpolation_function,interpol%windowing_function)
        endif
    end subroutine get_comp_from_image




    !>  \brief spread a complex value into a window
#if defined(skip_runtime_checks)
    pure &
#endif
    subroutine put_comp_into_window(cvalue,win,interpol_function,window_function)
        use interpolation, only : interpolation_function, windowing_function
        implicit none
        ! arguments
        complex,                    intent(in)          ::  cvalue              !< the value to spread into the window
        type(window),               intent(inout)       ::  win                 !< the window to spread the value into
        procedure(interpolation_function),  pointer     ::  interpol_function
        procedure(windowing_function),      pointer     ::  window_function
        ! private variables
        integer ::  i,j,k
        real    ::  dist(3), weight(3)

        ! diagnostics
        !write(*,'(a,2es12.3,a,3f6.2)') '**debug(put_comp_into_window): putting ', cvalue, ' into window center at ', win%center
        !write(*,'(a,3(i3,x))') '**debug(put_comp_into_window): window dimensions =', win%logical_dimensions
        !write(*,'(a,i3)') '**debug(put_comp_into_window): window width =', win%width

#if defined(skip_runtime_checks)
#else
        if (.not. win%own_memory) then
            write(*,'(a)') '**error(put_comp_into_window): can only work on windows with own memory'
            call terminate('fatal error in put_comp_into_window')
        endif
#endif

        ! start work
        do k=1,win%logical_dimensions(3)
            dist(3) = abs(real(k) - win%center(3))
            weight(3) = interpol_function(dist(3)) * window_function(dist(3),win%width)
            do j=1,win%logical_dimensions(2)
                dist(2) = abs(real(j) - win%center(2))
                weight(2) = interpol_function(dist(2)) * window_function(dist(2),win%width)
                do i=1,win%logical_dimensions(1)
                    dist(1) = abs(real(i) - win%center(1))
                    weight(1) = interpol_function(dist(1)) * window_function(dist(1),win%width)
                    win%cpointer(i,j,k) = cvalue * weight(1) * weight(2) * weight(3)
                enddo
            enddo
        enddo
    end subroutine put_comp_into_window





    !>  \brief  spread a value into a window
#if defined(skip_runtime_checks)
    pure &
#endif
    subroutine put_real_into_window(rvalue,win,interpol_function,window_function)
        use interpolation, only : interpolation_function, windowing_function
        implicit none
        ! arguments
        real,   intent(in)                              ::  rvalue                  !<  the value to spread into the window
        type(window),   intent(inout)                   ::  win                     !<  the window to spread the value into
        procedure(interpolation_function),  pointer     ::  interpol_function       !<  the interpolation function to use
        procedure(windowing_function),      pointer     ::  window_function         !<  the windowing function to use
        ! private variables
        integer ::  i,j,k
        real    ::  dist(3), weight(3)

        if (win%own_memory) then
            win%real_values = 0.0
        else
#if defined(skip_runtime_checks)
#else
            write(*,'(a)') '**error(put_real_into_window): can only work on windows with own memory'
            call terminate('fatal error in put_real_into_window')
#endif
        endif

        ! if input value is 0.0, no work to do
        if (rvalue .eq. 0.0) then
            return
        endif

        ! start work
        do k=1,win%logical_dimensions(3)
            dist(3) = abs(k - win%center(3))
            weight(3) = interpol_function(dist(3)) * window_function(dist(3),win%width)
            do j=1,win%logical_dimensions(2)
                dist(2) = abs(j - win%center(2))
                weight(2) = interpol_function(dist(2)) * window_function(dist(2),win%width)
                do i=1,win%logical_dimensions(1)
                    dist(1) = abs(i - win%center(1))
                    weight(1) = interpol_function(dist(1)) * window_function(dist(1),win%width)
                    !win%real_values(i,j,k) = rvalue * weight(1) * weight(2) * weight(3)
                    win%rpointer(i,j,k) = win%rpointer(i,j,k) + rvalue * weight(1) * weight(2) * weight(3)
                enddo
            enddo
        enddo
    end subroutine put_real_into_window


    !>  \brief  get a complex value from the window by interpolation
#if defined(skip_runtime_checks)
    pure &
#endif
    subroutine get_comp_from_window(cvalue,win,interpol_function,window_function,square)
        use interpolation, only : interpolation_function, windowing_function
        implicit none
        ! arguments
        complex,                intent(out)             ::  cvalue              !< the value to extract from the window
        type(window),           intent(in)              ::  win                 !< the window to extract the value from
        procedure(interpolation_function),  pointer     ::  interpol_function
        procedure(windowing_function),      pointer     ::  window_function
        logical,    optional,   intent(in)              ::  square              !<  whether the interpolated value should be squared
        ! private variables
        integer ::  i,j,k
        real    ::  dist(3), weight(3)
        logical ::  ssquare

        ! if all the values in the window are zero, don't need to do anything!
        if (sum(win%cpointer) .eq. (0.0,0.0)) then
            cvalue = (0.0,0.0)
            return
        endif

        ! will we be squaring?
        if (present(square)) then
            ssquare = square
        else
            ssquare = .false.
        endif

#if defined(skip_runtime_checks)
#else
        if (ssquare) then
            print *, '**warning(get_comp_from_window): squaring! are you sure you want to do that?'
            call terminate('get_comp_from_window')
        endif
#endif

        ! start work
        cvalue = (0.0,0.0)
        do k=1,win%logical_dimensions(3)
            dist(3) = abs(k - win%center(3))
            weight(3) = interpol_function(dist(3)) * window_function(dist(3),win%width)
            do j=1,win%logical_dimensions(2)
                dist(2) = abs(j - win%center(2))
                weight(2) = interpol_function(dist(2)) * window_function(dist(2),win%width)
                do i=1,win%logical_dimensions(1)
                    if (win%cpointer(i,j,k) .eq. (0.0,0.0)) cycle
                    dist(1) = abs(i - win%center(1))
                    weight(1) = interpol_function(dist(1)) * window_function(dist(1),win%width)
                    cvalue = cvalue + win%cpointer(i,j,k) * weight(1) * weight(2) * weight(3)
                enddo
            enddo
        enddo

        if (ssquare) then
            cvalue = cvalue**2
        endif
    end subroutine get_comp_from_window




    !>  \brief  get a real value from the window by interpolation
    pure subroutine get_real_from_window(rvalue,win,interpol_function,window_function,square)
        use interpolation, only : interpolation_function, windowing_function
        implicit none
        ! arguments
        real,                   intent(out)             ::  rvalue              !< the value to extract from the window
        type(window),           intent(in)              ::  win                 !< the window to extract the value from
        procedure(interpolation_function),  pointer     ::  interpol_function
        procedure(windowing_function),      pointer     ::  window_function
        logical,    optional,   intent(in)              ::  square              !<  whether the interpolated value should be squared
        ! private variables
        integer ::  i,j,k
        real    ::  dist(3), weight(3)

        ! start work
        rvalue = 0.0
        do k=1,win%logical_dimensions(3)
            dist(3) = abs(k - win%center(3))
            weight(3) = interpol_function(dist(3)) * window_function(dist(3),win%width)
            do j=1,win%logical_dimensions(2)
                dist(2) = abs(j - win%center(2))
                weight(2) = interpol_function(dist(2)) * window_function(dist(2),win%width)
                do i=1,win%logical_dimensions(1)
                    if (win%rpointer(i,j,k) .eq. 0.0) cycle
                    dist(1) = abs(i - win%center(1))
                    weight(1) = interpol_function(dist(1)) * window_function(dist(1),win%width)
                    rvalue = rvalue + win%rpointer(i,j,k) * weight(1) * weight(2) * weight(3)
                enddo
            enddo
        enddo
        if (present(square)) then
            if (square) then
                rvalue = rvalue**2
            endif
        endif
    end subroutine get_real_from_window

    !>  \brief  window part of an image and copy the contents into a window object
    subroutine window_from_image(img,win,coos)
        use Images
        implicit none
        ! arguments
        type(image),    intent(in)      ::  img     !<  image to extract data from
        type(window),   intent(inout)   ::  win     !<  window to store the data
        real,           intent(in)      ::  coos(3) !<  coordinates around which to extract the window
        ! private variables

        ! start work

        ! find the bounds of the window
        win%centered_on = coos
        call eval_window_bounds(win)

        if (.not. img%is_in_real_space) then
            call window_from_image_comp(img,win)
        else
            call window_from_image_real(img,win)
        endif
    end subroutine window_from_image

    !>  \brief  window object is populated with data from image object over which it is hovering
    !!  \todo   remove built-in assumption that image has even dimensions by calling origin_addr
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine window_from_image_real(img,win)
        use Images
        implicit none
        ! arguments
        type(image),    target, intent(in)      ::  img !<  image to extract data from
        type(window),   target, intent(inout)   ::  win !<  window to store the data
        ! private variables
        integer ::  i,j,k
        integer ::  ii,jj,kk
        integer ::  iii,jjj,kkk
        integer ::  coos_min(3),coos_max(3)
        integer ::  lbnd(3), ubnd(3), img_origin(3)
        ! start work
        coos_min = img%logical_lower_bound_real
        coos_max = img%logical_upper_bound_real
#ifndef skip_runtime_checks
        if (win%own_memory) then
#endif
            win%real_values = 0.0
            ! iterate over window voxels. i,j,k is physical address within window,
            ! ii,jj,kk is logical address within image (with origin 0,0,0 at center),
            ! and iii,jjj,kkk is physical address within image (with origin 1,1,1) at top-left
            do k=1,win%logical_dimensions(3)
                kk  = k + win%lbound(3) - 1
                if (kk .lt. coos_min(3) .or. kk .gt. coos_max(3)) cycle
                kkk = kk + img%logical_dimensions(3)/2 + 1
                if (img%logical_dimensions(3) .eq. 1) kkk = 1
                do j=1,win%logical_dimensions(2)
                    jj = j + win%lbound(2) - 1
                    if (jj .lt. coos_min(2) .or. jj .gt. coos_max(2)) cycle
                    jjj = jj + img%logical_dimensions(2)/2 + 1
                    do i=1,win%logical_dimensions(1)
                        ii = i + win%lbound(1) - 1
                        if (ii .lt. coos_min(1) .or. ii .gt. coos_max(1)) cycle
                        iii = ii + img%logical_dimensions(1)/2 + 1
                        ! copy voxel value over
                        win%real_values(i,j,k) = img%real_values(iii,jjj,kkk)
                    enddo
                enddo
            enddo
            win%rpointer => win%real_values
#ifndef skip_runtime_checks
        else
            if (img%is_in_real_space .and. img%object_is_centered_in_box) then
                img_origin = img%physical_address_of_box_center
            else
                img_origin = [1,1,1]
            endif
            !print *, '**debug(window_from_image_real): image dim, origin = ', img%logical_dimensions, img_origin
            !print *, '**debug(window_from_image_real): coos_min, coos_max = ', coos_min, coos_max
            !print *, '**debug(window_from_image_real): win%logical_dimensions = ', win%logical_dimensions
            !print *, '**debug(window_from_image_real): win%lbound = ', win%lbound
            do i=1,3
                lbnd(i) = max(coos_min(i),1+win%lbound(i)-1)
                lbnd(i) = lbnd(i) + img_origin(i)
                ubnd(i) = min(coos_max(i),win%logical_dimensions(i) + win%lbound(i) - 1)
                ubnd(i) = ubnd(i) + img_origin(i)
                if (lbnd(i) .eq. 1) then
                    ubnd(i) = lbnd(i) + win%logical_dimensions(i) - 1
                else if (ubnd(i) .eq. img%logical_dimensions(i)) then
                    lbnd(i) = ubnd(i) - win%logical_dimensions(i) + 1
                endif
            enddo
            if (img%logical_dimensions(3) .eq. 1) then
                lbnd(3) = 1
                ubnd(3) = 1
            endif
            !print *, '**debug(window_from_image_real): rpointer points to ', lbnd,ubnd
            win%rpointer => img%real_values(lbnd(1):ubnd(1),lbnd(2):ubnd(2),lbnd(3):ubnd(3))
        endif
#endif
    end subroutine window_from_image_real

    !>  \brief  window object is populated with data from image object over which it is hovering
#ifdef skip_runtime_checks
    pure &
#endif
    subroutine window_from_image_comp(img,win)
        use Images
        implicit none
        ! arguments
        type(image),    target, intent(in)      ::  img
        type(window),   intent(inout)   ::  win
        ! private variables
        integer ::  i,j,k
        integer ::  ii,jj,kk
        integer ::  iii,jjj,kkk
        integer ::  tmp_coos(3), tmp_coos_in(3)
        integer ::  coos_min(3),coos_max(3) !< the bounds of the image
        integer ::  lbnd(3), ubnd(3), img_origin(3)

        ! start work
        coos_min = img%logical_lower_bound_complex
        coos_max = img%logical_upper_bound_complex


        ! in the complex case, we always have to use own_memory for a window, because it is not possible to use a pointer when we
        ! are near the frequency origin in the first dimension (which is "folded"). this is why there is 1==1 here.
#ifndef skip_runtime_checks
        if (win%own_memory .or. 1==1) then
#endif
            win%complex_values = (0.0,0.0)
            ! iterate over window voxels. i,j,k is address within window, ii,jj,kk is address within image (with origin 0,0,0 at center),
            ! and iii,jjj,kkk is address within image (with origin 1,1,1) at top-left
            do k=1,win%logical_dimensions(3)
                kk  = k + win%lbound(3) - 1
                if (kk .lt. coos_min(3) .or. kk .gt. coos_max(3)) cycle
                do j=1,win%logical_dimensions(2)
                    jj = j + win%lbound(2) - 1
                    if (jj .lt. coos_min(2) .or. jj .gt. coos_max(2)) cycle
                    do i=1,win%logical_dimensions(1)
                        ii = i + win%lbound(1) - 1
                        if (ii .lt. coos_min(1) .or. ii .gt. coos_max(1)) cycle
                        ! get address within complex image which corresponds to voxel coordinate
                        tmp_coos_in = (/ii,jj,kk/)
                        call img%PhysicalAddressGivenLogicalAddressInFourierSpace(tmp_coos_in,tmp_coos)
                        iii = tmp_coos(1)
                        jjj = tmp_coos(2)
                        kkk = tmp_coos(3)
                        ! now copy the voxel value over
                        if ( ii .ge. 0) then
                            win%complex_values(i,j,k) = img%complex_values(iii,jjj,kkk)
                        else
                            win%complex_values(i,j,k) = conjg(img%complex_values(iii,jjj,kkk))  ! this is the line that cannot be done with pointers
                        endif
                    enddo
                enddo
            enddo
            !win%cpointer => win%complex_values ! 25-june-2012 - this pointer should already be associated (by window_allocate)
#ifndef skip_runtime_checks
        !
        ! the complex case without own_memory does not work, because of the case where only part of the window is in positive frequencies in the first dimension
        !
        else
            if (img%is_in_real_space .and. img%object_is_centered_in_box) then
                img_origin = img%physical_address_of_box_center
            else
                img_origin = [1,1,1]
            endif
            !print *, '**debug(window_from_image_comp): image dim, origin = ', img%logical_dimensions, img_origin
            !print *, '**debug(window_from_image_comp): coos_min, coos_max = ', coos_min, coos_max
            !print *, '**debug(window_from_image_comp): win%logical_dimensions = ', win%logical_dimensions
            !print *, '**debug(window_from_image_comp): win%lbound = ', win%lbound
            lbnd = max(coos_min,1+win%lbound-1)
            lbnd = lbnd + img_origin
            ubnd = min(coos_max,win%logical_dimensions + win%lbound - 1)
            ubnd = ubnd + img_origin
            !print *, '**debug(window_from_image_comp): lbnd, ubnd = ', lbnd, ubnd
            do i=1,3
                if (lbnd(i) .eq. 1) then
                    ubnd(i) = lbnd(i) + win%logical_dimensions(i) - 1
                else if (ubnd(i) .eq. img%logical_dimensions(i)) then
                    lbnd(i) = ubnd(i) - win%logical_dimensions(i) + 1
                endif
            enddo
            if (img%logical_dimensions(3) .eq. 1) then
                lbnd(3) = 1
                ubnd(3) = 1
            endif
            !print *, '**debug(window_from_image_comp): rpointer points to ', lbnd,ubnd
            win%cpointer => img%complex_values(lbnd(1):ubnd(1),lbnd(2):ubnd(2),lbnd(3):ubnd(3)) ! this does not work when the window hovers over the origin in the first dimension
        endif
#endif
    end subroutine window_from_image_comp

    !>  \brief  window object data is copied into image object
    !pure &
    subroutine window_to_image_real(img,win,subtract)
        use Images
        implicit none
        ! arguments
        type(image),                intent(inout)   ::  img         !<  image to extract data from
        type(window),               intent(in)      ::  win         !<  window to store the data
        logical,        optional,   intent(in)      ::  subtract    !<  if .true., the values in the window are subtracted from the image. .false. by default.
        ! private variables
        integer ::  i,j,k
        integer ::  ii,jj,kk
        integer ::  iii,jjj,kkk
        real    ::  sign
        ! start work
        if (present(subtract)) then
            if (subtract) then
                sign = -1.0
            else
                sign = 1.0
            endif
        else
            sign = 1.0
        endif

        !$omp   critical        (copy_window_to_volume_real_outer)
        if (win%own_memory) then
            ! iterate over window voxels. i,j,k is address within window, ii,jj,kk is address within image (with origin 0,0,0 at center),
            ! and iii,jjj,kkk is address within image (with origin 1,1,1) at top-left
            do k=1,win%logical_dimensions(3)
                kk  = k + win%lbound(3) - 1
                kkk = kk + img%logical_dimensions(3)/2 + 1
                if (img%logical_dimensions(3) .eq. 1) kkk = 1
                if (kkk .lt. 1 .or. kkk .gt. img%logical_dimensions(3)) cycle
                do j=1,win%logical_dimensions(2)
                    jj = j + win%lbound(2) - 1
                    jjj = jj + img%logical_dimensions(2)/2 + 1
                    if (jjj .lt. 1 .or. jjj .gt. img%logical_dimensions(2)) cycle
                    do i=1,win%logical_dimensions(1)
                        ii = i + win%lbound(1) - 1
                        iii = ii + img%logical_dimensions(1)/2 + 1
                        if (iii .lt. 1 .or. iii .gt. img%logical_dimensions(1)) cycle
                        ! copy voxel value over
                        img%real_values(iii,jjj,kkk) = img%real_values(iii,jjj,kkk) + win%rpointer(i,j,k) * sign
                    enddo
                enddo
            enddo
        else
            ! nothing to do
        endif
        !$omp   end critical    (copy_window_to_volume_real_outer)
    end subroutine window_to_image_real

    !>  \brief  window object data is copied into image object
    subroutine window_to_image_comp(img,win,subtract)
        use Images
        implicit none
        ! arguments
        type(image),                intent(inout)       ::  img
        type(window),               intent(in)          ::  win
        logical,        optional,   intent(in)          ::  subtract    !<  if .true., the values in the window are subtracted from the image. .false. by default.
        ! private variables
        integer ::  i,j,k
        integer ::  ii,jj,kk
        integer ::  iii,jjj,kkk
        integer ::  tmp_coos(3), tmp_coos_in(3)
        integer ::  coos_min(3),coos_max(3) !< the bounds of the iamge with origin 0,0,0 at center
        real    ::  sign

        ! start work
        coos_min = img%logical_lower_bound_complex
        coos_max = img%logical_upper_bound_complex

        ! are we subtracting or adding?
        if (present(subtract)) then
            if (subtract) then
                sign = -1.0
            else
                sign = 1.0
            endif
        else
            sign = 1.0
        endif

        !$omp   critical    (copy_window_to_volume_comp_outer)

        ! iterate over window voxels. i,j,k is address within window, ii,jj,kk is address within image (with origin 0,0,0 at center),
        ! and iii,jjj,kkk is address within image (with origin 1,1,1 at top-left)
        !write(*,*) '**debug: win%lbound =', win%lbound
        !write(*,*) '**debug: window%centered_on =', win%centered_on
        if (win%own_memory) then
            do k=1,win%logical_dimensions(3)
                kk  = k + win%lbound(3) - 1
                if (kk .lt. coos_min(3) .or. kk .gt. coos_max(3)) cycle
                do j=1,win%logical_dimensions(2)
                    jj = j + win%lbound(2) - 1
                    if (jj .lt. coos_min(2) .or. jj .gt. coos_max(2)) cycle
                    do i=1,win%logical_dimensions(1)
                        ii = i + win%lbound(1) - 1
                        if (ii .lt. coos_min(1) .or. ii .gt. coos_max(1)) cycle

                        ! get address within complex image which corresponds to voxel coordinate
                        tmp_coos_in = (/ii,jj,kk/)
                        !write(*,*) '**debug: coos =', tmp_coos
                        call img%PhysicalAddressGivenLogicalAddressInFourierSpace(tmp_coos_in,tmp_coos)
                        !write(*,*) '**debug: addr =', tmp_coos
                        iii = tmp_coos(1)
                        jjj = tmp_coos(2)
                        kkk = tmp_coos(3)

                        ! now copy the voxel value over
                        !write(*,'(a,2(f7.3,x))') '**debug(window_to_image_comp): cvalue =', win%complex_values(i,j,k)
                        !write(*,'(a,3(i3,x))') '**debug(window_to_image_comp): copying value over: iii,jjj,kkk =', iii,jjj,kkk
                        if (ii .ge. 0) then
                            img%complex_values(iii,jjj,kkk) = img%complex_values(iii,jjj,kkk) + win%cpointer(i,j,k) * sign
                        else
                            img%complex_values(iii,jjj,kkk) = img%complex_values(iii,jjj,kkk) + conjg(win%cpointer(i,j,k)) * sign
                        endif

                        ! at the origin of the first dimension, we have the unusual situation that friedel mates
                        ! are described explicitely, except for the nyquist components, which are only
                        ! described in the negative frequencies. therefore if a value lands anywhere on
                        ! the origin of the first dimension, we must make sure it is inserted a second time.
                        !if (ii .eq. 0 .and. jj .ne. 0 .and. (win%logical_dimensions(3) .eq. 1 .or. kk .ne. 0)) then
                        if (ii .eq. 0 .and. jj .ne. coos_min(2) .and. &
                                (win%logical_dimensions(3) .eq. 1 .or. kk .ne. coos_min(3))) then
                            ! need to also add a friedel/hermitian mate if we are on the origin in 1st dimension
                            tmp_coos_in = (/ii,-jj,-kk/)
                            call img%PhysicalAddressGivenLogicalAddressInFourierSpace(tmp_coos_in,tmp_coos)
                            iii = tmp_coos(1)
                            jjj = tmp_coos(2)
                            kkk = tmp_coos(3)
                            img%complex_values(iii,jjj,kkk) = img%complex_values(iii,jjj,kkk) + conjg(win%cpointer(i,j,k)) * sign
                        endif
                    enddo
                enddo
            enddo
        else
            ! nothing to do
        endif
        !$omp   end critical    (copy_window_to_volume_comp_outer)
    end subroutine window_to_image_comp

    !>  \brief  compute the lower and upper bound addresses of a window object. not the assumption that
    !!          central coordinates of window are given relative to origin (0,0,0) at center of image. also,
    !!          compute the coordinates of the window's center (i.e. where we're interpolating to / convoluting from),
    !!          with origin (1,1,1) at the top-left corner voxel.
    pure subroutine eval_window_bounds(win)
        implicit none
        ! arguments
        class(window),   intent(inout)   ::  win !<  window object
        ! private variables
        integer ::  l
        ! start work

        ! define the bounds of the window
        ! (note the window width is 2 for linear, .ge. 2 for sinc)
        ! these bounds can take any value, including 0, which means the address of the origin,
        ! and negative values
        do l=1,3
            win%lbound(l) = floor  (win%centered_on(l)) - (win%logical_dimensions(l)/2 - 1)
            win%ubound(l) = ceiling(win%centered_on(l)) + (win%logical_dimensions(l)/2 - 1)
            ! there is a special case in which the output coordinates are exactly on an output voxel. in this case, the center is
            ! at n/2+1
            if ( win%centered_on(l) .eq. anint(win%centered_on(l))) then ! we've fallen bang on a voxel coordinate
                win%lbound(l) = int(win%centered_on(l)) - (win%logical_dimensions(l)/2)
                win%ubound(l) = int(win%centered_on(l)) + (win%logical_dimensions(l)/2 - 1)
            endif
            ! now work out the coordinates of the "center"
            win%center(l) = win%centered_on(l) - real(win%lbound(l)) + 1.0
        enddo
        if (win%logical_dimensions(3) .eq. 1) then
            win%ubound(3) = 0
            win%lbound(3) = 0
            win%center(3) = 1.0
        endif

    end subroutine eval_window_bounds

    !impure elemental
    subroutine destroy_window(self)
        type(window), intent(inout) ::  self
        call self%DeallocateWindow()
    end subroutine destroy_window

    !>  \brief  deallocate memory assigned to a window object
    subroutine deallocate_window(self)
        use fftw33
        use runtime_parameters, only : terminate
        implicit none
        ! arguments
        class(window),   intent(inout)   ::  self
        ! private variables
        logical,        parameter       ::  debug   =   .false.
        ! start work
        if (debug)  write(*,'(a)') '**debug(deallocate_window): entering'
        if (self%is_in_memory) then
            if (self%own_memory) then
                if (associated(self%rpointer)) nullify(self%rpointer)
                if (associated(self%cpointer)) nullify(self%cpointer)
                if (associated(self%real_values)) nullify(self%real_values)
                if (associated(self%complex_values)) nullify(self%complex_values)
                !$omp   critical        (fftw_omp_crit)
                call fftwf_free(self%p)
                self%is_in_memory=.false.
                !$omp   end critical    (fftw_omp_crit)
            endif
            self%is_in_memory=.false.
        else
            !write(*,'(a)') '**info(deallocate_window): attempt at deallocating window which was not allocated'
            !call terminate('test deallocate_window')
        endif
        if (debug)  write(*,'(a)') '**debug(deallocate_window): exiting'
    end subroutine deallocate_window

    !>  \brief  allocate memory to the window object
    subroutine allocate_window(win,width,comp,threed,own_memory)
        use iso_c_binding
        use fftw33
        use runtime_parameters, only : is_odd, terminate
        implicit none
        ! arguments
        class(window),  target, intent(inout)   ::  win         !<  the window to allocate memory to
        integer,                intent(in)      ::  width       !<  window width in pixels
        logical,                intent(in)      ::  comp        !<  is it complex or real?
        logical,                intent(in)      ::  threed      !<  3-dimensional?
        logical,    optional,   intent(in)      ::  own_memory  !<  whether the window should have its own area in memory.
        ! private variables
        integer ::  alloc_dim(3)
        ! start work

        ! copy type over from image
        if (comp) then
            win%is_in_real_space = .false.
        else
            win%is_in_real_space = .true.
        endif

        ! will the window have its own area in memory?
        if (present(own_memory)) then
            win%own_memory = own_memory
        else
            if (win%is_in_real_space) then
                win%own_memory = .false.
            else
                win%own_memory = .true.
            endif
        endif


        ! own memory is compulsory for comp windows
        if (.not. win%is_in_real_space) then
            win%own_memory = .true.
        endif

        ! own memory is compulsory if we are skipping runtime checks, to make things faster
#ifdef skip_runtime_checks
        win%own_memory = .true.
#endif

        ! own memory is compulsory for comp windows
        ! if the window is complex, we cannot do the same tricks with pointers that we can with real, because of the "folding" of the
        ! first dimension, which cannot be accomodated by window_from_image_comp
        if (.not. win%own_memory .and. .not. win%is_in_real_space) then
            write(*,'(a)') '**error(allocate_window): complex windows must have their own memory'
            call terminate('fatal error in allocate_window')
        endif

        ! if window already allocated
        if (win%own_memory .and. win%is_in_memory) then
            write(*,'(a)') '**error(allocate_window): window already in memory'
            call terminate('fatal error in allocate_window')
        endif

        ! check width is even and positive
        if (is_odd(width) .or. width .le. 0) then
            write(*,'(a,i0)') '**error(allocate_window): window width must be even positive: ', width
            call terminate('fatal error in allocate_window')
        endif

        ! set dimensions of window
        win%width       =   width
        win%logical_dimensions(1:3)    =   width
        if (.not. threed) win%logical_dimensions(3) = 1

        ! work out memory allocation size (we will allocate complex array)
        alloc_dim = win%logical_dimensions
        if (win%is_in_real_space) then
            alloc_dim(1) = alloc_dim(1) / 2
        endif
        if (win%logical_dimensions(3) .eq. 1) alloc_dim(3) = 1

        ! allocate memory if necessary
        if (win%own_memory) then
            ! do the allocation
            !$omp   critical        (fftw_omp_crit)
            !win%p = fftwf_alloc_complex(int(product(alloc_dim),c_size_t))
            win%p = fftwf_malloc(int(product(alloc_dim)*8,c_size_t))
            win%is_in_memory = .true.
            !$omp   end critical    (fftw_omp_crit)
            call c_f_pointer(win%p,win%complex_values,alloc_dim)
            alloc_dim(1) = alloc_dim(1) * 2
            call c_f_pointer(win%p,win%real_values,alloc_dim)

            ! finally, initialise the contents
            win%real_values = 0.0

            ! and setup the pointers
            win%rpointer => win%real_values
            win%cpointer => win%complex_values
        else
            ! the window does not have its own memory. rather,
            ! when it comes to hovering over an image/volume,
            ! the r and c pointers will be associated with sections
            ! of the image/volume's data arrays, thus avoiding
            ! copying of data from image to window
        endif
        win%is_in_memory = .true.
    end subroutine allocate_window


end module InterpolatingWindows
