!>  \brief  Interpolation functions
!! The instrument function is the reciprocal of the interpolation function.
!! The arguments to the interpolation functions are given in pixels
!! The arguments to the instrument functions are given in boxsizes.
module interpolation
    use interpolation_kernels
    implicit none

    type :: interpolation_scheme
        procedure(interpolation_function),  pointer,    nopass  ::  interpolation_function          =>  null()
        procedure(windowing_function),      pointer,    nopass  ::  windowing_function              =>  null()
        integer                                                 ::  window_width
        procedure(interpolation_function),  pointer,    nopass  ::  interpolation_instr_function    =>  null()
        procedure(windowing_function),      pointer,    nopass  ::  windowing_instr_function        =>  null()
        ! With this interpolation scheme, this is the safe radius for the region of interest (roi), expressed
        ! as fraction of box size (i.e. maximum radius is 0.5)
        real                                                    ::  roi_safe_radius                 =   0.0
        ! With this interpolation scheme, this is an estimate of the expected power of aliases (where 1.0 would be power of signal)
        real                                                    ::  power_of_aliases                =   0.0
        ! Desired maximum power of aliases
        real                                                    ::  power_of_aliases_max_allowed    =   0.05
        ! How many aliases do we want to consider when evaluating the power of aliases etc?
        integer                                                 ::  num_aliases_to_consider         =   10
        !
        character(len=4)                                        ::  id                              =   '    '
    end type interpolation_scheme

    !>  \brief  This defines what an interpolation function should look like.
    abstract interface
        pure real function interpolation_function(arg)
            real,   intent(in)  ::  arg
        end function interpolation_function
    end interface

    !>  \brief  a windowing function is almost like an interpolating function, but its width is one of the arguments
    abstract interface
        pure real function windowing_function(arg,width)
            real,       intent(in)  ::  arg
            integer,    intent(in)  ::  width
        end function windowing_function
    end interface

    real,   parameter   ::  default_roi_safe_radius = 0.275


    contains

    !>  \brief  Return a reasonable default interpolation scheme
    pure type(interpolation_scheme) function default_interpol_scheme()
        implicit none
        ! Arguments
        ! Private variables
        ! Start work
        default_interpol_scheme%interpolation_function          =>  linear
        default_interpol_scheme%windowing_function              =>  rect
        default_interpol_scheme%window_width                    =   2
        default_interpol_scheme%interpolation_instr_function    =>  triangle_instr
        default_interpol_scheme%windowing_instr_function        =>  rect_instr
        default_interpol_scheme%roi_safe_radius                 =   default_roi_safe_radius
        default_interpol_scheme%id                              =   'LINE'
    end function default_interpol_scheme

    !>  \brief  Check whether the interpolation scheme object was setup properly
    pure logical function interpolation_scheme_check_setup(self)
        implicit none
        ! Arguments
        type(interpolation_scheme),     intent(in)      ::  self                    !<  Interpolation scheme object
        ! Private variables

        ! Start work

        if (            self%window_width .gt. 0                        &
                .and.   associated(self%interpolation_function)         &
                .and.   associated(self%windowing_function)             &
                .and.   associated(self%interpolation_instr_function)   &
                .and.   associated(self%windowing_instr_function)       &
                .and.   self%roi_safe_radius .gt. 0.0                   &
                .and.   self%power_of_aliases .gt. 0.0) then

            interpolation_scheme_check_setup = .true.
        else
            interpolation_scheme_check_setup = .false.
        endif
    end function interpolation_scheme_check_setup

    !>  \brief  Initialise an interpolation scheme, based on string descriptions of the functions to use
    subroutine interpolation_scheme_init(self,interpol_descr,window_descr,window_width,interpol_instr_descr,window_instr_descr)
        use parsers
        use runtime_parameters, only    :   word_len, terminate
        implicit none
        ! Arguments
        type(interpolation_scheme),     intent(inout)   ::  self                    !<  Interpolation scheme to be initialised
        character(len=*),               intent(in)      ::  interpol_descr          !<  String describing the interpolation function to be used (e.g. linear, SINC...)
        character(len=*),               intent(in)      ::  window_descr            !<  String describing the windowing function to be used (e.g. rect, hann)
        integer,                        intent(in)      ::  window_width            !<  Window width
        character(len=*),   optional,   intent(in)      ::  interpol_instr_descr    !<  String describing the instrument function for the interpolation function to be used (e.g. linear, SINC...)
        character(len=*),   optional,   intent(in)      ::  window_instr_descr      !<  String describing the instrument function for the windowing function to be used (e.g. rect, hann)
        ! Private variables
        character(len=WORD_LEN) ::  iinterpol_instr_descr
        character(len=WORD_LEN) ::  wwindow_instr_descr
        logical,    parameter   ::  debug = .false.
        ! Start work

        ! Output information to the user
        if (debug) then
            write(*,'(3(A,1x),i0)') '**debug(interpolation_scheme_init): initialising interpolation scheme: ', &
                            trim(interpol_descr), trim(window_descr), window_width
        endif

        ! If they were not explicitely given, work out the descriptors for the instrument functions
        if (present(interpol_instr_descr)) then
            iinterpol_instr_descr = trim(adjustl(interpol_instr_descr))
        else
            iinterpol_instr_descr = trim(adjustl(interpol_descr))//'_INSTR'
        endif
        if (present(window_instr_descr)) then
            wwindow_instr_descr = trim(adjustl(window_instr_descr))
        else
            wwindow_instr_descr = trim(adjustl(window_descr))//'_INSTR'
        endif

        ! Set the window width
        self%window_width = window_width

        if (debug) print '(5(A,1x),i3)', '**debug(interpolation_scheme_init): setting up with', trim(interpol_descr),    &
                                        trim(window_descr), trim(iinterpol_instr_descr), trim(wwindow_instr_descr), window_width

        ! Find the windowing function
        select case (upper_case(trim(adjustl(window_descr))))
            case ('HANN')
                self%windowing_function => hann
            case ('RECT')
                self%windowing_function => rect
            case ('UNITY')
                self%windowing_function => unity_window
            case default
                write(*,'(2a)') '**ERROR(interpolation_scheme_init): could not work out which windowing function you meant by ', &
                                upper_case(trim(adjustl(window_descr)))
                call terminate('Fatal error in INTERPOLATION_SCHEME_INIT')
        end select

        ! Find the interpolation function
        select case (upper_case(trim(adjustl(interpol_descr))))
            case ('SINC')
                self%interpolation_function => sinc
                self%id = 'SINC'
            case ('NEAREST')
                self%interpolation_function => nearest_neighbor
                self%id = 'NEAR'
            case ('TRIANGLE')
                self%interpolation_function => triangle
                self%id = 'LINE'
            case ('LINEAR')
                self%interpolation_function => linear
                self%id = 'LINE'
            case ('UNITY')
                self%interpolation_function => unity
                self%id = 'UNIT'
            case default
                write(*,'(3a)') '**ERROR(interpolation_scheme_init): could not work out which ',    &
                                'interpolation function you meant by ', trim(adjustl(interpol_descr))
                call terminate('Fatal error in INTERPOLATION_SCHEME_INIT')
        end select

        ! Find the windowing instrument function
        select case (upper_case(trim(adjustl(wwindow_instr_descr))))
            case ('HANN_INSTR')
                self%windowing_instr_function => hann_instr
            case ('RECT_INSTR')
                self%windowing_instr_function => rect_instr
            case ('UNITY_INSTR')
                self%windowing_instr_function => unity_window_instr
            case default
                write(*,'(3a)') '**ERROR(interpolation_scheme_init): could not work out which window',  &
                                ' instrument function you meant by ', trim(adjustl(wwindow_instr_descr))
                call terminate('Fatal error in INTERPOLATION_SCHEME_INIT')
        end select

        ! Find the interpolation instrument function
        select case (upper_case(trim(adjustl(iinterpol_instr_descr))))
            case ('SINC_INSTR')
                self%interpolation_instr_function => sinc_instr
            case ('NEAREST_INSTR')
                self%interpolation_instr_function => nearest_instr
            case ('TRIANGLE_INSTR')
                self%interpolation_instr_function => triangle_instr
            case ('LINEAR_INSTR')
                self%interpolation_instr_function => triangle_instr
            case ('UNITY_INSTR')
                self%interpolation_instr_function => unity_instr
            case default
                write(*,'(3a)') '**ERROR(interpolation_scheme_init): could not work out which interpolation ',  &
                                'instrument function you meant by ', trim(adjustl(iinterpol_instr_descr))
                call terminate('Fatal error in INTERPOLATION_SCHEME_INIT')
        end select

        ! Now try to work out a safe roi radius
        call interpol_find_roi_safe_radius(self)
    end subroutine interpolation_scheme_init

    !>  \brief  Heuristics to find safe radius of roi
    subroutine interpol_find_roi_safe_radius(self)
        implicit none
        ! Arguments
        type(interpolation_scheme), intent(inout)   ::  self    !<  Interpolation scheme
        ! Private variables
        real    ::  current_roi_radius, current_alias_power
        logical,    parameter   ::  debug   =   .false.
        real,       parameter   ::  search_step_size = 0.025
        logical,    parameter   ::  skip    =   .false.
        real,       parameter   ::  skip_value  =   default_roi_safe_radius
        ! Start work

        if (debug) then
            write(*,'(a)')  '**debug(interpol_find_roi_safe_radius): searching for the maximum safe radius for the '// &
                            'region of interest...'
        endif

        if (skip .or. self%id .eq. 'LINE') then
            self%roi_safe_radius = skip_value
            self%power_of_aliases = self%power_of_aliases_max_allowed
        else
            ! Loop over roi radius values (expressed as fraction of box size, so that 0.5 is maximum radius)
            current_roi_radius = 0.4
            do
                ! Estimate alias power with the given roi radius
                current_alias_power = estimate_power_of_aliases(self,current_roi_radius)
                ! Diagnostic
                if (debug) then
                    write(*,'(a,F0.4,1x,F0.4)')  '**debug(interpol_find_roi_safe_radius): roi radius, alias power: ', &
                                                current_roi_radius, current_alias_power
                endif
                if (current_alias_power .lt. self%power_of_aliases_max_allowed) then
                    exit
                endif
                ! Increment
                current_roi_radius = current_roi_radius - search_step_size
                ! Check if within bounds
                if (current_roi_radius .lt. 0.1) exit
            enddo

            ! Remember this value
            self%roi_safe_radius = current_roi_radius
            self%power_of_aliases = current_alias_power

            ! Diagnostics
            if (debug) then
                write(*,'(a,F0.4)') '**debug(interpol_find_roi_safe_radius): the maximum safe radius for the roi '// &
                                    'with this interpolation scheme is: ', self%roi_safe_radius
            endif
        endif
    end subroutine interpol_find_roi_safe_radius

    !>  \brief  Estimate the power of aliases
    function estimate_power_of_aliases(self,roi_radius)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(interpolation_scheme), intent(in)  ::  self        !<  Interpolation scheme
        real,                       intent(in)  ::  roi_radius  !<  Radius of the region of interest (as fraction of box size, so that 0.5 is the maximum radius)
        ! Result
        real(kind=8)    ::  estimate_power_of_aliases
        ! Private variables
        integer ::  current_alias
        real(kind=8)    ::  current_position
        real(kind=8)    ::  x
        real(kind=8),   parameter   ::  position_increment  =   0.001D0
        real(kind=8)    ::  atten_function_value, atten_function_value_alias    !   Values for the attenuation value
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        ! Check roi_radius is >0.0 and <=0.5
        if (roi_radius .le. 0.0 .or. roi_radius .gt. 0.5) then
            write(*,'(a,F0.4,a)') '**ERROR(estimate_power_of_aliases): roi radius (', roi_radius, ') out of range 0.0 - 0.5'
            call terminate('Fatal error in ESTIMATE_POWER_OF_ALIASES')
        endif

        ! Initialise the result
        estimate_power_of_aliases = 0.0D0

        ! Loop over the region of interest
        current_position = - dble(roi_radius)
        do
            ! Evaluate the attenuation function at the current position
            atten_function_value = interpol_window_convolution( real(current_position),             &
                                                                self%interpolation_instr_function,  &
                                                                self%windowing_instr_function,      &
                                                                self%window_width,                  &
                                                                BOX_SIZE=512    )
            ! Loop over the aliases
            do current_alias = -SELF%num_aliases_to_consider, self%num_aliases_to_consider
                ! We only consider aliases, not the central box
                if (current_alias .eq. 0) cycle
                ! The current aliased position
                x = current_position - dble(current_alias)
                ! Evaluate the attenuation function at the current aliased position
                atten_function_value_alias = interpol_window_convolution(   real(x),                            &
                                                                            self%interpolation_instr_function,  &
                                                                            self%windowing_instr_function,      &
                                                                            self%window_width,                  &
                                                                            BOX_SIZE=512    )
                !
                estimate_power_of_aliases = estimate_power_of_aliases + abs(atten_function_value_alias)!/abs(atten_function_value)

            enddo   ! End of loop over aliases
            ! Increment current_position
            current_position = current_position + position_increment
            ! Evaluate
            ! If we've gone over the edge, stop
            if (current_position .gt. roi_radius) exit
        enddo ! End of loop over roi

        ! The integral was in steps
        estimate_power_of_aliases = estimate_power_of_aliases * position_increment

        ! Diagnostic
        if (debug) then
            write(*,'(a,F0.3,a,F0.4)') '**debug(estimate_power_of_aliases): for radius ', roi_radius, &
                                        ' the estimated alias power is ', estimate_power_of_aliases
        endif
    end function estimate_power_of_aliases

    !>  \brief  Return the convolution of interpol function interpol with windowing function window by computing the actual convolution
    !!          integral.
    !!          This is useful when convolving instrument functions, e.g. to workout the falloff imposed in real-space on a 3d reconstruction
    !!          which was generated by inserting 2d slices in 3d reciprocal space using the interpolation and windowing schemes.
    pure real function interpol_window_convolution(arg,interpol,window,window_width,box_size)
        use runtime_parameters, only : is_even
        implicit none
        ! Arguments
        real,                               intent(in)      ::  arg             !<  Argument at which the convoluted function is evaluated
        procedure(interpolation_function),  pointer         ::  interpol        !<  Interpolation function
        procedure(windowing_function),      pointer         ::  window          !<  Windowing function
        integer,                            intent(in)      ::  window_width    !<  Window width
        integer,                            intent(in)      ::  box_size        !<  How many values (pixels) between and including ARG=0.0 and ARG=1.0. In the case of convolution of instrument functions, this is the box size of the 3d volume.
        ! Private variables
        integer,    parameter   ::  num_of_boxes    =   10  !<  How far will we consider the functions? For perfect convolution we need this to be infinity, but mostly we can approximate by using a relatively small number here.
                                                            !!  If num_of_boxes == 0, only the central box is considered.
        integer ::  i, j
        real    ::  x
        real(kind=8)    :: conv
        integer ::  orig
        ! Start work
        if (is_even(box_size)) then
            orig = box_size/2+1
        else
            orig = (BOX_SIZE-1)/2+1
        endif
        conv = 0.0D0
        ! We first do the origin (i.e. x = 0.0)
        !conv = dble(interpol(0.0)*WINDOW(arg,window_width))
        conv = dble(interpol(arg)*WINDOW(0.0,window_width))
        ! Then we do the rest of the integral. !Since our functions are all mirror symmetric, we can just do a half of the integral and *2?
        do I=-NUM_OF_BOXES,num_of_boxes
            do J=1,box_size
                !x = real((J+I*BOX_SIZE))/real(box_size)
                x = real((J+I*BOX_SIZE)-ORIG)/real(box_size) ! 30-dec-2010
                !conv = conv + dble(interpol(x)*WINDOW(ARG-X,window_width))
                conv = conv + dble(interpol(ARG-X)*WINDOW(x,window_width))
                !if (arg .EQ. 0.5) then
                !   print '(3(A,F7.2))', '**debug(interpol_window_convolution): x, interpol * window = ', x, ',', interpol(ARG-X), ' * ', window(x,window_width)
                !endif
            enddo
        enddo

        interpol_window_convolution = real(conv)/real(orig)
        !print '(2(A,F7.2))', '**debug(interpol_window_convolution): arg, result = ', arg, ',', interpol_window_convolution

    end function interpol_window_convolution

end module interpolation
