!>  \brief  a library of kernel functions used in interpolation
module interpolation_kernels

contains

    !>  \brief  Returns the normalised sinc of arg
    !!          If arg = 0, SINC = 1. If arg is integer-valued, SINC=0.
    pure real function sinc(arg)
        use runtime_parameters, only : pi
        implicit none
        real, intent(in)    ::  arg
        ! Private variables
        real    ::  arg2
        real,   parameter   ::  small   =   1.0e-6
        ! Start work
        arg2 = abs(pi * arg)
        if (arg2 .le. small) then
            sinc = 1.0
        else if (abs(ARG2-AINT(arg2)) .le. small) then
            sinc = 0.0
        else
            sinc = sin(arg2)/arg2
        endif
    end function sinc

    !>  \brief  Returns the square of the sinc of arg
    pure real function sinc2(arg)
        implicit none
        real,   intent(in)  ::  arg
        sinc2 = sinc(arg)**2
    end function sinc2

    !>  \brief  Returns the instrument function to sinc, which is a rectangle
    pure real function sinc_instr(arg)
        implicit none
        real,   intent(in)  ::  arg
        sinc_instr = rect(arg,1)
    end function sinc_instr

    !>  \brief  Returns the normalised linear function of arg
    !!          If arg = 0., linear = 1. If arg .GE. 1.0, linear = 0.
    pure real function linear(arg)
        implicit none
        real, intent(in)    ::  arg
        linear = triangle(arg)
    end function linear

    !>  \brief  The instrument function to linear interpolation kernel
    pure real function linear_instr(arg)
        implicit none
        real, intent(in)    ::  arg
        linear_instr = triangle_instr(arg)
    end function linear_instr

    !>  \brief The triangle function, used in linear interpolation.
    pure real function triangle(arg)
        implicit none
        real,   intent(in)  ::  arg
        ! Private variable
        real    ::  aarg
        ! Start work
        aarg = abs(arg)
        if (aarg .gt. 1.0) then
            triangle = 0.0
        else
            triangle = 1 - aarg
        endif
    end function triangle

    !>  \brief  Instrument function corresponding to the triangle function
    pure real function triangle_instr(arg)
        implicit none
        real,   intent(in)  ::  arg
        !triangle_instr = sinc2_with_aliases(arg)
        triangle_instr = sinc2(arg)
    end function triangle_instr

    !>  \brief  Returns 0.0 if abs(arg) .GT. 0.5, else 1.0
    !!          Warning: make sure the argument is positive!
    pure real function nearest_neighbor(arg)
        implicit none
        real, intent(in)    ::  arg
        nearest_neighbor = rect(arg,1)
    end function nearest_neighbor

    pure real function nearest_instr(arg)
        implicit none
        real,   intent(in)  ::  arg
        !nearest_instr = sinc_with_aliases(arg) ! needs to be DOUBLE-CHECKED
        nearest_instr = rect_instr(arg,1)
    end function nearest_instr

    !>  \brief  Returns the Hann function of arg (radians), given WIDTH. width is the full width
    !!          When arg = width, hann = 1.0 (this should never happen, since this function should not be called with arg .GT. (WIDTH-1)/2)
    !!          When arg = 0, HANN = 1.0
    !!          When arg = width/2, hann = 0
    pure real function hann(arg,width)
        use runtime_parameters, only: pi
        implicit none
        real, intent(in) :: arg
        integer, intent(in) :: width !< Full width of the window (centered at ARG=0)
        ! Private variables
        real    ::  rwidth, hwidth
        ! Start work
        rwidth = real(width)
        hwidth = rwidth/2.0
        if (arg .lt. hwidth) then
            hann = 0.5 * ( 1 + cos(PI*ARG/hwidth))
        else
            hann = 0.0
        endif
    end function

    !>  \brief  The instrument function corresponding to the Hann function
    !!  See http://mathworld.wolfram.com/ApodizationFunction.html and
    !!  http://mathworld.wolfram.com/HanningFunction.html
    !!  for more details
    pure real function hann_instr(arg,width)
        use runtime_parameters, only: pi
        implicit none
        real,       intent(in)  ::  arg
        integer,    intent(in)  ::  width
        ! Private variable
        real :: hwidth, rwidth, tmp
        ! Start work
        rwidth = real(width)
        hwidth = rwidth / 2.0
        !tmp = 2*PI*ARG*HWIDTH
        !tmp = 2*ARG*RWIDTH
        tmp = ARG*RWIDTH
        !hann_instr = HWIDTH*SINC(PI*RWIDTH*ARG)/(1-RWIDTH**2*ARG**2)   ! This is the quicker expression to evaluate, but it can lead to div/0.0 with arg = 0.5
        !hann_instr = HWIDTH*(sinc(tmp)+.5*SINC(TMP-PI)+.5*SINC(TMP+PI))
        !hann_instr = sinc(tmp)+.5*SINC(TMP-PI)+.5*SINC(TMP+PI)
        hann_instr = sinc(tmp)+.5*SINC(TMP-1.)+.5*SINC(TMP+1.)
        !print '(a,F8.3,x,i2,x,F10.3)', '**debug(hann_instr): arg, width, value = ', arg, width, hann_instr
    end function hann_instr

    !>  \brief  Top-hat / rectangle function is 1.0 if arg .le. width/2.0
    pure real function rect(arg,width)
        implicit none
        real,       intent(in) :: arg
        integer,    intent(in) :: width
        if (abs(arg) .gt. real(width)/2.) then
            rect = 0.0
        else
            rect = 1.0
        endif
    end function rect

    !>  \brief  The instrument function corresponding to the rectangle window function
    pure real function rect_instr(arg,width)
        implicit none
        real,       intent(in)  ::  arg
        integer,    intent(in)  ::  width
        !rect_instr = sinc_with_aliases(arg,real(width))
        rect_instr = sinc(arg*real(width))
    end function rect_instr

    pure real function unity_window(arg,width)
        implicit none
        real,       intent(in)  ::  arg
        integer,    intent(in)  ::  width
        unity_window = 1.0
    end function unity_window

    pure real function unity_window_instr(arg,width)
        implicit none
        real,       intent(in)  ::  arg
        integer,    intent(in)  ::  width
        unity_window_instr = dirac(arg)
    end function unity_window_instr

    !>  \brief  Always returns 1.0
    pure real function unity(arg)
        implicit none
        real, intent(in) :: arg
        unity = 1.0
    end function unity

    !>  \brief  The instrument function corresponding to the unity function
    pure real function unity_instr(arg)
        implicit none
        real,   intent(in)  ::  arg
        unity_instr = dirac(arg)
    end function unity_instr

    !>  \brief  The Dirac delta / impulse 'improper function'
    pure real function dirac(arg)
        implicit none
        real,   intent(in)  ::  arg
        if (arg .eq. 0.0) then
            dirac = 1.0
        else
            dirac = 0.0
        endif
    end function dirac


end module interpolation_kernels
