module linear_algebra
    implicit none

    contains

    !>  \brief  Solves a set of linear equations by calling lu decomposition routine from gsl
    !!
    !!  (a)*(x) = (b)
    !!
    !!  (a) is a matrix of doubles (known, input), (x) is the vector of unknowns, (b) is right-hand-side vector (known)
    !!
    !!  Internally, these are "converted" to fgsl data types
    !!
    subroutine linalg_solve_lu_fgsl(af,xf,bf,nf)
        use fgsl
        implicit none
        ! Arguments
        real(kind=8),   allocatable,    intent(in)      ::  af(:,:) !<  Matrix (N*N)
        real(kind=8),   allocatable,    intent(inout)   ::  xf(:)   !<  Unknown vector (output) (n)
        real(kind=8),   allocatable,    intent(in)      ::  bf(:)   !<  Right-hand-side of the equation (n)
        integer,                        intent(in)      ::  nf      !<  Dimension of matrix & vectors
        ! Private variables
        type(fgsl_matrix)       ::  a
        type(fgsl_vector)       ::  b,x
        type(fgsl_permutation)  ::  p
        integer(fgsl_int)       ::  status, signum
        integer(fgsl_size_t)    ::  n
        ! Start work
        n = nf
        ! Initialise arrays and matrix
        a = fgsl_matrix_init(TYPE=1.0_fgsl_double)
        b = fgsl_vector_init(TYPE=1.0_fgsl_double)
        x = fgsl_vector_init(TYPE=1.0_fgsl_double)
        p = fgsl_permutation_alloc(n)
        ! Convert fortran vectors / matrices to fgsl types
        status = fgsl_matrix_align(af,n,n,n,a)
        status = fgsl_vector_align(bf,n,b,n,0_fgsl_size_t,1_fgsl_size_t)
        status = fgsl_vector_align(xf,n,x,n,0_fgsl_size_t,1_fgsl_size_t)
        ! Decompose & solve
        status = fgsl_linalg_lu_decomp(a,p,signum)
        status = fgsl_linalg_lu_solve(a,p,b,x)
        ! Copy results over to Fortran double !! not sure if this correct way to do it
        xf = x
        ! Free up fgsl objects
        call fgsl_matrix_free(a)
        call fgsl_vector_free(b)
        call fgsl_vector_free(x)
        call fgsl_permutation_free(p)
    end subroutine

    !>  \brief  Solves a set of linear equations by calling lu decomposition routine from lapack
    subroutine linalg_solve_lapack(a,x,b)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        real(kind=8),   allocatable,    intent(inout)   ::  a(:,:) !<  Matrix (N*N)
        real(kind=8),   allocatable,    intent(inout)   ::  x(:)   !<  Unknown vector (output) (n)
        real(kind=8),   allocatable,    intent(in)      ::  b(:)   !<  Right-hand-side of the equation (n)
        ! Private variables
        integer ::  n, lda, ldb,nrhs
        integer, allocatable :: ipiv(:)
        integer ::  info
        ! Start work

        ! The lapack routine will overwrite the 'B' argument with the solution x
        x = b
        ! The other arguments to the lapack routine
        n = size(a,2)
        nrhs=1
        lda = size(a,1)
        ldb = size(b,1)
        allocate(ipiv(n))
        !
        call dgesv(n,nrhs,a,lda,ipiv,x,ldb,info)
        ! Check return status
        if (info .gt. 0) then
            write(*,'(2a,i0)') '**ERROR(linalg_solve_lapack): factorisation was completed', &
                            ' but the factor u is exactly singular so the solution could not be computed, i = ', info
            call terminate('Fatal error in LINALG_SOLVE_LAPACK')
        else if (info .lt. 0) then
            write(*,'(a,i0)') '**ERROR(linalg_solve_lapack): this argument had an illegal value: ', info
            call terminate('Fatal error in LINALG_SOLVE_LAPACK')
        endif
    end subroutine linalg_solve_lapack

    real function matrix_det(m)
        use fgsl
        implicit none
        ! Arguments
        real(kind=8),   allocatable,    intent(in)      ::  m(:,:)
        ! Private variables
        type(fgsl_matrix)   ::  mm
        integer(fgsl_int)       ::  status, signum
        integer(fgsl_size_t)    ::  n
        ! Start work
        mm = fgsl_matrix_init(TYPE=1.0_FGSL_DOUBLE)
        n = size(m,1)
        status = fgsl_matrix_align(m,n,n,n,mm)
        matrix_det = real(fgsl_linalg_lu_det(mm,signum))
    end function matrix_det
end module linear_algebra

