!>  \brief  Module with masking methods
!!  \todo   Improve naming convention for subroutines, and also consistency of arguments.
module masks

    contains


    !>  \brief  Convenience function to generate a Fourier mask suited to diffraction from projected amyloid fibrils
    subroutine fourier_mask_init_amyloid(self,mould,pixel_size,fil_max_rad,max_spacing_align,angles)
        use images
        use euleranglesmod, only : euler
        implicit none
        ! Arguments
        type(image),                intent(inout)   ::  self                !<  The Fourier mask
        type(image),                intent(in)      ::  mould               !<  The mask will be compatible with images of this dimension
        real,                       intent(in)      ::  pixel_size          !<  Pixel size in Angstroms
        real,                       intent(in)      ::  fil_max_rad         !<  Maximum radius of the filament (Angstroms)
        real,                       intent(in)      ::  max_spacing_align   !<  The maximum spacing (Angstroms) used during alignments
        type(euler),    optional,   intent(in)      ::  angles              !<  Euler angles describing the orientation of the filament
        ! Private variables
        real,       parameter   ::  layer_line_spacing              =   4.8             ! Angstroms
        real,       parameter   ::  layer_line_spacing_half_range   =   0.2             ! Angstroms
        !real,       parameter   ::  hp_spacing                      =   200.0           ! Angstroms - this is for Ab(40), which have diameter of ~180A
        !real,       parameter   ::  lp_spacing                      =   8.0             ! Angstroms
        logical,    parameter   ::  debug                           =   .false.
        type(image)             ::  self_copy   ! used when debugging
        ! Start work

        ! Initialise a Fourier mask
        call fourier_mask_init(self,mould)

        ! Band-pass the low frequencies
        !call fourier_mask_add_hollow_sphere(    self,FREQ_INNER=PIXEL_SIZE/hp_spacing,FREQ_OUTER=PIXEL_SIZE/lp_spacing)
        call fourier_mask_add_hollow_spheroid(  self,   &
                                                freq_outer_main     = pixel_size / 15.0,  &
                                                freq_inner_perp     = pixel_size / (FIL_MAX_RAD*1.2),    &
                                                freq_outer_perp     = pixel_size / max_spacing_align,   &
                                                main_axis_angles    = angles )

        ! Band-pass the layer line
        call fourier_mask_add_hollow_cylinder(  self,   &
                                                freq_inner = 0.0,   &
                                                freq_outer = pixel_size/25.0, &
                                                freq_base  = pixel_size/(LAYER_LINE_SPACING+LAYER_LINE_SPACING_HALF_RANGE), &
                                                freq_top   = pixel_size/(LAYER_LINE_SPACING-LAYER_LINE_SPACING_HALF_RANGE), &
                                                angles     = angles )

        ! Dump to disk a version of the Fourier mask easy to look at
        if (debug) then
            call self%WriteToDisk('dbg_fourier_mask_amyloid.mrc')
            self_copy = self
            call self_copy%SpectrumFromHalfToFull()
            call self_copy%WriteToDisk('dbg_fourier_mask_amyloid_full.mrc')
        endif
    end subroutine fourier_mask_init_amyloid

    !>  \brief  Intialise a real-space image with dimensions suitable for multiplication with Fourier-transformed images so that
    !!          it can act as a mask in Fourier space
    subroutine fourier_mask_init(self,mould)
        use images
        implicit none
        ! Arguments
        type(image),                intent(inout)   ::  self            !<  The Fourier mask to be initialised
        type(image),                intent(in)      ::  mould           !<  The mask will be suitable for images with the mould's dimensions
        ! Private variables

        ! Start work

        ! Deallocate if necessary
        if (self%IsAllocated()) call self%Deallocate()

        ! Work out logical dimensions
        self%logical_dimensions(1)   = mould%physical_upper_bound_complex(1)
        self%logical_dimensions(2:3) = mould%logical_dimensions(2:3)

        ! Mask will be real
        self%is_in_real_space = .true.

        ! Allocate memory (this also sets data arrays to 0)
        call self%Allocate()

    end subroutine fourier_mask_init

    !>  \brief  Add an ellipsoid
    !!
    subroutine fourier_mask_add_hollow_spheroid(self,freq_outer_main,freq_inner_perp,freq_outer_perp,   &
                                                main_axis_angles)
        use images
        use runtime_parameters, only : terminate
        use euleranglesmod, only : euler
        use rotation_matrices, only : euler_to_rotmat, print_rotmat
        implicit none
        ! Arguments
        type(image),                intent(inout)   ::  self                !<  Fourier mask to be modified
        real,                       intent(in)      ::  freq_outer_main     !<  Spatial frequency (1/pixel; Nyquist is 0.5) of outer radius along the main axis
        real,                       intent(in)      ::  freq_inner_perp     !<  Spatial frequency of inner radius along the two normal axes
        real,                       intent(in)      ::  freq_outer_perp     !<  Spatial frequency of outer radius along the two normal axes
        type(euler),    optional,   intent(in)      ::  main_axis_angles    !<  Euler angles defining the main spheroid axis
        ! Private variables
        integer ::  i,j,k
        real    ::  x,y,z
        real    ::  sph_sf
        logical ::  in_spheroid
        type(euler) :: aangles
        real    ::  coos(4)
        real    ::  rot(4,4)
        ! Start work

        ! The Fourier mask should be a real image
        if (.not. self%IsInRealSpace()) then
            write(*,'(a)') '**ERROR(fourier_mask_add_hollow_spheroid): Fourier mask images should be real'
            call terminate('Fatal error in FOURIER_MASK_ADD_HOLLOW_SPHEROID')
        endif

        ! Check frequencies make sense
        if (freq_outer_perp .lt. freq_inner_perp) then
            write(*,'(a)') '**ERROR(fourier_mask_add_hollow_spheroid): top frequency must be greater than base frequency'
            call terminate('Fatal error in FOURIER_MASK_ADD_HOLLOW_SPHEROID')
        endif

        ! The axis
        aangles%phi   = 0.0
        aangles%theta = 90.0
        aangles%psi   = 90.0
        if (present(main_axis_angles)) aangles = main_axis_angles
        rot = euler_to_rotmat(aangles%phi,aangles%theta,aangles%psi,REVERSE=.true.)

        ! Loop over voxels
        do K=1,self%logical_dimensions(3)
            z = real(self%LogicalIndexGivenPhysicalIndexInFourierSpace(k,3))*self%fourier_voxel_size(3)
            do J=1,self%logical_dimensions(2)
                y = real(self%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2))*self%fourier_voxel_size(2)
                do I=1,self%logical_dimensions(1)
                    x = real(I-1)/real(self%logical_dimensions(1)-1)*0.5
                    !
                    coos = matmul(rot,[X,y,0.0,1.0])
                    !
                    sph_sf = sqrt(   coos(1)**2  &
                                   + coos(2)**2  &
                                   + coos(3)**2 * (freq_outer_perp/freq_outer_main)**2)
                    !

                    in_spheroid = sph_sf .ge. freq_inner_perp .and. sph_sf .le. freq_outer_perp
                    !
                    if (in_spheroid) self%real_values(i,j,k) = 1.0
                enddo
            enddo
        enddo
    end subroutine fourier_mask_add_hollow_spheroid

    !>  \brief  Add a hollow sphere (ring in the 2d case) centered on the origin to a Fourier filter
    !!          (equivalent to a hard-edge band-pass filter)
    subroutine fourier_mask_add_hollow_sphere(self,freq_inner,freq_outer)
        use images
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),                intent(inout)   ::  self        !<  Fourier mask to be modified
        real,                       intent(in)      ::  freq_inner  !<  Spatial frequency (1/pixel; Nyquist is 0.5) of inner radius
        real,                       intent(in)      ::  freq_outer  !<  Spatial frequency (1/pixel; Nyquist is 0.5) of outer radius
        ! Private variables
        integer ::  i,j,k
        real    ::  x,y,z
        real    ::  current_sf
        ! Start work

        ! The Fourier mask should be a real image
        if (.not. self%IsInRealSpace()) then
            write(*,'(a)') '**ERROR(fourier_mask_add_hollow_sphere): Fourier mask images should be real'
            call terminate('Fatal error in FOURIER_MASK_ADD_HOLLOW_SPHERE')
        endif

        ! Check frequencies make sense
        if (freq_outer .lt. freq_inner) then
            write(*,'(a)') '**ERROR(fourier_mask_add_hollow_sphere): outer frequency must be greater than inner frequency'
            call terminate('Fatal error in FOURIER_MASK_ADD_HOLLOW_SPHERE')
        endif

        ! Loop over the image voxels
        do K=1,self%logical_dimensions(3)
            z = (real(self%LogicalIndexGivenPhysicalIndexInFourierSpace(k,3))*self%fourier_voxel_size(3))**2
            do J=1,self%logical_dimensions(2)
                y = (real(self%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2))*self%fourier_voxel_size(2))**2
                do I=1,self%logical_dimensions(1)
                    x = (real(I-1)/real(self%logical_dimensions(1)-1)*0.5)**2
                    ! The current spatial frequency
                    current_sf = sqrt(X+Y+Z)
                    !
                    if (current_sf .ge. freq_inner .and. current_sf .le. freq_outer) self%real_values(i,j,k) = 1.0

                enddo
            enddo
        enddo
    end subroutine fourier_mask_add_hollow_sphere

    !>  \brief  Add a hollow cylinder to a Fourier mask. The cylinder axis is defined by the Euler angles.
    subroutine fourier_mask_add_hollow_cylinder(self,freq_inner,freq_outer,freq_base,freq_top,angles)
        use Images
        use runtime_parameters, only : terminate
        use euleranglesmod, only : euler
        use rotation_matrices, only : euler_to_rotmat, print_rotmat
        implicit none
        ! Arguments
        type(image),                intent(inout)   ::  self        !<  Fourier mask to be modified
        real,                       intent(in)      ::  freq_inner  !<  Spatial frequency (1/pixel; Nyquist is 0.5) of inner radius
        real,                       intent(in)      ::  freq_outer  !<  Spatial frequency (1/pixel; Nyquist is 0.5) of outer radius
        real,                       intent(in)      ::  freq_base   !<  Spatial frequency (along the y axis) for the base of the cylinder
        real,                       intent(in)      ::  freq_top    !<  Spatial frequency (along the y axis) for the top of the cylinder
        type(euler),    optional,   intent(in)      ::  angles      !<  Euler angles defining the cylinder axis
        ! Private variables
        integer ::  i,j,k
        real    ::  x,y,z
        real    ::  cyl_sf
        logical ::  in_cylinder
        type(euler) :: aangles
        real    ::  coos(4)
        real    ::  rot(4,4)
        ! Start work

        ! The Fourier mask should be a real image
        if (.not. self%IsInRealSpace()) then
            write(*,'(a)') '**ERROR(fourier_mask_add_hollow_sphere): Fourier mask images should be real'
            call terminate('Fatal error in FOURIER_MASK_ADD_HOLLOW_SPHERE')
        endif

        ! Check frequencies make sense
        if (freq_outer .lt. freq_inner) then
            write(*,'(a)') '**ERROR(fourier_mask_add_hollow_cylinder): outer frequency must be greater than inner frequency'
            call terminate('Fatal error in FOURIER_MASK_ADD_HOLLOW_SPHERE')
        endif
        if (freq_top .lt. freq_base) then
            write(*,'(a)') '**ERROR(fourier_mask_add_hollow_cylinder): top frequency must be greater than base frequency'
            call terminate('Fatal error in FOURIER_MASK_ADD_HOLLOW_CYLINDER')
        endif

        ! The axis
        aangles%phi   = 0.0
        aangles%theta = 90.0
        aangles%psi   = 90.0
        if (present(angles)) aangles = angles
        rot = euler_to_rotmat(aangles%phi,aangles%theta,aangles%psi,REVERSE=.true.)

        ! Loop over voxels
        do K=1,self%logical_dimensions(3)
            z = real(self%LogicalIndexGivenPhysicalIndexInFourierSpace(k,3))*self%fourier_voxel_size(3)
            do J=1,self%logical_dimensions(2)
                y = real(self%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2))*self%fourier_voxel_size(2)
                do I=1,self%logical_dimensions(1)
                    x = real(I-1)/real(self%logical_dimensions(1)-1)*0.5
                    !
                    coos = matmul(rot,[X,y,0.0,1.0])
                    !
                    cyl_sf = sqrt(coos(1)**2+COOS(2)**2)
                    !
                    in_cylinder =           abs(coos(3)) .ge. freq_base  &
                                    .and.   abs(coos(3)) .le. freq_top   &
                                    .and.   cyl_sf       .ge. freq_inner &
                                    .and.   cyl_sf       .le. freq_outer
                    !
                    if (in_cylinder) self%real_values(i,j,k) = 1.0
                enddo
            enddo
        enddo
    end subroutine fourier_mask_add_hollow_cylinder

    !>  \brief  Mask the first and last few slices in a volume. This is useful for example to remove the discontinuity in a helical reconstruction
    !!          where the z dimension isn't an integer number of twists.
    !!  \todo   Check conversion of radius as fraction of maximum inner radius to number of pixels
    !!  \todo   Why isn't this done in place???
    subroutine  image_mask_z_ends(image_in,mask,rad,falloff,soften)
        use images
        use UsefulFunctions, only : IsEven
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),                intent(in)      ::  image_in        !<  Input image
        type(image),                intent(out)     ::  mask            !<  Output mask
        real,                       intent(in)      ::  rad             !<  Radius (from center of volume) for the masking. Pixels, or fraction of inner radius.
        real,                       intent(in)      ::  falloff         !<  Falloff width for soft edge of mask.
        logical,    optional,       intent(in)      ::  soften          !<  Whether to "soften" (~low-pass filter) the mask. True by default.
        ! Private variables
        real    ::  rrad, ffalloff
        integer ::  i, dist, origin
        logical ::  ssoften
        ! Start work

        ! This is really only meant for 3d images
        if (image_in%logical_dimensions(3) .le. 1) then
            write(*,'(a,3(i4,1x))') '**ERROR(image_mask_z_ends): this is only meant for 3d images. Input dimensions = ', &
                                     image_in%logical_dimensions
            call terminate('Fatal error in IMAGE_MASK_Z_ENDS')
        endif

        ! Will we be softening the mask?
        if (present(soften)) then
            ssoften = soften
        else
            ssoften = .true.
        endif

        ! Allocate the mask
        call mask%Allocate(image_in)

        ! Initialise the mask
        mask%real_values = 1.0

        ! Convert the radius from fraction to pixels if necessary
        if (rad .le. 1.0) then
            rrad = rad * (mask%logical_dimensions(3)/2-1)  ! Is this exactly the right conversion from radius as a fraction of internal radius?
        else
            rrad = rad
        endif

        ! Convert the falloff from fraction to pixels if necessary
        if (falloff .le. 1.0) then
            ffalloff = falloff * (mask%logical_dimensions(3)/2-1)
        else
            ffalloff = falloff
        endif

        ! Work out origin
        if (IsEven(mask%logical_dimensions(3))) then
            origin = mask%logical_dimensions(3)/2 +1
        else
            origin = (mask%logical_dimensions(3)-1)/2
        endif

        ! Loop along the z dimension, blank appropriate slices
        do I=1,mask%logical_dimensions(3)
            dist = abs(ORIGIN-I)
            if (dist .ge. rrad) mask%real_values(:,:,i) = 0.0
        enddo

        ! Now soften the mask
        call mask_add_cosine_falloff(mask,ffalloff)
    end subroutine image_mask_z_ends

    !>  \brief  Mask the first and last few slices in a volume. This is useful for example to remove the discontinuity
    !!          in a helical reconstruction where the z dimension isn't an integer number of twists.
    subroutine image_mask_z_ends_inplace(self,width,soft,mask_value)
        use images
        use interpolation, only : hann
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),                intent(inout)   ::  self           !<  Input/output image
        integer,                    intent(in)      ::  width          !<  Number of z slices to be masked at either end of the volume
        logical,        optional,   intent(in)      ::  soft           !<  Either width slices are blanked, or a WIDTH-wide
        real,           optional,   intent(in)      ::  mask_value     !<  Value to mask with
        ! Private variables
        logical ::  ssoft
        integer ::  i
        logical,    parameter   ::  debug   =   .false.
        real    ::  mmask_value, weight
        ! Start work

        ! This is really only meant for 3d images
        if (self%logical_dimensions(3) .le. 1) then
            write(*,'(a,3(i4,1x))') '**ERROR(image_mask_z_ends_inplace): this is only meant for 3d images. Input dimensions = ',&
                                     self%logical_dimensions
            call terminate('Fatal error in IMAGE_MASK_Z_ENDS_INPLACE')
        endif

        ! Will we be softening the mask?
        if (present(soft)) then
            ssoft = soft
        else
            ssoft = .true.
        endif

        if (debug) then
            write(*,'(a,i0,1x,l1)') '**debug(image_mask_z_ends_inplace): width, soft? ', width, ssoft
        endif

        ! Work out the mask value
        if (present(mask_value)) then
            mmask_value = mask_value
        else
            mmask_value = 0.0e0
            mmask_value = self%GetAverageOfValuesOnEdges()
        endif


        ! Loop along the z dimension, blank appropriate slices
        if (ssoft) then
            ! At i=1, we expect hann = 0.0
            ! At i=width+1, we expect hann = 1.0
            do i=1,width
                weight = hann(real(width-i+1),width=width*2)
                self%real_values(:,:,i) = self%real_values(:,:,i) * weight + mmask_value * (1.0e0-weight)
            enddo
            ! At i=self%logical_dimensions(3), we expect hann = 0.0
            ! At i=self%logical_dimensions(3)-width, we expect hann = 1.0
            do i=self%logical_dimensions(3)-width,self%logical_dimensions(3)
                weight = hann(real(i-self%logical_dimensions(3)+width),width=width*2)
                self%real_values(:,:,i) = self%real_values(:,:,i) * weight + mmask_value * (1.0e0-weight)
            enddo
        else
            do i=1,width
                self%real_values(:,:,i) = mask_value
            enddo
            do i=self%logical_dimensions(3)-width+1,self%logical_dimensions(3)
                self%real_values(:,:,i) = mmask_value
            enddo
        endif

    end subroutine image_mask_z_ends_inplace



    !>  \brief  Calculate a mask for the input image, based on a low-pass filtered version of the image
    subroutine image_low_pass_mask(image_in,mask, pct_in_mask,falloff_type,low_pass_frequency,falloff_width, soften, inversion)
        use Images
        use Parsers
        use runtime_parameters
        use Histograms

        implicit none
        ! Arguments
        type(image),                            intent(in)      ::  image_in            !<  Input image
        type(image),                            intent(out)     ::  mask                !<  Output mask
        real,                                   intent(in)      ::  pct_in_mask         !<  Percentage of voxels which need to end up with 1.0 in mask
        real,                       optional,   intent(in)      ::  low_pass_frequency  !<  Low-pass frequency. In reciprocal pixels (0.5 is Nyquist). Defaults to 0.25.
        real,                       optional,   intent(in)      ::  falloff_width       !<  Mask soft falloff width
        character(len=WORD_LEN),    optional,   intent(in)      ::  falloff_type        !<  How and if to have a falloff around the edge of the mask
        logical,                    optional,   intent(in)      ::  soften              !<  Whether to soften the mask. Defaults to TRUE.
        logical,                    optional,   intent(in)      ::  inversion           !<  Whether to invert the mask, letting through background.

        ! Private variables
        real                    ::  llow_pass_frequency
        !real,       allocatable ::  histogram_boundaries(:)
        !integer,    allocatable ::  histogram(:)
        type(Histogram)         ::  my_histogram
        real                    ::  threshold
        integer                 ::  i, j, k!, num_in_mask_target,num_in_mask_actual
        logical                 ::  ssoften
        logical                 ::  eextend
        logical                 ::  iinversion
        logical,    parameter   ::  debug   =   .false.
        ! Start work

        ! Assumption that incoming image is real
        if (.not. image_in%is_in_real_space) then
            write(*,'(a)') '**ERROR(image_low_pass_mask): expected real image'
            call terminate('Fatal error in IMAGE_LOW_PASS_MASK')
        endif

        ! Low-pass filter radius
        if (present(low_pass_frequency)) then
            llow_pass_frequency = low_pass_frequency
        else
            llow_pass_frequency = 0.25
        endif

        ! Soften the mask?
        if (present(soften)) then
            ssoften = soften
        else
            ssoften = .true.
        endif

        ! Extend / soften the mask?
        eextend = .false.
        if (present(falloff_type)) then
            select case (falloff_type)
                case('COS_EDGE')
                    SSOFTEN=.true.
                    EEXTEND=.false.
                case('HARD_EXTEND')
                    SSOFTEN=.false.
                    EEXTEND=.true.
                case('NONE')
                    SSOFTEN=.false.
                    EEXTEND=.false.
                case default
                write(*,'(2a)') '**ERROR(mode_mask): unrecognised falloff type: ', falloff_type
                call terminate('Fatal error in MODE_MASK')
            end select
        endif

        !Invert mask?
        if (present(inversion)) then
            iinversion = inversion
        else
            iinversion = .false.
        endif

        ! Copy incoming image
        call mask%Allocate(image_in)
        mask%real_values = image_in%real_values

        !
        if (debug) call mask%WriteToDisk('dbg_before_lp.mrc')

        ! Low pass filter
        if (debug) write(*,'(a,f0.3)') '**debug(image_low_pass_mask): llow_pass_freq = ', llow_pass_frequency
        call mask%ApplyBandPassFilter(0.0,llow_pass_frequency,width_of_cosine_falloff=0.025)

        ! Debug dump
        if (debug) call mask%WriteToDisk('dbg_lp.mrc')

        ! Get histogram for the image
        !call mask%ComputeHistogram(histogram,histogram_boundaries,NUM_OF_BINS=100, &
        !                           NEGS_IN_FIRST_BIN=.false.,BIN_CONDENSE_OUTLIERS=.false.)
        call mask%ComputeHistogramOfValues(my_histogram,number_of_bins=100)

        ! Target number of voxels in mask
        !num_in_mask_target = nint(real(product(mask%logical_dimensions))*pct_in_mask/100.)
        !if (debug) print *, '**debug(image_low_pass_mask): num_in_mask_target = ', num_in_mask_target

        ! Work out a threshold at which pct_in_mask % of voxels are included
        !threshold = 0.0
        !num_in_mask_actual = 0
        !do I=SIZE(histogram),2,-1   ! we skip the first bin, cos it only has negative values in it.
        !    !if (debug) print '(3(A,i10))', '**debug(image_low_pass_mask): on bin ', i, ' of ', size(histogram), ': num_in_mask_actual = ', num_in_mask_actual
        !
        !    if (num_in_mask_target .ge. num_in_mask_actual .and. num_in_mask_target .lt. num_in_mask_actual + histogram(i)) then
        !        ! The appropriate threshold is within the current bin - use linear interpolation to get a value
        !        threshold = histogram_boundaries(i) + (histogram_boundaries(I+1)-HISTOGRAM_BOUNDARIES(i))   &
        !                                                *(NUM_IN_MASK_TARGET-NUM_IN_MASK_ACTUAL)/histogram(i)
        !        !if (debug) print '(a,ES10.3,a,ES10.3)', '**debug(image_low_pass_mask): threshold should be between ', histogram_boundaries(i), ' and ', histogram_boundaries(I+1)
        !    else
        !        ! The appropriate threshold is not within the current bin, move on to the next bin
        !
        !    endif
        !    num_in_mask_actual = num_in_mask_actual + histogram(i)
        !enddo
        threshold = max(0.0,my_histogram%QuantileFunction(1.0-(pct_in_mask*0.01e0)))

        ! If the threshold hasn't be set, something went wrong
        !if (threshold .eq. 0.0) then
        !    write(*,'(a)') '**ERROR(image_low_pass_mask): threshold was not found'
        !    call terminate('Fatal error in IMAGE_LOW_PASS_MASK')
        !else
            if (debug) write(*,'(a,ES10.3)') '**debug(image_low_pass_mask): threshold set to ', threshold
        !endif

        ! Binarise the filtered map
        do K=1,mask%logical_dimensions(3)
            do J=1,mask%logical_dimensions(2)
                do I=1,mask%logical_dimensions(1)
                    if (mask%real_values(i,j,k) .lt. threshold) then
                        mask%real_values(i,j,k) = 0.0
                    else
                        mask%real_values(i,j,k) = 1.0
                    endif
                enddo
            enddo
        enddo

        ! Debug: dump the binary image
        if (debug) then
            print *, '**debug(image_low_pass_mask): dumping dbg_mask_out.mrc'
            print *, '**debug(image_low_pass_mask): fraction inside mask: ', &
                    sum(mask%real_values(1:mask%logical_dimensions(1),:,:))/product(mask%logical_dimensions)
            call mask%WriteToDisk('dbg_lp_mask_out.mrc')
        endif

        ! Soften the mask
        if (ssoften) then
            call mask_add_cosine_falloff(mask,5.0)
            if (debug) call mask%WriteToDisk('dbg_mask_out_soft.mrc')
        endif

        if (eextend) then
            write (*,'(a,F0.3)') '**debug(image_low_pass_mask): extending mask with the following falloff width: ', falloff_width
            call extend_mask(mask, falloff_width)
        endif

        if (iinversion) then
            call invert_mask(mask)
            write (*,*) 'inverting mask'
        endif

    end subroutine image_low_pass_mask

    !>  \brief  Compute the fraction of voxels within a mask image (binary or smooth)
    real function fraction_of_voxels_within_mask(mask) result(frac)
        use images
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),    intent(in)  ::  mask
        ! Private variables
        real(kind=8)    ::  f_mask
        integer         ::  i,j,k
        ! Start work

        ! Image is expected to be real
        if (.not. mask%is_in_real_space) then
            write(*,'(a)') '**ERROR(fraction_of_voxels_within_mask): image is not real'
            call terminate('Fatal error in FRACTION_OF_VOXELS_WITHIN_MASK')
        endif

        f_mask = 0.0D0
        ! Image should only have values between 0 and 1
        ! Loop over voxels
        do K=1,mask%logical_dimensions(3)
            do J=1,mask%logical_dimensions(2)
                do I=1,mask%logical_dimensions(1)
                    if (mask%real_values(i,j,k) .gt. 1.0 .or. mask%real_values(i,j,k) .lt. 0.0) then
                        write(*,'(a,G15.8)') '**ERROR(fraction_of_voxels_within_mask): unexpected value in mask volume: ', &
                                            mask%real_values(i,j,k)
                        call terminate('Fatal error FRACTION_OF_VOXELS_WITHIN_MASK')
                    endif
                    f_mask = f_mask + mask%real_values(i,j,k)**2
                enddo
            enddo
        enddo
        frac = f_mask / product(mask%logical_dimensions)
    end function fraction_of_voxels_within_mask

    !>  \brief  Add a cosine-edged falloff to a mask. The mask should consists of 1's and 0's only. The 1's will be left untouched.
    !!  \todo   Use stricter definition of max inner radius etc.
    pure subroutine mask_add_cosine_falloff(mask,falloff_width)
        use images
        use interpolation, only  :   hann
        implicit none
        ! Arguments
        type(image),    intent(inout)   ::  mask            !<  Input mask
        real,           intent(in)      ::  falloff_width   !<  Width of falloff. Pixels.
        ! Private variables
        integer ::  i,j,k
        integer ::  ii,jj,kk
        integer ::  imin,imax,jmin,jmax,kmin,kmax
        integer ::  window_radius
        integer ::  ifalloff_width
        real    ::  val, rad, rad_sq, rad_k, rad_j, rad_i
        logical ::  near_edge
        real    ::  ffalloff_width, ffalloff_width_sq
        ! Start work


        ! If the falloff parameter is <1.0, assume it's a fraction of inner radius
        if (falloff_width .lt. 1.0) then
            ffalloff_width = falloff_width * mask%logical_dimensions(1) / 2.0  ! This is done the lazy way, assuming square etc. Need to call proper functions to find out max inner radius etc.
        else
            ffalloff_width = falloff_width
        endif

        ffalloff_width_sq = ffalloff_width**2

        ! integer value used as parameter to the hann function
        ifalloff_width = nint(2.*ffalloff_width)

        ! work out the size of the "window" to check around each mask voxel
        window_radius = ceiling(abs(ffalloff_width))


        ! loop over the volume
        do k=1,mask%logical_dimensions(3)
            do j=1,mask%logical_dimensions(2)
                do i=1,mask%logical_dimensions(1)
                    ! If the current voxel is "in the mask" (i.e. has value 1.0), proceed
                    if (mask%real_values(i,j,k) .eq. 1.0) then
                        ! Find out whether the current voxel has an immediate neighbour which is 0.0
                        kmin = k-1
                        if (kmin .lt. 1) kmin = 1
                        jmin = j-1
                        if (jmin .lt. 1) jmin = 1
                        imin = i-1
                        if (imin .lt. 1) imin = 1
                        kmax = k+1
                        if (kmax .gt. mask%logical_dimensions(3)) kmax = mask%logical_dimensions(3)
                        jmax = j+1
                        if (jmax .gt. mask%logical_dimensions(2)) jmax = mask%logical_dimensions(2)
                        imax = i+1
                        if (imax .gt. mask%logical_dimensions(1)) imax = mask%logical_dimensions(1)
                        near_edge = .false.
                        edge_check: do KK=KMIN,kmax
                            do JJ=JMIN,jmax
                                do II=IMIN,imax
                                    if (mask%real_values(ii,jj,kk) .lt. 1.0) then
                                        near_edge = .true.
                                        exit edge_check
                                    endif
                                enddo
                            enddo
                        enddo edge_check
                        if (near_edge) then
                            ! Loop over the neighbours
                            kmin = k-window_radius
                            if (kmin .lt. 1) kmin = 1
                            jmin = j-window_radius
                            if (jmin .lt. 1) jmin = 1
                            imin = i-window_radius
                            if (imin .lt. 1) imin = 1
                            kmax = k+window_radius
                            if (kmax .gt. mask%logical_dimensions(3)) kmax = mask%logical_dimensions(3)
                            jmax = j+window_radius
                            if (jmax .gt. mask%logical_dimensions(2)) jmax = mask%logical_dimensions(2)
                            imax = i+window_radius
                            if (imax .gt. mask%logical_dimensions(1)) imax = mask%logical_dimensions(1)
                            do kk=kmin,kmax
                                rad_k = real((kk-k)**2)
                                do jj=jmin,jmax
                                    rad_j = real((jj-j)**2)
                                    do ii=imin,imax
                                        ! if the current neighbour has a value less than 1.0, we'll investigate further
                                        if (mask%real_values(ii,jj,kk) .lt. 1.0) then
                                            rad_i = real((ii-i)**2)
                                            ! how far the current neighbour is from the mask voxel
                                            !rad = sqrt(real((kk-k)**2+(jj-j)**2+(ii-i)**2))
                                            rad_sq = rad_i + rad_j + rad_k
                                            if (rad_sq .le. ffalloff_width_sq) then
                                                rad = sqrt(rad_sq)
                                                val = hann(rad,ifalloff_width)
                                                if (mask%real_values(ii,jj,kk) .lt. val) then
                                                    mask%real_values(ii,jj,kk) = val
                                                endif
                                            endif
                                        endif
                                    enddo
                                enddo
                            enddo
                        endif ! End of test of whether we are near the edge of the binary mask
                    endif ! End of test of whether current voxel is in mask (values of 1.0)
                enddo
            enddo
        enddo


    end subroutine mask_add_cosine_falloff

    !>  \brief  Extend binary mask
    subroutine extend_mask(mask, falloff_width)
        use images
        implicit none
        ! Arguments
        type(image),    intent(inout)   ::  mask            !<  Input mask
        real,           intent(in)      ::  falloff_width   !<  Width of extension. Pixels.
        ! Private variables
        integer ::  i,j,k
        integer ::  ii,jj,kk
        integer ::  imin,imax,jmin,jmax,kmin,kmax
        integer ::  window_radius
        real    ::  val, rad_sq, rad_k, rad_j, rad_i
        logical ::  near_edge
        real    ::  ffalloff_width, ffalloff_width_sq
        logical,    parameter   ::  debug   =   .false.



        if (debug) write(*,'(a)') '**debug(extend_mask): entering'

        ! If the falloff parameter is <1.0, assume it's a fraction of inner radius
        if (falloff_width .lt. 1.0) then
            ffalloff_width = falloff_width * mask%logical_dimensions(1) / 2.0  ! This is done the lazy way, assuming square etc. Need to call proper functions to find out max inner radius etc.
        else
            ffalloff_width = falloff_width
        endif

        ffalloff_width_sq = FFALLOFF_WIDTH**2

        ! Work out the size of the "window" to check around each mask voxel
        window_radius = ceiling(abs(ffalloff_width))

        ! Loop over the volume
        do K=1,mask%logical_dimensions(3)
            do J=1,mask%logical_dimensions(2)
                do I=1,mask%logical_dimensions(1)
                    ! If the current voxel is "in the mask" (i.e. has value 1.0), proceed
                    if (mask%real_values(i,j,k) .eq. 1.0) then
                        ! Find out whether the current voxel has an immediate neighbour which is 0.0
                        kmin = K-1
                        if (kmin .lt. 1) KMIN = 1
                        jmin = J-1
                        if (jmin .lt. 1) JMIN = 1
                        imin = I-1
                        if (imin .lt. 1) IMIN = 1
                        kmax = K+1
                        if (kmax .gt. mask%logical_dimensions(3)) KMAX = mask%logical_dimensions(3)
                        jmax = J+1
                        if (jmax .gt. mask%logical_dimensions(2)) JMAX = mask%logical_dimensions(2)
                        imax = I+1
                        if (imax .gt. mask%logical_dimensions(1)) IMAX = mask%logical_dimensions(1)
                        near_edge = .false.
                        edge_check: do KK=KMIN,kmax
                            do JJ=JMIN,jmax
                                do II=IMIN,imax
                                    if (mask%real_values(ii,jj,kk) .lt. 1.0) then
                                        near_edge = .true.
                                        exit edge_check
                                    endif
                                enddo
                            enddo
                        enddo edge_check
                        if (near_edge) then
                            ! Loop over the neighbours
                            kmin = K-WINDOW_RADIUS
                            if (kmin .lt. 1) KMIN = 1
                            jmin = J-WINDOW_RADIUS
                            if (jmin .lt. 1) JMIN = 1
                            imin = I-WINDOW_RADIUS
                            if (imin .lt. 1) IMIN = 1
                            kmax = K+WINDOW_RADIUS
                            if (kmax .gt. mask%logical_dimensions(3)) KMAX = mask%logical_dimensions(3)
                            jmax = J+WINDOW_RADIUS
                            if (jmax .gt. mask%logical_dimensions(2)) JMAX = mask%logical_dimensions(2)
                            imax = I+WINDOW_RADIUS
                            if (imax .gt. mask%logical_dimensions(1)) IMAX = mask%logical_dimensions(1)
                            do KK=KMIN,kmax
                                rad_k = real((KK-K)**2)
                                do JJ=JMIN,jmax
                                    rad_j = real((JJ-J)**2)
                                    do II=IMIN,imax
                                        ! If the current neighbour has a value less than 1.0, we'll investigate further
                                        if (mask%real_values(ii,jj,kk) .lt. 1.0) then
                                            rad_i = real((II-I)**2)
                                            ! How far the current neighbour is from the mask voxel
                                            !rad = sqrt(real((KK-K)**2+(JJ-J)**2+(II-I)**2))
                                            rad_sq = rad_i + rad_j + rad_k
                                            if (rad_sq .le. ffalloff_width_sq) then
                                                ! Distinguish new mask voxels from old mask voxels by giving them value of 2.0
                                                val = 2.0
                                                if (mask%real_values(ii,jj,kk) .lt. 1.0) then
                                                    mask%real_values(ii,jj,kk) = val
                                                endif
                                            endif
                                        endif
                                    enddo
                                enddo
                            enddo
                        endif
                    endif
                enddo
            enddo
        enddo

        ! New mask voxels are now 1.0
        do K=1,mask%logical_dimensions(3)
            do J=1,mask%logical_dimensions(2)
                do I=1,mask%logical_dimensions(1)
                    if (mask%real_values(i,j,k) .eq. 2.0) then
                        mask%real_values(i,j,k)=1.0
                    endif
                enddo
            enddo
        enddo


        if (debug) write(*,'(a)') '**debug(extend_mask): all done'
    end subroutine extend_mask


    !>  \brief Invert a mask
    subroutine invert_mask(mask)
        use images
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),    intent(inout)   ::  mask            !<  Input mask
        integer ::  i,j,k

        ! Check that the mask is binary, if not, crash out
        if (.not. all(  mask%real_values(1:MASK%logical_dimensions(1),1:mask%logical_dimensions(2),1:mask%logical_dimensions(3)) &
                                                              .eq. 1.0  .or. &
                        mask%real_values(1:MASK%logical_dimensions(1),1:mask%logical_dimensions(2),1:mask%logical_dimensions(3)) &
                                                              .eq. 0.0 ) ) then
            write(*,'(a)') '**ERROR(invert_mask): this subroutine cannot deal with soft masks'
            call terminate('Feature not implemented yet')
        endif

        !
        do K=1,mask%logical_dimensions(3)
            do J=1,mask%logical_dimensions(2)
                do I=1,mask%logical_dimensions(1)
                    if (mask%real_values(i,j,k) .lt. 0.5) then
                        mask%real_values(i,j,k)=1.0
                    else
                        mask%real_values(i,j,k)=0.0
                    endif
                enddo
            enddo
        enddo
    end subroutine invert_mask

end module masks
