!>  \brief  Methods to minimise functions
module minimisers
    use, intrinsic :: iso_c_binding
    use :: fgsl
    use :: differential_evolution_mod
    implicit none
    ! By default, stuff is hidden from the outside world
    private
    ! These interfaces are public
    public :: ds_run, ds_init, ds_free, ds_set_start_step
    public :: fgsl_minimisable_function, ds_print_progress_summary
    public :: bf_init, bf_run

    ! The various minimisers we may want to support
    integer,    public, parameter   ::  downhill_simplex        =   1
    integer,    public, parameter   ::  conjugate_gradient      =   2
    integer,    public, parameter   ::  differential_evolution  =   3
    integer,    public, parameter   ::  brute_force             =   4


    !
    ! Abstract interface for function used by the fgsl ds minimiser
    !

    !>  \brief  What a minimisable function should look like
    abstract interface
        function fgsl_minimisable_function(values,params) bind(c)
            import :: c_double, c_ptr   ! Because we're in an abstract interface, we normally don't have access to all the stuff from the module, so we need to import
            type(c_ptr), value  ::  values
            type(c_ptr), value  ::  params
            real(kind=C_DOUBLE) ::  fgsl_minimisable_function
        end function fgsl_minimisable_function
    end interface

    !
    ! Parameters for the downhill simplex
    !
    real,       parameter,  public  ::  search_accuracy_to_ds_size_ratio    =   1.0
    integer,    parameter,  public  ::  ds_max_iterations                   =   1000

    !
    ! Abstract interfaces for functions used by the fgsl sa minimiser
    !

    !>  \brief  What an energy function should look like (used by fgsl simulated annealing for scoring function, for example)
    abstract interface
        function fgsl_energy_function(params) bind(c)
            import              ::  c_double, c_ptr
            type(c_ptr), value  ::  params
            real(kind=C_DOUBLE) ::  fgsl_energy_function
        end function fgsl_energy_function
    end interface

    !>  \brief  What a stepping function should look like (used by fgsl simulated annealing for modifying the current configuration)
    abstract interface
        subroutine fgsl_stepping_function(rng,conf,max_step) bind(c)
            import  ::  c_ptr, fgsl_double
            type(c_ptr),        value   ::  rng         !<  Pointer to random number generator
            type(c_ptr),        value   ::  conf        !<  Pointer to configuration to alter
            real(fgsl_double),  value   ::  max_step    !<  Maximum step size
        end subroutine fgsl_stepping_function
    end interface

    !>  \brief  What a metric function should look like (used by fgsl sa to measure the distance between two configurations)
    abstract interface
        function fgsl_metric_function(conf1,conf2) bind(c)
            import              ::  fgsl_double, c_ptr
            type(c_ptr), value  ::  conf1
            type(c_ptr), value  ::  conf2
            real(fgsl_double)   ::  fgsl_metric_function
        end function fgsl_metric_function
    end interface

    !>  \brief  What a printing function should look like (used by fgsl sa to printout information about the current configuration)
    abstract interface
        subroutine fgsl_print_function(conf) bind(c)
            import  ::  c_ptr
            type(c_ptr),    value   ::  conf
        end subroutine fgsl_print_function
    end interface

    !>  \brief  What a configuration constructor should look like
    !abstract interface
    !   subroutine fgsl_conf_constructor(conf) bind(c)
    !       import  ::  c_ptr
    !       type(c_ptr),    value   ::  conf
    !   end subroutine
    !end interface

    !>  \brief  What a configuration constructor should look like
    !abstract interface
    !   subroutine fgsl_conf_destructor(conf) bind(c)
    !       import  ::  c_ptr
    !       type(c_ptr),    value   ::  conf
    !   end subroutine
    !end interface

    !>  \brief  What a configuration copy function should look like
    !abstract interface
    !   subroutine  fgsl_conf_copier(conf_source,conf_dest) bind(c)
    !       import  ::  c_ptr
    !       type(c_ptr),    value   ::  conf_source
    !       type(c_ptr),    value   ::  conf_dest
    !   end subroutine
    !end interface

    !
    ! Minimiser types
    !

    !>  \brief  The basic minimiser type
    type, abstract ::   minimiser
        integer ::  n   !<  Number of dimensions
        logical ::  in_memory = .false. !<  Whether allocated memory
    end type

    !>  \brief  Brute-force (exhaustive search) minimiser
    type, public, extends(minimiser)    ::  minimiser_bf
        procedure(fgsl_minimisable_function), pointer, nopass   ::  target_function => null()   !<  a pointer to the function the minimiser aims to minimise.
                                                                                                !!  no pass because there an assumption by default that type-bound procedures have an object of the type to which they are bound as first argument.
                                                                                                !!  this default assumption is overriden by no pass.
        type(c_ptr),    allocatable                             ::  parameters(:)               !<  pointer to object (usually fmc) with parameters needed by the function.
                                                                                                !!  When multithreading, each thread has a pointer to an area in memory. For FMCs, make those distinct/separate.
        real(kind=8),   allocatable                             ::  starting_value(:)           !<  for each dimension, the starting value
        real(kind=8),   allocatable                             ::  best_value(:)               !<  for each dimension, the starting value
        real(kind=8),   allocatable                             ::  half_range(:)               !<  for each dimension, the half-range to search over
        real(kind=8),   allocatable                             ::  step_size(:)                !<  for each dimension, the stepping size during the search
        logical,        allocatable                             ::  dimension_at_max(:)         !<  for each dimension, whether the search has already reached the maximum
        real                                                    ::  best_score
        integer                                                 ::  num_iterations              !<  the total number of iterations needed to complete the exhaustive search
    end type

    !
    ! Downhill simplex types
    !

    !>  \brief  Downhill simplex multi-dimensional minimiser
    type, public, extends(minimiser) :: minimiser_ds
        procedure(fgsl_minimisable_function), pointer, nopass   ::  target_function => null()  !<  a pointer to the function the minimiser aims to minimise.
                                                                                        !!  no pass because there an assumption by default that type-bound procedures have an object of the type to which they are bound as first argument.
                                                                                        !!  This default assumption is overriden by no PASS.
        real,   allocatable    ::   best_score(:)                          !<  For each iteration, a record of the best score
        real,   allocatable    ::   simplex_size(:)                        !<  For each iteration, a record of the simplex size
        integer                 ::  max_num_iter  =     ds_max_iterations  !<  Maximum number of iteration
        logical                 ::  double_runs   =     .true.             !<  Start a second minimisation run from the end of the first
    end type


    !>  \brief  ds minimiser which relies on the fgsl
    type, public, extends(minimiser_ds) :: minimiser_ds_fgsl
        integer(fgsl_size_t)                                ::  n_fgsl
        type(fgsl_multimin_fminimizer)                      ::  minimiser_fgsl
        character(len=FGSL_STRMAX,kind=FGSL_CHAR)           ::  name
        type(fgsl_multimin_function)                        ::  function_fgsl
        real(fgsl_double),                       pointer    ::  current_best_fgsl(:)        !<  The search parameter values that gave the best score
        real(fgsl_double)                                   ::  current_minimum_fgsl        !<  The value of the best score
    end type minimiser_ds_fgsl

    !
    ! Simulated annealing types
    !

    !>  \brief  Simulated annealing minimiser
    type, public, extends(minimiser)    ::  minimiser_sa
        type(fgsl_rng)      ::  rng             !<  Random number generator (provided by the fgsl)
        integer             ::  iters_fixed_t   !<  How many iterations for each temperature?
        real                ::  kb              !<  Boltzmann constant
        real                ::  t_initial       !<  Initial temperature
        real                ::  t_damping       !<  Damping factor
        real                ::  t_min           !<  Minimum temperature
        real                ::  current_energy  !<  Current energy
        real                ::  t_current       !<  Current temperature
    end type minimiser_sa

    !>  \brief  sa (simulated annealing) minimiser which relies on the fgsl
    type, public, extends(minimiser_sa) ::  minimiser_sa_fgsl
        integer(fgsl_size_t)        ::  n_fgsl                  !<  Number of dimensions (as an fgsl integer)
        type(fgsl_siman_params_t)   ::  sa_params               !<  Data structure that holds all parameters that control the sa run
        integer(fgsl_int)           ::  num_tries_fgsl          !<  How many points do we try before stepping
        integer(fgsl_int)           ::  iters_fixed_t_fgsl      !<  How many iterations for each temperature?
        real(fgsl_double)           ::  max_step_size_fgsl      !<  Maximum step size in random walk
        real(fgsl_double)           ::  kb_fgsl                 !<  Boltzmann constant
        real(fgsl_double)           ::  t_initial_fgsl          !<  Initial temperature
        real(fgsl_double)           ::  t_damping_fgsl          !<  Damping factor
        real(fgsl_double)           ::  t_min_fgsl              !<  Minimum temperature
        type(c_ptr)                 ::  current_conf            !<  Current configuration (pointer to object with all information needed to return an energy/score) (at initialisation, starting position; upon solvation, search result)
        procedure(fgsl_energy_function),    pointer,    nopass  ::  energy_function     !<  Pointer to energy function
        procedure(fgsl_stepping_function),  pointer,    nopass  ::  stepping_function   !<  Pointer to stepping function
        procedure(fgsl_metric_function),    pointer,    nopass  ::  metric_function     !<  Pointer to metric function
        procedure(fgsl_print_function),     pointer,    nopass  ::  print_function      !<  Pointer to a print function
        !procedure(fgsl_conf_constructor),  pointer,    nopass  ::  conf_constructor    !<  Function to allocate/construct a configuration (state) object
        !procedure(fgsl_conf_destructor),   pointer,    nopass  ::  conf_destructor     !<  Function to free/deallocate/destroy a configuration object
        !procedure(fgsl_conf_copier),       pointer,    nopass  ::  conf_copier         !<  Function to copy one configuration (state) to another
        integer                     ::  size_of_conf            !<  Size (in bytes) of the configuration object current_conf
    end type minimiser_sa_fgsl

    !
    ! Interfaces
    !

    interface ds_run
        module procedure fgsl_ds_run
    end interface

    interface  ds_init
        module procedure fgsl_ds_init
    end interface

    interface   ds_set_start_step
        module procedure fgsl_ds_set_start_step
    end interface

    interface ds_update_current_best
        module procedure fgsl_ds_update_current_best
    end interface

    interface ds_free
        module procedure fgsl_ds_free
    end interface

    interface sa_init
        module procedure minimiser_sa_fgsl_init
    end interface

    interface sa_solve
        module procedure minimiser_sa_fgsl_solve
    end interface

    contains


    !>  \brief  Solve the simulated annealing problem
    subroutine minimiser_sa_fgsl_solve(self)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(minimiser_sa_fgsl),    intent(inout)   ::  self
        ! Private variables
        integer(fgsl_size_t)    ::  size_of_conf_fgsl
        ! Start work
        size_of_conf_fgsl = self%size_of_conf
        !
        call terminate('Not implemented')
        ! The portland compiler (v 13.3) does not like this call below
        !call fgsl_siman_solve(  self%rng,self%current_conf, &
        !                        self%energy_function,self%stepping_function,self%metric_function,self%print_function,   &
        !                        ELEMENT_SIZE=SIZE_OF_CONF_FGSL,PARAMS=SELF%sa_params)
    end subroutine minimiser_sa_fgsl_solve

    !>  \brief  Initialise the sa minimiser
    subroutine minimiser_sa_fgsl_init(  self,n,starting_conf, size_of_conf, &
                                        energy_function, stepping_function, metric_function, &
                                        print_function, &
                                        !conf_constructor, conf_destructor, conf_copier,    &
                                        num_tries,iters_fixed_t,max_step_size,  &
                                        kb,t_initial,t_damping,t_min)
        use, intrinsic  ::  iso_c_binding
        implicit none
        ! Arguments
        type(minimiser_sa_fgsl),    intent(inout)   ::  self
        integer,                    intent(in)      ::  n                   !<  Number of dimensions
        type(c_ptr),                intent(in)      ::  starting_conf       !<  Starting configuration (pointer to an object that contains all information needed to score)
        integer,                    intent(in)      ::  size_of_conf        !<  Size (in bytes) of the configuration object (this can easily be worked out by using the sizeof() inquuiry intrinsic function
        procedure(fgsl_energy_function)             ::  energy_function     !<  Energy/scoring function
        procedure(fgsl_stepping_function)           ::  stepping_function   !<  Function to alter the configuration by performing a step
        procedure(fgsl_metric_function)             ::  metric_function     !<  Function to measure distance between two configurations / states
        procedure(fgsl_print_function)              ::  print_function      !<  Function to print out information about the current configuration / state
        !procedure(fgsl_conf_constructor)           ::  conf_constructor    !<  Function to allocate/construct a configuration (state) object
        !procedure(fgsl_conf_destructor)                ::  conf_destructor     !<  Function to free/deallocate/destroy a configuration object
        !procedure(fgsl_conf_copier)                    ::  conf_copier         !<  Function to copy one configuration (state) to another
        integer,                    intent(in)      ::  num_tries           !<  How many points do we try before stepping
        integer,                    intent(in)      ::  iters_fixed_t       !<  How many iterations for each temperature?
        real,                       intent(in)      ::  max_step_size       !<  Maximum step size in random walk
        real,                       intent(in)      ::  kb                  !<  Boltzmann constant
        real,                       intent(in)      ::  t_initial           !<  Initial temperature
        real,                       intent(in)      ::  t_damping           !<  Damping factor for temperature
        real,                       intent(in)      ::  t_min               !<  Minimum temperature
        ! Private variables
        type(fgsl_rng_type) ::  t

        ! Start work
        ! The dimensionality
        self%n      =   n
        self%n_fgsl =   self%n
        ! Setup the random number generator
        t           =   fgsl_rng_env_setup()
        self%rng    =   fgsl_rng_alloc(t)
        ! Setup sa parameters
        self%num_tries_fgsl     =   num_tries
        self%iters_fixed_t_fgsl =   iters_fixed_t
        self%max_step_size_fgsl =   max_step_size
        self%kb_fgsl            =   kb
        self%t_initial_fgsl     =   t_initial
        self%t_damping_fgsl     =   t_damping
        self%t_min_fgsl         =   t_min
        self%current_conf       =   starting_conf
        self%size_of_conf       =   size_of_conf
        self%energy_function    =>  energy_function
        self%stepping_function  =>  stepping_function
        self%metric_function    =>  metric_function
        self%print_function     =>  print_function
        !self%conf_constructor  =>  conf_constructor
        !self%conf_destructor   =>  conf_destructor
        !self%conf_copier       =>  conf_copier

        ! Do the initialisation
        call fgsl_siman_params_init(    self%sa_params, &
                                        self%num_tries_fgsl,self%iters_fixed_t_fgsl,self%max_step_size_fgsl,    &
                                        self%kb_fgsl, self%t_initial_fgsl,self%t_damping_fgsl,self%t_min_fgsl)

    end subroutine minimiser_sa_fgsl_init


    !>  \brief  Run the minimisation
    subroutine fgsl_ds_run(self,target_size)
        use ui, only: interval_print
        use runtime_parameters, only : terminate
        use DatesAndTimes
        use Globals
        use omp_lib
        implicit none
        ! Arguments
        type(minimiser_ds_fgsl),    intent(inout)   ::  self                !<  Minimiser object
        real,                       intent(in)      ::  target_size         !<  Target size that the
        ! Private variables
        integer(fgsl_int)   ::  status
        integer             ::  i
        real(fgsl_double)   ::  current_size
        real(fgsl_double)   ::  target_size_fgsl
        real                ::  current_best_score = 0.0
        logical,    parameter   ::  debug   =   .false.
        type(DateAndTime)   ::  myTime
        ! Start work
        if (debug) write(*,'(a)') '**debug(fgsl_ds_run): entered'
        ! Check the minimiser's been initialised
        if (.not. self%in_memory) then
            write(*,'(a)') '**ERROR(fgsl_ds_run): minimiser not initialised'
            call terminate('Fatal error in FGSL_DS_RUN')
        endif
        ! Target size of the Simplex. When the simplex has shrunk to this size, the iterations stop.
        target_size_fgsl = target_size
        ! Work out how often we'll print info
        !output_interval = nint(real(max_num_iterations)/100.)
        !if (output_interval .LT. 1) OUTPUT_INTERVAL = 1

        ! Start iterations
        i = 0
        do
            i = i + 1

            ! Do an iteration
            status = fgsl_multimin_fminimizer_iterate(self%minimiser_fgsl)
            if (status .ne. fgsl_success) then
                write(*,'(a,i4)') '**ERROR(fgsl_ds_run): failure at iteration ', i
                call terminate('Fatal error in FGSL_DS_RUN')
            endif
            !print '(a,i4,F12.6)', '**debug(fgsl_ds_run): iteration, best score: ', i, fgsl_multimin_fminimizer_minimum(self%minimiser_fgsl)

            ! Check the size of the simplex
            current_size = fgsl_multimin_fminimizer_size(self%minimiser_fgsl)

            ! Check the current best score
            current_best_score = fgsl_multimin_fminimizer_minimum(self%minimiser_fgsl)

            ! Diagnostics
            !if (mod(I-1,output_interval) .EQ. 0) THEN
            !if (interval_print(i,max_num_iterations)) then
            if (this_program%IsVerbose() .or. debug) then
#ifdef _OPENMP
                if (omp_get_active_level() .eq. 0 .and. (this_program%IsVerbose() .or. debug) ) then
#endif
                    !$omp   critical (fgsl_ds_run_omp_crit)
                    print '(2a,2(a,i5),a,E13.6,a,ES10.3)',  '**info(fgsl_ds_run): ', myTime%DateAndTimeAsString(), &
                                                            ': on iteration ', i , &
                                                            ' of a maximum of ', self%max_num_iter, '. Current best score = ', &
                                                            current_best_score, '. Current simplex size = ', current_size
                    !$omp   end critical (fgsl_ds_run_omp_crit)
#ifdef _OPENMP
                endif
#endif
            endif

            ! Keep track of best score and simplex size by growing the relevant arrays
            call grow_allocatable_array_real(self%best_score)
            call grow_allocatable_array_real(self%simplex_size)
            self%best_score(size(self%best_score))    =  -CURRENT_BEST_SCORE
            self%simplex_size(size(self%simplex_size))  =   current_size
            if (i .eq. 1) SELF%simplex_size(1) = current_size

            ! The best score should never decrease
            if (size(self%best_score) .gt. 1) then
                if (self%best_score(size(self%best_score)) - self%best_score(size(self%best_score)-1) .lt. -.01) then
                    write(*,'(a,F0.5,a,F0.5)') '**ERROR(fgsl_ds_run): best score went from ', &
                                                self%best_score(size(self%best_score)-1), ' to ', &
                                                self%best_score(size(self%best_score))
                    if (i .eq. 1) then
                        ! This is an ugly work-around for a situation encountered 2012-aug-23 whereby
                        ! the filament_get_path routine returns 1 more waypoint when called by the minimiser
                        ! than when called by fmc_eval when getting the starting_score. It looks like this is due
                        ! to changes of < 0.001 pixels in the path length calculation, but I've not been able to find
                        ! a proper fix
                        if (this_program%IsVerbose() .or. debug) then
                            write(*,'(a)') '**info(fgsl_ds_run): ignoring decreasing-score error'//&
                                            ' because we are in the first iteration'
                        endif
                    else
                        write(*,'(a,i0)') '**ERROR(fgsl_ds_run): terminating on iteration ', i
                        call ds_print_progress_summary(self)
                        call this_program%TerminateWithFatalError('fgsl_ds_run','Fatal error in FGSL_DS_RUN')
                    endif
                endif
            endif

            ! Check the status of the simplex
            status = fgsl_multimin_test_size(current_size,target_size_fgsl)
            if (status .eq. fgsl_success .or. i .eq. self%max_num_iter) then
                exit
            endif
        enddo
        ! Diagnostics
        if (this_program%IsVerbose() .or. debug) then
            if (i .ge. self%max_num_iter) then
                print '(2a,2(a,i0),a,F0.6)', '**info(fgsl_ds_run): ', myTime%DateAndTimeAsString(), ': reached iteration ', i , &
                                            ' of a maximum of ', self%max_num_iter, '. Best score = ', current_best_score
            else
                print '(2a,2(a,i0),a,F0.6)', '**info(fgsl_ds_run): ', myTime%DateAndTimeAsString(), ': &
                                             target size reached at iteration ', &
                                             i , ' of a maximum of ', self%max_num_iter, '. Best score = ', current_best_score
            endif
        endif

        ! Update the best values
        call fgsl_ds_update_current_best(self)
        !print *, '**debug(fgsl_ds_run): with current values for the variables = ', self%current_best_fgsl
    end subroutine fgsl_ds_run

    !>  \brief  Print out best score and size per iteration
    subroutine ds_print_progress_summary(self)
        implicit none
        ! Arguments
        class(minimiser_ds), intent(in)  ::  self
        ! Private variables
        integer ::  i
        ! Start work
        write(*,'(3(A12,1x))') 'Iteration #', 'Simplex size', 'Best score'
        do I=1,size(self%best_score)
            write(*,'(i12,1x,F12.5,1x,F12.5)') I-1, self%simplex_size(i), self%best_score(i)
        enddo
    end subroutine ds_print_progress_summary

    subroutine grow_allocatable_array_real(self)
        implicit none
        ! Arguments
        real,   allocatable,    intent(inout) ::  self(:)
        ! Private variables
        real,   allocatable ::  tmp(:)
        integer ::  new_size
        ! Start work
        if (allocated(self)) then
            new_size = size(self) + 1
        else
            new_size = 1
        endif
        allocate(tmp(new_size))
        if (allocated(self)) then
            tmp(1:size(self)) = self
        else
            tmp = 0.0
        endif
        call move_alloc(FROM=TMP,TO=SELF)
    end subroutine grow_allocatable_array_real

    !>  \brief  Initialise a fgsl downhill simplex minimiser
    subroutine fgsl_ds_init(self,function_to_minimise,n,parameters,initial_score,random,max_num_iter)
        use, intrinsic  ::  iso_c_binding
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(minimiser_ds_fgsl),    intent(inout)   ::  self
        procedure(fgsl_minimisable_function)        ::  function_to_minimise
        integer,                    intent(in)      ::  n                           !<  Number of dimensions
        type(c_ptr),                intent(in)      ::  parameters                  !<  Parameters (a c pointer to an object, an array, whatever... ;
                                                                                    !!  probably makes sense to have it point to a derived-type object that
                                                                                    !!  contains useful stuff for the function to be minimised)
        real                                        ::  initial_score
        logical,        optional,   intent(in)      ::  random                      !<  Whether the starting setup of the ds should be randomised. False by default.
        integer,        optional,   intent(in)      ::  max_num_iter                !<  Maximum number of iterations
        ! Private variables
        logical                     ::  simplex_random_init
        logical,    parameter       ::  debug   =   .false.
        ! Start work

        if (debug) write(*,'(a,F0.6)') '**debug(fgsl_ds_init): entry, with initial score = ', initial_score

        ! Will we use the random version of the ds
        if (present(random)) then
            simplex_random_init = random
        else
            simplex_random_init = .false.
        endif

        ! Initialise the minimizer
        self%n = n
        self%n_fgsl = self%n
        if (simplex_random_init) then
            write(*,'(a)') '**info(fgsl_ds_init): initialising the ds using random basis vectors instead of fixed coordinate axes'
            self%minimiser_fgsl = fgsl_multimin_fminimizer_alloc(fgsl_multimin_fminimizer_nmsimplex2rand,self%n_fgsl)
        else
            self%minimiser_fgsl = fgsl_multimin_fminimizer_alloc(fgsl_multimin_fminimizer_nmsimplex2,self%n_fgsl)
        endif
        self%name = fgsl_multimin_fminimizer_name(self%minimiser_fgsl)
        !print *, '**debug(fgsl_ds_init): initialising minimiser named ', trim(self%name)
        ! Initialise the function to be minimised
        self%function_fgsl = fgsl_multimin_function_init(function_to_minimise,self%n_fgsl,parameters)
        self%target_function => function_to_minimise

        !call fgsl_ds_set_start_step(self,starting_values,first_step)

        ! The gsl minimisers maintain a current best estimate of the minimum at all times, which can be accessed as follows:
        !starting_values_vec = fgsl_multimin_fminimizer_x(self%minimiser_fgsl)
        !status = fgsl_vector_align(self%current_best_fgsl,starting_values_vec)
        !if (status .NE. fgsl_success) call terminate('Failed vector align for current_best in FGSL_DS_INIT')
        !print *, '**debug(fgsl_ds_init): done. current best params = ', self%current_best_fgsl

        ! Remember the initial score
        if (allocated(self%best_score)) deallocate(self%best_score)
        if (allocated(self%simplex_size)) deallocate(self%simplex_size)
        if (.not. allocated(self%best_score)) then
            allocate(self%best_score(1))
            allocate(self%simplex_size(1))
            self%best_score(1) = initial_score
            self%simplex_size  = 0.0
        endif

        !
        if (present(max_num_iter)) self%max_num_iter = max_num_iter

        ! Set the in_memory flag
        self%in_memory = .true.
    end subroutine fgsl_ds_init

    !>  \brief  Set the starting point and first step for a minimiser object
    subroutine fgsl_ds_set_start_step(self,starting_values,first_step)
        use, intrinsic  ::  iso_c_binding
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(minimiser_ds_fgsl),    intent(inout)   ::  self
        real(kind=8),               intent(in)      ::  starting_values(:)
        real(kind=8),               intent(in)      ::  first_step(:)
        ! Private variables
        integer(fgsl_int)           ::  status
        type(fgsl_vector)           ::  starting_values_vec, first_step_vec
        real(kind=FGSL_DOUBLE), target      ::  starting_values_fgsl(self%n)
        real(kind=FGSL_DOUBLE), target      ::  first_step_fgsl(self%n)
        real    ::  starting_score
        logical, parameter :: debug = .false.
        ! Start work

        if (debug) then
            write (*,*) '**debug(fgsl_ds_set_start_step): starting values = ', starting_values
            write (*,*) '**debug(fgsl_ds_set_start_step):      first step = ', first_step
        endif

        ! Deal with best score history array
        if (allocated(self%best_score)) then
            if (debug) write(*,'(a)') '**debug(fgsl_ds_set_start_step): best score array already allocated, will reset it'
            starting_score = self%best_score(1)
            if (debug) write(*,'(a,F0.6)') '**debug(fgsl_ds_set_start_step): starting score = ', starting_score
            deallocate(self%best_score)
            if (allocated(self%simplex_size)) deallocate(self%simplex_size)
            allocate(self%best_score(1))
            allocate(self%simplex_size(1))
            self%best_score(1) = starting_score
            self%simplex_size(1) = 0.0
        endif

        ! Check sizes of start, step arrays
        if ((size(starting_values) .ne. self%n) .or. (size(first_step) .ne. self%n)) then
            write(*,'(a,3(i0,a))') '**ERROR(fgsl_ds_set_start_step): size of starting values (', size(starting_values), &
                                    ') and first step (', size(first_step), &
                                    ') arrays do not match the number of dimensions (', self%n, ')'
            call terminate('Fatal error in FGSL_DS_SET_START_STEP')
        endif

        ! Convert incoming starting values to fgsl types
        starting_values_fgsl = starting_values
        starting_values_vec = fgsl_vector_init(1.0_FGSL_DOUBLE)
        status = fgsl_vector_align(starting_values_fgsl,self%n_fgsl,starting_values_vec,self%n_fgsl,0_fgsl_size_t,1_fgsl_size_t)
        if (status .ne. fgsl_success) call terminate('Failed vector align in FGSL_DS_SET_START_STEP')
        ! Convert incoming first step to fgsl types
        first_step_fgsl = first_step
        first_step_vec = fgsl_vector_init(1.0_FGSL_DOUBLE)
        status = fgsl_vector_align(first_step_fgsl,self%n_fgsl,first_step_vec,self%n_fgsl,0_fgsl_size_t,1_fgsl_size_t)
        if (status .ne. fgsl_success) call terminate('Failed vector align in FGSL_DS_SET_START_STEP')
        ! Set the starting point
        !print *, '**debug(fgsl_ds_set_start_step): setting the starting point...'
        status = fgsl_multimin_fminimizer_set(self%minimiser_fgsl,self%function_fgsl,starting_values_vec,first_step_vec)    ! This is the line that needs to be called again on the same object to get a new, random first step direction.
        !print *, '**debug(fgsl_ds_set_start_step): setting the starting point... done'
        if (status .ne. fgsl_success) call terminate('Failed fminimizer_set in FGSL_DS_SET_START_STEP')
        ! Free up FGSL-versions of the starting value and first step
        call fgsl_vector_free(starting_values_vec)
        call fgsl_vector_free(first_step_vec)
    end subroutine fgsl_ds_set_start_step

    !>  \brief  Update the current best parameters in the minimiser object
    subroutine fgsl_ds_update_current_best(self)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(minimiser_ds_fgsl),    intent(inout)   ::  self
        ! Private variables
        type(fgsl_vector) :: values_vec
        real(fgsl_double)   ::  min
        integer(fgsl_int) :: status
        logical,    parameter   ::  debug = .false.
        character(len=6)    ::  n_c
        ! Start work
        values_vec = fgsl_multimin_fminimizer_x(self%minimiser_fgsl)
        status = fgsl_vector_align(self%current_best_fgsl,values_vec)
        if (status .ne. fgsl_success) call terminate('Failed vector align for current_best in FGSL_DS_UPDATE_CURRENT_BEST')

        min = fgsl_multimin_fminimizer_minimum(self%minimiser_fgsl)
        self%current_minimum_fgsl = min

        if (debug) then
            write(n_c,'(i0)') self%n
            print '(A'//n_c//'(F0.3,1x))', '**debug(fgsl_ds_update_current_best): updating best values to ', self%current_best_fgsl
            print '(a,F0.6)', '**debug(fgsl_ds_update_current_best): updating best score to ', self%current_minimum_fgsl
        endif
    end subroutine fgsl_ds_update_current_best

    !>  \brief Deallocate memory taken up by the minimiser
    subroutine fgsl_ds_free(self)
        implicit none
        ! Arguments
        type(minimiser_ds_fgsl),    intent(inout)   ::  self
        ! Private variables

        ! Start work
        nullify(self%target_function)
        nullify(self%current_best_fgsl)
        call fgsl_multimin_fminimizer_free(self%minimiser_fgsl)
        call fgsl_multimin_function_free(self%function_fgsl)
        !
        self%in_memory = .false.
    end subroutine fgsl_ds_free



    !>  \brief  Run the brute force minimiser's exhaustive search
    subroutine bf_run(self,expected_starting_score)
        use runtime_parameters
        use datesandtimes
        use ProgressBars
        use omp_lib
        use, intrinsic :: iso_c_binding
        implicit none
        ! arguments
        type(minimiser_bf), target,     intent(inout)   ::  self    !<  brute-force minimiser (must be initialised)
        real,               optional,   intent(in)      ::  expected_starting_score !<  useful for debugging
        ! private variables
        integer                             ::  ierr!, i
        real(kind=8),   target, allocatable ::  current_values(:), current_values_private(:)
        logical                             ::  search_completed
        integer                             ::  current_iteration, num_iterations_completed
        real                                ::  seconds_since_start, seconds_per_iteration, seconds_to_finish
        integer                             ::  clock_count_at_start, clock_count_current,clock_counts_per_second
        logical,            parameter       ::  output_at_every_iteration   =   .false.
        real                                ::  current_score
        !type(c_ptr)                         ::  current_values_cptr
        type(c_ptr)                         ::  starting_values_cptr
        type(c_ptr)                         ::  current_values_private_cptr
        type(dateandtime)                   ::  mytime
        type(ProgressBar)                   ::  my_progress_bar
        logical,            parameter       ::  debug = .false.
        logical                             ::  spawn_threads
        ! start work

        ! Check the minimiser is ready
        if (.not. self%in_memory) then
            write(*,'(a)') '**error(bf_run): minimiser was not initialised, call bf_init first!'
            call terminate('fatal error in bf_run')
        endif

        ! Allocate memory for the current values
        allocate(current_values(self%n),stat=ierr)
        if (ierr .ne. 0) call terminate('fatal error (memory allocation) in bf_run')


        ! Start work

        ! The time when we start
        call system_clock(clock_count_at_start,clock_counts_per_second)

        ! The incoming values & score
        starting_values_cptr = c_loc(self%starting_value)
        self%best_score = self%target_function(starting_values_cptr,self%parameters(1))
        self%best_value = self%starting_value
        write(*,'(a,f6.3)') '**info(bf_run): score on input = ', self%best_score

        ! Sanity check: is the score at the starting parameter values as expected?
        if (present(expected_starting_score)) then
            if (abs(expected_starting_score+self%best_score) .gt. 0.01) then
                write(*,'(a,2(f0.3,1x))') '**error(bf_run): starting score not as expected: ', &
                                            self%best_score, expected_starting_score
                call terminate('fatal error in bf_run. failed sanity check.')
            endif
        endif

        ! Will we be spawning threads?
        spawn_threads = .false.
#ifdef _OPENMP
        if (.not. omp_in_parallel()) then
            if (omp_get_max_threads() .gt. 1) spawn_threads = .true.
        endif
#endif


        ! The starting point for the brute-force search
        current_values = -huge(1.0d0)
        self%dimension_at_max = .false.

        if (.not. output_at_every_iteration) call my_progress_bar%Begin(self%num_iterations)

        ! How many iterations have we completed?
        num_iterations_completed = 0


        !$omp   parallel default(shared) &
        !$omp   private(current_iteration, current_values_private,current_score)

        ! were we given the correct number of parameter pointers?
#ifdef _OPENMP
        if (spawn_threads) then
            if (omp_get_num_threads() .ne. size(self%parameters)) then
                write(*,'(2(a,i0),a)') '**error(bf_run): want to spawn ',omp_get_num_threads(), &
                                        ' but have ', size(self%parameters), ' sets of parameters '
                call terminate('Fatal error in bf_run')
            endif
        endif
#endif

        ! allocate & initialise
        current_values_private = current_values

        ! make sure the c pointer to the current values is setup
        current_values_private_cptr = c_loc(current_values_private)

        ! start the brute-force search iterations
        !$omp   do
        do current_iteration=1,self%num_iterations

            !$omp critical (bf_run_crit_incr)
            ! what values for the parameters should we try now?
            call bf_increment_current_values(self,current_values,search_completed)
            if (search_completed) then
                call terminate('failed sanity check in bf_run')
                !exit
            endif
            current_values_private = current_values
            !$omp end critical (bf_run_crit_incr)


            ! Try the current parameters by calling the scoring function
            if (debug) then
                write(*,'(a)',advance='no') '**debug(bf_run): calling scoring function with values: '
                print *, current_values
            endif
            current_score =  self%target_function(current_values_private_cptr,self%parameters( &
#ifdef _OPENMP
            omp_get_thread_num()+1 &
#else
            1 &
#endif
            ))

            !$omp critical (bf_run_crit_best)
            num_iterations_completed = num_iterations_completed + 1
            ! if the score is the best we've seen so far, remember the values and the score
            if (current_score .lt. self%best_score) then
                self%best_score = current_score
                self%best_value = current_values_private
            endif
            ! Progress bar
            if (.not. output_at_every_iteration) call my_progress_bar%Update(num_iterations_completed)

            ! Output for the user
            if (output_at_every_iteration) then
                ! add a printout with a time estimate
                call system_clock(clock_count_current)
                seconds_since_start = real(clock_count_current - clock_count_at_start) / real(clock_counts_per_second)
                seconds_per_iteration = seconds_since_start / current_iteration
                seconds_to_finish = (self%num_iterations - current_iteration + 1) * seconds_per_iteration
                write(*,'(3a,i6,a,i6,a,f6.3,2a)') '**info(bf_run): ', mytime%dateandtimeasstring(), ': after iteration ', &
                                                    current_iteration, ' of ', self%num_iterations, '. current best score : ',&
                                                     self%best_score, '. estimated time to finish: ', &
                                                     mytime%durationasstring(seconds_to_finish)
            endif
            if (debug) then
                write(*,'(a)',advance='no') '**debug(bf_run): values: '
                write(*,*) current_values
                write(*,'(a,f0.4)') '**debug(bf_run): score: ', current_score
            endif
            !$omp end critical (bf_run_crit_best)

        enddo ! end of loop over exhaustive search iterations
        !$omp end do

        if (allocated(current_values_private)) deallocate(current_values_private)

        !$omp end parallel

        if (.not. output_at_every_iteration) call my_progress_bar%Finish()

        ! Search is done
        write(*,'(2(a,i6),a)') '**info(bf_run): brute-force search completed in ', num_iterations_completed, &
                            ' iterations (predicted number of iterations was ', self%num_iterations, ')'

        if (debug) then
            print *, '**debug(bf_run): best value = ', self%best_value
            print *, '**debug(bf_run): best score = ', self%best_score
        endif

    end subroutine bf_run

    !>  \brief  In the bf loop, this should be called before the scoring function. Before the bf loop, set all values in current_values to -huge
    subroutine bf_increment_current_values(self,current_values,search_is_now_completed)
        use runtime_parameters, only : terminate
        implicit none
        ! arguments
        type(minimiser_bf), intent(inout)   ::  self
        real(kind=8),       intent(inout)   ::  current_values(*)
        logical,            intent(out)     ::  search_is_now_completed
        ! private variables
        integer                             ::  i
        !

        ! do the increment
        search_is_now_completed= .true.
        do i=1,self%n
            ! if we haven't reached the max for this dimension, increment it, and reset all previous dimensions to their starting point
            if (.not. self%dimension_at_max(i)) then
                ! if we got here, it means our search is not over yet
                search_is_now_completed = .false.
                ! increment the ith dimension
                ! it could be that the ith dimension has a very large negative value, perhaps because the search hasn't even started yet
                if (current_values(i) .lt. self%starting_value(i) - self%half_range(i)) then
                    current_values(i) = self%starting_value(i) - self%half_range(i)
                else
                    current_values(i) = current_values(i) + self%step_size(i)
                endif
                ! reset all previous dimensions to their starting values, and reset their dimension_at_max flags to .false.
                if (i .gt. 1) then
                    current_values(1:i-1) = self%starting_value(1:i-1) - self%half_range(1:i-1)
                    if (any( .not. self%dimension_at_max(1:i-1))) then
                        call terminate('failed sanity check in bf_increment_current_values')
                    endif
                    self%dimension_at_max(1:i-1) = .false.
                endif
                ! if the ith dimension has reached or gone over its max, set it at the max, and set its logical flag dimension_at_max to .true.
                if (current_values(i) .ge. self%starting_value(i) + self%half_range(i)) then
                    current_values(i) = self%starting_value(i) + self%half_range(i)
                    self%dimension_at_max(i) = .true.
                endif
                exit
            endif
        enddo

        do i=1,self%n
            if (current_values(i) .lt. self%starting_value(i) - self%half_range(i)) then
                current_values(i) = self%starting_value(i) - self%half_range(i)
            endif
        enddo

    end subroutine bf_increment_current_values


    !>  \brief  initialise a brute-force minimiser
    subroutine bf_init(self,function_to_minimise,parameters,num_dim,starting_value,half_range,step_size)
        use runtime_parameters, only : terminate
        implicit none
        ! arguments
        type(minimiser_bf),         intent(inout)   ::  self                    !<  the brute-force minimiser to initialise / allocate
        procedure(fgsl_minimisable_function)        ::  function_to_minimise    !<  the function to minimise
        type(c_ptr),  allocatable,  intent(in)      ::  parameters(:)           !<  parameters (a c pointer to an object, an array, whatever... ;
                                                                                !!  probably makes sense to have it point to a derived-type object that
                                                                                !!  contains useful stuff for the function to be minimised).
                                                                                !!  One pointer per thread needs to be provided

        integer,                    intent(in)      ::  num_dim                 !<  number of dimensions
        real(kind=8), allocatable,  intent(in)      ::  starting_value(:)       !<  starting point for the minimiser
        real(kind=8), allocatable,  intent(in)      ::  half_range(:)           !<  half-range to search over
        real(kind=8), allocatable,  intent(in)      ::  step_size(:)            !<  step size of the exhaustive search
        ! private variables
        integer                 ::  ierr!, i
        real(kind=8)            ::  current_values(num_dim)
        logical                 ::  search_done
        logical,    parameter   ::  debug = .false.
        ! start work

        ! if already allocated, deallocate
        if (self%in_memory) then
            !call self%free()
            call bf_free(self)
        endif

        ! allocate memory
        self%n = num_dim
        allocate(self%starting_value(self%n),self%half_range(self%n),self%step_size(self%n), &
                 self%best_value(self%n),self%dimension_at_max(self%n),stat=ierr)
        if (ierr .ne. 0) then
            write(*,'(a,i3)') '**error(bf_init): memory allocation failed with error ', ierr
            call terminate('fatal error in bf_init')
        endif

        ! check the incoming starting values, half range and step size have expected dimensionality
        if (size(starting_value) .ne. num_dim .or. size(half_range) .ne. num_dim .or. size(step_size) .ne. num_dim) then
            write(*,'(a,i5,a,3i5)') '**error(bf_init): incoming arrays do not have expected size of ', num_dim, ': ',       &
                                    size(starting_value), size(half_range), size(step_size)
            call terminate('fatal error in bf_init')
        endif

        ! copy starting values, half range and step size over
        self%starting_value = starting_value
        self%half_range = half_range
        self%step_size = step_size

        !
        self%dimension_at_max = .false.

        ! work out how many iterations the exhaustive search will take
        ! we will start at lowest value for each dimension, then go in step sizes, until we get to upper limit.
        ! we will also do extra iterations for the starting value and the upper limit
        self%num_iterations = 0
        current_values = -huge(1.0d0)
        do
            call bf_increment_current_values(self,current_values,search_done)
            if (search_done) exit
            self%num_iterations = self%num_iterations + 1
        enddo
        !do i=1,self%n
        !    self%num_iterations = self%num_iterations * ceiling(self%half_range(i)*2/self%step_size(i))
        !enddo
        write(*,'(a,i0)') '**info(bf_init): predicted number of steps in the exhaustive search = ', self%num_iterations
        write(*,'(a,i0)') '**info(bf_init): number of dimensions = ', self%n

        if (debug) then
            print *, '**debug(bf_init): half-ranges = ', self%half_range
            print *, '**debug(bf_init): step sizes  = ', self%step_size
        endif

        ! set up pointer to function to minimise
        self%target_function => function_to_minimise

        !
        self%parameters = parameters

        ! set the memory flag
        self%in_memory = .true.
    end subroutine bf_init

    !>  \brief  Deallocate a brute-force minimiser
    subroutine bf_free(self)
        implicit none
        ! Argument
        class(minimiser_bf),    intent(inout)   ::  self    !<  Brute-force minimiser to deallocate
        ! Private variables

        ! Start work
        if (.not. self%in_memory) then
            write(*,'(a)') '**ERROR(bf_free): cannot free a minimiser not in memory'
            stop 'Fatal error in BF_FREE'
        endif
        ! Deallocate component arrays
        deallocate(self%starting_value,self%half_range,self%step_size)
        ! Reset pointer
        nullify(self%target_function)
        ! Reset values
        self%n = 0
        self%num_iterations = 0
        self%in_memory = .false.
    end subroutine bf_free




end module minimisers
