module modes
    use runtime_parameters, only : word_len, line_len
    use rtps
    use Globals
    implicit none
    private

    !>  \brief  Definition of a subroutine which constitutes a frealix2 mode
    abstract interface
        subroutine mode_subroutine()
        end subroutine mode_subroutine
    end interface

    type, private :: rtp_real_p
        type(rtp_real), pointer ::  p
    end type
    type, private :: rtp_intg_p
        type(rtp_intg), pointer ::  p
    end type
    type, private :: rtp_logi_p
        type(rtp_logi), pointer ::  p
    end type
    type, private :: rtp_char_p
        type(rtp_char), pointer ::  p
    end type

    type, public :: mode
        private
        character(len=WORD_LEN)                     ::  command
        procedure(mode_subroutine), pointer, nopass ::  subroutine                      !<  Subroutine to call when the mode is in effect
        character(len=LINE_LEN)                     ::  short_description               !<  a short description of what the mode does
        ! Run-time parameters associated with this mode
        type(rtp_real_p), allocatable               ::  real_rtps(:)                    !<  The runtime parameters needed to run the mode
        type(rtp_intg_p), allocatable               ::  intg_rtps(:)
        type(rtp_logi_p), allocatable               ::  logi_rtps(:)
        type(rtp_char_p), allocatable               ::  char_rtps(:)
        integer                                     ::  number_of_rtps      =   0       !<  The number of runtime parameters associated with this mode
        contains
            procedure   ::  init                        =>  mode_init
            procedure   ::  add_rtp                     =>  mode_add_rtp
            procedure   ::  print_info                  =>  mode_print_info
            procedure   ::  print_welcome               =>  mode_print_welcome
            procedure   ::  print_command_file          =>  mode_print_command_file
            procedure   ::  get_rtp_values_from_user    =>  mode_get_rtp_values_from_user
            procedure   ::  run                         =>  mode_run
    end type mode


contains

    !>  \brief  Run a frealix mode
    subroutine mode_run(self)
        implicit none
        ! Arguments
        class(mode),        intent(inout)      ::  self
        ! Private variables
        ! Start work
        if (this_program%IsVerbose()) then
            write(*,'(a)') '<mode command='//trim(adjustl(self%command))//'>'
        endif
        ! Get rtp values from user
        call self%get_rtp_values_from_user()
        ! Print welcome message
        call self%print_welcome()
        ! Run the actual routine
        call self%subroutine()
        if (this_program%IsVerbose()) then
            write(*,'(a)') '</mode>'
        endif
    end subroutine mode_run


    !>  \brief  Initialise a frealix mode
    subroutine mode_init(self,command,subroutine,short_description)
        implicit none
        ! Arguments
        class(mode),        intent(inout)   ::  self
        character(len=*),   intent(in)      ::  command
        procedure(mode_subroutine)          ::  subroutine
        character(len=*),   intent(in)      ::  short_description
        ! Private variables

        ! Start work
        self%command            =   command
        self%short_description  =   short_description
        self%subroutine         =>  subroutine
        self%number_of_rtps     =   0
        if (allocated(self%real_rtps)) deallocate(self%real_rtps)
    end subroutine mode_init


    !>  \brief  Get all values for RTPs from user
    subroutine mode_get_rtp_values_from_user(self)
        implicit none
        ! Argument
        class(mode),        intent(inout)      ::  self
        ! Private variables
        integer ::  current_rtp, rtp_type, rtp_index
        ! Start work
        do current_rtp=1,self%number_of_rtps
            ! We need to find the rtp with order value current_rtp
            call mode_find_rtp_from_order(self,current_rtp,rtp_type,rtp_index)
            !
            select case (rtp_type)
                case (1)
                    call self%real_rtps(rtp_index)%p%get_value_from_user()
                case (2)
                    call self%intg_rtps(rtp_index)%p%get_value_from_user()
                case (3)
                    call self%logi_rtps(rtp_index)%p%get_value_from_user()
                case (4)
                    call self%char_rtps(rtp_index)%p%get_value_from_user()
            end select
        enddo
    end subroutine mode_get_rtp_values_from_user

    !>  \brief  Print welcome message for the mode
    subroutine mode_print_welcome(self)
        implicit none
        ! Arguments
        class(mode),        intent(in)      ::  self
        ! Private variables

        ! Start work
        write(*,'(3a)') 'Frealix is running in its ', trim(adjustl(self%command)), ' mode'
        call mode_print_info(self)
        ! Flush standard output
        flush(6) ! This is a standard f2003 statement
    end subroutine mode_print_welcome

    !>  \brief  Print information about the mode
    subroutine mode_print_info(self)
        implicit none
        ! Arguments
        class(mode),        intent(in)      ::  self
        ! Private variables
        integer ::  current_rtp, rtp_type, rtp_index
        ! Start work
        !write(*,'(3a)') '** ', trim(adjustl(self%command)) ,' mode information **'
        !write(*,'(a)') ' '
        write(*,'(2a)') 'Mode description: ', trim(adjustl(self%short_description))
        write(*,'(a)') 'Run-time parameters:'

        call rtp_print_info_header()

        ! Print information about the mode's RTPs (loop over them)
        do current_rtp=1,self%number_of_rtps
            ! We need to find the rtp with order value current_rtp
            call mode_find_rtp_from_order(self,current_rtp,rtp_type,rtp_index)
            !
            select case (rtp_type)
                case (1)
                    call self%real_rtps(rtp_index)%p%print_info()
                case (2)
                    call self%intg_rtps(rtp_index)%p%print_info()
                case (3)
                    call self%logi_rtps(rtp_index)%p%print_info()
                case (4)
                    call self%char_rtps(rtp_index)%p%print_info()
            end select
        enddo

        write(*,'(a)') ' '
    end subroutine mode_print_info

    !>  \brief  Print out a command file for the mode
    subroutine mode_print_command_file(self,command_filename)
        use runtime_parameters, only : available_lun, lun_is_now_available, filename_len, terminate
        use DatesAndTimes
        implicit none
        ! Arguments
        class(mode),                    intent(in)  ::  self
        character(len=*),               intent(in)  ::  command_filename
        ! Private variables
        integer ::  current_rtp, rtp_type, rtp_index
        integer ::  lun
        character(len=32)   ::  frealix_mode
        integer ::  ios
        type(DateAndTime)   ::  myTime
        ! Start work
        ! Get logical unit number
        lun = available_lun()
        ! Open file
        open(unit=LUN,file=TRIM(adjustl(command_filename)),status='REPLACE',iostat=IOS)
        if (ios .ne. 0) then
            write(*,'(3a,i0)') '**ERROR(mode_print_command_file): file opening failed for file ', &
                                    trim(adjustl(command_filename)), ' with error ', ios
            call terminate('Fatal error in MODE_PRINT_COMMAND_FILE')
        endif
        frealix_mode = 'FREALIX_MODE'
        write(lun,'(a)') '!'
        write(lun,'(2a)') '! Example frealix command file, generated ', myTime%DateAndTimeAsString()
        write(lun,'(a)') '!'
        write(lun,'(a32,a)') adjustl(frealix_mode), self%command
        write(lun,'(a)') '!'
        ! Print information about the mode's RTPs (loop over them)
        do CURRENT_RTP=1,self%number_of_rtps
            ! We need to find the rtp with order value current_rtp
            call mode_find_rtp_from_order(self,current_rtp,rtp_type,rtp_index)
            !
            select case (rtp_type)
                case (1)
                    call self%real_rtps(rtp_index)%p%print_to_command_file(lun)
                case (2)
                    call self%intg_rtps(rtp_index)%p%print_to_command_file(lun)
                case (3)
                    call self%logi_rtps(rtp_index)%p%print_to_command_file(lun)
                case (4)
                    call self%char_rtps(rtp_index)%p%print_to_command_file(lun)
            end select
        enddo

        call lun_is_now_available(lun)
    end subroutine mode_print_command_file

    !>  \brief  Given the order index for an rtp, find it
    subroutine mode_find_rtp_from_order(self,order,rtp_type,rtp_index)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(mode),     intent(in)  ::  self
        integer,        intent(in)  ::  order       !<  The order value for the rtp we're looking for
        integer,        intent(out) ::  rtp_type    !<  Return 1, 2, 3, or 4 to indicate that the rtp is real, integer, logical or character
        integer,        intent(out) ::  rtp_index   !<  Return the index of the rtp within the appropriate array
        ! Private variables
        logical ::  found_rtp
        integer ::  i
        integer ::  sum_of_sizes
        ! Start work
        found_rtp = .false.
        ! No point looking if the order value is larger than the total number of rtps
        if (order .gt. self%number_of_rtps) then
            write(*,'(a,i0,a,i0,a)')    '**ERROR(mode_find_rtp_from_order): you asked to lookup rtp #', order, &
                                        ', but there are only ', self%number_of_rtps, ' rtps in this mode'
            call terminate('Fatal error in MODE_FIND_RTP_FROM_ORDER. Programming error.')
        endif
        ! Check the declared number of rtps matches the total of the rtp array sizes
        sum_of_sizes = 0
        if (allocated(self%real_rtps)) sum_of_sizes = sum_of_sizes + size(self%real_rtps)
        if (allocated(self%intg_rtps)) sum_of_sizes = sum_of_sizes + size(self%intg_rtps)
        if (allocated(self%logi_rtps)) sum_of_sizes = sum_of_sizes + size(self%logi_rtps)
        if (allocated(self%char_rtps)) sum_of_sizes = sum_of_sizes + size(self%char_rtps)
        if (sum_of_sizes .ne. self%number_of_rtps) then
            write(*,'(a,i0,a,i0)')  '**ERROR(mode_find_rtp_from_order): there are ', self%number_of_rtps,    &
                                    ' declared rtps in the mode, but the rtp arrays have a total size of ', sum_of_sizes
            call terminate('Fatal error in MODE_FIND_RTP_FROM_ORDER. Programming error.')
        endif

        if (allocated(self%real_rtps)) then
            ! First we try the real rtps
            do I=1,size(self%real_rtps)
                if (self%real_rtps(i)%p%order .eq. order) then
                    found_rtp = .true.
                    rtp_type = 1
                    rtp_index = i
                    return
                endif
            enddo
        endif
        if (allocated(self%intg_rtps)) then
            ! Next, we try integers
            do I=1,size(self%intg_rtps)
                if (self%intg_rtps(i)%p%order .eq. order) then
                    found_rtp = .true.
                    rtp_type = 2
                    rtp_index = i
                    return
                endif
            enddo
        endif
        if (allocated(self%logi_rtps)) then
            ! Next, we try logicals
            do I=1,size(self%logi_rtps)
                if (self%logi_rtps(i)%p%order .eq. order) then
                    found_rtp = .true.
                    rtp_type = 3
                    rtp_index = i
                    return
                endif
            enddo
        endif
        if (allocated(self%char_rtps)) then
            ! Next, we try characters
            do I=1,size(self%char_rtps)
                if (self%char_rtps(i)%p%order .eq. order) then
                    found_rtp = .true.
                    rtp_type = 4
                    rtp_index = i
                    return
                endif
            enddo
        endif

        if (.not. found_rtp) then
            write(*,'(a,i0,a,i0,a)') '**ERROR(mode_find_rtp_from_order): could not find rtp #', order, &
                                    '. There are ', self%number_of_rtps, ' rtps in this mode.'
            call terminate('Fatal error in MODE_FIND_RTP_FROM_ORDER')
        endif

    end subroutine mode_find_rtp_from_order

    !>  \brief  Add an rtp to the mode
    subroutine mode_add_rtp(self,new_rtp)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        class(mode),                    intent(inout)   ::  self
        class(rtp),     target,         intent(in)      ::  new_rtp
        ! Private variables
        logical,    parameter    :: debug   =   .false.
        ! Start work
        select type(new_rtp)
            type is (rtp_real)
                ! Grow array
                if (allocated(self%real_rtps)) then
                    self%real_rtps = reshape(self%real_rtps,SHAPE=[SIZE(self%real_rtps)+1],PAD=SELF%real_rtps)
                else
                    allocate(self%real_rtps(1))
                endif
                ! Copy rtp over
                self%real_rtps(size(self%real_rtps))%p => new_rtp
                ! Keep track of the order RTPs were added
                self%real_rtps(size(self%real_rtps))%p%order = self%number_of_rtps + 1
            type is (rtp_intg)
                ! Grow array
                if (allocated(self%intg_rtps)) then
                    self%intg_rtps = reshape(self%intg_rtps,SHAPE=[SIZE(self%intg_rtps)+1],PAD=SELF%intg_rtps)
                else
                    allocate(self%intg_rtps(1))
                endif
                ! Copy rtp over
                self%intg_rtps(size(self%intg_rtps))%p => new_rtp
                ! Keep track of the order RTPs were added
                self%intg_rtps(size(self%intg_rtps))%p%order = self%number_of_rtps + 1
                if (debug) then
                    write(*,'(a,3i0)') '**debug(mode_add_rtp): added intg rtp: ', self%intg_rtps(size(self%intg_rtps))%p%order, &
                                                                                  self%number_of_rtps, size(self%intg_rtps)
                endif
            type is (rtp_logi)
                ! Grow array
                if (allocated(self%logi_rtps)) then
                    self%logi_rtps = reshape(self%logi_rtps,SHAPE=[SIZE(self%logi_rtps)+1],PAD=SELF%logi_rtps)
                else
                    allocate(self%logi_rtps(1))
                endif
                ! Copy rtp over
                self%logi_rtps(size(self%logi_rtps))%p => new_rtp
                ! Keep track of the order RTPs were added
                self%logi_rtps(size(self%logi_rtps))%p%order = self%number_of_rtps + 1
            type is (rtp_char)
                ! Grow array
                if (allocated(self%char_rtps)) then
                    self%char_rtps = reshape(self%char_rtps,SHAPE=[SIZE(self%char_rtps)+1],PAD=SELF%char_rtps)
                else
                    allocate(self%char_rtps(1))
                endif
                ! Copy rtp over
                self%char_rtps(size(self%char_rtps))%p => new_rtp
                ! Keep track of the order RTPs were added
                self%char_rtps(size(self%char_rtps))%p%order = self%number_of_rtps + 1
            class default
                call terminate('RTP type not implemented')
        end select
        self%number_of_rtps = self%number_of_rtps + 1
    end subroutine mode_add_rtp

!
!    !
!    ! Routines below probably not useful
!    !
!
!
!    !>  \brief  Initialise an rtp and add it to the mode
!    subroutine mode_init_and_add_real_rtp(self,default_value,keyword,description,help)
!        implicit none
!        ! Arguments
!        class(mode),                    intent(inout)   ::  self
!        real,                           intent(in)      ::  default_value   !<  Default value for the parameter
!        character(len=*),               intent(in)      ::  keyword         !<  Keyword used in frealix command files to identify the parameter
!        character(len=*),   optional,   intent(in)      ::  description     !<  Description used in the question and as comment in frealix command files
!        character(len=*),   optional,   intent(in)      ::  help            !<  Help to be given to user to describe what the parameter means / does
!        ! Private variables
!        type(rtp_real)  ::  rtp
!        ! Start work
!        call rtp%init(default_value,keyword,DESCRIPTION=DESCRIPTION,HELP=HELP)
!        call mode_add_rtp(self,rtp)
!    end subroutine mode_init_and_add_real_rtp
!
!    !>  \brief  Initialise an rtp and add it to the mode
!    subroutine mode_init_and_add_intg_rtp(self,default_value,keyword,description,help)
!        implicit none
!        ! Arguments
!        class(mode),                    intent(inout)   ::  self
!        integer,                        intent(in)      ::  default_value   !<  Default value for the parameter
!        character(len=*),               intent(in)      ::  keyword         !<  Keyword used in frealix command files to identify the parameter
!        character(len=*),   optional,   intent(in)      ::  description     !<  Description used in the question and as comment in frealix command files
!        character(len=*),   optional,   intent(in)      ::  help            !<  Help to be given to user to describe what the parameter means / does
!        ! Private variables
!        type(rtp_intg)  ::  rtp
!        ! Start work
!        call rtp%init(default_value,keyword,DESCRIPTION=DESCRIPTION,HELP=HELP)
!        call mode_add_rtp(self,rtp)
!    end subroutine mode_init_and_add_intg_rtp
!
!    !>  \brief  Initialise an rtp and add it to the mode
!    subroutine mode_init_and_add_char_rtp(self,default_value,keyword,description,help)
!        implicit none
!        ! Arguments
!        class(mode),                    intent(inout)   ::  self
!        character(len=*),               intent(in)      ::  default_value   !<  Default value for the parameter
!        character(len=*),               intent(in)      ::  keyword         !<  Keyword used in frealix command files to identify the parameter
!        character(len=*),   optional,   intent(in)      ::  description     !<  Description used in the question and as comment in frealix command files
!        character(len=*),   optional,   intent(in)      ::  help            !<  Help to be given to user to describe what the parameter means / does
!        ! Private variables
!        type(rtp_char)  ::  rtp
!        ! Start work
!        call rtp%init(default_value,keyword,DESCRIPTION=DESCRIPTION,HELP=HELP)
!        call mode_add_rtp(self,rtp)
!    end subroutine mode_init_and_add_char_rtp
!
!    !>  \brief  Initialise an rtp and add it to the mode
!    subroutine mode_init_and_add_logi_rtp(self,default_value,keyword,description,help)
!        implicit none
!        ! Arguments
!        class(mode),                    intent(inout)   ::  self
!        logical,                        intent(in)      ::  default_value   !<  Default value for the parameter
!        character(len=*),               intent(in)      ::  keyword         !<  Keyword used in frealix command files to identify the parameter
!        character(len=*),   optional,   intent(in)      ::  description     !<  Description used in the question and as comment in frealix command files
!        character(len=*),   optional,   intent(in)      ::  help            !<  Help to be given to user to describe what the parameter means / does
!        ! Private variables
!        type(rtp_logi)  ::  rtp
!        ! Start work
!        call rtp%init(default_value,keyword,DESCRIPTION=DESCRIPTION,HELP=HELP)
!        call mode_add_rtp(self,rtp)
!    end subroutine mode_init_and_add_logi_rtp



end module modes
