!>  \brief  Mode to bin or unbin images, volumes, frealix2 paramter files or boxer coordinates files.
module mode_bfactor
    ! Modules
    use modes
    use rtps


    implicit none
    type(mode) :: bfactor

    type(rtp_char), target  ::  input_filename  =               &
        rtp_char(   value='input.mrc',                  &
                    keyword='INPUT_FILENAME',           &
                    description='')


    type(rtp_char), target  ::  output_filename  =               &
        rtp_char(   value='output.mrc',                  &
                    keyword='OUTPUT_FILENAME',           &
                    description='')


    type(rtp_real), target  ::  b_factor  =               &
        rtp_real(   value=0.0,                  &
                    keyword='BFACTOR',           &
                    description='B-factor, in A^2, following the crystallographic, aka Henderson, aka Grigorieff convention')


    type(rtp_logi), target  ::  low_pass  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='LOW_PASS',           &
                    description='Whether to also apply a cosine-edge low-pass filter.')


    type(rtp_real), target  ::  low_pass_resolution  =               &
        rtp_real(   value=1.0,                  &
                    keyword='LOW_PASS_RESOLUTION',           &
                    description='Low-pass filter resolution, in Angstroms. This is the spacing at which the filter reaches 0.')


    type(rtp_real), target  ::  low_pass_falloff  =               &
        rtp_real(   value=40.0,                  &
                    keyword='LOW_PASS_FALLOFF',           &
                    description='Low-pass filter falloff width, in Angstroms.')


    type(rtp_logi), target  ::  high_pass  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='HIGH_PASS',           &
                    description='Whether to also apply a cosine-edge high-pass filter')


    type(rtp_real), target  ::  high_pass_resolution  =               &
        rtp_real(   value=250.0,                  &
                    keyword='HIGH_PASS_RESOLUTION',           &
                    description='High-pass filter resolution, in Angstroms. This is the spacing at which the filter reaches 1.0.')


    type(rtp_real), target  ::  high_pass_falloff  =               &
        rtp_real(   value=250.0,                  &
                    keyword='HIGH_PASS_FALLOFF',           &
                    description='High-pass filter falloff width, in Angstroms.')


    type(rtp_real), target  ::  pixel_size  =               &
        rtp_real(   value=1.0,                  &
                    keyword='PIXEL_SIZE',           &
                    description='In Angstroms.')

    type(rtp_logi), target  ::  two_d  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='TWO_D',           &
                    description='Process 2d slices rather than the 3d volume')

contains

subroutine mode_bfactor_init_and_run
    ! Intialise the mode
    call bfactor%init(command='BFACTOR',subroutine=mode_bfactor_slave,short_description='Add description here')

    ! Add RTPs to the mode
    call bfactor%add_rtp(input_filename)
    call bfactor%add_rtp(output_filename)
    call bfactor%add_rtp(b_factor)
    call bfactor%add_rtp(low_pass)
    call bfactor%add_rtp(low_pass_resolution)
    call bfactor%add_rtp(low_pass_falloff)
    call bfactor%add_rtp(high_pass)
    call bfactor%add_rtp(high_pass_resolution)
    call bfactor%add_rtp(high_pass_falloff)
    call bfactor%add_rtp(pixel_size)
    call bfactor%add_rtp(two_d)

    ! Run the mode
    call bfactor%run()
end subroutine mode_bfactor_init_and_run

subroutine mode_bfactor_slave
    use runtime_parameters
    use images
    use imagefiles
    implicit none
    ! Variables
    type(image)     ::  vol, stack
    type(imagefile) ::  input_file, output_file
    integer         ::  i


    ! Start work
    call input_file%init(input_filename%value)


    ! We assume the image is square/cube
    if (two_d%value) then
        if (vol%logical_dimensions(1) .ne. vol%logical_dimensions(2)) then
            write(*,'(a,2(i0,1x))') '**error(mode_bfactor): input images are not square: ', vol%logical_dimensions(1:2)
            call terminate('Fatal error in MODE_BFACTOR')
        endif
    else
        if (.not. vol%issquare()) then
            write(*,'(a,3(i0,1x))') '**error(mode_bfactor): input volume is not cubic: ', vol%logical_dimensions
            call terminate('Fatal error in MODE_BFACTOR')
        endif
        call stack%readfromdisk(input_file,read_volume_not_slice=.true.)
        vol = stack
    endif

    call output_file%init(output_filename%value,mould=input_file,pixel_size=pixel_size%value)

    if (two_d%value) then
        do i=1,input_file%getstacksize()
            call vol%readfromdisk(input_file,wanted_location=i)
            call vol%ForwardFFT()
            call vol%applybfactor(bfactor=b_factor%value/pixel_size%value**2,   &
                                lp_cosine=low_pass%value,                  &
                                lp_cosine_freq=pixel_size%value/low_pass_resolution%value,  &
                                lp_cosine_falloff=pixel_size%value/low_pass_falloff%value,  &
                                hp_cosine=high_pass%value,                  &
                                hp_cosine_freq=pixel_size%value/high_pass_resolution%value,  &
                                hp_cosine_falloff=pixel_size%value/high_pass_falloff%value,  &
                                print_filter_shape=.true.)
            call vol%BackwardFFT()
            call vol%writetodisk(output_file,wanted_location=i)
        enddo
    else
        call vol%ForwardFFT()
        call vol%applybfactor(bfactor=b_factor%value/pixel_size%value**2,   &
                            lp_cosine=low_pass%value,                  &
                            lp_cosine_freq=pixel_size%value/low_pass_resolution%value,  &
                            lp_cosine_falloff=pixel_size%value/low_pass_falloff%value,  &
                            hp_cosine=high_pass%value,                  &
                            hp_cosine_freq=pixel_size%value/high_pass_resolution%value,  &
                            hp_cosine_falloff=pixel_size%value/high_pass_falloff%value,  &
                            print_filter_shape=.true.)
        call vol%BackwardFFT()
        call vol%writetodisk(output_file)
    endif

    call input_file%close()
    call output_file%printinfo()
    call output_file%close()

end subroutine mode_bfactor_slave

end module mode_bfactor
