!>  \brief  Mode to bin or unbin images, volumes, frealix parameter files or boxer coordinates files.
module mode_bin
    use modes
    use rtps
    implicit none
    ! Declare mode object
    type(mode)  ::  bin
    ! Declare RTPs
    type(rtp_char), target  ::  input_filename  =       &
        rtp_char(   value='input.mrc',                  &
                    keyword='INPUT_FILENAME',           &
                    description='Name of input file, to be binned. Can be image or frealix alignment file')

    type(rtp_char), target  ::  output_filename =       &
        rtp_char(   value='output.mrc',                 &
                    keyword='OUTPUT_FILENAME',          &
                    description='Name of output file')

    type(rtp_intg), target  ::  binning_factor  =       &
        rtp_intg(   value=4,                            &
                    keyword='BINNING_FACTOR',           &
                    description='Images or alignment parameters will be binned/decimated/coarsened '//  &
                    'by this factor in each dimension')

    type(rtp_logi), target  ::  unbin           =       &
        rtp_logi(   value=.false.,                      &
                    keyword='UNBIN',                    &
                    description='If true, the images or parameters will be unbinned, not binned')

    type(rtp_logi), target  ::  preserve_center  =      &
        rtp_logi(   value=.true.,                       &
                    keyword='PRESERVE_CENTER',          &
                    description='Ensure that the origin is still centered. Actual binning '//   &
                                '(real-space averaging of neighbouring pixels) would not do this.')
contains

subroutine mode_bin_init_and_run
    implicit none
    ! Initialise the mode
    call bin%init(  command='BIN',              &
                    subroutine=mode_bin_slave,  &
                    short_description='Bin or unbin images, volumes, frealix alignment files or boxer coordinates files')
    ! Add RTPs to the mode
    call bin%add_rtp(input_filename)
    call bin%add_rtp(output_filename)
    call bin%add_rtp(binning_factor)
    call bin%add_rtp(unbin)
    !call bin%add_rtp(preserve_center)
    ! Run the mode
    call bin%run()
end subroutine mode_bin_init_and_run

subroutine mode_bin_slave
    use images
    use imagefiles
    use parsers, only : image_filename, extension_from_filename
    use frealix_parameters
    use coos, only : waypoint, boxer_db, boxer_db_read_from_disk, boxer_db_bin, boxer_db_write_to_disk
    use ui, only : interval_print
    implicit none
    ! Variables
    type(image)     ::  volin,volout
    type(boxer_db)  ::  dbin,dbout
    type(imagefile) ::  imgfile, imgfileout
    integer         ::  current_loc, stack_size
    real            ::  pixel_size_old, pixel_size_new

    ! Start work
    ! If the input filename has an image extension (mrc,spi,hed or img), we treat it as an image/volume
    if (image_filename(input_filename%value)) then
        write(*,'(a)') '**info(mode_bin): input file is image file'
        call imgfile%init(input_filename%value)
        pixel_size_old = imgfile%GetPixelSize()
        if (unbin%value) then
            pixel_size_new = pixel_size_old / binning_factor%value
        else
            pixel_size_new = pixel_size_old * binning_factor%value
        endif
        call imgfile%PrintInfo()
        call imgfileout%init(output_filename%value,pixel_size=pixel_size_new)
        stack_size = imgfile%GetStackSize()
        if (imgfile%GetDimension(2) .eq. imgfile%GetDimension(1) .and. imgfile%GetDimension(1) .eq. imgfile%GetDimension(3)) then
            stack_size = 1
        endif
        do current_loc = 1,stack_size
            if (interval_print(current_loc,stack_size)) then
                write(*,'(2(A,i0))') '**info(mode_bin): working on location ', &
                                        current_loc, ' of ', stack_size
            endif
            call volin%ReadFromDisk(imgfile,wanted_location=current_loc,read_volume_not_slice=(STACK_SIZE.eq.1))
            if (unbin%value) then
                call volout%Allocate(dims=volin%GetLogicalDimensions()*binning_factor%value)
            else
                call volout%Allocate(dims=volin%GetLogicalDimensions()/binning_factor%value)
            endif
            call volin%ForwardFFT()
            call volin%ClipInto(volout)
            call volout%BackwardFFT()
            !call volin%BinTo(volout,binning_factor%value,unbin=unbin%value,preserve_center=preserve_center%value)
            call volout%WriteToDisk(imgfileout,wanted_location=current_loc)
        enddo
        call imgfile%close()
        call imgfileout%PrintInfo()
        call imgfileout%close()
    ! Maybe the input file was a boxer "database" (list of coordinates)
    else if (extension_from_filename(input_filename%value) .eq. 'box') then
        write(*,'(a)') '**info(mode_bin): input file is Boxer coordinates file'
        call boxer_db_read_from_disk(dbin,input_filename%value)
        call boxer_db_bin(dbin,dbout,binning_factor%value,unbin=unbin%value)
        call boxer_db_write_to_disk(dbout,output_filename%value)
    ! Otherwise, we treat the file as a parameter file
    else
        write(*,'(a)') '**info(mode_bin): input file is frealix alignment file'
        call frealix2_par_bin(input_filename%value,output_filename%value,binning_factor%value,unbin=unbin%value)
    endif

end subroutine mode_bin_slave

end module mode_bin
