!>  \brief  Given a set of waypoints file, recompute the sigma (estimate noise variance) and occupancy values for each waypoint
module mode_calc_occ
    use modes
    use rtps
    use Globals
    implicit none

    type(mode)              ::  calc_occ

    type(rtp_char), target  ::  input_ali_filenames  =               &
        rtp_char(   value='waypoints_r??.ali',                  &
                    keyword='input_ali_filenames',           &
                    description='Filename pattern for input ali filenames')

    type(rtp_char), target  ::  output_ali_filenames  =               &
        rtp_char(   value='waypoints_out_r??.ali',                  &
                    keyword='output_ali_filenames',           &
                    description='Filename pattern for output ali filenames')

    type(rtp_intg), target  ::  number_of_classes  =               &
        rtp_intg(   value=1,                  &
                    keyword='number_of_classes',           &
                    description='Number of classes / ali files')

contains

    subroutine mode_calc_occ_init_and_run()
        implicit none
        call calc_occ%init(command='calc_occ',subroutine=mode_calc_occ_slave,short_description='Compute sigma and occupancy')

        ! Add runtime parameters (RTPs) to the mode
        call calc_occ%add_rtp(input_ali_filenames)
        call calc_occ%add_rtp(number_of_classes)
        call calc_occ%add_rtp(output_ali_filenames)

        ! Run the mode
        call calc_occ%run()
    end subroutine mode_calc_occ_init_and_run

    subroutine mode_calc_occ_slave()
        use parsers, only : filename_from_wild
        use UsefulFunctions, only : FileExists
        use StringManipulations, only  : IntegerToString
        use filaments_base
        use frealix_parameters
        implicit none
        ! Private variables
        integer                         ::  counter_class, counter_fil, counter_wp
        character(len=filename_max_len) ::  ali_filename
        type(Filament), allocatable     ::  filaments(:,:), fils_tmp(:)
        real(kind=8),   allocatable     ::  average_probability(:) ! per class
        real                            ::  weighted_average_probability
        real                            ::  noise_sigma
        integer,        allocatable     ::  number_of_wps(:)
        ! Start work

        ! Load all data into memory
        do counter_class=1,number_of_classes%value
            ! Get filename for ali file
            call filename_from_wild(input_ali_filenames%value,counter_class,ali_filename)
            ! Check it exists
            if (.not. FileExists(ali_filename)) then
                call this_program%TerminateWithFatalError('mode_calc_occ','File does not exist: '//trim(adjustl(ali_filename)))
            endif
            ! Get filaments
            call filaments_from_frealix_par_noinit(fils_tmp,ali_filename,pixel_size=1.0)
            ! We may need to allocate
            if (counter_class .eq. 1) allocate(filaments(size(fils_tmp),number_of_classes%value))
            ! Check number of filaments checks out
            if (size(filaments,1) .ne. size(fils_tmp)) then
                call this_program%TerminateWithFatalError('mode_calc_occ', &
                    'Unexpected number of filaments in ali file '//trim(adjustl(ali_filename)))
            endif
            ! Copy into an array to keep all filaments for all classes
            do counter_fil=1,size(filaments,1)
                filaments(counter_fil,counter_class) = fils_tmp(counter_fil)
            enddo
        enddo


        ! Do some sanity checks
        do counter_fil=1,size(filaments,1)
            if (any(filaments(counter_fil,1)%id .ne. filaments(counter_fil,:)%id)) then
                call this_program%TerminateWithFatalError('mode_calc_occ',&
                            'The filaments are not in the same order in all ali files.')
            endif
            if (any(filaments(counter_fil,1)%nwp .ne. filaments(counter_fil,:)%nwp)) then
                write(*,*) filaments(counter_fil,:)%nwp
                call this_program%TerminateWithFatalError('mode_calc_occ','Filament '// &
                                        IntegerToString(filaments(counter_fil,1)%id)//&
                                        ' does not have same number of waypoints in all ali files')
            endif
        enddo


        ! For each class, work out the average probability
        allocate(number_of_wps(number_of_classes%value))
        allocate(average_probability(number_of_classes%value))
        number_of_wps = 0
        do counter_class=1,number_of_classes%value
            do counter_fil=1,size(filaments,1)
                do counter_wp=1,filaments(counter_fil,counter_class)%nwp
                    number_of_wps(counter_class) = number_of_wps(counter_class) + 1
                    average_probability(counter_class) = average_probability(counter_class) &
                                                       + exp(filaments(counter_fil,counter_class)%wp(counter_wp)%logp)
                enddo
            enddo
            average_probability(counter_class) = average_probability(counter_class) / real(number_of_wps(counter_class),kind=8)
        enddo

        ! Compute occupancy & noise sigma per waypoint
        do counter_fil=1,size(filaments,1)
            do counter_wp=1,filaments(counter_fil,1)%nwp
                weighted_average_probability = 0.0
                do counter_class=1,number_of_classes%value
                    weighted_average_probability = weighted_average_probability + average_probability(counter_class) &
                                                * exp(filaments(counter_fil,counter_class)%wp(counter_wp)%logp)
                enddo
                ! Now compute occupancy
                do counter_class=1,number_of_classes%value
                    filaments(counter_fil,counter_class)%wp(counter_wp)%occupancy = average_probability(counter_class) &
                                                    * exp(filaments(counter_fil,counter_class)%wp(counter_wp)%logp) &
                                                    / weighted_average_probability

                enddo
                ! Now compute a new estimate for the noise sigma
                noise_sigma = 0.0
                do counter_class=1,number_of_classes%value
                    noise_sigma = noise_sigma + filaments(counter_fil,counter_class)%wp(counter_wp)%noise_sigma &
                                                    * filaments(counter_fil,counter_class)%wp(counter_wp)%occupancy
                enddo
                do counter_class=1,number_of_classes%value
                    filaments(counter_fil,counter_class)%wp(counter_wp)%noise_sigma = noise_sigma
                enddo
            enddo
        enddo

        ! Write out the data
        do counter_class=1,number_of_classes%value
            ! Get filename for ali file
            call filename_from_wild(output_ali_filenames%value,counter_class,ali_filename)
            ! Write out
            fils_tmp = filaments(:,counter_class)
            call filaments_to_frealix_par(fils_tmp,ali_filename)
        enddo

    end subroutine mode_calc_occ_slave

end module mode_calc_occ
