!>  \brief  Estimate the persistence length by measuring mean tangent correlations
module mode_estimate_persistence_length
    use modes
    use rtps
    implicit none

    type(mode)  ::  estimate_pl

    type(rtp_char), target  ::  input_filename = &
                        rtp_char(   value = 'waypoints.ali',    &
                                    keyword = 'INPUT_FILENAME', &
                                    description = 'Filename for input waypoints')

    type(rtp_real), target  ::  pixel_size = &
                        rtp_real(   value = 0.0,    &
                                    keyword = 'PIXEL_SIZE',  &
                                    description = 'Pixel size (Angstroms)')

    type(rtp_logi), target  ::  ignore_fal_wps = &
                        rtp_logi(   value = .false.,    &
                                    keyword = 'IGNORE_FAL_WPS',  &
                                    description = 'Ignore the first and last waypoints of each filament')

    type(rtp_real), target  ::  maximum_filament_length = &
                        rtp_real(   value = 0.0,    &
                                    keyword = 'MAXIMUM_FILAMENT_LENGTH',  &
                                    description = 'Ignore filamnents longer than this length (in Angstroms)')

contains

subroutine mode_estimate_pl_init_and_run
    implicit none
    ! Intialise the mode
    call estimate_pl%init(command='ESTIMATE_PERSISTENCE_LENGTH',subroutine=mode_estimate_pl_slave,  &
                                short_description='Estimate the persistence length by measuring tangent correlations')

    ! Add RTPs to the mode
    call estimate_pl%add_rtp(input_filename)
    call estimate_pl%add_rtp(pixel_size)
    call estimate_pl%add_rtp(ignore_fal_wps)
    call estimate_pl%add_rtp(maximum_filament_length)

    ! Run the mode
    call estimate_pl%run()
end subroutine mode_estimate_pl_init_and_run

subroutine mode_estimate_pl_slave
    use coos, only : waypoints_add_disorder
    use filaments_base
    use filaments_path, only : filament_get_path_pixel_steps
    use filaments_geometry, only : filament_arc_length
    use filaments_parametrisation, only : filament_parametrisation_update
    use frealix_parameters
    use runtime_parameters, only : terminate
    implicit none
    ! Private variables
    type(filament), allocatable ::  fils(:)
    type(waypoint), allocatable ::  wp(:)
    integer                     ::  current_fil
    real                        ::  max_fil_length
    real(kind=8),   parameter   ::  s_min = 200.0D0
    real                        ::  s_max
    integer                     ::  num_intervals
    real(kind=8),   allocatable ::  ave(:),sq_ave(:)
    real,           allocatable ::  s(:)
    integer,        allocatable ::  n(:), nf(:)
    integer                     ::  i, current_wp,wp_from,wp_to
    real(kind=8)                ::  D
    ! Start work

    ! Parse the input waypoints file
    call filaments_from_frealix_par_noinit(fils,input_filename%value,pixel_size%value)

    ! Loop over filaments to initialise them and find the longest one
    max_fil_length = 0.0
    do current_fil=1,size(fils)
        ! Initialise the filament
        wp = fils(current_fil)%wp
        call fils(current_fil)%Allocate(size(wp))
        fils(current_fil)%wp = wp
        call filament_parametrisation_update(fils(current_fil))
        ! Print out information for this filament
        call fils(current_fil)%PrintInfo()
        ! Is this the longest filament we've seen so far?
        if (filament_arc_length(fils(current_fil),ignore_fal_wps%value) .gt. max_fil_length) then
            max_fil_length = filament_arc_length(fils(current_fil),ignore_fal_wps%value)
        endif
    enddo
    write(*,'(a,F0.1,a)') '**info(mode_estimate_persistence_length): longest filament is ', max_fil_length , ' pixels'

    ! If we were given a maximum filament length by the user, heed it
    if (maximum_filament_length%set_by_user) then
        max_fil_length = min(max_fil_length,maximum_filament_length%value/pixel_size%value)
        write(*,'(a,F0.1,a)') '**info(mode_estimate_persistence_length): will only use filaments up to ', &
                                                                        max_fil_length, ' pixels long'
    endif

    ! Allocate arrays for our measurements
    s_max = aint(max_fil_length / s_min) * s_min
    num_intervals = int((s_max - s_min) / s_min) + 1
    allocate(s(num_intervals),ave(num_intervals),sq_ave(num_intervals),n(num_intervals),nf(num_intervals))
    write(*,'(a,3(a,F0.1),a,i0,a)') '**info(mode_estimate_persistence_length): ', &
                                    'will measure tangent correlations with separations from ',&
                                    s_min,' to ', s_max , ' in ', s_min , ' increments (', &
                                    num_intervals, ' interval lengths)'

    ! Initialise the arrays
    s      = 0.0 ! length of this interval (in pixels)
    ave    = 0.0D0
    sq_ave = 0.0D0
    n      = 0  ! number of measurements that contributed to this interval's stats
    nf     = 0  ! number of filaments that contributed to this interval's stats

    ! Setup the s array - the length of each interval
    do i=1,num_intervals
        s(i) = s_min*i
    enddo

    ! Collect the data
    do current_fil=1,size(fils)
        if (maximum_filament_length%set_by_user) then
            if (filament_arc_length(fils(current_fil),ignore_fal_wps%value) .gt. max_fil_length) then
                write(*,'(a,2(i0,a))') '**info(mode_estimate_persistence_length): skipping filament ', current_fil, ' of ', &
                                                                                   size(fils), ' because it is too long'
                cycle
            endif
        endif
        write(*,'(a,i0,a,i0)') '**info(mode_estimate_persistence_length): working on filament ', current_fil, ' of ', size(fils)

        ! Get waypoints at regular intervals along the filament
        call filament_get_path_pixel_steps( fils(current_fil),s_min,wp,  &
                                            ignore_fal_wps=ignore_fal_wps%value,include_finish_wp=.false.)
        ! We should have at least 2 waypoints, if not something went wrong
        if (size(wp) .lt. 2) then
            write(*,'(a)') '**error(mode_estimate_persistence_length): less than 2 waypoints found'
            call terminate('Fatal error in MODE_ESTIMATE_PERSISTENCE_LENGTH. Failed sanity check.')
        endif

        ! This filament will contribute to all these interval's stats
        nf(1:size(wp)-1) = nf(1:size(wp)-1) + 1

        ! Loop over the intervals
        do i=1,size(wp)-1
            ! Loop over the waypoints
            do current_wp=1,size(wp)-I,i
                wp_from = current_wp
                wp_to   = wp_from + i
                ! Compute the dot product
                d = dot_product(wp(wp_from)%dcoo,wp(wp_to)%dcoo)
                ! Accumulate data
                n(i)      = n(i) + 1
                ave(i)    = ave(i) + d
                sq_ave(i) = sq_ave(i) + d**2
            enddo
        enddo

    enddo

    ! Print out results
    write(*,'(a)') '**info(mode_estimate_persistence_length): tangent correlation statistics follow'
    write(*,'(5(A17,1x))') 'Interval (pixels)', 'Num. measurements', 'Num. filaments', 'Sample mean', 'Sample variance'
    do i=1,num_intervals
        ave(i)    = ave(i) / n(i)
        sq_ave(i) = sq_ave(i) / n(i)
        write(*,'(F17.1,1x,2(i17,1x),1x,2(G19.5,1x))') s(i), n(i), nf(i), ave(i), sq_ave(i)-AVE(i)**2
    enddo

    ! All done
    write(*,'(a)') '**info(mode_estimate_persistence_length): all done.'

end subroutine mode_estimate_pl_slave


end module mode_estimate_persistence_length
