!>  \brief  Given a model 3d, refine way point parameters. The input and outputs are frealix2 parameter files.
!!
!!  Note: this mode can be used to just do a reconstruction without refinement, by setting the "refine?" flag in the input parameter file
!!  to f for each waypoint.
!!
module mode_filament_refine
!dec$ optimize:0
    use modes
    use rtps

    use minimisers, only : ds_max_iterations


    implicit none
    type(mode) :: filament_refine

    type(rtp_char), target  ::  input_par_filename  =               &
        rtp_char(   value='waypoints_old.ali',                  &
                    keyword='INPUT_PAR_FILENAME',           &
                    description='')


    type(rtp_char), target  ::  output_par_filename  =               &
        rtp_char(   value='waypoints.ali',                  &
                    keyword='OUTPUT_PAR_FILENAME',           &
                    description='')

    type(rtp_char), target  ::  output_par_filename_subunits =  &
        rtp_char(   value='none',                       &
                    keyword='OUTPUT_PAR_FILENAME_SUBUNITS', &
                    description='Unless ''none'', filename for output alignment parameter file, '//&
                                'with one line per helical asymetric unit')


    type(rtp_char), target  ::  model_3d_filename  =               &
        rtp_char(   value='model3d.mrc',                  &
                    keyword='MODEL_3D_FILENAME',           &
                    description='3D reconstruction to be used as reference, or model, during refinement')

    type(rtp_char), target  ::  model_3d_fsc_filename  =               &
        rtp_char(   value='model3d_fsc.txt',                  &
                    keyword='MODEL_3D_FSC_FILENAME',           &
                    description='FSC corresponding to the model 3d reconstruction. '//&
                                'This should be output by Frealix at the end of the previous round.')


    type(rtp_char), target  ::  output_3d_filename  =               &
        rtp_char(   value='model3d_out.mrc',                  &
                    keyword='OUTPUT_3D_FILENAME',           &
                    description='Filename for new 3d reconstruction following refinement')


    type(rtp_char), target  ::  ctftilt_files  =               &
        rtp_char(   value='none',                  &
                    keyword='CTFTILT_FILES',           &
                    description='Filename pattern (with ? characters replacing index number) for files '//&
                                'containing ctftilt output. Either this or ctffind_files must be specified.')


    type(rtp_char), target  ::  ctffind_files  =               &
        rtp_char(   value='none',                  &
                    keyword='CTFFIND_FILES',           &
                    description='Filename pattern (with ? characters replacing index number) for files '//&
                                'containing ctffind (<3.6.0) output or single filename for output from '//&
                                'CTFFIND (>=3.6.0). Either this or ctftilt_files must be specified')


    type(rtp_char), target  ::  film_files  =               &
        rtp_char(   value='film_????.mrc',                  &
                    keyword='FILM_FILES',           &
                    description='Filename pattern (with ? characters replacing index number) for micrographs '//&
                                'or single filename for stack of micrographs')

    type(rtp_intg), target  ::  number_of_frames_per_micrograph  =               &
        rtp_intg(   value=1,                  &
                    keyword='number_of_frames_per_micrograph',           &
                    description='If the input stack contains frames from dose-fractionated exposures, give the'//&
                                'number of frames. If not, give 1.')

    type(rtp_intg), target  ::  first_frame_for_reconstruction  =               &
        rtp_intg(   value=1,                  &
                    keyword='first_frame_for_reconstruction',           &
                    description='If the input stack contains frames from dose-fractionated exposures, give the'//&
                                'number of the first frame to include in the 3D reconstruction. If not, give 1.')


    type(rtp_intg), target  ::  last_frame_for_reconstruction  =               &
        rtp_intg(   value=1,                  &
                    keyword='last_frame_for_reconstruction',           &
                    description='If the input stack contains frames from dose-fractionated exposures, give the'//&
                                'number of the last frame to include in the 3D reconstruction. If not, give 1.')

    type(rtp_intg), target  ::  first_frame_for_refinement  =               &
        rtp_intg(   value=1,                  &
                    keyword='first_frame_for_refinement',           &
                    description='If the input stack contains frames from dose-fractionated exposures, give the'//&
                                'number of the first frame to include in the refinement. If not, give 1.')

    type(rtp_intg), target  ::  last_frame_for_refinement  =               &
        rtp_intg(   value=1,                  &
                    keyword='last_frame_for_refinement',           &
                    description='If the input stack contains frames from dose-fractionated exposures, give the'//&
                                'number of the last frame to include in the refinement. If not, give 1.')

    type(rtp_char), target  ::  unblur_result_filename  =               &
        rtp_char(   value='unblur_output.txt',                  &
                    keyword='unblur_result_filename',           &
                    description='Filename of results from unblur. Should contain one line per micrograph, with '//&
                                'relative shifts of frames')

    type(rtp_logi), target  ::  recompute_subframe_alignments  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='recompute_subframe_alignments',           &
                    description='For each waypoint, recompute local alignment parameters between movie frames, '//&
                                'starting from unblur results')

    type(rtp_real), target  ::  box_size_for_subframe_alignment = &
        rtp_real(   value=1000.0, &
                    keyword='box_size_for_subframe_alignment', &
                    description='Size of the box to extract from movie frames around each waypoint when '//&
                                'computing local alignments. Angstroms.')


    type(rtp_logi), target  ::  apply_dose_filter = &
        rtp_logi(   value=.false., &
                    keyword='apply_dose_filter', &
                    description='When summing frames from dose-fractionated exposures, filter each frames '//&
                                'as a function of the expected radiation damage from electron exposure')

    type(rtp_real), target  ::  exposure_per_frame = &
        rtp_real(   value=1.0, &
                    keyword='exposure_per_frame', &
                    description='For each frame of dose-fractionated exposures, the number of electrons '//&
                                'incident on the specimen per square Angstroms')

    type(rtp_char), target  ::  matching_projections_files  =               &
        rtp_char(   value='match_proj_??????.mrc',                  &
                    keyword='MATCHING_PROJECTIONS_FILES',           &
                    description='')


    type(rtp_char), target  ::  filament_trace_files  =               &
        rtp_char(   value='trace_??????_??????.box',                  &
                    keyword='FILAMENT_TRACE_FILES',           &
                    description='')


    type(rtp_real), target  ::  pixel_size  =               &
        rtp_real(   value=9.99,                  &
                    keyword='PIXEL_SIZE',           &
                    description='')


    type(rtp_real), target  ::  acceleration_voltage =  &
        rtp_real(   value=300.0,                        &
                    keyword='acceleration_voltage',     &
                    description='in kV')

    type(rtp_real), target  ::  spherical_aberration =  &
        rtp_real(   value=2.7,                          &
                    keyword='spherical_aberration',     &
                    description='in mm')

    type(rtp_real), target  ::  amplitude_contrast =    &
        rtp_real(   value=0.07,                         &
                    keyword='amplitude_contrast',       &
                    description='Fraction of amplitude contrast')

    type(rtp_real), target  ::  twist_per_subunit  =               &
        rtp_real(   value=0.0,                  &
                    keyword='TWIST_PER_SUBUNIT',           &
                    description='In degrees, the twist per subunit of the 1-start helix')


    type(rtp_real), target  ::  rise_per_subunit  =               &
        rtp_real(   value=0.0,                  &
                    keyword='RISE_PER_SUBUNIT',           &
                    description='In Angstroems, the rise per subunit of the 1-start helix')


    type(rtp_intg), target  ::  symmetry_axial  =               &
        rtp_intg(   value=1,                  &
                    keyword='SYMMETRY_AXIAL',           &
                    description='Rotational symmetry redundancy about the helical axis')


    type(rtp_intg), target  ::  symmetry_perp  =               &
        rtp_intg(   value=1,                  &
                    keyword='SYMMETRY_PERP',           &
                    description='Rotational symmetry redundancy perpendicular to the helical axis (e.g. 2 for dna)')


    type(rtp_logi), target  ::  wps_are_crossovers  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='WPS_ARE_CROSSOVERS',           &
                    description='The waypoints are crossovers (typically of an amyloid beta fibril or actin filament)')


    type(rtp_logi), target  ::  fal_wps_are_crossovers  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='FAL_WPS_ARE_CROSSOVERS',           &
                    description='The first and last waypoints are also crossovers. Typically, it''s recommended extra'//&
                                ' fal waypoints are picked before and after the first and last waypoint. '//&
                                'Therefore, this would typically false.')


    type(rtp_logi), target  ::  amyloid_beta  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='AMYLOID_BETA',           &
                    description='The filaments are amyloid beta fibrils')


    type(rtp_logi), target  ::  microtubule  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='MICROTUBULE',           &
                    description='The filaments are microtubules')


    type(rtp_intg), target  ::  mt_pf_number  =               &
        rtp_intg(   value=0,                  &
                    keyword='MT_PF_NUMBER',           &
                    description='For microtubules only: number of protofilaments (e.g. 13)')


    type(rtp_intg), target  ::  mt_rise  =               &
        rtp_intg(   value=0,                  &
                    keyword='MT_RISE',           &
                    description='For microtubules only: how much further along the filament (distance measured '//&
                                'in tubulin subunits) the end of a MT_PF_NUMBER-element "washer" is compared to '//&
                                'its beginning. The 13-pf microtobules have a value of 3.')


    type(rtp_real), target  ::  helix_max_radius  =               &
        rtp_real(   value=0.0,                  &
                    keyword='HELIX_MAX_RADIUS',           &
                    description='Maximum outer radius of the helix, in Angstroms')


    type(rtp_intg), target  ::  num_helical_starts  =               &
        rtp_intg(   value=1,                  &
                    keyword='NUM_HELICAL_STARTS',           &
                    description='Number of helical starts. The helical parameters given above must describe one '//&
                                'of these starts. Not including symmetry (e.g. could be describing a 1-start helix '//&
                                'with an additional 2-fold axial symmetry).')

    type(rtp_logi), target  ::  stack  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='STACK',           &
                    description='The assembly is a stack rather than a true helix')


    type(rtp_real), target  ::  asym_unit_mw  =               &
        rtp_real(   value=0.0,                  &
                    keyword='ASYM_UNIT_MW',           &
                    description='Expected molecular weight (Da) of the asymetric unit')

    type(rtp_logi), target  ::  cross_beta_filter  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='CROSS_BETA_FILTER',           &
                    description='Filter the data in Fourier space to use the cross-beta reflection at 4.8A')


    type(rtp_logi), target  ::  use_conjugate_gradient  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='USE_CONJUGATE_GRADIENT',           &
                    description='Use conjugate gradient for refinement')


    type(rtp_logi), target  ::  use_differential_evolution  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='USE_DIFFERENTIAL_EVOLUTION',           &
                    description='Use differential evolution for refinement')

    type(rtp_logi), target  ::  use_brute_force  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='USE_BRUTE_FORCE',           &
                    description='Use brute force (systematic search) for refinement')

    type(rtp_logi), target  ::  use_downhill_simplex  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='USE_DOWNHILL_SIMPLEX',           &
                    description='Use downhill simplex for refinement')



    type(rtp_logi), target  ::  add_fal_wps  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='ADD_FAL_WPS',           &
                    description='Add one waypoint at each end of the filament, by spline extrapolation.')


    type(rtp_logi), target  ::  fal_wps_in_score  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='FAL_WPS_IN_SCORE',           &
                    description='The parts of the filament between the first and second as well as between the'//&
                                ' penultimate and last waypoints should be included in the scoring function')


    type(rtp_logi), target  ::  fal_wps_in_threed  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='FAL_WPS_IN_THREED',           &
                    description='The parts of the filament between the first and second as well as between the '//&
                                'penultimate and last waypoints should be included in the 3d reconstruction')


    type(rtp_real), target  ::  force_wp_to_wp_distance  =               &
        rtp_real(   value=0.0,                  &
                    keyword='FORCE_WP_TO_WP_DISTANCE',           &
                    description='Each filament''s waypoints will be recalculated so that they are spaced by '//&
                                'this distance (in Angstroms). If set to 0.0, the waypoints will be left as in input.')

    type(rtp_real), target  ::  force_wp_to_wp_phi  =               &
        rtp_real(   value=0.0,                  &
                    keyword='FORCE_WP_TO_WP_PHI',           &
                    description='Each filament''s waypoints will be recalculated so that they are spaced by '//&
                                'this angle (in degrees). If set to 0.0, the waypoints will be left as in input.')


    type(rtp_logi), target  ::  reverse_film_contrast  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='REVERSE_FILM_CONTRAST',           &
                    description='Reverse micrograph contrast internally. It is assumed that protein density is '//&
                                'darker (lower values) than the background. If it isn''t, set this to T')


    type(rtp_logi), target  ::  calc_3d_reconstruction  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='CALC_3D_RECONSTRUCTION',           &
                    description='Calculate a 3d reconstruction using the refined waypoint parameters.')


    type(rtp_logi), target  ::  calc_fsc  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='CALC_FSC',           &
                    description='Calculate an fsc between two half-dataset reconstructions.')


    type(rtp_logi), target  ::  estimate_ssnr  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='ESTIMATE_SSNR',           &
                    description='Estimate the ssnr and use it during Wiener filtering of the reconstruction.')


!    type(rtp_logi), target  ::  estimate_phi_angles  =               &
!        rtp_logi(   value=.FALSE.,                  &
!                    keyword='ESTIMATE_PHI_ANGLES',           &
!                    description='Set this to true if the Phi (twist) angles are unknown. Except in the case of crossovers')


    type(rtp_logi), target  ::  extra_ccf_per_wp  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='EXTRA_CCF_PER_WP',           &
                    description='')


    type(rtp_logi), target  ::  bfactor_during_3d_reconstruction  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='BFACTOR_DURING_3D_RECONSTRUCTION',           &
                    description='Apply a B-factor filter to images when they are added to 3d reconstructions, '//&
                                'to reflect their quality (score)')

   type(rtp_logi), target  ::  apply_drift_filter  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='apply_drift_filter',           &
                    description='Apply a filter to images when they are inserted in the 3D reconstruction, to account '//&
                                'for any drift which was occuring when the micrograph frame was being recorded. Also '//&
                                'apply this filter when extracting images from micrograph frames during refinement.')

   type(rtp_logi), target  ::  bfactor_per_subunit  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='BFACTOR_PER_SUBUNIT',           &
                    description='During reconstruction, b factor is computer per subunit rather than per filament')


    type(rtp_real), target  ::  model_3d_filter_bfactor  =               &
        rtp_real(   value=0.0,                  &
                    keyword='MODEL_3D_FILTER_BFACTOR',           &
                    description='Apply this b factor (in A^2) to the reference before it is used by the scoring function')


    type(rtp_real), target  ::  average_score  =               &
        rtp_real(   value=0.0,                  &
                    keyword='AVERAGE_SCORE',           &
                    description='Average score of filaments after refinement. Used to calculate a Bfactor to '//&
                                'apply to images during reconstruction.')


    type(rtp_logi), target  ::  finalise_3d_reconstruction  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='FINALISE_3D_RECONSTRUCTION',           &
                    description='Compute a finalised 3d volume. If .FALSE., the unweighted 3d and a CTF^2 volume '//&
                                'will be dumped to the disk and can be finalised later.')

    type(rtp_logi), target  ::  reconstruct_amplitudes  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='RECONSTRUCT_AMPLITUDES',           &
                    description='Compute a reconstruction using only amplitudes.')


    type(rtp_logi), target  ::  apply_fom_filter  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='APPLY_FOM_FILTER',           &
                    description='Apply a Figure-Of-Merit (fom) filter to the final reconstruction (equivalent to '//&
                                'FCREF in Frealign; see Henderson & Rosenthal, 2003)')


    type(rtp_real), target  ::  min_score_fil_to_3d  =               &
        rtp_real(   value=-1.0,                  &
                    keyword='MIN_SCORE_FIL_TO_3D',           &
                    description='Only filaments with a score greather than this will be included in the 3d reconstruction')


!    type(rtp_intg), target  ::  num_minim_rounds  =               &
!        rtp_intg(   value=1,                  &
!                    keyword='NUM_MINIM_ROUNDS',           &
!                    description='')


    type(rtp_logi), target  ::  polarity_check  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='POLARITY_CHECK',           &
                    description='Test both filament polarities (the two polarities are related by a reversal of '//&
                                'the order of waypoints, equivalent to a change of 180 degrees in out-of-plane tilt)')


    type(rtp_logi), target  ::  phi_polarity_check  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='PHI_POLARITY_CHECK',           &
                    description='Test both twist polarities of the filament (the two polarities are related by an '//&
                                'axial rotation of 180 degrees)')


    !type(rtp_logi), target  ::  real_space_projection  =               &
    !    rtp_logi(   value=.false.,                  &
    !                keyword='REAL_SPACE_PROJECTION',           &
    !                description='Use real space projection. If .FALSE., reciprocal space slice extraction is '//&
    !                            'used instead (which is faster).')


    type(rtp_logi), target  ::  refine_x  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='REFINE_X',           &
                    description='Refine X')


    type(rtp_logi), target  ::  refine_y  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='REFINE_Y',           &
                    description='Refine Y')


    type(rtp_logi), target  ::  refine_z  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='REFINE_Z',           &
                    description='Refine z (defocus)')


    type(rtp_logi), target  ::  refine_phi  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='REFINE_PHI',           &
                    description='Refine phi (twist; axial rotation)')


    type(rtp_logi), target  ::  refine_theta  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='REFINE_THETA',           &
                    description='Refine theta (out-of-plane rotation)')


    type(rtp_logi), target  ::  refine_psi  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='REFINE_PSI',           &
                    description='Refine psi (in-plane rotation)')


    type(rtp_logi), target  ::  refine_subunit_num  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='REFINE_SUBUNIT_NUM',           &
                    description='Refine subunit number (axial positioning)')

    type(rtp_logi), target  ::  refine_helical_twist  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='refine_helical_twist',           &
                    description='Refine the helical twist (when doing single-particle refinement)')

    type(rtp_logi), target  ::  refine_helical_rise  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='refine_helical_rise',           &
                    description='Refine the helical rise (when doing single-particle refinement)')

    type(rtp_logi), target  ::  use_hermite_splines =    &
       rtp_logi(    value=.true.,               &
                    keyword='USE_HERMITE_SPLINES',  &
                    description='Use Hermite splines, where the tangents (i.e. psi & theta) are used as contraints')


    type(rtp_logi), target  ::  rigid_x  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='RIGID_X',           &
                    description='X is rigid (per-filament rather than per-waypoint)')


    type(rtp_logi), target  ::  rigid_y  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='RIGID_Y',           &
                    description='Y is rigid (per-filament rather than per-waypoint)')


    type(rtp_logi), target  ::  rigid_z  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='RIGID_Z',           &
                    description='Z is rigid (per-filament rather than per-waypoint)')


    type(rtp_logi), target  ::  rigid_phi  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='RIGID_PHI',           &
                    description='Phi is rigid (per-filament rather than per-waypoint)')


    type(rtp_logi), target  ::  rigid_theta  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='RIGID_THETA',           &
                    description='Theta is rigid (per-filament rather than per-waypoint)')


    type(rtp_logi), target  ::  rigid_psi  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='RIGID_PSI',           &
                    description='Psi is rigid (per-filament rather than per-waypoint)')


    type(rtp_logi), target  ::  rigid_subunit_num  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='RIGID_SUBUNIT_NUM',           &
                    description='Subunit number (~ axial phasing) is rigid (per-filament rather than per-waypoint).'//&
                                ' Rigid implies constant rise per subunit.')


    type(rtp_logi), target  ::  hold_z_average  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='HOLD_Z_AVERAGE',           &
                    description='The average z value of each filament will be held constant')


    type(rtp_logi), target  ::  refine_subunit_num_increments  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='REFINE_SUBUNIT_NUM_INCREMENTS',           &
                    description='Refine wp-on-wp increment in subunit_num rather than actual subunit_num values')


    type(rtp_logi), target  ::  subunit_num_integer_changes_only  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='SUBUNIT_NUM_INTEGER_CHANGES_ONLY',           &
                    description='Only allow integer shifts in subunit_num values')

    type(rtp_logi), target  ::  fragmented_phi  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='FRAGMENTED_PHI',           &
                    description='Phi function is discontinuous, with breaks between waypoints')

    type(rtp_logi), target  ::  fragmented_subunit_num  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='FRAGMENTED_SUBUNIT_NUM',           &
                    description='Subunit_num function is discontinuous, with breaks between waypoints')

    type(rtp_logi), target  ::  constant_twist  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='CONSTANT_TWIST',           &
                    description='The twist per unit length is constant.')


    type(rtp_logi), target  ::  constant_rise_per_subunit  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='CONSTANT_RISE_PER_SUBUNIT',           &
                    description='The rise per subunit is constant. Note that non-cumulative axial disorder is not accounted for.')


    type(rtp_real), target  ::  search_half_range_x  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_HALF_RANGE_X',           &
                    description='Angstroms. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    type(rtp_real), target  ::  search_half_range_y  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_HALF_RANGE_Y',           &
                    description='Angstroms. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    type(rtp_real), target  ::  search_half_range_z  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_HALF_RANGE_Z',           &
                    description='Angstroms. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    type(rtp_real), target  ::  search_half_range_phi  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_HALF_RANGE_PHI',           &
                    description='Degrees. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    type(rtp_real), target  ::  search_half_range_theta  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_HALF_RANGE_THETA',           &
                    description='Degrees. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    type(rtp_real), target  ::  search_half_range_psi  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_HALF_RANGE_PSI',           &
                    description='Degrees. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    type(rtp_real), target  ::  search_half_range_subunit_num  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_HALF_RANGE_SUBUNIT_NUM',           &
                    description='Number of subunits. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')

    type(rtp_real), target  ::  search_half_range_helical_rise  =               &
        rtp_real(   value=0.0,                  &
                    keyword='search_half_range_helical_rise',           &
                    description='Angstroms. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')

    type(rtp_real), target  ::  search_half_range_helical_twist  =               &
        rtp_real(   value=0.0,                  &
                    keyword='search_half_range_helical_twist',           &
                    description='Degrees. If not set to 0.0, this will override the default search half-range '//&
                                '(which is based on the target resolution).')


    type(rtp_real), target  ::  search_step_size_x  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_STEP_SIZE_X',           &
                    description='Pixels. If not set to 0.0, this will override the default search size '//&
                                '(which is based on the target resolution).')


    type(rtp_real), target  ::  search_step_size_y  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_STEP_SIZE_Y',           &
                    description='Pixels. If not set to 0.0, this will override the default search size '//&
                                '(which is based on the target resolution).')


    type(rtp_real), target  ::  search_step_size_z  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_STEP_SIZE_Z',           &
                    description='Pixels. If not set to 0.0, this will override the default search step size '//&
                                '(which is based on the target resolution).')


    type(rtp_real), target  ::  search_step_size_phi  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_STEP_SIZE_PHI',           &
                    description='Degrees. If not set to 0.0, this will override the default search step size '//&
                                '(which is based on the target resolution).')


    type(rtp_real), target  ::  search_step_size_theta  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_STEP_SIZE_THETA',           &
                    description='Degrees. If not set to 0.0, this will override the default search step size '//&
                                '(which is based on the target resolution).')


    type(rtp_real), target  ::  search_step_size_psi  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_STEP_SIZE_PSI',           &
                    description='Degrees. If not set to 0.0, this will override the default search step size '//&
                                '(which is based on the target resolution).')


    type(rtp_real), target  ::  search_step_size_subunit_num  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SEARCH_STEP_SIZE_SUBUNIT_NUM',           &
                    description='Number of subunits. If not set to 0.0, this will override the default search step size '//&
                                '(which is based on the target resolution).')

    type(rtp_real), target  ::  search_step_size_helical_rise  =               &
        rtp_real(   value=0.0,                  &
                    keyword='search_step_size_helical_rise',           &
                    description='Angstroms. If not set to 0.0, this will override the default search step size '//&
                                '(which is based on the target resolution).')

    type(rtp_real), target  ::  search_step_size_helical_twist  =               &
        rtp_real(   value=0.0,                  &
                    keyword='search_step_size_helical_twist',           &
                    description='Degrees. If not set to 0.0, this will override the default search step size '//&
                                '(which is based on the target resolution).')

    type(rtp_logi), target  ::  search_filament_tilt  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='SEARCH_FILAMENT_TILT',           &
                    description='Do a brute-force search for filament out-of-plane tilts '//&
                                '(whole filament is tilted as a rigid body)')


    type(rtp_logi), target  ::  subtract_fil_from_ref  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='SUBTRACT_FIL_FROM_REF',           &
                    description='Subtract the original state of each filament from the 3d reference model, '//&
                                'to reduce bias to original state, before the filament is processed.')

    type(rtp_logi), target  ::  single_filament_refinement  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='SINGLE_FILAMENT_REFINEMENT',           &
                    description='Refine each filament against itself.')


    type(rtp_intg), target  ::  num_minim_runs  =               &
        rtp_intg(   value=1,                  &
                    keyword='NUM_MINIM_RUNS',           &
                    description='Number of minimisations to run')


    type(rtp_logi), target  ::  ds_random_start_conf  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='DS_RANDOM_START_CONF',           &
                    description='Randomise the starting state of the dowhill simplex minimiser')


    type(rtp_logi), target  ::  minim_random_start_pos  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='MINIM_RANDOM_START_POS',           &
                    description='Randomise the starting position of the minimiser (with '//&
                                'Gaussian noise, sigma = search half-range)')


    type(rtp_intg), target  ::  ds_max_number_of_iterations  =               &
        rtp_intg(   value=ds_max_iterations,                  &
                    keyword='DS_MAX_NUMBER_OF_ITERATIONS',           &
                    description='Maximum number of iterations of the Downhill Simplex per run')


    type(rtp_intg), target  ::  path_oversampling  =               &
        rtp_intg(   value=1,                  &
                    keyword='PATH_OVERSAMPLING',           &
                    description='On average, each pixel will be used this number of times in the scoring function. 1 by default.')


    type(rtp_logi), target  ::  skip_refinement  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='SKIP_REFINEMENT',           &
                    description='Do not do any refinement.')


    type(rtp_logi), target  ::  model_3d_available  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='MODEL_3D_AVAILABLE',           &
                    description='A model 3d reconstruction is available as input.')


    type(rtp_logi), target  ::  refine_wps_simultaneously  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='REFINE_WPS_SIMULTANEOUSLY',           &
                    description='Refine all waypoints together, rather than sequentially. Default is simultaneous refinement.')


    type(rtp_logi), target  ::  search_subunit_num  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='SEARCH_SUBUNIT_NUM',           &
                    description='Do a brute-force search in integer steps around the current subunit number values.')


    type(rtp_logi), target  ::  search_mt_seam  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='SEARCH_MT_SEAM',           &
                    description='For microtubules only: search for the seam position')


    type(rtp_logi), target  ::  average_model_cylindrically  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='AVERAGE_MODEL_CYLINDRICALLY',           &
                    description='Average cylindrically the input 3d model before refinement starts')


    !type(rtp_logi), target  ::  refine_in_fourier_space  =               &
    !    rtp_logi(   value=.false.,                  &
    !                keyword='REFINE_IN_FOURIER_SPACE',           &
    !                description='Refine the filament model''s Fourier transform rather than the model itself')


    type(rtp_logi), target  ::  refine_normal_modes  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='REFINE_NORMAL_MODES',           &
                    description='Refine the filament model''s normal mode amplitudes rather than the model itself')


    type(rtp_logi), target  ::  straighten_filament_before_refinement  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='STRAIGHTEN_FIL_BEFORE_REFINEMENT',           &
                    description='Straighten the filament (and impose constant helical parameters) before refinement')


    type(rtp_logi), target  ::  move_waypoints_to_nearest_repeat  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='MOVE_WAYPOINTS_TO_NEAREST_REPEAT',           &
                    description='On output, move (refined) waypoints to the nearest helical subunit')

    type(rtp_logi), target  ::  average_z_to_ctftilt_plane  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='AVERAGE_Z_TO_CTFTILT_PLANE',           &
                    description='Move each filament in the z direction so that its average defocus matches '//&
                                'the expectation from ctftilt information')


    type(rtp_logi), target  ::  use_weighted_correlation  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='USE_WEIGHTED_CORRELATION',           &
                    description='Use a weighted (rather than linear) correlation function. See Stewart & Grigorieff (2004).')


    !type(rtp_logi), target  ::  use_average_of_cubed_segment_scores  =               &
    !    rtp_logi(   value=.false.,                  &
    !                keyword='USE_AVE_OF_CUBED_SEGMENT_SCORES',           &
    !                description='When in full-filament mode: if false, full-filament normalisation is used')


    type(rtp_logi), target  ::  use_curvature_restraint  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='USE_CURVATURE_RESTRAINT',           &
                    description='Use the curvature of each filament as a restraint on its refinement')

    type(rtp_logi), target  ::  use_torsion_restraint  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='USE_TORSION_RESTRAINT',           &
                    description='Use the torsion of each filament as a restraint on its refinement')

    type(rtp_logi), target  ::  use_stretch_restraint  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='USE_STRETCH_RESTRAINT',           &
                    description='Use the stretch of each filament as a restraint on its refinement')


    type(rtp_real), target  ::  persistence_length  =               &
        rtp_real(   value=0.0,                  &
                    keyword='PERSISTENCE_LENGTH',           &
                    description='Persistence length of the filament (Angstroms)')


    type(rtp_real), target  ::  torsion_persistence_length  =               &
        rtp_real(   value=0.0,                  &
                    keyword='TORSION_PERSISTENCE_LENGTH',           &
                    description='Torsional persistence length (Angstroms)')


    type(rtp_real), target  ::  stretch_persistence_length  =               &
        rtp_real(   value=0.0,                  &
                    keyword='STRETCH_PERSISTENCE_LENGTH',           &
                    description='Stretching persistence length (Angstroms)')


    type(rtp_real), target  ::  target_resolution  =               &
        rtp_real(   value=25.0,                  &
                    keyword='TARGET_RESOLUTION',           &
                    description='Target resolution for the refinements (Angstroms)')


    type(rtp_real), target  ::  low_resolution_cuton  =               &
        rtp_real(   value=1000.0,                  &
                    keyword='LOW_RESOLUTION_CUTON',           &
                    description='Lowest resolution to be used during refinements (Angstroms)')


    type(rtp_real), target  ::  high_resolution_cutoff  =               &
        rtp_real(   value=2.0,                  &
                    keyword='HIGH_RESOLUTION_CUTOFF',           &
                    description='Highest resolution to be used during refinements (Angstroms)')

    type(rtp_logi), target  ::  wps_as_sps  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='WPS_AS_SPS',           &
                    description='Treat filament waypoints (WPs) as independent single particles (SPs)')


    type(rtp_real), target  ::  sp_segment_length  =               &
        rtp_real(   value=700.0,                  &
                    keyword='SP_SEGMENT_LENGTH',           &
                    description='In Angstroms. Single particles are defined as segments of this length along '//&
                                'the filament. Only used if wps_as_sps is .TRUE..')


    !type(rtp_logi), target  ::  cache_projections  =               &
    !    rtp_logi(   value=.false.,                  &
    !                keyword='CACHE_PROJECTIONS',           &
    !                description='Cache 2d projections of 3d models. Could lead to significant time savings, '//&
    !                            'especially when using expensive interpolation schemes. However, could also '//&
    !                            'lead to loss in quality of alignments.')


    type(rtp_logi), target  ::  single_filament_reconstructions  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='SINGLE_FILAMENT_RECONSTRUCTIONS',           &
                    description='For each individual filament, write to disk the central slice of its 3d reconstruction')

    type(rtp_logi), target  ::  rotate_to_register_crossovers  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='ROTATE_TO_REGISTER_CROSSOVERS',           &
                    description='Rotate each filament about its axis such that its crossovers correspond to phi=0.0,180.0, etc.')

    type(rtp_logi), target  ::  align_map_to_helical_lattice  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='align_map_to_helical_lattice',           &
                    description='Align the 3D map so that most density is away from the borders of the helical unit cells')

    type(rtp_logi), target  ::  project_individual_subunits  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='project_individual_subunits',           &
                    description='Project individual subunits, one at a time in real space, '//&
                                'rather than segments of symmetrised model 3Ds')


    type(rtp_logi), target  ::  write_matching_projections  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='WRITE_MATCHING_PROJECTIONS',           &
                    description='Write stacks of images & matching projections to disk after refinement')


    type(rtp_logi), target  ::  write_filament_traces  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='WRITE_FILAMENT_TRACES',           &
                    description='Write coordinates (boxer) files showing waypoint positions and filament traces')


    type(rtp_char), target  ::  interpol_function  =               &
        rtp_char(   value='LINEAR',                  &
                    keyword='INTERPOL_FUNCTION',           &
                    description='The interpolation function to use')


    type(rtp_char), target  ::  window_function  =               &
        rtp_char(   value='RECT',                  &
                    keyword='WINDOW_FUNCTION',           &
                    description='The windowing function to use')


    type(rtp_intg), target  ::  window_width  =               &
        rtp_intg(   value=2,                  &
                    keyword='WINDOW_WIDTH',           &
                    description='Window width for interpolation (kernel support)')

	type(rtp_logi), target ::   sp_to_ff = 	&
		rtp_logi(	value=.false.,			&
					keyword='SP_TO_FF',		&
					description='Input parameters are from sp analysis, but ff analsysis should be done')

    type(rtp_logi), target ::   resample_data_for_refinement = 	&
		rtp_logi(	value=.true.,			&
					keyword='RESAMPLE_DATA_FOR_REFINEMENT',		&
					description='Resample the micrograph and model 3d so that the new Nyquist is just '//&
                                'beyond the high frequency used in refinement. This makes refinements much faster.')

    type(rtp_real), target  ::  resolution_of_model_3d  =               &
        rtp_real(   value=0.0,                  &
                    keyword='RESOLUTION_OF_MODEL_3D',           &
                    description='In Angstroms. This is normally computed automatically from the input fsc curve.')

    type(rtp_logi), target ::   boost_low_frequency_contrast = 	&
		rtp_logi(	value=.false.,			&
					keyword='boost_low_frequency_contrast',		&
					description='Leave low frequencies (until the first peak of the ctf) intact, rather than  '//&
                                'scaling them by the ctf value. May help in early stages of alignments. To be '//&
                                'superseded in future version of Frealix by ssnr weighting of the scoring function')

    type(rtp_real), target  ::  min_twist_per_subunit_for_rec  =               &
        rtp_real(   value=-179.9999,                  &
                    keyword='min_twist_per_subunit_for_rec',           &
                    description='In degrees. Waypoints or parts of filaments with twists per subunit lower than '//&
                                'this will not be included in the 3D reconstruction')

    type(rtp_real), target  ::  max_twist_per_subunit_for_rec  =               &
        rtp_real(   value=179.9999,                  &
                    keyword='max_twist_per_subunit_for_rec',           &
                    description='In degrees. Waypoints or parts of filaments with twists per subunit greater than '//&
                                'this will not be included in the 3D reconstruction')

    type(rtp_real), target  ::  min_rise_per_subunit_for_rec  =               &
        rtp_real(   value=0.0,                  &
                    keyword='min_rise_per_subunit_for_rec',           &
                    description='In Angstroms. Waypoints or parts of filaments with rises per subunit lower than '//&
                                'this will not be included in the 3D reconstruction')

    type(rtp_real), target  ::  max_rise_per_subunit_for_rec  =               &
        rtp_real(   value=huge(1.0)*0.0001,                  &
                    keyword='max_rise_per_subunit_for_rec',           &
                    description='In Angstroms. Waypoints or parts of filaments with rises per subunit greater than '//&
                                'this will not be included in the 3D reconstruction')


contains

    subroutine mode_filament_refine_init_and_run
    implicit none
    ! Intialise the mode
    call filament_refine%init(command='FILAMENT_REFINE',subroutine=mode_filament_refine_slave, &
                                short_description='Refine filament waypoint parameters')

    ! Add RTPs to the mode
    call filament_refine%add_rtp(input_par_filename)
    call filament_refine%add_rtp(output_par_filename)
    call filament_refine%add_rtp(output_par_filename_subunits)
    call filament_refine%add_rtp(model_3d_filename)
    call filament_refine%add_rtp(model_3d_fsc_filename)
    call filament_refine%add_rtp(output_3d_filename)
    call filament_refine%add_rtp(ctftilt_files)
    call filament_refine%add_rtp(ctffind_files)
    call filament_refine%add_rtp(film_files)
    call filament_refine%add_rtp(number_of_frames_per_micrograph)
    call filament_refine%add_rtp(first_frame_for_reconstruction)
    call filament_refine%add_rtp(last_frame_for_reconstruction)
    call filament_refine%add_rtp(first_frame_for_refinement)
    call filament_refine%add_rtp(last_frame_for_refinement)
    call filament_refine%add_rtp(recompute_subframe_alignments)
    call filament_refine%add_rtp(unblur_result_filename)
    call filament_refine%add_rtp(box_size_for_subframe_alignment)
    call filament_refine%add_rtp(apply_dose_filter)
    call filament_refine%add_rtp(exposure_per_frame)
    call filament_refine%add_rtp(matching_projections_files)
    call filament_refine%add_rtp(filament_trace_files)
    call filament_refine%add_rtp(pixel_size)
    call filament_refine%add_rtp(acceleration_voltage)
    call filament_refine%add_rtp(spherical_aberration)
    call filament_refine%add_rtp(amplitude_contrast)
    call filament_refine%add_rtp(twist_per_subunit)
    call filament_refine%add_rtp(rise_per_subunit)
    call filament_refine%add_rtp(symmetry_axial)
    call filament_refine%add_rtp(symmetry_perp)
    call filament_refine%add_rtp(wps_are_crossovers)
    call filament_refine%add_rtp(fal_wps_are_crossovers)
    call filament_refine%add_rtp(amyloid_beta)
    call filament_refine%add_rtp(microtubule)
    call filament_refine%add_rtp(mt_pf_number)
    call filament_refine%add_rtp(mt_rise)
    call filament_refine%add_rtp(helix_max_radius)
    !call filament_refine%add_rtp(num_helical_starts)
    call filament_refine%add_rtp(stack)
    call filament_refine%add_rtp(asym_unit_mw)
    call filament_refine%add_rtp(cross_beta_filter)
    call filament_refine%add_rtp(use_conjugate_gradient)
    call filament_refine%add_rtp(use_differential_evolution)
    call filament_refine%add_rtp(use_brute_force)
    call filament_refine%add_rtp(use_downhill_simplex)
    call filament_refine%add_rtp(add_fal_wps)
    call filament_refine%add_rtp(fal_wps_in_score)
    call filament_refine%add_rtp(fal_wps_in_threed)
    call filament_refine%add_rtp(force_wp_to_wp_distance)
    call filament_refine%add_rtp(force_wp_to_wp_phi)
    call filament_refine%add_rtp(reverse_film_contrast)
    call filament_refine%add_rtp(calc_3d_reconstruction)
    call filament_refine%add_rtp(calc_fsc)
    call filament_refine%add_rtp(estimate_ssnr)
    call filament_refine%add_rtp(bfactor_during_3d_reconstruction)
    call filament_refine%add_rtp(apply_drift_filter)
    call filament_refine%add_rtp(bfactor_per_subunit)
    call filament_refine%add_rtp(model_3d_filter_bfactor)
    call filament_refine%add_rtp(average_score)
    call filament_refine%add_rtp(finalise_3d_reconstruction)
    call filament_refine%add_rtp(reconstruct_amplitudes)
    !call filament_refine%add_rtp(apply_fom_filter)
    call filament_refine%add_rtp(min_score_fil_to_3d)
    call filament_refine%add_rtp(polarity_check)
    call filament_refine%add_rtp(phi_polarity_check)
    !call filament_refine%add_rtp(real_space_projection)
    call filament_refine%add_rtp(refine_x)
    call filament_refine%add_rtp(refine_y)
    call filament_refine%add_rtp(refine_z)
    call filament_refine%add_rtp(refine_phi)
    call filament_refine%add_rtp(refine_theta)
    call filament_refine%add_rtp(refine_psi)
    call filament_refine%add_rtp(refine_subunit_num)
    call filament_refine%add_rtp(refine_helical_rise)
    call filament_refine%add_rtp(refine_helical_twist)
    call filament_refine%add_rtp(use_hermite_splines)
    call filament_refine%add_rtp(rigid_x)
    call filament_refine%add_rtp(rigid_y)
    call filament_refine%add_rtp(rigid_z)
    call filament_refine%add_rtp(rigid_phi)
    call filament_refine%add_rtp(rigid_theta)
    call filament_refine%add_rtp(rigid_psi)
    call filament_refine%add_rtp(rigid_subunit_num)
    call filament_refine%add_rtp(hold_z_average)
    call filament_refine%add_rtp(refine_subunit_num_increments)
    call filament_refine%add_rtp(subunit_num_integer_changes_only)
    call filament_refine%add_rtp(fragmented_phi)
    call filament_refine%add_rtp(fragmented_subunit_num)
    call filament_refine%add_rtp(constant_twist)
    call filament_refine%add_rtp(constant_rise_per_subunit)
    call filament_refine%add_rtp(search_half_range_x)
    call filament_refine%add_rtp(search_half_range_y)
    call filament_refine%add_rtp(search_half_range_z)
    call filament_refine%add_rtp(search_half_range_phi)
    call filament_refine%add_rtp(search_half_range_theta)
    call filament_refine%add_rtp(search_half_range_psi)
    call filament_refine%add_rtp(search_half_range_subunit_num)
    call filament_refine%add_rtp(search_half_range_helical_rise)
    call filament_refine%add_rtp(search_half_range_helical_twist)
    call filament_refine%add_rtp(search_step_size_x)
    call filament_refine%add_rtp(search_step_size_y)
    call filament_refine%add_rtp(search_step_size_z)
    call filament_refine%add_rtp(search_step_size_phi)
    call filament_refine%add_rtp(search_step_size_theta)
    call filament_refine%add_rtp(search_step_size_psi)
    call filament_refine%add_rtp(search_step_size_subunit_num)
    call filament_refine%add_rtp(search_step_size_helical_rise)
    call filament_refine%add_rtp(search_step_size_helical_twist)
    call filament_refine%add_rtp(search_filament_tilt)
    call filament_refine%add_rtp(subtract_fil_from_ref)
    call filament_refine%add_rtp(single_filament_refinement)
    call filament_refine%add_rtp(num_minim_runs)
    call filament_refine%add_rtp(ds_random_start_conf)
    call filament_refine%add_rtp(minim_random_start_pos)
    call filament_refine%add_rtp(ds_max_number_of_iterations)
    call filament_refine%add_rtp(path_oversampling)
    call filament_refine%add_rtp(skip_refinement)
    call filament_refine%add_rtp(model_3d_available)
    call filament_refine%add_rtp(refine_wps_simultaneously)
    call filament_refine%add_rtp(search_subunit_num)
    call filament_refine%add_rtp(search_mt_seam)
    call filament_refine%add_rtp(average_model_cylindrically)
    !call filament_refine%add_rtp(refine_in_fourier_space)
    call filament_refine%add_rtp(refine_normal_modes)
    call filament_refine%add_rtp(straighten_filament_before_refinement)
    call filament_refine%add_rtp(move_waypoints_to_nearest_repeat)
    call filament_refine%add_rtp(average_z_to_ctftilt_plane)
    call filament_refine%add_rtp(use_weighted_correlation)
    !call filament_refine%add_rtp(use_average_of_cubed_segment_scores)
    call filament_refine%add_rtp(use_curvature_restraint)
    call filament_refine%add_rtp(use_torsion_restraint)
    call filament_refine%add_rtp(use_stretch_restraint)
    call filament_refine%add_rtp(persistence_length)
    call filament_refine%add_rtp(torsion_persistence_length)
    call filament_refine%add_rtp(stretch_persistence_length)
    call filament_refine%add_rtp(target_resolution)
    call filament_refine%add_rtp(low_resolution_cuton)
    call filament_refine%add_rtp(high_resolution_cutoff)
    call filament_refine%add_rtp(wps_as_sps)
    call filament_refine%add_rtp(sp_segment_length)
    !call filament_refine%add_rtp(cache_projections)
    call filament_refine%add_rtp(single_filament_reconstructions)
    call filament_refine%add_rtp(rotate_to_register_crossovers)
    call filament_refine%add_rtp(align_map_to_helical_lattice)
    call filament_refine%add_rtp(project_individual_subunits)
    call filament_refine%add_rtp(write_matching_projections)
    call filament_refine%add_rtp(write_filament_traces)
    !call filament_refine%add_rtp(interpol_function)
    !call filament_refine%add_rtp(window_function)
    !call filament_refine%add_rtp(window_width)
	call filament_refine%add_rtp(sp_to_ff)
	call filament_refine%add_rtp(resample_data_for_refinement)
	call filament_refine%add_rtp(resolution_of_model_3d)
	call filament_refine%add_rtp(boost_low_frequency_contrast)
	call filament_refine%add_rtp(min_twist_per_subunit_for_rec)
	call filament_refine%add_rtp(max_twist_per_subunit_for_rec)
    call filament_refine%add_rtp(min_rise_per_subunit_for_rec)
	call filament_refine%add_rtp(max_rise_per_subunit_for_rec)

    ! Run the mode
    call filament_refine%run()
end subroutine mode_filament_refine_init_and_run


subroutine mode_filament_refine_slave
    use runtime_parameters, only : terminate, psf_radius, psf_radius_pixels, pbc, filename_len, default_falloff_pixels
    use UsefulFunctions, only : IsOdd
    use parsers, only : string_equal, file_number_from_filename, filename_from_wild, filename_add_suffix, file_exists
    use parsers, only : filename_remove_extension, filename_contains_qmarks
    use frealix_parameters
    use procfs, only : print_memory_usage_stats
    ! "Science" modules
    use helices, only : helical_symmetry, helical_symmetry_init, helical_symmetry_print_info, helical_symmetry_print_extra_info
    use filaments_base, only : filament
    use filaments_model_base, only : filament_print_extra_info
    use filaments_model_modify, only : filament_shift_all_wps, filament_flip_polarity, filament_add_fal_wps, filament_straighten
    use filaments_model_modify, only : filament_move_wps_to_nearest_subunit
    use filaments, only : filament_fix_polarity_inconsistencies
    use filaments, only : filament_resample_waypoints, filament_print_diff
    use filaments, only : filament_shift_wps_to_obtain_3d_z_phi_shift
    use fmcs_evaluation, only : fmc_eval, fmc_model3d_prep
    use filaments_microtubules, only : image_impose_microtubule_symmetry
    use filaments_minimisation, only : filament_minimise_deviations_from_canonical_helix
    use images
    use images_arithm, only : cyl_mask
    use fmcs, only : fmc_bf_tilt, fmc_maximise, fmc_bf_find_mt_seam, fmc_bf_subunit_num_intg_refine
    use fmcs_base, only : filament_model_comparison, fmc_box_size, fmc_threed_size
    use fmcs_base, only : fmc_init, fmc_new_filament, fmc_update_fil_threed, fmc_print_info, fmc_deallocate
    use filaments, only : filament_boxer_db_to_disk, filament_calc_threed, filament_trace_to_disk_csv
    use interpolation, only : interpolation_scheme, interpolation_scheme_init
    use threeds_base, only : threed, threed_init, threed_deallocate
    use images_helices, only : image_impose_helical_symmetry_new, image_helix_mask, find_angle_of_crossover
    use images_helices, only : image_align_to_helical_lattice
    use point_symmetries, only : point_symmetry_symbol
    use threeds, only : threed_eval_reconstruction, threed_add, threed_write_to_disk, threed_assignment
    use fscs, only : compute_fsc
    use images_comparison
    use ImageFiles
    use filaments, only : filament_change_pixel_size
    use Curves
    use NumericTextFiles
    use fmcs_base, only : ssnr_weighting, linear_correlation, cubed_correlation
    use filaments_base, only : filament_print_subframe_shifts
    use parsers, only : grep
    use StringManipulations, only : StringsAreEqual

    implicit none
    !
    ! Private variables
    !
    type(helical_symmetry)                  ::  helix_average                   !   Average helical parameters
    type(helical_symmetry)                  ::  helix_min_for_rec               !   Minimum helical parameters for inclusion into 3D reconstruction
    type(helical_symmetry)                  ::  helix_max_for_rec               !   Maximum helical parameters for inclusion into 3D reconstruction
    logical                                 ::  use_ctftilt                     !   Whether we'll be using ctftilt or ctffind information
    type(filament), allocatable, target     ::  fils(:)                         !   Array of filament objects
    type(filament), allocatable             ::  fil_states(:,:,:)               !   Temporary array of filament states (1st dim: polarity, 2dim: phi polarity, 3dim: before/after refinement)
    type(filament), allocatable             ::  fils_final(:)                   !   Array of final states of filament objects
    type(filament)                          ::  temp_fil
    type(image), target                     ::  model_3dvol                     !   Input 3d model
    type(image)                             ::  new_3dvol(2)                    !   Output 3d reconstruction volume(s) (one for each half-reconstruction, if we're doing fsc)
    type(image)                             ::  new_3dvol_ampl(2)               !   Output 3d reconstruction volume(s) containing only amplitudes
    type(threed)                            ::  new_3drec                       !   3d reconstruction object
    type(threed)                            ::  current_3drec                   !   Temporary 3d reconstruction object for individual filaments
    type(threed), target                    ::  ref_3drec                       !   Input reference 3d reconstruction
    !type(threed)                            ::  tmp_3drec
    type(image)                             ::  current_3dvol                   !   Temporary finalised 3d reconstruction
    type(image)                             ::  current_3dvol_ampl
    type(image)                             ::  current_3dvol_slice             !   Central xy slice from temporary finalised 3d reconstruction
    integer                                 ::  ierr                            !   Status
    type(filament_model_comparison)         ::  fmc                             !   An object containing relevant information re. the comparison
    type(image),    allocatable, target     ::  current_micrograph_frames(:)    !   The images of the frames of micrograph being currently analysed
    logical                                 ::  new_film                        !   Whether we need to read a new film image from disk
    integer                                 ::  current_filament                !   Filament we're currently looping over
    logical                                 ::  refine_params(9)                !   Which parameters (x,y,z,Phi,Theta,Psi) will we refine?
    type(interpolation_scheme)              ::  interpol_scheme                 !   Interpolation scheme to be used
    integer                                 ::  i                               !   Loop counter
    character(len=filename_len)             ::  match_proj_filename             !   Filename for matching projection stacks
    integer                                 ::  current_film_number             !   Number of film to which current filament belongs
    character(len=filename_len)             ::  filament_trace_filename         !   Filename for coordinates of waypoints and trace
    character(len=filename_len)             ::  temp_filename                   !   Temporary file name
    character(len=filename_len)             ::  output_3d_ampl_filename         !   Filename for output amplitude reconstruction
    real                                    ::  film_hp_param                   !   High-pass filter parameter for film images
    real,           allocatable             ::  foushell(:)                     !   Fourier shell correlation
    real                                    ::  search_half_range_override(9)   !   Override of search half-range by the user
    real                                    ::  search_step_size_override(9)    !   Override of search step size by the user
    logical                                 ::  rrefine_normal_modes(9)         !   Some parameters may be refined via their normal modes
    real                                    ::  dummy                           !   For debugging
    logical,        parameter               ::  debug   =   .false.             !   Debugging flag
    integer                                 ::  crop_dim(3)
    ! Scores
    real,   allocatable                     ::  score(:,:,:)                    !   For each polarity (1st dim), for each phi polarity (2nd dim) and before/after minimisation (values 1 or 2 3rd dim), the score
    real                                    ::  score_tmp
    ! Polarity checks
    integer                                 ::  num_polarities_to_test          !   2 if we check both possible polarities of each filament
    integer                                 ::  num_phi_polarities_to_test      !   2 if we check both phi polarities, 1 otherwise
    integer                                 ::  current_polarity                !   Polarity of filament we're currently checking
    integer                                 ::  current_phi_polarity            !
    integer                                 ::  best_polarity   =   1           !   Keep track of best polarity for current filament
    integer                                 ::  best_phi_polarity(2)    =   1   !   For each filament polarity, the best phi polarity
    ! Imagefile objects
    type(imagefile)                         ::  film_file
    type(imagefile)                         ::  threed_file
    !type(imagefile)                         ::  slices_file
    !
    logical                                 ::  rigid_params(9)
    logical                                 ::  hold_param_average(9)
    logical                                 ::  refine_param_increments(9)
    logical                                 ::  integer_changes_only(9)
    type(point_symmetry_symbol)             ::  sym_symbol
    real                                    ::  current_filament_crossover_angle
    real                                    ::  estimated_resolution_of_model_3d      !<  In angstroms; This will need to be worked out from previous fsc or user
    type(Curve)                             ::  model_3d_fsc
    integer                                 ::  current_frame
    type(NumericTextFile)                   ::  unblur_result_file
    real,               allocatable         ::  unblur_shifts(:,:,:)            !<  First dimension: 1,2,3 for x,y,z; Second dimension: frame number; Third dimension: micrograph number
    real,               allocatable         ::  temp_x(:), temp_y(:)
    integer                                 ::  number_of_micrographs
    type(ImageFile)                         ::  temp_image_file
    integer                                 ::  type_of_scoring_function
    !
    ! Runtime binning
    !
    real,           parameter               ::  target_psize_factor         =   1.2          !<  Input data will be resampled such that the new Nyquist is 2.0 times the target frequency
    real                                    ::  refinement_pixel_size, reconstruction_pixel_size
    real                                    ::  target_refinement_pixel_size
    type(helical_symmetry)                  ::  helix_average_refinement
    integer                                 ::  model_3dvol_padding_dimensions(3)
    integer                                 ::  model_3dvol_refinement_dimensions(3)
    real                                    ::  actual_scaling_factor
    integer                                 ::  film_padding_dimensions(3)
    integer                                 ::  film_refinement_dimensions(3)
    integer                                 ::  film_original_dimensions(3)
    ! Alignment of 3d model to helical lattice
    real                                    ::  model3d_phi_shift
    real                                    ::  model3d_z_shift

    !
    ! Get variable values from control file
    !


    !
    ! Start work
    !

    ! Memory footprint diag
    if (debug) then
        call print_memory_usage_stats()
    endif

    ! This is a helical mode
    sym_symbol%symbol = 'H'
    if (symmetry_perp%value .eq. 2) then
        sym_symbol%symbol(1:1) = 'D'
    else
        sym_symbol%symbol(1:1) = 'C'
    endif
    if (symmetry_axial%value .gt. 1) then
        if (symmetry_axial%value .gt. 99) then
            call this_program%TerminateWithFatalError('mode_filament_refine','Axial symmetry > 99 not supported')
        else if (symmetry_axial%value .gt. 9) then
            write(sym_symbol%symbol(2:3),'(i2)') symmetry_axial%value
        else
            write(sym_symbol%symbol(2:2),'(i1)') symmetry_axial%value
            sym_symbol%symbol(3:3) = ' '
        endif
    else
        sym_symbol%symbol(2:3) = '1 '
    endif
    if (symmetry_perp%value .eq. 2 .or. symmetry_axial%value .gt. 1) then
        write(*,'(2a)') '**info(mode_filament_refine): point symmetry: ', sym_symbol%symbol
    endif

    !
    ! Sanity check: target resolution cannot be lower than the Nyquist!
    !
    if (target_resolution%value .lt. pixel_size%value * 2.0) then
        write(*,'(a,F0.2,a,F0.3,a)')    '**error(mode_filament_refine): target resolution (', target_resolution%value, &
                                        ') is incompatible with pixel size (', pixel_size%value, ')'
        call terminate('Fatal error in MODE_FILAMENT_REFINE')
    endif

    !
    ! Now we know the pixel size, convert the psf_radius to psf_radius_pixels
    !
    if (this_program%IsVerbose() .or. debug) then
        write(*,'(a,F0.1,a)') '**info(mode_filament_refine): will take psf radius to be ', psf_radius, ' A'
    endif
    psf_radius_pixels = psf_radius / pixel_size%value

    !
    ! Estimated resolution of input model 3d
    !
    if (model_3d_fsc_filename%set_by_user .and. file_exists(model_3d_fsc_filename%value)) then
        call fsc_curve_read_from_disk(model_3d_fsc,model_3d_fsc_filename%value)
        estimated_resolution_of_model_3d = fsc_curve_last_frequency_with_fsc_above_threshold(model_3d_fsc,0.5) ! We use 0.5 to get an understimate of the resolution
        if (estimated_resolution_of_model_3d .gt. 0.0) then
            estimated_resolution_of_model_3d = 1.0e0/estimated_resolution_of_model_3d
        else
            estimated_resolution_of_model_3d = 20.0e0
            if (this_program%IsVerbose() .or. debug) then
                write(*,'(2a)') '**warning(mode_filament_refine): something went wrong when estimating the resolution from ', &
                                model_3d_fsc_filename%value
            endif
        endif
        if (this_program%IsVerbose() .or. debug) then
            write(*,'(a,f0.1,a)') '**info(mode_filament_refine): model 3d has a resolution of ~', &
                                    estimated_resolution_of_model_3d, 'A'
        endif
    else
        estimated_resolution_of_model_3d = 20.0e0
        write(*,'(a,f0.1)') '**info(mode_filament_refine): no input fsc curve. Will assume the model 3d has a resolution of ',  &
                        estimated_resolution_of_model_3d
    endif

    if (resolution_of_model_3d%set_by_user) then
        estimated_resolution_of_model_3d = resolution_of_model_3d%value
        if (this_program%IsVerbose() .or. debug) then
            write(*,'(a,f0.1,a)') '**info(mode_filament_refine): resolution of model 3d overidden to ', &
                                    estimated_resolution_of_model_3d, 'A'
        endif
    endif

    !
    ! Initialise the interpolation scheme
    !
    call interpolation_scheme_init(interpol_scheme,interpol_function%value,window_function%value,window_width%value)


    !
    ! Are we given ctffind files or ctftilt files?
    !
    use_ctftilt = .false.
    if (string_equal(ctftilt_files%value,'none')) then
        if (string_equal(ctffind_files%value,'none')) then
            write(*,'(a)') '**error(mode_filament_refine): neither ctffind nor ctftilt files were supplied'
            call terminate('Fatal error in MODE_FILAMENT_REFINE')
        else
            use_ctftilt = .false.
        endif
    else
        use_ctftilt = .true.
    endif
    ! Diagnostic
    if (.not. use_ctftilt .and. (this_program%IsVerbose() .or. debug)) then
        print '(a)', '**info(mode_filament_refine): no ctftilt files given, will use ctffind info'
    endif

    !
    ! Will we be checking the polarity of the filament?
    !
    if (polarity_check%value) then
        num_polarities_to_test = 2
    else
        num_polarities_to_test = 1
    endif

    !
    ! Will we be checking the phi polarity of the filament?
    !
    if (phi_polarity_check%value) then
        num_phi_polarities_to_test = 2
    else
        num_phi_polarities_to_test = 1
    endif

    !
    ! Which parameters will we refine
    !
    refine_params = [refine_x%value,refine_y%value,refine_z%value,  &
                    refine_phi%value,refine_theta%value,refine_psi%value, &
                    refine_subunit_num%value,refine_helical_rise%value,refine_helical_twist%value]

    ! Can only refine psi & theta when using Hermite splines
    if ((refine_psi%value .or. refine_theta%value) .and. .not. use_hermite_splines%value) then
        write(*,'(a)') '**error(mode_filament_refine): psi and theta can only be refined when Hermite splines are used'
        call terminate('Fatal error in MODE_FILAMENT_REFINE. Bad choice of parameters.')
    endif

    !
    ! Which parameters will be held rigid (per-filament rather than per-waypoint)
    !
    rigid_params = [rigid_x%value,rigid_y%value,rigid_z%value,  &
                    rigid_phi%value,rigid_theta%value,rigid_psi%value,  &
                    rigid_subunit_num%value,.false.,.false.]

    !
    ! Will we be holding the average value for any parameter constant?
    !
    hold_param_average = [.false.,.false.,hold_z_average%value,.false.,.false.,.false.,.false.,.false.,.false.]

    !
    ! Will we refine subunit_num wp-on-wp increments rather than absolute values?
    !
    refine_param_increments = (/ .false., .false., .false., .false., .false., .false., refine_subunit_num_increments%value, &
                                 .false., .false. /)

    !
    ! Will we allow only integer-valued changes during refinement?
    !
    integer_changes_only = (/ .false., .false., .false., .false., .false., .false., subunit_num_integer_changes_only%value, &
                              .false., .false. /)

    if (any(integer_changes_only)) then
        if (.not. skip_refinement%value) then
            if (.not. use_differential_evolution%value) then
                write(*,'(a)') '**error(mode_filament_refine): integer-only changes in subunit_num are only implemented '// &
                                'in conjunction with the differential evolution minimiser'
                call terminate('Fatal error in MODE_FILAMENT_REFINE. Bad choice of parameters.')
            endif
            if (refine_normal_modes%value) then
                write(*,'(a)') '**error(mode_filament_refine): integer-only changes in subunit_num are not '// &
                                'compatible with normal mode refinement'
                call terminate('Fatal error in MODE_FILAMENT_REFINE. Bad choice of parameters.')
            endif
        endif
    endif

    ! It doesn't make sense to refine the defocus unless we are aiming for high resolution
    if (refine_z%value .and. .not. hold_z_average%value) then
        if (target_resolution%value .gt. 5.0) then
            write(*,'(a,i0,a)') '**warning(mode_filament_refine): refining z without fixing the average z per filament amounts '// &
                            'to defocus refinement, but your target resolution is ', target_resolution%value, ', which may be '// &
                            'too low. You may want to use option HOLD_Z_AVERAGE'
        endif
    endif

    ! I don't think it is possible to do sequential refinement of waypoints while refining a rigid parameter (I'm not sure)
    if (any(refine_params .and. rigid_params) .and. .not. refine_wps_simultaneously%value) then
        write(*,'(a)') '**error(mode_filament_refine): i do not know how to refine waypoints sequentially '//&
                                                     'while also refining a rigid parameters. Not implemented.'
        call terminate('Fatal error in in MODE_FILAMENT_REFINE. Feature not implemented.')
    endif

    ! Normal modes refinement implies all waypoints are refined simultaneously
    if (.not. refine_wps_simultaneously%value .and. refine_normal_modes%value) then
        write(*,'(a)') '**error(mode_filament_refine): normal mode refinement requires that waypoints be refined simultaneously'
        call terminate('Fatal error in MODE_FILAMENT_REFINE. Bad choice of run-time parameter values.')
    endif


    !
    ! Check for incompatible settings of constant_rise_per_subunit%value and rigid_subunit_num%value
    !
    if (constant_rise_per_subunit%value .and. &
       .not. (rigid_subunit_num%value .or. fragmented_subunit_num%value .or. wps_as_sps%value)) then
        write(*,'(2a)') '**error(mode_filament_refine): you have constant_rise_per_subunit ',&
                        'set to true and rigid_subunit_num set to false. These settings are incompatible.'
        write(*,'(2a)') '**error(mode_filament_refine): if you want to keep a constant rise per subunit,',   &
                        ' you should set rigid_subunit_num to true.'
        call terminate('Fatal error in MODE_FILAMENT_REFINE')
    endif

    !
    ! Which parameters will we refine via their normal modes?
    !
    rrefine_normal_modes = .false.
    if (refine_normal_modes%value) then
        rrefine_normal_modes = .true.
        if (fragmented_subunit_num%value) then
            rrefine_normal_modes(7) = .false.
        endif
        if (fragmented_phi%value) then
            rrefine_normal_modes(4) = .false.
        endif
    endif

    !
    ! We need to know the average score if we're to do B-factor filtering during
    ! 3d reconstruction
    !
    if (bfactor_during_3d_reconstruction%value) then
        if (.not. average_score%set_by_user) then
            write(*,'(2a)') '**error(mode_filament_refine): average score not specified on input. ', &
                            'Needed for b factor calculation during 3d reconstruction.'
            call terminate('Fatal error in MODE_FILAMENT_REFINE')
        endif
        if (pbc .eq. 0.0) then
            write(*,'(a)') '**error(mode_filament_refine): pbc value cannot be 0.0. Please set it in the input parameter file.'
            write(*,'(a)') 'Typical pbc values may range from 5.0 (strong weighting) to 100.0 (very weak weighting).'
            call terminate('Fatal error in MODE_FILAMENT_REFINE')
        endif
    endif

    !
    ! How many micrographs in the input stack?
    !
    if (filename_contains_qmarks(film_files%value)) then
        number_of_micrographs = 0 ! unknown
    else
        ! we have a stack
        call temp_image_file%Init(film_files%value)
        number_of_micrographs =  temp_image_file%GetStackSize() / number_of_frames_per_micrograph%value
        call temp_image_file%Close()
    endif

    !
    ! If we're re-computing local subframe alignments, we need unblur results
    !
    if (unblur_result_filename%set_by_user) then
        ! Read in unblur results
        call unblur_result_file%Init(unblur_result_filename%value,OPEN_TO_READ)
        allocate(temp_x(unblur_result_file%GetNumberOfRecordsPerLine()),temp_y(unblur_result_file%GetNumberOfRecordsPerLine()))
        allocate(unblur_shifts(3,number_of_frames_per_micrograph%value,number_of_micrographs))
        if (this_program%IsVerbose() .or. debug) then
            write(*,'(a,i0,a)') '**info(mode_filament_refine): parsing results from unblur for ', number_of_micrographs, &
                                                        ' micrographs ... '
        endif
        if (number_of_frames_per_micrograph%value .lt. unblur_result_file%GetNumberOfRecordsPerLine()) then
            write(*,'(a,i0,a,i0,a,i0,a)') &
                            '**warning(mode_filament_refine): you specified ', number_of_frames_per_micrograph%value, &
                            ' frames, but the unblur result file describes ', unblur_result_file%GetNumberOfRecordsPerLine(), &
                            '. Will only use the first part of the unblur results (frames 1 to ', &
                            number_of_frames_per_micrograph%value, ').'
        endif
        do current_film_number=1,number_of_micrographs
            call unblur_result_file%ReadNextDataLine(temp_x)
            call unblur_result_file%ReadNextDataLine(temp_y)
            unblur_shifts(1,:,current_film_number) = temp_x(1:number_of_frames_per_micrograph%value)
            unblur_shifts(2,:,current_film_number) = temp_y(1:number_of_frames_per_micrograph%value)
            unblur_shifts(3,:,current_film_number) = 0.0
        enddo
        deallocate(temp_x, temp_y)

        ! Work out whether unblur resulta are in pixels or Angstroms (early versions stored shifts in pixels)
        ! If we read in Angstrom shifts, convert to pixels for internal consumption
        temp_filename = grep(unblur_result_filename%value,'Angstroms')
        if (StringsAreEqual(temp_filename,' ')) then
            ! Unblur shifts were in pixels
            if (debug) write(*,'(a)') '**debug(mode_filament_refine): read in unblur shifts in pixels'
        else
            if (debug) then
                write(*,'(a)') '**debug(mode_filament_refine): read in unblur shifts in Angstroms, converting to pixels internally'
            endif
            unblur_shifts(1:2,:,:) = unblur_shifts(1:2,:,:) / pixel_size%value
        endif
    else if (number_of_micrographs .gt. 0) then
        allocate(unblur_shifts(3,1,number_of_micrographs))
        unblur_shifts = 0.0
    else
        if (allocated(unblur_shifts)) deallocate(unblur_shifts)
    endif

    !
    ! Initialise the average helical parameters
    !
    call helical_symmetry_init( helix_average,rise_per_subunit%value/pixel_size%value,twist_per_subunit%value,    &
                                helix_max_radius%value/pixel_size%value, &
                                symmetry_axial%value,symmetry_perp%value,num_starts=num_helical_starts%value, &
                                stack=stack%value)
    write(*,'(a)',advance='no') '**info(mode_filament_refine): initialised helical symmetry operator: '
    call helical_symmetry_print_info(helix_average)
    if (this_program%IsVerbose() .or. debug) call helical_symmetry_print_extra_info(helix_average)

    !
    ! Initialise the min/max helical parameters
    !
    call helical_symmetry_init( helix_min_for_rec, &
                                min_rise_per_subunit_for_rec%value/pixel_size%value,min_twist_per_subunit_for_rec%value, &
                                helix_max_radius%value/pixel_size%value, &
                                symmetry_axial%value,symmetry_perp%value,num_starts=num_helical_starts%value, &
                                stack=stack%value, do_not_setup_unit_cell=.true.)
    call helical_symmetry_init( helix_max_for_rec, &
                                max_rise_per_subunit_for_rec%value/pixel_size%value,max_twist_per_subunit_for_rec%value, &
                                helix_max_radius%value/pixel_size%value, &
                                symmetry_axial%value,symmetry_perp%value,num_starts=num_helical_starts%value, &
                                stack=stack%value, do_not_setup_unit_cell=.true.)

    !
    ! Read in the frealix parameter file and intialise an array of filament objects
    !
    if (this_program%IsVerbose() .or. debug) then
        write(*,'(a)') '<parse_ali_file>'
        write(*,'(/3a)') '**info(mode_filament_refine): reading input parameter file (', &
                                                adjustl(trim(input_par_filename%value)),') and initialising filaments'
    endif
    if (use_ctftilt) then
        call filaments_from_frealix_par(    fils,input_par_filename%value,film_files%value,pixel_size%value,helix_average,  &
                                            ctftilt_filename_wild = ctftilt_files%value,                                    &
                                            wps_are_crossovers = wps_are_crossovers%value,                                  &
                                            fal_wps_are_crossovers = fal_wps_are_crossovers%value,                          &
                                            max_radius = helix_max_radius%value/pixel_size%value,                           &
                                            microtubule=microtubule%value,mt_pf_number=mt_pf_number%value,mt_rise = mt_rise%value,&
                                            rigid_params = rigid_params,                                                    &
                                            constant_rise_per_subunit = constant_rise_per_subunit%value,                    &
                                            constant_twist_per_length = constant_twist%value,                               &
                                            wps_as_sps = wps_as_sps%value,refine_params = refine_params,                    &
                                            persistence_length = persistence_length%value/pixel_size%value,                       &
                                            torsion_persistence_length = torsion_persistence_length%value/pixel_size%value,       &
                                            stretch_persistence_length = stretch_persistence_length%value/pixel_size%value,       &
                                            fal_wps_in_score = fal_wps_in_score%value,fal_wps_in_threed = fal_wps_in_threed%value,&
                                            refine_normal_modes = rrefine_normal_modes,                                     &
                                            average_z_to_ctftilt_plane = average_z_to_ctftilt_plane%value,                  &
                                            use_hermite_splines = use_hermite_splines%value,                                &
                                            fragmented_phi = fragmented_phi%value,                                          &
                                            fragmented_subunit_num = fragmented_subunit_num%value,                          &
                                            amyloid_beta=amyloid_beta%value,                                                &
                                            acceleration_voltage=acceleration_voltage%value,                                &
                                            spherical_aberration=spherical_aberration%value,                                &
                                            amplitude_contrast=amplitude_contrast%value,                                    &
                                            add_fal_wps=add_fal_wps%value,                                                  &
                                            unblur_shifts=unblur_shifts,                                                    &
                                            box_size_for_subframe_alignment=nint(box_size_for_subframe_alignment%value      &
                                                                                                        /pixel_size%value), &
                                            recompute_subframe_alignments=recompute_subframe_alignments%value,              &
                                            number_of_frames=number_of_frames_per_micrograph%value,                         &
                                            exposure_per_frame=exposure_per_frame%value,                                    &
                                            apply_dose_filter=apply_dose_filter%value,                                      &
                                            helix_min_for_rec=helix_min_for_rec,                                            &
                                            helix_max_for_rec=helix_max_for_rec                                             &
                                            )
    else
        call filaments_from_frealix_par(    fils,input_par_filename%value,film_files%value,pixel_size%value,helix_average,  &
                                            ctffind_filename_wild = ctffind_files%value,                                    &
                                            wps_are_crossovers = wps_are_crossovers%value,                                  &
                                            fal_wps_are_crossovers = fal_wps_are_crossovers%value,                          &
                                            max_radius = helix_max_radius%value/pixel_size%value,                           &
                                            microtubule =microtubule%value,mt_pf_number=mt_pf_number%value,mt_rise=mt_rise%value,  &
                                            rigid_params = rigid_params,                                                    &
                                            constant_rise_per_subunit = constant_rise_per_subunit%value,                    &
                                            constant_twist_per_length = constant_twist%value,                               &
                                            wps_as_sps = wps_as_sps%value,refine_params =refine_params,                     &
                                            persistence_length = persistence_length%value/pixel_size%value,                       &
                                            torsion_persistence_length = torsion_persistence_length%value/pixel_size%value,       &
                                            stretch_persistence_length = stretch_persistence_length%value/pixel_size%value,       &
                                            fal_wps_in_score = fal_wps_in_score%value,fal_wps_in_threed = fal_wps_in_threed%value,&
                                            refine_normal_modes = rrefine_normal_modes,                                     &
                                            average_z_to_ctftilt_plane = average_z_to_ctftilt_plane%value,                  &
                                            use_hermite_splines = use_hermite_splines%value,                                &
                                            fragmented_phi = fragmented_phi%value,                                          &
                                            fragmented_subunit_num = fragmented_subunit_num%value,                          &
                                            amyloid_beta=amyloid_beta%value,                                                &
                                            acceleration_voltage=acceleration_voltage%value,                                &
                                            spherical_aberration=spherical_aberration%value,                                &
                                            amplitude_contrast=amplitude_contrast%value,                                    &
                                            add_fal_wps=add_fal_wps%value,                                                  &
                                            unblur_shifts=unblur_shifts,                                                    &
                                            box_size_for_subframe_alignment=nint(box_size_for_subframe_alignment%value      &
                                                                                                        /pixel_size%value), &
                                            recompute_subframe_alignments=recompute_subframe_alignments%value,              &
                                            number_of_frames=number_of_frames_per_micrograph%value,                         &
                                            exposure_per_frame=exposure_per_frame%value,                                    &
                                            apply_dose_filter=apply_dose_filter%value,                                      &
                                            helix_min_for_rec=helix_min_for_rec,                                            &
                                            helix_max_for_rec=helix_max_for_rec                                             &
                                            )
    endif
    if (size(fils) .gt. 1) then
        write(*,'(a,i0,a/)') '**info(mode_filament_refine): ', size(fils), ' filaments were initialised'
    else
        write(*,'(a,i0,a/)') '**info(mode_filament_refine): ', size(fils), ' filament was initialised'
    endif
    if (this_program%IsVerbose() .or. debug) then
        write(*,'(a)') '</parse_ali_file>'
    endif



    !
    ! Allocate memory to array of final filament states
    !
    allocate(fils_final(size(fils)),stat=ierr)
    if (ierr .ne. 0) then
        write(*,'(a,i0)') '**error(mode_filament_refine): memory allocation for fils_final array failed with error ', ierr
        call terminate('Fatal error in MODE_FILAMENT_REFINE')
    endif

    !
    ! Allocate memory to array which will keep track of scores & positions as we minimise
    !
    ! The score array will keep track of the score for for each oritentation (flip) and for each iteration of the refinement
    !
    allocate(       score(num_polarities_to_test,num_phi_polarities_to_test,2),        &
                    fil_states(num_polarities_to_test,num_phi_polarities_to_test,2),   &
                    stat=ierr   )
    if (ierr .ne. 0) then
        write(*,'(a,i0)') '**error(mode_filament_refine): memory allocation for score array failed with error ', ierr
        call terminate('Fatal error in MODE_FILAMENT_REFINE')
    endif

    ! Memory footprint diag
    if (debug) then
        write(*,'(a)') '**debug(mode_filament_refine): memory usage after filaments, scores etc, but before 3Ds'
        call print_memory_usage_stats()
    endif


    !
    ! Read the 3d model from disk
    !
    if (this_program%IsVerbose() .or. debug) then
        write(*,'(a)') '<read_model_3d_from_disk>'
    endif
    if (model_3d_available%value) then
        call threed_file%Init(model_3d_filename%value,pixel_size=pixel_size%value,read_only=.true.)
        call model_3dvol%ReadFromDisk(threed_file,read_volume_not_slice=.true.)
        call threed_file%close()
        ! Check the 3d model is indeed 3d
        if (any(model_3dvol%logical_dimensions .le. 1)) then
            write(*,'(a,3(i0,1x))') '**error(mode_filament_refine): input 3d model is not 3d! Dimensions = ', &
                                    model_3dvol%logical_dimensions
            call terminate('Fatal error in MODE_FILAMENT_REFINE')
        endif
        ! For now, we assume the 3d model will be cube
        if (.not. model_3dvol%IsSquare()) then
            write(*,'(a,3i5)')  '**error(mode_filament_refine): sorry, 3d model must be in a cubic box. '//&
                                'Image supplied is ', model_3dvol%logical_dimensions
            call terminate('Fatal error in MODE_FILAMENT_REFINE')
        endif
        ! Also read in the full 3d reconstruction information (required when subtract_fil_from_ref%value)
        if (subtract_fil_from_ref%value .or. single_filament_refinement%value) then
            call threed_init(   ref_3drec,                                                          &
                                filename_root       =   model_3d_filename%value,                    &
                                interpol_scheme     =   interpol_scheme,                            &
                                !oversampling_factor =   1,                                          &
                                sym_symbol          =   sym_symbol,                                 &
                                helix               =   helix_average,                              &
                                asym_unit_mw        =   asym_unit_mw%value,                         &
                                pixel_size          =   pixel_size%value,                           &
                                helix_max_rad       =   helix_max_radius%value / pixel_size%value,  &
                                compute_fsc         =   calc_fsc%value,                             &
                                use_ssnr_estimate   =   estimate_ssnr%value,                        &
                                fom_filter          =   apply_fom_filter%value,                     &
                                reconstruct_amplitudes = reconstruct_amplitudes%value,              &
                                estimated_resolution =  estimated_resolution_of_model_3d )
        endif
    else
        if (debug) write(*,'(a)') '**debug(mode_filament_refine): model 3d not available, therefore not read in!'
        if (.not. skip_refinement%value) then
            write(*,'(a)') '**info(mode_filament_refine): no model 3d is available, so no refinement will be done'
            skip_refinement%value = .true.
        endif
        if (write_matching_projections%value) then
            write(*,'(a)') '**warning(mode_filament_refine): since no model 3d volume is available, '// &
                            'no matching projections will be generated'
            write_matching_projections%value = .false.
        endif
    endif
    if (this_program%IsVerbose() .or. debug) write(*,'(a)') '</read_model_3d_from_disk>'

    ! Memory footprint diagnostics
    if (debug) then
        write(*,'(a)') '**debug(mode_filament_refine): memory usage after 3d read from file'
        call print_memory_usage_stats()
    endif

    !
    ! Ensure that the protein density in the 3D is roughly aligned with the primitive unit cell,
    ! and then make sure the filaments are modified accordingly
    !
    if (align_map_to_helical_lattice%value) then
        write(*,'(a)') '**debug(mode_filament_refine): experimental: aligning 3d to helical lattice'
        !call model_3dvol%WriteToDisk('dbg_filament_refine_3d_before_alignment.mrc')
        call image_align_to_helical_lattice(model_3dvol,helix_average,0.0,helix_max_radius%value/pixel_size%value, &
                                              model3d_phi_shift,model3d_z_shift)
        !call model_3dvol%WriteToDisk('dbg_filament_refine_aligned_3d.mrc')
        ! Apply corresponding shifts to the filaments
        do current_filament=1,size(fils)
            call filament_shift_wps_to_obtain_3d_z_phi_shift(fils(current_filament),model3d_z_shift,model3d_phi_shift)
        enddo
    endif



    !
    ! If necessary, initialise a 3d reconstruction object
    !
    if (this_program%IsVerbose() .or. debug) write(*,'(a)') '<initalise_3d_reconstruction>'
    if (calc_3d_reconstruction%value) then
        new_3drec%logical_dimensions(1:3) = fmc_threed_size(interpol_scheme,helix_max_radius%value/pixel_size%value)
        if (this_program%IsVerbose() .or. debug) then
            write(*,'(a,i0)') '**info(mode_filament_refine): initialising output 3d reconstruction. Box size: ', &
                                new_3drec%logical_dimensions(1)
        endif
        call threed_init(   new_3drec,                                                                  &
                            interpol_scheme     =   interpol_scheme,                                    &
                            sym_symbol          =   sym_symbol,                                         &
                            helix               =   helix_average,                                      &
                            asym_unit_mw        =   asym_unit_mw%value,                                 &
                            pixel_size          =   pixel_size%value,                                   &
                            helix_max_rad       =   helix_max_radius%value / pixel_size%value,          &
                            compute_fsc         =   calc_fsc%value,                                     &
                            use_ssnr_estimate   =   estimate_ssnr%value,                                &
                            fom_filter          =   apply_fom_filter%value,                             &
                            reconstruct_amplitudes = reconstruct_amplitudes%value,                      &
                            estimated_resolution =  estimated_resolution_of_model_3d )
        ! Memory footprint diag
        if (debug) then
            write(*,'(a)') '**debug(mode_filament_refine): memory usage after  new 3drec'
            call print_memory_usage_stats()
        endif
        ! Initialise another 3d rec object, to be used temporarily
        if (this_program%IsVerbose() .or. debug) then
            write(*,'(a)') '**info(mode_filament_refine): initialising temporary 3d reconstruction'
        endif
        current_3drec%logical_dimensions(1:3) = fmc_threed_size(interpol_scheme,helix_max_radius%value/pixel_size%value)
        call threed_init(   current_3drec,                                                              &
                            interpol_scheme     =   interpol_scheme,                                    &
                            sym_symbol          =   sym_symbol,                                         &
                            helix               =   helix_average,                                      &
                            asym_unit_mw        =   asym_unit_mw%value,                                 &
                            pixel_size          =   pixel_size%value,                                   &
                            helix_max_rad       =   helix_max_radius%value / pixel_size%value,          &
                            compute_fsc         =   calc_fsc%value,                                     &
                            use_ssnr_estimate   =   estimate_ssnr%value,                                &
                            fom_filter          =   apply_fom_filter%value,                             &
                            reconstruct_amplitudes = reconstruct_amplitudes%value,                      &
                            estimated_resolution =  estimated_resolution_of_model_3d     )
        ! Memory footprint diag
        if (debug) then
            write(*,'(a)') '**debug(mode_filament_refine): memory usage after  current 3drec'
            call print_memory_usage_stats()
        endif
        if (single_filament_reconstructions%value) then
            current_3dvol_slice%logical_dimensions(1:2) = fmc_threed_size(interpol_scheme,helix_max_radius%value/pixel_size%value)
            current_3dvol_slice%logical_dimensions(3) = 1
            call current_3dvol_slice%Allocate()
        endif
    endif
    if (this_program%IsVerbose() .or. debug) then
        write(*,'(a)') '</initalise_3d_reconstruction>'
    endif

    ! Memory footprint diag
    if (debug) then
        write(*,'(a)') '**debug(mode_filament_refine): memory usage after 3Ds loaded'
        call print_memory_usage_stats()
        write(*,'(a,i0,a)') '**debug(mode_filament_refine): 3Drec object storage size = ', storage_size(new_3drec)/8/1024/1024, 'MB'
    endif


    !
    ! If we're in "single-particle mode" (i.e. where waypoints are treated as independent single particles), let the user know
    !
    if (this_program%IsVerbose() .or. debug) then
        if (wps_as_sps%value) then
            write(*,'(2a)') '**info(mode_filament_refine): will be treating waypoints as independent single particles, ',    &
                            'so full-filament scoring & refinement will not take place'
        endif
    endif

    !
    ! Loop over filaments
    !
    if (this_program%IsVerbose() .or. debug) write(*,'(a,i0,a)') '<loop_over_filaments num_of_fils=',size(fils),'>'
    do current_filament=1,size(fils)

        if (this_program%IsVerbose() .or. debug) write(*,'(a,i0,a)') '<filament num=',current_filament,'>'

        ! Did the user ask us to straighten the filament on entry?
        if (straighten_filament_before_refinement%value) then
            if (this_program%IsVerbose() .or. debug) then
                write(*,'(a)') '**info(mode_filament_refine): straightening filament before refinement begins'
            endif
            call filament_straighten(fils(current_filament))
        endif

        ! When in single-particle mode, we may need to do some fixing of Euler angles, especially if they were determined
        ! with a program which doesn't follow our conventions
        if (wps_as_sps%value) then
            call filament_fix_polarity_inconsistencies(fils(current_filament),allow_polarity_flip=sp_to_ff%value)
        endif


        ! If requested, add new waypoints at either end of the filament
        ! This should now be done on filament initialisation
        !if (add_fal_wps%value) then
        !    write(*,'(a,i0)') '**info(mode_filament_refine): adding waypoints to the ends of filament ', fils(current_filament)%id
        !    call filament_add_fal_wps(fils(current_filament))
        !endif

        ! If requested, resample waypoints along the path of the filament
        if (force_wp_to_wp_distance%value .gt. 0.0) then
            call filament_resample_waypoints(fils(current_filament),real(force_wp_to_wp_distance%value/pixel_size%value,kind=8))
        endif
        if (force_wp_to_wp_phi%value .gt. 0.0) then
            call filament_resample_waypoints(fils(current_filament),real(force_wp_to_wp_phi%value,kind=8),step_in_phi=.true.)
        endif
        if (force_wp_to_wp_distance%value .gt. 0.0 .and. force_wp_to_wp_phi%value .gt. 0.0) then
            write(*,'(a)') '**error(mode_filament_refine): force_wp_to_wp_distance and force_wp_to_wp_phi can''t be used together'
            call terminate('Fatal error in MODE_FILAMENT_REFINE. Bad choice of parameters.')
        endif


        if (debug) then
            write(*,'(a)') '**debug(mode_filament_refine): filament info with statistics follows'
            call fils(current_filament)%PrintInfo()
            call filament_print_extra_info(fils(current_filament))
        endif


        ! Will we need to read a new film image from disk?
        if (current_filament .eq. 1) then
            new_film    =   .true.
        else if (fils(current_filament)%film%GetFilename() .ne. fils(current_filament-1)%film%GetFilename()) then
            new_film    =   .true.
        else if (fils(current_filament)%film%GetLocationInFile() .ne. fils(current_filament-1)%film%GetLocationInFile()) then
            new_film    =   .true.
        else
            new_film    =   .false.
        endif
        if (calc_3d_reconstruction%value .or. subtract_fil_from_ref%value .or. single_filament_refinement%value) then
            new_film = .true. ! This is because when 3d reconstruction is computed, the film is read in again but not filtered
        endif
        ! Note that some time could be saved by working out whether the correct film image is alread loaded in memory and
        ! therefore only the filtering remains to be done. However, to keep the code simpler for now, i sacrifice some run time to IO.
        ! In the case where we subtract fil 3d from reference, this will be done by fmc_init or fmc_new_filament, so the film
        ! must be bandpassed after that.

        ! If necessary, read a new film image from disk & high-pass filter it
        film_hp_param =   pixel_size%value / (3.0 * helix_max_radius%value)
        if (new_film .and. model_3d_available%value) then
            if (debug) print '(2a)', '**debug(mode_filament_refine): reading micrograph ', &
                                    trim(adjustl(fils(current_filament)%film%GetFilename()))
            call film_file%Init(fils(current_filament)%film%GetFilename(),read_only=.true.)
            if (.not. allocated(current_micrograph_frames)) then
                allocate(current_micrograph_frames(number_of_frames_per_micrograph%value))
            endif
            do current_frame=1,number_of_frames_per_micrograph%value
                call current_micrograph_frames(current_frame)%ReadFromDisk(film_file, &
                                   fils(current_filament)%film%GetFrameLocationInFile(current_frame))
            enddo
            call film_file%close()
            if (reverse_film_contrast%value) then
                do current_frame=1,number_of_frames_per_micrograph%value
                    current_micrograph_frames(current_frame)%real_values = &
                    current_micrograph_frames(current_frame)%real_values * (-1.0)
                enddo
            endif
            ! High-pass filter the film image to bring average density to 0.0 and remove density ramps
            if (this_program%IsVerbose() .or. debug) then
                write(*,'(a,F0.2,a)') '**info(mode_filament_refine): high-pass filtering the micrograph. Resolution = ', &
                                        pixel_size%value/film_hp_param, ' A'
            endif
            do current_frame=1,number_of_frames_per_micrograph%value
                call current_micrograph_frames(current_frame)%ApplyBandPassFilter(film_hp_param,0.0,film_hp_param/2.0)
            enddo
        endif

        !
        ! If necessary:
        ! - resample the model 3d (first pad in real space, then crop in Fourier space)
        ! - resample the film (first pad in real space, then crop in Fourier space)
        ! - setup helical symmetry object with new pixel size
        !
        refinement_pixel_size = pixel_size%value
        reconstruction_pixel_size = pixel_size%value
        helix_average_refinement = helix_average
		! What pixel size would we like to use during refinement?
		target_refinement_pixel_size = (high_resolution_cutoff%value* 0.5) / target_psize_factor
        !
        if (.not. skip_refinement%value .and. resample_data_for_refinement%value &
                                .and. target_refinement_pixel_size .gt. pixel_size%value ) then
            ! What pixel size would we like to use during refinement?
            target_refinement_pixel_size = (high_resolution_cutoff%value * 0.5) / target_psize_factor
            ! Let the user know
            write(*,'(2a,f0.2,2a)')  &
                            '**info(mode_filament_refine): to save time during refinement, the model 3d and the micrograph ', &
                            ' will be resampled to a new pixel size of ~ ', target_refinement_pixel_size, 'A. ', &
                            'The original micrograph will be used for 3d reconstruction once refinement is complete.'

            if (current_filament .eq. 1) then
                ! Resample the 3d reconstruction
                call find_dimensions_for_resampling(model_3dvol,pixel_size%value/target_refinement_pixel_size, &
                                                    model_3dvol_padding_dimensions,model_3dvol_refinement_dimensions, &
                                                    actual_scaling_factor)
                if (.not. model_3dvol%is_in_real_space) call model_3dvol%BackwardFFT()
                call model_3dvol%Resize(model_3dvol_padding_dimensions,padding_value_real=model_3dvol%GetAverageOfValuesOnEdges())
                call model_3dvol%ForwardFFT()
                call model_3dvol%Resize(model_3dvol_refinement_dimensions)
                call model_3dvol%BackwardFFT()
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a,f0.3)') '**info(mode_filament_refine): new pixel size for model3D = ', &
                                                                pixel_size%value/actual_scaling_factor
                endif
            endif

            if (new_film) then
                ! Resample the film
                film_original_dimensions = current_micrograph_frames(1)%logical_dimensions
                dummy = actual_scaling_factor
                call find_dimensions_for_resampling(current_micrograph_frames(1),dummy, &
                                                    film_padding_dimensions,film_refinement_dimensions, &
                                                    actual_scaling_factor)
                do current_frame=1,number_of_frames_per_micrograph%value
                    if (.not. current_micrograph_frames(current_frame)%is_in_real_space) then
                        call current_micrograph_frames(current_frame)%BackwardFFT()
                    endif
                    call current_micrograph_frames(current_frame)%Resize(film_padding_dimensions, &
                                           padding_value_real=current_micrograph_frames(current_frame)%GetAverageOfValuesOnEdges())
                    call current_micrograph_frames(current_frame)%ForwardFFT()
                    call current_micrograph_frames(current_frame)%Resize(film_refinement_dimensions)
                    call current_micrograph_frames(current_frame)%BackwardFFT()
                enddo
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a,f0.3)') '**info(mode_filament_refine): new pixel size for film   = ', &
                                                                pixel_size%value/actual_scaling_factor
                endif
            endif

            ! Set the refinement pixel size
            refinement_pixel_size = pixel_size%value / actual_scaling_factor

            ! Rescale the filament
            call filament_change_pixel_size(fils(current_filament), &
                                            film_original_dimensions,film_refinement_dimensions,actual_scaling_factor)


            ! Rescale helix_average object (change the rise and radius)
            call helical_symmetry_init( helix_average_refinement,rise_per_subunit%value/refinement_pixel_size, &
                                        twist_per_subunit%value,  &
                                        helix_max_radius%value/refinement_pixel_size, &
                                        symmetry_axial%value,symmetry_perp%value, &
                                        num_starts=num_helical_starts%value, &
                                        stack=stack%value)

            if (this_program%IsVerbose() .or. debug) then
                call helical_symmetry_print_info(helix_average)
                call helical_symmetry_print_info(helix_average_refinement)
                call fils(current_filament)%PrintInfo()
            endif

            ! psf radius
            psf_radius_pixels = psf_radius_pixels * actual_scaling_factor
        endif ! end of test for resampling

        !
        ! Any search half-range override from the user
        !
        search_half_range_override = [  search_half_range_x%value/refinement_pixel_size,&
                                        search_half_range_y%value/refinement_pixel_size, &
                                        search_half_range_z%value/refinement_pixel_size,  &
                                        search_half_range_phi%value,search_half_range_theta%value,&
                                        search_half_range_psi%value,&
                                        search_half_range_subunit_num%value, &
                                        search_half_range_helical_rise%value/refinement_pixel_size, &
                                        search_half_range_helical_twist%value ]

        search_step_size_override  = [  search_step_size_x%value/refinement_pixel_size,  &
                                        search_step_size_y%value/refinement_pixel_size, &
                                        search_step_size_z%value/refinement_pixel_size,  &
                                        search_step_size_phi%value,search_step_size_theta%value, &
                                        search_step_size_psi%value,&
                                        search_step_size_subunit_num%value, &
                                        search_step_size_helical_rise%value/refinement_pixel_size, &
                                        search_half_range_helical_twist%value  ]

        ! Let's work out what kind of scoring function the user wants us to use
        type_of_scoring_function = linear_correlation
        if (use_weighted_correlation%value) type_of_scoring_function = ssnr_weighting



        ! Initialise a filament-model comparison (fmc) object if necessary
        if (model_3d_available%value) then
            if (.not. fmc%initialised) then
                call fmc_init(                                                                                  &
                                fmc,fils(current_filament),current_micrograph_frames,                           &
                                model_3dvol,                                                                    &
                                helix_average_refinement,                                                       &
                                target_resolution           =   target_resolution%value,                        &
                                interpol                    =   interpol_scheme,                                &
                                asym_unit_mw                =   asym_unit_mw%value,                             &
                                !real_space_proj             =   real_space_projection%value,                    &
                                project_individual_subunits =   project_individual_subunits%value               &
                                                           .or. refine_helical_rise%value                       &
                                                           .or. refine_helical_twist%value,                     &
                                subtract_fil_from_ref       =   subtract_fil_from_ref%value,                    &
                                model3d                     =   ref_3drec,                                      &
                                wps_as_sps                  =   wps_as_sps%value,                               &
                                sp_segment_length           =   sp_segment_length%value,                        &
                                search_half_range_override  =   search_half_range_override,                     &
                                search_step_size_override   =   search_step_size_override,                      &
                                num_minim_runs              =   num_minim_runs%value,                           &
                                ds_random_start_conf        =   ds_random_start_conf%value,                     &
                                minim_random_start_pos      =   minim_random_start_pos%value,                   &
                                ds_max_num_iter             =   ds_max_number_of_iterations%value,              &
                                path_oversampling           =   path_oversampling%value,                        &
                                low_resolution_cuton        =   min(low_resolution_cuton%value,pixel_size%value/film_hp_param), &
                                high_resolution_cutoff      =   high_resolution_cutoff%value,                   &
                                type_of_scoring_function    =   type_of_scoring_function,                       &
                                !cubed_average               =   use_average_of_cubed_segment_scores%value,      &
                                curvature_restraint         =   use_curvature_restraint%value,                  &
                                torsion_restraint           =   use_torsion_restraint%value,                    &
                                stretch_restraint           =   use_stretch_restraint%value,                    &
                                model_ave_cyl               =   average_model_cylindrically%value,              &
                                model3dvol_bfactor          =   model_3d_filter_bfactor%value,                  &
                                hold_param_average          =   hold_param_average,                             &
                                use_downhill_simplex        =   use_downhill_simplex%value,                     &
                                use_conjugate_gradient      =   use_conjugate_gradient%value,                   &
                                use_differential_evolution  =   use_differential_evolution%value,               &
                                use_brute_force             =   use_brute_force%value,                          &
                                refine_param_increments     =   refine_param_increments,                        &
                                integer_changes_only        =   integer_changes_only,                           &
                                cross_beta_filter           =   cross_beta_filter%value,                        &
                                refine_wps_simultaneously   =   refine_wps_simultaneously%value,                &
                                single_filament_refinement  =   single_filament_refinement%value,               &
                                estimated_resolution_of_model_3d = estimated_resolution_of_model_3d,            &
                                first_frame_for_refinement  =   max(first_frame_for_refinement%value,1),        &
                                last_frame_for_refinement   =   min(last_frame_for_refinement%value,            &
                                                                    number_of_frames_per_micrograph%value),     &
                                boost_low_frequency_contrast=   boost_low_frequency_contrast%value,             &
                                apply_dose_filter           =   apply_dose_filter%value,                        &
                                apply_drift_filter          =   apply_drift_filter%value                        &
                            )
            else
                ! The fmc object is already initialised
                fmc%interpol = interpol_scheme
                call fmc_new_filament(fmc,fils(current_filament))
                if (fmc%subtract_fil_from_ref .or. fmc%single_filament_refinement) then
                    ! Compute the reconstruction of the filament from the original state, so that it can later
                    ! be subtracted from the reference 3d
                    call fmc_update_fil_threed(fmc)
                endif
            endif
            call fmc_print_info(fmc)
        endif



        if (debug .and. model_3d_available%value) then
            write(*,'(a)') '**debug(mode_filament_refine): filament info with statistics follows (2)'
            call fmc%fil%PrintInfo()
            call filament_print_extra_info(fmc%fil)
        endif

        ! Memory footprint diag
        if (debug) then
            write(*,'(a)') '**debug(mode_filament_refine): memory usage after FMC'
            call print_memory_usage_stats()
        endif


        !
        ! If none of the waypoints of the current filament are marked as to-be-refined, we can skip to the next filament
        !
        if (.not. any(fils(current_filament)%wp%refine) .or. skip_refinement%value) then

            write(*,'(a,i0,a)') '**info(mode_filament_refine): none of the waypoints of filament ', fils(current_filament)%id, &
                           ' need to be refined, or you indicated that refinement should be skipped. Skipping filament refinement.'
            fmc%interpol = interpol_scheme
            if (model_3d_available%value) call fmc_new_filament(fmc,fils(current_filament))
            call fils_final(current_filament)%Assign(fils(current_filament))

        !
        ! If at least one of the waypoints of the current filament are marked as to-be-refined
        !
        else
            !
            ! Loop over filament polarities
            !
            !
            ! todo: introduce OpenMP parallelisation around polarity loops?
            !

            if (this_program%IsVerbose() .or. debug) write(*,'(a)') '<polarity_loop>'
            polarity_loop : do current_polarity=1,num_polarities_to_test

                if (this_program%IsVerbose() .or. debug) write(*,'(a,i0,a)') '<polarity num=', current_polarity, '>'
                ! Let the user know if we're checking polarities
                if (num_polarities_to_test .gt. 1) then
                    write(*,'(2(A,i1),a,i0)') '**info(mode_filament_refine): checking polarity ', current_polarity, ' of ', &
                                                num_polarities_to_test, ' for filament ', fmc%fil%id
                endif


                ! Loop over phi polarities
                if (this_program%IsVerbose() .or. debug) write(*,'(a)') '<phi_polarity_loop>'
                do current_phi_polarity=1,num_phi_polarities_to_test

                    if (this_program%IsVerbose() .or. debug) write(*,'(a,i0,a)') '<phi_polarity num=', current_phi_polarity, '>'
                    if (current_phi_polarity .gt. 1) then
                        write(*,'(a,i0,a,i0)') '**info(mode_filament_refine): on phi polarity ', current_phi_polarity, &
                                                ' of ', num_phi_polarities_to_test
                    endif

                    ! If necessary, reset the filament to the starting point and flip its polarity
                    call fmc%fil%Assign(fils(current_filament))
                    if (debug) then
                        write(*,'(a)') '**debug(mode_filament_refinement): fil info after resetting to fils(current_filament):'
                        call fmc%fil%PrintInfo()
                    endif
                    if (current_polarity .eq. 2) then
                        !print '(a)', '**debug(mode_filament_startup): flipping polarity of filament'
                        call filament_flip_polarity(fmc%fil)
                        fmc%fil_original = fmc%fil
                    endif
                    if (current_phi_polarity .eq. 2) then
                        call filament_shift_all_wps(fmc%fil,[0.,0.,0.,180.,0.,0.],update_model=.true.)
                        fmc%fil_original = fmc%fil
                    endif

                    ! Print out filament info for current polarity
                    write(*,'(3(A,i0),a)') '**info(mode_filament_refine): info for filament ', fmc%fil%id, &
                                            ' (polarity ', current_polarity, '; phi polarity ', current_phi_polarity, &
                                            '), before refinement: '
                    call fmc%fil%PrintInfo()

                    ! Remember the starting state of the filament
                    fil_states(current_polarity,current_phi_polarity,1) = fmc%fil


                    ! Prepare the model 3d (pad/crop, impose helical symmetry, mask, subtract current fil's 3D...)
                    ! todo: Could this be done outside the polarity loops? (the model should be constant)
                    if (this_program%IsVerbose() .or. debug) write(*,'(a)') '<fmc_model3d_prep>'
                    call fmc_model3d_prep(fmc)
                    if (this_program%IsVerbose() .or. debug) write(*,'(a)') '</fmc_model3d_prep>'


                    ! Memory footprint diag
                    if (debug) then
                        write(*,'(a)') '**debug(mode_filament_refine): memory usage after fmc_model3d_prep'
                        call print_memory_usage_stats()
                    endif

                    !
                    ! Score the starting state
                    !
                    print '(/a,i0,a,i0,a,i0,a)',    '**info(mode_filament_refine): scoring starting state for filament ', &
                                                    fils(current_filament)%id, ' (polarity ', current_polarity, &
                                                    '; phi polarity ', current_phi_polarity, ')'
                    write(temp_filename,'(a,i1,a)') 'start_pol_', current_polarity, '.mrc'
                    score(current_polarity,current_phi_polarity,1) = fmc_eval(fmc,fmc%fil,write_stacks=.false.,&
                                                                                stacks_rootname=temp_filename,&
                                                                                detailled_score_info=.true.)
                    print '(a,i0,a,i1,a,F0.3/)',    '**info(mode_filament_refine): starting state score for filament ', &
                                                    fils(current_filament)%id, ' (polarity ', current_polarity, &
                                                    ') is ', score(current_polarity,current_phi_polarity,1)

                    ! Memory footprint diag
                    if (debug) then
                        write(*,'(a)') '**debug(mode_filament_refine): memory usage after first scoring function'
                        call print_memory_usage_stats()
                    endif




                    if (fmc%wps_as_sps) then
                        ! We've been told to treat WPs as single particles...
                        call fmc_maximise(fmc)
                        if (this_program%IsVerbose() .or. debug) then
                            write(*,'(2a,i0,a)') '**debug(mode_filament_refine): detailled score after single ', &
                                                'particle refinement, filament #', fmc%fil%id, ':'
                            dummy = fmc_eval(fmc,fmc%fil,detailled_score_info=.true.)
                        endif
                    else
                        ! We are not treating waypoints as single particles
                        if (search_subunit_num%value) then
                            print '(a)', '**info(mode_filament_refine): trying bf subunit_num search before DS. Experimental.'
                            call fmc_bf_subunit_num_intg_refine(fmc)
                        endif


                        ! If requested, do a filament-tilt brute-force search
                        ! todo: it should be possible to incorporate this search logic in the framework of the rigid_params
                        if (search_filament_tilt%value) then
                            write(*,'(2a,F0.2,a,F0.2,a)') '**info(mode_filament_refine): doing rigid-body filament ',&
                                                            'tilt brute force search over +/- ', fmc%search_half_range(5), &
                                                            ' degrees in ',fmc%search_sampling(5),' degree steps'
                            call fmc_bf_tilt(fmc,tilt_half_range=fmc%search_half_range(5),tilt_sampling=fmc%search_sampling(5))
						else
							! Now find the nearby maximum in the scoring function
	                        ! (this is the workhorse of refinements)
	                        call fmc_maximise(fmc)
						endif

                        ! If requested, do seam search
                        if (search_mt_seam%value) then
                            if (fmc%fil%microtubule) then
                                write(*,'(a)') '**info(mode_filament_refine): will search for microtubule seam'
                                call fmc_bf_find_mt_seam(fmc)
                            else
                                if (debug) then
                                    write(*,'(a)') '**debug(mode_filament_refine): filament is not microtubule, '// &
                                                    'will not search for seam.'
                                endif
                            endif
                        endif


                        ! Test: to give the flipped polarity a better chance, we do another maximisation
                        if (current_polarity .eq. 2) then
                            !write(*,'(a)') '**debug(mode_filament_refine): giving the flipped polarity a second and third run at the maximisation'
                            !call fmc_maximise(fmc)
                        endif
                    endif   ! End of test for wps_as_sps%value

                    ! Evaluate the score at the end of the iteration
                    write(*,'(a,i0,a)') '**info(mode_filament_refine): scoring information for filament ', &
                                            fils(current_filament)%id,  ' following minimisation follows'
                    write(temp_filename,'(a,i1)') 'after_ds_polarity', current_polarity
                    if (this_program%IsVerbose() .or. debug) then
                        write(*,'(2a)') '**debug(mode_filament_refine): will dump matching projections to files named ', &
                                                trim(temp_filename)
                    endif
                    score(current_polarity,current_phi_polarity,2) = fmc_eval(fmc,fmc%fil, &
                                                                        write_stacks=.false.,stacks_rootname=temp_filename, &
                                                                        detailled_score_info=.true.)

                    ! Store the state of the filament at the end of the refinement
                    fil_states(current_polarity,current_phi_polarity,2) = fmc%fil


                    ! It may be interesting to get a ff score when we're doing sp refinement, to see how "compatible"
                    ! the two modes are
                    if (fmc%wps_as_sps .and. .false.) then ! 141022: turning this off because causing problems with non-constant helical parameters when switching from SP to FF
                        write(*,'(2a)') '**info(mode_filament_refine): below is score with ',   &
                                        'full-filament (rather than wps as sps) function: '
                        fmc%wps_as_sps = .false.
                        temp_fil = fmc%fil
                        temp_fil%wps_as_sps = .false.
                        !fmc%fil%wps_as_sps = .false.
                        dummy = fmc_eval(fmc,temp_fil,write_stacks=.false.,  &
                                        stacks_rootname=temp_filename,detailled_score_info=.true.)
                        fmc%wps_as_sps = .true.
                        !fmc%fil%wps_as_sps = .true.
                    endif

                    print '(3(A,i0),a,F0.4)',   '**info(mode_filament_refine): score for filament ', fils(current_filament)%id , &
                                                ' (polarity ', current_polarity, '; phi polarity ', current_phi_polarity, '): ', &
                                                score(current_polarity,current_phi_polarity,2)
                    call fmc%fil%PrintInfo()
                    if (this_program%IsVerbose() .or. debug) then
                        if (.not. wps_as_sps%value) call filament_print_extra_info(fmc%fil)
                    endif

                    if (debug) then
                        write(*,'(a)') '**debug(mode_filament_refine): fil info for saved state (after assignment):'
                        call fil_states(current_polarity,current_phi_polarity,2)%PrintInfo()
                    endif


                    if (this_program%IsVerbose() .or. debug) write(*,'(a)') '</phi_polarity>'
                enddo   ! End of loop over filament phi polarities
                if (this_program%IsVerbose() .or. debug) write(*,'(a)') '</phi_polarity_loop>'

                ! Decide which phi polarity was better for the current filament in the current polarity
                if (phi_polarity_check%value) then
                    if (score(current_polarity,1,2) .gt. score(current_polarity,2,2)) then
                        best_phi_polarity(current_polarity) = 1
                    else
                        best_phi_polarity(current_polarity) = 2
                    endif

                    write(*,'(3(A,i0),a,F0.4)')     '**info(mode_filament_refine): for filament ', fils(current_filament)%id,   &
                                                    ' in polarity ', current_polarity,                                          &
                                                    ', phi polarity ', best_phi_polarity(current_polarity),                     &
                                                    ' was best, with a score of ', &
                                                    score(current_polarity,best_phi_polarity(current_polarity),2)
                else
                    best_phi_polarity(current_polarity) = 1
                endif

                if (this_program%IsVerbose() .or. debug) write(*,'(a)') '</polarity>'
            enddo   polarity_loop ! End of loop over filament polarities
            if (this_program%IsVerbose() .or. debug) write(*,'(a)') '</polarity_loop>'

            !
            ! Decide which polarity was best for the current filament
            !
            if (polarity_check%value) then
                if (score(1,best_phi_polarity(1),2) .gt. score(2,best_phi_polarity(2),2)) then
                    best_polarity = 1
                else
                    best_polarity = 2
                endif

                write(*,'(a,i0,a,i0,a,F0.4)')   '**info(mode_filament_refine): for filament ', fils(current_filament)%id,   &
                                                ', polarity ', best_polarity,                                       &
                                                ' was best, with a score of ', &
                                                score(best_polarity,best_phi_polarity(best_polarity),2)
                ! Reset the filament object in the fmc to the starting point of refinement for the winning polarity
                call fmc_new_filament(fmc,fil_states(best_polarity,best_phi_polarity(best_polarity),1))
            else
                best_polarity = 1
            endif

            !
            ! Remember the results of the search for the current filament
            !
            !fils_final(current_filament)        =   fil_states(best_polarity,best_phi_polarity(best_polarity),2)
            call fils_final(current_filament)%Assign(fil_states(best_polarity,best_phi_polarity(best_polarity),2))
            fils_final(current_filament)%score  =   score(best_polarity,best_phi_polarity(best_polarity),2)
        !
        ! End of test of whether any wps needed to be refined
        !
        endif


        !
        ! Now that we're all done with refinement, compute the final score and output any files the user may
        ! have asked for (matching projections etc.)
        !
        if (model_3d_available%value) then
            if (fmc%initialised) then
                ! Memory footprint diag
                if (debug) then
                    write(*,'(a)') '**debug(mode_filament_refine): memory before model3d prep'
                    call print_memory_usage_stats()
                endif
                ! Prepare the model 3d (pad/crop, impose helical symmetry, mask..). In case
                ! we didn't do any refinement, this won't have been done yet.
                call fmc_model3d_prep(fmc)
                ! Memory footprint diag
                if (debug) then
                    write(*,'(a)') '**debug(mode_filament_refine): memory usage after model3d prep'
                    call print_memory_usage_stats()
                endif
                call filename_from_wild(matching_projections_files%value,fils_final(current_filament)%id,match_proj_filename)
                if (write_matching_projections%value) then
                    write(*,'(a,i0,2a)') '**info(mode_filament_refine): writing stacks of images and matching projections '//&
                                        'for filament ', fils_final(current_filament)%id, &
                                        '. Filename root: ', trim(adjustl(match_proj_filename))
                endif
                ! Print out final state of filament
                write(*,'(a,i0,a)') '**info(mode_filament_refine): final state of filament ', fils(current_filament)%id, ':'
                call fils_final(current_filament)%PrintInfo()
                write(*,'(a,i0,a)') '**info(mode_filament_refine): final statistics for filament ', fils(current_filament)%id, ':'
                call filament_print_extra_info(fils_final(current_filament))
                ! The easiest way to get matching projection stacks is by calling
                ! the scoring function (threading off so that matching projections are in correct order)
                write(*,'(a,i0,a)') '**info(mode_filament_refine): detailled scoring information for final state of filament ', &
                                    fils_final(current_filament)%id, ' follows:'
                score_tmp = fmc_eval(   fmc,fils_final(current_filament),write_stacks=write_matching_projections%value,  &
                                        stacks_rootname=match_proj_filename,   &
                                        detailled_score_info=.true.,threading=.false.)
                if (fmc%step .lt. default_falloff_pixels + 1.) then
                    write(*,'(3a)') '**warning(mode_filament_refine): the step size for the scoring function was small, which ', &
                                   ' suggests filament deformations are significant relative to the pixel size. Check results ', &
                                   ' carefully and if applicable, move to a smaller pixel size'
                endif

                ! Memory footprint diag
                if (debug) then
                    write(*,'(a)') '**debug(mode_filament_refine): memory usage after scoring function'
                    call print_memory_usage_stats()
                endif

                ! Sanity check
                if (abs(score_tmp-fils_final(current_filament)%score) .gt. 0.02) then
                    write(*,'(a,2(F0.4))') '**warning(mode_filament_refine): final score is inconsistent: ', &
                                            score_tmp, fils_final(current_filament)%score
                    call terminate('Failed sanity check in MODE_FILAMENT_REFINE')
                endif
                ! If we used projection cache, also give detailled scoring info with the cache de-activated
                !if (fmc%cache_projections) then
                !    write(*,'(a)') '**info(mode_filament_refine): parameters were refined using cached projections. '// &
                !                    'Below is the detailled scoring information of the filament again, '//&
                !                    'this time without using cached projections.'
                !    dummy = fmc_eval(   fmc,fils_final(current_filament),write_stacks=write_matching_projections%value,  &
                !                        stacks_rootname=match_proj_filename,   &
                !                        detailled_score_info=.true.,threading=.false.,write_frc_plots=.false.,dont_use_cache=.true.)
                !endif
                ! We also do a sanity check
                if (any(fils(current_filament)%wp%refine) .and. .not. skip_refinement%value) then
                    if ( abs( score(best_polarity,best_phi_polarity(best_polarity),2)    &
                             -fils_final(current_filament)%score) .gt. 0.02 ) then
                        write(*,'(2(A,F8.5))')  '**warning(mode_filament_refine): mismatch between scores - '//&
                                                'programming error? Scores =', &
                                                score(best_polarity,best_phi_polarity(best_polarity),2), ';', &
                                                fils_final(current_filament)%score
                        if (any(refine_params(8:9))) then
                            write(*,'(2a)') '**info(mode_filament_refinement): ignoring this warning, since it''s possible', &
                                            ' the score got worse only because the helical parameters of the model 3D changed'
                        else if (use_curvature_restraint%value) then
                            write(*,'(3a)') '**info(mode_filament_refinement): ignoring this warning, since it''s possible', &
                                            ' the score got worse only because of a strong curvature restraint and numerical', &
                                            ' noise in the filament''s space curve'
                        else
                            call terminate('Failed sanity check in MODE_FILAMENT_REFINE')
                        endif
                    endif
                endif
            else
                write(*,'(a)')  '**warning(mode_filament_refine): unexpectedly, fmc was never initialised, '//&
                                'not doing the final scoring etc.'
            endif
        endif ! End of test for model 3d availability


        ! Print out changes in filament alignment
        write(*,'(a)') '**info(mode_filament_refine): summary of changes in filament during refinement: '
        call filament_print_diff(fils(current_filament),fils_final(current_filament))


        !
        ! Now's the time to convert the filament back to the bin1, not-resampled coordinates
        !
        if (.not. skip_refinement%value .and. resample_data_for_refinement%value &
                        .and. target_refinement_pixel_size .gt. pixel_size%value) then
            if (this_program%IsVerbose() .or. debug) then
                write(*,'(a)') '**info(mode_filament_refine): original pixel size will be used for 3d reconstruction.'
            endif
            call filament_change_pixel_size(fils_final(current_filament), &
                                            film_refinement_dimensions,film_original_dimensions,1.0/actual_scaling_factor)
            psf_radius_pixels = psf_radius_pixels / actual_scaling_factor
        endif

        if (this_program%IsVerbose() .or. debug) then
            write(*,'(a)') '**info(mode_filament_refine): filament info before 3d reconstruction'
            call fils_final(current_filament)%PrintInfo()
        endif


        !
        ! If the user asked, write boxer files with new waypoint positions, as well as traces of the filaments.
        !
        if (write_filament_traces%value) then
            if (filename_contains_qmarks(film_files%value)) then
                ! Find out the number of the film to which the current filament belongs
                current_film_number =  file_number_from_filename(fils_final(current_filament)%film%GetFilename(),film_files%value)
                ! Work out the filename we'll write to
                call filename_from_wild(filament_trace_files%value,current_film_number,filament_trace_filename)
                call filename_from_wild(filament_trace_filename,fils(current_filament)%id,filament_trace_filename)
                ! Write the coordinates to disk
                print '(2a)', '**info(mode_filament_refine): dumping trace to ', trim(filament_trace_filename)
                call filament_boxer_db_to_disk(fils_final(current_filament),filament_trace_filename)

                ! Write out trace to csv file
                call filename_remove_extension(filament_trace_filename,filament_trace_filename)
                filament_trace_filename = trim(filament_trace_filename)//'.csv'
                print '(2a)', '**info(mode_filament_refine): dumping per-subunit trace to csv file ', filament_trace_filename
                call filament_trace_to_disk_csv(fils_final(current_filament),filament_trace_filename)

                ! Write out just the way points without trace
                call filename_remove_extension(filament_trace_filename,temp_filename)
                filament_trace_filename = trim(adjustl(temp_filename))//'_wp.box'
                call filament_boxer_db_to_disk(fils_final(current_filament),filament_trace_filename,with_trace=.false.)
            else
                write(*,'(a)') '**warning(mode_filament_refine): writing of filament traces is disabled because input '//&
                                'micrograph stack was used, rather than set of separate files'
            endif
        endif




        !
        ! Calculate a new 3d reconstruction
        !
        ! Note (1-Nov-2010):
        ! Watch out: if all the odd-numbered filaments have low scores, one of the fsc volumes could end up empty!
        ! The logic below could be re-arranged to avoid this.
        ! The score would be checked against the threshold first, and then if the score is good enough,
        ! the filament would be added to the next group.
        if (calc_3d_reconstruction%value) then

            ! Some of the output 3Ds will be croped to save diskspace
            crop_dim = nint(2.*1.1*helix_max_radius%value/pixel_size%value)
            if (IsOdd(crop_dim(1))) crop_dim = crop_dim + 1 ! Not necessary for Frealix, but may be so for other software

            if (     (fils_final(current_filament)%score .gt. min_score_fil_to_3d%value)  &
                .or. (single_filament_reconstructions%value)    ) then
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a,i0,a)') '**info(mode_filament_refine): about to compute 3d reconstruction for filament ', &
                                        fils_final(current_filament)%id, '.'
                    write(*,'(2a)') '**info(mode_filament_refine): reading in the micrograph image. It will be used without', &
                                    ' any filtering for the filament reconstruction'
                endif
                call film_file%init(fils(current_filament)%film%GetFilename(),read_only=.true.)
                if (this_program%IsVerbose() .or. debug) call film_file%PrintInfo()
                ! Read in relevant frames
                if (.not. allocated(current_micrograph_frames)) then
                    allocate(current_micrograph_frames(number_of_frames_per_micrograph%value))
                endif
                do current_frame=1,number_of_frames_per_micrograph%value
                    call current_micrograph_frames(current_frame)%ReadFromDisk(film_file, &
                                              fils(current_filament)%film%GetFrameLocationInFile(current_frame))
                    if (reverse_film_contrast%value) then
                        current_micrograph_frames(current_frame)%real_values(:,:,:) = &
                        current_micrograph_frames(current_frame)%real_values(:,:,:) * (-1.0)
                    endif
                    call current_micrograph_frames(current_frame)%ZeroFloat()
                enddo
                call film_file%close()

                ! Memory footprint diag
                if (debug) then
                    write(*,'(a)') '**debug(mode_filament_refine): memory usage before 3d reconstruction'
                    call print_memory_usage_stats()
                endif
                if (wps_as_sps%value) then
                    call filament_calc_threed(  fils_final(current_filament),current_micrograph_frames,current_3drec,      &
                                                safe_box_size=fmc_box_size(interpol_scheme, &
                                                helix_max_radius%value/pixel_size%value), &
                                                finalise=.false.,wipe=.true.,       &
                                                sp_mode=wps_as_sps%value, &
                                                sp_segment_length=sp_segment_length%value/pixel_size%value, &
                                                !sp_helix=helix_average,                                             &
                                                apply_a_bfactor=bfactor_during_3d_reconstruction%value,   &
                                                average_score=average_score%value, &
                                                bfactor_per_subunit=bfactor_per_subunit%value, &
                                                first_frame_for_reconstruction=first_frame_for_reconstruction%value, &
                                                last_frame_for_reconstruction=last_frame_for_reconstruction%value, &
                                                apply_drift_filter=apply_drift_filter%value, &
                                                apply_dose_filter=apply_dose_filter%value)
                else
                    call filament_calc_threed(  fils_final(current_filament),current_micrograph_frames,current_3drec,      &
                                                safe_box_size=fmc_box_size(interpol_scheme, &
                                                helix_max_radius%value/pixel_size%value), &
                                                finalise=.false.,wipe=.true.,       &
                                                sp_mode=wps_as_sps%value, &
                                                sp_segment_length=sp_segment_length%value/pixel_size%value,         &
                                                apply_a_bfactor=bfactor_during_3d_reconstruction%value,   &
                                                average_score=average_score%value, &
                                                bfactor_per_subunit=bfactor_per_subunit%value, &
                                                first_frame_for_reconstruction=first_frame_for_reconstruction%value, &
                                                last_frame_for_reconstruction=last_frame_for_reconstruction%value, &
                                                apply_drift_filter=apply_drift_filter%value, &
                                                apply_dose_filter=apply_dose_filter%value)
                endif
                ! If the user asked for it, finalise this recontruction from the current filament
                ! and write the central slice to disk
                if (single_filament_reconstructions%value) then
                    ! Memory footprint diag
                    if (debug) then
                        write(*,'(a)') '**debug(mode_filament_refine): memory usage before single fil 3d eval'
                        call print_memory_usage_stats()
                    endif
                    write(*,'(a,i0,a)')'**info(mode_filament_refine): finalising single-filament 3d reconstruction for filament ',&
                                        fils_final(current_filament)%id, '.'
                    call threed_eval_reconstruction(current_3drec,current_3dvol,img_ampl=current_3dvol_ampl)
                    ! Memory footprint diag
                    if (debug) then
                        write(*,'(a)') '**debug(mode_filament_refine): memory usage after single fil 3d eval'
                        call print_memory_usage_stats()
                    endif
                    call current_3dvol%Resize(crop_dim)
                    ! Memory footprint diag
                    if (debug) then
                        write(*,'(a)') '**debug(mode_filament_refine): memory usage after single fil crop'
                        call print_memory_usage_stats()
                    endif
                    ! Extract yz slice
                    call current_3dvol%GetSlice(current_3dvol_slice,axis=1,address=current_3dvol%logical_dimensions(1)/2+1)
                    call filename_add_suffix(output_3d_filename%value,'_slice_yz_??????',temp_filename)
                    call filename_from_wild(temp_filename,fils_final(current_filament)%id,temp_filename)
                    print '(2a)', '**info(mode_filament_refine): writing slice to file ', trim(adjustl(temp_filename))
                    call current_3dvol_slice%WriteToDisk(temp_filename,pixel_size=pixel_size%value)
                    ! Extract xy slice
                    call current_3dvol%GetSlice(current_3dvol_slice,axis=3,address=current_3dvol%logical_dimensions(3)/2+1)
                    call filename_add_suffix(output_3d_filename%value,'_slice_xy_??????',temp_filename)
                    call filename_from_wild(temp_filename,fils_final(current_filament)%id,temp_filename)
                    print '(2a)', '**info(mode_filament_refine): writing slice to file ', trim(adjustl(temp_filename))
                    call current_3dvol_slice%WriteToDisk(temp_filename,pixel_size=pixel_size%value)

                    ! Memory footprint diag
                    if (debug) then
                        write(*,'(a)') '**debug(mode_filament_refine): memory usage after single fil 3d slicing'
                        call print_memory_usage_stats()
                    endif

                    ! Work out the crossover direction
                    !call image_helix_mask(current_3dvol,fils_final(current_filament)%helix, &
                    !                        asym_unit_mw%value,pixel_size%value,helix_max_radius%value, &
                    !                        mask_resolution=pixel_size%value/30.0)
                    if (rotate_to_register_crossovers%value) then
                        call cyl_mask(current_3dvol,helix_max_radius%value/pixel_size%value,10,zero_float_inside=.true.)
                        call find_angle_of_crossover(current_3dvol,fils_final(current_filament)%helix, &
                                                        current_filament_crossover_angle)
                        write(*,'(a,i0,a,f0.1,a)') '**info(mode_filament_refine): crossover angle for filament ', &
                                                    fils_final(current_filament)%id, ' is estimated to be ', &
                                                    current_filament_crossover_angle, ' degrees'
                        write(*,'(a)') '**info(mode_filament_refine): rotating filament about its axis to register its crossovers'
                        call filament_shift_all_wps(fils_final(current_filament),   &
                                                    (/0.,0.,0.,-current_filament_crossover_angle,0.,0.,0./),update_model=.true.)
                    endif
                else
                    if (rotate_to_register_crossovers%value) then
                        write(*,'(a)') '**warning(mode_filament_refine): was not able to register crossovers, '//&
                                        'because single_filament_reconstructions flag was F'
                    endif
                endif

                ! Add the current filament's reconstruction to the total reconstruction
                ! only if the filament's score is over the user-supplied threshold
                write(*,'(a,i0,a,F0.3,a)',advance='NO') '**info(mode_filament_refine): score for filament ',    &
                                                        fils_final(current_filament)%id, ' (',                  &
                                                        fils_final(current_filament)%score,') is '
                if (fils_final(current_filament)%score .gt. min_score_fil_to_3d%value) then
                    write(*,'(a)',advance='NO') 'above '
                else
                    write(*,'(a)',advance='NO') 'below '
                endif
                write(*,'(a,F0.3,a)') 'threshold for inclusion into final 3d (',min_score_fil_to_3d%value, ').'

                if (fils_final(current_filament)%score .gt. min_score_fil_to_3d%value) then
                    ! Memory footprint diag
                    if (debug) then
                        write(*,'(a)') '**debug(mode_filament_refine): memory usage before 3d add'
                        call print_memory_usage_stats()
                    endif
                    !call threed_add(new_3drec,current_3drec,tmp_3drec)
                    !call threed_assignment(new_3drec,tmp_3drec)
                    !call threed_deallocate(tmp_3drec)
                    call threed_add(new_3drec,current_3drec,new_3drec)
                    ! Memory footprint diag
                    if (debug) then
                        write(*,'(a)') '**debug(mode_filament_refine): memory usage after 3d add'
                        call print_memory_usage_stats()
                    endif
                endif
            endif ! End of test of whether filament should be reconstructed
        endif ! End of test of whether user asked for a 3d reconstruction

        ! Memory footprint diag
        if (debug) then
            write(*,'(a)') '**debug(mode_filament_refine): memory usage after 3d reconstruction'
            call print_memory_usage_stats()
        endif

        !
        ! Move waypoints to the nearest helical repeats
        !
        if (move_waypoints_to_nearest_repeat%value) then
            write(*,'(a)') '**info(mode_filament_refine): moving output waypoints to the nearest helical subunit'
            call filament_move_wps_to_nearest_subunit(fils_final(current_filament))
        endif

        if (this_program%IsVerbose() .or. debug) write(*,'(a)') '</filament>'

    enddo   ! End of loop over filaments
    if (this_program%IsVerbose() .or. debug) write(*,'(a)') '</loop_over_filaments>'

    ! Memory footprint diag
    if (debug) then
        write(*,'(a)') '**debug(mode_filament_refine): memory usage after loop over filaments'
        call print_memory_usage_stats()
    endif

    !
    ! Write out alignment parameter file
    !
    call filaments_to_frealix_par(fils_final,output_par_filename%value)

    !
    ! Write out alignment parameter file with each asymetric unit
    !
    if (.not. string_equal(output_par_filename_subunits%value,'none')) then
        call filaments_to_frealix_par(fils_final,output_par_filename_subunits%value,subunits=.true.)
    endif

    !
    ! Finalise the 3d reconstruction & write it to disk
    !
    if (calc_3d_reconstruction%value) then
        if (this_program%IsVerbose() .or. debug) write(*,'(a)') '<calc_3d_rec>'
        if (calc_fsc%value) then
            if (this_program%IsVerbose() .or. debug) write(*,'(a)') '<calc_fsc>'
            ! Evaluate / finalise the two half reconstructions
            call threed_eval_reconstruction(new_3drec,new_3dvol(1),rec_num=1,preserve_high_frequencies=.true.,skip_ampl=.true.)
            call threed_eval_reconstruction(new_3drec,new_3dvol(2),rec_num=2,preserve_high_frequencies=.true.,skip_ampl=.true.)
            ! Impose symmetry
            call image_impose_helical_symmetry_new( new_3dvol(1),helix_average,  &
                                                    interpol_scheme,max_radius=helix_max_radius%value / pixel_size%value, &
                                                    progress_report=.true.)
            call image_impose_helical_symmetry_new( new_3dvol(2),helix_average, &
                                                    interpol_scheme,max_radius=helix_max_radius%value / pixel_size%value, &
                                                    progress_report=.true.)
            ! Mask the reconstructions
            call image_helix_mask(new_3dvol(1),helix_average,asym_unit_mw%value,pixel_size%value, &
                                    helix_max_radius%value / pixel_size%value, &
                                    mask_resolution = pixel_size%value / 20.0 )
            call image_helix_mask(new_3dvol(2),helix_average,asym_unit_mw%value,pixel_size%value, &
                                    helix_max_radius%value / pixel_size%value, &
                                    mask_resolution = pixel_size%value / 20.0 )
            ! Dump reconstruction 1 to disk
            call filename_add_suffix(output_3d_filename%value,'_fsc1',temp_filename)
            call new_3dvol(1)%WriteToDisk(temp_filename,pixel_size=pixel_size%value)
            ! Dump reconstruction 2 to disk
            call filename_add_suffix(output_3d_filename%value,'_fsc2',temp_filename)
            call new_3dvol(2)%WriteToDisk(temp_filename,pixel_size=pixel_size%value)
            ! Calculate & printout the fsc
            call new_3dvol(1)%ForwardFFT()
            call new_3dvol(2)%ForwardFFT()
            call compute_fsc(new_3dvol(1),new_3dvol(2),foushell)
            call print_fsc(foushell)
            if (this_program%IsVerbose() .or. debug) write(*,'(a)') '</calc_fsc>'
        endif

        ! Dump the 3d reconstruction to disk, without finalising. The unweighted volume, the CTF^2 volumes are written separately.
        ! (We always write the unfinished 3d to disk, so that we have flexibility later on to subtract from it, for example)
        call threed_write_to_disk(new_3drec,output_3d_filename%value)

        ! If we're finalising, also dump to disk the final reconstruction
        if (finalise_3d_reconstruction%value) then
            if (this_program%IsVerbose() .or. debug) write(*,'(a)') '<finalise_3d>'
            ! Finalise the total recontruction
            if (single_filament_reconstructions%value .and. size(fils) .eq. 1) then
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(2a)') '**info(mode_filament_refine): since we were only refining 1 filament, ',    &
                                    'for which we also already evaluated a 3d reconstruction, we will use this ',   &
                                    'as the output 3d reconstruction, without recalculating one'
                endif
                new_3dvol(1) = current_3dvol
				if (reconstruct_amplitudes%value) new_3dvol_ampl(1) = current_3dvol_ampl
            else
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a,i0,a)') '**info(mode_filament_refine): will finalise the combined 3d reconstruction from the ', &
                                            size(fils), ' filaments'
                endif
                call threed_eval_reconstruction(new_3drec,new_3dvol(1),preserve_high_frequencies=.true.,img_ampl=new_3dvol_ampl(1))
            endif
            ! If microtubule, rebuild the seam
            if (microtubule%value) then
                write(*,'(a)') '**info(mode_filament_refine): imposing microtubule symmetry in real space'
                call image_impose_microtubule_symmetry(fils_final(1),new_3dvol(1))
            endif
            ! Write the final reconstruction to disk
            call new_3dvol(1)%WriteToDisk(output_3d_filename%value,pixel_size=pixel_size%value,print_file_info=.true.)
            if (reconstruct_amplitudes%value) then
                call filename_add_suffix(output_3d_filename%value,'_ampl',output_3d_ampl_filename)
                call new_3dvol_ampl(1)%WriteToDisk(output_3d_ampl_filename,print_file_info=.true.,pixel_size=pixel_size%value)
            endif
!
            !
            if (this_program%IsVerbose() .or. debug) then
                write(*,'(2a)') '**info(mode_filament_refine): preparing a cropped, symmetrised and masked ',    &
                                'version of the final 3d reconstruction'
            endif
            call new_3dvol(1)%Resize(crop_dim)
            call image_impose_helical_symmetry_new(new_3dvol(1),helix_average,interpol_scheme,  &
                                                    max_radius=helix_max_radius%value/pixel_size%value)
            call image_helix_mask(new_3dvol(1),helix_average,asym_unit_mw%value,pixel_size%value,helix_max_radius%value, &
                                mask_resolution=pixel_size%value/20.0e0) !!! todo: compute new estimated resolution for this mask? rather than use input model's resolution
            call filename_add_suffix(output_3d_filename%value,'_hm',output_3d_filename%value)
            call new_3dvol(1)%WriteToDisk(output_3d_filename%value,pixel_size=pixel_size%value,print_file_info=.true.)

            if (this_program%IsVerbose() .or. debug) write(*,'(a)') '</finalise_3d>'
        endif
        if (this_program%IsVerbose() .or. debug) write(*,'(a)') '</calc_3d_rec>'
    endif


    ! Cleanup
    if (allocated(fils)) then
        do i=1,size(fils)
            call fils(i)%Deallocate()
        enddo
        deallocate(fils)
    endif

    ! Deallocate image objects (i don't think this should be necessary,
    ! but it seems ifort is not runnin the finalisation subroutines as i
    ! expected)
    call new_3dvol(1)%Deallocate()
    call new_3dvol(2)%Deallocate()
    call model_3dvol%Deallocate()
    call current_3dvol%Deallocate()
    call current_3dvol_slice%Deallocate()
    do current_frame=1,number_of_frames_per_micrograph%value
        call current_micrograph_frames(current_frame)%Deallocate()
    enddo
    ! Deallocate threed objects
    call threed_deallocate(current_3drec)
    call threed_deallocate(new_3drec)
    !
    call fmc_deallocate(fmc)

    ! Memory footprint diag
    if (debug) then
        write(*,'(a)') '**debug(mode_filament_refine): memory usage on exit'
        call print_memory_usage_stats()
    endif

end subroutine mode_filament_refine_slave


!>  \brief  Find suitable dimensions for the image so that it can be resampled in both (or all three) dimensions using Fourier cropping.
!!          The assumption is that the image will first be padded in real space before being cropped in Fourier space to achieve the exact
!!          downsampling desired
subroutine find_dimensions_for_resampling(self,desired_scaling_factor,padding_dimensions,cropping_dimensions, &
                                          actual_scaling_factor)
    use UsefulFunctions, only : IsOdd, IsEven
    use Images
    !
    type(Image),    intent(in)  ::  self
    real,           intent(in)  ::  desired_scaling_factor
    integer,        intent(out) ::  padding_dimensions(3)
    integer,        intent(out) ::  cropping_dimensions(3)
    real,           intent(out) ::  actual_scaling_factor       !<  multiply old dimensions by this to get new dimensions
    ! private variables
    real,           parameter   ::  tolerance = 0.0001 ! tolerance for the scaling factor
    !real                        ::  current_error(3)
    integer                     ::  current_dimension
    integer                     ::  number_of_dimensions
    integer                     ::  maximum_size_per_dimension
    ! start work


    padding_dimensions = self%logical_dimensions
    cropping_dimensions = self%logical_dimensions
    actual_scaling_factor = desired_scaling_factor

    ! How many dimensions?
    number_of_dimensions = 3
    if (self%logical_dimensions(3) .eq. 1) number_of_dimensions = 2

    ! Maximum size per dimension
    maximum_size_per_dimension = (huge(1)/8)**(1/number_of_dimensions)

    !
    ! We treat each dimension in turn
    !
    do current_dimension=1,number_of_dimensions
        do
            cropping_dimensions(current_dimension) = nint(real(padding_dimensions(current_dimension)) * desired_scaling_factor)
            if (IsEven(self%logical_dimensions(current_dimension)) .and. IsOdd(cropping_dimensions(current_dimension))) then
                cropping_dimensions(current_dimension) = cropping_dimensions(current_dimension) + 1
            endif
            actual_scaling_factor = real(cropping_dimensions(current_dimension))/real(padding_dimensions(current_dimension))
            if (abs( actual_scaling_factor - desired_scaling_factor) .lt. tolerance) then
                ! we're done
                exit
            else
                if (padding_dimensions(current_dimension) .le. maximum_size_per_dimension-2) then
                    padding_dimensions(current_dimension) = padding_dimensions(current_dimension) + 2 ! we add a pixel on either end of that dimension
                else
                    exit ! We can't go any larger
                endif
            endif
        enddo
    enddo

end subroutine find_dimensions_for_resampling

end module mode_filament_refine
