!>  \brief  Combine 3d reconstructions obtained from subsets of the input data into a single 3d reconstruction.
!!          This mode is meant for use with frealix2, whereas mode COMBINE_3Ds is meant for use with frealix1 & chuff.
module mode_finalise_3ds
    use modes
    use rtps
    implicit none


    type(rtp_char), target  ::  input_threed_files  =               &
        rtp_char(   value='threed_???.mrc',                  &
                    keyword='INPUT_THREED_FILES',           &
                    description='Root filename (with wildcard) for input 3d reconstruction files')


    type(rtp_char), target  ::  output_threed_filename  =               &
        rtp_char(   value='threed.mrc',                  &
                    keyword='OUTPUT_THREED_FILENAME',           &
                    description='Output 3d reconstruction filename')


    type(rtp_char), target  ::  input_threed_selection  =               &
        rtp_char(   value='1-100',                  &
                    keyword='INPUT_THREED_SELECTION',           &
                    description='Selection of 3d reconstructions to be summed & finalised')


    type(rtp_logi), target  ::  calc_fsc  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='CALC_FSC',           &
                    description='Whether to calculate an fsc between halves of the dataset')


    type(rtp_logi), target  ::  apply_fom_filter  =               &
        rtp_logi(   value=.true.,                  &
                    keyword='APPLY_FOM_FILTER',           &
                    description='Whether a Figure-Of-Merit filter should be applied to the final reocnstruction')


    type(rtp_char), target  ::  interpol_function  =               &
        rtp_char(   value='LINEAR',                  &
                    keyword='INTERPOL_FUNCTION',           &
                    description='Which interpolation function to use')


    type(rtp_char), target  ::  window_function  =               &
        rtp_char(   value='RECT',                  &
                    keyword='WINDOW_FUNCTION',           &
                    description='Which windowing function to use')


    type(rtp_intg), target  ::  window_width  =               &
        rtp_intg(   value=2,                  &
                    keyword='WINDOW_WIDTH',           &
                    description='Window width for interpolation (kernel support)')


    type(rtp_logi), target  ::  helical_masking  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='HELICAL_MASKING',           &
                    description='')


    type(rtp_real), target  ::  twist_per_subunit  =               &
        rtp_real(   value=0.0,                  &
                    keyword='TWIST_PER_SUBUNIT',           &
                    description='')


    type(rtp_real), target  ::  rise_per_subunit  =               &
        rtp_real(   value=0.0,                  &
                    keyword='RISE_PER_SUBUNIT',           &
                    description='')


    type(rtp_intg), target  ::  symmetry_axial  =               &
        rtp_intg(   value=1,                  &
                    keyword='SYMMETRY_AXIAL',           &
                    description='')


    type(rtp_intg), target  ::  symmetry_perp  =               &
        rtp_intg(   value=1,                  &
                    keyword='SYMMETRY_PERP',           &
                    description='')

    type(rtp_logi), target  ::  stack  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='STACK',           &
                    description='The assembly is a stack rather than a true helix')


    type(rtp_real), target  ::  helix_max_radius  =               &
        rtp_real(   value=0.0,                  &
                    keyword='HELIX_MAX_RADIUS',           &
                    description='Maximum outer radius of the helix, in Angstroms')


    type(rtp_real), target  ::  asym_unit_mw  =               &
        rtp_real(   value=0.0,                  &
                    keyword='ASYM_UNIT_MW',           &
                    description='Expected molecular weight (Da) of the asymetric unit')


    type(rtp_real), target  ::  pixel_size  =               &
        rtp_real(   value=1.0,                  &
                    keyword='PIXEL_SIZE',           &
                    description='Pixel size (Angstrom)')


    type(rtp_char), target  ::  model_3d_fsc_filename  =               &
        rtp_char(   value='model3d_fsc.txt',                  &
                    keyword='MODEL_3D_FSC_FILENAME',           &
                    description='FSC corresponding to the model 3d reconstruction. '//&
                                'This should be output by Frealix at the end of the previous round.')


contains
subroutine mode_finalise_3ds_init_and_run
    implicit none
    type(mode) :: finalise_3ds

    ! Intialise the mode
    call finalise_3ds%init(command='FINALISE_3DS',subroutine=mode_finalise_3ds_slave,short_description='Add description here')

    ! Add RTPs to the mode
    call finalise_3ds%add_rtp(input_threed_files)
    call finalise_3ds%add_rtp(output_threed_filename)
    call finalise_3ds%add_rtp(input_threed_selection)
    !call finalise_3ds%add_rtp(calc_fsc)
    !call finalise_3ds%add_rtp(apply_fom_filter)
    !call finalise_3ds%add_rtp(interpol_function)
    !call finalise_3ds%add_rtp(window_function)
    !call finalise_3ds%add_rtp(window_width)
    call finalise_3ds%add_rtp(helical_masking)
    call finalise_3ds%add_rtp(twist_per_subunit)
    call finalise_3ds%add_rtp(rise_per_subunit)
    call finalise_3ds%add_rtp(symmetry_axial)
    call finalise_3ds%add_rtp(symmetry_perp)
    call finalise_3ds%add_rtp(stack)
    call finalise_3ds%add_rtp(helix_max_radius)
    call finalise_3ds%add_rtp(asym_unit_mw)
    call finalise_3ds%add_rtp(pixel_size)
    call finalise_3ds%add_rtp(model_3d_fsc_filename)

    ! Run the mode
    call finalise_3ds%run()
end subroutine mode_finalise_3ds_init_and_run

subroutine mode_finalise_3ds_slave
    use runtime_parameters      ! This gives us access to user-defined or default values for runtime parameters
    use parsers                 ! This contains the parser for lists of files
    use threeds
    use fscs
    use images_comparison
    use helices
    use images_helices
    use interpolation
    use imagefiles
    use DatesAndTimes
    use Curves
    implicit none
    ! Private variables
    integer,            allocatable ::  list_of_threeds(:)
    integer                         ::  i, j, threed_num, central_slice
    character(len=filename_len)     ::  input_threed_base_filename, tmp_filename, output_ampl_filename
    type(threed)                    ::  input_threed, threed_sum(2)
    type(image)                     ::  threed_final(2), central_slices(2), ampl_final(2)
    type(image)                     ::  threed_final_no_fom
    real,   allocatable             ::  foushell(:)
    type(helical_symmetry)          ::  helix
    logical,    parameter           ::  helical_symmetrise  =   .true.
    type(interpolation_scheme)      ::  interpol_scheme
    logical,            parameter   ::  debug   =   .false.
    integer                         ::  crop_dim(3)
    integer                         ::  input_threed_num_threeds, num_sum_threeds
    type(image)                     ::  fsc_volumes(2), fsc_volumes_raw(2), fsc_volumes_no_mask_fom(2)
    type(point_symmetry_symbol)     ::  sym_symbol
    character(len=6)                ::  size_list_c
    type(DateAndTime)               ::  myTime
    type(Curve)                     ::  fsc_curve
    type(Curve)                     ::  model_3d_fsc
    real                            ::  estimated_resolution_of_model_3d
    logical                         ::  reconstruct_amplitudes
    ! Get runtime parameters from control file


    !
    ! Estimated resolution of input model 3d
    !
    if (model_3d_fsc_filename%set_by_user .and. file_exists(model_3d_fsc_filename%value)) then
        call fsc_curve_read_from_disk(model_3d_fsc,model_3d_fsc_filename%value)
        estimated_resolution_of_model_3d = fsc_curve_last_frequency_with_fsc_above_threshold(model_3d_fsc,0.5) ! We use 0.5 to get an understimate of the resolution
        if (estimated_resolution_of_model_3d .gt. 0.0) then
            estimated_resolution_of_model_3d = 1.0e0/estimated_resolution_of_model_3d
        else
            estimated_resolution_of_model_3d = 20.0e0
            write(*,'(2a)') '**warning(mode_finalise_3ds): something went wrong when estimating the resolution from ', &
                            model_3d_fsc_filename%value
        endif
        write(*,'(a,f0.1,a)') '**info(mode_finalise_3ds): model 3d has a resolution of ~', estimated_resolution_of_model_3d, 'A'
    else
        estimated_resolution_of_model_3d = 20.0e0 !pixel_size%value * 10.0e0
        write(*,'(a,f0.1)') '**info(mode_finalise_3ds): no input fsc curve. Will assume the model 3d has a resolution of ',  &
                        estimated_resolution_of_model_3d
    endif


    ! Work out the list of 3d reconstructions
    list_of_threeds = set_to_list(input_threed_selection%value)
    write(size_list_c,'(i6)') size(list_of_threeds)
#ifdef nagfor_compiler
    if (size(list_of_threeds) .lt. 100) then
#endif
    print '(a,i0,a,'//size_list_c//'(i0,1x))', '**info(mode_finalise_3ds): The following ', size(list_of_threeds), &
                                                    ' 3d reconstructions are selected: ', list_of_threeds
#ifdef nagfor_compiler
    endif
#endif

    ! Initialise point symmetry
    sym_symbol%symbol = 'H'

    ! Initialise helical parameters
    call helical_symmetry_init( helix,rise_per_subunit%value/pixel_size%value,twist_per_subunit%value,    &
                                helix_max_radius%value/pixel_size%value, &
                                symmetry_axial%value,symmetry_perp%value, &
                                stack=stack%value)

    ! Intialise interpolation scheme in case it is needed
    call interpolation_scheme_init(interpol_scheme,'LINEAR','RECT',2)


    !!$omp parallel default(shared) &
    !!$omp private(threed_num,input_threed_base_filename,input_threed,input_threed_num_threeds) &
    !!$omp private(i,j)

    ! Initialise a 3d reconstruction object for the input 3Ds
    threed_num = list_of_threeds(1)
    call filename_from_wild(input_threed_files%value,threed_num,input_threed_base_filename)
    !!$omp single
    reconstruct_amplitudes = threed_on_disk_ampl_vol(input_threed_base_filename)
    !!$omp end single
    call threed_init(       input_threed,                                           &
                            filename_root       =   input_threed_base_filename,     &
                            interpol_descr      =   interpol_function%value,              &
                            window_descr        =   window_function%value,                &
                            window_width        =   window_width%value,                   &
                            sym_symbol          =   sym_symbol,                     &
                            helix               =   helix,                          &
                            asym_unit_mw        =   asym_unit_mw%value,                   &
                            pixel_size          =   pixel_size%value,                     &
                            helix_max_rad       =   helix_max_radius%value/pixel_size%value,    &
                            reconstruct_amplitudes =  reconstruct_amplitudes, &
                            estimated_resolution = estimated_resolution_of_model_3d  )




    !!$omp single
    input_threed_num_threeds = threed_num_of_recs(input_threed)

    ! How many sums do we need?
    if (size(list_of_threeds) .gt. 1) then
        if (input_threed_num_threeds .eq. 1) then
            ! There are multiple input 3Ds, each with 1 internal 3d
            if (calc_fsc%value .or. apply_fom_filter%value) then
                ! We need 2 running sums
                num_sum_threeds = 2
            else
                num_sum_threeds = 1
            endif
        else
            ! There are multiple input 3Ds, each with 2 internal 3Ds
            num_sum_threeds = 1
        endif
    else
        if (input_threed_num_threeds .eq. 1) then
            ! There is only one input 3d, and it only has 1 internal 3d
            num_sum_threeds = 1
            if (calc_fsc%value .or. apply_fom_filter%value) then
                ! fsc/fom not possible if only one 3d
                write(*,'(a)') '**info(mode_finalise_3ds): only one 3d reconstruction, therefore no fsc or fom filter'
                calc_fsc%value = .false.
                apply_fom_filter%value = .false.
            endif
        else
            ! There is only one input 3d, with 2 internal 3Ds
            num_sum_threeds = 1
        endif
    endif

    ! Initialise the "sum" 3Ds
    do j=1,num_sum_threeds
        threed_sum(j)%logical_dimensions = input_threed%logical_dimensions
        call threed_init(   threed_sum(j),                                          &
                            interpol_descr      =   interpol_function%value,              &
                            window_descr        =   window_function%value,                &
                            window_width        =   window_width%value,                   &
                            sym_symbol          =   sym_symbol,                     &
                            helix               =   helix,                          &
                            asym_unit_mw        =   asym_unit_mw%value,                   &
                            pixel_size          =   pixel_size%value,                     &
                            helix_max_rad       =   helix_max_radius%value/pixel_size%value,    &
                            num_of_recs         =   input_threed_num_threeds,       &
                            fom_filter          =   apply_fom_filter%value,&! .AND. (num_sum_threeds .EQ. 1),    &
                            compute_fsc         =   calc_fsc%value,    &! .AND. (num_sum_threeds .EQ. 1)             &
                            reconstruct_amplitudes = input_threed%reconstruct_amplitudes, &
                            estimated_resolution = estimated_resolution_of_model_3d)
        call threed_wipe(threed_sum(j))
    enddo
    !!$omp end single

    ! Loop over the 3Ds and sum them
    !!$omp do schedule(static,1)
    do i=1,size(list_of_threeds)
        threed_num = list_of_threeds(i)

        ! Read in the 3d reconstruction
        call filename_from_wild(input_threed_files%value,threed_num,input_threed_base_filename)

        !$omp critical (finalise_read_from_disk)
        write(*,'(2a,3(a,i0),2a)') '**info(mode_finalise_3ds): ', myTime%DateAndTimeAsString(), ': adding 3d ', threed_num, &
                                    ' (', i, ' of ', size(list_of_threeds), '): ', trim(input_threed_base_filename)
        call threed_read_from_disk(input_threed,input_threed_base_filename)
        if (debug) write(*,'(3a)') '**debug(mode_finalise_3ds): ', myTime%DateAndTimeAsString(), &
                                    ': reading from disk completed. About to add.'
        !$omp end critical (finalise_read_from_disk)

        ! If we're accumulating 2 sum reconstructions, the even numbered 3Ds go into a second volume
        j = 1
        if (num_sum_threeds .eq. 2) then
            if (mod(i,2) .eq. 0) then
                j = 2
            endif
        endif
        ! Accumulate sums of the volumes
        !!$omp critical (finalise_add)
        call threed_add(threed_sum(j),input_threed,threed_sum(j))
        !!$omp end critical (finalise_add)
    enddo
    !!$omp enddo

    call threed_deallocate(input_threed)

    !!$omp end parallel

    ! We can now finalise the 3d
    if (num_sum_threeds .eq. 2) then
        ! There were 2 sums collected. They should each contain only one reconstruction
        if (input_threed_num_threeds .eq. 1) then
            write(*,'(a)') '**info(mode_finalise_3ds): about to merge the two half-dataset reconstructions'
            write(*,'(a,2(i0,3x))') '**info(mode_finalise_3ds): number of contributing elements: ', &
                                                            threed_sum(1)%rec(1)%num_contrib_elems, &
                                                            threed_sum(2)%rec(1)%num_contrib_elems
            call threed_merge(threed_sum(1),threed_sum(2))
        else
            write(*,'(a)') '**error(mode_finalise_3ds): programming error'
            call terminate('Fatal error in MODE_FINALISE_3DS')
        endif
    endif
    if (calc_fsc%value) then
        call threed_eval_reconstruction(threed_sum(1),threed_final(1),img_no_fom=threed_final_no_fom, &
                                        fsc_volumes=fsc_volumes,img_ampl=ampl_final(1), &
                                        fsc_volumes_no_mask=fsc_volumes_raw, &
                                        fsc_volumes_no_mask_fom=fsc_volumes_no_mask_fom)
    else
        call threed_eval_reconstruction(threed_sum(1),threed_final(1),img_no_fom=threed_final_no_fom, &
                                        img_ampl=ampl_final(1))
    endif

    ! Deal with fsc
    if (calc_fsc%value) then
        ! Write the 2 FSC volumes to disk
        call filename_add_suffix(output_threed_filename%value,'_fsc1',tmp_filename)
        print '(2a)', '**info(mode_finalise_3ds): writing fsc volume to ', trim(tmp_filename)
        call fsc_volumes(1)%WriteToDisk(tmp_filename)

        call filename_add_suffix(output_threed_filename%value,'_fsc2',tmp_filename)
        print '(2a)', '**info(mode_finalise_3ds): writing fsc volume to ', trim(tmp_filename)
        call fsc_volumes(2)%WriteToDisk(tmp_filename)

        ! Write the 2 FSC volumes to disk, without masking or helical symmetrysation
        call filename_add_suffix(output_threed_filename%value,'_fsc1_no_mask_no_hsym',tmp_filename)
        print '(2a)', '**info(mode_finalise_3ds): writing fsc volume (without masking or symmetrisation) to ', trim(tmp_filename)
        call fsc_volumes_raw(1)%WriteToDisk(tmp_filename)

        call filename_add_suffix(output_threed_filename%value,'_fsc2_no_mask_no_hsym',tmp_filename)
        print '(2a)', '**info(mode_finalise_3ds): writing fsc volume (without masking or symmetrisation) to ', trim(tmp_filename)
        call fsc_volumes_raw(2)%WriteToDisk(tmp_filename)

        ! Write the 2 FSC volumes to disk, without masking or helical symmetrysation, but with FOM filtering
        call filename_add_suffix(output_threed_filename%value,'_fsc1_no_mask_no_hsym_fom',tmp_filename)
        print '(2a)', '**info(mode_finalise_3ds): writing fsc volume (without masking or symmetrisation) to ', trim(tmp_filename)
        call fsc_volumes_no_mask_fom(1)%WriteToDisk(tmp_filename)

        call filename_add_suffix(output_threed_filename%value,'_fsc2_no_mask_no_hsym_fom',tmp_filename)
        print '(2a)', '**info(mode_finalise_3ds): writing fsc volume (without masking or symmetrisation) to ', trim(tmp_filename)
        call fsc_volumes_no_mask_fom(2)%WriteToDisk(tmp_filename)

        ! Print out the fsc
        call threed_print_fsc(threed_sum(1))

        ! Do an frc between the central slices
        central_slice = threed_final(1)%GetPhysicalIndexOfBoxCenter(3)
        call fsc_volumes(1)%GetSlice(central_slices(1),3,central_slice)
        call fsc_volumes(2)%GetSlice(central_slices(2),3,central_slice)
        call compute_fsc(central_slices(1),central_slices(2),foushell)
        write(*,'(a)') '**info(mode_finalise_3ds): frc between central slices of two half-dataset reconstructions follows'
        call print_fsc(foushell,fourier_bin_size=(1.0/fsc_volumes(1)%logical_dimensions(1)/pixel_size%value))
        call fsc_to_fsc_curve(foushell,fsc_curve,pixel_size%value)
    endif


    ! We can write it to disk
    call threed_write_to_disk(threed_sum(1),output_threed_filename%value)
    call threed_final(1)%WriteToDisk(output_threed_filename%value,pixel_size=pixel_size%value,print_file_info=.true.)
    if (input_threed%reconstruct_amplitudes) then
        call filename_add_suffix(output_threed_filename%value,'_ampl',output_ampl_filename)
        call ampl_final(1)%WriteToDisk(output_ampl_filename,pixel_size=pixel_size%value,print_file_info=.true.)
    endif
    call filename_add_suffix(output_threed_filename%value,'_no_filter',output_ampl_filename)
    call threed_final_no_fom%WriteToDisk(output_ampl_filename,pixel_size=pixel_size%value,print_file_info=.true.)

    ! Let's also prepare a version which is cropped, symmetrised & masked
    write(*,'(a)') '**info(mode_finalise_3ds): preparing a cropped, symmetrised and masked version of the final 3d reconstruction'
    crop_dim = nint(2.*1.1*helix_max_radius%value/pixel_size%value)
    if (is_odd(crop_dim(1))) crop_dim = crop_dim + 1 ! Not necessary for Frealix, but may be so for other software
    call threed_final(1)%Resize(crop_dim)
    call image_impose_helical_symmetry_new(threed_final(1),helix,interpol_scheme,max_radius=helix_max_radius%value/pixel_size%value)
    if (helical_masking%value) then
        call image_helix_mask(threed_final(1),helix,asym_unit_mw%value,pixel_size%value,helix_max_radius%value/pixel_size%value, &
                              mask_resolution=pixel_size%value/estimated_resolution_of_model_3d)
        call filename_add_suffix(output_threed_filename%value,'_hm',output_threed_filename%value)
    else
        call filename_add_suffix(output_threed_filename%value,'_h',output_threed_filename%value)
    endif
    call threed_final(1)%WriteToDisk(output_threed_filename%value,pixel_size=pixel_size%value)

end subroutine mode_finalise_3ds_slave

end module mode_finalise_3ds
