module mode_frealign_stack_to_frealix_ali
    use modes
    use rtps
    implicit none
    ! Declare the mode
    type(mode)  ::  m

    ! Declare run-time parameters
    type(rtp_char), target      ::  frealign_par_filename   =   rtp_char(   &
                                    value='frealign.par',                   &
                                    keyword='FREALIGN_PAR_FILENAME',        &
                                    description='Name of input frealign parameter file')

    type(rtp_char), target      ::  stack_filename          =   rtp_char(   &
                                    value='stack.mrc',                      &
                                    keyword='STACK_FILENAME',               &
                                    description='The stack of 2d images processed by Frealign')

    type(rtp_char), target      ::  film_filenames          =   rtp_char(   &
                                    value='film_????.mrc',                  &
                                    keyword='FILM_FILENAMES',               &
                                    description='Pattern for film filenames. Question marks replace single characters.')

    type(rtp_char), target      ::  film_selection          =   rtp_char(   &
                                    value='1,6,7-9',                        &
                                    keyword='FILM_SELECTION',               &
                                    description='Which films to analyse. Comma-separated list where - indicates range.')

    type(rtp_char), target      ::  list_of_film_numbers    =   rtp_char(   &
                                    value='none',                  &
                                    keyword='list_of_film_numbers',               &
                                    description='Flat text file. One line per particle, with micrograph number.')

    type(rtp_char), target      ::  ali_filename            =   rtp_char(   &
                                    value='waypoints.ali',                  &
                                    keyword='ALI_FILENAME',                 &
                                    description='Output alignment file name')

    type(rtp_real), target      ::  pixel_size              =   rtp_real(   &
                                    value=1.0,                              &
                                    keyword='PIXEL_SIZE',                   &
                                    description='Pixel size of the Frealign parameter file, in Angstroms')

    type(rtp_intg), target      ::  y_max                   =   rtp_intg(   &
                                    value=0,                                &
                                    keyword='Y_MAX',                        &
                                    description='Maximum y address at which to look for single particle. '//&
                                    'Useful to avoid film label.')

    type(rtp_intg), target      ::  binning_factor          =   rtp_intg(   &
                                    value=1,                                &
                                    keyword='BINNING_FACTOR',               &
                                    description='Images (stack & films) are binned by this factor relative to the '//&
                                    'Frealign parameter file')

    type(rtp_logi), target      ::  check_180_rotation      =   rtp_logi(   &
                                    value=.false.,                          &
                                    keyword='CHECK_180_ROTATION',           &
                                    description='For each image, look for its 180-degree-rotated version in the film images also')

contains

subroutine mode_frealign_stack_to_frealix_ali_init_and_run
    implicit none
    ! Initialise the mode
    call m%init(    command='FREALIGN_STACK_TO_FREALIX_ALI',                &
                    subroutine=mode_frealign_stack_to_frealix_ali_slave,    &
                    short_description='Convert alignment parameters from Frealign to Frealix')

    ! Add RTPs to the mode
    call m%add_rtp(frealign_par_filename)
    call m%add_rtp(stack_filename)
    call m%add_rtp(film_filenames)
    call m%add_rtp(film_selection)
    call m%add_rtp(list_of_film_numbers)
    call m%add_rtp(ali_filename)
    call m%add_rtp(pixel_size)
    !call m%add_rtp(y_max)
    call m%add_rtp(binning_factor)
    call m%add_rtp(check_180_rotation)
    ! Run the mode
    call m%run()
end subroutine mode_frealign_stack_to_frealix_ali_init_and_run


subroutine mode_frealign_stack_to_frealix_ali_slave
    use single_particles, only : single_particle
    use frealign_parameters
    use frealix_parameters
    use parsers
    use runtime_parameters, only : filename_len
    use images
    use images_arithm
    use Peaks
    use ui, only : interval_print
    use filaments
    use rotation_matrices, only : euler_to_rotmat
    use interpolation
    use ImageFiles
    use InterpolatingWindows
    use ctfs, only : z_from_ctf, ctf_ang_to_pix
    use coos, only : waypoints_grow_array, wp_print_info
    use ProgressBars
	implicit none
    ! Private variables
    type(single_particle),  allocatable ::  sp(:)
    integer,                allocatable ::  list_of_films(:)
    integer                             ::  i, j, k
    integer                             ::  film_num
    character(len=filename_len)         ::  film_filename
    type(image)                         ::  film_img
    integer                             ::  current_sp
    type(imagefile)                     ::  sp_imagefile, film_imagefile
    type(image)                         ::  sp_img
    type(image)                         ::  tmp_img
    type(image)                         ::  corr
    type(peak)                          ::  best_peak
    type(peak),             allocatable ::  peaks_found(:,:,:)  ! First dimension: location in stack, second: film index, third: 180-deg rotation (1 if not, 2 if so)
    logical,                parameter   ::  debug   =   .false.
    type(image)                         ::  film_local_average
    type(image)                         ::  film_local_std_dev
    type(image)                         ::  film_squared
    integer                             ::  sp_dim(3), film_ori(3)
    integer                             ::  found_film_index
    integer                             ::  number_of_filaments, current_filament
    type(filament),         allocatable ::  fils(:)
    type(waypoint)                      ::  current_wp
    type(waypoint),         allocatable ::  wps(:)
    logical                             ::  new_filament
    logical                             ::  last_wp_in_filament
    integer                             ::  rotations_to_check, current_rotation, best_rotation
    real                                ::  rotmat(4,4)
    type(image)                         ::  tmpimg
    type(interpolation_scheme)          ::  interpol
    type(window)                        ::  interpol_window
    type(ProgressBar)                   ::  my_progress_bar
    integer,    allocatable             ::  film_number(:)
    logical                             ::  sp_may_be_on_current_micrograph
    logical,                parameter   ::  extract_box_from_micrograph = .false.
    integer                             ::  box_coos(3)
    ! Start work

    ! Grab values from the input Frealign parameter file
    sp = frealign_par_to_sp(frealign_par_filename%value)

    ! Work out how many filaments are described (Frealign 8.09 stores
    ! filament number in the film column)
    number_of_filaments = 1
    do current_sp=2,size(sp)
        if (sp(current_sp)%film_num .ne. sp(current_sp-1)%film_num) number_of_filaments = number_of_filaments + 1
    enddo
    write(*,'(a,i0)') '**info(mode_frealign_stack_to_frealix_ali): number of filaments = ', number_of_filaments
    allocate(fils(number_of_filaments))

    ! Work out how many rotations we'll check
    rotations_to_check = 1
    if (check_180_rotation%value) rotations_to_check=2
    rotmat = euler_to_rotmat(0.0D0,0.0D0,180.0D0,reverse=.true.)

    ! Initialise interpolation
    call interpolation_scheme_init(interpol,'LINEAR','RECT',2)


    ! Open the stack file
    call sp_imagefile%init(filename=stack_filename%value)

    !
    list_of_films = set_to_list(film_selection%value)

    allocate(peaks_found(size(sp),size(list_of_films),rotations_to_check))

    !
    ! Parse list of film numbers
    !
    if (.not. string_equal(list_of_film_numbers%value,'none')) then
        call get_ints_from_column(list_of_film_numbers%value,1,film_number)
        if (size(film_number) .lt. size(sp)) then
            write(*,'(a,i0,3a,i0)') '**error(mode_frealign_stack_to_frealix_ali): expected ', size(sp), ' lines in ', &
                                    trim(list_of_film_numbers%value), ' but only found ', size(list_of_films)
            call terminate('Fatal error in mode_frealign_stack_to_frealix_ali')
        endif
    endif

    ! Progress bar
    if (allocated(film_number)) then
        call my_progress_bar%Begin(size(sp)*rotations_to_check)
    else
        call my_progress_bar%Begin(size(list_of_films)*size(sp)*rotations_to_check)
    endif

    !
    ! Loop over films / micrographs, and over single particles
    ! and find the best correlation peak between all combinations
    !
    do i=1,size(list_of_films)

        if (debug) write(*,'(2(A,i0))') '**info(mode_frealign_stack_to_frealix_ali): on film ', i, ' of ', size(list_of_films)

        film_num = list_of_films(i)

        ! Check the micrograph file exists
        call filename_from_wild(film_filenames%value,film_num,film_filename)
        if (.not. file_exists(film_filename)) then
            if (debug) write(*,'(2a)') '**debug(mode_frealign_stack_to_frealix_ali): file does not exist: ', film_filename
            cycle
            !call terminate('Fatal error in MODE_FREALIGN_STACK_TO_FREALIX_ALI')
        endif

        ! Read the micrograph into memory
        if (debug) write(*,'(a)') '**info(mode_frealign_stack_to_frealix_ali): reading film into memory'
        call film_imagefile%init(film_filename)
        if (debug) call film_imagefile%PrintInfo()
        call film_img%ReadFromDisk(film_imagefile)
        call film_imagefile%close()
        if (film_img%is_in_real_space .and. film_img%object_is_centered_in_box) then
            film_ori = film_img%physical_address_of_box_center
        else
            film_ori = [1,1,1]
        endif

		! Pre-process the micrograph
		!call film_img%ApplySoftRectangularMask(real(film_img%logical_dimensions(1:2)-26),    &
		!                 azimuth=0.0,falloff=10.0,wipe_input=.false.,mask_with_average=.true.)
		if (y_max%set_by_user .and. y_max%value .ne. 0 .and. y_max%value .lt. film_img%logical_dimensions(2)) then
			film_img%real_values(:,y_max%value:film_img%logical_dimensions(2),:) = film_img%GetAverageOfValues()
		endif
		if (debug) call film_img%WriteToDisk('dbg_film_rect.mrc')

		! Compute the square of the film image
		call film_squared%Allocate(film_img) ! We will need this later
        film_squared%real_values = film_img%real_values ** 2

        if (debug) then
            write(*,'(2a)') '**info(mode_frealign_stack_to_frealix_ali): computing the micrograph''s local', &
                                                    ' average and standard deviation'
        endif

        ! Prepare the film & its squared version (ft, shift origin)
        call film_img%ForwardFFT()
        call film_squared%ForwardFFT()
        call film_img%SwapRealSpaceQuadrants()
        call film_squared%SwapRealSpaceQuadrants()

        ! Compute the local average at each point on the film
        ! (where "local" is defined by the mask over the single particles)
        ! (note: we are going to compute the local normalised correlation coefficient, as
        ! described in Roseman, 2003)

        ! First, we create an image of the "mask", in our case the full box
        call sp_img%ReadFromDisk(sp_imagefile,wanted_location=1)
        sp_img%real_values = 1.0
        sp_img%object_is_centered_in_box = .true.
        sp_dim = sp_img%logical_dimensions

        ! Prepare the mask for convolution with the film
        call sp_img%Resize(film_img%logical_dimensions)
        call sp_img%ForwardFFT()
        call sp_img%SwapRealSpaceQuadrants()

        ! Compute the convolution
        if (debug) write(*,'(a)') '**debug(mode_frealign_stack_to_frealix_ali): about to do convolution'
        call film_local_average%Allocate(film_img)
        film_local_average%complex_values = film_img%complex_values * sp_img%complex_values

        ! Scale to give local average
        if (debug) write(*,'(a)') '**debug(mode_frealign_stack_to_frealix_ali): about to compute average'
        call film_local_average%SwapRealSpaceQuadrants()
        call film_local_average%BackwardFFT()
        film_local_average%real_values = film_local_average%real_values * &
                                        real(product(film_img%logical_dimensions)) / real(product(sp_dim)) ! multiplication by film%logical_dimensions is to get correct scaling of the convolution
        if (debug) then
            call film_local_average%WriteToDisk('dbg_film_local_average.mrc')
        endif

        ! Compute the local standard deviation at each point on the film
        if (debug) write(*,'(a)') '**debug(mode_frealign_stack_to_frealix_ali): about to compute std dev'
        call film_local_std_dev%Allocate(film_img)
        film_local_std_dev%complex_values = film_squared%complex_values * sp_img%complex_values
        call film_local_std_dev%SwapRealSpaceQuadrants()
        call film_local_std_dev%BackwardFFT()
        film_local_std_dev%real_values = film_local_std_dev%real_values * &
                                        real(product(film_img%logical_dimensions)) / real(product(sp_dim)) &
                                        - film_local_average%real_values
        if (debug) write(*,'(a)') '**debug(mode_frealign_stack_to_frealix_ali): about to do sqrt'
        do k=1,film_local_std_dev%GetLogicalDimension(2)
            do j=1,film_local_std_dev%GetLogicalDimension(1)
                if (film_local_std_dev%real_values(j,k,1) .gt. 0.0) then
                    film_local_std_dev%real_values(j,k,1) = sqrt(film_local_std_dev%real_values(j,k,1))
                else
                    film_local_std_dev%real_values(j,k,1) = 0.0
                endif
            enddo
        enddo
        if (debug) then
            call film_local_std_dev%WriteToDisk('dbg_film_local_std_dev.mrc')
        endif
        call film_squared%Deallocate()  ! Won't need this anymore

        call sp_img%Deallocate()   ! This will be re-allocated later, in OpenMP parallel region

        !$omp   parallel            &
        !$omp   default(shared)           &
        !$omp   private(current_sp,sp_img,corr,current_rotation,tmp_img,tmpimg,interpol_window,sp_may_be_on_current_micrograph)

        call corr%Reset()
        call sp_img%Reset()
        call tmp_img%Reset()
        call tmpimg%Reset()
        call interpol_window%AllocateWindow(interpol%window_width,comp=.false.,threed=.false.)

        ! Loop over the stack of single-particle images
        !$omp   do
        do current_sp=1,size(sp)
            if (debug .and. interval_print(current_sp,size(sp))) then
                !$omp   critical    (write_statement_omp_crit)
                write(*,'(3(A,i0),a)') '**info(mode_frealign_stack_to_frealix_ali): on particle ', current_sp, ' of ', size(sp), &
                                     ' (loc in stack: ', sp(current_sp)%loc_in_stack, ')'
                !$omp   end critical    (write_statement_omp_crit)
            endif

            ! We only need to do work if this particle is on this micrograph
            sp_may_be_on_current_micrograph = .true.
            if (.not. string_equal(list_of_film_numbers%value,'none')) then
                if (film_number(current_sp) .ne. film_num) sp_may_be_on_current_micrograph = .false.
            endif

            if (sp_may_be_on_current_micrograph) then

                do current_rotation=1,rotations_to_check

                    ! Load the single particle image into memory, and give it 0.0 mean and std dev of 1.0 (see Roseman 2003)
                    !$omp   critical    (read_image_omp_crit)
                    call sp_img%ReadFromDisk(sp_imagefile,wanted_location=sp(current_sp)%loc_in_stack)
                    call my_progress_bar%Update(my_progress_bar%GetCurrentTick()+1)
                    !$omp   end critical    (read_image_omp_crit)
                    !call cosine_disc(sp_img,rad=sp_img%logical_dimensions(1)*0.2,falloff=5.0,wipe_input=.FALSE.,mask_with_average=.TRUE.)
                    sp_img%object_is_centered_in_box = .true.
                    sp_img%real_values = sp_img%real_values - sp_img%GetAverageOfValues()
                    sp_img%real_values = sp_img%real_values / sp_img%GetSigmaOfValues()
                    if (debug) then
                        call sp_img%WriteToDisk('dbg_sp_image_normalised.mrc')
                    endif
                    if (current_rotation .eq. 2) then
                        ! Rotate the single particle image by 180 degrees
                        !call image_rotate_180(sp_img) ! This doesn't preserve the centering of the image
                        tmp_img = sp_img
                        call tmp_img%ApplyAffineTransformation(sp_img,real(euler_to_rotmat(0.0D0,0.0D0,180.0D0)))
                    endif
                    call sp_img%Resize(film_img%logical_dimensions)
                    call sp_img%ForwardFFT()
                    call sp_img%SwapRealSpaceQuadrants()

                    ! Compute the correlation
                    call corr%Allocate(film_img)
                    corr%is_in_real_space = .false.
                    corr%object_is_centered_in_box = .false.
                    corr%complex_values = film_img%complex_values * conjg(sp_img%complex_values)
                    call corr%SwapRealSpaceQuadrants()
                    call corr%BackwardFFT()
                    corr%real_values = corr%real_values * real(product(corr%logical_dimensions))

                    ! Normalise the correlation
                    if (.not. corr%is_in_real_space) call terminate('OOPS')
                    do k=1,corr%GetLogicalDimension(2)
                        do j=1,corr%GetLogicalDimension(1)
                            if (film_local_std_dev%real_values(j,k,1) .lt. 0.001) then
                                corr%real_values(j,k,1) = 0.0
                            else
                                corr%real_values(j,k,1) = corr%real_values(j,k,1) / real(product(sp_dim)) &
                                                                                 / film_local_std_dev%real_values(j,k,1)
                            endif
                        enddo
                    enddo


                    ! Find the peak
                    !$omp   critical    (record_peak_omp_crit)
                    !peaks_found(current_sp,i,current_rotation) = peak_from_image_intg(corr,y_max=y_max%value,y_min=30)
                    !peaks_found(current_sp,i,current_rotation) = peak_from_image(corr,y_max=y_max%value,y_min=10)
                    ! This used to be done with restrictions to avoid the film label
                    ! and with a center-of-mass rather than parabola fit to get sub-pixel accuracy
                    !peaks_found(current_sp,i,current_rotation) = corr%FindPeakWithParabolaFit()
                    peaks_found(current_sp,i,current_rotation) = corr%FindPeakWithIntegerCoordinates()
                    !$omp   end critical    (record_peak_omp_crit)

                    !
                    if (debug) then
                        !$omp   critical    (write_statement_omp_crit)
                        write(*,'(4(A,i0),a)',advance='NO') '**debug(mode_frealign_stack_to_frealix_ali): single particle ', &
                                                            current_sp, ' of ', size(sp), ' vs film # ', film_num, ' (', &
                                                            i , '); peak info (relative to center): '
                        call peaks_found(current_sp,i,current_rotation)%PrintInfo()
                        !$omp   end critical    (write_statement_omp_crit)
                    endif

                    ! Debug: window out image
                    if (debug) then
                        tmpimg%logical_dimensions = sp_dim
                        call film_img%SwapRealSpaceQuadrants()
                        call film_img%BackwardFFT()
                        print *, '**debug(): boxing out at coos', nint(peaks_found(current_sp,i,current_rotation)%CoOrdinates)
                        call film_img%ClipInto(tmpimg,nint(peaks_found(current_sp,i,current_rotation)%CoOrdinates &
                                                           + film_img%physical_address_of_box_center))
                        call tmpimg%WriteToDisk('dbg_out.mrc')
                        call film_img%ForwardFFT()
                        call film_img%SwapRealSpaceQuadrants()
                        call corr%WriteToDisk('dbg_corr.mrc')
                        call sp_img%SwapRealSpaceQuadrants()
                        call sp_img%BackwardFFT()
                        call sp_img%WriteToDisk('dbg_sp.mrc')
                        print *, '**debug(): dumped correlation function and boxed image'
                        !call terminate('debug')
                    endif

                enddo   ! End of loop over rotations
            else
                ! The current particle is definitely not on the current micrograph, let's not bother
            endif

        enddo   ! End of loop over single particles
        !$omp   enddo

        ! Deallocate any objects that were allocated within parallel region
        call sp_img%Deallocate()
        call corr%Deallocate()
        call tmp_img%Deallocate()
        call tmpimg%Deallocate()
        call interpol_window%DeallocateWindow()

        !$omp   end parallel

    enddo   ! End of loop over films

    ! Close image files
    call sp_imagefile%close()

    ! Stop progress bar
    call my_progress_bar%Finish()

    !
    ! Now look for the best peak for each location in the stack of single particles,
    ! and turn single particles into filament waypoints
    !
    current_filament = 0
    do current_sp=1,size(sp)

        ! Work out whether we're starting a new filament
        new_filament = .false.
        if (current_sp .eq. 1) then
            new_filament = .true.
        else
            if (sp(current_sp)%film_num .ne. sp(current_sp-1)%film_num) then
                ! We've just moved on to a new filament
                new_filament = .true.
            endif
        endif

        ! If we're starting a new filament, reset/prepare array of waypoints
        if (new_filament) then
            ! Increment
            if (debug) write(*,'(a,i0)') '**debug(mode_frealign_stack_to_frealix_ali): working on new filament, id ', &
                                            sp(current_sp)%film_num
            current_filament = current_filament + 1
            ! Reset the array of waypoints
            if (allocated(wps)) deallocate(wps)
        endif

        ! Look for the film to which this single particle belongs
        best_peak%PeakHeight = -10.0
        best_rotation = 1
        do i=1,size(list_of_films)
            do current_rotation=1,rotations_to_check
                if (peaks_found(current_sp,i,current_rotation)%PeakHeight .gt. best_peak%PeakHeight) then
                    best_peak = peaks_found(current_sp,i,current_rotation)
                    found_film_index = i
                    best_rotation =  current_rotation
                endif
            enddo
        enddo ! End of loop over films

        ! Print out result
        write(*,'(3(a,i0),a)',advance='NO') '**info(mode_frealign_stack_to_frealix_ali): particle ', &
                                                current_sp, ' belongs to film ', found_film_index, &
                                                ' ( ', list_of_films(found_film_index), ' ) '
        if (best_rotation .eq. 2) write(*,'(a)',advance='NO') ' (with 180-deg rotation) '
        write(*,'(a)',advance='NO')  ' peak info (relative to center): '
        call best_peak%PrintInfo()


        ! Convert the peak to a waypoint
        current_wp%coo = best_peak%CoOrdinates
        current_wp%coo(1:2) = current_wp%coo(1:2) + film_img%physical_address_of_box_center(1:2)
        call ctf_ang_to_pix(sp(current_sp)%ctf,pixel_size=pixel_size%value*real(binning_factor%value))
        current_wp%coo(3) = z_from_ctf(sp(current_sp)%ctf)
        ! Apply the shifts from the single particle (note that we assume the single particle shifts are in Angstroms, a la Frealign 9, so we divide by pixel size
        current_wp%coo(1) = current_wp%coo(1) - sp(current_sp)%shift_x/real(binning_factor%value)/pixel_size%value
        current_wp%coo(2) = current_wp%coo(2) - sp(current_sp)%shift_y/real(binning_factor%value)/pixel_size%value
        current_wp%euler  = sp(current_sp)%angles
        if (best_rotation .eq. 2) then
            current_wp%coo(1) = - current_wp%coo(1)
            current_wp%coo(2) = - current_wp%coo(2)
            current_wp%euler%psi = current_wp%euler%psi + 180.
        endif
        if (.not. allocated(current_wp%local_subframe_shift)) then
            allocate(current_wp%local_subframe_shift(3,1))
        endif
        current_wp%local_subframe_shift = 0.0

        ! Extract particle if required
        if (extract_box_from_micrograph) then
            call filename_from_wild(film_filenames%value,list_of_films(found_film_index),film_filename)
            call film_img%ReadFromDisk(film_filename)
            call sp_img%Allocate(dims=sp_dim)
            box_coos = int(best_peak%CoOrdinates)
            box_coos(1:2) = box_coos(1:2) + film_img%physical_address_of_box_center(1:2)
            box_coos(3) = 1
            call film_img%ClipInto(sp_img,box_coos)
            if (best_rotation .eq. 2) then
                call sp_img%ApplyAffineTransformation(tmp_img,real(euler_to_rotmat(0.0D0,0.0D0,180.0D0)))
                sp_img = tmp_img
            endif
            call sp_img%WriteToDisk('dbg_found_boxes.mrc',current_sp)
        endif


        ! Add the waypoint to the array
        call waypoints_grow_array(wps)
        wps(size(wps)) = current_wp

        ! Work out whether we've just processed the last waypoint in a filament
        last_wp_in_filament = .false.
        if (current_sp .eq. size(sp)) then
            last_wp_in_filament = .true.
        else
            if (sp(current_sp)%film_num .ne. sp(current_sp+1)%film_num) then
                last_wp_in_filament = .true.
            endif
        endif

        ! If that was the last waypoint in the filament, we can populate the filament object
        if (last_wp_in_filament) then
            call fils(current_filament)%PopulateWithWaypoints(wps, &
                                    id=sp(current_sp)%film_num, &
                                    film_id=list_of_films(found_film_index  ),   &
                                    pixel_size=pixel_size%value*real(binning_factor%value))
            if (debug) then
                write(*,'(a)') '**debug(mode_frealign_stack_to_frealix_ali): just populated the following filament:'
                call fils(current_filament)%PrintInfo()
            endif
        endif

    enddo ! End of loop over particles

    ! Now that we have prepared the array of filaments, let's write a .ali file to disk
    call filaments_to_frealix_par(fils,ali_filename%value)

end subroutine mode_frealign_stack_to_frealix_ali_slave

end module mode_frealign_stack_to_frealix_ali
