!>  \brief  Prepare files for Frealign
module mode_frealix_to_frealign
    use modes
    use rtps
    implicit none
    type(mode)  ::  frealix_to_frealign

    type(rtp_char), target  ::  input_par_filename  =               &
        rtp_char(   value='waypoints.ali',                  &
                    keyword='INPUT_PAR_FILENAME',           &
                    description='Waypoints file describing the filaments')

    type(rtp_char), target  ::  output_par_filename  =               &
        rtp_char(   value='waypoints.par',                  &
                    keyword='OUTPUT_PAR_FILENAME',           &
                    description='Frealign parameter file for the output')

    type(rtp_char), target  ::  output_stack_filename  =               &
        rtp_char(   value='stack.mrc',                  &
                    keyword='OUTPUT_STACK_FILENAME',           &
                    description='Stack of images to be used as input to Frealign')

    type(rtp_real), target  ::  segment_length  =               &
        rtp_real(   value=128,                  &
                    keyword='SEGMENT_LENGTH',           &
                    description='Length of filament segments to be extracted (Angstroms)')


    type(rtp_char), target  ::  film_files  =               &
        rtp_char(   value='film_????.mrc',                  &
                    keyword='FILM_FILES',           &
                    description='Filename pattern for the micrographs where the filaments are found')

    type(rtp_real), target  ::  pixel_size  =               &
        rtp_real(   value=9.99,                  &
                    keyword='PIXEL_SIZE',           &
                    description='Pixel size of the micrographs, in Angstroms')

    type(rtp_real), target  ::  magnification  =               &
        rtp_real(   value=60000.,                  &
                    keyword='MAGNIFICATION',           &
                    description='Magnification value to be written to Frealign parameter file')

    type(rtp_char), target  ::  ctftilt_files  =               &
        rtp_char(   value='none',                  &
                    keyword='CTFTILT_FILES',           &
                    description='')


    type(rtp_char), target  ::  ctffind_files  =               &
        rtp_char(   value='none',                  &
                    keyword='CTFFIND_FILES',           &
                    description='')

    type(rtp_logi), target  ::  apply_shifts_to_images = &
        rtp_logi(   value=.true.,              &
                    keyword='APPLY_SHIFTS_TO_IMAGES',   &
                    description='Shift the images so that Frealign will just have to rotate them for reconstruction')

    type(rtp_logi), target  ::  apply_ctf_to_images = &
        rtp_logi(   value=.false.,              &
                    keyword='APPLY_CTF_TO_IMAGES',   &
                    description='Apply ctf to the images. Should be false if stack will be used by Frealign.')

    type(rtp_logi), target  ::  ignore_subunit_num = &
        rtp_logi(   value=.false.,              &
                    keyword='IGNORE_SUBUNIT_NUM',   &
                    description='Ignore subunit_num value of waypoints when converting them to single particles')

    type(rtp_real), target  ::  helix_max_radius = &
        rtp_real(   value=100., &
                    keyword='HELIX_MAX_RADIUS', &
                    description='Maximum radius of filament, in Angstroms')

contains

subroutine mode_frealix_to_frealign_init_and_run
    implicit none
    ! Intialise the mode
    call frealix_to_frealign%init(command='FREALIX_TO_FREALIGN',subroutine=mode_frealix_to_frealign_slave,  &
                                    short_description='Prepare a set of input files for Frealign')

    ! Add RTPs to the mode
    call frealix_to_frealign%add_rtp(input_par_filename)
    call frealix_to_frealign%add_rtp(output_par_filename)
    call frealix_to_frealign%add_rtp(output_stack_filename)
    call frealix_to_frealign%add_rtp(segment_length)
    call frealix_to_frealign%add_rtp(film_files)
    call frealix_to_frealign%add_rtp(pixel_size)
    call frealix_to_frealign%add_rtp(magnification)
    call frealix_to_frealign%add_rtp(ctftilt_files)
    call frealix_to_frealign%add_rtp(ctffind_files)
    call frealix_to_frealign%add_rtp(apply_shifts_to_images)
    call frealix_to_frealign%add_rtp(apply_ctf_to_images)
    call frealix_to_frealign%add_rtp(ignore_subunit_num)
    call frealix_to_frealign%add_rtp(helix_max_radius)

    ! Run the mode
    call frealix_to_frealign%run()
end subroutine mode_frealix_to_frealign_init_and_run


subroutine mode_frealix_to_frealign_slave
    use filaments, only : wp_to_sp
    use frealix_parameters
    use Images
    use ctftilts
    use filaments_base
    use single_particles
    use frealign_parameters
    use parsers, only : string_equal, filename_from_wild
    use ui, only : interval_print
    use runtime_parameters, only : is_odd, filename_len, terminate
    use ImageFiles
    use coos, only : wp_print_info
    implicit none
    ! Private variables
    type(filament), allocatable ::  fils(:)
    type(single_particle),  allocatable ::  sp(:), sp_tmp(:)
    integer ::  current_fil
    integer ::  current_wp
    integer ::  number_of_waypoints
    type(image) ::  sp_image
    character(len=filename_len) :: current_film_filename, current_ctf_filename
    type(image) ::  current_film_image(1)
    type(imagefile) ::  output_stack_file
    integer ::  current_output_loc, start_output_loc
    logical ::  use_ctftilt
    type(ctftilt) :: ctftilt_params
    logical,    parameter   ::  debug   =   .false.
    integer ::  box_dim(2)
    ! Start work

    call terminate('This mode is deprecated. Use stack_from_par instead.')

    ! Are we given ctffind files or ctftilt files?
    use_ctftilt = .false.
    if (string_equal(ctftilt_files%value,'none')) then
        if (string_equal(ctffind_files%value,'none')) then
            write(*,'(a)') '**error(mode_frealix_to_frealign_slave): neither ctffind nor ctftilt files were supplied'
            call terminate('Fatal error in MODE_FREALIX_TO_FREALIGN_SLAVE')
        else
            use_ctftilt = .false.
        endif
    else
        use_ctftilt = .true.
    endif

    ! Get filaments from waypoints file
    call filaments_from_frealix_par_noinit(fils,input_par_filename%value,pixel_size%value)
    fils%max_radius = helix_max_radius%value / pixel_size%value ! This is normally done by filament_init

    ! Work out total number of waypoints
    number_of_waypoints = 0

    ! Loop over filaments
    do current_fil=1,size(fils)
        number_of_waypoints = number_of_waypoints + size(fils(current_fil)%wp)
    enddo

    ! Allocate array of single particles
    allocate(sp(number_of_waypoints))

    ! Open the output image stack file for writing
    box_dim = max ( ceiling(3.*helix_max_radius%value/pixel_size%value), &
                    ceiling(segment_length%value/pixel_size%value) )
    if (is_odd(box_dim(1))) box_dim = BOX_DIM+1
    call output_stack_file%init(filename=output_stack_filename%value,  &
                                dim_x=box_dim(1),dim_y=box_dim(2),  &
                                pixel_size=pixel_size%value)

    ! Loop over filaments
    current_output_loc = 0
    do current_fil=1,size(fils)
        if (interval_print(current_fil,size(fils))) then
            write(*,'(a,i0,a,i0)') '**info(mode_frealix_to_frealign): on filament ', current_fil, ' of ', size(fils)
        endif
        ! Work out micrograph filename and the ctftilt filename for this filament
        call filename_from_wild(film_files%value,fils(current_fil)%film%GetID(),current_film_filename)
        if (use_ctftilt) then
            call filename_from_wild(ctftilt_files%value,fils(current_fil)%film%GetID(),current_ctf_filename)
            call ctftilt_read_from_ctftilt(ctftilt_params,current_ctf_filename)
        else
            call filename_from_wild(ctffind_files%value,fils(current_fil)%film%GetID(),current_ctf_filename)
            call ctftilt_read_from_ctffind(ctftilt_params,current_ctf_filename)
        endif
        ! Read the micrograph image from disk (we should check whether this is necessary -
        ! we could have the correct image in memory already; won't do this for now - os caching
        ! probably good enough in most cases)
        call current_film_image(1)%ReadFromDisk(current_film_filename)
        ! Allocate temporary array of single particles for this filament's waypoints
        if (allocated(sp_tmp)) deallocate(sp_tmp)
        allocate(sp_tmp(size(fils(current_fil)%wp)))
        ! Loop over waypoints
        start_output_loc = current_output_loc + 1
        do current_wp=1,size(fils(current_fil)%wp)
            current_output_loc = current_output_loc + 1
            call wp_to_sp(  fils(current_fil)%wp(current_wp),sp_tmp(current_wp),  &
                            ctf=ctftilt_params%ctf, &
                            sp_image=sp_image,segment_length=box_dim(1),  &
                            micrograph_frames=current_film_image,  &
                            pixel_size=pixel_size%value,    &
                            apply_shift_to_image=apply_shifts_to_images%value, &
                            apply_ctf_to_image=apply_ctf_to_images%value,&
                            ignore_subunit_num=ignore_subunit_num%value,&
                            fil=fils(current_fil))
            if (debug) then
                write(*,'(2(A,i0))') '**debug(mode_frealix_to_frealign_slave): on wp ', current_wp, ' of filament ', current_fil
                call wp_print_info(fils(current_fil)%wp(current_wp))
                call sp_print_info(sp_tmp(current_wp))
            endif
            call sp_image%WriteToDisk(output_stack_file,wanted_location=current_output_loc)
        enddo
        ! Now that we've filled the temporary array of SPs, add it to the output array of SPs
        sp_tmp%film_num = fils(current_fil)%id
        sp(start_output_loc:START_OUTPUT_LOC+SIZE(sp_tmp)-1) = sp_tmp(1:size(sp_tmp))
    enddo

    ! Close the output stack file
    call output_stack_file%PrintInfo()
    call output_stack_file%close()

    ! Write out the Frealign parameter file
    call sp_to_frealign_par(sp,output_par_filename%value,magnification%value,pixel_size%value)

    ! Sanity check
    if (.not. current_output_loc .eq. number_of_waypoints) then
        write(*,'(a)') '**error(mode_frealix_to_frealign_slave): failed sanity check'
        call terminate('Failed sanity check in MODE_FREALIX_TO_FREALIGN_SLAVE')
    endif
end subroutine mode_frealix_to_frealign_slave

end module mode_frealix_to_frealign
