!>  \brief  mode to measure the helical symmetry of an input image
module mode_hmeasure
    ! modules
    use modes
    use rtps
    implicit none
    type(mode) :: hmeasure

    type(rtp_char), target  ::  input_filename  =               &
        rtp_char(   value='test_in',                  &
                    keyword='input_filename',           &
                    description='')


    type(rtp_real), target  ::  twist  =               &
        rtp_real(   value=0.0,                  &
                    keyword='twist_per_subunit',           &
                    description='degrees')


    type(rtp_real), target  ::  rise  =               &
        rtp_real(   value=0.0,                  &
                    keyword='rise_per_subunit',           &
                    description='angstroms')

    type(rtp_real), target  ::  twist_half_range  =               &
        rtp_real(   value=2.0,                  &
                    keyword='twist_half_range',           &
                    description='degrees')


    type(rtp_real), target  ::  rise_half_range  =               &
        rtp_real(   value=2.0,                  &
                    keyword='rise_half_range',           &
                    description='angstroms')


    type(rtp_real), target  ::  twist_step_size  =               &
        rtp_real(   value=1.0,                  &
                    keyword='twist_step_size',           &
                    description='degrees')


    type(rtp_real), target  ::  rise_step_size  =               &
        rtp_real(   value=1.0,                  &
                    keyword='rise_step_size',           &
                    description='angstroms')


    type(rtp_real), target  ::  pixel_size  =               &
        rtp_real(   value=0.0,                  &
                    keyword='pixel_size',           &
                    description='')


    type(rtp_real), target  ::  max_rad  =               &
        rtp_real(   value=0.0,                  &
                    keyword='helix_max_radius',           &
                    description='maximum radius (angstroms)')


    type(rtp_intg), target  ::  symmetry_axial  =               &
        rtp_intg(   value=1,                  &
                    keyword='symmetry_axial',           &
                    description='')


    type(rtp_intg), target  ::  symmetry_perp  =               &
        rtp_intg(   value=1,                  &
                    keyword='symmetry_perp',           &
                    description='')

    type(rtp_intg), target  ::  helical_starts  =               &
        rtp_intg(   value=1,                  &
                    keyword='num_helical_starts',           &
                    description='number of helical starts')

     type(rtp_logi), target  ::  stack  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='stack',           &
                    description='The assembly is a stack rather than a true helix')


  contains

subroutine mode_hmeasure_init_and_run
    ! intialise the mode
    call hmeasure%init(  command='hmeasure',subroutine=mode_hmeasure_slave,    &
                        short_description='Estimate the helical symmetry parameters of a 3d density volume')

    ! add rtps to the mode
    call hmeasure%add_rtp(input_filename)
    call hmeasure%add_rtp(twist)
    call hmeasure%add_rtp(rise)
    call hmeasure%add_rtp(twist_half_range)
    call hmeasure%add_rtp(rise_half_range)
    call hmeasure%add_rtp(twist_step_size)
    call hmeasure%add_rtp(rise_step_size)
    call hmeasure%add_rtp(pixel_size)
    call hmeasure%add_rtp(max_rad)
    call hmeasure%add_rtp(symmetry_axial)
    call hmeasure%add_rtp(symmetry_perp)
    call hmeasure%add_rtp(helical_starts)
    call hmeasure%add_rtp(stack)

    ! run the mode
    call hmeasure%run()
end subroutine mode_hmeasure_init_and_run

subroutine mode_hmeasure_slave
    use runtime_parameters
    use images
    use helices
    use images_helices, only : image_impose_helical_symmetry_new
    use images_arithm, only : cyl_mask
    use imagefiles
    use interpolation
    use images_comparison, only : correlation
    !use BruteForceSearches
    implicit none
    ! variables
    !type(BruteForceSearch)      ::  my_bf_search
    type(image)                 ::  imgin, imgout, imgmask
    type(helical_symmetry)      ::  helix
    type(interpolation_scheme)  ::  interpol_scheme
    !type(imagefile)             ::  input_file, output_file, asym_unit_file
    !type(image)                 ::  asym_unit_vol
    real                        ::  score
    !integer                     ::  rise_step, twist_step
    real(kind=8)                ::  current_rise, current_twist
    real                        ::  best_score, best_rise, best_twist
    integer                     ::  crop_dimensions(3)


    ! start work

    ! convert the maximum radius to pixels
    if (max_rad%value .ne. 0.0) then
        max_rad%value = max_rad%value / pixel_size%value
    endif

    ! setup brute-force search
    !call my_bf_search%Init()



    ! setup interpolation scheme
    call interpolation_scheme_init(interpol_scheme,'linear','rect',2)

    ! read the image from disk
    call imgin%readfromdisk(input_filename%value,read_volume_not_slice=.true.)

    ! Crop the volume
    crop_dimensions = ceiling(max_rad%value*2.0/pixel_size%value)
    write(*,'(a,i0)') '**info(mode_hmeasure): cropping volume to ', crop_dimensions(1)
    call imgin%Resize(crop_dimensions)

    ! create a mask for the correlation measurement
    call imgmask%Allocate(mould=imgin)
    imgmask%real_values = 1.0
    call cyl_mask(imgmask,max_rad%value,3)

    !
    best_score = -huge(1.0)
    current_twist = twist%value - twist_half_range%value
    do while (current_twist .le. twist%value + twist_half_range%value)
        current_rise = rise%value - rise_half_range%value
        do while (current_rise .le. rise%value + rise_half_range%value)

            ! setup the helical symmetry operator
            call helical_symmetry_init( helix,real(current_rise)/pixel_size%value,real(current_twist), &
                                        max_rad%value/pixel_size%value, &
                                        symmetry_axial%value,symmetry_perp%value,&
                                        num_starts=helical_starts%value, &
                                        stack=stack%value)

            ! symmetrise
            imgout = imgin
            call image_impose_helical_symmetry_new(imgout,helix,interpol_scheme,max_radius=max_rad%value,quiet=.true.)

            ! measure the correlation between the volumes before and after symmetrisation
            score =  correlation(imgin,imgout,imgmask,normalise=.true.)

            !$omp critical
            write(*,'(a,3(f0.5,1x))') '**info(mode_hmeasure_slave): twist, rise, score = ', current_twist, &
                                                                                            current_rise , score
            !$omp end critical

            if (score .gt. best_score) then
                best_rise = current_rise
                best_twist = current_twist
                best_score = score
            endif
            current_rise = current_rise + rise_step_size%value
        enddo ! loop over rise
        current_twist = current_twist + twist_step_size%value
    enddo ! loop over twist

    write(*,'(a,3(f0.5,1x))') '**info(mode_hmeasure_slave): best twist, rise, score = ', best_twist, best_rise , best_score

end subroutine mode_hmeasure_slave

end module mode_hmeasure
