module mode_mask

    !
    ! Modules we'll need
    !
    use runtime_parameters
    use modes
    use rtps

    implicit none
    type(mode) :: mask

    type(rtp_char), target  ::  input_filename  =               &
        rtp_char(   value='input.mrc',                  &
                    keyword='INPUT_FILENAME',           &
                    description='Input image / volume')


    type(rtp_char), target  ::  output_filename  =               &
        rtp_char(   value='threed.mrc',                  &
                    keyword='OUTPUT_FILENAME',           &
                    description='Output image / volume')


    type(rtp_char), target  ::  mask_type  =               &
        rtp_char(   value='CIRCLE',                  &
                    keyword='MASK_TYPE',           &
                    description='One of circle, cylinder, z_edges, low_pass, cosine_disc or AUTO')


    type(rtp_char), target  ::  falloff_type  =               &
        rtp_char(   value='NONE',                  &
                    keyword='FALLOFF_TYPE',           &
                    description='One of cos_edge, hard_extend, or NONE. Used by low_pass mask only.')


    type(rtp_real), target  ::  mask_radius  =               &
        rtp_real(   value=0.5,                  &
                    keyword='MASK_RADIUS',           &
                    description='Mask radius (pixels or reciprocal pixels)')

    type(rtp_real), target  ::  mask_inner_radius  =               &
        rtp_real(   value=0.0,                  &
                    keyword='MASK_INNER_RADIUS',           &
                    description='Mask inner radius (pixels or reciprocal pixels) (used to give hollow cylinder)')


    type(rtp_real), target  ::  mask_falloff  =               &
        rtp_real(   value=0.05,                  &
                    keyword='MASK_FALLOFF',           &
                    description='Mask soft falloff width (pixels or reciprocal pixels)')


    type(rtp_real), target  ::  pct_in_mask  =               &
        rtp_real(   value=2.0,                  &
                    keyword='PCT_IN_MASK',           &
                    description='Percentage of voxels in the mask (not including the voxels in soft falloff around the mask)')


    type(rtp_logi), target  ::  inverted  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='INVERTED',           &
                    description='If true, flips 1 to 0 and vice versa at the end of masking routine to let through only background')

    type(rtp_logi), target  ::  mask_with_average =     &
        rtp_logi(   value=.true.,                  &
                    keyword='MASK_WITH_AVERAGE',           &
                    description='Use average value within the mask to fill the areas outside of the mask aperture. '//&
                                'Not implemented for all types of masks')
contains

subroutine mode_mask_init_and_run
    implicit none
    ! Intialise the mode
    call mask%init(command='MASK',subroutine=mode_mask_slave,short_description='Add description here')

    ! Add RTPs to the mode
    call mask%add_rtp(input_filename)
    call mask%add_rtp(output_filename)
    call mask%add_rtp(mask_type)
    call mask%add_rtp(falloff_type)
    call mask%add_rtp(mask_radius)
    call mask%add_rtp(mask_inner_radius)
    call mask%add_rtp(mask_falloff)
    call mask%add_rtp(pct_in_mask)
    call mask%add_rtp(inverted)
    call mask%add_rtp(mask_with_average)

    ! Run the mode
    call mask%run()
end subroutine mode_mask_init_and_run

subroutine mode_mask_slave
    use images
    use images_arithm
    use masks
    implicit none
    !
    ! Private variables
    !
    type(image) ::  input_image, mask_image
    real    ::  rad,falloff, rad_in

    !
    ! Get variable values from control file
    !


    !
    ! Start work
    !

    ! Read input image
    call input_image%ReadFromDisk(input_filename%value,read_volume_not_slice=.true.)

    ! Depending on the mask_type%value chosen by the user, apply the appropriate mask
    select case (mask_type%value)
        case ('CIRCLE')
            write(*,'(2a)') '**info(mode_mask): circular mask'
            if (mask_radius%value .lt. 1.0) then
                rad = mask_radius%value*(input_image%logical_dimensions(1)/2.)
                falloff = mask_falloff%value*(input_image%logical_dimensions(1)/2.)
            else
                rad = mask_radius%value
                falloff = mask_falloff%value
            endif
            call input_image%ApplyCircularMask(rad)
            call input_image%WriteToDisk(output_filename%value)
        case ('CYLINDER')
            write(*,'(2a)') '**info(mode_mask): cylindrical mask'
            if (mask_radius%value .lt. 1.0) then
                rad = mask_radius%value*(input_image%logical_dimensions(1)/2.)
                rad_in = mask_inner_radius%value*(input_image%logical_dimensions(1)/2.)
                falloff = mask_falloff%value*(input_image%logical_dimensions(1)/2.)
            else
                rad = mask_radius%value
                rad_in = mask_inner_radius%value
                falloff = mask_falloff%value
            endif
            call cyl_mask(input_image,rad,nint(falloff),rad_in)
            call input_image%WriteToDisk(output_filename%value)
        case ('Z_EDGES')
            ! Apply the mask
            call image_mask_z_ends_inplace(input_image,nint(mask_radius%value),soft=.true.)
            ! Write output
            call input_image%WriteToDisk(output_filename%value)
        case ('LOW_PASS')
            write(*,'(a)') '**info(mode_mask): low pass mask'
            call image_low_pass_mask(input_image,mask_image, pct_in_mask%value,falloff_type%value,  &
                                    low_pass_frequency=mask_radius%value, falloff_width=mask_falloff%value, &
                                    inversion=inverted%value)
            ! Apply the mask
            input_image%real_values = input_image%real_values * mask_image%real_values
            ! Write the masked image to disk
            call input_image%WriteToDisk(output_filename%value)
        case ('COSINE_DISC')
            write(*,'(a)') '**info(mode_mask): cosine-edged disc mask'
            call input_image%ApplySoftCircularMask( radius_after_falloff=mask_radius%value, &
                                                    falloff=mask_falloff%value,wipe_input=.false.,   &
                                                    mask_with_average=mask_with_average%value)
            call input_image%WriteToDisk(output_filename%value)
        case default
            write(*,'(2a)') '**error(mode_mask): unrecognised mask type: ', trim(mask_type%value)
            call terminate('Fatal error in MODE_MASK')
    end select

end subroutine mode_mask_slave

end module mode_mask
