!>  \brief  Mode to test the rotate_image subroutine
!!  \todo   Make the interpolation scheme and the real- Fourier-space choice runtime parameters
module mode_rotate
    use modes
    use rtps
    implicit none
    type(mode) :: rotate

    type(rtp_char), target  ::  filename_input  =               &
        rtp_char(   value='in.mrc',                  &
                    keyword='INPUT_FILENAME',           &
                    description='')


    type(rtp_char), target  ::  filename_output  =               &
        rtp_char(   value='out.mrc',                  &
                    keyword='OUTPUT_FILENAME',           &
                    description='')


    type(rtp_real), target  ::  phi  =               &
        rtp_real(   value=0.0,                  &
                    keyword='PHI',           &
                    description='')


    type(rtp_real), target  ::  theta  =               &
        rtp_real(   value=0.0,                  &
                    keyword='THETA',           &
                    description='')


    type(rtp_real), target  ::  psi  =               &
        rtp_real(   value=0.0,                  &
                    keyword='PSI',           &
                    description='')


    type(rtp_real), target  ::  shift_x  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SHIFT_X',           &
                    description='')


    type(rtp_real), target  ::  shift_y  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SHIFT_Y',           &
                    description='')


    type(rtp_real), target  ::  shift_z  =               &
        rtp_real(   value=0.0,                  &
                    keyword='SHIFT_Z',           &
                    description='')

contains

subroutine mode_rotate_init_and_run
    ! Intialise the mode
    call rotate%init(command='ROTATE',subroutine=mode_rotate_slave,short_description='Add description here')

    ! Add RTPs to the mode
    call rotate%add_rtp(filename_input)
    call rotate%add_rtp(filename_output)
    call rotate%add_rtp(phi)
    call rotate%add_rtp(theta)
    call rotate%add_rtp(psi)
    call rotate%add_rtp(shift_x)
    call rotate%add_rtp(shift_y)
    call rotate%add_rtp(shift_z)

    ! Run the mode
    call rotate%run()
end subroutine mode_rotate_init_and_run


subroutine mode_rotate_slave
    use runtime_parameters
    use images
    use interpolation, only : interpolation_scheme, interpolation_scheme_init
    use rotation_matrices
    use ImageFiles
    use InterpolatingWindows
    implicit none
    ! Variables
    type(image)             ::  imgin,imgout
    real                    ::  matrix(4,4)
    logical,    parameter   ::  debug               =   .true.
    logical,    parameter   ::  real_space_rotation =   .true.
    type(interpolation_scheme)  ::  interpol_scheme
    type(imagefile)         ::  input_file
    type(Window)            ::  interpol_window


    ! Start work
    call input_file%init(filename_input%value)
    call imgin%ReadFromDisk(input_file,read_volume_not_slice=.true.)
    call input_file%PrintInfo()
    call input_file%close()
    write(*,'(a)') '**warning(mode_rotate): reading image in as 3d, without checking whether it actually is'

    ! Set up interpolation
    call interpolation_scheme_init(interpol_scheme,'LINEAR','RECT',2)
    call interpol_window%AllocateWindow(interpol_scheme%window_width,comp=.not.REAL_SPACE_ROTATION,threed=imgin%IsAVolume())

    ! Get the appropriate rotation matrix (rotate_image needs to matrix to go from output back to input, which is why we
    ! use reverse=.TRUE.)
    matrix = euler_to_rotmat(dble(phi%value),dble(theta%value),dble(psi%value),reverse=.true.)
    matrix(1,4) = shift_x%value
    matrix(2,4) = shift_y%value
    matrix(3,4) = shift_z%value
    if (debug) then
        write(*,'(a)') '**debug(mode_rotate): rotation matrix will be: '
        call print_rotmat(matrix)
    endif


    call imgin%PrintInfo()

    ! Rotate the image
    if (.not. real_space_rotation) then
        call imgin%ForwardFFT()
        call imgin%SwapRealSpaceQuadrants()
    endif
    call imgout%Allocate(imgin)
    call imgin%ApplyAffineTransformation(imgout,matrix)
    if (.not. real_space_rotation) then
        call imgout%SwapRealSpaceQuadrants()
        call imgout%BackwardFFT()
    endif

    call imgout%PrintInfo()


    ! Write output to disk
    call imgout%WriteToDisk(filename_output%value,pixel_size=input_file%GetPixelSize(),print_file_info=.true.)

end subroutine mode_rotate_slave

end module mode_rotate
