module mode_stack_align_to_sum

    ! Modules
    use runtime_parameters
    use modes
    use rtps
    implicit none
    type(mode) :: stack_align_to_sum

    type(rtp_char), target  ::  input_filename  =       &
        rtp_char(   value='input.mrc',                  &
                    keyword='INPUT_STACK',              &
                    description='')


    type(rtp_char), target  ::  output_filename  =      &
        rtp_char(   value='output.mrc',                 &
                    keyword='OUTPUT_STACK',             &
                    description='')


    type(rtp_intg), target  ::  num_iterations  =       &
        rtp_intg(   value=3,                            &
                    keyword='NUM_ITERATIONS',           &
                    description='Number of alignment & averaging iterations')


    type(rtp_logi), target  ::  subtract_from_sum  =    &
        rtp_logi(   value=.true.,                       &
                    keyword='SUBTRACT_FROM_SUM',        &
                    description='Whether to substract the input image from the average image before doing the alignment')

    type(rtp_logi), target  ::  amyloid_beta  =               &
        rtp_logi(   value=.false.,                  &
                    keyword='AMYLOID_BETA',           &
                    description='Input stack consists of projections of amyloid-beta fibrils')

    type(rtp_real), target  ::  pixel_size  =               &
        rtp_real(   value=1.17,                  &
                    keyword='PIXEL_SIZE',           &
                    description='Pixel size (Angstroms)')

contains

subroutine mode_stack_align_to_sum_init_and_run
    implicit none
    ! Intialise the mode
    call stack_align_to_sum%init(command='STACK_ALIGN_TO_SUM',subroutine=mode_stack_align_to_sum_slave, &
                             short_description='Iteratively compute the average of a stack and align each image of the stack to it')

    ! Add RTPs to the mode
    call stack_align_to_sum%add_rtp(input_filename)
    call stack_align_to_sum%add_rtp(output_filename)
    call stack_align_to_sum%add_rtp(num_iterations)
    call stack_align_to_sum%add_rtp(subtract_from_sum)
    call stack_align_to_sum%add_rtp(amyloid_beta)
    call stack_align_to_sum%add_rtp(pixel_size)

    ! Run the mode
    call stack_align_to_sum%run()


end subroutine mode_stack_align_to_sum_init_and_run

subroutine mode_stack_align_to_sum_slave
    use Images
    use imagefiles
    use euleranglesmod, only : euler
    use rotation_matrices, only : euler_to_rotmat, identity_rotmat
    use images_comparison, only : correlation, find_shift_euler
    use interpolation, only : interpolation_scheme, interpolation_scheme_init
    use images_arithm, only : images_average
    use masks, only : fourier_mask_init_amyloid
    use InterpolatingWindows
    implicit none
    ! Variables
    type(image)                     ::  stack_ave, mask, ref
    type(image),    allocatable     ::  stack(:), stack_original(:)
    type(image)                     ::  fmask, fmask_copy
    integer                         ::  i, ierr, iteration
    real                            ::  shift(3), matrix(4,4)
    real,           allocatable     ::  stack_matrices(:,:,:)
    type(euler)                     ::  rot
    integer                         ::  current_shift, best_shift
    real                            ::  cc, best_cc
    integer                         ::  par_lun
    type(interpolation_scheme)      ::  interpol_scheme
    type(window)                    ::  interpol_window
    type(imagefile)                 ::  input_stack_file, output_stack_file
    integer                         ::  locs_in_stack
    integer                         ::  current_loc
    logical,    parameter           ::  debug                   = .true.
    logical,    parameter           ::  real_space_alignment    = .false.


    ! Start work

    par_lun =   available_lun()

    ! Setup interpolation scheme
    call interpolation_scheme_init(interpol_scheme,'LINEAR','RECT',2)
    call interpol_window%AllocateWindow(interpol_scheme%window_width,comp=.false.,threed=.false.)


    ! Open the input file & find out the number of locations in the stack
    call input_stack_file%init(input_filename%value)
    call input_stack_file%PrintInfo()
    locs_in_stack = input_stack_file%GetStackSize()

    ! Allocate memory so that we can hold the stack in core twice
    allocate(stack(locs_in_stack),stack_original(locs_in_stack),stat=ierr)
    if (ierr .ne. 0) then
        write(*,'(a,i0)') '**error(mode_stack_align_to_sum): memory allocation failed for stack, with error ', ierr
        call terminate('Fatal error in MODE_STACK_ALIGN_TO_SUM')
    endif

    ! Read the stack image data from disk
    do current_loc=1,locs_in_stack
        call stack(i)%ReadFromDisk(input_stack_file,wanted_location=current_loc)
        stack_original(i) = stack(i)
    enddo

    ! Open the output file
    call output_stack_file%init(output_filename%value,mould=input_stack_file)

    ! Close the input file
    call input_stack_file%close()


    ! Allocate memory for & initialise the rotation matrices. We will store the transformation matrix
    ! for each image from the original stack to the latest alignment
    allocate(stack_matrices(4,4,size(stack)),stat=ierr)
    if (ierr .ne. 0) then
        write(*,'(a,i0)') '**error(mode_stack_align_to_sum): memory allocation failed for rotation matrices with error ', ierr
        call terminate('Fatal error in MODE_STACK_ALIGN_TO_SUM')
    endif
    forall(i=1:size(stack))
        stack_matrices(:,:,i) = identity_rotmat()
    end forall


    ! Prepare a Fourier mask
    if (amyloid_beta%value) then
        call fourier_mask_init_amyloid(fmask,stack(1),pixel_size%value,fil_max_rad=100.0,max_spacing_align=8.0)
        fmask_copy = fmask
        call fmask_copy%SpectrumFromHalfToFull()
        call fmask_copy%WriteToDisk('dbg_fourier_mask.mrc')
        call fmask_copy%Deallocate()
        if (real_space_alignment) then
            write(*,'(a)') '**error(mode_stack_align_to_sum): Fourier masks not implemented yet in real-space ccc calculation'
            call terminate('Fatal error in MODE_STACK_ALIGN_TO_SUM. Feature not implemented yet')
        endif
    endif


    !
    ! Start iterative alignment. stack keeps latest aligned state of the stack, while stack_original keeps original input stack
    !
    do iteration=1,num_iterations%value

        write(*,'(/a,i0,a,i0)') '**info(mode_stack_align_to_sum): on iteration ', iteration, ' of ', num_iterations%value

        ! Calculate new average, from current alignment
        call images_average(stack,stack_ave)
        stack_ave%real_values = 0.0

        ! Align each image
        do i=1,size(stack)
            ! Calculate the reference for this image
            ref = stack_ave
            if (subtract_from_sum%value) then
                ref%real_values = stack_ave%real_values - stack(i)%real_values
            endif

            ! Do the alignment
            if (real_space_alignment) then
                ! Find alignment the Chuck way, by doing shifts and calculating real-space normalised ccc
                best_cc = -1.0
                do current_shift=0,3
                    stack(i)%real_values = stack_original(i)%real_values
                    call stack(i)%ForwardFFT()
                    call stack(i)%PhaseShift(0.0,real(current_shift),0.0)
                    call stack(i)%BackwardFFT()
                    cc = correlation(stack(i),ref,mask,normalise=.true.)
                    if (cc .gt. best_cc) then
                        best_cc = cc
                        best_shift = current_shift
                    endif
                    !print '(a,i2,a,i2,x,F12.8)', '**debug(): loc ', i, '; current shift, cc = ', current_shift, cc
                enddo
                shift = [0.0,real(best_shift),0.0]
                !print '(a,i2,a,i2,x,F12.8)', '**debug(): loc ', i, '; best shift by real space, cc = ', best_shift, best_cc
            else
                !call find_shift_euler(stack(i),ref,shift,rot,shift_only=.true.,max_rad_shift=4.0,vertical_only=.true.)
                call find_shift_euler(stack(i),ref,shift,rot,shift_only=.true.,max_rad_shift=4.0)
            endif

            ! Print out result of the alignemnt
            write(*,'(2(A,i0),a,3(F0.2,1x))') '**info(mode_stack_align_to_sum): image ', i, ' of ', size(stack), &
                                                '; shift & rot = ', shift(1:2), rot%psi

            ! Shift & rotate the original image
            matrix = euler_to_rotmat(rot%phi,rot%theta,rot%psi,shift=shift,reverse=.true.)
            stack_matrices(:,:,i) = matmul(matrix,stack_matrices(:,:,i))    ! Not 100% whether this is the correct order for the matrix multiplication
            call stack_original(i)%ApplyAffineTransformation(stack(i),stack_matrices(:,:,i))

        enddo

    enddo

    !
    ! Calculate new sum, from final alignment
    !
    call images_average(stack,stack_ave)

    ! Write out the final stack
    do i=1,size(stack)
        call stack(i)%WriteToDisk(output_stack_file,wanted_location=i)
    enddo
    call output_stack_file%PrintInfo()
    call output_stack_file%close()

end subroutine mode_stack_align_to_sum_slave


end module mode_stack_align_to_sum
