!>  \brief  Module with various parsers
!!
module parsers

    use iso_c_binding

    implicit none


    interface parse_filelist_file
        module procedure parse_filelist_file
    end interface parse_filelist_file

    interface file_get_value
        module procedure file_get_char
        module procedure file_get_intg
        module procedure file_get_real
        module procedure file_get_logi
    end interface

    character(len=*), private, parameter :: lower_case_letters = 'abcdefghijklmnopqrstuvwxyz'
    character(len=*), private, parameter :: upper_case_letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    character(len=*), private, parameter :: integers = '1234567890'
    character(kind=C_CHAR,len=*), private, parameter :: new_lines_c = c_form_feed // c_new_line // c_carriage_return &
                                                                            // c_vertical_tab
    character(kind=C_CHAR,len=*), private, parameter :: blank_c_characters = c_null_char // c_horizontal_tab
    character(len=*), private, parameter :: blank_characters =  ' '//blank_c_characters
    character(len=*), private, parameter :: blanks_and_new_lines = blank_characters // new_lines_c

    character(len=*), private, parameter :: integer_digits = '10'   ! Maximum number of digits expected when reading an integer

    contains

    !>  \brief  Get global run-time parameters from the control file
    subroutine get_global_rtps()
        use runtime_parameters
        implicit none
        if (control_filename .ne. '') then
            call file_get_value(control_filename,'DEFAULT_FILE_FORMAT',default_file_format)
            call file_get_value(control_filename,'USE_FFTW_WISDOM',use_fftw_wisdom)
            call file_get_value(control_filename,'USE_FFTW_OPENMP',use_fftw_openmp)
            call file_get_value(control_filename,'PSF_RADIUS',psf_radius)
            call file_get_value(control_filename,'MT_PHI_FIRST_PF',mt_phi_first_pf)
            call file_get_value(control_filename,'PBC',pbc)
        endif
    end subroutine get_global_rtps

    !>  \brief  Parse flat-text file and retrieve a run-time parameter (rtp) integer value
    subroutine file_get_intg(filename,parameter,rtp_value,found)
        use runtime_parameters
        implicit none
        character(len=FILENAME_LEN),    intent(in)      ::  filename
        character(len=*),               intent(in)      ::  parameter
        integer,                        intent(inout)   ::  rtp_value
        logical,            optional,   intent(out)     ::  found
        ! Private variable
        character(len=FILENAME_LEN)                 ::  tmpchar
        ! Start work
        write(tmpchar,'(I'//integer_digits//')') rtp_value
        call file_get_char(filename,parameter,tmpchar,found)
        !print *, '**debug(file_get_intg):  tmpchar = ', tmpchar
        if (.not. is_blank(tmpchar)) then
            if (is_intg(tmpchar)) then
                read(tmpchar,'(I'//integer_digits//')') rtp_value
            else
                write(*,'(4a)') '**ERROR(file_get_intg): the value given for ', trim(adjustl(parameter)), ' is not an integer: ', &
                                trim(adjustl(tmpchar))
                call terminate('Fatal error FILE_GET_INTG')
            endif
        endif
    end subroutine file_get_intg

    !>  \brief  Parse flat-text file and retrieve a run-time parameter (rtp) real value
    subroutine file_get_real(filename,parameter,rtp_value,found)
        use runtime_parameters
        implicit none
        character(len=FILENAME_LEN),    intent(in)      ::  filename
        character(len=*),               intent(in)      ::  parameter
        real,                           intent(inout)   ::  rtp_value
        logical,            optional,   intent(out)     ::  found
        ! Private variable
        character(len=FILENAME_LEN)                 ::  tmpchar
        ! Start work
        write(tmpchar,*) rtp_value
        call file_get_char(filename,parameter,tmpchar,found)
        if (.not. is_blank(tmpchar)) then
            if (string_is_real(tmpchar)) then
                read(tmpchar,*) rtp_value
            else
                write(*,'(4a)') '**ERROR(file_get_real): the value given for ', trim(adjustl(parameter)), ' is not a real: ', &
                                trim(adjustl(tmpchar))
                call terminate('Fatal error FILE_GET_REAL')
            endif
        endif
    end subroutine file_get_real

    !>  \brief  Parse flat-text file and retrieve a run-time parameter (rtp) logical value
    subroutine file_get_logi(filename,parameter,rtp_value,found)
        use runtime_parameters
        implicit none
        character(len=FILENAME_LEN),    intent(in)      ::  filename
        character(len=*),               intent(in)      ::  parameter
        logical,                        intent(inout)   ::  rtp_value
        logical,            optional,   intent(out)     ::  found
        ! Private variable
        character(len=FILENAME_LEN)                 ::  tmpchar
        ! Start work
        write(tmpchar,*) rtp_value
        call file_get_char(filename,parameter,tmpchar,found)
        if (.not. is_blank(tmpchar)) then
            read(tmpchar,*) rtp_value
        endif
    end subroutine file_get_logi


    !>  \brief  Parse flat-text file and retrieve a run-time parameter (rtp) character string
    subroutine file_get_char(filename,parameter,rtp_value,found)
        use runtime_parameters
        use StringManipulations, only : StringsAreEqual
        implicit none
        ! Arguments
        character(len=FILENAME_LEN),    intent(in)      ::  filename    !<  Name of file to be parsed
        character(len=*),               intent(in)      ::  parameter   !<  Label of parameter to be read in
        character(len=*),               intent(inout)   ::  rtp_value   !<  Variable which will hold rtp value
        logical,            optional,   intent(out)     ::  found       !<  Indicates whether the parameter was found in the file
        ! Private variables
        character(len=FILENAME_LEN+500)                 ::  buffer      !< Will hold a line from the file
        character(len=RTP_LEN)                          ::  label       !< Will hold the label part of the line
        character(len=LINE_LEN)                         ::  value       !< Will hold the value part of the line
        character(len=WORD_LEN),        allocatable     ::  words(:)
        integer                                         ::  file_handle !< Handle for the file
        integer                                         ::  ios         !< ios is negative if an end of record condition is encountered or if
                                                                        !! an endfile condition was detected.  It is positive if an error was
                                                                        !! detected.  ios is zero otherwise.
        integer                                         ::  line        !< line number
        logical                                         ::  ffound
        logical,                        parameter       ::  debug   =   .false.

        ! Initialise private variables
        line    = 0
        ios     = 0
        ffound  = .false.


        ! Debug
        if (debug) write(*,'(6a)') '**debug(file_get_char): looking for ', trim(adjustl(parameter)),    &
                                    ', which has current value of ', trim(adjustl(rtp_value)), ' in file ', &
                                    trim(adjustl(filename))

        ! Maybe this needs to be a specific number?
        file_handle = available_lun()

        ! Open the file
        open(file_handle,file=FILENAME,iostat=IOS)
        if (ios .ne. 0) then
            write(*,'(a)') '**ERROR(file_get_char): failed to open file', trim(adjustl(filename))
            call terminate('Fatal error in FILE_GET_CHAR')
        endif

        ! Rewind to first line of file
        rewind(file_handle, iostat=IOS)
        if (ios .ne. 0) then
            write(*,'(a)') '**ERROR(file_get_char): failed to rewind file', trim(adjustl(filename))
            call terminate('Fatal error in FILE_GET_CHAR')
        endif

        if (debug) write(*,'(a)') '**debug(file_get_char): about to start looping through lines'

        ! Iterate through the lines of the file (when eof, ios will be set to non-zero)
        ios = 0
        line = 0
        do while (ios == 0)
            read(file_handle, '(a)', iostat=IOS) buffer
            if ( ios == 0 ) then
                line    =   line + 1

                !
                ! If the line is a comment or is blank, we can cycle
                !
                if (is_comment(buffer) .or. is_blank(buffer)) then
                    !print *, '**debug(fil_get_char): cycling line: ', trim(buffer), '!'
                    cycle
                endif

                !
                ! Split the line into words, the first of which is the label, and the second the value
                !
                if (debug) write(*,'(a,i0)') '**debug(file_get_char): about to split line ', line
                words = split(buffer)
                !if (debug) print '(a,i0,1x,a)', '**debug(file_get_char): size, words array = ', size(words), words
                label = words(1)
                if (size(words) .gt. 1) then
                    value = words(2)
                else
                    value = ''
                endif

                if (debug) print '(a,i0,5a)', '**debug(file_get_char): line, label, value: |', line,    &
                                                '|', trim(label), '|', trim(value), '|'


                !
                ! Is this the label we're looking for?
                !
                if ( StringsAreEqual(label,parameter,case_sensitive=.false.)) then
                    ! We've found the label we were looking for - transfer it to output variable, and stop scanning through the file
                    read(value,'(a)') rtp_value
                    if (debug) write(*,'(4a)') '**debug(file_get_char): found ', trim(parameter), '=', trim(value)
                    ffound = .true.
                    exit
                else
                    ! This line didn't contain the rtp we were looking for, do nothing
                endif
            elseif ( ios .gt. 0) then
                write(*,'(a,i3,a,a)') '**ERROR(file_get_char): io error encountered while parsing line',    &
                                        line, ' of file ', trim(filename)
                call terminate('Fatal error in FILE_GET_CHAR')
            elseif ( ios .lt. 0) then
                ! We reached the end of the file and still haven't found what we're looking for
                if (debug) write(*,'(2a)') '**debug(file_get_char): reached end of file, did not find parameter ', trim(parameter)
            endif
        enddo

        close(file_handle)
        call lun_is_now_available(file_handle)

        ! Cleanup
        if (allocated(words)) deallocate(words)

        !
        if (.not. ffound) then
            !rtp_value = ' '
        endif
        if (present(found)) then
            found = ffound
        endif

        ! Debug
        !write(*,*) '**debug(file_get_char): was looking for ', trim(parameter), ', which now has value of ', trim(rtp_value), '.'
    end subroutine file_get_char

    !>  \brief  Return the contents of desired line
    subroutine file_read_line_num(filename,line_num,line)
        use runtime_parameters
        implicit none
        ! Arguments
        character(len=*),       intent(in)  ::  filename    !<  Input filename
        integer,                intent(in)  ::  line_num    !<  Number of the line to be read in. First line in the file is numbered 1.
        character(len=*),       intent(out) ::  line        !<  Line read in
        ! Private variables
        integer ::  current_line_num, ios, lun
        logical ::  success
        ! Start work

        !print *, '**debug(file_read_line_num): line length = ', len(line)

        ! Check the file actually exists
        if (.not. file_exists(filename)) then
            write(*,'(a)') '**ERROR(file_read_line_num): file does not exist: ', trim(adjustl(filename))
            call terminate('Fatal error in FILE_READ_LINE_NUM')
        endif

        ! Check the line number makes sense
        if (line_num .lt. 1) then
            write(*,'(a,I'//integer_digits//')') '**ERROR(file_read_line_num): cannot read line number <1 : ', line_num
            call terminate('Fatal error in FILE_READ_LINE_NUM')
        else if (line_num .gt. number_of_lines_in_file(filename)) then
            write(*,'(a,i5,a,i5,a,a)')  '**ERROR(file_read_line_num): requested line number (', line_num,           &
                                        ') is greater than number of lines (', number_of_lines_in_file(filename),   &
                                        ') in file :', trim(adjustl(filename))
            call terminate('Fatal error in FILE_READ_LINE_NUM')
        endif

        ! Open the file
        lun = available_lun()
        open (unit=LUN,file=TRIM(adjustl(filename)),status='OLD',iostat=IOS)
        if (ios .ne. 0) then
            write(*,'(3a,i3)')  '**ERROR(file_read_line_num): file opening failed for file ', trim(adjustl(filename)),   &
                                ' with error ', ios
            call terminate('Fatal error in FILE_READ_LINE_NUM')
        endif

        ! Loop over the lines
        current_line_num = 0
        ios = 0
        success = .false.
        do while (ios == 0)
            current_line_num = current_line_num + 1
            read(lun,'(a)',iostat=IOS) line
            !print '(a,i3,a)', '**debug(file_read_line_num): line #, line : ', current_line_num, trim(adjustl(line))
            if (current_line_num == line_num) then
                success = .true.
                exit
            endif
        enddo

        if (.not. success) then
            line = ' '
            write(*,'(a,i5,a,a,a)') '**ERROR(file_read_line_num): did not find line number ', line_num, ' in file ', &
                                    trim(adjustl(filename)), '. Programming error?'
            call terminate('Fatal error in FILE_READ_LINE_NUM')
        endif

        close(lun)
        call lun_is_now_available(lun)
    end subroutine file_read_line_num


    !>  \brief  Find the first non-blank character in a string and return its position.
    pure integer function first_non_blank(string,back)
        implicit none
        ! Arguments
        character(len=*),       intent(in)  ::  string
        logical,    optional,   intent(in)  ::  back        !<  If .TRUE., we'll look for the last non-blank character in the string
        ! Private variables
        logical ::  bback
        ! Start work
        ! Reverse?
        if (present(back)) then
            bback = back
        else
            bback = .false.
        endif

        first_non_blank =   verify(string,blanks_and_new_lines,back=bback)
    end function first_non_blank

    !>  \brief  Find the first blank character in a string and return its position.
    pure integer function first_blank(string,back)
        implicit none
        ! Arguments
        character(len=*),       intent(in)  ::  string
        logical,    optional,   intent(in)  ::  back        !<  If .TRUE., we'll look for the last blank character in the string
        ! Private variables
        logical ::  bback
        ! Start work
        ! Reverse?
        if (present(back)) then
            bback = back
        else
            bback = .false.
        endif

        first_blank =   scan(string,blanks_and_new_lines,back=bback)
    end function first_blank






    !>  \brief Parse the command-line with which the program was invoked. Return relevant argument values.
    !!
    !!  Flags
    !!  -i  Run Frealix in interactive mode
    !!
    !!  The expected arguments:
    !!  (1) Control file name (optional)
    !!
    subroutine parse_command_line_arguments()

        use runtime_parameters

        implicit none

        ! Local variables
        character(200)      ::  COMMAND_LINE    ! Complete command line
        integer             ::  command_line_len    ! Number of characters in command line
        integer             ::  arg_count   ! Number of command-line arguments
        integer             ::  max_arg_count   ! Maximum number of arguments
        integer             ::  stat    ! Status variable for error-checking
        character(200)      ::  CURRENT_ARG ! Argument currently being parsed
        integer             ::  i

        max_arg_count   =   2

        control_filename = ''


        call get_command(command_line,command_line_len,stat)
        if (stat .eq. -1) then
            write(*,'(a)') '**ERROR (parse_command_line_arguments): command_line variable not long enough to hold the command line!'
            call terminate('Fatal error in PARSE_COMMAND_LINE_ARGUMENTS')
        elseif (stat .gt. 0) then
            write(*,'(a)') '**ERROR (parse_command_line_arguments): command line could not be retrieved'
            call terminate('Fatal error in PARSE_COMMAND_LINE_ARGUMENTS')
        elseif (stat .lt. -1) then
            write(*,'(a)') '**ERROR (parse_command_line_arguments): unknown error'
            call terminate('Fatal error in PARSE_COMMAND_LINE_ARGUMENTS')
        endif

        arg_count   =   command_argument_count()

        if (arg_count .gt. max_arg_count) then
            write(*,'(a,i1,a,i1,a,i1,a)')   '**WARNING (parse_command_line_arguments): ', arg_count,                            &
                                            ' command-line arguments were supplied, but only expecting ', max_arg_count,    &
                                            '. ', arg_count - max_arg_count, ' arguments will be ignored.'
        elseif (arg_count .eq. 0) then
            !write(*,'(a)') '**debug (parse_command_line_arguments): no arguments found, setting interactive mode'
            interactive = .true.
        endif

        ! If there is more than one argument, we need to parse them
        if (arg_count .gt. 0) then
            do I=1,arg_count

                call get_command_argument(i,current_arg,STATUS=STAT)
                if(stat .eq. -1) then
                    write(*,'(3a)')     '**ERROR (parse_command_line_arguments): argument supplied is longer', &
                                        ' than maximum allowable. Either give a shorter argument, or recompile ',   &
                                        'program with larger character string.'
                    call terminate('Fatal error in PARSE_COMMAND_LINE_ARGUMENTS')
                elseif(stat .ne. 0) then
                    call terminate('Unknown fatal error in PARSE_COMMAND_LINE_ARGUMENTS')
                endif

                ! Check if this is a flag
                if (current_arg(1:1) .eq. '-') then
                    if (current_arg(2:2) .eq. 'i') then
                        interactive = .true.
                    else
                        write(*,'(2a)') '**ERROR (parse_command_line_arguments): unrecognised flag: ', current_arg
                    endif
                else ! Not a flag
                    if (file_exists(current_arg)) then
                        control_filename = current_arg
                    else
                        write(*,'(2a)') '**warning(parse_command_line_arguments): this file does not exist: ', &
                                        trim(adjustl(current_arg))
                        write(*,'(a)')  '**warning(parse_command_line_argumnets): will run frealix in interactive mode'
                        control_filename = ''
                        interactive = .true.
                    endif
                endif
            enddo
        endif

        if ( control_filename .eq. '' .and. .not. interactive ) then
            write(*,'(a)') '**ERROR (parse_command_line_arguments): not in interactive mode, but no control filename either'
            call terminate('Fatal error in PARSE_COMMAND_LINE_ARGUMENTS')
        endif

    end subroutine parse_command_line_arguments


    !>  \brief  Get integer values from column col of file FILENAME. See the real equivalent for more details.
    subroutine get_ints_from_column(filename,col,array)
        use runtime_parameters
        implicit none
        ! Arguments
        character(len=FILENAME_LEN),    intent(in)      ::  filename    !< The flat-text file we're reading from
        integer,                        intent(in)      ::  col !< Which column
        integer,        allocatable,    intent(inout)   ::  array(:)    !< The array to fill with values from the file
        ! Private variables
        real, allocatable   ::  temparray(:)
        integer ::  i

        ! Check allocation status
        if (allocated(array)) then
            allocate(temparray(size(array)))
        endif
        call get_reals_from_column(filename,col,temparray)
        if (.not. allocated(array)) then
            allocate(array(size(temparray)))
        endif
        forall (I=1:size(temparray))
            array(i) = int(temparray(i))
        end forall
        deallocate(temparray)
    end subroutine get_ints_from_column

        !>  \brief  Get real values from column col of file FILENAME. The filename is assumed to contain space- or tab-seperated
    !!          values. The output array should be allocatable. If it is allocated already, it needs to be of correct size,
    !!          or an error may occur. Lines where the first character is a 'C' or a '#' are ignored.
    !!
    subroutine get_logicals_from_column(filename,col,array)
        use runtime_parameters
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  filename    !< The flat-text file we're reading from
        integer, intent(in) ::  col !< Which column
        logical, allocatable, intent(inout) ::  array(:)    !< The array to fill with values from the file
        ! Private variables
        integer ::  pos, pos2, current_col, number_of_lines, ierr, file_handle, ios, line
        character(len=FILENAME_LEN+500) ::  buffer      !< Will hold a line from the file

        file_handle = available_lun()

        ! Check the file exists
        if (.not. file_exists(filename)) then
            write(*,'(a,a)') ' **ERROR: this file does not exist: ', filename
            call terminate('Error in GET_LOGICALS_FROM_COLUMN')
        endif

        ! Count the number of lines
        number_of_lines = number_of_lines_in_file(filename,IGNORE_COMMENTS=.true.)

        ! If necessary, allocate the output array
        if (.not. allocated(array) ) then
            allocate(array(number_of_lines),stat=IERR)
            if (ierr .ne. 0) then
                write(*,*) '**ERROR (get_logicals_from_column): memory allocation failed for list of filenames'
                call terminate('Fatal error in GET_LOGICALS_FROM_COLUMN')
            endif
        endif

        ! Open the flat-text file
        open(file_handle,file=FILENAME)
        ! Go back to beginning of the file if we weren't there already
        rewind (file_handle, iostat=IOS)
        if (ios .gt. 0) write(*,*) 'Error in rewind'
        !
        ios = 0
        line = 0
        do while (ios == 0)
            read(file_handle, '(a)', iostat=IOS) buffer
            if (ios .eq. 0) then
                if ( is_comment(buffer) .or. is_blank(buffer) ) then
                    ! Ignore current line, it must be a comment.
                    cycle
                endif
                line = line + 1
                do current_col = 1, COL
                    !write(*,'(a,i3)') '**debug: processing column', current_col
                    !write(*,'(a,a)')  '**debug: buffer =', trim(buffer)
                    ! Find the first non-blank character
                    pos  = first_non_blank(buffer)
                    !write(*,'(a,a,a,i3)')  '**debug: first non-blank character is ', buffer(pos:pos), ' at position ', pos
                    ! Find the last non-blank character after that
                    pos2 = pos + first_blank(buffer(pos:)) - 1
                    !write(*,'(a,a,a,i3)')  '**debug: last non-blank character is ', buffer(pos2:pos2), ' at position ', pos2
                    if (pos .ne. 0 .and. pos2 .ne. 0) then ! if we found a column
                        !write(*,'(a,i3,a)') '**debug: column ', current_col, ' is ', buffer(pos:pos2)
                        if ( current_col .eq. col ) then ! We're on the column we're looking for
                            buffer = buffer(pos:pos2)
                            read(buffer,*) array(line)
                        else
                            buffer = buffer(pos2+1:)
                        endif
                    else
                        write(*,'(a,i0,a,i0,a,a)') '**ERROR (get_reals_from_column): Could not find field #', current_col,  &
                                    ' on line', line, ' of file ', trim(filename)
                        call terminate('Fatal error in GET_REALS_FROM_COLUM')
                    endif
                enddo ! End loop over columns
            endif
        enddo   ! End loop over lines
        ! Close the file
        close(file_handle)
        call lun_is_now_available(file_handle)
    end subroutine get_logicals_from_column


    subroutine get_reals_from_column(filename,col,array)
        use runtime_parameters
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  filename    !< The flat-text file we're reading from
        integer, intent(in) ::  col !< Which column
        real(kind=4), allocatable, intent(inout)    ::  array(:)    !< The array to fill with values from the file
        !
        real(kind=8), allocatable :: tmp_array(:)
        integer ::  i
        !
        call get_dbles_from_column(filename,col,tmp_array)

        ! this should not be necessary but gfortran 5.1.1 chokes on the auto realloc lhs
        ! when also doing a type conversion. See bug https://gcc.gnu.org/bugzilla/show_bug.cgi?id=52162
        if (allocated(array)) then
            if (size(array) .ne. size(tmp_array)) then
                deallocate(array)
                allocate(array(size(tmp_array)))
            endif
        else
            allocate(array(size(tmp_array)))
        endif
        do i=1,size(tmp_array)
            array(i) = real(tmp_array(i))
        enddo
        ! array = tmp_array
    end subroutine get_reals_from_column


    !>  \brief  Get real values from column col of file FILENAME. The filename is assumed to contain space- or tab-seperated
    !!          values. The output array should be allocatable. If it is allocated already, it needs to be of correct size,
    !!          or an error may occur. Lines where the first character is a 'C' or a '#' or a ';' are ignored.
    !!
    subroutine get_dbles_from_column(filename,col,array)
        use runtime_parameters
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  filename    !< The flat-text file we're reading from
        integer, intent(in) ::  col !< Which column
        real(kind=8), allocatable, intent(inout)    ::  array(:)    !< The array to fill with values from the file
        ! Private variables
        integer ::  pos, pos2, current_col, number_of_lines, ierr, file_handle, ios, line
        character(len=8192) ::  buffer      !< Will hold a line from the file
        logical,    parameter   ::  use_split = .false.
        character(len=word_len),    allocatable :: words(:)

        file_handle = available_lun()

        ! Check the file exists
        if (.not. file_exists(filename)) then
            write(*,'(a,a)') ' **ERROR: this file does not exist: ', filename
            call terminate('Error in GET_DBLES_FROM_COLUMN')
        endif

        ! Count the number of lines
        number_of_lines = number_of_lines_in_file(filename,IGNORE_COMMENTS=.true.)

        ! If necessary, allocate the output array
        if (.not. allocated(array) ) then
            allocate(array(number_of_lines),stat=IERR)
            if (ierr .ne. 0) then
                write(*,*) '**ERROR (get_dbles_from_column): memory allocation failed for list of filenames'
                call terminate('Fatal error in GET_DBLES_FROM_COLUMN')
            endif
        endif

        ! Open the flat-text file
        open(file_handle,file=FILENAME)
        ! Go back to beginning of the file if we weren't there already
        rewind (file_handle, iostat=IOS)
        if (ios .gt. 0) write(*,*) 'Error in rewind'
        !
        ios = 0
        line = 0
        do while (ios == 0)
            read(file_handle, '(a)', iostat=IOS) buffer
            if (ios .eq. 0) then
                if ( is_comment(buffer) .or. is_blank(buffer) ) then
                    ! Ignore current line, it must be a comment.
                    cycle
                endif
                line = line + 1
                if (use_split) then
                    words = split(buffer)
                    !print *, '**debug(): found ', size(words), 'words'
                    read(words(col),*) array(line)
                else
                    do current_col = 1, col
                        !write(*,'(a,i0)') '**debug: processing column', current_col
                        !write(*,'(a,a)')  '**debug: buffer =', trim(buffer)
                        !write(*,'(a,i0)') '**debug: line =', line
                        ! Find the first non-blank character
                        pos  = first_non_blank(buffer)
                        !write(*,'(a,a,a,i3)')  '**debug: first non-blank character is ', buffer(pos:pos), ' at position ', pos
                        ! Find the last non-blank character after that
                        if (pos .eq. 0) then
                            pos2 = 0
                        else
                            pos2 = pos + first_blank(buffer(pos:)) - 1
                        endif
                        !write(*,'(a,a,a,i3)')  '**debug: last non-blank character is ', buffer(pos2:pos2), ' at position ', pos2
                        if (pos .ne. 0 .and. pos2 .ne. 0) then ! if we found a column
                            !write(*,'(a,i3,a)') '**debug: column ', current_col, ' is ', buffer(pos:pos2)
                            if ( current_col .eq. col ) then ! We're on the column we're looking for
                                buffer = buffer(pos:pos2)
                                read(buffer,*) array(line)
                            else
                                buffer = buffer(pos2+1:)
                            endif
                        else
                            write(*,'(a,i0,a,i0,a,a)') '**ERROR (get_dbles_from_column): Could not find field #', current_col,     &
                                        ' on line ', line, ' of file ', trim(filename)
                            print *, pos, pos2, col
                            write(*,'(a,a)')  '**debug: buffer =', trim(buffer)
                            call terminate('Fatal error in get_dbles_from_column')
                        endif
                    enddo ! End loop over columns
                endif ! End of use_split?
            endif ! End of ios check
        enddo   ! End loop over lines
        ! Close the file
        close(file_handle)
        call lun_is_now_available(file_handle)
    end subroutine get_dbles_from_column

    !>  \brief  Emulates Perl's split routine, which returns an array of strings
    function split(line,separators) result(strings)
        use runtime_parameters, only : line_len, terminate, word_len
        implicit none
        ! Arguments
        character(len=*)                        ::  line        !<  Line to be split
        character(len=*),           optional    ::  separators  !<  Characters which separate words, if not present, default is blank characters (space, tabs...)
        ! Results
        character(len=word_len),    allocatable ::  strings(:)
        ! Private variables
        character(len=line_len)                 ::  buffer
        character(len=word_len),    allocatable ::  strings_tmp(:)
        integer                                 ::  pos1, pos2, word_count, ierr
        logical,                    parameter   ::  debug = .false.
        ! Start work
        if (is_blank(line)) then
            ! Line is blank, crash out
            write(*,'(a)') '**ERROR(split): cannot split a blank line'
            call terminate('Fatal error in SPLIT')
        endif
        if (debug) print '(3a)', '**debug(split): line = |', trim(line), '|'
        ! Start work in earnest
        buffer = trim(line)
        word_count = 0
        do
            if (present(separators)) then
                ! Find the first non-separator character
                pos1 = verify(buffer,separators,BACK=.false.)
                if (pos1 .ne. 0) then
                    ! Find the last non-separator character after that
                    pos2 = pos1 + scan(buffer(pos1:),separators,BACK=.false.) - 2
                endif
            else
                ! Find the first non-blank character
                pos1 = first_non_blank(buffer)
                if (pos1 .ne. 0) then
                    ! Find the last non-blank character after that
                    pos2 = pos1 + first_blank(buffer(pos1:)) - 2
                    if (debug) print '(a,2(i0,1x))', '**debug(split): pos1, pos2 = ', pos1, pos2
                endif
            endif
            if (pos1 .ne. 0 .and. pos2 .ne. 0) then ! if we found a word
                if (word_count .gt. 0) strings_tmp = strings
                word_count = word_count + 1
                if (allocated(strings)) deallocate(strings)
                allocate(strings(word_count),stat=ierr)
                if (ierr .ne. 0) then
                    write(*,'(a)') '**ERROR(split): could not allocate memory to output array'
                    call terminate('Fatal error in SPLIT')
                endif
                if (word_count .gt. 1) strings(1:word_count-1) = strings_tmp(1:word_count-1)
                strings(word_count) = ''
                if (debug) print '(3a)', '**debug(split): adding this word: |', buffer(pos1:pos2), '|'
                write(strings(word_count),'(a)') buffer(pos1:pos2)
                if (debug) print '(a)', '**debug(split): internal write was done'
                if ( POS2+1 .ge. len_trim(buffer)) then ! if we reached end of the line
                    if (debug) write(*,'(a)') '**debug(split): reached end of line'
                    exit
                else
                    if (debug) write(*,'(a)') '**debug(split): shortening the buffer'
                    buffer = buffer(POS2+1:)
                    if (debug) write(*,'(a)') '**debug(split): done shortening the buffer'
                endif
            else
                exit
            endif
        enddo

        if (debug) then
            print *, '**debug(split): result array = ', strings
        endif

    end function split

    !>  \brief  Parse a flat-text file containing a list of filenames, one per line. An array of file names is returned.
    !!
    subroutine parse_filelist_file(filename,array_of_filenames)

        use runtime_parameters
        implicit none

        ! Arguments
        character(len=FILENAME_LEN),                intent(in)  ::  filename                !< The flat-text file containing the list
        character(len=FILENAME_LEN), allocatable,   intent(out) ::  array_of_filenames(:)   !< The output list of files

        ! Private variables
        character(len=FILENAME_LEN+100) ::  buffer  !< Will hold a line from the file
        integer ::  line            =   0
        integer ::  number_of_lines
        integer ::  file_handle
        integer ::  ierr
        integer ::  ios             =   0           !< ios is negative if an end of record condition is encountered or if
                                                    !! an endfile condition was detected.  It is positive if an error was
                                                    !! detected.  ios is zero otherwise.

        ! Maybe this should be set more cleverly?
        file_handle =   11

        ! Check the file exists
        if (.not. file_exists(filename)) then
            write(*,'(a,a)') ' **ERROR: this file does not exist:', filename
            call terminate('Error in PARSE_FILELIST_FILE')
        endif

        ! Count the number of lines (maybe add some syntax checking in this loop to check filenames look kosher?)
        number_of_lines = number_of_lines_in_file(filename)

        ! Allocate the output array
        allocate(array_of_filenames(number_of_lines),stat=IERR)
        if (ierr .ne. 0) then
            write(*,*) '**ERROR (parse_filelist_file): memory allocation failed for list of filenames'
            call terminate('Fatal error in PARSE_FILELIST_FILE')
        endif

        ! Open the flat-text file
        open(file_handle,file=FILENAME)
        ! Go back to beginning of the file
        rewind (file_handle, iostat=IOS)
        if (ios .gt. 0) write(*,*) 'Error in rewind'
        ! Fill the array with filenames
        line = 0
        do while (ios == 0)
            read(file_handle, '(a)', iostat=IOS) buffer
            if (ios .eq. 0) then
                line = line + 1
                array_of_filenames(line) = trim(buffer)
            endif
        enddo
        ! Close the file
        close(file_handle)
    end subroutine parse_filelist_file

    logical function file_exists(filename)
        use runtime_parameters
        implicit none
        character(len=*), intent(in)    ::  filename
        inquire(file=TRIM(adjustl(filename)),exist=FILE_EXISTS)
        !print '(2a,l)', '**debug(file_exists): checking file ', trim(adjustl(filename)), file_exists
    end function file_exists

    !>  \brief  Equivalent to the unix grep command, except it returns only the first line that contains pattern
    function grep(filename,pattern,shift) result(line)
        use runtime_parameters, only : tmp_file_handle, terminate
        implicit none
        ! Arguments
        character(len=*),       intent(in)  ::  filename
        character(len=*),       intent(in)  ::  pattern
        integer,    optional,   intent(in)  ::  shift   !<  If shift is n, the line returned will the line n lines after that with the pattern
        ! Result
        character(len=512)  ::  line
        ! Private variables
        character(len=512)  ::  buffer
        integer ::  ios
        logical ::  pattern_found
        integer ::  sshift
        integer ::  shifted
        ! Start work
        ! Shift?
        if (present(shift)) then
            if (shift .lt. 0) then
                write(*,'(a,i3)') '**ERROR(grep): shift parameter has to be positive ', shift
                call terminate('Fatal error in GREP')
            endif
            sshift = shift
        else
            sshift = 0
        endif
        pattern_found = .false.
        ! Open the flat-text file
        open(tmp_file_handle,file=FILENAME,status='OLD',iostat=IOS)
        if (ios .gt. 0) then
            write(*,'(a,a,a)')  '**ERROR(grep): Error opening file ', trim(filename), '.'
        else if (ios .lt. 0) then
            write(*,'(a,a,a)')  '**ERROR(grep): Error opening file ', trim(filename), '. eof or EOR.'
        endif
        ! Make sure we are at beginning of the file
        rewind (tmp_file_handle,iostat=IOS)
        if (ios .ne. 0) then
            write(*,*) '**ERROR(grep): Error when REWINDing file ', trim(filename)
            call terminate('Fatal error in GREP')
        endif
        ! Read through the file, line by line
        ios = 0
        shifted = 0
        do while (ios == 0)
            read(tmp_file_handle, '(a)', iostat=IOS) buffer
            if (ios .eq. 0) then
                if (index(buffer,pattern) .ne. 0 .or. pattern_found) then
                    ! Found pattern in buffer
                    line = trim(buffer)
                    pattern_found = .true.
                    if (shifted .eq. sshift) then
                        exit
                    endif
                    shifted = shifted + 1
                endif
            endif
        enddo
        ! If we get here, it means the pattern was not found, so we return a blank
        if (pattern_found) then

        else
            line = ' '
        endif
        close(tmp_file_handle)
    end function grep


    !>  \brief  Return the number of lines in a file, optionally ignoring comment lines
    integer function number_of_lines_in_file(filename,ignore_comments) result(number_of_lines)
        use runtime_parameters
        implicit none
        ! Arguments
        character(len=FILENAME_LEN),    intent(in)  ::  filename
        logical,    optional,           intent(in)  ::  ignore_comments
        ! Private variables
        integer ::  file_handle = tmp_file_handle ! Maybe this should be set more cleverly?
        integer ::  ios = 0
        character(len=FILENAME_LEN+200) ::  buffer
        logical ::  iignore_comments

        ! Check whether we are ignoring comments. By default, we are not.
        if (present(ignore_comments)) then
            iignore_comments = ignore_comments
        else
            iignore_comments = .false.
        endif

        !write(*,*) '**debug: Counting lines in: ', trim(filename)
        ! Open the flat-text file
        open(file_handle,file=FILENAME,status='OLD',iostat=IOS)
        if (ios .gt. 0) then
            write(*,'(a,a,a)')  '**ERROR(number_of_lines_in_file): Error opening file ', trim(filename), '.'
        else if (ios .lt. 0) then
            write(*,'(a,a,a)')  '**ERROR(number_of_lines_in_file): Error opening file ', trim(filename), '. eof or EOR.'
        endif
        ! Make sure we are at beginning of the file
        rewind (file_handle,iostat=IOS)
        if (ios .ne. 0) then
            write(*,*) '**ERROR(number_of_lines_in_file): Error when REWINDing file ', trim(filename)
            call terminate('Fatal error in NUMBER_OF_LINES_IN_FILE')
        endif
        ! Count the number of lines
        number_of_lines = 0
        ios = 0
        do while (ios == 0)
            read(file_handle, *, iostat=IOS) buffer
            if (ios .eq. 0) then
                if (.not. is_blank(buffer)) then
                    if (iignore_comments) then
                        if (is_comment(buffer)) then
                            ! do nothing, this line is a comment
                        else
                            number_of_lines = number_of_lines + 1
                        endif
                    else
                        number_of_lines = number_of_lines + 1
                    endif
                endif
            else if (ios .gt. 0) then
                write(*,'(a,i3,a,a)')   '**ERROR(number_of_lines_in_file): Error reading line',     &
                                        NUMBER_OF_LINES+1, ' of file ', trim(filename)
            endif
        enddo
        ! Close the file
        close(file_handle)
        !write(*,*) '**debug:number_of_lines_in_file: result=',number_of_lines
    end function number_of_lines_in_file

    !>  \brief  Works out whether a character string is a comment
    logical function is_comment(line)
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  line
        ! Private variables
        integer ::  pos1
        ! Find the first non-blank character
        pos1    =   first_non_blank(line)

        ! If we didn't find a non-blank character, then the line must be blank!
        if (pos1 .eq. 0) then
            is_comment = .false.
        else if (scan(line(pos1:),'#!C;',BACK=.false.) .eq. 1) then
            ! The first non-blank character is a comment character. If that's a c, we need to check that the following character is a space.
            if (string_equal(line(pos1:pos1),'C')) then
                if (string_equal(line(POS1+1:POS1+1),' ')) then
                    is_comment = .true.
                else
                    is_comment = .false.
                endif
            else
                is_comment = .true.
            endif
        else
            is_comment = .false.
        endif
    end function is_comment

    !>  \brief Works out whether a character string is blank
    pure logical function is_blank(line)
        implicit none
        ! Argument
        character(len=*),   intent(in)  ::  line
        ! Start work
        if (trim(line) .eq. '') then
            is_blank = .true.
        else if (first_non_blank(line) .eq. 0) then
            is_blank = .true.
        else
            is_blank = .false.
        endif
    end function is_blank

    !>  \brief  Works out whether a character string is an integer
    pure logical function is_intg(line)
        implicit none
        ! Argument
        character(len=*),   intent(in)  ::  line
        ! Local variables
        integer ::  first_non_blank, last_non_blank
        ! Start work
        if (string_is_real(line)) then
            first_non_blank = verify(line,blank_characters)
            last_non_blank  = verify(line,blank_characters,BACK=.true.)
            if (scan(line(first_non_blank:last_non_blank),'.') .eq. 0) then
                is_intg = .true.
            else
                is_intg = .false.
            endif
        else
            is_intg = .false.
        endif
    end function is_intg

    !>  \brief  Works out whether a character string is a real
    pure logical function string_is_real(line)
        implicit none
        ! Argument
        character(len=*),   intent(in)  ::  line
        ! Local variables
        integer ::  first_non_blank, last_non_blank
        ! Start work
        first_non_blank = verify(line,blank_characters)
        last_non_blank  = verify(line,blank_characters,BACK=.true.)
        if (verify(line(first_non_blank:last_non_blank),'E+-0123456789.') .eq. 0) then
            string_is_real = .true.
        else
            string_is_real = .false.
        endif
    end function string_is_real

    !>  \brief  Works out whether a character string is a logical
    pure logical function is_logi(line)
        implicit none
        ! Argument
        character(len=*),   intent(in)  ::  line
        ! Local variables
        integer ::  first_non_blank, last_non_blank
        ! Start work
        first_non_blank = verify(line,blank_characters)
        last_non_blank  = verify(line,blank_characters,BACK=.true.)
        if (line(first_non_blank:last_non_blank) .eq. 'T' .or. line(first_non_blank:last_non_blank) .eq. 'F') then
            is_logi = .true.
        else
            is_logi = .false.
        endif
    end function is_logi

    !>  \brief  Works out whether a character string begins with a question mark
    pure logical function starts_with_question_mark(line)
        implicit none
        ! Argument
        character(len=*),   intent(in)  ::  line
        ! Local variables
        integer ::  first_non_blank
        ! Start work
        first_non_blank = verify(line,blank_characters)
        if (line(first_non_blank:first_non_blank) .eq. '?') then
            starts_with_question_mark = .true.
        else
            starts_with_question_mark = .false.
        endif
    end function starts_with_question_mark



    !>  \brief  Replace wildcard characters in a filename by integer string to yield an actual filename
    subroutine filename_from_wild(filename_wild,num,filename_out)
        use runtime_parameters
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  filename_wild   !<  Filename containing wildcards (only consecutive question marks are allowed)
        integer,            intent(in)  ::  num             !<  Number to be placed in the filename pattern
        character(len=*),   intent(out) ::  filename_out    !<  Filename
        ! Private variables
        integer ::  pos1, pos2, num_wildcards
        character(len=64) :: fmt
        ! Find where the question marks are
        call filename_wild_find_qmarks(filename_wild,pos1,pos2)
        ! How many question marks
        num_wildcards = pos2 - pos1 + 1
        write(fmt,'(i0,a,i0)') num_wildcards, '.', num_wildcards
        ! The actual filename
        write(filename_out,'(a,I'//trim(adjustl(fmt))//',a)')   filename_wild(1:POS1-1), num,   &
                                                                filename_wild(POS2+1:len_trim(filename_wild))
    end subroutine filename_from_wild

    !>  \brief  Find the integer in a filename, given the filename and the wildcard pattern
    !!  Example: If FILENAME='micrograph0010_bin4.mrc' and FILENAME_WILD='micrograph????_bin4.mrc', then num = 10
    function file_number_from_filename(filename,filename_wild) result (num)
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  filename        !<  Filename to grab the integer from
        character(len=*),   intent(in)  ::  filename_wild   !<  Filename pattern, including ? wildcards
        ! Result
        integer     ::  num             !<  Integer number found in the filename
        ! Private variables
        integer ::  pos1,pos2, num_wildcards
        character(len=64) :: fmt
        ! Start work
        ! Find where the question marks are in the filename pattern
        call filename_wild_find_qmarks(filename_wild,pos1,pos2)
        num_wildcards = pos2 - pos1 + 1
        write(fmt,'(i0)') num_wildcards
        ! Diagnostics
        !print '(2a)', '**debug(file_number_from_filename): filename      = ', trim(adjustl(filename))
        !print '(2a)', '**debug(file_number_from_filename): filename_wild = ', trim(adjustl(filename_wild))
        !print '(a,2(i3,x))', '**debug(file_number_from_filename): pos1, pos2 = ', pos1, pos2
        ! Read off the relevant part of the filename
        read(filename(pos1:pos2),'(I'//fmt//')') num
    end function file_number_from_filename

    !>  \brief  Check whether filename has question marks in it
    logical function filename_contains_qmarks(filename)
        ! Arguments
        character(len=*),   intent(in)  ::  filename
        ! Private variables
        integer ::  first_qmark_position
        ! Start work
        first_qmark_position = scan(filename,'?',BACK=.false.)
        filename_contains_qmarks = (first_qmark_position .gt. 0)
    end function filename_contains_qmarks

    !>  \brief  Given a filename with wildcards, find the first and last consecutive ?
    subroutine filename_wild_find_qmarks(filename_wild,pos_first,pos_last)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  filename_wild
        integer,            intent(out) ::  pos_first       !<  Position of the first question mark
        integer,            intent(out) ::  pos_last        !<  Position of the last question mark
        ! Private variables

        ! Start work

        ! Find the first question mark
        pos_first   =   scan(filename_wild,'?',BACK=.false.)
        if (pos_first .eq. 0) then
            write(*,'(2a)') '**ERROR(filename_wild_find_qmarks): cannot find question mark in filename ',   &
                             trim(adjustl(filename_wild))
            call terminate('Fatal error in FILENAME_WILD_FIND_QMARKS')
        endif

        ! Find the first non-? character following the first
        pos_last    =   verify(filename_wild(pos_first:),'?',BACK=.false.) + pos_first - 2

        ! Check there are only ? in between
        if (verify(filename_wild(pos_first:pos_last),'?') .ne. 0) then
            write(*,'(2a)') '**ERROR(filename_wild_find_qmarks): wrong wildcard formatting in filename ',   &
                            trim(adjustl(filename_wild))
            call terminate('Fatal error in FILENAME_WILD_FIND_QMARKS')
        endif
    end subroutine filename_wild_find_qmarks

    !>  \brief  Add a suffix to a filename, before its extension if it has one
    subroutine filename_add_suffix(filename_in,suffix,filename_out)
        use runtime_parameters, only : filename_len
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  filename_in     !<  Input filename
        character(len=*),   intent(in)  ::  suffix          !<  a string to be added onto the filename (e.g. '_fsc1')
        character(len=*),   intent(out) ::  filename_out    !<  Output filename
        ! Private variables
        character(len=FILENAME_LEN)     ::  filename_ext, filename_tmp
        ! Start work
        filename_ext = extension_from_filename(filename_in)
        call filename_remove_extension(filename_in,filename_tmp)
        filename_out = trim(adjustl(filename_tmp)) // trim(adjustl(suffix)) // '.' // trim(adjustl(filename_ext))
    end subroutine filename_add_suffix

    !>  \brief  Test whether two strings are equal, ignoring blank characters
    pure logical function string_equal(string1,string2)
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  string1
        character(len=*),   intent(in)  ::  string2
        ! Private variables
        integer ::  first_non_blank_1, first_non_blank_2
        integer ::  last_non_blank_1,  last_non_blank_2
        ! Start work
        first_non_blank_1 = verify(string1,blank_characters)
        last_non_blank_1  = verify(string1,blank_characters,BACK=.true.)
        first_non_blank_2 = verify(string2,blank_characters)
        last_non_blank_2  = verify(string2,blank_characters,BACK=.true.)

        !print *, '**debug(string_equal): |', string1(first_non_blank_1:last_non_blank_1), '|'
        !print *, '**debug(string_equal): |', string2(first_non_blank_2:last_non_blank_2), '|'
        !print '(4i)', first_non_blank_1, last_non_blank_1, first_non_blank_2, last_non_blank_2


        if (first_non_blank_1 .eq. 0 .and. first_non_blank_2 .eq. 0) then
            ! Both strings are blank
            string_equal = .true.
        else if (first_non_blank_1 .eq. 0 .or. first_non_blank_2 .eq. 0) then
            ! One string is blank, the other isn't
            string_equal = .false.
        else if (index(string1(first_non_blank_1:last_non_blank_1),string2(first_non_blank_1:last_non_blank_2)) .eq. 1  &
                .and. LAST_NON_BLANK_1-FIRST_NON_BLANK_1 .eq. LAST_NON_BLANK_2-FIRST_NON_BLANK_2) then
            ! Neither string is blank, and the strings match
            string_equal = .true.
        else
            ! All other cases
            string_equal = .false.
        endif

        !print *, '**debug(string_equal): result = ', string_equal
    end function string_equal



    !>  \brief Return a one letter code for the file format designated by the extension in the filename
    !!
    !!  If .mrc: m
    !!  If .spi: s
    !!  If .img: i
    !!  If .hed: i
    !!  Else: n
    pure function format_from_filename(filename)
        use runtime_parameters
        ! Arguments
        character(len=*),   intent(in)  ::  filename
        ! Return value
        character(len=1)    ::  format_from_filename
        ! Private variable
        character(len=3)    ::  extension
        ! Start work
        extension = extension_from_filename(filename)
        select case(extension)
            case ('img')
                format_from_filename = 'I'
            case ('hed')
                format_from_filename = 'I'
            case ('mrc')
                format_from_filename = 'M'
            case ('map')
                format_from_filename = 'M'
            case ('spi')
                format_from_filename = 'S'
            case default
                format_from_filename = 'N'
        end select
    end function format_from_filename

    !>  \brief  Return the 3-letter extension of a filename if present (without the period)
    pure function extension_from_filename(filename)
        use runtime_parameters
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  filename
        ! Return value
        character(len=3)    ::  extension_from_filename
        ! Private variables
        integer ::  length
        ! Start work - we look for a simple extension, where the filename ends with .???
        length = len_trim(filename)
        if (scan(filename(1:length),'.',BACK=.true.) .eq. LENGTH-3) then
            extension_from_filename = filename(LENGTH-2:length)
        else
            extension_from_filename = '   '
        endif
    end function extension_from_filename

    !>  \brief  Return the path part of a filename. For example if the filename is /a/b/c/d.e, return /a/b/c/
    function path_from_filename(filename) result(path)
        use runtime_parameters
        ! Arguments
        character(len=*),   intent(in)  ::  filename
        ! Return
        character(len=LINE_LEN)         ::  path
        ! Private variables
        integer ::  length
        integer ::  last_slash
        !
        length = len_trim(filename)
        last_slash = scan(filename(1:length),'/',BACK=.true.)
        if (last_slash .ne. 0) then
            path = filename(1:last_slash)
        else
            path = ''
        endif
    end function path_from_filename

    !>  \brief  Check whether the filename is that of an image, based on the extension in the filename
    pure logical function image_filename(filename)
        ! Argument
        character(len=*),   intent(in)  ::  filename
        ! Private variables
        character(len=3)    ::  ext
        ! Start work
        ext = extension_from_filename(trim(adjustl(filename)))
        if (    string_equal(ext,'mrc') .or.    &
                string_equal(ext,'spi') .or.    &
                string_equal(ext,'hed') .or.    &
                string_equal(ext,'img') ) then
            image_filename = .true.
        else
            image_filename = .false.
        endif
    end function image_filename

    !>  \brief  Replace the filename extension
    !!
    !!  Note, if the new extension is longer than the current extension, behaviour is not defined.
    subroutine filename_replace_extension(filename,new_extension)
        use runtime_parameters
        implicit none
        ! Arguments
        character(len=*),   intent(inout)   ::  filename        !<  Filename to be modified
        character(len=*),   intent(in)      ::  new_extension   !<   The filename's extension will be replaced by this
        ! Private variables
        character(len=:),   allocatable     ::  nnew_extension  ! Cleaner version of the new extension, without the first dot, and without trailing blanks
        character(len=:),   allocatable     ::  temp
        character(len=LEN(filename))        ::  temp2
        ! Start work

        !
        ! Clean up the new extension
        !
        allocate(character(LEN=LEN_TRIM(new_extension)) :: temp)
        temp = trim(adjustl(new_extension))
        if (temp(1:1) .eq. '.') then
            allocate(character(LEN=LEN(temp)-1) :: nnew_extension)
            nnew_extension = temp(2:len(temp))
        else
            allocate(character(LEN=LEN(temp)) :: nnew_extension)
            nnew_extension = temp
        endif

        !
        ! If the new extension is longer than 3 characters, this won't work
        !
        if (len(nnew_extension) .gt. 3) then
            write(*,'(3a)') '**ERROR(filename_replace_extension): can only deal with 3-character-long', &
                            ' extensions for now. New extension supplied: ', nnew_extension
            call terminate('Fatal error in FILENAME_REPLACE_EXTENSION')
        endif

        !
        ! Remove the extension from the filename
        !
        call filename_remove_extension(filename,temp2)

        !
        ! Add the new extension to the filename
        !
        filename = trim(adjustl(temp2)) // '.' // nnew_extension

    end subroutine filename_replace_extension

    !>  \brief  Return the filename without the .??? extension, if present
    subroutine filename_remove_extension(filename_in,filename_out)
        use runtime_parameters
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  filename_in     !<  Filename to be modified
        character(len=*),   intent(out) ::  filename_out    !<  Modified filename
        ! Private variables
        integer ::  length
        ! Start work
        filename_out = trim(adjustl(filename_in))
        if (extension_from_filename(filename_in) .eq. '   ') then
            ! Input filename does not have an extension, nothing to do
        else
            ! Remove the extension
            length = len_trim(filename_out)
            filename_out = filename_out(1:LENGTH-4)
        endif
    end subroutine filename_remove_extension

    !>  \brief  Convert string to all upper case
    !!  Adapted from
    !!  Figure 3.5B, pg 80, "Upgrading to Fortran 90", by Cooper Redwine,
    !!  1995 Springer-Verlag, New York.
    function upper_case(string)
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  string
        ! Result
        character(len=LEN(string))  ::  upper_case
        ! Private variables
        integer ::  i,n
        ! Start work
        upper_case = string
        do I=1,len(string)
            n = index(lower_case_letters,upper_case(i:i))
            if (n .ne. 0) UPPER_CASE(i:i) = upper_case_letters(n:n)
        enddo
    end function upper_case

    !>  \brief  Convert string to all upper case
    !!  Adapted from
    !!  Figure 3.5B, pg 80, "Upgrading to Fortran 90", by Cooper Redwine,
    !!  1995 Springer-Verlag, New York.
    function lower_case(string)
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  string
        ! Result
        character(len=LEN(string))  ::  lower_case
        ! Private variables
        integer ::  i,n
        ! Start work
        lower_case = string
        do I=1,len(string)
            n = index(upper_case_letters,lower_case(i:i))
            if (n .ne. 0) LOWER_CASE(i:i) = lower_case_letters(n:n)
        enddo
    end function lower_case

    !>  \brief  Convert a set specification to a list
    !!
    !!  For example, if the user gives "1-4,12-13,15,23", then the output will be an integer array: [1,2,3,4,12,13,15,23]
    function set_to_list(set)   result(list)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  set
        ! Result
        integer,    allocatable ::  list(:)
        ! Private variable
        integer ::  pos1,pos2,first_non_blank, pos_minus
        character(len=64)   ::  current_segment
        integer,    allocatable :: list_tmp(:)
        integer ::  num_in_list, new_value, new_value_start, new_value_finish
        integer ::  i
        logical ::  last_segment
        ! Start work

        ! Find the first non-blank character
        first_non_blank = verify(set,blank_characters)

        ! The first non-blank character had better be an integer
        pos1 = first_non_blank
        if (scan(set(pos1:pos1),integers) .eq. 0) then
            write(*,'(2a)') '**ERROR(set_to_list): the set specification does not start with an integer: ', trim(set)
            call terminate('Fatal error in SET_TO_LIST')
        endif

        ! We will loop from comma to comma
        do

            ! Find the next comma
            pos2 = scan(set(pos1:),',') + POS1-1
            if (pos2 .eq. POS1-1) then
                ! There was no remaining comma
                current_segment = set(pos1:)
                last_segment = .true.
            else
                ! We found a comma
                ! The character immediately following the comma should be an integer
                first_non_blank = verify(set(POS2+1:),blank_characters)
                if (scan(set(first_non_blank:first_non_blank),integers) .eq. 0) then
                    write(*,'(a,i0,a,a)') '**ERROR(set_to_list): the first non-blank character following the comma at position ', &
                                            pos2, ' is not an integer: ', trim(set(first_non_blank:first_non_blank))
                    call terminate('Fatal error in SET_TO_LIST')
                endif
                current_segment = set(pos1:POS2-1)
                last_segment = .false.
            endif

            ! Diagnostic
            !print *, '**debug(set_to_list): last segment = ', last_segment
            !print *, '**debug(set_to_list): pos1, pos2 = ', pos1, pos2
            !print *, '**debug(set_to_list): current segment = ', trim(current_segment)

            ! The current segment is only allowed to have integers or -
            if (verify(trim(current_segment),'1234567890-') .ne. 0) then
                write(*,'(2a)') '**ERROR(set_to_list): the current segment should only contain integer or - characters,',   &
                                ' but i found something else: ', trim(current_segment)
                call terminate('Fatal error in SET_TO_LIST')
            endif

            ! Find out whether the current segment contains a -
            pos_minus = scan(current_segment,'-')
            if (pos_minus .eq. 0) then
                ! Current segment does not contain a -, so we can read off the value directly
                read(current_segment,'(I'//integer_digits//')') new_value
                ! Diagnostic
                !print '(a,i3,a,a)', '**debug(set_to_list): found value ', new_value, ' from segment ', trim(current_segment)
                ! Add the new value to the end of the array, or to element 1 if we're at the beginning
                if (allocated(list)) then
                    num_in_list = size(list)
                    if (allocated(list_tmp)) deallocate(list_tmp)
                    allocate(list_tmp(num_in_list))
                    list_tmp = list
                    deallocate(list)
                    num_in_list = num_in_list + 1
                else
                    num_in_list = 1
                endif
                allocate(list(num_in_list))
                if (num_in_list .gt. 1) then
                    list(1:NUM_IN_LIST-1) = list_tmp
                endif
                list(num_in_list) = new_value
            else
                ! Current segment contains a -
                ! So let's read the two integers either side!
                read(current_segment(:POS_MINUS-1),'(I'//integer_digits//')') new_value_start
                read(current_segment(POS_MINUS+1:),'(I'//integer_digits//')') new_value_finish

                ! Diagnostic
                !print '(a,2(i3,x))', '**debug(set_to_list): start, finsh for current segment = ', new_value_start, new_value_finish

                ! How many new values will we add
                new_value = new_value_finish - new_value_start + 1

                ! Add the new value to the end of the array
                if (allocated(list)) then
                    num_in_list = size(list)
                    if (allocated(list_tmp)) deallocate(list_tmp)
                    allocate(list_tmp(num_in_list))
                    list_tmp = list
                    deallocate(list)
                else
                    num_in_list = 0
                endif
                num_in_list = num_in_list + new_value
                !print *, '**debug(set_to_list): num_in_list grown from ', NUM_IN_LIST-NEW_VALUE, ' to ', num_in_list
                allocate(list(num_in_list))
                if (size(list) .gt. new_value) then
                    list(1:NUM_IN_LIST-NEW_VALUE) = list_tmp
                endif
                do I=NEW_VALUE_START,new_value_finish
                    list(NUM_IN_LIST-NEW_VALUE+I-NEW_VALUE_START+1) = i
                enddo
            endif

            ! Diagnostic
            !print '(a,<NUM_IN_LIST>(i3,x))', '**debug(set_to_list): list has grown to: ', list

            ! Set pos1 to the beginning of the next segment (pos2 had the position of the comma)
            if (last_segment) then
                exit
            else
                pos1 = POS2+1
            endif
        enddo

    end function set_to_list

end module parsers

