!>  \brief  Plotting functionality
module plots

#if defined(have_plplot)
    use plplot
#endif

    use runtime_parameters, only : filename_len, word_len

    type plot_series
        real,   allocatable ::  xyz(:,:)    !<  Second index: 1-X, 2-Y, 3-Z
        integer             ::  rgb(3) = (/0,0,0/) !<  rgb colour to use when plotting the series (0 - 255)
        real                ::  alpha   =   1.0 !<  Transparency to use when plotting the series (0.0 - 1.0)
        logical             ::  draw_shadow =   .false. !<      Whether to draw shadow of this series on the 3 planes
        contains
            final   ::  plot_series_deallocate
    end type plot_series

    interface assignment(=)
        module procedure plot_series_copy_to_other
    end interface

    type plot
        integer                         ::  dimensionality  =   2               !<  Whether 3d or 2d
        logical                         ::  greyscale       =   .false.         !<    Whether greyscale or color
        character(len=FILENAME_LEN)     ::  filename        =   'test.png'
        type(plot_series),  allocatable ::  series(:)
        !
        character(len=WORD_LEN)         ::  x_axis_title    =   'X axis'
        character(len=WORD_LEN)         ::  y_axis_title    =   'Y axis'
        ! 3D-specific settings
        real                            ::  altitude        =   40.0                !<  The viewing altitude in degrees above the xy plane.
        real                            ::  azimuth         =   45.0                !<  The viewing azimuth in degrees. When az=0, the observer is
																					!!  looking face onto the zx plane, and as az is increased, the observer
																					!!  moves clockwise around the box when viewed from above the xy plane.
        character(len=WORD_LEN)         ::  z_axis_title    =   'Z axis'
        ! Range of values contained in series
        real                            ::  xyz_min(3)      =   -1.0
        real                            ::  xyz_max(3)      =   1.0
    end type plot

contains

    !>  \brief  Convenience function. Add a 2d series to a plot
    subroutine plot_add_xy(self,x,y)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(plot),                     intent(inout)   ::  self
        real,                           intent(in)      ::  x(:)
        real,                           intent(in)      ::  y(:)
        ! Private variables
        type(plot_series)   ::  series
        ! Start work
        if (size(x) .eq. size(y)) then
            if (allocated(series%xyz)) deallocate(series%xyz)
            allocate(series%xyz(size(x),2))
            series%xyz(:,1) = x
            series%xyz(:,2) = y
            call plot_add_series(self,series)
        else
            write(*,'(a)') '**ERROR(plot_add_twod): x and y series have different sizes'
            call terminate('Fatal error in PLOT_ADD_XY')
        endif
    end subroutine plot_add_xy

    subroutine plot_add_xyz(self,x,y,z,rgb,alpha,draw_shadow)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(plot),                     intent(inout)   ::  self
        real,                           intent(in)      ::  x(:)
        real,                           intent(in)      ::  y(:)
        real,                           intent(in)      ::  z(:)
        integer,        optional,       intent(in)      ::  rgb(3)
        real,           optional,       intent(in)      ::  alpha
        logical,        optional,       intent(in)      ::  draw_shadow
        ! Private variables
        type(plot_series)   ::  series
        ! Start work
        if (size(x) .eq. size(y) .and. size(x) .eq. size(z)) then
            if (allocated(series%xyz)) deallocate(series%xyz)
            allocate(series%xyz(size(x),3))
            series%xyz(:,1) = x
            series%xyz(:,2) = y
            series%xyz(:,3) = z
            if (present(rgb)) series%rgb = rgb
            if (present(alpha)) series%alpha = alpha
            if (present(draw_shadow)) series%draw_shadow = draw_shadow
            call plot_add_series(self,series)
        else
            write(*,'(a)') '**ERROR(plot_add_twod): x, y and z series have different sizes'
            call terminate('Fatal error in PLOT_ADD_XYZ')
        endif
    end subroutine plot_add_xyz


    !>  \brief  Add a series to a plot
    subroutine plot_add_series(self, series)
        implicit none
        ! Arguments
        type(plot),                     intent(inout)   ::  self
        type(plot_series),              intent(in)      ::  series
        ! Private variables
        type(plot_series),  allocatable ::  series_backup(:)
        integer ::  i
        ! Start work
#if defined (have_plplot)
        ! Allocate memory for the incoming series
        if (allocated(self%series)) then
            ! Backup current series
            series_backup = self%series
            ! Grow array of series by 1
            call plot_deallocate_all_series(self)
            deallocate(self%series)
            allocate(self%series(size(series_backup)+1))
            ! Restore the backup
            self%series(1:size(series_backup)) = series_backup(:)
        else
            ! Allocate memory for the array of series
            allocate(self%series(1))
        endif
        ! Copy the data over
        self%series(size(self%series)) = series
        ! Update the plot's min/max values
        if (size(self%series) .eq. 1) then
            do I=1,3
                self%xyz_min(i) = minval(self%series(1)%xyz(:,i))
                self%xyz_max(i) = maxval(self%series(1)%xyz(:,i))
            enddo
        else
            do I=1,3
                self%xyz_min(i) = min(self%xyz_min(i),minval(self%series(size(self%series))%xyz(:,i)))
                self%xyz_max(i) = max(self%xyz_max(i),maxval(self%series(size(self%series))%xyz(:,i)))
            enddo
        endif
#endif
    end subroutine plot_add_series

    !>  \brief  Copy all series from one plot object to the other
    subroutine plot_copy_all_series_to_other(self,other)
        implicit none
        ! Arguments
        type(plot),     intent(in)      ::  self
        type(plot),     intent(inout)   ::  other
        ! Private variables
        integer ::  i
        ! Start work

        ! Deallocate target's series
        call plot_deallocate_all_series(other)
        ! Allocate target's series
        allocate(other%series(size(self%series)))
        ! Copy series over
        do I=1,size(self%series)
            other%series(i) = self%series(i)
        enddo
    end subroutine plot_copy_all_series_to_other

    !>  \brief  Deallocate memory of all plot series contained in the plot object
    subroutine plot_deallocate_all_series(self)
        implicit none
        ! Arguments
        type(plot),     intent(inout)   ::  self
        ! Private variables
        integer :: i
        ! Start work
        if (allocated(self%series)) then
            do I=1,size(self%series)
                call plot_series_deallocate(self%series(i))
            enddo
        endif
    end subroutine plot_deallocate_all_series

    !>  \brief  Copy a plot series to another plot series
    subroutine plot_series_copy_to_other(other,self)
        implicit none
        ! Arguments
        type(plot_series),  intent(inout)   ::  other
        type(plot_series),  intent(in)      ::  self
        ! Private variables

        ! Start work
        ! Deallocate target
        call plot_series_deallocate(other)
        ! Copy
        other%xyz = self%xyz
        other%rgb = self%rgb
        other%alpha = self%alpha
        other%draw_shadow = self%draw_shadow
    end subroutine plot_series_copy_to_other

    !>  \brief  Deallocate memory of a plot series
    !elemental impure
    subroutine plot_series_deallocate(self)
        implicit none
        ! Arguments
        type(plot_series),  intent(inout)   ::  self
        ! Private variables

        ! Start work
        if (allocated(self%xyz)) deallocate(self%xyz)
    end subroutine plot_series_deallocate

    !>  \brief  Actually do the drawing
    subroutine plot_draw(self)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(plot),     intent(in)  ::  self
#if defined(have_plplot)
        ! Private varibles
        integer ::  i, current_shadow
        real,   allocatable ::  xyz_shadow(:,:)
        ! Start work

        if (self%dimensionality .eq. 3) then
            ! We'll be doing a png
            print *, '**DEBUG(plot_draw): filename = ', self%filename
            call png_plot_init(trim(adjustl(self%filename)))
            ! Advance to new (sub)page
            call pladv(0)
            ! Set up the viewport to take up the full (sub)page
            call plvpor(0.0_PLFLT,1.0_PLFLT,0.0_PLFLT,1.0_PLFLT)
            ! Specify world coordinates of viewport boundaries
            call plwind(-1.0_PLFLT,1.0_PLFLT,-1.0_PLFLT,1.0_PLFLT)
            ! Select color 1 (probably black)
            call plcol0(1)
            ! Set up a window for 3d plotting within the currently defined 2d window
            write(*,*) 'SELF%xyz_min = ', self%xyz_min
            write(*,*) 'SELF%xyz_max = ', self%xyz_max
            call plw3d( 1.0_PLFLT,  &   ! basex
                        1.0_PLFLT,  &   ! basey
                        1.0_PLFLT,  &   ! height
                        real(self%xyz_min(1),plflt),    &   ! xmin
                        real(self%xyz_max(1),plflt),    &   ! xmax
                        real(self%xyz_min(2),plflt),    &   ! ymin
                        real(self%xyz_max(2),plflt),    &   ! ymax
                        real(self%xyz_min(3),plflt),    &   ! zmin
                        real(self%xyz_max(3),plflt),    &   ! zmax
                        real(self%altitude,plflt),real(self%azimuth,plflt))
            ! Set up the 3d box with axes etc
            call threed_plot_box_init(self%x_axis_title,self%y_axis_title,self%z_axis_title)
            ! Draw all the lines
            do I=1,size(self%series)
                ! Set up color 2 to the color for that series
                call plscol0a(2,self%series(i)%rgb(1),    &
                                self%series(i)%rgb(2),    &
                                self%series(i)%rgb(3),    &
                                real(self%series(i)%alpha,plflt))
                ! Select color 2 for plotting
                call plcol0(2)
                ! Do the actual plotting
                call threed_plot_add_line(  self%series(i)%xyz(:,1), &
                                            self%series(i)%xyz(:,2), &
                                            self%series(i)%xyz(:,3))
                if (self%series(i)%draw_shadow) then
                    if (allocated(xyz_shadow)) deallocate(xyz_shadow)
                    allocate(xyz_shadow,MOLD=SELF%series(i)%xyz)
                    ! Set up color 2 to the color for that series
                    call plscol0a(2,self%series(i)%rgb(1),    &
                                    self%series(i)%rgb(2),    &
                                    self%series(i)%rgb(3),    &
                                    real(self%series(i)%alpha/10.,plflt))
                    ! Select color 2 for plotting
                    call plcol0(2)
                    ! Do the actual plotting
                    do CURRENT_SHADOW=1,3
                        xyz_shadow = self%series(i)%xyz
                        if (current_shadow .le. 2) then
                            xyz_shadow(:,current_shadow) = self%xyz_max(current_shadow)
                        else
                            xyz_shadow(:,current_shadow) = self%xyz_min(current_shadow)
                        endif
                        call threed_plot_add_line(  xyz_shadow(:,1), &
                                                    xyz_shadow(:,2), &
                                                    xyz_shadow(:,3))
                    enddo
                endif
            enddo
            ! Done
            call plot_complete()
        else
            call terminate('2D plots not yet implemented')
        endif

#else
        write(*,'(2a)') '**warning(plot_draw): plotting is not available because the PLplot library',    &
                        ' was not available at configure time'
#endif
    end subroutine plot_draw


    !>  \brief  Initialise a plot object
    subroutine plot_init(   self,dimensionality,greyscale,filename,x_axis_title,y_axis_title,  &
                            altitude,azimuth,z_axis_title )
        implicit none
        ! Arguments
        type(plot),                     intent(inout)   ::  self
        integer,            optional,   intent(in)      ::  dimensionality
        logical,            optional,   intent(in)      ::  greyscale
        character(len=*),   optional,   intent(in)      ::  filename
        character(len=*),   optional,   intent(in)      ::  x_axis_title
        character(len=*),   optional,   intent(in)      ::  y_axis_title
        real,               optional,   intent(in)      ::  altitude
        real,               optional,   intent(in)      ::  azimuth
        character(len=*),   optional,   intent(in)      ::  z_axis_title
        ! Private variables

        ! Start work
        if (present(dimensionality)) self%dimensionality = dimensionality
        if (present(greyscale)) self%greyscale = greyscale
        if (present(filename)) self%filename = filename
        if (present(x_axis_title)) self%x_axis_title = x_axis_title
        if (present(y_axis_title)) self%y_axis_title = y_axis_title
        if (present(altitude)) self%altitude = altitude
        if (present(azimuth)) self%azimuth = azimuth
        if (present(z_axis_title)) self%z_axis_title = z_axis_title
    end subroutine plot_init


    !
    ! routines below are the OLD-FASHIONED kind
    !
    !




    !>  \brief  Set up a 3d plot
    subroutine threed_plot_init(filename,altitude,azimuth,x_axis_title,y_axis_title,z_axis_title)
#if defined(have_plplot)
        use plplot
#endif
        implicit none
        ! Arguments
        character(len=*),               intent(in)  ::  filename        !<  Filename for the output png file
        real,               optional,   intent(in)  ::  altitude        !<  The viewing altitude in degrees above the xy plane.
        real,               optional,   intent(in)  ::  azimuth         !<  The viewing azimuth in degrees. When az=0, the observer is
																		!!  looking face onto the zx plane, and as az is increased, the
																		!!  observer moves clockwise around the box when viewed from above the xy plane.
        character(len=*),   optional,   intent(in)  ::  x_axis_title, y_axis_title, z_axis_title
#if defined(have_plplot)
        ! Private variables
        real(kind=PLFLT)    ::  pl_alt, pl_az
        ! Start work

        ! Work out altitude and azimuth
        if (present(altitude)) then
            pl_alt = altitude
        else
            pl_alt = 40.0
        endif
        if (present(azimuth)) then
            pl_az = azimuth
        else
            pl_az = 45.0
        endif

        ! We'll be doing
        call png_plot_init(filename)

        ! Advance to new (sub)page
        call pladv(0)
        ! Set up the viewport to take up the full (sub)page
        call plvpor(0.0_PLFLT,1.0_PLFLT,0.0_PLFLT,1.0_PLFLT)
        ! Specify world coordinates of viewport boundaries
        call plwind(-1.0_PLFLT,1.0_PLFLT,-1.0_PLFLT,1.0_PLFLT)
        ! Select color 1 (probably black)
        call plcol0(1)
        ! Set up a window for 3d plotting within the currently defined 2d window
        call plw3d(1.0_PLFLT,1.0_PLFLT,1.0_PLFLT,-1.0_PLFLT,1.0_PLFLT,-1.0_PLFLT,1.0_PLFLT,-1.0_PLFLT,1.0_PLFLT,pl_alt,pl_az)
        ! Set up the 3d box with axes etc
        call threed_plot_box_init(x_axis_title,y_axis_title,z_axis_title)

#else
        write(*,'(a)') '**warning(): plotting is not available because the PLplot library was not available at configure time'
        !call terminate('This routine should not have been called, since PLplot is not available')
#endif

    end subroutine threed_plot_init

    !>  \brief  Draws axes, numeric and text labels for a three-dimensional surface plot.
    subroutine threed_plot_box_init(x_axis_title,y_axis_title,z_axis_title)
        use runtime_parameters, only : line_len
#if defined(have_plplot)
        use plplot
#endif
        implicit none
        ! Arguments
        character(len=*),   optional,   intent(in)  ::  x_axis_title, y_axis_title, z_axis_title
        ! Private variables
        character(len=LINE_LEN) ::  pl_x_axis_title, pl_y_axis_title, pl_z_axis_title
        ! Start work

        ! Work out axes titles
        if (present(x_axis_title)) then
            pl_x_axis_title = x_axis_title
        else
            pl_x_axis_title = 'X axis'
        endif
        if (present(y_axis_title)) then
            pl_y_axis_title = y_axis_title
        else
            pl_y_axis_title = 'Y axis'
        endif
        if (present(z_axis_title)) then
            pl_z_axis_title = z_axis_title
        else
            pl_z_axis_title = 'Z axis'
        endif

!        call plbox3('bnstu', pl_x_axis_title, 0.0_plflt, 0, &
!        'bnstu', pl_y_axis_title, 0.0_plflt, 0, &
!        'bcdmnstuv', pl_z_axis_title, 0.0_plflt, 0)
#if defined(have_plplot)
        call plbox3('bntu', trim(adjustl(pl_x_axis_title)), 0.0_plflt, 0, &
        'bntu', trim(adjustl(pl_y_axis_title)), 0.0_plflt, 0, &
        'bcdfntu', trim(adjustl(pl_z_axis_title)), 0.0_plflt, 0)
#endif
    end subroutine threed_plot_box_init

    !>  \brief  Add a solid line to a 3d plot
    subroutine threed_plot_add_line(x,y,z)
#if defined(have_plplot)
        use plplot
#endif
        implicit none
        ! Arguments
        real,       intent(in)  ::  x(:), y(:), z(:)
        ! Private variables
#if defined(have_plplot)
        real(kind=PLFLT), allocatable    ::  pl_x(:), pl_y(:), pl_z(:)
        ! Start work
        pl_x = x
        pl_y = y
        pl_z = z
        ! Draw the line
        call plline3(pl_x,pl_y,pl_z)
#endif
    end subroutine threed_plot_add_line

    !>  \brief  Complete the plot
    subroutine plot_complete()
#if defined(have_plplot)
        use plplot
#endif
        implicit none
        ! Start work
#if defined(have_plplot)
        call plend()
#endif
    end subroutine plot_complete


    !>  \brief  Setup a plot to be written to png file
    subroutine png_plot_init(filename)
#if defined(have_plplot)
        use plplot
#endif
        implicit none
        ! Arguments
        character(len=*),       intent(in)      ::  filename        !<  Filename for the output png file

        ! Start work
#if defined(have_plplot)
        ! We'll write out using the png driver
        call plsdev('png')
        ! Set the filename
        call plsfnam(filename)
        ! Setup default colormap the way i like it
        call modify_default_colormap(GREY=.true.)
        ! Initialise the plot
        call plinit
#endif
    end subroutine png_plot_init

    !>  \brief  Plot curves together in a single png image file.
    subroutine threed_plot_to_png(filename,ys,xs,curve_titles,plot_title,x_min,x_max,y_min,y_max,x_title,y_title,overlay_ys)
#if defined(have_plplot)
        use plplot
#endif
        use runtime_parameters, only : terminate, line_len
        implicit none
        ! Arguments
        character(len=*),                               intent(in)  ::  filename        !<  Filename for the output png file
        real,                                           intent(in)  ::  ys(:,:)         !<  The first dimension is for the y value (per x value). The second dimension is used to index different curves.
        real,                               optional,   intent(in)  ::  xs(:)           !<  The ordinates (x values). The size of this array should be identical to the size of the ys array in the first dimension.
        character(len=*),                   optional,   intent(in)  ::  curve_titles(:) !<  For each series, a title. This array must have dimension equal to the 2nd dimension of the ys array.
        character(len=*),                   optional,   intent(in)  ::  plot_title      !<  a title for the plot
        real,                               optional,   intent(in)  ::  x_min,x_max,y_min,y_max !<  Override the ranges covered by the x and y axes
        character(len=*),                   optional,   intent(in)  ::  x_title, y_title        !<  Override the default titles fo the x and y axes
        real,                               optional,   intent(in)  ::  overlay_ys(:)           !<  An extra curve to be plotted in light grey, dashed
        ! Private variables
        logical,    parameter   ::  debug   =   .false.

#if defined(have_plplot)
        ! Private variables
        real(kind=PLFLT)    ::  ys_pl(size(ys,1),size(ys,2))
        real(kind=PLFLT)    ::  xs_pl(size(ys,1))
        real(kind=PLFLT)    ::  overlay_ys_pl(size(ys,1))
        real(kind=PLFLT)    ::  x_axis_min, x_axis_max, y_axis_min, y_axis_max
        integer             ::  i
        character(len=LINE_LEN) ::  x_axis_title, y_axis_title, pplot_title
        integer,    parameter   ::  max_number_of_series    =   16
        type(plot)          :: fig
        ! Private variables to do with the legend
        real :: x(50), y(50), z(50), xx(50), yy(50), zz(50)



        ! Start work
        print *, 'NEW'
        call plot_init(fig,DIMENSIONALITY=3,FILENAME=FILENAME,   &
                        ALTITUDE=20.0,AZIMUTH=45.0)
        !
        do I=1,50
            x(i)=COS(real(i)*3.14*2./49.)
            y(i)=SIN(real(i)*3.14*2./49.)
            z(i)=0.0
        enddo
        call plot_add_xyz(fig,x,y,z,DRAW_SHADOW=.true.)
        do I=1,50
            z(i)=COS(real(i)*3.14*2./49.)
            x(i)=SIN(real(i)*3.14*2./49.)
            y(i)=0.0
        enddo
        call plot_add_xyz(fig,x,y,z,DRAW_SHADOW=.true.)
        call plot_draw(fig)
#else
        write(*,'(2a)') '**warning(series_to_png): plotting is not available because the PLplot library ',  &
                        'was not available at configure time'
        !call terminate('This routine should not have been called, since PLplot is not available')
#endif
    end subroutine threed_plot_to_png


    !>  \brief  Set up my preferred colormap by modifying the PLplot default
    subroutine modify_default_colormap(grey)
        implicit none
        ! Arguments
        logical,    optional,   intent(in)  ::  grey    !<  Whether the colormap should be modified to be greyscale
        ! Private variables
        logical ::  ggrey
        ! Start work
        if (present(grey)) then
            ggrey = grey
        else
            ggrey = .false.
        endif
#if defined(have_plplot)
        ! i prefer white as a background color, so i replace color 0 in the default colormap
        call plscol0 (0,255, 255, 255)
        if (ggrey) then
            call plscol0(1,0,0,0)
            call plscol0(2,200,200,200)
        else
            ! Since color 15 in the colormap was white, i set that to red
            call plscol0(15,255,0,0)
            ! And we set the default foreground color 1 to black
            call plscol0(1,0,0,0)
            ! Yellow is not visible on white background so we replace it with dark orange
            call plscol0(2,255,140,0)
        endif
#endif
    end subroutine modify_default_colormap

    !>  \brief  Plot curves together in a single png image file.
    subroutine series_to_png(filename,ys,xs,curve_titles,plot_title,x_min,x_max,y_min,y_max,x_title,y_title,overlay_ys)
#if defined(have_plplot)
        use plplot
#endif
        use runtime_parameters, only : terminate, line_len
        implicit none
        ! Arguments
        character(len=*),                               intent(in)  ::  filename        !<  Filename for the output png file
        real,                                           intent(in)  ::  ys(:,:)         !<  The first dimension is for the y value (per x value). The second dimension is used to index different curves.
        real,                               optional,   intent(in)  ::  xs(:)           !<  The ordinates (x values). The size of this array should be identical to the size of the ys array in the first dimension.
        character(len=*),                   optional,   intent(in)  ::  curve_titles(:) !<  For each series, a title. This array must have dimension equal to the 2nd dimension of the ys array.
        character(len=*),                   optional,   intent(in)  ::  plot_title      !<  a title for the plot
        real,                               optional,   intent(in)  ::  x_min,x_max,y_min,y_max !<  Override the ranges covered by the x and y axes
        character(len=*),                   optional,   intent(in)  ::  x_title, y_title        !<  Override the default titles fo the x and y axes
        real,                               optional,   intent(in)  ::  overlay_ys(:)           !<  An extra curve to be plotted in light grey, dashed
        ! Private variables
        logical,    parameter   ::  debug   =   .false.

#if defined(have_plplot)
        ! Private variables
        real(kind=PLFLT)    ::  ys_pl(size(ys,1),size(ys,2))
        real(kind=PLFLT)    ::  xs_pl(size(ys,1))
        real(kind=PLFLT)    ::  overlay_ys_pl(size(ys,1))
        real(kind=PLFLT)    ::  x_axis_min, x_axis_max, y_axis_min, y_axis_max
        integer             ::  i
        character(len=LINE_LEN) ::  x_axis_title, y_axis_title, pplot_title
        integer,    parameter   ::  max_number_of_series    =   16
        ! Private variables to do with the legend
        integer             ::  nlegend

        real(kind=PLFLT)    ::  legend_width, legend_height
        integer             ::  opt_array(size(ys_pl,2)), text_colors(size(ys_pl,2)), line_colors(size(ys_pl,2)), &
                                line_styles(size(ys_pl,2)), line_widths(size(ys_pl,2)), symbol_colors(size(ys_pl,2)), &
                                symbol_numbers(size(ys_pl,2))
        real(kind=PLFLT)    ::  symbol_scales(size(ys_pl,2)), box_scales(0)
        integer             ::  box_colors(0), box_patterns(0)
        integer             ::  box_line_widths(0)
        character(len=20)   ::  text(size(ys_pl,2))
        character(len=1)    ::  symbols(size(ys_pl,2))



        ! Start work

        ! Check input arrays have matching dimensions
        if (present(xs)) then
            if (size(xs) .ne. size(ys,1)) then
                write(*,'(a,i0,a,i0,a)')    '**ERROR(series_to_png): x values array extent (', size(xs), &
                                            ') does not match extent of y array in first dimension (', size(ys,1), ')'
                call terminate('Fatal error in SERIES_TO_PNG')
            endif
        endif
        if (present(curve_titles)) then
            if (size(curve_titles) .ne. size(ys,2)) then
                write(*,'(a,i0,a,i0)')  '**ERROR(series_to_png): ', size(curve_titles), ' titles were provided for the ', &
                                        size(ys,2), ' series'
                call terminate('Fatal error in SERIES_TO_PNG')
            endif
        endif
        if (present(overlay_ys)) then
            if (size(overlay_ys) .ne. size(ys,1)) then
                write(*,'(a,i0,a,i0,a)')    '**ERROR(series_to_png): overlay y values array extent (', size(xs),    &
                                            ') does not match extent of y array in first dimension (', size(ys,1), ')'
                call terminate('Fatal error in SERIES_TO_PNG')
            endif
        endif

        ! Warn the user if we're not going to plot all the series
        if (size(ys,2) .gt. max_number_of_series) then
            write(*,'(a,i0,a,i0,a)')    '**warning(series_to_png): will only plot the first ', max_number_of_series, ' of the ', &
                                        size(ys,2), ' series you submitted'
        endif

        ! Default labels
        if (present(plot_title)) then
            pplot_title = trim(adjustl(plot_title))
        else
            pplot_title = 'Frealix-generated plot'
        endif
        if (present(x_title)) then
            x_axis_title = x_title
        else
            x_axis_title = 'X axis'
        endif
        if (present(y_title)) then
            y_axis_title = y_title
        else
            y_axis_title = 'Y axis'
        endif
        !

        ! Copy values over to arrays of correct type
        if (present(xs)) then
            xs_pl = xs
        else
            do I=1,size(xs_pl)
                xs_pl(i) = i
            enddo
        endif
        ys_pl = ys

        ! We'll write out using the png driver
        call plsdev('png')
        ! Set the filename
        call plsfnam(filename)
        ! i prefer white as a background color, so i replace color 0 in the default colormap
        call plscol0 (0,255, 255, 255)
        ! Since color 15 in the colormap was white, i set that to red
        call plscol0(15,255,0,0)
        ! And we set the default foreground color 1 to black
        call plscol0(1,0,0,0)
        ! Yellow is not visible on white background so we replace it with dark orange
        call plscol0(2,255,140,0)
        ! Initialise the plot
        call plinit

        ! Work out the axes min max
        if (present(x_min)) then
            x_axis_min = x_min
        else
            x_axis_min = minval(xs_pl)
        endif
        if (present(x_max)) then
            x_axis_max = x_max
        else
            x_axis_max = maxval(xs_pl)
        endif
        if (present(y_min)) then
            y_axis_min = y_min
        else
            y_axis_min = minval(ys_pl(:,:))
        endif
        if (present(y_max)) then
            y_axis_max = y_max
        else
            y_axis_max = maxval(ys_pl(:,:))
        endif

        !
        call plenv(x_axis_min,x_axis_max,y_axis_min,y_axis_max,0,0)
        ! Do the labels
        call pllab(trim(adjustl(x_axis_title)),trim(adjustl(y_axis_title)),trim(adjustl(pplot_title)))
        ! Do the actual drawing of the lines
        do I=1,minval([SIZE(ys_pl,2),MAX_NUMBER_OF_SERIES])
            if (debug) print '(a,i0)', '**debug(series_to_png): drawing line ', i
            ! Set the color
            line_colors(i) = mod(i,14)+1
            call plcol0(line_colors(i))
            ! Set the line style
            line_styles(i) = minval([CEILING(i/15.),8])
            call pllsty(line_styles(i))
            ! Set the line width
            line_widths(i) = 1
            ! Deal with the legend entry for this line
            opt_array(i)   = pl_legend_line
            text_colors(i) = line_colors(i)
            if (present(curve_titles)) then
                write(text(i),'(a)') adjustl(trim(curve_titles(i)))
            else
                write(text(i),'(i0)') i
            endif
            ! Draw the line
            !if (debug) then
            !   write(*,'(a,<SIZE(xs_pl)>(F0.3,x))') '**debug(series_to_png): x values = ', xs_pl
            !   write(*,'(a,<SIZE(xs_pl)>(F0.3,x))') '**debug(series_to_png): y values = ', ys_pl(:,i)
            !endif
            call plline(xs_pl,ys_pl(:,i))
            !if (debug) write(*,'(a,i0,a)') '**debug(series_to_png): line ', i, ' is done'
        enddo
        ! If the user asked for an overlay, draw it now
        if (present(overlay_ys)) then
            if (debug) print '(a)', '**debug(series_to_png): drawing overlay line'
            call plcol0(7)  ! grey
            call pllsty(3)  ! long dashes
            overlay_ys_pl = overlay_ys
            call plline(xs_pl,overlay_ys_pl)
        endif
        ! Draw the legend
        if (debug) print '(a)', '**debug(series_to_png): drawing legend'
        nlegend = minval([SIZE(ys_pl,2),MAX_NUMBER_OF_SERIES])
        call pllegend(  legend_width, legend_height, &
                        0, 0, &
                        0.0_PLFLT, 0.0_PLFLT, 0.1_PLFLT, 15, &
                        1, 1, 0, 0, &
                        nlegend, opt_array, &
                        1.0_PLFLT, 1.0_PLFLT, 2.0_PLFLT, &
                        1.0_PLFLT, text_colors, text, &
                        box_colors, box_patterns, box_scales, box_line_widths, &
                        line_colors, line_styles, line_widths, &
                        symbol_colors, symbol_scales, symbol_numbers, symbols )
        ! We're done
        call plend
        if (debug) print '(a)', '**debug(series_to_png): all done'
#else
        write(*,'(2a)') '**warning(series_to_png): plotting is not available because the PLplot library',   &
                        ' was not available at configure time'
        !call terminate('This routine should not have been called, since PLplot is not available')
#endif
    end subroutine series_to_png

end module plots
