!>  \brief  Classes and methods to deal with single particle point group symmetries
!!
module point_symmetries
    use rotation_matrices

    implicit none
    private
    public  ::  point_symmetry, point_symmetry_symbol, point_symmetry_init, point_symmetry_print

    !>  \brief  Describes point group symmetry using Sch�nflies notation
    !!
    type    ::  point_symmetry_symbol
        character(len=3)    ::  symbol = 'C1 '      !<  Point group symmetry using Sch�nflies notation
    end type

    !>  \brief  Describes point group symmetry as a set of rotation matrices
    !!
    type    ::  point_symmetry
        integer                     ::  iredun = 0                  !<  Total symmetry redundancy
        type(rotation_matrix),      allocatable ::  symop(:)                    !<  Symmetry operators (rotation matrices)
        type(point_symmetry_symbol) ::  symbol                      !<  Sch�nflies notation
        logical     ::  helical = .false.
    end type

    contains


        !>  \brief  Print out point group symmetry matrices
        subroutine point_symmetry_print(self)
            implicit none
            ! Arguments
            type(point_symmetry),   intent(in)  ::  self
            ! Private variables
            integer ::  i,nsym
            ! Start work
            nsym = size(self%symop)
            write(*,'(/3a)') '**info(point_symmetry_print): symmetry operator matrices for ',   &
                                self%symbol%symbol, ' point group follow'
            do I=1,nsym
                write(*,'(3(A,i2),a)') 'Symmetry operator', i, ' of ', nsym, ' with ', self%symop(i)%jmax, '-fold redundancy:'
                call rotation_matrix_print(self%symop(i))
            enddo
            write(*,'(3a,i3/)') '**info(point_symmetry_print): Total symmetry redundancy for point group ', &
                                    self%symbol%symbol, ' is ', self%iredun
        end subroutine point_symmetry_print

        !>  \brief  Initialise a point_symmetry object, given Sch�nflies symmetry symbol
        !!
        subroutine point_symmetry_init(self,sym_symbol, helical_symmetry)
            use runtime_parameters, only : pi,terminate
            use rotation_matrices, only : identity_rotmat
            implicit none
            ! Arguments
            type(point_symmetry),       intent(out) ::  self                !<  Point group symmetry matrices to initialise
            type(point_symmetry_symbol),    intent(in)  ::  sym_symbol          !<  Point group symmetry using Sch�nflies notation
            logical, optional,              intent(in)  ::  helical_symmetry    !<  If this symmetry operator is used in the context of helical work (frealix), set this to true
            ! Private variables
            integer                 ::  j,k,l
            integer                 ::  nget,nsym,nasym
            integer                 ::  n_cdtoi
            integer                 ::  iredun
            integer,parameter,dimension(4,6)    ::  nsymtype =  reshape(    SOURCE=    &
                                                                                            (/  8,0,0,0,  &
                                                                                                8,5,0,0,  &
                                                                                                4,5,6,0,  &
                                                                                                3,7,4,0,  &
                                                                                                1,4,5,6,  &
                                                                                                2,4,5,6   /), &
                                                                            shape =         (/  4,6/) &
                                                                        )
            integer,parameter,dimension(6)      ::  NSYM_CDTOI = [1,2,3,3,4,4]
            real                    ::  angle
            real,dimension(3,3,8)           ::  SYMSTORE =      reshape(    SOURCE= &
                                        (/  0.309017,0.809017,-0.5,     -0.809017,0.5,0.309017,         0.5,0.309017,0.809017,  &
                                            0.809017,0.309017,0.5,      0.309017,0.5,-0.809017,         -0.5,0.809017,0.309017, &
                                            0.,  1.,  0.,               -1.,  0.,  0.,                  0.,  0.,  1.,           &
                                            0.,  0.,  1.,               1.,  0.,  0.,                   0.,  1.,  0.,           &
                                            1.,  0.,  0.,               0., -1.,  0.,                   0.,  0., -1.,           &
                                            -1.,  0.,  0.,              0.,  1.,  0.,                   0.,  0., -1.,           &
                                            0.,  1.,  0.,               1.,  0.,  0.,                   0.,  0., -1.,           &
                                            0.,  0.,  0.,               0.,  0.,  0.,                   0.,  0.,  1.    /),     &
                                        SHAPE= (/3,3,8/) &
                                                                        )
            character(len=1)            ::  symbol1
            character(len=1),parameter,dimension(16)::  asymtest = [' ','C','D','T','O','I','0','1','2','3','4','5','6','7','8','9']
            ! Start work
            self%symbol = sym_symbol
            call parse_sym_symbol(sym_symbol,symbol1,nasym)
            if (nasym.ge.0 .and. symbol1 .ne. 'H') then
                write(*,'(2a)') '**info(point_symmetry_init): initializing symmetry matrices for point group ', sym_symbol%symbol
                n_cdtoi = 0
                iredun = 1
                do j = 2,5
                    if(symbol1.eq.asymtest(j)) n_cdtoi = J-1
                enddo
                if (symbol1.eq.asymtest(6)) then
                    ! We have icosahedral symmetry defined as in the Int Tables...
                    n_cdtoi = 5
                    ! .. or as defined by Tony Crowther
                    if (nasym.eq.2) n_cdtoi = 6
                endif

                ! Calculate N-fold rotation matrix (symbols c and d)
                if ((N_CDTOI.eq.1).or.(N_CDTOI.eq.2)) then
                    angle = 2.0*PI/nasym
                    symstore(1,1,8) = cos(angle)
                    symstore(1,2,8) = sin(angle)
                    symstore(2,1,8) =-sin(angle)
                    symstore(2,2,8) = cos(angle)
                endif

                nsym = nsym_cdtoi(n_cdtoi)
                allocate(self%symop(nsym))
                do j = 1,nsym
                    nget = nsymtype(j,n_cdtoi)
                    do k = 1,4
                        do l = 1,4
                            if ((K.le.3).and.(L.le.3)) then
                                self%symop(j)%rmatrix(k,l) = symstore(k,l,nget)
                            else
                                self%symop(j)%rmatrix(k,l) = 0.0
                                if (K.eq.L) self%symop(j)%rmatrix(k,l) = 1.0
                            endif
                        enddo
                    enddo
                enddo
                call check_point_symmetry(self)
                if (self%iredun.lt.0) then
                    do j = 1,nsym
                        if (self%symop(j)%jmax.lt.0) then
                            write(*,'(a,i0)') '**ERROR(point_symmetry_print): Repeated application of symmetry ',    &
                                            'matrix does not produce identity, isym = ', j
                            call terminate('Fatal error in POINT_SYMMETRY_PRINT')
                        endif
                    enddo
                endif
            else if (symbol1 .eq. 'H') then
                iredun = 1
                allocate(self%symop(1))
                self%symop(1)%rmatrix = identity_rotmat()
                call check_point_symmetry(self)
            else if (symbol1 .ne. 'H') then
                write(*,'(a,a)') '**ERROR(point_symmetry_init): Invalid symmetry, sym_symbol = ', sym_symbol%symbol
                call terminate('Fatal error in POINT_SYMMETRY_INIT')
            endif

            ! Is this symmetry operator also used with helical symmetry (frealix)
            self%helical = (sym_symbol%symbol .eq. 'H')
            if (present(helical_symmetry)) then
                self%helical = helical_symmetry
            endif
        end subroutine point_symmetry_init


        subroutine parse_sym_symbol(self,symbol1,nasym)
            use runtime_parameters, only : terminate
            implicit none
            ! Arguments
            type(point_symmetry_symbol),    intent(in)  ::  self                !<  Point group symmetry using Sch�nflies notation
            character(len=1),       intent(out) ::  symbol1             !<  First character of the symmetry symbol
            integer,            intent(out) ::  nasym               !<  Integer number contained in the symmetry symbol
            ! Private variables
            integer                 ::  j
            integer                 ::  ksymn1,ksymn2,ksymn3
            character(len=1)            ::  symbol2
            character(len=1)            ::  symbol3
            character(len=1),parameter,dimension(17)::  ASYMTEST =  &
                                                        [' ','C','D','T','O','I','0','1','2','3','4','5','6','7','8','9','H']
            logical                 ::  asymstore
            ! Start work
            ksymn1 = -1
            ksymn2 = -1
            ksymn3 = -1
            symbol1 = self%symbol(1:1)
            symbol2 = self%symbol(2:2)
            symbol3 = self%symbol(3:3)

            asymstore = .false.

            nasym = 0

            do j = 2,17
                if ((SYMBOL3.eq.ASYMTEST(j)).and.(J.le.6)) then
                    ! No letters allowed
                    write(*,'(a)') '**ERROR(parse_sym_symbol): 3rd character in symmetry symbol invalid'
                    call terminate('Fatal error in PARSE_SYM_SYMBOL')
                endif
                if ((SYMBOL3.eq.ASYMTEST(j)).and.(J.gt.6)) ksymn3 = J-7
                if ((SYMBOL2.eq.ASYMTEST(j)).and.(J.le.6)) then
                    ! No letters allowed
                    write(*,'(a)') '**ERROR(parse_sym_symbol): 2nd character in symmetry symbol invalid'
                    call terminate('Fatal error in PARSE_SYM_SYMBOL')
                            endif
                if ((SYMBOL2.eq.ASYMTEST(j)).and.(J.gt.6)) ksymn2 = J-7
                if (SYMBOL1.eq.ASYMTEST(j)) then
                    ksymn1 = J-1
                    if(J.le.6) asymstore = .true.
                    if(J.gt.6) ksymn1 = J-7
                endif
            enddo

            if (KSYMN1.eq.-1) then
                write(*,'(a)') '**ERROR(parse_sym_symbol): Symmetry symbol invalid'
                call terminate('Fatal error in PARSE_SYM_SYMBOL')
            endif

            if (asymstore ) then
                if(KSYMN3.ne.-1) then
                    NASYM=KSYMN3+10*KSYMN2
                else
                    if(KSYMN2.ne.-1) then
                        NASYM=KSYMN2
                    else
                        NASYM=0
                    endif
                endif
                if ((KSYMN1.eq.1).or.(KSYMN1.eq.2)) then
                    if (NASYM.eq.0) then
                        write(*,'(a)') '**ERROR(parse_sym_symbol): Point groups c and d must be followed by an integer'
                        call terminate('Fatal error in PARSE_SYM_SYMBOL')
                    endif
                endif
            else if (symbol1 .ne. 'H') then
                write(*,'(a)') '**ERROR(parse_sym_symbol): Input of symmetry matrices not yet implemented'
                call terminate('Fatal error in PARSE_SYM_SYMBOL')
            endif
        end subroutine parse_sym_symbol


        subroutine check_point_symmetry(self)
            implicit none
            ! Arguments
            type(point_symmetry),       intent(inout)   ::  self        !<  Point group symmetry matrices
            ! Private variables
            integer                     ::  i,j,k,n,nsym
            integer                     ::  symcount,ir
            real                        ::  diff
            real,parameter                  ::  delta = 0.01
            real                        ::  tm(4,4)
            ! Start work
            nsym = size(self%symop)
            ir = 1
            do n = 1,NSYM
                self%symop(n)%jmax = 0
            enddo
            do n = 1,NSYM
                do i = 1,4
                    do j = 1,4
                        tm(i,j) = 0.0
                    enddo
                enddo
                tm(1,1) = 1.0
                tm(2,2) = 1.0
                tm(3,3) = 1.0
                tm(4,4) = 1.0

                symcount = 0
                do k = 1,100
                    symcount = SYMCOUNT+1
                    tm = matmul(tm,self%symop(n)%rmatrix)
                    diff = 0.0
                    do i = 1,3
                        do j = 1,3
                            if (I.eq.J) then
                                diff = DIFF+ABS(tm(i,j)-1.0)
                            else
                                diff = DIFF+ABS(tm(i,j))
                            endif
                        enddo
                    enddo
                    if (DIFF.lt.DELTA) exit
                enddo
                if (SYMCOUNT.eq.100) then
                    ir = -1
                    self%symop(n)%jmax = -1
                    exit
                else
                    ir = IR*SYMCOUNT
                    self%symop(n)%jmax = symcount
                endif
            enddo
            self%iredun = ir
        end subroutine check_point_symmetry


end module point_symmetries
