!>  \brief  Module to interface with the proc filesystem, if available
module procfs
    character(len=*),   parameter   ::  procfs_path = '/proc'

contains

    !>  \brief  Printout memory usage statistics
    subroutine print_memory_usage_stats()
        use DatesAndTimes
        implicit none
        ! Arguments

        ! Private variables
        integer ::  peak_vm
        type(DateAndTime) :: myTime
        integer ::  vm_size, peak_vm_size
        integer ::  my_pid
        ! Start work
        write(*,'(2a)')         '**info(print_memory_usage_stats): date & time = ', myTime%DateAndTimeAsString()
        vm_size = get_vm_size()
        peak_vm_size = get_peak_vm_size()
        write(*,'(a,i12,a)')    '**info(print_memory_usage_stats): current virtual memory size = ', vm_size,         ' kB'
        peak_vm = get_peak_vm_size()
        if (peak_vm .ne. 0) then
        write(*,'(a,i12,a)')    '**info(print_memory_usage_stats):    peak virtual memory size = ', peak_vm_size,    ' kB'
        endif
    end subroutine print_memory_usage_stats

    !>  \brief  Get current virtual memory size, in kB
    integer function get_vm_size(pid )
        implicit none
        ! Arguments
        integer,    optional,   intent(in)  ::  pid     !<  Process id
        ! Private variables

        ! Start work
        get_vm_size = procfs_status_get_value('VmSize',pid)
    end function get_vm_size

    !>  \brief  Get current virtual memory size, in kB
    integer function get_peak_vm_size(pid)
        implicit none
        ! Arguments
        integer,    optional,   intent(in)  ::  pid     !<  Process id
        ! Private variables

        ! Start work
        get_peak_vm_size = procfs_status_get_value('VmPeak',pid)
    end function get_peak_vm_size

    !>  \brief  Get current virtual memory size, in kB
    integer function get_swap_size(pid)
        implicit none
        ! Arguments
        integer,    optional,   intent(in)  ::  pid     !<  Process id
        ! Private variables

        ! Start work
        get_swap_size = procfs_status_get_value('VmSwap',pid)
    end function get_swap_size

    !>  \brief  Guess system swap usage by parsing procfs's meminfo file
    integer function procfs_meminfo_system_swap()
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        ! Private variables
        ! Start work
        procfs_meminfo_system_swap = 0
        call terminate('Not implemented yet')
        ! Relevant keywords in /proc/meminfo
        ! are SwapTotal and SwapFree

    end function procfs_meminfo_system_swap

    !>  \brief  Get value for given parameter in procfs status file
    integer function procfs_status_get_value(key,pid) result(value)
        use parsers, only : grep, split, is_blank
        use runtime_parameters, only : get_process_id, line_len, word_len
        implicit none
        ! Arguments
        character(len=*),               intent(in)  ::  key     !<  Keyword we're looking for in the status file
        integer,            optional,   intent(in)  ::  pid     !<  Process id
        ! Private variables
        character(len=LINE_LEN)                 ::  line
        character(len=WORD_LEN)                 ::  value_c
        character(len=LINE_LEN),   allocatable  ::  words(:)
        integer ::  ppid
        logical,    parameter   ::  debug   =   .false.
        ! Start work
        if (present(pid)) then
            ppid = pid
        else
            ppid = get_process_id()
        endif
        value = 0
        if (procfs_available()) then
            line = grep(procfs_status_file_path(ppid),key)
            if (is_blank(line)) then
                value = 0
            else
                words = split(line)
                value_c = trim(adjustl(words(2)))
                if (debug) write(*,'(2a)') '**debug(procfs_status_get_value): found value of ', trim(adjustl(value_c))
                read(value_c,'(i12)') value
            endif
        else
            value = 0
        endif
    end function procfs_status_get_value

    !>  \brief  Get the path of the status file for a process
    function procfs_status_file_path(pid) result(path)
        use runtime_parameters, only : get_process_id, line_len
        implicit none
        ! Arguments
        integer,    optional,   intent(in)  ::  pid     !<  Process id
        ! Result
        character(len=LINE_LEN)             ::  path
        ! Private variables
        character(len=6)    ::  pidc
        integer             ::  ppid
        ! Start work
        if (present(pid)) then
            ppid = pid
        else
            ppid = get_process_id()
        endif
        ! Convert pid to character string
        write(pidc,'(i0)') ppid

        path = procfs_path // '/' // trim(adjustl(pidc)) // '/status'
    end function procfs_status_file_path

    !>  \brief  Check whether the proc filesystem is available
    logical function procfs_available()
        use parsers, only : file_exists
        use runtime_parameters, only : line_len
        implicit none
        ! Arguments
        ! Private variables
        logical,    parameter   ::  debug   =   .false.
        character(len=LINE_LEN) ::  file_path
        ! Start work
        file_path = procfs_path // '/1/status'

        procfs_available = file_exists(file_path)
        if (debug) write(*,'(3a,l1)') '**debug(procfs_available): file, exists?: ', file_path, ', ', procfs_available
    end function procfs_available
end module procfs
