!>  \brief  Rotation matrices & related functions
module rotation_matrices

    type    ::  rotation_matrix
        integer ::  jmax            !<  Symmetry redundancy of operator (set to 0 if repeated application of rotation will not reproduce identity)
        real    ::  rmatrix(4,4)        !<  Rotation matrix (4th row/column not used - they are kept to make the matrix compatible with general rotation/translation operations)
    end type

    !rotmat_90deg(1,:) = (/ 1.0, 0.0, 0.0, 0.0/)
    !rotmat_90deg(2,:) = (/ 0.0, 0.0,-1.0, 0.0/)
    !rotmat_90deg(3,:) = (/ 0.0, 1.0, 0.0, 0.0/)
    !rotmat_90deg(4,:) = (/ 0.0, 0.0, 0.0, 1.0/)

    !rotmat_m90deg(1,:) = (/ 1.0, 0.0, 0.0, 0.0/)
    !rotmat_m90deg(2,:) = (/ 0.0, 0.0, 1.0, 0.0/)
    !rotmat_m90deg(3,:) = (/ 0.0,-1.0, 0.0, 0.0/)
    !rotmat_m90deg(4,:) = (/ 0.0, 0.0, 0.0, 1.0/)

    interface rotmat
        module procedure rotmat_2d
    end interface

    interface degrees
        module procedure degrees_single
        module procedure degrees_double
    end interface

    interface radians
        module procedure radians_single
        module procedure radians_double
    end interface

    contains

    !>  \brief  Print properties of rotation matrix object
    subroutine rotation_matrix_print(rotmat)
        implicit none
        type(rotation_matrix),  intent(in)  ::  rotmat
        call print_rotmat(rotmat%rmatrix)
        !write(*,'(a,i5/)') 'Symmetry redundancy for this operator: ',rotmat%jmax
    end subroutine rotation_matrix_print

    !>  \brief  Print a rotation matrix to terminal nicely
    subroutine print_rotmat(rotmat)
        implicit none
        real,   intent(in)  ::  rotmat(4,4)
        integer ::  i
        do I=1,4
            write(*,'(4(F6.2,2x))') rotmat(i,1), rotmat(i,2), rotmat(i,3), rotmat(i,4)
        enddo
    end subroutine print_rotmat

    !>  \brief  Return a 4x4 identity matrix
    pure function identity_rotmat() result(rotmat)
        implicit none
        ! Arguments
        ! Result
        real    ::  rotmat(4,4)
        ! Private variables
        integer ::  i
        ! Start work
        !rotmat(1,:)   = (/ 1.0, 0.0, 0.0, 0.0/)
        !rotmat(2,:)   = (/ 0.0, 1.0, 0.0, 0.0/)
        !rotmat(3,:)   = (/ 0.0, 0.0, 1.0, 0.0/)
        !rotmat(4,:)   = (/ 0.0, 0.0, 0.0, 1.0/)
        rotmat = 0.0
        do I=1,4
            rotmat(i,i) = 1.0
        enddo
    end function identity_rotmat


    !>  \brief  Return a 4x4 rotation matrix, given Euler angles following the Spider / Frealign convention,
    !!          as decribed in http://www.wadsworth.org/spider_doc/spider/docs/euler.html
    !!
    !!
    !!      By Niko, on the lab forum:
    !!      "Here are the Frealign conventions. The directions i use here assume that the particle image origin is in the lower left corner (as is the case in the program Ximdisp).
    !!      1) In-plane rotation (psi angle in Frealign): a positive value describes a counterclockwise rotation of the particle. To align it with the reference, the negative value of this angle needs to be applied, i.e. a clockwise rotation.
    !!      2) X,y shifts: Positive values describe shifts to the right and up. The negative values have to be applied to align the particle with a reference.
    !!      3) Frealign applies the shifts first and then the rotations. There are no mirrors used in Frealign."
    !!
    !!      The reverse flag can be used to obtain the reverse transformation described by the Euler angles. In other words,
    !!      multiplying the matrix returned when using REVERSE=.FALSE. by the matrix returned when using REVERSE=.TRUE. should
    !!      give the identity matrix.
    !!
    !!      Example uses:
    !!
    !!      1/ Apply a rotation to a 3d volume
    !!      r = euler_to_rotamt(phi,theta,psi)
    !!      coos_out = matmul(r,coos_in)
    !!
    !!      2/ Insert a 2d slice (ft of projection at Euler angles) into 3d volume (typical during reconstruction by Fourier inversion)
    !!      r = euler_to_rotmat(phi,theta,psi,REVERSE=.TRUE.)
    !!      coos_3d = matmul(r,coos_2d)
    !!
    !!  \todo   Use euler derived type as input rather than separate REALs
    pure function euler_to_rotmat(phi,theta,psi,shift,reverse) result(rotmat)
        implicit none
        ! Arguments
        real(kind=8),           intent(in)  ::  phi         !<  Euler angle phi (degrees)
        real(kind=8),           intent(in)  ::  theta       !<  Euler angle theta (degrees)
        real(kind=8),           intent(in)  ::  psi         !<  Euler angle psi (degrees)
        real,       optional,   intent(in)  ::  shift(3)    !<  Shift (applied after the rotation in forward direction)
        logical,    optional,   intent(in)  ::  reverse     !<  Whether the inverse matrix should be returned. See subroutine documentation for more detail.
        ! Result
        real(kind=8)    ::  rotmat(4,4)
        ! Private variables
        real    ::  pphi,ttheta,ppsi
        real    ::  rotmat_phi(4,4), rotmat_theta(4,4), rotmat_psi(4,4)
        logical ::  rreverse
        integer ::  i
        ! Start work

        ! Initialise resut matrix (to the identity matrix)
        rotmat = 0.0
        forall(I=1:4)
            rotmat(i,i) = 1.0
        end forall

        ! Check whether we'll be reversing (by default, we don't)
        if (present(reverse)) then
            rreverse = reverse
        else
            rreverse = .false.
        endif

        pphi = radians(phi)
        ttheta = radians(theta)
        ppsi = radians(psi)

        ! Reverse the angles  if this was asked.
        !
        ! Note the way frealign thinks about Euler angles:
        ! To it, you apply the Euler angles to the 3d, then shift the projection to obtain the original image,
        ! whereas to Imagic (and Spider i think), the Euler angles are applied after the image is shifted to
        ! back-project it into the 3D.
        if (rreverse) then
            pphi = -PPHI
            ttheta = -TTHETA
            ppsi = -PPSI
        endif

        ! Set the matrix elements.
        rotmat_psi(1,:)   = (/ cos(ppsi),     sin(ppsi), 0.0,         0.0/)
        rotmat_psi(2,:)   = (/-SIN(ppsi),     cos(ppsi), 0.0,         0.0/)
        rotmat_psi(3,:)   = (/ 0.0,           0.0,       1.0,         0.0/)
        rotmat_psi(4,:)   = (/ 0.0,           0.0,       0.0,         1.0/)

        rotmat_theta(1,:) = (/ cos(ttheta), 0.0,      -SIN(ttheta), 0.0/)
        rotmat_theta(2,:) = (/ 0.0,         1.0,       0.0,         0.0/)
        rotmat_theta(3,:) = (/ sin(ttheta), 0.0,       cos(ttheta), 0.0/)
        rotmat_theta(4,:) = (/ 0.0,         0.0,       0.0,         1.0/)

        rotmat_phi(1,:)   = (/ cos(pphi),     sin(pphi), 0.0,         0.0/)
        rotmat_phi(2,:)   = (/-SIN(pphi),     cos(pphi), 0.0,         0.0/)
        rotmat_phi(3,:)   = (/ 0.0,           0.0,       1.0,         0.0/)
        rotmat_phi(4,:)   = (/ 0.0,           0.0,       0.0,         1.0/)

        if (rreverse) then
            if (present(shift)) then
                rotmat_psi(1:3,4) = -SHIFT(1:3)
            endif
            rotmat = matmul(rotmat_theta,rotmat_psi)
            rotmat = matmul(rotmat_phi,rotmat)
        else
            rotmat = matmul(rotmat_theta,rotmat_phi)
            rotmat = matmul(rotmat_psi,rotmat)
            if (present(shift)) then
                rotmat(1:3,4) = shift(1:3)
            endif
        endif
    end function euler_to_rotmat


    !>  \brief  Return a 4x4 rotation matrix, given rotation in degrees and shifts in pixels. The rotation is applied first, the
    !!          shifts afterwards, unless shifts_first is .TRUE.
    !!  \todo   This may not be compliant with Frealign/frealix convention - check
    pure function rotmat_2d(rot,shx,shy,shifts_first)
        implicit none
        ! Function result
        real(kind=8)    ::  rotmat_2d(4,4)
        ! Arguments
        real,                   intent(in)  ::  rot             !<  Rotation angle (around z, since we're in 2d) (degrees)
        real,                   intent(in)  ::  shx             !<  Shift in x direction (pixels)
        real,                   intent(in)  ::  shy             !<  Shift in y direction (pixels)
        logical,    optional,   intent(in)  ::  shifts_first    !<  .FALSE. by default. If .TRUE., the matrix applies shifts first, then
        ! Private variables
        real    ::  rrot
        logical ::  sshifts_first
        real    ::  rotmat_rot(4,4), rotmat_shifts(4,4)
        ! Start work
        ! Check whether shifts are first
        if (present(shifts_first)) then
            sshifts_first = shifts_first
        else
            sshifts_first = .false.
        endif
        ! Convert rotation to radians
        rrot = radians(rot)

        ! Set the matrix elements for the rotation
        rotmat_rot(1,:) = (/ cos(rrot),-SIN(rrot), 0.0, 0.0/)
        rotmat_rot(2,:) = (/ sin(rrot), cos(rrot), 0.0, 0.0/)
        rotmat_rot(3,:) = (/ 0.0,       0.0,       1.0, 0.0/)
        rotmat_rot(4,:) = (/ 0.0,       0.0,       0.0, 1.0/)
        ! The translation matrix
        rotmat_shifts(1,:) = (/ 1.0, 0.0, 0.0, shx/)
        rotmat_shifts(2,:) = (/ 0.0, 1.0, 0.0, shy/)
        rotmat_shifts(3,:) = (/ 0.0, 0.0, 1.0, 0.0/)
        rotmat_shifts(4,:) = (/ 0.0, 0.0, 0.0, 1.0/)
        if (sshifts_first) then
            rotmat_2d = matmul(rotmat_shifts,rotmat_rot)
        else
            rotmat_2d = matmul(rotmat_rot,rotmat_shifts)
        endif
    end function rotmat_2d


    !>  Convert degrees to radians
    elemental pure real function radians_single(degrees)
        use runtime_parameters, only: pi
        implicit none
        real, intent(in) :: degrees
        radians_single = degrees * pi / 180.
    end function radians_single

    !>  Convert radians to degrees
    elemental pure real function degrees_single(radians)
        use runtime_parameters, only: pi
        implicit none
        real, intent(in) :: radians
        degrees_single = radians / pi * 180.
    end function degrees_single

    !>  Convert degrees to radians (double precision)
    pure elemental function radians_double(degrees)
        use runtime_parameters, only: dpi
        implicit none
        real(kind=8), intent(in) :: degrees
        real(kind=8)    ::  radians_double
        radians_double = degrees * dpi / 180.
    end function radians_double

    !>  Convert radians to degrees (double precision)
    pure elemental function degrees_double(radians)
        use runtime_parameters, only: dpi
        implicit none
        real(kind=8), intent(in) :: radians
        real(kind=8)    ::  degrees_double
        degrees_double = radians / dpi * 180.
    end function degrees_double


end module rotation_matrices
