!>  \brief  Class & methods defining run-time parameters and their behaviour
module rtps


    use runtime_parameters
    implicit none
    private
    public :: rtp_init, rtp_get_value, rtp_print_info_header

    integer,            parameter   ::  colwidth(5)         =   [36,45,12,64,15]
    character(len=*),   parameter   ::  colwidth_c(5)       =   ['36','45','12','64','15']
    integer,            parameter   ::  com_colwidth(3)     =   [32,filename_len,line_len]
    character(len=4),   parameter   ::  com_colwidth_c(3)   =   ['32  ',filename_len_c,line_len_c]

    type,   abstract,   public  ::  rtp
        character(len=64)       ::  keyword                 !<  Keyword used in frealix command files to identify the parameter
        character(len=line_len) ::  description             !<  Description used in the question and as comment in frealix command files
        logical                 ::  set_by_user =   .false. !<  Flag to indicate whether the rtp has been set by the user
        integer                 ::  order       =   0       !<  Runtime parameters used within modes are indexed
        contains
            procedure   ::  get_value_from_user =>  rtp_get_value
    end type

    type, extends(rtp), public  ::  rtp_real
        !real    ::  default_value   !<  Default value for the parameter
        real            ::  value           !<  Current value for the parameter
        contains
            procedure   ::  init                =>  rtp_init_real
            procedure   ::  print_info          =>  rtp_real_print_info
            procedure   ::  print_to_command_file   =>  rtp_real_print_to_command_file
    end type

    type, extends(rtp), public  ::  rtp_intg
        !integer ::  default_value   !<  Default value for the parameter
        integer ::  value           !<  Current value for the parameter
        contains
            procedure   ::  init                =>  rtp_init_intg
            procedure   ::  print_info          =>  rtp_intg_print_info
            procedure   ::  print_to_command_file   =>  rtp_intg_print_to_command_file
    end type

    type, extends(rtp), public  ::  rtp_logi
        !logical ::  default_value   !<  Default value for the parameter
        logical ::  value           !<  Current value for the parameter
        contains
            procedure   ::  init                =>  rtp_init_logi
            procedure   ::  print_info          =>  rtp_logi_print_info
            procedure   ::  print_to_command_file   =>  rtp_logi_print_to_command_file
    end type

    type, extends(rtp), public  ::  rtp_char
        !character(LEN=LINE_LEN)    ::  default_value   !<  Default value for the parameter
        character(len=LINE_LEN)    ::  value           !<  Current value for the parameter
        contains
            procedure   ::  init                =>  rtp_init_char
            procedure   ::  print_info          =>  rtp_char_print_info
            procedure   ::  print_to_command_file   =>  rtp_char_print_to_command_file
    end type

    interface   rtp_assign
        module procedure    rtp_assign_real
        module procedure    rtp_assign_char
        module procedure    rtp_assign_intg
        module procedure    rtp_assign_logi
    end interface

    interface rtp_init
        module procedure    rtp_init_real
        module procedure    rtp_init_intg
        module procedure    rtp_init_char
        module procedure    rtp_init_logi
    end interface rtp_init

    contains

    !>  \brief  Print rtp information in format suitable for command file
    subroutine rtp_real_print_to_command_file(self,lun)
        implicit none
        ! Arguments
        class(rtp_real),    intent(in)  ::  self
        integer,            intent(in)  ::  lun
        ! Private variables

        ! Start work
        write(lun,'(A'//com_colwidth_c(1)//',F0.6,10x,a,1x,a)')   adjustl(self%keyword), self%value,    &
                                                                '! ', trim(adjustl(self%description))
    end subroutine rtp_real_print_to_command_file

    !>  \brief  Print rtp information in format suitable for command file
    subroutine rtp_intg_print_to_command_file(self,lun)
        implicit none
        ! Arguments
        class(rtp_intg),    intent(in)  ::  self
        integer,            intent(in)  ::  lun
        ! Private variables

        ! Start work
        write(lun,'(A'//com_colwidth_c(1)//',i0,10x,a,1x,a)') adjustl(self%keyword), self%value, '! ', &
                                                            trim(adjustl(self%description))
    end subroutine rtp_intg_print_to_command_file

    !>  \brief  Print rtp information in format suitable for command file
    subroutine rtp_char_print_to_command_file(self,lun)
        implicit none
        ! Arguments
        class(rtp_char),    intent(in)  ::  self
        integer,            intent(in)  ::  lun
        ! Private variables

        ! Start work
        write(lun,'(A'//com_colwidth_c(1)//',a,10x,a,1x,a)') adjustl(self%keyword), trim(adjustl(self%value)), '! ', &
                                                            trim(adjustl(self%description))
    end subroutine rtp_char_print_to_command_file

    !>  \brief  Print rtp information in format suitable for command file
    subroutine rtp_logi_print_to_command_file(self,lun)
        implicit none
        ! Arguments
        class(rtp_logi),    intent(in)  ::  self
        integer,            intent(in)  ::  lun
        ! Private variables

        ! Start work
        write(lun,'(A'//com_colwidth_c(1)//',l1,10x,a,1x,a)') adjustl(self%keyword), self%value, '! ',   &
                                                            trim(adjustl(self%description))
    end subroutine rtp_logi_print_to_command_file

    !>  \brief  Print information about a real rtp
    subroutine rtp_real_print_info(self)
        implicit none
        ! Arguments
        class(rtp_real), intent(in)  ::  self
        ! Private variables

        ! Start work
        call rtp_print_info_slave_1(self)
        write(*,'(F'//colwidth_c(2)//'.4,1x)',advance='NO') self%value!, self%default_value
        call rtp_print_info_slave_2(self)
        call rtp_print_info_slave_3(self)
        write(*,'(a)') ' '
    end subroutine rtp_real_print_info

    !>  \brief  Print information about a real rtp
    subroutine rtp_intg_print_info(self)
        implicit none
        ! Arguments
        class(rtp_intg), intent(in)  ::  self
        ! Private variables

        ! Start work
        call rtp_print_info_slave_1(self)
        write(*,'(I'//colwidth_c(2)//',1x)',advance='NO') self%value!, self%default_value
        call rtp_print_info_slave_2(self)
        call rtp_print_info_slave_3(self)
        write(*,'(a)') ' '
    end subroutine rtp_intg_print_info

    !>  \brief  Print information about a real rtp
    subroutine rtp_logi_print_info(self)
        implicit none
        ! Arguments
        class(rtp_logi), intent(in)  ::  self
        ! Private variables

        ! Start work
        call rtp_print_info_slave_1(self)
        write(*,'(L'//colwidth_c(2)//',1x)',advance='NO') self%value!, self%default_value
        call rtp_print_info_slave_2(self)
        call rtp_print_info_slave_3(self)
        write(*,'(a)') ' '
    end subroutine rtp_logi_print_info

    !>  \brief  Print information about a real rtp
    subroutine rtp_char_print_info(self)
        implicit none
        ! Arguments
        class(rtp_char), intent(in)  ::  self
        ! Private variables

        ! Start work
        call rtp_print_info_slave_1(self)
        write(*,'(A'//colwidth_c(2)//',1x)',advance='NO') trim(self%value)
        call rtp_print_info_slave_2(self)
        call rtp_print_info_slave_3(self)
        write(*,'(a)') ' '
    end subroutine rtp_char_print_info

    !>  \brief  Print basic information (not the value) of an rtp
    subroutine rtp_print_info_slave_1(self)
        implicit none
        ! Arguments
        class(rtp),     intent(in)  ::  self
        ! Private variables

        ! Start work

        write(*,'(A'//colwidth_c(1)//',1x)',advance='NO') trim(adjustl(self%keyword))
    end subroutine rtp_print_info_slave_1

    !>  \brief  Print basic information (not the value) of an rtp
    subroutine rtp_print_info_slave_2(self)
        implicit none
        ! Arguments
        class(rtp),     intent(in)  ::  self
        ! Private variables

        ! Start work

        write(*,'(L'//colwidth_c(3)//',1x)',advance='NO') self%set_by_user
    end subroutine rtp_print_info_slave_2

    !>  \brief  Print the description of an rtp, over multiple lines if necessary
    subroutine rtp_print_info_slave_3(self)
        implicit none
        ! Arguments
        class(rtp),     intent(in)  ::  self
        ! Private variables
        integer ::  descr_length, descr_lead_blanks
        integer :: num_lines, current_line, num_spaces
        character(len=5)    ::  num_spaces_here
        integer ::  begin,end
        character(len=:), allocatable :: tmp_str
        logical ::  no_space
        ! Start work

!        write(*,'(A<COLWIDTH(4)>,x)',ADVANCE='NO') adjustl(self%description)

        descr_length = len_trim(adjustl(self%description))
        descr_lead_blanks = len(self%description) - len_trim(self%description)
        num_lines = ceiling(real(descr_length)/real(colwidth(4)))
        num_spaces = colwidth(1) + 1 + colwidth(2) + 1 + colwidth(3) + 1

        allocate(character(LEN=COLWIDTH(4)) :: TMP_STR)


        do CURRENT_LINE=1,num_lines
            if (current_line .gt. 1) then
                ! We need a new line
                write(*,'(a)') ' '
            endif
            !
            begin = (CURRENT_LINE-1)* colwidth(4) + 1
            end   =  current_line   * colwidth(4)
            !
            !begin = begin + descr_lead_blanks
            !end   = end   + descr_lead_blanks
            !
            end   = min(end,DESCR_LEAD_BLANKS+DESCR_LENGTH)
            tmp_str = self%description(begin:end)
            no_space = .false.
            if (current_line .gt. 1) then
                write(num_spaces_here,'(i5)') num_spaces
            else
                num_spaces_here = '0'
                no_space = .true.
            endif
            if (no_space) then
                write(*,                       '(A'//colwidth_c(4)//',1x)',advance='NO') adjustl(tmp_str)
            else
                write(*,'('//num_spaces_here//'X,A'//colwidth_c(4)//',1x)',advance='NO') adjustl(tmp_str)
            endif
        enddo
    end subroutine rtp_print_info_slave_3

    !>  \brief  Print out header for rtp descriptions
    subroutine rtp_print_info_header()
        implicit none
        ! Arguments

        ! Private variables
        character(len=COLWIDTH(4))  ::  DESCRIPTION
        character(len=COLWIDTH(5))  ::  HELP
        ! Start work
        description = 'Description'
        help = 'Help'
        write(*,'(A'//colwidth_c(1)//',1x,A'//colwidth_c(2)//',1x,A'//colwidth_c(3)//',1x,A'//colwidth_c(4)//')') &
                                                                                'Key', 'Value', 'Set by user?', adjustl(description)
    end subroutine rtp_print_info_header

    !>  \brief  Initialise a runtime parameter
    subroutine rtp_init_real(self,default_value,keyword,description,help)
        implicit none
        ! Arguments
        class(rtp_real),                intent(inout)   ::  self
        real,                           intent(in)      ::  default_value   !<  Default value for the parameter
        character(len=*),               intent(in)      ::  keyword         !<  Keyword used in frealix command files to identify the parameter
        character(len=*),   optional,   intent(in)      ::  description     !<  Description used in the question and as comment in frealix command files
        character(len=*),   optional,   intent(in)      ::  help            !<  Help to be given to user to describe what the parameter means / does
        ! Start work
        call rtp_init_slave(self,keyword,DESCRIPTION=DESCRIPTION,HELP=HELP)
        !self%default_value  =   default_value
        self%value          =   default_value
        self%set_by_user    =   .false.
    end subroutine rtp_init_real

    !>  \brief  Initialise a runtime parameter
    subroutine rtp_init_intg(self,default_value,keyword,description,help)
        implicit none
        ! Arguments
        class(rtp_intg),                intent(inout)   ::  self
        integer,                        intent(in)      ::  default_value   !<  Default value for the parameter
        character(len=*),               intent(in)      ::  keyword         !<  Keyword used in frealix command files to identify the parameter
        character(len=*),   optional,   intent(in)      ::  description     !<  Description used in the question and as comment in frealix command files
        character(len=*),   optional,   intent(in)      ::  help            !<  Help to be given to user to describe what the parameter means / does
        ! Start work
        call rtp_init_slave(self,keyword,DESCRIPTION=DESCRIPTION,HELP=HELP)
        !self%default_value  =   default_value
        self%value          =   default_value
        self%set_by_user    =   .false.
    end subroutine rtp_init_intg

    !>  \brief  Initialise a runtime parameter
    subroutine rtp_init_logi(self,default_value,keyword,description,help)
        implicit none
        ! Arguments
        class(rtp_logi),                intent(inout)   ::  self
        logical,                        intent(in)      ::  default_value   !<  Default value for the parameter
        character(len=*),               intent(in)      ::  keyword         !<  Keyword used in frealix command files to identify the parameter
        character(len=*),   optional,   intent(in)      ::  description     !<  Description used in the question and as comment in frealix command files
        character(len=*),   optional,   intent(in)      ::  help            !<  Help to be given to user to describe what the parameter means / does
        ! Start work
        call rtp_init_slave(self,keyword,DESCRIPTION=DESCRIPTION,HELP=HELP)
        !self%default_value  =   default_value
        self%value          =   default_value
        self%set_by_user    =   .false.
    end subroutine rtp_init_logi

    !>  \brief  Initialise a runtime parameter
    subroutine rtp_init_char(self,default_value,keyword,description,help)
        implicit none
        ! Arguments
        class(rtp_char),                intent(inout)   ::  self
        character(len=*),               intent(in)      ::  default_value   !<  Default value for the parameter
        character(len=*),               intent(in)      ::  keyword         !<  Keyword used in frealix command files to identify the parameter
        character(len=*),   optional,   intent(in)      ::  description     !<  Description used in the question and as comment in frealix command files
        character(len=*),   optional,   intent(in)      ::  help            !<  Help to be given to user to describe what the parameter means / does
        ! Start work
        call rtp_init_slave(self,keyword,DESCRIPTION=DESCRIPTION,HELP=HELP)
        !self%default_value  =   default_value
        self%value          =   default_value
        self%set_by_user    =   .false.
    end subroutine rtp_init_char



    !>  \brief  Initialise a runtime parameter
    subroutine rtp_init_slave(self,keyword,description,help)
        implicit none
        ! Arguments
        class(rtp),                     intent(inout)   ::  self
        character(len=*),               intent(in)      ::  keyword     !<  Keyword used in frealix command files to identify the parameter
        character(len=*),   optional,   intent(in)      ::  description !<  Description used in the question and as comment in frealix command files
        character(len=*),   optional,   intent(in)      ::  help        !<  Help to be given to user to describe what the parameter means / does
        ! Private variables

        ! Start work
        self%keyword    =   keyword
        if (present(description)) then
            self%description = description
        endif

    end subroutine rtp_init_slave

    !>  \brief  The RTP's value pointer is pointed at the relevant variable
    subroutine rtp_assign_real(self,var)
        implicit none
        ! Arguments
        class(rtp),         intent(inout)   ::  self
        real,               intent(in)      ::  var
        ! Private variables
        ! Start work
        select type(self)
            type is (rtp_real)
                self%value = var
            class default
                write(*,'(a)') '**ERROR(rtp_assign_real): rtp / variable type mismatch'
                call terminate('Fatal error in RTP_ASSIGN_REAL')
        end select
    end subroutine rtp_assign_real

    !>  \brief  The RTP's value pointer is pointed at the relevant variable
    subroutine rtp_assign_intg(self,var)
        implicit none
        ! Arguments
        class(rtp),             intent(inout)   ::  self
        integer,        target, intent(in)      ::  var
        ! Private variables
        ! Start work
        select type(self)
            type is (rtp_intg)
                self%value = var
            class default
                write(*,'(a)') '**ERROR(rtp_assign_intg): rtp / variable type mismatch'
                call terminate('Fatal error in RTP_ASSIGN_INTG')
        end select
    end subroutine rtp_assign_intg

    !>  \brief  The RTP's value pointer is pointed at the relevant variable
    subroutine rtp_assign_logi(self,var)
        implicit none
        ! Arguments
        class(rtp),             intent(inout)   ::  self
        logical,        target, intent(in)      ::  var
        ! Private variables
        ! Start work
        select type(self)
            type is (rtp_logi)
                self%value = var
            class default
                write(*,'(a)') '**ERROR(rtp_assign_logi): rtp / variable type mismatch'
                call terminate('Fatal error in RTP_ASSIGN_LOGI')
        end select
    end subroutine rtp_assign_logi

    !>  \brief  The RTP's value pointer is pointed at the relevant variable
    subroutine rtp_assign_char(self,var)
        implicit none
        ! Arguments
        class(rtp),                     intent(inout)   ::  self
        character(len=*),       target, intent(in)      ::  var
        ! Private variables
        ! Start work
        select type(self)
            type is (rtp_char)
                self%value = var
            class default
                write(*,'(a)') '**ERROR(rtp_assign_char): rtp / variable type mismatch'
                call terminate('Fatal error in RTP_ASSIGN_CHAR')
        end select
    end subroutine rtp_assign_char



    !>  \brief  Fetch the value for a runtime parameter, either via user interface or by parsing command file
    subroutine rtp_get_value(self)
        use ui
        use parsers
        implicit none
        ! Arguments
        class(rtp),                 intent(inout)   ::  self
        ! Private variables
        character(len=LINE_LEN) ::  question
        ! Start work
        question = 'Please give ' // trim(self%keyword)
        if (interactive) then
            ! Eventually, the ui routines should be rewritten to take rtp objects as arguments, so that the select type will be done at a lower level
            select type(self)
                type is (rtp_char)
                    call ui_get_char(question,self%value)
                type is (rtp_logi)
                    call ui_get_logi(question,self%value)
                type is (rtp_real)
                    call ui_get_real(question,self%value)
                type is (rtp_intg)
                    call ui_get_intg(question,self%value)
            end select
            self%set_by_user = .true.
        else
            select type(self)
                type is (rtp_char)
                    call file_get_char(control_filename,self%keyword,self%value,FOUND=SELF%set_by_user)
                type is (rtp_logi)
                    call file_get_logi(control_filename,self%keyword,self%value,FOUND=SELF%set_by_user)
                type is (rtp_real)
                    call file_get_real(control_filename,self%keyword,self%value,FOUND=SELF%set_by_user)
                type is (rtp_intg)
                    call file_get_intg(control_filename,self%keyword,self%value,FOUND=SELF%set_by_user)
            end select
        endif
    end subroutine rtp_get_value

    !>  \brief  Given an array of RTPs, and a string of characters, find the rtp value in the string if possible
    subroutine  rtp_from_string(rtps,string)
        use parsers
        implicit none
        ! Arguments
        class(rtp),         allocatable,    intent(inout)   ::  rtps(:) !<  Array of runtime parameters to check the line against
        character(len=WORD_LEN),            intent(in)      ::  string  !<  a line (string) of characters
        ! Private variables
        character(len=WORD_LEN),    allocatable ::  words(:)
        character(len=WORD_LEN),    allocatable ::  label, value
        integer ::  i
        ! Start work

        !
        ! Analyse the line. The first word is the label, the second is the value
        !
        words = split(string)
        label = words(1)
        value = words(2)

        !
        ! Loop over the array of rtps, to find the one, if any, which is designated by the string
        !
        do I=1,size(rtps)
            if (string_equal(label,rtps(i)%keyword)) then
                call rtp_value_from_string(rtps(i),value)
                rtps(i)%set_by_user = .true.
            endif
        enddo
    end subroutine rtp_from_string

    !>  \brief  Set RTP's value to that described by the supplied string
    subroutine rtp_value_from_string(self,string)
        implicit none
        ! Arguments
        class(rtp),                         intent(inout)   ::  self    !<  Runtime parameter for which the value should be set
        character(len=WORD_LEN),            intent(in)      ::  string  !<  Character string with value for the rtp
        ! Private variables

        ! Start work
        select type(self)
            type is (rtp_char)
                read(string,*) self%value
            type is (rtp_logi)
                read(string,*) self%value
            type is (rtp_real)
                read(string,*) self%value
            type is (rtp_intg)
                read(string,*) self%value
        end select
    end subroutine rtp_value_from_string
end module rtps
