!>    \brief Collection of all parameters that affect runtime behaviour
!!
!!    This module contains declarations for all variables which control run-time behaviour,
!!    most of which can be amended at run time by the user by using a control file.
!!    Default values for all variables are specified such that the user does not need
!!    to supply their own values.
!!
module runtime_parameters
    use DatesAndTimes
    implicit none

    ! The user cannot change the value of these variables
    integer,            parameter               ::  filename_len            =   200             !< Maximum number of characters in a file name
    character(len=4),   parameter               ::  filename_len_c          =   '200 '
    integer,            parameter               ::  rtp_len                 =   64              !< Maximum number of characters in a run-time parameter (rtp) name
    integer,            parameter               ::  line_len                =   4096            !< Maximum number of characters in a line of a file being read/written
    character(len=4),   parameter               ::  line_len_c              =   '4096'
    integer,            parameter               ::  word_len                =   200             !< Maximum number of characters in a word (from a flat-text file)
    character(len=4),   parameter               ::  word_len_c              =   '200 '

    ! The control file name will be set by the first command-line argument
    character(len=FILENAME_LEN)                 ::  control_filename

    ! Frealix mode. See documentation for allowed values.
    character(len=FILENAME_LEN)                 ::  frealix_mode

    ! Default file format
    character(len=1)                            ::  default_file_format     =   'I'             !< The default file format:
                                                                                                !! 'I', 'M' or 'S' for imagic, mrc, spider
    ! Version string
    character(len=15),  parameter               ::  version                 =   '1.2.0'         !< Version number   (0.0.0 indicates development version)

    ! svn revision info
#if defined(SVN_REV)
    character(len=*),   parameter               ::  svn_revision            =   SVN_REV
#else
    character(len=*),   parameter               ::  svn_revision            =   'NoSVNinfo'
#endif

    ! Compile date
#ifdef NAGFOR_COMPILER
    character(len=*),   parameter               ::  compile_date            =    'Unknown'
#else
    character(len=*),   parameter               ::  compile_date            =    __DATE__       !<    The preprocessor subsitutes the date here at compile time
#endif
    ! Whether this is an interactive session
    logical                                     ::  interactive             =   .false.         !< Whether we are running an interactive frealix session

    ! fftw parameters
    logical                                     ::  use_fftw_wisdom         =   .false.         !< Whether fftw wisdom should be used
    logical                                     ::  use_fftw_openmp         =   .false.         !< Whether fftw OpenMP threading should be used

    ! Constants
    real,               parameter               ::  pi                      =   acos(-1.0)
    real(kind=8),       parameter               ::  dpi                     =   acos(-1.0D0)
    !real,               parameter               ::  protein_density         =   0.86            !<    Protein density in Dalton / cubic Angstrom. Converted from 1.43 g/cm^3, see
    !                                                                                            !!  Quillin, M. L. and Matthews, B. W., "Accurate calculation of the density of proteins.", Acta Crystallogr d Biol Crystallogr, vol. 56, pp. 791--794, 2000
    real                                        ::  psf_radius              =   200             !<    In Angstroms. The approximate radius of point spread function.
    real                                        ::  psf_radius_pixels       =   0.0             !<    To be set at runtime, when the pixel size is known
    real,               parameter               ::  default_falloff_pixels  =   10.0            !< Default falloff, in pixels

    real                                        ::  pbc                     =   0.0             !<  "Fudge" parameter which relates an alignment score to a b factor to be applied during 3d reconstruction


    real,               parameter               ::  h_asym_unit_center(3)   = (/ 1.0, 0.0, 0.0 /)   !< The unit vector that points from the center of the
                                                                                                    !! volume to the origin of the helical unit cell


    ! Start and finish time
    integer                                     ::  date_start(8)
    integer                                     ::  date_finish(8)

    ! File handles, a.k.a. LUNs (logical unit number)
    integer,            parameter               ::  tmp_file_handle         =   99
    integer,            parameter               ::  max_lun                 =   400             !<    This is imposed by the frealign/frealix io routines (which could do with a rewrite!)
    integer,            parameter               ::  min_lun                 =   201              !<    This is to stay out of the way of default LUNs, which for ifort are 0,5,6. I'm not sure what the correct behaviour is, but I'm pretty sure 10 is a safe number.
    logical                                     ::  luns(min_lun:max_lun)   =   .true.          !<    For each lun at our disposal, .TRUE. if it is available, .FALSE. if it is already used.

    !
    ! The direction of the z axis.
    !
    ! There are 2 possibilities:
    !
    ! A. z_axis_direction is -1.0. Increasing z reduces underfocus. a cryo-EM micrograph collected at 2.3 um underfocus has a z height of -23000pxl if the pixel size is 1 A/pxl. As z increases, features become closer to focus.
    ! or
    ! B. z_axis_direction is 1.0. Increasing z increases underfocus.
    !
    ! Since underfocus corresponds to the case where the objective focal plain is further away from the electron source than the specimen is, a and b can be described as:
    !
    ! A. The z axis goes from the electron source to the detector. Z=0 is the focal plain.
    ! or
    ! B. The z axis goes from the detector to the electron source. Z=0 is the focal plain.
    !
    !
    real,               parameter            ::    z_axis_direction         =   -1.0

    ! Microtubule-specific parameters
    real                                    ::    mt_phi_first_pf           =   0.0             !<    For microtubules only: the latitude (phi angle; degrees) at which protofilament 1 is to be found in the geometry of the 3d model/reconstruction.
                                                                                                !!Protofilaments are labelled 1 to MT_PF_NUMBER. For a given "washer", the tubulin of pf 1 are nearer the - end of the filament.
    ! Endenianness of the system (not checked)
    logical,            parameter            ::     big_endian              =   ichar(transfer(1,'a')) == 0 !<  Test whether the system is big-endian

    interface safe_for_sqrt
        module procedure safe_for_sqrt_real
        module procedure safe_for_sqrt_dble
    end interface

    contains

!        !>    Convert degrees to radians
!        elemental pure real function radians(degrees)
!            implicit none
!            real, intent(in) :: degrees
!            radians = degrees * pi / 180.
!        end function radians
!
!        !>    Convert radians to degrees
!        elemental pure real function degrees(radians)
!            implicit none
!            real, intent(in) :: radians
!            degrees = radians / pi * 180
!        end function degrees


    !>    \brief    Return the next available lun
    integer function available_lun()
        implicit none
        ! Private variables
        integer    ::    i
        logical    ::    success
        !$omp    critical        (lun_omp_crit)
        ! Start work
        success = .false.
        ! Loop through the luns array until an available unit is found
        do I=MIN_LUN,max_lun
            if (luns(i)) then
                ! lun i is available
                success = .true.
                available_lun = i
                exit
            endif
        enddo
        if (success) then
            ! Now we have found an available lun, make it unavailable for the next call before we return
            luns(available_lun) = .false.
        else
            write(*,'(a)') '**ERROR(available_lun): was unable to find an available LUN'
            call terminate('Fatal error in AVAILABLE_LUN')
        endif
        !$omp    end critical    (lun_omp_crit)
    end function available_lun

    !>    \brief    Mark a given lun as available again
    subroutine lun_is_now_available(lun)
        implicit none
        ! Arguments
        integer,    intent(in)    ::    lun    !<    Number of the lun to "free up"
        ! Start work
        if (lun .lt. min_lun .or. lun .gt. max_lun) then
            write(*,'(a,i4)') '**WARNING: attempt to free up a lun which is not in range: ', lun
        else
            !$omp    critical        (lun_omp_crit)
            luns(lun) = .true.
            !$omp    end critical    (lun_omp_crit)
        endif
    end subroutine lun_is_now_available

    !>    \brief    Crash out with error message and useful information, like time/date and traceback
    subroutine terminate(string,dont_stop)
        use DatesAndTimes
#if defined(__INTEL_COMPILER)
        use ifcore
        use ifport
#elif defined(NAGFOR_COMPILER)
        use f90_unix, only : flush
#endif
        implicit none
        ! Arguments
        character(len=*),   optional,   intent(in)    ::    string      !<  Message to be printed out
        logical,            optional,   intent(in)    ::    dont_stop   !<  Whether an abort signal should be raised so that a core dump occurs
        ! Private variables
        logical    ::    ddont_stop
        type(DateAndTime)   ::  myTime
#if defined(NAGFOR_COMPILER)
        real                ::  dummy
#endif
        ! Start work
        ddont_stop = .false.
        if (present(dont_stop)) ddont_stop = dont_stop

        date_finish = myTime%DateAndTimeAsIntegers()

#if defined(__GFORTRAN__)
        if (.not. ddont_stop) then
            if (present(string)) then
                write(*,'(/a/)') string
            else
                write(*,'(/2a/)')   'Error termination of frealix, traceback below ', &
                '(if you do not see a stack trace below, check the environment variable for_disable_stack_trace).'
            endif
            write(*,'(a)') '</frealix>'
            call backtrace()
            stop
        endif
#elif defined(__INTEL_COMPILER)
        if (present(string)) then
            write(*,'(/a/)') string
        else
            write(*,'(/2a/)')   'Error termination of frealix, traceback below ', &
                                '(if you do not see a stack trace below, check the environment variable for_disable_stack_trace).'
        endif
        call tracebackqq(USER_EXIT_CODE=-1)
        write(*,'(/2a/)') 'Termination date & time : ', myTime%DateAndTimeAsString()

        write(*,'(/4a/)') myTime%DateAndTimeAsString(), ': frealix says sorry (total execution time: ', &
                                                myTime%DurationAsString(real(myTime%SecondsBetweenDates(date_start,date_finish))), ')'
        if (ddont_stop) then
            write(*,'(a)') '**debug(terminate): abnormal termination, see error messages above'
        else
            write(*,'(a)') '</frealix>'
            stop
        endif
#elif defined(NAGFOR_COMPILER)
        call flush(6)
        if (present(string)) then
            write(*,'(/a/)') string
        else
            write(*,'(/2a/)')   'Error termination of frealix, traceback below ', &
                                '(if you do not see a stack trace below, check the environment variable for_disable_stack_trace).'
        endif
        ! Now we generate a trace back the ugly way
        write(*,'(a)') '**info(terminate): Frealix is terminating in error state, see error message(s) above for details.'
        write(*,'(a)') '**info(terminate): will now attempt to generate a back trace by trigerring a floating point exception'
        write(*,'(/4a/)') myTime%DateAndTimeAsString(), ': frealix says sorry (total execution time: ', &
                                        myTime%DurationAsString(real(myTime%SecondsBetweenDates(date_start,date_finish))), ')'
        call flush(6)
        dummy = 0.0
        !dummy = 100.0/dummy
		dummy = atan2(dummy,dummy)
		if (ddont_stop) then
            write(*,'(a)') '**debug(terminate): abnormal termination, see error messages above'
        else
            write(*,'(a)') '</frealix>'
            stop
        endif
#else
        if (.not. ddont_stop) then
            write(*,'(a)') '</frealix>'
            stop
        endif
#endif
    end subroutine terminate

    !>    \brief    Register signal handlers
    subroutine register_signal_handlers()
#if defined(__INTEL_COMPILER)
        use ifport
#endif
        implicit none
        ! Arguments

        ! Private variables
#if defined(__INTEL_COMPILER)
        integer    ::    h_ret
        ! Start work
        h_ret = signalqq(sig$abort,signal_handler)
        h_ret = signalqq(sig$fpe,fpe_handler) + h_ret
        h_ret = signalqq(sig$ill,signal_handler) + h_ret
        h_ret = signalqq(sig$int,signal_handler) + h_ret
        h_ret = signalqq(sig$segv,signal_handler) + h_ret
        h_ret = signalqq(sig$term,signal_handler) + h_ret
        if (h_ret .ne. 0) then
            write(*,'(a,i0)') '**ERROR(register_signal_handlers): errors reported by signalqq. h_ret = ', h_ret
            call terminate('Failed when initialising abort signal handler')
        endif
#else
        ! Signal handlers not implemented (yet)
#endif
    end subroutine register_signal_handlers


    !>    \brief    Signal handler for abort signal (only available when built with Intel compiler)
#if defined(__INTEL_COMPILER)
    function signal_handler(signum)
        use ifport
        implicit none
        ! Argument
        integer(kind=2),    intent(in)    ::    signum    !< Identifies the signal
        ! Result
        !dec$ attributes c :: signal_handler
        integer(kind=4)    ::    signal_handler
        ! Private variables
        logical,    parameter    ::    debug    =    .false.
        ! Start work
        write(*,'(a)') ' '
        if (debug) then
            write(*,'(a,i0)') '**debug(signal_handler): caught signal # ', signum
        endif
        ! Work out which signal it is
        ! Here is the full list, from ifport.f90:
        !integer(4), parameter :: sig$err   = -1
        !integer(4), parameter :: sig$int   =  2
        !integer(4), parameter :: sig$ill   =  4
        !integer(4), parameter :: sig$fpe   =  8
        !integer(4), parameter :: sig$segv  = 11
        !integer(4), parameter :: sig$term  = 15
        !integer(4), parameter :: sig$usr1  = 16
        !integer(4), parameter :: sig$usr2  = 17
        !integer(4), parameter :: sig$usr3  = 20
        !integer(4), parameter :: sig$break = 21
        !integer(4), parameter :: sig$abort = 22
        !integer(4), parameter :: sig$nsig  = 23
        select case (signum)
            case (sig$abort)
                write(*,'(a)') '**info(signal_handler): caught abnormal termination signal (sigabort). Will terminate.'
                call terminate('Abnormal termination signal received',DONT_STOP=.true.)
            case (sig$fpe)
                write(*,'(2a)') '**ERROR(signal_handler): received floating-point error signal (sigfpe), ', &
                                'but this should have been handled by another routine!'
                write(*,'(a)') '**info(signal_handler): caught floating-point error signal (sigfpe). Will terminate.'
                call terminate('Floating-point error signal received by wrong signal handler',DONT_STOP=.true.)
            case (sig$ill)
                write(*,'(a)')  '**info(signal_handler): caught illegal instruction error signal (sigill). Will terminate.'
                call terminate('Illegal instruction signal received',DONT_STOP=.true.)
            case (sig$int)
                write(*,'(2a)') '**info(signal_handler): caught interrupt signal (sigint). ',    &
                                'This may be because you used Ctrl-C. Will terminate.'
                call terminate('Interrupt signal received')
            case (sig$segv)
                write(*,'(2a)') '**info(signal_handler): caught illegal storage access signal (sigsegv), ', &
                                'aka segmentation fault. Will terminate.'
                call terminate('Segmentation fault signal received',DONT_STOP=.true.)
            case (sig$term)
                write(*,'(a)') '**info(signal_handler): caught termination request signal (sigterm). Will terminate.'
                call terminate('Termination request received')
            case default
                write(*,'(a,i0,a)') '**ERROR(signal_handler): caught unknown signal number ', signum, '. Will terminate.'
                call terminate('Unknown signal received')
        end select
        !
        signal_handler = 1
    end function signal_handler
#endif

    !>    \brief    Signal handler for floating point error signals (only available when built with Intel compiler)
#if defined(__INTEL_COMPILER)
    function fpe_handler(signum,siginfo)
        use ifcore
        use ifport
        implicit none
        ! Argument
        integer(kind=2),    intent(in)    ::    signum    !< Identifies the signal
        integer(kind=2),    intent(in)    ::    siginfo
        ! Result
        !dec$ attributes c :: fpe_handler
        integer(kind=4)    ::    fpe_handler
        ! Private variables
        logical,    parameter    ::    debug    =    .false.
        ! Start work
        write(*,'(a)') ' '
        if (debug) then
            write(*,'(a,2(i0,1x))') '**debug(fpe_handler): caught signal #, info ', signum, siginfo
        endif
        ! Work out which signal it is
        select case (signum)
            case (sig$fpe)
                write(*,'(a)',advance='NO') '**info(fpe_handler): caught the following floating-point error signal (sigfpe): '
                select case (siginfo)
                    case (fpe$invalid)
                        write(*,'(a)',advance='NO') 'invalid operation'
                    case (fpe$zerodivide)
                        write(*,'(a)',advance='NO') 'division by zero'
                    case (fpe$overflow)
                        write(*,'(a)',advance='NO') 'numeric overflow'
                    case (fpe$underflow)
                        write(*,'(a)',advance='NO') 'numeric underflow'
                    case (fpe$inexact)
                        write(*,'(a)',advance='NO') 'inexact precision'
                    case default
                        write(*,'(a)',advance='NO') 'unknown'
                end select
                write(*,'(a)') '. Will terminate.'
                call core_dump_instructions()
                call terminate('Floating-point error signal received',DONT_STOP=.true.)
            case default
                write(*,'(a,i0,a)') '**ERROR(fpe_handler): caught unknown signal number ', signum, '. Will terminate.'
                call terminate('Unknown signal received')
        end select
        !
        fpe_handler = 1
    end function fpe_handler
#endif

    !>    \brief    Print out instructions on how to get a core dump
    subroutine core_dump_instructions()
        implicit none
        ! Argumnets

        ! Private variables

        ! Start work
        write(*,'(/2a)')'**info(core_dump_instructions): if you would like to get a core dump for debugging purposes, ',   &
                        'issue the following commands and re-run the program:'
        write(*,'(a)')  '**info(core_dump_instructions): C-shell: limit coredumpsize unlimited; setenv decfort_dump_flag y'
        write(*,'(a/)') '**info(core_dump_instructions): Bash: ulimit -c unlimited; export decfort_dump_flag=y'
    end subroutine core_dump_instructions

    !>    \brief    Check whether the real supplied as argument is safe for square-rooting. If not, crash out.
    !!    \todo    Come up with a proper tolerance
    elemental subroutine safe_for_sqrt_real(self,negative_to_zero)
!#if defined(__INTEL_COMPILER)
!        use ifport
!#endif
        implicit none
        ! Arguments
        real(kind=4),                intent(inout)  ::    self
        logical,        optional,    intent(in)     ::    negative_to_zero    !<    If true and if the input value is negative, it'll be zeroed
        ! Private variables
        real(kind=4),   parameter        ::    tolerance    =    -.0000001
        logical    ::    nnegative_to_zero
        ! Start work
        if (present(negative_to_zero)) then
            nnegative_to_zero = negative_to_zero
        else
            nnegative_to_zero = .false.
        endif

        ! We only need to worry about the variable being negative
        if (self .lt. 0.) then
            if (self .gt. tolerance .or. nnegative_to_zero) then
                self = 0.
            !else
            !    write(*,'(a,G15.7)') '**ERROR(safe_for_sqrt_real): cannot compute square root of ', self
            !    call terminate('Fatal error in SAFE_FOR_SQRT_REAL'    &
!#if defined(__INTEL_COMPILER)
!                , DONT_STOP=.TRUE.    &
!#endif
            !    )
!#if defined(__INTEL_COMPILER)
!                idummy = raiseqq(sig$fpe)
!#endif
            endif
        endif
    end subroutine safe_for_sqrt_real

    !>    \brief    Check whether the double supplied as argument is safe for square-rooting. If not, crash out.
    !!    \todo    Come up with a proper tolerance
    subroutine safe_for_sqrt_dble(self,negative_to_zero)
!#if defined(__INTEL_COMPILER)
!        use ifport
!#endif
        implicit none
        ! Arguments
        real(kind=8),                intent(inout)    ::    self
        logical,        optional,    intent(in)        ::    negative_to_zero    !<    If true and if the input value is negative, it'll be zeroed
        ! Private variables
        real(kind=8),    parameter        ::    tolerance    =    -.000000001D0
        logical    ::    nnegative_to_zero
        ! Start work
        if (present(negative_to_zero)) then
            nnegative_to_zero = negative_to_zero
        else
            nnegative_to_zero = .false.
        endif

        ! We only need to worry about the variable being negative
        if (self .lt. 0.D0) then
            if (self .gt. tolerance .or. nnegative_to_zero) then
                self = 0.D0
            !else
            !    write(*,'(a,G15.7)') '**ERROR(safe_for_sqrt_dble): cannot compute square root of ', self
            !    call terminate('Fatal error in SAFE_FOR_SQRT_DBLE'    &
!#if defined(__INTEL_COMPILER)
!                , DONT_STOP=.TRUE.    &
!#endif
            !    )
!#if defined(__INTEL_COMPILER)
!                idummy = raiseqq(sig$fpe)
!#endif
            endif
        endif
    end subroutine safe_for_sqrt_dble


    !>    \brief    Print a welcome message to the terminal
    subroutine print_welcome()
        implicit none
        ! Arguments

        ! Private variables
        character(len=2),    parameter    ::    text_width    =    '30'
        type(DateAndTime)   ::  myTime
        ! Start work
        write(*,'(/18x,a/)')                '** Welcome to Frealix **'
        write(*,'(A'//text_width//',1x,a)')        'Version:',                    version
        write(*,'(A'//text_width//',1x,a)')        'SVN revision:',            SVN_REVision
        write(*,'(A'//text_width//',1x,a)')        'Build date:',                compile_date
        write(*,'(A'//text_width//',1x)',advance='NO')        'Debug symbols?:'
#if defined (with_debug_symbols)
        write(*,'(a)')                        'Yes'
#else
        write(*,'(a)')                        'No'
#endif
        write(*,'(a)')                         ' '
        if (interactive) then
        write(*,'(A'//text_width//',1x,a)')        'Interactive mode?:', 'Yes'
        else
        write(*,'(A'//text_width//',1x,a)')        'Control filename:',    trim(adjustl(control_filename))
        endif
        write(*,'(a)')                         ' '
        call print_environment_info()
        write(*,'(a)')                         ' '
        write(*,'(A'//text_width//',1x,a)')        'Date & time:', myTime%DateAndTimeAsString()
        write(*,'(a)')                         ' '
        ! Flush standard output
        flush(6) ! This is a standard f2003 statement
    end subroutine print_welcome

    !>    \brief    Find values for enviroment variables which may affect frealix at runtime and print them out
    subroutine print_environment_info()
#ifdef __INTEL_COMPILER
        use ifport
#elif defined(NAGFOR_COMPILER)
        use f90_unix_env
#endif
        implicit none
        ! Arguments

        ! Private variables
        character(len=2),    parameter          ::    text_width    =    '30'
        character(len=WORD_LEN)                 ::    home, hostname, pwd, username, hosttype, machtype
        character(len=LINE_LEN)                 ::    command_line
        integer    ::    ierr,cmd_len
#ifdef __INTEL_COMPILER
        integer     ::  res
#endif
        ! Start work
        call get_env_var('HOME',home)
#ifdef __INTEL_COMPILER
        res = hostnam(hostname)
        if (res .ne. 0) call get_env_var('HOST',hostname)
#elif defined(NAGFOR_COMPILER)
        call gethostname(hostname)
#else
        call get_env_var('HOST',hostname)
#endif
        call get_env_var('PWD',pwd)
        call get_env_var('USER',username)
        call get_env_var('HOSTTYPE',hosttype)
        call get_env_var('MACHTYPE',machtype)
        call get_command(command_line,LENGTH=CMD_LEN,STATUS=IERR)
        if (ierr .ne. 0) call terminate('Could not get command line')

        write(*,'(A'//text_width//',1x,a,1x,4a)')   'User name (home directory):',      trim(adjustl(username)), '(', &
                                                                                        trim(adjustl(home)), ')'
        write(*,'(A'//text_width//',1x,a)')         'Working directory:',               trim(adjustl(pwd))
        write(*,'(A'//text_width//',1x,a)')         'Command line:',                    trim(adjustl(command_line))
        write(*,'(A'//text_width//',1x,a)')         'Host name:',                       trim(adjustl(hostname))
        write(*,'(A'//text_width//',1x,2a)')        'Host, machine type:',              trim(adjustl(hosttype))//', '//&
                                                                                        trim(adjustl(machtype))

        call print_openmp_info()

    end subroutine print_environment_info

    !>    \brief    Print information on OpenMP which may affect runtime behaviour
    subroutine print_openmp_info()
        use omp_lib
        implicit none
        ! Arguments

        ! Private variables
        character(len=2),  parameter     ::     text_width      =   '30'
#ifdef _OPENMP
        character(len=WORD_LEN)          ::     env_omp_num_threads
        logical    ::    env_omp_num_threads_is_set
#endif
        logical    ::    omp_is_on
        ! Start work

        omp_is_on = openmp_is_available()

        if (omp_is_on) then
#ifdef _OPENMP
            ! We manually turn thread nesting off, to keep things simple
            call omp_set_nested(.false.)
            write(*,'(A'//text_width//',1x,a)')    'OpenMP threading available?:', 'Yes'
            ! First printout relevant environment variables the user may have set
            call get_env_var('OMP_NUM_THREADS',env_omp_num_threads,IS_SET=ENV_OMP_NUM_THREADS_IS_SET,QUIET=.true.)
            if (env_omp_num_threads_is_set) then
                write(*,'(A'//text_width//',1x,a)')        'Omp_num_threads env. var.:',     trim(adjustl(env_omp_num_threads))
            endif
            write (*,'(A'//text_width//',1x,i0)') 'Max number of OpenMP threads:', omp_get_max_threads()
            write (*,'(A'//text_width//',1x,i0)') 'Number of processors:', omp_get_num_procs()
            if (omp_get_nested()) then
                write(*,'(A'//text_width//',1x,a)')    'Nested threading available?:', 'Yes'
            else
                write(*,'(A'//text_width//',1x,a)')    'Nested threading available?:', 'No'
            endif
#endif
        else
            write(*,'(A'//text_width//',1x,a)')    'OpenMP threading available?:', 'No'
        endif


    end subroutine print_openmp_info

    !>    \brief    Test whether OpenMP is available
    logical function openmp_is_available()
        use omp_lib
        implicit none
        ! Arguments

        ! Private variables

        ! Start work
#ifdef _OPENMP
        !$omp    parallel    &
        !$omp          shared(openmp_is_available)
        !$omp    single
        openmp_is_available = omp_get_max_threads() .gt. 1
        !$omp    end single
        !$omp    end parallel
#else
        openmp_is_available = .false.
#endif
    end function openmp_is_available

    !>  \brief  Return the id of the current process
    integer function get_process_id()
#if defined(__INTEL_COMPILER)
        use ifport
#elif defined(NAGFOR_COMPILER)
        use f90_unix_env
#endif
        implicit none
        ! Arguments
        ! Private variables
        ! Start work
#if defined(__INTEL_COMPILER)
        get_process_id = getpid()
#elif defined(NAGFOR_COMPILER)
        get_process_id = getpid()
#else
        get_process_id = 0
#endif
    end function get_process_id

    !>    \brief    Find out the value of an environment variable at runtime
    subroutine get_env_var(name,value,is_set,quiet)
        implicit none
        ! Arguments
        character(len=*),                   intent(in)      ::    name      !<    Status code returned by get_environment_variable
        character(len=*),                   intent(out)     ::    value
        logical,            optional,       intent(out)     ::    is_set    !<    Whether the environment variable exists / is set
        logical,            optional,       intent(in)      ::    quiet     !<    Do not print warnings. True by default.
        ! Private variables
        integer                            ::    length
        integer                            ::    status
        logical                            ::    iis_set, qquiet
        character(len=WORD_LEN)    ::    val
        ! Start work

        ! Will we be quiet?
        if (present(quiet)) then
            qquiet = quiet
        else
            qquiet = .true.
        endif

        !
        call get_environment_variable(name,val,length,status,.true.)
        iis_set = .false.
        if (status .ge. 2) then
            write(*,'(a,i0)') '**ERROR(get_env_var): get_environment_variable failed with error ', status
            call terminate('Fatal error')
        else if (status .eq. 1) then
            if (.not. qquiet) write(*,'(3a)') '**warning(get_env_var): environment variable "', &
                                                trim(adjustl(name)), '" does not exist'
        else if (status .eq. -1) then
            if (.not. qquiet) write(*,'(2a,2(a,i0))') '**warning(get_env_var): environment variable "', &
                                                trim(adjustl(name)), '" length (', length, ') was trimmed to ', len(val)
        else if (length .eq. 0) then
            if (.not. qquiet) write(*,'(2a,2(a,i0))') '**warning(get_env_var): environment variable "', &
                                                trim(adjustl(name)), '" exists but has no value'
        else
            iis_set = .true.
        endif

        !
        val = adjustl(val)
        write(value,'(a)') trim(adjustl(val))

        ! The user asked to know whether that variable was set, so we let them know
        if (present(is_set)) then
            is_set = iis_set
        endif
    end subroutine get_env_var


    !>    \brief    Returns true if the argument is even
    pure logical function is_even_slow(int)
        implicit none
        ! Argument
        integer,    intent(in)    ::    int
        ! Start work
        if (mod(int,2) .eq. 0.) then
            is_even_slow = .true.
        else
            is_even_slow = .false.
        endif
    end function is_even_slow

    !>    \brief    Returns true if the argument is even
    pure elemental logical function is_even(int)
        implicit none
        ! Argument
        integer,    intent(in)    ::    int
        ! Start work
        ! Test bit 0 of number. If it is 0, the number is even
        is_even = .not. btest(int,0)
    end function is_even

    !>    \brief    Returns true if the argument is odd
    pure elemental logical function is_odd(int)
        implicit none
        ! Argument
        integer,    intent(in)    ::    int
        ! Start work
        is_odd = btest(int,0)
    end function is_odd

    !>    \brief    Returns true if the argument is odd
    pure logical function is_odd_slow(int)
        implicit none
        ! Argument
        integer,    intent(in)    ::    int
        ! Start work
        if (is_even_slow(int)) then
            is_odd_slow = .false.
        else
            is_odd_slow = .true.
        endif
    end function is_odd_slow

    !>    \brief    Return the array of unique elements present in the input array
    function int_unique(array)
        implicit none
        ! Argument
        integer,    intent(in)    ::    array(:)
        ! Result
        integer,    allocatable    ::    int_unique(:)
        ! Private variables
        integer    ::    number_of_unique_elements
        integer    ::    tmp_result(size(array))
        integer    ::    i, j, ierr
        logical    ::    unique
        ! Start work

        ! Initialise the output
        number_of_unique_elements = 1
        tmp_result(1) = array(1)

        ! Loop over the input array
        loop_over_input: do I=2,size(array)
            unique = .true.
            ! Loop over the input array
            do J=1,number_of_unique_elements
                if (array(i) .eq. tmp_result(j)) then
                    unique = .false.
                    cycle loop_over_input
                endif
            enddo
            ! If we got here, it means the element is unique - add it to the output
            number_of_unique_elements = number_of_unique_elements + 1
            tmp_result(number_of_unique_elements) = array(i)
        enddo loop_over_input

        ! Allocate memory for the output
        allocate(int_unique(number_of_unique_elements),stat=IERR)
        if (ierr .ne. 0) then
            call terminate('Memory allocation failed in INT_UNIQUE')
        endif

        ! Copy values over to the output array
        int_unique(1:number_of_unique_elements) = tmp_result(1:number_of_unique_elements)

    end function int_unique




end module runtime_parameters
