!>  \brief  Classes & methods for dealing with single particles
module single_particles

    use euleranglesmod
    use ctfs
    use helices

    ! Portland compiler (v 13.3) does not like extending a type across modules, it seems
    !type, extends(point) :: single_particle
    type single_particle
        real(kind=8)            ::  coo(3)  =   [0.0D0,0.0D0,0.0D0] !<  x, y, z coordinates
        integer                 ::  loc_in_stack            !<  Location number in stack; first image in stack has number 1
        type(euler)             ::  angles
        type(ctf_parameters)    ::  ctf
        real(kind=8)            ::  shift_x
        real(kind=8)            ::  shift_y
        integer                 ::  film_num
        real(kind=8)            ::  subunit_num =   0.0D0   !<  When dealing with helical assemblies, the position of subunit nearest to the origin
        type(helical_symmetry)  ::  helix                   !<  Helical symmetry operator when dealing with segments of helical assembles
        real                    ::  score
    end type single_particle

    interface operator (+)
        module procedure sp_plus_sp
    end interface

    interface operator (-)
        module procedure sp_minus_sp
    end interface

    contains

    pure elemental function sp_plus_sp(sp1,sp2)
        implicit none
        !
        type(single_particle), intent(in)  ::  sp1, sp2
        !
        type(single_particle)  ::  sp_plus_sp
        !
        sp_plus_sp%angles      = sp1%angles + sp2%angles
        sp_plus_sp%ctf         = sp1%ctf + sp2%ctf
        sp_plus_sp%shift_x     = sp1%shift_x + sp2%shift_x
        sp_plus_sp%shift_y     = sp1%shift_y + sp2%shift_y
        sp_plus_sp%film_num    = sp1%film_num + sp2%film_num
        sp_plus_sp%subunit_num = sp1%subunit_num + sp2%subunit_num
        sp_plus_sp%helix       = sp1%helix + sp2%helix
        sp_plus_sp%score       = sp1%score + sp2%score
    end function sp_plus_sp

    pure elemental function sp_minus_sp(sp1,sp2)
        implicit none
        !
        type(single_particle), intent(in)  ::  sp1, sp2
        !
        type(single_particle)  ::  sp_minus_sp
        !
        sp_minus_sp%angles      = sp1%angles - sp2%angles
        sp_minus_sp%ctf         = sp1%ctf - sp2%ctf
        sp_minus_sp%shift_x     = sp1%shift_x - sp2%shift_x
        sp_minus_sp%shift_y     = sp1%shift_y - sp2%shift_y
        sp_minus_sp%film_num    = sp1%film_num - sp2%film_num
        sp_minus_sp%subunit_num = sp1%subunit_num - sp2%subunit_num
        sp_minus_sp%helix       = sp1%helix - sp2%helix
        sp_minus_sp%score       = sp1%score - sp2%score
    end function sp_minus_sp



    !>  \brief  Initialise a single particle object
    subroutine sp_init(self,ctf,angles,shift_x,shift_y,phi,theta,psi,subunit_num,helix)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        class(single_particle),             intent(inout)   ::  self        !<  Single particle object
        type(ctf_parameters),               intent(in)      ::  ctf         !<  ctf parameters object
        type(euler),            optional,   intent(in)      ::  angles      !<  Euler angles. If not given, and separate angles are not given either, angles will be set to 0.0.
        real(kind=8),           optional,   intent(in)      ::  shift_x     !<  x shift (pixels)
        real(kind=8),           optional,   intent(in)      ::  shift_y     !<  y shift (pixels)
        real(kind=8),           optional,   intent(in)      ::  phi         !<  Phi Euler angles (degrees)
        real(kind=8),           optional,   intent(in)      ::  theta       !<  Theta Euler angles (degree)
        real(kind=8),           optional,   intent(in)      ::  psi         !<  Psi Euler angles (degrees)
        real(kind=8),           optional,   intent(in)      ::  subunit_num !<  Subunit number at the origin (always 0.0 for non-helical objects)
        type(helical_symmetry), optional,   intent(in)      ::  helix       !<  Helical operator. Required when doing sp processing of filament segments.
        ! Private variables

        ! Start work

        ! Deal with angles
        if (present(angles)) then
            self%angles = angles
            if (present(phi) .or. present(theta) .or. present(psi)) then
                write(*,'(a)') '**ERROR(sp_init): either give an Euler object, or separate angles, not both'
                call terminate('Fatal error in sp_init - programming error')
            endif
        else
            if (present(phi)) then
                self%angles%phi = phi
            else
                self%angles%phi = 0.0D0
            endif
            if (present(theta)) then
                self%angles%theta = theta
            else
                self%angles%theta = 0.0D0
            endif
            if (present(psi)) then
                self%angles%psi = psi
            else
                self%angles%psi = 0.0D0
            endif
        endif

        ! Deal with shifts
        if (present(shift_x)) then
            self%shift_x = shift_x
        else
            self%shift_x = 0.0D0
        endif
        if (present(shift_y)) then
            self%shift_y = shift_y
        else
            self%shift_y = 0.0D0
        endif

        ! Deal with subunit_num
        if (present(subunit_num)) then
            self%subunit_num = subunit_num
        else
            self%subunit_num = 0.0D0
        endif

        ! Deal with helical operator
        if (present(helix)) then
            self%helix = helix
        else
            call helical_symmetry_init(self%helix,rise=0.0,twist=0.0,outer_radius=0.0)
        endif

        ! Copy the ctf parameters over
        self%ctf = ctf
    end subroutine sp_init

    subroutine sps_print_info(self)
        implicit none
        type(single_particle),  intent(in)  ::  self(:)
        integer ::  i
        do I=1,size(self)
            call sp_print_info(self(i))
        enddo
    end subroutine sps_print_info

    !>  \brief  Print useful info to the terminal which describes the single-particle parameters
    subroutine sp_print_info(self)
        implicit none
        ! Arguments
        class(single_particle), intent(in)  ::  self    !<  Single particle object
        ! Private variables
        logical,    parameter   ::  annotate = .true.
        ! Start work
        if (annotate) then
            write(*,'(3(F0.3,1x,a,1x),5(F0.3,1x,a,1x),2(F0.3,1x,a,1x),(F0.3,1x,a,1x))')                 &
                                                        self%angles%phi,    '(phi)',            &
                                                        self%angles%theta,  '(theta)',          &
                                                        self%angles%psi,    '(psi)',            &
                                                        self%shift_x,       '(x)',              &
                                                        self%shift_y,       '(y)',              &
                                                        self%subunit_num,   '(subunit_num)',    &
                                                        self%helix%rise,    '(h. rise)',        &
                                                        self%helix%twist,   '(h. twist)',       &
                                                        self%ctf%dfmid1,    '(df1)',            &
                                                        self%ctf%dfmid2,    '(df2)',            &
                                                        self%score,         '(score)'
        else
            write(*,'(3(F8.3,1x),3(F8.3,1x),2(F8.3,1x))') self%angles, self%shift_x, self%shift_y, self%subunit_num,   &
                                                        self%ctf%dfmid1, self%ctf%dfmid2
        endif
    end subroutine sp_print_info

end module single_particles

!>  \brief  Classes & methods for dealing with comparisons between singles particles and 3d models
module sp_model_comparisons

    use single_particles, only : single_particle
    use Images
    use InterpolatingWindows
    use interpolation, only : interpolation_scheme
    use helices, only : helical_symmetry
    use, intrinsic :: iso_c_binding
    use minimisers, only : downhill_simplex, conjugate_gradient, differential_evolution, brute_force

    implicit none

    private

    public  ::  sp_model_comparison, spmc_init, spmc_print_info, spmc_eval, spmc_get_sp, spmc_ds_maximise
    public  ::  spmc_deallocate_work_memory, spmc_bf_maximise


    type sp_model_comparison
        private
        type(single_particle)       ::  sp                                      !<  Single-particle object, describes the single particle parameters (angles, shifts, CTFs)
        type(image)                 ::  model3d                                 !<  3d image of the model
        type(helical_symmetry)      ::  model_helix                             !<  Helical symmetry of the model 3D. This can be different from the helical symmetry described by the single-particle object.
        type(image)                 ::  sp_image                                !<  Image of the single particle (before shifts are applied)
        real                        ::  pixel_size                              !<  Pixel size (in Angstroms)
        ! We hold in memory a few intermediate steps to avoid recalculating them everytime we want to score a comparison
        logical                     ::  first_use               =   .true.      !<  Whether intermediate images have already been computed. They still
                                                                                !!  may need to be recomputed even if this is set to .FALSE., for example
                                                                                !!  if we now need to score parameters which have changed.
        type(single_particle)       ::  sp_ready                                !<  The single particle parameters (ctf, shifts, Euler angles) for which the intermediate images were calculated
        type(image)                 ::  ctf_image_ready                         !<  Image of the ctf function, which will be applied to the image and the projection
        type(image), public         ::  proj_ready                              !<  Projection image ready for similarity measure
        type(image), public         ::  sp_image_ready                          !<  Single particle image ready for similarity measure
        type(image)                 ::  env_mask_ready                          !<  Envelope mask prepared from the projection
        ! The components below affect behaviour when optimising the sp/model comparison
        integer                     ::  minimiser_algorithm     =   downhill_simplex !<  What kind of minimiser will we use
        real                        ::  target_resolution       =   0.0         !<  Target resolution (a) for the refinement
        real                        ::  freq_max                =   0.5         !<  Highest frequency (1/pixel units, 0.5 is Nyquist) to use when scoring using correlation functions.
        real                        ::  freq_min                =   0.0         !<  Lowest frequency (1/pixel units, 0.5 is Nyquist) to use when scoring using correlation functions.
        real                        ::  obj_max_rad                             !<  Maximum (spherical) radius of the 3d object (pixels)
        real                        ::  obj_max_rad_cyl         =   0.0         !<  Maximum (cylindrical) radius of the 3d object (pixels). Will be ignored if set to 0.0.
        logical                     ::  refine_params(9)        =   .false.     !<  Which of the 6 degrees of freedom to refine (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi)
        real                        ::  param_scale(9)          =   0.0         !<  For each parameter, a scaling factor to ensure that all parameters have comparable values
        real                        ::  search_accuracy(9)      =   0.0         !<  The desired accuracy of the search result (pixels, degrees).
        real                        ::  search_half_range(9)    =   0.0         !<  The half range (degrees, pixels) of exhaustive search (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi)
        real                        ::  search_sampling(9)      =   0.0         !<  The sampling step size (pixels, degrees) during exhaustive search (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi)
        logical                     ::  real_space_proj         =   .false.     !<  Whether to use real space projection. If .FALSE., reciprocal space slice extraction is used instead (which is faster).
        integer                     ::  num_minim_runs          =   1           !<  The number of times the minimiser should be run. Only meant for dowhill simplex (ds), which can be started several times from the same position but with a random first step.
        logical                     ::  ds_random_start         =   .false.     !<  Whether the starting setup of the ds should be randomised (see gsl documentation for gsl_multimin_fminimizer_nmsimplex2rand
                                                                                !!  at http://www.gnu.org/software/gsl/manual/html_node/Multimin-Algorithms-without-Derivatives.html)
        !
        logical                     ::  use_ccf_to_find_xy      =   .false.     !<  Use peak search in cross-correlation function to find the best xy shift to align the sp image to the projection
        real                        ::  latest_ccf_xy(2)        =   0.0         !<  Latest results of the ccf peak search
        !
        logical                     ::  project_individual_subunits = .false.   !<  Project individual subunits rather than segments of helically symmetrised volume
        logical                     ::  boost_low_frequency_contrast = .false.  !<  Boost low frequencies in the scoring funciton by only multiplying by the ctf after the first peak of the ctf
        integer                     ::  model3d_oversampling    =   1           !<  If greater than 1, the model 3d will be oversampled. This can be used to improve the quality of real-space projections.
        !
        type(interpolation_scheme)  ::  interpol                                !<  Interpolation scheme
        ! Work area memory
        type(window)                ::  interpol_window                         !<  Window image for interpolation
        type(image)                 ::  sp_image_tmp
        !
        type(c_ptr)                 ::  cptr                                    !<  a C-pointer pointing to itself. This is needed when passing the object through the FGSL.
        !contains
        !    procedure,  public      ::  init        =>  spmc_init
        !    procedure,  public      ::  print_info  =>  spmc_print_info
        !    procedure,  public      ::  eval        =>  spmc_eval
        !    procedure,  public      ::  get_sp      =>  spmc_get_sp
        !    procedure,  public      ::  ds_maximise =>  spmc_ds_maximise
    end type sp_model_comparison

    contains

    !>  \brief  Brute-force maximisation of the similarity measure between the single particle
    !!  image and the 3d mdel by varying the refinement paramenters
    subroutine spmc_bf_maximise(self)
        use BruteForceSearches
        implicit none
        ! Arguments
        type(sp_model_comparison),      intent(inout)   ::  self
        ! Private variables
        type(BruteForceSearch)      ::  my_search
        real(kind=8), allocatable   ::  starting_values(:), half_ranges(:), step_sizes(:)
        real(kind=8), allocatable   ::  final_values(:)
        real                        ::  score
        ! Start work


        ! Initialise the brute-force search
        call spmc_get_values(self,starting_values)
        call spmc_get_half_ranges(self,half_ranges)
        call spmc_get_step_sizes(self,step_sizes)
        call my_search%Init(spmc_eval_wrap_nofgsl,  &
                            [self%cptr], &
                            count(self%refine_params), &
                            starting_values,half_ranges,step_sizes, &
                            minimise_at_every_step=.false., &
                            print_progress_bar=.true.)

        ! Run the search
        call my_search%Run()

        ! Debug
        write(*,'(a,f0.4)') '**debug(spmc_bf_maximise): final score = ', my_search%GetBestScore()
        write(*,*) '**debug(spmc_bf_maximise): final values = ', my_search%GetBestValues()

        ! Get results out
        call spmc_set_values(self,my_search%GetBestValues())

        ! If we were doing ccf x/y peak search, we need to call the scoring function again, to update self%latest_xy
        if (self%use_ccf_to_find_xy) then
            score = spmc_eval(self)
            self%sp%shift_x = self%sp%shift_x + self%latest_ccf_xy(1)
            self%sp%shift_y = self%sp%shift_y + self%latest_ccf_xy(2)
        endif

    end subroutine spmc_bf_maximise

    !>  \brief  Return a single-particle object with the current state of the SP-model comparison
    function spmc_get_sp(self) result(sp)
        implicit none
        ! Arguments
        class(sp_model_comparison),     intent(in)  ::  self    !<  SP-model comparison object
        ! Result
        type(single_particle)   ::  sp
        ! Private variables

        ! Start work
        sp = self%sp

    end function spmc_get_sp

    !>  \brief  Dowhill simplex maximisation of the similarity measure between the single particle image and the 3d model by varying the refinement parameters
    subroutine spmc_ds_maximise(self)
        use minimisers, only : minimiser_ds_fgsl, search_accuracy_to_ds_size_ratio, ds_init, ds_run, ds_set_start_step
        use single_particles
        use images_comparison, only : score_got_worse
        use runtime_parameters, only : terminate
        use images
        use Globals
        implicit none
        ! Arguments
        class(sp_model_comparison),     intent(inout)   ::  self    !<  Single particle (sp) model comparison object
        ! Private variables
        real                        ::  ds_target_size
        real(kind=8),   allocatable ::  start_values(:), final_values(:,:), best_final_values(:), final_score(:)
        real(kind=8),   allocatable ::  first_step(:)
        integer                     ::  i, ierr
        type(minimiser_ds_fgsl)     ::  ds_minimiser    !<  a minimiser object (we use a fgsl minimiser - this is all we have as of Jan2010)
        integer                     ::  num_dim_total
        integer                     ::  current_minim_run
        logical,        parameter   ::  debug   =   .false.
        integer                     ::  best_minim_run(1)
        real                        ::  initial_score
        type(single_particle)       ::  sp_start
        character(len=6)            ::  num_dim_c
        logical,    parameter       ::  randomise_simplex_axes_orientation = .true. !< If false, the axes will be held in place, but the first step size will be randomised
        real                        ::  score
        ! Start work

        ! Do we have any work to do?
        if (.not. any(self%refine_params)) then
            write(*,'(a)') '**WARNING(spmc_ds_maximise): nothing to do, since no parameters are to be refined'
            return
        endif

        ! If only the in-plane shifts are to be optimised, then the user should call a CCF-based routine
        if (all(self%refine_params .eqv. [.true.,.true.,.false.,.false.,.false.,.false.,.false.,.false.,.false.])) then
            write(*,'(a)') '**ERROR(spmc_ds_maximise): refinement of x and y shifts only is done more ',    &
                            'efficiently using cross-correlation functions'
            call terminate('Fatal error in SPMC_DS_MAXIMISE')
        endif

        ! We may need to prepare the reference volume
        !$omp   critical
        if (.not. self%real_space_proj) then
            if (self%model3d%is_in_real_space) then
                write(*,'(a)') '**warning(spmc_ds_maximise): FWD_FTing the volume'
                call self%model3d%ForwardFFT()
            endif
            if (self%model3d%object_is_centered_in_box) then
                write(*,'(a)') '**warning(spmc_ds_maximise): shifting the real-space origin of the volume'
                call self%model3d%SwapRealSpaceQuadrants()
            endif
        endif
        !$omp   end critical

        ! Work out a target size for the simplex
        ds_target_size = 0.0
        do i=1,9
            if (self%refine_params(i)) then
                ds_target_size = ds_target_size + (self%search_accuracy(i)/self%param_scale(i))/search_accuracy_to_ds_size_ratio
            endif
        enddo
        ds_target_size = ds_target_size / count(self%refine_params)
        !$omp   single
        if (this_program%IsVerbose() .or. debug) then
            write(*,'(a,F0.4)') '**info(spmc_ds_maximise): target size for simplex = ', ds_target_size
        endif
        !$omp   end single  nowait

        ! Get the start values for each parameter
        call spmc_get_values(self,start_values)

        ! Starting score?
        initial_score = spmc_eval(self)

        ! Print out initial single particle state
        !$omp   critical        (print_out_starting_sp_ds_omp_crit)
        sp_start = self%sp
        write(*,'(2a)',advance='NO') '**info(spmc_ds_maximise): prior to downhill simplex minimisation, ',   &
                                    'the single particle parameters are: '
        call sp_print_info(self%sp)
        !write(*,'(a,F0.5)') '**info(spmc_ds_maximise): and the starting score is ', initial_score
        !$omp   end critical    (print_out_starting_sp_ds_omp_crit)

        ! Initialise the ds minimiser
        call ds_init(ds_minimiser,spmc_eval_wrap,count(self%refine_params),self%cptr,   &
                    initial_score=initial_score, &
                    random=(self%ds_random_start .or. (self%num_minim_runs .gt. 1)) .and. randomise_simplex_axes_orientation)

        ! Allocate memory for the results of the minimiser runs
        num_dim_total = count(self%refine_params)
        write(num_dim_c,'(i6)') num_dim_total
        allocate(   final_values(num_dim_total,self%num_minim_runs),    &
                    final_score(self%num_minim_runs),   &
                    best_final_values(num_dim_total),stat=IERR)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(spmc_ds_maximise): memory allocation failed with error ', ierr
            call terminate('Fatal error in SPMC_DS_MAXIMISE')
        endif


        do current_minim_run = 1, self%num_minim_runs

            ! The first step is essentially the half range, scaled appropriately
            first_step = spmc_ds_get_first_step(self,randomise_step=(self%ds_random_start .or. (self%num_minim_runs .gt. 1)) &
                                                                     .and. .not. randomise_simplex_axes_orientation &
                                                                     .and. current_minim_run .gt. 1)

            ! Set the starting position and the first step for the ds
            call ds_set_start_step(ds_minimiser,start_values,first_step)

            ! Run the ds minimiser
            call ds_run(ds_minimiser,ds_target_size)

            ! Get the last best values from the minimiser
            final_values(:,current_minim_run) = ds_minimiser%current_best_fgsl(:)
            final_score(current_minim_run) = ds_minimiser%current_minimum_fgsl
            if (debug) then
                print '(a,i0,a,i0,a,'//num_dim_c//'(F0.3,1x))',  &
                        '**debug(spmc_ds_maximise): final score for run ', current_minim_run, ' of ' , self%num_minim_runs, &
                        ' = ', final_score(current_minim_run)
                print '(a,i0,a,i0,a,'//num_dim_c//'(F0.3,1x))',  &
                        '**debug(spmc_ds_maximise): final values for run ', current_minim_run, ' of ' , self%num_minim_runs, &
                        ' = ', final_values(:,current_minim_run)
            endif
        enddo

        ! Decide which run of the minimiser gave the best results
        best_minim_run = minloc(final_score)
        if (self%num_minim_runs .gt. 1) then
            write(*,'(a,i0,a)') '**info(spmc_ds_maximise): minimisation run ', best_minim_run, ' gave the best results'
        endif
        best_final_values = final_values(:,best_minim_run(1))

        ! Plug those back into the single particle
        if (score_got_worse(initial_score,real(-final_score(best_minim_run(1))))) then
            if (self%ds_random_start .or. (self%num_minim_runs .gt. 1)) then
                ! Let's just forget about it and go back to input values
                call spmc_set_values(self,start_values)
                self%sp = sp_start
            else
                ! Something must have gone wrong
                write(*,'(a,2(f0.4,1x))') '**error(spmc_ds_maximise): the score did not improve: ', &
                                                initial_score, -final_score(best_minim_run(1))
                call terminate('Fatal error in spmc_ds_maximise')
            endif
        else
            call spmc_set_values(self,best_final_values)
            !self%sp%score = -final_score(best_minim_run(1))
            self%sp%score = spmc_eval(self)
#ifndef SKIP_RUNTIME_CHECKS
            if (abs(self%sp%score + final_score(best_minim_run(1))) .gt. 0.01) then
                write(*,'(a,f0.4,a,f0.4)') '**error(spmc_ds_maximise): recorded final score is not reproduced: ', &
                                            -final_score(best_minim_run(1)), self%sp%score
                call this_program%TerminateWithFatalError('spmc_ds_maximise','Failed sanity check')
            endif
#endif

            ! If we were doing ccf x/y peak search, we need to call the scoring function again, to update self%latest_xy
            if (self%use_ccf_to_find_xy) then
                score = spmc_eval(self)
                self%sp%shift_x = self%sp%shift_x - self%latest_ccf_xy(1)
                self%sp%shift_y = self%sp%shift_y - self%latest_ccf_xy(2)
                if (debug) write(*,'(a,2(f0.3,1x))') '**debug(spmc_ds_maximise): new sp shifts: ', self%sp%shift_x, self%sp%shift_y
            endif
        endif

        ! Return the reference volume to its original state
        !!$omp  critical
        !if (.NOT. self%real_space_proj) then
        !   if (.NOT. self%model3d%object_is_centered_in_box) call self%model3d%ShiftRealSpaceOrigin()
        !   if (self%model3d%type .NE. 'REAL') call self%model3d%BackwardFFT()
        !endif
        !!$omp  end critical

        ! Print out final single particle step
        !$omp   critical        (print_out_final_sp_ds)
        write(*,'(2a)',advance='NO') '**info(spmc_ds_maximise):following downhill simplex minimisation, ',    &
                                    'the single particle parameters are: '
        call sp_print_info(self%sp)
        sp_start = self%sp - sp_start
        write(*,'(a)',advance='NO') '**info(spmc_ds_maximise): The single-particle parameters changed by this much: '
        call sp_print_info(sp_start)
        !$omp   end critical    (print_out_final_sp_ds)
    end subroutine spmc_ds_maximise

    !>  \brief  Return the first step for the ds minimiser
    function spmc_ds_get_first_step(self,randomise_step) result(first_step)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(sp_model_comparison),              intent(in)  ::  self
        logical,                    optional,   intent(in)  ::  randomise_step  !<  Multiply the step by a random number between -1 and 1
        ! Result
        real(kind=8),   allocatable ::  first_step(:)
        ! Private variables
        integer ::  current_param
        integer ::  current_value
        integer ::  ierr
        logical ::  rrandomise_step
        real    ::  my_random_number
        ! Start work

        !
        rrandomise_step = .false.
        if (present(randomise_step)) rrandomise_step = randomise_step

        ! Allocate memory
        allocate(first_step(count(self%refine_params)),stat=ierr)
        if (ierr .ne. 0) call terminate('Memory allocation failed in SPMC_DS_GET_FIRST_STEP')

        !
        current_value = 0
        do current_param = 1,9
            if (self%refine_params(current_param)) then
                current_value = current_value + 1
                first_step(current_value) = self%search_half_range(current_param) / self%param_scale(current_param)
                if (rrandomise_step) then
                    call random_number(my_random_number)
                    first_step(current_value) = first_step(current_value) * (my_random_number-0.5)*2.0
                endif
            endif
        enddo

        ! Sanity
        if (current_value .ne. count(self%refine_params)) then
            call terminate('Failed sanity check in SPMC_DS_GET_FIRST_STEP')
        endif
    end function spmc_ds_get_first_step

    !>  \brief  Wrapper to be called by the fgsl ds minimiser
    !!  The values and parameters are passed as c pointers so that this subroutine can be called by the (f)GSL.
    function spmc_eval_wrap(values,params) bind(c)
        use, intrinsic :: iso_c_binding
        use runtime_parameters, only : terminate
        use fgsl, only : fgsl_double, fgsl_vector, fgsl_obj_c_ptr, fgsl_vector_align, fgsl_int, fgsl_success
        implicit none
        ! Arguments
        type(c_ptr),    value   ::  values              !<  c pointer to current values to try out
        type(c_ptr),    value   ::  params              !<  c pointer to a spmc object
        ! Result
        real(kind=c_double)     ::  spmc_eval_wrap
        ! Private variables
        type(sp_model_comparison),  pointer     ::  spmc
        real(fgsl_double),          pointer     ::  val_fgsl(:)
        real(kind=8),               allocatable ::  val(:)
        type(fgsl_vector)                       ::  val_vec
        integer                                 ::  expected_num_values
        integer(fgsl_int)                       ::  status
        integer                                 ::  ierr
        ! Start work

        ! Unpack the spmc object
        call c_f_pointer(params,spmc)

        ! Work out how many dimensions we're minimising
        expected_num_values = count(spmc%refine_params)

        ! Unpack the values we are going to try
        call fgsl_obj_c_ptr(val_vec,values)
        status = fgsl_vector_align(val_fgsl,val_vec)
        if (status .ne. fgsl_success) call terminate('FGSL_VECTOR_ALIGN failed in SPMC_EVAL_WRAP')

        ! Check we got the expected number of variables
        if (size(val_fgsl) .ne. expected_num_values) then
            write(*,'(a,2i5)') '**ERROR(spmc_eval_wrap): did not get expected number of variables ', &
                                size(val_fgsl), expected_num_values
            call terminate('Fatal error in SPMC_EVAL_WRAP')
        endif

        ! Convert from fgsl_double to real(KIND=8)
        allocate(val(expected_num_values),stat=IERR)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(spmc_eval_wrap): allocation of real values array failed with error', ierr
            call terminate('Fatal error in SPMC_EVAL_WRAP')
        endif
        val = val_fgsl

        ! Modify the single particle object to have the desired values
        call spmc_set_values(spmc,val)

        ! Now call the actual scoring function
        spmc_eval_wrap = - spmc_eval(spmc)

        !write(*,*) '**debug(spmc_eval_wrap): ', spmc_eval_wrap, val

    end function spmc_eval_wrap

    !>  \brief  Wrapper to be called by the fgsl ds minimiser
    !!  The values and parameters are passed as c pointers so that this subroutine can be called by the (f)GSL.
    function spmc_eval_wrap_nofgsl(values,params) bind(c)
        use, intrinsic :: iso_c_binding
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(c_ptr),    value   ::  values              !<  c pointer to current values to try out
        type(c_ptr),    value   ::  params              !<  c pointer to a spmc object
        ! Result
        real(kind=c_double)     ::  spmc_eval_wrap_nofgsl
        ! Private variables
        type(sp_model_comparison),  pointer     ::  spmc
        real(kind=8),               pointer     ::  val(:)
        integer                                 ::  expected_num_values
        integer                                 ::  ierr
        ! Start work

        ! Unpack the spmc object
        call c_f_pointer(params,spmc)

        ! Work out how many dimensions we're minimising
        expected_num_values = count(spmc%refine_params)

        ! Unpack the values we are going to try
        call c_f_pointer(values,val,[expected_num_values])

        ! Modify the single particle object to have the desired values
        call spmc_set_values(spmc,val)

        ! Now call the actual scoring function
        spmc_eval_wrap_nofgsl = - spmc_eval(spmc)

    end function spmc_eval_wrap_nofgsl


    !>  \brief  Change the values of sp parameters in a spmc object
    subroutine spmc_set_values(self,values)
        use ctfs
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        class(sp_model_comparison),                 intent(inout)   ::  self
        real(kind=8),                               intent(in)      ::  values(:)
        ! Private variables
        integer ::  current_param, current_value
        ! Start work
        if (size(values) .ne. count(self%refine_params)) then
            write(*,'(a)') '**ERROR(spmc_set_values): number of values and number of refinement parameters ', &
                            size(values), count(self%refine_params)
            call terminate('Fatal error in SPMC_SET_VALUES')
        endif

        ! Loop over parameters
        current_value = 0
        do current_param = 1,9
            if (self%refine_params(current_param)) then
                current_value = current_value + 1
                select case (current_param)
                    case(1)
                        self%sp%shift_x         =   values(current_value) * self%param_scale(current_param)
                    case(2)
                        self%sp%shift_y         =   values(current_value) * self%param_scale(current_param)
                    case(3)
                        call ctf_at_z(self%sp%ctf,values(current_value) * self%param_scale(current_param))
                    case(4)
                        self%sp%angles%phi      =   values(current_value) * self%param_scale(current_param)
                    case(5)
                        self%sp%angles%theta    =   values(current_value) * self%param_scale(current_param)
                    case(6)
                        self%sp%angles%psi      =   values(current_value) * self%param_scale(current_param)
                    case(7)
                        self%sp%subunit_num     =   values(current_value) * self%param_scale(current_param)
                    case(8)
                        self%sp%helix%rise      =   values(current_value) * self%param_scale(current_param)
                    case(9)
                        self%sp%helix%twist     =   values(current_value) * self%param_scale(current_param)
                end select
            endif
        end do

        ! Sanity check
        if (current_value .ne. size(values)) then
            write(*,'(a)') '**ERROR(spmc_set_values): likely programming error'
            call terminate('Fatal error in SPMC_SET_VALUES')
        endif
    end subroutine spmc_set_values

    !>  \brief  Read out the values of ps parameters in a spmc object
    subroutine spmc_get_half_ranges(self,half_ranges)
        use ctfs
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        class(sp_model_comparison),                     intent(inout)   ::  self
        real(kind=8),                   allocatable,    intent(out)     ::  half_ranges(:)
        ! Private variables
        integer ::  current_param, current_value
        integer ::  ierr
        ! Start work

        ! Allocate memory to the values array
        allocate(half_ranges(count(self%refine_params)),stat=IERR)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(spmc_get_values): memory allocation failed with ierr = ', ierr
            call terminate('Fatal error in SPMC_GET_half_ranges')
        endif

        ! Loop over parameters
        current_value = 0
        do current_param = 1,9
            if (self%refine_params(current_param)) then
                current_value = current_value + 1
                half_ranges(current_value) = self%search_half_range(current_param)
            endif
        enddo

    end subroutine spmc_get_half_ranges

    !>  \brief
    subroutine spmc_get_step_sizes(self,step_sizes)
        use ctfs
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        class(sp_model_comparison),                     intent(inout)   ::  self
        real(kind=8),                   allocatable,    intent(out)     ::  step_sizes(:)
        ! Private variables
        integer ::  current_param, current_value
        integer ::  ierr
        ! Start work

        ! Allocate memory to the values array
        allocate(step_sizes(count(self%refine_params)),stat=IERR)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(spmc_get_values): memory allocation failed with ierr = ', ierr
            call terminate('Fatal error in SPMC_GET_step_sizes')
        endif

        ! Loop over parameters
        current_value = 0
        do current_param = 1,9
            if (self%refine_params(current_param)) then
                current_value = current_value + 1
                step_sizes(current_value) = self%search_sampling(current_param)
            endif
        enddo

    end subroutine spmc_get_step_sizes


    !>  \brief  Read out the values of ps parameters in a spmc object
    subroutine spmc_get_values(self,values)
        use ctfs
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        class(sp_model_comparison),                     intent(inout)   ::  self
        real(kind=8),                   allocatable,    intent(out)     ::  values(:)
        ! Private variables
        integer ::  current_param, current_value
        integer ::  ierr
        ! Start work

        ! Allocate memory to the values array
        allocate(values(count(self%refine_params)),stat=IERR)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(spmc_get_values): memory allocation failed with ierr = ', ierr
            call terminate('Fatal error in SPMC_GET_VALUES')
        endif

        ! Loop over parameters
        current_value = 0
        do current_param = 1,9
            if (self%refine_params(current_param)) then
                current_value = current_value + 1
                select case(current_param)
                    case(1)
                        values(current_value)   =   self%sp%shift_x
                    case(2)
                        values(current_value)   =   self%sp%shift_y
                    case(3)
                        call ctf_ang_to_pix(self%sp%ctf,PIXEL_SIZE=SELF%pixel_size)
                        values(current_value)   =   z_from_ctf(self%sp%ctf)
                    case(4)
                        values(current_value)   =   self%sp%angles%phi
                    case(5)
                        values(current_value)   =   self%sp%angles%theta
                    case(6)
                        values(current_value)   =   self%sp%angles%psi
                    case(7)
                        values(current_value)   =   self%sp%subunit_num
                    case(8)
                        values(current_value)   =   self%sp%helix%rise
                    case(9)
                        values(current_value)   =   self%sp%helix%twist
                end select
                values(current_value) = values(current_value) / self%param_scale(current_param)
            endif
        enddo

        ! Sanity check
        if (current_value .ne. size(values)) then
            write(*,'(a)') '**ERROR(spmc_get_values): likely programming error'
            call terminate('Fatal error in SPMC_GET_VALUES')
        endif

    end subroutine spmc_get_values


    !>  \brief  Score the current state of the single particle model comparison object. Essentially,
    !!          this involves comparing the ft of the single particle image with the corresponding
    !!          central section of the model's ft
    real function spmc_eval(self)
        use images_comparison
        use images_ctf
        use euleranglesmod
        use single_particles, only : single_particle, sp_print_info
        use coos, only : waypoint
        use runtime_parameters, only : terminate
        use Peaks
        use Globals
#ifndef __GFORTRAN__
        use, intrinsic :: ieee_arithmetic, only : ieee_is_nan
#endif
        use images_helices, only : image_mask_helical_subunit ! debug only
        implicit none
        ! Arguments
        class(sp_model_comparison),     intent(inout)   ::  self
        ! Private variables
        logical                 ::  new_params(9)       !<  For each parameter (1-X, 2-Y, 3-Z/defocus/ctf, 4-Phi,
                                                        !!  5-Theta,6.Psi,7.SubunitNum,8.HRise,9.HTwist), whether we are dealing
                                                        !!  with new values compared to the prepared intermediate
                                                        !!  steps already in memory
        logical,    parameter   ::  debug           =   .false.
        real(kind=8)            ::  subunit_num_shift, phi_shift, z_shift
        !type(euler)             ::  angles
        type(waypoint)          ::  projection_wp
        type(Peak)              ::  ccf_peak
        ! Start work


        ! If we're doing slice extraction (rather than real space projection), check the 3d model is in Fourier space, with real-space origin at the corner
        !!$omp   critical
        !if (self%model3d%is_in_real_space) then
        !    write(*,'(2a)') '**warning(spmc_eval): model 3d was not in Fourier space - apply FT. ',  &
        !                    'Do this externally to save time if possible.'
        !    call self%model3d%ForwardFFT()
        !endif
        !if (self%model3d%object_is_centered_in_box) then
        !    write(*,'(2a)') '**warning(spmc_eval): model 3d had real-space origin at center - apply phase shift. ',  &
        !                    'Do this externally to save time if possible.'
        !    call self%model3d%SwapRealSpaceQuadrants()
        !endif
        !!$omp   end critical

        if (debug) then
            write(*,'(a)',advance='no') '**debug(spmc_eval): sp      : '
            call sp_print_info(self%sp)
            write(*,'(a)',advance='no') '**debug(spmc_eval): sp_ready: '
            call sp_print_info(self%sp_ready)
        endif


        !
        ! Check what parameters have changed
        !
        if (self%first_use) then
            ! All parameters are new
            new_params = .true.
            self%first_use = .false.
        else
            new_params = .false.
            if (self%sp%angles .ne. self%sp_ready%angles) then
                new_params(3:6) = .true.
            endif
            if (ctf_is_not_equal(self%sp%ctf,self%sp_ready%ctf)) then
                new_params(3) = .true.
            endif
            if (self%sp%shift_x .ne. self%sp_ready%shift_x .or. self%sp%shift_y .ne. self%sp_ready%shift_y) then
                new_params(1:2) = .true.
            endif
            if (self%sp%subunit_num .ne. self%sp_ready%subunit_num) then
                new_params(7) = .true.
            endif
            if (self%sp%helix%rise .ne. self%sp_ready%helix%rise) then
                new_params(8) = .true.
            endif
            if (self%sp%helix%twist .ne. self%sp_ready%helix%twist) then
                new_params(9) = .true.
            endif
        endif

        if (debug) write(*,'(a,9(l1,1x))') '**debug(spmc_eval): new params = ', new_params

        !
        ! There are some conditions under which we want to return -1.0,
        ! especially when invalid parameters are being tried by the minimiser
        !
        if (self%sp%helix%rise .le. 0.0d0) then
            spmc_eval = -1.0
        else ! non-pathological values are being tested

            !
            ! Depending on which parameters have changed, we may need to recalculate intermediate images
            !
            if (any(new_params)) then
                if (new_params(3)) then
                    ! We need to (re)calcuate the ctf image
                    if (debug) write(*,'(a)') '**debug(spmc_eval): recomputing ctf image'
                    self%ctf_image_ready%logical_dimensions = self%sp_image%logical_dimensions
                    call ctf_to_img(self%sp%ctf,self%ctf_image_ready,pixel_size=self%pixel_size, &
                                    one_at_origin=self%boost_low_frequency_contrast)
                endif
                if (any(new_params(1:2)) .or. any(new_params(3:6)) .or. new_params(7) .or. any(new_params(8:9))) then
                    ! We need to compute a new projection and a new envelope mask
                    if (debug) write(*,'(a)') '**debug(spmc_eval): recomputing projection and envelope mask'
                    ! Work out any changes due to the non-integer value of subunit_num
                    ! (we may have to change phi and z)
                    subunit_num_shift = self%sp%subunit_num - nint(self%sp%subunit_num)
                    phi_shift = - subunit_num_shift * self%sp%helix%twist
                    z_shift = - subunit_num_shift * self%sp%helix%rise
                    !angles = self%sp%angles
                    !angles%phi = angles%phi + phi_shift
                    ! Set up a waypoint for the projection
                    projection_wp%coo(1) = self%sp%shift_x
                    projection_wp%coo(2) = self%sp%shift_y
                    projection_wp%coo(3) = 0.0
                    projection_wp%euler  = self%sp%angles
                    projection_wp%euler%phi = projection_wp%euler%phi + phi_shift
                    projection_wp%helix  = self%sp%helix
                    allocate(projection_wp%local_subframe_shift(1,1))
                    projection_wp%local_subframe_shift = 0.0
                    ! Compute the projection
                    call projection_prepare(    self%model3d,self%proj_ready,self%sp_image%logical_dimensions,&
                                                projection_wp,self%ctf_image_ready,  &
                                                pixel_size=self%pixel_size, &
                                                proj_max_rad=self%obj_max_rad,  &
                                                interpol=self%interpol,win=self%interpol_window,  &
                                                env_mask=self%env_mask_ready,real_space_proj=self%real_space_proj,&
                                                vol_shift_z = z_shift,rect_mask_width=2.*self%obj_max_rad_cyl, &
                                                segment_with_helical_subunits=any(self%refine_params(8:9)) &
                                                                                .or. self%project_individual_subunits, &
                                                helix_max_rad=self%obj_max_rad_cyl, &
                                                helix=self%model_helix, &
                                                volume_oversampling=self%model3d_oversampling &
                                                )
                endif
                if (new_params(3)) then
                    ! We need to (re) apply the ctf to the image, and remask it with the envelope mask
                    if (debug) write(*,'(a)') '**debug(spmc_eval): apply ctf & mask to the sp image'
                    self%sp_image_ready = self%sp_image
                    call image_prepare( self%sp_image_ready,self%sp_image_tmp,self%ctf_image_ready,self%obj_max_rad,    &
                                        apply_ctf=.true., &
                                        env_mask=self%env_mask_ready)
                endif

                ! Now that we've recalculated intermediate steps, remember we've done so
                self%sp_ready = self%sp

                !
                !call self%sp_image_ready%WriteToDisk('dbg_sp_image_ready.mrc')
                !call self%proj_ready%WriteToDisk('dbg_proj_ready.mrc')

                ! We may need to do a CCF peak search to find X,Y shifts
                if (self%use_ccf_to_find_xy) then
                    self%sp_image_tmp = self%sp_image_ready
                    if (self%proj_ready%IsInRealSpace()) call self%proj_ready%ForwardFFT()
                    if (self%sp_image_tmp%IsInRealSpace()) call self%sp_image_tmp%ForwardFFT()
                    call self%sp_image_tmp%CalculateCrossCorrelationImageWith(self%proj_ready)
                    ccf_peak = self%sp_image_tmp%FindPeakWithParabolaFit(max_radius= &
                                                       sqrt(self%search_half_range(1)**2+self%search_half_range(2)**2))
                    !ccf_peak = self%sp_image_tmp%FindPeakWithIntegerCoordinates(max_radius= &
                    !                                   sqrt(self%search_half_range(1)**2+self%search_half_range(2)**2))
                    self%latest_ccf_xy = ccf_peak%CoOrdinates(1:2)
                    call self%sp_image_ready%PhaseShift(self%latest_ccf_xy(1),self%latest_ccf_xy(2),0.0)
                    if (debug) then
                        write(*,'(a,2(f0.2,1x))') '**debug(spmc_eval): latest xy shifts by ccf peak finding = ', self%latest_ccf_xy
                    endif
                    if (debug) call self%sp_image_tmp%WriteToDisk('dbg_ccf.mrc')
                endif

            else
                if (this_program%IsVerbose() .or. debug) then
                    write(*,'(a)') '**warning(spmc_eval): evaluating something that was already evaluated - no parameters changed!'
                endif
            endif



            !
            ! Now all we have to do is measure the similarity between proj_ready and sp_image_ready
            !
            !if (debug) write(*,'(a,i0)') '**debug(spmc_eval): about to call cc3 with rad_max = ', nint(self%sp_image_ready%logical_dimensions(1)*SELF%pixel_size/self%target_resolution)
            if (self%sp_image_ready%IsInRealSpace()) call self%sp_image_ready%ForwardFFT()
            if (self%proj_ready%IsInRealSpace()) call self%proj_ready%ForwardFFT()
            spmc_eval = cclin(  self%sp_image_ready,self%proj_ready,  &
                                rad_max=real(self%sp_image_ready%logical_dimensions(1))*self%freq_max,   &
                                rad_min=real(self%sp_image_ready%logical_dimensions(1))*self%freq_min)

        endif ! End of test for pathological cases which return -1.0

        self%sp%score = spmc_eval
        self%sp_ready%score = spmc_eval
        ! Debug
        if (debug) then
            write(*,'(a,F0.5,a)',advance='NO') '**debug(spmc_eval): score of ', spmc_eval, &
                                                ' for the following single-particle parameters: '
            call sp_print_info(self%sp)
#ifdef __GFORTRAN__
            if (isnan(spmc_eval)) then
#else
            if (ieee_is_nan(spmc_eval)) then
#endif
                call terminate('Fatal error in spmc_eval: returning NaN')
            endif
        endif
    end function spmc_eval

    !>  \brief  Deallocate work memory for the fmc object
    subroutine spmc_deallocate_work_memory(self)
        implicit none
        ! Arguments
        type(sp_model_comparison),    intent(inout)   ::  self
        ! Private variables

        ! Start work
        call self%interpol_window%DeallocateWindow()
        call self%sp_image_tmp%Deallocate()
        call self%proj_ready%Deallocate()
        call self%ctf_image_ready%Deallocate()
        call self%sp_image_ready%Deallocate()
        call self%env_mask_ready%Deallocate()
    end subroutine spmc_deallocate_work_memory

    !>  \brief  Allocate work memory for the fmc object
    subroutine spmc_allocate_work_memory(self)
        use images
        implicit none
        ! Arguments
        type(sp_model_comparison),    intent(inout)   ::  self
        ! Private variables

        ! Start work
        call spmc_deallocate_work_memory(self)
        call self%interpol_window%AllocateWindow(self%interpol%window_width,COMP=.not. self%real_space_proj,THREED=.true.)
        call self%sp_image_tmp%Allocate(self%sp_image)
        call self%proj_ready%Allocate(self%sp_image)
        call self%ctf_image_ready%Allocate(self%sp_image)
        call self%sp_image_ready%Allocate(self%sp_image)
        call self%env_mask_ready%Allocate(self%sp_image)
    end subroutine spmc_allocate_work_memory

    !>  \brief  Initialise a single particle model comparison object
    subroutine spmc_init(   self,sp,model3d,sp_image,refine_params,pixel_size, &
                            target_resolution,obj_max_rad,interpol,real_space_proj, &
                            project_individual_subunits, &
                            search_half_range_override,num_minim_runs,ds_random_start, &
                            low_resolution_cuton,high_resolution_cutoff,obj_max_rad_cyl, &
                            boost_low_frequency_contrast,model_helix, &
                            model3d_oversampling,minimiser_algorithm, &
                            search_step_size)
        use images_comparison
        use ctfs, only : ctf_pix_to_ang
        use helices, only : helical_symmetry
        use,    intrinsic   :: iso_c_binding
        use runtime_parameters, only : terminate
        use Globals
        implicit none
        ! Arguments
        type(sp_model_comparison),  target,     intent(inout)   ::  self                            !<  Single particle - model comparison object to be initialised
        type(single_particle),                  intent(in)      ::  sp                              !<  Single particle object
        type(image),                target,     intent(in)      ::  model3d                         !<  3d model to which the 2d image will be compared
        type(image),                            intent(in)      ::  sp_image                        !<  2d projection image of the single particle (presumably boxed out from a noisy micrograph)
        logical,                                intent(in)      ::  refine_params(9)                !<  Refine parameters (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi)
        real,                                   intent(in)      ::  pixel_size                      !<  Pixel size (Angstroms)
        real,                                   intent(in)      ::  target_resolution               !<  Target resolution (Angstroms)
        real,                                   intent(in)      ::  obj_max_rad                     !<  Maximum radius of the object (pixels)
        type(interpolation_scheme),             intent(in)      ::  interpol                        !<  Interpolation scheme
        logical,                    optional,   intent(in)      ::  real_space_proj                 !<  Whether to use real space projection. If .FALSE., reciprocal space slice extraction is used instead (which is faster). Defaults to .FALSE.
        logical,                    optional,   intent(in)      ::  project_individual_subunits     !<  Project individual subunits, rather than segments of helically-symmetrised volume
        real,                       optional,   intent(in)      ::  search_half_range_override(9)   !<  For each refinement dimension (1-X, 2-Y, 3-Z/defocus, 4-Phi, 5-Theta,6.Psi), 0.0 if default search half-range
                                                                                                    !!  should be used, or other value to override the default.
        integer,                    optional,   intent(in)      ::  num_minim_runs                  !<  The number of times the minimiser should be run. Only meant for dowhill simplex (ds), which can be started several times
                                                                                                    !!  from the same position but with a random first step.
        logical,                    optional,   intent(in)      ::  ds_random_start                 !<  Whether the starting setup of the ds should be randomised (see gsl documentation for gsl_multimin_fminimizer_nmsimplex2rand
                                                                                                    !!  at http://www.gnu.org/software/gsl/manual/html_node/Multimin-Algorithms-without-Derivatives.html)
        real,                       optional,   intent(in)      ::  low_resolution_cuton            !<  Lowest resolution (in Angstroms) to be used by scoring functions
        real,                       optional,   intent(in)      ::  high_resolution_cutoff          !<  Highest resolution (in Angstroms) to be used by scoring functions
        real,                       optional,   intent(in)      ::  obj_max_rad_cyl                 !<  Maximum (cylindrical) radius of the 3d object (pixels). Useful refining images of filaments.
        logical,                    optional,   intent(in)      ::  boost_low_frequency_contrast    !<  by not multiplying by the ctf near the spectral origin and until the first peak of the ctf
        type(helical_symmetry),     optional,   intent(in)      ::  model_helix                     !<  Helical symmetry which describes (and has been applied to) the model 3D
        integer,                    optional,   intent(in)      ::  model3d_oversampling            !<  The model3d is oversampled by this factor
        integer,                    optional,   intent(in)      ::  minimiser_algorithm               !<  What kind of minimiser should we use?
        real,                       optional,   intent(in)      ::  search_step_size(9)
        ! Private variables
        logical ::  bad_dimensions
        logical,    parameter   ::  debug = .false.
        logical,    parameter   ::  allow_use_of_ccf_for_xy = .false.
        ! Start work

        ! The 3d model and the single particle image are expected to have the same dimensions
        self%model3d_oversampling = 1
        if (present(model3d_oversampling)) self%model3d_oversampling = model3d_oversampling
        bad_dimensions = .false.
        if (.not. model3d%IsSquare() .or. .not. sp_image%IsSquare()) then
            bad_dimensions = .true.
        endif
        if (model3d%logical_dimensions(1) .ne. sp_image%logical_dimensions(1)*self%model3d_oversampling) then
            bad_dimensions = .true.
        endif
        if (bad_dimensions) then
            write(*,'(a)') '**ERROR(spmc_init): image and 3d model need to be square and of same dimensions, but they are not.'
            write(*,'(a,3(i5,1x))') '**ERROR(spmc_init): image dimensions = ', sp_image%logical_dimensions
            write(*,'(a,3(i5,1x))') '**ERROR(spmc_init): model dimensions = ', model3d%logical_dimensions
            call terminate('Fatal error in SPMC_INIT')
        endif

        self%sp                     =   sp
        self%model3d                =   model3d
        self%sp_image               =   sp_image
        call self%sp_image_ready%Reset()
        self%refine_params          =   refine_params
        self%pixel_size             =   pixel_size
        self%obj_max_rad            =   obj_max_rad
        self%first_use              =   .true.
        ! Min and max resolutions
        self%freq_max               =   0.5
        self%freq_min               =   0.0
        if (present(low_resolution_cuton))      self%freq_min = pixel_size / low_resolution_cuton
        if (present(high_resolution_cutoff))    self%freq_max = pixel_size / high_resolution_cutoff
        ! Target resolution
        self%target_resolution      =   target_resolution

        ! Work out the desired accuracy of the search, as well as the search half-range and sampling (we need to the wavelength in angstroms)
        if (self%sp%ctf%pix_units) call ctf_pix_to_ang(self%sp%ctf,self%pixel_size)
        self%search_accuracy        =   accuracy_from_resolution(   self%target_resolution,self%obj_max_rad*self%pixel_size,   &
                                                                    self%model3d%logical_dimensions(1),    &
                                                                    self%sp%ctf%wl,self%pixel_size,real(self%sp%helix%rise), &
                                                                    akv_is_wl=.true.)

        self%search_half_range      =   self%search_accuracy * 2.
        self%search_sampling        =   self%search_accuracy / 4.

        ! If the user supplied overrides, apply them
        if (present(search_half_range_override)) then
            where (search_half_range_override .ne. 0.0)
                self%search_half_range = search_half_range_override
            end where
        endif

        ! For subunit_num, it doesn't make sense to look beyond +- 0.5 subunits
        !self%search_half_range(7)   =   min(self%search_half_range(7),0.5)
        if (this_program%IsVerbose() .or. debug) then
            write(*,'(a)') '**warning(spmc_init): setting half-range for subunit_num to 0.5'
        endif
        self%search_half_range(7)   =   0.5

        ! Step size
        if (present(search_step_size)) self%search_sampling = search_step_size

        ! Work out the scaling of the various parameters
        self%param_scale(1:6)       =   param_scaling_from_accuracy(self%search_accuracy(1:6),self%obj_max_rad)
        if (this_program%IsVerbose() .or. debug) then
            write(*,'(a)') '**warning(spmc_init): setting scale for subunit_num to 1.0'
        endif
        self%param_scale(7)         =   1.0

        !>  \todo   Set this properly
        self%param_scale(8:9)       =   1.0     ! TODO

        ! Interpolation scheme
        self%interpol               =   interpol

        ! Real space projection?
        if (present(real_space_proj)) then
            self%real_space_proj    =   real_space_proj
        else
            self%real_space_proj    =   .false.
        endif

        ! Project individual subunits?
        self%project_individual_subunits = .false.
        if (present(project_individual_subunits)) self%project_individual_subunits = project_individual_subunits

        ! Deal with ds options for random & multiple starting conditions
        if (present(num_minim_runs)) self%num_minim_runs = num_minim_runs
        if (present(ds_random_start)) self%ds_random_start = ds_random_start

        !
        if (present(obj_max_rad_cyl)) self%obj_max_rad_cyl = obj_max_rad_cyl

        !
        self%boost_low_frequency_contrast = .false.
        if (present(boost_low_frequency_contrast)) self%boost_low_frequency_contrast = boost_low_frequency_contrast

        ! Helix
        if (present(model_helix)) self%model_helix = model_helix

        ! Minimiser type
        if (present(minimiser_algorithm)) then
            self%minimiser_algorithm = minimiser_algorithm
        endif

        ! Use ccf peak search for xy search
        self%use_ccf_to_find_xy = .false.
        if (allow_use_of_ccf_for_xy) then
            !write(*,'(a)') '**WARNING(spmc_init): As of 160426, ccf for xy is broken'
            call this_program%TerminateWithFatalError('spmc_init','As of 160426, ccf for xy is broken')
            ! More precisely, it gives significantly worse results in testing on synthetic images
            if (all(self%refine_params(1:2))) then
                self%use_ccf_to_find_xy = .true.
                self%refine_params(1:2) = .false.
            endif
        endif


        ! Allocate work area memory
        call spmc_allocate_work_memory(self)

        ! Set up the c pointer to itself
        self%cptr                   =   c_loc(self)
    end subroutine spmc_init

    !>  \brief  Print out information describing the single particle model comparison object
    subroutine spmc_print_info(self)
        use minimisers, only : brute_force, downhill_simplex
        implicit none
        ! Arguments
        class(sp_model_comparison), intent(in)  ::  self        !<  Single particle - model comparison object
        ! Private variables
        character(len=2),    parameter   ::  text_width  =   '35'
        ! Start work

        !$omp   critical        (print_out_spmc_info)

        ! Print header line
        write(*,'(/a)') 'Single-particle-model comparison parameters'
        ! Print information
        write(*,'(A'//text_width//',1x,F7.3)')       'Pixel size (Angstroems)',          self%pixel_size
        write(*,'(A'//text_width//',1x,F7.3)')       'Target resolution (A)',            self%target_resolution
        write(*,'(A'//text_width//',1x,F7.3)')       'Minimum frequency (1/pixel)',      self%freq_min
        write(*,'(A'//text_width//',1x,F7.3)')       'Maximum frequency (1/pixel)',      self%freq_max
        write(*,'(A'//text_width//',1x,l1)')         'Individual subunit projection?',   self%project_individual_subunits
        write(*,'(A'//text_width//',1x,l1)')         'Real-space projection?',           self%real_space_proj
        write(*,'(A'//text_width//',1x,l1)',advance='no')         'Minimiser type'
        select case (self%minimiser_algorithm)
            case (downhill_simplex)
                write(*,'(a)') 'downhill simplex'
            case (brute_force)
                write(*,'(a)') 'brute force'
            case default
                write(*,'(a,i0)') 'unknown: ', self%minimiser_algorithm
        end select

        write(*,'(A'//text_width//',1x,9(a8,1x))')   'Search paramameters',              'X','Y','Z/def','Phi','Theta','Psi',&
                                                                                       'Subunit#','Rise','Twist'
        write(*,'(A'//text_width//',1x,9(l8,1x))')   'Parameter to be refined?',         self%refine_params
        write(*,'(A'//text_width//',1x,9(F8.3,1x))') 'Search accuracy',                  self%search_accuracy
        write(*,'(A'//text_width//',1x,9(F8.3,1x))') 'Search half-range',                self%search_half_range
        write(*,'(A'//text_width//',1x,9(F8.3,1x))') 'Search sampling',                  self%search_sampling
        write(*,'(A'//text_width//',1x,9(F8.3,1x))') 'Parameter scales',                 self%param_scale
        write(*,'(A'//text_width//',1x,l1)')         'Randomised ds start?',             self%ds_random_start
        write(*,'(A'//text_width//',1x,i0)')         'Number of (ds) minimisation runs', self%num_minim_runs
        if (self%obj_max_rad_cyl .ne. 0.0) then
            write(*,'(A'//text_width//',1x,F7.3)')       'Maximum (cylindrical) radius (pixel)',      self%obj_max_rad_cyl
        endif
        write(*,'(A'//text_width//',1x,l1)')         'Boost low frequencies?',             self%boost_low_frequency_contrast
        write(*,'(A'//text_width//',1x,l1)')         'Refine X and Y using CCF?',          self%use_ccf_to_find_xy

        ! Add blank line
        write(*,'(a)') ' '

        !$omp   end critical    (print_out_spmc_info)
    end subroutine spmc_print_info


end module sp_model_comparisons


