!>  \brief  Module gathering spline-fitting / interpolating routines
!!
!! For details, see file "Splines.pdf" in the documents folder.
!!  \todo   Allocate fgsl objects necessary for solving splines directly within the spline object so that we don't have allocation and deallocation everytime we solve splines (also, spline solving can then become pure subroutine)
module Splines

    use fgsl

    implicit none

    private

    public   :: interpol_spline



    ! Spline object based on the gnu scientific library (gsl) and its Fortran interface fgsl
    type gsl_spline
        private
        type(fgsl_interp_accel)                 ::  acc
        type(fgsl_spline)                       ::  spline
        integer                                 ::  status
        integer                                 ::  num_points
        real(fgsl_double), allocatable, public  ::  x(:),y(:)
    end type gsl_spline

    ! Homegrown spline object for interpolation, built following
    ! http://online.redwoods.cc.ca.us/instruct/darnold/laproj/Fall98/SkyMeg/Proj.PDF
    type interpol_spline
        private
        integer                             ::  n   = 0                         !< number of points
        integer                             ::  k   = 3                         !< degree of spline (assumed to be 3, for cubic)
        real(kind=8), allocatable,  public  ::  x(:), y(:), dy(:), ddy(:)       !< ordinates, values and first and seconds derivatives
        real(kind=8), allocatable,  public  ::  w(:,:)                          !< weights - first dim is point number, second is degree of the weight. So a cubic polynomial is y(:) = w(:,1)*X(:)**3 + w(:,2)*X(:)**2 + w(:,3)*X(:) + w(:,4)
        logical                             ::  natural = .true.                !< Spline is natural, that is ddy(1) = ddy(n) = 0
        logical                             ::  parabolic = .false.             !< Spline has parabolic runout, as in ddy(1) = ddy(2) and ddy(n) = ddy(N-1)
        logical                             ::  cubic = .false.                 !< Spline has cubic runout, as in ddy(1) = 2*DDY(2)-DDY(3)
        logical                             ::  known_d  = .false.              !< First derivatives are known at all knots
        logical                             ::  linear = .false.                !< Not really a spline, just a straight line (not tested)
        logical                             ::  in_memory = .false.             !< Whether spline is allocated
        integer                             ::  last_segment_used   =   1       !< Remember in which segment the spline was last evaluated
        real(kind=8), allocatable           ::  m(:,:)                          !< Matrix for linear equations (spline%N-2,spline%N-2)
        real(kind=8), allocatable           ::  ddv(:)                          !< second derivative vector (unknown)
        real(kind=8), allocatable           ::  rhs(:)                          !< right-hand-side of equation when solving the spline
        ! Commenting all TBPs for now
        contains
            procedure,  public  ::  SetType
            procedure,  public  ::  SetWhetherFirstDerivativesAreKnown
            procedure,  public  ::  Allocate
            procedure,  public  ::  Deallocate
            procedure           ::  IsAllocated
            final               ::  Destroy
            procedure           ::  interpol_spline_assign
            generic,    public  ::  assignment(=) => interpol_spline_assign
            procedure,  public  ::  PrintInfo => interpol_spline_print_info
            procedure           ::  Solve
            !
            procedure,  public  ::  FirstDerivativesAreKnown
            procedure,  public  ::  NumberOfKnots
            procedure,  public  ::  Degree
            !
            procedure,  public  ::  IsMonotonic
            procedure,  public  ::  FindSegment
            procedure,  public  ::  FindRoot
            procedure,  public  ::  FindRootGivenSegment
            procedure,  public  ::  FindRootsOfMultiple
            procedure,  public  ::  Evaluate
            procedure,  public  ::  EvaluateDerivative
            procedure,  public  ::  EvaluateSecondDerivative
            procedure,  public  ::  EvaluateThirdDerivative
            procedure,  public  ::  HornerEvaluateAll
    end type interpol_spline

contains

!>  \brief  Are the values of the first derivatives known at all knots?
pure logical function FirstDerivativesAreKnown(self)
    class(interpol_spline), intent(in)  ::  self
    FirstDerivativesAreKnown = self%known_d
end function FirstDerivativesAreKnown

!>  \brief  Return the number of knots
pure integer function NumberOfKnots(self)
    class(interpol_spline), intent(in)  ::  self
    NumberOfKnots = self%n
end function NumberOfKnots

!>  \brief  Return the degree
pure integer function Degree(self)
    class(interpol_spline), intent(in)  ::  self
    Degree = self%k
end function Degree

!>  \brief  Override the known_d flag
subroutine SetWhetherFirstDerivativesAreKnown(self,first_derivatives_are_known)
    class(interpol_spline), intent(inout)   ::  self
    logical,                intent(in)      ::  first_derivatives_are_known
    self%known_d = first_derivatives_are_known
end subroutine SetWhetherFirstDerivativesAreKnown

!>	\brief	Return the value of the spline's third derivative at x
pure function EvaluateThirdDerivative(self,x) result(dddy)
    ! Arguments
    class(interpol_spline), intent(in)  ::  self
    real(kind=8),           intent(in)  ::  x
    ! Result
    real(kind=8)    ::  dddy
    ! Private variables
    integer                 ::  seg
    ! Start work


    ! Find which function we need to use
    if (x .lt. self%x(1)) then
		seg = 1
	else if (x .gt. self%x(self%n)) then
        seg = self%n-1
    else
        seg = self%FindSegment(x)
    endif

    ! Now calculate the value
    dddy = 6.*self%w(seg,1)
end function EvaluateThirdDerivative

!>  \brief  Return the value of the spline's second derivative at X.
pure function EvaluateSecondDerivative(self,x) result(ddy)
    ! Arguments
    class(interpol_spline), intent(in)  ::  self
    real(kind=8),           intent(in)  ::  x
    ! Result
    real(kind=8)    ::  ddy
    ! Private variables
    integer                 ::  seg
    ! Start work

    ! Find which function we need to use
    if (x .lt. self%x(1)) then
        seg = 1
    else if (x .gt. self%x(self%n)) then
        seg = self%n-1
    else
        seg = self%FindSegment(x)
    endif

    ! Now calculate the value
    ddy = 0.0D0
    ddy = 6.*self%w(seg,1)*(x-self%x(seg)) + 2.*self%w(seg,2)
    ! This should be rewritten to work with any degree of spline (see the other eval function)
end function EvaluateSecondDerivative

!>  \brief  Return the value of the spline's first derivative at X. Can do extrapolation too.
pure function EvaluateDerivative(self,x) result(dy)
    ! Arguments
    class(interpol_spline), intent(in)  ::  self
    real(kind=8),           intent(in)  ::  x
    ! Result
    real(kind=8)    ::  dy
    ! Private variables
    integer                 ::  seg
    ! Start work

    if (self%n .ge. 3) then
        ! Find which function we need to use
        if (x .lt. self%x(1)) then
            seg = 1
        else if (x .gt. self%x(self%n)) then
            seg = self%n-1
        else
            seg = self%FindSegment(x)
        endif

        ! Now calculate the value
        dy = 0.0d0
        dy = 3.*self%w(seg,1)*(x-self%x(seg))**2 + 2.*self%w(seg,2)*(x-self%x(seg)) + self%w(seg,3)
        ! This should be rewritten to work with any degree of spline (see the other eval function)
    else if (self%n .eq. 2) then
        ! Linear
        dy = self%w(1,3)
    else
        dy = 0.0d0
    endif
end function EvaluateDerivative

!>  \brief  Return the value of the spline at X. Can do extrapolation too.
pure subroutine Evaluate(spline,x,y,first_and_last_unknown)
    ! Arguments
    class(interpol_spline),     intent(inout)   ::  spline
    real(kind=8),               intent(in)      ::  x
    real(kind=8),               intent(out)     ::  y                       !<  Result of the evaluation
    logical,        optional,   intent(in)      ::  first_and_last_unknown  !<  If .TRUE., the first and last segments' polynomial are unknown and we reuse the second and penultimate's segments polynomials instead.
    ! Private variables
    integer                 ::  sseg                            ! the relevant segment of the spline
    integer                 ::  deg                             ! degree
    logical, parameter      ::  horner          =   .true.      ! Whether to use Horner's method for evaluating the polynomial (default) or the naive method
    logical                 ::  ffirst_and_last_unknown
    ! Start work

    ! Do we have well defined first/last segments?
    ffirst_and_last_unknown = .false.
    if (present(first_and_last_unknown)) ffirst_and_last_unknown = first_and_last_unknown


    ! If x is before the first point (or after the last point) of the interpolation, we will extrapolate using the first polynomial
    if (x .lt. spline%x(1)) then
        sseg = 1
    else if (x .gt. spline%x(spline%n)) then
        sseg = spline%n-1
    else
        ! Find which function we need to use
        sseg = spline%FindSegment(x)
        spline%last_segment_used = sseg
    endif

    ! If we don't know the first and last segment properly
    if (ffirst_and_last_unknown) then
        !print *, '**debug(interpol_spline_eval): spline%n =', spline%n
        if (sseg .eq. 1) sseg = 2
        if (sseg .eq. spline%n-1) sseg = spline%n-2
    endif

    ! Now calculate the value
    y = 0.0D0
    if (horner) then
        y = spline%w(sseg,1)
        do deg = 1, spline%k
            y = y*(x-spline%x(sseg)) + spline%w(sseg,deg+1)
        enddo
    else
        do deg = 1, spline%k+1
            y = y + spline%w(sseg,deg) * (x-spline%x(sseg))**dble(spline%k-deg+1)
        enddo
    endif
end subroutine Evaluate

!>  \brief  Return the value of the spline at x, as well as its first and second derivatives. Uses Horner's method.
!!
!!  Algorithm from
!!  Horner's Method for Evaluating and Deflating Polynomials
!!  C. Sidney Burrus, James W. Fox, Gary A. Sitton, and Sven Treitel
!!  Rice University
!!  November 26, 2003
pure subroutine HornerEvaluateAll(spline,x,y,d,dd,first_and_last_unknown)
    ! Arguments
    class(interpol_spline),     intent(inout)   ::  spline
    real(kind=8),               intent(in)      ::  x
    real(kind=8),               intent(out)     ::  y                       !<  Result of the evaluation
    real(kind=8),               intent(out)     ::  d                       !<  First derivative
    real(kind=8),               intent(out)     ::  dd                      !<  Second derivative
    logical,        optional,   intent(in)      ::  first_and_last_unknown  !<  If .TRUE., the first and last segments' polynomial are unknown and we reuse the second and penultimate's segments polynomials instead.
    ! Private variables
    integer                 ::  seg                             ! the relevant segment of the spline
    integer                 ::  deg                             ! degree
    logical                 ::  ffirst_and_last_unknown
    logical,    parameter   ::  debug           =   .false.
    real(kind=8)            ::  xx
    ! Start work

    ! Do we have well defined first/last segments?
    ffirst_and_last_unknown = .false.
    if (present(first_and_last_unknown)) ffirst_and_last_unknown = first_and_last_unknown

    ! If x is before the first point (or after the last point) of the interpolation, we will extrapolate using the first polynomial
    if (x .lt. spline%x(1)) then
        seg = 1
    else if (x .gt. spline%x(spline%n)) then
        seg = spline%n-1
    else
        ! find which function we need to use
        seg = spline%FindSegment(x)
        spline%last_segment_used = seg
    endif

    ! If we don't know the first and last segment properly
    if (ffirst_and_last_unknown) then
        if (seg .eq. 1) seg = 2
        if (seg .eq. spline%n-1) seg = spline%n-2
    endif

    ! Now calculate the value, the derivative and the second derivative
    y  = 0.0D0
    d  = 0.0D0
    dd = 0.0D0
    xx = x-spline%x(seg)
    do deg = 1, spline%k-1
        y  = xx*y  + spline%w(seg,deg)
        d  = xx*d  + y
        dd = xx*dd + d
    enddo
    y  = xx*y + spline%w(seg,spline%k)
    d  = xx*d + y
    y  = xx*y + spline%w(seg,spline%k+1)
    dd = 2*dd

end subroutine HornerEvaluateAll


!>  \brief  Find the roots x of the spline when its value is an integer multiple of y
function FindRootsOfMultiple(self,y,y_offset) result(x)
    use runtime_parameters, only : terminate
    ! Arguments
    class(interpol_spline),             intent(in)  ::  self        !<  Interpolating spline
    real(kind=8),                       intent(in)  ::  y           !<  Value of the spline. The roots of the spline when its value is a multiple of y will be returned
    real(kind=8),           optional,   intent(in)  ::  y_offset    !<  If present, the roots of the spline when its value is y_offset + n*Y will be returned.
    ! Result
    real(kind=8),           allocatable ::  x(:)
    ! Private variables
    !real(KIND=8)   ::  min_x,max_x
    real(kind=8)    ::  min_y,max_y
    real(kind=8)    ::  current_y
    integer         ::  number_of_roots, current_root, current_y_mul
    integer         ::  min_y_mul, max_y_mul
    integer         ::  ierr
    real(kind=8)    ::  yy_offset
    ! Start work

    ! y offset
    if (present(y_offset)) then
        yy_offset = y_offset
    else
        yy_offset = 0.0D0
    endif

    ! For simplification, we only treat the case where the spline is monotonic
    if (.not. (interpol_spline_is_monotonic_increasing(self) .or. interpol_spline_is_monotonic_decreasing(self))) then
        write(*,'(a)') '**ERROR(interpol_spline_find_roots_of_multiple): spline is not monotonic increasing or decreasing'
        write(*,'(a,2l1)') '**ERROR(interpol_spline_find_roots_of_multiple):  all(self%dy .GE. 0.0D0), all(self%ddy .GE. 0.0D0): ',&
                            all(self%dy .ge. 0.0D0), all(self%ddy .ge. 0.0D0)
        call interpol_spline_print_info(self)
        call terminate('Fatal error in INTERPOL_SPLINE_FIND_ROOTS_OF_MULTIPLE')
    endif

    ! Find the minimum and maximum values of the spline
    min_y = min(self%y(1),self%y(self%n))
    max_y = max(self%y(1),self%y(self%n))

    ! Find the multiple of y nearest to, but greater than, min_y
    min_y_mul = ceiling((MIN_Y-YY_OFFSET)/y)
    max_y_mul = floor((MAX_Y-YY_OFFSET)/y)

    ! How many roots?
    number_of_roots = max_y_mul - min_y_mul + 1

    ! Allocate memory for the results
    if (allocated(x)) deallocate(x)
    allocate(x(number_of_roots),stat=IERR)
    if (ierr .ne. 0) then
        call terminate('Memory allocation failed in INTERPOL_SPLINE_FIND_ROOTS_OF_MULTIPLE')
    endif

    ! Find the solutions
    do current_root = 1, number_of_roots
        current_y_mul = min_y_mul + current_root - 1
        current_y = dble(current_y_mul) * y + yy_offset
        x(current_root) = self%FindRoot(current_y)
    enddo

end function FindRootsOfMultiple

!>  \brief  Find the root x of the spline when its value is y, given the relevant segment of the spline
function FindRootGivenSegment(self,y,seg) result(x)
    use runtime_parameters, only : terminate
#ifndef __GFORTRAN__
    use, intrinsic :: ieee_arithmetic, only : ieee_is_nan
#endif
    ! Arguments
    class(interpol_spline),  intent(in)  ::  self
    real(kind=8),           intent(in)  ::  y
    integer,                intent(in)  ::  seg
    ! Result
    real(kind=8)                        ::  x
    ! Private variables
    logical,    parameter   ::  debug   =   .false.
    ! Start work

    !
    x = self%x(seg) + gsl_solve_cubic_polynomial(self%w(seg,1),self%w(seg,2),self%w(seg,3),self%w(seg,4),y, &
                                                xmin=self%x(seg)-self%x(seg),xmax=self%x(seg+1)-self%x(seg))

#ifndef __GFORTRAN__
    if (ieee_is_nan(x)) then
#else
    if (isnan(x)) then
#endif
        write(*,'(a)') '**ERROR(interpol_spline_find_root): found NaN value'
        call terminate('Fatal error in INTERPOL_SPLINE_FIND_ROOT')
    endif
end function FindRootGivenSegment

!>  \brief  Find the root x of the spline when its value is y
function FindRoot(self,y) result(x)
    use runtime_parameters, only : terminate
#ifndef __GFORTRAN__
    use, intrinsic :: ieee_arithmetic, only : ieee_is_nan
#endif
    ! Arguments
    class(interpol_spline), intent(in)  ::  self
    real(kind=8),           intent(in)  ::  y
    ! Result
    real(kind=8)                        ::  x
    ! Private variables
    integer ::  i, seg
    logical,    parameter   ::  debug   =   .false.
    ! Start work

    if (debug) then
        write(*,'(a)') '**debug(interpol_spline_find_root): spline info follows'
        call interpol_spline_print_info(self)
    endif

    ! Find which segment we're interested in
    ! While we're at it, check for special case where the solution is one of the waypoints
    seg = 0
    if (interpol_spline_is_monotonic_increasing(self)) then
        do I=1,self%n-1
            if (y .eq. self%y(i)) then
                x = self%x(i)
                return
            else if (y .eq. self%y(I+1)) then
                x = self%x(I+1)
                return
            else if (y .ge. self%y(i) .and. y .le. self%y(I+1)) then
                seg = i
                exit
            endif
        enddo
    else if (interpol_spline_is_monotonic_decreasing(self)) then
        do I=1,self%n-1
            if (y .eq. self%y(i)) then
                x = self%x(i)
                return
            else if (y .eq. self%y(I+1)) then
                x = self%x(I+1)
                return
            else if (y .le. self%y(i) .and. y .ge. self%y(I+1)) then
                seg = i
                exit
            endif
        enddo
    else
        write(*,'(a)') '**ERROR(interpol_spline_find_root): spline is not monotonic increasing or decreasing'
        write(*,'(a,2l1)') '**ERROR(interpol_spline_find_root):  all(self%dy .GE. 0.0D0), all(self%ddy .GE. 0.0D0): ', &
                            all(self%dy .ge. 0.0D0), all(self%ddy .ge. 0.0D0)
        call interpol_spline_print_info(self)
        call terminate('Fatal error in INTERPOL_SPLINE_FIND_ROOT')
    endif

    if (seg .eq. 0) then
        write(*,'(3(A,F0.3),a)')    '**ERROR(interpol_spline_find_root): value y (', y, &
                                    ') does not appear to lie within the values of the spline (from ', &
                                    self%y(1), ' to ', self%y(self%n), ')'
        call interpol_spline_print_info(self)
        call terminate('Fatal error in INTERPOL_SPLINE_FIND_ROOT')
    endif


    !
    x = self%x(seg) + gsl_solve_cubic_polynomial(self%w(seg,1),self%w(seg,2),self%w(seg,3),self%w(seg,4),y, &
                                                XMIN=SELF%x(seg)-SELF%x(seg),XMAX=SELF%x(SEG+1)-SELF%x(seg))

#ifndef __GFORTRAN__
    if (ieee_is_nan(x)) then
#else
    if (isnan(x)) then
#endif
        write(*,'(a)') '**ERROR(interpol_spline_find_root): found NaN value'
        call terminate('Fatal error in INTERPOL_SPLINE_FIND_ROOT')
    endif
end function FindRoot



!>  \brief  Return whether the spline is strictly monotonic
logical function IsMonotonic(self)
    ! Arguments
    class(interpol_spline),  intent(in)  ::  self
    ! Private variables
    real(kind=8)    ::  d_roots(2)
    integer         ::  num_d_roots
    integer         ::  seg
    ! Start work

    IsMonotonic = .true.
    ! a function is definitely monotonic if its derivative is never 0.0.
    ! So for each segment, let's check whether we can find roots for the derivative.
    do seg=1,self%n-1
        num_d_roots = gsl_solve_quadratic(3*self%w(seg,1),2*self%w(seg,2),self%w(seg,3),d_roots,  &
                                            xmin=0.0d0,xmax=self%x(seg+1)-self%x(seg))
        if (num_d_roots .gt. 0) then
			! The function may still be monotonic, but not strictly so
			IsMonotonic = .false.
		endif
    enddo

end function IsMonotonic

!>  \brief  Return whether the spline is strictly monotonic increasing
logical function interpol_spline_is_monotonic_increasing(self)
    ! Arguments
    type(interpol_spline),  intent(in)  ::  self
    ! Private variables
    ! Start work
    interpol_spline_is_monotonic_increasing = self%IsMonotonic() .and. self%y(self%n) .gt. self%y(1)
end function interpol_spline_is_monotonic_increasing

logical function interpol_spline_is_monotonic_decreasing(self)
    ! Arguments
    type(interpol_spline),  intent(in)  ::  self
    ! Private variables
    ! Start work
    interpol_spline_is_monotonic_decreasing =  self%IsMonotonic() .and. self%y(self%n) .lt. self%y(1)
end function interpol_spline_is_monotonic_decreasing


!>  \brief  Given an ordinate x, find out which segment of the spline it "belongs" to
pure function FindSegment(self,x) result(seg)
    ! Arguments
    class(interpol_spline),             intent(in)  ::  self
    real(kind=8),                       intent(in)  ::  x
    ! Result
    integer ::  seg
    ! Private variables
    integer ::  i
    ! Start work

    seg = 0
    do i=self%last_segment_used,self%n-1
        if (x .gt. self%x(i) .and. x .le. self%x(i+1)) then
            seg = i
            exit
        endif
    enddo
    ! If we still haven't found the segment, start again from the beginning
    if (seg .eq. 0) then
        do i=1,self%last_segment_used-1
            if (x .gt. self%x(i) .and. x .le. self%x(i+1)) then
                seg = i
                exit
            endif
        enddo
    endif
	if (seg .eq. 0) then
		if (x .eq. self%x(1)) seg = 1
	endif
end function FindSegment


!>  \brief  Return the value of the spline at x, but don't use weights - instead use second derivatives
function interpol_spline_eval2(spline,x) result(y)
    use runtime_parameters, only : terminate
    implicit none
    ! Arguments
    type(interpol_spline),  intent(in)  ::  spline
    real(kind=8),           intent(in)  ::  x
    real(kind=8)    ::  y
    ! Private variables
    integer ::  i
    integer ::  seg ! the relevant segment of the spline
    real(kind=8)    ::  a,b,c,d
    ! Start work
    if (spline%k .ne. 3) then
        write(*, '(a)') '**ERROR(interpol_spline_eval2): routine only works on cubic splines'
        call terminate('Fatal error in INTERPOL_SPLINE_EVAL2')
    endif
    if (all(spline%x .eq. 0.0d0)) then
        y = 0.0d0
    else
        ! Find which function we need to use
        do i=1,spline%n-1
            if (x .ge. spline%x(i) .and. x .le. spline%x(i+1)) then
                seg = i
                exit
            endif
        enddo
        ! Now calculate the value (see numerical recipes chapter on cublic spline interpolation)
        a = (spline%x(seg+1) - x) / (spline%x(seg+1)-spline%x(seg))
        b = 1.0d0 - a
        c = (a**3-a)*(spline%x(seg+1)-spline%x(seg))**2/6.0d0
        d = (b**3-b)*(spline%x(seg+1)-spline%x(seg))**2/6.0d0
        y = a*spline%y(seg) + b*spline%y(seg+1) + c*spline%ddy(seg) + d*spline%ddy(seg+1)
    endif
end function interpol_spline_eval2

!>  \brief  Set type of spline
subroutine SetType(spline,type)
    class(interpol_spline), intent(inout)   ::  spline
    character(len=*),       intent(in)      ::  type    !<  Character string - we'll only need to check first letter
    ! Private variable

    ! Start work
    spline%natural      =   .false.
    spline%parabolic    =   .false.
    spline%cubic        =   .false.
    spline%known_d      =   .false.
    select case(type(1:1))
        case('N')
            spline%natural      =   .true.
        case ('P')
            spline%parabolic    =   .true.
        case ('C')
            spline%cubic        =   .true.
        case ('K')
            spline%known_d      =   .true.
        case default
            write(*,'(2a)') '**WARNING(spline_set_type): did not understand what type you meant: ', type
            write(*,'(a)')  '**WARNING(spline_set_type): setting type to natural'
            spline%natural      =   .true.
    end select
end subroutine SetType

!>  \brief  Solve the interpolation spline, using the fgsl for the linear algebra and following the method described in
!!      pdf in the doc folder
subroutine Solve(self)
    use linear_algebra
    use runtime_parameters, only : terminate
    ! Arguments
    class(interpol_spline),  intent(inout)  ::  self  !<  Spline object
    ! Private variables

    real(kind=8)    ::  h, hm
    integer ::  i,j
    character(len=6) :: n_c
    logical, parameter :: debug = .false.
    logical, parameter :: use_lapack = .true.
    ! Start work
#ifndef skip_runtime_checks
    write(n_c,'(i6)') self%n
    ! Check spline is allocated
    if (.not. self%in_memory) then
        write(*,'(a)') '**ERROR(solve_interpol_spline): spline not allocated on entry', self%IsAllocated()
        call terminate('Fatal error in SOLVE_INTERPOL_SPLINE')
    endif
    ! Check values of x are increasing
    do i=2,self%n
        if (self%x(i) .le. self%x(i-1)) then
            write(*,'(a)') '**ERROR(solve_interpol_spline): values of x must increase'
            write(*,'('//n_c//'(F0.2,1x))') self%x
            call terminate('Fatal error in SOLVE_INTERPOL_SPLINE')
        endif
    enddo
#endif


    if (self%n .eq. 1) then
        self%w(:,:) = 0.0
        self%dy     = 0.0
        self%ddy    = 0.0
    else if (self%n .eq. 2) then
        ! Linear
        self%w(:,:) = 0.0
        self%w(1,3) = (self%y(2) - self%y(1))/(self%x(2)-self%x(1))
        self%dy(:)  = (self%y(2) - self%y(1))/(self%x(2)-self%x(1))
        self%ddy(:) = 0.0
    else

        ! Setup the matrix
        ! This is done according to the document included in the documentation.
        !
        ! The spline is solved by solving this set of linear equations:
        !   (m)*(ddv) = (rhs)
        !
        if (self%natural .or. self%parabolic .or. self%cubic) then
            do i=2,self%n-1
                h  = self%x(i+1)-self%x(i)
                hm = self%x(i)-self%x(i-1)
                do j=2,self%n-1
                    if (j .eq. i) then
                        self%m(i-1,j-1) = 2.0d0 * (hm + h)
                    else if (j .eq. i+1) then
                        self%m(i-1,j-1) = h
                    else if (j .eq. i-1) then
                        self%m(i-1,j-1) = hm
                    else
                        self%m(i-1,j-1) = 0.0d0
                    endif
                enddo
                self%rhs(i-1) = 6.0d0*((self%y(i+1)-self%y(i))/h-(self%y(i)-self%y(i-1))/hm)
            enddo
            if (self%parabolic) then
                ! 1st modification
                i  = 2
                hm = self%x(i)-self%x(i-1)
                j  = 2
                self%m(i-1,j-1) = self%m(i-1,j-1) + hm
                ! 2nd modification
                i = self%n-1
                hm = self%x(i)-self%x(i-1)
                h  = self%x(i+1)-self%x(i)
                j  = i
                self%m(i-1,j-1) = self%m(i-1,j-1) + h
            else if (self%cubic) then
                ! 1st modification
                i  = 2
                hm = self%x(i)-self%x(i-1)
                h  = self%x(i+1)-self%x(i)
                j  = 2
                self%m(i-1,j-1) = self%m(i-1,j-1) + 2.0d0*hm
                j  = 3
                self%m(i-1,j-1) = self%m(i-1,j-1) - hm
                ! 2nd modification
                i = self%n-1
                hm = self%x(i)-self%x(i-1)
                h  = self%x(i+1)-self%x(i)
                j  = i
                self%m(i-1,j-1) = self%m(i-1,j-1) + 2.0d0*h
                j  = i-1
                self%m(i-1,j-1) = self%m(i-1,j-1) - h
            endif ! End of modifications for non-natural
            ! Initialise the unknown vector
            self%ddv(1:self%n-2) = 0.0D0

#ifndef skip_runtime_checks
            ! Diagnostics
            if (debug) then
                print *, '**debug(solve_interpol_spline): self%m   = ', self%m
                print *, '**debug(solve_interpol_spline): det = ', matrix_det(self%m)
                print *, '**debug(solve_interpol_spline): self%N-2 = ', self%N-2
                print *, '**debug(solve_interpol_spline): self%rhs = ', self%rhs
            endif
#endif

            ! To solve the equations, we will use lu decomposition, as provided by gsl
            ! (Gaussian decomposition may be more efficient?)
            if (self%n .eq. 3) then
                self%ddv(1) = self%rhs(1)/self%m(1,1)
            else
                if (use_lapack) then
                    call linalg_solve_lapack(self%m,self%ddv,self%rhs)
                else
                    call linalg_solve_lu_fgsl(self%m,self%ddv,self%rhs,self%N-2)
                endif
            endif
            !print *, '**debug(solve_interpol_spline): ddv = ', ddv

            ! Copy the results (second derivatives) over into the spline object; the boundaries are worked out separately.
            do i=2,self%n-1
                self%ddy(i) = self%ddv(i-1)
            enddo
            ! The second derivative at the boundaries
            if (self%natural) then
                self%ddy(1) = 0.0D0
                self%ddy(self%n) = 0.0D0
            else if (self%parabolic) then
                self%ddy(1) = self%ddy(2)
                self%ddy(self%n) = self%ddy(self%n-1)
            else if (self%cubic) then
                self%ddy(1) = 2*self%ddy(2) - self%ddy(3)
                self%ddy(self%n) = 2*self%ddy(self%n-1) - self%ddy(self%n-2)
            endif


            ! Now we know the second derivatives at each knot (ddy), we can work out the weights
            do i=1,self%n-1
                h = self%x(i+1)-self%x(i)

                self%w(i,4) = self%y(i)
                if (self%known_d) then
                    self%w(i,3) = self%dy(i)
                else
                    self%w(i,3) = (self%y(i+1)-self%y(i))/h       &
                                - (self%ddy(i+1) + 2*self%ddy(i))*h/6.0d0
                endif
                self%w(i,2) = self%ddy(i) / 2.0d0
                self%w(i,1) = (self%ddy(i+1)-self%ddy(i))/(6.0d0*h)
            enddo

            ! Now we can work out the first derivative
            do i=1,self%n-1
                self%dy(i) = self%w(i,3)
            enddo

            ! Work out the first derivative on the last point
            self%dy(self%n) = self%EvaluateDerivative(self%x(self%n))

        else if (self%known_d) then
            if (self%k .eq. 3) then
                do i=1,self%n-1
                    h = self%x(i+1)-self%x(i)
                    self%w(i,1) = -2.0d0*(self%y(i+1)-self%y(i)-(self%dy(i+1)-self%dy(i))*h/2.0d0 - self%dy(i)*h) /h**3
                    self%w(i,2) = ( self%dy(i+1) - self%dy(i) - 3.0d0*h**2*self%w(i,1) ) / (2.0d0*h)
                    self%w(i,3) = self%dy(i)
                    self%w(i,4) = self%y(i)
                    self%ddy(i) = 2.0d0 * self%w(i,2)
                enddo
            else
#ifndef skip_runtime_checks
                write(*,'(a)') '**ERROR(solve_interpol_spline): known_d only available for cubic spline'
                call terminate('Fatal error in SOLVE_INTERPOL_SPLINE')
#endif
            endif
        else
#ifndef skip_runtime_checks
            write(*,'(a)') '**ERROR(solve_interpol_spline): unknown spline type!'
            call terminate('Fatal error in SOLVE_INTERPOL_SPLINE')
#endif
        endif !

        ! Now that we have solved the spline, we know its first derivative (added 18-Aug-2011)
        self%known_d      =   .true.
        self%natural      =   .false.
        self%parabolic    =   .false.
        self%cubic        =   .false.
        self%linear       =   .false.
    endif

    ! Diagnostics
    !call interpol_spline_print_info(spline)
end subroutine Solve

!>  \brief  Print summary info to terminal
subroutine interpol_spline_print_info(self,check_mono)
    ! Arguments
    class(interpol_spline), intent(in)  ::  self
	logical, optional,      intent(in)  ::  check_mono
    ! Private variables
    real    ::  eval2
	logical ::  ccheck_mono
	integer ::  i
    ! Start work
	ccheck_mono = .true.
	if (present(check_mono)) ccheck_mono =  check_mono
    print *
    print '(a)', '**Spline information**'
    write(*,'(a,i0)') 'Number of points: ', self%n
    write(*,'(a,5(l1,1x))') 'Natural? Parabolic? Cubic? Known D? In memory?', self%natural, self%parabolic, &
                                                                            self%cubic, self%known_d, self%in_memory
    write(*,'(a,i0)') 'Last segment used = ', self%last_segment_used
    write(*,'(a7,1x,9(a12,1x))') 'Point #','W1', 'W2', 'W3', 'W4', 'X', 'Y', 'S(x)(2)', 'DY', 'DDY'
    do I=1,self%n
        !call interpol_spline_eval(self,self%x(i),eval)
        eval2 = interpol_spline_eval2(self,self%x(i))
        write(*,'(i7,1x,4(G12.5,1x) ,2(F12.2,1x),3(G12.5,1x))') i, self%w(i,1:4), self%x(i), self%y(i),     &
                                                eval2,    &
                                                self%dy(i), self%ddy(i)
    enddo
	if (ccheck_mono) then
	    write(*,'(a,2(l1,1x))') 'Monotonic increasing? decreasing? ', interpol_spline_is_monotonic_increasing(self), &
                            interpol_spline_is_monotonic_decreasing(self)
	endif
	print *
end subroutine interpol_spline_print_info

!>  \brief  Subroutine used to overload the assignment operator
subroutine interpol_spline_assign(lhs,rhs)
    ! Arguments
    class(interpol_spline), intent(inout)   ::  lhs
    type(interpol_spline),  intent(in)      ::  rhs
    ! Private variables

    ! Start work
    if (lhs%IsAllocated() .and. ((lhs%n .ne. rhs%n) .or. (lhs%k .ne. rhs%k))) call lhs%Deallocate()
    if (.not. lhs%IsAllocated()) call lhs%Allocate(rhs%n,rhs%k)
    lhs%k           =   rhs%k
    if (allocated(rhs%x)) then
        lhs%x(:)    =   rhs%x  ! We use (:) on the lhs to avoid reallocation by the compiler
    endif
    if (allocated(rhs%y)) then
        lhs%y(:)    =   rhs%y
    endif
    if (allocated(rhs%dy)) then
        lhs%dy(:)   =   rhs%dy
    endif
    if (allocated(rhs%ddy)) then
    lhs%ddy(:)      =   rhs%ddy
    endif
    if (allocated(rhs%w)) then
    lhs%w(:,:)      =   rhs%w
    endif
    lhs%natural     =   rhs%natural
    lhs%parabolic   =   rhs%parabolic
    lhs%cubic       =   rhs%cubic
    lhs%known_d     =   rhs%known_d
    lhs%in_memory   =   rhs%in_memory
    lhs%last_segment_used   =   rhs%last_segment_used
    if (allocated(rhs%m)) then
        lhs%m       =   rhs%m
    endif
    if (allocated(rhs%ddv)) then
        lhs%ddv     =   rhs%ddv
    endif
    if (allocated(rhs%rhs)) then
        lhs%rhs     =   rhs%rhs
    endif
end subroutine interpol_spline_assign

!>  \brief  Deallocate memory from interpol_spline object
subroutine Deallocate(self)
    class(interpol_spline),  intent(inout)   ::  self
    ! Start work
    if (self%IsAllocated()) then
        deallocate(self%x,self%y,self%dy,self%ddy,self%w)
        deallocate(self%m,self%ddv,self%rhs)
		self%in_memory = .false.
    endif
end subroutine Deallocate

!impure elemental
subroutine Destroy(self)
    type(interpol_spline),  intent(inout)   ::  self
    call self%Deallocate()
end subroutine Destroy

!>  \brief  Check wither an interpol_spline object is currently allocated
pure logical function IsAllocated(self)
    class(interpol_spline),  intent(in)  ::  self
    ! Start work
    if (self%in_memory) then
        if (allocated(self%x) .and. allocated(self%y) .and. allocated(self%dy) .and. allocated(self%ddy)) then
            IsAllocated = .true.
        else
            !call terminate('Incoherent allocation status')
        endif
    else
        IsAllocated = .false.
    endif
end function IsAllocated

!>  \brief  Allocate memory to interpol_spline object
subroutine Allocate(self,num_points,degree)
    use runtime_parameters, only: terminate
    ! Arguments
    class(interpol_spline), intent(inout)   ::  self        !< Spline object
    integer,                intent(in)      ::  num_points  !< Number of points to interpolate through
    integer,                intent(in)      ::  degree      !< Degree of the splines
    ! Private variables
    integer ::  ierr
    character(len=128) :: error_message
    ! Start work
    !print *, '**debug(allocate_interpol_spline): on entry, ', allocated(self), allocated(self%x), size(self%x), allocated(self%y)
    if (degree .ne. 3) then
        write(*, '(a)') '**ERROR(allocate_interpol_spline): only cubic splines supported'
        call terminate('Fatal error in ALLOCATE_INTERPOL_SPLINE')
    endif
    ! Check whether already allocated
    if (self%IsAllocated()) then
        print '(a)', '**WARNING(allocate_interpol_spline): already allocated, will deallocate first'
        !call self%deallocate()
        call self%Deallocate()
    endif
    ! Check a sensible number of points was requested
    if (num_points .lt. 1) then
#ifdef WITH_DEBUG_SYMBOLS
        write(*,'(a,i0,a)') '**WARNING(allocate_interpol_spline): allocating spline with ', num_points, ' knots'
#endif
    else if (num_points .gt. 10000) then
        write(*,'(a,i0,a)') '**warning(allocate_interpol_spline): allocating spline with ', num_points, ' knots - this seems big'
        call terminate('Too much memory being requested. Spline::Allocate')
    endif
    ! Allocate
    allocate(   self%x(num_points),self%y(num_points),self%dy(num_points),self%ddy(num_points),         &
                self%w(num_points,degree+1),    &
                stat=ierr)
    ! Set n and k
    self%n = num_points
    self%k = degree
    !
    self%last_segment_used = 1
    ! Error case
    if (ierr .ne. 0) then
        write(*,'(a,i3,1x,i3)')  '**ERROR(allocate_interpol_spline): error in memory allocation; num_points, degree = ',     &
                                num_points, degree
        call terminate('Fatal error in ALLOCATE_INTERPOL_SPLINE')
    endif
    ! Allocate memory for local variables used when solving the spline
    if (num_points .ge. 3) then
        allocate(self%m(self%n-2,self%n-2),self%ddv(self%n-2),self%rhs(self%n-2),stat=ierr,errmsg=error_message)
        if (ierr .ne. 0) then
            write(*,'(a)') '**ERROR(allocate_interpol_spline): memory allocation failed for local variables'
            print *, error_message, num_points
            call terminate('Fatal error in ALLOCATE_INTERPOL_SPLINE')
        endif
    else
        allocate(self%m(1,1),self%ddv(1),self%rhs(1),stat=ierr)
    endif
    ! Initialise all the arrays
    self%x = 0.0D0
    self%y = 0.0D0
    self%dy = 0.0D0
    self%ddy = 0.0D0
    self%w = 0.0D0
    self%m = 0.0D0
    self%ddv = 0.0D0
    self%rhs = 0.0D0
    ! Set allocated flag
    self%in_memory = .true.

end subroutine Allocate

!>  \brief  Use the gsl to solve (find the roots of) a quadratic equation
!!
!!  The equation is: a x^2 + b x + c = 0. The number of roots is returned. The roots themselves are returned in the roots array, if there's only one root, it'll be the first element of the array.
integer function gsl_solve_quadratic(a,b,c,roots,xmin,xmax) result(num_roots)
    use runtime_parameters, only : terminate
    implicit none
    ! Arguments
    real(kind=8),               intent(in)  ::  a,b,c
    real(kind=8),               intent(out) ::  roots(2)
    real(kind=8),   optional,   intent(in)  ::  xmin,xmax       !<  The minimum and maximum values allowed for the roots to be returned
    ! Private variables
    real(kind=FGSL_DOUBLE)      ::  aa,bb,cc
    logical                     ::  roots_in_range(2)
    ! Start work
    roots_in_range = .true.
    aa = a
    bb = b
    cc = c
    num_roots = fgsl_poly_solve_quadratic(aa,bb,cc,roots(1),roots(2))



    ! Now we check whether the roots are within range
    if (num_roots .eq. 0) then
        roots_in_range = .false.
    elseif (num_roots .eq. 1) then
        roots_in_range = [.true.,.false.]
    elseif (num_roots .eq. 2) then
        roots_in_range = .true.
    endif

    ! Check the second root
    if (num_roots .eq. 2) then
        if (present(xmin)) then
            if (roots(2) .lt. xmin) roots_in_range(2) = .false.
        endif
        if (present(xmax)) then
            if (roots(2) .gt. xmax) roots_in_range(2) = .false.
        endif
    endif

    ! Check the first root
    if (num_roots .ge. 1) then
        if (present(xmin)) then
            if (roots(1) .lt. xmin) roots_in_range(1) = .false.
        endif
        if (present(xmax)) then
            if (roots(1) .gt. xmax) roots_in_range(1) = .false.
        endif
    endif

    ! The final number of roots
    num_roots = count(roots_in_range)

    ! If the only root in range is second one, move it to the first position
    if (num_roots .eq. 1 .and. roots_in_range(2)) then
        roots(1) = roots(2)
        roots_in_range = [.true.,.false.]
    endif

end function gsl_solve_quadratic

!>  \brief  Use the gsl to solve (find a root of) a cubic polynomial
!!
!!  The  equation is: a x^3 + b  x^2 + c  x + d  = y
!!  the gsl wants   :   x^3 + aa x^2 + bb x + cc = 0
function gsl_solve_cubic_polynomial(a,b,c,d,y,xmin,xmax) result(x)
    use runtime_parameters, only : terminate
#ifndef __GFORTRAN__
    use, intrinsic :: ieee_arithmetic, only : ieee_is_nan
#endif
    implicit none
    ! Arguments
    real(kind=8),   intent(in)  ::  a,b,c,d,y
    real(kind=8),   intent(in)  ::  xmin,xmax   !<  Minimum and maximum values allowed/expected for the solution
    ! Result
    real(kind=8)    ::  x
    ! Private variables
    integer ::  num_roots
    real(kind=FGSL_DOUBLE)      ::  tmpx(3)
    !complex(KIND=FGSL_DOUBLE)  ::  tmpz(3)
    real(kind=FGSL_DOUBLE)      ::  aa,bb,cc
    logical,        parameter   ::  debug   =   .false.
    logical                     ::  roots_in_range(3), roots_in_range_with_tolerance(3)
    real,           parameter   ::  range_tolerance = .05
    integer                     ::  i
    real(kind=8),   parameter   ::  small_number = 0.000000000001
    logical                     ::  cubic
    logical                     ::  print_error, print_error_tol
    ! Start work

    tmpx = -999999.9999
    print_error = .false.
    print_error_tol = .false.
    roots_in_range = .false.
    roots_in_range_with_tolerance = .false.

    if (a .ne. 0.0D0) then
        cubic = .true.
        aa = b/a
        bb = c/a
        cc = (D-Y)/a
        ! Solve the equation
        num_roots =  fgsl_poly_solve_cubic(aa,bb,cc,tmpx(1),tmpx(2),tmpx(3))
        ! Diagnostics
        if (debug) then
            write(*,'(a,5(F0.8,1x))') '**debug(gsl_solve_cubic_polynomial): a,b,c,d,y = ', a,b,c,d,y
            write(*,'(a,3(F0.8,1x))') '**debug(gsl_solve_cubic_polynomial): aa, bb, cc = ', aa, bb, cc
            write(*,'(a,3(F0.8,1x))') '**debug(gsl_solve_cubic_polynomial): tmpx = ', tmpx
            !write(*,'(a,3(2(F0.3,x),x))') '**debug(gsl_solve_cubic_polynomial): tmpz = ', tmpz
        endif
    else
        if (debug) then
            write(*,'(a)') '**debug(gsl_solve_cubic_polynomial): actually dealing with a quadratic'
        endif
        cubic = .false.
        ! If a == 0.0, we are dealing with a quadratic equation
        aa = b
        bb = c
        cc = (D-Y)
        ! Solve the equation
        num_roots =  fgsl_poly_solve_quadratic(aa,bb,cc,tmpx(1),tmpx(2))
    endif



    !
    if (num_roots .eq. 1) then
        x = tmpx(1)
    else if (.not. cubic .and. num_roots .eq. 2) then
        ! We found two solutions to the quadratic equation
        ! Are they within range?
        roots_in_range(1:2) = tmpx(1:2) .ge. xmin .and. tmpx(1:2) .le. xmax
        if (count(roots_in_range(1:2)) .eq. 1) then
            ! Only one of the two solutions were within range
            do I=1,2
                if (roots_in_range(i)) then
                    x = tmpx(i)
                endif
            enddo
        else if (count(roots_in_range(1:2)) .eq. 0) then
            ! None of the solutions were within the range, so we try to see if they're within the range with some added tolerance
            roots_in_range_with_tolerance(1:2) =        tmpx(1:2) .ge. XMIN-(XMAX-XMIN)*RANGE_TOLERANCE &
                                                .and.   tmpx(1:2) .le. XMAX+(XMAX-XMIN)*RANGE_TOLERANCE
            if(count(roots_in_range_with_tolerance(1:2)) .eq. 1) then
                ! Great, only one of them was within range with tolerance
                do I=1,2
                    if (roots_in_range_with_tolerance(i)) then
                        x = tmpx(i)
                    endif
                enddo
            else
                ! ok, we give up and printout information
                print_error_tol = .true.
            endif
        else if (count(roots_in_range(1:2)) .eq. 2) then
            print_error = .true.
        else
            write(*,'(a,i0)') '**ERROR(gsl_solve_cubic_polynomial): unexpected number of roots found: ', count(roots_in_range(1:2))
            call terminate('Fatal error in GSL_SOLVE_CUBIC_POLYNOMIAL')
        endif
    else if (cubic .and. num_roots .eq. 3) then
        ! We found three solutions to the cubic
        roots_in_range = tmpx .ge. xmin .and. tmpx .le. xmax
        ! We need to look for the correct root. Chances are only one of them will be within tolerable values of x
        if (count(roots_in_range) .eq. 1) then
            do I=1,3
                if (roots_in_range(i)) then
                    x = tmpx(i)
                endif
            enddo
        else if (count(roots_in_range) .eq. 0) then
            roots_in_range_with_tolerance =       tmpx .ge. XMIN-(XMAX-XMIN)*RANGE_TOLERANCE  &
                                            .and. tmpx .le. XMAX+(XMAX-XMIN)*RANGE_TOLERANCE
            if (count(roots_in_range_with_tolerance) .eq. 1) then
                do I=1,3
                    if (roots_in_range_with_tolerance(i)) then
                        x = tmpx(i)
                    endif
                enddo
            else
                print_error_tol = .true.
            endif
        else
            print_error = .true.
        endif
    else
        write(*,'(a,i0)') '**ERROR(gsl_solve_cubic_polynomial): unexpected number of roots: ', num_roots
        write(*,'(a,l1)') '**ERROR(gsl_solve_cubic_polynomial): cubic? ', cubic
        call terminate('Fatal error in GSL_SOLVE_CUBIC_POLYNOMIAL')
    endif

    ! Printout errors
    if (print_error) then
        write(*,'(a,i0,a,3(F0.8,1x))')  '**ERROR(gsl_solve_cubic_polynomial): ', num_roots, ' roots were returned: ', tmpx
        write(*,'(a,i0,a,2(F0.8,1x))')  '**ERROR(gsl_solve_cubic_polynomial): and ', count(roots_in_range), &
                                        ' of them were found to be within range: ', xmin, xmax
        write(*,'(a,5(F0.8,1x))') '**debug(gsl_solve_cubic_polynomial): a,b,c,d,y = ', a,b,c,d,y
        write(*,'(a,3(F0.8,1x))') '**debug(gsl_solve_cubic_polynomial): aa, bb, cc = ', aa, bb, cc
        write(*,'(a,l1)') '**ERROR(gsl_solve_cubic_polynomial): cubic? ', cubic
        call terminate('Fatal error in GSL_SOLVE_CUBIC_POLYNOMIAL')
    endif
    if (print_error_tol) then
        write(*,'(a,i0,a,3(F0.8,1x))')  '**ERROR(gsl_solve_cubic_polynomial): ', num_roots, ' roots were returned: ', tmpx
        write(*,'(a,i0,a,2(F0.8,1x))')  '**ERROR(gsl_solve_cubic_polynomial): and ', count(roots_in_range_with_tolerance), &
                                        ' of them were found to be within range with tolerance: ', &
                                        XMIN-(XMAX-XMIN)*RANGE_TOLERANCE, XMAX+(XMAX-XMIN)*RANGE_TOLERANCE
        write(*,'(a,5(F0.8,1x))') '**debug(gsl_solve_cubic_polynomial): a,b,c,d,y = ', a,b,c,d,y
        write(*,'(a,3(F0.8,1x))') '**debug(gsl_solve_cubic_polynomial): aa, bb, cc = ', aa, bb, cc
        write(*,'(a,l1)') '**ERROR(gsl_solve_cubic_polynomial): cubic? ', cubic
        call terminate('Fatal error in GSL_SOLVE_CUBIC_POLYNOMIAL')
    endif

    ! Check for NaN
#ifndef __GFORTRAN__
    if (ieee_is_nan(x)) then
#else
    if (isnan(x)) then
#endif
        write(*,'(a)') '**ERROR(gsl_solve_cubic_polynomial): found NaN'
        call terminate('Fatal error in GSL_SOLVE_CUBIC_POLYNOMIAL')
    endif
end function gsl_solve_cubic_polynomial

function gsl_interpol_spline_eval(spline,x) result(y)
    implicit none
    ! Arguments
    type(gsl_spline),   intent(inout)   ::  spline
    real,               intent(in)      ::  x
    real                                ::  y
    ! Private variables
    real(fgsl_double)   ::  xx
    ! Start work
    xx = x
    y = fgsl_spline_eval(spline%spline,xx,spline%acc)
end function gsl_interpol_spline_eval

subroutine gsl_interpol_spline_init(spline,num_points)
    implicit none
    ! Arguments
    type(gsl_spline),   intent(inout)   ::  spline
    integer,            intent(in)      ::  num_points
    ! Private variables
    integer(fgsl_size_t)    ::  nnum_points
    ! Start work
    if (.not. (allocated(spline%x) .and. allocated(spline%y))) then
        print *, '**ERROR(gsl_interpol_spline_init): x and y arrays are not allocated'
        stop 'Fatal error in GSL_INTERPOL_SPLINE_INIT'
    endif
    if (size(spline%x) .ne. size(spline%y)) then
        print *, '**ERROR(gsl_interpol_spline_init): x and y arrays are not allocated with same sizes'
        stop 'Fatal error in GSL_INTERPOL_SPLINE_INIT'
    endif
    if (size(spline%x) .ne. num_points) then
        print *, '**WARNING(gsl_interpol_spline_init): num_points (', num_points, ') .NE. size(x) (', size(spline%x), ')'
        stop 'Fatal error in GSL_INTERPOL_SPLINE_INIT'
    endif
    ! Convert integer to fgsl type
    nnum_points = num_points
    ! Initialise the spline
    spline%acc = fgsl_interp_accel_alloc()
    spline%spline = fgsl_spline_alloc(fgsl_interp_cspline,nnum_points)
    spline%status = fgsl_spline_init(spline%spline,spline%x,spline%y,nnum_points)
end subroutine gsl_interpol_spline_init

subroutine gsl_interpol_spline_free(spline)
    ! Arguments
    type(gsl_spline),   intent(inout)   ::  spline
    ! Start work
    call fgsl_spline_free(spline%spline)
    call fgsl_interp_accel_free(spline%acc)
end subroutine gsl_interpol_spline_free

subroutine gsl_cspline_test
    use fgsl
    implicit none
    integer :: i,index
    real(fgsl_double) :: x(7),y(7),dy(7),xa
    type(fgsl_interp_accel) :: acc
    type(fgsl_spline) :: spline
    character(len=fgsl_strmax) :: name
    integer(fgsl_int) :: status
    real(fgsl_double) :: ra, rda, rd2a, ri
    integer, parameter :: nstep = 100
    type(interpol_spline) :: home_spline
    !integer(fgsl_int)  ::  npoints
    integer(fgsl_size_t)    ::  npoints

    ! build the signal
    print *, '#m=0,S=2'
    !npoints = 7
    !do i=1,npoints
    !   x(i) = i + 0.5 * sin(real(i))
    !   y(i) = i + cos (real(i)**2)
    !end do
    npoints = 7
    x  = [1.0,2.0,3.0,4.6,5.5,6.8,7.4]
    y  = [1.0,3.7,-4.0,3.0,14.1,11.0,15.60]
    dy = [0.0,0.0,10.0,-5.0,0.0 ,0.0,0.0]


    ! print out the signal
    do i=1,npoints
        print '(2g15.5)', x(i), y(i)
    end do

    !
    acc = fgsl_interp_accel_alloc()
    spline = fgsl_spline_alloc(fgsl_interp_cspline,npoints)
    name = fgsl_spline_name(spline)
    status = fgsl_spline_init(spline,x,y,npoints)
    !print *, 'FGSL spline init status = ', status

    print *, '#m=2,S=0'
    do i=1,nstep
        xa = x(1) + (x(npoints)-x(1))*i/nstep
        ra = fgsl_spline_eval(spline, xa, acc)
        rda = fgsl_spline_eval_deriv(spline, xa, acc)
        rd2a = fgsl_spline_eval_deriv2(spline, xa, acc)
        ri = fgsl_spline_eval_integ(spline, 0.0_fgsl_double, xa, acc)
        !print *, i, ra, rda, rd2a, ri
        print '(2g15.5)', xa, ra
    enddo
    index = fgsl_spline_min_size(spline)
    !print *, 'index = ', index
    call fgsl_spline_free(spline)

    ! Now do the same with the homegrown spline object to see whether we get the same
    call home_spline%Allocate(int(npoints),3)
    call home_spline%SetType('NATURAL')
    home_spline%x = x
    home_spline%y = y
    home_spline%dy = dy
    call home_spline%Solve()
    print *, '#m=1,S=0'
    do i=1,nstep
        xa = x(1) + (x(npoints)-x(1))*i/nstep
        ra = interpol_spline_eval2(home_spline,xa)
        print '(2g15.5)', xa, ra
    enddo
end subroutine gsl_cspline_test


end module Splines
