!>  \brief  Defines the threed class, which handles 3d reconstruction from a set of images,
!!          including appropriate filters etc.
!!
module threeds
    use threeds_base
    use runtime_parameters, only : terminate
    implicit none

    contains

    !>  \brief  Test functionality of the 3d reconstruction module
    subroutine threed_unit_tests()
        implicit none
        call threed_unit_test_1()
        call threed_unit_test_1b()
        call threed_unit_test_2()
    end subroutine threed_unit_tests

    !>  \brief  Single slice insertion with no interpolation
    subroutine threed_unit_test_1()
        use masks, only : mask_add_cosine_falloff
        use ctfs, only : ctf_parameters_init, ctf_parameters, ctf_ang_to_pix
        use images_ctf, only : image_apply_ctf
        use images_comparison, only : correlation
        use euleranglesmod
        use images
        implicit none
        ! Private variables
        type(image) ::  test_2d, test_2d_original, test_3d, check_2d, mask
        integer     ::  ori(3), test_dim
        type(ctf_parameters) :: ctf
        type(euler) ::  angles
        type(threed)    ::  self
        real,   parameter   ::  pixel_size = 1.0
        real    ::  cc
        ! Start work

        write(*,'(a)') '**info(threed_unit_test_1): entering'

        test_dim = 128

        ! Prepare a ctf object
        call ctf_parameters_init(ctf,akv=200.0,cs=2.0,ampl_cont=0.07,dfmid1=1.3)
        call ctf_ang_to_pix(ctf,pixel_size=pixel_size)

        ! Create a test pattern (a cosine bell) and apply the microscope ctf
        test_2d%logical_dimensions = [test_dim,test_dim,1]
        call test_2d%Allocate()
        test_2d%real_values = 0.0
        if (test_2d%is_in_real_space .and. test_2d%object_is_centered_in_box) then
            ori = test_2d%physical_address_of_box_center
        else
            ori = [1,1,1]
        endif
        test_2d%real_values(ori(1),ori(2),ori(3)) = 1.0
        call mask_add_cosine_falloff(test_2d,5.0)
        call test_2d%WriteToDisk('threed_unit_test_1_test_2d.mrc')
        test_2d_original = test_2d
        call image_apply_ctf(test_2d,ctf)
        if (test_2d%IsInFourierSpace()) call test_2d%BackwardFFT()
        call test_2d%WriteToDisk('threed_unit_test_1_test_2d_ctf.mrc')



        ! Prepare Euler angles
        angles%phi = 0.0
        angles%theta = 0.0
        angles%psi = 0.0

        ! Prepare a 3d reconstruction
        self%logical_dimensions = [test_dim,test_dim,test_dim]
        call threed_init(self)

        ! Add slice(s) to 3d reconstruction
        if (test_2d%IsInRealSpace()) call test_2d%ForwardFFT()
        if (test_2d%object_is_centered_in_box) call test_2d%SwapRealSpaceQuadrants()
        call threed_add_slice(self,test_2d,ctf,angles)



        ! Write 3d object to disk
        call threed_write_to_disk(self,'threed_unit_test_1_threed.mrc')

        ! Finalise the 3d reconstruction
        call threed_eval_reconstruction(self,test_3d)
        call test_3d%WriteToDisk('threed_unit_test_1_test_3d.mrc')


        ! Check we got the expected result: extract the central slice in real space and compare it to our test pattern
        call test_3d%GetSlice(check_2d,3,ori(1))
        call check_2d%WriteToDisk('threed_unit_test_1_check_2d.mrc')
        call mask%Allocate(test_2d_original)
        mask%real_values = 1.0
        call mask%ApplyCircularMask(RAD=TEST_DIM/3.0)
        cc = correlation(test_2d_original,check_2d,mask,normalise=.true.)
        write(*,'(a,F0.4)') '**info(threed_unit_test_1): this normalised cross-correlation score should be very high: ', cc
        if (cc .lt. 0.95) then
            write(*,'(a)') '**ERROR(threed_unit_test_1): failed unit test'
            call terminate('Fatal error in THREED_UNIT_TESTS')
        endif

        write(*,'(a)') '**info(threed_unit_test_1): test passed'

    end subroutine threed_unit_test_1

    !>  \brief  Single slice insertion with no interpolation
    subroutine threed_unit_test_1b()
        use masks, only : mask_add_cosine_falloff
        use ctfs, only : ctf_parameters_init, ctf_parameters, ctf_ang_to_pix
        use images_ctf, only : image_apply_ctf
        use images_comparison, only : correlation
        use euleranglesmod
        use images
        implicit none
        ! Private variables
        type(image) ::  test_2d, test_2d_original, test_3d, check_2d, mask
        integer     ::  ori(3), test_dim
        type(ctf_parameters) :: ctf
        type(euler) ::  angles
        type(threed)    ::  self
        real,   parameter   ::  pixel_size = 1.0
        real    ::  cc
        ! Start work

        write(*,'(a)') '**info(threed_unit_test_1b): entering'

        test_dim = 128

        ! Prepare a ctf object
        call ctf_parameters_init(ctf,akv=200.0,cs=2.0,ampl_cont=0.07,dfmid1=1.3)
        call ctf_ang_to_pix(ctf,pixel_size=pixel_size)

        ! Create a test pattern (a cosine bell) and apply the microscope ctf
        test_2d%logical_dimensions = [test_dim,test_dim,1]
        call test_2d%Allocate()
        test_2d%real_values = 0.0
        if (test_2d%is_in_real_space .and. test_2d%object_is_centered_in_box) then
            ori = test_2d%physical_address_of_box_center
        else
            ori = [1,1,1]
        endif
        test_2d%real_values(ori(1),ori(2),ori(3)) = 1.0
        !call mask_add_cosine_falloff(test_2d,5.0)
        call test_2d%WriteToDisk('threed_unit_test_1b_test_2d.mrc')
        test_2d_original = test_2d
        call image_apply_ctf(test_2d,ctf)
        if (test_2d%IsInFourierSpace()) call test_2d%BackwardFFT()
        call test_2d%WriteToDisk('threed_unit_test_1b_test_2d_ctf.mrc')



        ! Prepare Euler angles
        angles%phi = 0.0
        angles%theta = 90.0
        angles%psi = 0.0

        ! Prepare a 3d reconstruction
        self%logical_dimensions = [test_dim,test_dim,test_dim]
        call threed_init(self)

        ! Add slice(s) to 3d reconstruction
        if (test_2d%IsInRealSpace()) call test_2d%ForwardFFT()
        if (test_2d%object_is_centered_in_box) call test_2d%SwapRealSpaceQuadrants()
        call threed_add_slice(self,test_2d,ctf,angles)



        ! Write 3d object to disk
        call threed_write_to_disk(self,'threed_unit_test_1b_threed.mrc')

        ! Finalise the 3d reconstruction
        call threed_eval_reconstruction(self,test_3d)
        call test_3d%WriteToDisk('threed_unit_test_1b_test_3d.mrc')


        ! Check we got the expected result: extract the central slice in real space and compare it to our test pattern
        call test_3d%GetSlice(check_2d,1,ori(1))
        call check_2d%WriteToDisk('threed_unit_test_1b_check_2d.mrc')
        call mask%Allocate(test_2d_original)
        mask%real_values = 1.0
        call mask%ApplyCircularMask(rad=test_dim/3.0)
        cc = correlation(test_2d_original,check_2d,mask,normalise=.true.)
        write(*,'(a,F0.4)') '**info(threed_unit_test_1b): this normalised cross-correlation score should be very high: ', cc
        if (cc .lt. 0.86) then
            write(*,'(a)') '**ERROR(threed_unit_test_1b): failed unit test'
            call terminate('Fatal error in THREED_UNIT_TESTS')
        endif

        write(*,'(a)') '**info(threed_unit_test_1b): test passed'

    end subroutine threed_unit_test_1b

    !>  \brief  Many slices inserted, requiring interpolation. Comparison of interpolation-free direction
    subroutine threed_unit_test_2()
        use masks, only : mask_add_cosine_falloff
        use ctfs, only : ctf_parameters_init, ctf_parameters, ctf_ang_to_pix
        use images_ctf, only : image_apply_ctf
        use images_comparison, only : correlation
        use euleranglesmod
        use images
        implicit none
        ! Private variables
        type(image) ::  test_2d, test_2d_original, test_3d, check_2d, mask
        integer     ::  ori(3), test_dim
        type(ctf_parameters) :: ctf
        type(euler) ::  angles
        type(threed)    ::  self
        real,   parameter   ::  pixel_size = 1.0
        real    ::  cc
        integer, parameter ::  number_of_insertions = 100
        integer ::  current_insertion
        ! Start work

        write(*,'(a)') '**info(threed_unit_test_2): entering'

        test_dim = 128

        ! Prepare a ctf object
        call ctf_parameters_init(ctf,AKV=200.0,CS=2.0,AMPL_CONT=0.07,DFMID1=1.3)
        call ctf_ang_to_pix(ctf,PIXEL_SIZE=PIXEL_SIZE)

        ! Create a test pattern (a cosine bell) and apply the microscope ctf
        test_2d%logical_dimensions = [TEST_DIM,test_dim,1]
        call test_2d%Allocate()
        test_2d%real_values = 0.0
        if (test_2d%is_in_real_space .and. test_2d%object_is_centered_in_box) then
            ori = test_2d%physical_address_of_box_center
        else
            ori = [1,1,1]
        endif
        test_2d%real_values(ori(1),ori(2),ori(3)) = 1.0
        call mask_add_cosine_falloff(test_2d,5.0)
        call test_2d%WriteToDisk('threed_unit_test_2_test_2d.mrc')
        test_2d_original = test_2d
        call image_apply_ctf(test_2d,ctf)
        if (test_2d%IsInFourierSpace()) call test_2d%BackwardFFT()
        call test_2d%WriteToDisk('threed_unit_test_2_test_2d_ctf.mrc')



        ! Prepare Euler angles
        angles%phi = 0.0
        angles%theta = 0.0
        angles%psi = 0.0

        ! Prepare a 3d reconstruction
        self%logical_dimensions = [TEST_DIM,test_dim,TEST_DIM]
        call threed_init(self)

        ! Add slice(s) to 3d reconstruction
        if (test_3d%IsInRealSpace()) call test_2d%ForwardFFT()
        if (test_2d%object_is_centered_in_box) call test_2d%SwapRealSpaceQuadrants()
        do current_insertion=1,number_of_insertions
            angles%theta = real(current_insertion) / real(number_of_insertions) * 180. + 1.23
            write(*,'(a,i0,a,i0,a,F0.2)') '**info(threed_unit_test_2): inserting slice ', current_insertion, ' of ', &
                                            number_of_insertions, ' with theta = ', angles%theta
            call threed_add_slice(self,test_2d,ctf,angles)
        enddo



        ! Write 3d object to disk
        call threed_write_to_disk(self,'threed_unit_test_2_threed.mrc')

        ! Finalise the 3d reconstruction
        call threed_eval_reconstruction(self,test_3d)
        call test_3d%WriteToDisk('threed_unit_test_2_test_3d.mrc')


        ! Check we got the expected result: extract the central slice in real space and compare it to our test pattern
        call test_3d%GetSlice(check_2d,3,ori(1))
        call check_2d%WriteToDisk('threed_unit_test_2_check_2d.mrc')
        call mask%Allocate(test_2d_original)
        mask%real_values = 1.0
        call mask%ApplyCircularMask(rad=test_dim/3.0)
        cc = correlation(test_2d_original,check_2d,mask,normalise=.true.)
        write(*,'(a,F0.4)') '**info(threed_unit_test_2): this normalised cross-correlation score should be very high: ', cc
        if (cc .lt. 0.99) then
            write(*,'(a)') '**ERROR(threed_unit_test_2): failed unit test'
            call terminate('Fatal error in THREED_UNIT_TESTS')
        endif

        write(*,'(a)') '**info(threed_unit_test_2): test passed'

    end subroutine threed_unit_test_2

    !>  \brief  Return a real 3d volume reconstruction from the threed object - i.e. apply weighting & correct for the interpolation
    !!          artefacts (a la correct3d / gridding)
    recursive subroutine threed_eval_reconstruction(  self,img,img_no_fom, &
                                            !wiener_filter, &
                                            rec_num, &
                                            !pssnr_fudge, &
                                            preserve_high_frequencies,    &
                                            cylinder_mask,skip_fom_filter, &
                                            fsc_volumes,fsc_volumes_no_mask, &
                                            fsc_volumes_no_mask_fom, &
                                            skip_fsc,img_ampl,skip_ampl, &
                                            estimated_resolution_of_fsc_volumes)
        use imagefiles ! for debug
        use procfs, only : print_memory_usage_stats ! debug
        use InterpolatingWindows, only : correct_3d
        use images
        implicit none
        ! Arguments
        type(threed),               intent(inout)   ::  self
        type(image),                intent(inout)   ::  img                         !<  Output reconstruction image
        type(image),    optional,   intent(inout)   ::  img_no_fom                  !<  Output reconstruction, without fom filtering
        !logical,        optional,   intent(in)      ::  wiener_filter               !<  Whether a Wiener filter should be used. Defaults to true. If not, a simple division by CTF^2 is performed.
        integer,        optional,   intent(in)      ::  rec_num                     !<  Which internal reconstruction should be used. If not specified and there are two internal reconstructions, the sum will be returned.
        !real,           optional,   intent(in)      ::  pssnr_fudge                 !<  Scaling fudge factor for the pssnr term
        logical,        optional,   intent(in)      ::  preserve_high_frequencies   !<  Even if they are dominated by noise (i.e. the ssnr is low), do not zero high frequencies.
                                                                                    !!  This will probably lead to more noisy reconstructions, but may be used if the Wiener filter is thought to be too stringent. Defaults to false.
        logical,        optional,   intent(in)      ::  cylinder_mask               !<  Cylindrical, rather than spherical, mask
        logical,        optional,   intent(in)      ::  skip_fom_filter             !<  Do not apply fom filter
        type(image),    optional,   intent(inout)   ::  fsc_volumes(2)              !<  If present, the volumes between which the fsc is computed will be returned (as used to estimate SSNR etc)
        type(image),    optional,   intent(inout)   ::  fsc_volumes_no_mask(2)      !<  If present, the volumes between which the fsc is computed will be returned, before symmetrisation and masking
        type(image),    optional,   intent(inout)   ::  fsc_volumes_no_mask_fom(2)  !<  If present, the volumes between which the fsc is computed will be returned, before symmetrisation and masking, but after FOM filtering
        logical,        optional,   intent(in)      ::  skip_fsc                    !<  Do not compute fsc
        type(image),    optional,   intent(inout)   ::  img_ampl                    !<  Output amplitudes reconstruction image
        logical,        optional,   intent(in)      ::  skip_ampl                   !<  Do not evaluate the amplitude reconstruction
        real,           optional,   intent(in)      ::  estimated_resolution_of_fsc_volumes        !<  Estimated resolution of fsc volumes. This is used to do masking.

        ! Private variables
        logical,        parameter       ::  debug   =   .false.
        logical     ::  wwiener_filter
        integer     ::  rrec_num    ! 0 indicates sum of both
        logical     ::  ppreserve_high_frequencies
        logical     ::  sskip_fom_filter
        logical     ::  sskip_fsc
        logical     ::  ccylinder_mask
        logical     ::  sskip_ampl
        ! Start work

        ! Memory footprint diag
        !if (debug) then
        !    write(*,'(a)') '**debug(threed_eval_reconstruction): memory usage on entry'
        !    call print_memory_usage_stats()
        !endif

        ! Debug
        if (debug) then
            call threed_write_to_disk(self,'dbg_3deval_on_entry.mrc')
        endif

        ! Will we be using a Wiener filter?
        wwiener_filter = .true.
        !if (present(wiener_filter)) wwiener_filter = wiener_filter

        sskip_fom_filter = .false.
        if (present(skip_fom_filter)) sskip_fom_filter = skip_fom_filter

        sskip_fsc = .false.
        if (present(skip_fsc)) sskip_fsc = skip_fsc

        ccylinder_mask = .false.
        if (present(cylinder_mask)) ccylinder_mask = cylinder_mask

        sskip_ampl = .false.
        if (present(skip_ampl)) sskip_ampl = skip_ampl

        ! Which reconstruction are we asked to evaluate?
        if (present(rec_num)) then
            rrec_num = rec_num
        else
            if (threed_num_of_recs(self) .eq. 2) then
                rrec_num = 0    ! we'll evaluate the sum of both
            else
                rrec_num = 1    ! there is only one reconstruction available, that's the one we'll eval
            endif
        endif

        ! By default, we will not preserve high-frequencies if their ssnr is too low
        ppreserve_high_frequencies = .false.
        if (present(preserve_high_frequencies)) ppreserve_high_frequencies = preserve_high_frequencies

        ! Check that the reconstruction we are asked to evaluate exists
        if (rrec_num .gt. threed_num_of_recs(self)) then
            write(*,'(a,i0,a,i0,a)') '**ERROR(threed_eval_reconstruction): was asked to evaluate reconstruction ', rrec_num, &
                                        ' yet there are only ', threed_num_of_recs(self), ' reconstructions available'
            call terminate('Fatal error in THREED_EVAL_RECONSTRUCTION. Programming error?')
        else if (rrec_num .eq. 0 .and. threed_num_of_recs(self) .eq. 1) then
            write(*,'(2a,i0,a)') '**ERROR(threed_eval_reconstruction): was asked to evaluate the sum of',    &
                                ' internal reconstructions yet there is only ', &
                                    threed_num_of_recs(self), ' reconstruction available'
            call terminate('Fatal error in THREED_EVAL_RECONSTRUCTION. Programming error?')
        endif

        !
        if (threed_num_of_recs(self) .eq. 2) then
            if (.not. threed_two_recs(self)) then
                write(*,'(a)') '**ERROR(threed_eval_reconstruction): expecting, but did not find, two internal reconstructions'
                call terminate('Fatal error in THREED_EVAL_RECONSTRUCTION')
            endif
        endif

        !
        if (self%reconstruct_amplitudes .and. .not. present(img_ampl) .and. .not. sskip_ampl) then
            write(*,'(2a)') '**ERROR(threed_eval_reconstruction): was asked to reconstruct amplitudes, ',    &
                                'but not given image volume to write to'
            call terminate('Fatal error in THREED_EVAL_RECONSTRUCTION')
        endif

        ! Memory footprint diag
        !if (debug) then
        !    write(*,'(a)') '**debug(threed_eval_reconstruction): memory usage before output alloc'
        !    call print_memory_usage_stats()
        !    print *, '**() already allocated?', img%is_in_memory
        !endif

        ! Allocate memory for output
        call img%Allocate(self%rec(1)%vol_image)
        if (self%reconstruct_amplitudes .and. present(img_ampl)) then
            call img_ampl%Allocate(self%rec(1)%vol_ampl)
        endif

        ! Initialise the output
        img = (0.0e0,0.0e0)

        ! Memory footprint diag
        !if (debug) then
        !    write(*,'(a)') '**debug(threed_eval_reconstruction): memory usage after output alloc'
        !    call print_memory_usage_stats()
        !endif

        ! If the incoming threed object had no data, we return a blank volume
        if (sum(self%rec%num_contrib_elems) .eq. 0) then
            print '(a)', '**info(threed_eval_reconstruction): empty 3d reconstruction!'
            img%real_values = 0.0
            img%is_in_real_space = .true.
            if (self%reconstruct_amplitudes .and. present(img_ampl)) then
                img_ampl%real_values = 0.0
                img_ampl%is_in_real_space = .true.
            endif
        ! The normal case, where the 3d reconstruction isn't blank
        else

            ! Diag
            !if (debug) then
            !    if (self%get2d_put3d) then
            !        print '(a)', '**debug(threed_eval_reconstruction): using get2d_put3d logic'
            !    else
            !        print '(a)', '**debug(threed_eval_reconstruction): using put3d logic'
            !    endif
            !endif

            ! Debug dumps
            if (debug) then
                img = self%rec(1)%vol_image
                if (img%IsInRealSpace()) call img%ForwardFFT()
                call img%SwapRealSpaceQuadrants()
                !call img%ComputeAmplitudeSpectrum(tmpvol)
                !call tmpvol%WriteToDisk('debug_before_3dctf_ampl.mrc')
                !tmpvol%real_values = tmpvol%real_values**2
                !call tmpvol%WriteToDisk('debug_before_3dctf_pow.mrc')
                call img%BackwardFFT()
                call img%WriteToDisk('debug_before_3dctf.mrc')
                if (debug) write(*,'(a,l1)') '**debug(threed_eval_reconstruction): before 3dctf, volume origin at center? ', &
                                                img%object_is_centered_in_box
                call img%ForwardFFT()
                call img%SwapRealSpaceQuadrants()
            endif

            ! Prepare the CTF^2 volume
            if (debug) call self%rec(1)%vol_amplitude_weights%WriteToDisk('debug_3dctf.mrc')

            !call ctfvol_shift(self%rec(1)%vol_amplitude_weights,ctf2_halfvol)
            !if (debug) then
            !    call ctf2_halfvol%WriteToDisk('debug_3dctf2.mrc')
            !endif

            ! Memory footprint diag
            !if (debug) then
            !    write(*,'(a)') '**debug(threed_eval_reconstruction): memory usage before FSC'
            !    call print_memory_usage_stats()
            !endif

            ! Compute fsc
            if (threed_num_of_recs(self) .eq. 2 .and. .not. sskip_fsc) then
                if (.not. self%fsc_is_up_to_date .or. .not. self%ssnr_is_up_to_date) then
                    write(*,'(a)') '**info(threed_eval_reconstruction): computing FSC'
                    call threed_update_fsc(self,fsc_volumes=fsc_volumes,fsc_volumes_no_mask=fsc_volumes_no_mask)
                endif
            endif

            ! Memory footprint diag
            !if (debug) then
            !    write(*,'(a)') '**debug(threed_eval_reconstruction): memory usage after FSC'
            !    call print_memory_usage_stats()
            !endif

            ! Apply Wiener filter
            if (wwiener_filter) then
                write(*,'(a)') '**info(threed_eval_reconstruction): will apply a Wiener-like filter'
                call threed_apply_wiener_filter(self,img,rec_num=rrec_num, &
                                                !pssnr_fudge=pssnr_fudge,  &
                                                preserve_high_frequencies=ppreserve_high_frequencies)
                if (self%reconstruct_amplitudes .and. present(img_ampl)) then
                    call threed_apply_wiener_filter(self,img_ampl,rec_num=rrec_num, &
                                                    !pssnr_fudge=pssnr_fudge, &
                                                    preserve_high_frequencies=ppreserve_high_frequencies)
                endif
            else
                write(*,'(a)') '**info(threed_eval_reconstruction): no Wiener filter - will divide by CTF^2 instead'
                call threed_ctf2_division(self,img,rec_num=rrec_num)
                if (self%reconstruct_amplitudes .and. present(img_ampl)) then
                     call threed_ctf2_division(self,img,rec_num=rrec_num)
                endif
            endif

            ! Memory footprint diag
            !if (debug) then
            !    write(*,'(a)') '**debug(threed_eval_reconstruction): memory usage after Wiener'
            !    call print_memory_usage_stats()
            !endif

            if (present(img_no_fom)) then
                img_no_fom = img
                if (img_no_fom%is_in_real_space) call img_no_fom%ForwardFFT()
            endif

            ! Debug
            if (debug) then
                call img%WriteToDisk('dbg_3deval_before_fom.mrc')
            endif


            ! Apply fom filter
            if (self%fom_filter .and. .not. sskip_fom_filter) then
                write(*,'(a)') '**info(threed_eval_reconstruction): applying fom filter'
                call threed_apply_fom_filter(self,img,print_filter=.false.)
                if (self%reconstruct_amplitudes) then
                    write(*,'(a)') '**info(threed_eval_reconstruction): not applying fom filter to amplitudes reconstruction'
                endif
                if (present(fsc_volumes_no_mask_fom)) then
                    if (present(fsc_volumes_no_mask)) then
                        fsc_volumes_no_mask_fom(1) = fsc_volumes_no_mask(1)
                        fsc_volumes_no_mask_fom(2) = fsc_volumes_no_mask(2)
                        call threed_apply_fom_filter(self,fsc_volumes_no_mask_fom(1),print_filter=.false.)
                        call threed_apply_fom_filter(self,fsc_volumes_no_mask_fom(2),print_filter=.false.)
                    else
                        write(*,'(a)') '**warning(threed_eval_reconstruction): cannot do fsc_volumes_no_mask_fom'//&
                                                                                    ' without fsc_volumes_no_mask'
                    endif
                endif
            endif

            ! Memory footprint diag
            !if (debug) then
            !    write(*,'(a)') '**debug(threed_eval_reconstruction): memory usage after FOM'
            !    call print_memory_usage_stats()
            !endif

            ! Debug
            if (debug) then
                call img%WriteToDisk('dbg_3deval_after_fom.mrc')
            endif

            ! The volume should be in Fourier space
            if (img%is_in_real_space) call img%ForwardFFT()
            if (present(img_no_fom)) then
                if (img_no_fom%is_in_real_space) call img_no_fom%ForwardFFT()
                call img_no_fom%ApplyCircularMask(0.5)
                img_no_fom%complex_values(1,1,1) = (0.,0.)
            endif

            ! The Nyquist elements are not reliable, so we remove them
            call img%ApplyCircularMask(0.5)

            ! Also, zero the average
            img%complex_values(1,1,1) = (0.,0.)

            ! The image is ready to be back-FT'ed, but first may need to be shifted
            if (.not. img%object_is_centered_in_box) call img%SwapRealSpaceQuadrants()
            call img%BackwardFFT()
            if (present(img_no_fom)) then
                if (.not. img_no_fom%object_is_centered_in_box) call img_no_fom%SwapRealSpaceQuadrants()
                call img_no_fom%BackwardFFT()
            endif

            ! Correct for falloff due to interpolation in reciprocal space
            write(*,'(a)') '**info(threed_eval_reconstruction): will correct damping artefact due to interpolation kernel'
            if (debug) call img%WriteToDisk('debug_before_correct_3d.mrc')
            !if (self%get2d_put3d) then
            !    if (debug) print '(a)', '**debug(threed_eval_reconstruction): correcting for interpolation kernel linear rect 2'
            !    call correct_3d(img,self%interpol_scheme%interpolation_instr_function,  &
            !                        self%interpol_scheme%windowing_instr_function,    &
            !                        self%interpol_scheme%window_width,reverse=.false.)
            !    if (self%reconstruct_amplitudes .and. present(img_ampl) .and. .not. sskip_ampl) then
            !        call correct_3d(img_ampl,self%interpol_scheme%interpolation_instr_function, &
            !                        self%interpol_scheme%windowing_instr_function,    &
            !                        self%interpol_scheme%window_width,reverse=.false.)
            !    endif
            !else
            if (debug) print '(a)', '**debug(threed_eval_reconstruction): correcting for appropriate interpolation kernel'
            ! Generalised correct_3d
            call correct_3d(img,self%interpol_scheme%interpolation_instr_function,  &
                                self%interpol_scheme%windowing_instr_function,    &
                                self%interpol_scheme%window_width,reverse=.false.)
            if (self%reconstruct_amplitudes .and. present(img_ampl) .and. .not. sskip_ampl) then
                call correct_3d(img_ampl,self%interpol_scheme%interpolation_instr_function, &
                                self%interpol_scheme%windowing_instr_function,    &
                                self%interpol_scheme%window_width,reverse=.false.)
            endif
            if (present(img_no_fom)) then
                call correct_3d(img_no_fom,self%interpol_scheme%interpolation_instr_function,  &
                                self%interpol_scheme%windowing_instr_function,    &
                                self%interpol_scheme%window_width,reverse=.false.)
            endif
            !endif

            ! Memory footprint diag
            !if (debug) then
            !    write(*,'(a)') '**debug(threed_eval_reconstruction): memory usage after correct3d'
            !    call print_memory_usage_stats()
            !endif


            ! Add a mask - correct_3d can leave funny artefacts, and we can flatten the solvent anyway. a cosine-edged sphere should do the trick.
            call img%ApplySoftCircularMask(self%mask_rad,self%mask_falloff,wipe_input=.false.,mask_with_average=.false.,  &
                                cylinder=ccylinder_mask .or. self%point_sym%helical)
            if (present(img_no_fom)) then
                call img_no_fom%ApplySoftCircularMask(  self%mask_rad,self%mask_falloff, &
                                                        wipe_input=.false.,mask_with_average=.false.,  &
                                                        cylinder=(ccylinder_mask .or. self%point_sym%helical))
            endif
            if (self%reconstruct_amplitudes .and. .not. sskip_ampl) then
                write(*,'(a)') '**info(threed_eval_reconstruction): not applying cosine-edge disc mask to amplitudes reconstruction'
            endif
        endif   ! End of test for blank reconstruction

        ! Memory footprint diag
        !if (debug) then
        !    write(*,'(a)') '**debug(threed_eval_reconstruction): memory usage on exit'
        !    call print_memory_usage_stats()
        !endif
    end subroutine threed_eval_reconstruction

    !>  \brief  Apply a Figure-Of-Merit (fom) filter to the reconstruction, a la Frealign (flag fcref) and Henderson & Rosenthal (2003)
    !!
    !!  We actually use Eq 11 of Sindelar & Grigorieff (2012), which is the square of Henderson & Rosenthal (2003). Apparently, this is
    !!  better suited to our use, where the fsc is calcuated between masked volumes
    subroutine threed_apply_fom_filter(self,img,print_filter)
        use images
        implicit none
        ! arguments
        type(threed),               intent(inout)   ::  self            !<  3d reconstruction object
        type(image),                intent(inout)   ::  img             !<  Evaluated 3d reconstruction volume (should already have been wiener filtered)
        logical,        optional,   intent(in)      ::  print_filter    !<  Print out filter values
        ! private variables
        logical,    parameter   ::  dnf(3) = (/.true.,.false.,.false./)
        integer                 ::  i,j,k, current_shell
        integer                 ::  last_printed_shell
        logical                 ::  pprint_filter, print_filter_value
        real                    ::  filter_value
        logical,    parameter   ::  use_sindelar_2012   =   .true.
        logical,    parameter   ::  oonce_at_zero_stay_there = .true.
        real,       parameter   ::  stay_at_one_until_this_resolution = 30.0 ! In Angstroms
        integer                 ::  stay_at_one_until_this_shell
        integer                 ::  first_shell_where_filter_is_zero
        integer                 ::  dont_zero_before_this_shell
        real                    ::  x,y,z
        ! start work

        pprint_filter = .false.
        if (present(print_filter)) pprint_filter = print_filter
        last_printed_shell = -1

        ! update the fsc if necessary
        if (.not. self%fsc_is_up_to_date) then
            call threed_update_fsc(self)
        endif

        ! the image needs to be in fourier space
        if (img%is_in_real_space) call img%ForwardFFT()

        if (pprint_filter) then
            write(*,'(a)') '**info(threed_apply_fom_filter): fsc, fom, filter values for reconstruction:'
            write(*,'(3(a6,1x))') 'Shell', 'FSC', 'FOM'
        endif

        first_shell_where_filter_is_zero = 0
        !dont_zero_before_this_shell = size(self%fsc) / 4 ! We're allowed to zero before then, but not to be forced to stay at 0
        dont_zero_before_this_shell = self%fsc_curve%GetNumberOfDataPoints() / 4 ! We're allowed to zero before then, but not to be forced to stay at 0

        stay_at_one_until_this_shell = nint(self%pixel_size / stay_at_one_until_this_resolution * self%logical_dimensions(1))

        ! loop over the volume and apply the filter
        do k=1,img%logical_dimensions(3)
            z = real(img%LogicalIndexGivenPhysicalIndexInFourierSpace(k,3))**2
            do j=1,img%logical_dimensions(2)
                y = real(img%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2))**2
                do i=1,img%physical_index_of_first_negative_frequency(1)-1
                    x = real(i-1)**2
                    current_shell = nint(sqrt(x+y+z))+1
                    print_filter_value = (pprint_filter .and. current_shell-1 .eq. last_printed_shell+1)
                    if (print_filter_value) last_printed_shell = current_shell-1
                    !if (current_shell .le. size(self%fsc)) then
                    if (current_shell .le. self%fsc_curve%GetNumberOfDataPoints()) then
                        if (current_shell .le. stay_at_one_until_this_shell) then
                            filter_value = 1.0e0
                        !else if (self%fsc(current_shell) .le. 0.0e0) then
                        else if (self%fsc_curve%GetY(current_shell) .le. 0.0e0) then
                            filter_value = 0.0e0
                            if (first_shell_where_filter_is_zero .eq. 0 .and. current_shell .ge. dont_zero_before_this_shell) then
                                first_shell_where_filter_is_zero = current_shell
                            endif
                        else
                            !filter_value =  2.*self%fsc(current_shell)/(1.+self%fsc(current_shell))
                            filter_value =  2.*self%fsc_curve%GetY(current_shell)/(1.+self%fsc_curve%GetY(current_shell))
                            if (.not. use_sindelar_2012) filter_value = sqrt(filter_value)
                        endif
                        if (oonce_at_zero_stay_there .and. current_shell .ge. first_shell_where_filter_is_zero &
                                                     .and. first_shell_where_filter_is_zero .ne. 0) then
                            filter_value = 0.0e0
                        endif
                        img%complex_values(i,j,k) = img%complex_values(i,j,k) * filter_value
                        if (pprint_filter) then
                            !write(*,'(i6,1x,2(f6.3,1x))') current_shell, self%fsc(current_shell), filter_value
                            write(*,'(i6,1x,2(f6.3,1x))') current_shell, self%fsc_curve%GetY(current_shell), filter_value
                        endif
                    else
                        img%complex_values(i,j,k) = (0.0,0.0)
                    endif
                enddo
            enddo
        enddo
        if (pprint_filter) write(*,'(a)') ''
    end subroutine threed_apply_fom_filter

    !>  \brief  Divide reconstruction by CTF^2 volume to obtain least-squares estimate of the reconstruction in the absence of snr estimates
    subroutine threed_ctf2_division(self,img,rec_num)
        use images_ctf
        use images
        implicit none
        ! Arguments
        type(threed),               intent(inout)   ::  self            !<  3d reconstruction object (input)
        type(image),                intent(inout)   ::  img             !<  Resulting volume
        integer,                    intent(in)      ::  rec_num         !<  Which internal reconstruction should be used. If 0, and there are two internal reconstructions, the sum will be returned.
        ! Private variables
        integer                         ::  i,j,k,i_real,j_real,k_real
        logical,    parameter           ::  divide_the_sachse_way = .false.  !  Use maxval(ctfsqavg)*0.01 as an added constant in the denominator to avoid division by zero
                                                                             !  Tested on 20-March-2013 and it makes reconstructions worse, so leaving this off.
        real                            ::  max_val
        ! Start work

        ! Make sure the image is already allocated
        if (.not. img%IsAllocated()) then
            img%logical_dimensions = self%logical_dimensions
            call img%Allocate()
        endif

        ! Ensure the reconstruction volume is in Fourier space
        if (self%rec(1)%vol_image%IsInRealSpace()) then
            call self%rec(1)%vol_image%ForwardFFT()
            call self%rec(2)%vol_image%ForwardFFT()
        endif

        ! Ensure the origin is not in center (is this really necessary?)
        if (self%rec(1)%vol_image%object_is_centered_in_box) then
            call self%rec(1)%vol_image%SwapRealSpaceQuadrants()
            call self%rec(2)%vol_image%SwapRealSpaceQuadrants()
        endif

        ! Work out the constant value to avoid divide by zero
        if (divide_the_sachse_way) then
            ! Work out the maximum value
            max_val = -HUGE(1.0)
            do K=1,img%logical_dimensions(3)
                k_real = ctfvol_addr_real_from_comp_dim2or3(img,k)
                if (k_real .eq. img%logical_dimensions(3)+1) cycle
                do J=1,img%logical_dimensions(2)
                    j_real = ctfvol_addr_real_from_comp_dim2or3(img,j)
                    if (j_real .eq. img%logical_dimensions(2)+1) cycle
                    do I=1,img%logical_dimensions(1)/2+1
                        i_real = ctfvol_addr_real_from_comp_dim1(img,i)
                        if (rec_num .eq. 0) then
                            if (  self%rec(1)%vol_amplitude_weights%real_values(i_real,j_real,k_real)+   &
                                  self%rec(2)%vol_amplitude_weights%real_values(i_real,j_real,k_real)    &
                                  .gt. max_val) then
                                max_val =   self%rec(1)%vol_amplitude_weights%real_values(i_real,j_real,k_real)+   &
                                            self%rec(2)%vol_amplitude_weights%real_values(i_real,j_real,k_real)
                            endif
                        else
                            if (self%rec(rec_num)%vol_amplitude_weights%real_values(i_real,j_real,k_real) .gt. max_val) then
                                max_val =   self%rec(rec_num)%vol_amplitude_weights%real_values(i_real,j_real,k_real)
                            endif
                        endif
                    enddo
                enddo
            enddo
        endif

        ! Do the division
        do K=1,img%logical_dimensions(3)
            k_real = ctfvol_addr_real_from_comp_dim2or3(img,k)
            if (k_real .eq. img%logical_dimensions(3)+1) cycle
            do J=1,img%logical_dimensions(2)
                j_real = ctfvol_addr_real_from_comp_dim2or3(img,j)
                if (j_real .eq. img%logical_dimensions(2)+1) cycle
                do I=1,img%logical_dimensions(1)/2+1
                    i_real = ctfvol_addr_real_from_comp_dim1(img,i)
                    if (rec_num .eq. 0) then
                        img%complex_values(i,j,k) =    self%rec(1)%vol_image%complex_values(i,j,k) + &
                                                       self%rec(2)%vol_image%complex_values(i,j,k)
                        if (divide_the_sachse_way) then
                            img%complex_values(i,j,k) =    img%complex_values(i,j,k) / &
                                                            (self%rec(1)%vol_amplitude_weights%real_values(i_real,j_real,k_real) + &
                                                             self%rec(2)%vol_amplitude_weights%real_values(i_real,j_real,k_real) + &
                                                             MAX_VAL*0.01)
                        else
                            img%complex_values(i,j,k) =    img%complex_values(i,j,k) / &
                                                   max(self%rec(1)%vol_amplitude_weights%real_values(i_real,j_real,k_real)+   &
                                                       self%rec(2)%vol_amplitude_weights%real_values(i_real,j_real,k_real),  &
                                                       0.0001)    ! To avoid division by zero, we use the max function
                        endif
                    else
                        img%complex_values(i,j,k) =    self%rec(rec_num)%vol_image%complex_values(i,j,k)
                        if (divide_the_sachse_way) then
                            img%complex_values(i,j,k) =    img%complex_values(i,j,k)    &
                                        / (self%rec(rec_num)%vol_amplitude_weights%real_values(i_real,j_real,k_real)+MAX_VAL*0.01)
                        else
                            ! To avoid division by zero, we use the max function
                            img%complex_values(i,j,k) =    img%complex_values(i,j,k)    &
                                          / max(self%rec(rec_num)%vol_amplitude_weights%real_values(i_real,j_real,k_real),0.0001)
                        endif
                    endif
                enddo
            enddo
        enddo
    end subroutine threed_ctf2_division


    !>  \brief  Apply Wiener filter and amplitude weighting to the 3D reconstruction
    subroutine threed_apply_wiener_filter(self,img,rec_num, &
                                          !pssnr_fudge, &
                                          preserve_high_frequencies)
        use imagefiles ! for debug
        use images_ctf
        use images
        implicit none
        ! Arguments
        type(threed),               intent(inout)   ::  self                        !<  3d reconstruction object
        type(image),                intent(inout)   ::  img                         !<  Resulting volume
        integer,                    intent(in)      ::  rec_num                     !<  Which internal reconstruction should be used. If 0, and there are two internal reconstructions, the sum will be returned.
        !real,           optional,   intent(in)      ::  pssnr_fudge                 !<  a scaling fudge factor applied to the Wiener term
        logical,        optional,   intent(in)      ::  preserve_high_frequencies   !<  Even if they are dominated by noise (i.e. the ssnr is low), do not zero high frequencies.
                                                                                    !!  This will probably lead to more noisy reconstructions, but may be used if the Wiener filter is thought to be too stringent. Defaults to false.
        ! Private variables
        logical,        parameter       ::  debug   =   .false.
        !type(image)                    ::  tmpvol
        integer                         ::  i,j,k
        integer                         ::  i_real,j_real,k_real
        real,           allocatable     ::  wiener_filter_term(:)
        logical,        parameter       ::  dnf(3) = [.true.,.false.,.false.]
        integer                         ::  current_shell, num_of_shells
        !real                            ::  ppssnr_fudge
        logical                         ::  ppreserve_high_frequencies
        integer                         ::  rad_x, rad_y, rad_z
        ! Start work

        ! Fudge factor
        !if (present(pssnr_fudge)) then
        !    if (debug) write(*,'(a,G15.3)') '**debug(threed_apply_wiener_filter): applying fudge factor of ', pssnr_fudge
        !    ppssnr_fudge = pssnr_fudge
        !else
        !    ppssnr_fudge = 1.0
        !    if (debug) write(*,'(a,G15.3)') '**debug(threed_apply_wiener_filter): no fudge factor specified, defaulting to ', &
        !                    ppssnr_fudge
        !endif

        if (.not. img%HasEvenDimensions()) then
            write(*,'(a,3(i0,1x))') '**ERROR(threed_apply_wiener_filter): odd dimensions not supported: ', img%logical_dimensions
            call terminate('Feature not yet implemented')
        endif

        ! Debug dumps
        !if (debug) then
        !    img = self%vol_image
        !    call img%ShiftRealSpaceOrigin()
        !    call img%ComputeAmplitudeSpectrum(tmpvol)
        !    call tmpvol%WriteToDisk('debug_before_3dctf_ampl.mrc')
        !    tmpvol%real_values = tmpvol%real_values**2
        !    call tmpvol%WriteToDisk('debug_before_3dctf_pow.mrc')
        !   call img%BackwardFFT()
        !    call img%WriteToDisk('debug_before_3dctf.mrc')
        !    call img%ForwardFFT()
        !    call img%SwapRealSpaceQuadrants()
        !endif

        ! Make sure the image is already allocated
        if (.not. img%IsAllocated()) then
            img%logical_dimensions = self%logical_dimensions
            call img%Allocate()
        endif

        ! We assume the image is square/cube
        if (.not. img%IsSquare()) then
            write(*,'(a,3(i0,1x))') '**ERROR(threed_apply_wiener_filter): output image is not cubic: ', img%logical_dimensions
            call terminate('Fatal error in THREED_APPLY_WIENER_FILTER')
        endif

        ! By default, we will not preserve high-frequencies if their ssnr is too low
        if (present(preserve_high_frequencies)) then
            ppreserve_high_frequencies = preserve_high_frequencies
        else
            ppreserve_high_frequencies = .false.
        endif

        ! Ensure the reconstruction volume is in Fourier space
        if (self%rec(1)%vol_image%is_in_real_space) then
            call self%rec(1)%vol_image%ForwardFFT()
            if (size(self%rec) .eq. 2) call self%rec(2)%vol_image%ForwardFFT()
        endif

        ! Ensure the origin is not in center (is this really necessary?)
        if (self%rec(1)%vol_image%object_is_centered_in_box) then
            call self%rec(1)%vol_image%SwapRealSpaceQuadrants()
            if (size(self%rec) .eq. 2) call self%rec(2)%vol_image%SwapRealSpaceQuadrants()
        endif

        if (debug) print *, '**debug(threed_apply_wiener_filter): about to work out filter'

        ! Work out the number of shells to the Nyquist
        ! (maybe should instead work out the number of shells to the corners of the cube?)
        num_of_shells = img%GetMaximumRadius()

        ! Calculate the Wiener filter term (constant or per-resolution-shell)
        if (self%use_ssnr_estimate) then
            if (rec_num .ne. 0 .and. size(self%rec) .gt. 1) then
                write(*,'(2a)') '**warning(threed_apply_wiener_filter): will apply a Wiener filter based on ',   &
                                'FSC-estimated ssnr, even though it will only be applied to one of the reconstructions'
            endif
            if (self%ssnr_is_up_to_date) then
                write(*,'(2a)') '**info(threed_apply_wiener_filter): we already have an up-to-date estimate of ',   &
                                'the ssnr, so we will not re-estimate it'
            else
                write(*,'(2a)') '**info(threed_apply_wiener_filter): we do not have an up-to-date estimate of ',    &
                                'the ssnr, so we will estimate it now'
                !call threed_estimate_ssnr_from_half_dataset_threeds(self,fsc_estimate_above_noise=ppreserve_high_frequencies)
                call threed_update_fsc(self,keep_fsc_above_noise_when_estimating_ssnr=ppreserve_high_frequencies)
            endif
            !wiener_filter_term = self%ssnr_estimate
            call self%ssnr_curve%CopyYData(wiener_filter_term)
            ! The Wiener filter term is actually 1/PSSNR
            do i=1,size(wiener_filter_term)
                wiener_filter_term(i) = 1.0 / max(wiener_filter_term(i),.001)
            enddo
        else
            wiener_filter_term = threed_guess_wiener_filter_term(self,rec_num=rec_num)
        endif

        ! Apply fudge
        !wiener_filter_term = wiener_filter_term * ppssnr_fudge

        if (debug) then
            !write(*,'(a,<SIZE(wiener_filter_term)>(F0.3,x))') '**debug(threed_apply_wiener_filter): wiener filter term = ', &
            !                                                            wiener_filter_term
            !call image_print_info(self%rec(1)%vol_image)
            !call image_print_info(self%rec(2)%vol_image)
            !call image_print_info(self%rec(1)%vol_amplitude_weights)
            !call image_print_info(self%rec(2)%vol_amplitude_weights)
        endif

        if (debug) then
            print *, '**debug(threed_apply_wiener_filter): about to apply filter'
            print *, wiener_filter_term
        endif


        ! Loop over the complex elements of the reconstruction volume and apply the filter
        do k=1,img%logical_dimensions(3)
            k_real = ctfvol_addr_real_from_comp_dim2or3(img,k)
            if (k_real .eq. img%logical_dimensions(3)+1) cycle
            rad_z = img%LogicalIndexGivenPhysicalIndexInFourierSpace(k,3)**2
            do j=1,img%logical_dimensions(2)
                j_real = ctfvol_addr_real_from_comp_dim2or3(img,j)
                if (j_real .eq. img%logical_dimensions(2)+1) cycle
                rad_y = img%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2)**2
                do i=1,img%logical_dimensions(1)/2+1
                    i_real = ctfvol_addr_real_from_comp_dim1(img,i)
                    rad_x = (i-1)**2
                    current_shell = nint(sqrt(real(rad_x+rad_y+rad_z)))
                    if (current_shell+1 .gt. num_of_shells) cycle
                    if (rec_num .eq. 0) then
                        if (self%rec(1)%vol_amplitude_weights%real_values(i_real,j_real,k_real) + &
                                     self%rec(2)%vol_amplitude_weights%real_values(i_real,j_real,k_real) .eq. 0.0) then
                            img%complex_values(i,j,k) = (0.0,0.0)
                        else
                            img%complex_values(i,j,k) = (self%rec(1)%vol_image%complex_values(i,j,k) + &
                                                         self%rec(2)%vol_image%complex_values(i,j,k)) / &
                                                        (self%rec(1)%vol_amplitude_weights%real_values(i_real,j_real,k_real) + &
                                                         self%rec(2)%vol_amplitude_weights%real_values(i_real,j_real,k_real) &
                                                       + wiener_filter_term(current_shell+1))
                        endif
                    else
                        if (self%rec(rec_num)%vol_amplitude_weights%real_values(i_real,j_real,k_real) .eq. 0.0) then
                            img%complex_values(i,j,k) = (0.0,0.0)
                        else
                            img%complex_values(i,j,k) =  self%rec(rec_num)%vol_image%complex_values(i,j,k) / &
                                                        (self%rec(rec_num)%vol_amplitude_weights%real_values(i_real,j_real,k_real) &
                                                       + wiener_filter_term(current_shell+1))
                        endif
                    endif
                enddo
            enddo
        enddo

        if (debug) print *, '**debug(threed_apply_wiener_filter): done applying filter'

        ! This leaves the 3d reconstruction with origin not at center
        img%object_is_centered_in_box = .false.

        ! The Nyquist elements are not reliable, so we remove them
        call img%ApplyCircularMask(0.5)

        ! The image is ready to be back-FT'ed, but first needs to be shifted
        call img%SwapRealSpaceQuadrants()
        call img%BackwardFFT()

    end subroutine threed_apply_wiener_filter

!    !>  \brief  Estimate the ssnr of a reconstruction based on the comparison of two reconstructions from half datasets.
!    subroutine threed_estimate_ssnr_from_half_dataset_threeds(self,fsc_estimate_above_noise)
!        use images_helices
!        use images_ctf
!        implicit none
!        ! Arguments
!        type(threed),                   intent(inout)       ::  self
!        logical,        optional,       intent(in)          ::  fsc_estimate_above_noise        !<  If true, the fsc estimate is not allowed below a sigma threshold, which will prevent subsequent Wiener filters from being too harsh. Defaults to false.
!        ! Private variables
!        integer                     ::  current_rec
!        type(image)                 ::  half_dataset_rec(2)
!        real(kind=8),   allocatable ::  half_dataset_fsc(:), sum_ctf2(:), fsc_num(:),fsc_denom1(:),fsc_denom2(:)
!        integer,        allocatable ::  voxels_in_shell(:)
!        integer                     ::  i,j,k, i_real,j_real,k_real
!        integer                     ::  num_of_shells, current_shell
!        logical,        parameter   ::  iignore_friedel_mates   =   .true.  ! Some Friedel mates are described explicelty by the fftw array, we should not count them twice!
!        integer                     ::  ierr
!        logical,        parameter   ::  dnf(3) = [.true.,.false.,.false.]
!        real                        ::  current_fsc
!        real                        ::  current_frequency
!        real                        ::  current_ssnr
!        real,           parameter   ::  threed_pssnr_fudge_factor    =  1.0    !1.8    ! To be replaced by proper equation later (when Chuck works it all out)
!        integer                     ::  fsc_below_thresh_shell  ! Shell number at which the fsc falls below 0.143
!        real,           parameter   ::  sigma_factor    =   2.
!        real                        ::  noise_fsc_sigma
!        logical                     ::  ffsc_estimate_above_noise    ! If true, the fsc estimate is not allowed below a sigma threshold, which will prevent subsequent Wiener filters from being too harsh
!        logical,        parameter   ::  debug   =   .false.
!        integer                     ::  rad_x, rad_y, rad_z
!        ! Start work
!
!        ! Check that two reconstructions have been prepared
!        if (threed_num_of_recs(self) .ne. 2) then
!            write(*,'(a)') '**ERROR(threed_estimate_ssnr_from_half_dataset_threeds): number of internal reconstructions = ',  &
!                                threed_num_of_recs(self)
!            call terminate('Fatal error in THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS. Programming error.')
!        endif
!
!        ! Check neither reconstruction is blank
!        if (any(self%rec%num_contrib_elems .eq. 0)) then
!            write(*,'(2a,2(i0,1x))') '**ERROR(threed_estimate_ssnr_from_half_dataset_threeds): at least one of the internal ',    &
!                                    'reconstructions is blank: ', self%rec%num_contrib_elems
!            call terminate('Fatal error in THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS')
!        endif
!
!        ! Will we allow the fsc to go below n sigma?
!        if (present(fsc_estimate_above_noise)) then
!            ffsc_estimate_above_noise = fsc_estimate_above_noise
!        else
!            ffsc_estimate_above_noise = .false.
!        endif
!
!
!
!        ! Check whether we are re-calculating ssnr unnecessarily
!        if (self%ssnr_is_up_to_date) then
!            write(*,'(2a)') '**warning(threed_estimate_ssnr_from_half_dataset_threeds): ssnr estimate appears to ',  &
!                            'already be up-to-date. You may be wasting time by calling this again.'
!        endif
!
!        write(*,'(2a)') '**info(threed_estimate_ssnr_from_half_dataset_threeds): will estimate the ssnr by ',    &
!                        'computing the fsc between the two internal reconstructions'
!
!        if (ffsc_estimate_above_noise) then
!            write(*,'(a,F0.1,a)') '**info(threed_estimate_ssnr_from_half_dataset_threeds): the fsc will not be allowed below ', &
!                                        sigma_factor, ' standard derivations of noise FSCs'
!        endif
!
!        ! Debug
!        if (debug) then
!            call threed_write_to_disk(self,'dbg_3d_raw.mrc')
!        endif
!
!
!        ! Loop over each reconstruction
!        do current_rec=1,threed_num_of_recs(self)
!            write(*,'(2(A,i0))') '**info(threed_estimate_ssnr_from_half_dataset_threeds): evaluating internal reconstruction ', &
!                                    current_rec, ' of ', threed_num_of_recs(self)
!            ! Compute two reconstructions from the half datasets, by straightforward division by CTF^2
!            call threed_eval_reconstruction(self,half_dataset_rec(current_rec),wiener_filter=.false.,rec_num=current_rec)
!            if (.not. half_dataset_rec(current_rec)%object_is_centered_in_box) then
!                call half_dataset_rec(current_rec)%SwapRealSpaceQuadrants()
!            endif
!            if (debug) then
!                if (current_rec==1) call half_dataset_rec(current_rec)%WriteToDisk('dbg_3d_ssnrest_1_nomask.mrc')
!                if (current_rec==2) call half_dataset_rec(current_rec)%WriteToDisk('dbg_3d_ssnrest_2_nomask.mrc')
!            endif
!            ! Compute  & apply a soft mask for each reconstruction
!            write(*,'(2(A,i0))') '**info(threed_estimate_ssnr_from_half_dataset_threeds): masking internal reconstruction ', &
!                                current_rec, ' of ', threed_num_of_recs(self)
!            if (self%point_sym%helical) then
!                ! The mask corresponds to molecular envelope at resolution of ~20A
!                call image_helix_mask(  half_dataset_rec(current_rec),self%helix,self%asym_unit_mw,self%pixel_size,   &
!                                        self%helix_max_rad,F_MASK=SELF%f_mask,F_SMOOTH=SELF%f_smooth,   &
!                                        MASK_RESOLUTION=SELF%pixel_size/20.)
!            else
!                write(*,'(a)') '**ERROR(threed_estimate_ssnr_from_half_dataset_threeds): masking not yet implemented',  &
!                                ' for single-particle reconstructions'
!                call terminate('Feature not implemented yet')
!            endif
!            if (debug) then
!                if (current_rec==1) call half_dataset_rec(current_rec)%WriteToDisk('dbg_3d_ssnrest_1_mask.mrc')
!                if (current_rec==2) call half_dataset_rec(current_rec)%WriteToDisk('dbg_3d_ssnrest_2_mask.mrc')
!            endif
!            ! Go to Fourier space
!            call half_dataset_rec(current_rec)%ForwardFFT()
!        enddo
!
!        ! Check that f_mask and f_smooth have reasonable values
!        if (self%f_mask .eq. 0.0 .or. self%f_smooth .eq. 0.0) then
!            write(*,'(a,2(F0.3,1x))') '**ERROR(threed_estimate_ssnr_from_half_dataset_threeds): f_smooth, f_mask = ', &
!                                    self%f_smooth, self%f_mask
!            call terminate('Fatal error in THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS')
!        endif
!        if (debug) then
!            write(*,'(a,2(F0.3,1x))') '**debug(threed_estimate_ssnr_from_half_dataset_threeds): f_smooth, f_mask = ', &
!                                        self%f_smooth, self%f_mask
!        endif
!
!
!        ! For the following steps, the reconstructions need to be in Fourier space
!        if (half_dataset_rec(1)%is_in_real_space) call half_dataset_rec(1)%ForwardFFT()
!        if (half_dataset_rec(2)%is_in_real_space) call half_dataset_rec(2)%ForwardFFT()
!
!        ! Also, i think we need them to be with real space origin not at center
!        if (half_dataset_rec(1)%object_is_centered_in_box) call half_dataset_rec(1)%SwapRealSpaceQuadrants()
!        if (half_dataset_rec(2)%object_is_centered_in_box) call half_dataset_rec(2)%SwapRealSpaceQuadrants()
!
!        ! Allocate memory for output arrays
!        num_of_shells = half_dataset_rec(1)%GetMaximumRadius()
!        !if (allocated(self%ssnr_estimate)) deallocate(self%ssnr_estimate)
!        call self%ssnr_curve%ClearData()
!        allocate(half_dataset_fsc(num_of_shells), sum_ctf2(num_of_shells),&! self%ssnr_estimate(num_of_shells),&
!                fsc_num(num_of_shells),fsc_denom1(num_of_shells),fsc_denom2(num_of_shells),voxels_in_shell(num_of_shells),stat=IERR)
!        if (ierr .ne. 0) then
!            write(*,'(2a,i0)') '**ERROR(threed_estimate_ssnr_from_half_dataset_threeds): memory allocation for', &
!                                ' intermediate and result arrays failed with error ', ierr
!            call terminate('Fatal error in THREED_ESTIMATE_SSNR_FROM_HALF_DATASET_THREEDS')
!        endif
!        half_dataset_fsc    =   0.0
!        sum_ctf2            =   0.0
!        !self%ssnr_estimate  =   0.0
!        fsc_num             =   0.0
!        fsc_denom1          =   0.0
!        fsc_denom2          =   0.0
!        voxels_in_shell     =   0
!
!        ! Loop over the volume and accumulate terms for the estimation of the ssnr as in Equation 19 of Sindelar (2011):
!        ! ssnr(r) = (1/F_MASK) * n(r) / sum_ctf2(r) * 2. * f_smooth * fsc(r) / (1.-FSC(r))
!        write(*,'(2a)') '**info(threed_estimate_ssnr_from_half_dataset_threeds): accumulating ',   &
!                                'statistics for fsc and ssnr calculations'
!        do i=1,self%logical_dimensions(1)/2+1
!            i_real = ctfvol_addr_real_from_comp_dim1(self%rec(1)%vol_image,i)
!            rad_x = (i-1)**2
!            do j=1,self%logical_dimensions(2)
!                if (iignore_friedel_mates .and. i == 1 .and. j .gt. 1 .and. j .lt. self%logical_dimensions(2)/2+1) cycle   ! Skip voxels whose Friedel mates are explicit in the fftw description
!                j_real = ctfvol_addr_real_from_comp_dim2or3(self%rec(1)%vol_image,j)
!                rad_y = self%rec(1)%vol_image%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2)
!                if (j_real .eq. self%rec(1)%vol_image%logical_dimensions(2)+1) cycle ! Positive Nyquist component is not described in the vol_ctf2 array
!                do k=1,self%logical_dimensions(3)
!                    if (iignore_friedel_mates .and. i == 1 .and. j == 1 .and. k .gt. 1 .and. k .lt. self%logical_dimensions(3)/2+1) cycle
!                    k_real = ctfvol_addr_real_from_comp_dim2or3(self%rec(1)%vol_image,k)
!                    if (k_real .eq. self%rec(1)%vol_image%logical_dimensions(3)+1) cycle
!                    rad_z = self%rec(1)%vol_image%LogicalIndexGivenPhysicalIndexInFourierSpace(k,3)
!
!                    current_shell = 1 + nint(sqrt(real(rad_z+rad_y+rad_x)))
!                    if (current_shell .le. num_of_shells) then
!                        fsc_num(current_shell)          =   fsc_num(current_shell)    &
!                                                         + dble(half_dataset_rec(1)%complex_values(i,j,k) &
!                                                         * conjg(half_dataset_rec(2)%complex_values(i,j,k)))
!                        fsc_denom1(current_shell)       =   fsc_denom1(current_shell) &
!                                                         + dble(abs(half_dataset_rec(1)%complex_values(i,j,k)))**2
!                        fsc_denom2(current_shell)       =   fsc_denom2(current_shell) &
!                                                         + dble(abs(half_dataset_rec(2)%complex_values(i,j,k)))**2
!                        sum_ctf2(current_shell)         =   sum_ctf2(current_shell)     &
!                                                      + dble(self%rec(1)%vol_amplitude_weights%real_values(i_real,j_real,k_real)) &
!                                                      + dble(self%rec(2)%vol_amplitude_weights%real_values(i_real,j_real,k_real))
!                        voxels_in_shell(current_shell)  =   voxels_in_shell(current_shell) + 1
!                    endif
!                enddo
!            enddo
!        enddo
!
!        ! Loop over shells
!        fsc_below_thresh_shell = -1
!        write(*,'(2(A,i0))') '**info(threed_estimate_ssnr_from_half_dataset_threeds): computing SSNR'
!        write(*,'(a5,1x,a6,1x,a6,1x,a6)') 'Shell', 'FSC', 'Sigma', 'PSSNR'
!        do current_shell=1,num_of_shells
!            !>  \todo double-check I'm computing the spatial frequency correctly
!            current_frequency = 0.5 * real(current_shell-1) / real(num_of_shells)
!            current_fsc = fsc_num(current_shell) / sqrt(fsc_denom1(current_shell)*fsc_denom2(current_shell))
!            ! To avoid division by 0 later on, we cap the fsc at 0.999
!            current_fsc = min(current_fsc,0.999)
!            ! Work out the expected variance of fsc between noise volumes at this shell (see for example Eq19 in Van Heel 2005)
!            noise_fsc_sigma = sigma_factor / sqrt(real(voxels_in_shell(current_shell))) / sqrt(2.) / self%f_mask
!            ! If we've just fallen below 0.143, remember what shell we're on
!            if (current_fsc .lt. 0.143 .and. fsc_below_thresh_shell .eq. -1) fsc_below_thresh_shell = current_shell
!            ! If requested, do not allow fsc to fall below n standard deviations of the expected fsc between noise volumes
!            if (ffsc_estimate_above_noise) current_fsc = max(current_fsc,min(noise_fsc_sigma,0.999))
!            if (sum_ctf2(current_shell) .gt. tiny(1.0)) then
!                ! Equation 19 of Sindelar (2011)
!                !self%ssnr_estimate(current_shell) = (1/self%f_mask) * (voxels_in_shell(current_shell)/sum_ctf2(current_shell)) &
!                !                                                * (2.*self%f_smooth*current_fsc/(1-current_fsc))
!                current_ssnr = (1/self%f_mask) * (voxels_in_shell(current_shell)/sum_ctf2(current_shell)) &
!                                                                * (2.*self%f_smooth*current_fsc/(1-current_fsc))
!            else
!                !self%ssnr_estimate(current_shell) = 0.0
!                current_ssnr = 0.0
!            endif
!            ! Apply fudge factor
!            current_ssnr = current_ssnr * threed_pssnr_fudge_factor
!            ! ssnr cannot be negative
!            current_ssnr = max(0.0,current_ssnr)
!            !
!            !self%ssnr_estimate(current_shell) = current_ssnr
!            call self%ssnr_curve%AddPoint(current_frequency,current_ssnr)
!            ! Output
!            write(*,'(i5,1x,F6.3,1x,F6.3,1x,F0.3)') current_shell, current_fsc, noise_fsc_sigma, current_ssnr
!        enddo
!
!        ! Let the user know the estimated internal resolution
!        write(*,'(a,i0,a,i0)') '**info(threed_estimate_ssnr_from_half_dataset_threeds): the fsc fell below 0.143 in shell ', &
!                                fsc_below_thresh_shell-1, ' of ', num_of_shells-1
!
!        ! Deallocate arrays holding intermediate results
!        deallocate(half_dataset_fsc, sum_ctf2,fsc_num,fsc_denom1,fsc_denom2,voxels_in_shell)
!
!        ! Deallocate
!        call half_dataset_rec(1)%Deallocate()
!        call half_dataset_rec(2)%Deallocate()
!
!        ! Now we have an updated pssnr estimate
!        self%ssnr_is_up_to_date = .true.
!
!        write(*,'(2(A,i0))') '**info(threed_estimate_ssnr_from_half_dataset_threeds): all done'
!
!    end subroutine threed_estimate_ssnr_from_half_dataset_threeds

    !>  \brief  Compute the fsc between half-dataset reconstructions and update the estimated PSSNR
    subroutine threed_update_fsc(self,fsc_volumes,fsc_volumes_no_mask,keep_fsc_above_noise_when_estimating_ssnr)
        use images_helices
        use images_ctf, only : ctfvol_addr_real_from_comp_dim2or3, ctfvol_addr_real_from_comp_dim1
        implicit none
        ! Arguments
        type(threed),               intent(inout)   ::  self
        type(image),    optional,   intent(inout)   ::  fsc_volumes(2)          !<  If given, the final fsc volumes will be returned
        type(image),    optional,   intent(inout)   ::  fsc_volumes_no_mask(2)  !<  If given, the final fsc volumes will be returned before any symmetrisation or masking
        logical,        optional,   intent(in)      ::  keep_fsc_above_noise_when_estimating_ssnr
        ! Private variables
        integer                     ::  current_rec
        type(image)                 ::  half_dataset_rec(2)
        real(kind=8),   allocatable ::  fsc_num(:),fsc_denom1(:),fsc_denom2(:), sum_ctf2(:)
        integer,        allocatable ::  voxels_in_shell(:)
        integer                     ::  num_of_shells, current_shell, i,j,k, i_real,j_real,k_real
        integer                     ::  i_logi,j_logi,k_logi
        integer                     ::  fsc_below_thresh_shell
        real                        ::  current_fsc, current_ssnr, current_frequency
        real                        ::  noise_fsc_sigma
        real,           parameter   ::  sigma_factor = 2.0
        logical,        parameter   ::  iignore_friedel_mates   =   .true.  ! Some Friedel mates are described explicelty by the fftw array, we should not count them twice!
        logical,        parameter   ::  dnf(3) = (/.true.,.false.,.false./)
        logical,        parameter   ::  debug   =   .false.
        logical,        parameter   ::  helical_symmetrise = .true.
        logical,        parameter   ::  also_update_ssnr = .false. ! to be true soon
        logical                     ::  kkeep_fsc_above_noise_when_estimating_ssnr
        real                        ::  denom
        ! Start work

        kkeep_fsc_above_noise_when_estimating_ssnr = .false.
        if (present(keep_fsc_above_noise_when_estimating_ssnr)) then
            kkeep_fsc_above_noise_when_estimating_ssnr = keep_fsc_above_noise_when_estimating_ssnr
        endif

        ! Check that two reconstructions have been prepared
        if (threed_num_of_recs(self) .ne. 2) then
            write(*,'(a,i0)') '**ERROR(threed_update_fsc): number of internal reconstructions = ',  &
                                threed_num_of_recs(self)
            call terminate('Fatal error in THREED_UPDATE_FSC. Programming error.')
        endif

        ! Check neither reconstruction is blank
        if (any(self%rec%num_contrib_elems .eq. 0)) then
            write(*,'(a,2(i0,1x))') '**ERROR(threed_update_fsc): at least one of the internal reconstructions is blank: ', &
                                    self%rec%num_contrib_elems
            call terminate('Fatal error in THREED_UPDATE_FSC')
        endif

        write(*,'(a)') '**info(threed_update_fsc): will compute fsc between half-dataset reconstructions'

        ! Check whether we need to be here in the first place
        if (self%fsc_is_up_to_date) then
            write(*,'(a)') '**warning(threed_update_fsc): fsc appears to be up-to-date already'
        endif

        ! Loop over each reconstruction
        do current_rec=1,threed_num_of_recs(self)
            write(*,'(2(A,i0))') '**info(threed_update_fsc): evaluating internal reconstruction ', &
                                    current_rec, ' of ', threed_num_of_recs(self)
            ! Compute two reconstructions from the half datasets
            call threed_eval_reconstruction(self,half_dataset_rec(current_rec),rec_num=current_rec, &
                                            skip_fom_filter=.true.,skip_fsc=.true.,skip_ampl=.true.)
            if (.not. half_dataset_rec(current_rec)%object_is_centered_in_box) then
                call half_dataset_rec(current_rec)%SwapRealSpaceQuadrants()
            endif
            if (debug) then
                if (current_rec==1) call half_dataset_rec(current_rec)%WriteToDisk('dbg_3d_ssnrest_1_nomask.mrc')
                if (current_rec==2) call half_dataset_rec(current_rec)%WriteToDisk('dbg_3d_ssnrest_2_nomask.mrc')
            endif
            if (present(fsc_volumes_no_mask)) then
                fsc_volumes_no_mask(current_rec) = half_dataset_rec(current_rec)
            endif
            ! Compute  & apply a soft mask for each reconstruction
            if (self%point_sym%helical) then
                if (helical_symmetrise) then
                    if (debug) write(*,'(a,F0.3)') '**debug(threed_update_fsc): about to hextend with max rad = ', &
                                                    self%helix_max_rad
                    write(*,'(2(A,i0))') '**info(threed_update_fsc): symmetrising internal reconstruction ', &
                                    current_rec, ' of ', threed_num_of_recs(self)
                    call image_impose_helical_symmetry_new(half_dataset_rec(current_rec),self%helix,self%interpol_scheme,&
                                                        max_radius=self%helix_max_rad)
                endif
                ! The mask corresponds to molecular envelope at resolution of ~20A
                if (debug) write(*,'(a,F0.3)') '**debug(threed_update_fsc): about to hmask with max rad = ', self%helix_max_rad
                write(*,'(2(A,i0))') '**info(threed_update_fsc): masking internal reconstruction ', &
                                    current_rec, ' of ', threed_num_of_recs(self)
                if (debug) then
                    if (current_rec==1) call half_dataset_rec(current_rec)%WriteToDisk('dbg_3d_ssnrest_1_beforehmask.mrc')
                    if (current_rec==2) call half_dataset_rec(current_rec)%WriteToDisk('dbg_3d_ssnrest_2_beforehmask.mrc')
                endif

                call image_helix_mask(  half_dataset_rec(current_rec),self%helix,self%asym_unit_mw,self%pixel_size,   &
                                        self%helix_max_rad,f_mask=self%f_mask,f_smooth=self%f_smooth,   &
                                        mask_resolution=(self%pixel_size/max(10.0,self%estimated_resolution+2.0)))
                if (debug) then
                    if (current_rec==1) call half_dataset_rec(current_rec)%WriteToDisk('dbg_3d_ssnrest_1_hmask.mrc')
                    if (current_rec==2) call half_dataset_rec(current_rec)%WriteToDisk('dbg_3d_ssnrest_2_hmask.mrc')
                endif
            else
                write(*,'(a)') '**ERROR(threed_update_fsc): masking not yet implemented for single-particle reconstructions'
                call terminate('Feature not implemented yet')
            endif
            if (debug) then
                if (current_rec==1) call half_dataset_rec(current_rec)%WriteToDisk('dbg_3d_ssnrest_1_mask.mrc')
                if (current_rec==2) call half_dataset_rec(current_rec)%WriteToDisk('dbg_3d_ssnrest_2_mask.mrc')
            endif
            if (present(fsc_volumes)) then
                fsc_volumes(current_rec) = half_dataset_rec(current_rec)
            endif
            ! Go to Fourier space
            call half_dataset_rec(current_rec)%ForwardFFT()
        enddo

        ! Allocate memory for fsc array
        num_of_shells = half_dataset_rec(1)%GetMaximumRadius()
        !if (allocated(self%fsc)) deallocate(self%fsc)
        if (self%fsc_curve%HasBeenInitialized()) then
            call self%fsc_curve%ClearData()
        else
            call self%fsc_curve%Init(num_of_shells,'Spatial frequency (1/A)','FSC')
        endif
        !if (also_update_ssnr .and. allocated(self%ssnr_estimate)) deallocate(self%ssnr_estimate)
        if (also_update_ssnr) then
            if (self%ssnr_curve%HasBeenInitialized()) then
                call self%ssnr_curve%ClearData()
            else
                call self%ssnr_curve%Init(num_of_shells,'Spatial frequency (1/A)','SSNR')
            endif
        endif
        !allocate(self%fsc(num_of_shells),fsc_num(num_of_shells),fsc_denom1(num_of_shells),fsc_denom2(num_of_shells))
        allocate(fsc_num(num_of_shells),fsc_denom1(num_of_shells),fsc_denom2(num_of_shells))
        !if (also_update_ssnr) allocate(self%ssnr_estimate(num_of_shells))
        allocate(sum_ctf2(num_of_shells),voxels_in_shell(num_of_shells))
        !self%fsc = 0.0
        !if (also_update_ssnr) self%ssnr_estimate = 0.0
        if (also_update_ssnr) call self%ssnr_curve%ClearData()
        fsc_num = 0.0d0
        fsc_denom1 = 0.0d0
        fsc_denom2 = 0.0d0
        sum_ctf2   = 0.0d0
        voxels_in_shell = 0
        write(*,'(2(A,i0))') '**info(threed_update_fsc): accumulating statistics for FSC & PSSNR computation'
        do i=1,self%rec(1)%vol_image%physical_index_of_first_negative_frequency(1)
            i_logi = i-1
            i_real = ctfvol_addr_real_from_comp_dim1(self%rec(1)%vol_image,i)
            do j=1,self%logical_dimensions(2)
                if (iignore_friedel_mates .and. i == 1 .and. j .gt. 1 &
                   .and. j .lt. self%rec(1)%vol_image%physical_index_of_first_negative_frequency(2)) cycle   ! Skip voxels whose Friedel mates are explicit in the fftw description
                j_logi = self%rec(1)%vol_image%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2)
                j_real = ctfvol_addr_real_from_comp_dim2or3(self%rec(1)%vol_image,j)
                if (j_logi .eq. self%rec(1)%vol_image%logical_dimensions(2)+1) cycle ! Positive Nyquist component is not described in the vol_ctf2 array
                do k=1,self%logical_dimensions(3)
                    if (iignore_friedel_mates .and. i == 1 .and. j == 1 .and. k .gt. 1 &
                        .and. k .lt. self%rec(1)%vol_image%physical_index_of_first_negative_frequency(3)) cycle
                    k_logi = self%rec(1)%vol_image%LogicalIndexGivenPhysicalIndexInFourierSpace(k,3)
                    k_real = ctfvol_addr_real_from_comp_dim2or3(self%rec(1)%vol_image,k)
                    if (k_logi .eq. self%rec(1)%vol_image%logical_dimensions(3)+1) cycle

                    current_shell = 1 + nint(sqrt(real(i_logi**2+j_logi**2+k_logi**2)))
                    if (current_shell .le. num_of_shells) then
                        fsc_num(current_shell)          =   fsc_num(current_shell)    &
                                                            + dble(half_dataset_rec(1)%complex_values(i,j,k) &
                                                            * conjg(half_dataset_rec(2)%complex_values(i,j,k)))
                        fsc_denom1(current_shell)       =   fsc_denom1(current_shell) &
                                                            + dble(abs(half_dataset_rec(1)%complex_values(i,j,k)))**2
                        fsc_denom2(current_shell)       =   fsc_denom2(current_shell) &
                                                            + dble(abs(half_dataset_rec(2)%complex_values(i,j,k)))**2
                        sum_ctf2(current_shell)         =   sum_ctf2(current_shell)     &
                                                      + dble(self%rec(1)%vol_amplitude_weights%real_values(i_real,j_real,k_real)) &
                                                      + dble(self%rec(2)%vol_amplitude_weights%real_values(i_real,j_real,k_real))
                        voxels_in_shell(current_shell)  =   voxels_in_shell(current_shell) + 1
                    endif
                enddo
            enddo
        enddo

        ! Loop over shells
        fsc_below_thresh_shell = -1
        write(*,'(2(A,i0))') '**info(threed_update_fsc): computing FSC & PSSNR'
        if (debug) then
            print *, 'voxels_in_shell', voxels_in_shell
            print *, 'fmask', self%f_mask
        endif
        do current_shell=1,num_of_shells
            !>  \todo double-check I'm computing the spatial frequency correctly
            current_frequency = 0.5 * real(current_shell-1) / real(num_of_shells) / self%pixel_size
            !
            denom = sqrt(fsc_denom1(current_shell)*fsc_denom2(current_shell))
            if (denom .eq. 0.0) then
                write(*,'(a,i0,a)') '**warning(threed_update_fsc): fsc denominator for shell ', current_shell, ' is 0.0'
                !self%fsc(current_shell) = 0.0
                current_fsc = 0.0
            else
                !self%fsc(current_shell) = fsc_num(current_shell) / denom
                current_fsc = fsc_num(current_shell) / denom
            endif
            !
            call self%fsc_curve%AddPoint(current_frequency,current_fsc)
            ! Work out the expected variance of fsc between noise volumes at this shell (see for example Eq19 in Van Heel 2005)
            noise_fsc_sigma = sigma_factor / sqrt(real(voxels_in_shell(current_shell))) / sqrt(2.) / self%f_mask
            ! If we've just fallen below 0.143 for the first time, remember what shell we're on
            !if (self%fsc(current_shell) .lt. 0.143 .and. fsc_below_thresh_shell .eq. -1) fsc_below_thresh_shell = current_shell
            if (current_fsc .lt. 0.143 .and. fsc_below_thresh_shell .eq. -1) fsc_below_thresh_shell = current_shell
            ! When computing the SSNR, perhaps need to make sure we don't go to 0.0
            if (kkeep_fsc_above_noise_when_estimating_ssnr) then
                current_fsc = max(current_fsc,min(noise_fsc_sigma,0.999))
            else
                current_fsc = min(current_fsc,0.999)
            endif
            ! Estimate the PSSNR following Sindelar & Grigorieff (2012), Eqns. 3 & 6
            ! Note that in our case, we assume that self%f_mask is also f_particle, i.e. that the masking was done very tightly,
            ! so that the f_mask / f_particle ratio is 1.0 in our case
            if (also_update_ssnr) then
                if (sum_ctf2(current_shell) .gt. tiny(1.0)) then
                    current_ssnr = (voxels_in_shell(current_shell)/sum_ctf2(current_shell)) &
                                                                    * (2.*current_fsc/(1.-current_fsc))
                    !self%ssnr_estimate(current_shell) = (voxels_in_shell(current_shell)/sum_ctf2(current_shell)) &
                    !                                                * (2.*current_fsc/(1.-current_fsc))
                else
                    current_ssnr = 0.0
                    !self%ssnr_estimate(current_shell) = 0.0
                endif
                ! SSNR cannot be negative
                !self%ssnr_estimate(current_shell) = max(0.0,self%ssnr_estimate(current_shell))
                current_ssnr = max(0.0,current_ssnr)
                !
                call self%ssnr_curve%AddPoint(current_frequency,current_ssnr)
            endif
        enddo

        ! Deallocate things
        call half_dataset_rec(1)%Deallocate()
        call half_dataset_rec(2)%Deallocate()

        ! Remember we did this
        self%fsc_is_up_to_date = .true.
        if (also_update_ssnr) self%ssnr_is_up_to_date = .true.

        ! Update estimated resolution
        self%estimated_resolution = fsc_below_thresh_shell / num_of_shells * 0.5 ! In reciprocal pixels
        if (self%estimated_resolution .ne. 0.0) self%estimated_resolution = self%pixel_size / self%estimated_resolution ! In Angstroms
    end subroutine threed_update_fsc

    !>  \brief  Guess a constant (i.e. not frequency-dependent) Wiener filter term, as in Grigorieff (1998) and Frealign or as in Sachse (2007)
    function threed_guess_wiener_filter_term(self,rec_num) result(wiener_filter_term)
        implicit none
        ! Arguments
        type(threed),               intent(in)  ::  self
        integer,                    intent(in)  ::  rec_num
        ! Results
        real,               allocatable         ::  wiener_filter_term(:)
        ! Private variables
        logical,            parameter   ::  debug = .false.
        !logical,            parameter   ::  include_nsr_term_in_wiener  =   .false.
        logical,            parameter   ::  a_la_sachse_2007 = .false. !  Tested on 20-March-2013 and it makes reconstructions worse, so leaving this off.
        real(kind=8)                    ::  sig_ctf
        real                            ::  max_val
        real                            ::  sym_redun
        integer                         ::  num_of_shells
        integer ::  i,j,k, ierr
        integer ::  ori(3)
        ! Start work

        ! Allocate memory for output arrays
        num_of_shells = self%rec(1)%vol_image%GetMaximumRadius()
        if (allocated(wiener_filter_term)) deallocate(wiener_filter_term)
        allocate(wiener_filter_term(num_of_shells),stat=IERR)
        if (ierr .ne. 0) then
            write(*,'(a,i0)') '**ERROR(threed_guess_wiener_filter_term): memory allocation for result arrays failed with error ', &
                                ierr
            call terminate('Fatal error in THREED_GUESS_WIENER_FILTER_TERM')
        endif


        if (a_la_sachse_2007) then
            ! The constant will be 0.01* the maximum value in the ctf2 vol
            max_val = -HUGE(1.0)
            if (    self%rec(1)%vol_amplitude_weights%is_in_real_space &
              .and. self%rec(1)%vol_amplitude_weights%object_is_centered_in_box) then
                ori = self%rec(1)%vol_amplitude_weights%physical_address_of_box_center
            else
                ori = [1,1,1]
            endif
            do K=1,self%rec(1)%vol_amplitude_weights%logical_dimensions(3)
                do J=1,self%rec(1)%vol_amplitude_weights%logical_dimensions(2)
                    do I=1,ori(1)
                        if (i .eq. ori(1) .and. (j .gt. ori(2) .or. k .gt. ori(3))) cycle
                        if (rec_num .eq. 0) then
                            if (  (self%rec(1)%vol_amplitude_weights%real_values(i,j,k) + &
                                   self%rec(2)%vol_amplitude_weights%real_values(i,j,k)) .gt. max_val ) then
                                max_val = self%rec(1)%vol_amplitude_weights%real_values(i,j,k) &
                                        + self%rec(2)%vol_amplitude_weights%real_values(i,j,k)
                            endif
                        else
                            if (  self%rec(rec_num)%vol_amplitude_weights%real_values(i,j,k) .gt. max_val ) then
                                max_val = self%rec(rec_num)%vol_amplitude_weights%real_values(i,j,k)
                            endif
                        endif
                    enddo
                enddo
            enddo
            wiener_filter_term = 0.01*MAX_VAL
        else
            ! Accumulate some statistics which we will use for our fake 1/ssnr term
            sig_ctf = 0.0D0
            if (    self%rec(1)%vol_amplitude_weights%is_in_real_space &
              .and. self%rec(1)%vol_amplitude_weights%object_is_centered_in_box) then
                ori = self%rec(1)%vol_amplitude_weights%physical_address_of_box_center
            else
                ori = [1,1,1]
            endif
            do K=1,self%rec(1)%vol_amplitude_weights%logical_dimensions(3)
                do J=1,self%rec(1)%vol_amplitude_weights%logical_dimensions(2)
                    do I=1,ori(1)
                        if (i .eq. ori(1) .and. (j .gt. ori(2) .or. k .gt. ori(3))) cycle
                        if (rec_num .eq. 0) then
                            sig_ctf = sig_ctf + self%rec(1)%vol_amplitude_weights%real_values(i,j,k) &
                                              + self%rec(2)%vol_amplitude_weights%real_values(i,j,k)
                        else
                            sig_ctf = sig_ctf + self%rec(rec_num)%vol_amplitude_weights%real_values(i,j,k)
                        endif
                    enddo
                enddo
            enddo
            if (debug) then
                print '(a,G15.6)', '**debug(threed_guess_wiener_filter_term): real(sig_ctf) before div by nce = ', real(sig_ctf)
            endif
            if (rec_num .eq. 0) then
                sig_ctf = sig_ctf / dble(self%rec(1)%num_contrib_elems + self%rec(2)%num_contrib_elems)
            else
                sig_ctf = sig_ctf / dble(self%rec(rec_num)%num_contrib_elems)
            endif


            sym_redun = self%point_sym%iredun
            if (debug) then
                print '(a,i0)',  '**debug(threed_guess_wiener_filter_term): self%num_contrib_elems = ',self%rec(1)%num_contrib_elems
                !print '(a,G0.6)', '**debug(threed_guess_wiener_filter_term): sum_double(ctf2_vol) = ', sum_double(ctf2_vol)
                print '(a,G15.6)', '**debug(threed_guess_wiener_filter_term): real(sig_ctf) = ', real(sig_ctf)
                print '(a,G15.3)', '**debug(threed_guess_wiener_filter_term): real(sym_redum) = ', real(sym_redun)
                !print '(a,G15.3)', '**debug(threed_guess_wiener_filter_term): real(self%non_symmetry_redundancy) = ', &
                !                                                            real(self%non_symmetry_redundancy)
            endif

            ! The actual calculation
            wiener_filter_term = real(sig_ctf)*REAL(sym_redun)*0.1

        endif   ! End of test for "a la Sachse 2007"

        ! We may or may not want to include an extra term for non-symmetry redundancy
        !if (include_nsr_term_in_wiener) then
        !    write(*,'(a)') '**debug(threed_guess_wiener_filter_term): including non-symmetry redundancy',   &
        !                    ' in calculation of Wiener-filter term'
        !    wiener_filter_term = WIENER_FILTER_TERM*REAL(self%non_symmetry_redundancy)
        !endif

        if (debug) then
            !write(*,'(a,<SIZE(wiener_filter_term)>(G0.3,x))') '**debug(threed_guess_wiener_filter_term): wiener_filter_term = ', &
            !                                                    wiener_filter_term
        endif
    end function threed_guess_wiener_filter_term



    !>  \brief  Add a slice (image) to a 3d reconstruction
    !!  \todo   Optimise. We are calculating the CTF values twice, and we are phase-shifting the dose filter image twice
    subroutine threed_add_slice(self,slice_in,ctf,angles,shift_x,shift_y,subtract,bfactor,apply_ctf_to_slice, &
                                add_to_threed_num,drift_x,drift_y, &
                                apply_dose_filter,dose_filter_image,dose_filter_image_swapped)
        use ctfs
        use images_ctf
        use images_threed
        use images
        use euleranglesmod
        use images_arithm
        use rotation_matrices
        use runtime_parameters, only : terminate
        use omp_lib
        use imagefiles  ! debug only
        use UsefulFunctions, only : IsOdd, IsEven
        use InterpolatingWindows, only : correct_3d
        implicit none
        ! Arguments
        type(threed),               intent(inout)   ::  self                !<  3d reconstruction to update
        type(image),                intent(in)      ::  slice_in            !<  Image data to insert (will be FT'ed internally if necessary)
        type(ctf_parameters),       intent(in)      ::  ctf                 !<  ctf parameters for that slice
        type(euler),                intent(in)      ::  angles              !<  Euler angles to insert slice
        real(kind=8),   optional,   intent(in)      ::  shift_x             !<  Shift to apply in x to slice before insertion
        real(kind=8),   optional,   intent(in)      ::  shift_y             !<  Shift to apply in y to slice before insertion
        logical,        optional,   intent(in)      ::  subtract            !<  If .TRUE., the slice (image) is subtracted from the reconstruction,
                                                                            !!  rather than added to it. .FALSE. by default.
        real,           optional,   intent(in)      ::  bfactor             !<  Optionally, apply this Bfactor (unitless) to the image when
                                                                            !!  inserting it into the reconstruction.
                                                                            !!  See Grigorieff (1998), Eq 3, and Grigorieff (2007), Eq 7.
                                                                            !!  b factor is expected to be unitless (to get this, take the b factor
                                                                            !!  in A^2 and divide by the square of the pixel size)
        logical,        optional,   intent(in)      ::  apply_ctf_to_slice  !<  Set this to false if you've already applied the ctf
        integer,        optional,   intent(in)      ::  add_to_threed_num   !<  Can be 1 or 2 (ignored unless an fsc is being computed)
        real,           optional,   intent(in)      ::  drift_x,drift_y     !<  Drift vector (if given, corresponding filter will be applied to the slice and recorded in the 3D reconstruction)
        logical,        optional,   intent(in)      ::  apply_dose_filter   !<  Apply dose (radiation damage) filter
        type(Image),    optional,   intent(in)      ::  dose_filter_image   !<  A real-space image of same dimensions as slice_in, with dose filter values. DC component in central pixel.
        type(Image),    optional,   intent(in)      ::  dose_filter_image_swapped   !<  A real-space image of same dimensions as slice_in, with dose filter values. DC component in first pixel.
        ! Private variables
        type(Image)             ::  slice_ctf
        type(Image)             ::  proj_ctf
        type(rotation_matrix)   ::  matrix                      !   Rotation matrix to go from slice coordinates to 3d coordinates
        type(rotation_matrix)   ::  mmatrix                     !   Rotation matrix modified with point group symmetry for multiple slice insertion
        type(rotation_matrix)   ::  matrix_psi_zero, mmatrix_psi_zero
        logical                 ::  ssubtract
        integer                 ::  subtract_factor
        integer                 ::  i,j,nsym
        integer                 ::  counter
        integer,allocatable     ::  isym(:)
        integer                 ::  slice_dim(3)
        logical,    parameter   ::  wipe_threed_on_input    =   .false.
        real                    ::  sshift_x,sshift_y
        logical,    parameter   ::  debug                   =   .false.
        logical,    parameter   ::  ignore_bfactor          =   .false.
        logical,    parameter   ::  fsc_every_other_slice   =   .false.
        integer                 ::  add_slice_to_rec_num
        type(Image)             ::  slice
        integer                 ::  nce
        logical                 ::  aapply_ctf_to_slice
        logical                 ::  aapply_dose_filter
        ! Start work

        slice = slice_in

        ! Will we apply ctf to the slice internally?
        aapply_ctf_to_slice = .true.
        if (present(apply_ctf_to_slice)) aapply_ctf_to_slice = apply_ctf_to_slice

        ! Dose filter?
        aapply_dose_filter = .false.
        if (present(apply_dose_filter)) aapply_dose_filter = apply_dose_filter
        if (aapply_dose_filter .and. (.not. present(dose_filter_image) .or. .not. present(dose_filter_image_swapped))) then
            call terminate('filament_calc_threed: Need dose filter image')
        endif

        ! Work out which reconstruction this slice will be added to
        if (self%compute_fsc) then
            if (threed_num_of_recs(self) .ne. 2) then
                write(*,'(2a,i0)') '**ERROR(threed_add_slice): compute_fsc flag is t, but number of internal ', &
                                                            'reconstructions is not 2: ', &
                                                            threed_num_of_recs(self)
                call terminate('Fatal error in THREED_ADD_SLICE')
            endif
            !$omp   critical (threed_add_slice_decide_which_threed)
            if (present(add_to_threed_num)) then
                add_slice_to_rec_num = add_to_threed_num
            else
                add_slice_to_rec_num = self%add_next_slice_to_rec_num
                ! Update object so that next slice is added to other reconstruction
                if (self%add_next_slice_to_rec_num .eq. 1) then
                    self%add_next_slice_to_rec_num = 2
                else if (self%add_next_slice_to_rec_num .eq. 2) then
                    self%add_next_slice_to_rec_num = 1
                else
                    write(*,'(a)') '**ERROR(threed_add_slice): unexpected value of add_next_slice_to_rec_num = ', &
                                    self%add_next_slice_to_rec_num
                    call terminate('Fatal error in THREED_ADD_SLICE. Programming error.')
                endif
            endif
            !$omp   end critical (threed_add_slice_decide_which_threed)
        else
            add_slice_to_rec_num = 1
        endif

        if (debug) write(*,'(a,3(F0.2,1x),a,i0)') '**debug(threed_add_slice): adding slice with euler angles ', &
                                                angles%phi, angles%theta, angles%psi, ' to reconstruction ', add_slice_to_rec_num

        ! Will we be ignoring the b factor? (debugging)
        if (ignore_bfactor .and. present(bfactor)) then
            write(*,'(a)') '**debug(threed_add_slice): will ignore b factor'
        endif

        ! Check ctf is in pixels
        if (.not. ctf%pix_units) then
            write(*,'(a)') '**ERROR(threed_add_slice): need ctf in pixel units.'
            call terminate('Fatal error in THREED_ADD_SLICE')
        endif
        ! Check the threed object is allocated
        if (.not. threed_allocated(self)) then
            write(*,'(a)') '**ERROR(threed_add_slice): 3d object not allocated'
            call terminate('Fatal error in THREED_ADD_SLICE')
        endif
        ! Check the image object is allocated
        if (.not. slice%IsAllocated()) then
            write(*,'(a)') '**ERROR(threed_add_slice): slice object not allocated'
            call terminate('Fatal error in THREED_ADD_SLICE')
        endif
        ! Check the first dimension is even
        if (IsOdd(slice%logical_dimensions(1))) then
            write(*,'(a,i0)') '**ERROR(threed_add_slice): odd first dimension for slice is not supported: ', &
                                slice%logical_dimensions(1)
            call terminate('Fatal error in THREED_ADD_SLICE')
        endif
        ! Are we subtracting?
        ssubtract = .false.
        if (present(subtract))  ssubtract = subtract

        ! If we're subtracting, we need to have a -1 later
        subtract_factor = 1
        if (ssubtract) subtract_factor = -1

        ! Shifts in x and y default to 0.0 if not given
        sshift_x = 0.0
        sshift_y = 0.0
        if (present(shift_x)) sshift_x = shift_x
        if (present(shift_y)) sshift_y = shift_y


        !
        ! Wipe 3d on input if required
        !
        if (wipe_threed_on_input) then
            ! Note: i am not sure that the single OpenMP directive garantees that the wipe will be done only by the first thread to get here.
            ! Therefore, i wouldn't necessarily trust results from this routine when run with threading
            if (omp_get_active_level() .gt. 0) then
                write(*,'(3a)') '**WARNING(threed_add_slice): this portion of the code has not been tested in parallel. ',   &
                                    'Results may be wrong. ', &
                                    'See in-source comment for more details.'
            endif
            !$omp   single
            call threed_wipe(self)
            !$omp   end single
        endif


        ! If the input image needs to be FT'ed, do it
        if (slice%is_in_real_space) then
            call slice%ForwardFFT()
        endif

        !
        ! Apply the ctf to the slice
        !
        if (aapply_ctf_to_slice) then
            proj_ctf%logical_dimensions(1) = slice%physical_upper_bound_complex(1)
            proj_ctf%logical_dimensions(2) = slice%logical_dimensions(2)
            proj_ctf%logical_dimensions(3) = 1
            proj_ctf%is_in_real_space = .true.
            proj_ctf%object_is_centered_in_box = .false.
            call proj_ctf%Allocate()
            call ctf_to_img(ctf,proj_ctf,object_is_centered_in_box = .false.)
            forall(i=1:proj_ctf%logical_dimensions(1),j=1:proj_ctf%logical_dimensions(2))
                slice%complex_values(i,j,1) = slice%complex_values(i,j,1) * proj_ctf%real_values(i,j,1)
            end forall
            call proj_ctf%Deallocate()
        endif

        !
        ! Apply drift filter to the slice
        !
        if (present(drift_x) .and. present(drift_y)) then
            call slice%ApplyDriftFilter(drift_x,drift_y)
        endif

        !
        ! Apply dose filter to the slice
        !
        if (aapply_dose_filter) then
            do j=1,slice%GetPhysicalUpperBoundComplex(2)
                do i=1,slice%GetPhysicalUpperBoundComplex(1)
                    slice%complex_values(i,j,1) = slice%complex_values(i,j,1) * dose_filter_image_swapped%real_values(i,j,1)
                enddo
            enddo
        endif


        !
        ! Apply B-factor to the slice
        !
        if (present(bfactor) .and. .not. ignore_bfactor) then
            if (bfactor .ne. 0.) then
                if (slice%is_in_real_space) call slice%ForwardFFT()
                if (debug) write(*,'(a,F0.3,a)') '**debug(threed_add_slice): applying b factor of ', bfactor, &
                                                                ' to slice before adding it to the reconstruction'
                call slice%ApplyBFactor(bfactor,lp_cosine=.false.,print_filter_shape=.false.,squared=.true.)
            endif
        endif

        !
        ! Remove anything beyond the Nyquist frequency in the slice
        !
        if (.not. IsEven(slice%logical_dimensions(1))) call terminate('Code only written for even dimensions')
        if (slice%is_in_real_space) call slice%ForwardFFT()
        call slice%ApplyCircularMask(0.5)

        !
        ! Apply shift to the slice
        !
        if ((abs(sshift_x) + abs(sshift_y)) .gt. 0.0) then
            call slice%PhaseShift(sshift_x,sshift_y,0.0)
        endif


        ! For the next few operations, we need the slice to have its real-space origin in the center
        if (.not. slice%ObjectIsCenteredInBox()) call slice%SwapRealSpaceQuadrants()

        !
        ! Reverse ft
        !
        call slice%BackwardFFT()

        !
        ! Crop or pad the box if necessary to match the dimension of the 3d reconstruction
        !
        slice_dim = self%logical_dimensions
        slice_dim(3) = 1
        call slice%Resize(slice_dim)

        ! Check the slice has even dimensions
        if (.not. slice%HasEvenDimensions()) then
            write(*,'(a,2(i0,1x))') '**ERROR(threed_add_slice): slice does not have even dimensions: ', &
                                    slice%logical_dimensions(1:2)
            call terminate('Fatal error in THREED_ADD_SLICE')
        endif

        !
        ! Apply "degridding" if necessary
        !
        !if (self%get2d_put3d) then
        !    call correct_3d(slice,self%interpol_scheme%interpolation_instr_function,    &
        !                    self%interpol_scheme%windowing_instr_function,  &
        !                    self%interpol_scheme%window_width)
        !endif


        !
        ! Forward ft back to reciprocal space
        !
        call slice%ForwardFFT()

        !
        ! Phase shift the slice so its real-space origin is not in the center (needed for the insertion into the 3d vol)
        !
        if (slice%ObjectIsCenteredInBox()) call slice%SwapRealSpaceQuadrants()

        !
        ! Generate the ctf image corresponding to that slice
        !
        slice_ctf%logical_dimensions = slice%logical_dimensions
        slice_ctf%object_is_centered_in_box = .true.
        slice_ctf%is_in_real_space = .true.
        call slice_ctf%Allocate()
        call ctf_to_img(ctf,slice_ctf,object_is_centered_in_box=.true.)
        if (debug) then
            call ctf_print_info(ctf)
            call slice_ctf%WriteToDisk('debug_slice_ctf1.mrc')
        endif
        ! Apply the drift filter (before squaring also)
        if (present(drift_x) .and. present(drift_y)) then
            call slice_ctf%ApplyDriftFilter(drift_x,drift_y)
        endif
        ! Apply dose filter to the CTF
        if (aapply_dose_filter) then
            do j=1,slice_ctf%GetLogicalDimension(2)
                do i=1,slice_ctf%GetLogicalDimension(1)
                    slice_ctf%real_values(i,j,1) = slice_ctf%real_values(i,j,1) * dose_filter_image%real_values(i,j,1)
                enddo
            enddo
        endif
        ! Square the ctf
        !! \todo add an optional argument to ctf_to_img so that it returns squared ctf already
        slice_ctf%real_values = slice_ctf%real_values**2
        ! Apply Bfactor to the CTF^2 slice
        if (present(bfactor) .and. .not. ignore_bfactor) then
            if (bfactor .ne. 0.) then
                if (debug) write(*,'(a,F0.3,a)') '**debug(threed_add_slice): applying b factor of ', bfactor, &
                                                ' to slice ctf before adding it to the reconstruction'
                call slice_ctf%ApplyBFactor(bfactor,lp_cosine=.false.,print_filter_shape=.false.,squared=.true.)
            endif
        endif
        ! Mask the ctf to remove stuff beyond Nyq
        call slice_ctf%ApplyCircularMask(real(slice_ctf%logical_dimensions(1))*0.5e0)

        if (debug) then
            call slice_ctf%WriteToDisk('debug_slice_ctf.mrc')
        endif

        !
        ! Work out the rotation matrix for the insertion
        !
        matrix%rmatrix          = euler_to_rotmat(angles%phi,angles%theta,angles%psi,reverse=.true.)
        matrix_psi_zero%rmatrix = euler_to_rotmat(angles%phi,angles%theta,0.0d0     ,reverse=.true.)

        !
        ! Do the insertions; insert iredun times for symmetries not equal to c1
        !
        nsym = size(self%point_sym%symop)
        allocate(isym(nsym))
        do i = 1,nsym
            isym(i) = 0
        enddo


        counter = 1
        do while (counter.le.nsym)
            mmatrix = matrix
            mmatrix_psi_zero = matrix_psi_zero
            do i = 1,nsym
                if (isym(i).ne.self%point_sym%symop(i)%jmax) then
                    do j = 1,isym(i)
                        mmatrix%rmatrix          =  matmul(self%point_sym%symop(i)%rmatrix,mmatrix%rmatrix)
                        mmatrix_psi_zero%rmatrix =  matmul(self%point_sym%symop(i)%rmatrix,mmatrix_psi_zero%rmatrix)
                    enddo
                endif
            enddo
            if (debug) then
                write(*,'(2(a,i0),2(a,2(i0,1x)))') '**debug(threed_add_slice): on symmetry axis  ', counter, ' of ', nsym, &
                        ' rotation ', isym(:), ' of ', self%point_sym%symop(:)%jmax
                call rotation_matrix_print(mmatrix)
            endif

            ! Debug: we expect the volume to have its real-space origin in the corner
            if (self%rec(add_slice_to_rec_num)%vol_image%object_is_centered_in_box) then
                write(*,'(a)') '**warning(threed_add_slice): unexpectedly, volume has real-space origin in center'
            endif


            ! We are putting/spreading into 3d
            call self%rec(add_slice_to_rec_num)%vol_image%InsertComplexSlice(mmatrix%rmatrix,slice,&
                            0.0,0.0,0.0,0.5,nce)
            if (self%reconstruct_amplitudes) then
                call insert_slice_put(self%rec(add_slice_to_rec_num)%vol_ampl, slice,     mmatrix, &
                                           self%interpol_scheme,subtract=ssubtract,num_contrib_elems=nce,amplitudes_only=.true.)
            endif
            call self%rec(add_slice_to_rec_num)%vol_amplitude_weights%InsertRealSlice(mmatrix%rmatrix,slice_ctf)


            if (debug) then
                call self%rec(add_slice_to_rec_num)%vol_amplitude_weights%WriteToDisk('debug_vol_ctf2.mrc')
            endif


            ! Update the number of contributing elements
            !$omp   critical        (threed_sum_num_contrib_elems)
            self%rec(add_slice_to_rec_num)%num_contrib_elems = self%rec(add_slice_to_rec_num)%num_contrib_elems + nce
            !$omp   end critical    (threed_sum_num_contrib_elems)

            counter = 1
            do while (counter.le.nsym)
                isym(counter) = isym(counter)+1
                if (isym(counter).ge.self%point_sym%symop(counter)%jmax) then
                    isym(counter) = 0
                    counter = counter+1
                else
                    exit
                endif
            enddo
        enddo

        ! Now that we've added data, our pssnr estimate is probably thrown off and so is our fsc
        self%ssnr_is_up_to_date = .false.
        self%fsc_is_up_to_date = .false.
    end subroutine threed_add_slice


end module threeds
