!>  \brief  Defines the threed class, which handles 3d reconstruction from a set of images,
!!          including appropriate filters etc.
!!
module threeds_base
    use Images
    use interpolation
    use point_symmetries
    use helices, only : helical_symmetry
    use runtime_parameters, only : terminate
    use Curves
    implicit none

    !private :: threed_reconstruction

    type threed_reconstruction
        !private
        type(image)                 ::  vol_image
        type(image)                 ::  vol_amplitude_weights
        type(image)                 ::  vol_ampl                                !<  Volume for the reconstruction of the amplitudes (optional)
        integer(kind=8)             ::  num_contrib_elems       =   0           !<  Number of contributing elements
    end type threed_reconstruction

    !>  \brief  a threed object includes all necessary data to generate a final 3d reconstruction, but keep track of intermediate
    !!          stages, for example, CTF**2 volume, half-dataset reconstructions etc.
    !!
    !!  \todo   Make FSC and SSNR_estimate Curve objects to make this more portable (otherwise just changing the box size makes them invalid
    type threed
        !private

        type(threed_reconstruction),    allocatable ::  rec(:)                  !<  Component reconstructions. Usually 2 (one for each half-dataset reconstructions) or just 1.

        integer                     ::  logical_dimensions(3)   =   [1,1,1]     !<  Dimensions of the 3d reconstrction
        type(point_symmetry)        ::  point_sym                               !<  Point group symmetry matrices
        !integer                     ::  non_symmetry_redundancy =   0           !<  Non-symmetry redundancy. For example, if the same slice is inserted several times for pseudo-symmetry.
        logical                     ::  in_memory               =   .false.     !<  Whether memory has been allocated to the 3d object
        type(interpolation_scheme)  ::  interpol_scheme                         !<  The interpolation functions & parameters used to calculate the 3d reconstruction
        real                        ::  mask_rad                =   .99         !<  The radius of the mask to impose on the 3d reconstruction. In pixels.
        real                        ::  mask_falloff            =   .1          !<  The width (in pixels) of the dropoff in the mask from 1.0 to 0.0. The mask reaches 0.0 and MASK_RAD.
        ! ssnr
        logical                     ::  compute_fsc             =   .false.     !<  Whether an fsc should be calculated. If so, two reconstructions will be accumulated.
        logical                     ::  fom_filter              =   .false.     !<  Whether a Figure of Merit filter should be applied to the reconstruction (see Henderson & Rosenthal 2003; equivalent flag in Frealign = fcref)
        logical                     ::  use_ssnr_estimate       =   .false.     !<  Whether an estimate of the ssnr should be used as part of the Wiener filter
        integer                     ::  add_next_slice_to_rec_num   =   1       !<  To which internal reconstruction should the next slice be added? Only useful in the context of half-dataset reconstructions, FSCs, SSNR estimation etc.
        !real,       allocatable     ::  fsc(:)                                  !<  The fsc computed between half-dataset reconstructions
        type(Curve)                 ::  fsc_curve                               !<  The FSC computed between half-dataset reconstructions
        logical                     ::  fsc_is_up_to_date       =   .false.     !<  Whether the fsc array is up to date
        !real,       allocatable     ::  ssnr_estimate(:)                        !<  The particle SSNR, as defined by Sindelar & Grigorieff (2012), Eqns. 3 and 6. An estimate of the SSNR of particle images that went into the 3d reconstruction.
        type(Curve)                 ::  ssnr_curve                              !<  The particle SSNR, as defined by Sindelar & Grigorieff (2012), Eqns. 3 and 6. An estimate of the SSNR of particle images that went into the 3d reconstruction.
        logical                     ::  ssnr_is_up_to_date      =   .false.     !<  Whether the PSSNR estimate is up-to-date
        real                        ::  f_mask                  =   0.0         !<  The fraction of voxels within the binary mask
        real                        ::  f_smooth                =   0.0         !<  The fraction of voxels within the smooth mask (computed a la Sindelar, 2011)
        real                        ::  estimated_resolution    =   0.0         !<  Estimated resolution of the 3d reconstruciton (Angstroms)
        ! Masking
        type(helical_symmetry)      ::  helix                                   !<  Helical symmetry of the reconstruction
        real                        ::  asym_unit_mw            =   0.0         !<  Expected molecular weight (Da) of the asymmetric unit
        real                        ::  pixel_size              =   0.0         !<  Voxel size (Angstroms)
        real                        ::  helix_max_rad           =   0.0         !<  Maximum radius of the helical assembly (pixels)
        ! Amplitudes reconstruction
        logical                     ::  reconstruct_amplitudes  =   .false.     !<  Compute a reconstruction using only the input's amplitudes
        ! Experimental properties
        !integer                     ::  oversampling_factor     =   1           !<  Oversampling factor during 3d reconstruction.
                                                                                !!  This can be achieved, for example, by padding in real space (if reconstruction done in reciprocal space). Value of 1 means no oversampling.
        !logical                     ::  get2d_put3d             =   .false.     !<  Whether to use the experimental way, which interpolates in the slice's grid, and then samples in 3d using nearest-neighbour. If .FALSE., the regular way is used.
        contains
        !   procedure   ::  add_slice   =>  threed_add_slice
        !   procedure   ::  eval_reconstruction =>  threed_eval_reconstruction
        !    procedure   ::  get_interpol_scheme =>  threed_get_interpol_scheme
        !    procedure   ::  get_window_width    =>  threed_get_window_width
            final       ::  threed_deallocate
    end type threed

    interface is_allocated
        module procedure threed_allocated
    end interface

    !interface deallocate
    !   module procedure threed_deallocate
    !end interface

    !interface assignment(=)
    !    module procedure threed_assignment
    !    module procedure threed_reconstruction_assignment
    !end interface

    interface init
        module procedure threed_init
    end interface

    contains

    !>  \brief  Print out basic information about the threed object
    subroutine threed_print_info(self)
        implicit none
        ! Arguments
        type(threed),                   intent(in)      ::  self
        ! Private variables

        ! Start work
        write(*,'(a,3(i0,1x))') '**info(threed_print_info): dim = ', self%logical_dimensions
    end subroutine threed_print_info

    subroutine threed_assignment(lhs,rhs)
        implicit none
        ! Arguments
        type(threed),                   intent(inout)   ::  lhs
        type(threed),                   intent(in)      ::  rhs
        ! Private variables
        integer ::  current_rec
        ! Start work

        !
        lhs%logical_dimensions                         =   rhs%logical_dimensions
        call threed_allocate(lhs,threed_num_of_recs(rhs))

        do current_rec=1,size(lhs%rec)
            call threed_reconstruction_assignment(lhs%rec(current_rec),rhs%rec(current_rec))
        enddo

        lhs%point_sym                   =   rhs%point_sym
        !lhs%non_symmetry_redundancy     =   rhs%non_symmetry_redundancy
        lhs%in_memory                   =   rhs%in_memory
        lhs%interpol_scheme             =   rhs%interpol_scheme
        lhs%mask_rad                    =   rhs%mask_rad
        lhs%mask_falloff                =   rhs%mask_falloff


        lhs%compute_fsc                 =   rhs%compute_fsc
        lhs%fom_filter                  =   rhs%fom_filter
        lhs%use_ssnr_estimate           =   rhs%use_ssnr_estimate
        lhs%add_next_slice_to_rec_num   =   rhs%add_next_slice_to_rec_num
        !if (allocated(rhs%fsc)) then
        !    if (allocated(lhs%fsc)) deallocate(lhs%fsc)
        !    allocate(lhs%fsc(size(rhs%fsc)))
        !    lhs%fsc                         =   rhs%fsc
        !endif
        lhs%fsc_curve                   =   rhs%fsc_curve
        lhs%fsc_is_up_to_date           =   rhs%fsc_is_up_to_date
        !if (allocated(rhs%ssnr_estimate)) then
        !    if (allocated(lhs%ssnr_estimate)) deallocate(lhs%ssnr_estimate)
        !    allocate(lhs%ssnr_estimate(size(rhs%ssnr_estimate)))
        !    lhs%ssnr_estimate               =   rhs%ssnr_estimate
        !endif
        lhs%ssnr_curve                  =   rhs%ssnr_curve
        lhs%ssnr_is_up_to_date          =   rhs%ssnr_is_up_to_date
        lhs%f_mask                      =   rhs%f_mask
        lhs%f_smooth                    =   rhs%f_smooth


        lhs%helix                       =   rhs%helix
        lhs%asym_unit_mw                =   rhs%asym_unit_mw
        lhs%pixel_size                  =   rhs%pixel_size
        lhs%helix_max_rad               =   rhs%helix_max_rad

        lhs%reconstruct_amplitudes      =   rhs%reconstruct_amplitudes

        !lhs%oversampling_factor         =   rhs%oversampling_factor

        !lhs%get2d_put3d                 =   rhs%get2d_put3d

    end subroutine threed_assignment

    subroutine threed_reconstruction_assignment(lhs,rhs)
        implicit none
        ! Arguments
        type(threed_reconstruction),    intent(inout)   ::  lhs
        type(threed_reconstruction),    intent(in)      ::  rhs
        ! Private variables
        logical     ::  reconstruct_amps
        ! Start work
        reconstruct_amps = product(rhs%vol_ampl%logical_dimensions) .gt. 1
        call threed_reconstruction_deallocate(lhs)
        call threed_reconstruction_allocate(lhs,rhs%vol_image%logical_dimensions,reconstruct_amps)
        lhs%vol_image                       = rhs%vol_image
        lhs%vol_amplitude_weights           = rhs%vol_amplitude_weights
        if (reconstruct_amps) lhs%vol_ampl  = rhs%vol_ampl
        lhs%num_contrib_elems               = rhs%num_contrib_elems
    end subroutine threed_reconstruction_assignment

    subroutine threed_reconstruction_deallocate(self)
        implicit none
        ! Arguments
        type(threed_reconstruction),    intent(inout)   ::  self
        ! Private variables

        ! Start work
        call self%vol_image%Deallocate()
        call self%vol_amplitude_weights%Deallocate()
        call self%vol_ampl%Deallocate()
        self%num_contrib_elems = 0
    end subroutine threed_reconstruction_deallocate

    subroutine threed_reconstruction_allocate(self,dims,reconstruct_amplitudes)
        use images
        implicit none
        ! Arguments
        type(threed_reconstruction),    intent(inout)   ::  self
        integer,                        intent(in)      ::  dims(3)
        logical,                        intent(in)      ::  reconstruct_amplitudes
        ! Private variables
        logical,                        parameter       ::  debug   =   .false.
        ! Start work
        ! Allocate memory for the image volume, which is complex
        if (debug) write(*,'(a,3(i0,1x))') '**debug(threed_reconstruction_allocate): on entry, dims= ', dims
        self%vol_image%logical_dimensions = dims
        self%vol_image%is_in_real_space = .false.
        if (debug) then
            write(*,'(a)') '**debug(threed_reconstruction_allocate): about to allocate image vol'
            call self%vol_image%PrintInfo()
        endif
        call self%vol_image%Allocate()
        ! Allocate memory to the CTF**2 volume, which is real
        self%vol_amplitude_weights%logical_dimensions = dims
        self%vol_amplitude_weights%is_in_real_space = .true.
        if (debug) write(*,'(a)') '**debug(threed_reconstruction_allocate): about to allocate ctf2 vol'
        call self%vol_amplitude_weights%Allocate()
        ! Allocate memory to the amplitudes volume, which is real
        if (reconstruct_amplitudes) then
            self%vol_ampl%logical_dimensions = dims
            self%vol_ampl%is_in_real_space = .false.
            if (debug) write(*,'(a)') '**debug(threed_reconstruction_allocate): about to allocate ampl vol'
            call self%vol_ampl%Allocate()
        endif

    end subroutine threed_reconstruction_allocate

    pure logical function threed_reconstruction_allocated(self)
        implicit none
        ! Arguments
        type(threed_reconstruction),    intent(in)  ::  self
        ! Private variables

        ! Start work
        threed_reconstruction_allocated = self%vol_image%IsAllocated() .and. &
                                          self%vol_amplitude_weights%IsAllocated()
    end function threed_reconstruction_allocated


    pure logical function threed_allocated(self)
        implicit none
        ! Argument
        type(threed),   intent(in)  ::  self
        ! Start work
        if (allocated(self%rec)) then
            threed_allocated = threed_reconstruction_allocated(self%rec(1))
            if (size(self%rec) .eq. 2) then
                threed_allocated = threed_allocated .and. threed_reconstruction_allocated(self%rec(2))
            endif
            threed_allocated = threed_allocated .and. self%in_memory
        else
            threed_allocated = .false.
        endif
    end function threed_allocated

    !>  \brief  Allocate memory and initialise a 3d reconstruction object. The interpolation can be passed as an object, or descriptors can be supplied.
    subroutine threed_init( self,mask_rad,mask_falloff,interpol_descr,window_descr,window_width,interpol_scheme,    &
                            !oversampling_factor, &
                            sym_symbol, &
                            !non_symmetry_redundancy, &
                            filename_root,use_ssnr_estimate,compute_fsc, &
                            helix,asym_unit_mw,pixel_size,helix_max_rad,num_of_recs,fom_filter,reconstruct_amplitudes, &
                            estimated_resolution)
        use runtime_parameters, only : is_odd, terminate, default_falloff_pixels
        implicit none
        ! Argument
        type(threed),                               intent(inout)   ::  self                    !<  3d reconstruction object to allocate & initialise
        real,                           optional,   intent(in)      ::  mask_rad                !<  The radius of the mask to impose on the slices and the 3d reconstruction. In pixels.
        real,                           optional,   intent(in)      ::  mask_falloff            !<  The width (in pixels) of the dropoff in the mask from 1.0 to 0.0. The mask reaches 0.0 and MASK_RAD.
        character(len=*),               optional,   intent(in)      ::  interpol_descr          !<  a string which describes the interpolation function to use for the 3d
        character(len=*),               optional,   intent(in)      ::  window_descr            !<  a string which describes the windowing function to use in the 3d reconstruction
        integer,                        optional,   intent(in)      ::  window_width            !<  The window width to use during 3d reconstruction
        type(interpolation_scheme),     optional,   intent(in)      ::  interpol_scheme         !<  Interpolation scheme (contains information about interpolation
        !integer,                        optional,   intent(in)      ::  oversampling_factor     !<  Oversampling factor (only used when following get2d_put3d logic)
        type(point_symmetry_symbol),    optional,   intent(in)      ::  sym_symbol              !<  Point group symmetry using Schonflies notation
        !integer,                        optional,   intent(in)      ::  non_symmetry_redundancy !<  Non-symmetry redundancy. For example, if the same slice is inserted several times for pseudo-symmetry.
        character(len=*),               optional,   intent(in)      ::  filename_root           !<  Root filename for the 3d reconstruction. If given, will be read from disk.
        logical,                        optional,   intent(in)      ::  use_ssnr_estimate       !<  Whether an estimate of the ssnr should be used in the Wiener filter
        logical,                        optional,   intent(in)      ::  compute_fsc             !<  Whether an fsc should be computed internally.
        type(helical_symmetry),         optional,   intent(in)      ::  helix                   !<  Helical symmetry of the reconstruction
        real,                           optional,   intent(in)      ::  asym_unit_mw            !<  Expected molecular weight (Da) of the asymmetric unit
        real,                           optional,   intent(in)      ::  pixel_size              !<  Voxel size (Angstroms)
        real,                           optional,   intent(in)      ::  helix_max_rad           !<  Maximum radius of the helical assembly (pixels)
        integer,                        optional,   intent(in)      ::  num_of_recs             !<  Use this to override the number of internal reconstructions
        logical,                        optional,   intent(in)      ::  fom_filter              !<  Whether a fom filter should be applied to the 3d reconstruction (a la Frelaign FCREF; Henderson & Rosenthal 2003)
        logical,                        optional,   intent(in)      ::  reconstruct_amplitudes  !<  Compute a 3d reconstruction using only the 2d amplitudes
        real,                           optional,   intent(in)      ::  estimated_resolution    !<  Estimated resolution (Angstroms) of the 3d reconstruction
        ! Private variables
        type(point_symmetry_symbol)                                 ::  ssym_symbol             !<  Point group symmetry using Schonflies notation, local
        integer                                                     ::  num_of_recs_needed, current_rec
        integer                                                     ::  num_of_recs_on_disk, nnum_of_recs
        integer                                                     ::  ierr
        logical,                        parameter                   ::  debug   =   .false.
        ! Start work

        ! Check how the user defined the interpolation scheme, and if necessary initialise the interpolation_scheme object
        if (present(interpol_scheme)) then
            self%interpol_scheme = interpol_scheme
        else
            if (present(window_descr) .and. present(window_width) .and. present(interpol_descr)) then
                call interpolation_scheme_init(self%interpol_scheme,interpol_descr,window_descr,window_width)
            else
                write(*,'(2a)') '**info(threed_init): no interpolation scheme information supplied.',    &
                                ' Make sure all required arguments are given. Defaulting to linear, rect, 2'
                call interpolation_scheme_init(self%interpol_scheme,'LINEAR','RECT',2)
            endif
        endif

        ! Use ssnr estimate?
        if (present(use_ssnr_estimate))  self%use_ssnr_estimate = use_ssnr_estimate

        ! Compute FSC?
        if (present(compute_fsc))  self%compute_fsc = compute_fsc

        ! Apply fom filter?
        if (present(fom_filter)) self%fom_filter = fom_filter
        if (self%fom_filter) self%compute_fsc = .true.

        ! Reconstruct amplitudes?
        if (present(reconstruct_amplitudes)) self%reconstruct_amplitudes = reconstruct_amplitudes

        ! We will need to compute an fsc to estimate the ssnr
        if (self%use_ssnr_estimate .and. .not. self%compute_fsc) then
            write(*,'(a)') '**warning(threed_init): will need to compute an fsc to estimate the SSNR.', &
                            ' Overriding and setting compute_fsc to true'
            self%compute_fsc = .true.
        endif

        ! Will we need two reconstructions (e.g. to compute fsc)?
        if (self%compute_fsc) then
            num_of_recs_needed = 2
        else
            num_of_recs_needed = 1
        endif

        ! If we'll be reading from disk, how many internal reconstructions are stored there?
        if (present(filename_root)) then
            num_of_recs_on_disk = threed_on_disk_num_of_recs(filename_root)
        else
            num_of_recs_on_disk = 0
        endif

        ! How many internal reconstructions will the 3d have?
        if (present(num_of_recs)) then
            nnum_of_recs = num_of_recs
        else
            nnum_of_recs = max(num_of_recs_needed,num_of_recs_on_disk)
        endif

        ! Now that we know how many internal reconstructions we have, we may overwrite the default value for use_ssnr_estimate
!        if (.NOT. present(use_ssnr_estimate)) then
!            if (nnum_of_recs .EQ. 2) THEN
!                write(*,'(a)') '**warning(threed_init): 2 internal reconstructions therefore will use ssnr estimate via FSC. Overriding compute_fsc and USE_SSNR_ESTIMATE'
!                self%use_ssnr_estimate = .TRUE.
!                self%compute_fsc = .TRUE.
!            else
!                self%use_ssnr_estimate = .FALSE.
!            endif
!        endif

        ! Deallocate memory
        if (debug) write(*,'(a)') '**debug(threed_init): about to call threed_deallocate'
        call threed_deallocate(self)

        ! Read from disk if a filename was given
        if (present(filename_root)) then
            if (num_of_recs_on_disk .lt. num_of_recs_needed) then
                write(*,'(a,i0,a,a,a,i0,a)') '**ERROR(threed_init): ', num_of_recs_needed , &
                                            ' 3d reconstructions are needed, but the reconstruction on disk (root filename ', &
                                            trim(adjustl(filename_root)), ') only has ', num_of_recs_on_disk , ' reconstruction'
                call terminate('Fatal error in THREED_INIT')
            endif
            if (debug) write(*,'(a)') '**debug(threed_init): about to allocate rec array'
            allocate(self%rec(nnum_of_recs),stat=IERR)
            if (ierr .ne. 0) then
                write(*,'(a,i0)') '**ERROR(threed_init): rec array allocation failed with error ', ierr
                call terminate('Failed memory allocation in THREED_INIT')
            endif
            ! Read from disk
            call threed_read_from_disk(self,filename_root)
        else
            ! If the 3d reconstruction is not being read from disk, we need to set it up manually
            ! Check the dimensions have been set
            if (all(self%logical_dimensions .eq. [1,1,1])) then
                write(*,'(a)') '**ERROR(threed_init): dimensions for threed object not set'
                call terminate('Fatal error in THREED_INIT')
            else if ((any(self%logical_dimensions .gt. 1000000)) .or. any(self%logical_dimensions .le. 0)) then
                write(*,'(a,3(i0,1x))') '**ERROR(threed_init): strange dimensions for threed object: ', &
                                        self%logical_dimensions
                call terminate('Fatal error in THREED_INIT')
            endif
            ! Allocate memory
            if (debug) write(*,'(a)') '**debug(threed_init): about to call threed_allocate'
            call threed_allocate(self,nnum_of_recs)
            ! Zero the volumes
            call threed_wipe(self)
        endif


        ! We'll need to keep track of the origin of the reconstruction
        do current_rec=1,threed_num_of_recs(self)
            self%rec(current_rec)%vol_image%object_is_centered_in_box = .false.
            self%rec(current_rec)%vol_amplitude_weights%object_is_centered_in_box = .true.
            self%rec(current_rec)%vol_ampl%object_is_centered_in_box = .false.
        enddo

        ! Only cubic volumes with even dimensions are supported
        if (self%logical_dimensions(1) .ne. self%logical_dimensions(2) .or. &
            self%logical_dimensions(1) .ne. self%logical_dimensions(3)) then
            write(*,'(a,3(i0,1x))') '**ERROR(threed_init): non-cubic reconstruction volumes are not yet supported: ', &
                                    self%logical_dimensions
            call terminate('Fatal error in THREED_INIT')
        endif
        if (any(is_odd(self%logical_dimensions))) then
            write(*,'(a,3(i0,1x))') '**ERROR(threed_init): reconstruction volumes with odd dimensions are not yet supported: ', &
                                    self%logical_dimensions
            call terminate('Fatal error in THREED_INIT')
        endif

        ! Set the masking parameters
        if (present(mask_rad)) then
            self%mask_rad       =   mask_rad
        else
            self%mask_rad       =   real(minval(self%logical_dimensions))/2. - 1.0
        endif
        if (debug) print *, '**debug(threed_init): mask_rad = ', self%mask_rad
        if (present(mask_falloff)) then
            self%mask_falloff   =   mask_falloff
        else
            self%mask_falloff   =   default_falloff_pixels
        endif

        ! Set oversampling factor
        !if (present(oversampling_factor)) then
        !    self%oversampling_factor    =   oversampling_factor
        !else
        !    self%oversampling_factor    =   1
        !endif
        ! Oversampling is currently only used in the case of get2d_put3d
        !if (self%oversampling_factor .gt. 1 .and. .not. self%get2d_put3d) then
        !    print *, '**warning(threed_init): oversampling will be ignored, since not using get2d_put3d logic'
        !endif
        ! Set symmetry matrices
        if (present(sym_symbol)) then
            ssym_symbol = sym_symbol
        else
            ssym_symbol%symbol = 'C1 '
        endif
        call point_symmetry_init(self%point_sym,ssym_symbol,helical_symmetry=present(helix))
        if (.not. self%point_sym%helical) call point_symmetry_print(self%point_sym)
        ! Non-symmetry redundancy
        !if (present(non_symmetry_redundancy)) then
        !    self%non_symmetry_redundancy = non_symmetry_redundancy
        !else
        !    self%non_symmetry_redundancy = 1
        !endif
        !if (debug) print '(a,i0)', '**info(threed_init): non-symmetry redundancy = ', self%non_symmetry_redundancy

        ! Parameters needed for helical masking
        if (present(helix)) self%helix = helix
        if (present(asym_unit_mw)) self%asym_unit_mw = asym_unit_mw
        if (present(pixel_size)) self%pixel_size = pixel_size
        if (present(helix_max_rad)) self%helix_max_rad = helix_max_rad

        ! If the user asked to estimate ssnr, and the reconstruction is helical, we should check all relevant parameters have been set
        if (self%use_ssnr_estimate .and. self%compute_fsc) then
            write(*,'(2a)') '**info(threed_init): two 3d reconstruction will be accumulated internally, ',   &
                            'and their masked fsc will be used to evaluate the SSNR'
            if (self%point_sym%helical) then
                if (.not. (present(helix) .and. present(asym_unit_mw) .and. present(pixel_size) .and. present(helix_max_rad))    &
                    .or. self%asym_unit_mw .eq. 0. .or. self%pixel_size .eq. 0. .or. self%helix_max_rad .eq. 0.) then
                    write(*,'(2a)') '**ERROR(threed_init): ssnr estimation via fsc for helical reconstructions ',    &
                                    'requires that helical parameters be given to the threed object.'
                    call terminate('Fatal error in THREED_INIT. Programming error')
                endif
            else
                write(*,'(a)') '**ERROR(threed_init): non-helical ssnr estimation via fsc not implemented yet'
                call terminate('Feature not implemented yet')
            endif
        endif

        ! Estimated resolution
        self%estimated_resolution = 20.0
        if (present(estimated_resolution)) then
            self%estimated_resolution = estimated_resolution
        endif

        ! We don't have an ssnr estimate or fsc  yet
        self%ssnr_is_up_to_date = .false.
        self%fsc_is_up_to_date = .false.

        if (debug) write(*,'(a,i0)') '**debug(threed_init): all done. Number of internal 3d reconstructions = ', &
                                    threed_num_of_recs(self)

    end subroutine threed_init


    !impure elemental
    subroutine threed_deallocate(self)
        implicit none
        ! Argument
        type(threed),   intent(inout)   ::  self
        ! Private variables
        integer ::  current_rec
        logical, parameter :: debug = .false.
        ! Start work
        if (debug) write(*,'(a)') '**debug(threed_deallocate): entering'
        if (allocated(self%rec)) then
            do CURRENT_REC=1,size(self%rec)
                if (debug) write(*,'(2(A,i0))') '**debug(threed_deallocate): about to deallocate rec ', &
                                            current_rec, ' of ', threed_num_of_recs(self)
                call threed_reconstruction_deallocate(self%rec(current_rec))
            enddo
            if (debug) write(*,'(a,i0)') '**debug(threed_deallocate): deallocating rec array, size ', size(self%rec)
            if (allocated(self%rec)) deallocate(self%rec)
            if (debug) write(*,'(a)') '**debug(threed_deallocate): deallocated rec array'
        !else
        !    if (debug) write(*,'(a)') '**debug(threed_deallocate): attempt to deallocate threed which was not allocated'
        endif
        if (debug) write(*,'(a)') '**debug(threed_deallocate): about to deallocate rec array'
        if (allocated(self%rec)) deallocate(self%rec)
        ! Set the allocated flag
        self%in_memory = .false.
        !
        self%ssnr_is_up_to_date = .false.
        self%fsc_is_up_to_date = .false.
        if (debug) write(*,'(a)') '**debug(threed_deallocate): deallocating real arrays'
        !if (allocated(self%ssnr_estimate)) deallocate(self%ssnr_estimate)
        !if (allocated(self%fsc)) deallocate(self%fsc)
        call self%ssnr_curve%Destroy()
        call self%fsc_curve%Destroy()
        if (debug) write(*,'(a)') '**debug(threed_deallocate): exiting'
    end subroutine threed_deallocate

    subroutine threed_allocate(self,num_recs)
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(threed),   intent(inout)   ::  self
        integer,        intent(in)      ::  num_recs
        ! Private variables
        integer                         ::  current_rec
        logical,        parameter       ::  debug   =   .false.
        ! Start work
        if (num_recs .lt. 1 .or. num_recs .gt. 2) then
            write(*,'(a,i0)') '**ERROR(threed_allocate): number of reconstructions should be 1 or 2, not ', num_recs
            call terminate('Fatal error in THREED_ALLOCATE')
        endif
        if (debug) then
            write(*,'(a)') '**debug(threed_allocate): threed info on entry'
            call threed_print_info(self)
        endif
        if (debug) write(*,'(a)') '**debug(threed_allocate): about to call threed_deallocate'
        call threed_deallocate(self)
        if (debug) write(*,'(a)') '**debug(threed_allocate): about to allocate rec array'
        allocate(self%rec(num_recs))
        do current_rec=1,num_recs
            if (debug) write(*,'(a,i0)') '**debug(threed_allocate): about to allocate rec # ', current_rec
            call threed_reconstruction_allocate(self%rec(current_rec),self%logical_dimensions,self%reconstruct_amplitudes)
        enddo
        self%in_memory = .true.
    end subroutine threed_allocate

        !>  \brief  Return the window width of the interpolation scheme used by the threed object
    pure integer function threed_get_window_width(self)
        implicit none
        ! Arguments
        type(threed),  intent(in)      ::  self
        ! Start work
        threed_get_window_width = self%interpol_scheme%window_width
    end function threed_get_window_width

    !>  \brief  Return the interpolation scheme used by the threed object
    !!  This function segfaults with ifort 11.1
    function threed_get_interpol_scheme(self)
        implicit none
        ! Arguments
        type(threed),  intent(in)      ::  self
        ! Result
        type(interpolation_scheme)      ::  threed_get_interpol_scheme
        ! Private variables

        ! Start work
        threed_get_interpol_scheme = self%interpol_scheme
    end function threed_get_interpol_scheme


    !>  \brief  Return the number of reconstructions accumulated (1 or 2)
    pure integer function threed_num_of_recs(self)
        implicit none
        ! Arguments
        type(threed),       intent(in)  ::  self
        ! Start work
        threed_num_of_recs = size(self%rec)
    end function threed_num_of_recs

    !>  \brief  Check the number of internal 3d reconstructions in a threed object stored on disk
    integer function threed_on_disk_num_of_recs(filename_root)
        use parsers
        use runtime_parameters, only : filename_len
        implicit none
        ! Argument
        character(len=*),   intent(in)      ::  filename_root   !<  Root filename for the files
        ! Private variables
        character(len=FILENAME_LEN) ::  filename_image, filenames_image(2)
        character(len=FILENAME_LEN) ::  filename_ctf2!, filenames_ctf2(2)
        character(len=FILENAME_LEN) ::  filename_ampl
        character(len=FILENAME_LEN) ::  filename_stats!, filenames_stats(2)  ! this file will contain two lines, 1. the number of contributing elements, 2. the non-symmetry redundancy
        character(len=FILENAME_LEN) ::  filename_fsc
        character(len=FILENAME_LEN) ::  filename_ssnr

        ! Start work
        ! Get filenames for component files
        call threed_filenames_from_root(filename_root,filename_image,filename_ctf2,filename_stats,filename_ampl,&
                                        filename_fsc,filename_ssnr)
        ! Check whether the input files are for one or two reconstructions
        call threed_filename_add_suffixes(filename_image,filenames_image)
        if (file_exists(filenames_image(1)) .and. file_exists(filenames_image(2))) then
            threed_on_disk_num_of_recs = 2
        else
            threed_on_disk_num_of_recs = 1
        endif
    end function threed_on_disk_num_of_recs

    !>  \brief  Check whether a 3d object stored on disk includes the amplitude reconstruction
    logical function threed_on_disk_ampl_vol(filename_root)
        use parsers
        use runtime_parameters, only : filename_len
        implicit none
        ! Argument
        character(len=*),   intent(in)  ::  filename_root
        ! Private variables
        character(len=FILENAME_LEN) ::  filename_image
        character(len=FILENAME_LEN) ::  filename_ctf2
        character(len=FILENAME_LEN) ::  filename_ampl, filenames_ampl(2)
        character(len=FILENAME_LEN) ::  filename_stats
        character(len=FILENAME_LEN) ::  filename_fsc
        character(len=FILENAME_LEN) ::  filename_ssnr
        ! Start work
        ! Get filenames for component files
        call threed_filenames_from_root(filename_root,filename_image,filename_ctf2,filename_stats,filename_ampl, &
                                        filename_fsc,filename_ssnr)
        if (threed_on_disk_num_of_recs(filename_root) .eq. 2) then
            call threed_filename_add_suffixes(filename_ampl,filenames_ampl)
            threed_on_disk_ampl_vol = file_exists(filenames_ampl(1)) .and. file_exists(filenames_ampl(2))
        else
            threed_on_disk_ampl_vol = file_exists(filename_ampl)
        endif
    end function threed_on_disk_ampl_vol


    !>  \brief  Set the non-symmetry redundancy (nsr) of the 3d reconstruction
    !subroutine threed_set_nsr(self,nsr)
    !    implicit none
    !    ! Arguments
    !    type(threed),  intent(inout)   ::  self
    !    integer,        intent(in)      ::  nsr
    !    ! Private variables!
    !
    !    ! Start work
    !    ! The 3d reconstruction object keeps track of the nsr
    !    if (self%non_symmetry_redundancy .ne.1 .and. self%non_symmetry_redundancy .ne. nsr) then
    !        write(*,'(2a,2(i0,1x))') '**ERROR(threed_set_nsr): mismatch between argument non-symmetry redundancy ',   &
    !                                '(nsr) and nsr of incoming 3d reconstruction: ', &
    !                                nsr, self%non_symmetry_redundancy
    !        call terminate('Fatal error in THREED_SET_NSR')
    !    endif
    !    self%non_symmetry_redundancy = nsr
    !end subroutine threed_set_nsr

    !>  \brief  Check whether two threed reconstructions have been accumulated
    pure logical function threed_two_recs(self)
        implicit none
        ! Arguments
        type(threed),       intent(in)  ::  self    !<  3d reconstruction object
        ! Private variables

        ! Start work
        threed_two_recs  =                      all((self%rec%vol_image%IsAllocated()))   &
                                        .and.   all((self%rec%vol_amplitude_weights%IsAllocated()))    &
                                        .and.   all(self%rec%num_contrib_elems .ne. 0)
    end function threed_two_recs

    !>  \brief  Check the dimensions of the components of the threed object are consistent
    subroutine threed_check_dimensions(self)
        implicit none
        ! Arguments
        type(threed),       intent(in)  ::  self
        ! Private variables
        integer ::  current_rec
        ! Start work
        do CURRENT_REC=1,threed_num_of_recs(self)
            if (any(self%rec(current_rec)%vol_image%logical_dimensions .ne. self%logical_dimensions)) then
                write(*,'(a,i0,a,3(i0,1x),a,3(i0,1x),a)') '**ERROR(threed_check_dimensions): dimensions for image volume ', &
                                                        current_rec, ' (',self%rec(current_rec)%vol_image%logical_dimensions, &
                                                        ') do not match dimensions of threed object (', self%logical_dimensions, ')'
                call terminate('Fatal error in THREED_CHECK_DIMENSIONS')
            endif
            if (any(self%rec(current_rec)%vol_amplitude_weights%logical_dimensions .ne. self%logical_dimensions)) then
                write(*,'(a,i0,a,3(i0,1x),a,3(i0,1x),a)') '**ERROR(threed_check_dimensions): dimensions for CTF^2 volume ', &
                                                        current_rec, ' (',self%rec(current_rec)%vol_image%logical_dimensions, &
                                                        ') do not match dimensions of threed object (', self%logical_dimensions, ')'
                call terminate('Fatal error in THREED_CHECK_DIMENSIONS')
            endif
        enddo
    end subroutine threed_check_dimensions

    !>  \brief  Merge the two internal reconstructions into a single one
    subroutine threed_to_single(self)
        use images
        implicit none
        ! Arguments
        type(threed),                   intent(inout)   ::  self
        ! Private variables
        type(threed_reconstruction) ::  bak(2)
        ! Start work

        ! Backup data
        call threed_reconstruction_assignment(bak(1),self%rec(1))
        if (size(self%rec) .eq. 2) call threed_reconstruction_assignment(bak(2),self%rec(2))

        ! Reallocate
        call threed_reconstruction_deallocate(self%rec(1))
        call threed_reconstruction_deallocate(self%rec(2))
        deallocate(self%rec)
        allocate(self%rec(1))
        call self%rec(1)%vol_image%Allocate(bak(1)%vol_image)
        call self%rec(1)%vol_amplitude_weights%Allocate(bak(1)%vol_amplitude_weights)


        self%rec(1)%vol_image%real_values  = bak(1)%vol_image%real_values  + bak(2)%vol_image%real_values
        self%rec(1)%vol_amplitude_weights%real_values   = bak(1)%vol_amplitude_weights%real_values &
                                                        + bak(2)%vol_amplitude_weights%real_values
        self%rec(1)%num_contrib_elems = bak(1)%num_contrib_elems + bak(2)%num_contrib_elems
        if (self%reconstruct_amplitudes) then
        call self%rec(1)%vol_ampl%Allocate(bak(1)%vol_ampl)
        self%rec(1)%vol_ampl%real_values   = bak(1)%vol_ampl%real_values   + bak(2)%vol_ampl%real_values
        endif
        self%in_memory = .true.
    end subroutine threed_to_single

    !>  \brief  Merge two threeds each with one internal reconstruction
    !!          into a single threed object with two internal
    subroutine threed_merge(input1,input2)
        implicit none
        ! Arguments
        type(threed),                   intent(inout)   ::  input1
        type(threed),                   intent(in)      ::  input2
        ! Private variables
        type(threed_reconstruction) ::  tmp_rec(2)
        ! Start work

        ! Check the input threeds only have one internal reconstruction
        if (threed_num_of_recs(input1) .ne. 1 .or. &
            threed_num_of_recs(input2) .ne. 1 ) then
            write(*,'(2a,2(i0,1x))')  '**ERROR(threed_merge): input 3d reconstruction objects should have one ',  &
                                    'and only one internal reconstruction, but they have: ', &
                                    threed_num_of_recs(input1), &
                                    threed_num_of_recs(input2)
            call terminate('Fatal error in THREED_MERGE. Programming error.')
        endif

!        ! Initialise the output threed
!        call threed_init(   output,                                 &
!                            MASK_RAD=INPUT1%mask_rad,               &
!                            MASK_FALLOFF=INPUT1%mask_falloff,       &
!                            INTERPOL_SCHEME=INPUT1%interpol_scheme, &
!                            OVERSAMPLING_FACTOR=INPUT1%oversampling_factor,&
!                            SYM_SYMBOL=INPUT1%point_sym%symbol,     &
!                            NON_SYMMETRY_REDUNDANCY=INPUT1%non_symmetry_redundancy, &
!                            USE_SSNR_ESTIMATE=INPUT1%use_ssnr_estimate, &
!                            COMPUTE_FSC=INPUT1%compute_fsc,         &
!                            HELIX=INPUT1%helix,                     &
!                            ASYM_UNIT_MW=INPUT1%asym_unit_mw,       &
!                            PIXEL_SIZE=INPUT1%pixel_size,           &
!                            HELIX_MAX_RAD=INPUT1%helix_max_rad,     &
!                            NUM_OF_RECS=2,                          &
!                            FOM_FILTER=INPUT1%fom_filter  )

        ! Copy the data over
        tmp_rec(1)%vol_image = input1%rec(1)%vol_image
        tmp_rec(1)%vol_amplitude_weights = input1%rec(1)%vol_amplitude_weights
        tmp_rec(1)%num_contrib_elems = input1%rec(1)%num_contrib_elems
        if (input1%reconstruct_amplitudes) tmp_rec(1)%vol_ampl = input1%rec(1)%vol_ampl
        tmp_rec(2)%vol_image = input2%rec(1)%vol_image
        tmp_rec(2)%vol_amplitude_weights = input2%rec(1)%vol_amplitude_weights
        tmp_rec(2)%num_contrib_elems = input2%rec(1)%num_contrib_elems
        if (input2%reconstruct_amplitudes) tmp_rec(2)%vol_ampl = input2%rec(1)%vol_ampl

        ! Reallocate
        call input1%rec(1)%vol_image%Deallocate()
        call input1%rec(1)%vol_amplitude_weights%Deallocate()
        if (input1%reconstruct_amplitudes) call input1%rec(1)%vol_ampl%Deallocate()
        deallocate(input1%rec)
        allocate(input1%rec(2))
        input1%rec(1)%vol_image = tmp_rec(1)%vol_image
        input1%rec(1)%vol_amplitude_weights = tmp_rec(1)%vol_amplitude_weights
        input1%rec(1)%num_contrib_elems = tmp_rec(1)%num_contrib_elems
        input1%rec(2)%vol_image = tmp_rec(2)%vol_image
        input1%rec(2)%vol_amplitude_weights = tmp_rec(2)%vol_amplitude_weights
        input1%rec(2)%num_contrib_elems = tmp_rec(2)%num_contrib_elems
        if (input1%reconstruct_amplitudes) input1%rec(1)%vol_ampl = tmp_rec(1)%vol_ampl
        if (input2%reconstruct_amplitudes) input1%rec(2)%vol_ampl = tmp_rec(2)%vol_ampl


    end subroutine threed_merge

    !>  \brief  Sum two threeds
    !!
    !!  Input 2 is added to (or subtracted from) input1, to give output
    subroutine threed_add(input1,input2,output,subtract)
        implicit none
        ! Arguments
        type(threed),                   intent(in)      ::  input1
        type(threed),                   intent(in)      ::  input2
        type(threed),                   intent(inout)   ::  output
        logical,        optional,       intent(in)      ::  subtract    !<  Whether to subtract input2 from input1, rather than add. .FALSE. by default.
        ! Private variables
        real            ::  ssign
        integer         ::  num_of_recs
        integer         ::  i,j,k, current_rec
        type(threed)    ::  input1_tmp
        ! Start work

        ! Are the two 3Ds of same dimensions
        if (any(input1%logical_dimensions .ne. input2%logical_dimensions)) then
            write(*,'(2(A,3(I0,1x)))') '**ERROR(threed_add): the two 3d reconstructions do not have the same dimensions: ', &
                                        input1%logical_dimensions, ' and ', input2%logical_dimensions
        endif

        ! Either both or neither reconstructions has an amplitude volume
        if ((input1%reconstruct_amplitudes .and. .not. input2%reconstruct_amplitudes) .or. &
            (input2%reconstruct_amplitudes .and. .not. input1%reconstruct_amplitudes)) then
            write(*,'(a)') '**ERROR(threed_add): only one of the 3d reconstructions includes an amplitude-only volume'
            call terminate('Fatal error in THREED_ADD')
        endif

        ! Are we subtracting or adding? (adding by default)
        ssign = 1.0
        if (present(subtract)) then
            if (subtract) then
                ssign = -1.0
            endif
        endif

        ! Maybe we need to allocate the output?
        if (threed_num_of_recs(output) .eq. 0 .or. .not. threed_allocated(output)) then
            output%logical_dimensions = input1%logical_dimensions
            call threed_allocate(output,threed_num_of_recs(input1))
        endif

        ! How many reconstructions?
        num_of_recs = threed_num_of_recs(input1)
        if (num_of_recs .eq. 2 .and. (threed_num_of_recs(input2) .eq. 1) .and. (threed_num_of_recs(output) .eq. 1)) then
            !write(*,'(a)') '**debug(threed_add): first but not second 3d has two reconstructions. Merging to single rec'
            call threed_assignment(input1_tmp,input1)
            call threed_to_single(input1_tmp)
        else
            call threed_assignment(input1_tmp,input1)
        endif
        num_of_recs = threed_num_of_recs(input1_tmp)
        if (threed_num_of_recs(input1_tmp) .ne. threed_num_of_recs(input2) .or. num_of_recs .ne. threed_num_of_recs(output)) then
            write(*,'(2a,3(i0,1x))') '**ERROR(threed_add): input and output 3d reconstruction objects do not contain ',   &
                                    'the same number of 3d reconstructions: ', &
                                    threed_num_of_recs(input1_tmp), threed_num_of_recs(input2), threed_num_of_recs(output)
            call terminate('Fatal error in THREED_ADD')
        endif

        ! Check allocation
        if (.not. threed_allocated(input1_tmp) .or. .not. threed_allocated(input2) .or. .not. threed_allocated(output)) then
            write(*,'(a,3l1)') '**ERROR(threed_add): not all threed objects allocated: ',   threed_allocated(input1_tmp),   &
                                                                                            threed_allocated(input2),       &
                                                                                            threed_allocated(output)
            call terminate('Fatal error in THREED_ADD')
        endif

        ! Do the actual addition
        !write(*,'(a,i0)') '**debug(threed_add): about to do actual addition, num_of_recs = ', num_of_recs
        !write(*,'(a,3(i0,1x))') '**debug(threed_add): dims = ', input1_tmp%logical_dimensions
        do current_rec=1,num_of_recs
            do k=1,input1_tmp%logical_dimensions(3)
                do j=1,input1_tmp%logical_dimensions(2)
                    do i=1,input1_tmp%logical_dimensions(1)
                        if (i .le. input1_tmp%rec(current_rec)%vol_image%physical_upper_bound_complex(1)) then
                        output%rec(current_rec)%vol_image%complex_values(i,j,k) = &
                                input1_tmp%rec(current_rec)%vol_image%complex_values(i,j,k) &
                              + input2%rec(current_rec)%vol_image%complex_values(i,j,k) * ssign
                        if (input1_tmp%reconstruct_amplitudes) then
                        output%rec(current_rec)%vol_ampl%complex_values(i,j,k) =  &
                                input1_tmp%rec(current_rec)%vol_ampl%complex_values(i,j,k)     &
                              + input2%rec(current_rec)%vol_ampl%complex_values(i,j,k) * ssign
                        endif
                        endif
                        output%rec(current_rec)%vol_amplitude_weights%real_values(i,j,k) = &
                                input1_tmp%rec(current_rec)%vol_amplitude_weights%real_values(i,j,k) + &
                                input2%rec(current_rec)%vol_amplitude_weights%real_values(i,j,k) * ssign

                    enddo
                enddo
            enddo
            output%rec(current_rec)%num_contrib_elems = input1_tmp%rec(current_rec)%num_contrib_elems + &
                                                        input2%rec(current_rec)%num_contrib_elems * int(ssign)
        enddo

        ! Deal with the non-symmetry redundancy
        !if (input1_tmp%non_symmetry_redundancy .eq. 1 .or. input2%non_symmetry_redundancy .eq. 1) then
        !    if (input1_tmp%non_symmetry_redundancy .eq. 1) then
        !        output%non_symmetry_redundancy = input2%non_symmetry_redundancy
        !    else if (input2%non_symmetry_redundancy .eq. 1) then
        !        output%non_symmetry_redundancy = input1_tmp%non_symmetry_redundancy
        !    endif
        !else
        !    if (input1_tmp%non_symmetry_redundancy .ne. input2%non_symmetry_redundancy) then
        !        write(*,'(a,2(i5,1x))') '**ERROR(threed_add): the two input 3Ds do not have the same non-symmetry redundancy: ', &
        !                                input1_tmp%non_symmetry_redundancy, input2%non_symmetry_redundancy
        !        call terminate('Fatal error in THREED_ADD')
        !    endif
        !endif
        !if (    input1_tmp%non_symmetry_redundancy .ne. output%non_symmetry_redundancy .and. &
        !        input1_tmp%non_symmetry_redundancy .eq. input2%non_symmetry_redundancy) then
        !    write(*,'(2a,2(i5,1x))') '**ERROR(threed_add): the input 3Ds do not have the same ',  &
        !                            'non-symmetry redundancy as the output 3d: ',    &
        !                            input1_tmp%non_symmetry_redundancy, output%non_symmetry_redundancy
        !    call terminate('Fatal error in THREED_ADD')
        !endif

        call threed_deallocate(input1_tmp)

        ! The output's pssnr estimate & fsc are not up to date anymore
        output%ssnr_is_up_to_date = .false.
        output%fsc_is_up_to_date = .false.
    end subroutine threed_add

    !>  \brief  Fills the 3D's volumes and data to 0.
    subroutine threed_wipe(self)
        implicit none
        ! Argument
        type(threed),   intent(inout)   ::  self
        ! Private variables
        integer ::  current_rec
        ! Start work
        do CURRENT_REC=1,threed_num_of_recs(self)
            self%rec(current_rec)%vol_amplitude_weights%real_values = 0.0
            self%rec(current_rec)%vol_image%complex_values = (0.0,0.0)
            self%rec(current_rec)%num_contrib_elems = 0
            if (self%reconstruct_amplitudes) self%rec(current_rec)%vol_ampl%complex_values = (0.0,0.0)
        enddo
        self%ssnr_is_up_to_date = .false.
    end subroutine threed_wipe

    subroutine threed_filename_add_suffixes(filename,filenames)
        use parsers
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  filename
        character(len=*),   intent(out) ::  filenames(2)
        ! Private variables

        ! Start work
        call filename_add_suffix(filename,'_1of2',filenames(1))
        call filename_add_suffix(filename,'_2of2',filenames(2))
    end subroutine threed_filename_add_suffixes

    !>  \brief  Given a root filename for a 3d reconstruction object, return the actual filenames for the component files
    subroutine threed_filenames_from_root(filename_root,filename_image,filename_ctf2,filename_stats,filename_ampl, &
                                          filename_fsc,filename_ssnr)
        use parsers
        use runtime_parameters, only : filename_len
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  filename_root   !<  Root filename
        character(len=*),   intent(out) ::  filename_image  !<  Filename for volume containing image data
        character(len=*),   intent(out) ::  filename_ctf2   !<  Filename for volume containing squared ctf data
        character(len=*),   intent(out) ::  filename_stats  !<  Filename for text file containing 3d reconstruction statistics
        character(len=*),   intent(out) ::  filename_ampl   !<  Filename for the amplitudes volume
        character(len=*),   intent(out) ::  filename_fsc    !<  Filename for the fsc curve
        character(len=*),   intent(out) ::  filename_ssnr   !<  Filename for the ssnr curve
        ! Private variables
        character(len=FILENAME_LEN)     ::  tmp_string, tmp_string_noext
        character(len=3)                ::  tmp_string_ext
        ! Start work

        tmp_string = trim(adjustl(filename_root))
        tmp_string_ext = extension_from_filename(tmp_string)
        if (string_equal(tmp_string_ext,'   ')) then
            ! If the filename_root has no extension
            tmp_string_noext = tmp_string
            tmp_string_ext = 'mrc'
        else
            tmp_string_noext = tmp_string(1:len_trim(tmp_string)-4)
        endif
        write(filename_image,'(a,a,a1,a3)') trim(adjustl(tmp_string_noext)), '_unweighted', '.', tmp_string_ext
        write(filename_ctf2 ,'(a,a,a1,a3)') trim(adjustl(tmp_string_noext)), '_ctfweights', '.', tmp_string_ext
        write(filename_ampl,'(a,a,a1,a3)') trim(adjustl(tmp_string_noext)), '_ampl','.', tmp_string_ext
        write(filename_stats  ,'(a,a,a1,a3)') trim(adjustl(tmp_string_noext)), '_stats', '.', 'dat'
        write(filename_fsc  ,'(a,a,a1,a3)') trim(adjustl(tmp_string_noext)), '_fsc', '.', 'dat'
        write(filename_ssnr  ,'(a,a,a1,a3)') trim(adjustl(tmp_string_noext)), '_ssnr', '.', 'dat'
    end subroutine threed_filenames_from_root


    !>  \brief  Read a (non-finalised) 3d reconstruction from disk (i.e. both parts, the unweighted image data, and the CTF^2 weights).
    !!          You need to supply a filename with extension.
    !!  Warning: the resulting 3d reconstruction object is not fully initialised (it is missing the interpolation scheme, as well as masking parameters etc.)
    subroutine threed_read_from_disk(self,filename_root)
        use parsers
        use runtime_parameters, only : filename_len, available_lun, lun_is_now_available
        use Images
        implicit none
        ! Arguments
        type(threed),      intent(inout)   ::  self            !<  3d reconstruction object to save to disk
        character(len=*),   intent(in)      ::  filename_root   !<  Root filename for the files
        ! Private variables
        character(len=FILENAME_LEN) ::  filename_image, filenames_image(2)
        character(len=FILENAME_LEN) ::  filename_ctf2, filenames_ctf2(2)
        character(len=FILENAME_LEN) ::  filename_ampl, filenames_ampl(2)
        character(len=FILENAME_LEN) ::  filename_stats, filenames_stats(2)  ! this file will contain two lines, 1. the number of contributing elements, 2. the non-symmetry redundancy
        character(len=FILENAME_LEN) ::  filename_fsc, filename_ssnr

        integer ::  ios ! i/o status
        integer ::  current_rec
        integer ::  file_handle
        logical,    parameter   ::  debug   =   .false.
        ! Start work
        if (is_blank(extension_from_filename(filename_root))) then
            write(*,'(a)') '**ERROR(threed_write_to_disk): code not written for case where filename_root does not have an extension'
            call terminate('Fatal error in THREED_WRITE_TO_DISK')
        else
            ! Get an available file handle
            file_handle = available_lun()
            ! Get filenames for component files
            call threed_filenames_from_root(filename_root,filename_image,filename_ctf2,filename_stats,filename_ampl, &
                                            filename_fsc,filename_ssnr)
            ! Check whether the input files are for one or two reconstructions
            if (threed_on_disk_num_of_recs(filename_root) .eq. 2) then
                if (threed_num_of_recs(self) .ne. 2) then
                    write(*,'(2a,i0)') '**ERROR(threed_read_from_disk): reading from files describing two internal ',    &
                                        'reconstructions into a threed object with ', &
                                        threed_num_of_recs(self)
                    call terminate('Fatal error in THREED_READ_FROM_DISK')
                endif
            endif
            !
            if ((threed_on_disk_ampl_vol(filename_root) .and. .not. self%reconstruct_amplitudes) .or. &
                (.not. threed_on_disk_ampl_vol(filename_root) .and. self%reconstruct_amplitudes)) then
                if (self%reconstruct_amplitudes) then
                    write(*,'(a)') '**ERROR(threed_read_from_disk): amplitude reconstruction(s) on disk expected but not found'
                else
                    write(*,'(a)') '**ERROR(threed_read_from_disk): amplitude reconstruction(s) on disk found but not expected'
                endif
                call terminate('Fatal error in THREED_READ_FROM_DISK')
            endif

            ! If needed, get filenames for both reconstructions
            if (threed_num_of_recs(self) .eq. 2) then
                call threed_filename_add_suffixes(filename_image,filenames_image)
                call threed_filename_add_suffixes(filename_ctf2,filenames_ctf2)
                call threed_filename_add_suffixes(filename_stats,filenames_stats)
                call threed_filename_add_suffixes(filename_ampl,filenames_ampl)
            else if (threed_num_of_recs(self) .eq. 1) then
                filenames_image = filename_image
                filenames_ctf2 = filename_ctf2
                filenames_stats = filename_stats
                filenames_ampl = filename_ampl
            else
                write(*,'(a)') '**ERROR(threed_read_from_disk): unexpected threed_num_of_recs(self) = ', threed_num_of_recs(self)
                call terminate('Programming error')
            endif
            do current_rec=1,threed_num_of_recs(self)
                if (debug) write(*,'(a,i0,a,i0)') '**debug(threed_read_from_disk): on internal reconstruction ', current_rec, &
                                                                                ' of ', threed_num_of_recs(self)
                if (debug) write(*,'(5(A,1x))') '**debug(threed_read_from_disk): filenames:',    &
                                                        trim(adjustl(filenames_image(current_rec))), &
                                                        trim(adjustl(filenames_ctf2(current_rec))), &
                                                        trim(adjustl(filenames_stats(current_rec))), &
                                                        trim(adjustl(filenames_ampl(current_rec)))
                ! Read image data
                !call read_from_disk(self%rec(current_rec)%vol_image,filenames_image(current_rec))
                call self%rec(current_rec)%vol_image%ReadFromDisk(filenames_image(current_rec),read_volume_not_slice=.true.)
                if (debug) write(*,'(a,3(i0,1x))') '**debug(threed_read_from_disk): image dimensions: ', &
                                                        self%rec(current_rec)%vol_image%logical_dimensions
                ! Check whether it is blank
                if ((self%rec(current_rec)%vol_image%IsConstant())) then
                    write(*,'(3a)') '**warning(threed_read_from_disk): image volume (', &
                                    trim(adjustl(filenames_image(current_rec))),&
                                    ') is blank'
                endif
                ! Check whether it has NaN value(s)
                if ((self%rec(current_rec)%vol_image%HasNan())) then
                    write(*,'(3a)') '**warning(threed_read_from_disk): image volume (', &
                                    trim(adjustl(filenames_image(current_rec))),&
                                    ') has at least one NaN voxel value'
                endif
                call self%rec(current_rec)%vol_image%ForwardFFT()
                call self%rec(current_rec)%vol_image%SwapRealSpaceQuadrants()

                ! Read in amplitudes
                if (self%reconstruct_amplitudes) then
                    !call read_from_disk(self%rec(current_rec)%vol_ampl, filenames_ampl(current_rec))
                    call self%rec(current_rec)%vol_ampl%ReadFromDisk(filenames_ampl(current_rec),read_volume_not_slice=.true.)
                    ! Check whether it is blank
                    if ((self%rec(current_rec)%vol_ampl%IsConstant())) then
                        write(*,'(3a)') '**warning(threed_read_from_disk): ampl volume (', &
                                        trim(adjustl(filenames_ampl(current_rec))),&
                                        ') is blank'
                    endif
                    ! Check whether it has NaN value(s)
                    if ((self%rec(current_rec)%vol_ampl%HasNan())) then
                        write(*,'(3a)') '**warning(threed_read_from_disk): ampl volume (', &
                                        trim(adjustl(filenames_ampl(current_rec))),&
                                        ') has at least one NaN voxel value'
                    endif
                    call self%rec(current_rec)%vol_ampl%ForwardFFT()
                    call self%rec(current_rec)%vol_ampl%SwapRealSpaceQuadrants()
                endif

                ! Read ctf2 data
                !call read_from_disk(self%rec(current_rec)%vol_amplitude_weights, filenames_ctf2(current_rec))
                call self%rec(current_rec)%vol_amplitude_weights%ReadFromDisk(filenames_ctf2(current_rec), &
                                                                              read_volume_not_slice=.true.)
                ! Check whether it is blank
                if ((self%rec(current_rec)%vol_amplitude_weights%IsConstant())) then
                    write(*,'(3a)') '**warning(threed_read_from_disk): ctf2 volume (', trim(adjustl(filenames_ctf2(current_rec))),&
                                    ') is blank'
                endif
                ! Check whether it has NaN value(s)
                if ((self%rec(current_rec)%vol_amplitude_weights%HasNan())) then
                    write(*,'(3a)') '**warning(threed_read_from_disk): ctf2 volume (', trim(adjustl(filenames_ctf2(current_rec))),&
                                    ') has at least one NaN voxel value'
                endif

                ! Read in stats
                open(unit=file_handle,file=TRIM(adjustl(filenames_stats(current_rec))),iostat=IOS)
                if (ios .ne. 0) then
                    write(*,'(a)') '**ERROR(threed_read_from_disk): error opening stats file ', ios
                    call terminate('Fatal error in THREED_READ_FROM_DISK')
                else
                    read(file_handle,'(i12)') self%rec(current_rec)%num_contrib_elems
                    !read(file_handle,'(i12)') self%non_symmetry_redundancy
                endif
                close(unit=file_handle)

                ! Read in FSC

            enddo
            if (debug) write(*,'(a,3(i0,1x))') '**debug(threed_read_from_disk): setting threed dimensions to ', &
                                                self%rec(1)%vol_image%logical_dimensions
            self%logical_dimensions = self%rec(1)%vol_image%logical_dimensions
            if (debug) write(*,'(a,3(i0,1x))') '**debug(threed_read_from_disk): set threed dimensions to ', &
                                                self%rec(1)%vol_image%logical_dimensions
            call threed_check_dimensions(self)
            self%in_memory = .true.
            ! Release file handle
            call lun_is_now_available(file_handle)
        endif
    end subroutine threed_read_from_disk

    !>  \brief  Check whether the 3d reconstruction has any NaN values in there
    logical function threed_has_nan(self)
        implicit none
        ! Arguments
        type(threed),   intent(in)  ::  self
        ! Private variables
        integer ::  current_rec
        ! Start work
        threed_has_nan = .false.
        do CURRENT_REC=1,threed_num_of_recs(self)
            ! Check whether it has NaN value(s)
            if (self%rec(current_rec)%vol_image%HasNan()) then
                threed_has_nan = .true.
                return
            endif
            if ((self%rec(current_rec)%vol_amplitude_weights%HasNan())) then
                threed_has_nan = .true.
                return
            endif
        enddo
    end function threed_has_nan


    !>  \brief  Dump a (non-finalised) 3d reconstruction to disk (i.e. both parts, the unweighted image data, and the CTF^2 weights).
    !!          You need to supply a filename with extension. The number of contributing elements is written into a separate text file.
    subroutine threed_write_to_disk(self,filename_root)
        use parsers, only : is_blank, extension_from_filename
        use imagefiles, only : imagefile
        use runtime_parameters, only : filename_len, available_lun, lun_is_now_available
        use Curves
        use images_comparison
        !use images_ctf, only : ctfvol_shift ! debug
        use images
        implicit none
        ! Arguments
        type(threed),      intent(inout)   ::  self            !<  3d reconstruction object to save to disk
        character(len=*),   intent(in)      ::  filename_root   !<  Root filename for the files
        ! Private variables
        character(len=FILENAME_LEN)     ::  filename_image, filenames_image(2)
        character(len=FILENAME_LEN)     ::  filename_ctf2, filenames_ctf2(2)
        character(len=FILENAME_LEN)     ::  filename_ampl, filenames_ampl(2)
        character(len=FILENAME_LEN)     ::  filename_stats, filenames_stats(2)  ! this file will contain two lines,
                                                                                ! 1. the number of contributing elements,
                                                                                ! 2. the non-symmetry redundancy
        character(len=FILENAME_LEN)     ::  filename_fsc, filenames_fsc(2)
        character(len=FILENAME_LEN)     ::  filename_ssnr, filenames_ssnr(2)
        integer                         ::  ios ! i/o status

        integer                         ::  current_rec
        integer                         ::  file_handle
        type(Curve)                     ::  fsc_curve, ssnr_curve
        !type(image)                     ::  tmpctf
        ! Start work
        if (is_blank(extension_from_filename(filename_root))) then
            write(*,'(a)') '**ERROR(threed_write_to_disk): code not written for case where filename_root does not have an extension'
            call terminate('Fatal error in THREED_WRITE_TO_DISK')
        else
            ! Get an available file handle
            file_handle = available_lun()
            ! Work out filenames to write to
            call threed_filenames_from_root(filename_root,filename_image,filename_ctf2,filename_stats,filename_ampl, &
                                            filename_fsc,filename_ssnr)
            ! If needed, get filenames for both reconstructions
            if (threed_num_of_recs(self) .eq. 2) then
                call threed_filename_add_suffixes(filename_image,filenames_image)
                call threed_filename_add_suffixes(filename_ctf2,filenames_ctf2)
                call threed_filename_add_suffixes(filename_ampl,filenames_ampl)
                call threed_filename_add_suffixes(filename_stats,filenames_stats)
                call threed_filename_add_suffixes(filename_fsc,filenames_fsc)
                call threed_filename_add_suffixes(filename_ssnr,filenames_ssnr)
            else if (threed_num_of_recs(self) .eq. 1) then
                filenames_image = filename_image
                filenames_ctf2 = filename_ctf2
                filenames_stats = filename_stats
                filenames_ampl = filename_ampl
                filenames_fsc  = filename_fsc
                filenames_ssnr  = filename_ssnr
            else
                write(*,'(a)') '**ERROR(threed_write_to_disk): unexpected threed_num_of_recs(self) = ', threed_num_of_recs(self)
                call terminate('Programming error')
            endif

            do current_rec=1,threed_num_of_recs(self)
                ! Write the unweighted 3d reconstruction to disk
                !write(*,'(a,l1)') '**debug(threed_write_to_disk): about to shift origin. Currently at center? ', self%rec(current_rec)%vol_image%object_is_centered_in_box
                call self%rec(current_rec)%vol_image%SwapRealSpaceQuadrants()
                call self%rec(current_rec)%vol_image%BackwardFFT()
                ! Check whether it is blank
                if ((self%rec(current_rec)%vol_image%IsConstant())) then
                    write(*,'(3a)') '**warning(threed_write_to_disk): image volume (', trim(adjustl(filenames_image(current_rec))),&
                                    ') is blank'
                endif
                ! Check whether it has NaN value(s)
                if ((self%rec(current_rec)%vol_image%HasNan())) then
                    write(*,'(3a)') '**warning(threed_write_to_disk): image volume (', trim(adjustl(filenames_image(current_rec))),&
                                    ') has at least one NaN voxel value'
                endif
                call self%rec(current_rec)%vol_image%WriteToDisk(filenames_image(current_rec),    &
                                            pixel_size=self%pixel_size)
                call self%rec(current_rec)%vol_image%ForwardFFT()
                call self%rec(current_rec)%vol_image%SwapRealSpaceQuadrants()

                ! Write the amplitudes reconstruction to disk
                if (self%reconstruct_amplitudes) then
                    call self%rec(current_rec)%vol_ampl%SwapRealSpaceQuadrants()
                    call self%rec(current_rec)%vol_ampl%BackwardFFT()
                    ! Check whether it is blank
                    if ((self%rec(current_rec)%vol_ampl%IsConstant())) then
                        write(*,'(3a)') '**warning(threed_write_to_disk): ampl volume (', &
                                        trim(adjustl(filenames_ampl(current_rec))),&
                                        ') is blank'
                    endif
                    ! Check whether it has NaN value(s)
                    if ((self%rec(current_rec)%vol_ampl%HasNan())) then
                        write(*,'(3a)') '**warning(threed_write_to_disk): ampl volume (', &
                                        trim(adjustl(filenames_ampl(current_rec))),&
                                        ') has at least one NaN voxel value'
                    endif

                    call self%rec(current_rec)%vol_ampl%WriteToDisk(filenames_ampl(current_rec),&
                                                pixel_size=self%pixel_size)
                    call self%rec(current_rec)%vol_ampl%ForwardFFT()
                    call self%rec(current_rec)%vol_ampl%SwapRealSpaceQuadrants()
                endif

                ! Write the CTF^2 volume to disk
                ! Check whether it is blank
                if ((self%rec(current_rec)%vol_amplitude_weights%IsConstant())) then
                    write(*,'(3a)') '**warning(threed_write_to_disk): ctf2 volume (', trim(adjustl(filenames_ctf2(current_rec))),&
                                    ') is blank'
                endif
                ! Check whether it has NaN value(s)
                if (self%rec(current_rec)%vol_amplitude_weights%HasNan()) then
                    write(*,'(3a)') '**warning(threed_write_to_disk): ctf2 volume (', trim(adjustl(filenames_ctf2(current_rec))),&
                                    ') has at least one NaN voxel value'
                endif
                !call ctfvol_shift(self%rec(current_rec)%vol_amplitude_weights,tmpctf)
                !call ctf2_file%write_image(tmpctf,FILENAME=FILENAMES_CTF2(current_rec),&
                !                            PIXEL_SIZE=SELF%pixel_size)
                call self%rec(current_rec)%vol_amplitude_weights%WriteToDisk(filenames_ctf2(current_rec),&
                                            pixel_size=self%pixel_size)


                ! Write the number of contributing elements to a flat text file
                open(unit=file_handle,file=trim(adjustl(filenames_stats(current_rec))),status='REPLACE',iostat=ios)
                if (ios .ne. 0) then
                    write(*,'(a)') '**ERROR(threed_write_to_disk): error opening stats file ', ios
                    call terminate('Fatal error in THREED_WRITE_TO_DISK')
                else
                    write(file_handle,'(i0)') self%rec(current_rec)%num_contrib_elems
                    !write(file_handle,'(i0)') self%non_symmetry_redundancy
                endif
                close(unit=file_handle)


            enddo
            ! Write the fsc curve to a flat text file
            if (self%fsc_curve%HasBeenInitialized()) then
                !call fsc_curve_write_to_disk(self%fsc_curve,self%ssnr_curve,filename_fsc)
                call self%fsc_curve%WriteToDisk(filename_fsc)
            else
                write(*,'(a)') '**debug(threed_write_to_disk): FSC curve not initialised, not writing to disk'
            endif
            if (self%ssnr_curve%HasBeenInitialized()) then
                call self%ssnr_curve%WriteToDisk(filename_ssnr)
            endif
            !if (allocated(self%fsc)) then
            !    if (.not. allocated(self%ssnr_estimate)) then
            !        write(*,'(a)') '**warning(threed_write_to_disk): ssnr_estimate array not allocated'
            !        call fsc_to_fsc_curve(self%fsc,fsc_curve,self%pixel_size)
            !        call fsc_curve_write_to_disk(fsc_curve,ssnr_curve,filename_fsc)
            !    else
            !        call fsc_to_fsc_curve(self%ssnr_estimate,ssnr_curve,self%pixel_size)
            !        call fsc_to_fsc_curve(self%fsc,fsc_curve,self%pixel_size)
            !        call fsc_curve_write_to_disk(fsc_curve,ssnr_curve,filename_fsc)
            !    endif
            !endif
            ! Release file handle
            call lun_is_now_available(file_handle)
        endif
    end subroutine threed_write_to_disk

    !>  \brief  Print the 3D's fsc to terminal
    subroutine threed_print_fsc(self)
        use images_comparison, only : print_fsc
        implicit none
        ! Arguments
        type(threed),   intent(inout)  ::  self
        ! Private variables

        ! Start work

        if (.not. self%fsc_is_up_to_date) then
            write(*,'(a)') '**warning(threed_print_fsc): fsc is not up-to-date'
        endif

        write(*,'(a)') '**info(threed_print_fsc): fsc between half-dataset reconstructions follows'
        !call print_fsc(self%fsc,1.0/self%logical_dimensions(1)/self%pixel_size)
        call self%fsc_curve%PrintCurve()

    end subroutine threed_print_fsc





end module threeds_base
