!>  \brief  Functionality for interactive use
module ui

    interface get_rtp_value
        module procedure get_rtp_value_logi
        module procedure get_rtp_value_real
        module procedure get_rtp_value_intg
        module procedure get_rtp_value_char
    end interface get_rtp_value

    interface ui_get_value
        module procedure ui_get_char
        module procedure ui_get_intg
        module procedure ui_get_real
        module procedure ui_get_logi
    end interface

contains



    !>  brief   Find out which mode the user wants frealix to run by parsing the control file
    subroutine get_frealix_mode()
        use runtime_parameters
        use parsers
        implicit none
        if (interactive) then
            call ui_get_char('Please give FREALIX_MODE', frealix_mode, '(todo: add list of possible modes)')
        else
            call file_get_char(control_filename,'FREALIX_MODE',frealix_mode)
        endif
    end subroutine get_frealix_mode


    !>  Get a runtime parameter value, either from the control file or from the user interactively
    subroutine get_rtp_value_logi(parameter,value,question,help)
        use runtime_parameters
        use parsers
        implicit none
        ! Arguments
        character(len=*),               intent(in)      ::  parameter   !< Label of parameter
        logical,                        intent(inout)   ::  value       !< Value of parameter
        character(len=*),   optional,   intent(in)      ::  question    !< Question to ask user for parameter
        character(len=*),   optional,   intent(in)      ::  help        !< Help to give user

        ! Private variables
        character(len=200)  ::  qquestion, hhelp

        ! Start work
        if (present(question)) then
            qquestion = question
        else
            qquestion = 'Please give ' // trim(parameter)
        endif
        if (present(help)) then
            hhelp = help
        else
            hhelp = 'Sorry, no help available'
        endif

        if (interactive) then
            call ui_get_logi(qquestion,value,hhelp)
        else
            call file_get_logi(control_filename,parameter,value)
        endif

    end subroutine get_rtp_value_logi

    !>  Get a runtime parameter value, either from the control file or from the user interactively
    subroutine get_rtp_value_real(parameter,value,question,help)
        use runtime_parameters
        use parsers
        implicit none
        ! Arguments
        character(len=*),               intent(in)      ::  parameter   !< Label of parameter
        real,                           intent(inout)   ::  value       !< Value of parameter
        character(len=*),   optional,   intent(in)      ::  question    !< Question to ask user for parameter
        character(len=*),   optional,   intent(in)      ::  help        !< Help to give user

        ! Private variables
        character(len=200)  ::  qquestion, hhelp

        ! Start work
        if (present(question)) then
            qquestion = question
        else
            qquestion = 'Please give ' // trim(parameter)
        endif
        if (present(help)) then
            hhelp = help
        else
            hhelp = 'Sorry, no help available'
        endif

        if (interactive) then
            call ui_get_real(qquestion,value,hhelp)
        else
            call file_get_real(control_filename,parameter,value)
        endif

    end subroutine get_rtp_value_real


    !>  Get a runtime parameter value, either from the control file or from the user interactively
    subroutine get_rtp_value_intg(parameter,value,question,help)
        use runtime_parameters
        use parsers
        implicit none
        ! Arguments
        character(len=*),               intent(in)      ::  parameter   !< Label of parameter
        integer,                        intent(inout)   ::  value       !< Value of parameter
        character(len=*),   optional,   intent(in)      ::  question    !< Question to ask user for parameter
        character(len=*),   optional,   intent(in)      ::  help        !< Help to give user

        ! Private variables
        character(len=200)  ::  qquestion, hhelp

        ! Start work
        if (present(question)) then
            qquestion = question
        else
            qquestion = 'Please give ' // trim(parameter)
        endif
        if (present(help)) then
            hhelp = help
        else
            hhelp = 'Sorry, no help available'
        endif

        !print *, '**debug(get_rtp_value_intg): looking for ', trim(parameter), ', which has current value of ', value

        if (interactive) then
            call ui_get_intg(qquestion,value,hhelp)
        else
            call file_get_intg(control_filename,parameter,value)
        endif

        !print *, '**debug(get_rtp_value_intg): was looking for ', trim(parameter), ', which now has value of ', value

    end subroutine get_rtp_value_intg


    !>  Get a runtime parameter value, either from the control file or from the user interactively
    subroutine get_rtp_value_char(parameter,value,question,help)
        use runtime_parameters
        use parsers
        implicit none
        ! Arguments
        character(len=*),               intent(in)      ::  parameter   !< Label of parameter
        character(len=*),               intent(inout)   ::  value       !< Value of parameter
        character(len=*),   optional,   intent(in)      ::  question    !< Question to ask user for parameter
        character(len=*),   optional,   intent(in)      ::  help        !< Help to give user

        ! Private variables
        character(len=200)  ::  qquestion, hhelp

        ! Start work
        if (present(question)) then
            qquestion = question
        else
            qquestion = 'Please give ' // trim(parameter)
        endif
        if (present(help)) then
            hhelp = help
        else
            hhelp = 'Sorry, no help available'
        endif


        if (interactive) then
            call ui_get_char(qquestion,value,hhelp)
        else
            call file_get_char(control_filename,parameter,value)
        endif

    end subroutine get_rtp_value_char



    !>  brief   Writes the prompt to terminal
    subroutine prompt
        implicit none

        write(*,'(a)',advance='NO') '<frealix> '
    end subroutine prompt


    subroutine ui_get_logi(question,value,help)
        use parsers
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  question    !<  The question to which the user will reply
        logical,            intent(out) ::  value       !<  The value the user provided as an answer
        character(len=*),   intent(in), optional    ::  help    !<  Help string given to the user if they type '?'
        ! Private variables
        logical ::  answered = .false.
        character(len=128)  ::  answer, hhelp

        ! We start off not knowing the answer
        answered = .false.

        ! Start work
        !value = 0.0
        if (present(help)) then
            hhelp = help
        else
            hhelp = 'No help available, sorry'
        endif

        do while (.not. answered)
            call prompt()
            write (*,'(2a)',advance='NO') trim(question), ': '
            read (*,'(a)') answer
            if (is_blank(answer)) cycle
            if (starts_with_question_mark(answer)) then
                write(*,'(/a/)') hhelp
                cycle
            endif
            if (is_logi(answer)) then
                answered = .true.
                read(answer,*) value
            else
                cycle
            endif
        enddo
    end subroutine ui_get_logi


    subroutine ui_get_real(question,value,help)
        use parsers
        implicit none
        ! Arguments
        character(len=*), intent(in)    ::  question    !<  The question to which the user will reply
        real,               intent(out) ::  value       !<  The value the user provided as an answer
        character(len=*),   intent(in), optional    ::  help    !<  Help string given to the user if they type '?'
        ! Private variables
        logical ::  answered
        character(len=128)  ::  answer, hhelp

        ! We start off not knowing the answer
        answered = .false.

        ! Start work
        value = 0.0
        if (present(help)) then
            hhelp = help
        else
            hhelp = 'No help available, sorry'
        endif

        do while (.not. answered)
            call prompt()
            write (*,'(2a)',advance='NO') trim(question), ': '
            read (*,'(a)') answer
            if (is_blank(answer)) cycle
            if (starts_with_question_mark(answer)) then
                write(*,'(/a/)') hhelp
                cycle
            endif
            if (string_is_real(answer)) then
                answered = .true.
                read(answer,*) value
            else
                cycle
            endif
        enddo
    end subroutine ui_get_real

    subroutine ui_get_intg(question,value,help)
        use parsers
        implicit none
        ! Arguments
        character(len=*),   intent(in)  ::  question    !<  The question to which the user will reply
        integer,            intent(out) ::  value       !<  The value the user provided as an answer
        character(len=*),   intent(in), optional    ::  help    !<  Help string given to the user if they type '?'
        ! Private variables
        logical ::  answered = .false.
        character(len=128)  ::  answer, hhelp

        ! We start off not knowing the answer
        answered = .false.

        ! Start work
        value = 0
        if (present(help)) then
            hhelp = help
        else
            hhelp = 'No help available, sorry'
        endif

        do while (.not. answered)
            call prompt()
            write (*,'(2a)',advance='NO') trim(question), ': '
            read (*,'(a)') answer
            if (is_blank(answer)) cycle
            if (starts_with_question_mark(answer)) then
                write(*,'(/a/)') hhelp
                cycle
            endif
            if (is_intg(answer)) then
                answered = .true.
                read(answer,*) value
            else
                cycle
            endif
        enddo
    end subroutine ui_get_intg

    subroutine ui_get_char(question,value,help)
        use parsers
        implicit none
        ! Arguments
        character(len=*), intent(in)    ::  question    !<  The question to which the user will reply
        character(len=*),   intent(out) ::  value       !<  The value the user provided as an answer
        character(len=*),   intent(in), optional    ::  help    !<  Help string given to the user if they type '?'
        ! Private variables
        logical ::  answered = .false.
        character(len=128)  ::  answer, hhelp

        ! We start off not knowing the answer
        answered = .false.

        ! Start work
        value = ' '
        if (present(help)) then
            hhelp = help
        else
            hhelp = 'No help available, sorry'
        endif

        do while (.not. answered)
            call prompt()
            write (*,'(2a)',advance='NO') trim(question), ': '
            read (*,'(a)') answer
            !print *, '**debug (ui_get_char): got answer ', answer
            if (is_blank(answer)) cycle
            if (starts_with_question_mark(answer)) then
                write(*,'(/a/)') hhelp
                cycle
            endif
            answered = .true.
            read(answer,'(a)') value
            !print *, '**debug (ui_get_char): got value ', value
        enddo
    end subroutine ui_get_char

    !>  \brief  Use this function to work out whether you should print output for iteration current_iteration of a total of num_iterations iterations
    logical function interval_print(current_iteration,num_iterations)
        implicit none
        ! Arguments
        integer,    intent(in)  ::  current_iteration           !<  Current iteration number
        integer,    intent(in)  ::  num_iterations              !<  Total number of iterations
        ! Private variables
        integer,    parameter   ::  num_messages    =   10      !<  How many messages should be printed over the course of the iterations?
        ! Start work

        !
        ! We want to print on the first iteration, and in total 10 times, at regular intervals
        !
        if (num_iterations .ge. num_messages) then
            if (mod(current_iteration,num_iterations/num_messages) .eq. 0.0 .or. current_iteration .eq. 1) then
                interval_print = .true.
            else
                interval_print = .false.
            endif
        else
            interval_print = .true.
        endif
    end function interval_print


end module ui
