#!/bin/bash
#
# Test Frealix' multi-reference refinement.
#
# First we create 9 micrographs, which each contain one of 3 types of filaments
# following one of 3 possible paths:
#
# Film # as a function of structure and path #s:
#
# path \ structure    1       2       3
# 1                  #1      #4      #7
# 2                  #2      #5      #8
# 3                  #3      #6      #9
#
# Then, we do a single round of refinement in Frealix using random occupancy assignments
# for the 3 structures to see whether it can "recover" the correct assignments.

pixel_size="1.2"
h_rise="4.7"
h_twist="0.77"
h_rad="60.0"
h_stack="t"
h_axial_symmetry="1"
h_asym_unit_mw="5000.0"
snr="0.5"
model_resolution="2.5"
micrograph_dim_x="3072"
micrograph_dim_y="3072"

if [ $# -eq 2 ]; then
  frealix_exec=$1
  frealix_wrapper_script=$2
else
  echo "Usage: $(basename $0) frealix_binary frealix_wrapper_script"
  exit 1
fi

if [ ! -x $frealix_exec ]; then
  echo "Frealix is not executable: $frealix_exec"
  exit 1
fi

log_file="test.log"
rm -f $log_file
#rm -fr scratch
test -d scratch || mkdir -v scratch

#
# We will have several reference structures from which to generate
# synthetic micrographs
#
pdb_files[0]="Ab42_model_1_5units_0.77_4.7.pdb"
pdb_files[1]="Ab42_model_2_5units_0.77_4.7.pdb"
pdb_files[2]="Ab42_model_3_5units_0.77_4.7.pdb"

cp "${pdb_files[@]}" scratch/
cd scratch

main() {

  # Convert the PDBs to density and impose helical symmetry
  for pdb_file in "${pdb_files[@]}"; do
    bgex -verbose 7 -size 128,128,128 -origin 64.5,65.8,65 -sampling ${pixel_size} -resolution $model_resolution -coordinates ${pdb_file} ${pdb_file%%.pdb}.mrc >> $log_file 2>&1
    cat > hextend.flx <<eof
frealix_mode       hextend
input_filename     ${pdb_file%%.pdb}.mrc
output_filename    ${pdb_file%%.pdb}_h.mrc
twist_per_subunit  $h_twist
rise_per_subunit   $h_rise
symmetry_axial     $h_axial_symmetry
helix_max_radius   $h_rad
stack              $h_stack
pixel_size         $pixel_size
eof
    $frealix_exec hextend.flx >> $log_file 2>&1
  done

  if [ ! -f micrographs.mrc ]; then
    cat > synth_r001.flx <<eof
frealix_mode          synth_micrograph
model_3d_filename     Ab42_model_1_5units_0.77_4.7_h.mrc
input_par_filename    waypoints_r001_truth.ali
ctffind_filenames     ctffind.txt
output_film_filenames micrographs.mrc
output_film_size_x    $micrograph_dim_x
output_film_size_y    $micrograph_dim_y
twist_per_subunit     $h_twist
rise_per_subunit      $h_rise
helix_max_radius      $h_rad
symmetry_axial        $h_axial_symmetry
asym_unit_mw          $h_asym_unit_mw
stack                 $h_stack
pixel_size            $pixel_size
target_resolution     $( echo "$pixel_size * 2.1 | bc -l" )
snr                   1000000.0
estimated_model_resolution  $model_resolution
overwrite_micrographs t
apply_ctf             f
eof
    $frealix_exec synth_r001.flx >> $log_file 2>&1

    cat > synth_r002.flx <<eof
frealix_mode          synth_micrograph
model_3d_filename     Ab42_model_2_5units_0.77_4.7_h.mrc
input_par_filename    waypoints_r002_truth.ali
ctffind_filenames     ctffind.txt
output_film_filenames micrographs.mrc
output_film_size_x    $micrograph_dim_x
output_film_size_y    $micrograph_dim_y
twist_per_subunit     $h_twist
rise_per_subunit      $h_rise
helix_max_radius      $h_rad
symmetry_axial        $h_axial_symmetry
asym_unit_mw          $h_asym_unit_mw
stack                 $h_stack
pixel_size            $pixel_size
target_resolution     $( echo "$pixel_size * 2.1 | bc -l" )
snr                   1000000.0
estimated_model_resolution  $model_resolution
overwrite_micrographs f
apply_ctf             f
eof
    $frealix_exec synth_r002.flx >> $log_file 2>&1

    cat > synth_r003.flx <<eof
frealix_mode          synth_micrograph
model_3d_filename     Ab42_model_3_5units_0.77_4.7_h.mrc
input_par_filename    waypoints_r003_truth.ali
ctffind_filenames     ctffind.txt
output_film_filenames micrographs.mrc
output_film_size_x    $micrograph_dim_x
output_film_size_y    $micrograph_dim_y
twist_per_subunit     $h_twist
rise_per_subunit      $h_rise
helix_max_radius      $h_rad
symmetry_axial        $h_axial_symmetry
asym_unit_mw          $h_asym_unit_mw
stack                 $h_stack
pixel_size            $pixel_size
target_resolution     $( echo "$pixel_size * 2.1 | bc -l" )
snr                   $snr
estimated_model_resolution  $model_resolution
overwrite_micrographs f
apply_ctf             t
eof
    $frealix_exec synth_r003.flx >> $log_file 2>&1
  fi

  # Use the frealix wrapper to pretend we are starting a new project without a 3D reconstruction
  cat > parameters.flx <<eof
  film_files            micrographs.mrc
  ctffind_files         ctffind.txt
  twist_per_subunit     $h_twist
  rise_per_subunit      $h_rise
  helix_max_radius      $h_rad
  symmetry_axial        $h_axial_symmetry
  asym_unit_mw          $h_asym_unit_mw
  stack                 $h_stack
  pixel_size            $pixel_size
eof

  rm -fr frealix_round_00?
  ln -s $frealix_exec frealix
  ${frealix_wrapper_script} -r 1 -h 10.0 -n 12 -c 3 -a waypoints_start.ali parameters.flx

} # end of main


function generate_ctffind_file {
cat <<eof > ctffind.txt
# Output from CTFFind version 4.0.16, run on 2015-06-16 11:34:19
# Input file: /scratch/rohoua/tmp.dM3GveLPr/May26_16.41.20_magdistcorr_bin.mrc; Number of micrographs: 1
# Pixel size: 1.680; acceleration voltage: 300.0; spherical aberration: .0; amplitude contrast: .10
# Box size: 512; min res: 30.0; max. res.: 3.5; min. def.: 3000.0; max. def.: 30000.0
# Columns: #1 - micrograph number; #2 - defocus 1 [Angstroms]; #3 - defocus 2; #4 - azimuth of astigmatism; #5 - additional phase shift [radians]; #6 - cross correlation; #7 - spacing (in Angstroms) up to which CTF rings were fit successfully
1 10894.79 11931.05       89.26165       0.000000      0.8047181E-01   15.09198
2 19231.33 16864.08 -89.55347 0.000000 0.9204271E-01 26.88259
3 10945.33 9254.810 -83.97608 0.000000 0.1037194      7.480374
4 10894.79 11931.05       89.26165       0.000000      0.8047181E-01   15.09198
5 19231.33 16864.08 -89.55347 0.000000 0.9204271E-01 26.88259
6 10945.33 9254.810 -83.97608 0.000000 0.1037194      7.480374
7 10894.79 11931.05       89.26165       0.000000      0.8047181E-01   15.09198
8 19231.33 16864.08 -89.55347 0.000000 0.9204271E-01 26.88259
9 10945.33 9254.810 -83.97608 0.000000 0.1037194      7.480374
eof
}

function generate_starting_waypoints_file {
  cat <<eof > waypoints_start.ali
  !Fil.#   WP # Film #           X           Y           Z  PathLength          Phi        Theta          Psi   Subunit #        Twist         Rise    Score    LogP    Sigma    Occupancy     Refine? Units Angstr    # frames    1   par_vers. 6
       1      1      1    2348.345      98.243  -16584.531       0.000       -4.532       90.000       84.650       3.149        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       1      2      1    2297.923     243.749  -16584.531     154.546        6.984       90.000       74.228      36.031        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       1      3      1    2106.999     781.402  -16584.531     725.312       80.920       90.000       69.017     157.471        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       1      4      1    1929.529    1257.522  -16584.531    1233.631      189.643       90.000       65.059     265.624        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       1      5      1    1739.116    1708.026  -16584.531    1722.868      279.775       90.000       63.302     369.717        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       1      6      1    1539.679    2116.521  -16584.531    2177.520      376.830       90.000       66.516     466.451        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       1      7      1    1331.009    2563.042  -16584.531    2670.465      460.400       90.000       67.545     571.333        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       1      8      1    1126.211    3029.909  -16584.531    3180.433      547.798       90.000       70.333     679.837        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       1      9      1    1094.961    3106.053  -16584.531    3262.759      557.710       90.000       70.479     697.353        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       2      1      2     240.518    2898.969  -18047.705       0.000      -13.440      270.000       67.865       0.727        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       2      2      2     418.302    2662.202  -18047.705     298.136        3.207      270.000       45.054      64.160        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       2      3      2     837.154    2226.040  -18047.705     902.969      105.482      270.000       48.855     192.848        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       2      4      2    1236.058    1773.168  -18047.705    1506.526      183.874      270.000       50.749     321.264        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       2      5      2    1567.182    1348.320  -18047.705    2045.231      284.452      270.000       53.703     435.882        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       2      6      2    1875.316     908.353  -18047.705    2582.613      368.044      270.000       59.228     550.219        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       2      7      2    2141.474     417.879  -18047.705    3141.214      432.444      270.000       67.642     669.070        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       2      8      2    2224.166     163.619  -18047.705    3409.147      479.041      270.000       80.190     726.077        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       3      1      3    1886.281    3078.170  -17158.502       0.000      -70.123      270.000      130.960      25.658        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       3      2      3    1771.356    2913.349  -17158.502     201.076      -58.718      270.000      125.484      68.440        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       3      3      3    1515.551    2518.754  -17158.502     671.393       -0.911      270.000      124.058     168.508        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       3      4      3    1314.043    2138.754  -17158.502    1101.839       97.458      270.000      118.065     260.092        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       3      5      3    1098.284    1687.518  -17158.502    1602.065      179.860      270.000      116.041     366.523        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       3      6      3     905.430    1058.924  -17158.502    2261.035      280.160      270.000      103.504     506.730        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       3      7      3     755.309     210.383  -17158.502    3122.982      366.211      270.000      102.239     690.123        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       3      8      3     735.063      97.686  -17158.502    3237.494      375.680      270.000      101.346     714.487        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       4      1      4    2348.345      98.243  -16584.531       0.000       -4.532       90.000       84.650       3.149        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       4      2      4    2297.923     243.749  -16584.531     154.546        6.984       90.000       74.228      36.031        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       4      3      4    2106.999     781.402  -16584.531     725.312       80.920       90.000       69.017     157.471        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       4      4      4    1929.529    1257.522  -16584.531    1233.631      189.643       90.000       65.059     265.624        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       4      5      4    1739.116    1708.026  -16584.531    1722.868      279.775       90.000       63.302     369.717        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       4      6      4    1539.679    2116.521  -16584.531    2177.520      376.830       90.000       66.516     466.451        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       4      7      4    1331.009    2563.042  -16584.531    2670.465      460.400       90.000       67.545     571.333        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       4      8      4    1126.211    3029.909  -16584.531    3180.433      547.798       90.000       70.333     679.837        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       4      9      4    1094.961    3106.053  -16584.531    3262.759      557.710       90.000       70.479     697.353        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       5      1      5     240.518    2898.969  -18047.705       0.000      -13.440      270.000       67.865       0.727        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       5      2      5     418.302    2662.202  -18047.705     298.136        3.207      270.000       45.054      64.160        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       5      3      5     837.154    2226.040  -18047.705     902.969      105.482      270.000       48.855     192.848        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       5      4      5    1236.058    1773.168  -18047.705    1506.526      183.874      270.000       50.749     321.264        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       5      5      5    1567.182    1348.320  -18047.705    2045.231      284.452      270.000       53.703     435.882        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       5      6      5    1875.316     908.353  -18047.705    2582.613      368.044      270.000       59.228     550.219        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       5      7      5    2141.474     417.879  -18047.705    3141.214      432.444      270.000       67.642     669.070        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       5      8      5    2224.166     163.619  -18047.705    3409.147      479.041      270.000       80.190     726.077        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       6      1      6    1886.281    3078.170  -17158.502       0.000      -70.123      270.000      130.960      25.658        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       6      2      6    1771.356    2913.349  -17158.502     201.076      -58.718      270.000      125.484      68.440        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       6      3      6    1515.551    2518.754  -17158.502     671.393       -0.911      270.000      124.058     168.508        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       6      4      6    1314.043    2138.754  -17158.502    1101.839       97.458      270.000      118.065     260.092        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       6      5      6    1098.284    1687.518  -17158.502    1602.065      179.860      270.000      116.041     366.523        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       6      6      6     905.430    1058.924  -17158.502    2261.035      280.160      270.000      103.504     506.730        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       6      7      6     755.309     210.383  -17158.502    3122.982      366.211      270.000      102.239     690.123        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       6      8      6     735.063      97.686  -17158.502    3237.494      375.680      270.000      101.346     714.487        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       7      1      7    2348.345      98.243  -16584.531       0.000       -4.532       90.000       84.650       3.149        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       7      2      7    2297.923     243.749  -16584.531     154.546        6.984       90.000       74.228      36.031        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       7      3      7    2106.999     781.402  -16584.531     725.312       80.920       90.000       69.017     157.471        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       7      4      7    1929.529    1257.522  -16584.531    1233.631      189.643       90.000       65.059     265.624        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       7      5      7    1739.116    1708.026  -16584.531    1722.868      279.775       90.000       63.302     369.717        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       7      6      7    1539.679    2116.521  -16584.531    2177.520      376.830       90.000       66.516     466.451        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       7      7      7    1331.009    2563.042  -16584.531    2670.465      460.400       90.000       67.545     571.333        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       7      8      7    1126.211    3029.909  -16584.531    3180.433      547.798       90.000       70.333     679.837        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       7      9      7    1094.961    3106.053  -16584.531    3262.759      557.710       90.000       70.479     697.353        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
       8      1      8     240.518    2898.969  -18047.705       0.000      -13.440      270.000       67.865       0.727        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       8      2      8     418.302    2662.202  -18047.705     298.136        3.207      270.000       45.054      64.160        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       8      3      8     837.154    2226.040  -18047.705     902.969      105.482      270.000       48.855     192.848        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       8      4      8    1236.058    1773.168  -18047.705    1506.526      183.874      270.000       50.749     321.264        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       8      5      8    1567.182    1348.320  -18047.705    2045.231      284.452      270.000       53.703     435.882        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       8      6      8    1875.316     908.353  -18047.705    2582.613      368.044      270.000       59.228     550.219        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       8      7      8    2141.474     417.879  -18047.705    3141.214      432.444      270.000       67.642     669.070        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       8      8      8    2224.166     163.619  -18047.705    3409.147      479.041      270.000       80.190     726.077        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
       9      1      9    1886.281    3078.170  -17158.502       0.000      -70.123      270.000      130.960      25.658        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       9      2      9    1771.356    2913.349  -17158.502     201.076      -58.718      270.000      125.484      68.440        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       9      3      9    1515.551    2518.754  -17158.502     671.393       -0.911      270.000      124.058     168.508        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       9      4      9    1314.043    2138.754  -17158.502    1101.839       97.458      270.000      118.065     260.092        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       9      5      9    1098.284    1687.518  -17158.502    1602.065      179.860      270.000      116.041     366.523        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       9      6      9     905.430    1058.924  -17158.502    2261.035      280.160      270.000      103.504     506.730        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       9      7      9     755.309     210.383  -17158.502    3122.982      366.211      270.000      102.239     690.123        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
       9      8      9     735.063      97.686  -17158.502    3237.494      375.680      270.000      101.346     714.487        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
eof
}

function generate_waypoints_file {
# Generate waypoints files
cat <<eof > waypoints_r001_truth.ali
!Fil.#   WP # Film #           X           Y           Z  PathLength          Phi        Theta          Psi   Subunit #        Twist         Rise    Score    LogP    Sigma    Occupancy     Refine? Units Angstr    # frames    1   par_vers. 6
     1      1      1    2348.345      98.243  -16584.531       0.000       -4.532       90.000       84.650       3.149        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     1      2      1    2297.923     243.749  -16584.531     154.546        6.984       90.000       74.228      36.031        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     1      3      1    2106.999     781.402  -16584.531     725.312       80.920       90.000       69.017     157.471        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     1      4      1    1929.529    1257.522  -16584.531    1233.631      189.643       90.000       65.059     265.624        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     1      5      1    1739.116    1708.026  -16584.531    1722.868      279.775       90.000       63.302     369.717        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     1      6      1    1539.679    2116.521  -16584.531    2177.520      376.830       90.000       66.516     466.451        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     1      7      1    1331.009    2563.042  -16584.531    2670.465      460.400       90.000       67.545     571.333        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     1      8      1    1126.211    3029.909  -16584.531    3180.433      547.798       90.000       70.333     679.837        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     1      9      1    1094.961    3106.053  -16584.531    3262.759      557.710       90.000       70.479     697.353        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     2      1      2     240.518    2898.969  -18047.705       0.000      -13.440      270.000       67.865       0.727        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     2      2      2     418.302    2662.202  -18047.705     298.136        3.207      270.000       45.054      64.160        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     2      3      2     837.154    2226.040  -18047.705     902.969      105.482      270.000       48.855     192.848        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     2      4      2    1236.058    1773.168  -18047.705    1506.526      183.874      270.000       50.749     321.264        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     2      5      2    1567.182    1348.320  -18047.705    2045.231      284.452      270.000       53.703     435.882        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     2      6      2    1875.316     908.353  -18047.705    2582.613      368.044      270.000       59.228     550.219        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     2      7      2    2141.474     417.879  -18047.705    3141.214      432.444      270.000       67.642     669.070        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     2      8      2    2224.166     163.619  -18047.705    3409.147      479.041      270.000       80.190     726.077        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     3      1      3    1886.281    3078.170  -17158.502       0.000      -70.123      270.000      130.960      25.658        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     3      2      3    1771.356    2913.349  -17158.502     201.076      -58.718      270.000      125.484      68.440        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     3      3      3    1515.551    2518.754  -17158.502     671.393       -0.911      270.000      124.058     168.508        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     3      4      3    1314.043    2138.754  -17158.502    1101.839       97.458      270.000      118.065     260.092        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     3      5      3    1098.284    1687.518  -17158.502    1602.065      179.860      270.000      116.041     366.523        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     3      6      3     905.430    1058.924  -17158.502    2261.035      280.160      270.000      103.504     506.730        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     3      7      3     755.309     210.383  -17158.502    3122.982      366.211      270.000      102.239     690.123        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     3      8      3     735.063      97.686  -17158.502    3237.494      375.680      270.000      101.346     714.487        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     4      1      4    2348.345      98.243  -16584.531       0.000       -4.532       90.000       84.650       3.149        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      2      4    2297.923     243.749  -16584.531     154.546        6.984       90.000       74.228      36.031        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      3      4    2106.999     781.402  -16584.531     725.312       80.920       90.000       69.017     157.471        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      4      4    1929.529    1257.522  -16584.531    1233.631      189.643       90.000       65.059     265.624        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      5      4    1739.116    1708.026  -16584.531    1722.868      279.775       90.000       63.302     369.717        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      6      4    1539.679    2116.521  -16584.531    2177.520      376.830       90.000       66.516     466.451        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      7      4    1331.009    2563.042  -16584.531    2670.465      460.400       90.000       67.545     571.333        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      8      4    1126.211    3029.909  -16584.531    3180.433      547.798       90.000       70.333     679.837        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      9      4    1094.961    3106.053  -16584.531    3262.759      557.710       90.000       70.479     697.353        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      1      5     240.518    2898.969  -18047.705       0.000      -13.440      270.000       67.865       0.727        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      2      5     418.302    2662.202  -18047.705     298.136        3.207      270.000       45.054      64.160        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      3      5     837.154    2226.040  -18047.705     902.969      105.482      270.000       48.855     192.848        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      4      5    1236.058    1773.168  -18047.705    1506.526      183.874      270.000       50.749     321.264        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      5      5    1567.182    1348.320  -18047.705    2045.231      284.452      270.000       53.703     435.882        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      6      5    1875.316     908.353  -18047.705    2582.613      368.044      270.000       59.228     550.219        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      7      5    2141.474     417.879  -18047.705    3141.214      432.444      270.000       67.642     669.070        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      8      5    2224.166     163.619  -18047.705    3409.147      479.041      270.000       80.190     726.077        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      1      6    1886.281    3078.170  -17158.502       0.000      -70.123      270.000      130.960      25.658        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      2      6    1771.356    2913.349  -17158.502     201.076      -58.718      270.000      125.484      68.440        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      3      6    1515.551    2518.754  -17158.502     671.393       -0.911      270.000      124.058     168.508        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      4      6    1314.043    2138.754  -17158.502    1101.839       97.458      270.000      118.065     260.092        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      5      6    1098.284    1687.518  -17158.502    1602.065      179.860      270.000      116.041     366.523        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      6      6     905.430    1058.924  -17158.502    2261.035      280.160      270.000      103.504     506.730        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      7      6     755.309     210.383  -17158.502    3122.982      366.211      270.000      102.239     690.123        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      8      6     735.063      97.686  -17158.502    3237.494      375.680      270.000      101.346     714.487        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      1      7    2348.345      98.243  -16584.531       0.000       -4.532       90.000       84.650       3.149        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      2      7    2297.923     243.749  -16584.531     154.546        6.984       90.000       74.228      36.031        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      3      7    2106.999     781.402  -16584.531     725.312       80.920       90.000       69.017     157.471        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      4      7    1929.529    1257.522  -16584.531    1233.631      189.643       90.000       65.059     265.624        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      5      7    1739.116    1708.026  -16584.531    1722.868      279.775       90.000       63.302     369.717        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      6      7    1539.679    2116.521  -16584.531    2177.520      376.830       90.000       66.516     466.451        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      7      7    1331.009    2563.042  -16584.531    2670.465      460.400       90.000       67.545     571.333        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      8      7    1126.211    3029.909  -16584.531    3180.433      547.798       90.000       70.333     679.837        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      9      7    1094.961    3106.053  -16584.531    3262.759      557.710       90.000       70.479     697.353        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      1      8     240.518    2898.969  -18047.705       0.000      -13.440      270.000       67.865       0.727        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      2      8     418.302    2662.202  -18047.705     298.136        3.207      270.000       45.054      64.160        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      3      8     837.154    2226.040  -18047.705     902.969      105.482      270.000       48.855     192.848        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      4      8    1236.058    1773.168  -18047.705    1506.526      183.874      270.000       50.749     321.264        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      5      8    1567.182    1348.320  -18047.705    2045.231      284.452      270.000       53.703     435.882        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      6      8    1875.316     908.353  -18047.705    2582.613      368.044      270.000       59.228     550.219        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      7      8    2141.474     417.879  -18047.705    3141.214      432.444      270.000       67.642     669.070        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      8      8    2224.166     163.619  -18047.705    3409.147      479.041      270.000       80.190     726.077        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      1      9    1886.281    3078.170  -17158.502       0.000      -70.123      270.000      130.960      25.658        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      2      9    1771.356    2913.349  -17158.502     201.076      -58.718      270.000      125.484      68.440        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      3      9    1515.551    2518.754  -17158.502     671.393       -0.911      270.000      124.058     168.508        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      4      9    1314.043    2138.754  -17158.502    1101.839       97.458      270.000      118.065     260.092        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      5      9    1098.284    1687.518  -17158.502    1602.065      179.860      270.000      116.041     366.523        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      6      9     905.430    1058.924  -17158.502    2261.035      280.160      270.000      103.504     506.730        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      7      9     755.309     210.383  -17158.502    3122.982      366.211      270.000      102.239     690.123        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      8      9     735.063      97.686  -17158.502    3237.494      375.680      270.000      101.346     714.487        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
eof

cat <<eof > waypoints_r002_truth.ali
!Fil.#   WP # Film #           X           Y           Z  PathLength          Phi        Theta          Psi   Subunit #        Twist         Rise    Score    LogP    Sigma    Occupancy     Refine? Units Angstr    # frames    1   par_vers. 6
     1      1      1    2348.345      98.243  -16584.531       0.000       -4.532       90.000       84.650       3.149        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      2      1    2297.923     243.749  -16584.531     154.546        6.984       90.000       74.228      36.031        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      3      1    2106.999     781.402  -16584.531     725.312       80.920       90.000       69.017     157.471        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      4      1    1929.529    1257.522  -16584.531    1233.631      189.643       90.000       65.059     265.624        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      5      1    1739.116    1708.026  -16584.531    1722.868      279.775       90.000       63.302     369.717        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      6      1    1539.679    2116.521  -16584.531    2177.520      376.830       90.000       66.516     466.451        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      7      1    1331.009    2563.042  -16584.531    2670.465      460.400       90.000       67.545     571.333        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      8      1    1126.211    3029.909  -16584.531    3180.433      547.798       90.000       70.333     679.837        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      9      1    1094.961    3106.053  -16584.531    3262.759      557.710       90.000       70.479     697.353        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      1      2     240.518    2898.969  -18047.705       0.000      -13.440      270.000       67.865       0.727        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      2      2     418.302    2662.202  -18047.705     298.136        3.207      270.000       45.054      64.160        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      3      2     837.154    2226.040  -18047.705     902.969      105.482      270.000       48.855     192.848        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      4      2    1236.058    1773.168  -18047.705    1506.526      183.874      270.000       50.749     321.264        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      5      2    1567.182    1348.320  -18047.705    2045.231      284.452      270.000       53.703     435.882        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      6      2    1875.316     908.353  -18047.705    2582.613      368.044      270.000       59.228     550.219        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      7      2    2141.474     417.879  -18047.705    3141.214      432.444      270.000       67.642     669.070        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      8      2    2224.166     163.619  -18047.705    3409.147      479.041      270.000       80.190     726.077        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      1      3    1886.281    3078.170  -17158.502       0.000      -70.123      270.000      130.960      25.658        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      2      3    1771.356    2913.349  -17158.502     201.076      -58.718      270.000      125.484      68.440        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      3      3    1515.551    2518.754  -17158.502     671.393       -0.911      270.000      124.058     168.508        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      4      3    1314.043    2138.754  -17158.502    1101.839       97.458      270.000      118.065     260.092        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      5      3    1098.284    1687.518  -17158.502    1602.065      179.860      270.000      116.041     366.523        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      6      3     905.430    1058.924  -17158.502    2261.035      280.160      270.000      103.504     506.730        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      7      3     755.309     210.383  -17158.502    3122.982      366.211      270.000      102.239     690.123        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      8      3     735.063      97.686  -17158.502    3237.494      375.680      270.000      101.346     714.487        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      1      4    2348.345      98.243  -16584.531       0.000       -4.532       90.000       84.650       3.149        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     4      2      4    2297.923     243.749  -16584.531     154.546        6.984       90.000       74.228      36.031        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     4      3      4    2106.999     781.402  -16584.531     725.312       80.920       90.000       69.017     157.471        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     4      4      4    1929.529    1257.522  -16584.531    1233.631      189.643       90.000       65.059     265.624        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     4      5      4    1739.116    1708.026  -16584.531    1722.868      279.775       90.000       63.302     369.717        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     4      6      4    1539.679    2116.521  -16584.531    2177.520      376.830       90.000       66.516     466.451        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     4      7      4    1331.009    2563.042  -16584.531    2670.465      460.400       90.000       67.545     571.333        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     4      8      4    1126.211    3029.909  -16584.531    3180.433      547.798       90.000       70.333     679.837        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     4      9      4    1094.961    3106.053  -16584.531    3262.759      557.710       90.000       70.479     697.353        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     5      1      5     240.518    2898.969  -18047.705       0.000      -13.440      270.000       67.865       0.727        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     5      2      5     418.302    2662.202  -18047.705     298.136        3.207      270.000       45.054      64.160        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     5      3      5     837.154    2226.040  -18047.705     902.969      105.482      270.000       48.855     192.848        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     5      4      5    1236.058    1773.168  -18047.705    1506.526      183.874      270.000       50.749     321.264        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     5      5      5    1567.182    1348.320  -18047.705    2045.231      284.452      270.000       53.703     435.882        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     5      6      5    1875.316     908.353  -18047.705    2582.613      368.044      270.000       59.228     550.219        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     5      7      5    2141.474     417.879  -18047.705    3141.214      432.444      270.000       67.642     669.070        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     5      8      5    2224.166     163.619  -18047.705    3409.147      479.041      270.000       80.190     726.077        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     6      1      6    1886.281    3078.170  -17158.502       0.000      -70.123      270.000      130.960      25.658        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     6      2      6    1771.356    2913.349  -17158.502     201.076      -58.718      270.000      125.484      68.440        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     6      3      6    1515.551    2518.754  -17158.502     671.393       -0.911      270.000      124.058     168.508        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     6      4      6    1314.043    2138.754  -17158.502    1101.839       97.458      270.000      118.065     260.092        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     6      5      6    1098.284    1687.518  -17158.502    1602.065      179.860      270.000      116.041     366.523        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     6      6      6     905.430    1058.924  -17158.502    2261.035      280.160      270.000      103.504     506.730        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     6      7      6     755.309     210.383  -17158.502    3122.982      366.211      270.000      102.239     690.123        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     6      8      6     735.063      97.686  -17158.502    3237.494      375.680      270.000      101.346     714.487        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     7      1      7    2348.345      98.243  -16584.531       0.000       -4.532       90.000       84.650       3.149        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      2      7    2297.923     243.749  -16584.531     154.546        6.984       90.000       74.228      36.031        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      3      7    2106.999     781.402  -16584.531     725.312       80.920       90.000       69.017     157.471        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      4      7    1929.529    1257.522  -16584.531    1233.631      189.643       90.000       65.059     265.624        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      5      7    1739.116    1708.026  -16584.531    1722.868      279.775       90.000       63.302     369.717        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      6      7    1539.679    2116.521  -16584.531    2177.520      376.830       90.000       66.516     466.451        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      7      7    1331.009    2563.042  -16584.531    2670.465      460.400       90.000       67.545     571.333        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      8      7    1126.211    3029.909  -16584.531    3180.433      547.798       90.000       70.333     679.837        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      9      7    1094.961    3106.053  -16584.531    3262.759      557.710       90.000       70.479     697.353        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      1      8     240.518    2898.969  -18047.705       0.000      -13.440      270.000       67.865       0.727        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      2      8     418.302    2662.202  -18047.705     298.136        3.207      270.000       45.054      64.160        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      3      8     837.154    2226.040  -18047.705     902.969      105.482      270.000       48.855     192.848        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      4      8    1236.058    1773.168  -18047.705    1506.526      183.874      270.000       50.749     321.264        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      5      8    1567.182    1348.320  -18047.705    2045.231      284.452      270.000       53.703     435.882        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      6      8    1875.316     908.353  -18047.705    2582.613      368.044      270.000       59.228     550.219        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      7      8    2141.474     417.879  -18047.705    3141.214      432.444      270.000       67.642     669.070        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     8      8      8    2224.166     163.619  -18047.705    3409.147      479.041      270.000       80.190     726.077        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      1      9    1886.281    3078.170  -17158.502       0.000      -70.123      270.000      130.960      25.658        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      2      9    1771.356    2913.349  -17158.502     201.076      -58.718      270.000      125.484      68.440        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      3      9    1515.551    2518.754  -17158.502     671.393       -0.911      270.000      124.058     168.508        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      4      9    1314.043    2138.754  -17158.502    1101.839       97.458      270.000      118.065     260.092        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      5      9    1098.284    1687.518  -17158.502    1602.065      179.860      270.000      116.041     366.523        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      6      9     905.430    1058.924  -17158.502    2261.035      280.160      270.000      103.504     506.730        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      7      9     755.309     210.383  -17158.502    3122.982      366.211      270.000      102.239     690.123        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     9      8      9     735.063      97.686  -17158.502    3237.494      375.680      270.000      101.346     714.487        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
eof

cat <<eof > waypoints_r003_truth.ali
!Fil.#   WP # Film #           X           Y           Z  PathLength          Phi        Theta          Psi   Subunit #        Twist         Rise    Score    LogP    Sigma    Occupancy     Refine? Units Angstr    # frames    1   par_vers. 6
     1      1      1    2348.345      98.243  -16584.531       0.000       -4.532       90.000       84.650       3.149        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      2      1    2297.923     243.749  -16584.531     154.546        6.984       90.000       74.228      36.031        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      3      1    2106.999     781.402  -16584.531     725.312       80.920       90.000       69.017     157.471        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      4      1    1929.529    1257.522  -16584.531    1233.631      189.643       90.000       65.059     265.624        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      5      1    1739.116    1708.026  -16584.531    1722.868      279.775       90.000       63.302     369.717        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      6      1    1539.679    2116.521  -16584.531    2177.520      376.830       90.000       66.516     466.451        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      7      1    1331.009    2563.042  -16584.531    2670.465      460.400       90.000       67.545     571.333        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      8      1    1126.211    3029.909  -16584.531    3180.433      547.798       90.000       70.333     679.837        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     1      9      1    1094.961    3106.053  -16584.531    3262.759      557.710       90.000       70.479     697.353        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      1      2     240.518    2898.969  -18047.705       0.000      -13.440      270.000       67.865       0.727        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      2      2     418.302    2662.202  -18047.705     298.136        3.207      270.000       45.054      64.160        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      3      2     837.154    2226.040  -18047.705     902.969      105.482      270.000       48.855     192.848        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      4      2    1236.058    1773.168  -18047.705    1506.526      183.874      270.000       50.749     321.264        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      5      2    1567.182    1348.320  -18047.705    2045.231      284.452      270.000       53.703     435.882        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      6      2    1875.316     908.353  -18047.705    2582.613      368.044      270.000       59.228     550.219        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      7      2    2141.474     417.879  -18047.705    3141.214      432.444      270.000       67.642     669.070        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     2      8      2    2224.166     163.619  -18047.705    3409.147      479.041      270.000       80.190     726.077        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      1      3    1886.281    3078.170  -17158.502       0.000      -70.123      270.000      130.960      25.658        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      2      3    1771.356    2913.349  -17158.502     201.076      -58.718      270.000      125.484      68.440        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      3      3    1515.551    2518.754  -17158.502     671.393       -0.911      270.000      124.058     168.508        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      4      3    1314.043    2138.754  -17158.502    1101.839       97.458      270.000      118.065     260.092        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      5      3    1098.284    1687.518  -17158.502    1602.065      179.860      270.000      116.041     366.523        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      6      3     905.430    1058.924  -17158.502    2261.035      280.160      270.000      103.504     506.730        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      7      3     755.309     210.383  -17158.502    3122.982      366.211      270.000      102.239     690.123        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     3      8      3     735.063      97.686  -17158.502    3237.494      375.680      270.000      101.346     714.487        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      1      4    2348.345      98.243  -16584.531       0.000       -4.532       90.000       84.650       3.149        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      2      4    2297.923     243.749  -16584.531     154.546        6.984       90.000       74.228      36.031        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      3      4    2106.999     781.402  -16584.531     725.312       80.920       90.000       69.017     157.471        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      4      4    1929.529    1257.522  -16584.531    1233.631      189.643       90.000       65.059     265.624        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      5      4    1739.116    1708.026  -16584.531    1722.868      279.775       90.000       63.302     369.717        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      6      4    1539.679    2116.521  -16584.531    2177.520      376.830       90.000       66.516     466.451        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      7      4    1331.009    2563.042  -16584.531    2670.465      460.400       90.000       67.545     571.333        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      8      4    1126.211    3029.909  -16584.531    3180.433      547.798       90.000       70.333     679.837        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     4      9      4    1094.961    3106.053  -16584.531    3262.759      557.710       90.000       70.479     697.353        0.810        4.700    0.316     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      1      5     240.518    2898.969  -18047.705       0.000      -13.440      270.000       67.865       0.727        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      2      5     418.302    2662.202  -18047.705     298.136        3.207      270.000       45.054      64.160        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      3      5     837.154    2226.040  -18047.705     902.969      105.482      270.000       48.855     192.848        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      4      5    1236.058    1773.168  -18047.705    1506.526      183.874      270.000       50.749     321.264        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      5      5    1567.182    1348.320  -18047.705    2045.231      284.452      270.000       53.703     435.882        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      6      5    1875.316     908.353  -18047.705    2582.613      368.044      270.000       59.228     550.219        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      7      5    2141.474     417.879  -18047.705    3141.214      432.444      270.000       67.642     669.070        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     5      8      5    2224.166     163.619  -18047.705    3409.147      479.041      270.000       80.190     726.077        0.679        4.700    0.314     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      1      6    1886.281    3078.170  -17158.502       0.000      -70.123      270.000      130.960      25.658        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      2      6    1771.356    2913.349  -17158.502     201.076      -58.718      270.000      125.484      68.440        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      3      6    1515.551    2518.754  -17158.502     671.393       -0.911      270.000      124.058     168.508        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      4      6    1314.043    2138.754  -17158.502    1101.839       97.458      270.000      118.065     260.092        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      5      6    1098.284    1687.518  -17158.502    1602.065      179.860      270.000      116.041     366.523        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      6      6     905.430    1058.924  -17158.502    2261.035      280.160      270.000      103.504     506.730        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      7      6     755.309     210.383  -17158.502    3122.982      366.211      270.000      102.239     690.123        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     6      8      6     735.063      97.686  -17158.502    3237.494      375.680      270.000      101.346     714.487        0.647        4.700    0.320     0.0      0.5          0.0           T  0.0 0.0 0.0
     7      1      7    2348.345      98.243  -16584.531       0.000       -4.532       90.000       84.650       3.149        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     7      2      7    2297.923     243.749  -16584.531     154.546        6.984       90.000       74.228      36.031        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     7      3      7    2106.999     781.402  -16584.531     725.312       80.920       90.000       69.017     157.471        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     7      4      7    1929.529    1257.522  -16584.531    1233.631      189.643       90.000       65.059     265.624        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     7      5      7    1739.116    1708.026  -16584.531    1722.868      279.775       90.000       63.302     369.717        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     7      6      7    1539.679    2116.521  -16584.531    2177.520      376.830       90.000       66.516     466.451        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     7      7      7    1331.009    2563.042  -16584.531    2670.465      460.400       90.000       67.545     571.333        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     7      8      7    1126.211    3029.909  -16584.531    3180.433      547.798       90.000       70.333     679.837        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     7      9      7    1094.961    3106.053  -16584.531    3262.759      557.710       90.000       70.479     697.353        0.810        4.700    0.316     0.0      0.5        100.0           T  0.0 0.0 0.0
     8      1      8     240.518    2898.969  -18047.705       0.000      -13.440      270.000       67.865       0.727        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     8      2      8     418.302    2662.202  -18047.705     298.136        3.207      270.000       45.054      64.160        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     8      3      8     837.154    2226.040  -18047.705     902.969      105.482      270.000       48.855     192.848        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     8      4      8    1236.058    1773.168  -18047.705    1506.526      183.874      270.000       50.749     321.264        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     8      5      8    1567.182    1348.320  -18047.705    2045.231      284.452      270.000       53.703     435.882        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     8      6      8    1875.316     908.353  -18047.705    2582.613      368.044      270.000       59.228     550.219        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     8      7      8    2141.474     417.879  -18047.705    3141.214      432.444      270.000       67.642     669.070        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     8      8      8    2224.166     163.619  -18047.705    3409.147      479.041      270.000       80.190     726.077        0.679        4.700    0.314     0.0      0.5        100.0           T  0.0 0.0 0.0
     9      1      9    1886.281    3078.170  -17158.502       0.000      -70.123      270.000      130.960      25.658        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     9      2      9    1771.356    2913.349  -17158.502     201.076      -58.718      270.000      125.484      68.440        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     9      3      9    1515.551    2518.754  -17158.502     671.393       -0.911      270.000      124.058     168.508        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     9      4      9    1314.043    2138.754  -17158.502    1101.839       97.458      270.000      118.065     260.092        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     9      5      9    1098.284    1687.518  -17158.502    1602.065      179.860      270.000      116.041     366.523        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     9      6      9     905.430    1058.924  -17158.502    2261.035      280.160      270.000      103.504     506.730        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     9      7      9     755.309     210.383  -17158.502    3122.982      366.211      270.000      102.239     690.123        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
     9      8      9     735.063      97.686  -17158.502    3237.494      375.680      270.000      101.346     714.487        0.647        4.700    0.320     0.0      0.5        100.0           T  0.0 0.0 0.0
eof


} # end of function


generate_waypoints_file
generate_starting_waypoints_file
generate_ctffind_file
main
