#!/bin/tcsh

if ($#argv != 1 ) then
echo "No arguments supplied...exiting"
exit 1
endif

if (! -f $argv[1]) then
echo File does not exist: $argv[1]
exit 1
endif

if (! -x $argv[1]) then
echo File is not executable: $argv[1]
exit 1
endif

set pdb_file = 2OM3.pdb1
set log_file = test.log
set frealix_exec = $1
set plot_fsc_exec = ${1:h}/../scripts/plot_fsc.rb
set pixel_size = 4.652
set h_rise = 1.390
set h_rad = 105


rm -f $log_file
test -d scratch || mkdir -v scratch
cd scratch

echo -n `date` Downlading PDB if necessary...
test -f ~/data/tmv_casa/${pdb_file}.gz  && cp ~/data/tmv_casa/${pdb_file}.gz .
test -f ${pdb_file}.gz || test -f ${pdb_file} || wget http://www.rcsb.org/pdb/files/${pdb_file}.gz >>& $log_file
test -f $pdb_file || gzip -d ${pdb_file}.gz >>& $log_file
echo " done."

echo -n `date` Converting PDB to density volume...
#test -f ${pdb_file:r}.hdf || e2pdb2em.py -A 1.163 -B 192 --center n --O ${pdb_file} ${pdb_file:r}.hdf >>& $log_file
#test -f ${pdb_file:r}.hdf || e2pdb2em.py -A 4.652 -B 48 --center n --O ${pdb_file} ${pdb_file:r}.hdf >>& $log_file
#test -f ${pdb_file:r}_z.hdf || e2proc3d.py --meanshrink 8 --trans=0,0,4 ${pdb_file:r}.hdf ${pdb_file:r}_z.hdf >>& $log_file
test -f ${pdb_file:r}.mrc || bgex -verbose 7 -size 48,48,48 -origin 24,24,18 -sampling ${pixel_size} -resolution 5.0 -coordinates ${pdb_file} ${pdb_file:r}.mrc >>& $log_file
echo " done."

#echo -n `date` Converting HDF to MRC...
#test -f ${pdb_file:r}.mrc || e2proc3d.py ${pdb_file:r}_z.hdf ${pdb_file:r}.mrc >>& $log_file
#echo " done."

#echo -n `date` Masking to remove artefacts...
#if (! -f ${pdb_file:r}_m.mrc) then
#cat > mask.flx << EOF
#FREALIX_MODE MASK
#INPUT_FILENAME 		${pdb_file:r}.mrc
#OUTPUT_FILENAME 	${pdb_file:r}_m.mrc
#MASK_TYPE 		CYLINDER
#FALLOFF_TYPE 		COS_EDGE
#MASK_RADIUS		23.0
#MASK_FALLOFF		5.0
#EOF
#$frealix_exec mask.flx  >>& $log_file
#endif
#echo "done."

echo -n `date` "Imposing helical symmetry (new algorithm)..."
if (! -f ${pdb_file:r}_h.mrc) then
cat > hextend.flx << EOF
FREALIX_MODE HEXTEND
INPUT_FILENAME ${pdb_file:r}.mrc
OUTPUT_FILENAME ${pdb_file:r}_h.mrc
TWIST_PER_SUBUNIT -22.03
RISE_PER_SUBUNIT $h_rise
PIXEL_SIZE ${pixel_size}
HELIX_MAX_RADIUS $h_rad
NEW_ALGORITHM T
EOF
$frealix_exec hextend.flx >>& $log_file
endif
echo " done."

echo -n `date` "Extracting the central slices... "
test -f ${pdb_file:r}_c.mrc   || e2proc2d.py --threed2twod --first 24 --last 24 ${pdb_file:r}.mrc ${pdb_file:r}_c.mrc >>& $log_file
test -f ${pdb_file:r}_h_c.mrc || e2proc2d.py --threed2twod --first 24 --last 24 ${pdb_file:r}_h.mrc ${pdb_file:r}_h_c.mrc >>& $log_file
echo " done."

echo -n `date` "Computing FRC... "
cat > calc_fsc.flx << EOF
FREALIX_MODE CALC_FSC
INPUT_VOLUME_1 ${pdb_file:r}_c.mrc
INPUT_VOLUME_2 ${pdb_file:r}_h_c.mrc
EOF
$frealix_exec calc_fsc.flx >& fsc.log
cat fsc.log >>& $log_file
$plot_fsc_exec -p ${pixel_size} fsc.log fsc.png >>& $log_file
echo " done."

echo -n `date` "Checking FRC... "
cat > check_fsc.rb << EOF
#!/usr/bin/env ruby

require_relative "../../../scripts/rubylix.rb"
require_relative "../../../scripts/FrealixOutput.rb"

# Initialise log file & find out the frealix mode that was run
log = FrealixOutput.new("fsc.log")

fsc_values = []
fsc_values = log.getFSCValues()

fsc_values.delete_if {|x| x > 0.9}

if (fsc_values.count > 3) then
  puts "FSC looks bad"
  exit 1
else
  puts "FSC looks good"
  exit 0
end
EOF

chmod +x check_fsc.rb
./check_fsc.rb >>& $log_file

if ($?) then
	echo "Test failed"
	exit 1
endif


egrep "sorry|Fatal" $log_file >& /dev/null

if ($?) then
  echo "Test passed"
  exit 0
else
  echo "Test failed"
  echo "Have a look at log files and fsc.png"
  exit 1
endif

cd ..
