#!/bin/bash
#
#
# Check that the runtime resampling of input data for refinement does not affect the quality of the final reconstruction.
#
pdb_file=2OM3.pdb1
pixel_size=1.163
h_twist="-22.03"
h_rise="1.390"
h_rad=105
asym_unit_mw="17589.66"
high_res_cutoff="9.304"
frealix_exec=$1

test -d scratch || mkdir scratch
cd scratch

# Download PDB
test -f ${pdb_file}.gz || test -f ${pdb_file} || wget http://www.rcsb.org/pdb/files/${pdb_file}.gz > wget.log 2>&1
test -f $pdb_file || gzip -d ${pdb_file}.gz > gzip.log 2>&1

# Convert to density
test -f ${pdb_file%%.*}.mrc || bgex -verbose 7 -size 192,192,192 -origin 96,96,72 -sampling ${pixel_size} -resolution 3.0 -coordinates ${pdb_file} ${pdb_file%%.*}.mrc > bgex.log 2>&1

# Hextend
cat > hextend.flx << EOF
FREALIX_MODE HEXTEND
INPUT_FILENAME  ${pdb_file%%.*}.mrc
OUTPUT_FILENAME  ${pdb_file%%.*}_h.mrc
TWIST_PER_SUBUNIT $h_twist
RISE_PER_SUBUNIT $h_rise
PIXEL_SIZE $pixel_size
HELIX_MAX_RADIUS $h_rad
NEW_ALGORITHM T
EOF
test -f ${pdb_file%%.*}_h.mrc || $frealix_exec hextend.flx > hextend.log 2>&1

# Generate synthetic micrograph
cat <<eof > waypoints_made_up.ali
!Fil.#   WP # Film #           X           Y           Z  PathLength          Phi        Theta          Psi   Subunit #    Score Refine? Units Angstr   par_vers. 4
    34      1      1   11119.998    6936.677   00000.000   00000.000    00000.000    00000.000    00000.000   00000.000    0.223       T
    34      2      1   10629.840    6505.115  -00000.000   00000.000    00000.000    00000.000    00000.000   00000.000    0.223       T
    34      3      1   10128.577    6007.739  -00000.000   00000.000    00000.000    00000.000    00000.000   00000.000    0.223       T
eof
cat > synth.flx <<eof
FREALIX_MODE        SYNTH_MICROGRAPH
MODEL_3D_FILENAME   ${pdb_file%%.*}_h.mrc
INPUT_PAR_FILENAME  waypoints_made_up.ali
OUTPUT_PAR_FILENAME waypoints_straight.ali
CTFTILT_FILENAMES   ../A??_bin2_ctftilt.log
OUTPUT_FILM_FILENAMES   micrograph_????.mrc
STRAIGHTEN_FILAMENT    T
TWIST_PER_SUBUNIT   $h_twist
RISE_PER_SUBUNIT    $h_rise
HELIX_MAX_RADIUS    $h_rad
ASYM_UNIT_MW        $asym_unit_mw
PIXEL_SIZE      	$pixel_size
TARGET_RESOLUTION   4.0
SNR         		10000.0
REAL_SPACE_PROJ         T
ESTIMATED_MODEL_RESOLUTION 5.0
eof
test -f micrograph_0001.mrc || $frealix_exec synth.flx > synth.log 2>&1

# Do the refinement with resampling
cat > refine_resample.flx <<eof
FREALIX_MODE        FILAMENT_REFINE
MODEL_3D_FILENAME   ${pdb_file%%.*}_h.mrc
OUTPUT_3D_FILENAME  3drec_resample.mrc
INPUT_PAR_FILENAME  waypoints_straight.ali
OUTPUT_PAR_FILENAME junk_resample.ali
FORCE_WP_TO_WP_DISTANCE  200.0
CTFTILT_FILES       ../A??_bin2_ctftilt.log
FILM_FILES      	micrograph_????.mrc
TWIST_PER_SUBUNIT   $h_twist
RISE_PER_SUBUNIT    $h_rise
HELIX_MAX_RADIUS    $h_rad
ASYM_UNIT_MW        $asym_unit_mw
PIXEL_SIZE      	$pixel_size
SKIP_REFINEMENT     F
WPS_AS_SPS      	F
FAL_WPS_IN_SCORE    F
FAL_WPS_IN_THREED   F
SP_SEGMENT_LENGTH   100.0
CALC_3D_RECONSTRUCTION  T
FINALISE_3D_RECONSTRUCTION  T
PSF_RADIUS      200
TARGET_RESOLUTION   4.0
HIGH_RESOLUTION_CUTOFF $high_res_cutoff
RESAMPLE_DATA_FOR_REFINEMENT T
CALC_FSC F
APPLY_FOM_FILTER F
eof
test -f junk_resample.ali || $frealix_exec refine_resample.flx > refine_resample.log 2>&1

# Do the refinement without resampling (should take much longer, but will be our reference)
cp refine_{,no}resample.flx
awk '$1 ~ /RESAMPLE_DATA_FOR_REFINEMENT/ {$2="F"} $1 ~ /OUTPUT_3D_FILENAME/ {$2="output_3d_noresample.mrc"} $1 ~ /OUTPUT_PAR_FILENAME/ {$2="junk_noresample.ali"} $1 ~ /OUTPUT_3D_FILENAME/ {$2="3drec_noresample.mrc"} $1 ~ /SKIP_REFINEMENT/ {$2="T"} {print $0}' refine_resample.flx > refine_noresample.flx
test -f junk_noresample.ali || $frealix_exec refine_noresample.flx > refine_noresample.log 2>&1

# Compare the two 3D reconstructions
cat > crop_resample.flx <<eof
FREALIX_MODE      PAD
INPUT_FILENAME    3drec_resample.mrc
OUTPUT_FILENAME   3drec_resample_200.mrc
NEW_DIM           200
eof
$frealix_exec crop_resample.flx > crop_resample.log 2>&1

cat > crop_noresample.flx <<eof
FREALIX_MODE      PAD
INPUT_FILENAME    3drec_noresample.mrc
OUTPUT_FILENAME   3drec_noresample_200.mrc
NEW_DIM           200
eof
$frealix_exec crop_noresample.flx > crop_noresample.log 2>&1

cat > cyl_resample.flx << eof
FREALIX_MODE      MASK
INPUT_FILENAME    3drec_resample_200.mrc
OUTPUT_FILENAME   3drec_resample_200_cyl.mrc
MASK_TYPE         CYLINDER
MASK_RADIUS       85.0
MASK_INNER_RADIUS 14.0
MASK_FALLOFF      10.0
MASK_WITH_AVERAGE T
eof
$frealix_exec cyl_resample.flx > cyl_resample.log 2>&1

cat > cyl_noresample.flx << eof
FREALIX_MODE      MASK
INPUT_FILENAME    3drec_noresample_200.mrc
OUTPUT_FILENAME   3drec_noresample_200_cyl.mrc
MASK_TYPE         CYLINDER
MASK_RADIUS       85.0
MASK_INNER_RADIUS 14.0
MASK_FALLOFF      10.0
MASK_WITH_AVERAGE T
eof
$frealix_exec cyl_noresample.flx > cyl_noresample.log 2>&1



cat > fsc.flx <<eof
FREALIX_MODE      CALC_FSC
INPUT_VOLUME_1    3drec_resample_200_cyl.mrc
INPUT_VOLUME_2    3drec_noresample_200_cyl.mrc
eof
$frealix_exec fsc.flx > fsc.log 2>&1

number_of_bad_shells=$(awk '$1 =="1", $1 == "100" {if ($2 < 0.95) {print $2}}'  fsc.log | wc -l)
echo "Number of bad shells = $number_of_bad_shells"

if [ $number_of_bad_shells -gt 10 ]; then
	echo "Test failed"
	exit 1
else
	echo "Test passed"
	exit 0
fi
