!>  \brief  Module grouping functions and subroutines to do with phase contrast transfer functions
module images_ctf

    use ctfs

    contains

    !>  \brief  Unit tests
    subroutine image_ctf_unit_tests()
        implicit none
        call image_ctf_unit_test_1()
    end subroutine image_ctf_unit_tests

    !>  \brief  Test the thread safety of ctf_to_img
    subroutine image_ctf_unit_test_1()
        use images
        implicit none
        ! Private variables
        integer                             ::  i
        integer,                parameter   ::  num_tests = 20
        type(image),            allocatable ::  ctf_image(:)
        type(ctf_parameters),   allocatable ::  ctf(:)
        ! Start work
        write(*,'(a)') '**info(image_ctf_unit_test_1): entering'
        allocate(ctf_image(num_tests),ctf(num_tests))

        !$omp   parallel do default(none) &
        !$omp   private(i) &
        !$omp   shared(ctf,ctf_image)
        do I=1,num_tests
            call ctf_parameters_init(ctf(i),AKV=200.0,CS=2.0,AMPL_CONT=0.07,DFMID1=2.3,DFMID2=2.1,ANGAST=43.0)
            ctf_image(i)%logical_dimensions = [128,128,1]
            call ctf_to_img(ctf(i),ctf_image(i),PIXEL_SIZE=2.3)
        end do
        !$omp   end parallel do
        ! Deallocate
        do i=1,num_tests
            call ctf_image(i)%Deallocate()
        enddo
        deallocate(ctf_image,ctf)
        write(*,'(a)') '**info(image_ctf_unit_test_1): done'
    end subroutine image_ctf_unit_test_1

    !>  \brief  Apply ctf to an image
    subroutine image_apply_ctf(self,ctf,pixel_size,ctf_squared,flip_only)
        use images
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),                        intent(inout)   ::  self
        type(ctf_parameters),               intent(inout)   ::  ctf
        real,                   optional,   intent(in)      ::  pixel_size
        logical,                optional,   intent(in)      ::  ctf_squared
        logical,                optional,   intent(in)      ::  flip_only       !<  Multiply by sign(ctf) rather than ctf itself
        ! Private variables
        integer ::  i,j,x,y,xx,yy
        real    ::  ang
        integer ::  pow
        real    ::  rad
        ! Start work

        ! We can only deal with 2d images
        if (self%logical_dimensions(3) .ne. 1) then
            write(*,'(a)') '**ERROR(image_apply_ctf): this subroutine is only meant for 2d images'
            call terminate('Fatal error in IMAGE_APPLY_CTF')
        endif

        ! Check whether image is allocated
        if (.not. self%is_in_memory) then
            write(*,'(a)') '**ERROR(image_apply_ctf): input image not allocated'
            call terminate('Fatal error in IMAGE_APPLY_CTF')
        endif

        ! We need the ctf in pixel units
        if ( .not. ctf%pix_units) then
            if (present(pixel_size)) then
                call ang_to_pix(ctf,pixel_size)
            else
                write(*,'(a)') '**ERROR(image_apply_ctf): ctf parameters in Angstrom but no pixel size given'
                call terminate('Fatal error in IMAGE_APPLY_CTF')
            endif
        endif

        ! Power of the ctf
        pow = 1
        if (present(ctf_squared)) then
            if (ctf_squared) pow = 2
        endif

        ! We need the image in Fourier space
        if (self%is_in_real_space) call self%ForwardFFT()

        ! The eval_ctf routine currently assumes square images
        if (.not. self%IsSquare()) then
            write(*,'(a)') '**ERROR(image_apply_ctf): not implemented for non-square images'
            call terminate('Feature in image_apply_ctf not yet implemented')
        endif

        ! Loop over the image pixels
        do J=1,self%logical_dimensions(2)
            y = self%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2)
            yy = Y**2
            do I=1,self%physical_upper_bound_complex(1)
                x = I-1
                xx = X**2
                rad = sqrt(real(XX+YY))
                if (rad .gt. 0.0) ang = atan2(real(y),real(x))
                self%complex_values(i,j,1) = self%complex_values(i,j,1) * &
                                            eval_ctf(ctf,self%logical_dimensions(2),RAD,ang,SIGN_ONLY=FLIP_ONLY) ** pow
            enddo
        enddo

    end subroutine image_apply_ctf


    !>  \brief Image is filled with ctf
    subroutine ctf_to_img(ctf1,img,pixel_size,object_is_centered_in_box,one_at_origin)
        use images
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(ctf_parameters),   intent(in)      ::  ctf1                        !<  ctf object specifying ctf parameters
        type(image),            intent(inout)   ::  img                         !<  Image object to be filled with ctf values
        real,       optional,   intent(in)      ::  pixel_size                  !<  Pixel size (required if the ctf is supplied in Angstroms)
        logical,    optional,   intent(in)      ::  object_is_centered_in_box   !<  Whether to center the origin. .FALSE. by default.
        logical,    optional,   intent(in)      ::  one_at_origin               !<  Near the origin, return 1.0's rather than actual ctf value. Return to ctf value as soon as the first peak is met
        ! Private variables
        integer ::  i,j
        real    ::  rad, ang(2), max_rad_sq
        logical ::  do_not_fold(3)
        type(ctf_parameters) :: ctf2 ! internal parameters, in pixels
        logical ::  half_ctf
        integer ::  x,y,xx,yy
        integer ::  origin(3)
        logical,    parameter   ::  debug    = .false.
        logical ::  oone_at_origin
        real,   allocatable ::  arg_of_peaks(:)
        real    ::  freq_of_first_peak

        ! We can only deal with 2d images
        if (img%logical_dimensions(3) .ne. 1) then
            write(*,'(a)') '**ERROR(ctf_to_img): this subroutine is only meant for 2d images'
            call terminate('Fatal error in CTF_TO_IMG')
        endif

        ! We can only deal with real images
        if (.not. img%is_in_real_space) then
            write(*,'(a)') '**ERROR(ctf_to_img): this subroutine is only meant for real images'
            call terminate('Fatal error in CTF_TO_IMG')
        endif

        ! Will we be setting values near the origin to 1.0?
        oone_at_origin = .false.
        if (present(one_at_origin)) oone_at_origin = one_at_origin


        ! Check whether we need to allocate the image
        if (.not. img%is_in_memory) then
            call img%Allocate()
        endif

        ! Check whether we need to center the origin
        if (present(object_is_centered_in_box)) then
            img%object_is_centered_in_box = object_is_centered_in_box
        else
            img%object_is_centered_in_box = .false.
        endif

        ! Check whether first dimension = half of second dimension
        if (img%logical_dimensions(1) .eq. img%logical_dimensions(2)/2+1) then
            half_ctf = .true.
        else
            half_ctf = .false.
        endif


        ! If we have corner origins, make sure the first dimension only has one origin
        if (half_ctf) then
            do_not_fold = [.true., .false.,.false.]
        else
            do_not_fold = [.false.,.false.,.false.]
        endif

        ! If we need the pixel size, check the user gave it
        if ( .not. ctf1%pix_units) then
            if (present(pixel_size)) then
                ctf2 = ctf1
                call ang_to_pix(ctf2,pixel_size)
            else
                write(*,'(a)') '**ERROR(ctf_to_img): ctf parameters in Angstrom but no pixel size given'
                call terminate('Fatal error in CTF_TO_IMG')
            endif
        else
            ctf2 = ctf1
        endif

        ! First peak of the ctf (note, this will not work well for highly astigmatic CTFs
        if (oone_at_origin) then
            allocate(arg_of_peaks(10))
            call ctf_solve_for_arg(ctf2%ampl_cont,arg_of_peaks,-1.0e0)
            call ctf_sf_from_arg(ctf2%cs,ctf2%wl,ctf2%dfmid1,ctf2%dfmid2,ctf1%angast,arg_of_peaks(1), &
                                ctf2%angast+45.0,freq_of_first_peak)
        endif

        ! The square of the maximum radius
        max_rad_sq = (img%logical_dimensions(2)/2)**2

        ! Initialise output (note added June 2011 - this is wasteful as it loops over the array; determine if this is safe and remove if possible)
        !img%real_values = 0.0

        if (debug) print *, '**debug(ctf_to_img): dnf, half_ctf, origin at center: ', &
                            do_not_fold, half_ctf, img%object_is_centered_in_box

        if (img%is_in_real_space .and. img%object_is_centered_in_box) then
            origin = img%physical_address_of_box_center
        else
            origin = [1,1,1]
        endif

        ! Loop over pixels
        if (img%object_is_centered_in_box) then
            do J=1,img%logical_dimensions(2)
                y = J-ORIGIN(2)
                yy = Y**2
                do I=1,img%logical_dimensions(1)
                    if (do_not_fold(1)) then
                        x = I-1
                    else
                        x = I-ORIGIN(1)
                    endif
                    xx = X**2
                    rad = sqrt(real(XX+YY))
                    if (rad .gt. 0.0) ang(1) = atan2(real(y),real(x))
                    if (oone_at_origin .and. rad .le. freq_of_first_peak*img%logical_dimensions(2)) then
                        img%real_values(i,j,1) = -1.0e0
                    else
                        img%real_values(i,j,1) = eval_ctf(ctf2,img%logical_dimensions(2),rad,ang(1))
                    endif
                enddo
            enddo
        else
            do j=1,img%logical_dimensions(2)
                y = img%LogicalIndexGivenPhysicalIndexInFourierSpace(j,2)
                yy = y**2
                do i=1,img%logical_dimensions(1)
                    if (do_not_fold(1)) then
                        x = i-1
                    else
                        x = img%LogicalIndexGivenPhysicalIndexInFourierSpace(i,1)
                    endif
                    xx = x**2
                    rad = sqrt(real(xx+yy))
                    if (rad .gt. 0.0) ang(1) = atan2(real(y),real(x))
                    if (oone_at_origin .and. rad .le. freq_of_first_peak*img%logical_dimensions(2)) then
                        img%real_values(i,j,1) = -1.0e0
                    else
                        img%real_values(i,j,1) = eval_ctf(ctf2,img%logical_dimensions(2),rad,ang(1))
                    endif
                enddo
            enddo
        endif

        img%real_values(img%logical_dimensions(1)+1:,:,:) = 0.0e0

    end subroutine ctf_to_img

    !>  \brief  Given a CTF**2 volume with the CTFs centered in the center of the volume, apply the appropriate shifts
    !!          so that the output volume is half the size of the input and has the origin in the top-left corner
    subroutine ctfvol_shift(vol_in,vol_out)
        use images
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),    intent(in)      ::  vol_in
        type(image),    intent(inout)   ::  vol_out
        ! Private variables
        integer ::  i,j,k,ii,jj,kk,iii,jjj,kkk
        integer ::  origin(3)
        ! Start work
        if (.not. vol_in%is_in_real_space) then
            write(*,'(a)') '**ERROR(ctfvol_shift): input volume is not real'
            call terminate('Fatal error in CTFVOL_SHIFT')
        endif
        if (.not. vol_in%IsAllocated()) then
            write(*,'(a)') '**ERROR(ctfvol_shift): input volume not allocated'
            call terminate('Fatal error in CTFVOL_SHIFT')
        endif
        if (vol_in%object_is_centered_in_box) then
            ! Prepare the output volume
            vol_out%logical_dimensions(1)              =   vol_in%physical_upper_bound_complex(1)
            vol_out%logical_dimensions(2:3)            =   vol_in%logical_dimensions(2:3)
            vol_out%is_in_real_space                =   vol_in%is_in_real_space
            vol_out%object_is_centered_in_box    =   .false.
            call vol_out%Allocate()
            vol_out%real_values = 0.0 ! May not be necessary, i think allocate does this anyway...?

            ! Physical address of origin voxel
            if (vol_in%is_in_real_space .and. vol_in%object_is_centered_in_box) then
                origin = vol_in%physical_address_of_box_center
            else
                origin = [1,1,1]
            endif

            ! Loop over the input volume
            do K=1,vol_in%logical_dimensions(3)
                kk = k - origin(3)  ! relative to origin at center
                kk = -KK    ! Hermitian symmetry mate (relative to origin at center)
                if (kk .ge. 0) then
                    kkk = kk + 1    ! Address within output volume (with origin at corner)
                else
                    kkk = kk + vol_out%logical_dimensions(3) + 1   ! Address within output volume (with origin at corner), negative frequencies
                endif
                do J=1,vol_in%logical_dimensions(2)
                    jj = j - origin(2)
                    jj = -JJ
                    if (jj .ge. 0) then
                        jjj = jj + 1
                    else
                        jjj = jj + vol_out%logical_dimensions(2) + 1
                    endif
                    !if (jjj .EQ. 86) THEN
                    !   print '(a,3(i3,x))', '**debug(ctfvol_shift): jjj = 86 for i,j,k = ', i,j,k
                    !endif
                    ! For the first dimensions, we only need to loop over the "negative frequencies"
                    do I=1,origin(1)
                        ii = i - origin(1) ! relative to origin at center
                        ii = -II    ! Hermitian symmetry mate (relative to origin at center)
                        iii = ii + 1
                        ! Now copy value over
                        vol_out%real_values(iii,jjj,kkk) = vol_in%real_values(i,j,k)
                    enddo
                enddo
            enddo
        else
            write(*,'(a)') '**warning(ctfvol_shift): input volume origin not at center, not doing anything'
            call terminate('debug') ! remove this when testing has been done
        endif
    end subroutine ctfvol_shift

    !>  \brief Along first dimension, return address within real-formatted array given address within complex-formatted array.
    !!
    !!
    !!  Useful when looping over a complex volume (say a 3d reconstruction) and want to find out corresponding address in real array with, say, CTF^2 terms.
    !!  "Real formatted array" : same extent in all dimensions. Origin at n/2+1 in every dimension.
    !!  "Comp formatted array" : first dimension extent = second dimension/2+1; origin at address (1,1,1). Only positive frequencies in first dimension. Only negative Nyquist frequency in 2nd and 3rd dimension.
    pure integer function ctfvol_addr_real_from_comp_dim1(vol_real,addr_comp_dim1) result(addr_real)
        use images
        use runtime_parameters, only : terminate
        implicit none
        ! Arguments
        type(image),    intent(in)  ::  vol_real            !<  Volume with correct logical dimensions corresponding to real addressing (all dimensions equal). Origin at center. Explicit Friedel mates.
        integer,        intent(in)  ::  addr_comp_dim1      !<  Address in 1st dimension within output (comp-arranged) volume. Must be .le. vol_in%logical_dimensions(1)/2+1.
        ! Private variables
        integer ::  coo
        ! Start work
        coo = addr_comp_dim1 - 1
        addr_real = coo + vol_real%GetPhysicalIndexOfBoxCenter(1)
        ! At this point, addr_real could be .GT. vol_real%logical_dimensions(3).
        ! This would happen if we're on the +ve Nyquist frequency, which is not represented. We'll just have to catch that case later on and go the Hermitian symmetry mate once all frequencies are known.
    end function ctfvol_addr_real_from_comp_dim1

    !>  \todo   Pass dimension number as argument and make this work for non cubes
    pure integer function ctfvol_addr_real_from_comp_dim2or3(vol_real,addr_comp_dim2or3) result(addr_real)
        use images
        implicit none
        ! Arguments
        type(image),    intent(in)  ::  vol_real                !<  Volume with correct logical dimensions corresponding to real addressing. Origin at center. Explicit Friedel mates.
        integer,        intent(in)  ::  addr_comp_dim2or3       !<  Address in 3rd dimension within output volume
        ! Private variables
        integer ::  coo
        ! Start work
        coo = vol_real%LogicalIndexGivenPhysicalIndexInFourierSpace(addr_comp_dim2or3,3)
        addr_real = coo + vol_real%GetPhysicalIndexOfBoxCenter(3)
        ! At this point, addr_real could be .GT. vol_real%logical_dimensions(3).
        ! This would happen if we're on the +ve Nyquist frequency, which is not represented. We'll just have to catch that case later on and go the Hermitian symmetry mate once all frequencies are known.
    end function ctfvol_addr_real_from_comp_dim2or3



end module images_ctf
