!>  \brief  Modify input image by apply a ctf to it
module mode_apply_ctf
    use modes
    use rtps
    implicit none
    ! Declare mode object
    type(mode)  ::  apply_ctf
    ! Declare RTPs
    type(rtp_char), target  ::  input_filename  =               &
        rtp_char(   value='input.mrc',                  &
                    keyword='INPUT_FILENAME',           &
                    description='Name of input image file.')

    type(rtp_char), target  ::  output_filename =               &
        rtp_char(   value='output.mrc',                 &
                    keyword='OUTPUT_FILENAME',          &
                    description='Name of output image file')

    type(rtp_real), target  ::  pixel_size  =               &
        rtp_real(   value=9.99,                  &
                    keyword='PIXEL_SIZE',           &
                    description='Size of pixel, in Angstroms')

    type(rtp_real), target  ::  acceleration_voltage  =               &
        rtp_real(   value=200.,                  &
                    keyword='ACCELERATION_VOLTAGE',           &
                    description='Acceleration voltage, in kV')

    type(rtp_real), target  ::  spherical_aberration  =               &
        rtp_real(   value=2.0,                  &
                    keyword='SPHERICAL_ABERRATION',           &
                    description='Objective lens spherical aberration, in mm.')

    type(rtp_real), target  ::  amplitude_contrast  =               &
        rtp_real(   value=0.07,                  &
                    keyword='AMPLITUDE_CONTRAST',           &
                    description='Amplitude contrast, as a fraction')


    type(rtp_real), target  ::  defocus_1  =               &
        rtp_real(   value=3.0,                  &
                    keyword='DEFOCUS_1',           &
                    description='Underfocus along the long axis, in microns')

    type(rtp_real), target  ::  defocus_2  =               &
        rtp_real(   value=3.0,                    &
                    keyword='DEFOCUS_2',           &
                    description='Underfocus along the short axis, in microns')

    type(rtp_real), target  ::  astigmatism_angle  =               &
        rtp_real(   value=0.0,                              &
                    keyword='ASTIGMATISM_ANGLE',           &
                    description='Angle of the long axis orientation')

contains

subroutine mode_apply_ctf_init_and_run
    implicit none
    ! Initialise the mode
    call apply_ctf%init(  command='APPLY_CTF',              &
                    subroutine=mode_apply_ctf_slave,  &
                    short_description='Apply a microscope ctf to an image')
    ! Add RTPs to the mode
    call apply_ctf%add_rtp(input_filename)
    call apply_ctf%add_rtp(output_filename)
    call apply_ctf%add_rtp(pixel_size)
    call apply_ctf%add_rtp(acceleration_voltage)
    call apply_ctf%add_rtp(spherical_aberration)
    call apply_ctf%add_rtp(amplitude_contrast)
    call apply_ctf%add_rtp(defocus_1)
    call apply_ctf%add_rtp(defocus_2)
    call apply_ctf%add_rtp(astigmatism_angle)
    ! Run the mode
    call apply_ctf%run()

end subroutine mode_apply_ctf_init_and_run

subroutine mode_apply_ctf_slave
    ! Modules
    use images
    use imagefiles
    use ctfs
    use images_ctf
    implicit none
    ! Private variables
    type(ctf_parameters)    ::  ctf
    type(imagefile)         ::  input_file, output_file
    type(image)             ::  img
    ! Start work

    ! Initialise ctf parameters
    call ctf_parameters_init(   ctf,   &
                                akv=acceleration_voltage%value, &
                                cs=spherical_aberration%value,  &
                                ampl_cont=amplitude_contrast%value,  &
                                dfmid1=defocus_1%value, &
                                dfmid2=defocus_2%value, &
                                angast=astigmatism_angle%value )

    ! Print out information about the ctf parameters
    call ctf_print_info(ctf)

    ! Read in the input image
    call input_file%init(filename=input_filename%value)
    call img%ReadFromDisk(input_file)
    call input_file%PrintInfo()
    call input_file%close()

    ! Apply the ctf
    call image_apply_ctf(img,ctf,pixel_size=pixel_size%value)
    if (img%IsInFourierSpace()) call img%BackwardFFT()

    ! Write out the output image
    call output_file%init(filename=output_filename%value)
    call img%WriteToDisk(output_file)
    call output_file%PrintInfo()
    call output_file%close()


end subroutine mode_apply_ctf_slave

end module mode_apply_ctf
