#!/bin/csh
#
# Check whether Frealix does CTF operations like other packages

if ($#argv != 1 ) then
echo "No arguments supplied...exiting"
exit 1
endif

if (! -f $argv[1]) then
echo File does not exist: $argv[1]
exit 1
endif

if (! -x $argv[1]) then
echo File is not executable: $argv[1]
exit 1
endif

set e2_python = $EMAN2DIR/Python/bin/python
set flx_exec = $1
set log_file = test.log
set defocus = 2.5
set ampcont_pc = 7
set ampcont = 0.07
set akv = 200
set cs = 2.0
set apix = 4.652
set fsc_thresh = 0.99
set fsc_tol = 0

rm -f $log_file
test -d scratch || mkdir scratch
cd scratch/
rm -f $log_file

echo -n `date` "Generating test image..."
$e2_python << eof >>& $log_file
from EMAN2 import *
img=EMData(256,256)
img.process_inplace("testimage.noise.gauss",{"sigma":10.0})
img.process_inplace("mask.gaussian",{"inner_radius":100.0,"outer_radius":5.0})
img.write_image("test_in.mrc")
eof
echo "done."

echo -n `date` "Applying CTF (EMAN2)..."
$e2_python << eof >>& $log_file
from EMAN2 import *
ctf=EMAN2Ctf()
ctf.from_dict({"defocus":$defocus,"dfdiff":0.0,"dfang":0.0,"ampcont":$ampcont_pc,"voltage":$akv,"cs":$cs,"apix":$apix})
img=EMData()
img.read_image("test_in.mrc")
fft1=img.do_fft()
ctfim=fft1.copy()
ctf.compute_2d_complex(ctfim,Ctf.CtfType.CTF_AMP)
fft1.mult(ctfim)
out=fft1.do_ift()
out.mult(-1.0)
out.write_image("test_out_e2.mrc",0,IMAGE_MRC)
eof
echo "done."

echo -n `date` "Applying CTF (Frealix)..."
cat > tmp.flx <<eof
FREALIX_MODE APPLY_CTF
INPUT_FILENAME test_in.mrc
OUTPUT_FILENAME test_out_flx.mrc
PIXEL_SIZE $apix
ACCELERATION_VOLTAGE $akv
SPHERICAL_ABERRATION $cs
AMPLITUDE_CONTRAST $ampcont 
DEFOCUS_1 $defocus
DEFOCUS_2 $defocus
eof
$flx_exec tmp.flx >>& $log_file
echo "done."

echo -n `date` "Computing FRC (E2 v FLX)..."
cat > tmp.flx <<eof
FREALIX_MODE CALC_FSC
INPUT_VOLUME_1 test_out_e2.mrc
INPUT_VOLUME_2 test_out_flx.mrc
eof
$flx_exec tmp.flx |& tee fsc.log >>& $log_file
echo "done."

echo -n `date` "Checking FRC (E2 v FLX)..."
cat > check_fsc.rb << EOF
#!/usr/bin/env ruby
require_relative "../../../scripts/rubylix.rb"
require_relative "../../../scripts/FrealixOutput.rb"

# Initialise log file & find out the frealix mode that was run
log = FrealixOutput.new("fsc.log")
fsc_values = []
fsc_values = log.getFSCValues()
fsc_values.delete_if {|x| x > 0.99}
if (fsc_values.count > 0) then
  #puts "FSC looks bad"
  exit 1
else
  #puts "FSC looks good"
  exit 0
end
EOF

chmod +x check_fsc.rb
./check_fsc.rb >>& $log_file

if ($?) then
	echo "Test failed"
	exit 1
endif


egrep "sorry|Fatal" $log_file >& /dev/null

if ($?) then
  echo "Test passed"
  exit 0
else
  echo "Test failed"
  exit 1
endif

cd ..
